##############################################################################
## CDM.tcl -                                                                ##
## This file contains various utilities offered by CDM, implemented in Tcl. ##
##  Copyright 1998-2008 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

## CDM_RegisterCollectionHandler
#  This function registers a Collection handler. A Collection handler consists
#  of a label (name) and a Tcl command. For every Document that is appended in
#  a Collection, this handler will be called, if selected by the user.
#  The arguments that will be appended to this handler are the Collection and
#  Document over which the handler has to be applied...
proc CDM_RegisterCollectionHandler {Label Command} {
  set ::CDM::RegisteredCollectionHandlers($Label) $Command
};# CDM_RegisterCollectionHandler

## CDM_GetRegisteredCollectionHadlerLabels
#  This procedure returns the labels describing all the available registered
#  handlers...
proc CDM_GetRegisteredCollectionHadlerLabels {} {
  if {[info exists ::CDM::RegisteredCollectionHandlers]} {
    return [lsort [array names ::CDM::RegisteredCollectionHandlers]]
  }
  return {}
};# CDM_GetRegisteredCollectionHadlerLabels

## CDM_GetRegisteredCollectionHadlerCommand
#  This function will return the Command associated with a specific Collection
#  handler, specified through its label...
proc CDM_GetRegisteredCollectionHadlerCommand {Label} {
  if {[info exists ::CDM::RegisteredCollectionHandlers($Label)]} {
    return $::CDM::RegisteredCollectionHandlers($Label)
  }
  return {}
};# CDM_GetRegisteredCollectionHadlerCommand

## CDM_GetRegisteredCollectionHadlerLabel
#  This function will return the Label associated with a specific Collection
#  handler, specified through its command...
proc CDM_GetRegisteredCollectionHadlerLabel {Command} {
  if {[info exists ::CDM::RegisteredCollectionHandlers]} {
    foreach label [array names ::CDM::RegisteredCollectionHandlers] {
      if {[string equal $Command \
                        $::CDM::RegisteredCollectionHandlers($label)]} {
        return $label
      }
    }
  }
  return $Command
};# CDM_GetRegisteredCollectionHadlerLabel

#################################
## Viewer Registration
#################################

## CDM_RegisterViewer
#  This function registers a Viewer.
proc CDM_RegisterViewer {Label Command {image ITViewText}} {
  ## Do not allow more than one viewer to have a specific label...
  if {[info exists ::CDM::RegisteredViewers($Label)]} {
    if {[string equal $Command $::CDM::RegisteredViewers($Label)]} {return}
    error "Label \"$Label\" already registered to viewer\
           \"$::CDM::RegisteredViewers($Label)\"!"
  }
  set ::CDM::RegisteredViewers($Label) $Command
  set ::CDM::RegisteredViewerImages($Label) $image
};# CDM_RegisterViewer

## CDM_GetRegisteredViewerLabels
#  This procedure returns the labels describing all the available registered
#  handlers...
proc CDM_GetRegisteredViewerLabels {} {
  ggi_InitialiseViewersSubsystem
  if {[info exists ::CDM::RegisteredViewers]} {
    return [lsort [array names ::CDM::RegisteredViewers]]
  }
  return {}
};# CDM_GetRegisteredViewerLabels

## CDM_GetRegisteredViewerCommand
#  This function will return the Command associated with a specific Collection
#  handler, specified through its label...
proc CDM_GetRegisteredViewerCommand {Label} {
  if {[info exists ::CDM::RegisteredViewers($Label)]} {
    return $::CDM::RegisteredViewers($Label)
  }
  return {}
};# CDM_GetRegisteredViewerCommand

## CDM_GetRegisteredViewerLabel
#  This function will return the Label associated with a specific Collection
#  handler, specified through its command...
proc CDM_GetRegisteredViewerLabel {Command} {
  if {[info exists ::CDM::RegisteredViewers]} {
    foreach label [array names ::CDM::RegisteredViewers] {
      if {[string equal $Command \
                        $::CDM::RegisteredViewers($label)]} {
        return $label
      }
    }
  }
  return $Command
};# CDM_GetRegisteredViewerLabel

proc CDM_GetRegisteredViewerImage {Label} {
  if {[info exists ::CDM::RegisteredViewerImages($Label)]} {
    return $::CDM::RegisteredViewerImages($Label)
  }
  return ITBlank
};# CDM_GetRegisteredViewerImage

#################################
## Error Reporting Facilities
#################################

proc CDM_ErrorInfoAsCaller {find replace} {
  set info $::errorInfo
  set i [string last "\n    (\"$find" $info]
  if {$i == -1} {return $info}
  set result [string range $info 0 [incr i 6]]	;# keep "\n    (\""
  append result $replace			;# $find -> $replace
  incr i [string length $find]
  set j [string first ) $info [incr i]]	;# keep rest of parenthetical
  append result [string range $info $i $j]
  return $result
};# CDM_ErrorInfoAsCaller

proc CDM_TkHtmlTruncationError {msg} {
  catch {console show}
  puts stderr "TkHTML: Truncation error: \"$msg\""
};# CDM_TkHtmlTruncationError
