/*****************************************************************************
 *  CDMFileIO.cpp:                                                           *
 * ----------------                                                          *
 *  This file contains various functions that relate to file handling and    *
 *  Tcl object serialisation...                                              *
 *                                                                           *
 *  Copyright 1998-2010 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licencing information, please refer to the licence conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/

#include "CDMIO.h"
#include <string.h>
#include "CDM_GlobalObjects.h"

extern Tcl_ThreadDataKey dataKey;

/*
 * CDMIO_OpenCollectionSaveFile:
 * This function opens the Collection index file, according to the type of the
 * Collection. The newly opened file will be set to the proper encoding...
 * In case of an error, NULL will be returned and an error will be left at the
 * Collection interpreter...
 */
Tcl_Channel CDMIO_OpenCollectionSaveFile(CDM_Collection Collection) {
  Tcl_Channel File;
  Tcl_DString ds;
  CDM_THREAD_SPECIFIC_DATA;
  char *argv[2], *encoding = (char *) CDMIO_DefaultEncoding,
       *translation = (char *) CDMIO_DefaultTranslation;

  /*
   * According to the Collection type, try to build the full path to the
   * Collection index file...
   */
  Tcl_DStringInit(&ds);
  argv[0] = Collection->Name; argv[1] = CDM_CollectionFilename;
  if (Collection->FileFormat == CDM_FORMAT_GATE_1x ||
      Collection->FileFormat == CDM_FORMAT_GATE_2x) {
    argv[1]  = (char *) "GuT_mAgIc";
    encoding = (char *) Collection->Encoding;
  } else if (Collection->FileFormat >= CDM_FORMAT_REGISTERED &&
             Collection->FileFormat <  CDM_FORMAT_REGISTERED +
                                       CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER) {
    argv[1]  = CDM_CollectionTypeFilename;
    encoding = (char *) CDMIO_DefaultEncoding;
  }
  Tcl_JoinPath(2, argv, &ds);

  /*
   * Try to open the Collection Index file...
   */
#ifdef DEBUGIO
  CDM_Debug("CDMIO_OpenCollectionSaveFile: Opening Collection Index file: "
            "\"%s\"\n", Tcl_DStringValue(&ds));
#endif
  File = 
      Tcl_OpenFileChannel(Collection->Interp, Tcl_DStringValue(&ds), "w", 0666);
  /*
   * If we failed to open the Collection Index File, make sure that the
   * Collection directory that will contain the Documents exists.
   * Then retry to open the Collection Index File...
   */
  if (File == NULL) {
    /*
     * Create the Collection's Host directory, if it doesn't already exist...
     */
    if (CDM_Mkdir(Collection->Name) != TCL_OK) {
      CDM_ReportError("CDMIO_OpenCollectionSaveFile: Cannot create directory "
                      "\"%s\"", Collection->Name);
      Tcl_DStringFree(&ds);
      return NULL;
    }
#ifdef DEBUGIO
    CDM_Debug("CDMIO_OpenCollectionSaveFile: Creating Collection Host "
              "Directory: \"%s\"\n", Collection->Name);
#endif
    File =
      Tcl_OpenFileChannel(Collection->Interp, Tcl_DStringValue(&ds), "w", 0666);
  }
  if (File == NULL) {
    /*
     * Place an error message on the Collection's interpreter...
     */
    CDM_ReportError("CDMIO_OpenCollectionSaveFile: Cannot open Collection "
                    "Index file: \"%s\"", Tcl_DStringValue(&ds));
    Tcl_DStringFree(&ds);
    return NULL;
  }
  Tcl_DStringFree(&ds);

  /*
   * Set the correct encoding. IF the Collection type is a GATE format, set
   * the encoding to the Collection encoding. Else, use the default CDM
   * encoding...
   */
  if (Tcl_SetChannelOption(Collection->Interp, File, "-encoding", encoding)
                                                                    != TCL_OK) {
     Tcl_Close(Collection->Interp, File);
     CDM_ReportError("CDMIO_OpenCollectionSaveFile: Cannot change the encoding "
                     "of the Collection Index file to \"%s\"", encoding);
     return NULL;
  }
  /*
   * Set the correct translation mode. Under all operating systems we use '\n'
   * as the character denoting end-of-line...
   */
  if (Tcl_SetChannelOption(Collection->Interp, File, "-translation",
                                                       translation) != TCL_OK) {
    Tcl_Close(Collection->Interp, File);
    CDM_ReportError("CDMIO_OpenCollectionSaveFile: Cannot change the "
                    "translation of the Document file to \"%s\"", translation);
    return NULL;
  }
#ifdef DEBUGIO
  CDM_Debug("CDMIO_OpenCollectionSaveFile: Collection Index file encoding "
            "set to \"%s\"\n", encoding);
#endif
  return File;
} /* CDMIO_OpenCollectionSaveFile */

/*
 * CDMIO_OpenCollectionReadFile:
 * This function opens the Collection index file, making decisions upon the
 * type of the Collection. The newly opened file will be set to the proper
 * encoding...
 * In case of an error, NULL will be returned and an error will be left at the
 * CDM_Interp interpreter...
 */
Tcl_Channel CDMIO_OpenCollectionReadFile(CDM_CONST char *path, int *format) {
  Tcl_Channel File;
  Tcl_DString ds;
  CDM_CONST char *argv[2];
  char *encoding = (char *) CDMIO_DefaultEncoding;
  CDM_THREAD_SPECIFIC_DATA;
  /* 
   * First of all, check for the native Collection type.
   * Check whether "CDM_CollectionFilename" exists in the given Path. If it
   * exists, open it...
   */
  *format = CDM_FORMAT_NATIVE2;
  Tcl_DStringInit(&ds);
  argv[0] = path; argv[1] = CDM_CollectionFilename;
  Tcl_JoinPath(2, argv, &ds);
  File = Tcl_OpenFileChannel(CDM_Interp, Tcl_DStringValue(&ds), "r", 0666);
  Tcl_DStringFree(&ds);

  if (File == NULL) {
    /*
     * We have failed. Is in GATE format?
     */
    *format = CDM_FORMAT_GATE_1x;
    Tcl_DStringInit(&ds);
    argv[0] = path; argv[1] = "GuT_mAgIc";
    Tcl_JoinPath(2, argv, &ds);
    File = Tcl_OpenFileChannel(CDM_Interp, Tcl_DStringValue(&ds), "r", 0666);
    Tcl_DStringFree(&ds);
    if (File) encoding = NULL; // Use the system encoding...
  }

  if (File == NULL) {
    /*
     * Finally, is it a registered format?
     */
    *format = CDM_FORMAT_REGISTERED;
    Tcl_DStringInit(&ds);
    argv[0] = path; argv[1] = CDM_CollectionTypeFilename;
    Tcl_JoinPath(2, argv, &ds);
    File = Tcl_OpenFileChannel(CDM_Interp, Tcl_DStringValue(&ds), "r", 0666);
    Tcl_DStringFree(&ds);
    if (File) encoding = (char *) CDMIO_DefaultEncoding;
  }
  
  if (File == NULL) {
    if (CDM_FileExists(path)) {
      CDM_ReportError("CDMIO_OpenCollectionReadFile: Unknown Collection Type!");
    } else {
      CDM_ReportError("CDMIO_OpenCollectionReadFile: Directory "
                      "\"%s\" does not exist!", path);
    }
    return NULL;
  }

  /*
   * Set the proper encoding...
   */
  if (encoding != NULL &&
      Tcl_SetChannelOption(CDM_Interp, File, "-encoding", encoding) != TCL_OK) {
    Tcl_Close(CDM_Interp, File);
    CDM_ReportError("CDMIO_OpenCollectionReadFile: Cannot change the encoding "
                    "of the Collection Index file to \"%s\"", encoding);
    return NULL;
  }

#ifdef WINDOWS
  /* Under windows, we have to disable ^Z (\x1a) as the EOF character,
     as our files can contain binary data... */
  Tcl_SetChannelOption(CDM_Interp, File, "-eofchar", "");
#endif /* WINDOWS */
  return File;
} /* CDMIO_OpenCollectionReadFile */

/*
 * CDMIO_OpenDocumentSaveFile:
 * This function opens a file suitable for storing a Document.
 * The newly opened file will be set to the proper encoding...
 * In case of an error, NULL will be returned and an error will be left at the
 * Document interpreter...
 */
Tcl_Channel CDMIO_OpenDocumentSaveFile(CDM_Document Document, char *Directory) {
  Tcl_Channel File;
  Tcl_DString ds;
  char *argv[2], *encoding = (char *) CDMIO_DefaultEncoding,
       *translation = (char *) CDMIO_DefaultTranslation;

  /*
   * Try to build the full path to the Document file...
   */
  Tcl_DStringInit(&ds);
  argv[0] = Directory; argv[1] = Document->DocumentPtr->ptr[CDM_DOCUMENT_ID];
  encoding = (char *) CDMIO_DefaultEncoding;
  Tcl_JoinPath(2, argv, &ds);

  /*
   * Try to open the Document Index file...
   */
#ifdef DEBUGIO
  CDM_Debug("CDMIO_OpenDocumentSaveFile: Opening Document file: "
            "\"%s\"\n", Tcl_DStringValue(&ds));
#endif
  File = 
      Tcl_OpenFileChannel(Document->Interp, Tcl_DStringValue(&ds), "w", 0666);
  /*
   * If we failed to open the Document Index File, make sure that the
   * Document directory that will contain the Documents exists.
   * Then retry to open the Document Index File...
   */
  if (File == NULL) {
    /*
     * Create the Document's Host directory, if it doesn't already exist...
     */
    if (CDM_Mkdir(Directory) != TCL_OK) {
      CDM_ReportError("CDMIO_OpenDocumentSaveFile: Cannot create directory "
                      "\"%s\"", Directory);
      Tcl_DStringFree(&ds);
      return NULL;
    }
#ifdef DEBUGIO
    CDM_Debug("CDMIO_OpenDocumentSaveFile: Creating Document Host "
              "Directory: \"%s\"\n", Directory);
#endif
    File =
      Tcl_OpenFileChannel(Document->Interp, Tcl_DStringValue(&ds), "w", 0666);
  }
  if (File == NULL) {
    /*
     * Place an error message on the Document's interpreter...
     */
    CDM_ReportError("CDMIO_OpenDocumentSaveFile: Cannot open Document "
                    "Index file: \"%s\"", Tcl_DStringValue(&ds));
    Tcl_DStringFree(&ds);
    return NULL;
  }
  Tcl_DStringFree(&ds);

  /*
   * Set the correct encoding. IF the Document type is a GATE format, set
   * the encoding to the Document encoding. Else, use the default CDM
   * encoding...
   */
  if (Tcl_SetChannelOption(Document->Interp, File, "-encoding", encoding)
                                                                   != TCL_OK) {
    Tcl_Close(Document->Interp, File);
    CDM_ReportError("CDMIO_OpenDocumentSaveFile: Cannot change the encoding "
                    "of the Document file to \"%s\"", encoding);
    return NULL;
  }
  /*
   * Set the correct translation mode. Under all operating systems we use '\n'
   * as the character denoting end-of-line...
   */
  if (Tcl_SetChannelOption(Document->Interp, File, "-translation", translation)
                                                                   != TCL_OK) {
    Tcl_Close(Document->Interp, File);
    CDM_ReportError("CDMIO_OpenDocumentSaveFile: Cannot change the translation"
                    " of the Document file to \"%s\"", translation);
    return NULL;
  }
#ifdef DEBUGIO
  CDM_Debug("CDMIO_OpenDocumentSaveFile: Document file encoding "
            "set to \"%s\"\n", encoding);
#endif
  return File;
} /* CDMIO_OpenDocumentSaveFile */

/*
 * CDMIO_WriteCollectionHeader:
 * This function embeds a specific header, identifying file as a Collection
 * file...
 */
int CDMIO_WriteCollectionHeader(Tcl_Channel File) {
  register int bytes;

  bytes = Tcl_WriteChars(File, CDM_CollectionHeader, -1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
  }
  return bytes;
} /* CDMIO_WriteCollectionHeader */

/*
 * CDMIO_WriteDocumentHeader:
 * This function embeds a specific header, identifying file as a Document
 * file...
 */
int CDMIO_WriteDocumentHeader(Tcl_Channel File) {
  register int bytes;

  bytes = Tcl_WriteChars(File, CDM_DocumentHeader, -1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
  }
  return bytes;
} /* CDMIO_WriteDocumentHeader */

/*
 * CDMIO_WriteField:
 * This function writes a field in the provided file...
 */
int CDMIO_WriteField(Tcl_Channel File, const char *name, const char *value) {
  char *ptr;
  register int bytes;

  bytes = Tcl_WriteChars(File, name, -1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
    return bytes;
  }
  /*
   * Is the value to be written containing a '\n'?
   */
  ptr = (char *) Tcl_UtfFindFirst(value, '\n');
  while (ptr) {
    bytes = Tcl_WriteChars(File, value, ptr - value);
    if (bytes == -1) {
      CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
      return bytes;
    }
    bytes = Tcl_WriteChars(File, "\\n", 2);
    if (bytes == -1) {
      CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
      return bytes;
    }
    value = (char *) Tcl_UtfNext(ptr);
    ptr   = (char *) Tcl_UtfFindFirst(value, '\n');
  }
  bytes = Tcl_WriteChars(File, value, -1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
  }
  return bytes;
} /* CDMIO_WriteField */
int CDMIO_WriteField(Tcl_Channel File, const char *name, int value) {
  register int bytes;
  CDM_THREAD_SPECIFIC_DATA;

  bytes = Tcl_WriteChars(File, name, -1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
    return bytes;
  }
  snprintf(CDM_Temp, MAX_TEMP_STRING, "%d", value);
  bytes = Tcl_WriteChars(File, CDM_Temp, -1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
  }
  return bytes;
} /* CDMIO_WriteField */

int CDMIO_WriteField(Tcl_Channel File, const char *name, long value) {
  register int bytes;
  CDM_THREAD_SPECIFIC_DATA;

  bytes = Tcl_WriteChars(File, name, -1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
    return bytes;
  }
  snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", value);
  bytes = Tcl_WriteChars(File, CDM_Temp, -1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
  }
  return bytes;
} /* CDMIO_WriteField */

/*
 * CDMIO_WriteLengthedObject:
 * This function writes a label and an object. Before the object, the lenth of
 * its value is calculated and written...
 */
int CDMIO_WriteLengthedObject(Tcl_Channel File, const char *fmt, Tcl_Obj *value) {
  const char* string;
  register int bytes;
  CDM_THREAD_SPECIFIC_DATA;
  
  string = Tcl_GetStringFromObj(value, &bytes);
  snprintf(CDM_Temp, MAX_TEMP_STRING, fmt, 
           Tcl_NumUtfChars(string, bytes));
  bytes = Tcl_WriteChars(File, CDM_Temp, -1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
    return bytes;
  }
  bytes = Tcl_WriteObj(File, value);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
  }
  return bytes;
} /* CDMIO_WriteLengthedObject */

/*
 * CDMIO_WriteAttributes:
 * This function writes an object holding Attributes into the specified
 * file...
 */
int CDMIO_WriteAttributes(Tcl_Channel File, Tcl_Obj *Attributes) {
  Tcl_Obj **Attribute;
  char *string, *ptr;
  int items;
  register int bytes, i;
  CDM_THREAD_SPECIFIC_DATA;

  if (Tcl_ListObjGetElements(CDM_Interp, Attributes, &items, &Attribute) 
     != TCL_OK) return -1;
  snprintf(CDM_Temp, MAX_TEMP_STRING, "\nAttributes::%d\n", items);
  bytes = Tcl_WriteChars(File, CDM_Temp, -1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
    return bytes;
  }
  for (i=0; i<items; i++) {
    string = Tcl_GetString(Attribute[i]);
    /*
     * Is the string to be written containing a '\n'?
     */
    ptr = (char *) Tcl_UtfFindFirst(string, '\n');
    while (ptr) {
      bytes = Tcl_WriteChars(File, string, ptr - string);
      if (bytes == -1) {
        CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
        Tcl_InvalidateStringRep(Attribute[i]);
        return bytes;
      }
      bytes = Tcl_WriteChars(File, "\\n", 2);
      if (bytes == -1) {
        CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
        Tcl_InvalidateStringRep(Attribute[i]);
        return bytes;
      }
      string = (char *) Tcl_UtfNext(ptr);
      ptr    = (char *) Tcl_UtfFindFirst(string, '\n');
    }
    bytes = Tcl_WriteChars(File, string, -1);
    //Tcl_WriteObj(File, Attribute[i]);
    Tcl_InvalidateStringRep(Attribute[i]);
    if (bytes == -1) {
      CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
      return bytes;
    }
    bytes = Tcl_WriteChars(File, "\n", 1);
    if (bytes == -1) {
      CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
      return bytes;
    }
  }
  return bytes;
} /* CDMIO_WriteAttributes */

/*
 * CDMIO_WriteAnnotations:
 * This function writes an object holding Annotations into the specified
 * file...
 */
int CDMIO_WriteAnnotations(Tcl_Channel File, Tcl_Obj *Annotations) {
  Tcl_Obj **Annotation;
  char *string, *ptr;
  int items;
  register int bytes, i;
  CDM_THREAD_SPECIFIC_DATA;

  if (Tcl_ListObjGetElements(CDM_Interp, Annotations, &items, &Annotation) 
     != TCL_OK) return -1;
  snprintf(CDM_Temp, MAX_TEMP_STRING, "\nAnnotations::%d\n", items);
  bytes = Tcl_WriteChars(File, CDM_Temp, -1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
    return bytes;
  }
  for (i=0; i<items; i++) {
    string = Tcl_GetString(Annotation[i]);
    /*
     * Is the string to be written containing a '\n'?
     */
    ptr = (char *) Tcl_UtfFindFirst(string, '\n');
    while (ptr) {
      bytes = Tcl_WriteChars(File, string, ptr - string);
      if (bytes == -1) {
        CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
        Tcl_InvalidateStringRep(Annotation[i]);
        return bytes;
      }
      bytes = Tcl_WriteChars(File, "\\n", 2);
      if (bytes == -1) {
        CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
        Tcl_InvalidateStringRep(Annotation[i]);
        return bytes;
      }
      string = (char *) Tcl_UtfNext(ptr);
      ptr    = (char *) Tcl_UtfFindFirst(string, '\n');
    }
    bytes = Tcl_WriteChars(File, string, -1);
    //Tcl_WriteObj(File, Attr[i]);
    Tcl_InvalidateStringRep(Annotation[i]);
    if (bytes == -1) {
      CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
      return bytes;
    }
    bytes = Tcl_WriteChars(File, "\n", 1);
    if (bytes == -1) {
      CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
      return bytes;
    }
  }
  return bytes;
} /* CDMIO_WriteAnnotations */

/*
 * CDMIO_WriteRawData:
 * This function writes an object holding RawData into the specified
 * file...
 */
int CDMIO_WriteRawData(Tcl_Channel File, Tcl_Obj *RawData) {
  return CDMIO_WriteLengthedObject(File, "\nRawData::%d\n", RawData);
#if 0
  char *string, *ptr;
  register int bytes;

  string = Tcl_GetStringFromObj(RawData, &bytes);
  snprintf(CDM_Temp, MAX_TEMP_STRING, "\nRawData::%d\n", 
           Tcl_NumUtfChars(string, bytes));
  bytes = Tcl_WriteChars(File, CDM_Temp, -1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
    return bytes;
  }
  
  /*
   * Is the string to be written containing a '\r'?
   */
  ptr = (char *) Tcl_UtfFindFirst(string, '\r');
  while (ptr) {
    bytes = Tcl_WriteChars(File, string, ptr - string);
    if (bytes == -1) {
      CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
      return bytes;
    }
    bytes = Tcl_WriteChars(File, "\\r", 2);
    if (bytes == -1) {
      CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
      return bytes;
    }
    string = (char *) Tcl_UtfNext(ptr);
    ptr    = (char *) Tcl_UtfFindFirst(string, '\r');
  }
  bytes = Tcl_WriteChars(File, string, -1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
    return bytes;
  }
  bytes = Tcl_WriteChars(File, "\n", 1);
  if (bytes == -1) {
    CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
    return bytes;
  }
  
  return bytes;
#endif
} /* CDMIO_WriteRawData */

/*
 * CDMIO_DisplayAppendProgressDialog:
 * This function will display a progress dialog displaying a customised,
 * localised message, if found...
 */
int CDMIO_DisplayAppendProgressDialog(Tcl_Interp *interp, const char *key,
                                      const char *defMessage) {
  Tcl_Obj *Message, *objv[4];
  int status, i;

  /*
   * Try to get the requested localised message from the "GUI_Text" global
   * array. If this fails, use the provided fall-back message...
   */
  Message = Tcl_GetVar2Ex(interp, "GUI_Text", key, TCL_GLOBAL_ONLY);
  if (Message == NULL) Message = CDM_NewStringObj(defMessage);
  
  objv[0] = CDM_NewStringObj("ggi_wait_append");
  objv[1] = Message;
  objv[2] = CDM_NewLongObj(0);
  for(i=0; i<3; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(interp, 3, objv, TCL_EVAL_GLOBAL);
  for(i=0; i<3; i++) Tcl_DecrRefCount(objv[i]);
  return status;
} /* CDMIO_DisplayAppendProgressDialog */

/*
 * CDMIO_UpdateProgressDialog:
 * This function will update a progress dialog displayed by 
 * CDMIO_DisplayAppendProgressDialog...
 */
int CDMIO_UpdateProgressDialog(Tcl_Interp *interp, int progress,
                               Tcl_Obj *win) {
  Tcl_Obj *objv[4];
  int status, i;
  CDM_THREAD_SPECIFIC_DATA;

  objv[0] = CDM_NewStringObj("ggi_wait_update");
  objv[1] = &CDM_Obj_NULL;
  objv[2] = CDM_NewLongObj(progress);
  objv[3] = win;
  for(i=0; i<4; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(interp, 4, objv, TCL_EVAL_GLOBAL);
  for(i=0; i<4; i++) Tcl_DecrRefCount(objv[i]);
  Tcl_DoOneEvent(TCL_IDLE_EVENTS|TCL_WINDOW_EVENTS|TCL_DONT_WAIT);
  return status;
} /* CDMIO_UpdateProgressDialog */

/*
 * CDMIO_DestroyAppendProgressDialog:
 * This function destroys a progress dialog created with 
 * CDMIO_DisplayAppendProgressDialog...
 */
int CDMIO_DestroyAppendProgressDialog(Tcl_Interp *interp, Tcl_Obj *win) {
  Tcl_Obj *objv[2];
  int status, i;
  CDM_THREAD_SPECIFIC_DATA;

  objv[0] = CDM_NewStringObj("ggi_destroy_append");
  objv[1] = win;
  for(i=0; i<1; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(CDM_Interp, 2, objv, TCL_EVAL_GLOBAL);
  for(i=0; i<2; i++) Tcl_DecrRefCount(objv[i]);
  return status;
} /* CDMIO_DestroyAppendProgressDialog */

/*
 * End of File - EOF
 */
