/*****************************************************************************
 *  CDMIO.cpp:                                                               *
 * ------------                                                              *
 *  This file implements the File I/O of CDM. It implements the CDM file     *
 *  storage format (version 1 & 2), as well as the GATE storage format (1.x).*
 *  Additionally, it offers support for embedable components that implement  *
 *  additional storage formats. An example component is implemented, able to *
 *  support an XML-like format (file XMLIO.cpp)...                           *
 *                                                                           *
 *  Copyright 1998-2008 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licencing information, please refer to the licence conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/

#include "CDMIO.h"
#include "CDM_GlobalObjects.h"
#include <string.h>
extern Tcl_ThreadDataKey dataKey;

/*
 * This file contains Functions that Handle the Physical Storring and
 * Retrieval of Documents & Collections. They are internal parts of CDM: The
 * end-user is not supposed to use them.
 */
extern int          CDM_GetAttributeTypeFromObject(Tcl_Obj *Type);
extern void         CDM_DocumentInternalAddAnnotation(CDM_Document Document,
                                                      CDM_Annotation Ann);

/*
 * CDM I/O Function Prototypes....
 */
int            CDM_SaveCollection(CDM_Collection Collection);
CDM_Collection CDM_LoadCollection(CDM_CONST char *Path);
int            CDM_SaveDocument(CDM_Document Document, char *Directory);
CDM_Document   CDM_LoadDocument(CDM_Collection Collection, char *Path);

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     C O L L E C T I O N            S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 * -- CDM_SaveCollection:
 *    Saves given Collection in disk.
 *    Arguments: 
 *      Collection: The Collection's that will be saved pointer (CDM_Collection)
 */
int CDM_SaveCollection(CDM_Collection Collection) {
  CDM_ListNode *ptr;
  Tcl_Channel File;
  Tcl_Obj *Obj, **listv, *Win=NULL;
  Tcl_Obj *DocumentIDs, *DocumentXIDs;
  double progress=0.0, last_prgrss=0.0, step=0.0;
  int  i, status, listc, last_progress;
  CDM_THREAD_SPECIFIC_DATA;

  if (Collection == NULL) 
    return CDM_ReportError("CDM_SaveCollection: NULL Collection Pointer");
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    return CDM_ReportError("CDM_SaveCollection: Invalid Collection Type");
  }
  Tcl_ResetResult(Collection->Interp);
  CDM_Interp = Collection->Interp;

  /*
   * Open the Collection Index File...
   */
  File = CDMIO_OpenCollectionSaveFile(Collection);
  if (File == NULL) goto error;
  
  /* Display Progress Dialog... */
  if (Collection->FileFormat == CDM_FORMAT_GATE_1x ||
      Collection->FileFormat == CDM_FORMAT_GATE_2x) {
    status = CDMIO_DisplayAppendProgressDialog(CDM_Interp,
                   "SavingGateCollection",
                   "Saving Collection (GATE Format).\nPlease Wait...");
  } else {
    status = CDMIO_DisplayAppendProgressDialog(CDM_Interp, "SavingCollection",
                   "Saving Collection.\nPlease Wait...");
  }
  if (status == TCL_OK) {
    Win = Tcl_GetObjResult(CDM_Interp);
    Tcl_IncrRefCount(Win);
  }

  switch (Collection->FileFormat) {
    case CDM_FORMAT_NATIVE:
    case CDM_FORMAT_NATIVE2: {
      /* Write Collection Header... */
      if (CDMIO_WriteCollectionHeader(File) == -1) goto error;
      /* Write various Collection fields... */
      if (CDMIO_WriteField(File, "\nEncoding::",   Collection->Encoding) == -1)
         goto error;
      if (CDMIO_WriteField(File, "\nFileFormat::", CDM_FORMAT_NATIVE2)   == -1)
         goto error;
      if (CDMIO_WriteField(File, "\nTranslation::", CDMIO_DefaultTranslation)
         == -1) goto error;
      if (CDMIO_WriteField(File, "\nName::",       Collection->Name)     == -1)
         goto error;
      if (CDMIO_WriteField(File, "\nOwner::",      Collection->Owner)    == -1)
         goto error;
      /* Write Comment, if one exists... */
      Obj = CDM_GetComment(Collection);
      if (Obj) {
        if (CDMIO_WriteLengthedObject(File, "\nComment::%d\n", Obj)      == -1)
           goto error;
      }
      /* Write Collection Documents IDs & XIDs... */
      DocumentIDs = CDM_ListGetObject(Collection->Documents, CDM_DOCUMENT_ID);
      if (CDMIO_WriteLengthedObject(File, "\nDocumentIDs::%d\n", DocumentIDs)
         == -1) goto error;
      Tcl_DecrRefCount(DocumentIDs);
      DocumentXIDs = CDM_ListGetObject(Collection->Documents, CDM_DOCUMENT_XID);
      if (CDMIO_WriteLengthedObject(File, "\nDocumentXIDs::%d\n", DocumentXIDs)
         == -1) goto error;
      Tcl_DecrRefCount(DocumentXIDs);
      /* Write Attributes */
      Obj = CDM_GetAttributes(Collection);
      if (Obj == NULL) goto error;
      if (CDMIO_WriteAttributes(File, Obj) == -1) goto error;
      Tcl_DecrRefCount(Obj);
      
      Tcl_WriteChars(File, "\n## - End of File -", -1);
      break;
    }
    case CDM_FORMAT_GATE_1x:
    case CDM_FORMAT_GATE_2x: {
      status = CDM_SaveGateCollection(Collection, File);
      if (status != TCL_OK) goto error;
      break;
    }
    default: {
      char *type;
      int format = 0;
      /*
       * This is a registered format. First, try to get the format type
       * name...
       */
      if (Collection->FileFormat < CDM_FORMAT_REGISTERED)  goto error;
      format = Collection->FileFormat - CDM_FORMAT_REGISTERED;
      if (format >= CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER) goto error;
      if (CDM_AdditionalCollectionTypes[format].CollectionType == NULL) {
        CDM_ReportError("CDM_SaveCollection: Collection has an unregistered "
                        "format!");
        goto error;
      }
      type = CDM_AdditionalCollectionTypes[format].CollectionType;
      Tcl_WriteChars(File, type, -1);
      Tcl_WriteChars(File, "\n",  1);
      /*
       * We have found the entry matching our type. If the corresponding
       * function pointer is not NULL, call it (and hope it doesn't crash
       * :-))
       */
      if (CDM_AdditionalCollectionTypes[format].saveCollectionPtr != NULL) {
        status = (CDM_AdditionalCollectionTypes[format].saveCollectionPtr)
                  (Collection);
        if (status != TCL_OK) goto error;
      }
      break;
    }
  } /* switch (Collection->FileFormat) */

  /*
   * We have written the Collection Index file...
   */
#ifdef DEBUGIO
  CDM_Debug("CDM_SaveCollection: File Written - Closing...\n", (char *) NULL);
#endif
  if (Tcl_Close(CDM_Interp, File) != TCL_OK) return TCL_ERROR;
  if (Win != NULL) {
    CDMIO_UpdateProgressDialog(CDM_Interp, 5, Win);
  }

  /*
   * There is the possibility that some Documents were removed from the
   * Collection. Remove them also from the disk, if available...
   */
#ifdef DEBUGIO
  CDM_Debug("CDM_SaveCollection: Start Deleting Removed Documents...\n",
            (char *) NULL);
#endif
  /* Remove Documents that meant to be removed from disk... */
  if (Collection->RemoveDiskDocuments != NULL) {
    status = Tcl_ListObjGetElements(CDM_Interp, Collection->RemoveDiskDocuments,
                                    &listc, &listv);
    if (status != TCL_OK) return TCL_ERROR;
    for (i = 0; i < listc; i++) {
      /*
       * If this Collection is new and has not been saved yet, then the 
       * Document should not exist in disk. So, don't report an error on
       * failure...
       */
#ifdef DEBUGIO
      CDM_Debug("CDM_SaveCollection: Deleting Document \"%s\"...\n", 
                Tcl_GetString(listv[i]));
#endif
      CDM_DeleteFile(Tcl_GetString(listv[i]));
    }
    /* We have removed all Documents that were meant to be removed. Clear the
     * list of Documents to be removed. */
    Tcl_DecrRefCount(Collection->RemoveDiskDocuments);
    Collection->RemoveDiskDocuments = NULL;
  }
  
#ifdef DEBUGIO
  CDM_Debug("CDM_SaveCollection: Saving Opened Documents...\n");
#endif
  /*
   * Save Opened Documents...
   */
  ptr   = Collection->Documents;
  listc = CDM_Length(Collection);
  /* Update Status Dialog... */
  last_progress = 5;
  last_prgrss = progress = float(last_progress);
  if (listc) {
    step = (100.0 - last_progress)/float(listc);
  } else {
    step = (100.0 - last_progress);
  }
  while (ptr != NULL) {
    if (ptr->ptr[CDM_DOCUMENT_OPENED] != NULL) {
#ifdef DEBUGIO
      CDM_Debug("CDM_SaveCollection: Saving Opened Document %s...\n",
                ptr->ptr[CDM_DOCUMENT_ID], (char *) NULL);
#endif
      if (CDM_SaveDocument((CDM_Document) ptr->ptr[CDM_DOCUMENT_POINTER],
                           Collection->Name) != TCL_OK) return TCL_ERROR;
    }
    progress += step;
    if (Win != NULL && progress > last_prgrss+5.0) {
      CDMIO_UpdateProgressDialog(CDM_Interp, int(progress), Win);
      last_prgrss = progress;
    }
    ptr = ptr->next;
  } /* while (ptr != NULL) */
  
  Collection->Modified = Collection->ModifiedDocuments = 0;
  if (Win != NULL) {
    CDMIO_DestroyAppendProgressDialog(CDM_Interp, Win);
    Win = NULL;
  }
  Tcl_ResetResult(CDM_Interp);
  return TCL_OK;
error:
  if (File) Tcl_Close(CDM_Interp, File);
  if (Win != NULL) {
    CDMIO_DestroyAppendProgressDialog(CDM_Interp, Win);
  }
  return TCL_ERROR;
} /* CDM_SaveCollection */

/*
 * -- CDM_LoadCollection:
 *    Loads requested Collection from the disk.
 *    Arguments: 
 *      Path:  The Collection's Disk address                 (char *)
 */
CDM_Collection CDM_LoadCollection(CDM_CONST char *Path) {
  CDM_Collection NewCollection;
  Tcl_Channel File;
  Tcl_Obj *LineObj, *ListObj, **elemPtrs;
  Tcl_Obj *DocumentIDs, *DocumentXIDs;
  char *Encoding = (char *) CDMIO_DefaultEncoding, *Line = NULL, 
        CmdName[MAX_TEMP_STRING];
  int Obj_ID, chars,  status, i, listLen, format, items;
  CDM_THREAD_SPECIFIC_DATA;


  if (Path == NULL) {
    CDM_ReportError("CDM_LoadCollection: NULL Path");
    return NULL;
  }
  
  /*
   * Open the Collection Index file. According to the file found, the
   * Collection type (format) is set to a proper value...
   */
  File = CDMIO_OpenCollectionReadFile(Path, &format);
  if (File == NULL) return NULL; 

  /*
   * Create the New Collection Object...
   */
  NewCollection                       = CDM_NewCollectionObj();
  NewCollection->Interp               = CDM_Interp;
  Obj_ID = ID.CollectionUniqID();
  snprintf(CmdName, MAX_TEMP_STRING, "Col%d", Obj_ID);
  NewCollection->Name                 = NULL;
  NewCollection->Owner                = NULL;
  NewCollection->Documents            = NULL;
  NewCollection->RemoveDiskDocuments  = Tcl_NewListObj(0, NULL);
  if (NewCollection->RemoveDiskDocuments == NULL) 
      CDM_Panic("CDM_LoadCollection: out of memory!\n");
  Tcl_IncrRefCount(NewCollection->RemoveDiskDocuments);
  NewCollection->CurrentDocumentIndex = NULL;
  NewCollection->Modified             = NewCollection->ModifiedDocuments = 0;
  NewCollection->FileFormat           = (CDM_UCHAR) format;
  
  /*
   * Initialise Objects...
   */
  LineObj     = Tcl_NewObj(); ListObj      = Tcl_NewListObj(0, NULL);
  DocumentIDs = Tcl_NewObj(); DocumentXIDs = Tcl_NewObj();
  if (LineObj     == NULL || ListObj      == NULL || 
      DocumentIDs == NULL || DocumentXIDs == NULL) {
    CDM_Panic("CDM_LoadCollection: Tcl_NewObj() returns NULL - out of memory?");
  }
  Tcl_IncrRefCount(LineObj);     Tcl_IncrRefCount(ListObj);
  Tcl_IncrRefCount(DocumentIDs); Tcl_IncrRefCount(DocumentXIDs);
  
  switch (format) {
    case CDM_FORMAT_NATIVE:
    case CDM_FORMAT_NATIVE2: {
      chars = Tcl_GetsObj(File, LineObj);
      if (chars > 0) Line = Tcl_GetString(LineObj);
      if (chars < 1 || (strncmp(Line, "#Ellogon:", 9) != 0 && 
                        strncmp(Line, "#Logos:"  , 7) != 0)) {
        Tcl_DecrRefCount(LineObj);
        Tcl_DecrRefCount(ListObj);
        Tcl_Close(CDM_Interp, File);
        CDM_Close(NewCollection);
        CDM_ReportError("CDM_LoadCollection: Invalid Collection File: \"%s\"",
                        Path);
        return NULL;
      }
      
      /*
       * Read File data...
       */
      while (chars > -1) {
        // Empty object, but keep allocated space...
        Tcl_SetObjLength(LineObj, 0);
        chars = Tcl_GetsObj(File, LineObj);
        if (chars < 0 || Tcl_Eof(File)) break; // EOF or ERROR...
        Line = Tcl_GetString(LineObj);
        if (strlen(Line) == 0 || Line[0] == '#' || Line[0] == ' ')
           continue; // Skip unwanted lines
        
        /***************** Encoding *****************/
        if (strncmp(Line, "Encoding", 8) == 0) {
          NewCollection->Encoding = CDM_Strdup(&Line[10]);
#ifdef DEBUGIO
          CDM_Debug("Encoding = \"%s\"\n", NewCollection->Encoding,
                    (char *) NULL);
#endif /* DEBUGIO */
          continue;
        }

        /***************** FileFormat *****************/
        if (strncmp(Line, "FileFormat", 10) == 0) {
          sscanf(&Line[12], "%d", &status);
          NewCollection->FileFormat = (CDM_UCHAR) status;
#ifdef DEBUGIO
          CDM_Debug("FileFormat = %d\n", (int) NewCollection->FileFormat,
                    (char *) NULL);
#endif /* DEBUGIO */
          continue;
        }

        /***************** Translation *****************/
        if (strncmp(Line, "Translation", 11) == 0) {
#ifdef DEBUGIO
          CDM_Debug("Translation = \"%s\"\n", &Line[13], (char *) NULL);
#endif /* DEBUGIO */
          Tcl_SetChannelOption(NULL, File, "-translation", &Line[13]);
          continue;
        } /* Translation */

        /***************** Name *****************/
        if (strncmp(Line, "Name", 4) == 0) {
          NewCollection->Name = CDM_Strdup(&Line[6]);
#ifdef DEBUGIO
          CDM_Debug("Name = \"%s\"\n", NewCollection->Name, (char *) NULL);
#endif /* DEBUGIO */
          continue;
        }

        /***************** Owner *****************/
        if (strncmp(Line, "Owner", 5) == 0) {
          NewCollection->Owner = CDM_Strdup(&Line[7]);
#ifdef DEBUGIO
          CDM_Debug("Owner = \"%s\"\n", NewCollection->Owner, (char *) NULL);
#endif /* DEBUGIO */
          continue;
        }

        /***************** Comment *****************/
        if (strncmp(Line, "Comment", 7) == 0) {
          if (sscanf(Line, "Comment::%d", &items) < 1) {
            CDM_ReportError("CDM_LoadCollection: Cannot read number of Comment "
                            "characters!");
            continue;
          }
          if (items) {
            chars = Tcl_ReadChars(File, LineObj, items, 0);
            CDM_SetComment(NewCollection, Tcl_DuplicateObj(LineObj));
            NewCollection->Modified = 0;
          } else {
            chars = Tcl_GetsObj(File, LineObj);
          }
          if (chars < 0) break;
          continue;
        }

        /***************** DocumentIDs *****************/
        if (strncmp(Line, "DocumentIDs", 11) == 0) {
          if (sscanf(Line, "DocumentIDs::%d", &items) < 1) {
            CDM_ReportError("CDM_LoadCollection: Cannot read length of "
                            "Document IDs");
            continue;
          }
          if (items) {
            chars = Tcl_ReadChars(File, DocumentIDs, items, 0);
          }
#ifdef DEBUGIO
          CDM_Debug("DocumentIDs = \"%s\"\n", 
                    Tcl_GetString(DocumentIDs), (char *) NULL);
#endif /* DEBUGIO */
          continue;
        }
        
        /***************** DocumentXIDs *****************/
        if (strncmp(Line, "DocumentXIDs", 12) == 0) {
          register CDM_ListNode *ptr, *lastNode;
          Tcl_Obj **Doc_IDs, **Doc_XIDs;
          int id_c, xid_c;

          if (sscanf(Line, "DocumentXIDs::%d", &items) < 1) {
            CDM_ReportError("CDM_LoadCollection: Cannot read length of Document"
                            " XIDs");
            continue;
          }
          if (items) {
            chars = Tcl_ReadChars(File, DocumentXIDs, items, 0);
          }
#ifdef DEBUGIO
          CDM_Debug("DocumentXIDs = \"%s\"\n",
                    Tcl_GetString(DocumentXIDs), (char *) NULL);
#endif /* DEBUGIO */
          
          /*
           * We have read the Document IDs & XIDs in two objects.
           * Time to populate the Collection->Documents slot...
           */
          if (Tcl_ListObjGetElements(NULL, DocumentIDs, &id_c, &Doc_IDs) !=
              TCL_OK) {
            CDM_ReportError("CDM_LoadCollection: Invalid Document ID list!");
            goto error;
          }
          if (Tcl_ListObjGetElements(NULL, DocumentXIDs, &xid_c, &Doc_XIDs) !=
              TCL_OK) {
            CDM_ReportError("CDM_LoadCollection: Invalid Document XID list!");
            goto error;
          }
          if (id_c != xid_c) {
            CDM_ReportError("CDM_LoadCollection: Number of Document Ids is not "
                            "equal with number of Document XIds!");
            goto error;
          }
          /*
           * Foreach entry, add a list node...
           */
          if (id_c > 0) {
            ptr = CDM_NewListNode();
            ptr->ptr[CDM_DOCUMENT_ID]  =
                     CDM_Strdup(Tcl_GetString(Doc_IDs[0]));
            ptr->ptr[CDM_DOCUMENT_XID] = 
                     CDM_Strdup(Tcl_GetString(Doc_XIDs[0]));
            NewCollection->Documents = lastNode = CDM_AppendListNode(NULL, ptr);
            if (id_c > 1) {
              for (xid_c=1; xid_c<id_c; xid_c++) {
                ptr = CDM_NewListNode();
                ptr->ptr[CDM_DOCUMENT_ID]  =
                         CDM_Strdup(Tcl_GetString(Doc_IDs[xid_c]));
                ptr->ptr[CDM_DOCUMENT_XID] = 
                         CDM_Strdup(Tcl_GetString(Doc_XIDs[xid_c]));
                lastNode = CDM_AppendListNode(lastNode, ptr);
              }
            } /* if (id_c > 1) */
          } /* if (id_c > 0) */
          continue;
        }

        /***************** Attributes *****************/
        if (strncmp(Line, "Attributes", 10) == 0) {
          switch (NewCollection->FileFormat) {
            case CDM_FORMAT_NATIVE: {
              /*
               * Native version 1 format:
               */
              if (sscanf(Line, "Attributes::%d", &items) < 1) {
                CDM_ReportError("CDM_LoadCollection: Cannot read length of "
                                "Collection Attributes");
                continue;
              }
              if (items) {
                chars = Tcl_ReadChars(File, LineObj, items, 0);
                Tcl_SetStringObj(ListObj, Tcl_GetString(LineObj), -1);
              } else {
                chars = Tcl_GetsObj(File, LineObj);
              }
              if (chars == -1) {
                if (Tcl_Eof(File)) {
                  CDM_ReportError("CDM_LoadCollection: Unexpected EOF, while "
                                  "processing Collection Attributes");
                  goto error;
                } else {
                  CDM_ReportError("CDM_LoadCollection: Error reading file, "
                                  "while processing Collection Attributes");
                  goto error;
                }
              }
              if (items) {
                status = Tcl_ListObjGetElements(CDM_Interp, ListObj,
                                                &listLen, &elemPtrs);
                if (status != TCL_OK) {
                  snprintf(CDM_Temp, MAX_TEMP_STRING, "CDM_LoadCollection: "
                      "Invalid Attribute List: \"%s\"", Tcl_GetString(LineObj));
                  CDM_ReportError(CDM_Temp);
                  Tcl_SetListObj(ListObj, 0, NULL);
                  //Tcl_SetStringObj(ListObj, NULL, 0); // Empty object...
                  continue;
                }
                for (i=0; i<listLen; i++) {
#ifdef RECONSTRUCT_ATTRIBUTES
                  CDM_Attribute RecAttr;
                  RecAttr = CDM_ReconstructAttribute(elemPtrs[i]);
                  if (RecAttr == NULL) continue;
                  Tcl_IncrRefCount(RecAttr);
                  CDM_PutAttribute(NewCollection, RecAttr);
                  Tcl_DecrRefCount(RecAttr);
#else  /* RECONSTRUCT_ATTRIBUTES */
                  CDM_PutAttribute(NewCollection, elemPtrs[i]);
#endif /* RECONSTRUCT_ATTRIBUTES */
                }
                Tcl_SetListObj(ListObj, 0, NULL);
#ifdef DEBUGIO
                CDM_Debug("Attributes = \"%s\"\n", Tcl_GetString(LineObj));
#endif /* DEBUGIO */
              }
              break;
            }
            case CDM_FORMAT_NATIVE2: {
              /*
               * Native version 2 format:
               */
              if (sscanf(Line, "Attributes::%d", &items) < 1) {
                CDM_ReportError("CDM_LoadCollection: Cannot read number of "
                                "Attributes");
                goto error;
              }
#ifdef DEBUGIO
              CDM_Debug("Number of Attributes: %d\n", items, (char *) NULL);
#endif /* DEBUGIO */
              /*
               * Add Attributes into Collection...
               */
              Tcl_SetObjLength(LineObj, 0);
              for (i=0; i<items; i++) {
                chars = Tcl_GetsObj(File, LineObj);
                if (chars < 0) {
                  if (Tcl_Eof(File)) {
                    CDM_ReportError("CDM_LoadCollection: Unexpected EOF, while "
                                    "processing Collection Attributes");
                    goto error;
                  } else {
                    CDM_ReportError("CDM_LoadCollection: Error reading File, "
                                    "while processing Collection Attributes");
                    goto error;
                  }
                }
#ifdef RECONSTRUCT_ATTRIBUTES
                CDM_Attribute AttrObj;
                char *AttrName;
                int   AttrType, attrValLen;
                Tcl_Obj **AttrElem;

                Tcl_SetStringObj(ListObj, Tcl_GetString(LineObj), -1);
                /*
                 * ****************************************************
                 *   Re-create Attribute Set... (CDM_FORMAT_NATIVE2)
                 * ****************************************************
                 */
                status = Tcl_ListObjGetElements(CDM_Interp, ListObj,
                                                &attrValLen, &AttrElem);
                if (status != TCL_OK || attrValLen != 2) continue;
                AttrName = Tcl_GetString(AttrElem[0]);
                status = Tcl_ListObjGetElements(CDM_Interp, AttrElem[1],
                                                &attrValLen, &AttrElem);
                if (status != TCL_OK || attrValLen < 2) continue;
                AttrType = CDM_GetAttributeTypeFromObject(AttrElem[0]);
                if (AttrType < CDM_STRING) continue;
                AttrObj = CDM_CreateAttributeValue(AttrType, AttrElem[1]);
                if (AttrObj == NULL) continue;
                AttrObj = CDM_CreateAttribute(AttrName, AttrObj);
                if (AttrObj == NULL) continue;
                Tcl_IncrRefCount(AttrObj);
                CDM_PutAttribute(NewCollection, AttrObj);
                Tcl_DecrRefCount(AttrObj);
#else  /* RECONSTRUCT_ATTRIBUTES */
                CDM_PutAttribute(NewCollection, LineObj);
#endif /* RECONSTRUCT_ATTRIBUTES */
                Tcl_SetObjLength(LineObj, 0);
                Tcl_SetListObj(ListObj, 0, NULL);
              }
              break;
            } /* case CDM_FORMAT_NATIVE2 */
          } /* switch (NewCollection->FileFormat) */
          break; // We have red anything. Stop reading further...
        } /* if (strncmp(Line, "Attributes", 10) == 0) */
      } /* while (chars > -1) */
      break;
    } /* case CDM_FORMAT_NATIVE */
    case CDM_FORMAT_GATE_1x:
    case CDM_FORMAT_GATE_2x: {
      status = CDM_LoadGateCollection(NewCollection, File);
      //if (status != TCL_OK) goto error;
      break;
    }
    case CDM_FORMAT_REGISTERED: {
      int found = 0;
      /*
       * This is a registered format. First, try to get the format type
       * name...
       */
      chars = Tcl_GetsObj(File, LineObj);
      if (chars <= 0) {
        CDM_ReportError("Cannot read registered Collection type!");
        goto error;
      }
      Line = Tcl_GetString(LineObj);
      /*
       * Search all registered plugins for this type...
       */
      for (i=0; i<CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER; i++) {
        if (CDM_AdditionalCollectionTypes[i].CollectionType == NULL) continue;
        if (CDM_Strcmp(CDM_AdditionalCollectionTypes[i].CollectionType, Line)
                                                            != 0)    continue;
                                                           
        /*
         * We have found the entry matching our type. If the corresponding
         * function pointer is not NULL, call it (and hope it doesn't crash
         * :-))
         */
        found = 1;
        NewCollection->FileFormat = (CDM_UCHAR) (i + CDM_FORMAT_REGISTERED);
        if (CDM_AdditionalCollectionTypes[i].loadCollectionPtr != NULL) {
          NewCollection->Name = CDM_Strdup(Path);
          status = (CDM_AdditionalCollectionTypes[i].loadCollectionPtr)
                        (NewCollection);
          if (status != TCL_OK) goto error;
        }
        break;
      }
      /*
       * It is an error if a plugin of the wanted type is not found...
       */
      if (!found) {
        CDM_ReportError("CDM_LoadCollection: The Collection with External Id "
                        "\"%s\" is of type \"%s\". A plug-in component for "
                        "this type has not been registered.", Path, Line);
        goto error;
      }
      break;
    }
  }
  /*
   * Do some sanity! checks, in order to (partially!) recover from damaged
   * files...
   */
  if (NewCollection->Encoding == NULL) {
    if (NewCollection->FileFormat == CDM_FORMAT_NATIVE ||
        NewCollection->FileFormat == CDM_FORMAT_NATIVE2) 
         NewCollection->Encoding = CDM_Strdup(Encoding);
    else NewCollection->Encoding = CDM_Strdup("iso8859-1");
  }
  if (NewCollection->Owner == NULL) NewCollection->Owner = CDM_Strdup("CDM");
  
  Tcl_DecrRefCount(LineObj);     Tcl_DecrRefCount(ListObj);
  Tcl_DecrRefCount(DocumentIDs); Tcl_DecrRefCount(DocumentXIDs);
  Tcl_Close(CDM_Interp, File);
  
  // Register the new Collection's command...
  NewCollection->Tcl_Cmd = Tcl_CreateObjCommand(CDM_Interp, CmdName, 
                           CDM_CollectionCallback,
                           (ClientData) NewCollection,
                           (Tcl_CmdDeleteProc *) NULL);
  if (NewCollection->Tcl_Cmd == NULL) {
     snprintf(CDM_Temp, MAX_TEMP_STRING, 
          "Cannot create Collection's command \"%s\"", CmdName);
     CDM_ReportError(CDM_Temp);
     return NULL;
  }
  /* Upgrade Collections to our new native format... */
  if (NewCollection->FileFormat == CDM_FORMAT_NATIVE) {
    NewCollection->FileFormat = CDM_FORMAT_NATIVE2;
    printf("CDM_LoadCollection: Upgrading Collection \"%s\" from "
           "\"native v1\" format to \"native v2\"\n", Path);
  }
  
  /* Ensure that the modified flag is false, as adding Attributes may have
   * turned it to true... */
  NewCollection->Modified = NewCollection->ModifiedDocuments = 0;
  return NewCollection;
error:
  Tcl_DecrRefCount(LineObj);     Tcl_DecrRefCount(ListObj);
  Tcl_DecrRefCount(DocumentIDs); Tcl_DecrRefCount(DocumentXIDs);
  Tcl_Close(CDM_Interp, File);
  CDM_Close(NewCollection);
  return NULL;
} /* CDM_LoadCollection */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |  E N D   O F    C O L L E C T I O N    S E C T I O N  |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     D O C U M E N T                S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 * -- CDM_SaveDocument:
 *    Saves given Document in disk.
 *    Arguments: 
 *      Document:  The Document's that will be saved pointer (CDM_Document)
 *      Directory: Collection's Home Directory               (char *)
 */
int CDM_SaveDocument(CDM_Document Document, char *Directory) {
  Tcl_Channel File = NULL;
  Tcl_Obj *Obj, *Win = NULL;
  int status;
  CDM_THREAD_SPECIFIC_DATA;

  if (Document == NULL) 
          return CDM_ReportError("CDM_SaveDocument: NULL Document Pointer!");
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    return CDM_ReportError("CDM_SaveDocument: Invalid Document Type");
  }
  /* No need to save the Document if it is not modified... */
  if (!Document->Modified) return TCL_OK;
  Tcl_ResetResult(Document->Interp);
  CDM_Interp = Document->Interp;
  
  /*
   * Open the Document File...
   */
  File = CDMIO_OpenDocumentSaveFile(Document, Directory);
  if (File == NULL) goto error;

  /*
   * If the Document format is a native one, display a Progress dialog...
   */
  if (Document->FileFormat == CDM_FORMAT_NATIVE || 
      Document->FileFormat == CDM_FORMAT_NATIVE2) {
    status = CDMIO_DisplayAppendProgressDialog(CDM_Interp, "SavingDocument",
                   "Saving Document.\nPlease Wait...");
    if (status == TCL_OK) {
      Win = Tcl_GetObjResult(CDM_Interp);
      Tcl_IncrRefCount(Win);
    }
  }
  
  switch (Document->FileFormat) {
    case CDM_FORMAT_NATIVE:
    case CDM_FORMAT_NATIVE2: {
      /*
       * Currently, there are two versions of the native format, although the
       * first version (CDM_FORMAT_NATIVE) is obsolete: It is not accessible
       * from the Tcl level and if a Document of CDM_FORMAT_NATIVE is found,
       * is silently updated to CDM_FORMAT_NATIVE2 by CDM. CDM_FORMAT_NATIVE
       * is only held for compatibility reasons, like to open older Documents
       * stored in this format.
       *
       * The main problem with the version 1 format was related to the
       * Attribute/Annotation saving process. It simply reguested them in a
       * single object (through CDM_GetAttributes/CDM_GetAnnotations) and then
       * created and save a *string* representation of this objects, requiring
       * a large amount of memory... Version 2 simply iterates over individual
       * items (Attributes/Annotations) and saves a sinle item in each line.
       * Embedded new-lines are converted to "\n"...
       * Version 2 format was designed/developed by G. Petasis, 06/04/2001.
       */
      /* Write Document Header... */
      if (CDMIO_WriteDocumentHeader(File) == -1) goto error;

      /* Write various Document fields... */
      if (CDMIO_WriteField(File, "\nEncoding::",   Document->Encoding)   == -1)
         goto error;
      if (CDMIO_WriteField(File, "\nFileFormat::", (int) Document->FileFormat)
         == -1) goto error;
      if (CDMIO_WriteField(File, "\nTranslation::", CDMIO_DefaultTranslation)
         == -1) goto error;
      if (CDMIO_WriteField(File, "\nExternalID::", 
               Document->DocumentPtr->ptr[CDM_DOCUMENT_XID]) == -1) goto error;
      if (CDMIO_WriteField(File, "\nID::", 
               Document->DocumentPtr->ptr[CDM_DOCUMENT_ID])  == -1) goto error;
      if (CDMIO_WriteField(File, "\nAnnotationMaxID::", 
               (long) Document->AnnotationMaxID)             == -1) goto error;
      /* Write Attributes */
      Obj = CDM_GetAttributes(Document);
      if (Obj == NULL) goto error;
      if (Document->FileFormat == CDM_FORMAT_NATIVE) {
        if (CDMIO_WriteLengthedObject(File, "\nAttributes::%d\n", Obj) == -1)
             goto error;
      } else {
        if (CDMIO_WriteAttributes(File, Obj) == -1) goto error;
      }
      Tcl_DecrRefCount(Obj);
  
      /* Update our Progress dialog... */
      if (Win != NULL) {
        CDMIO_UpdateProgressDialog(Document->Interp, 30, Win);
      }
      
      /* Write Annotations */
      Obj = CDM_GetAnnotations(Document);
      if (Obj == NULL) goto error;
      if (Document->FileFormat == CDM_FORMAT_NATIVE) {
        if (CDMIO_WriteLengthedObject(File, "\nAnnotations::%d\n", Obj) == -1)
             goto error;
      } else {
        if (CDMIO_WriteAnnotations(File, Obj) == -1) goto error;
      }
      Tcl_DecrRefCount(Obj);
      
      /* Update our Progress dialog... */
      if (Win != NULL) {
        CDMIO_UpdateProgressDialog(Document->Interp, 90, Win);
      }
      
      /* Finally, write Document text...
       * 02/Sep/2002: BUG reported by Sigletos, corrected by Petasis:
       *   CDMIO_WriteLengthedObject was used to write the RawData, but when
       *   it contained the "\r" character, it was written as such. Now it is
       *   replaced with the two characters '\' & 'r'. This was done to be
       *   able to read the character back, without converting it into "\n"
       *   and without modifying the translation of the file, which would have
       *   coused a major incompatibility in collections generated under
       *   windows...
       */
      if (CDMIO_WriteRawData(File, Document->RawData) == -1) goto error;
  
      Tcl_WriteChars(File, "\n## - End of File -", -1);
      break;
    } /* CDM_FORMAT_NATIVE - CDM_FORMAT_NATIVE2 */
    case CDM_FORMAT_GATE_1x:
    case CDM_FORMAT_GATE_2x: {
      if (CDM_SaveGateDocument(Document, File) != TCL_OK) goto error;
      break;
    } /* CDM_FORMAT_GATE_1x - CDM_FORMAT_GATE_2x */
    default: {
      int format, status;
      /*
       * This is a registered format. First, try to get the format type
       * name...
       */
      if (Document->FileFormat < CDM_FORMAT_REGISTERED)  goto error;
      format = Document->FileFormat - CDM_FORMAT_REGISTERED;
      if (format >= CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER) goto error;
      if (CDM_AdditionalCollectionTypes[format].CollectionType == NULL) {
        CDM_ReportError("Collection has an unregistered format!");
        goto error;
      }
      /*
       * We have found the entry matching our type. If the corresponding
       * function pointer is not NULL, call it (and hope it doesn't crash
       * :-))
       */
      if (CDM_AdditionalCollectionTypes[format].saveDocumentPtr != NULL) {
        status = (CDM_AdditionalCollectionTypes[format].saveDocumentPtr)
                  (Document, File);
        if (status != TCL_OK) goto error;
      }
      break;
    } /* default */
  } /* switch (Document->FileFormat) */

  Document->Modified = 0;
  if (Win != NULL) {
    CDMIO_DestroyAppendProgressDialog(Document->Interp, Win);
  }
  Tcl_ResetResult(Document->Interp);
  return Tcl_Close(Document->Interp, File);
error:
  if (File) Tcl_Close(Document->Interp, File);
  if (Win != NULL) {
    CDMIO_DestroyAppendProgressDialog(Document->Interp, Win);
  }
  return TCL_ERROR;
} /* CDM_SaveDocument */

/*
 * -- CDM_LoadDocument:
 *    Loads requested Document from the disk.
 *    Arguments: 
 *      Collection:  The Document's Parent                   (CDM_Collection)
 *      Path:        Document's location                     (char *)
 */
CDM_Document CDM_LoadDocument(CDM_Collection Collection, char *Path) {
  CDM_THREAD_SPECIFIC_DATA;
  CDM_Document NewDocument;
  CDM_ListNode *ptr;
  Tcl_Channel File = NULL;
  Tcl_Obj *LineObj = NULL, *ListObj = NULL, **elemPtrs, *objv[4], *Win=NULL,
          *Message=NULL;
  char *Encoding = (char *) CDMIO_DefaultEncoding, *Line=NULL, CmdName[MAX_TEMP_STRING];
  char *Id = NULL, *ExternalId = NULL;
  int Obj_ID, chars, permissions = 0666, status, i, listLen, items;

  if (Path == NULL) {
    CDM_ReportError("CDM_LoadDocument: NULL Path");
    return NULL;
  }
#ifdef DEBUGIO
  CDM_Debug(" ** Loading Document from Disk (CDM_LoadDocument called...)\n",
            (char *) NULL);
  CDM_Debug(" **   Path: \"%s\"\n", Path, (char *) NULL);
#endif
  
  CDM_Interp = Collection->Interp;
  /* 
   * Check whether the file exists in the given Path. If it
   * exists, open it...
   */
  File = Tcl_OpenFileChannel(CDM_Interp, Path, "r", permissions);
  if (File == NULL) return NULL;

  // Set the correct Encoding... (UTF-8)
  if (Tcl_SetChannelOption(CDM_Interp, File, "-encoding", Encoding) != TCL_OK) {
     snprintf(CDM_Temp, MAX_TEMP_STRING,
                   "CDM_LoadDocument: cannot change encoding to \"%s\"",
                    Encoding);
     CDM_ReportError(CDM_Temp);
     Tcl_Close(CDM_Interp, File);
     return NULL;
  }

#ifdef WINDOWS
  /* Under windows, we have to disable ^Z (\x1a) as the EOF character,
     as our files can contain binary data... */
  Tcl_SetChannelOption(CDM_Interp, File, "-eofchar", "");
#endif /* WINDOWS */

  /*
   * Create the New Document Object...
   */
  Obj_ID = ID.DocumentUniqID();
  snprintf(CmdName, MAX_TEMP_STRING, "Doc%d", Obj_ID);
  NewDocument = CDM_NewDocumentObj();
  NewDocument->Encoding        = NULL;
  NewDocument->ParentPtr       = (void *) Collection;
  NewDocument->RawData         = NULL;
  NewDocument->AnnotationMaxID = 0;
  NewDocument->DocumentPtr     = NULL;
  NewDocument->Modified        = 0;
  NewDocument->Interp          = Collection->Interp;
  
  switch (Collection->FileFormat) {
  case CDM_FORMAT_NATIVE:
  case CDM_FORMAT_NATIVE2: {
    /* Initialise LineObj */
    LineObj = Tcl_NewObj(); ListObj = Tcl_NewListObj(0, NULL);
    if (LineObj == NULL || ListObj == NULL) 
      CDM_Panic("CDM_LoadDocument: Tcl_NewObj() returns NULL - out of memory?");
    Tcl_IncrRefCount(LineObj); Tcl_IncrRefCount(ListObj);

    /* Check to see if the Document is valid... */
    chars = Tcl_GetsObj(File, LineObj);
    if (chars > 0) Line = Tcl_GetString(LineObj);
    if (chars < 1 || (strncmp(Line, "#Ellogon:", 9) != 0 &&
                      strncmp(Line, "#Logos:"  , 7) != 0)) {
      Tcl_DecrRefCount(LineObj); Tcl_DecrRefCount(ListObj);
      Tcl_Close(CDM_Interp, File);
      CDM_Free(NewDocument);
      CDM_ReportError("CDM_LoadDocument: Invalid Document File: ");
      CDM_AppendError(Path);
      return NULL;
    }
    while (chars > -1) {
      Tcl_SetObjLength(LineObj, 0);// Empty object, but keep allocated space...
      chars = Tcl_GetsObj(File, LineObj); if (chars<0 || Tcl_Eof(File)) break;
      Line = Tcl_GetString(LineObj);
      if (strlen(Line) == 0 || Line[0] == '#' || Line[0] == ' ')
        continue; // Skip unwanted lines
    
      /***************** Encoding *****************/
      if (strncmp(Line, "Encoding", 8) == 0) {
        NewDocument->Encoding = CDM_Strdup(&Line[10]);
#ifdef DEBUGIO
        CDM_Debug("Encoding = \"%s\"\n", NewDocument->Encoding, (char *) NULL);
#endif /* DEBUGIO */
        continue;
      } /* Encoding */

      /***************** FileFormat *****************/
      if (strncmp(Line, "FileFormat", 10) == 0) {
        if (sscanf(&Line[12], "%d", &status) < 1) {
          CDM_ReportError("CDM_LoadDocument: "
                          "Cannot Read Document's File Format");
          goto error;
        }
        NewDocument->FileFormat = (CDM_UCHAR) status;
#ifdef DEBUGIO
        CDM_Debug("FileFormat = %d\n", (int) NewDocument->FileFormat,
                  (char *) NULL);
#endif /* DEBUGIO */
        continue;
      } /* FileFormat */

      /***************** Translation *****************/
      if (strncmp(Line, "Translation", 11) == 0) {
#ifdef DEBUGIO
        CDM_Debug("Translation = \"%s\"\n", &Line[13], (char *) NULL);
#endif /* DEBUGIO */
        Tcl_SetChannelOption(NULL, File, "-translation", &Line[13]);
        continue;
      } /* Translation */

      /***************** ExternalID *****************/
      if (strncmp(Line, "ExternalID", 10) == 0) {
        ExternalId = CDM_Strdup(&Line[12]);
#ifdef DEBUGIO
        CDM_Debug("ExternalID = \"%s\"\n", ExternalId, (char *) NULL);
#endif /* DEBUGIO */
        continue;
      } /* ExternalID */

      /***************** ID *****************/
      if (strncmp(Line, "ID", 2) == 0) {
        Id = CDM_Strdup(&Line[4]);
#ifdef DEBUGIO
        CDM_Debug("ID = \"%s\"\n", Id, (char *) NULL);
#endif /* DEBUGIO */
        continue;
      } /* ID */
    
      /***************** AnnotationMaxID *****************/
      if (strncmp(Line, "AnnotationMaxID", 15) == 0) {
        if (sscanf(&Line[17], "%ld", &NewDocument->AnnotationMaxID) < 1) {
          CDM_ReportError("CDM_LoadDocument: "
                          "Cannot Read Annotations Maximum ID");
          goto error;
        }
#ifdef DEBUGIO
        CDM_Debug("AnnotationMaxID = %ld\n", NewDocument->AnnotationMaxID,
                  (char *) NULL);
#endif /* DEBUGIO */
        continue;
      } /* AnnotationMaxID */

      /***************** Attributes *****************/
      if (strncmp(Line, "Attributes", 10) == 0) {
        switch (NewDocument->FileFormat) {
          case CDM_FORMAT_NATIVE: {
            /*
             * Native version 1 format:
             */
            if (sscanf(Line, "Attributes::%d", &items) < 1) {
              CDM_ReportError("CDM_LoadDocument: "
                              "Cannot Read Length of Attribute Object");
              goto error;
            }
#ifdef DEBUGIO
            CDM_Debug("Length of Attribute Object: %d\n", items, (char *) NULL);
#endif /* DEBUGIO */
            if (items) {
              chars = Tcl_ReadChars(File, LineObj, items, 0);
              Tcl_SetStringObj(ListObj, Tcl_GetString(LineObj), -1);
            } else {
              chars = Tcl_GetsObj(File, LineObj);
            }
            if (chars <= 0) {
              if (chars < 0) {
                CDM_ReportError("CDM_LoadDocument: "
                                "Error reading File, while processing "
                                "Document's Attributes");
                goto error;
              }
              if (Tcl_Eof(File)) {
                CDM_ReportError("CDM_LoadDocument: "
                                "Unexpected EOF, while processing Document's "
                                "Attributes");
                goto error;
              }
            }

            // Add Attributes into Document...
            if (items) {
              status = Tcl_ListObjGetElements(CDM_Interp, ListObj,
                                              &listLen, &elemPtrs);
              if (status != TCL_OK) {
                snprintf(CDM_Temp, MAX_TEMP_STRING, 
                         "CDM_LoadDocument: Invalid Attribute List: %s",
                         Tcl_GetString(ListObj));
                CDM_ReportError(CDM_Temp);
                goto error;
              }
              for (i=0; i<listLen; i++) {
#ifdef RECONSTRUCT_ATTRIBUTES
                CDM_Attribute RecAttr;
                RecAttr = CDM_ReconstructAttribute(elemPtrs[i]);
                if (RecAttr == NULL) continue;
                Tcl_IncrRefCount(RecAttr);
                CDM_PutAttribute(NewDocument, RecAttr);
                Tcl_DecrRefCount(RecAttr);
#else  /* RECONSTRUCT_ATTRIBUTES */
                CDM_PutAttribute(NewDocument, elemPtrs[i]);
#endif /* RECONSTRUCT_ATTRIBUTES */
              }
            }
#ifdef DEBUGIO
            CDM_Debug("Attributes = \"%s\"\n", Tcl_GetString(ListObj),
                                               (char *) NULL);
#endif /* DEBUGIO */
            Tcl_SetListObj(ListObj, 0, NULL);
            break;
          } /* case CDM_FORMAT_NATIVE */
          case CDM_FORMAT_NATIVE2: {
            /*
             * Native version 2 format:
             */
            if (sscanf(Line, "Attributes::%d", &items) < 1) {
              CDM_ReportError("CDM_LoadDocument: "
                              "Cannot Read Number of Attributes");
              goto error;
            }
#ifdef DEBUGIO
            CDM_Debug("Number of Attributes: %d\n", items, (char *) NULL);
#endif /* DEBUGIO */
            /*
             * Add Attributes into Document...
             */
            Tcl_SetObjLength(LineObj, 0);
            for (i=0; i<items; i++) {
              chars = Tcl_GetsObj(File, LineObj);
              if (chars <= 0) {
                if (chars < 0) {
                  CDM_ReportError("CDM_LoadDocument: "
                                  "Error reading File, while processing "
                                  "Document's Attributes");
                  goto error;
                }
                if (Tcl_Eof(File)) {
                  CDM_ReportError("CDM_LoadDocument: "
                                  "Unexpected EOF, while processing Document's "
                                  "Attributes");
                  goto error;
                }
              }
#ifdef RECONSTRUCT_ATTRIBUTES
              /*
               * ****************************************************
               *   Re-create Attribute Set... (CDM_FORMAT_NATIVE2)
               * ****************************************************
               */
              //CDM_AttributeSet NewAttributeSet;
              CDM_Attribute RecAttr;
              Tcl_SetStringObj(ListObj, Tcl_GetString(LineObj), -1);
              RecAttr = CDM_ReconstructAttribute(ListObj);
              if (RecAttr == NULL) continue;
              /*
               * Now, the attribute set is in the variable "AttrObj".
               * If memory compression is enabled, we have to check whether
               * the same attribute set is used by some other annotation in
               * order to share the previous one and free this...
               */
              //if (CDM_EnableMemoryCompression) {
              //  NewAttributeSet = CDM_SearchForAttributeSet(RecAttr);
              //  if (NewAttributeSet != RecAttr) {
              //    Tcl_DecrRefCount(RecAttr);
              //    RecAttr = NewAttributeSet;
              //  }
              //}
              Tcl_IncrRefCount(RecAttr);
              CDM_PutAttribute(NewDocument, RecAttr);
              Tcl_DecrRefCount(RecAttr);
#else  /* RECONSTRUCT_ATTRIBUTES */
              /*
               * Tcl_DuplicateObj is required, else our LineObj will be turned
               * into a shared object and any further action on it will be
               * fail...
               */
              CDM_PutAttribute(NewDocument, Tcl_DuplicateObj(LineObj));
#endif /* RECONSTRUCT_ATTRIBUTES */
              Tcl_SetObjLength(LineObj, 0);
              Tcl_SetListObj(ListObj, 0, NULL);
            }
            break;
          } /* case CDM_FORMAT_NATIVE2 */
        } /* switch (NewDocument->FileFormat) */
        continue;
      } /* Attributes */

      /***************** Annotations *****************/
      if (strncmp(Line, "Annotations", 11) == 0) {
        switch (NewDocument->FileFormat) {
          case CDM_FORMAT_NATIVE: {
            /*
             * Native version 1 format:
             */
            if (sscanf(&Line[13], "%d", &items) < 1) {
              CDM_ReportError("CDM_LoadDocument: "
                              "Cannot Read Length of Annotation Object");
              goto error;
            }
#ifdef DEBUGIO
            CDM_Debug("Length of Annotation Object: %d\n", items,
                                                            (char *) NULL);
#endif /* DEBUGIO */
            if (items) {
              //chars = Tcl_ReadChars(File, ListObj, items, 0);
              chars = Tcl_ReadChars(File, LineObj, items, 0);
              Tcl_SetStringObj(ListObj, Tcl_GetString(LineObj), -1);
            } else {
              chars = Tcl_GetsObj(File, LineObj);
            }
            if (chars <= 0) {
              if (chars < 0) {
                CDM_ReportError("CDM_LoadDocument: "
                                "Error reading File, while processing "
                                "Document's Annotations"); goto error;
              }
              if (Tcl_Eof(File)) {
                CDM_ReportError("CDM_LoadDocument: "
                                "Premature EOF, while processing Document's "
                                "Annotations"); goto error;
              }
            }

            // Add Annotations into Document...
            if (items) {
#ifdef RECONSTRUCT_ANNOTATIONS
              CDM_Annotation Ann;
#endif /* RECONSTRUCT_ANNOTATIONS */
              double progress = 0.0, step;
              int last_progress = 0;

              if (items > 20000) {
                Message = Tcl_GetVar2Ex(CDM_Interp, "GUI_Text",
                                        "LoadingDocument", TCL_GLOBAL_ONLY);
                if (Message == NULL) Message =
                  CDM_NewStringObj("Loading Document.\nPlease Wait...");
                objv[0] = CDM_NewStringObj("ggi_wait_append");
                objv[1] = Message;
                objv[2] = CDM_NewLongObj(0);
                for(i=0; i<3; i++) Tcl_IncrRefCount(objv[i]);
                status = Tcl_EvalObjv(CDM_Interp, 3, objv, TCL_EVAL_GLOBAL);
                for(i=0; i<3; i++) Tcl_DecrRefCount(objv[i]);
                if (status == TCL_OK) {
                  Win = Tcl_GetObjResult(CDM_Interp);
                  Tcl_IncrRefCount(Win);
                }
              }
              
              status = Tcl_ListObjGetElements(CDM_Interp, ListObj,
                                              &listLen, &elemPtrs);
              if (status != TCL_OK) {
                snprintf(CDM_Temp, MAX_TEMP_STRING, 
                         "CDM_LoadDocument: Invalid Annotation List: %s",
                         Tcl_GetString(ListObj));
                CDM_ReportError(CDM_Temp);
                goto error;
              }
              
              if (listLen > 0.0) {
                step = (90.0 - last_progress)/float(listLen);
              } else {
                step = (90.0 - last_progress);
              }
              for (i=0; i<listLen; i++) {
                progress += step;
                if (Win != NULL && (int(progress) - last_progress) > 9) {
                  int j;
                  last_progress = int(progress);
                  objv[0] = CDM_NewStringObj("ggi_wait_update");
                  objv[1] = &CDM_Obj_NULL;
                  objv[2] = CDM_NewLongObj(last_progress);
                  objv[3] = Win;
                  for(j=0; j<4; j++) Tcl_IncrRefCount(objv[j]);
                  Tcl_EvalObjv(CDM_Interp, 4, objv, TCL_EVAL_GLOBAL);
                  for(j=0; j<4; j++) Tcl_DecrRefCount(objv[j]);
                  Tcl_DoOneEvent(TCL_IDLE_EVENTS | TCL_WINDOW_EVENTS |
                                 TCL_DONT_WAIT);
                }
#ifdef RECONSTRUCT_ANNOTATIONS
                /* 
                 * Well, it would be perfect to check for validity, but its
                 * slower :-) ...
                 */
                /* if (!CDM_ValidAnnotation(elemPtrs[i])) goto error; */
                Ann = CDM_ReconstructAnnotation(elemPtrs[i]);
                if (Ann == NULL) {
                  /* We have just red an invalid Annotation object ! */
                  goto error;
                }
                /*
                 * Avoid to use CDM_AddAnnotation, as this function does various
                 * optimisations regarding cached memory. We don't need all this
                 * fuss, as the Annotation was already reconstructed...
                 * CDM_AddAnnotation(NewDocument, Ann);
                 */
                CDM_DocumentInternalAddAnnotation(NewDocument, Ann);
#else  /* RECONSTRUCT_ANNOTATIONS */
                CDM_AddAnnotation(NewDocument, elemPtrs[i]);
#endif /* RECONSTRUCT_ANNOTATIONS */
              } 
#ifdef DEBUGIO
              CDM_Debug("Annotations = \"%s\"\n", Tcl_GetString(ListObj),
                                                   (char *) NULL);
#endif /* DEBUGIO */
            }
#ifdef UNIX
            /* This is a BUG under windows */
            //Tcl_SetListObj(ListObj, 0, NULL);
#endif /* UNIX */
            break;
          } /* case CDM_FORMAT_NATIVE */
          case CDM_FORMAT_NATIVE2: {
#ifdef RECONSTRUCT_ANNOTATIONS
            CDM_Annotation Ann;
#endif /* RECONSTRUCT_ANNOTATIONS */
            double progress = 0.0, step = 0.0;
            int last_progress = 0;

            /*
             * Native version 2 format:
             */
            if (sscanf(&Line[13], "%d", &items) < 1) {
              CDM_ReportError("CDM_LoadDocument: "
                              "Cannot Read Number of Annotations");
              goto error;
            }
#ifdef DEBUGIO
            CDM_Debug("Number of Annotations: %d\n", items, (char *) NULL);
#endif /* DEBUGIO */
            /*
             * If we have many Annotations, we have to show our progress...
             */
            if (items > 5000) {
              Message = Tcl_GetVar2Ex(CDM_Interp, "GUI_Text",
                                      "LoadingDocument", TCL_GLOBAL_ONLY);
              if (Message == NULL) Message =
                CDM_NewStringObj("Loading Document.\nPlease Wait...");
              objv[0] = CDM_NewStringObj("ggi_wait_append");
              objv[1] = Message;
              objv[2] = CDM_NewLongObj(0);
              for(i=0; i<3; i++) Tcl_IncrRefCount(objv[i]);
              status = Tcl_EvalObjv(CDM_Interp, 3, objv, TCL_EVAL_GLOBAL);
              for(i=0; i<3; i++) Tcl_DecrRefCount(objv[i]);
              if (status == TCL_OK) {
                Win = Tcl_GetObjResult(CDM_Interp);
                Tcl_IncrRefCount(Win);
              }
            }
            /*
             * Add Annotations into Document...
             */
            if (items) {
              step = (90.0 - last_progress)/float(items);
            } else {
              step = (90.0 - last_progress);
            }
            Tcl_SetObjLength(LineObj, 0);
            for (i=0; i<items; i++) {
              chars = Tcl_GetsObj(File, LineObj);
              if (chars <= 0) {
                if (chars < 0) {
                  CDM_ReportError("CDM_LoadDocument: "
                                  "Error reading File, while processing "
                                  "Document's Annotations"); goto error;
                }
                if (Tcl_Eof(File)) {
                  CDM_ReportError("CDM_LoadDocument: "
                                  "Premature EOF, while processing Document's "
                                  "Annotations"); goto error;
                }
              }

              progress += step;
              if (Win != NULL && (int(progress) - last_progress) > 9) {
                int j;
                last_progress = int(progress);
                objv[0] = CDM_NewStringObj("ggi_wait_update");
                objv[1] = &CDM_Obj_NULL;
                objv[2] = CDM_NewLongObj(last_progress);
                objv[3] = Win;
                for(j=0; j<4; j++) Tcl_IncrRefCount(objv[j]);
                Tcl_EvalObjv(CDM_Interp, 4, objv, TCL_EVAL_GLOBAL);
                for(j=0; j<4; j++) Tcl_DecrRefCount(objv[j]);
                Tcl_DoOneEvent(TCL_IDLE_EVENTS | TCL_WINDOW_EVENTS |
                               TCL_DONT_WAIT);
              }
#ifdef RECONSTRUCT_ANNOTATIONS
              /* 
               * Well, it would be perfect to check for validity, but its
               * slower :-) ...
               */
              /* if (!CDM_ValidAnnotation(LineObj)) goto error; */
              Tcl_SetStringObj(ListObj, Tcl_GetString(LineObj), -1);
              Ann = CDM_ReconstructAnnotation(ListObj);
              if (Ann == NULL) {
                /* We have just red an invalid Annotation object ! */
                goto error;
              }
              /*
               * Avoid to use CDM_AddAnnotation, as this function does various
               * optimisations regarding cached memory. We don't need all this
               * fuss, as the Annotation was already reconstructed...
               * CDM_AddAnnotation(NewDocument, Ann);
               */
              Tcl_IncrRefCount(Ann);
              CDM_DocumentInternalAddAnnotation(NewDocument, Ann);
              Tcl_DecrRefCount(Ann);
#else  /* RECONSTRUCT_ANNOTATIONS */
              /*
               * Tcl_DuplicateObj is required, else our LineObj will be turned
               * into a shared object and any further action on it will be
               * fail...
               */
              CDM_AddAnnotation(NewDocument, Tcl_DuplicateObj(LineObj));
#endif /* RECONSTRUCT_ANNOTATIONS */
              Tcl_SetObjLength(LineObj, 0);
              Tcl_SetListObj(ListObj, 0, NULL);
            } /* for (i=0; i<items; i++) */
            Tcl_SetStringObj(LineObj, "", 0);
            //CDM_SearchForAttributeSet(NULL);
            break;
          } /* case CDM_FORMAT_NATIVE2 */
        } /* switch (Collection->FileFormat) */
        continue;
      } /* Annotations */

      if (Win != NULL) {
        int j;
        objv[0] = CDM_NewStringObj("ggi_wait_update");
        objv[1] = &CDM_Obj_NULL;
        objv[2] = CDM_NewLongObj(90);
        objv[3] = Win;
        for(j=0; j<4; j++) Tcl_IncrRefCount(objv[j]);
        Tcl_EvalObjv(CDM_Interp, 4, objv, TCL_EVAL_GLOBAL);
        for(j=0; j<4; j++) Tcl_DecrRefCount(objv[j]);
        Tcl_DoOneEvent(TCL_IDLE_EVENTS|TCL_WINDOW_EVENTS|TCL_DONT_WAIT);
      }

      /* RawData */
      if (strncmp(Line, "RawData", 7) == 0) {
        if (Tcl_GetInt(CDM_Interp, &Line[9], &chars) != TCL_OK) {
          CDM_ReportError("CDM_LoadDocument: "
                          "Cannot locate Document's RawData (Text) Length");
          goto error;
        }
#ifdef DEBUGIO
        CDM_Debug("Located Text of size %d chars...\n", chars, (char *) NULL);
#endif /* DEBUGIO */
        Tcl_SetObjLength(LineObj, 0); // Empty object, but keep allocated space!
        chars = Tcl_ReadChars(File, LineObj, chars, 0);
        if (chars <= 0) {
         if (chars < 0) {
           CDM_ReportError("CDM_LoadDocument: "
                           "Cannot Read Document's RawData (Text) (Error)");
            goto error;
         }
         if (Tcl_Eof(File)) {
           CDM_ReportError("CDM_LoadDocument: "
                           "Cannot Read Document's RawData (Text) (EOF)");
            goto error;
         }
        } 
        NewDocument->RawData = Tcl_NewStringObj(Tcl_GetString(LineObj), -1);
        Tcl_IncrRefCount(NewDocument->RawData);
#ifdef DEBUGIO
        CDM_Debug("RawData = \"%s\"\n", Tcl_GetString(NewDocument->RawData),
                  (char *) NULL);
#endif /* DEBUGIO */
        break;
      } /* RawData */
    }
    Tcl_DecrRefCount(LineObj); LineObj = NULL;
    Tcl_DecrRefCount(ListObj); ListObj = NULL;
    if (Win != NULL) {
      objv[0] = CDM_NewStringObj("ggi_destroy_append");
      objv[1] = Win;
      for(i=0; i<1; i++) Tcl_IncrRefCount(objv[i]);
      Tcl_EvalObjv(CDM_Interp, 2, objv, TCL_EVAL_GLOBAL);
      for(i=0; i<2; i++) Tcl_DecrRefCount(objv[i]);
      Tcl_DoOneEvent(TCL_IDLE_EVENTS|TCL_WINDOW_EVENTS|TCL_DONT_WAIT);
      Win = NULL;
    }
    break;
  }
  case CDM_FORMAT_GATE_1x:
  case CDM_FORMAT_GATE_2x: {
    status = CDM_LoadGateDocument(NewDocument, File, &Id, &ExternalId);
    if (status != TCL_OK) goto error;
    break;
  }
  default: {
    int format = 0;
    /*
     * This is a registered format. First, try to get the format type
     * name...
     */
    if (Collection->FileFormat < CDM_FORMAT_REGISTERED)  goto error;
    format = Collection->FileFormat - CDM_FORMAT_REGISTERED;
    if (format >= CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER) goto error;
    NewDocument->FileFormat = Collection->FileFormat;
    if (CDM_AdditionalCollectionTypes[format].CollectionType == NULL) {
      CDM_ReportError("CDM_LoadDocument: "
                      "Collection has an unregistered format!");
      goto error;
    }
    /*
     * We have found the entry matching our type. If the corresponding
     * function pointer is not NULL, call it (and hope it doesn't crash
     * :-))
     */
    if (CDM_AdditionalCollectionTypes[format].loadDocumentPtr != NULL) {
      status = (CDM_AdditionalCollectionTypes[format].loadDocumentPtr)
                (NewDocument, File, &Id, &ExternalId);
      if (status != TCL_OK) goto error;
    }
    break;
  }
  } /* switch */
  Tcl_Close(CDM_Interp, File);
  File = NULL;
  NewDocument->Modified = 0;

  /*
   * Make some "sanity" checks...
   */
  if (NewDocument->RawData == NULL) {
    CDM_ReportError("CDM_LoadDocument: "
                    "Document has no Text? - Error Loading Document!");
    goto error;
  }
  if (Id == NULL || ExternalId == NULL) {
    CDM_ReportError("CDM_LoadDocument: "
                    "Document has no External Id? - Error Loading Document!");
    goto error;
  }
  /* We shouldn't find either '\n' or '\r' in the document Id/XId... */
  if ((strchr(ExternalId, '\r') != NULL) || (strchr(ExternalId, '\n') != NULL)
      || (strchr(Id, '\r') != NULL) || (strchr(Id, '\n') != NULL)) {
    CDM_ReportError("CDM_LoadDocument: Corrupted Document detected: New-line "
                    "representation does not match with the representation "
                    "declared in the Document.");
    goto error;
  }

  /*
   * Update Parent Collection (report this Document as Opened...)
   */
  ptr = CDM_ListFind(Collection->Documents, Id, CDM_DOCUMENT_ID);
  if (ptr == NULL) {
    CDM_ReportError("CDM_LoadDocument: Cannot locate the Document in its "
                    "parent Collection");
    goto error;
  }
  if (ptr->ptr[CDM_DOCUMENT_ID]  != NULL) Tcl_Free(ptr->ptr[CDM_DOCUMENT_ID]);
  if (ptr->ptr[CDM_DOCUMENT_XID] != NULL) Tcl_Free(ptr->ptr[CDM_DOCUMENT_XID]);
  ptr->ptr[CDM_DOCUMENT_ID]      = Id;
  ptr->ptr[CDM_DOCUMENT_XID]     = ExternalId;
  ptr->ptr[CDM_DOCUMENT_OPENED]  = (char *) 1;
  ptr->ptr[CDM_DOCUMENT_POINTER] = (char *) NewDocument;
  NewDocument->DocumentPtr = ptr;
  
  // Register the new Document's command...
  NewDocument->Tcl_Cmd = Tcl_CreateObjCommand(CDM_Interp, CmdName, 
                           /*(Tcl_ObjCmdProc*)*/ CDM_DocumentCallback,
                           (ClientData) NewDocument,
                           (Tcl_CmdDeleteProc *) NULL);
  if (NewDocument->Tcl_Cmd == NULL) {
     snprintf(CDM_Temp, MAX_TEMP_STRING, 
                    "CDM_LoadDocument: Cannot create Document's command \"%s\"", CmdName);
     CDM_ReportError(CDM_Temp);
     goto error;
  }

#ifdef DEBUGIO
  CDM_Debug(" ** Document Loaded - Registered as %s...\n", CmdName,
            (char *) NULL);
#endif
  if (NewDocument->FileFormat == CDM_FORMAT_NATIVE) {
    NewDocument->FileFormat = CDM_FORMAT_NATIVE2;
    printf("CDM_LoadDocument: Upgrading Document \"%s\" from "
          "\"native v1\" format to \"native v2\"\n", Path);
  }
  return NewDocument;
error: // Free Allocated Resources...
  if (File) Tcl_Close(CDM_Interp, File);
  if (Win != NULL) {
    objv[0] = CDM_NewStringObj("ggi_destroy_append");
    objv[1] = Win;
    for(i=0; i<1; i++) Tcl_IncrRefCount(objv[i]);
    Tcl_EvalObjv(CDM_Interp, 2, objv, TCL_EVAL_GLOBAL);
    for(i=0; i<2; i++) Tcl_DecrRefCount(objv[i]);
    Win = NULL;
  }
  if (LineObj != NULL) Tcl_DecrRefCount(LineObj);
  if (ListObj != NULL) Tcl_DecrRefCount(ListObj);
  ID.CancelDocumentUniqID(/*NewDocument->*/Obj_ID);
  CDM_Free(NewDocument); 
  return NULL;
} /* CDM_LoadDocument */


/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     G U I     R E L A T E D      R O U T I N E S      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 * -- CDM_GetCollectionFileXIDs:
 *    Read Specified File and return a Tcl List with all Document XIDs.
 *    Arguments: 
 *      FileName:    The File's Path                         (char *)
 *      Format:      Collection's Format                       (int)
 */
CDM_EXPORT
Tcl_Obj *CDM_GetCollectionFileXIDs(char *FileName, int Format) {
  Tcl_Channel File = NULL;
  Tcl_Obj *LineObj=NULL, *Result=NULL, *Obj=NULL;
  char *Line;
  int chars, File_nu, i;
  CDM_THREAD_SPECIFIC_DATA;

  if (Format < CDM_FORMAT_REGISTERED) {
    File = Tcl_OpenFileChannel(CDM_Interp, FileName, "r", 0666);
    if (File == NULL) return NULL;
#ifdef WINDOWS
    /* Under windows, we have to disable ^Z (\x1a) as the EOF character,
       as our files can contain binary data... */
    Tcl_SetChannelOption(CDM_Interp, File, "-eofchar", "");
#endif /* WINDOWS */    
  }

  switch (Format) {
    case CDM_FORMAT_NATIVE: {
      if (Tcl_SetChannelOption(CDM_Interp, File, "-encoding", 
                               CDMIO_DefaultEncoding) != TCL_OK) {
        snprintf(CDM_Temp, MAX_TEMP_STRING, "cannot change encoding to \"%s\"",
                                            CDMIO_DefaultEncoding);
        CDM_ReportError(CDM_Temp);
        Tcl_Close(CDM_Interp, File);
        return NULL;
      }
      chars = 1; LineObj = Tcl_NewObj(); if (LineObj == NULL) 
        CDM_Panic("CDM_GetCollectionFileXIDs: out of memory");
      while (chars > -1) {
        Tcl_SetObjLength(LineObj, 0); chars = Tcl_GetsObj(File, LineObj);
        if (chars < 0) break; Line = Tcl_GetString(LineObj);
        if (strncmp(Line, "DocumentXIDs", 12) == 0) {
          Tcl_SetObjLength(LineObj, 0);
          if (Tcl_GetsObj(File, LineObj) < 0) {
            CDM_ReportError("Cannot locate Collection's Document XIDs");
            Tcl_Close(CDM_Interp, File);
          }
          Tcl_Close(CDM_Interp, File);
          return LineObj;
        }
      }
      break;
    }
    case CDM_FORMAT_GATE_1x:
    case CDM_FORMAT_GATE_2x: {
      Result = Tcl_NewListObj(0, NULL);
      LineObj = Tcl_NewObj(); if (LineObj == NULL || Result == NULL) 
        CDM_Panic("CDM_GetCollectionFileXIDs: out of memory");
      /* Skip 5 lines... */
      chars = Tcl_GetsObj(File, LineObj); if (chars < 0) break;
      chars = Tcl_GetsObj(File, LineObj); if (chars < 0) break;
      chars = Tcl_GetsObj(File, LineObj); if (chars < 0) break;
      chars = Tcl_GetsObj(File, LineObj); if (chars < 0) break;
      Tcl_SetObjLength(LineObj, 0);
      chars = Tcl_GetsObj(File, LineObj); if (chars < 0) break;
      Line = Tcl_GetString(LineObj); 
      if (sscanf(Line, "GutSequence %d", &File_nu) < 1) goto error;
      for (i=0; i<File_nu; i++) {
        Line = CDM_ReadGateString(File); if (Line == NULL) goto error;
        Tcl_Free(Line);
        Line = CDM_ReadGateString(File); if (Line == NULL) goto error;
        Obj = Tcl_NewStringObj(Line, -1); Tcl_Free(Line);
        if (Obj == NULL) CDM_Panic("CDM_LoadGateCollection: Tcl_NewObj() "
                                   "returns NULL - out of memory?");
        Tcl_ListObjAppendList(CDM_Interp, Result, Obj);
        Obj = NULL;
      }
      Tcl_DecrRefCount(LineObj);
      Tcl_Close(CDM_Interp, File);
      return Result;
    }
    default: {
      char *dir;
      int format = 0;
      /*
       * This is a registered format. First, try to get the format type
       * name...
       */
      if (Format < CDM_FORMAT_REGISTERED)  goto error;
      format = Format - CDM_FORMAT_REGISTERED;
      if (format >= CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER) goto error;
      if (CDM_AdditionalCollectionTypes[format].CollectionType == NULL) {
        CDM_ReportError("Collection has an unregistered format!");
        goto error;
      }
      /*
       * We have found the entry matching our type. If the corresponding
       * function pointer is not NULL, call it (and hope it doesn't crash
       * :-))
       */
      if (CDM_AdditionalCollectionTypes[format].CollectionXIdsPtr != NULL) {
        dir = CDM_FileNameDirname(FileName);
        if (dir == NULL) goto error;
        Result= (CDM_AdditionalCollectionTypes[format].CollectionXIdsPtr) (dir);
        CDM_Free(dir);
        return Result;
      }
      break;
    }
  }
error:
  CDM_ReportError("Cannot Process Collection ");
  switch (Format) {
    case CDM_FORMAT_NATIVE: {
      CDM_AppendError("(Not in Native Format?)");
      break;
    }
    case CDM_FORMAT_GATE_1x:
    case CDM_FORMAT_GATE_2x: {
      CDM_AppendError("(Not in GATE Format?)");
      break;
    }
    default: {
      CDM_AppendError("(Not in a registered Format?)");
      break;
    }
  }
  if (Result  != NULL) Tcl_DecrRefCount(Result);
  if (Obj     != NULL) Tcl_DecrRefCount(Obj);
  if (LineObj != NULL) Tcl_DecrRefCount(LineObj);
  if (File) Tcl_Close(CDM_Interp, File);
  return NULL;
} /* CDM_GetCollectionFileXIDs */


/*
 * -- CDM_GetCollectionFileIDs:
 *    Read Specified File and return a Tcl List with all Document IDs.
 *    Arguments: 
 *      FileName:    The File's Path                         (char *)
 *      Format:      Collection's Format                       (int)
 */
CDM_EXPORT
Tcl_Obj *CDM_GetCollectionFileIDs(char *FileName, int Format) {
  Tcl_Channel File = NULL;
  Tcl_Obj *LineObj=NULL, *Result=NULL, *Obj=NULL;
  char *Line;
  int chars, File_nu, i;
  CDM_THREAD_SPECIFIC_DATA;

  if (Format < CDM_FORMAT_REGISTERED) {
    File = Tcl_OpenFileChannel(CDM_Interp, FileName, "r", 0666);
    if (File == NULL) return NULL;
#ifdef WINDOWS
  /* Under windows, we have to disable ^Z (\x1a) as the EOF character,
     as our files can contain binary data... */
  Tcl_SetChannelOption(CDM_Interp, File, "-eofchar", "");
#endif /* WINDOWS */
  }

  switch (Format) {
    case CDM_FORMAT_NATIVE: {
      if (Tcl_SetChannelOption(CDM_Interp, File, "-encoding", 
                               CDMIO_DefaultEncoding) != TCL_OK) {
        snprintf(CDM_Temp, MAX_TEMP_STRING, "cannot change encoding to \"%s\"",
                                            CDMIO_DefaultEncoding);
        CDM_ReportError(CDM_Temp);
        Tcl_Close(CDM_Interp, File);
        return NULL;
      }
      chars = 1; LineObj = Tcl_NewObj(); if (LineObj == NULL) 
        CDM_Panic("CDM_GetCollectionFileIDs: out of memory");
      while (chars > -1) {
        Tcl_SetObjLength(LineObj, 0); chars = Tcl_GetsObj(File, LineObj);
        if (chars < 0) break; Line = Tcl_GetString(LineObj);
        if (strncmp(Line, "DocumentIDs", 11) == 0) {
          Tcl_SetObjLength(LineObj, 0);
          if (Tcl_GetsObj(File, LineObj) < 0) {
            CDM_ReportError("Cannot locate Collection's Document IDs");
            Tcl_Close(CDM_Interp, File);
          }
          Tcl_Close(CDM_Interp, File);
          return LineObj;
        }
      }
      break;
    }
    case CDM_FORMAT_GATE_1x:
    case CDM_FORMAT_GATE_2x: {
      Result = Tcl_NewListObj(0, NULL);
      LineObj = Tcl_NewObj(); if (LineObj == NULL || Result == NULL) 
        CDM_Panic("CDM_GetCollectionFileIDs: out of memory");
      /* Skip 5 lines... */
      chars = Tcl_GetsObj(File, LineObj); if (chars < 0) break;
      chars = Tcl_GetsObj(File, LineObj); if (chars < 0) break;
      chars = Tcl_GetsObj(File, LineObj); if (chars < 0) break;
      chars = Tcl_GetsObj(File, LineObj); if (chars < 0) break;
      Tcl_SetObjLength(LineObj, 0);
      chars = Tcl_GetsObj(File, LineObj); if (chars < 0) break;
      Line = Tcl_GetString(LineObj); 
      if (sscanf(Line, "GutSequence %d", &File_nu) < 1) goto error;
      for (i=0; i<File_nu; i++) {
        Line = CDM_ReadGateString(File); if (Line == NULL) goto error;
        Obj = Tcl_NewStringObj(Line, -1); Tcl_Free(Line);
        if (Obj == NULL) CDM_Panic("CDM_LoadGateCollection: Tcl_NewObj() "
                                   "returns NULL - out of memory?");
        Tcl_ListObjAppendList(CDM_Interp, Result, Obj);
        Obj = NULL;
        Line = CDM_ReadGateString(File); if (Line == NULL) goto error;
        Tcl_Free(Line);
      }
      Tcl_DecrRefCount(LineObj);
      Tcl_Close(CDM_Interp, File);
      return Result;
    }
    default: {
      char *dir;
      int format = 0;
      /*
       * This is a registered format. First, try to get the format type
       * name...
       */
      if (Format < CDM_FORMAT_REGISTERED)  goto error;
      format = Format - CDM_FORMAT_REGISTERED;
      if (format >= CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER) goto error;
      if (CDM_AdditionalCollectionTypes[format].CollectionType == NULL) {
        CDM_ReportError("Collection has an unregistered format!");
        goto error;
      }
      /*
       * We have found the entry matching our type. If the corresponding
       * function pointer is not NULL, call it (and hope it doesn't crash
       * :-))
       */
      if (CDM_AdditionalCollectionTypes[format].CollectionIdsPtr != NULL) {
        dir = CDM_FileNameDirname(FileName);
        if (dir == NULL) goto error;
        Result = (CDM_AdditionalCollectionTypes[format].CollectionIdsPtr) (dir);
        CDM_Free(dir);
        return Result;
      }
      break;
    }
  }
error:
  CDM_ReportError("Cannot Process Collection ");
  switch (Format) {
    case CDM_FORMAT_NATIVE: {
      CDM_AppendError("(Not in Native Format?)");
      break;
    }
    case CDM_FORMAT_GATE_1x:
    case CDM_FORMAT_GATE_2x: {
      CDM_AppendError("(Not in GATE Format?)");
      break;
    }
    default: {
      CDM_AppendError("(Not in a registered Format?)");
      break;
    }
  }
  if (Result  != NULL) Tcl_DecrRefCount(Result);
  if (Obj     != NULL) Tcl_DecrRefCount(Obj);
  if (LineObj != NULL) Tcl_DecrRefCount(LineObj);
  if (File) Tcl_Close(CDM_Interp, File);
  return NULL;
} /* CDM_GetCollectionFileIDs */


/*
 * -- CDM_GetDocumentAttributes:
 *    Read Specified File and return a Tcl List with all Document Attributes.
 *    Arguments: 
 *      FileName:    The File's Path                         (char *)
 *      Format:      Document's Format                       (int)
 */
CDM_EXPORT
Tcl_Obj *CDM_GetDocumentAttributes(char *FileName, int Format) {
  Tcl_Channel File;
  Tcl_Obj *LineObj=NULL, *Result=NULL, *Attr, **elemPtrs;
  char *Line;
  int chars, i, items, listLen, FileFormat;
  CDM_THREAD_SPECIFIC_DATA;
  
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetDocumentAttributes: Aborting...");
    return NULL;
  }

  File = Tcl_OpenFileChannel(CDM_Interp, FileName, "r", 0666);
  if (File == NULL) return NULL;

#ifdef WINDOWS
  /* Under windows, we have to disable ^Z (\x1a) as the EOF character,
     as our files can contain binary data... */
  Tcl_SetChannelOption(CDM_Interp, File, "-eofchar", "");
#endif /* WINDOWS */

  switch (Format) {
    case CDM_FORMAT_NATIVE: {
      if (Tcl_SetChannelOption(CDM_Interp, File, "-encoding", 
                               CDMIO_DefaultEncoding) != TCL_OK) {
        snprintf(CDM_Temp, MAX_TEMP_STRING, "cannot change encoding to \"%s\"",
                                            CDMIO_DefaultEncoding);
        CDM_ReportError(CDM_Temp);
        Tcl_Close(CDM_Interp, File);
        return NULL;
      }
      chars      = 1;
      FileFormat = CDM_FORMAT_NATIVE;
      LineObj    = Tcl_NewObj();
      if (LineObj == NULL) {
        CDM_Panic("CDM_GetDocumentAttributes: out of memory");
      }
      while (chars > -1) {
        Tcl_SetObjLength(LineObj, 0); chars = Tcl_GetsObj(File, LineObj);
        if (chars <= 0) break; Line = Tcl_GetString(LineObj);
        /*
         * There is always the possibility some of the Collection Documents to
         * have been upgraded to the new format, but not the Collections as a
         * whole. If this is the case, handle the Document approprietly...
         */
        if (strncmp(Line, "FileFormat", 10) == 0) {
          sscanf(Line, "FileFormat::%d", &FileFormat);
          if (FileFormat == CDM_FORMAT_NATIVE2) {
            goto native2;
          }
        }
        if (strncmp(Line, "Attributes", 10) == 0) {
          if (sscanf(Line, "Attributes::%d", &chars) < 1) {
            CDM_ReportError("Cannot Read Length of Attribute Object");
            Tcl_Close(CDM_Interp, File);
            Tcl_DecrRefCount(LineObj);
            return NULL;
          }
          Tcl_SetObjLength(LineObj, 0);
          if (chars) chars = Tcl_ReadChars(File, LineObj, chars, 0);
          else       chars = Tcl_GetsObj(File, LineObj);
          if (chars < 0) {
            CDM_ReportError("Cannot locate Document's Attributes");
            Tcl_Close(CDM_Interp, File);
            Tcl_DecrRefCount(LineObj);
            return NULL;
          }
          i = Tcl_ListObjGetElements(CDM_Interp, LineObj, &listLen, &elemPtrs);
          if (i != TCL_OK) {
            snprintf(CDM_Temp, MAX_TEMP_STRING, "Invalid Attribute List: %s",
                     Tcl_GetString(LineObj));
            CDM_ReportError(CDM_Temp);
            Tcl_Close(CDM_Interp, File);
            Tcl_DecrRefCount(LineObj);
            return NULL;
          }
          Tcl_Close(CDM_Interp, File);
          return LineObj;
        }
      }
      Tcl_DecrRefCount(LineObj);
      Tcl_Close(CDM_Interp, File);
      break;
    } /* case CDM_FORMAT_NATIVE */
    case CDM_FORMAT_NATIVE2: {
      if (Tcl_SetChannelOption(CDM_Interp, File, "-encoding", 
                               CDMIO_DefaultEncoding) != TCL_OK) {
        snprintf(CDM_Temp, MAX_TEMP_STRING, "cannot change encoding to \"%s\"",
                                            CDMIO_DefaultEncoding);
        CDM_ReportError(CDM_Temp);
        Tcl_Close(CDM_Interp, File);
        return NULL;
      }
      LineObj = Tcl_NewObj();
native2:
      Result  = Tcl_NewListObj(0, NULL);
      chars   = 1; 
      if (LineObj == NULL || Result == NULL) {
        CDM_Panic("CDM_GetDocumentAttributes: out of memory");
      }
      while (chars > -1) {
        Tcl_SetObjLength(LineObj, 0); chars = Tcl_GetsObj(File, LineObj);
        if (chars <= 0) break; Line = Tcl_GetString(LineObj);
        if (strncmp(Line, "Attributes", 10) == 0) {
          if (sscanf(Line, "Attributes::%d", &items) < 1) {
            CDM_ReportError("Cannot Read Number of Attributes");
            Tcl_Close(CDM_Interp, File);
            Tcl_DecrRefCount(LineObj);
            Tcl_DecrRefCount(Result);
            return NULL;
          }
          Tcl_SetObjLength(LineObj, 0);
          for (i=0; i<items; i++) {
            chars = Tcl_GetsObj(File, LineObj);
            if (chars <= 0) {
              if (chars < 0) {
                CDM_ReportError("Error reading File, while processing "
                                "Document's Attributes");
                goto error;
              }
              if (Tcl_Eof(File)) {
                CDM_ReportError("Unexpected EOF, while processing Document's "
                                "Attributes");
                goto error;
              }
            }
            Tcl_ListObjAppendElement(CDM_Interp, Result,
                                                 Tcl_DuplicateObj(LineObj));
            Tcl_SetObjLength(LineObj, 0);
          }
          Tcl_Close(CDM_Interp, File);
          Tcl_DecrRefCount(LineObj);
          return Result;
        }
      }
      Tcl_Close(CDM_Interp, File);
      Tcl_DecrRefCount(LineObj);
      Tcl_DecrRefCount(Result);
      break;
    } /* case CDM_FORMAT_NATIVE2 */
    case CDM_FORMAT_GATE_1x:
    case CDM_FORMAT_GATE_2x: {
      LineObj = Tcl_NewObj(); Result = Tcl_NewListObj(0, NULL);
      if (LineObj == NULL || Result == NULL) 
        CDM_Panic("CDM_GetDocumentAttributes: out of memory");
      chars = Tcl_GetsObj(File, LineObj); if (chars <= 0) goto error;
      Line = Tcl_GetString(LineObj);
      if (sscanf(Line, "GdmDocument GutSequence %d", &chars) < 1) goto error;
      Tcl_DecrRefCount(LineObj); LineObj = NULL;
      for (i=0; i<chars; i++) {
        Attr = CDM_ReadGateAttribute(File);
        if (Attr == NULL) goto error;
        Tcl_ListObjAppendElement(CDM_Interp, Result, Attr);
      }
      Tcl_Close(CDM_Interp, File);
      return Result;
    }
    default: {
      int format = 0;
      if (Tcl_SetChannelOption(CDM_Interp, File, "-encoding", 
                               CDMIO_DefaultEncoding) != TCL_OK) {
        snprintf(CDM_Temp, MAX_TEMP_STRING, "cannot change encoding to \"%s\"",
                                            CDMIO_DefaultEncoding);
        CDM_ReportError(CDM_Temp);
        Tcl_Close(CDM_Interp, File);
        return NULL;
      }
      /*
       * This is a registered format. First, try to get the format type
       * name...
       */
      if (Format < CDM_FORMAT_REGISTERED)  goto error;
      format = Format - CDM_FORMAT_REGISTERED;
      if (format >= CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER) goto error;
      if (CDM_AdditionalCollectionTypes[format].CollectionType == NULL) {
        CDM_ReportError("Collection has an unregistered format!");
        goto error;
      }
      /*
       * We have found the entry matching our type. If the corresponding
       * function pointer is not NULL, call it (and hope it doesn't crash
       * :-))
       */
      if (CDM_AdditionalCollectionTypes[format].DocumentAttrsPtr != NULL) {
        Result = (CDM_AdditionalCollectionTypes[format].DocumentAttrsPtr)
                 (File);
        Tcl_Close(CDM_Interp, File);
        return Result;
      }
      break;
    }
  }
  return NULL;
error:
  CDM_ReportError("Cannot Process Document ");
  switch (Format) {
    case CDM_FORMAT_NATIVE: {
      CDM_AppendError("(Not in Native Format?)");
      break;
    }
    case CDM_FORMAT_GATE_1x:
    case CDM_FORMAT_GATE_2x: {
      CDM_AppendError("(Not in GATE Format?)");
      break;
    }
    default: {
      CDM_AppendError("(Not in a registered Format?)");
      break;
    }
  }
  if (LineObj != NULL) Tcl_DecrRefCount(LineObj);
  if (Result  != NULL) Tcl_DecrRefCount(Result);
  Tcl_Close(CDM_Interp, File);
  return NULL;
} /* CDM_GetDocumentAttributes */


/*
 * CDMIO.cpp - EOF -
 */
