/*****************************************************************************
 *  CDM_Macros.h:                                                            *
 * ---------------------------------                                         *
 *  This file implements the macro subsystem of the Collection and Document  *
 *  Manager (CDM) Public API.                                                *
 *  Version: 2.0                                                             *
 *  For more details regarding CDM Version consult file "CDM_Version.h"      *
 *  CDM is the Collection and Document Manager and the core component of the *
 *  Ellogon Text Engineering Platform.                                       *
 *  CDM was designed and implemented by Georgios Petasis,                    *
 *  e-mail: petasis@iit.demokritos.gr                                        *
 *  URL:    http://www.iit.demokritos.gr/~petasis                            *
 *                                                                           *
 *  Copyright 1998-2010 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licencing information, please refer to the licence conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/
#include "CDM_Macros.h"

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |             M A C R O S       S E C T I O N           |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/
 
/*
 * ---- CDM_MACRO_GetContainerSystem:
 *  DESCRIPTION:
 *   <English:
 *    This function will return a token for the System window in the graphical
 *    interface that contains this Collection.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    col: The Collection object. (const CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to ::Macro::GetContainerSystem (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value from this function will be a token (of type 
 *    CDM_MACRO_SYSTEM) representing the opened System window.
 *    CDM does not own the returned token: the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the token and the memory
 *    associated with it.
 *
 *    In case of an error, NULL will be returned and an error message will be
 *    left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MACRO_EditSystem, CDM_MACRO_GetOpenedCollection,
 *    CDM_MACRO_GetOpenedDocument, CDM_MACRO_OpenInSystem,
 *    CDM_MACRO_RunSystemComponent, CDM_MACRO_RunSystemTool,
 *    CDM_MACRO_DisplayViewer,
 *    CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *    ::Macro::OpenSystem
 *   >:&:
 */
CDM_EXPORT
CDM_MACRO_SYSTEM CDM_MACRO_GetContainerSystem(const CDM_Collection col) {
  CDM_MACRO_SYSTEM token;
  int status = CDM_Eval(CDM_GetActiveInterpreter(),
      "::Macro::GetContainerSystem",
      Tcl_GetCommandName(CDM_GetActiveInterpreter(),col->Tcl_Cmd), NULL);
  if (status != TCL_OK) return NULL;
  token = Tcl_GetObjResult(CDM_GetActiveInterpreter());
  Tcl_IncrRefCount(token);
  return token;
}; /* CDM_MACRO_GetContainerSystem */

/*
 * ---- CDM_MACRO_OpenSystem:
 *  DESCRIPTION:
 *   <English:
 *    This function will open the System window in the graphical interface.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    name: The name of the System to be displayed. (const char *)
 *    folder: The name of the Folder the System resided in (optional). (const char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to ::Macro::OpenSystem (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value from this function will be a token (of type 
 *    CDM_MACRO_SYSTEM) representing the opened System window.
 *    CDM does not own the returned token: the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the token and the memory
 *    associated with it.
 *
 *    In case of an error, NULL will be returned and an error message will be
 *    left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MACRO_EditSystem, CDM_MACRO_GetOpenedCollection,
 *    CDM_MACRO_GetOpenedDocument, CDM_MACRO_OpenInSystem,
 *    CDM_MACRO_RunSystemComponent, CDM_MACRO_RunSystemTool,
 *    CDM_MACRO_DisplayViewer,
 *    CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *    ::Macro::OpenSystem
 *   >:&:
 */
CDM_EXPORT
CDM_MACRO_SYSTEM CDM_MACRO_OpenSystem(const char *name,
                                      const char *folder) {
  CDM_MACRO_SYSTEM token;
  int status = CDM_Eval(CDM_GetActiveInterpreter(),
                        "::Macro::OpenSystem", name, folder, NULL);
  if (status != TCL_OK) return NULL;
  token = Tcl_GetObjResult(CDM_GetActiveInterpreter());
  Tcl_IncrRefCount(token);
  return token;
}; /* CDM_MACRO_OpenSystem */

/*
 * ---- CDM_MACRO_EditSystem:
 *  DESCRIPTION:
 *   <English:
 *    This function will open a System for editing in the graphical interface.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    name: The name of the System to be displayed for editing. (const char *)
 *    folder: The name of the Folder the System resided in (optional). (const char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to ::Macro::EditSystem (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value from this function will be a token (of type 
 *    CDM_MACRO_SYSTEM) representing the opened System window.
 *    CDM does not own the returned token: the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the token and the memory
 *    associated with it.
 *
 *    In case of an error, NULL will be returned and an error message will be
 *    left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MACRO_OpenSystem, CDM_MACRO_GetOpenedCollection,
 *    CDM_MACRO_GetOpenedDocument, CDM_MACRO_OpenInSystem,
 *    CDM_MACRO_RunSystemComponent, CDM_MACRO_RunSystemTool,
 *    CDM_MACRO_DisplayViewer,
 *    CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *    ::Macro::EditSystem
 *   >:&:
 */
CDM_EXPORT
CDM_MACRO_SYSTEM CDM_MACRO_EditSystem(const char *name,
                                      const char *folder) {
  CDM_MACRO_SYSTEM token;
  int status = CDM_Eval(CDM_GetActiveInterpreter(),
                        "::Macro::EditSystem", name, folder, NULL);
  if (status != TCL_OK) return NULL;
  token = Tcl_GetObjResult(CDM_GetActiveInterpreter());
  Tcl_IncrRefCount(token);
  return token;
}; /* CDM_MACRO_EditSystem */

/*
 * ---- CDM_MACRO_OpenInSystem:
 *  DESCRIPTION:
 *   <English:
 *    This function will open a whole Collection or a Document inside a System.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    system: A token representing an opened System. (const CDM_MACRO_SYSTEM)
 *    collection_name: The name (path on disk) of the Collection to be opened. (const char *)
 *    document_id: The Document ID of the Document to be opened (optional). (const char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to ::Macro::OpenInSystem (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value from this function will be a token (of type 
 *    CDM_MACRO_SYSTEM) representing the opened Collection/Document.
 *    CDM does not own the returned token: the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the token and the memory
 *    associated with it.
 *
 *    In case of an error, NULL will be returned and an error message will be
 *    left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MACRO_OpenSystem, CDM_MACRO_EditSystem,
 *    CDM_MACRO_GetOpenedCollection,
 *    CDM_MACRO_GetOpenedDocument, CDM_MACRO_OpenInSystem,
 *    CDM_MACRO_RunSystemComponent, CDM_MACRO_RunSystemTool,
 *    CDM_MACRO_DisplayViewer,
 *    CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *    ::Macro::OpenInSystem
 *   >:&:
 */
CDM_EXPORT
CDM_MACRO_SYSTEM_COLLECTION CDM_MACRO_OpenInSystem(
                                      const CDM_MACRO_SYSTEM system,
                                      const char *collection_name,
                                      const char *document_id) {
  CDM_MACRO_SYSTEM_COLLECTION token;
  if (system == NULL) return NULL;
  int status = CDM_Eval(CDM_GetActiveInterpreter(),
                        "::Macro::OpenInSystem",
                        Tcl_GetString(system),
                        collection_name, document_id, NULL);
  if (status != TCL_OK) return NULL;
  token = Tcl_GetObjResult(CDM_GetActiveInterpreter());
  Tcl_IncrRefCount(token);
  return token;
}; /* CDM_MACRO_OpenInSystem */

/*
 * ---- CDM_MACRO_GetOpenedCollection:
 *  DESCRIPTION:
 *   <English:
 *    This function will return the Collection object (of type CDM_Collection)
 *    that is represented by the provided token. The token must have been
 *    retrieved with a previous call to either CDM_MACRO_OpenSystem or
 *    CDM_MACRO_EditSystem.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    token: A token representing an opened Collection/Document in a System. (const const CDM_MACRO_SYSTEM_COLLECTION)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the Collection object (of type
 *     CDM_Collection) that is represented by the provided token.
 *
 *     In case of an error NULL will be returned and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never
 *     be freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MACRO_OpenSystem, CDM_MACRO_EditSystem,
 *    CDM_MACRO_GetOpenedDocument
 *   >:&:
 */
CDM_EXPORT
CDM_Collection CDM_MACRO_GetOpenedCollection(
                           const CDM_MACRO_SYSTEM_COLLECTION token) {
  if (token == NULL) return NULL;
  return CDM_GetCollectionPointer(CDM_Nth(token, 0));
}; /* CDM_MACRO_GetOpenedCollection */

/*
 * ---- CDM_MACRO_GetOpenedDocument:
 *  DESCRIPTION:
 *   <English:
 *    This function will return the Document object (of type CDM_Document)
 *    that is represented by the provided token. The token must have been
 *    retrieved with a previous call to either CDM_MACRO_OpenSystem or
 *    CDM_MACRO_EditSystem.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    token: A token representing an opened Document/Document in a System. (const const CDM_MACRO_SYSTEM_COLLECTION)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the Document object (of type
 *     CDM_Document) that is represented by the provided token.
 *
 *     In case of an error NULL will be returned and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never
 *     be freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MACRO_OpenSystem, CDM_MACRO_EditSystem,
 *    CDM_MACRO_GetOpenedCollection
 *   >:&:
 */
CDM_EXPORT
CDM_Document CDM_MACRO_GetOpenedDocument(
                           const CDM_MACRO_SYSTEM_COLLECTION token) {
  if (token == NULL) return NULL;
  return CDM_GetDocumentPointer(CDM_Nth(token, 1));
}; /* CDM_MACRO_GetOpenedDocument */

/*
 * ---- CDM_MACRO_RunSystemComponent:
 *  DESCRIPTION:
 *   <English:
 *    This function will apply the specified Annotator (through the Annotator
 *    parameter) on the Collection/Document represented by "token". If the
 *    perameter "force" is false (the default), the Annotator will be applied
 *    only if all the previous Annotators in the system have already run. If
 *    "force" is true, no check will be performed and the Annotator will be 
 *    applied.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    system: A token representing an opened System. (const CDM_MACRO_SYSTEM)
 *    token: A token representing an opened Document/Document in a System. (const const CDM_MACRO_SYSTEM_COLLECTION)
 *    Annotator: The name of the Annotator (component) to apply. (const char *)
 *    force: A boolean flag, specifying whether the Annotator will be applied anyway (true), or whether its applicability should be first examined (false). (const int)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to ::Macro::RunSystemComponent (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value from this function will be a standart Tcl completion
 *    code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *    error occurs, the return value will be TCL_ERROR and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). If all annotators run without errors, TCL_OK will be
 *    returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MACRO_OpenSystem, CDM_MACRO_EditSystem,
 *    CDM_MACRO_GetOpenedCollection,
 *    CDM_MACRO_GetOpenedDocument, CDM_MACRO_OpenInSystem,
 *    CDM_MACRO_RunSystemComponent, CDM_MACRO_RunSystemTool,
 *    CDM_MACRO_DisplayViewer,
 *    ::Macro::RunSystemComponent
 *   >:&:
 */
CDM_EXPORT
int CDM_MACRO_RunSystemComponent(
                           const CDM_MACRO_SYSTEM system,
                           const CDM_MACRO_SYSTEM_COLLECTION token,
                           const char *Annotator,
                           const int force) {
  char tmp[24];
  int status;
  if (system == NULL) {
    return CDM_ReportError("CDM_MACRO_RunSystemComponent: "
                           "NULL system paremeter");
  }
  if (token == NULL) {
    return CDM_ReportError("CDM_MACRO_RunSystemComponent: "
                           "NULL token paremeter");
  }
  snprintf(tmp, 23, "%d", force);
  status = CDM_Eval(CDM_GetActiveInterpreter(),
                    "::Macro::RunSystemComponent", 
                    Tcl_GetString(system), Annotator,
                    Tcl_GetString(CDM_Nth(token, 0)),
                    Tcl_GetString(CDM_Nth(token, 1)), tmp, NULL);
  return status;
}; /* CDM_MACRO_RunSystemComponent */

/*
 * ---- CDM_MACRO_RunSystemComponent:
 *  DESCRIPTION:
 *   <English:
 *    This function will apply the specified Annotator (through the Annotator
 *    parameter) on the Collection/Document specified through the "col" and
 *    "doc" parameters. If the
 *    perameter "force" is false (the default), the Annotator will be applied
 *    only if all the previous Annotators in the system have already run. If
 *    "force" is true, no check will be performed and the Annotator will be 
 *    applied.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    system: A token representing an opened System. (const CDM_MACRO_SYSTEM)
 *    col: A Collection object (CDM_Collection)
 *    doc: A Document object. (CDM_Document)
 *    Annotator: The name of the Annotator (component) to apply. (const char *)
 *    force: A boolean flag, specifying whether the Annotator will be applied anyway (true), or whether its applicability should be first examined (false). (const int)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to ::Macro::RunSystemComponent (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value from this function will be a standart Tcl completion
 *    code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *    error occurs, the return value will be TCL_ERROR and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). If all annotators run without errors, TCL_OK will be
 *    returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MACRO_OpenSystem, CDM_MACRO_EditSystem,
 *    CDM_MACRO_GetOpenedCollection,
 *    CDM_MACRO_GetOpenedDocument, CDM_MACRO_OpenInSystem,
 *    CDM_MACRO_RunSystemComponent, CDM_MACRO_RunSystemTool,
 *    CDM_MACRO_DisplayViewer,
 *    ::Macro::RunSystemComponent
 *   >:&:
 */
CDM_EXPORT
int CDM_MACRO_RunSystemComponent(
                           const CDM_MACRO_SYSTEM system,
                           CDM_Collection col,
                           CDM_Document doc,
                           const char *Annotator,
                           const int force) {
  char tmp[24];
  int status;
  if (system == NULL) {
    return CDM_ReportError("CDM_MACRO_RunSystemComponent: "
                           "NULL system paremeter");
  }
  if (col == NULL) {
    return CDM_ReportError("CDM_MACRO_RunSystemComponent: "
                           "NULL col paremeter");
  }
  snprintf(tmp, 23, "%d", force);
  if (doc) {
    status = CDM_Eval(CDM_GetActiveInterpreter(),
                    "::Macro::RunSystemComponent",
                    Tcl_GetString(system), Annotator,
                    Tcl_GetCommandName(CDM_GetActiveInterpreter(),col->Tcl_Cmd),
                    Tcl_GetCommandName(CDM_GetActiveInterpreter(),doc->Tcl_Cmd),
                    tmp, NULL);
  } else {
    status = CDM_Eval(CDM_GetActiveInterpreter(),
                    "::Macro::RunSystemComponent", Annotator,
                    Tcl_GetCommandName(CDM_GetActiveInterpreter(),col->Tcl_Cmd),
                    "", tmp, NULL);
  }
  return status;
}; /* CDM_MACRO_RunSystemComponent */

/*
 * ---- CDM_MACRO_RunSystemTool:
 *  DESCRIPTION:
 *   <English:
 *    This function will apply the specified Tool (through the Annotator
 *    parameter) on the Collection/Document represented by "token". If the
 *    perameter "force" is false (the default), the Annotator will be applied
 *    only if all the previous Annotators in the system have already run. If
 *    "force" is true, no check will be performed and the Annotator will be 
 *    applied.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    system: A token representing an opened System. (const CDM_MACRO_SYSTEM)
 *    token: A token representing an opened Document/Document in a System. (const const CDM_MACRO_SYSTEM_COLLECTION)
 *    Annotator: The name of the Annotator (component) to apply. (const char *)
 *    force: A boolean flag, specifying whether the Annotator will be applied anyway (true), or whether its applicability should be first examined (false). (const int)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to ::Macro::RunSystemTool (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value from this function will be a standart Tcl completion
 *    code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *    error occurs, the return value will be TCL_ERROR and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). If all annotators run without errors, TCL_OK will be
 *    returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MACRO_OpenSystem, CDM_MACRO_EditSystem,
 *    CDM_MACRO_GetOpenedCollection,
 *    CDM_MACRO_GetOpenedDocument, CDM_MACRO_OpenInSystem,
 *    CDM_MACRO_RunSystemTool, CDM_MACRO_RunSystemTool,
 *    CDM_MACRO_DisplayViewer,
 *    ::Macro::RunSystemTool
 *   >:&:
 */
CDM_EXPORT
int CDM_MACRO_RunSystemTool(
                           const CDM_MACRO_SYSTEM system,
                           const CDM_MACRO_SYSTEM_COLLECTION token,
                           const char *Annotator,
                           const int force) {
  char tmp[24];
  int status;
  if (system == NULL) {
    return CDM_ReportError("CDM_MACRO_RunSystemTool: "
                           "NULL system paremeter");
  }
  if (token == NULL) {
    return CDM_ReportError("CDM_MACRO_RunSystemTool: "
                           "NULL token paremeter");
  }
  snprintf(tmp, 23, "%d", force);
  status = CDM_Eval(CDM_GetActiveInterpreter(),
                    "::Macro::RunSystemTool",
                    Tcl_GetString(system), Annotator,
                    Tcl_GetString(CDM_Nth(token, 0)),
                    Tcl_GetString(CDM_Nth(token, 1)), tmp, NULL);
  return status;
}; /* CDM_MACRO_RunSystemTool */

/*
 * ---- CDM_MACRO_RunSystemTool:
 *  DESCRIPTION:
 *   <English:
 *    This function will apply the specified Tool (through the Annotator
 *    parameter) on the Collection/Document specified through the "col" and
 *    "doc" parameters. If the
 *    perameter "force" is false (the default), the Annotator will be applied
 *    only if all the previous Annotators in the system have already run. If
 *    "force" is true, no check will be performed and the Annotator will be 
 *    applied.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    system: A token representing an opened System. (const CDM_MACRO_SYSTEM)
 *    col: A Collection object (CDM_Collection)
 *    doc: A Document object. (CDM_Document)
 *    Annotator: The name of the Annotator (component) to apply. (const char *)
 *    force: A boolean flag, specifying whether the Annotator will be applied anyway (true), or whether its applicability should be first examined (false). (const int)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to ::Macro::RunSystemTool (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value from this function will be a standart Tcl completion
 *    code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *    error occurs, the return value will be TCL_ERROR and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). If all annotators run without errors, TCL_OK will be
 *    returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MACRO_OpenSystem, CDM_MACRO_EditSystem,
 *    CDM_MACRO_GetOpenedCollection,
 *    CDM_MACRO_GetOpenedDocument, CDM_MACRO_OpenInSystem,
 *    CDM_MACRO_RunSystemTool, CDM_MACRO_RunSystemTool,
 *    CDM_MACRO_DisplayViewer,
 *    ::Macro::RunSystemTool
 *   >:&:
 */
CDM_EXPORT
int CDM_MACRO_RunSystemTool(
                           const CDM_MACRO_SYSTEM system,
                           CDM_Collection col,
                           CDM_Document doc,
                           const char *Annotator,
                           const int force) {
  char tmp[24];
  int status;
  if (system == NULL) {
    return CDM_ReportError("CDM_MACRO_RunSystemTool: "
                           "NULL system paremeter");
  }
  if (col == NULL) {
    return CDM_ReportError("CDM_MACRO_RunSystemTool: "
                           "NULL col paremeter");
  }
  snprintf(tmp, 23, "%d", force);
  if (doc) {
    status = CDM_Eval(CDM_GetActiveInterpreter(),
                    "::Macro::RunSystemTool",
                    Tcl_GetString(system), Annotator,
                    Tcl_GetCommandName(CDM_GetActiveInterpreter(),col->Tcl_Cmd),
                    Tcl_GetCommandName(CDM_GetActiveInterpreter(),doc->Tcl_Cmd),
                    tmp, NULL);
  } else {
    status = CDM_Eval(CDM_GetActiveInterpreter(),
                    "::Macro::RunSystemTool",
                    Tcl_GetString(system), Annotator,
                    Tcl_GetCommandName(CDM_GetActiveInterpreter(),col->Tcl_Cmd),
                    "", tmp, NULL);
  }
  return status;
}; /* CDM_MACRO_RunSystemTool */

/*
 * ---- CDM_MACRO_DisplayViewer:
 *  DESCRIPTION:
 *   <English:
 *    This function will display a graphical viewer over a Document opened in a
 *    System.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    system: A token representing an opened System. (const CDM_MACRO_SYSTEM)
 *    viewer_name: The name of the viewer to be shown. (const char *)
 *    doc: A Document object. Either the Document or the Collection owning the Document must have been opened in the System specified through the "system" parameter.  (CDM_Document)
 *    parameter1: A paremeter that depends on the viewer to be displayed. It usually corrsponds to Annotation Type. (const char *)
 *    parameter2: A paremeter that depends on the viewer to be displayed. It usually corrsponds to an Attribute. (const char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to ::Macro::DisplayViewer (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value from this function will be a standart Tcl completion
 *    code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *    error occurs, the return value will be TCL_ERROR and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). If all annotators run without errors, TCL_OK will be
 *    returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    ::Macro::DisplayViewer
 *   >:&:
 */
CDM_EXPORT
int CDM_MACRO_DisplayViewer(const CDM_MACRO_SYSTEM system,
                            const char *viewer_name,
                            CDM_Document doc,
                            const char *parameter1,
                            const char *parameter2) {
  int status;
  if (system == NULL) {
    return CDM_ReportError("CDM_MACRO_DisplayViewer: "
                           "NULL system paremeter");
  }
  if (doc == NULL) {
    return CDM_ReportError("CDM_MACRO_DisplayViewer: "
                           "NULL doc paremeter");
  }
  status = CDM_Eval(CDM_GetActiveInterpreter(),
                    "::Macro::DisplayViewer",
                    Tcl_GetString(system), viewer_name,
                    Tcl_GetCommandName(CDM_GetActiveInterpreter(),doc->Tcl_Cmd),
                    parameter1, parameter2, NULL);
  return status;
}; /* CDM_MACRO_DisplayViewer */

/*
 * ---- CDM_MACRO_NFoldExecution:
 *  DESCRIPTION:
 *   <English:
 *    This function will perform a N-Fold evaluation on the specified
 *    Collection.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    col: The Collection on which N-Fold evaluation will be performed. (CDM_Collection)
 *    int: The number of folds. (const int)
 *    TrainingCommands: A set of commands to be applied on the "training" set of Documents. (CDM_ByteSequenceSet or CDM_RawData)
 *    EvaluationCommands: A set of commands to be applied on the "training" set of Documents. (CDM_ByteSequenceSet or CDM_RawData)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to ::Macro::NFoldExecution (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value from this function will be a standart Tcl completion
 *    code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *    error occurs, the return value will be TCL_ERROR and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). If all annotators run without errors, TCL_OK will be
 *    returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    ::Macro::NFoldExecution
 *   >:&:
 */
CDM_EXPORT
int CDM_MACRO_NFoldExecution(CDM_Collection col,
                             const int folds,
                             CDM_ByteSequenceSet TrainingCommands,
                             CDM_ByteSequenceSet EvaluationCommands) {
  Tcl_Obj *objv[5];
  int status;
  if (col == NULL) {
    return CDM_ReportError("CDM_MACRO_NFoldExecution: "
                           "NULL col paremeter");
  }
  /* Construct the command to be executed. */
  objv[0] = Tcl_NewStringObj("::Macro::NFoldExecution", -1);
  if (objv[0] == NULL) {
    return CDM_ReportError("CDM_MACRO_NFoldExecution: out of memory!");
  }
  objv[1] = Tcl_NewStringObj(Tcl_GetCommandName(CDM_GetActiveInterpreter(),
                                                col->Tcl_Cmd), -1);
  if (objv[1] == NULL) {
    Tcl_DecrRefCount(objv[0]);
    return CDM_ReportError("CDM_MACRO_NFoldExecution: out of memory!");
  }
  objv[2] = Tcl_NewIntObj(folds);
  if (objv[2] == NULL) {
    Tcl_DecrRefCount(objv[0]);
    Tcl_DecrRefCount(objv[1]);
    return CDM_ReportError("CDM_MACRO_NFoldExecution: out of memory!");
  }
  objv[3] = objv[4] = Tcl_NewListObj(0, NULL);
  if (objv[3] == NULL) {
    Tcl_DecrRefCount(objv[0]);
    Tcl_DecrRefCount(objv[1]);
    Tcl_DecrRefCount(objv[2]);
    return CDM_ReportError("CDM_MACRO_NFoldExecution: out of memory!");
  }
  Tcl_IncrRefCount(objv[3]);
  if (TrainingCommands)   objv[3] = TrainingCommands;
  if (EvaluationCommands) objv[4] = EvaluationCommands;
  for (int i = 0; i < 5; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(CDM_GetActiveInterpreter(), 5, objv, TCL_EVAL_GLOBAL);
  for (int i = 0; i < 5; i++) Tcl_DecrRefCount(objv[i]);
  Tcl_DecrRefCount(objv[3]);
  return status;
}; /* CDM_MACRO_NFoldExecution */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |      E N D   O F   M A C R O S    S E C T I O N       |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/* The following is needed by the "ExCppDoc" Utility. Do not remove! */
/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |        N O N E                     S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     E N D     O F     A L L     S E C T I O N S       |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/
/*
 * CDM_Macros.cpp - EOF -
 */
