/*****************************************************************************
 *  CDM_MemoryCache.cpp:                                                     *
 * ----------------------                                                    *
 *  This file contains various functions that relate to reducing the memory  *
 *  required by CDM in order to store string, Annotation and Attribute       *
 *  objects...                                                               *
 *                                                                           *
 *  Copyright 1998-2010 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licencing information, please refer to the licence conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/

#include "CDM_MemoryCache.h"
#include "CDM_GlobalObjects.h"
#include <string.h>
extern Tcl_ThreadDataKey dataKey;
extern int CDM_GetAttributeTypeFromObject(Tcl_Obj *Type);

/*
 * CDM_ObjectCache:
 * -------------------
 * This is a dummy constructor of the class. The caller should call Init() in
 * order to initialise the class.
 */
CDM_ObjectCache::CDM_ObjectCache() {
  CDM_THREAD_SPECIFIC_DATA;
  Initialised       = 0;
  LongTablePtr      = &LongTable;
  AttributeSetTable = NULL;
  DialogWindow      = NULL;

  AttributeHashTableType.version         = TCL_HASH_KEY_TYPE_VERSION;
  AttributeHashTableType.flags           = 0;
  AttributeHashTableType.hashKeyProc     = &HashObjKey;
  AttributeHashTableType.compareKeysProc = &CompareObjKeys;
  AttributeHashTableType.allocEntryProc  = &AllocObjEntry;
  AttributeHashTableType.freeEntryProc   = &FreeObjEntry;
} /* CDM_ObjectCache::CDM_ObjectCache() */

/*
 * Init:
 * -----------------
 * This function initialises the internal structures of the class. It should
 * be called before any of the other methods are called.
 */
void CDM_ObjectCache::Init() {
  if (!Initialised) {
    CDM_THREAD_SPECIFIC_DATA;
    LongTablePtr                           = &LongTable;
    AttributeHashTableType.version         = TCL_HASH_KEY_TYPE_VERSION;
    AttributeHashTableType.flags           = 0;
    AttributeHashTableType.hashKeyProc     = &HashObjKey;
    AttributeHashTableType.compareKeysProc = &CompareObjKeys;
    AttributeHashTableType.allocEntryProc  = &AllocObjEntry;
    AttributeHashTableType.freeEntryProc   = &FreeObjEntry;
    AttributeSetTable                      = NULL;
    DialogWindow                           = NULL;
#ifdef CDM_OBJ_HASH_KEYS
    /*
     * If the Tcl version is > 8.4a1, we use the new Tcl feature of using Tcl
     * objects directly as Hash table keys. We save the memory used for keeping
     * internally in the hash table copies of the key strings...
     */
    Tcl_InitObjHashTable(&StringTable);
    Tcl_InitObjHashTable(&AttributeValueTable);
    /* Tcl_InitObjHashTable(&AttributeTable); */
    Tcl_InitCustomHashTable(&AttributeTable, TCL_CUSTOM_PTR_KEYS,
                            &AttributeHashTableType);
#else  /* CDM_OBJ_HASH_KEYS */
    /*
     * Petasis, 30/09/2000
     * Ok, 8.4a2 got too long to be released, and we need CDM also under
     * windows :-) Modifications in order for CDM to compile with lower
     * versions of Tcl... (<= 8.4a1)
     */
    Tcl_InitHashTable(&StringTable, TCL_STRING_KEYS);
    Tcl_InitHashTable(&AttributeValueTable, TCL_STRING_KEYS);
    Tcl_InitHashTable(&AttributeTable, TCL_STRING_KEYS);
#endif /* CDM_OBJ_HASH_KEYS */
    Tcl_InitHashTable(&LongTable, TCL_ONE_WORD_KEYS);
    Tcl_DStringInit(&LastRequestedAnnotationTypeStr);
    Tcl_DStringInit(&LastRequestedAttributeNameStr);
    Tcl_DStringInit(&LastRequestedAttributeValueStr);
    LastRequestedAnnotationTypeObj = LastRequestedAttributeNameObj =
    LastRequestedAttributeValueObj = LastRequestedLongObj = NULL;
    LastRequestedLong = -1;
    NumberOfStrings         = NumberOfLongs      =
    NumberOfAttributeValues = NumberOfAttributes = 0;
    Initialised = 1;
  }
} /*CDM_ObjectCache::Init */

/*
 * Clear:
 * -------------------
 * This method removes all stored objects from the class and re-initialises
 * all internal structures.
 */
void CDM_ObjectCache::Clear() {
  register Tcl_HashEntry *entryPtr;
  register Tcl_HashSearch searchPtr;
  register Tcl_Obj *ObjPtr;
  
#ifdef DEBUG_HASHED_VALUES
  CDM_Debug("+++++++ CDM_ObjectCache::Clear Called!\n");
#endif /* DEBUG_HASHED_VALUES */
  CreateProgressInfoDialog("Clearing Object Memory Cache...");

  CDM_ClearAttributeSetHash();
  UpdateProgressInfoDialog(20);
  /* Free the objects stored in StringTable. Note that if objects are used as
   * keys, they will be freed when the table is deleted, so there is no need
   * to be freed... */
#ifndef CDM_OBJ_HASH_KEYS
  for (entryPtr = Tcl_FirstHashEntry(&StringTable, &searchPtr);
       entryPtr != NULL; entryPtr = Tcl_NextHashEntry(&searchPtr)) {
    ObjPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    if (ObjPtr != NULL)  Tcl_DecrRefCount(ObjPtr);
  }
#endif /* CDM_OBJ_HASH_KEYS */
  Tcl_DeleteHashTable(&StringTable);
  UpdateProgressInfoDialog(40);
  
  /* Free the objects stored in AttributeTable. Note that if objects are used as
   * keys, they will be freed when the table is deleted, so there is no need
   * to be freed... */
#ifndef CDM_OBJ_HASH_KEYS
  for (entryPtr = Tcl_FirstHashEntry(&AttributeTable, &searchPtr);
       entryPtr != NULL; entryPtr = Tcl_NextHashEntry(&searchPtr)) {
    ObjPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    if (ObjPtr != NULL)  Tcl_DecrRefCount(ObjPtr);
  }
#endif /* CDM_OBJ_HASH_KEYS */
  Tcl_DeleteHashTable(&AttributeTable);
  UpdateProgressInfoDialog(60);
  
  /* Free the objects stored in AttributeValueTable.
     This table has key/value nismatch. Values must be absolutelly freed...*/
  for (entryPtr = Tcl_FirstHashEntry(&AttributeValueTable, &searchPtr);
       entryPtr != NULL; entryPtr = Tcl_NextHashEntry(&searchPtr)) {
    ObjPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    if (ObjPtr != NULL)  Tcl_DecrRefCount(ObjPtr);
  }
  Tcl_DeleteHashTable(&AttributeValueTable);
  UpdateProgressInfoDialog(80);

  /* Free the objects stored in LongTable. This table also has key/value
     nismatch and values must be absolutelly freed...*/
  for (entryPtr = Tcl_FirstHashEntry(&LongTable, &searchPtr);
       entryPtr != NULL; entryPtr = Tcl_NextHashEntry(&searchPtr)) {
    ObjPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    if (ObjPtr != NULL)  Tcl_DecrRefCount(ObjPtr);
  }
  /* Reset Hash Table... */
  Tcl_DeleteHashTable(&LongTable);
  UpdateProgressInfoDialog(100);

  Tcl_DStringFree(&LastRequestedAnnotationTypeStr);
  Tcl_DStringFree(&LastRequestedAttributeNameStr);
  Tcl_DStringFree(&LastRequestedAttributeValueStr);

  /*
   * Finally, call Init to re-initialise the class...
   */
  Initialised = 0;
  Init();
  /* printf(" ->Hashed Objects Cleared...\n"); */
  DestroyProgressInfoDialog();
} /* CDM_ObjectCache::Clear */

/*
 * RemoveUnusedObjects:
 * ----------------------
 * This method removes from the cache only the objects that are not used by
 * either CDM (in Annotations or Attributes), Tcl or components. It is a good
 * idea to be frequently called when large sets of Annotations/Attributes are
 * not any more wanted, like when large Documents or Collections are closed.
 * But note that this method may require some processing time, especially if
 * the cache contains many stored objects...
 */
void CDM_ObjectCache::RemoveUnusedObjects() {
  Tcl_HashTable *tablePtr[4];
  register Tcl_HashEntry *entryPtr, *nextPtr;
  register Tcl_HashSearch searchPtr;
  register Tcl_Obj *ObjPtr;
  long *NumberOf[4];
  int i, total, removed;
  
  CreateProgressInfoDialog("Clearing Unused Objects from the Object Memory "
                           "Cache...");
#ifdef DEBUG_HASHED_VALUES
  CDM_Debug("+++++++ CDM_ObjectCache::RemoveUnusedObjects Called!\n");
#endif /* DEBUG_HASHED_VALUES */

  /*
   * Store the table pointers on an array, so as to use a loop to free unused
   * objects...
   */
  tablePtr[0] = &StringTable;         NumberOf[0] = &NumberOfStrings;
  tablePtr[1] = &AttributeTable;      NumberOf[1] = &NumberOfAttributes;
  tablePtr[2] = &AttributeValueTable; NumberOf[2] = &NumberOfAttributeValues;
  tablePtr[3] = &LongTable;           NumberOf[3] = &NumberOfLongs;
  /* printf("RemoveUnusedObjects: Clearing tables...\n"); */
  for (i=0; i<4; i++) {
    UpdateProgressInfoDialog(i*25);
    total = 0; removed = 0;
    entryPtr = Tcl_FirstHashEntry(tablePtr[i], &searchPtr); 
    while (entryPtr != NULL) {
      total++;
      ObjPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
      if (ObjPtr != NULL) {
        if (!Tcl_IsShared(ObjPtr)) {
          /* This object is used only by the table. Remove it... */
#ifndef CDM_OBJ_HASH_KEYS
          Tcl_DecrRefCount(ObjPtr);
#else
          /* Only the AttributeValueTable & LongTable do not use the value as
           * key. As a result the value object needs to be freed... */
          if (i > 1) Tcl_DecrRefCount(ObjPtr);
#endif /* CDM_OBJ_HASH_KEYS */
          nextPtr  = Tcl_NextHashEntry(&searchPtr);
          Tcl_DeleteHashEntry(entryPtr);
          entryPtr = nextPtr; removed++; (*NumberOf[i])--;
        } else {
          entryPtr = Tcl_NextHashEntry(&searchPtr);
        } /* if (!Tcl_IsShared(ObjPtr)) */
      } else {
        /* This entry is not associated with an object. Remove it... */
        nextPtr  = Tcl_NextHashEntry(&searchPtr);
        Tcl_DeleteHashEntry(entryPtr);
        entryPtr = nextPtr; removed++;
      } /* if (ObjPtr != NULL) */
    } /* while (entryPtr != NULL) */
    /* printf("Table %d: removed %d out of %d...\n", i, removed, total); */
  } /* for (i=0; i<4; i++) */

  Tcl_DStringSetLength(&LastRequestedAnnotationTypeStr, 0);
  Tcl_DStringSetLength(&LastRequestedAttributeNameStr, 0);
  Tcl_DStringSetLength(&LastRequestedAttributeValueStr, 0);
  LastRequestedAnnotationTypeObj = LastRequestedAttributeNameObj =
  LastRequestedAttributeValueObj = LastRequestedLongObj = NULL;
  LastRequestedLong = -1;
  /* printf(" ->Hashed Objects Cleared...\n"); */
  DestroyProgressInfoDialog();
} /* CDM_ObjectCache::RemoveUnusedObjects */

/*
 * RemoveUnusedLongs:
 * ----------------------
 * This method removes from the cache the unused objects storing long
 * integers. It is useful to be called after massive displacements of
 * Annotations, as the objects used in spans would have changes, so many
 * objects holding long integers may not be used any more...
 */
int CDM_ObjectCache::RemoveUnusedLongs(void) {
  register Tcl_HashEntry *entryPtr, *nextPtr;
  register Tcl_HashSearch searchPtr;
  register Tcl_Obj *ObjPtr;
  register int removed = 0;
  
  CreateProgressInfoDialog("Clearing Unused Long Integer Objects from the "
                           "Object Memory Cache...");
#ifdef DEBUG_HASHED_VALUES
  CDM_Debug("+++++++ CDM_ObjectCache::RemoveUnusedLongs Called!\n");
#endif /* DEBUG_HASHED_VALUES */

  entryPtr = Tcl_FirstHashEntry(&LongTable, &searchPtr); 
  while (entryPtr != NULL) {
    ObjPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    if (ObjPtr != NULL) {
      if (!Tcl_IsShared(ObjPtr)) {
        /* This object is used only by the table. Remove it... */
        Tcl_DecrRefCount(ObjPtr);
        nextPtr  = Tcl_NextHashEntry(&searchPtr);
        Tcl_DeleteHashEntry(entryPtr);
        entryPtr = nextPtr;
        removed++; NumberOfLongs--;
      } else {
        entryPtr = Tcl_NextHashEntry(&searchPtr);
      } /* if (!Tcl_IsShared(ObjPtr)) */
    } else {
      nextPtr  = Tcl_NextHashEntry(&searchPtr);
      Tcl_DeleteHashEntry(entryPtr);
      entryPtr = nextPtr;
      removed++;
    } /* if (ObjPtr != NULL) */
  }
  LastRequestedLong    = -1;
  LastRequestedLongObj = NULL;
  DestroyProgressInfoDialog();
  return removed;
} /* CDM_ObjectCache::RemoveUnusedLongs */

/*
 * Info:
 * ----------------------
 * This method will print the contents of the CDM_ObjectCache into the
 * provided file...
 */
void CDM_ObjectCache::Info(FILE *file) {
  Tcl_HashTable *tablePtr[4];
  const char *TableName[] = {"StringTable",         "AttributeTable",
                             "AttributeValueTable", "LongTable"};
  register Tcl_HashEntry *entryPtr;
  register Tcl_HashSearch searchPtr;
  register Tcl_Obj *ObjPtr;
  long *NumberOf[4];
  int i, total;
  
  /*
   * Store the table pointers on an array, so as to use a loop to free unused
   * objects...
   */
  tablePtr[0] = &StringTable;         NumberOf[0] = &NumberOfStrings;
  tablePtr[1] = &AttributeTable;      NumberOf[1] = &NumberOfAttributes;
  tablePtr[2] = &AttributeValueTable; NumberOf[2] = &NumberOfAttributeValues;
  tablePtr[3] = &LongTable;           NumberOf[3] = &NumberOfLongs;
  fprintf(file, "++ CDM_ObjectCache::Info: Printing Cached Objects...\n");
  fprintf(file, "++ Location of the object in memory: %p\n", this);
  for (i=0; i<4; i++) {
    fprintf(file, "++ Contents of table \"%s\" (%ld items)...\n",
            TableName[i], *NumberOf[i]);
    total = 0;
    entryPtr = Tcl_FirstHashEntry(tablePtr[i], &searchPtr); 
    while (entryPtr != NULL) {
      total++;
      ObjPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
      if (ObjPtr != NULL) {
        fprintf(file, "     %s -> %d (%d)\n", Tcl_GetString(ObjPtr),
                                              ObjPtr->refCount, total);
      } else {
        fprintf(file, "     %s -> %d (%d)\n", "NULL", 0, total);
      }
      entryPtr = Tcl_NextHashEntry(&searchPtr);
    }
  } /* for (i=0; i<4; i++) */
  fprintf(file, "++ CDM_ObjectCache::Info: Printing Cached Objects done!\n");
} /* CDM_ObjectCache::Info */

/*
 * Find:
 * ----------------------
 * This method will search the memory cache for the string contained in the
 * specified object. If an object that holds the same string is found in the
 * cache, it is returned. Otherwise the provided object is added to the cache
 * and returned.
 * Before the provided object is appended to the cache and returned,
 * this method will also check whether the string represents an integer
 * or a long integer. In such a case, it will create and return an object
 * that holds a long, as FindLong would have done.
 * The new object will not be appended to the string cache, but will be
 * appended to the long integer cache instead...
 */
Tcl_Obj *CDM_ObjectCache::Find(Tcl_Obj *strObj) {
  Tcl_HashEntry *EntryPtr;
  long Long;
  int newEntry;

  if (strObj == NULL || Tcl_GetCharLength(strObj) == 0) {
    CDM_THREAD_SPECIFIC_DATA;
    return &CDM_Obj_NULL;
  }
  /*
   * Search Hash table for given string...
   */
#ifdef CDM_OBJ_HASH_KEYS
  EntryPtr = Tcl_FindHashEntry(&StringTable, (char *) strObj);
#else  /* CDM_OBJ_HASH_KEYS */
  EntryPtr = Tcl_FindHashEntry(&StringTable, (char *) Tcl_GetString(strObj));
#endif /* CDM_OBJ_HASH_KEYS */
  if (EntryPtr != NULL) {
    /*
     * The specified string was found in the cache. Return the cached
     * object...
     */
    strObj = (Tcl_Obj *) Tcl_GetHashValue(EntryPtr);
#ifdef ENABLE_HASHED_VALUES_CHECKS
    if (strObj != (Tcl_Obj *) Tcl_GetHashKey(&StringTable, EntryPtr)) {
      snprintf(CDM_Temp, MAX_TEMP_STRING, "CDM_ObjectCache::Find:\n"
         "Internal inconsistency regarding key/value for entry \"%s\"\n",
          Tcl_GetString(strObj));
      CDM_Panic(CDM_Temp);
    }
#endif /* ENABLE_HASHED_VALUES_CHECKS */
#ifdef DEBUG_HASHED_VALUES
    CDM_Debug("CDM_ObjectCache::Find: Returning Hashed Obj (%s) (%d)\n",
              Tcl_GetString(strObj), strObj->refCount);
#endif /* DEBUG_HASHED_VALUES */
    return strObj;
  }

  /*
   * We have not fount given object. Check to see if requested object should
   * be a long integer object...
   * This was a BUG! (jkoutsi, 8/12/1999)
   * if (sscanf(string, "%ld", &Long) > 0) return CDM_NewLongObj(Long);
   */
  if (Tcl_GetLongFromObj(NULL, strObj, &Long) == TCL_OK) {
     return CDM_NewLongObj(Long);
  }
  
  /*
   * We have not found given object. So, we create a new one, we add it to the
   * table and we return it...
   */
#ifdef CDM_OBJ_HASH_KEYS
  EntryPtr = Tcl_CreateHashEntry(&StringTable, (char *) strObj, &newEntry);
#else  /* CDM_OBJ_HASH_KEYS */
  EntryPtr = Tcl_CreateHashEntry(&StringTable, (char *) Tcl_GetString(strObj),
                                 &newEntry);
#endif /* CDM_OBJ_HASH_KEYS */
  if (EntryPtr == NULL) {
    CDM_THREAD_SPECIFIC_DATA;
    snprintf(CDM_Temp, MAX_TEMP_STRING, "CDM_ObjectCache::Find:\n"
             "Tcl_CreateHashEntry returned NULL while creating entry \"%s\"\n",
             Tcl_GetString(strObj));
    CDM_Panic(CDM_Temp);
  }
  Tcl_SetHashValue(EntryPtr, (ClientData) strObj);
  NumberOfStrings++;
#ifndef CDM_OBJ_HASH_KEYS
  /*
   * If we don't use Tcl objects as keys, then we should keep a reference to
   * this object. But if we use Tcl objects as keys, this reference is not
   * needed, as the hash table places one for us...
   */
  Tcl_IncrRefCount(strObj);
#endif /* CDM_OBJ_HASH_KEYS */

#ifdef ENABLE_HASHED_VALUES_CHECKS
  if (strObj != (Tcl_Obj *) Tcl_GetHashKey(&StringTable, EntryPtr)) {
    snprintf(CDM_Temp, MAX_TEMP_STRING, "CDM_ObjectCache::Find:\n"
       "Internal inconsistency regarding newly created key for entry \"%s\"\n",
        Tcl_GetString(strObj));
    CDM_Panic(CDM_Temp);
  }
#endif /* ENABLE_HASHED_VALUES_CHECKS */
#ifdef DEBUG_HASHED_VALUES
  CDM_Debug("CDM_ObjectCache::Find: Returning New Obj (%s) (%d)\n",
            Tcl_GetString(strObj), strObj->refCount);
#endif /* DEBUG_HASHED_VALUES */
  return strObj;
} /* CDM_ObjectCache::Find */

/*
 * Find:
 * ----------------------
 * This method is the overloaded version of the previous one. It simply
 * accepts a char* to hold the desired string instead of a Tcl object...
 */
Tcl_Obj *CDM_ObjectCache::Find(const char *string) {
  Tcl_Obj *result, *strObj;

  if (string == NULL || strlen(string) == 0) {
    CDM_THREAD_SPECIFIC_DATA;
    return &CDM_Obj_NULL;
  }
  strObj = Tcl_NewStringObj(string, -1);
  if (strObj == NULL) CDM_Panic("CDM_ObjectCache::Find: Tcl_NewStringObj "
                                "returned NULL - out of memory?\n");
  Tcl_IncrRefCount(strObj);
  result = Find(strObj);
  Tcl_DecrRefCount(strObj);
  return result;
} /* CDM_ObjectCache::Find */

/*
 * FindLong:
 * ----------------------
 * This method will search the memory cache for the specified long integer.
 * If an object that holds the same long integer is found in the
 * cache, it is returned. Otherwise a new object that holds the long integer
 * is created, added to the cache and returned.
 */
Tcl_Obj *CDM_ObjectCache::FindLong(long value) {
  Tcl_HashEntry *EntryPtr;
  Tcl_Obj *ObjPtr;
  int newEntry;

  /*
   * Is value the same as previous request? If that is the case,
   * don't bother looking the hash table. Just return the object...
   */
  if (value == LastRequestedLong && LastRequestedLongObj != NULL)
      return LastRequestedLongObj;
  LastRequestedLong = value;

  /*
   * Search the Hash Table we cache long integers...
   */
  EntryPtr = Tcl_FindHashEntry(&LongTable, (char *) value);
  if (EntryPtr != NULL) {
    ObjPtr = (Tcl_Obj *) Tcl_GetHashValue(EntryPtr);
#ifdef ENABLE_HASHED_VALUES_CHECKS
    long storedValue = -1;
    if (Tcl_GetLongFromObj(NULL, ObjPtr, &storedValue) != TCL_OK ||
        storedValue != value) {
      snprintf(CDM_Temp, MAX_TEMP_STRING, "CDM_ObjectCache::FindLong:\n"
        "Internal inconsistency regarding stored value (%ld) for entry (%ld)\n",
         storedValue, value);
      CDM_Panic(CDM_Temp);
    }
#endif /* ENABLE_HASHED_VALUES_CHECKS */
#ifdef DEBUG_HASHED_VALUES
    CDM_Debug("CDM_ObjectCache::FindLong: Returning Hashed Obj (%ld) (%d)\n",
              value, ObjPtr->refCount);
#endif /* DEBUG_HASHED_VALUES */
    LastRequestedLongObj = ObjPtr;
    return ObjPtr;
  }
  
  /*
   * We have not found given value. So, we create a new one, we add it to the
   * table and we return it...
   */
  ObjPtr = Tcl_NewLongObj(value);
  if (ObjPtr == NULL)
      CDM_Panic("CDM_ObjectCache::FindLong: Cannot Allocate Memory!\n");
  EntryPtr = Tcl_CreateHashEntry(&LongTable, (char *) value, &newEntry);
  if (EntryPtr == NULL) {
    CDM_THREAD_SPECIFIC_DATA;
    snprintf(CDM_Temp, MAX_TEMP_STRING, "CDM_ObjectCache::FindLong:\n"
             "Tcl_CreateHashEntry returned NULL while creating entry \"%ld\"\n",
             value);
    CDM_Panic(CDM_Temp);
  }
  Tcl_SetHashValue(EntryPtr, (ClientData) ObjPtr);
  Tcl_IncrRefCount(ObjPtr); NumberOfLongs++;
#ifdef DEBUG_HASHED_VALUES
  CDM_Debug("CDM_ObjectCache::FindLong: Returning New Obj (%ld) (%d)\n", value,
          ObjPtr->refCount);
#endif /* DEBUG_HASHED_VALUES */
  LastRequestedLongObj = ObjPtr;
  return ObjPtr;
} /* CDM_ObjectCache::FindLong */

/*
 * FindAnnotationType:
 * ----------------------
 * This method is equivalent to calling directly Find. The only difference is
 * that this method "remembers" the last string requested, so it returns the
 * object directly if requested. It should be faster on repetitive requests of
 * the same string...
 */
Tcl_Obj *CDM_ObjectCache::FindAnnotationType(const char *string) {
  Tcl_Obj *ObjPtr;
  
  /* Is string NULL or ""? */
  if (string == NULL || strlen(string) == 0) {
    CDM_THREAD_SPECIFIC_DATA;
    return &CDM_Obj_NULL;
  }
  /*
   * Is the string the same as the previous request? If that is the case,
   * don't bother looking the hash table. Just return the Object...
   */
  if (CDM_Strcmp(Tcl_DStringValue(&LastRequestedAnnotationTypeStr),string)==0) {
#ifdef DEBUG_HASHED_VALUES
   CDM_Debug("CDM_ObjectCache::FindAnnotationType: Returning Last Obj"
             "(%s) (%d)\n", string, LastRequestedAnnotationTypeObj->refCount);
#endif /* DEBUG_HASHED_VALUES */
    return LastRequestedAnnotationTypeObj;
  }
  
  ObjPtr = Find(string);
  Tcl_DStringSetLength(&LastRequestedAnnotationTypeStr, 0);
  Tcl_DStringAppend(&LastRequestedAnnotationTypeStr, string, -1);
  LastRequestedAnnotationTypeObj = ObjPtr;
  return ObjPtr;
} /* CDM_ObjectCache::FindAnnotationType */

/*
 * FindAttributeName:
 * ----------------------
 * This method is equivalent to calling directly Find. The only difference is
 * that this method "remembers" the last string requested, so it returns the
 * object directly if requested. It should be faster on repetitive requests of
 * the same string...
 */
Tcl_Obj *CDM_ObjectCache::FindAttributeName(const char *string) {
  Tcl_Obj *ObjPtr;
  
  /* Is string NULL or ""? */
  if (string == NULL || strlen(string) == 0) {
    CDM_THREAD_SPECIFIC_DATA;
    return &CDM_Obj_NULL;
  }
  /*
   * Is the string the same as the previous request? If that is the case,
   * don't bother looking the hash table. Just return the Object...
   */
  if (CDM_Strcmp(Tcl_DStringValue(&LastRequestedAttributeNameStr), string)==0) {
#ifdef DEBUG_HASHED_VALUES
    CDM_Debug("CDM_ObjectCache::FindAttributeName: Returning Last Obj"
             "(%s) (%d)\n", string, LastRequestedAttributeNameObj->refCount);
#endif /* DEBUG_HASHED_VALUES */
    return LastRequestedAttributeNameObj;
  }
  
  ObjPtr = Find(string);
  Tcl_DStringSetLength(&LastRequestedAttributeNameStr, 0);
  Tcl_DStringAppend(&LastRequestedAttributeNameStr, string, -1);
  LastRequestedAttributeNameObj = ObjPtr;
  return ObjPtr;
} /* CDM_ObjectCache::FindAttributeName */

/*
 * FindAttributeValue:
 * ----------------------
 * This method is equivalent to calling directly Find. The only difference is
 * that this method "remembers" the last string requested, so it returns the
 * object directly if requested. It should be faster on repetitive requests of
 * the same string...
 */
Tcl_Obj *CDM_ObjectCache::FindAttributeValue(const char *string) {
  Tcl_Obj *ObjPtr;
  
  /* Is string NULL or "" ? */
  if (string == NULL || strlen(string) == 0) {
    CDM_THREAD_SPECIFIC_DATA;
    return &CDM_Obj_NULL;
  }
  /*
   * Is the string the same as the previous request? If that is the case,
   * don't bother looking the hash table. Just return the Object...
   */
  if (CDM_Strcmp(Tcl_DStringValue(&LastRequestedAttributeValueStr),string)==0) {
#ifdef DEBUG_HASHED_VALUES
    CDM_Debug("CDM_ObjectCache::FindAttributeValue: Returning Last Obj"
             "(%s) (%d)\n", string, LastRequestedAttributeValueObj->refCount);
#endif /* DEBUG_HASHED_VALUES */
    return LastRequestedAttributeValueObj;
  }
  
  ObjPtr = Find(string);
  Tcl_DStringSetLength(&LastRequestedAttributeValueStr, 0);
  Tcl_DStringAppend(&LastRequestedAttributeValueStr, string, -1);
  LastRequestedAttributeValueObj = ObjPtr;
  return ObjPtr;
} /* CDM_ObjectCache::FindAttributeValue */

/*
 * FindAttributeValueObj:
 * ----------------------
 * This method caches objects that represent Attribute Values. An Attribute
 * Value is a list object that holds two elements (also objects): the first
 * element holds the value type (such as GDM_STRING) while the second object
 * holds the value.
 * All caching semantics apply also here, i.e. this method will search the
 * cache to locate a suitable object. If one is found, it is returned. Else, a
 * new object is created, added to the cache and returned...
 * Note that this function will decide whether to store an object or not.
 * Currently, returned objects are cached only if their type is GDM_STRING and
 * the value length does not exceed 50 characters. When the type is
 * GDM_STRING_SET, the value (which is a list - set) will be reconstructed and
 * all the set elements will be cached (as it is highly probable that these
 * elements are Annotation Ids). In all other cases, suitable objects will be
 * created and returned, but they won't be cached...
 */
Tcl_Obj *CDM_ObjectCache::FindAttributeValueObj(int type, Tcl_Obj *Value) {
  register Tcl_HashEntry *EntryPtr;
  register CDM_AttributeValue AttrVal = NULL;
  Tcl_Obj *argv[2];
  register int newEntry;
  Tcl_Obj **Elem;
  int elements, i;
  CDM_THREAD_SPECIFIC_DATA;

  switch (type) {
    case CDM_STRING: {
      /*
       * Search Hash table for given string...
       */
#ifdef CDM_OBJ_HASH_KEYS
      EntryPtr = Tcl_FindHashEntry(&AttributeValueTable, (char *) Value);
#else  /* CDM_OBJ_HASH_KEYS */
      EntryPtr = Tcl_FindHashEntry(&AttributeValueTable,
                                   (char *) Tcl_GetString(Value));
#endif /* CDM_OBJ_HASH_KEYS */
      if (EntryPtr != NULL) {
        AttrVal = (Tcl_Obj *) Tcl_GetHashValue(EntryPtr);
#ifdef ENABLE_HASHED_VALUES_CHECKS
        if (Tcl_ListObjGetElements(NULL, AttrVal, &elements, &Elem) != TCL_OK ||
            CDM_Strcmp(Elem[1], (Tcl_Obj *) 
            Tcl_GetHashKey(&AttributeValueTable, EntryPtr)) != 0) {
          snprintf(CDM_Temp, MAX_TEMP_STRING,
            "CDM_ObjectCache::FindAttributeValueObj:\n"
            "Internal inconsistency regarding key/value for entry \"%s\"\n",
             Tcl_GetString(Value));
          CDM_Panic(CDM_Temp);
        }
#endif /* ENABLE_HASHED_VALUES_CHECKS */
#ifdef DEBUG_HASHED_VALUES
        CDM_Debug("CDM_ObjectCache::FindAttributeValueObj: Returning Hashed "
                  "Obj (%s) (%d)\n", Tcl_GetString(AttrVal), AttrVal->refCount);
#endif /* DEBUG_HASHED_VALUES */
        return AttrVal;
      }
      /*
       * We have not found given string. So, we create a new one, we add it to 
       * the table and we return it...
       */
      argv[0] = &CDM_Obj_GDM_STRING;
      argv[1] = Value;
      AttrVal = Tcl_NewListObj(2, argv);
      if (AttrVal == NULL) 
        CDM_Panic("CDM_ObjectCache::FindAttributeValueObj: Tcl_NewListObj "
                  "returned NULL - out of memory?\n");
      
      /*
       * Petasis, 08/04/2001
       * This hash table may grow significantly, containing information that
       * is not very repetitive. In this case, hashing these objects may
       * result in wasting memory instead of saving.
       * So, place a restriction over the size of the value. If the string is
       * longer than this limit, do not add it to the hash table. Usual such
       * cases are for example sentence parses stored as attributes, lists
       * that incorectly have as type CDM_STRING, etc. Usually, values that
       * tend to repeat have small lengths (I hope...)
       */
      if (Tcl_GetCharLength(Value) > 50) {
        return AttrVal;
      }

#ifdef CDM_OBJ_HASH_KEYS
      EntryPtr = Tcl_CreateHashEntry(&AttributeValueTable,
                                     (char *) Value, &newEntry);
#else  /* CDM_OBJ_HASH_KEYS */
      EntryPtr = Tcl_CreateHashEntry(&AttributeValueTable,
                                     (char *) Tcl_GetString(Value), &newEntry);
#endif /* CDM_OBJ_HASH_KEYS */
      Tcl_SetHashValue(EntryPtr, (ClientData) AttrVal);
      Tcl_IncrRefCount(AttrVal); NumberOfAttributeValues++;
#ifdef DEBUG_HASHED_VALUES
      CDM_Debug("CDM_ObjectCache::FindAttributeValueObj: Returning New Obj"
             " (%s) (%d)\n", Tcl_GetString(AttrVal), AttrVal->refCount);
#endif /* DEBUG_HASHED_VALUES */
      break;
    }
    case CDM_STRING_SET: {
      argv[0] = &CDM_Obj_GDM_STRING_SET;
      argv[1] = Value;
      /*
       * We try to create a new set of values, by re-using elements...
       */
      if (Tcl_ListObjGetElements(NULL, Value, &elements, &Elem) == TCL_OK) {
        argv[1] = Tcl_NewListObj(0, NULL);
        if (argv[1] != NULL) {
          for (i=0; i<elements; i++) {
            Tcl_ListObjAppendElement(NULL, argv[1], Find(Elem[i]));
          }
        } else {
          argv[1] = Value;
        }
      }
      AttrVal = Tcl_NewListObj(2, argv);
      if (AttrVal == NULL) 
        CDM_Panic("CDM_ObjectCache::FindAttributeValueObj: Tcl_NewListObj "
                  "returned NULL - out of memory?\n");
      break;
    }
    case CDM_IMAGE: {
      argv[0] = &CDM_Obj_CDM_IMAGE;
      argv[1] = Value;
      AttrVal = Tcl_NewListObj(2, argv);
      if (AttrVal == NULL) 
        CDM_Panic("CDM_ObjectCache::FindAttributeValueObj: Tcl_NewListObj "
                  "returned NULL - out of memory?\n");
      break;
    }
    case CDM_BASE64_IMAGE: {
      argv[0] = &CDM_Obj_CDM_BASE64_IMAGE;
      argv[1] = Value;
      AttrVal = Tcl_NewListObj(2, argv);
      if (AttrVal == NULL) 
        CDM_Panic("CDM_ObjectCache::FindAttributeValueObj: Tcl_NewListObj "
                  "returned NULL - out of memory?\n");
      break;
    }
    default: {
      CDM_ReportError("CDM_ObjectCache::FindAttributeValueObj: Invalid type "
                      "%d", type);
      return NULL;
    }
  }
  return AttrVal;
} /* CDM_ObjectCache::FindAttributeValueObj */

/*
 * FindAttributeValueObj:
 * ----------------------
 * This method is the overloaded version of the previous one. It simply
 * accepts a char* to hold the desired string instead of a Tcl object...
 */
Tcl_Obj *CDM_ObjectCache::FindAttributeValueObj(int type, const char *string) {
  register Tcl_Obj *result, *strObj;

  switch (type) {
    case CDM_STRING: {
      strObj = Find(string);
      break;
    }
    default: {
      strObj = Tcl_NewStringObj(string, -1);
    }
  }
  if (strObj == NULL) CDM_Panic("CDM_ObjectCache::FindAttributeValueObj: "
                      "Tcl_NewStringObj returned NULL - out of memory?\n");
  Tcl_IncrRefCount(strObj);
  result = FindAttributeValueObj(type, strObj);
  Tcl_DecrRefCount(strObj);
  return result;
} /* FindAttributeValueObj */

/* 
 * -- FindAttributeObj:
 *  This function tries to share Attribute Objects as much as possible.
 *  It takes as input the type of Attribute and its value Object, and
 *  tries to find an Object that already holds these values! This lookup 
 *  currently works only for values of type "CDM_STRING"(as values of type
 *  "CDM_STRING_SET" do not seem to repeat often. If we keep them, maybe we
 *  will loose more memory than if we don't)...
 */
/*
 * FindAttributeObj:
 * ----------------------
 * This method caches objects that represent Attributes. An Attribute
 * is a list object that holds two elements (also objects): the first
 * element holds the Attribute name (such as language) while the second object
 * is an Attribute Value.
 * All caching semantics apply also here, i.e. this method will search the
 * cache to locate a suitable object. If one is found, it is returned. Else, a
 * new object is created, added to the cache and returned...
 * Note that this function will decide whether to store an object or not.
 * Currently, returned objects are cached only if their type is GDM_STRING.
 * In all other cases, suitable objects will be
 * created and returned, but they won't be cached...
 */
Tcl_Obj *CDM_ObjectCache::FindAttributeObj(const char *Name,
                                           CDM_AttributeValue Value) {
  CDM_Attribute Attr;
  register Tcl_HashEntry *EntryPtr;
  Tcl_Obj **elements, *argv[2];
  int items, result, type, newEntry;
  
  if (Value == NULL) return NULL;
  /*
   * Is the value valid?
   */
  result = Tcl_ListObjGetElements(NULL, Value, &items, &elements);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("Mallformed Attribute Value \"%s\"\n"
                    "  Valid Attribute Value form: {Type Value}",
                    Tcl_GetString(Value));
    return NULL;
  }
  /*
   * Retrieve the value type...
   */
  type = CDM_GetAttributeTypeFromObject(elements[0]);
  if (type < CDM_STRING) {
    CDM_ReportError("Invalid Value Type \"%s\" while "
                    "creating Attribute with Name=\"%s\", Value=\"%s\"",
                    Tcl_GetString(elements[0]), Name, Tcl_GetString(Value));
    return NULL;
  }

  switch (type) {
    case CDM_STRING: {
      /*
       * Case of CDM_STRING
       */
#ifdef CDM_OBJ_HASH_KEYS
      argv[0] = FindAttributeName(Name); /* Cache the Attribute name... */
      argv[1] = Value;
      Attr = Tcl_NewListObj(2, argv);
      if (Attr == NULL) 
        CDM_Panic("CDM_ObjectCache::FindAttributeObj: Tcl_NewListObj returned"
                  " NULL - out of memory?\n");
      EntryPtr = Tcl_FindHashEntry(&AttributeTable, (char *) Attr);
      if (EntryPtr != NULL) {
        Tcl_DecrRefCount(Attr);
        Attr = (Tcl_Obj *) Tcl_GetHashValue(EntryPtr);
#ifdef DEBUG_HASHED_VALUES
        CDM_Debug("CDM_ObjectCache::FindAttributeObj: Returning Hashed Obj"
               "(%s) (%d)\n", Tcl_GetString(Attr), Attr->refCount);
#endif /* DEBUG_HASHED_VALUES */
        return Attr;
      }
      /* We have not found given object. Added to the table... */
      EntryPtr = Tcl_CreateHashEntry(&AttributeTable, (char *) Attr,
                                     &newEntry);
      Tcl_SetHashValue(EntryPtr, (ClientData) Attr);
#else /* CDM_OBJ_HASH_KEYS */
      char *value;
      Tcl_DString ds;
      value = Tcl_GetString(elements[1]);
      Tcl_DStringInit(&ds);
      Tcl_DStringAppend(&ds, value, -1);
      Tcl_DStringAppend(&ds, ":\1", 2);
      Tcl_DStringAppend(&ds, Name, -1);
      // Search Hash table for given string...
      EntryPtr=Tcl_FindHashEntry(&AttributeTable, Tcl_DStringValue(&ds));
      if (EntryPtr != NULL) {
        Attr = (Tcl_Obj *) Tcl_GetHashValue(EntryPtr);
#ifdef DEBUG_HASHED_VALUES
        CDM_Debug("CDM_ObjectCache::FindAttributeObj: Returning Hashed Obj"
               "(%s) (%d)\n", Tcl_GetString(Attr), Attr->refCount);
#endif /* DEBUG_HASHED_VALUES */
        Tcl_DStringFree(&ds);
        return Attr;
      }
      // We have not found given object. So, we create a new one, we add it to 
      // the table and we return it...
      argv[0] = FindAttributeName(Name);
      argv[1] = Value;
      Attr = Tcl_NewListObj(2, argv);
      if (Attr == NULL) 
        CDM_Panic("CDM_ObjectCache::FindAttributeObj: Tcl_NewListObj returned"
                  " NULL - out of memory?\n");
      EntryPtr = Tcl_CreateHashEntry(&AttributeTable, 
                                     Tcl_DStringValue(&ds), &newEntry);
      Tcl_SetHashValue(EntryPtr, (ClientData) Attr);
      Tcl_IncrRefCount(Attr);
      Tcl_DStringFree(&ds);
#ifdef DEBUG_HASHED_VALUES
      CDM_Debug("CDM_ObjectCache::FindAttributeObj: Returning New Obj"
               "(%s) (%d)\n", Tcl_GetString(Attr), Attr->refCount);
#endif /* DEBUG_HASHED_VALUES */
#endif /* CDM_OBJ_HASH_KEYS */
       NumberOfAttributes++;
      return Attr;
    }
    default: {
      /* 
       * All other cases...
       */
      argv[0] = FindAttributeName(Name);
      argv[1] = Value;
      Attr = Tcl_NewListObj(2, argv);
      if (Attr == NULL) 
        CDM_Panic("CDM_ObjectCache::FindAttributeObj: Tcl_NewListObj returned "
                  "NULL - out of memory?\n");
      return Attr;
    }
  }
  return NULL;
} /* CDM_ObjectCache::FindAttributeObj */

void CDM_ObjectCache::SetAttributeSetTablePtr(Tcl_HashTable *tablePtr) {
  AttributeSetTable = tablePtr;
}; /* CDM_ObjectCache::SetAttributeSetTablePtr */

/*
 * ExposeCache:
 * ----------------------
 * This function exposes the cache managed by this class to the Tcl level,
 * so as the cache contents to be viewable by the Tcl level.
 */
int CDM_ObjectCache::ExposeCache(Tcl_Interp *interp) {
  Tcl_Obj *StringTableObj, *LongTableObj, *AttributeValueTableObj,
          *AttributeTableObj, *AttributeSetTableObj, *ValObj;

  /*
   * Create the "::CDM::CDM_Cache" namespace...
   */
  CDM_Eval(interp, "namespace", "eval", "::CDM::CDM_Cache", "{}", (char *)NULL);
  /*
   * Create array variables by defining an element into each array...
   */
  ValObj = Tcl_NewStringObj(" ", -1); Tcl_IncrRefCount(ValObj);
  StringTableObj = Tcl_NewStringObj("::CDM::CDM_Cache::StringTable", -1);
  Tcl_IncrRefCount(StringTableObj);
  if (Tcl_ObjSetVar2(interp, StringTableObj, ValObj,
               CDM_NewLongObj(0), TCL_LEAVE_ERR_MSG) == NULL) return TCL_ERROR;
  Tcl_DecrRefCount(StringTableObj);
  LongTableObj = Tcl_NewStringObj("::CDM::CDM_Cache::LongTable", -1);
  Tcl_IncrRefCount(LongTableObj);
  if (Tcl_ObjSetVar2(interp, LongTableObj, ValObj,
               CDM_NewLongObj(0), TCL_LEAVE_ERR_MSG) == NULL) return TCL_ERROR;
  Tcl_DecrRefCount(LongTableObj);
  AttributeValueTableObj = 
               Tcl_NewStringObj("::CDM::CDM_Cache::AttributeValueTable", -1);
  Tcl_IncrRefCount(AttributeValueTableObj);
  if (Tcl_ObjSetVar2(interp, AttributeValueTableObj, ValObj,
               CDM_NewLongObj(0), TCL_LEAVE_ERR_MSG) == NULL) return TCL_ERROR;
  Tcl_DecrRefCount(AttributeValueTableObj);
  AttributeTableObj = Tcl_NewStringObj("::CDM::CDM_Cache::AttributeTable", -1);
  Tcl_IncrRefCount(AttributeTableObj);
  if (Tcl_ObjSetVar2(interp, AttributeTableObj, ValObj,
               CDM_NewLongObj(0), TCL_LEAVE_ERR_MSG) == NULL) return TCL_ERROR;
  Tcl_DecrRefCount(AttributeTableObj);
  AttributeSetTableObj = 
               Tcl_NewStringObj("::CDM::CDM_Cache::AttributeSetTable", -1);
  Tcl_IncrRefCount(AttributeSetTableObj);
  if (Tcl_ObjSetVar2(interp, AttributeSetTableObj, ValObj,
               CDM_NewLongObj(0), TCL_LEAVE_ERR_MSG) == NULL) return TCL_ERROR;
  Tcl_DecrRefCount(AttributeSetTableObj);
  Tcl_DecrRefCount(ValObj);

  /*
   * Add traces to each array, so as when the variables are accessed or
   * written Tcl to call us so as to set the proper values...
   */
  if (Tcl_TraceVar2(interp, "::CDM::CDM_Cache::StringTable", NULL,
           TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS |
           TCL_TRACE_ARRAY | TCL_TRACE_RESULT_OBJECT,
           (Tcl_VarTraceProc *) &ExposeCacheCallback,
           &StringTable) != TCL_OK) return TCL_ERROR;
  if (Tcl_TraceVar2(interp, "::CDM::CDM_Cache::LongTable", NULL,
           TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS |
           TCL_TRACE_ARRAY | TCL_TRACE_RESULT_OBJECT,
           (Tcl_VarTraceProc *) &ExposeCacheCallback,
           &LongTable) != TCL_OK) return TCL_ERROR;
  if (Tcl_TraceVar2(interp, "::CDM::CDM_Cache::AttributeValueTable", NULL,
           TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS |
           TCL_TRACE_ARRAY | TCL_TRACE_RESULT_OBJECT,
           (Tcl_VarTraceProc *) &ExposeCacheCallback,
           &AttributeValueTable) != TCL_OK) return TCL_ERROR;
  if (Tcl_TraceVar2(interp, "::CDM::CDM_Cache::AttributeTable", NULL,
           TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS |
           TCL_TRACE_ARRAY | TCL_TRACE_RESULT_OBJECT,
           (Tcl_VarTraceProc *) &ExposeCacheCallback,
           &AttributeTable) != TCL_OK) return TCL_ERROR;
  if (Tcl_TraceVar2(interp, "::CDM::CDM_Cache::AttributeSetTable", NULL,
           TCL_TRACE_READS | TCL_TRACE_WRITES | TCL_TRACE_UNSETS |
           TCL_TRACE_ARRAY | TCL_TRACE_RESULT_OBJECT,
           (Tcl_VarTraceProc *) &ExposeCacheCallback,
           AttributeSetTable) != TCL_OK) return TCL_ERROR;
  return TCL_OK;
} /* CDM_ObjectCache::ExposeCache */

/*
 * ExposeCacheCallback:
 * ----------------------
 * This function exposes the cache managed by this class to the Tcl level,
 * so as the cache contents to be viewable by the Tcl level.
 */
char *CDM_ObjectCache::ExposeCacheCallback(ClientData clientData,
          Tcl_Interp *interp, const char *name1, const char *name2, int flags) {
  Tcl_HashTable *table = (Tcl_HashTable *) clientData;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch searchPtr;
  Tcl_Obj *Result, *Name1, *Name2;
  const char *ptr;
  if (table == NULL) return NULL;

  if (flags & TCL_TRACE_DESTROYED || flags & TCL_INTERP_DESTROYED) return NULL;
  /*
   * Is it a request on the whole array or on a single element?
   */
  if (name2 != NULL) {
    /*
     * Access to an array element...
     */
    if (flags & TCL_TRACE_WRITES || flags & TCL_TRACE_UNSETS) {
      /*
       * Elements cannot be modified!
       */
      Result = Tcl_NewStringObj("Elements from this array cannot be "
                                "modified/unset", -1);
      Tcl_IncrRefCount(Result);
      return (char *) Result;
    } else if (flags & TCL_TRACE_READS) {
      CDM_THREAD_SPECIFIC_DATA;
      /*
       * An attempt to read one of our elements. Redefine its value or return
       * an error if it does not exist...
       */
      if (table == LongTablePtr) {
        int value;
        if (Tcl_GetInt(NULL, name2, &value) != TCL_OK) {
          Result = Tcl_NewStringObj("no such element in array: only integer "
                                    "keys are valid", -1);
          Tcl_IncrRefCount(Result);
          return (char *) Result;
        }
        entryPtr = Tcl_FindHashEntry(table, (char *) value);
      } else {
        Result = Tcl_NewStringObj(name2, -1);
        entryPtr = Tcl_FindHashEntry(table, (char *) Result);
        Tcl_DecrRefCount(Result);
        /*
         * If this is the AttributeValueTable, then we should remove the value
         * type from the name2...
         */
        if (entryPtr == NULL) {
          ptr = strchr(name2, ' ');
          if (ptr && ++ptr != NULL) {
            if (strcmp(ptr, "{}") == 0) ptr +=2;
            Result = Tcl_NewStringObj(ptr, -1);
            entryPtr = Tcl_FindHashEntry(table, (char *) Result);
            Tcl_DecrRefCount(Result);
          }
        }
      }
      
      if (entryPtr == NULL) {
        Result = Tcl_NewStringObj("no such element in array", -1);
        Tcl_IncrRefCount(Result);
        return (char *) Result;
      } else {
        Name1 = Tcl_NewStringObj(name1, -1);
        Name2 = Tcl_NewStringObj(name2, -1);
        Result = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
        Tcl_ObjSetVar2(interp, Name1, Name2,
             Tcl_NewIntObj(Result->refCount), TCL_LEAVE_ERR_MSG);
        Tcl_DecrRefCount(Name1);
        Tcl_DecrRefCount(Name2);
        return (char *) NULL;
      }
    } else {
      Result = Tcl_NewStringObj("don't know what to do :-)", -1);
      Tcl_IncrRefCount(Result);
      return (char *) Result;
    }
  } else {
    /*
     * Access to the whole array... First, uset the whole array...
     */
    if (flags & TCL_TRACE_WRITES || flags & TCL_TRACE_UNSETS) {
      /*
       * Elements cannot be modified!
       */
      Result = Tcl_NewStringObj("Elements from this array cannot be "
                                "modified/unset", -1);
      Tcl_IncrRefCount(Result);
      return (char *) Result;
    }
    CDM_Eval(interp, "array", "unset", name1, "*", (char *) NULL);
    /*
     * Iterate over the table elements and define the proper entries...
     */
    Name1 = Tcl_NewStringObj(name1, -1);
    for (entryPtr = Tcl_FirstHashEntry(table, &searchPtr);
         entryPtr != NULL;
         entryPtr = Tcl_NextHashEntry(&searchPtr)) {
      /* Name2  = (Tcl_Obj *) Tcl_GetHashKey(table, entryPtr); */
      Result = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
      Tcl_ObjSetVar2(interp, Name1, Result,
             Tcl_NewIntObj(Result->refCount), TCL_LEAVE_ERR_MSG);
    }
    Tcl_DecrRefCount(Name1);
  }

  return (char *) NULL;
} /* CDM_ObjectCache::ExposeCacheCallback */

/*
 *----------------------------------------------------------------------
 *
 * HashObjKey --
 *
 *        Compute a one-word summary of the string representation of the
 *        Tcl_Obj, which can be used to generate a hash index.
 *
 * Results:
 *        The return value is a one-word summary of the information in
 *        the string representation of the Tcl_Obj.
 *
 * Side effects:
 *        None.
 *
 *----------------------------------------------------------------------
 */
unsigned
int CDM_ObjectCache::HashObjKey(Tcl_HashTable *tablePtr, VOID *keyPtr) {
  Tcl_Obj *objPtr = (Tcl_Obj *) keyPtr;
  register CONST char *string;
  register int length;
  register unsigned int result;
  register int c;

  string = Tcl_GetString(objPtr);
  length = objPtr->length;
  
  /*
   * I tried a zillion different hash functions and asked many other
   * people for advice.  Many people had their own favorite functions,
   * all different, but no-one had much idea why they were good ones.
   * I chose the one below (multiply by 9 and add new character)
   * because of the following reasons:
   *
   * 1. Multiplying by 10 is perfect for keys that are decimal strings,
   *    and multiplying by 9 is just about as good.
   * 2. Times-9 is (shift-left-3) plus (old).  This means that each
   *    character's bits hang around in the low-order bits of the
   *    hash value for ever, plus they spread fairly rapidly up to
   *    the high-order bits to fill out the hash value.  This seems
   *    works well both for decimal and non-decimal strings.
   */

  result = 0;
  while (length) {
      c = *string;
      string++;
      length--;
      if (length == 0) {
          break;
      }
      result += (result<<3) + c;
  }
  return result;
} /* CDM_ObjectCache::HashObjKey */

/*
 *----------------------------------------------------------------------
 *
 * CompareObjKeys --
 *
 *        Compares two Tcl_Obj * keys.
 *
 * Results:
 *        The return value is 0 if they are different and 1 if they are
 *        the same.
 *
 * Side effects:
 *        None.
 *
 *----------------------------------------------------------------------
 */
int CDM_ObjectCache::CompareObjKeys(VOID *keyPtr, Tcl_HashEntry *hPtr) {
  Tcl_Obj *objPtr1 = (Tcl_Obj *) keyPtr;
  Tcl_Obj *objPtr2 = (Tcl_Obj *) hPtr->key.oneWordValue;
  register CONST char *p1, *p2;
  register int l1, l2;
  Tcl_Obj **elements1, **elements2;
  int result, items;

  /*
   * If the object pointers are the same then they match.
   */
  if (objPtr1 == objPtr2) {
      return 1;
  }

  /*
   * The objects that we receive are Attributes. First we compare the
   * Attribute names...
   */
  result = Tcl_ListObjGetElements(NULL, objPtr1, &items, &elements1);
  if (result != TCL_OK || items != 2) goto obj_comparison;
  result = Tcl_ListObjGetElements(NULL, objPtr2, &items, &elements2);
  if (result != TCL_OK || items != 2) goto obj_comparison;
  if (elements1[0] != elements2[0]) {
    p1 = Tcl_GetString(elements1[0]);
    l1 = elements1[0]->length;
    p2 = Tcl_GetString(elements2[0]);
    l2 = elements2[0]->length;
    if (l1 != l2) return 0;
    for (;l1 >= 0; p1++, p2++, l1--) {
      if (*p1 != *p2) return 0;
    }
  }
  
  /*
   * If we got here, the two Attribute names are the same. We have to compare
   * the Attribute values...
   */
  result = Tcl_ListObjGetElements(NULL, elements1[1], &items, &elements1);
  if (result != TCL_OK || items != 2) goto obj_comparison;
  result = Tcl_ListObjGetElements(NULL, elements2[1], &items, &elements2);
  if (result != TCL_OK || items != 2) goto obj_comparison;

  /*
   * Compare the value types...
   */
  if (elements1[0] != elements2[0]) {
    p1 = Tcl_GetString(elements1[0]);
    l1 = elements1[0]->length;
    p2 = Tcl_GetString(elements2[0]);
    l2 = elements2[0]->length;
    if (l1 != l2) return 0;
    for (;l1 >= 0; p1++, p2++, l1--) {
      if (*p1 != *p2) return 0;
    }
  }

  /*
   * Compare the value value...
   */
  if (elements1[1] != elements2[1]) {
    p1 = Tcl_GetString(elements1[1]);
    l1 = elements1[1]->length;
    p2 = Tcl_GetString(elements2[1]);
    l2 = elements2[1]->length;
    if (l1 != l2) return 0;
    for (;l1 >= 0; p1++, p2++, l1--) {
      if (*p1 != *p2) return 0;
    }
  }

  /*
   * The two objects are the same...
   */
  return 1;
  
obj_comparison:
  /*
   * Don't use Tcl_GetStringFromObj as it would prevent l1 and l2 being
   * in a register.
   */
  p1 = Tcl_GetString (objPtr1);
  l1 = objPtr1->length;
  p2 = Tcl_GetString (objPtr2);
  l2 = objPtr2->length;
  
  /*
   * Only compare if the string representations are of the same length.
   */
  if (l1 == l2) {
      for (;; p1++, p2++, l1--) {
          if (*p1 != *p2) {
              break;
          }
          if (l1 == 0) {
              return 1;
          }
      }
  }

  return 0;
} /* CDM_ObjectCache::CompareObjKeys */

/*
 *----------------------------------------------------------------------
 *
 * AllocObjEntry --
 *
 *        Allocate space for a Tcl_HashEntry containing the Tcl_Obj * key.
 *
 * Results:
 *        The return value is a pointer to the created entry.
 *
 * Side effects:
 *        Increments the reference count on the object.
 *
 *----------------------------------------------------------------------
 */
Tcl_HashEntry *
CDM_ObjectCache::AllocObjEntry(Tcl_HashTable *tablePtr, VOID *keyPtr) {
  Tcl_Obj *objPtr = (Tcl_Obj *) keyPtr;
  Tcl_HashEntry *hPtr;

  hPtr = (Tcl_HashEntry *) ckalloc((unsigned) (sizeof(Tcl_HashEntry)));
  hPtr->key.oneWordValue = (char *) objPtr;
  Tcl_IncrRefCount (objPtr);

  return hPtr;
} /* CDM_ObjectCache::AllocObjEntry */

/*
 *----------------------------------------------------------------------
 *
 * FreeObjEntry --
 *
 *        Frees space for a Tcl_HashEntry containing the Tcl_Obj * key.
 *
 * Results:
 *        The return value is a pointer to the created entry.
 *
 * Side effects:
 *        Decrements the reference count of the object.
 *
 *----------------------------------------------------------------------
 */
void CDM_ObjectCache::FreeObjEntry(Tcl_HashEntry *hPtr) {
  Tcl_Obj *objPtr = (Tcl_Obj *) hPtr->key.oneWordValue;

  Tcl_DecrRefCount (objPtr);
  ckfree ((char *) hPtr);
} /* CDM_ObjectCache::FreeObjEntry */

/*
 * Functions for displaying and updating a progress dialog...
 */
void CDM_ObjectCache::CreateProgressInfoDialog(const char *message) {
  Tcl_Obj *objv[2];
  int status;
  CDM_THREAD_SPECIFIC_DATA;

  if (DialogWindow != NULL) DestroyProgressInfoDialog();
  objv[0] = Tcl_NewStringObj("::CDM::DisplayProgressInfoDialog", -1);
  objv[1] = Tcl_NewStringObj(message, -1);
  for (status=0; status < 2; status++) Tcl_IncrRefCount(objv[status]);
  status = Tcl_EvalObjv(CDM_Interp, 2, objv, TCL_EVAL_GLOBAL);
  if (status == TCL_OK) {
    DialogWindow = Tcl_GetObjResult(CDM_Interp);
    Tcl_IncrRefCount(DialogWindow);
  }
  for (status=0; status < 2; status++) Tcl_DecrRefCount(objv[status]);
} /* CDM_ObjectCache::CreateProgressInfoDialog */

void CDM_ObjectCache::UpdateProgressInfoDialog(int percent) {
  Tcl_Obj *objv[3];
  int status;
  if (DialogWindow == NULL) return;
  CDM_THREAD_SPECIFIC_DATA;

  objv[0] = Tcl_NewStringObj(":CDM::UpdateProgressInfoDialog", -1);
  objv[1] = DialogWindow;
  objv[2] = Tcl_NewIntObj(percent);
  for (status=0; status < 3; status++) Tcl_IncrRefCount(objv[status]);
  status = Tcl_EvalObjv(CDM_Interp, 3, objv, TCL_EVAL_GLOBAL);
  for (status=0; status < 3; status++) Tcl_DecrRefCount(objv[status]);
} /* CDM_ObjectCache::UpdateProgressInfoDialog */

void CDM_ObjectCache::DestroyProgressInfoDialog(void) {
  Tcl_Obj *objv[2];
  int status;

  if (DialogWindow != NULL) {
    CDM_THREAD_SPECIFIC_DATA;
    objv[0] = Tcl_NewStringObj("::CDM::DestroyProgressInfoDialog", -1);
    objv[1] = DialogWindow;
    for (status=0; status < 2; status++) Tcl_IncrRefCount(objv[status]);
    status = Tcl_EvalObjv(CDM_Interp, 2, objv, TCL_EVAL_GLOBAL);
    for (status=0; status < 2; status++) Tcl_DecrRefCount(objv[status]);
    Tcl_DecrRefCount(DialogWindow);
  }
  DialogWindow = NULL;
} /* CDM_ObjectCache::DestroyProgressInfoDialog */
