/*****************************************************************************
 *  CDM_MemoryCache.h:                                                       *
 * ----------                                                                *
 *  This file contains various function definitions that relate to reducing  *
 *  the memory required by CDM in order to store string, Annotation and      *
 *  Attribute objects...                                                     *
 *                                                                           *
 *  Copyright 1998-2010 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licencing information, please refer to the licence conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/

#ifndef _CDM_MEMORY_CACHE
#define _CDM_MEMORY_CACHE

#include "CDM.h"

/*
 * Class CDM_ObjectCache:
 * --------------------------------
 * This class implements a cache for Tcl objects of specific structures, that
 * are frequently needed and stored by CDM. The idea behind this cache is that
 * we try to "compress" memory, by trying to reuse as much as possible parts
 * of an object across objects.
 *
 * Typical example is Annotations: the Annotation type (such as
 * "token") tends to not change often and to be used by a large number of
 * Annotations. So, instead of allocating lets say 2000 memory segments that
 * that will hold the same information (in this case the word "token"), only
 * one is allocated and shared among all Annotations...
 *
 * The provided functions by this class are able to cache the following object
 * types:
 */
class CDM_ObjectCache {
  /*
   * Structures private to this class only:
   */
  private:
    int Initialised;                   /* A flag that is true if the class has
                                          been initialised */
    Tcl_HashTable StringTable;         /* A table caching objects holding plain
                                          strings, like Annotation types... */
    Tcl_HashTable LongTable;           /* A table caching objects holding 
                                          long integers, like Annotation Ids */
    Tcl_HashTable AttributeValueTable; /* A table caching objects holding 
                                          Attribute Values... */
    Tcl_HashTable AttributeTable;      /* A table caching objects holding 
                                          whole Attribute Objects... */
    Tcl_HashTable *AttributeSetTable;  /* A pointer to the external table where
                                          Attribute Sets are cached... */
    /*
     * Although Hash table lookup is fast, it is faster if we save the last
     * requested object. Usually CDM makes subsequent requests of the same
     * object...
     */
    Tcl_Obj      *LastRequestedAnnotationTypeObj;
    Tcl_DString   LastRequestedAnnotationTypeStr;
    Tcl_Obj      *LastRequestedAttributeNameObj;
    Tcl_DString   LastRequestedAttributeNameStr;
    Tcl_Obj      *LastRequestedAttributeValueObj;
    Tcl_DString   LastRequestedAttributeValueStr;
    Tcl_Obj      *LastRequestedLongObj;
    long          LastRequestedLong;
    
    /*
     * A few integers that keep track of how many objects currently the cache
     * holds. An quick statistics figure...
     */
    long          NumberOfStrings;
    long          NumberOfLongs;
    long          NumberOfAttributeValues;
    long          NumberOfAttributes;
    
    /*
     * The following structure is needed in order to create a Hash table of
     * a custom type. In this new Hash table we will store whole Attributes.
     * The Hash table that accepts objects check their string representation
     * in order to compare two objects. Since we know the object structure, we
     * can do better than that...
     */
    Tcl_HashKeyType AttributeHashTableType;

    /*
     * An object holding the name of a window, used as our progress dialog...
     */
    Tcl_Obj *DialogWindow;
  public:
    static unsigned int   HashObjKey(Tcl_HashTable *tablePtr, VOID *keyPtr);
    static int            CompareObjKeys(VOID *keyPtr, Tcl_HashEntry *hPtr);
    static Tcl_HashEntry *AllocObjEntry(Tcl_HashTable *tablePtr, VOID *keyPtr);
    static void           FreeObjEntry(Tcl_HashEntry *hPtr);
    
    /*
     * Public API to retrieve stored objects. All functions will store an
     * object if not already in the cache. Also, methods are provided that can
     * eliminate memory cache from unused objects, and return them back to Tcl
     * in order tot be reused. These functions are called by CDM periodically.
     * Note that some of them may require some processing time, especially if
     * the memory cache is very large. These functions for example are the
     * cause that Ellogon systems take some time to close, if a large
     * Collections has been opened within them...
     */
                  CDM_ObjectCache(); /* Class constructor. A dummy one, as it
                                        does nothing. The caller has to call
                                        Init() in order to initialise the
                                        class */
    void          Init();            /* Initialise class */
    void          Clear();           /* Reset class and clear memory cache
                                        completely (remove all stored objects)
                                      */
    void          RemoveUnusedObjects(void);
    int           RemoveUnusedLongs(void);
    void          Info(FILE *file);
    Tcl_Obj      *Find(const char *string);
    Tcl_Obj      *Find(Tcl_Obj *strObj);
    Tcl_Obj      *FindLong(long value);
    Tcl_Obj      *FindAnnotationType(const char *string);
    Tcl_Obj      *FindAttributeName(const char *string);
    Tcl_Obj      *FindAttributeValue(const char *string);
    Tcl_Obj      *FindAttributeValueObj(int type, const char *string);
    Tcl_Obj      *FindAttributeValueObj(int type, Tcl_Obj *string);
    Tcl_Obj      *FindAttributeObj(const char *name, Tcl_Obj *Value);

    void          SetAttributeSetTablePtr(Tcl_HashTable *tablePtr);

    /*
     * Simple Statistics...
     */
    long          GetNumberOfStrings(void)         {return NumberOfStrings;}
    long          GetNumberOfLongs(void)           {return NumberOfLongs;}
    long          GetNumberOfAttributeValues(void) {return
                                                      NumberOfAttributeValues;}
    long          GetNumberOfAttributes(void)      {return NumberOfAttributes;}

    /*
     * The following command exposes the cache contents to the Tcl level...
     */
    int           ExposeCache(Tcl_Interp *interp);
    static char  *ExposeCacheCallback(ClientData clientData, Tcl_Interp *interp,
                               const char *name1, const char *name2, int flags);

    /*
     * Display and Update a dialog showing our progress...
     */
    void          CreateProgressInfoDialog(const char *message);
    void          UpdateProgressInfoDialog(int percent);
    void          DestroyProgressInfoDialog(void);
}; /* class CDM_ObjectCache */

#endif /* _CDM_MEMORY_CACHE */
