/*****************************************************************************
 *  CDM_ObjectTypes.cpp:                                                     *
 * ------------------------------------                                      *
 *  This file implements a set of new Tcl object types. These types are used *
 *  by CDM in order to further reduce the memory required to store them.     *
 *  This memory reduction method works better if the TIPSTER API is used to  *
 *  get the sub-elements. If the various Tcl list handling facilities are    *
 *  convertions among types will be made, which is usually a bad thing...    *
 *                                                                           *
 *  Copyright 1998-2010 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licensing information, please refer to the license conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/

#ifdef ENABLE_ANNOTATION_OBJECT

#include <string.h>
#include <limits.h>
#include "CDM.h"
#include "CDM_ObjectTypes.h"
#include "CDM_GlobalObjects.h"
extern Tcl_ThreadDataKey dataKey;

#ifndef USHRT_MAX
#define USHRT_MAX 65535
#endif /* USHRT_MAX */

/*
 * Prototypes for procedures defined later in this file:
 */

static void DupAnnotationInternalRep _ANSI_ARGS_((Tcl_Obj *srcPtr,
               Tcl_Obj *cpPtr));
static void FreeAnnotationInternalRep _ANSI_ARGS_((Tcl_Obj *listPtr));
static int  SetAnnotationFromAny _ANSI_ARGS_((Tcl_Interp *interp,
               Tcl_Obj *objPtr));
static void UpdateStringOfAnnotation _ANSI_ARGS_((Tcl_Obj *listPtr));

//static void ConvertToShort _ANSI_ARGS_((Tcl_Obj *obj));
static void ConvertToLong _ANSI_ARGS_((Tcl_Obj *obj));

/*
 *  Annotation Object Type
 * ------------------------
 * The structure below defines the CDM_Annotation Tcl object type by means of
 * procedures that can be invoked by generic object code.
 *
 */
Tcl_ObjType cdmAnnotationType = {
    (char *) "cdmAnnotation",                /* name */
    FreeAnnotationInternalRep,               /* freeIntRepProc */
    DupAnnotationInternalRep,                /* dupIntRepProc */
    UpdateStringOfAnnotation,                /* updateStringProc */
    SetAnnotationFromAny                     /* setFromAnyProc */
};
Tcl_ObjType cdmCompactAnnotationType = {
    (char *) "cdmCompactAnnotation",         /* name */
    FreeAnnotationInternalRep,               /* freeIntRepProc */
    DupAnnotationInternalRep,                /* dupIntRepProc */
    UpdateStringOfAnnotation,                /* updateStringProc */
    SetAnnotationFromAny                     /* setFromAnyProc */
};

/*
 *----------------------------------------------------------------------
 *
 * CDM_NewAnnotationObj --
 *
 *        Create a new Annotation object. 
 *
 * Results:
 *        The newly created Annotation object or NULL in case of error.
 *
 * Side effects:
 *        None.
 *
 *----------------------------------------------------------------------
 */

Tcl_Obj *CDM_NewAnnotationObj(long id, Tcl_Obj *type, long start, long end,
                              CDM_AttributeSet attributes) {
  register Tcl_Obj *objPtr = Tcl_NewObj();
  /* Invalidate the internal string representation... */
  objPtr->bytes = NULL;

  /*
   * Can we use the compact internal representation?
   */
  if (id < USHRT_MAX && start < USHRT_MAX && end < USHRT_MAX) {
    register cdmCompactAnnotationObjInternalRep *annRepPtr;
    annRepPtr = (cdmCompactAnnotationObjInternalRep *)
                      ckalloc(sizeof(cdmCompactAnnotationObjInternalRep));
    annRepPtr->id         = (unsigned short) id+1;
    annRepPtr->type       = type;
    annRepPtr->attributes = attributes;
    annRepPtr->start      = (unsigned short) start+1;
    annRepPtr->span.end   = (unsigned short) end+1;
    objPtr->typePtr       = &cdmCompactAnnotationType;
    objPtr->internalRep.twoPtrValue.ptr1 = (VOID *) annRepPtr;
  } else {
    register cdmAnnotationObjInternalRep *annRepPtr;
    annRepPtr = (cdmAnnotationObjInternalRep *)
                      ckalloc(sizeof(cdmAnnotationObjInternalRep));
    annRepPtr->id         = id;
    annRepPtr->type       = type;
    annRepPtr->attributes = attributes;
    annRepPtr->start      = start;
    annRepPtr->span.end   = end;
    objPtr->typePtr       = &cdmAnnotationType;
    objPtr->internalRep.twoPtrValue.ptr1 = (VOID *) annRepPtr;
  }
  Tcl_IncrRefCount(type);
  Tcl_IncrRefCount(attributes);
  objPtr->internalRep.twoPtrValue.ptr2 = NULL;
  return objPtr;
}; /* CDM_NewAnnotationObj */

/*
 *----------------------------------------------------------------------
 *
 * CDM_NewAnnotationObj --
 *
 *        Create a new Annotation object. 
 *
 * Results:
 *        The newly created Annotation object or NULL in case of error.
 *
 * Side effects:
 *        None.
 *
 *----------------------------------------------------------------------
 */

Tcl_Obj *CDM_NewAnnotationObj(long id, Tcl_Obj *type, CDM_SpanSet spans,
                              CDM_AttributeSet attributes) {
  register Tcl_Obj *objPtr = Tcl_NewObj();
  /* Invalidate the internal string representation... */
  objPtr->bytes = NULL;

  /*
   * Can we use the compact internal representation?
   */
  if (id < USHRT_MAX) {
    register cdmCompactAnnotationObjInternalRep *annRepPtr;
    annRepPtr = (cdmCompactAnnotationObjInternalRep *)
                      ckalloc(sizeof(cdmCompactAnnotationObjInternalRep));
    annRepPtr->id           = (unsigned short) id+1;
    annRepPtr->type         = type;
    annRepPtr->attributes   = attributes;
    annRepPtr->start        = 0;
    annRepPtr->span.spanSet = spans;
    objPtr->typePtr         = &cdmCompactAnnotationType;
    objPtr->internalRep.twoPtrValue.ptr1 = (VOID *) annRepPtr;
  } else {
    register cdmAnnotationObjInternalRep *annRepPtr;
    annRepPtr = (cdmAnnotationObjInternalRep *)
                      ckalloc(sizeof(cdmAnnotationObjInternalRep));
    annRepPtr->id           = id;
    annRepPtr->type         = type;
    annRepPtr->attributes   = attributes;
    annRepPtr->start        = -1;
    annRepPtr->span.spanSet = spans;
    objPtr->typePtr         = &cdmAnnotationType;
    objPtr->internalRep.twoPtrValue.ptr1 = (VOID *) annRepPtr;
  }
  Tcl_IncrRefCount(type);
  Tcl_IncrRefCount(attributes);
  Tcl_IncrRefCount(spans);
  // printf("Spans: %p %d '%s'\n", spans, spans->refCount,
  //                               CDMTCL_GetString(spans));
  objPtr->internalRep.twoPtrValue.ptr2 = NULL;
  return objPtr;
}; /* CDM_NewAnnotationObj */

/*
 *----------------------------------------------------------------------
 *
 * SetAnnotationFromAny --
 *
 *        Attempt to generate an Annotation internal form for the Tcl object
 *        "objPtr".
 *
 * Results:
 *        The return value is TCL_OK or TCL_ERROR. If an error occurs during
 *        conversion, an error message is left in the interpreter's result
 *        unless "interp" is NULL.
 *
 * Side effects:
 *        If no error occurs, an Annotation is stored as "objPtr"s internal
 *        representation.
 *
 *----------------------------------------------------------------------
 */

static int
SetAnnotationFromAny (Tcl_Interp *interp, Tcl_Obj *objPtr)
  //Tcl_Interp *interp;          /* Used for error reporting if not NULL. */
  //Tcl_Obj *objPtr;             /* The object to convert. */
{
  const Tcl_ObjType *oldTypePtr = objPtr->typePtr;
  Tcl_Obj **span, **attrs, **elements, *type, *spans, *attributes;
  long id = -1L, start, end;
  int result, items;
  register int i;
  
  /* A valid Annotation is a 4-list object... */
  result = Tcl_ListObjGetElements(NULL, objPtr, &items, &elements);
  if (result != TCL_OK || items != 4) {
    CDM_SetActiveInterpreter(interp);
    CDM_ReportError("Mallformed Annotation \"%s\"", Tcl_GetString(objPtr));
    return TCL_ERROR;
  }

  /* Get the id: Id can be either empty or a long... */
  if (Tcl_GetCharLength(elements[0]) != 0) {
    if (Tcl_GetLongFromObj(interp, elements[0], &id) != TCL_OK) {
      CDM_SetActiveInterpreter(interp);
      return CDM_ReportError("Mallformed Annotation: Invalid Id \"%s\"",
                    Tcl_GetString(elements[0]));
    }
  }

  /* The span set is a list of spans. Each span is a list of two elements... */
  result = Tcl_ListObjGetElements(interp, elements[2], &items, &span);
  if (result != TCL_OK || items < 0) {
    CDM_SetActiveInterpreter(interp);
    return CDM_ReportError("Mallformed SpanList in Annotation \"%s\"",
                    Tcl_GetString(objPtr));
  }
  for (i=0; i<items; i++) {
    result = CDM_GetSpanOffsets(span[i], &start, &end);
    if (result != TCL_OK) {
      CDM_SetActiveInterpreter(interp);
      CDM_AppendError(": mallformed Span in position %d in Annotation \"%s\"",
                      i, Tcl_GetString(objPtr));
      return TCL_ERROR;
    }
  }
  /* If only one span, we are lacky. We can get rid of the span set! */
  if (items != 1) {
    start = end = -1L;
  }

  /* Attributes are stored in a list... */
  result = Tcl_ListObjGetElements(interp, elements[3], &items, &attrs);
  if (result != TCL_OK || items < 0) {
    CDM_SetActiveInterpreter(interp);
    CDM_ReportError("Mallformed Attribute Set in Annotation \"%s\"",
                    Tcl_GetString(objPtr));
    return TCL_ERROR;
  }
  for (i=0; i<items; i++) {
    if (!CDM_ValidAttribute(attrs[i])) {
      CDM_SetActiveInterpreter(interp);
      CDM_AppendError(": mallformed Attribute in position %d in Annotation "
                      "\"%s\"", i, Tcl_GetString(objPtr));
      return TCL_ERROR;
    }
  }

  /* Allocate and fill our internal representation structure... */
  type       = elements[1]; Tcl_IncrRefCount(type);
  /* Actually, the folowing uses more memory/time. Avoid it... */
  /* attributes = CDM_SearchForAttributeSet(elements[3]); */
  attributes = elements[3]; Tcl_IncrRefCount(attributes);
  spans      = elements[2];
  if (start != -1) {
    /*
     * Free the old internalRep before setting the new one. We do this as
     * late as possible to allow the conversion code, in particular
     * Tcl_GetStringFromObj, to use that old internalRep.
     */
    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
      oldTypePtr->freeIntRepProc(objPtr);
    }
    
    if (id < USHRT_MAX && start < USHRT_MAX && end < USHRT_MAX) {
      register cdmCompactAnnotationObjInternalRep *annRepPtr;
      annRepPtr = (cdmCompactAnnotationObjInternalRep *)
                        ckalloc(sizeof(cdmCompactAnnotationObjInternalRep));
      annRepPtr->id           = (unsigned short) id+1;
      annRepPtr->type         = type;
      annRepPtr->attributes   = attributes;
      annRepPtr->start        = (unsigned short) start+1;
      annRepPtr->span.end     = (unsigned short) end+1;
      objPtr->typePtr         = &cdmCompactAnnotationType;
      objPtr->internalRep.twoPtrValue.ptr1 = (VOID *) annRepPtr;
    } else {
      register cdmAnnotationObjInternalRep *annRepPtr;
      annRepPtr = (cdmAnnotationObjInternalRep *)
                        ckalloc(sizeof(cdmAnnotationObjInternalRep));
      annRepPtr->id           = id;
      annRepPtr->type         = type;
      annRepPtr->attributes   = attributes;
      annRepPtr->start        = start;
      annRepPtr->span.end     = end;
      objPtr->typePtr         = &cdmAnnotationType;
      objPtr->internalRep.twoPtrValue.ptr1 = (VOID *) annRepPtr;
    }
  } else {
    Tcl_IncrRefCount(spans);
    /*
     * Free the old internalRep before setting the new one. We do this as
     * late as possible to allow the conversion code, in particular
     * Tcl_GetStringFromObj, to use that old internalRep.
     */
    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
      oldTypePtr->freeIntRepProc(objPtr);
    }

    if (id < USHRT_MAX) {
      register cdmCompactAnnotationObjInternalRep *annRepPtr;
      annRepPtr = (cdmCompactAnnotationObjInternalRep *)
                        ckalloc(sizeof(cdmCompactAnnotationObjInternalRep));
      annRepPtr->id           = (unsigned short) id+1;
      annRepPtr->type         = type;
      annRepPtr->attributes   = attributes;
      annRepPtr->start        = 0;
      annRepPtr->span.spanSet = spans;
      objPtr->typePtr         = &cdmCompactAnnotationType;
      objPtr->internalRep.twoPtrValue.ptr1 = (VOID *) annRepPtr;
    } else {
      register cdmAnnotationObjInternalRep *annRepPtr;
      annRepPtr = (cdmAnnotationObjInternalRep *)
                        ckalloc(sizeof(cdmAnnotationObjInternalRep));
      annRepPtr->id           = id;
      annRepPtr->type         = type;
      annRepPtr->attributes   = attributes;
      annRepPtr->start        = -1;
      annRepPtr->span.spanSet = spans;
      objPtr->typePtr         = &cdmAnnotationType;
      objPtr->internalRep.twoPtrValue.ptr1 = (VOID *) annRepPtr;
    }
  }
  objPtr->internalRep.twoPtrValue.ptr2 = NULL;
  
  /* Keeping a string representation is a waste of space. Discard it... */
  Tcl_InvalidateStringRep(objPtr);
  return TCL_OK;
}; /* SetAnnotationFromAny */

/*
 *----------------------------------------------------------------------
 *
 * FreeAnnotationInternalRep --
 *
 *        Deallocate the storage associated with an Annotation object's internal
 *        representation.
 *
 * Results:
 *        None.
 *
 * Side effects:
 *        Frees annPtr's List* internal representation and sets annPtr's
 *        internalRep.twoPtrValue.ptr1 to NULL. Decrements the ref counts
 *        of all element objects, which may free them.
 *
 *----------------------------------------------------------------------
 */

static void
FreeAnnotationInternalRep(Tcl_Obj *annPtr)
  //Tcl_Obj *annPtr;        /* Annotation object with internal rep to free. */
{
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) annPtr->internalRep.twoPtrValue.ptr1;
  if (annPtr->typePtr == &cdmCompactAnnotationType) {
    Tcl_DecrRefCount(intRepPtr->type);
    Tcl_DecrRefCount(intRepPtr->attributes);
    if (intRepPtr->start == 0) {
      Tcl_DecrRefCount(intRepPtr->span.spanSet);
    }
  } else {
    register cdmAnnotationObjInternalRep *lintRepPtr =
             (cdmAnnotationObjInternalRep *) intRepPtr;
    Tcl_DecrRefCount(lintRepPtr->type);
    Tcl_DecrRefCount(lintRepPtr->attributes);
    if (lintRepPtr->start == -1) {
      Tcl_DecrRefCount(lintRepPtr->span.spanSet);
    }
  }
  ckfree((char *) intRepPtr);
  if (annPtr->internalRep.twoPtrValue.ptr2) {
    Tcl_DecrRefCount((Tcl_Obj *) annPtr->internalRep.twoPtrValue.ptr2);
  }

  annPtr->internalRep.twoPtrValue.ptr1 = NULL;
  annPtr->internalRep.twoPtrValue.ptr2 = NULL;
}; /* FreeAnnotationInternalRep */

/*
 *----------------------------------------------------------------------
 *
 * DupAnnotationInternalRep --
 *
 *        Initialize the internal representation of an Annotation Tcl_Obj to a
 *        copy of the internal representation of an existing object. 
 *
 * Results:
 *        None.
 *
 * Side effects:
 *        "srcPtr"s list internal rep pointer should not be NULL and we assume
 *        it is not NULL. We set "copyPtr"s internal rep to a pointer to a
 *        newly allocated List structure that, in turn, points to "srcPtr"s
 *        element objects. Those element objects are not actually copied but
 *        are shared between "srcPtr" and "copyPtr". The ref count of each
 *        element object is incremented.
 *
 *----------------------------------------------------------------------
 */

static void
DupAnnotationInternalRep(Tcl_Obj *srcPtr, Tcl_Obj *copyPtr)
  //Tcl_Obj *srcPtr;               /* Object with internal rep to copy. */
  //Tcl_Obj *copyPtr;             /* Object with internal rep to set. */
{
  register cdmCompactAnnotationObjInternalRep *srcIntRepPtr =
    (cdmCompactAnnotationObjInternalRep *) srcPtr->internalRep.twoPtrValue.ptr1;
  if (srcPtr->typePtr == &cdmCompactAnnotationType) {
    register cdmCompactAnnotationObjInternalRep *copyIntRepPtr;

    copyIntRepPtr = (cdmCompactAnnotationObjInternalRep *)
                        ckalloc(sizeof(cdmCompactAnnotationObjInternalRep));
    copyIntRepPtr->id         = srcIntRepPtr->id;
    copyIntRepPtr->type       = srcIntRepPtr->type;
    Tcl_IncrRefCount(copyIntRepPtr->type);
    copyIntRepPtr->attributes = srcIntRepPtr->attributes;
    Tcl_IncrRefCount(copyIntRepPtr->attributes);
    copyIntRepPtr->start      = srcIntRepPtr->start;
    if (srcIntRepPtr->start != 0) {
      copyIntRepPtr->span.end = srcIntRepPtr->span.end;
    } else {
      copyIntRepPtr->span.spanSet = srcIntRepPtr->span.spanSet;
      Tcl_IncrRefCount(copyIntRepPtr->span.spanSet);
    }
    copyPtr->internalRep.twoPtrValue.ptr1 = (VOID *) copyIntRepPtr;
    copyPtr->typePtr = &cdmCompactAnnotationType;
  } else {
    register cdmAnnotationObjInternalRep *lsrcIntRepPtr, *copyIntRepPtr;
    lsrcIntRepPtr = (cdmAnnotationObjInternalRep *) srcIntRepPtr;

    copyIntRepPtr = (cdmAnnotationObjInternalRep *)
                        ckalloc(sizeof(cdmAnnotationObjInternalRep));
    copyIntRepPtr->id         = lsrcIntRepPtr->id;
    copyIntRepPtr->type       = lsrcIntRepPtr->type;
    Tcl_IncrRefCount(copyIntRepPtr->type);
    copyIntRepPtr->attributes = lsrcIntRepPtr->attributes;
    Tcl_IncrRefCount(copyIntRepPtr->attributes);
    copyIntRepPtr->start      = lsrcIntRepPtr->start;
    if (lsrcIntRepPtr->start != -1) {
      copyIntRepPtr->span.end = lsrcIntRepPtr->span.end;
    } else {
      copyIntRepPtr->span.spanSet = lsrcIntRepPtr->span.spanSet;
      Tcl_IncrRefCount(copyIntRepPtr->span.spanSet);
    }
    copyPtr->internalRep.twoPtrValue.ptr1 = (VOID *) copyIntRepPtr;
    copyPtr->typePtr = &cdmAnnotationType;
  }
  
  if (srcPtr->internalRep.twoPtrValue.ptr2) {
    copyPtr->internalRep.twoPtrValue.ptr2 =
             srcPtr->internalRep.twoPtrValue.ptr2;
    Tcl_IncrRefCount((Tcl_Obj *) (copyPtr->internalRep.twoPtrValue.ptr2));
  } else {
    copyPtr->internalRep.twoPtrValue.ptr2 = NULL;
  }
}; /* DupAnnotationInternalRep */

/*
 *----------------------------------------------------------------------
 *
 * UpdateStringOfAnnotation --
 *
 *        Update the string representation for an Annotation object.
 *        Note: This procedure does not invalidate an existing old string rep
 *        so storage will be lost if this has not already been done. 
 *
 * Results:
 *        None.
 *
 * Side effects:
 *        The object's string is set to a valid string that results from
 *        the Annotation-to-string conversion. This string will be empty if the
 *        list has no elements. The Annotation internal representation
 *        should not be NULL and we assume it is not NULL.
 *
 *----------------------------------------------------------------------
 */

static void
UpdateStringOfAnnotation(Tcl_Obj *annPtr)
  //Tcl_Obj *annPtr;       /* Annotation object with string rep to update. */
{
  Tcl_Obj *tmp, *objv[4];
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) annPtr->internalRep.twoPtrValue.ptr1;
  register cdmAnnotationObjInternalRep *lintRepPtr = 
    (cdmAnnotationObjInternalRep *) intRepPtr;
  char *buffer;
  int len;
  CDM_THREAD_SPECIFIC_DATA;
  
  if (annPtr->typePtr == &cdmCompactAnnotationType) {
    /* We construct a list object from the internal representation and we
     * request its string representation... */
    if (intRepPtr->id == 0) {
      objv[0] = &CDM_Obj_NULL;
    } else {
      objv[0] = CDM_NewLongObj((long) intRepPtr->id-1);
    }
    objv[1] = intRepPtr->type;
    if (intRepPtr->start != 0) {
      /* Maybe a span set object is already allocated? */
      if (annPtr->internalRep.twoPtrValue.ptr2) {
        objv[2] = (Tcl_Obj *) annPtr->internalRep.twoPtrValue.ptr2;
      } else {
        objv[2] = CDM_CreateSpanSet(intRepPtr->start-1, intRepPtr->span.end-1);
      }
    } else {
      objv[2] = intRepPtr->span.spanSet;
    }
    objv[3] = intRepPtr->attributes;
  } else if (annPtr->typePtr == &cdmAnnotationType) {
    /* We construct a list object from the internal representation and we
     * request its string representation... */
    if (lintRepPtr->id == -1) {
      objv[0] = &CDM_Obj_NULL;
    } else {
      objv[0] = CDM_NewLongObj(lintRepPtr->id);
    }
    objv[1] = lintRepPtr->type;
    if (lintRepPtr->start != -1) {
      /* Maybe a span set object is already allocated? */
      if (annPtr->internalRep.twoPtrValue.ptr2) {
        objv[2] = (Tcl_Obj *) annPtr->internalRep.twoPtrValue.ptr2;
      } else {
        objv[2] = CDM_CreateSpanSet(lintRepPtr->start, lintRepPtr->span.end);
      }
    } else {
      objv[2] = lintRepPtr->span.spanSet;
    }
    objv[3] = lintRepPtr->attributes;
  } else {
    /* Why Tcl core has called us? */
    CDM_Panic("CDM Annotation Object Type: UpdateStringOfAnnotation was called"
              " with an object of unknown type!\n");
  }

  tmp = Tcl_NewListObj(4, objv);
  Tcl_IncrRefCount(tmp);
  buffer = Tcl_GetStringFromObj(tmp, &len);
  annPtr->bytes = (char *) ckalloc((unsigned) (len + 1));
  memcpy((VOID *) annPtr->bytes, (VOID *) buffer, (unsigned) len);
  annPtr->bytes[len] = '\0';
  annPtr->length = len;
  Tcl_DecrRefCount(tmp);
  /* Although this seems time consuming, it saved 7 MB and 1 CPU sec for our
   * benchmark! */
  if (annPtr->typePtr == &cdmCompactAnnotationType) {
    if(intRepPtr->attributes != &CDM_Obj_NULL) {
      Tcl_InvalidateStringRep(intRepPtr->attributes);
    }
    if (intRepPtr->start == 0) {
      Tcl_ListObjLength(NULL, intRepPtr->span.spanSet, &len);
      if (len) Tcl_InvalidateStringRep(intRepPtr->span.spanSet);
    } else if (annPtr->internalRep.twoPtrValue.ptr2) {
      tmp = (Tcl_Obj *) annPtr->internalRep.twoPtrValue.ptr2;
      Tcl_ListObjLength(NULL, tmp, &len);
      if (len) Tcl_InvalidateStringRep(tmp);
    }
  } else {
    if(lintRepPtr->attributes != &CDM_Obj_NULL) {
      Tcl_InvalidateStringRep(lintRepPtr->attributes);
    }
    if (lintRepPtr->start == -1) {
      Tcl_ListObjLength(NULL, intRepPtr->span.spanSet, &len);
      if (len) Tcl_InvalidateStringRep(lintRepPtr->span.spanSet);
    } else if (annPtr->internalRep.twoPtrValue.ptr2) {
      tmp = (Tcl_Obj *) annPtr->internalRep.twoPtrValue.ptr2;
      Tcl_ListObjLength(NULL, tmp, &len);
      if (len) Tcl_InvalidateStringRep(tmp);
    }
  }
}; /* UpdateStringOfAnnotation */

/*
 * Various functions for making our life easier :-)
 */
void CDMAnnObj_SetId(Tcl_Obj *Ann, long id) {
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
  if (id < USHRT_MAX) {
    if (Ann->typePtr == &cdmCompactAnnotationType) {
      intRepPtr->id = (unsigned short) id+1;
    } else {
      ((cdmAnnotationObjInternalRep *) 
        Ann->internalRep.twoPtrValue.ptr1)->id = id;
    }
  } else {
    if (Ann->typePtr == &cdmCompactAnnotationType) {
      ConvertToLong(Ann);
    }
    ((cdmAnnotationObjInternalRep *) 
      Ann->internalRep.twoPtrValue.ptr1)->id = id;
  }
  Tcl_InvalidateStringRep(Ann);
} /* CDMAnnObj_SetId */

void CDMAnnObj_SetType(Tcl_Obj *Ann, Tcl_Obj *type) {
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
  Tcl_IncrRefCount(type);
  if (Ann->typePtr == &cdmCompactAnnotationType) {
    if (intRepPtr->type) Tcl_DecrRefCount(intRepPtr->type);
    intRepPtr->type = type;
  } else {
    register cdmAnnotationObjInternalRep *lintRepPtr =
      (cdmAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
    if (lintRepPtr->type) Tcl_DecrRefCount(lintRepPtr->type);
    lintRepPtr->type = type;
  }
  Tcl_InvalidateStringRep(Ann);
} /* CDMAnnObj_SetType */

void CDMAnnObj_SetStart(Tcl_Obj *Ann, long start) {
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
  if (start < USHRT_MAX) {
    if (Ann->typePtr == &cdmCompactAnnotationType) {
      if (intRepPtr->start == 0) {
        if (intRepPtr->span.spanSet) Tcl_DecrRefCount(intRepPtr->span.spanSet);
        intRepPtr->span.end = 0;
      }
      intRepPtr->start = (unsigned short) start+1;
    } else {
      register cdmAnnotationObjInternalRep *lintRepPtr =
        (cdmAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
      if (lintRepPtr->start == -1) {
        if (lintRepPtr->span.spanSet) Tcl_DecrRefCount(lintRepPtr->span.spanSet);
        lintRepPtr->span.end = -1;
      }
      lintRepPtr->start = start;
    }
  } else {
    register cdmAnnotationObjInternalRep *lintRepPtr;
    if (Ann->typePtr == &cdmCompactAnnotationType) {
      ConvertToLong(Ann);
    }
    lintRepPtr = (cdmAnnotationObjInternalRep *)
                 Ann->internalRep.twoPtrValue.ptr1;
    if (lintRepPtr->start == -1) {
      if (lintRepPtr->span.spanSet) Tcl_DecrRefCount(lintRepPtr->span.spanSet);
        lintRepPtr->span.end = -1;
      }
    lintRepPtr->start = start;
  }
  /* We have changed the start. If a span set has been allocated, free it... */
  if (Ann->internalRep.twoPtrValue.ptr2) {
    Tcl_DecrRefCount((Tcl_Obj *) Ann->internalRep.twoPtrValue.ptr2);
    Ann->internalRep.twoPtrValue.ptr2 = NULL;
  }
  Tcl_InvalidateStringRep(Ann);
} /* CDMAnnObj_SetStart */

void CDMAnnObj_SetEnd(Tcl_Obj *Ann, long end) {
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
  if (end < USHRT_MAX) {
    if (Ann->typePtr == &cdmCompactAnnotationType) {
      if (intRepPtr->start == 0) {
        if (intRepPtr->span.spanSet) Tcl_DecrRefCount(intRepPtr->span.spanSet);
      }
      intRepPtr->span.end = (unsigned short) end+1;
    } else {
      register cdmAnnotationObjInternalRep *lintRepPtr =
        (cdmAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
      if (lintRepPtr->start == -1) {
        if (lintRepPtr->span.spanSet) Tcl_DecrRefCount(lintRepPtr->span.spanSet);
      }
      lintRepPtr->span.end = end;
    }
  } else {
    register cdmAnnotationObjInternalRep *lintRepPtr;
    if (Ann->typePtr == &cdmCompactAnnotationType) {
      ConvertToLong(Ann);
    }
    lintRepPtr = (cdmAnnotationObjInternalRep *)
                 Ann->internalRep.twoPtrValue.ptr1;
    if (lintRepPtr->start == -1) {
      if (lintRepPtr->span.spanSet) Tcl_DecrRefCount(lintRepPtr->span.spanSet);
    }
    lintRepPtr->span.end = end;
  }
  /* We have changed the end. If a span set has been allocated, free it... */
  if (Ann->internalRep.twoPtrValue.ptr2) {
    Tcl_DecrRefCount((Tcl_Obj *) Ann->internalRep.twoPtrValue.ptr2);
    Ann->internalRep.twoPtrValue.ptr2 = NULL;
  }
  Tcl_InvalidateStringRep(Ann);
} /* CDMAnnObj_SetEnd */

void CDMAnnObj_SetSpans(Tcl_Obj *Ann, CDM_SpanSet spans) {
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
  Tcl_IncrRefCount(spans); /* We will keep a reference to this object... */
  if (Ann->typePtr == &cdmCompactAnnotationType) {
    if (intRepPtr->start == 0) {
      if (intRepPtr->span.spanSet) Tcl_DecrRefCount(intRepPtr->span.spanSet);
    }
    intRepPtr->span.spanSet = spans;
    intRepPtr->start = 0;
  } else {
    register cdmAnnotationObjInternalRep *lintRepPtr =
      (cdmAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
    if (lintRepPtr->start == -1) {
      if (lintRepPtr->span.spanSet) Tcl_DecrRefCount(lintRepPtr->span.spanSet);
    }
    lintRepPtr->span.spanSet = spans;
    lintRepPtr->start = -1;
  }
  /* If a span set has been allocated, free it... */
  if (Ann->internalRep.twoPtrValue.ptr2) {
    Tcl_DecrRefCount((Tcl_Obj *) Ann->internalRep.twoPtrValue.ptr2);
    Ann->internalRep.twoPtrValue.ptr2 = NULL;
  }
  Tcl_InvalidateStringRep(Ann);
} /* CDMAnnObj_SetSpans */

void CDMAnnObj_SetAttributes(Tcl_Obj *Ann, CDM_AttributeSet attibutes) {
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
  Tcl_IncrRefCount(attibutes); /* We will keep a reference of this object */
  if (Ann->typePtr == &cdmCompactAnnotationType) {
    if (intRepPtr->attributes) Tcl_DecrRefCount(intRepPtr->attributes);
    intRepPtr->attributes = attibutes;
  } else {
    register cdmAnnotationObjInternalRep *lintRepPtr =
      (cdmAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
    if (lintRepPtr->attributes) Tcl_DecrRefCount(lintRepPtr->attributes);
    lintRepPtr->attributes = attibutes;
  }
  Tcl_InvalidateStringRep(Ann);
} /* CDMAnnObj_SetAttributes */

/*
 * This function will store a span set in the second unused pointer of an
 * Annotation object. This field, if exists is a list representation of a
 * single span that has as start/end the longs hold by the object...
 */
void CDMAnnObj_KeepSpanSet(Tcl_Obj *Ann, CDM_SpanSet spans) {
  Tcl_IncrRefCount(spans); /* We will keep a reference of this object */
  if (Ann->internalRep.twoPtrValue.ptr2) {
    Tcl_DecrRefCount((Tcl_Obj *) Ann->internalRep.twoPtrValue.ptr2);
  }
  Ann->internalRep.twoPtrValue.ptr2 = (VOID *) spans;
} /* CDMAnnObj_KeepSpanSet */

void CDMAnnObj_FreeKeptSpanSet(Tcl_Obj *Ann) {
  if (Ann->internalRep.twoPtrValue.ptr2) {
    Tcl_DecrRefCount((Tcl_Obj *) Ann->internalRep.twoPtrValue.ptr2);
    Ann->internalRep.twoPtrValue.ptr2 = NULL;
  }
} /* CDMAnnObj_FreeKeptSpanSet */

long CDMAnnObj_GetId(Tcl_Obj *Ann) {
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
  if (Ann->typePtr == &cdmCompactAnnotationType) {
    return (long) intRepPtr->id-1;
  }
  return ((cdmAnnotationObjInternalRep *)
          Ann->internalRep.twoPtrValue.ptr1)->id;
} /* CDMAnnObj_GetId */

Tcl_Obj *CDMAnnObj_GetType(Tcl_Obj *Ann) {
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
  if (Ann->typePtr == &cdmCompactAnnotationType) {
    return intRepPtr->type;
  }
  return ((cdmAnnotationObjInternalRep *)
          Ann->internalRep.twoPtrValue.ptr1)->type;
} /* CDMAnnObj_GetType */

long CDMAnnObj_GetStart(Tcl_Obj *Ann) {
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
  if (Ann->typePtr == &cdmCompactAnnotationType) {
    return (long) intRepPtr->start-1;
  }
  return ((cdmAnnotationObjInternalRep *)
          Ann->internalRep.twoPtrValue.ptr1)->start;
} /* CDMAnnObj_GetStart */

long CDMAnnObj_GetEnd(Tcl_Obj *Ann) {
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
  if (Ann->typePtr == &cdmCompactAnnotationType) {
    if (intRepPtr->start == 0) return -1;
    return (long) intRepPtr->span.end-1;
  }
  if (((cdmAnnotationObjInternalRep *)
       Ann->internalRep.twoPtrValue.ptr1)->start == -1) return -1;
  return ((cdmAnnotationObjInternalRep *)
          Ann->internalRep.twoPtrValue.ptr1)->span.end;
} /* CDMAnnObj_GetEnd */

Tcl_Obj *CDMAnnObj_GetSpans(Tcl_Obj *Ann) {
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
  if (Ann->typePtr == &cdmCompactAnnotationType) {
    if (intRepPtr->start != 0) {
      /* Perhaps a span set already stored? */
      if (Ann->internalRep.twoPtrValue.ptr2) {
        return ((Tcl_Obj *) Ann->internalRep.twoPtrValue.ptr2);
      }
      return NULL;
    }
    return intRepPtr->span.spanSet;
  }
  if (((cdmAnnotationObjInternalRep *)
       Ann->internalRep.twoPtrValue.ptr1)->start != -1) {
    /* Perhaps a span set already stored? */
    if (Ann->internalRep.twoPtrValue.ptr2) {
      return ((Tcl_Obj *) Ann->internalRep.twoPtrValue.ptr2);
    }
    return NULL;
  }
  return ((cdmAnnotationObjInternalRep *)
          Ann->internalRep.twoPtrValue.ptr1)->span.spanSet;
} /* CDMAnnObj_GetSpans */

Tcl_Obj *CDMAnnObj_GetAttributes(Tcl_Obj *Ann) {
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) Ann->internalRep.twoPtrValue.ptr1;
  if (Ann->typePtr == &cdmCompactAnnotationType) {
    return intRepPtr->attributes;
  }
  return ((cdmAnnotationObjInternalRep *)
          Ann->internalRep.twoPtrValue.ptr1)->attributes;
} /* CDMAnnObj_GetAttributes */

#if 0
/*
 * Change the internal representation to CDM_ANN_OBJ_REP_SHORT
 */
void ConvertToShort(Tcl_Obj *obj) {
  register cdmAnnotationObjInternalRep *intRepPtr =
    (cdmAnnotationObjInternalRep *) obj->internalRep.twoPtrValue.ptr1;
  register cdmCompactAnnotationObjInternalRep *sintRepPtr;
  if (obj->typePtr == &cdmCompactAnnotationType) return;
  sintRepPtr= (cdmCompactAnnotationObjInternalRep *)
                    ckalloc(sizeof(cdmCompactAnnotationObjInternalRep));
  sintRepPtr->id         = (unsigned short) intRepPtr->id+1;
  sintRepPtr->type       = intRepPtr->type;
  sintRepPtr->attributes = intRepPtr->attributes;
  sintRepPtr->start      = (unsigned short) intRepPtr->start+1;
  if (sintRepPtr->start != 0) {
    sintRepPtr->span.end = (unsigned short) intRepPtr->span.end+1;
  } else {
    sintRepPtr->span.spanSet = intRepPtr->span.spanSet;
  }
  obj->internalRep.twoPtrValue.ptr1 = (VOID *) sintRepPtr;
  obj->typePtr = &cdmCompactAnnotationType;
  ckfree((char *) intRepPtr);
} /* ConvertToShort */
#endif

/*
 * Change the internal representation to CDM_ANN_OBJ_REP_LONG
 */
void ConvertToLong(Tcl_Obj *obj) {
  register cdmCompactAnnotationObjInternalRep *intRepPtr =
    (cdmCompactAnnotationObjInternalRep *) obj->internalRep.twoPtrValue.ptr1;
  register cdmAnnotationObjInternalRep *lintRepPtr;
  if (obj->typePtr == &cdmAnnotationType) return;
  lintRepPtr= (cdmAnnotationObjInternalRep *)
                    ckalloc(sizeof(cdmAnnotationObjInternalRep));
  lintRepPtr->id         = (long int) intRepPtr->id-1;
  lintRepPtr->type       = intRepPtr->type;
  lintRepPtr->attributes = intRepPtr->attributes;
  lintRepPtr->start      = (long int) intRepPtr->start-1;
  if (intRepPtr->start != 0) {
    lintRepPtr->span.end = (long int) intRepPtr->span.end-1;
  } else {
    lintRepPtr->span.spanSet = intRepPtr->span.spanSet;
  }
  obj->internalRep.twoPtrValue.ptr1 = (VOID *) lintRepPtr;
  obj->typePtr = &cdmAnnotationType;
  ckfree((char *) intRepPtr);
} /* ConvertToLong */
#endif /* ENABLE_ANNOTATION_OBJECT */
