/*****************************************************************************
 *  CDM_Signals.cpp:                                                         *
 * ----------------------------------                                        *
 *  This file adds to CDM ability to catch (mainly Unix) signals and         *
 *  execute registered Tcl callbacks upon receival.                          *
 *  Code in this file is based upon the "expect" "trap" command from         *
 *  Don Libes, NIST, 9/1/93                                                  *
 *                                                                           *
 *  Copyright 1998-2010 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licensing information, please refer to the license conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/
/*****************************************************************************
 *                                                                           *
 * -- File: CDM_Signals.cpp                                                  *
 *    Collection & Document Manager (CDM):                                   *
 *    This file adds to CDM ability to catch (mainly Unix) signals and       *
 *    execute registered Tcl callbacks upon receival.                        *
 *    Code in this file is based upon the "expect" "trap" command from       *
 *    Don Libes, NIST, 9/1/93                                                *
 *    Version: 2.0                                                           *
 *      Author: Giorgos Petasis                                              *
 *    Last Revised: Mon 26 Feb 2001                                          *
 *****************************************************************************/

/*
 * Some signals:
 * Signal     Value     Action   Comment
 * -------------------------------------------------------------------------     * SIGHUP        1        A      Hangup detected on controlling terminal
 *                               or death of controlling process
 * SIGINT        2        A      Interrupt from keyboard
 * SIGQUIT       3        C      Quit from keyboard
 * SIGILL        4        C      Illegal Instruction
 * SIGABRT       6        C      Abort signal from abort(3)
 * SIGFPE        8        C      Floating point exception
 * SIGKILL       9       AEF     Kill signal
 * SIGSEGV      11        C      Invalid memory reference
 * SIGPIPE      13        A      Broken pipe: write to pipe with no readers
 * SIGALRM      14        A      Timer signal from alarm(2)
 * SIGTERM      15        A      Termination signal
 * SIGUSR1   30,10,16     A      User-defined signal 1
 * SIGUSR2   31,12,17     A      User-defined signal 2
 * SIGCHLD   20,17,18     B      Child stopped or terminated
 * SIGCONT   19,18,25            Continue if stopped
 * SIGSTOP   17,19,23    DEF     Stop process
 * SIGTSTP   18,20,24     D      Stop typed at tty
 * SIGTTIN   21,21,26     D      tty input for background process
 * SIGTTOU   22,22,27     D      tty output for background process
 *
 * The letters in the  "Action"  column  have  the  following
 *     meanings:
 *     A      Default action is to terminate the process.
 *     B      Default action is to ignore the signal.
 *     C      Default action is to terminate the process and dump core.
 *     D      Default action is to stop the process.
 *     E      Signal cannot be caught.
 *     F      Signal cannot be ignored.
 */
 
#include "CDM.h"
#ifdef CDM_TRAP_SIGNALS

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <sys/types.h>

#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#if defined(SIGCLD) && !defined(SIGCHLD)
#define SIGCHLD SIGCLD
#endif
#define NO_SIG 0
#include "CDM_GlobalObjects.h"
extern Tcl_ThreadDataKey dataKey;

static struct CDM_Trap {
  char *action;         /* Tcl command to execute upon sig */
                        /* Each is handled by the CDM_EvalTrapAction */
  int mark;             /* 1 if signal has occurred */
  Tcl_Interp *interp;   /* interp to use or 0 if we should use the */
                        /* interpreter active at the time the sig */
                        /* is processed */
  int code;             /* return our new code instead of code */
                        /* available when signal is processed */
  char *name;           /* name of signal */
  int reserved;         /* if unavailable for trapping */
} CDM_Traps[NSIG];

static int CDM_SigchldCount=0; /* # of sigchlds caught but not yet processed */

static int CDM_EvalTrapAction(Tcl_Interp *interp, int sig,
                              struct CDM_Trap *trap, int oldcode);

static int CDM_GotSig;  /* this records the last signal received */
                        /* it is only a hint and can be wiped out */
                        /* by multiple signals, but it will always */
                        /* be left with a valid signal that is */
                        /* pending */

static Tcl_AsyncHandler CDM_AsyncHandler;

static char *CDM_SignalToString(int sig) {
  if (sig <= 0 || sig > NSIG) return("SIGNAL OUT OF RANGE");
  return(CDM_Traps[sig].name);
}; /* CDM_SignalToString */

/* current sig being processed by user sig handler */
static int CDM_CurrentSignal = NO_SIG;

int CDM_NoStackDump = 0;    /* 1 if user has requested unrolling of */
                                /* stack with no trace */


int streq(const char *s1, const char *s2) {
  if (strcmp(s1, s2) == 0) return 1;
  return 0;
}; /* streq */

/*ARGSUSED*/
static int CDM_TopHalf(ClientData clientData, Tcl_Interp *interp, int code) {
  struct CDM_Trap *trap;        /* last trap processed */
  int rc;
  int i;
  Tcl_Interp *sig_interp;

  //expDiagLog("sighandler: handling signal(%d)\r\n",CDM_GotSig);

  if (CDM_GotSig <= 0 || CDM_GotSig >= NSIG) {
    CDM_ReportError("caught impossible signal %d\r\n",CDM_GotSig);
    //abort();
  }

  /* start to work on this sig.  CDM_GotSig can now be overwritten */
  /* and it won't cause a problem */
  CDM_CurrentSignal = CDM_GotSig;
  trap = &CDM_Traps[CDM_CurrentSignal];

  trap->mark = 0;

  /* decrement below looks dangerous */
  /* Don't we need to temporarily block CDM_BottomHalf? */
  if (CDM_CurrentSignal == SIGCHLD) {
    CDM_SigchldCount--;
    //expDiagLog("CDM_SigchldCount-- == %d\n",CDM_SigchldCount);
  }

  if (!trap->action) {
    /* In this one case, we let ourselves be called when no */
    /* signaler predefined, since we are calling explicitly */
    /* from another part of the program, and it is just simpler */
    if (CDM_CurrentSignal == 0) return code;
    CDM_ReportError("caught unexpected signal: %s (%d)",
              CDM_SignalToString(CDM_CurrentSignal), CDM_CurrentSignal);
    //abort();
  }

  if (trap->interp) {
    /* if trap requested original interp, use it */
    sig_interp = trap->interp;
  } else if (!interp) {
    /* else if another interp is available, use it */
    sig_interp = interp;
  } else {
    /* fall back to CDM_Interp */
    sig_interp = CDM_Interp;
  }

  rc = CDM_EvalTrapAction(sig_interp, CDM_CurrentSignal, trap, code);
  CDM_CurrentSignal = NO_SIG;

  /*
   * scan for more signals to process
   */

  /* first check for additional SIGCHLDs */
  if (CDM_SigchldCount) {
    CDM_GotSig = SIGCHLD;
    CDM_Traps[SIGCHLD].mark = 1;
    Tcl_AsyncMark(CDM_AsyncHandler);
  } else {
    CDM_GotSig = -1;
    for (i=1;i<NSIG;i++) {
      if (CDM_Traps[i].mark) {
        CDM_GotSig = i;
        Tcl_AsyncMark(CDM_AsyncHandler);
        break;
      }
    }
  }
  return rc;
}; /* CDM_TopHalf */

#ifdef REARM_SIG
int CDM_SigchldSleep;
static int CDM_RearmSigchld   = 0;  /* 1 if sigchld needs to be */
                                        /* rearmed (i.e., because it has */
                                        /* just gone off) */
static int CDM_RearmingSigchld = 0;
#endif

/* called upon receipt of a user-declared signal */
static void CDM_BottomHalf(int sig) {
#ifdef REARM_SIG
  /*
   * tiny window of death if same signal should arrive here
   * before we've reinstalled it
   */

  /* In SV, sigchld must be rearmed after wait to avoid recursion */
  if (sig != SIGCHLD) {
    signal(sig, CDM_BottomHalf);
  } else {
    /* request rearm */
    CDM_RearmSigchld = 1;
    if (CDM_RearmingSigchld) CDM_SigchldSleep = 1;
  }
#endif

  CDM_Traps[sig].mark = 1;
  CDM_GotSig = sig;             /* just a hint - can be wiped out by another */
  Tcl_AsyncMark(CDM_AsyncHandler);

  /* if we are called while this particular async is being processed */
  /* original async_proc will turn off "mark" so that when async_proc */
  /* is recalled, it will see that nothing was left to do */

  /* In case of SIGCHLD though, we must recall it as many times as
   * we have received it.
   */
  if (sig == SIGCHLD) {
    CDM_SigchldCount++;
  }
#if 0
  /* if we are doing an i_read, restart it */
  if (env_valid && (sig != 0)) longjmp(env,2);
#endif
}; /* CDM_BottomHalf */

/*ARGSUSED*/
void CDM_CDMRearmSigchld(Tcl_Interp *interp) {
#ifdef REARM_SIG
  if (CDM_RearmSigchld) {
    CDM_RearmSigchld = 0;
    CDM_RearmingSigchld = 1;
    signal(SIGCHLD,CDM_BottomHalf);
  }

  CDM_RearmingSigchld = 0;

  /* if the rearming immediately caused another SIGCHLD, slow down */
  /* It's probably one of Tcl's intermediary pipeline processes that */
  /* Tcl hasn't caught up with yet. */
  if (CDM_SigchldSleep) {
    exp_dsleep(interp,0.2);
    CDM_SigchldSleep = 0;
  }
#endif
}; /* CDM_CDMRearmSigchld */


void CDM_InitTrap() {
  int i;

  for (i=1;i<NSIG;i++) {
    CDM_Traps[i].name = Tcl_SignalId(i);
    CDM_Traps[i].action = 0;
    CDM_Traps[i].reserved = 0;
  }

  /*
   * fix up any special cases
   */

#if defined(SIGCLD)
  /* Tcl names it SIGCLD, not good for portable scripts */
  CDM_Traps[SIGCLD].name = "SIGCHLD";
#endif
#if defined(SIGALRM)
  CDM_Traps[SIGALRM].reserved = 1;
#endif
#if defined(SIGKILL)
  CDM_Traps[SIGKILL].reserved = 1;
#endif
#if defined(SIGSTOP)
  CDM_Traps[SIGSTOP].reserved = 1;
#endif

  CDM_AsyncHandler = Tcl_AsyncCreate(CDM_TopHalf, (ClientData) 0);
}; /* CDM_InitTrap */

/* given signal index or name as string, */
/* returns signal index or -1 if bad arg */
int CDM_StringToSignal(Tcl_Interp *interp, char * s) {
  int sig;
  char *name;

  /* try interpreting as an integer */
  if (1 == sscanf(s,"%d",&sig)) {
    if (sig > 0 && sig < NSIG) return sig;
  } else {
    /* try interpreting as a string */
    for (sig=1; sig<NSIG; sig++) {
      name = CDM_Traps[sig].name;
      if (streq(s, name) || streq(s, name+3)) return(sig);
    }
  }
  CDM_ReportError("invalid signal %s",s);
  return -1;
}; /* CDM_StringToSignal */

/*ARGSUSED*/
int CDM_TrapObjCmd(ClientData clientData, Tcl_Interp *interp,
                   int objc, Tcl_Obj *CONST objv[]) {
  char *action = 0;
  int n;                   /* number of signals in list */
  Tcl_Obj **list;          /* list of signals */
  char *arg = NULL;
  int len;                 /* length of action */
  int i;
  int show_name = 0;   /* if user asked for current sig by name */
  int show_number = 0; /* if user asked for current sig by number */
  int show_max = 0;    /* if user asked for NSIG-1 */
  int rc = TCL_OK;
  int new_code = 0;    /* if action result should overwrite orig */
  Tcl_Interp *new_interp = interp;/* interp in which to evaluate */
                                  /* action when signal occurs */

  objc--; objv++;

  while (objc) {
    arg = Tcl_GetString(*objv);
    if (streq(arg,"-code")) {
      objc--; objv++; 
      new_code = 1;
    } else if (streq(arg,"-interp")) {
      objc--; objv++; 
      new_interp = 0;
    } else if (streq(arg,"-name")) {
      objc--; objv++;
      show_name = 1;
    } else if (streq(arg,"-number")) {
      objc--; objv++;
      show_number = 1;
    } else if (streq(arg,"-max")) {
      objc--; objv++;
      show_max = 1;
    } else break;
  }

  if (show_name || show_number || show_max) {
    if (objc > 0) goto usage_error;
    if (show_max) {
      Tcl_SetObjResult(interp,Tcl_NewIntObj(NSIG-1));
    }

    if (CDM_CurrentSignal == NO_SIG) {
      Tcl_SetResult(interp,"no signal in progress",TCL_STATIC);
      return TCL_ERROR;
    }
    if (show_name) {
      /* skip over "SIG" */
      Tcl_SetResult(interp, CDM_SignalToString(CDM_CurrentSignal) + 3,
                    TCL_STATIC);
    } else {
      Tcl_SetObjResult(interp, Tcl_NewIntObj(CDM_CurrentSignal));
    }
    return TCL_OK;
  }

  if (objc == 0 || objc > 2) goto usage_error;

  if (objc == 1) {
    int sig = CDM_StringToSignal(interp, arg);
    if (sig == -1) return TCL_ERROR;

    if (CDM_Traps[sig].action) {
      Tcl_SetResult(interp, CDM_Traps[sig].action, TCL_STATIC);
    } else {
      Tcl_SetResult(interp, "SIG_DFL", TCL_STATIC);
    }
      return TCL_OK;
  }

  action = arg;

  /* objv[1] is the list of signals - crack it open */
  if (TCL_OK != Tcl_ListObjGetElements(interp,objv[1],&n,&list)) {
    CDM_ReportError(Tcl_GetStringResult(interp));
    goto usage_error;
  }

  for (i=0;i<n;i++) {
    char *s;
    int sig;

    s = Tcl_GetString(list[i]);

    sig = CDM_StringToSignal(interp,s);
    if (sig == -1) {
      rc = TCL_ERROR;
      break;
    }

    if (CDM_Traps[sig].reserved) {
      CDM_ReportError("cannot trap %s",CDM_SignalToString(sig));
      rc = TCL_ERROR;
      break;
    }

    //expDiagLog("trap: setting up signal %d (\"%s\")\r\n",sig,s);
    if (CDM_Traps[sig].action) ckfree(CDM_Traps[sig].action);
    if (streq(action,"SIG_DFL")) {
      /* should've been free'd by now if nec. */
      CDM_Traps[sig].action = 0;
      signal(sig,SIG_DFL);
#ifdef REARM_SIG
      if (sig == SIGCHLD)
        CDM_RearmSigchld = 0;
#endif /*REARM_SIG*/
      } else {
        len = 1 + strlen(action);
        CDM_Traps[sig].action = ckalloc(len);
        memcpy(CDM_Traps[sig].action,action,len);
        CDM_Traps[sig].interp = new_interp;
        CDM_Traps[sig].code = new_code;
        if (streq(action,"SIG_IGN")) {
          signal(sig,SIG_IGN);
        } else signal(sig,CDM_BottomHalf);
      }
    }
    /* It is no longer necessary to free the split list since it */
    /* is still owned by Tcl, yes? */
    /*        ckfree((char *)list); */
    return(rc);
 usage_error:
    CDM_ReportError("usage: CDM_Trap [command or SIG_DFL or SIG_IGN] {list of signals}");
    return TCL_ERROR;
}; /* CDM_TrapObjCmd */

/* called by CDM_TopHalf() to process the given signal */
static int CDM_EvalTrapAction(Tcl_Interp *interp, int sig,
                              struct CDM_Trap *trap, int oldcode) {
  int code_flag;
  int newcode;
  Tcl_Obj *eip = NULL;   /* errorInfo */
  Tcl_Obj *ecp = NULL;   /* errorCode */
  Tcl_Obj *irp = NULL;   /* interp->result */

  //expDiagLogU("async event handler: Tcl_Eval(");
  //expDiagLogU(trap->action);
  //expDiagLogU(")\r\n");

  /* save to prevent user from redefining trap->code while trap */
  /* is executing */
  code_flag = trap->code;

  if (!code_flag) {
    /* 
     * save return values
     */

    eip = Tcl_GetVar2Ex(interp,"errorInfo","",TCL_GLOBAL_ONLY);
    if (eip) eip = Tcl_DuplicateObj(eip);
    ecp = Tcl_GetVar2Ex(interp,"errorCode","",TCL_GLOBAL_ONLY);
    if (ecp) ecp = Tcl_DuplicateObj(ecp);
    irp = Tcl_GetObjResult(interp);
    if (irp) irp = Tcl_DuplicateObj(irp);
  }

  newcode = Tcl_GlobalEval(interp,trap->action);

  /*
   * if new code is to be ignored (usual case - see "else" below)
   *        allow only OK/RETURN from trap, otherwise complain
   */

  if (code_flag) {
    //expDiagLog("return value = %d for trap %s, action ",newcode,CDM_SignalToString(sig));
    //expDiagLogU(trap->action);
    //expDiagLogU("\r\n");
    if (0 != strcmp(Tcl_GetStringResult(interp),"")) {
      CDM_ReportError(interp->result);

      /*
       * Check errorinfo and see if it contains -nostack.
       * This shouldn't be necessary, but John changed the
       * top level interp so that it distorts arbitrary
       * return values into TCL_ERROR, so by the time we
       * get back, we'll have lost the value of errorInfo
       */

      eip = Tcl_GetVar2Ex(interp,"errorInfo","",TCL_GLOBAL_ONLY);
      if (eip) {
        CDM_NoStackDump = (0 == strncmp("-nostack",Tcl_GetString(eip),8));
      }
    }
  } else if (newcode != TCL_OK && newcode != TCL_RETURN) {
    if (newcode != TCL_ERROR) {
      CDM_ReportError("return value = %d for trap %s, action %s",
                      newcode, CDM_SignalToString(sig), trap->action);
    }
    Tcl_BackgroundError(interp);
  }

  if (!code_flag) {
    /*
     * restore values
     */
    Tcl_ResetResult(interp);        /* turns off Tcl's internal */
    /* flags: ERR_IN_PROGRESS, ERROR_CODE_SET */
    /* This also wipes clean errorInfo/Code/result which is why */
    /* all the calls to Tcl_Dup earlier */

    if (eip) {
      /* odd that Tcl doesn't have a call that does all this at once */
      int len;
      char *s = Tcl_GetStringFromObj(eip,&len);
      Tcl_AddObjErrorInfo(interp,s,len);
      Tcl_DecrRefCount(eip);
      /* we never incr'd it, but the code allows this */
    } else {
      Tcl_UnsetVar(interp,"errorInfo",0);
    }

    /* restore errorCode.  Note that Tcl_AddErrorInfo (above) */
    /* resets it to NONE.  If the previous value is NONE, it's */
    /* important to avoid calling Tcl_SetErrorCode since this */
    /* with cause Tcl to set its internal ERROR_CODE_SET flag. */
    if (ecp) {
      if (!streq("NONE",Tcl_GetString(ecp)))
        Tcl_SetErrorCode(interp,ecp);
      /* we're just passing on the errorcode obj */
      /* presumably, Tcl will incr ref count */
    } else {
      Tcl_UnsetVar(interp,"errorCode",0);
    }

    newcode = oldcode;

    /* note that since newcode gets overwritten here by old code */
    /* it is possible to return in the middle of a trap by using */
    /* "return" (or "continue" for that matter)! */
  }
  return newcode;
}; /* CDM_EvalTrapAction */

#endif /* CDM_TRAP_SIGNALS */
