/*****************************************************************************
 *  CDM_Utils.cpp:                                                           *
 * --------------------                                                      *
 *  This file implements various utilities used internally by the            *
 *  Collection & Document Manager.                                           *
 *                                                                           *
 *  Copyright 1998-2007 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licensing information, please refer to the license conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/

#include "CDM.h"
#include "CDM_GlobalObjects.h"
#include <string.h>
extern Tcl_ThreadDataKey dataKey;
static int AttrSetHash_CompareObjKeys(VOID *keyPtr, Tcl_HashEntry *hPtr);

/*
 * Linked List data structure...
 */

/*
 * CDM_NewListNode --
 *  Create a new list node.
 */
CDM_ListNode *CDM_NewListNode(void) {
  CDM_ListNode *ptr = NULL;

  ptr = (CDM_ListNode *) Tcl_Alloc(sizeof(CDM_ListNode));
  if (ptr == NULL) CDM_Panic("CDM_NewListNode: out of memory!\n");
  memset((void *)ptr, '\0', sizeof(CDM_ListNode));
  return ptr;
} /* CDM_NewListNode */

/*
 * CDM_NewListNodeArray --
 *  Create a new array of list nodes, in a continues memory space...
 */
CDM_ListNode **CDM_NewListNodeArray(int elements) {
  CDM_ListNode **ptr = NULL;

  ptr = (CDM_ListNode **) Tcl_Alloc(sizeof(CDM_ListNode)*elements);
  if (ptr == NULL) CDM_Panic("CDM_NewListNode: out of memory!\n");
  memset((void *)ptr, '\0', sizeof(CDM_ListNode)*elements);
  return ptr;
} /* CDM_NewListNodeArray */
/*
 * CDM_AppendListNode --
 *  Create a new list node.
 */
CDM_ListNode *CDM_AppendListNode(CDM_List *list, CDM_ListNode *node) {
  register CDM_ListNode *ptr = list;
  
  if (node == NULL) return list;
  if (list == NULL) {
    node->prev = NULL;
    node->next = NULL;
    return node;
  }
  
  while (ptr->next != NULL) ptr = ptr->next;

  ptr->next  = node;
  node->prev = ptr;
  node->next = NULL;
  return node;
} /* CDM_AppendListNode */

/*
 * CDM_DeleteListNode --
 *  Deletes a list node...
 */
CDM_List *CDM_DeleteListNode(CDM_List *list, CDM_ListNode *node) {
  CDM_ListNode *prev, *next;

  /*
   * Are we deleting the first node of the list? Then we should return a new
   * list...
   */
  if (node == list) {
    list = node->next;
    if (list != NULL) list->prev = NULL;
    if (node->ptr[CDM_DOCUMENT_ID]  != NULL)
      Tcl_Free(node->ptr[CDM_DOCUMENT_ID]);
    if (node->ptr[CDM_DOCUMENT_XID] != NULL)
      Tcl_Free(node->ptr[CDM_DOCUMENT_XID]);
    Tcl_Free((char *) node);
    return list;
  }

  /*
   * We are deleting something in the middle or the end of the list...
   */
  prev = node->prev;
  next = node->next;

  prev->next = next;
  if (next != NULL) next->prev = prev;
  if (node->ptr[CDM_DOCUMENT_ID]  != NULL)
    Tcl_Free(node->ptr[CDM_DOCUMENT_ID]);
  if (node->ptr[CDM_DOCUMENT_XID] != NULL)
    Tcl_Free(node->ptr[CDM_DOCUMENT_XID]);
  Tcl_Free((char *) node);
  return list;
} /* CDM_DeleteListNode */

/*
 * CDM_DeleteList --
 *  Frees a list...
 */
void CDM_DeleteList(CDM_List *list) {
  register CDM_ListNode *next, *ptr = list;
  /*register int i;*/
  
  while (ptr != NULL) {
    next = ptr->next;
    if (ptr->ptr[CDM_DOCUMENT_ID]  != NULL)
      Tcl_Free(ptr->ptr[CDM_DOCUMENT_ID]);
    if (ptr->ptr[CDM_DOCUMENT_XID] != NULL)
      Tcl_Free(ptr->ptr[CDM_DOCUMENT_XID]);
    Tcl_Free((char *) ptr);
    ptr = next;
  }
} /* CDM_DeleteList */

/*
 * CDM_ListFind --
 *  Find the first node that contains the string str...
 */
CDM_ListNode *CDM_ListFind(CDM_List *list, char *str, int slot) {
  register CDM_ListNode *ptr = list;
  size_t length;

  if (str == NULL) return NULL;
  if (slot < 0 || slot > CDM_LIST_POINTER_NUMBER) return NULL;
  length = strlen(str);
  
  while (ptr != NULL) {
    if (ptr->ptr[slot] != NULL) {
      if (strlen(ptr->ptr[slot]) == length &&
          strcmp(ptr->ptr[slot], str) == 0) {
        return ptr;
      }
    }
    ptr = ptr->next;
  }
  return NULL;
} /* CDM_ListFind */

/*
 * CDM_ListLength --
 *  Return the number of nodes that the list has...
 */
int CDM_ListLength(CDM_List *list) {
  register CDM_ListNode *ptr = list;
  register int items = 0;

  while (ptr != NULL) {
    items++;
    ptr = ptr->next;
  }
  return items;
} /* CDM_ListLength */

/*
 * CDM_ListGetObject --
 *  Create a Tcl Object that contains the values of the specified list slot...
 */
Tcl_Obj *CDM_ListGetObject(CDM_List *list, int slot) {
  register CDM_ListNode *ptr = list;
  Tcl_Obj *Obj, *strObj;

  Obj = Tcl_NewListObj(0, NULL);
  if (Obj == NULL) return NULL;
  
  while (ptr != NULL) {
    strObj = Tcl_NewStringObj(ptr->ptr[slot], -1);
    if (strObj != NULL) Tcl_ListObjAppendElement(NULL, Obj, strObj);
    ptr = ptr->next;
  }
  return Obj;
} /* CDM_ListGetObject */

/******************************************************************************
 ******************************************************************************
 ****       A t t r i b u t e   S e t   O b j e c t   C a c h e            ****
 ******************************************************************************
 *****************************************************************************/
/*
 * Functions used internally by the CDM in order to further compress memory.
 * Of cource it has a small penalty in execution time...
 */

/*
 * CDM_SearchForAttributeSet --
 *  This function keeps an internal cache of all used attribute sets in
 *  annotations. If the given annotation set is already in the cache, then the
 *  cached object is returned. If the attribute set is not in the cache, it is
 *  added and returned (its refcount is also increased). If the value NULL is
 *  given, then the whole cache is cleared, meaning that objects held only by
 *  the cache (have refcount < 2) are freed from the cache.
 */
void CDM_AttributeSetHashInitialise(void) {
  CDM_THREAD_SPECIFIC_DATA;
  if (!AttrSetHash_Initialised) {
#ifdef CDM_OBJ_HASH_KEYS
    /* Tcl_InitObjHashTable(&AttrSetHash); */
    AttrSetHash_TypeStruct.version         = TCL_HASH_KEY_TYPE_VERSION;
    AttrSetHash_TypeStruct.flags           = 0;
    AttrSetHash_TypeStruct.hashKeyProc     = &(CDM_ObjectCache::HashObjKey);
    AttrSetHash_TypeStruct.compareKeysProc = AttrSetHash_CompareObjKeys;
    AttrSetHash_TypeStruct.allocEntryProc  = &(CDM_ObjectCache::AllocObjEntry);
    AttrSetHash_TypeStruct.freeEntryProc   = &(CDM_ObjectCache::FreeObjEntry);
    Tcl_InitCustomHashTable(&AttrSetHash, TCL_CUSTOM_PTR_KEYS,
                            &AttrSetHash_TypeStruct);
#else  /* CDM_OBJ_HASH_KEYS */
    Tcl_InitHashTable(&AttrSetHash, TCL_STRING_KEYS);
#endif /* CDM_OBJ_HASH_KEYS */
    Tcl_IncrRefCount(&CDM_Obj_NULL);
    ObjectCache.SetAttributeSetTablePtr(&AttrSetHash);
    AttrSetHash_Initialised = 1;
  }
} /* CDM_AttributeSetHashInitialise */

int CDM_ClearAttributeSetHash(void) {
  register Tcl_HashEntry *entryPtr, *nextPtr;
  Tcl_HashSearch searchPtr;
  register Tcl_Obj *ObjPtr;
  register int i = 0;
  CDM_THREAD_SPECIFIC_DATA;
   
  ObjectCache.CreateProgressInfoDialog("Clearing Attribute Set Objects from "
                                       "the Object Memory Cache...");
  entryPtr = Tcl_FirstHashEntry(&AttrSetHash, &searchPtr); 
  while (entryPtr != NULL) {
    ObjPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    if (ObjPtr != NULL) {
      nextPtr  = Tcl_NextHashEntry(&searchPtr);
      Tcl_DeleteHashEntry(entryPtr);
#ifndef CDM_OBJ_HASH_KEYS
      Tcl_DecrRefCount(ObjPtr);
#endif
      entryPtr = nextPtr;
      i++;
    } else {
      entryPtr = Tcl_NextHashEntry(&searchPtr);
    }
  }
#ifdef CDM_AYTOCLEAN_COMPRESSEDMEMORYCACHE
  AttrSetHash_NumberOfCalls  = 0;
  AttrSetHash_NumberOfClears = 0;
#endif /* CDM_AYTOCLEAN_COMPRESSEDMEMORYCACHE */
  Tcl_DeleteHashTable(&AttrSetHash);
#ifdef CDM_OBJ_HASH_KEYS
  /* Tcl_InitObjHashTable(&AttrSetHash); */
  Tcl_InitCustomHashTable(&AttrSetHash, TCL_CUSTOM_PTR_KEYS,
                            &AttrSetHash_TypeStruct);
#else  /* CDM_OBJ_HASH_KEYS */
  Tcl_InitHashTable(&AttrSetHash, TCL_STRING_KEYS);
#endif /* CDM_OBJ_HASH_KEYS */
  ObjectCache.DestroyProgressInfoDialog();
  return i;
}; /* CDM_ClearAttributeSetHash */

int CDM_ClearUnusedAttributeSetHash(void) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * We have to iterate over all entries and remove those that are not
   * shared any more...
   */
  register Tcl_HashEntry *entryPtr, *nextPtr;
  register Tcl_HashSearch searchPtr;
  register Tcl_Obj *ObjPtr;
  register int i = 0;
   
#ifdef CDM_AYTOCLEAN_COMPRESSEDMEMORYCACHE
  if (AttrSetHash_NumberOfClears > 3) {
    /*
     * After 4 partial clears, do a full clear...
     */
    return CDM_ClearAttributeSetHash();
  }
#endif /* CDM_AYTOCLEAN_COMPRESSEDMEMORYCACHE */
  ObjectCache.CreateProgressInfoDialog("Clearing Unused Attribute Set Objects "
                                       "from the Object Memory Cache...");
  entryPtr = Tcl_FirstHashEntry(&AttrSetHash, &searchPtr); 
  while (entryPtr != NULL) {
    ObjPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    if (ObjPtr != NULL && !Tcl_IsShared(ObjPtr)) {
      nextPtr  = Tcl_NextHashEntry(&searchPtr);
#ifndef CDM_OBJ_HASH_KEYS
      Tcl_DecrRefCount(ObjPtr);
#endif
      Tcl_DeleteHashEntry(entryPtr);
      entryPtr = nextPtr;
      i++;
    } else {
      entryPtr = Tcl_NextHashEntry(&searchPtr);
    }
  }
#ifdef CDM_AYTOCLEAN_COMPRESSEDMEMORYCACHE
  AttrSetHash_NumberOfCalls  = 0;
  AttrSetHash_NumberOfClears ++;
#endif /* CDM_AYTOCLEAN_COMPRESSEDMEMORYCACHE */
  ObjectCache.DestroyProgressInfoDialog();
  return i;
} /* CDM_ClearUnusedAttributeSetHash */

CDM_AttributeSet CDM_SearchForAttributeSet(CDM_AttributeSet Set) {
  Tcl_HashEntry *entryPtr;
  Tcl_Obj **Element;
  int newPtr, elements;
  CDM_THREAD_SPECIFIC_DATA;

  if (Tcl_ListObjGetElements(CDM_Interp, Set, &elements, &Element) != TCL_OK) {
    /*
     * Now, what we can do? This is not a valid Attribute Set... Return NULL
     * or the empty set?
     */
    return NULL;
  }

  /*
   * If the Attribute set is empty, return the CDM's default empty object. No
   * need to go through the hash table...
   */
  if (elements == 0) return &CDM_Obj_NULL;

  /*
   * We don't want to store sets that contain attributes with type
   * CDM_STRING_SET...
   */
  /*
   * I changed my mind. We don't save more than a few kilobytes but its more
   * processing -> time...
   */
#if 0
  for (i=0; i<elements; i++) {
    newPtr = CDM_GetValueType(Element[i]);
    if (newPtr < 0) return NULL; /* This was not a valid element... */
    if (newPtr == CDM_STRING_SET) return Set;
  }
#endif
  
#ifdef CDM_AYTOCLEAN_COMPRESSEDMEMORYCACHE
  /*
   * After a ten thousand insertions in the hash table, check if we can free
   * anything...
   */
  if (AttrSetHash_NumberOfCalls >= 10000) {
    printf("Freeing CompressedMemoryCache (%d)...\n",
           AttrSetHash_NumberOfClears);
    printf("Freed %d\n", CDM_ClearUnusedAttributeSetHash());
  }
#endif /* CDM_AYTOCLEAN_COMPRESSEDMEMORYCACHE */
  
#ifdef CDM_OBJ_HASH_KEYS
  entryPtr = Tcl_CreateHashEntry(&AttrSetHash, (char *) Set, &newPtr);
#else  /* CDM_OBJ_HASH_KEYS */
  entryPtr = Tcl_CreateHashEntry(&AttrSetHash, (char *) Tcl_GetString(Set),
                                 &newPtr);
#endif /* CDM_OBJ_HASH_KEYS */
  if (newPtr) {
    /*
     * This is a new entry. The object was not in the cache...
     */
#ifdef CDM_AYTOCLEAN_COMPRESSEDMEMORYCACHE
    AttrSetHash_NumberOfCalls ++;
#endif /* CDM_AYTOCLEAN_COMPRESSEDMEMORYCACHE */
    Tcl_SetHashValue(entryPtr, (void *) Set);
#ifndef CDM_OBJ_HASH_KEYS
    Tcl_IncrRefCount(Set);
#endif /* CDM_OBJ_HASH_KEYS */
  } else {
    /*
     * This is an old entry. The object was in the cache...
     */
    Tcl_InvalidateStringRep(Set);
    Set = (CDM_AttributeSet) Tcl_GetHashValue(entryPtr);
  }

  /* 
   * Usually this argument is an Attribute Set, that possibly holds a list of
   * Attributes. The process of adding this set in the hash table may have
   * generated a string representation of this set object, that occupies
   * unnecessary memory. So, it's better to free this string representation,
   * in order to free the associated memory. But we have to take care the
   * special case where an empty set is passed: If we free the string
   * representation we may crash in Tcl_GetString. We must never invalidate
   * the string representation of an empty set object!
   */
  Tcl_InvalidateStringRep(Set);
  return Set;
} /* CDM_SearchForAttributeSet */

/*
 *----------------------------------------------------------------------
 *
 * AttrSetHash_CompareObjKeys --
 *
 *        Compares two Tcl_Obj * keys. The objects stored in the Hash table
 *        are assumed to be Attribute Sets...
 *
 * Results:
 *        The return value is 0 if they are different and 1 if they are
 *        the same.
 *
 * Side effects:
 *        None.
 *
 *----------------------------------------------------------------------
 */
static int AttrSetHash_CompareObjKeys(VOID *keyPtr, Tcl_HashEntry *hPtr) {
  Tcl_Obj *objPtr1 = (Tcl_Obj *) keyPtr;
  Tcl_Obj *objPtr2 = (Tcl_Obj *) hPtr->key.oneWordValue;
  register CONST char *p1, *p2;
  register int l1, l2;
  Tcl_Obj **elements1, **elements2;
  int result, items1, items2;

  /*
   * If the object pointers are the same then they match.
   */
  if (objPtr1 == objPtr2) {
      return 1;
  }

  /*
   * The objects that we receive are Attribute Set. First we compare the
   * number of elements...
   */
  result = Tcl_ListObjGetElements(NULL, objPtr1, &items1, &elements1);
  if (result != TCL_OK) goto obj_comparison;
  result = Tcl_ListObjGetElements(NULL, objPtr2, &items2, &elements2);
  if (result != TCL_OK) goto obj_comparison;
  if (items1 != items2) return 0;
  
obj_comparison:
  /*
   * Don't use Tcl_GetStringFromObj as it would prevent l1 and l2 being
   * in a register.
   */
  p1 = Tcl_GetString (objPtr1);
  l1 = objPtr1->length;
  p2 = Tcl_GetString (objPtr2);
  l2 = objPtr2->length;
  
  /*
   * Only compare if the string representations are of the same length.
   */
  if (l1 == l2) {
      for (;; p1++, p2++, l1--) {
          if (*p1 != *p2) {
              break;
          }
          if (l1 == 0) {
              /*
               * We don't need the string representation any more...
               */
              Tcl_InvalidateStringRep(objPtr1);
              Tcl_InvalidateStringRep(objPtr2);
              return 1;
          }
      }
  }

  Tcl_InvalidateStringRep(objPtr1);
  Tcl_InvalidateStringRep(objPtr2);
  return 0;
} /* AttrSetHash_CompareObjKeys */

/*
 * EOF - End of File
 */
