/*****************************************************************************
 *  GateIO.cpp:                                                              *
 * --------------------                                                      *
 *  This file implements various utilities used internally by the            *
 *  Collection & Document Manager to handle GATE's 1.x storage format.       *
 *                                                                           *
 *  Copyright 1998-2010 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licensing information, please refer to the license conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/

#include "CDM.h"
#include <string.h>
#ifndef WINDOWS
#include <strings.h>
#endif /* WINDOWS */
#include <ctype.h>
#include "CDM_GlobalObjects.h"
extern Tcl_ThreadDataKey dataKey;

/*
 * Prototypes...
 */
int         CDM_SaveGateCollection(CDM_Collection Collection, Tcl_Channel File);
int         CDM_LoadGateCollection(CDM_Collection Collection, Tcl_Channel File,
                                   char **Id, char **ExternalId);
int         CDM_SaveGateDocument(CDM_Document Document, Tcl_Channel File);
int         CDM_LoadGateDocument(CDM_Document NewDocument, Tcl_Channel File,
                                   char **Id, char **ExternalId);

char          *CDM_ReadGateString(Tcl_Channel File);
CDM_Annotation CDM_ReadGateAnnotation(Tcl_Channel File);
CDM_Attribute  CDM_ReadGateAttribute(Tcl_Channel File);
void  CDM_WriteGateString(Tcl_Channel File, const char *string);
void  CDM_WriteGateString(Tcl_Channel File, const char *string, int EmmitNewLine);
int   CDM_WriteGateAttribute(Tcl_Channel File, CDM_Attribute Attr);
int   CDM_WriteGateAnnotationSet(Tcl_Channel File, CDM_AnnotationSet Anns);
int   CDM_WriteGateAnnotationSet(Tcl_Channel File, CDM_AnnotationSet Anns, 
                                 Tcl_Obj *Win, int percent);

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     C O L L E C T I O N            S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 * -- CDM_SaveGateCollection:
 *    Saves given Collection in disk (in Gate's format).
 *    Arguments: 
 *      Collection: The Collection's that will be saved pointer (CDM_Collection)
 *      File:       File Object                                 (Tcl_Channel)
 */
int CDM_SaveGateCollection(CDM_Collection Collection, Tcl_Channel File) {
  CDM_ListNode *ptr;
  int listc;
  CDM_THREAD_SPECIFIC_DATA;

  if (Collection->FileFormat != CDM_FORMAT_GATE_1x &&
      Collection->FileFormat != CDM_FORMAT_GATE_2x)
    return CDM_ReportError("Collection in format other than gate[2]");

  Tcl_WriteChars(File, "GdmCollection GutString ", -1);
  snprintf(CDM_Temp, MAX_TEMP_STRING, "%zd ", strlen(Collection->Name));
  Tcl_WriteChars(File, CDM_Temp, -1);
  Tcl_WriteChars(File, Collection->Name, -1);
  Tcl_WriteChars(File, "\nGutSequence 0\n\nGutString 0 \nGutSequence ", -1);
  listc = CDM_ListLength(Collection->Documents);
  snprintf(CDM_Temp, MAX_TEMP_STRING, "%d\n", listc);
  Tcl_WriteChars(File, CDM_Temp, -1);
  ptr = Collection->Documents;
  while (ptr != NULL) {
    Tcl_WriteChars(File, "GdmCollection::Doc GutString ", -1);
    snprintf(CDM_Temp, MAX_TEMP_STRING, "%zd ",
                                        strlen(ptr->ptr[CDM_DOCUMENT_ID]));
    Tcl_WriteChars(File, CDM_Temp, -1);
    Tcl_WriteChars(File, ptr->ptr[CDM_DOCUMENT_ID], -1);
    Tcl_WriteChars(File, "\nGutString ", -1);
    snprintf(CDM_Temp, MAX_TEMP_STRING, "%zd ",
                                        strlen(ptr->ptr[CDM_DOCUMENT_XID]));
    Tcl_WriteChars(File, CDM_Temp, -1);
    Tcl_WriteChars(File, ptr->ptr[CDM_DOCUMENT_XID], -1);
    Tcl_WriteChars(File, "\n", -1);
    ptr = ptr->next;
  }

  /* Write Attributes */
  /* TODO: We don't actually write Collection Attributes?? */
  Tcl_WriteChars(File, "\n \n \n##CDM::Encoding GutString ", -1);
  snprintf(CDM_Temp, MAX_TEMP_STRING, "%zd ", strlen(Collection->Encoding));
  Tcl_WriteChars(File, CDM_Temp, -1);
  Tcl_WriteChars(File, Collection->Encoding, -1);
  return TCL_OK;
} /* CDM_SaveGateCollection */


/*
 * -- CDM_LoadGateCollection:
 *    Loads requested Collection from the disk (in Gate's Format).
 *    Arguments: 
 *      Path:  The Collection's Disk address                 (char *)
 */
int CDM_LoadGateCollection(CDM_Collection NewCollection, Tcl_Channel File) {
  CDM_ListNode *ptr, *lastNode;
  CDM_Attribute Attr=NULL;
  Tcl_Obj *LineObj, *Obj=NULL;
  char *Line, *value, *Encoding;
  int chars;
  long Attr_nu, File_nu, i;
  CDM_THREAD_SPECIFIC_DATA;
  
  /* Initialise Line Object */
  LineObj = Tcl_NewObj();
  if (LineObj == NULL) CDM_Panic("CDM_LoadGateCollection: Tcl_NewObj() "
                                 "returns NULL - out of memory?");
  Tcl_IncrRefCount(LineObj);
  
  NewCollection->Name = CDM_ReadGateString(File);
  if (NewCollection->Name == NULL) goto error;
  NewCollection->FileFormat = CDM_FORMAT_GATE_1x;
  Tcl_SetObjLength(LineObj, 0); chars = Tcl_GetsObj(File, LineObj);
  if (chars <= 0) goto error; Line = Tcl_GetString(LineObj);
  sscanf(Line, "GutSequence %ld", &Attr_nu);
  for (i=0; i<Attr_nu; i++) {
    Attr = CDM_ReadGateAttribute(File);
    if (Attr == NULL) goto error;
    CDM_PutAttribute(NewCollection, Attr);
    Attr = NULL;
  }

  Tcl_Free(CDM_ReadGateString(File));

  Tcl_SetObjLength(LineObj, 0); chars = Tcl_GetsObj(File, LineObj);
  if (chars <= 0) goto error; Line = Tcl_GetString(LineObj);
  sscanf(Line, "GutSequence %ld", &File_nu);
  
  if (File_nu > 0) {
    value = CDM_ReadGateString(File);
    if (value == NULL) goto error;
    ptr = CDM_NewListNode();
    ptr->ptr[CDM_DOCUMENT_ID] = value;
    value = CDM_ReadGateString(File);
    if (value == NULL) goto error;
    ptr->ptr[CDM_DOCUMENT_XID] = value;
    NewCollection->Documents = lastNode = CDM_AppendListNode(NULL, ptr);
    if (File_nu > 1) {
      for (i=1; i<File_nu; i++) {
        value = CDM_ReadGateString(File);
        if (value == NULL) goto error;
        ptr = CDM_NewListNode();
        ptr->ptr[CDM_DOCUMENT_ID] = value;
        value = CDM_ReadGateString(File);
        if (value == NULL) goto error;
        ptr->ptr[CDM_DOCUMENT_XID] = value;
        lastNode = CDM_AppendListNode(lastNode, ptr);
      }
    } /* if (File_nu > 1) */
  } /* if (File_nu > 0) */

  /* Get Encoding (if it exists!) */
  Encoding = CDM_ReadGateString(File);
  if (Encoding == NULL) {
    NewCollection->Encoding = CDM_Strdup(CDM_DefaultEncoding);
  } else {
    NewCollection->Encoding = Encoding;
  }
  Tcl_DecrRefCount(LineObj);
  return TCL_OK;
error:
  if (Attr != NULL) Tcl_DecrRefCount(Attr);
  if (Obj  != NULL) Tcl_DecrRefCount(Obj);
  Tcl_DecrRefCount(LineObj);
  return CDM_ReportError("Invalid Collection Specification (Gate Format)");
}


/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |  E N D   O F    C O L L E C T I O N    S E C T I O N  |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/


/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     D O C U M E N T                S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/


/*
 * -- CDM_SaveGateDocument:
 *    Saves given Document in disk (in Gate's Format).
 *    Arguments: 
 *      Document:  The Document's that will be saved pointer (CDM_Document)
 *      File:      File Object                               (Tcl_Channel)
 */
int CDM_SaveGateDocument(CDM_Document Document, Tcl_Channel File) {
  Tcl_Interp *interp = Document->Interp;
  CDM_AttributeSet  Attrs;
  CDM_AnnotationSet Anns;
  Tcl_Obj **listv, *objv[4], *Win=NULL, *Message=NULL;
  int i, status, listc;
  CDM_THREAD_SPECIFIC_DATA;

  if (Document->FileFormat != CDM_FORMAT_GATE_1x &&
      Document->FileFormat != CDM_FORMAT_GATE_2x)
    return CDM_ReportError("Document in format other than GATE 1.x/2.x");

  if (Tcl_SetChannelOption(interp, File, "-encoding", Document->Encoding)
     != TCL_OK) {
     snprintf(CDM_Temp, MAX_TEMP_STRING, "cannot change encoding to \"%s\"",
                        Document->Encoding);
     return CDM_ReportError(CDM_Temp);
  }

  /* Try to Display Progress Dialog... */
  Message = Tcl_GetVar2Ex(CDM_Interp, "GUI_Text", "SavingGateDocument",
                             TCL_GLOBAL_ONLY);
  if (Message == NULL) 
    Message=CDM_NewStringObj("Saving Document (GATE Format).\nPlease Wait...");
  objv[0] = CDM_NewStringObj("ggi_wait_append");
  objv[1] = Message;
  objv[2] = CDM_NewLongObj(0);
  for(i=0; i<3; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(CDM_Interp, 3, objv, TCL_EVAL_GLOBAL);
  for(i=0; i<3; i++) Tcl_DecrRefCount(objv[i]);
  if (status == TCL_OK) {
    Win = Tcl_GetObjResult(CDM_Interp);
    Tcl_IncrRefCount(Win);
  }
  /* Write Document Attributes */
  Attrs = CDM_GetAttributes(Document);
  if (Attrs == NULL) return CDM_ReportError("Cannot Find Document Attributes");
  status = Tcl_ListObjGetElements(interp, Attrs, &listc, &listv);
  if (status != TCL_OK) return status;
  snprintf(CDM_Temp, MAX_TEMP_STRING, "GdmDocument GutSequence %d\n", listc);
  Tcl_WriteChars(File, CDM_Temp, -1);
  for (i=0; i<listc; i++) {
#ifdef DEBUGGATEIO
    CDM_Debug("Writting Document Attribute:\"%s\"\n", Tcl_GetString(listv[i]));
#endif /* DEBUGGATEIO */
    CDM_WriteGateAttribute(File, listv[i]);
#ifdef DEBUGGATEIO
    CDM_Debug("  --> Attribute written ok...\n");
#endif /* DEBUGGATEIO */
  }
  Tcl_WriteChars(File, "\n", 1);
  Tcl_DecrRefCount(Attrs);

  /* Update Progress... */
  if (Win != NULL) {
    objv[0] = CDM_NewStringObj("ggi_wait_update");
    objv[1] = &CDM_Obj_NULL;
    objv[2] = CDM_NewLongObj(10);
    objv[3] = Win;
    for(i=0; i<4; i++) Tcl_IncrRefCount(objv[i]);
    Tcl_EvalObjv(CDM_Interp, 4, objv, TCL_EVAL_GLOBAL);
    for(i=0; i<4; i++) Tcl_DecrRefCount(objv[i]);
    Tcl_DoOneEvent(TCL_IDLE_EVENTS|TCL_WINDOW_EVENTS|TCL_DONT_WAIT);
  }

  /* Write Document ID's */
  CDM_WriteGateString(File, Document->DocumentPtr->ptr[CDM_DOCUMENT_ID]);
  CDM_WriteGateString(File, Document->DocumentPtr->ptr[CDM_DOCUMENT_XID]);

  /* Write Annotations */
  Anns = CDM_GetAnnotations(Document);
  if (Anns==NULL) return CDM_ReportError("Cannot Find Document's Annotations");
#ifdef DEBUGGATEIO
  CDM_Debug("====> Saving Document's Annotations...\n");
#endif /* DEBUGGATEIO */
  CDM_WriteGateAnnotationSet(File, Anns, Win, 10);
#ifdef DEBUGGATEIO
  CDM_Debug("====> Saving Document's Annotations Done...\n");
#endif /* DEBUGGATEIO */
  Tcl_DecrRefCount(Anns);
  snprintf(CDM_Temp, MAX_TEMP_STRING,"%ld\nCOPIED ",Document->AnnotationMaxID);
  Tcl_WriteChars(File, CDM_Temp, -1);

  /* Write Text... */
  CDM_WriteGateString(File, Tcl_GetString(Document->RawData));
  
  if (Win != NULL) {
    objv[0] = CDM_NewStringObj("ggi_destroy_append");
    objv[1] = Win;
    for(i=0; i<1; i++) Tcl_IncrRefCount(objv[i]);
    Tcl_EvalObjv(CDM_Interp, 2, objv, TCL_EVAL_GLOBAL);
    for(i=0; i<2; i++) Tcl_DecrRefCount(objv[i]);
    Win = NULL;
  }
  return TCL_OK;
} /* CDM_SaveGateDocument */


/*
 * -- CDM_LoadGateDocument:
 *    Loads requested Document from the disk (Convert form Gate Format).
 *    Arguments: 
 *      NewDocument:  The New Document's Pointer       (CDM_Document)
 *      File:         File Object                      (Tcl_Channel)
 */
int CDM_LoadGateDocument(CDM_Document NewDocument, Tcl_Channel File,
                         char **Id, char **ExternalId) {
  register CDM_Collection Parent;
  register CDM_Annotation Ann=NULL;
  register CDM_Attribute  Attr=NULL;
  register Tcl_Interp *interp = NewDocument->Interp;
  register Tcl_Obj *LineObj, *objv[4], *Win=NULL, *Message=NULL;
  register char *Encoding, *Line;
  double progress=0.0, step=0.0;
  register int chars, i, j, items, status, last_progress;
  CDM_THREAD_SPECIFIC_DATA;

  if (NewDocument->FileFormat == CDM_FORMAT_GATE_1x &&
      NewDocument->FileFormat != CDM_FORMAT_GATE_2x) 
    return CDM_ReportError("Document in format other than gate[2]");

  /* Initialise Line Object */
  LineObj = Tcl_NewObj();
  if (LineObj == NULL) 
  CDM_Panic("CDM_LoadGateDocument: Tcl_NewObj() returns NULL - out of memory?");
  Tcl_IncrRefCount(LineObj);

  /* Change File's encoding... */
  Parent = (CDM_Collection) (NewDocument->ParentPtr);
  NewDocument->Encoding = Encoding = CDM_Strdup(Parent->Encoding);
  if (Tcl_SetChannelOption(interp, File, "-encoding", Encoding) != TCL_OK) {
     snprintf(CDM_Temp, MAX_TEMP_STRING, "cannot change encoding to \"%s\"",
                    Encoding);
     CDM_ReportError(CDM_Temp);
     goto error;
  }

  /* Get First Line form File and Check if it is a valid Gate Document...*/
  Tcl_SetObjLength(LineObj, 0); chars = Tcl_GetsObj(File, LineObj);
  if (chars <= 0) {
    CDM_ReportError("CDM_LoadGateDocument: Empty Document!"); goto error;}
  Line = Tcl_GetString(LineObj);
  if (strncmp(Line, "GdmDocument", 11) != 0) {
    CDM_ReportError("CDM_LoadGateDocument: Document not in GATE format!");
    goto error;
  }
  NewDocument->FileFormat = CDM_FORMAT_GATE_1x;

  /* Get Number of Attributes... */
  if (sscanf(Line, "GdmDocument GutSequence %d", &items) < 1) {
    CDM_ReportError("Cannot Read Number of Attributes"); goto error; }
  /* Get Attributes... */
  for (i=0; i<items; i++) {
    Attr = CDM_ReadGateAttribute(File);
    if (Attr == NULL) {
      CDM_ReportError("CDM_LoadGateDocument: Error Reading Attributes");
      goto error;
    }
    CDM_PutAttribute(NewDocument, Attr);
  }

  /*
   * Read Document Id...
   */
  *Id = CDM_ReadGateString(File);
  if (*Id == NULL) {
    CDM_ReportError("CDM_LoadGateDocument: Cannot Read Document's ID!");
    goto error;
  }
  
  /*
   * Read Document External Id...
   */
  *ExternalId = CDM_ReadGateString(File);
  if (*ExternalId == NULL) {
   CDM_ReportError("CDM_LoadGateDocument: Cannot Read Document's External ID!");
   goto error;
  }

  Tcl_SetObjLength(LineObj, 0); chars = Tcl_GetsObj(File, LineObj);
  if (chars <= 0) {
    CDM_ReportError("CDM_LoadGateDocument: Premature EOF - Corrupted File?");
    goto error;
  }
  Line = Tcl_GetString(LineObj);
  if (strncmp(Line, "GdmAnnotationSet", 16) != 0) {
    CDM_ReportError("CDM_LoadGateDocument: Cannot Locate Annotation Set!");
    goto error;
  }
  
  /* Get Number of Annotations... */
  sscanf(&Line[29], "%d", &items);

  /* If there are quite many Annotations, Display Progress! */
  if (items > 100) {
    Message = Tcl_GetVar2Ex(CDM_Interp, "GUI_Text", "LoadingGateDocument",
                             TCL_GLOBAL_ONLY);
    if (Message == NULL) Message = 
      CDM_NewStringObj("Loading Document (GATE Format).\nPlease Wait...");
    objv[0] = CDM_NewStringObj("ggi_wait_append");
    objv[1] = Message;
    objv[2] = CDM_NewLongObj(0);
    for(j=0; j<3; j++) Tcl_IncrRefCount(objv[j]);
    status = Tcl_EvalObjv(CDM_Interp, 3, objv, TCL_EVAL_GLOBAL);
    for(j=0; j<3; j++) Tcl_DecrRefCount(objv[j]);
    if (status == TCL_OK) {
      Win = Tcl_GetObjResult(CDM_Interp);
      Tcl_IncrRefCount(Win);
    }
    step = 100.0/float(items);
  }
  last_progress = 0;
  /* Get Annotations... */
  for (i=0; i<items; i++) {
#ifdef DEBUGGATEIO
    CDM_Debug("Processing Annotation: %d\n", i);
#endif /* DEBUGGATEIO */
    Ann = CDM_ReadGateAnnotation(File);
    if (Ann == NULL) {
      snprintf(CDM_Temp, MAX_TEMP_STRING, "\nCDM_LoadGateDocument: Cannot Read "
        "Annotation!\nNumber of Correctly Processed Annotations: %d\n", i);
      CDM_AppendError(CDM_Temp);
      goto error;
    }
    CDM_AddAnnotation(NewDocument, Ann);
    progress += step;
    if (Win != NULL && (int(progress) - last_progress == 10)) {
      last_progress = int(progress);
      objv[0] = CDM_NewStringObj("ggi_wait_update");
      objv[1] = &CDM_Obj_NULL;
      objv[2] = CDM_NewLongObj(last_progress);
      objv[3] = Win;
      for(j=0; j<4; j++) Tcl_IncrRefCount(objv[j]);
      Tcl_EvalObjv(CDM_Interp, 4, objv, TCL_EVAL_GLOBAL);
      for(j=0; j<4; j++) Tcl_DecrRefCount(objv[j]);
      Tcl_DoOneEvent(TCL_IDLE_EVENTS|TCL_WINDOW_EVENTS|TCL_DONT_WAIT);
    }
  }
  if (Win != NULL) {
    objv[0] = CDM_NewStringObj("ggi_destroy_append");
    objv[1] = Win;
    for(i=0; i<1; i++) Tcl_IncrRefCount(objv[i]);
    Tcl_EvalObjv(CDM_Interp, 2, objv, TCL_EVAL_GLOBAL);
    for(i=0; i<2; i++) Tcl_DecrRefCount(objv[i]);
    Win = NULL;
  }

  /* Read Document's Text */
  Line = CDM_ReadGateString(File);
  if (Line == NULL) {
    CDM_ReportError("CDM_LoadGateDocument: Cannot Read Document's Text!");
    goto error;
  }
  NewDocument->RawData = Tcl_NewStringObj(Line, -1);
  Tcl_Free(Line);
  if (NewDocument->RawData == NULL) CDM_Panic("CDM_LoadGateDocument: "
      "Tcl_NewStringObj returns NULL - out of memory?");
  Tcl_IncrRefCount(NewDocument->RawData);
  NewDocument->Modified = 0;
  Tcl_DecrRefCount(LineObj);
  return TCL_OK;
error:
  Tcl_DecrRefCount(LineObj);
  LineObj = Tcl_GetObjResult(CDM_Interp);
  Tcl_IncrRefCount(LineObj);
  if (Win != NULL) {
    objv[0] = CDM_NewStringObj("ggi_destroy_append");
    objv[1] = Win;
    for(i=0; i<1; i++) Tcl_IncrRefCount(objv[i]);
    Tcl_EvalObjv(CDM_Interp, 2, objv, TCL_EVAL_GLOBAL);
    for(i=0; i<2; i++) Tcl_DecrRefCount(objv[i]);
  }
  Tcl_SetObjResult(CDM_Interp, LineObj);
  Tcl_DecrRefCount(LineObj);
  return TCL_ERROR;
} /* CDM_LoadGateDocument */


/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     U T I L I T I E S              S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

CDM_Annotation CDM_ReadGateAnnotation(Tcl_Channel File) {
  CDM_THREAD_SPECIFIC_DATA;
  CDM_AttributeSet Attrs=&CDM_Obj_NULL;
  CDM_SpanSet      Spans=NULL;
  CDM_Annotation   Ann=NULL;
  CDM_Attribute    Attr=NULL;
  CDM_Span         Span=NULL;
  Tcl_Obj *LineObj, *listv[4];
  const char *Line;
  char *name=NULL, *tmpLine;
  int chars, attr_nu, span_nu, i;
  long Id, start, end;

  LineObj = Tcl_NewObj();
  if (LineObj == NULL) CDM_Panic("CDM_ReadGateAnnotation: Tcl_NewObj() "
                                 "returns NULL - out of memory?");
  Tcl_IncrRefCount(LineObj);

  /* Skip white space */
#ifdef DEBUGGATEIO
  CDM_Debug("CDM_ReadGateAnnotation: skipping white space...\n");
#endif /* DEBUGGATEIO */
  Line = " ";
  while (isspace(*Line)) {
    chars = Tcl_ReadChars(File, LineObj, 1, 0);
    if (chars <= 0) {
      if (chars < 0) {
        CDM_ReportError("CDM_ReadGateAnnotation: Error Reading File!");
        goto error;
      }
      if (Tcl_Eof(File)) {
        CDM_ReportError("CDM_ReadGateAnnotation: Premature EOF!");
        goto error;
      }
    }
    Line = Tcl_GetString(LineObj);
  }
#ifdef DEBUGGATEIO
  CDM_Debug("CDM_ReadGateAnnotation: skipping white space DONE...\n");
#endif /* DEBUGGATEIO */

  /* Get Attribute Number */
  Tcl_SetObjLength(LineObj, 0); chars = Tcl_GetsObj(File, LineObj);
  if (chars <= 0) goto error; Line = Tcl_GetString(LineObj);
#ifdef DEBUGGATEIO
  CDM_Debug("Searching for Attribute number: '%s'\n", Line);
#endif /* DEBUGGATEIO */
  if (strncmp(Line, "dmAnnotation GutSequence ", 25) != 0) {
    CDM_ReportError("Malformed Document File: Annotation Descriptor:\n"
                    "Missing information about number of Attributes!\n");
    goto error;
  }
  
  /* Get Number of Attributes... */
  sscanf(Line, "dmAnnotation GutSequence %d", &attr_nu);
#ifdef DEBUGGATEIO
  CDM_Debug("Number of Attributes to Read: %d\n", attr_nu);
#endif /* DEBUGGATEIO */

  /* Get Attributes */
  if (attr_nu > 0) {
    Attrs = Tcl_NewListObj(0, NULL);
    if (Attrs == NULL) CDM_Panic("CDM_ReadGateAnnotation: out of memory");

    for (i=0; i<attr_nu; i++) {
      Attr = CDM_ReadGateAttribute(File);
      if (Attr == NULL) goto error;
      if (Tcl_ListObjAppendElement(NULL, Attrs, Attr) != TCL_OK) goto error;
    }
  }

  /* Get ID */
  tmpLine = CDM_ReadGateString(File); if (Line == NULL) goto error;
  sscanf(tmpLine, "%ld", &Id);
  Tcl_Free(tmpLine);

  /* Get Name */
  name = CDM_ReadGateString(File); if (name == NULL) goto error;

  /* Get Span Set */
  /* Get Span Number... */
  Tcl_SetObjLength(LineObj, 0); chars = Tcl_GetsObj(File, LineObj);
  if (chars <= 0) goto error; Line = Tcl_GetString(LineObj);
  if (strncmp(Line, "GutSequence ", 12) != 0) goto error; 
  sscanf(&Line[12], "%d", &span_nu);
  
  Spans = Tcl_NewListObj(0, NULL);
  if (Spans == NULL) CDM_Panic("CDM_ReadGateAnnotation: out of memory");
  
  for (i=0; i<span_nu; i++) {
    Tcl_SetObjLength(LineObj, 0); chars = Tcl_GetsObj(File, LineObj);
    if (chars <= 0) goto error; Line = Tcl_GetString(LineObj);
    if (sscanf(Line, "GdmSpan %ld %ld", &start, &end) == 0) {
      if (sscanf(Line, "GdmSpan {%ld} {%ld}", &start, &end) == 0) goto error;
    }
    Span = CDM_CreateSpan(start, end);
    if (Span == NULL) goto error;
    if (Tcl_ListObjAppendElement(NULL, Spans, Span) != TCL_OK) goto error;
  }
  
  listv[0] = CDM_NewLongObj(Id);
  listv[1] = CDM_NewStringObj(name);
  listv[2] = Spans;
  listv[3] = Attrs;
  Tcl_Free(name);
  Tcl_DecrRefCount(LineObj);
  Ann = Tcl_NewListObj(4, listv);
  if (Ann == NULL) CDM_Panic("CDM_ReadGateAnnotation: out of memory");
  return Ann;
error:
#ifdef DEBUGGATEIO
  CDM_Debug("CDM_ReadGateAnnotation: Error Trapped!\n");
#endif /* DEBUGGATEIO */
  Tcl_DecrRefCount(LineObj);
  if (Attrs != NULL) Tcl_DecrRefCount(Attrs);
  if (Spans != NULL) Tcl_DecrRefCount(Spans);
  if (Ann   != NULL) Tcl_DecrRefCount(Ann);
  if (Span  != NULL) Tcl_DecrRefCount(Span);
  if (name  != NULL) Tcl_Free(name);
  return NULL;
} /* CDM_ReadGateAnnotation */

int CDM_WriteGateAnnotationSet(Tcl_Channel File, CDM_AnnotationSet Anns) {
  return CDM_WriteGateAnnotationSet(File, Anns, NULL, 0);
}
int CDM_WriteGateAnnotationSet(Tcl_Channel File, CDM_AnnotationSet Anns,
                               Tcl_Obj *Win, int percent) {
  register CDM_AttributeSet Attrs;
  register CDM_SpanSet      Spans;
  register Tcl_Obj **Ann, **Attr, **Span, *objv[4], *Message=NULL;;
  double progress = 0.0, step = 0.0;
  register int i, j, status, listc, attrc, last_progress = percent;
  register long start, end;
  CDM_THREAD_SPECIFIC_DATA;

#ifdef DEBUGGATEIO
  CDM_Debug("Writting Document's Annotations...\n");
#endif /* DEBUGGATEIO */

  Tcl_WriteChars(File, "GdmAnnotationSet GutSequence ", 29);
  status = Tcl_ListObjGetElements(NULL, Anns, &listc, &Ann);
  if (status != TCL_OK) return TCL_ERROR;
  snprintf(CDM_Temp, MAX_TEMP_STRING, "%d\n", listc);
  Tcl_WriteChars(File, CDM_Temp, -1);

  if (Win != NULL && listc > 100) {
    progress = (float) percent;
    step = (90.0 - progress)/float(listc);
    Message = Tcl_GetVar2Ex(CDM_Interp, "GUI_Text", "SavingGateDocument",
                             TCL_GLOBAL_ONLY);
    if (Message == NULL) Message = 
      CDM_NewStringObj("Saving Document (GATE Format).\nPlease Wait...");
  }

  /* Write Annotations... */
  for (i=0; i<listc; i++) {
    progress += step;
    Attrs = CDM_GetAttributes(Ann[i]);
    Tcl_ListObjGetElements(NULL, Attrs, &attrc, &Attr);
    /* Write Attribute Number... */
    Tcl_WriteChars(File,"GdmAnnotation GutSequence ", 26);
    snprintf(CDM_Temp, MAX_TEMP_STRING, "%d\n", attrc);
    Tcl_WriteChars(File, CDM_Temp, -1);
    for (j=0; j<attrc; j++) {
      CDM_WriteGateAttribute(File, Attr[j]);
    }
    Tcl_WriteChars(File, "\n", 1);
 
    /* Write ID & Type */
    snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", CDM_GetId(Ann[i]));
    CDM_WriteGateString(File, CDM_Temp);
    CDM_WriteGateString(File, CDM_GetType(Ann[i]));

    /* Write SpanSet... */
    Spans = CDM_GetSpans(Ann[i]);
    Tcl_ListObjGetElements(NULL, Spans, &attrc, &Span);
    Tcl_WriteChars(File, "GutSequence ", -1);
    snprintf(CDM_Temp, MAX_TEMP_STRING, "%d\n", attrc);
    Tcl_WriteChars(File, CDM_Temp, -1);
    for (j=0; j<attrc; j++) {
      Tcl_WriteChars(File, "GdmSpan ", -1);
      start = CDM_GetStart(Span[j]); end = CDM_GetEnd(Span[j]);
      snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld %ld\n", start, end);
      Tcl_WriteChars(File, CDM_Temp, -1);
    }
    Tcl_WriteChars(File, "\n", 1);
    if (Win != NULL && (int(progress) - last_progress == 10)) {
      last_progress = int(progress);
      objv[0] = CDM_NewStringObj("ggi_wait_update");
      objv[1] = &CDM_Obj_NULL;
      objv[2] = CDM_NewLongObj(last_progress);
      objv[3] = Win;
      for(j=0; j<4; j++) Tcl_IncrRefCount(objv[j]);
      Tcl_EvalObjv(CDM_Interp, 4, objv, TCL_EVAL_GLOBAL);
      for(j=0; j<4; j++) Tcl_DecrRefCount(objv[j]);
      Tcl_DoOneEvent(TCL_IDLE_EVENTS|TCL_WINDOW_EVENTS|TCL_DONT_WAIT);
    }
  }
  Tcl_WriteChars(File, "\n", 1);
  return TCL_OK;
} /* CDM_WriteGateAnnotationSet */

CDM_Attribute CDM_ReadGateAttribute(Tcl_Channel File) {
  CDM_AttributeValue Val=NULL;
  CDM_Attribute Attr;
  Tcl_Obj *LineObj, *Set;
  const char *Line;
  char *name, *value;
  long Long;
  int chars, set_nu, i;

  LineObj = Tcl_NewObj();
  if (LineObj == NULL) CDM_Panic("CDM_ReadGateAttribute: Tcl_NewObj() "
                                 "returns NULL - out of memory?");
  Tcl_IncrRefCount(LineObj);

  /* Read Attribute Name */
  name = CDM_ReadGateString(File);
  if (name == NULL) goto error;

  /* Get Attribute Value */
  /* Skip white space */
  Line = " ";
  while (isspace(*Line)) {
    chars = Tcl_ReadChars(File, LineObj, 1, 0);
    if (chars <= 0) goto error; Line = Tcl_GetString(LineObj);
  }
  chars = Tcl_ReadChars(File, LineObj, 16, 1);
  if (chars <= 0) goto error; Line = Tcl_GetString(LineObj);
  if (strcmp(Line, "GdmAttributeValue") != 0) goto error;
  chars = Tcl_ReadChars(File, LineObj, 12, 0); 
  if (chars <= 0) goto error; Line = Tcl_GetString(LineObj);
  if (Line[11] == '_') {        /* GDM_STRING_SET */
    Tcl_SetObjLength(LineObj, 0); chars = Tcl_GetsObj(File, LineObj);
    if (chars <= 0) goto error; Line = Tcl_GetString(LineObj);
    if (sscanf(Line, "SET GutSequence %d", &set_nu) < 1) goto error;
    Set = Tcl_NewListObj(0, NULL);
    if (Set == NULL) CDM_Panic("CDM_ReadGateAttribute: Tcl_NewListObj() "
                                 "returns NULL - out of memory?");
    for (i=0; i<set_nu; i++) {
      value = CDM_ReadGateString(File);
      if (value == NULL) goto error;
      /* Is the item an integer? */
      if (sscanf(value, "%ld", &Long) > 0) Val = CDM_NewLongObj(Long);
      else                                 Val = CDM_NewStringObj(value);
      Tcl_Free(value);
      if (Tcl_ListObjAppendElement(NULL, Set, Val) != TCL_OK) goto error;
    }
    Val = CDM_CreateAttributeValue(CDM_STRING_SET, Set);
    if (Val == NULL) goto error;
  } else if (Line[11] == ' ') { /*GDM_STRING */
    value = CDM_ReadGateString(File);
    if (value == NULL) goto error;
    Val = CDM_CreateAttributeValue(CDM_STRING, value);
    Tcl_Free(value);
    if (Val == NULL) goto error;
  } else goto error;
  
  Tcl_DecrRefCount(LineObj);
  Attr = CDM_CreateAttribute(name, Val);
  Tcl_Free(name);
  return Attr;
error:
  CDM_ReportError("Cannot Process Attribute");
  Tcl_DecrRefCount(LineObj);
  if (Val != NULL) Tcl_DecrRefCount(Val);
  return NULL;
} /* CDM_ReadGateAttribute */

int CDM_WriteGateAttribute(Tcl_Channel File, CDM_Attribute Attr) {
  CDM_AttributeValue Val;
  Tcl_Obj **listv;
  char *name;
  int i, status, listc;
  CDM_THREAD_SPECIFIC_DATA;
  
#ifdef DEBUGGATEIO
  CDM_Debug("*** Writting Attribute \"%s\"\n", Tcl_GetString(Attr));
#endif /* DEBUGGATEIO */

  /* Write Attribute Name */
  name = CDM_GetName(Attr);
  if (name == NULL) return TCL_ERROR;
  Tcl_WriteChars(File, "GdmAttribute ", 13); 
#ifdef DEBUGGATEIO
  CDM_Debug("GdmAttribute ");
#endif /* DEBUGGATEIO */
  CDM_WriteGateString(File, name);
  Tcl_WriteChars(File, "GdmAttributeValue ", 18);
#ifdef DEBUGGATEIO
  CDM_Debug("GdmAttributeValue ");
#endif /* DEBUGGATEIO */

  /* Write Attribute Type */
  Val = CDM_GetValue(Attr);
  if (Val == NULL) return TCL_ERROR;
  status = Tcl_ListObjGetElements(NULL, Val, &listc, &listv);
  if (status != TCL_OK || listc != 2) return TCL_ERROR;
  name = Tcl_GetString(listv[0]);
  // CDM_WriteGateString(File, name, 0);
  Tcl_WriteChars(File, name, -1);
  Tcl_WriteChars(File, " ", 1);

  /* Write Attribute Value */
  if (strcmp(name, "GDM_STRING") == 0) {
    CDM_WriteGateString(File, Tcl_GetString(listv[1]));
  } else if (strcmp(name, "GDM_STRING_SET") == 0) {
    status = Tcl_ListObjGetElements(NULL, listv[1], &listc, &listv);
    if (status != TCL_OK) return TCL_ERROR;
    Tcl_WriteChars(File, "GutSequence ", 12);
#ifdef DEBUGGATEIO
    CDM_Debug("GutSequence ");
#endif /* DEBUGGATEIO */
    snprintf(CDM_Temp, MAX_TEMP_STRING, "%d\n", listc);
    Tcl_WriteChars(File, CDM_Temp, -1);
#ifdef DEBUGGATEIO
    CDM_Debug("%s", CDM_Temp);
#endif /* DEBUGGATEIO */
    for (i=0; i<listc; i++) {
      CDM_WriteGateString(File, Tcl_GetString(listv[i]));
    }
  } else if (strcmp(name, "CDM_IMAGE") == 0) {
    CDM_WriteGateString(File, "[binary data]");
  } else if (strcmp(name, "CDM_BASE64_IMAGE") == 0) {
    CDM_WriteGateString(File, Tcl_GetString(listv[1]));
  } else {
    return TCL_ERROR;
  }
  return TCL_OK;
} /* CDM_WriteGateAttribute */

char *CDM_ReadGateString(Tcl_Channel File) {
  Tcl_Obj *LineObj, *Obj;
  const char *Line;
  char *string;
  int chars;
  long length;

  LineObj = Tcl_NewObj(); Obj = Tcl_NewObj();
  if (LineObj == NULL || Obj == NULL) 
    CDM_Panic("CDM_ReadGateString: Tcl_NewObj() returns NULL - out of memory?");
  Tcl_IncrRefCount(LineObj); Tcl_IncrRefCount(Obj);

  /* Skip unwanted space... */
  while (1) {
    Line = " "; Tcl_SetObjLength(LineObj, 0);
    while (*Line != 'G') {
      chars = Tcl_ReadChars(File, LineObj, 1, 0 /* replace obj contents */);
      if (chars <= 0) goto error;
      Line = Tcl_GetString(LineObj);
    }
    /* We have found a "G", get the whole word to see if is "GutString"... */
    while (!isspace(*Line)) {
      chars = Tcl_ReadChars(File, Obj, 1, 0);
      if (chars <= 0) goto error;
      Line = Tcl_GetString(Obj);
      Tcl_AppendToObj(LineObj, Line, 1);
    }
    /* Have we found "GutString" ? */
    Line = Tcl_GetString(LineObj);
    if (strncmp(Line , "GutString", 9) == 0) break;
  }

  /* We are just to read the first digit character! */
  Line = ""; Tcl_SetObjLength(LineObj, 0);
  while (!isspace(*Line)) {
    chars = Tcl_ReadChars(File, Obj, 1, 0);
    if (chars <= 0) goto error;
    Line = Tcl_GetString(Obj);
    Tcl_AppendToObj(LineObj, Line, 1);
  }
  
  Line = Tcl_GetString(LineObj);
  /* Get the string length... */
  if (sscanf(Line, "%ld", &length) < 1) goto error;
  if (length > 0) {
    chars = Tcl_ReadChars(File, LineObj, length, 0);
    if (chars <= 0) goto error; Line = Tcl_GetString(LineObj);
#ifdef WINDOWS
    /* Under Windows, it is possible some file paths to be in windows stule.
     * (C:\etc...) So, wrap text in {} toavoid substitutions...
     */
      if (strchr(Line, '\\') != NULL) {
        Tcl_DString ds;

        Tcl_DStringInit(&ds);
        Tcl_DStringAppend(&ds, "{", 1);
        Tcl_DStringAppend(&ds, Line, length);
        Tcl_DStringAppend(&ds, "}", 1);
        string = CDM_Strdup(Tcl_DStringValue(&ds));
        Tcl_DStringFree(&ds);
      } else 
#endif /* WINDOWS */
      string = CDM_Strdup(Line);
  } else {
    string = CDM_Strdup("");
  }

  /* Finally, read the last newline! */
  chars = Tcl_ReadChars(File, Obj, 1, 0);
  Tcl_DecrRefCount(LineObj);
  Tcl_DecrRefCount(Obj);
  return string;
error:
  Tcl_DecrRefCount(LineObj);
  Tcl_DecrRefCount(Obj);
  return NULL;
} /* CDM_ReadGateString */

void CDM_WriteGateString(Tcl_Channel File, const char *string) {
  CDM_WriteGateString(File, string, 1);
}
void CDM_WriteGateString(Tcl_Channel File, const char *string, int EmmitNewLine) {
  char tmp[50];
  
  Tcl_WriteChars(File, "GutString ", 10);
#ifdef DEBUGGATEIO
    CDM_Debug("GutString ");
#endif /* DEBUGGATEIO */

  if (string == NULL || *string == '\0') {
    Tcl_WriteChars(File, "0 ", 2);
#ifdef DEBUGGATEIO
    CDM_Debug("0 \n");
#endif /* DEBUGGATEIO */
    if (EmmitNewLine) Tcl_WriteChars(File, "\n", 1);
    else              Tcl_WriteChars(File, " ", 1);
    return;
  }
  snprintf(tmp, 50, "%d ", Tcl_NumUtfChars(string, -1));
  Tcl_WriteChars(File, tmp, -1);
  Tcl_WriteChars(File, string, -1);
#ifdef DEBUGGATEIO
    CDM_Debug("%s%s\n", tmp, string);
#endif /* DEBUGGATEIO */
  if (EmmitNewLine) Tcl_WriteChars(File, "\n", 1);
  else              Tcl_WriteChars(File, " ", 1);
} /* CDM_WriteGateString */

/*
 * EOF -End Of File
 */
