##############################################################################
##  Java.tcl:                                                               ##
## ------------------------------                                           ##
##  This file implements various Java related procedures.                   ##
##                                                                          ##
##  Copyright 2002-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

package require XpUtils
namespace eval ::CDM_Java {
  variable JavaLoaded 0
  variable ClassPathSeparator \;
  variable ComponentClassPath {}
  variable TclBlendHome {}
  variable PerformLoadInDirectory {}
  variable JRE {}
}

## ::CDM_Java::appendJVMOptions
#  This function appends a set of options into the start-up options of the
#  JVM. Adding start-up options has no meaning after the JVM has been
#  initialised.
proc ::CDM_Java::appendJVMOptions {args} {
  foreach option $args {
    switch -glob -- $option {
      -*      { lappend ::tclblend_init [string range $option 1 end]}
      default { lappend ::tclblend_init $option }
    }
  }
};# ::CDM_Java::appendJVMOptions

## ::CDM_Java::appendPathsToClassPath
#  This function appends a set of paths into the CLASSPATH environmental
#  variable. Adding paths has no meaning after the JVM has been initialised.
proc ::CDM_Java::appendPathsToClassPath {args} {
  variable ComponentClassPath
  foreach path $args {
    XpUtils::appendpath ComponentClassPath $path
  }
};# ::CDM_Java::appendPathsToClassPath

## ::CDM_Java::appendPathToClassPath
#  This function appends a single path into the CLASSPATH environmental
#  variable. Adding paths has no meaning after the JVM has been initialised.
proc ::CDM_Java::appendPathToClassPath {path} {
  variable ComponentClassPath
  XpUtils::appendpath ComponentClassPath $path
};# ::CDM_Java::appendPathToClassPath

## ::CDM_Java::InitialiseJava
#  Use this function to create a JVM inside Ellogon.
proc ::CDM_Java::InitialiseJava {} {
  global HomeDir ApplicationInfo GUI_Options ApplicationName SplashPercent
  variable JavaLoaded
  variable TclBlendHome
  variable PerformLoadInDirectory
  variable JRE
  set TclBlendHome [file native [file normalize \
                     [glob -nocomplain -type d $HomeDir/lib/tcljava1.*]]]
  switch $::tcl_platform(platform) {
    windows {
      InitialiseJava_Windows
    }
    unix {
      InitialiseJava_Unix
    }
    default {
    }
  }
  ## Construct a proper CLASSPATH...
  GetComponentsClassPaths
  set CWD [pwd]
  catch {cd $PerformLoadInDirectory}
  if {[catch {package require java} error]} {
    ## The Java package failed to initialise. We will make another attempt
    ## before we report an error...
    if {[catch [list loadtclblend [file native $TclBlendHome]]]} {
      if {$ApplicationInfo(EnableGUI)} {
        catch {puts $::errorInfo}
        set VisScript [bind .splashscr <Visibility>]
        bind .splashscr <Visibility> {}
        lower .splashscr
        tk_messageBox -title Error... -type ok -default ok -icon error -message\
           "Cannot initialise Java Bindings:\n\
            The error was: $error" -parent .splashscr
        ## Maybe the user wants to disable Java support?
        set ans [gut_question "Do you want to disable Java support for the next\
          time $ApplicationName will start?" Question... 1 .splashscr]
        if {[string equal $ans yes]} {
          set GUI_Options(EnableJava) 0
        }
        unset ans
        bind .splashscr <Visibility> $VisScript
        raise .splashscr
        unset VisScript
        set GUI_Options(JavaRunTimePath) {}
        GUI_SaveOptions
      }
    } else {
      set JavaLoaded 1
    }
  } else {
    set JavaLoaded 1
  }
  if {$JavaLoaded} {
    ## The tcljava was successfully loaded. Load the CDM Java bindings (JCDM)
    LDM_SetSplashWindowStatus {Loading CDM Java bindings (JCDM)...} \
                                                           $SplashPercent 200
    ## We need to load the library from the directory that it is located
    ## (or the one above it) as we have set with -rpath ../../$(OS) & ../$(OS)
    ## for locating libCDM.so...
    cd $HomeDir/CDM/$ApplicationInfo(LoadBinaryFolderName)
    if {[catch {
      java::call System load [file join $HomeDir CDM Java \
                 $ApplicationInfo(LoadBinaryFolderName) libJavaCDM[info shared]]
    } error]} {
      set VisScript [bind .splashscr <Visibility>]
      bind .splashscr <Visibility> {}
      lower .splashscr
      tk_messageBox -title Error... -type ok -default ok -icon error -message \
         "Cannot initialise CDM Java Bindings (JCDM):\n\
          The error was: $error" -parent .splashscr
      bind .splashscr <Visibility> $VisScript
      raise .splashscr
      unset VisScript
      set JavaLoaded 0
      set GUI_Options(JavaRunTimePath) {}
      GUI_SaveOptions
    } else {
      ## Everything went fine: we have loaded both Java & JCDM...
      set GUI_Options(JavaRunTimePath) $JRE
      catch {LDM_SplashWindowNotify language java}
    }
    if {$GUI_Options(JavaLinkStreams)} {
      catch {
        set __JavaTmpObj [java::new \
             Ellogon.JCDM.OCDM_EllogonRegisterStdChannels [java::getinterp]]
        rename $__JavaTmpObj {}
      } error
      puts stderr $error
    }
  }
  cd $CWD
};# ::CDM_Java::InitialiseJava

## ::CDM_Java::InitialiseJava_Windows
#  Initialise Java under Windows
proc ::CDM_Java::InitialiseJava_Windows {} {
  package require registry
  global HomeDir ApplicationInfo GUI_Options
  variable TclBlendHome
  variable JRE
  variable ClassPathSeparator
  set ClassPathSeparator \;
  variable PerformLoadInDirectory
  set PerformLoadInDirectory \
      $TclBlendHome/$ApplicationInfo(LoadBinaryFolderName)

  set EnvironmentKeys {{HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Session Manager\Environment} {HKEY_CURRENT_USER\Environment}}

  ## =========================================================================
  ## Step 1: Try to locate an installed JRE/JDK from registry...
  ## =========================================================================
  ## First check if JRE is installed. If it is, we can run with it...

  if {[catch {
    set JRE [lindex [lsort -decreasing [registry keys \
      {HKEY_LOCAL_MACHINE\SOFTWARE\JavaSoft\Java Runtime Environment}]] 0]
    set JRE [registry get \
     "HKEY_LOCAL_MACHINE\\SOFTWARE\\JavaSoft\\Java Runtime Environment\\$JRE"\
      JavaHome]
  }]} {
    ## The JRE was not found. Perhaps we can locate a JDK?
    if {[catch {
       set JDK [lindex [lsort -decreasing [registry keys \
         {HKEY_LOCAL_MACHINE\SOFTWARE\JavaSoft\Java Development Kit}]] 0]
       set JDK [registry get \
        "HKEY_LOCAL_MACHINE\\SOFTWARE\\JavaSoft\\Java Development Kit\\$JDK"\
       JavaHome]
    }]} {
      ## Both JRE & JDK were not found. Java support must be disabled...
      if {$ApplicationInfo(EnableGUI)} {
        set VisScript [bind .splashscr <Visibility>]
        bind .splashscr <Visibility> {}
        lower .splashscr
        tk_messageBox -title Error... -type ok -default ok -icon error -message\
           "Cannot locate the Java JRE/JDK directory!" -parent .splashscr
        bind .splashscr <Visibility> $VisScript
        raise .splashscr
        unset VisScript
        return
      }
    } else {
      ## We have found the JDK!
      set JRE $JDK/jre
    }
  }

  ##
  ## Step 1 ==>> JRE location in var: JRE
  ##

  ## =========================================================================
  ## Step 2: Ensure the PATH environmental variable contains the needed paths
  ##         to locate the needed dlls...
  ## =========================================================================

  if {$::ApplicationInfo(EnableGUI) &&
      ![string equal $GUI_Options(JavaRunTimePath) $JRE]} {
    ## Either this is the first time the Java support was loaded, or a new
    ## JRE/JDK was installed!
    ## Make sure that the PATH contains some of the needed files, as well as
    ## the "."...
    set PathElements [split $::env(PATH) \
                            $ApplicationInfo(PathSplitCharacters)]
    set PathAddition {}
    set registry_updated    0
    set restart_application 0
    ## We could also add the paths to JCDM & tclblend, but is not really
    ## necessary, as we cd to the containing directory before loading...
    ## libJCDM.dll libtclblend.dll
    foreach file {libCDM.dll java.dll jvm.dll libtclblend*.dll} {
      set found 0
      foreach dir $PathElements {
        set files [glob -nocomplain -directory $dir $file]
        if {[llength $files]} {
          #gut_info "$file was found in $dir!"
          set found 1; break
        }
      }
      if {!$found} {
        switch -- $file {
          libCDM.dll {lappend PathAddition \
             [file native $HomeDir/CDM/$ApplicationInfo(LoadBinaryFolderName)]}
          libJCDM.dll {lappend PathAddition [file native \
             $HomeDir/CDM/Java/$ApplicationInfo(LoadBinaryFolderName)]}
          tclblend.dll {lappend PathAddition \
             [file native $TclBlendHome/$ApplicationInfo(LoadBinaryFolderName)]}
          java.dll {lappend PathAddition $JRE/bin}
          jvm.dll  {lappend PathAddition $JRE/bin/client}
        }
      }
    }
    if {[llength $PathAddition]} {
      set PathAddition [string map {\\ /} $PathAddition]
      set JRE [string map {\\ /} $JRE]
      set VisScript [bind .splashscr <Visibility>]
      bind .splashscr <Visibility> {}
      lower .splashscr
      set ans [gut_question "The needed paths to run Java do not seem to be\
        defined in your system.\n\nThe Java Runtime Environment has been found\
        in directory:\n -- $JRE\n\nThe missing directories from the\
        \"PATH\" environmental variable are:\
        \n -- [join $PathAddition \n\ --\ ]\n\nDo you want to add these\
        directories to your \"PATH\" environmental variable?\
        \n(Note that $::ApplicationName will be automatically\
        restarted after the update!)"]
      if {$ans == "yes"} {
        ## The update is done by changing the value of the 
        ## "HKEY_CURRENT_USER\Environment\Path" registry key...
        ## Changing for all user would require changing the key
        ## "HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\
        ## Session Manager\Environment\Path"...
        ## The idea is first to change the path for all users, and if that
        ## fails to change the path for individual users...
        foreach root $EnvironmentKeys {
          if {![catch {
            set PATH [registry get $root Path]
            append PATH \;[string map {/ \\} [join $PathAddition \;]]
            gut_info "Updating Registry Key \"$root\\Path\" with value:\
                      \n\"$PATH\"..."
            registry set $root Path $PATH
            set ::env(PATH) $PATH
            gut_info "Registry successfully updated.\n\nNow, $::ApplicationName\
              will be restarted. If Java again fails to load, please log-off\
              and log-in again or restart the operating system."
            set registry_updated 1
            set restart_application 1
          }]} {break}
        };# foreach root
      }
      bind .splashscr <Visibility> $VisScript
      raise .splashscr
    };# if {[llength $PathAddition]}

  ## =========================================================================
  ## Step 3: Ensure the CLASSPATH environmental variable contains the needed
  ##         paths to locate the needed JAR files...
  ## =========================================================================
    set PathAddition {}
    if {[info exists ::env(CLASSPATH)]} {
      set PathElements [split $::env(CLASSPATH) $ClassPathSeparator]
      ## Check paths like "." & $JRE/lib...
      foreach path [list . [file normalize $JRE/lib]] {
        if {[lsearch $PathElements $path] == -1} {lappend PathAddition $path}
      }
      ## Check whether our JAR files can be found...
      foreach file {tcljava.jar tclblend.jar JCDM.jar} {
        set found 0
        foreach dir $PathElements {
          if {[string match -nocase */$file [file normalize $dir]]} {
            #gut_info "$file was found in $dir!"
            set found 1; break
          } elseif {[file exists $dir/$file]} {
            #gut_info "$file was found in $dir!"
            #set found 1; break
          }
        }
        if {!$found} {
          switch -- $file {
            tcljava.jar  -
            tclblend.jar {
              lappend PathAddition [file normalize $TclBlendHome/$file]
            }
            JCDM.jar {
              lappend PathAddition [file normalize $HomeDir/CDM/Java/JCDM.jar]
            }
          }
        }
      }
    } else {
      ## ::env(CLASSPATH) does not exist!
      set PathAddition [list .]
      foreach path [list $JRE/lib \
                         $TclBlendHome/tcljava.jar $TclBlendHome/tclblend.jar \
                         $HomeDir/CDM/Java/JCDM.jar] {
        lappend PathAddition [file normalize $path]
      }
    };# if {[info exists ::env(CLASSPATH)]}
    if {[llength $PathAddition]} {
      set VisScript [bind .splashscr <Visibility>]
      bind .splashscr <Visibility> {}
      lower .splashscr
      set ans [gut_question "The \"CLASSPATH\" environmental variable needs to\
        be updated with the following paths:\
        \n -- [join $PathAddition \n\ --\ ]\n\nDo you want to add these\
        directories to your \"CLASSPATH\" environmental variable?\
        \n(Note that $::ApplicationName will be automatically\
        restarted after the update!)"]
      if {$ans == "yes"} {
        ## The update is done by changing the value of the 
        ## "HKEY_CURRENT_USER\Environment\Path" registry key...
        ## Changing for all user would require changing the key
        ## "HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\
        ## Session Manager\Environment\Path"...
        ## The idea is first to change the path for all users, and if that
        ## fails to change the path for individual users...
        foreach root $EnvironmentKeys {
          if {![catch {
            if {[catch {registry get $root CLASSPATH} PATH]} {
              set PATH {}
            }
            if {[string length $PATH]} {append PATH \;}
            append PATH [join $PathAddition \;]
            gut_info "Updating Registry Key \"$root\\CLASSPATH\" with value:\
                      \n\"$PATH\"..."
            registry set $root CLASSPATH $PATH
            set ::env(CLASSPATH) $PATH
            gut_info "Registry successfully updated.\n\nNow, $::ApplicationName\
              will be restarted. If Java again fails to load, please log-off\
              and log-in again or restart the operating system."
            set registry_updated 1
            set restart_application 1
          }]} {break}
        };# foreach root
      }
      bind .splashscr <Visibility> $VisScript
      raise .splashscr
    }

    if {$registry_updated} {
      ## Try to force Windows Explorer to re-initialise environmental
      ## variables...
      registry broadcast "Environment"
    }
    if {$restart_application} {
      GUI_SaveOptions
      exec [info nameofexecutable] $HomeDir/ellogon &
      exit
    }
  };# if $::ApplicationInfo(EnableGUI) && ...
};# ::CDM_Java::InitialiseJava_Windows

## ::CDM_Java::InitialiseJava_Unix
#  Initialise Java under Unix
proc ::CDM_Java::InitialiseJava_Unix {} {
  global HomeDir ApplicationInfo
  variable TclBlendHome
  variable JRE
  variable ClassPathSeparator
  variable PerformLoadInDirectory
  set ClassPathSeparator \:
  set PerformLoadInDirectory \
      $TclBlendHome/$ApplicationInfo(LoadBinaryFolderName)
  ## Try to locate JDK...
  if {[catch {exec which java} JavaDir]} {
    if {$ApplicationInfo(EnableGUI)} {
      set VisScript [bind .splashscr <Visibility>]
      bind .splashscr <Visibility> {}
      lower .splashscr
      tk_messageBox -title Error... -type ok -default ok -icon error -message\
         "Cannot locate the Java JRE/JDK directory:\n\
          Executing \"which java\" returned: $JavaDir" -parent .splashscr
      bind .splashscr <Visibility> $VisScript
      raise .splashscr
      unset VisScript
      return
    }
  }
  set JDK [file dirname [file dirname [file normalize $JavaDir]]]
  set folder $ApplicationInfo(LoadBinaryFolderName)
  append ::env(PATH) ":$TclBlendHome/$folder:$TclBlendHome:$HomeDir/CDM/$folder:$HomeDir/CDM"
  append ::env(LD_LIBRARY_PATH) ":$JDK/jre/lib/i386:$JDK/jre/lib/i386/client:$TclBlendHome/$folder:$TclBlendHome:$JDK/jre/lib/i386/native_threads:$HomeDir/CDM/$folder"
  append ::env(CLASSPATH) ":$JDK/jre/lib/rt.jar:$TclBlendHome:$HomeDir/CDM/Java/JCDM.jar"
  set JRE $JDK/jre
};# ::CDM_Java::InitialiseJava_Unix

## ::CDM_Java::LoadJavaComponent
proc ::CDM_Java::LoadJavaComponent {classpath module_name} {
  if {!$::CDM_Java::JavaLoaded} {error "java bindings disabled"}
  set compClassPath $classpath
  ## Retrieve additional directories to be added to the current classpath. A
  ## Component can store additional paths in the "classplath" entry in
  ## creole_config.tcl
  if {![catch {PUL_plugin_module_config $module_name classpath} \
          additional_class_path]} {
    ## The separator among different paths can be either ";" or ":"...
    if {[llength $additional_class_path]} {
      variable ClassPathSeparator
      set paths [split $additional_class_path \
                       $::ApplicationInfo(PathSplitCharacters)]
      ## TclBlend seems to want path in Unix notation...
      foreach path $paths {
        if {[catch {string map {\\ /} [uplevel #0 [list subst $path]]} path]} {
          puts "  Error: invalid value found in classpath:\n  $path"
          continue
        }
        lappend compClassPath $path
      }
      puts "  ++ CLASSPATH: \"$compClassPath\""
    }
  }
  puts "java::load -classpath $compClassPath $module_name"
  java::load -classpath $compClassPath $module_name
  ## Set the creole_${module_name}_home field of the component Java classes...
  puts "java::call $module_name SetComponentHome [file normalize $classpath]"
  catch {java::call $module_name SetComponentHome [file normalize $classpath]}
  catch {java::load -classpath $compClassPath \
                     Creole_${module_name}_Initialize}
  catch {java::load -classpath $compClassPath Creole_${module_name}_Finish}
};# ::CDM_Java::LoadJavaComponent

## ::CDM_Java::GetComponentsClassPaths
#  This function will examine all components in all directories of
#  PluginModulePath, to locate java components. For each java component, collect
#  the additional classpaths declared by the component...
proc ::CDM_Java::GetComponentsClassPaths {} {
  variable ComponentClassPath
  ## Get the CLASSPATH environmental variable, and initialise our
  ## ComponentClassPath variable.
  if {[info exists ::env(CLASSPATH)]} {
    set ComponentClassPath $::env(CLASSPATH)
  } else {
    set ComponentClassPath {}
  }

  ## Locate all component configuration files...
  foreach dir $::PluginModulePath {
    foreach config [glob -nocomplain $dir/*/creole_config.tcl] {
      set module_name [file tail [file dirname $config]]
      set ::creole_${module_name}_home [file dirname $config]
      catch {source $config}
    }
  }
  ## Load the located configuration files...
  foreach component [array names creole_config] {
    array unset key *
    array set key $creole_config($component)
    if {[string equal $key(coupling) java]} {
      if {[info exists key(classpath)] && [string length $key(classpath)]} {
        foreach path [split $key(classpath) \;] {
          set path [uplevel #0 subst $path]
          XpUtils::appendpath ComponentClassPath $path
        }
      }
    }
  }
  puts "Updating CLASSPATH: '$ComponentClassPath'"
  set ::env(CLASSPATH) $ComponentClassPath
};# ::CDM_Java::GetComponentsClassPaths
