/*****************************************************************************
 *  OCDM.cpp:                                                                *
 * ---------------------------------                                         *
 *  This file contains the implementation of all CDM Object Oriented         *
 *  Public API.                                                              *
 *  CDM is the Collection and Document Manager and the core component of the *
 *  Ellogon Text Engineering Platform.                                       *
 *  CDM was designed and implemented by Georgios Petasis,                    *
 *  e-mail: petasis@iit.demokritos.gr                                        *
 *  URL:    http://www.iit.demokritos.gr/~petasis                            *
 *                                                                           *
 *  Copyright 1998-2010 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licencing information, please refer to the licence conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/

#include "OCDM.h"

/*! \mainpage
 *
 * \section intro Introduction
 *
 * This documentation section describes the object-oriented interface provided
 * by the Ellogon Text Engineering Platform. Special care has been taken so as
 * the provided object oriented API to be as consistent as possible across
 * different programming languages (just as the procedural oriented API).
 *
 * Currently, three object oriented languages are supported: C++, Java and
 * Perl.
 * Since the object oriented API is almost identical for the three languages,
 * the same documentation can be used for programming in either of these 
 * languages.
 *
 * The only difference between the langauges is the inability of Java to
 * support operator overloading. As a result, the C++ overloaded operator \c []
 * (that retrieves an element of a set) has been converted to the method
 * \c get.
 */

/**
 * \file
 * \brief This file contains the implementation of the object oriented CDM
 * Public API (OCDM).
 *
 * This file contains the implementation of the object oriented CDM
 * Public API (OCDM).
 * CDM is the Collection and Document Manager and the core component of the
 * \b "Ellogon" Text Engineering Platform.
 * 
 * Copyright 1998-2004 by:                                             
 * Software and Knowledge Engineering Laboratory,                      
 * Institute of Informatics and Telecommunications,                    
 * National Center for Scientific Research (N.C.S.R.) "Demokritos",    
 * Athens, Greece.                                                     
 *                                                                     
 * This file is part of the \b "Ellogon" Text Engineering Platform.       
 * For licencing information, please refer to the licence conditions of
 * \b "Ellogon" Text Engineering Platform.
 */

/******************************************************************************
 ******************************************************************************
 *******   Helper Functions for CDM -> OCDM mapping                     *******
 ******************************************************************************
 *****************************************************************************/
template<class Type, class ParentObj, class Pointer>
const Type  OCDM_GetReference(const ParentObj& This, const Pointer obj) {
  if (obj == NULL) throw OCDM_Exception();
  /* Under SWIG generate and return a new object */
  Type *tmp = new Type(obj);
  This.storeObject(tmp);
  return *tmp;
}; /* OCDM_GetReference */

template<class ResultType, class ObjectType>
const ResultType OCDM_GetNth(const ObjectType& obj, const long index) {
  obj.ensureValidity();
  CDM_Object elementObj = CDM_Nth(obj.getObject(), index);
  if (elementObj == NULL) throw OCDM_Exception();
  // int objc;
  // CDM_Object *objv;
  // CDM_Object elementObj;
  // CDMTCL_ListObjGetElements(obj.getObject(), &objc, &objv);
  // if (index >= objc) throw OCDM_Exception("invalid index");
  // elementObj = objv[index];
  // CDM_Log("OCDM_GetNth -> (%p) [%d] %s\n", elementObj, elementObj->refCount, CDMTCL_GetString(elementObj));
  return OCDM_GetReference<ResultType>(obj, elementObj);
}; /* OCDM_GetNth */

CDM_Object OCDM_STL_ContainerToCDM_Object(const std::vector<std::string>
                                           &stl_obj) {
  CDM_Object tcl_obj = CDMTCL_NewListObj(0, NULL), tmp;
  if (tcl_obj == NULL) return NULL;
  for (std::vector<std::string>::const_iterator i  = stl_obj.begin();
                                                i != stl_obj.end(); ++i) {
    tmp = CDMTCL_NewStringObj(i->c_str(), -1);
    if (tmp) {
      CDMTCL_ListObjAppendElement(NULL, tcl_obj, tmp);
    }
  }
  return tcl_obj;
}; /* OCDM_STL_ContainerToCDM_Object */

int OCDM_AppendElement(OCDM_Object &obj, CDM_Object element) {
  if (element == NULL) return TCL_ERROR;
  CDM_Object tobj = (CDM_Object) obj.getObject();
  if (tobj == NULL) return TCL_ERROR;
  return CDMTCL_ListObjAppendElement(CDM_GetActiveInterpreter(), tobj, element);
}; /* OCDM_AppendElement */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_Object                                          *******
 ******************************************************************************
 *****************************************************************************/

/*
 * ---------------------------------------------------------------------------
 *                       C o n s t r u c t o r s 
 * ---------------------------------------------------------------------------
 */

/**
 * \brief This is the empty constructor.
 * \par Description:
 * This is the empty constructor. It creates an object that is empty.
 */
OCDM_Object::OCDM_Object() {
  cdmObj    = NULL;
  storedObj = NULL;
#ifdef LOG_CONSTRUCTOR_CALLS
  className = "OCDM_Object";
  //CDM_Log("OCDM_Object.cdmObj = (%p); ", cdmObj);
  //CDM_Log("Reference Count = %d\n", (cdmObj == NULL)? -100 : cdmObj->refCount);
  //CDM_Log("cdmObj.cstr = \"%s\"\n", CDMTCL_GetString(cdmObj));
#endif /* LOG_CONSTRUCTOR_CALLS */
}; /* OCDM_Object::OCDM_Object */

/**
* \brief This is a constructor that creates an OCDM object from a CDM object.
* \par Description:
* This is a constructor that creates an OCDM object from a CDM object.
* \par Arguments:
* \arg \c obj: the object to be reflected to the object oriented (OCDM) API.
*/
OCDM_Object::OCDM_Object(const CDM_Object obj) {
  storedObj = NULL;
  cdmObj    = (CDM_Object) obj;
  if (cdmObj != NULL) CDMTCL_IncrRefCount(cdmObj);
#ifdef LOG_CONSTRUCTOR_CALLS
  className = "OCDM_Object";
  //CDM_Log("OCDM_Object.cdmObj = (%p); ", cdmObj);
  //CDM_Log("Reference Count = %d\n", (cdmObj == NULL)? -100 : cdmObj->refCount);
  //CDM_Log("cdmObj.cstr = \"%s\"\n", CDMTCL_GetString(cdmObj));
#endif /* LOG_CONSTRUCTOR_CALLS */
}; /* OCDM_Object::OCDM_Object */

/**
* \brief  This is the default copy constructor.
* \par Description:
* This is the default copy constructor.
* \par Arguments:
* \arg \c obj: the object to be copied.
*/
OCDM_Object::OCDM_Object(const class OCDM_Object& obj) {
  storedObj = NULL;
  cdmObj    = obj.cdmObj;
  if (cdmObj != NULL) CDMTCL_IncrRefCount(cdmObj);
#ifdef LOG_CONSTRUCTOR_CALLS
  className = "OCDM_Object";
  //CDM_Log("OCDM_Object.cdmObj = (%p); ", cdmObj);
  //CDM_Log("Reference Count = %d\n", (cdmObj == NULL)? -100 : cdmObj->refCount);
  //CDM_Log("cdmObj.cstr = \"%s\"\n", CDMTCL_GetString(cdmObj));
#endif /* LOG_CONSTRUCTOR_CALLS */
}; /* OCDM_Object::OCDM_Object */

/*
 * ---------------------------------------------------------------------------
 *                         D e s t r u c t o r s 
 * ---------------------------------------------------------------------------
 */

/**
 * \brief This is the default destructor.
 * \par Description: * This is the default destructor.
 */
OCDM_Object::~OCDM_Object() {
  deleteObject();
}; /* OCDM_Object::~OCDM_Object */

/*
 * ---------------------------------------------------------------------------
 *                           O p e r a t o r s 
 * ---------------------------------------------------------------------------
 */

/**
 * \brief This is the default assignment operator.
 * \par Description: This is the default sssignment operator. Note that under
 * languages that do not support operator overloading (i.e. Java, Perl,
 * Python) this operator appears as a plain method named as "set".
 * \par Arguments:
 * \arg \c obj: the object to be copied.
 * \return This method returns a reference to the current object.
 */

class OCDM_Object& OCDM_Object::operator=  (const class OCDM_Object& obj) {
  if (obj.cdmObj != NULL) CDMTCL_IncrRefCount(obj.cdmObj);
  releaseObject();
  cdmObj = obj.cdmObj;
  return *this;
}; /* OCDM_Object::operator= */

#ifdef ENABLE_INDEX_OPERATOR_ON_ALL_OBJECTS
/**
* \brief This operator returns the element of a set pointed by the "index"
* parameter.
* \par Description: This operator returns the element of a set pointed by the
* "index" parameter. Note that under
* languages that do not support operator overloading (i.e. Java, Perl,
* Python) this operator appears as a plain method named as "get".
* \par Arguments:
* \arg \c index: the index (position) of the element to be retrieved. The
* first element of the set has as index the value 0.
* \note In case of an error, an exception of type OCDM_Exception
* will be thrown. 
*/

OCDM_INDEX_OPERATOR_REF(OCDM_Object) OCDM_Object::operator[] (const long index)
                                                                         const {
  return OCDM_GetNth<OCDM_Object>(*this, index);
}; /* OCDM_Object::operator[] */
#endif /* ENABLE_INDEX_OPERATOR_ON_ALL_OBJECTS */

/*
 * ---------------------------------------------------------------------------
 *              P r i v a t e   H e l p e r   f u n c t i o n s 
 * ---------------------------------------------------------------------------
 */

void OCDM_Object::initObject(const CDM_Object obj) const {
  storedObj = NULL; cdmObj = NULL;
  setObject(obj);
}; /* OCDM_Object::initObject */

void OCDM_Object::setObject(const CDM_Object obj) const {
#ifdef LOG_CDM_OBJECT
  CDM_Log(" %s.setObject: this: %p, obj: %p, count: %d\n",
          this, obj, (obj == NULL)? -100 : obj->refCount);
#endif /* LOG_CDM_OBJECT */
  /* First increment the reference of the object to be finally kept and then
   * release the older stored one. This will allow us to even replace an
   * object with itsself in a safe manner... */
  if (obj != NULL) CDMTCL_IncrRefCount((CDM_Object) obj);
  releaseObject();
  cdmObj = (CDM_Object) obj;
}; /* OCDM_Object::setObject */

const CDM_Object OCDM_Object::getObject(void) const {
  return (const CDM_Object) cdmObj;
}; /* OCDM_Object::getObject */

void  OCDM_Object::releaseObject(void) const {
#ifdef LOG_CDM_OBJECT
  CDM_Log(" %s.releaseObject(): this: %p, cdmObj: %p, count: %d, "
          "storedObj: %p\n",
          className, this, cdmObj, (cdmObj == NULL)? -100 : cdmObj->refCount-1,
          storedObj);
#endif /* LOG_CDM_OBJECT */
  if (cdmObj    != NULL) CDMTCL_DecrRefCount(cdmObj);
  if (storedObj != NULL) delete storedObj;
  cdmObj    = NULL;
  storedObj = NULL;
}; /* OCDM_Object::releaseObject */

void OCDM_Object::deleteObject(void) const {
  releaseObject();
#ifdef LOG_DESTRUCTOR_CALLS
  CDM_Log("~%s.deleteObject(): this: %p << Deleted\n", className, this);
#endif /* LOG_DESTRUCTOR_CALLS */
  OCDM_COMMON_DESTRUCTOR_CODE
} /* OCDM_Object::deleteObject */

const class OCDM_Object *OCDM_Object::getStoredObject(void) const {
  return storedObj;
} /* OCDM_Object::getStoredObject */

void  OCDM_Object::releaseStoredObject(void) const {
#ifdef LOG_CDM_OBJECT
  CDM_Log(" %s.releaseStoredObject(): this: %p, storedObj: %p\n",
          className, this, storedObj);
#endif /* LOG_CDM_OBJECT */
  if (storedObj != NULL) delete storedObj;
  storedObj = NULL;
}; /* OCDM_Object::releaseStoredObject */

void  OCDM_Object::storeObject(const class OCDM_Object *objPtr) const {
  releaseStoredObject();
  storedObj = objPtr;
#ifdef LOG_CDM_OBJECT
  CDM_Log(" %s.storeObject(): this: %p, storedObj: %p\n",
          className, this, storedObj);
#endif /* LOG_CDM_OBJECT */
}; /* OCDM_Object::storeObject */

/*
 * ---------------------------------------------------------------------------
 *               P u b l i c   H e l p e r   f u n c t i o n s 
 * ---------------------------------------------------------------------------
 */

/**
* \brief This method logs information. It is equivalent to
* OCDM_Utilities::Log()
*/
void OCDM_Object::Log(const char *str, ...) const {
  CDM_Log(str);
}; /* OCDM_Object::Log */

/**
* \brief This method returns the number of elements contained in the
* object.
*/
long OCDM_Object::size(void) const {
  long _size;
  if (cdmObj == NULL) return 0;
  _size =  CDM_Length(cdmObj);
  if (_size < 0) throw OCDM_Exception();
  return _size;
}; /* OCDM_Object::size */

/**
* \brief This method returns true if the OCDM object is valid.
*/
OCDM_BOOL OCDM_Object::Valid(void) const {
  if (cdmObj != NULL) return true;
  return false;
} /* OCDM_Object::Valid */

/**
*  \brief This method returns a string representation of the object. 
*/
const char *OCDM_Object::toString(void) const {
#ifdef LOG_OCDM_TOSTRING
  CDM_Log(" %s.toString(): this: %p, cdmObj: %p\n", className, this, cdmObj);
#endif /* LOG_OCDM_TOSTRING */
  if (cdmObj != NULL) return CDMTCL_GetString(cdmObj);
  return "";
} /* OCDM_Object::toString */

/**
* This method returns the type of the object as a string.
*/
const char *OCDM_Object::objectType(void) const {
  return "OCDM_Object";
} /* OCDM_Object::toString */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_Exception                                       *******
 ******************************************************************************
 *****************************************************************************/

OCDM_Exception::OCDM_Exception() throw() {
#ifdef LOG_CONSTRUCTOR_CALLS
  CDM_Log(">>> OCDM_Exception::OCDM_Exception() throw()\n");
#endif /* LOG_CONSTRUCTOR_CALLS */
  /* Retrieve and store the Tcl Object holding the error message... */
  Tcl_Interp *interp = CDM_GetActiveInterpreter();
  errorObj = NULL;
  if (interp != NULL) errorObj = CDMTCL_GetObjResult(interp);
  if (errorObj == NULL)
    errorObj = CDMTCL_NewStringObj("<error message not available>", -1);
  if (errorObj) CDMTCL_IncrRefCount(errorObj);
#ifdef LOG_DESTRUCTOR_CALLS
  CDM_Log("OCDM_Exception: \"%s\"\n", CDMTCL_GetString(errorObj));
#endif /* LOG_DESTRUCTOR_CALLS */
} /* OCDM_Exception::OCDM_Exception */

OCDM_Exception::OCDM_Exception(const char *msg) throw() {
#ifdef LOG_CONSTRUCTOR_CALLS
  CDM_Log(">>> OCDM_Exception::OCDM_Exception(\"%s\") throw()\n", msg);
#endif /* LOG_CONSTRUCTOR_CALLS */
  errorObj = CDMTCL_NewStringObj(msg, -1);
  if (errorObj) CDMTCL_IncrRefCount(errorObj);
} /* OCDM_Exception::OCDM_Exception */

OCDM_Exception::~OCDM_Exception() throw() {
#ifdef LOG_DESTRUCTOR_CALLS
  CDM_Log("~OCDM_Exception() ");
#endif /* LOG_DESTRUCTOR_CALLS */
  if (errorObj) CDMTCL_DecrRefCount(errorObj);
  errorObj = NULL;
#ifdef LOG_DESTRUCTOR_CALLS
  CDM_Log("<<< Freed\n");
#endif /* LOG_DESTRUCTOR_CALLS */
} /* OCDM_Exception::~OCDM_Exception */

const char *OCDM_Exception::what() const throw() {
  if (errorObj) {
    int len;
    const char *msg = CDMTCL_GetStringFromObj(errorObj, &len);

    if (len == 0) {
      return "<error message not available>";
    }
    return msg;
  }
  return "<error message not available>";
} /* OCDM_Exception::what */

OCDM_Exception::OCDM_Exception(const class OCDM_Exception& obj) {
#ifdef LOG_CONSTRUCTOR_CALLS
  CDM_Log(">>> OCDM_Exception::OCDM_Exception(const class OCDM_Exception& obj)\n");
#endif /* LOG_CONSTRUCTOR_CALLS */
  errorObj = obj.errorObj;
  if (errorObj) CDMTCL_IncrRefCount(errorObj);
} /* OCDM_Exception::OCDM_Exception */

class OCDM_Exception& OCDM_Exception::operator= (
                                          const class OCDM_Exception& obj) {
  if (obj.errorObj != NULL) CDMTCL_IncrRefCount(obj.errorObj);
  if (errorObj) CDMTCL_DecrRefCount(errorObj);
  errorObj = obj.errorObj;
  return *this;
} /* OCDM_Exception::operator = */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_ByteSequence                                    *******
 ******************************************************************************
 *****************************************************************************/

/**
 * \brief This is the main constructor of the OCDM_ByteSequence class. 
 * \par Description:
 * This is the main constructor of the OCDM_ByteSequence class.
 * It creates a OCDM_ByteSequence object from a string stored in a char *.
 * The string is expected to be in UTF-8 format and must be NULL terminated.
 * \par Arguments:
 * \arg \c str: a pointer to a NULL terminated string that will be copied
 * as the object's value.
 * \see OCDM_RawData::OCDM_RawData, Tcl_NewStringObj
 */
OCDM_ByteSequence::OCDM_ByteSequence(const char *str) {
  initObject(CDMTCL_NewStringObj(str, -1));
  OCDM_COMMON_CONSTRUCTOR_CODE
} /* OCDM_ByteSequence::OCDM_ByteSequence */

/**
 * \brief This is the main constructor of the OCDM_ByteSequence class. 
 * \par Description:
 * This is the main constructor of the OCDM_ByteSequence class.
 * It creates a OCDM_ByteSequence object from a string stored in a char *.
 * The string is expected to be in UTF-8 format.
 * \a len specifies the number of bytes to copy from \a str.
 * If negative, all bytes up to the first NULL are used.
 * \par Arguments:
 * \arg \c str: a pointer to a string that will be copied as the object's
 * value.
 * \arg \c len: an integer specifying the number of bytes to be copied
 * from \c str.
 * \see OCDM_RawData::OCDM_RawData, Tcl_NewStringObj
 */
OCDM_ByteSequence::OCDM_ByteSequence(const char *str, const int len) {
  initObject(CDMTCL_NewStringObj(str, len));
  OCDM_COMMON_CONSTRUCTOR_CODE
} /* OCDM_ByteSequence::OCDM_ByteSequence */

/*
 * ---------------------------------------------------------------------------
 *               P u b l i c   H e l p e r   f u n c t i o n s 
 * ---------------------------------------------------------------------------
 */

/**
 * \brief Set the object to hold a particular string. 
 * \par Description:
 * This function can be used to change the textual data that is held by 
 * an object. The older value (if any) will be deleted and the object will
 * be set to hold a copy of the provided string.
 * The string is expected to be in UTF-8 format and must be NULL terminated.
 * \par Arguments:
 * \arg \c str: a pointer to a NULL terminated string that will be copied
 * as the object's value.
 * \see OCDM_RawData::OCDM_RawData, Tcl_NewStringObj  
 */
void OCDM_ByteSequence::Set(const char *str) {
  setObject(CDMTCL_NewStringObj(str, -1));
} /* OCDM_ByteSequence::Set */

/**
 * \brief Set the object to hold a particular string. 
 * \par Description:
 * This function can be used to change the textual data that is held by 
 * an object. The older value (if any) will be deleted and the object will
 * be set to hold a copy of the provided string.
 * The string is expected to be in UTF-8 format.
 * \a len specifies the number of bytes to copy from \a str.
 * If negative, all bytes up to the first NULL are used.
 * \par Arguments:
 * \arg \c str: a pointer to a NULL terminated string that will be copied
 * as the object's value.
 * \arg \c len: an integer specifying the number of bytes to be copied
 * from \c str.
 * \see OCDM_RawData::OCDM_RawData, Tcl_NewStringObj  
 */
void OCDM_ByteSequence::Set(const char *str, const int len) {
  setObject(CDMTCL_NewStringObj(str, len));
} /* OCDM_ByteSequence::Set */

/**
 * \brief This function returns the number of characters contained in the
 * object.
 */
long OCDM_ByteSequence::length(void) const {
  int _size;
  if (getObject() == NULL) return 0;
  CDMTCL_GetStringFromObj(getObject(), &_size);
  return _size;
} /* OCDM_ByteSequence::length */

/**
 * \brief Changes every character in the object to upper-case.
 * \return Returns true if the convertion was succesfull, false otherwise.
 */
OCDM_BOOL OCDM_ByteSequence::toUpper(void) const {
  Tcl_Obj *newObj;
  const char *string;
  int len;
  if (getObject() == NULL) return false;
  string = CDMTCL_GetStringFromObj(getObject(), &len);
  newObj = CDMTCL_NewStringObj(string, len);
  if (newObj == NULL) return false;
  /*
   * Do the
   * conversion in place.  The conversion may change the len
   * of the string, so reset the len after conversion.
   */
  len = Tcl_UtfToUpper(Tcl_GetString(newObj));
  Tcl_SetObjLength(newObj, len);
  setObject(newObj);
  return true;
} /* OCDM_ByteSequence::toUpper */

/**
 * \brief Changes every character in the object to upper-case.
 * \return Returns true if the convertion was succesfull, false otherwise.
 */
OCDM_BOOL OCDM_ByteSequence::toLower(void) const {
  Tcl_Obj *newObj;
  char *string;
  int len;
  if (getObject() == NULL) return false;
  string = Tcl_GetStringFromObj(getObject(), &len);
  newObj = Tcl_NewStringObj(string, len);
  if (newObj == NULL) return false;
  /*
   * Do the
   * conversion in place.  The conversion may change the len
   * of the string, so reset the len after conversion.
   */
  len = Tcl_UtfToLower(Tcl_GetString(newObj));
  Tcl_SetObjLength(newObj, len);
  setObject(newObj);
  return true;
} /* OCDM_ByteSequence::toLower */

/**
 * \brief Changes every character in the object to upper-case.
 * \return Returns true if the convertion was succesfull, false otherwise.
 */
OCDM_BOOL OCDM_ByteSequence::toTitle(void) const {
  Tcl_Obj *newObj;
  char *string;
  int len;
  if (getObject() == NULL) return false;
  string = Tcl_GetStringFromObj(getObject(), &len);
  newObj = Tcl_NewStringObj(string, len);
  if (newObj == NULL) return false;
  /*
   * Do the
   * conversion in place.  The conversion may change the len
   * of the string, so reset the len after conversion.
   */
  len = Tcl_UtfToTitle(Tcl_GetString(newObj));
  Tcl_SetObjLength(newObj, len);
  setObject(newObj);
  return true;
} /* OCDM_ByteSequence::toTitle */


/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_ByteSequenceSet                                 *******
 ******************************************************************************
 *****************************************************************************/

/*
 * ---------------------------------------------------------------------------
 *                           O p e r a t o r s 
 * ---------------------------------------------------------------------------
 */

/**
* \brief This operator returns the element of a set pointed by the "index"
* parameter.
* \par Description: This operator returns the element of a set pointed by the
* "index" parameter. Note that under
* languages that do not support operator overloading (i.e. Java, Perl,
* Python) this operator appears as a plain method named as "get".
* \par Arguments:
* \arg \c index: the index (position) of the element to be retrieved. The
* first element of the set has as index the value 0.
* \note In case of an error, an exception of type OCDM_Exception
* will be thrown. 
*/

OCDM_INDEX_OPERATOR_REF(OCDM_ByteSequence) OCDM_ByteSequenceSet::operator[] (
                                                       const long index) const {
  return OCDM_GetNth<OCDM_ByteSequence>(*this, index);
}; /* OCDM_ByteSequenceSet::operator[] */

void OCDM_ByteSequenceSet::add(const OCDM_ByteSequence& obj) {
  if (!Valid()) setObject(CDMTCL_NewListObj(0, NULL));
  if (OCDM_AppendElement(*this, (CDM_Object) obj.getObject()) != TCL_OK) {
    throw OCDM_Exception();
  }
}; /* OCDM_ByteSequenceSet::add */

void OCDM_ByteSequenceSet::add(const char *str) {
  if (!Valid()) setObject(CDMTCL_NewListObj(0, NULL));
  if (OCDM_AppendElement(*this, CDMTCL_NewStringObj(str, -1)) != TCL_OK) {
    throw OCDM_Exception();
  }
}; /* OCDM_ByteSequenceSet::add */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_RawData                                         *******
 ******************************************************************************
 *****************************************************************************/

/**
 * \brief This is the main constructor of the OCDM_RawData class. 
 * \par Description:
 * This is the main constructor of the OCDM_RawData class.
 * It creates a OCDM_RawData object from a string stored in a char *.
 * The string is expected to be in UTF-8 format and must be NULL terminated.
 * \par Arguments:
 * \arg \c str: a pointer to a NULL terminated string that will be copied
 * as the object's value.
 * \see OCDM_RawData::OCDM_RawData, Tcl_NewStringObj
 */
OCDM_RawData::OCDM_RawData(const char *str) {
  initObject(CDMTCL_NewStringObj(str, -1));
  OCDM_COMMON_CONSTRUCTOR_CODE
} /* OCDM_RawData::OCDM_RawData */

/**
 * \brief This is the main constructor of the OCDM_RawData class. 
 * \par Description:
 * This is the main constructor of the OCDM_RawData class.
 * It creates a OCDM_RawData object from a string stored in a char *.
 * The string is expected to be in UTF-8 format.
 * \a len specifies the number of bytes to copy from \a str.
 * If negative, all bytes up to the first NULL are used.
 * \par Arguments:
 * \arg \c str: a pointer to a string that will be copied as the object's
 * value.
 * \arg \c len: an integer specifying the number of bytes to be copied
 * from \c str.
 * \see OCDM_RawData::OCDM_RawData, Tcl_NewStringObj
 */
OCDM_RawData::OCDM_RawData(const char *str, const int len) {
  initObject(CDMTCL_NewStringObj(str, len));
  OCDM_COMMON_CONSTRUCTOR_CODE
} /* OCDM_RawData::OCDM_RawData */

/**
 * \brief Set the object to hold a particular string. 
 * \par Description:
 * This function can be used to change the textual data that is held by 
 * an object. The older value (if any) will be deleted and the object will
 * be set to hold a copy of the provided string.
 * The string is expected to be in UTF-8 format and must be NULL terminated.
 * \par Arguments:
 * \arg \c str: a pointer to a NULL terminated string that will be copied
 * as the object's value.
 * \see OCDM_RawData::OCDM_RawData, Tcl_NewStringObj  
 */
void OCDM_RawData::Set(const char *str) {
  setObject(CDMTCL_NewStringObj(str, -1));
} /* OCDM_RawData::Set */

/**
 * \brief Set the object to hold a particular string. 
 * \par Description:
 * This function can be used to change the textual data that is held by 
 * an object. The older value (if any) will be deleted and the object will
 * be set to hold a copy of the provided string.
 * The string is expected to be in UTF-8 format.
 * \a len specifies the number of bytes to copy from \a str.
 * If negative, all bytes up to the first NULL are used.
 * \par Arguments:
 * \arg \c str: a pointer to a NULL terminated string that will be copied
 * as the object's value.
 * \arg \c len: an integer specifying the number of bytes to be copied
 * from \c str.
 * \see OCDM_RawData::OCDM_RawData, Tcl_NewStringObj  
 */
void OCDM_RawData::Set(const char *str, const int len) {
  setObject(CDMTCL_NewStringObj(str, len));
} /* OCDM_RawData::Set */

/**
 * \brief This function returns the number of characters contained in the
 * object.
 */
long OCDM_RawData::length(void) const {
  int _size;
  if (getObject() == NULL) return 0;
  CDMTCL_GetStringFromObj(getObject(), &_size);
  return _size;
} /* OCDM_RawData::length */

/**
 * \brief Changes every character in the object to upper-case.
 * \return Returns true if the convertion was succesfull, false otherwise.
 */
OCDM_BOOL OCDM_RawData::toUpper(void) const {
  Tcl_Obj *newObj;
  char *string;
  int len;
  if (getObject() == NULL) return false;
  string = Tcl_GetStringFromObj(getObject(), &len);
  newObj = Tcl_NewStringObj(string, len);
  if (newObj == NULL) return false;
  /*
   * Do the
   * conversion in place.  The conversion may change the len
   * of the string, so reset the len after conversion.
   */
  len = Tcl_UtfToUpper(Tcl_GetString(newObj));
  Tcl_SetObjLength(newObj, len);
  setObject(newObj);
  return true;
} /* OCDM_RawData::toUpper */

/**
 * \brief Changes every character in the object to upper-case.
 * \return Returns true if the convertion was succesfull, false otherwise.
 */
OCDM_BOOL OCDM_RawData::toLower(void) const {
  Tcl_Obj *newObj;
  char *string;
  int len;
  if (getObject() == NULL) return false;
  string = Tcl_GetStringFromObj(getObject(), &len);
  newObj = Tcl_NewStringObj(string, len);
  if (newObj == NULL) return false;
  /*
   * Do the
   * conversion in place.  The conversion may change the len
   * of the string, so reset the len after conversion.
   */
  len = Tcl_UtfToLower(Tcl_GetString(newObj));
  Tcl_SetObjLength(newObj, len);
  setObject(newObj);
  return true;
} /* OCDM_RawData::toLower */

/**
 * \brief Changes every character in the object to upper-case.
 * \return Returns true if the convertion was succesfull, false otherwise.
 */
OCDM_BOOL OCDM_RawData::toTitle(void) const {
  Tcl_Obj *newObj;
  char *string;
  int len;
  if (getObject() == NULL) return false;
  string = Tcl_GetStringFromObj(getObject(), &len);
  newObj = Tcl_NewStringObj(string, len);
  if (newObj == NULL) return false;
  /*
   * Do the
   * conversion in place.  The conversion may change the len
   * of the string, so reset the len after conversion.
   */
  len = Tcl_UtfToTitle(Tcl_GetString(newObj));
  Tcl_SetObjLength(newObj, len);
  setObject(newObj);
  return true;
} /* OCDM_RawData::toTitle */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_RawDataSet                                 *******
 ******************************************************************************
 *****************************************************************************/

/*
 * ---------------------------------------------------------------------------
 *                           O p e r a t o r s 
 * ---------------------------------------------------------------------------
 */

/**
* \brief This operator returns the element of a set pointed by the "index"
* parameter.
* \par Description: This operator returns the element of a set pointed by the
* "index" parameter. Note that under
* languages that do not support operator overloading (i.e. Java, Perl,
* Python) this operator appears as a plain method named as "get".
* \par Arguments:
* \arg \c index: the index (position) of the element to be retrieved. The
* first element of the set has as index the value 0.
* \note In case of an error, an exception of type OCDM_Exception
* will be thrown. 
*/

OCDM_INDEX_OPERATOR_REF(OCDM_RawData) OCDM_RawDataSet::operator[] (
                                                       const long index) const {
  return OCDM_GetNth<OCDM_RawData>(*this, index);
}; /* OCDM_RawDataSet::operator[] */

void OCDM_RawDataSet::add(const OCDM_RawData& obj) {
  if (!Valid()) setObject(CDMTCL_NewListObj(0, NULL));
  if (OCDM_AppendElement(*this, (CDM_Object) obj.getObject()) != TCL_OK) {
    throw OCDM_Exception();
  }
}; /* OCDM_RawDataSet::add */

void OCDM_RawDataSet::add(const char *str) {
  if (!Valid()) setObject(CDMTCL_NewListObj(0, NULL));
  if (OCDM_AppendElement(*this, CDMTCL_NewStringObj(str, -1)) != TCL_OK) {
    throw OCDM_Exception();
  }
}; /* OCDM_RawDataSet::add */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_Span                                            *******
 ******************************************************************************
 *****************************************************************************/

/**
*  \brief This function will create a Span object having the start and
*  end values defined in the parameters.
*  \par Description:
*  This function will create a Span object having a start and an end value 
*  defined in the parameters. Bare in mind that each span object is a set
*  of exactly two integers, named as "start" and "end".  
*  In case of an error, an Exception of type OCDM_Exception will
*  be thrown and a message will be left at the current active Tcl
*  interpreter (CDM_Interp).
*  \par Arguments:
*  \arg \c start: The starting value of the Span object to be created.
*  \arg \c   end: The ending value of the Span object to be created.   
*
*/
OCDM_Span::OCDM_Span(const long start, const long end) {
  CDM_Span obj = CDM_CreateSpan(start, end);
  if (obj == NULL) throw OCDM_Exception();
  initObject(obj);
  OCDM_COMMON_CONSTRUCTOR_CODE
} /* OCDM_Span::OCDM_Span */

/**
*  \brief This function returns the start of the span object in use.
*  \par Description:
*  This function returns the starting value of the span object in use.
*  It takes no other parameters. 
*
*/
long OCDM_Span::GetStart(void) const {
  ensureValidity();
  long start = CDM_GetStart(getObject());
  //if (start < 0) throw OCDM_Exception();
  return start;
} /* OCDM_Span::GetStart */

/**
*  \brief This function returns the end of the span object in use. 
*  \par Description:
*  This function returns the ending value of the span object in use.
*  It takes no other parameters.
*
*/
long OCDM_Span::GetEnd(void) const {
  ensureValidity();
  long end = CDM_GetEnd(getObject());
  //if (end < 0) throw OCDM_Exception();
  return end;
} /* OCDM_Span::GetEnd */

/**
*  This function will place the values of the both span start and end of 
*  the span object in use, in the two specified long integer pointers.  
*
*/
int OCDM_Span::GetSpanOffsets(long &start, long &end) const {
  ensureValidity();
  int status = CDM_GetSpanOffsets(getObject(), &start, &end);
  if (status != TCL_OK) throw OCDM_Exception();
  return status;
} /* OCDM_Span::GetSpanOffsets */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_SpanSet                                         *******
 ******************************************************************************
 *****************************************************************************/

/**
*  \brief This function will create a new SpanSet object 
*  \par Description:
*  The newly created spanset object will contain a single Span object (of type 
*  OCDM_Span) with the specified start and end values.
*  \par Arguments:
*  \arg \c start: The starting value of the span object.
*  \arg \c   end: The ending value of the span object.
*
*  \note In case of an error, an exception of type OCDM_Exception
*  will be thrown. 
*
*/
OCDM_SpanSet::OCDM_SpanSet(const long start, const long end) {
  CDM_SpanSet obj = CDM_CreateSpanSet(start, end);
  if (obj == NULL) throw OCDM_Exception();
  initObject(obj);
  OCDM_COMMON_CONSTRUCTOR_CODE
} /* OCDM_SpanSet::OCDM_SpanSet */

/*
 * ---------------------------------------------------------------------------
 *                           O p e r a t o r s 
 * ---------------------------------------------------------------------------
 */

/**
* \brief This operator returns the element of a set pointed by the "index"
* parameter.
* \par Description: This operator returns the element of a set pointed by the
* "index" parameter. Note that under
* languages that do not support operator overloading (i.e. Java, Perl,
* Python) this operator appears as a plain method named as "get".
* \par Arguments:
* \arg \c index: the index (position) of the element to be retrieved. The
* first element of the set has as index the value 0.
* \note In case of an error, an exception of type OCDM_Exception
* will be thrown. 
*/

OCDM_INDEX_OPERATOR_REF(OCDM_Span) OCDM_SpanSet::operator[] (const long index)
                                                                         const {
  return OCDM_GetNth<OCDM_Span>(*this, index);
}; /* OCDM_SpanSet::operator[] */
/*
 * ---------------------------------------------------------------------------
 *              P r i v a t e   H e l p e r   f u n c t i o n s 
 * ---------------------------------------------------------------------------
 */

/**
*  \brief It appends the given Span to the Span set object in use.  
*  \par Arguments:
*  \arg \c Span: The span object to be appended to the span set object in use.  
*  \note In case of an error, an exception of type OCDM_Exception
*  will be thrown. 
*
*/
void OCDM_SpanSet::AddSpan(const OCDM_Span& obj) {
  if (getObject() == NULL) {
    setObject(CDM_CreateSpanSet());
  }
  CDM_SpanSet newSpanSet = CDM_AddSpan(getObject(), obj.getObject());
  if (newSpanSet == NULL) throw OCDM_Exception();
  setObject(newSpanSet);
} /* OCDM_SpanSet::AddSpan */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_AttributeValue                                  *******
 ******************************************************************************
 *****************************************************************************/

/**
*  \brief Constructor
*  \par Description:
*  It Creates a new AttributeValue object.
*  \par Arguments:
*  \arg \c Type: The type of the attribute value as defined in the users
*  guide to \b Ellogon.
*  \arg \c Value: The value of the attribute value.
*/
OCDM_AttributeValue::OCDM_AttributeValue(const int Type, const char *Value) {
  CDM_AttributeValue obj = CDM_CreateAttributeValue(Type, (char *) Value);
  if (obj == NULL) throw OCDM_Exception();
  initObject(obj);
  OCDM_COMMON_CONSTRUCTOR_CODE
} /* OCDM_AttributeValue::OCDM_AttributeValue */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_Attribute                                       *******
 ******************************************************************************
 *****************************************************************************/

/**
*  \par Description:
*  An overloaded version of the above function.
*  It is the main constructor of the OCDM_Attribute class.
*  
*  Needless to say that this function will create an Attribute object
*
*  Special attention should be given to the Attribute value object 
*  (specified by the "Value" parameter) so as not to leak memory. This 
*  function internally will always increase the reference count.
*
*  The value of the "Name" parameter will not be modified 
*  or referenced in any way, as CDM will create and manipulate a local 
*  copy of this value.
*
*  \par Arguments:
*  \arg \c    name: The name of the OCDM_Attribute object to be created.
*  \arg \c Attrval: The value of the Atribute to be created.
*
*/
OCDM_Attribute::OCDM_Attribute(const char *name,
                                           const OCDM_AttributeValue& AttrVal) {
  CDM_Attribute obj = CDM_CreateAttribute((char *) name, AttrVal.getObject());
  if (obj == NULL) throw OCDM_Exception();
  initObject(obj);
  OCDM_COMMON_CONSTRUCTOR_CODE
} /* OCDM_Attribute::OCDM_Attribute */

/**
* \par Description:
* This constructor will create an OCDM_Attribute object with the provided
* name (as specified through the "Name" parameter). However, an Attribute
* Value object is not required: instead an Attribute Value object will be
* created by using the provided parameters "Type" and "Value".
*
* \par Arguments:
* \arg \c name:  The name of the OCDM_Attribute object to be created.
* \arg \c Type:  The type of the Attribute Value.
* \arg \c Value: The value to be placed in the Attribute.
*/
OCDM_Attribute::OCDM_Attribute(const char *name,
                               const int Type, const char *Value) {
  CDM_AttributeValue valObj = CDM_CreateAttributeValue(Type, (char *) Value);
  if (valObj == NULL) throw OCDM_Exception();
  CDM_Attribute obj = CDM_CreateAttribute((char *) name, valObj);
  if (obj == NULL) throw OCDM_Exception();
  initObject(obj);
  OCDM_COMMON_CONSTRUCTOR_CODE
} /* OCDM_Attribute::OCDM_Attribute */

/**
*  \brief Returns Attribute's name.
*  \note In case of Error an Exception of type OCDM_Exception will be thrown
*
*/
const char *OCDM_Attribute::GetName(void) const {
  ensureValidity();
  char *valueStr = CDM_GetName(getObject());
  if (valueStr == NULL) throw OCDM_Exception();
  return valueStr;
} /* OCDM_Attribute::GetName */

/**
*  \brief Returns Attribute's Value String.
*  \note In case of Error an Exception of type OCDM_Exception will be thrown
*
*/
const char *OCDM_Attribute::GetValueString(void) const {
  ensureValidity();
  char *valueStr = CDM_GetValueString(getObject());
  if (valueStr == NULL) throw OCDM_Exception();
  return valueStr;
} /* OCDM_Attribute::GetValueString */

/**
*  \brief Returns Attribute's Value.  
*  \note In case of Error an Exception of type OCDM_Exception will be thrown
*/
OCDM_REF(OCDM_AttributeValue) OCDM_Attribute::GetValue(void) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AttributeValue>(*this,
                                                CDM_GetValue(getObject()));
} /* OCDM_Attribute::GetValue */

/**
*  \brief Returns Attribute's Type.
*  \note In case of Error an Exception of type OCDM_Exception will be thrown
*/
int OCDM_Attribute::GetValueType(void) const {
  ensureValidity();
  int valueType = CDM_GetValueType(getObject());
  if (valueType < 0) throw OCDM_Exception();
  return valueType;
} /* OCDM_Attribute::GetValueType */

/**
*  \brief This function will return the value of the Attribute value of the 
*  Attribute object in use. 
*  \par Description:
*  This function will return the value of the Attribute value of the 
*  Attribute object in use. An Attribute value consists of two elements:
*  the type of the information stored as Attribute value (for example 
*  GDM_STRING or GDM_STRING_SET) and the information that is stored as a 
*  value of the Attribute. This function returns the latter object.  
*  This function is different from GetValue, as GetValue returns a 
*  list object that holds two elements, the value type and the value. On 
*  the other hand, GetValueValue returns directly the value, by 
*  ommiting the object specifying the type.  If the caller wants to also
*  get the type of the value, he can either call GetValue or 
*  GetValueType.
*
*  \note In case of Error an Exception of type OCDM_Exception will be thrown
*
*/
const char *OCDM_Attribute::GetValueValue(void) const {
  ensureValidity();
  CDM_AttributeValue value = CDM_GetValueValue(getObject());
  if (value == NULL) throw OCDM_Exception();
  return CDMTCL_GetString(value);
} /* OCDM_Attribute::GetValueValue */

/**
*  \brief This function will construct and return a new Attribute 
*  object that will be an exact copy of the provided one.
*  \par Description:
*  This function will construct and return a new Attribute 
*  object that will be an exact copy of the provided one.
*  The difference will be that the returned Attribute object will utilise 
*  elements from the object cache that is managed internally by CDM. As a 
*  result the returned object will normally require less memory to be 
*  stored than the provided one, as it will re-use elements already in use 
*  by other CDM objects.
*
*  \note This function should be used on Attribute objects that originate from 
*  an unknown source.  For example, if an Attribute object is stored in a 
*  file and then red back as a string, it will not use cached by the CDM 
*  objects.  It will be a good idea if the caller created a new object by 
*  calling this function and delete the object red from the file by either 
*  calling Tcl_DecrRefCount or 
*  CDM_Free.
*
*/
void OCDM_Attribute::ReconstructAttribute(void) const {
  ensureValidity();
  CDM_Attribute newAttribute = CDM_ReconstructAttribute(getObject());
  if (newAttribute) setObject(newAttribute);
} /* OCDM_Attribute::ReconstructAttribute */

/**
*  \brief
*  This function will construct and return a new Attribute 
*  object that will be an exact copy of the provided one.
*  \par Description:
*  This function will construct and return a new Attribute 
*  object that will be an exact copy of the provided one. The only 
*  difference will be that the returned Attribute object will utilise 
*  elements from the object cache that is managed internally by CDM. As a 
*  result the returned object will normally require less memory to be 
*  stored than the provided one, as it will re-use elements already in use 
*  by other CDM objects.
*
*  \note This function should be used on Attribute objects that originate from 
*  an unknown source.  For example, if an Attribute object is stored in a 
*  file and then red back as a string, it will not use cached by the CDM 
*  objects.  It will be a good idea if the caller created a new object by 
*  calling this function and delete the object red from the file by either 
*  calling Tcl_DecrRefCount or 
*  CDM_Free.
*
*/
void OCDM_Attribute::ReconstructAttribute(int* type) const {
  ensureValidity();
  CDM_Attribute newAttribute = CDM_ReconstructAttribute(getObject(), type);
  if (newAttribute) setObject(newAttribute);
} /* OCDM_Attribute::ReconstructAttribute */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_AttributeSet                                    *******
 ******************************************************************************
 *****************************************************************************/

/**
*  \brief This is the constructor function of the OCDM_AttributeSet class.
*  
*  \par Description:
*  This is the constructor function of the OCDM_AttributeSet class.
*  It will create a new Attribute set object (of type OCDM_AttributeSet)
*  that will contain a newly created Attribute 
*  object, created according to the values specified as parameters.
*  \arg \c    name: The name of the OCDM_AttributeSet object to be created.
*  \arg \c Attrval: The value of the Atribute to be created.
*  \note This function is equivelant to creating an empty Attribute set 
*  object with CDM_CreateAttributeSet (void), creating a new Attribute 
*  object with CDM_CreateAttribute (char *Name, CDM_AttributeValue Value) 
*  and adding this Attribute object to the newly created Annotation set.  
*
*/
OCDM_AttributeSet::OCDM_AttributeSet(const char *name,
                                     const OCDM_AttributeValue& AttrVal) {
  CDM_AttributeSet obj =
             CDM_CreateAttributeSet((char *) name, AttrVal.getObject());
  if (obj == NULL) throw OCDM_Exception();
  initObject(obj);
  OCDM_COMMON_CONSTRUCTOR_CODE
} /* OCDM_AttributeSet::OCDM_AttributeSet */

/*
 * ---------------------------------------------------------------------------
 *                           O p e r a t o r s 
 * ---------------------------------------------------------------------------
 */

/**
* \brief This operator returns the element of a set pointed by the "index"
* parameter.
* \par Description: This operator returns the element of a set pointed by the
* "index" parameter. Note that under
* languages that do not support operator overloading (i.e. Java, Perl,
* Python) this operator appears as a plain method named as "get".
* \par Arguments:
* \arg \c index: the index (position) of the element to be retrieved. The
* first element of the set has as index the value 0.
* \note In case of an error, an exception of type OCDM_Exception
* will be thrown. 
*/

OCDM_INDEX_OPERATOR_REF(OCDM_Attribute) OCDM_AttributeSet::operator[] (
                                                       const long index) const {
  return OCDM_GetNth<OCDM_Attribute>(*this, index);
}; /* OCDM_AttributeSet::operator[] */
/*
 * ---------------------------------------------------------------------------
 *              P u b l i c    H e l p e r   f u n c t i o n s 
 * ---------------------------------------------------------------------------
 */

/**
*  \brief This function will append the given Attribute (the value of the 
*  "Attribute" parameter) to the Attribute set object in use.
*  \par Arguments:
*  \arg \c Attr: The Attribute object to be appended to the attribute set.
*    
*  \note In case of an Error an Exception of type OCDM_Exception
*  will be thrown. 
*
*/
void OCDM_AttributeSet::AddAttribute(const OCDM_Attribute& Attr) {
  if (getObject() == NULL) {
    setObject(CDM_CreateAttributeSet());
  }
  CDM_AttributeSet newAttrSet = CDM_AddAttribute(getObject(), Attr.getObject());
  if (newAttrSet == NULL) throw OCDM_Exception();
  setObject(newAttrSet);
} /* OCDM_AttributeSet::AddAttribute */

/**
*  \brief This function will return true if an Attribute with the specified
*  name exists in the Attribute set.
*  \par Arguments:
*  \arg \c name: The Attribute name to be found.
*    
*  \note In case of an Error an Exception of type OCDM_Exception
*  will be thrown. 
*
*/
OCDM_BOOL OCDM_AttributeSet::AttributeExists(const char *name) const {
  ensureValidity();
  CDM_AttributeSet set = getObject();
  CDM_Attribute *Attr;
  int items;
  if (CDM_SetGetElements(set, &items, &Attr) != TCL_OK) throw OCDM_Exception();
  for (int i = 0; i < items; i++) {
    if (CDM_Strcmp(CDM_GetName(Attr[i]), (char *) name) == 0) return true;
  }
  return false;
} /* OCDM_AttributeSet::AttributeExists */

/**
*  \brief This function will return an OCDM_Attribute object having as name
*  the provided name through the "Name" parameter.
*  \par Arguments:
*  \arg \c Name: The name of the OCDM_Attribute object to be returned.
*    
*  \note In case of an Error an Exception of type OCDM_Exception
*  will be thrown. 
*
*/
OCDM_REF(OCDM_Attribute) OCDM_AttributeSet::GetAttribute(const char *name)
                                                                         const {
  ensureValidity();
  CDM_AttributeSet set = getObject();
  CDM_Attribute *Attr, attr = NULL;
  int items;
  if (CDM_SetGetElements(set, &items, &Attr) != TCL_OK) throw OCDM_Exception();
  for (int i = 0; i < items; i++) {
    attr = Attr[i];
    if (CDM_Strcmp(CDM_GetName(attr), (char *) name) == 0) break;
    attr = NULL;
  }
  return OCDM_GetReference<OCDM_Attribute>(*this, attr);
} /* OCDM_AttributeSet::GetAttribute */

/**
*  \brief This function will remove all Attributes having the provided name
*  from the Attribute set. The number of removed Annotations will be returned.
*  \par Arguments:
*  \arg \c Name: The name of the OCDM_Attribute object to be removed.
*    
*  \note In case of an Error an Exception of type OCDM_Exception
*  will be thrown. 
*
*/
int OCDM_AttributeSet::RemoveAttribute(const char *name) {
  ensureValidity();
  CDM_AttributeSet set = CDM_CreateAttributeSet();
  CDM_Attribute *Attr;
  int items, removed = 0;
  if (CDM_SetGetElements(getObject(), &items, &Attr) != TCL_OK)
                                                        throw OCDM_Exception();
  for (int i = 0; i < items; i++) {
    if (CDM_Strcmp(CDM_GetName(Attr[i]), (char *) name) != 0) {
      if (CDM_AddAttribute(set, Attr[i]) == NULL) throw OCDM_Exception();
    } else {
      removed++;
    }
  }
  setObject(set);
  return removed;
} /* OCDM_AttributeSet::RemoveAttribute */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_Annotation                                      *******
 ******************************************************************************
 *****************************************************************************/

/**
*  /brief This is the constructor function fo the OCDM_Annotation class.
*  
*  \par Description:
*  This is the constructor function fo the OCDM_Annotation class.
*  It will create a new Annotation object. The newly created Annotation object
*  will not have an empty Annotation Id and the Annotation type, Span set
*  and Annotation set will be set to the values 
*  provided by the caller through the corresponding parameters.
*  This function will always use the provided Span set object in the 
*  returned Annotation object.  However, this function may or may not use 
*  the provided Attribute set object: this function will internally 
*  increment the reference count of the provided Attribute set 
*  object.  If the reference count of the provided Attribute set object 
*  equals 0 (i.e.  the Attribute set object has been created with a call to 
*  OCDM_CreateAttribute), this function may free the object.  As a result,
*  if 
*  the caller wants to ensure that the provided Attribute set object is 
*  valid after calling this function, he should increment its reference 
*  count with Tcl_IncrRefCount.
*
*  \par Arguments:
*  \arg \c type: Annotation types are textual values that are used to classify
*  annotations into categories
*  \arg \c Spans: This is a set of pairs of two byte offsets that denote 
*  the range(s) of the annotated textual data. 
*  \arg \c Attributes: These attributes usually contain the necessary linguistic
*  information. 
*
*/
OCDM_Annotation::OCDM_Annotation(const char* type, const OCDM_SpanSet& Spans,
                            const OCDM_AttributeSet& Attributes) {
  CDM_Annotation obj = CDM_CreateAnnotation((char *) type, Spans.getObject(),
                                                      Attributes.getObject());
  if (obj == NULL) throw OCDM_Exception();
  initObject(obj);
  OCDM_COMMON_CONSTRUCTOR_CODE
} /* OCDM_Annotation::OCDM_Annotation */

/**
*  /brief This is the constructor function fo the OCDM_Annotation class.
*  
*  \par Description:
*  This constructor can be used in order to create an Annotation object which
*  will have only one span (although the number of Spans can be increased
*  later) and we don't want to create an OCDM_Span object. This constructor
*  can be more efficient, especially if the Annotation will always have only
*  one Span object (a quite common situation).
*
*  \par Arguments:
*  \arg \c type: Annotation types are textual values that are used to classify annotations
*  into categories
*  \arg \c start: This is the start character offset to be placed in the Span
*  of the Annotation.
*  \arg \c end: This is the end character offset to be placed in the Span
*  of the Annotation.
*  \arg \c Attributes: These attributes usually contain the necessary linguistic
*  information. (Can be also NULL)
*
*/
OCDM_Annotation::OCDM_Annotation(const char* type,
                                 const long start, const long end,
                                 const OCDM_AttributeSet& Attributes) {
  CDM_Annotation obj = CDM_CreateAnnotation(type, start, end,
                                                      Attributes.getObject());
  if (obj == NULL) throw OCDM_Exception();
  initObject(obj);
  OCDM_COMMON_CONSTRUCTOR_CODE
} /* OCDM_Annotation::OCDM_Annotation */

/**
*  \brief This function will compare the Annotation in use and the Annotation
*  defined
*  \par Description:
*  This function will compare the Annotation in use and the Annotation
*  defined in the parameter Ann, according to their Span sets.
*
*  \par Arguments:
*  \arg \c Ann: The Annotation to be compared with the annotation object in use.
*  
*  \note This function returns -1 if the Annotation in use is lower than 
*  Ann. If Ann has the same text ranges as the Annotation in use
*  then a zero is returned.  Finally, 1 
*  is returned if the current Annotation is greater than Ann 
*  In case of an error an Exception of type OCDM_Exception will be thrown
*
*/
int OCDM_Annotation::CompareAnnotations(const OCDM_Annotation &Ann) const {
  ensureValidity();
  int status = CDM_CompareAnnotations(getObject(), Ann.getObject());
  if (status > 10) throw OCDM_Exception();
  return status;
} /* OCDM_Annotation::CompareAnnotations */

/**
* \brief This function will check if the Annotation contains in any of its
* spans the specified position.
* \par Description:
* This function will check if the Annotation contains in any of its
* spans the specified position.
*
* \par Arguments:
* \arg \c Position: The position to me checked.
*  
*/
int OCDM_Annotation::AnnotationContainsPosition(const long Position) const {
  ensureValidity();
  int status = CDM_AnnotationContainsPosition(getObject(), Position);
  if (status == -1) throw OCDM_Exception();
  return status;
} /* OCDM_Annotation::AnnotationContainsPosition */

/**
* \brief This function will check if the Annotation contains a span that has
* the specified start and end offsets.
* \par Description:
* This function will check if the Annotation contains a span that has
* the specified start and end offsets.
*
* \par Arguments:
* \arg \c Start: The span start to me matched.
* \arg \c End:   The span end to me matched.
*  
*/
int OCDM_Annotation::AnnotationMatchesRange(const long Start, const long End)
                                                                         const {
  ensureValidity();
  int status = CDM_AnnotationMatchesRange(getObject(), Start, End);
  if (status == -1) throw OCDM_Exception();
  return status;
} /* OCDM_Annotation::AnnotationMatchesRange */

/**
*  \brief This function will return true if an Attribute with the specified
*  name exists in the Annotation object.
*  \par Arguments:
*  \arg \c name: The Attribute name to be found.
*    
*  \note In case of an Error an Exception of type OCDM_Exception
*  will be thrown. 
*
*/
OCDM_BOOL OCDM_Annotation::AttributeExists(const char *name) const {
  ensureValidity();
  if (CDM_GetAttribute(getObject(), (char *) name) != NULL) return true;
  return false;
} /* OCDM_Annotation::AttributeExists */

/**
*  \brief This function displases or "moves" the specified Annotation by 
*  "displacement" characters.
*  \par Description:
*  This function will iterate over all spans contained in the Annotation.
*  For each span all the offsets contained in it (either start or end) 
*  will be examined: 
*  \par Arguments:
*  \arg \c offset: The function checks whether any of this offsets is equal or greater 
*  than the value specified through the "offset" parameter,
*   \arg \c  displacement: If any of this offsets is equal or greater 
*  than the value specified through the "offset" parameter, then the value of the 
*  "displacement" parameter will be added to it.
*
*/
void OCDM_Annotation::DisplaceAnnotation(const long offset,
                                         const long displacement) {
  ensureValidity();
  CDM_Annotation newAnnotation =
                 CDM_DisplaceAnnotation(getObject(), offset, displacement);
  if (newAnnotation  == NULL) throw OCDM_Exception();
  setObject(newAnnotation);
} /* OCDM_Annotation::DisplaceAnnotation */

/**
* \brief This function will search all the Attributes that the 
* Annotation object in use has for an Attribute that has the same
* name as the "Name" parameter.
* \par Description:
* This function will search all the Attributes that the 
* Annotation object in use has for an Attribute that has the same
* name as the "Name" parameter. If such an Attribute is found,
* the Attribute object is return.
* If an Attribute with the specified name does not exist,
* an Exception of type OCDM_Exception will be thrown
* \par Arguments:
* \arg \c name: the name of the attribute to be returned
*/
OCDM_REF(OCDM_Attribute) OCDM_Annotation::GetAttribute(const char *name) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_Attribute>(*this,
                            CDM_GetAttribute(getObject(), (char *) name));
} /* OCDM_Annotation::GetAttribute */

/**
*  \return This function returns a new object (of type OCDM_AttributeSet) that 
*  contains all the Attributes that the specified Annotation object in use 
*  has.
*   
*/
OCDM_REF(OCDM_AttributeSet) OCDM_Annotation::GetAttributes(void) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AttributeSet>(*this,
                                              CDM_GetAttributes(getObject()));
} /* OCDM_Annotation::GetAttributes */

/**
*  \return This function will return the first span range of the span list 
*  contained in the Annotation object in use.  
*  
*/
OCDM_REF(OCDM_Span) OCDM_Annotation::GetFirstSpan(void) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_Span>(*this, CDM_GetFirstSpan(getObject()));
} /* OCDM_Annotation::GetFirstSpan */

/**
*  \return This function will place the values of both span start and end of
*  the Annotation's first span in the two specified long integer pointers. 
*  
*/
int OCDM_Annotation::GetFirstSpanOffsets(long &start, long &end) const {
  ensureValidity();
  int status = CDM_GetSpanOffsets(getObject(), &start, &end);
  if (status != TCL_OK) throw OCDM_Exception();
  return status;
} /* OCDM_Annotation::GetFirstSpanOffsets */

/**
*  \return This function will return the span list 
*  contained in the specified Annotation object in use.   
*
*/
OCDM_REF(OCDM_SpanSet) OCDM_Annotation::GetSpans(void) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_SpanSet>(*this, CDM_GetSpans(getObject()));
} /* OCDM_Annotation::GetSpans */

/**
*  \brief This function adds a new Attribute in the Annotation object
*  we are currently use
*  \note In case of an error an Exception of type OCDM_Exception will be thrown
*  \par Description:
*  This function adds a new Attribute in the Annotation object
*  we are currently use
*  \par Arguments:
*  \arg \c Attr: The new attribute to be added to the Annotation object we are currently use
*  
*/
void OCDM_Annotation::PutAttribute(const OCDM_Attribute &Attr) {
  ensureValidity();
  CDM_Annotation newAnn = CDM_PutAttribute(getObject(), Attr.getObject());
  if (newAnn == NULL) throw OCDM_Exception();
  setObject(newAnn);
} /* OCDM_Annotation::PutAttribute */

/**
*  \return This function returns the type of the Annotation object in use. 
*  The return value will be of type char*  
*  \note The returned value will be encoded using the UTF-8 encoding.
*  
*
*/
const char* OCDM_Annotation::GetType(void) const {
  ensureValidity();
  char *type = CDM_GetType(getObject());
  if (type == NULL) throw OCDM_Exception();
  return type;
} /* OCDM_Annotation::GetType */

/**
*  \brief This function will construct a new Annotation 
*  object an replace the Annotation object in use with the new one
*  \par Description:
*  This function will construct a new Annotation 
*  object an replace the Annotation object in use with the new one
*  The only difference will be that the returned Annotation object
*  will utilise elements from the object cache that is managed internally by CDM. 
*  As a result the returned object will normally require less memory to be 
*  stored than the provided one, as it will re-use elements already in use 
*  by other CDM objects.
*  This function should be used on Annotation objects that originate from 
*  an unknown source.  For example, if an Annotation object is stored in a 
*  file and then red back as a string, it will not use cached by the CDM 
*  objects.  It will be a good idea if the caller created a new object by 
*  calling this function and delete the object red from the file by either 
*  calling Tcl_DecrRefCount or CDM_Free.
* 
*  \note The new Annotation object will be an exact copy of
*  the Annotation object.
*
*/
void OCDM_Annotation::ReconstructAnnotation(void) const {
  ensureValidity();
  CDM_Attribute newAnnotation = CDM_ReconstructAnnotation(getObject());
  if (newAnnotation == NULL) throw OCDM_Exception();
  setObject(newAnnotation);
} /* OCDM_Annotation::ReconstructAnnotation */

/**
*  \brief This function will remove the Attribute having as name the value
*  of the 
*  "Name" parameter from the Annotation object in use, if such an Attribute 
*  exists.
*
*  \par Arguments:
*  \arg \c name: The name of the attribute to be removed.
*
*  \note The altered Annotation will replace the one in use  
*
*/
void OCDM_Annotation::RemoveAttribute(const char *name) {
  ensureValidity();
  CDM_Annotation newAnn = CDM_RemoveAttribute(getObject(), (char *) name);
  if (newAnn == NULL) throw OCDM_Exception();
  setObject(newAnn);
} /* OCDM_Annotation::RemoveAttribute */

/**
*  \par Description:
*  This function will add the provided Span object to the Annotation object in
*  use.  
*  The altered Annotation will replace the one in use.  
*  In the case of an error 
*  an Exception of type OCDM_Exception will be thrown
*  \par Arguments:
*  \arg \c span: The Span object that should be added.
*
*/
void OCDM_Annotation::AddSpan(const OCDM_Span &span) {
  ensureValidity();
  span.ensureValidity();
  CDM_Annotation newAnn = CDM_AddAnnotationSpan(getObject(), span.getObject());
  if (newAnn == NULL) throw OCDM_Exception();
  setObject(newAnn);
} /* OCDM_Annotation::AddSpan */

/**
*  \par Description:
*  This function will remove the Span having as start and end offsets 
*  the values specified through the "start" and "end" parameters 
*  from the Annotation object in use, if such a Span object exists.  
*  The altered Annotation will replace the one in use.  
*  If a Span object with the proper start and end offsets was not found,
*  the the replacemnt will not take place. Finally, in the case of an error 
*  an Exception of type OCDM_Exception will be thrown
*  \par Arguments:
*  \arg \c start: The starting value of the span to be removed
*  \arg \c   end: The ending value of the span to be removed
*
*/
void OCDM_Annotation::RemoveSpan(const long start, const long end) {
  ensureValidity();
  CDM_Annotation newAnn = CDM_RemoveSpan(getObject(), start, end);
  if (newAnn == NULL) throw OCDM_Exception();
  setObject(newAnn);
} /* OCDM_Annotation::RemoveSpan */

/**
*  \brief This function will return the Id of the Annotation object in use.
*  \par Description:
*  This function will return the Id of the Annotation object in use.
*  The Id of an Annotation is a unique integer (of type long) that 
*  unambigiously specifies an Annotation object inside a Document. CDM 
*  ensures that each Annotation of the Annotations contained in a Document 
*  will have a unique Id.
*  
*/
long OCDM_Annotation::GetId(void) const {
  ensureValidity();
  long id = CDM_GetId(getObject());
  if (id < 0) throw OCDM_Exception();
  return id;
} /* OCDM_Annotation::GetId */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_AnnotationSet                                   *******
 ******************************************************************************
 *****************************************************************************/

/**
*  \brief This function will add the given Annotation (specified by the "Ann" 
*  parameter) into the Annotation set object in use.
*  \par Description:
*  This function will add the given Annotation (specified by the "Ann" 
*  parameter) into the Annotation set object in use.
*  This function (unlike OCDM_AddAnnotation for Documents) will not 
*  modify in any way the Id of the provided Annotation: it is the 
*  responsibility of the caller to set a proper Id to the provided 
*  Annotation.  The Annotation object will be appended to the Annotation 
*  set even if it does not have an Id (Id = -1).
*  This function will simply append the provided Annotation object to the 
*  provided Annotation set.  It will not check if an Annotation with the 
*  same Id already exists in order to replace it, as OCDM_AddAnnotation 
*  will do when it is applied on a Document object.
*  \par Arguments:
*  \arg \c Ann: The name of the annotation to be added.
*
*/
void OCDM_AnnotationSet::AddAnnotation(const OCDM_Annotation& Ann) {
  if (getObject() == NULL) {
    setObject(CDM_CreateAnnotationSet());
  }
  CDM_AnnotationSet newAnnSet =
                   CDM_AddAnnotation(getObject(), Ann.getObject());
  if (newAnnSet == NULL) throw OCDM_Exception();
  setObject(newAnnSet);
} /* OCDM_AnnotationSet::AddAnnotation */

/*
 * ---------------------------------------------------------------------------
 *                           O p e r a t o r s 
 * ---------------------------------------------------------------------------
 */

/**
* \brief This operator returns the element of a set pointed by the "index"
* parameter.
* \par Description: This operator returns the element of a set pointed by the
* "index" parameter. Note that under
* languages that do not support operator overloading (i.e. Java, Perl,
* Python) this operator appears as a plain method named as "get".
* \par Arguments:
* \arg \c index: the index (position) of the element to be retrieved. The
* first element of the set has as index the value 0.
* \note In case of an error, an exception of type OCDM_Exception
* will be thrown. 
*/

OCDM_INDEX_OPERATOR_REF(OCDM_Annotation) OCDM_AnnotationSet::operator[] (
                                                       const long index) const {
  return OCDM_GetNth<OCDM_Annotation>(*this, index);
}; /* OCDM_AnnotationSet::operator[] */
/*
 * ---------------------------------------------------------------------------
 *              P r i v a t e   H e l p e r   f u n c t i o n s 
 * ---------------------------------------------------------------------------
 */

/**
*  \brief This function will compare two Annotations, Ann1 and Ann2,
*  according to their Span sets. 
*  \par Arguments:
*  \arg \c Ann1: The first annotation to be compared.  
*  \arg \c Ann2: The second annotation to be compared according to their
*                span set
*
*  \return This function returns -1 if Ann1 is lower than 
*  Ann2, i.e.  Ann1 spans a text range that starts earlier than all the 
*  text ranges spanned by the Ann2 Annotation.  If the two Annotations span 
*  the same text ranges, zero is returned.
*  Finally, 1 is returned if Ann1 is greater than Ann2, i.e.
*  when Ann2 spans a text range starts earlier than all the text ranges spanned 
*  by the Ann1 Annotation.
*
*  \note In case of an error, an Exception of type OCDM_Exception will
*  be thrown and an error message will be left at the current
*  Tcl interpreter (CDM_Interp).
*
*/
int OCDM_AnnotationSet::CompareAnnotations(const OCDM_Annotation &Ann1,
                                           const OCDM_Annotation &Ann2) const {
  ensureValidity();
  int status = CDM_CompareAnnotations(Ann1.getObject(), Ann2.getObject());
  if (status > 10) throw OCDM_Exception();
  return status;
} /* OCDM_AnnotationSet::CompareAnnotations */

/**
*  \brief The goal of this function is to concatenate the Annotation set
*  object in
*  use with the object having being defined in the parameter AnnSet.
*  \par Arguments:
*  \arg \c AnnSet: The name of the annotation to be concatenated.
*  \note In case of an error, an OCDM_Exception will be thrownn
*  and an error message will be left at the current active interpeter 
*  (CDM_Interp).
*
*/
void OCDM_AnnotationSet::MergeAnnotations(OCDM_AnnotationSet &AnnSet) {
  if (getObject() == NULL) {
    setObject(CDM_CreateAnnotationSet());
  }
  CDM_AnnotationSet newAnnSet =
                  CDM_MergeAnnotations(getObject(), AnnSet.getObject());
  if (newAnnSet == NULL) throw OCDM_Exception();
  setObject(newAnnSet);
} /* OCDM_AnnotationSet::MergeAnnotations */

/**
*  \par Description:
*  This function will return a new Annotation set object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  Annotation set in use that have the specified text position (as given 
*  by the parameter "Position") as a start position in any of their spans.  
*  \note This function will always create and return a new Annotation set object, 
*  even when the Annotation set has no Annotations defined or 
*  when no Annotation is found that includes the specified text position 
*  into its span set.
*  (In this case an empty Annotation set will be returned.)
*  
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_AnnotationSet::AnnotationsAt(
                                                    const long position) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsAt(getObject(), position));
} /* OCDM_AnnotationSet::AnnotationsAt */

/**
*  \brief This function will return a new Annotation set object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  specified Annotation set that include the specified text position 
*  (as given by the parameter "Position") into their spans.  
*  \par Description:
*  This function will return a new Annotation set object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  specified Annotation set that include the specified text position 
*  (as given by the parameter "Position") into their spans.  
*  \note This function will always create and return a new Annotation set object, 
*  even when the original Annotations set did not have any Annotations 
*  defined, or when no Annotation is found that includes the specified text 
*  position into its span.  
*  (In this case an empty Annotation set will be returned.)
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_AnnotationSet::AnnotationsContaining(
                                                    const long position) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsContaining(getObject(), position));
} /* OCDM_AnnotationSet::AnnotationsContaining */

/**
*  \brief This function will return a new Annotation set object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  specified Annotation set that include the specified text position 
*  (as given by the parameter "position1" and "position2") into their spans. 
*  \par Arguments:
*  
*  \note This function will always create and return a new Annotation set object, 
*  even when the original Annotations set did not have any Annotations 
*  defined, or when no Annotation is found that includes the specified text 
*  position into its span.  
*  (In this case an empty Annotation set will be returned.)
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_AnnotationSet::AnnotationsContaining(
                             const long position1, const long position2) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsContaining(getObject(),
                                                      position1, position2));
} /* OCDM_AnnotationSet::AnnotationsContaining */

/**
*  \par Description:
*  This function will return a new Annotation set object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  Annotation Set object in use that have at least one span
*  within the specified range (["Start", "End"]). 
*  \par Arguments:
*  \arg \c Start: The starting position of the range to be checked.
*  \arg \c   End: The ending position of th range to be checked.
*
*  \note This function will always create and return a new AnnotationSet
*  object, 
*  even when the original Annotations set did not have any Annotations 
*  defined, or when no Annotation from the given Annotation set 
*  is found to reside inside the specified range.  
*  (In this case an empty Annotation set will be returned.)
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_AnnotationSet::AnnotationsInRange(
                                       const long Start, const long End) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsInRange(getObject(), Start, End));
} /* OCDM_AnnotationSet::AnnotationsInRange */

/**
*  \par Description:
*  This function will return a new Annotation set object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  Annotation Set object in use that have at least one span
*  within the first span of the specified Annotation object. 
*  \par Arguments:
*  \arg \c Ann: The Annotation object whose first span will be used as a search range.
*
*  \note This function will always create and return a new AnnotationSet
*  object, 
*  even when the original Annotations set did not have any Annotations 
*  defined, or when no Annotation from the given Annotation set 
*  is found to reside inside the specified range.  
*  (In this case an empty Annotation set will be returned.)
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_AnnotationSet::AnnotationsInRange(
                                       const OCDM_Annotation& Ann) const {
  ensureValidity();
  long Start, End;
  Ann.GetFirstSpanOffsets(Start, End);
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsInRange(getObject(), Start, End));
} /* OCDM_AnnotationSet::AnnotationsInRange */

/**
*  \par Description:
*  This function will return a new Annotation set object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  specified Annotation Set that have at least one span that exactly 
*  matches the  range (["Start", "End"]). 
*
*  \par Arguments:
*  \arg \c Start: The starting position of the range to be checked.
*  \arg \c   End: The ending position of th range to be checked.
*
*  \note This function will always create and return a new AnnotationSet
*  object, 
*  even when the original Annotations set did not have any Annotations 
*  defined, or when no Annotation from the given Annotation set 
*  is found to reside inside the specified range.  
*  (In this case an empty Annotation set will be returned.)
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_AnnotationSet::AnnotationsMatchingRange(
                                                   long Start, long End) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsMatchingRange(getObject(),
                                                         Start, End));
} /* OCDM_AnnotationSet::AnnotationsMatchingRange */

/**
*  \par Description:
*  This function will return a new Annotation set object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  specified Annotation Set that have at least one span that exactly 
*  matches the first span of the specified Annotation object. 
*
*  \par Arguments:
*  \arg \c Ann: The Annotation object whose first span will be used as a search range.
*
*  \note This function will always create and return a new AnnotationSet
*  object, 
*  even when the original Annotations set did not have any Annotations 
*  defined, or when no Annotation from the given Annotation set 
*  is found to reside inside the specified range.  
*  (In this case an empty Annotation set will be returned.)
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_AnnotationSet::AnnotationsMatchingRange(
                                             const OCDM_Annotation& Ann) const {
  ensureValidity();
  long Start, End;
  Ann.GetFirstSpanOffsets(Start, End);
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsMatchingRange(getObject(),
                                                         Start, End));
} /* OCDM_AnnotationSet::AnnotationsMatchingRange */

/**
*  \par Description:
*  This function will return a new Annotation object (of type
*  OCDM_Annotation) that will be the first Annotation (with Annotations 
*  sorted by their Id) of the
*  Annotation Set that includes the specified text position (as given
*  by the parameter "Position") into its spans.
*
*  \par Arguments:
*  \arg \c Position: Position in text. (long)
*/
OCDM_REF(OCDM_Annotation) OCDM_AnnotationSet::FirstAnnotationContaining(
                                                   const long Position) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_Annotation>(*this,
                            CDM_FirstAnnotationContaining(getObject(),
                                                          Position));
} /* OCDM_AnnotationSet::FirstAnnotationContaining */

/**
*  \par Description:
*  This function will return a new Annotation object (of type
*  OCDM_Annotation) that will be the first Annotation (with Annotations 
*  sorted by their Id) of the
*  Annotation Set that include the two specified text positions
*  (as given by the parameters "Position1" and "Position2")
*  into its spans.
*
*  \par Arguments:
*  \arg \c Position1: Position in text. (long)
*  \arg \c Position2: Position in text. (long)
*/
OCDM_REF(OCDM_Annotation) OCDM_AnnotationSet::FirstAnnotationContaining(
                             const long Position1, const long Position2) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_Annotation>(*this,
                            CDM_FirstAnnotationContaining(getObject(),
                                                         Position1, Position2));
} /* OCDM_AnnotationSet::FirstAnnotationContaining */

/**
*  \brief This function will return the Annotation object that has as Id
*  the value 
*  specified by the "Id" parameter.  If the requested Annotation does not 
*  exist, an Exception of type OCDM_Exception will be thrown.
*  \par Arguments:
*  \arg \c Id:  The id of the annotation to be returned
*
*/
OCDM_REF(OCDM_Annotation) OCDM_AnnotationSet::GetAnnotation(const long Id)
                                                                         const {
  ensureValidity();
  return OCDM_GetReference<OCDM_Annotation>(*this,
                            CDM_GetAnnotation(getObject(), Id));
} /* OCDM_AnnotationSet::GetAnnotation */

/**
*  \brief This function will remove the Annotation object that has
*  as Id the value 
*  specified by the "Id" parameter from the Annotation set object
*  we are currently use.  
*  \par Arguments:
*  \arg \c Id:  The id of the annotation to be removed
*   
*  \note If the provided Annotation set does not contain an Annotation with the 
*  specified Id, an Exception of tpe OCDM_Exception will be thrown
*  signing an error.
*
*/
void OCDM_AnnotationSet::RemoveAnnotation(const long Id) {
  ensureValidity();
  CDM_AnnotationSet newSet = CDM_RemoveAnnotation(getObject(), Id);
  if (newSet == NULL) throw OCDM_Exception();
  setObject(newSet);
} /* OCDM_AnnotationSet::RemoveAnnotation */

/**
*  \par Description:
*  This function will create and return a new object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  Annotation set object in use that their type is the
*  same as the value of the "Type" parameter. 
*  object will be returned.
*  \par Arguments:
*  \arg \c type: The type of the annotations to selected.
*  
*  \note This function will always create and return a new 
*  object, even when no Annotations of the requested type are found or 
*  when the Annotation set is empty.  In this case an empty 
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_AnnotationSet::SelectAnnotations(
                                                       const char* Type) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_SelectAnnotations(getObject(), (char *) Type));
} /* OCDM_AnnotationSet::SelectAnnotations */

/**
*  \par Description:
*  This function will create and return a new object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  specified Annotation set that their type is the same as the value of the 
*  "Type" parameter and the provided OCDM_BOOLean expression (specified through 
*  the "Constraints" parameter) is true for every one of these Annotations.  
*  This function will always create and return a new 
*  object, even when no Annotations of the requested type are found or 
*  when the Annotation set is empty.  In this case an empty 
*  object will be returned.
*  The OCDM_BOOLean expression that can be specified through the "Constraints" 
*  parameter can be any valid Tcl OCDM_BOOLean expression (i.e.  a OCDM_BOOLean 
*  expression that will be accepted by the "expr" Tcl command), with the 
*  following excepion:
*  The OCDM_BOOLean expression can contain references to the values of 
*  Atttributes of the Annotation, by utilising the notation 
*  "ann::<Attribute Name>".  The only limitation is that 
*  "<Attribute Name>" cannot exceed 120 characters.  
*
*  For example, the following code will create a new Annotation set that 
*  will contain all Annotations from the specified Annotation set
*  that their type is 
*  token, 
*  they have a "type" Attribute which has as value "EFW" and also have a 
*  "pos" Attribute that has a value other than "NN": 
*   
* 
*  AnnSet2 =AnnSet1.SelectAnnotations("token",
*  "ann::type == \"EFW\" && ann::pos != \"NN\""); 
*  where AnnSet1 and AnnSet2 are objects of type AnnotationSet 
*
*  \par Arguments:
*  \arg \c        Type: The type of the annotations to selected.
*  \arg \c Constraints: A OCDM_BOOLean expression accordind to which
*          the annotations will be selected 
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_AnnotationSet::SelectAnnotations(
                              const char* Type, const char* Constraints) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_SelectAnnotations(getObject(),
                                      (char *) Type, (char *) Constraints));
} /* OCDM_AnnotationSet::SelectAnnotations */

/**
*  \par Description:
*  This function will create and return a new object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  specified Annotation set that their type is the same as the value of the 
*  "Type" parameter. The Annotations will be sorted according to their 
*  first span range, in ascending order.
*  \par Arguments:
*  \arg \c Type: The type of the annotations to selected.
*  
*  \note This function will always create and return a new 
*  object, even when no Annotations of the requested type are found or 
*  when the Annotation set is empty.  In this case an empty 
*  object will be returned.
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_AnnotationSet::SelectAnnotationsSorted(
                                                       const char* Type) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_SelectAnnotationsSorted(getObject(),
                                                        (char *) Type));
} /* OCDM_AnnotationSet::SelectAnnotationsSorted */

/**
*  \par Description:
*  This function will create and return a new object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  Annotation set object in use that their type is the same as the value of the 
*  "Type" parameter and the provided OCDM_BOOLean expression (specified through 
*  the "Constrains" parameter) is true for every one of these Annotations.  
*  The Annotations will be sorted according to their 
*  first span range, in ascending order.
*  This function will always create and return a new 
*  object, even when no Annotations of the requested type are found or 
*  when the Annotation set contains no Annotations.  In this case an empty 
*  object will be returned.
*
*  The OCDM_BOOLean expression that can be specified through the "Constrains" 
*  parameter can be any valid Tcl OCDM_BOOLean expression (i.e.  a OCDM_BOOLean 
*  expression that will be accepted by the "expr" Tcl command), with the 
*  following excepion:
*  The OCDM_BOOLean expression can contain references to the values of 
*  Atttributes of the Annotation, by utilising the notation 
*  "ann::<Attribute Name>".  The only limitation is that 
*  "<Attribute Name>" cannot exceed 120 characters.  For example, 
*  the following code will create a new Annotation set that will contain 
*  all Annotations from the specified Annotation set that their type is 
*  token, 
*  they have a "type" Attribute which has as value "EFW" and also have a 
*  "pos" Attribute that has a value other than "NN": 
*
*  
*  AnnSet2 =AnnSet1.SelectAnnotationsSorted("token",
*  "ann::type == \"EFW\" && ann::pos != \"NN\""); 
*  where AnnSet1 and AnnSet2 are objects of type AnnotationSet  
*
*  \par Arguments: 
*  \arg \c        Type: The type of the annotations to selected.
*  \arg \c Constraints: A OCDM_BOOLean expression accordind to which the
*          annotations will be selected 
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_AnnotationSet::SelectAnnotationsSorted(
                              const char* Type, const char* Constraints) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_SelectAnnotationsSorted(getObject(),
                                          (char *) Type, (char *) Constraints));
} /* OCDM_AnnotationSet::SelectAnnotationsSorted */

/**
*  \brief This function will sort all the annotations, contained in the  
*  Annotation Set object in use, according to their span ranges. 
*  
*  For more information about the sorting order, please refer to 
*  the CompareAnnotations function.
*
*  In case of an error, an Exception of type OCDM_Exception will be thrown
*
*/
void OCDM_AnnotationSet::SortAnnotationSet(void) {
  ensureValidity();
  CDM_AnnotationSet newAnnSet = CDM_SortAnnotationSet(getObject());
  if (newAnnSet == NULL) throw OCDM_Exception();
  setObject(newAnnSet);
} /* OCDM_AnnotationSet::SortAnnotationSet */

/**
*  \brief This function displases or "moves" the Annotations of the  
*  Annotation set object in use by "displacement" characters.  
*  \par Description:
*  This function displases or "moves" the Annotations of the  
*  Annotation set object in use by "displacement" characters. 
*  This function will iterate over all Annotations contained in
*  the Annotation set.
*  For each Annotation, all spans contained in the span set will be
*  examined: 
*  for each offset in the span (either start or end), if it is equal
*  or greater than the value provided by the "offset" parameter,
*  then the value of the 
*  "displacement" parameter will be added.
*
*  \note In case of an error, an Exception of type OCDM_Exception will be thrown
*/
void OCDM_AnnotationSet::DisplaceAnnotations(const long offset,
                                             const long displacement) {
  ensureValidity();
  CDM_AnnotationSet newAnnSet = 
               CDM_DisplaceAnnotations(getObject(), offset, displacement);
  if (newAnnSet == NULL) throw OCDM_Exception();
  setObject(newAnnSet);
} /* OCDM_AnnotationSet::DisplaceAnnotations */

/**
*  \brief This function returns an Annotation set object (of type 
*  CDM_AnnotationSet).  
*  
*
*  \note In case of an error, an Exception of type OCDM_Exception will be thrown
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_AnnotationSet::NextAnnotations(
                                                    const long Position) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_NextAnnotations(getObject(),Position));
} /* OCDM_AnnotationSet::NextAnnotations */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_Document                                        *******
 ******************************************************************************
 *****************************************************************************/

/*
 * ---------------------------------------------------------------------------
 *              P r i v a t e   H e l p e r   f u n c t i o n s 
 * ---------------------------------------------------------------------------
 */

int OCDM_Document::CheckStatus(const int status) const {
  if (status != TCL_OK) throw OCDM_Exception();
  return status;
} /* OCDM_Document::CheckStatus */

void OCDM_Document::ensureValidity(void) const {
  if (Document == NULL)
    throw OCDM_Exception("object does not hold a Document!");
} /* OCDM_Document::ensureValidity */

const char *OCDM_Document::ReturnString(const char *str) const {
  if (str == NULL) throw OCDM_Exception();
  return str;
} /* OCDM_Document::ReturnString */

const class OCDM_Object *OCDM_Document::getStoredObject(void) const {
  return storedObj;
} /* OCDM_Document::getStoredObject */

void  OCDM_Document::releaseStoredObject(void) const {
  if (storedObj != NULL) delete storedObj;
  storedObj = NULL;
  if (storedCollectionObj != NULL) delete storedCollectionObj;
  storedCollectionObj = NULL;
}; /* OCDM_Document::releaseStoredObject */

void  OCDM_Document::storeObject(const class OCDM_Object *objPtr) const {
  if (storedObj != NULL) delete storedObj;
  storedObj = objPtr;
}; /* OCDM_Document::storeObject */

void  OCDM_Document::storeObject(const class OCDM_Collection *objPtr) const {
  if (storedCollectionObj != NULL) delete storedCollectionObj;
  storedCollectionObj = objPtr;
}; /* OCDM_Document::storeObject */

/*
 * ---------------------------------------------------------------------------
 *                       C o n s t r u c t o r s 
 * ---------------------------------------------------------------------------
 */

/** 
* \brief This is a null constructor of an OCDM_Document Object.
* The purpose of this function is to create a null document,
* just to be used for programming reasons.  
*/
OCDM_Document::OCDM_Document() {
  Document            = NULL;
  storedObj           = NULL;
  storedCollectionObj = NULL;
#ifdef LOG_CONSTRUCTOR_CALLS
  CDM_Log(" OCDM_Document.Document = (NULL); Reference Count = 0\n");
#ifdef LOG_THREAD_INFO
  CDM_Log("Current Thread Id = %p\n", Tcl_GetCurrentThread());
#endif /* LOG_THREAD_INFO */
#endif /* LOG_CONSTRUCTOR_CALLS */
} /* OCDM_Document::OCDM_Document */

/** 
*   \brief This function servers as the main constructor of an
*   OCDM_Document object.
*   \par Arguments:
*   \arg \c TclCmdName: The parameter TclCmdName represents the name of the
*   document to be created.
*
*/
OCDM_Document::OCDM_Document(const char *TclCmdName) {
#ifdef LOG_CONSTRUCTOR_CALLS
  CDM_Log(" OCDM_Document(%s)\n", TclCmdName);
#endif /* LOG_CONSTRUCTOR_CALLS */
  Tcl_Obj *tmp = CDMTCL_NewStringObj(TclCmdName, -1);
  Document            = NULL;
  storedObj           = NULL;
  storedCollectionObj = NULL;
  if (tmp == NULL) 
     throw OCDM_Exception("Tcl_NewStringObj failed: out of memory");
  CDMTCL_IncrRefCount(tmp);
  Document = CDM_GetDocumentPointer(tmp);
  CDMTCL_DecrRefCount(tmp);
  if (Document == NULL) throw OCDM_Exception();
  Document->RefCount++; /* We will close it on the destructor... */
#ifdef LOG_CONSTRUCTOR_CALLS
  CDM_Log(" OCDM_Document.Document = (%p); Reference Count = %d (%s)\n",
                          Document, Document->RefCount, TclCmdName);
#ifdef LOG_THREAD_INFO
  CDM_Log("Current Thread Id = %p\n", Tcl_GetCurrentThread());
#endif /* LOG_THREAD_INFO */
#endif /* LOG_CONSTRUCTOR_CALLS */
} /* OCDM_Document::OCDM_Document */

/**
* \brief This is the default copy constructor.
*/
OCDM_Document::OCDM_Document(const class OCDM_Document& obj) {
  Document            = obj.Document;
  storedObj           = NULL;
  storedCollectionObj = NULL;
  if (Document) {
    Document->RefCount++; /* We will close it on the destructor... */
  }
#ifdef LOG_CONSTRUCTOR_CALLS
  CDM_Log(" OCDM_Document.Document = (%p); Reference Count = %d\n",
                                     Document, Document?Document->RefCount:0);
#ifdef LOG_THREAD_INFO
  CDM_Log("Current Thread Id = %p\n", Tcl_GetCurrentThread());
#endif /* LOG_THREAD_INFO */
#endif /* LOG_CONSTRUCTOR_CALLS */
} /* OCDM_Document::OCDM_Document */

/** 
* \brief This function servers as the main constructor of an
* OCDM_Document object.
* 
* \arg \c TclCmdName: The parameter TclCmdName represents the name of the
* document to be created 
*/
OCDM_Document::OCDM_Document(const Tcl_Obj *TclCmdName) {
  storedObj           = NULL;
  storedCollectionObj = NULL;
  Document  = CDM_GetDocumentPointer( (Tcl_Obj *) TclCmdName);
  if (Document == NULL) throw OCDM_Exception();
  Document->RefCount++; /* We will close it on the destructor... */
#ifdef LOG_CONSTRUCTOR_CALLS
  CDM_Log(" OCDM_Document.Document = (%p); Reference Count = %d\n",
                                           Document, Document->RefCount);
#ifdef LOG_THREAD_INFO
  CDM_Log("Current Thread Id = %p\n", Tcl_GetCurrentThread());
#endif /* LOG_THREAD_INFO */
#endif /* LOG_CONSTRUCTOR_CALLS */
} /* OCDM_Document::OCDM_Document */

/**
* \brief A constructor that maps a CDM_Document to an OCDM_Document object.
*/
OCDM_Document::OCDM_Document(const CDM_Document obj) {
  storedObj           = NULL;
  storedCollectionObj = NULL;
  Document  = obj;
  if (Document) {
    Document->RefCount++; /* We will close it on the destructor... */
  }
#ifdef LOG_CONSTRUCTOR_CALLS
  CDM_Log(" OCDM_Document.Document = (%p); Reference Count = %d\n",
                                     Document, Document?Document->RefCount:0);
#ifdef LOG_THREAD_INFO
  CDM_Log("Current Thread Id = %p\n", Tcl_GetCurrentThread());
#endif /* LOG_THREAD_INFO */
#endif /* LOG_CONSTRUCTOR_CALLS */
} /* OCDM_Document::OCDM_Document */

/*
 * ---------------------------------------------------------------------------
 *                           O p e r a t o r s 
 * ---------------------------------------------------------------------------
 */

/**
 * \brief This is the default Assignment operator.
 */
class OCDM_Document& OCDM_Document::operator= (const class OCDM_Document& obj) {
  releaseStoredObject();
  /* We will close it on the destructor... */
  if (obj.Document) obj.Document->RefCount++;
  if (Document) CDM_Close(Document);
  Document = obj.Document;
#ifdef LOG_CONSTRUCTOR_CALLS
  CDM_Log(" OCDM_Document::operator= (%p); Reference Count = %d\n",
                                     Document, Document?Document->RefCount:0);
#ifdef LOG_THREAD_INFO
  CDM_Log("Current Thread Id = %p\n", Tcl_GetCurrentThread());
#endif /* LOG_THREAD_INFO */
#endif /* LOG_CONSTRUCTOR_CALLS */
  return *this;
} /* OCDM_Document::operator = */

/*
 * ---------------------------------------------------------------------------
 *                         D e s t r u c t o r s 
 * ---------------------------------------------------------------------------
 */

/** 
* \brief This is the destructor of an OCDM_document object. 
* \par Description:
* The goal of this function is to close the given (opened)
* Document object. All the 
* objects that the Document to be closed owns, as well the Document 
* object itself, will be deleted from memory and will be unregistered 
* from the current Tcl interpreter (CDM_Interp). Note that this
* function will not save the Document object in use. 
* In order for the Document to be saved, the caller must previously
* use the function Sync of the Class OCDM_Document.
*/
OCDM_Document::~OCDM_Document() {
#ifdef LOG_DESTRUCTOR_CALLS
  CDM_Log("~OCDM_Document(%p); ", Document);
#endif /* LOG_DESTRUCTOR_CALLS */
  if (Document) CDM_Close(Document);
  Document = NULL;
  releaseStoredObject();
#ifdef LOG_DESTRUCTOR_CALLS
  CDM_Log("~OCDM_Document();\n");
#endif /* LOG_DESTRUCTOR_CALLS */
} /* OCDM_Document::~OCDM_Document */

/** 
* \brief This method will remove a Document from memory.. 
* \par Description:
* This method can be used to force a Document object to release the memory
* hold by the Document structure, in order to save memory.
* The Document object will not be released, but after calling this method will
* contain an invalid internal representation: any method (except the destructor)
* called after Close has beeen called will throw an exception.
* Note that this
* function will not save the Document object in use. 
* In order for the Document to be saved, the caller must previously
* use the function Sync of the Class OCDM_Document.
* Usually, the user does not have to call this method. A Document will be closed
* when the object representing this Document will be deleted. However, there are
* situation where such a method is need: in java for example, the user does not
* have control on when an object will be deleted. This method allows the
* programmer to release the memory occupied internally by the object and let the
* object get garbage-collected at a later time.
*/
void OCDM_Document::Close(void) {
  if (Document) CDM_Close(Document);
  Document = NULL;
} /* OCDM_Document::~OCDM_Document */

/*
 * ---------------------------------------------------------------------------
 *               P u b l i c   H e l p e r   f u n c t i o n s 
 * ---------------------------------------------------------------------------
 */

/**
*  \arg \c
*  This function will add the given Annotation (specified by the "Ann" 
*  parameter) into the Document object that we are currently use.
*  If the Annotation does not already have an Id, a new unique Id will be 
*  assigned to that Annotation and the Annotation will be added to the 
*  current Annotation set of the Document.
*  If the Annotation already has an Id, then this Id will left intact.  
*  If an Annotation with the same Id already exists in the current 
*  Annotation set, then the new Annotation will replace the old one. If 
*  an Annotation with the same Id does not exist, the given Annotation will 
*  be appended to the existing Annotation set.
*  
*  \par Arguments
*  \arg \c Ann: The Annotation object to be added.
*  \note This function returns the Id of the Annotation that was added to the 
*  Document object in use. Whether the Annotation had or had not an Id 
*  makes no difference to the returned value.  Note that after using this 
*  function, CDM will keep a reference to the given Annotation.  This will 
*  transform the given Annotation into a shared 
*  object, that cannot be manipulated any more by the caller.  Always 
*  remember that an attempt to modify a shared object will abort the 
*  execution of the whole platform.  If the caller wants to modify a shared 
*  object, the following steps must be followed: 
*  A check should be made in order to find if an object is shared 
*  or not, by using the Tcl function Tcl_IsShared. 
*  If the object is not shared (i.e.  Tcl_IsShared returns an 
*  integer value < 2), the object can be safely modified. 
*  If the object is shared, a new copy of the object must be 
*  created by using the Tcl function Tcl_DuplicateObj.  The new object will 
*  be an exact copy of the original one but it will not be shared.  The 
*  returned object can be safely modified.
*  In case of an error the return value will be -1 and 
*  an error message describing the error will be left at the current Tcl 
*  interpreter (CDM_Interp).
*/
long OCDM_Document::AddAnnotation(const OCDM_Annotation &Ann) {
  ensureValidity();
  long id = CDM_AddAnnotation(Document, Ann.getObject());
  if (id < 0) throw OCDM_Exception();
  return id;
} /* OCDM_Document::AddAnnotation */

/** 
*   \par Description:
*   This function will return a new Annotation set object (of type 
*   OCDM_AnnotationSet) that will contain all the Annotations of the 
*   current Document that have the specified text position (as given 
*   by the parameter "position") as a start position in any of their
*   spans.  
*   \par Arguments:
*   \arg \c position: it specifies the text position
*
*   \note This function will always create and return a new Annotation set
*   object,even when the Document has no Annotations defined or when
*   no Annotation 
*   is found that includes the specified text position into its span set.  
*   (In this case an empty Annotation set will be returned.)
*   In case of an error, NULL will be return and an error message will be 
*   left at the current active Tcl interpreter (CDM_Interp).
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_Document::AnnotationsAt(const long position)
                                                                         const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsAt(Document, position));
} /* OCDM_Document::AnnotationsAt*/

/** 
*   \par Description:
*   This function will return a new Annotation set object (of type 
*   OCDM_AnnotationSet) that will contain all the Annotations of the 
*   Document object that include the specified text position (as given 
*   by the parameter "Position") into their spans.
*   In this new release of Ellogon, the user does not 
*   have to manualy free the returned object as well as the memory 
*   associated with it (for example by the use  of the
*   Tcl_DecrRefCount or CDM_Free) since java performs it own
*   garbage collection methods.   
*   
*   \note This function will always create and return a new Annotation
*   set object, 
*   even when the Document has no Annotations defined or when no Annotation 
*   is found that includes the specified text position into its span.  
*   (In this case an empty Annotation set will be returned.)
*   In case of an error, NULL will be return and an error message will be 
*   left at the current active Tcl interpreter (CDM_Interp).
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_Document::AnnotationsContaining(
                                                    const long position) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsContaining(Document, position));
} /* OCDM_Document::AnnotationsContaining */

/** 
*   \par Description:
*   This function will return a new Annotation set object (of type 
*   OCDM_AnnotationSet) that will contain all the Annotations of the 
*   Document object that include the specified text position (as given 
*   by the parameters "Position1" and "Position2") into their spans.
*   In this new release of Ellogon, the user does not 
*   have to manualy free the returned object as well as the memory 
*   associated with it (for example by the use  of the
*   Tcl_DecrRefCount or CDM_Free) since java performs it own
*   garbage collection methods.   
*   
*   \note This function will always create and return a new Annotation
*   set object, even when the Document has no Annotations defined or
*   when no Annotation 
*   is found that includes the specified text position into its span.  
*   (In this case an empty Annotation set will be returned.)
*   In case of an error, NULL will be return and an error message will be 
*   left at the current active Tcl interpreter (CDM_Interp).
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_Document::AnnotationsContaining(
                             const long position1, const long position2) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsContaining(Document,
                                                      position1, position2));
} /* OCDM_Document::AnnotationsContaining */

/**  
*   \par Description:
*   This function will return a new Annotation set object (of type 
*   OCDM_AnnotationSet) that will contain all the Annotations of the 
*   Document object in use that have at least one span within
*   the specified range (["Start", "End"]). In this new release of Ellogon,
*   the user does not have to manualy free the returned object as well
*   as the memory 
*   associated with it (for example by the use of the Tcl_DecrRefCount
*   or CDM_Free) 
*   since java performs it own garbage collection methods.     
*   \par Arguments:
*   \arg \c Start: The starting position of the range to be checked.
*   \arg \c   End: The ending position of the range to be checked.
*   \note This function will always create and return a new AnnotationSet
*   object, 
*   even when the Document has no Annotations defined or
*   when no Annotation is found that resides inside the specified range.  
*   (In this case an empty Annotation set will be returned.)
*   In case of an error, NULL will be return and an error message will be 
*   left at the current active Tcl interpreter (CDM_Interp).
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_Document::AnnotationsInRange(const long Start,
                                                         const long End) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsInRange(Document, Start, End));
} /* OCDM_Document::AnnotationsInRange */

/**  
*   \par Description:
*   This function will return a new Annotation set object (of type 
*   OCDM_AnnotationSet) that will contain all the Annotations of the 
*   Document object in use that have at least one span within
*   the first span of the specified Annotation object.     
*   \par Arguments:
*  \arg \c Ann: The Annotation object whose first span will be used as a search range.
*   \note This function will always create and return a new AnnotationSet
*   object, 
*   even when the Document has no Annotations defined or
*   when no Annotation is found that resides inside the specified range.  
*   (In this case an empty Annotation set will be returned.)
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_Document::AnnotationsInRange(
                                             const OCDM_Annotation& Ann) const {
  ensureValidity();
  long Start, End;
  Ann.GetFirstSpanOffsets(Start, End);
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsInRange(Document, Start, End));
} /* OCDM_Document::AnnotationsInRange */

/**
*  \par Description:
*  This function will return a new Annotation set object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  Document object in use, that have at least one span that 
*  exactly matches the specified range (["Start", "End"]). 
*  \par Arguments:
*  \arg \c Start: It defines the starting range 
*  \arg \c Start: It defines the starting range 
*  In this new release of Ellogon, the user does not 
*  have to manualy free the returned object as well as the memory 
*  associated with it (for example by the use of the
*  Tcl_DecrRefCount or CDM_Free) 
*  since java performs it own garbage collection methods.  
*  \note This function will always create and return a new AnnotationSet object, 
*  even when the Document has no Annotations defined or when no Annotation 
*  is found that resides inside the specified range.  
*  (In this case an empty Annotation set will be returned.)
*  In case of an error, NULL will be return and an error message will be 
*  left at the current active Tcl interpreter (CDM_Interp).
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_Document::AnnotationsMatchingRange(
                                       const long Start, const long End) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsMatchingRange(Document, Start, End));
} /* OCDM_Document::AnnotationsMatchingRange */

/**
*  \par Description:
*  This function will return a new Annotation set object (of type 
*  OCDM_AnnotationSet) that will contain all the Annotations of the 
*  Document object in use, that have at least one span that 
*  exactly matches the first span of the specified Annotation object. 
*  \par Arguments:
*  \arg \c Ann: The Annotation object whose first span will be used as a search range.
*  \note This function will always create and return a new AnnotationSet
*  object, 
*  even when the Document has no Annotations defined or when no Annotation 
*  is found that resides inside the specified range.  
*  (In this case an empty Annotation set will be returned.)
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_Document::AnnotationsMatchingRange(
                                             const OCDM_Annotation& Ann) const {
  ensureValidity();
  long Start, End;
  Ann.GetFirstSpanOffsets(Start, End);
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_AnnotationsMatchingRange(Document, Start, End));
} /* OCDM_Document::AnnotationsMatchingRange */

/**
*  \brief This function will return true if an Attribute with the specified
*  name exists in the Document object.
*  \par Arguments:
*  \arg \c name: The Attribute name to be found.
*    
*  \note In case of an Error an Exception of type OCDM_Exception
*  will be thrown. 
*
*/
OCDM_BOOL OCDM_Document::AttributeExists(const char *name) const {
  ensureValidity();
  if (CDM_GetAttribute(Document, (char *) name) != NULL) return true;
  return false;
} /* OCDM_Document::AttributeExists */

/**
*   \par Description:
*   This function displases or "moves" the Annotations of the Document
*   object in use by "displacement" characters.  This function will 
*   iterate over all Annotations contained in the Document.  
*   For each Annotation, all spans contained in the span set
*   will be examined: 
*   for each offset in the span (either start or end),
*   if it is equal or greater
*   than the value provided by the "offset" parameter,
*   then the value of the 
*   "displacement" parameter will be added.
*
*    \note In case of illegal manipulation, an exception of type
*    OCDM_Exception will be
*    thrown
*/
int OCDM_Document::DisplaceAnnotations(long offset, long displacement) {
  ensureValidity();
  return CheckStatus(CDM_DisplaceAnnotations(Document, offset, displacement));
} /* OCDM_Document::DisplaceAnnotations */

/**
*  \par Description:
*  This function will return a new Annotation object (of type
*  OCDM_Annotation) that will be the first Annotation (with Annotations 
*  sorted by their Id) of the
*  Document that includes the specified text position (as given
*  by the parameter "Position") into its spans.
*
*  \par Arguments:
*  \arg \c Position: Position in text. (long)
*/
OCDM_REF(OCDM_Annotation) OCDM_Document::FirstAnnotationContaining(
                                                   const long Position) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_Annotation>(*this,
                            CDM_FirstAnnotationContaining(Document, Position));
} /* OCDM_Document::FirstAnnotationContaining */

/**
*  \par Description:
*  This function will return a new Annotation object (of type
*  OCDM_Annotation) that will be the first Annotation (with Annotations 
*  sorted by their Id) of the
*  Document that include the two specified text positions
*  (as given by the parameters "Position1" and "Position2")
*  into its spans.
*
*  \par Arguments:
*  \arg \c Position1: Position in text. (long)
*  \arg \c Position2: Position in text. (long)
*/
OCDM_REF(OCDM_Annotation) OCDM_Document::FirstAnnotationContaining(
                             const long Position1, const long Position2) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_Annotation>(*this,
                            CDM_FirstAnnotationContaining(Document, Position1,
                                                                    Position2));
} /* OCDM_Document::FirstAnnotationContaining */

/**  
* \brief This function will return the set of all text portions annotated by
* the provided Annotation object.
*/
OCDM_REF(OCDM_RawDataSet) OCDM_Document::GetAnnotatedTextRanges(
                                             const OCDM_Annotation& Ann) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_RawDataSet>(*this,
                            CDM_GetAnnotatedTextRanges(Document,
                                                       Ann.getObject()));
} /* OCDM_Document::GetAnnotatedTextRanges */

/**  
*  \brief This function will return the Annotation object that has as
*  Id the value 
*  specified by the "Id" parameter. 
*  \note In case that the specified Annotation does not exist, NULL is returned.
*/
OCDM_REF(OCDM_Annotation) OCDM_Document::GetAnnotation(const long Id) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_Annotation>(*this,
                            CDM_GetAnnotation(Document, Id));
} /* OCDM_Document::GetAnnotation */

/**
*  \par Description:
*  This function will search all the Attributes that the
*  Document object in use has, for an Attribute that has the same name as the 
*  "Name" parameter. If such an Attribute is found, the Attribute object is 
*  return.  
*  \par Arguments:
*  \arg \c name: The name of the attribute to be returned from the Document.
*  \note If an Attribute with the specified name does not exist,
*  NULL is  returned.
*  
*/
OCDM_REF(OCDM_Attribute) OCDM_Document::GetAttribute(const char *name) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_Attribute>(*this,
                            CDM_GetAttribute(Document, (char *) name));
} /* OCDM_Document::GetAttribute */

/**
*  \par Description:
*  This function returns a new object (of type OCDM_AttributeSet) that 
*  contains all the Attributes that the specified Document
*  object currently has. 
*   
*  \return This function will always create and return a new Attribute
*  set object, even when the Document has no Attributes defined. 
*  (In this case an empty Attribute set will be returned...)
*/
OCDM_REF(OCDM_AttributeSet) OCDM_Document::GetAttributes(void) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AttributeSet>(*this,
                            CDM_GetAttributes(Document));
} /* OCDM_Document::GetAttributes */

/**
*  \brief This function will return the text range from the text of the 
*  Document object we are currently using that is annotated by
*  the FIRST span of the provided Annotation.  
*  \par Description:
*  This function will return the text range from the text of the 
*  Document object we are currently using that is annotated by
*  the FIRST span of the provided Annotation. 
*  In simple words, this function will retrieve the first span from the 
*  provided Annotation and return the text range (from the text of the 
*  Document object) that is indexed by this span.
*  \par Arguments:
*  \arg \c Ann: The provided Annotation object 
*  
*  \note In this new release of Ellogon, the user does not 
*  have to manualy free the returned object as well as the memory 
*  associated with it (for example by the use  of the
*  Tcl_DecrRefCount or CDM_Free) 
*  since java performs it own garbage collection methods.     
*/
OCDM_REF(OCDM_ByteSequence) OCDM_Document::GetFirstAnnotatedTextRange(
                                             const OCDM_Annotation& Ann) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_ByteSequence>(*this,
                            CDM_GetFirstAnnotatedTextRange(Document,
                                                           Ann.getObject()));
} /* OCDM_Document::GetFirstAnnotatedTextRange */

/**
*   \par Description
*   This function will return a new Annotation set object (of type 
*   OCDM_AnnotationSet) that will contain all the Annotations of the 
*   Document, for which the document object was created, that
*   their spans ranges are after the specified text position 
*   Bare in mind that he user does not have to manualy free the returned 
*   object as well as the memory associated with it 
*   (for example by the use  of the Tcl_DecrRefCount or CDM_Free) 
*   since java performs it own garbage collection methods.  
*   as given by the parameter "Position").
*   \par Arguments
*   \arg \c Position: The specified text position
*
*   \note This function will always create and return a new
*   Annotation set object, 
*   even when the Document has no Annotations defined or no Annotations 
*   exist that all their spans are beneath the specified text position.  
*   In this case an empty Annotation set will be returned.
*   In case of an error, NULL will be return and an error message will be 
*   left at the current active Tcl interpreter (CDM_Interp).
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_Document::NextAnnotations(const long Position)
                                                                         const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_NextAnnotations(Document, Position));
} /* OCDM_Document::NextAnnotations */

/**
*   \brief This function will add a given Attribute to the specified Document 
*   object. 
*   \par Description
*   This function will add a given Attribute to the specified Document 
*   object. If an Attribute with the same name already exists in the 
*   Document, then it will be overwritten by the new Attribute. Else, the 
*   new Attribute will be appended to the existing Attribute set of the 
*   specified Document.
*   \par Arguments
*   \arg \c Attr: The attribute to be added to the Document in use.
*
*   \note In case of illegal manipulation, an exception of type
*   OCDM_Exception will be thrown
*/
int OCDM_Document::PutAttribute(const OCDM_Attribute &Attr) {
  ensureValidity();
  return CheckStatus(CDM_PutAttribute(Document, Attr.getObject()));
} /* OCDM_Document::PutAttribute */

/**
*   \brief As the name implies this function will remove the Annotation object 
*   that has as Id the value specified by the "Id" parameter.
*   \par Description
*   This function removes the Annotation object
*   that has as Id the value specified by the "Id" parameter.
*   \par Arguments
*   \arg \c Id: The Id value of the parameter to be removed 
*   \note If the requested Annotation does not exist, an exception
*   of type OCDM_Exception 
*   will be thrown.
*/
int OCDM_Document::RemoveAnnotation(const long Id) {
  ensureValidity();
  return CheckStatus(CDM_RemoveAnnotation(Document, Id));
} /* OCDM_Document::RemoveAnnotation */

/**
*   \brief This function will return the Id of the Document object in use.
*   \par Description
*   This function returns the Id of the Document object in use.
*   The Id is the file name of the initial file 
*   that contained the text which was used in order
*   to create the Document object. 
*     
*   \note The returned value will be encoded using the UTF-8 encoding 
*   (thus enabling the existance of non Latin characters in the value).
*/
const char* OCDM_Document::GetId(void) const {
  ensureValidity();
  return ReturnString(CDM_GetId(Document));
} /* OCDM_Document::GetId */

/**
*   \brief This function will remove the Attribute named exactly as the "Name" 
*   parameter.
*   \par Arguments
*   \arg \c name: The name of the attribute to be removed from the document.
*   
*   \note If the requested Attribute does not exist, an exception 
*   of type OCDM_Exception will be thrown
*   
*/
int OCDM_Document::RemoveAttribute(const char *name) {
  ensureValidity();
  return CheckStatus(CDM_RemoveAttribute(Document, (char *) name));
} /* OCDM_Document::RemoveAttribute */

/**
*   \par Description
*   This function will create and return a new object (of type 
*   OCDM_AnnotationSet) that will contain all the Annotations of the 
*   Document object in use that their type is the same as the value of the 
*   "Type" parameter.
*   \par Arguments
*   \arg \c Type: The type of the Annotations to be selected from the document.
*   \note This function will always create and return a new 
*   object, even when no Annotations of the requested type are found or 
*   when the Document has no Annotations defined. In this case an empty 
*   object will be returned.
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_Document::SelectAnnotations(const char* Type)
                                                                         const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_SelectAnnotations(Document, (char *) Type));
} /* OCDM_Document::SelectAnnotations */

/**
*  \par Description
*   This function is an overloaded version of the above one.
*   The difference between them lies in the use of the parameter Constraints
*   In other words this function will create and return
*   a new object (of type OCDM_AnnotationSet) that will contain
*   all the Annotations of the 
*   Document that their type is the same as the value of the 
*   "Type" parameter and the provided OCDM_BOOLean expression (specified through 
*   the "Constraints" parameter) is true for every one of these Annotations. 
*   \par Arguments
*   \arg \c        Type: The type of the Annotations to be selected from
*   the document.
*   \arg \c Constraints: A OCDM_BOOLean expression.
*   It should become true in order for the annotation to be selected.
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_Document::SelectAnnotations(const char* Type,
                                      const char* Constraints) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_SelectAnnotations(Document, (char *) Type,
                                                  (char *) Constraints));
} /* OCDM_Document::SelectAnnotations */

/** 
*  \par Description
*  This function will create and return a new object (of type 
*  ODM_AnnotationSet) that will contain all the Annotations of the 
*  Document object in use that their type is the same as the value of the 
*  "Type" parameter.  The Annotations will be sorted according to their 
*  first span range, in ascending order.
*  
*  This function will always create and return a new 
*  object, even when no Annotations of the requested type are found or 
*  when the Document has no Annotations defined.  In this case an empty 
*  object will be returned.
*  \par Arguments
*  \arg \c Type: The type of the Annotations to be selected from the document.
*  
*  \note In this new release of Ellogon, the user does not 
*  have to manualy free the returned object as well as the memory 
*  associated with it (for example by the use of the
*  Tcl_DecrRefCount or CDM_Free) 
*  since java performs it own garbage collection methods.
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_Document::SelectAnnotationsSorted(
                                                       const char* Type) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_SelectAnnotationsSorted(Document,
                                                        (char *) Type));
} /* OCDM_Document::SelectAnnotationsSorted */

/**
*  \par Description
*  The difference between this function and the above one is the use of the 
*  constraint parameter.
*  As mentioned before this function will create and return a new object
*  (of type OCDM_AnnotationSet) that will contain all the Annotations of the 
*  Document object in use that their type is the same as the value of the 
*  "Type" parameter and the provided OCDM_BOOLean expression (specified through 
*  the "Constrains" parameter) is true for every one of these Annotations.  
*  The Annotations will be sorted according to their 
*  first span range, in ascending order.
*
*  The OCDM_BOOLean expression that can be specified through the "Constrains" 
*  parameter can be any valid Tcl OCDM_BOOLean expression (i.e.  a OCDM_BOOLean 
*  expression that will be accepted by the "expr" Tcl command), with the 
*  following exception:
*  The OCDM_BOOLean expression can contain references to the values of 
*  Atttributes of the Annotation, by utilising the notation 
*  "ann::<Attribute Name>".  The only limitation is that 
*  "<Attribute Name>" cannot exceed 120 characters.
*  For example,  the following code will create a new Annotation set 
*  that will contain all Annotations from the Document
*  object in use that their
*  type is token, they have a "type" Attribute which has as value "EFW" 
*  and also have a "pos" Attribute that has a value other than "NN":
*  \par Arguments
*  \arg \c Type: The type of the Annotations to be selected from the document.
*  \arg \c Constraints: A OCDM_BOOLean expression.
*  It should become true in order for the annotation to be selected.
*  \note This function will always create and return a new 
*  object, even when no Annotations of the requested type are found or 
*  when the Document object has no Annotations defined.
*  In this case an empty 
*  object will be returned.
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_Document::SelectAnnotationsSorted(
                                            const char* Type,
                                            const char* Constraints) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_SelectAnnotationsSorted(Document,
                                (char *) Type, (char *) Constraints));
} /* OCDM_Document::SelectAnnotationsSorted */

/**
*   \par Description
*   This function will modify the text contained in the Document object in use, 
*   by inserting a string at the positions specified by the "pos" 
*   parameter.  This position "pos" is the number of characters that have to 
*   be skipped before the string is inserted, where the first character of 
*   the Document text has a position equal to 0.  If position is lower than 
*   zero, it is treaded as the last character of the Document 
*   text.
*
*   For example, if the Document contains as text the 
*   string "This is a test", the following code:
* 
*         CDM_ByteSequenceInsertString(Doc, 10, "simple "); 
*   
*   will modify the text of the Document to 
*   "This is a simple test".
*
*   The most important aspect of this function is that it will make the 
*   proper arrangements so as the Annotations contained inside the Document 
*   to be displaced by a proper amound of characters so as Annotations 
*   spanning text ranges after the insertion point to span the same text 
*   portions, after the string has been inserted.
*   \par Arguments
*   \arg \c   pos: The position for the string to be inserted.
*   \arg \c sting: The string sequence itself.
*
*   \note This function will return the new Document text (as an object of type 
*   OCDM_ByteSequence) if the insertion was succesful. 
*   In case of an error, an exception of type OCDM_Exception will be 
*   thrown. 
*          
*/
OCDM_REF(OCDM_ByteSequence) OCDM_Document::ByteSequenceInsertString(
                                           const long pos, const char *string) {
  ensureValidity();
  return OCDM_GetReference<OCDM_ByteSequence>(*this,
                            CDM_ByteSequenceInsertString(Document,
                                                         pos, (char *) string));
} /* OCDM_Document::ByteSequenceInsertString */

/**
*   \par Description
*   This function will modify the text contained in the Document object in use
*   by removing a range of consecutive characters from the Document Text, 
*   starting with the character whose index is the value of the parameter 
*   "first" and ending with the character whose index is the value of the 
*   parameter "last".  An index of 0 refers to the first character of the 
*   Document Text, while a negative value (i.e.  -1) refers to the last 
*   character of the Document Text.  If the value of the parameter "newstring" 
*   is not NULL, then it is placed in the removed character range.  
*   If "first" is less than zero then it is treated as if it was zero, and 
*   if "last" is greater than the length of the Document Text 
*   then it is treated as if it was equal to the length of the Document Text.  
*   It is an error if "first" is greater than "last".
*   
*   The most important aspect of this function is that it will make the 
*   proper arrangements so as the Annotations contained inside the Document 
*   to be displaced by a proper amound of characters so as Annotations 
*   spanning text ranges after the replacement point to span the same text 
*   portions, after the string has replaced the desired characters.
*   \par Arguments
*   \arg \c first: The starting range of the of the consecutive characters
*   to be replaced with the newstring.
*   \arg \c last: The ending range of the of the consecutive characters
*   to be replaced with the newstring.
*   \arg \c newstring: the newstring string itself.
*
*   \note This function will return the new Document text (as an object of type 
*   OCDM_ByteSequence) if the replacement was succesful.  
*   
*   In case of an error, an exception of type OCDM_Exception will be 
*   thrown. 
*
*/
OCDM_REF(OCDM_ByteSequence) OCDM_Document::ByteSequenceReplace(const long first,
                                       const long last, const char *newstring) {
  ensureValidity();
  return OCDM_GetReference<OCDM_ByteSequence>(*this,
                            CDM_ByteSequenceReplace(Document,
                                              first, last, (char *) newstring));
} /* OCDM_Document::ByteSequenceReplace */

/**
*   \par Description
*   This function will modify the text contained in the Document object in use,
*   by replacing the specified number of characters at the position 
*   specified by the "pos" parameter by the provided string.  
*   This position "pos" is the number of characters that have to 
*   be skipped before the characters are replaced by the string, 
*   where the first character of  the Document text has a position equal to 0. 
*   If position is lower than zero, it is treaded as the last character of 
*   the Document text. In the case the last character is specified as position, then 
*   the last "characters" characters are replaced.
*   For example, if the Document contains as text the 
*   string "This is a test", the following code:
* 
*          ByteSequenceReplaceCharacters(10, 4, "simple test"); 
*   
*   will modify the text of the Document in use to "This is a simple test".
*   The most important aspect of this function is that it will make the 
*   proper arrangements so as the Annotations contained inside the Document 
*   to be displaced by a proper amound of characters so as Annotations 
*   spanning text ranges after the replacement point to span the same text 
*   portions, after the string has replaced the desired characters.
*   
*    \return This function will return the new Document text (as an object of type 
*   OCDM_ByteSequence) if the replacement was succesful.  
*/
OCDM_REF(OCDM_ByteSequence) OCDM_Document::ByteSequenceReplaceCharacters(
                        const long first, const long last, const char *string) {
  ensureValidity();
  return OCDM_GetReference<OCDM_ByteSequence>(*this,
                            CDM_ByteSequenceReplaceCharacters(Document,
                                                 first, last, (char *) string));
} /* OCDM_Document::ByteSequenceReplaceCharacters */

/**
*   \par Description
*   This function accepts an argument of type const char :
*   This is a UTF-8 string that defines the Annotations to be deleted. 
*   According to the value specified by the "Type" parameter,
*   this function will delete Annotations or Annotation Attributes from 
*   the specified Document.
*
*   If the value of the "Type" parameter is an empty string or NULL, then 
*   ALL the Annotations that the specified Document currently has will 
*   be deleted from this Document object. 
*   The used Annotation Ids will be also reseted.  
*   This means that the first Annotation that will be added to this 
*   Document will take the value "0" as Annotation Id, if it does not 
*   already have an Id.  If the deletion is done successfully, 
*   the total number of the deleted Annotations will be returned.
*
*   If Annotations having as type the value of the "Type" parameter exist, 
*   then all these Annotations will be deleted from the Document.  If the 
*   deletion is done successfully, the total number of the deleted 
*   Annotations will be returned.
*  
*   If all Annotations have been searched and no Annotation having as type 
*   the value of the "Type" parameter was found, the value of the "Type" 
*   parameter is examined whether it can be splitted in two words.   If this 
*   is the case (the value of the "Type" parameter contains at least one 
*   space character), then the value is splitted in two parts on the first 
*   space character that is found.  The first word will be used as an 
*   Annotation type and the rest of the value (second word) 
*   will be used as an Attribute 
*   name.  All Annotations will be seached.  If Annotations having as type 
*   the first word are found, they are examined whether they contain an 
*   Attribute named as the second word.  If such an Attribute is found in an 
*   Annotation, it will be deleted from the Annotation.  The total number of 
*   deleted Attributes will be returned as the return value of this 
*   function.
* 
*   \note If this function fails to delete any Annotation or any Annotation 
*   Attribute, then the value "0" will be deleted.  In case of an error, 
*   an exception of type OCDM_Exception will be thrown.
*
*/
int OCDM_Document::DeleteAnnotations(const char *Type) {
  ensureValidity();
  int status = CDM_DeleteAnnotations(Document, (char *) Type);
  if (status < 0) throw OCDM_Exception();
  return status;
} /* OCDM_Document::DeleteAnnotations */

/**
*   \par Description
*   This function is an oveloaded version of the above one
*   It accepts two arguments: a UTF-8 string  that defines the Annotations 
*   to be deleted and a OCDM_BOOLean expression.
*   This expression must be true in order for an Annotation to be deleted.  
*   According to the value specified by the "Type" parameter, this function 
*   will delete Annotations or Annotation Attributes from the specified 
*   Document.
*
*   If all Annotations have been searched and no Annotation having as type 
*   the value of the "Type" parameter was found, the value of the "Type" 
*   parameter is examined whether it can be splitted in two words. If this 
*   is the case (the value of the "Type" parameter contains at least one 
*   space character), then the value is splitted in two parts on the first 
*   space character that is found.  The first word will be used as an 
*   Annotation type and the rest of the value (second word) 
*   will be used as an Attribute name. 
*   
*   All Annotations will be seached. If Annotations having as type 
*   the first word are found, they are examined whether they contain an 
*   Attribute named as the second word.  If such an Attribute is found in an 
*   Annotation and the OCDM_BOOLean expression is true for this Annotation, 
*   it will be deleted from the Annotation.  The total number of 
*   deleted Attributes will be returned as the return value of this 
*   function.
*
*   The OCDM_BOOLean expression that can be specified through the "Constraints" 
*   parameter will be evaluated before an Annotation or an Annotation 
*   Attribute is deleted.  If the expression is true, then the deletion will 
*   be done.  Else, this Annotation or Annotation Attribute will not be 
*   deleted.  This expression can be any valid Tcl OCDM_BOOLean expression (i.e.  a 
*   OCDM_BOOLean expression that will be accepted by the "expr" Tcl command), 
*   with the following excepion:
*   The OCDM_BOOLean expression can contain references to the values of 
*   Atttributes of the Annotation, by utilising the notation 
*   "ann::<Attribute Name>".  The only limitation is that 
*   "<Attribute Name>" cannot exceed 120 characters.  For example, 
*   the following code will delete 
*   all Annotations from the specified Document that their type is token, 
*   they have a "type" Attribute which has as value "EFW" and also have a 
*   "pos" Attribute that has a value other than "NN": 
*   
* 
*   DeleteAnnotations("token", "ann::type == \"EFW\" && ann::pos != \"NN\""); 
*   
*
*   If the value of the "Type" parameter is an empty string or NULL, then 
*   all Annotations will be searched for deletion, regardless their type.  
*   If the value of the "Constraints" parameter is an empty string or NULL, 
*   then the expression is ignored and the deletions will be based only on 
*   the value of the "Type" parameter.  
*   Note that if all Annotations are deleted from the Document object in use 
*   then the set of utilised Ids will also be reseted.  As a result, 
*   the first Annotation that will be added to this Document after 
*   the deletion will take the value "0" as Annotation Id, if it does not 
*   already have an Id.  
*   If any deletion are done successfully, the total number of the deleted 
*   (or modified) Annotations will be returned.
*   If this function fails to delete any Annotation or any Annotation 
*   Attribute, then the value "0" will be deleted. In case of an error, 
*   an exception of type OCDM_Exception will be thrown.
*
*/
int OCDM_Document::DeleteAnnotations(const char *Type,const char *Constraints) {
  ensureValidity();
  int status = CDM_DeleteAnnotations(Document, (char *) Type,
                                               (char *) Constraints);
  if (status < 0) throw OCDM_Exception();
  return status;
} /* OCDM_Document::DeleteAnnotations */

/**
*   \par Description
*   This function will examine the Annotations stored in the Document object
*   in use to find the maximum used Annotation Id. This function will 
*   make the proper arrangements so as the next Annotation that will be 
*   added to this Document to receive as Id the following Id after the 
*   identified maximum Ids.
*
*/
long OCDM_Document::FindMaxUsedAnnotationId(void) const {
  ensureValidity();
  long id = CDM_FindMaxUsedAnnotationId(Document);
  if (id < 0) throw OCDM_Exception();
  return id;
} /* OCDM_Document::FindMaxUsedAnnotationId */

/**
*   \par Description
*   This function will create and return a new object 
*   (of type CDM_AnnotationSet) that contains all the Annotations 
*   that the Document object in use.  
*
*   \note This function will always create and return a new Annotation 
*   set object, even when the Document has no Annotations defined.  
*   (In this case an empty Annotation set will be returned.)
*
*/
OCDM_REF(OCDM_AnnotationSet) OCDM_Document::GetAnnotations(void) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AnnotationSet>(*this,
                            CDM_GetAnnotations(Document));
} /* OCDM_Document::GetAnnotations */

/**
*   \par Description
*   This function will return the Text of the specified Document. It will 
*   return an object (of type OCDM_ByteSequence) that will contain the text 
*   of the Document object.
*/
OCDM_REF(OCDM_ByteSequence) OCDM_Document::GetByteSequence(void) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_ByteSequence>(*this,
                            CDM_GetByteSequence(Document));
} /* OCDM_Document::GetByteSequence */

/**
*   \par Description
*   This function will return the Text of the Document object in use It will 
*   return a C pointer (of type char*) that will contain the
*   text of the Document.
*   \par Arguments
*   \arg \c encoding: The text will be stored in a newly allocated memory 
*   space, using the requested by the value of the "Encoding" parameter 
*   encoding.  The pointer to that new memory segment will be returned to 
*   the caller.
*/
const char *OCDM_Document::GetByteSequence(const char *encoding) const {
  ensureValidity();
  return ReturnString(CDM_GetByteSequence(Document, (char *) encoding));
} /* OCDM_Document::GetByteSequence */

/**
*   \par Description
*   This function will return the encoding of the Document object in useThe 
*   return value will be of type char* and will be owned by the
*   new version of CDM. 
*   Its value will be a standart Tcl encoding value
*   (like iso8859-7 or cp1253).
*   For all available Tcl encodings please refer to the Tcl manuals.
*
*/
const char *OCDM_Document::GetEncoding(void) const {
  ensureValidity();
  return ReturnString(CDM_GetEncoding(Document));
} /* OCDM_Document::GetEncoding */

/**
*   \par Description
*   This function will return the external Id (XID) of the
*   Document object in use.
*   The external Id is the (absolute) path of the initial file 
*   that contained the text which was used in order to create
*   the Document object object in use.
*   The returned value will be encoded using the UTF-8 encoding 
*   (thus enabling the existance of non Latin characters in the value).
*   The return value will be of type char* and will be owned by the CDM.
*
*/
const char *OCDM_Document::GetExternalId(void) const {
  ensureValidity();
  return ReturnString(CDM_GetExternalId(Document));
} /* OCDM_Document::GetExternalId */

/**
*   \par Description
*   This function will start an iteration over the Annotations of the 
*   specified type that the Document object in use has. This function will 
*   return the first Annotation object that has as type the value 
*   specified by the "Type" parameter. If the requested Annotation does not 
*   exist, an error will be returned. Subsequent calls to 
*   OCDM_GetNextAnnotation will return the rest of the Annotations of the 
*   same type, sorted according to their Id.
*   An exception of type OCDM_Exception will be thrown if an error occurs.
*
*/
OCDM_REF(OCDM_Annotation) OCDM_Document::GetFirstAnnotation(const char *Type)
                                                                        const {
  ensureValidity();
  return OCDM_GetReference<OCDM_Annotation>(*this,
                            CDM_GetFirstAnnotation(Document, (char *) Type));
} /* OCDM_Document::GetFirstAnnotation */

/**
*   \par Description
*   This function will return the next Annotation of the iteration started 
*   with a call to OCDM_GetFirstAnnotation, or of the Annotation returned by 
*   a previous call to OCDM_GetNextAnnotation. If no more Annotations of the 
*   same type (which was specified through a call to OCDM_GetFirstAnnotation) 
*   exist in the Document object in use, NULL will be returned.
*   An exception of type 
*   OCDM_Exception will be thrown if an error occurs.
*
*/
OCDM_REF(OCDM_Annotation) OCDM_Document::GetNextAnnotation(void) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_Annotation>(*this,
                            CDM_GetNextAnnotation(Document));
} /* OCDM_Document::GetNextAnnotation */

/**
*  \brief This function will return the parent of the specified 
*  Document object in use. 
*  \par Description
*  his function will return the parent of the specified 
*  Document object in use. The parent of a Document is the
*  Collection object that holds the Document object in use.
*  The return value will be the parent Collection 
*  object (of type OCDM_Collection) and will be owned by CDM.
*
*/
OCDM_REF(OCDM_Collection) OCDM_Document::GetParent(void) const {
  CDM_Collection parent = CDM_GetParent(Document);
  if (parent == NULL) throw OCDM_Exception();
  return OCDM_GetReference<OCDM_Collection>(*this, parent);
} /* OCDM_Document::GetParent */

/**
*  \brief This function will return the Text of the Document object in use. 
*  \par Description
*  This function will return the Text of the Document object in use.
*  It will return an object (of type OCDM_RawData) that will contain the text 
*  of the Document object. The caller can use the Tcl function Tcl_GetString in 
*  order to get the text stored in the returned object as a C pointer (of 
*  type char*). The text will be stored using the UTF-8 encoding.  
*  In order to convertthe returned C pointer into a specific encoding 
*  (like iso8859-1 or iso8859-7), the function CDM_UtfToExternal can be used.
*
*/
OCDM_REF(OCDM_RawData) OCDM_Document::GetRawData(void) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_RawData>(*this, CDM_GetRawData(Document));
} /* OCDM_Document::GetRawData */

/**
*  \par Description
*  This function will return a C pointer (of type char*) that 
*  will contain the text (or byte sequence or raw data) of the Document
*  object in use. 
*  \par Arguments
*  \arg \ encoding: the parameter specifing the requested encoding.  
*
*/
const char *OCDM_Document::GetRawData(const char *encoding) const {
  ensureValidity();
  return ReturnString(CDM_GetRawData(Document, (char *) encoding));
} /* OCDM_Document::GetRawData */

/**
*  \brief This function will change the text of a Document object to the text 
*  specified by the "Text" parameter. 
*  \par Description
*  This function will change the text of a Document object to the text 
*  specified by the "Text" parameter.The "Text" parameter must be an 
*  object of type OCDM_ByteSequence. Such an object can be created either
*  by the use of Tcl_NewStringObj or by creating a OCDM_ByteSequence object 
*  using the default constructor. The text that this function accepts, is assumed to be 
*  encoding according to the UTF-8 encoding.  If the text is stored using a 
*  different encoding, the caller must use CDM_ExternalToUtf in order to 
*  convert the text to UTF. CDM will internally keep a reference to this 
*  object.
*
*/
int OCDM_Document::SetByteSequence(const OCDM_ByteSequence &Text) {
  ensureValidity();
  return CheckStatus(CDM_SetByteSequence(Document, Text.getObject()));
} /* OCDM_Document::SetByteSequence */

/**
*  \brief Updates the Percent Bar displayed when an Annotator runs...  
*
*/
void OCDM_Document::SetProcessStatus(const int value) const {
  ensureValidity();
  CDM_SetProcessStatus(Document, value);
} /* OCDM_Document::SetProcessStatus */

/**
*  \par Description
*  This function will change the encoding of a Document object to the 
*  encoding specified by the "Encoding" parameter.
*
*/
const char *OCDM_Document::SetEncoding(const char *encoding) {
  ensureValidity();
  return ReturnString(CDM_SetEncoding(Document, (char *) encoding));
} /* OCDM_Document::SetEncoding */

/**
*  \brief  This function will modify the external Id (XID) of the
*  Document object in use
*  \par Description
*  This function will modify the external Id (XID) of the Document object in use
*  The external Id is the (absolute) path of the initial file 
*  that contained the text which was used in order to create the Document. 
*  The value is expected to be encoded using the UTF-8 encoding 
*  (thus enabling the existance of non Latin characters in the value).  The 
*  caller can use the function CDM_ExternalToUft in order to convert a 
*  string from any supported encoding to UTF.
*
*/
const char *OCDM_Document::SetExternalId(const char *ExternalId) {
  ensureValidity();
  return ReturnString(CDM_SetExternalId(Document, (char *) ExternalId));
} /* OCDM_Document::SetExternalId */

/**
*  \brief This function will return information about the Document
*  object in use.  
*  \par Description
*  This function will return information about the Document object in use.  
*  This information will be stored as text in a newly created Tcl object.  
*
*  This function will always create and return a new Tcl string object.  
*  In case of an error, an exception of type OCDM_Exception will br thrown
*
*/
OCDM_REF(OCDM_ByteSequence) OCDM_Document::Status(void) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_ByteSequence>(*this, CDM_Status(Document));
} /* OCDM_Document::Status */

/**
*  \brief This function will save the Document object in disk.
*  \par Description
*  This function will save the Document object in disk.
*  If the Document object has never been saved again, 
*  then this function will create a file according to the Id the Document in use.
*  A representation of the Document will be saved in this file.  If the file 
*  already exists it will be overwritten. If the file cannot be created an error condition will be returned
*
*/
int OCDM_Document::Sync(void) const {
  ensureValidity();
  return CheckStatus(CDM_Sync(Document));
} /* OCDM_Document::Sync */

/**
*  \par Description
*  This function will modify the text contained in the Document object in use, 
*  by removing a range of consecutive characters from the Text, 
*  starting with the character whose index is the value of the parameter 
*  "first" and ending with the character whose index is the value of the 
*  parameter "last".  An index of 0 refers to the first character of the 
*  Document's Text, while a negative value (i.e.  -1) refers to the last 
*  character of the Document's Text. If the value of the parameter "newstring" 
*  is not NULL, then it is placed in the removed character range.  
*  If "first" is less than zero then it is treated as if it was zero, and 
*  if "last" is greater than the length of the Document Text 
*  then it is treated as if it was equal to the length of the Document's Text.  
*
*  It is an error if "first" is greater than "last".
*  The most important aspect of this function is that it will make the 
*  proper arrangements so as the Annotations contained inside the Document 
*  to be displaced by a proper amound of characters so as Annotations 
*  spanning text ranges after the replacement point to span the same text 
*  portions, after the string has replaced the desired characters.
*  This function will return the new Document text (as an object of type 
*  OCDM_RawData) if the replacement was succesful.  
*
*/
OCDM_REF(OCDM_RawData) OCDM_Document::RawDataReplace(const long first,
                                       const long last, const char *newstring) {
  ensureValidity();
  return OCDM_GetReference<OCDM_RawData>(*this, CDM_RawDataReplace(Document,
                                              first, last, (char *) newstring));
} /* OCDM_Document::RawDataReplace */

/**
*  \par Description
*  This function will modify the text contained in the Document object in use, 
*  by inserting a string at the positions specified by the "pos" 
*  parameter.  This position "pos" is the number of characters that have to 
*  be skipped before the string is inserted, where the first character of 
*  the Document's text has a position equal to 0.  If position is lower than 
*  zero, it is treaded as the last character of the Document 
*  text.
*
*  For example, if the Document in use contains as text the 
*  string "This is a test", the following code:
*
*          RawDataInsertString(10, "simple "); 
*   
*  will modify the text of the Document to 
*  "This is a simple test".
*  The most important aspect of this function is that it will make the 
*  proper arrangements so as the Annotations contained inside the Document 
*  to be displaced by a proper amound of characters so as Annotations 
*  spanning text ranges after the insertion point to span the same text 
*  portions, after the string has been inserted.
*
*  \par Arguments
*  \arg \c    pos: The position of the string to be inserted
*  \arg \c string: The string itself
*
*  \return This function will return the new Document's text (as an object of type 
*  OCDM_ByteSequence) if the insertion was succesful. 
*  
*/
OCDM_REF(OCDM_RawData) OCDM_Document::RawDataInsertString(const long pos,
                                                           const char *string) {
  ensureValidity();
  return OCDM_GetReference<OCDM_RawData>(*this, 
                            CDM_RawDataInsertString(Document,
                                                    pos, (char *) string));
} /* OCDM_Document::RawDataInsertString */

/**
*  \par Description
*  This function will modify the text contained in the Document object in use, 
*  by replacing the specified number of characters at the position 
*  specified by the "pos" parameter by the provided string.  
*  This position "pos" is the number of characters that have to 
*  be skipped before the characters are replaced by the string, 
*  where the first character of 
*  the Document's text has a position equal to 0.  If position is lower than 
*  zero, it is treaded as the last character of the Document 
*  text.  In the case the last character is specified as position, then 
*  the last "characters" characters are replaced.
*   
*  For example, if the Document contains as text the 
*  string "This is a test", the following code:
*
*          RawDataReplaceCharacters(10, 4, "simple test"); 
*   
*  will modify the text of the Document to "This is a simple test".
*
*  The most important aspect of this function is that it will make the 
*  proper arrangements so as the Annotations contained inside the Document 
*  to be displaced by a proper amound of characters so as Annotations 
*  spanning text ranges after the replacement point to span the same text 
*  portions, after the string has replaced the desired characters.
*
*  \return This function will return the new Document text (as an object of type 
*  OCDM_RawData) if the replacement was succesful.  
*  
*  \note In case of an error, an exception of type OCDM_Exception
*  will be thrown.
*
*/
OCDM_REF(OCDM_RawData) OCDM_Document::RawDataReplaceCharacters(const long first,
                                     const int characters, const char *string) {
  ensureValidity();
  return OCDM_GetReference<OCDM_RawData>(*this,
                            CDM_RawDataReplaceCharacters(Document, first,
                                                  characters, (char *) string));
} /* OCDM_Document::RawDataReplaceCharacters */

/**
* \brief This method logs information. It is equivalent to
* OCDM_Utilities::Log()
*/
void OCDM_Document::Log(const char *str, ...) const {
  CDM_Log(str);
}; /* OCDM_Document::Log */

/**
*  \brief As the name implies this function checks for a valid Document.
*  \par Description
*  As the name implies this function checks for a valid Document.
*  In other words, if a Document exists it returns a true OCDM_BOOLean variable.
*  Otherwise it returns false
*
*/
OCDM_BOOL OCDM_Document::Valid(void) const {
  if (Document) return true;
  return false;
} /* OCDM_Document::Valid */

/**
*  \brief Return object as a formatted string.
*
*/
const char *OCDM_Document::toString(void) const {
  if (Document) {
    return CDMTCL_GetCommandName(Document->Interp, Document->Tcl_Cmd);
  }
   return "";
} /* OCDM_Document::toString */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_Collection                                      *******
 ******************************************************************************
 *****************************************************************************/

/*
 * ---------------------------------------------------------------------------
 *              P r i v a t e   H e l p e r   f u n c t i o n s 
 * ---------------------------------------------------------------------------
 */

int OCDM_Collection::CheckStatus(const int status) const {
  if (status != TCL_OK) throw OCDM_Exception();
  return status;
} /* OCDM_Collection::CheckStatus */

void OCDM_Collection::ensureValidity(void) const {
  if (Collection == NULL)
    throw OCDM_Exception("object does not hold a Collection!");
} /* OCDM_Collection::ensureValidity */

const char *OCDM_Collection::ReturnString(const char *str) const {
  if (str == NULL) throw OCDM_Exception();
  return str;
} /* OCDM_Collection::ReturnString */

const class OCDM_Object *OCDM_Collection::getStoredObject(void) const {
  return storedObj;
} /* OCDM_Collection::getStoredObject */

void  OCDM_Collection::releaseStoredObject(void) const {
  if (storedObj != NULL) delete storedObj;
  storedObj         = NULL;
  if (storedDocumentObj != NULL) delete storedDocumentObj;
  storedDocumentObj = NULL;
}; /* OCDM_Collection::releaseStoredObject */

void  OCDM_Collection::storeObject(const class OCDM_Object *objPtr) const {
  if (storedObj != NULL) delete storedObj;
  storedObj = objPtr;
}; /* OCDM_Collection::storeObject */

void  OCDM_Collection::storeObject(const class OCDM_Document *objPtr) const {
  if (storedDocumentObj != NULL) delete storedObj;
  storedDocumentObj = objPtr;
}; /* OCDM_Collection::storeObject */
/*
 * ---------------------------------------------------------------------------
 *                       C o n s t r u c t o r s 
 * ---------------------------------------------------------------------------
 */

/**
*  \brief This is a null constructor of an OCDM_Document Object.
*  \par Description
*  The purpose of this function is to create a null document, just to be used
*  for programming reasons.    
*
*/
OCDM_Collection::OCDM_Collection() {
  Collection        = NULL;
  storedObj         = NULL;
  storedDocumentObj = NULL;
} /* OCDM_Collection::OCDM_Collection */

/**
*  \brief This is the overloaded version of the above constructor 
*   with a different usability
*   
*  \par Description
*  \par Arguments
*  \arg \c name: It is the Collection's name which must be the absolute path of 
*  the Collection on disk,If name is of the form "Col*", then it refers to an 
*  already open Collection. 
*  Else, we assume that name is a path, and the Collection is loaded
*  from the disk...
*
*/
OCDM_Collection::OCDM_Collection(const char *name) {
  storedObj         = NULL;
  storedDocumentObj = NULL;
  // This is a double constructor: if name is of the form "Col*", then it
  // refers to an already open Collection. Else, we assume that name is a
  // path, and the Collection is loaded from the disk...
  if (CDM_Strncmp((char *)name, "Col", 3) == 0) {
    // This is an already opened Collection reference...
    Tcl_Obj *tmp = CDMTCL_NewStringObj(name, -1);
    Collection = CDM_GetCollectionPointer(tmp);
    CDMTCL_DecrRefCount(tmp);
    if (Collection == NULL) throw OCDM_Exception();
    Collection->RefCount++; /* We will close it on the destructor... */
    return;
  }
  // Request to open a Collection from disk...
  Collection = CDM_OpenCollection((char *) name);
  if (Collection == NULL) throw OCDM_Exception();
} /* OCDM_Collection::OCDM_Collection */

/**
* \brief This is the default Copy constructor.
*/
OCDM_Collection::OCDM_Collection(const class OCDM_Collection& obj) {
  storedObj         = NULL;
  storedDocumentObj = NULL;
  Collection = obj.Collection;
  if (Collection) {
    Collection->RefCount++; /* We will close it on the destructor... */
  }
} /* OCDM_Collection::OCDM_Collection */

/**
*  \brief This is a constructor of the class OCDM_Collection.
*  \par Description
*  It's main purpose is to create a new Collection object. It accepts three 
*  arguments: the Collection's name which must be the absolute path of 
*  the Collection on disk, an object of type OCDM_AttributeSet, representing 
*  a set of Attributes that will be inserted into the new Collection 
*  and the encoding of the Collection. 
*  The Collection's encoding will be the default encoding that will be used 
*  when a new Document will be created in this Collection. 
*  Note that each Document can have a different encoding than the encoding of
*  the container Collection. The only place that this encoding will be used, is in the 
*  case where a new Document of unspecified encoding will be created.
*  If the Collection encoding argument is ommited, it defaults to 
*  "CDM_DefaultEncoding".  "CDM_DefaultEncoding" is a global variable (of 
*  type char *), that has as initial value the system encoding.
*
*  This variable is also exported at the Tcl level under the same name and as a result 
*  its value can easily be changed by the end user.
*  
*  \par Arguments
*  \arg \c name: It is the Collection's name which must be the absolute path of 
*  the Collection on disk,
*  \arg \c AttrSet: It is an object of type OCDM_AttributeSet which represents
*  a set of Attributes that will be inserted into the new Collection
*  \arg \c encoding: The encoding of the Collection 
*
*  \note In the new Collection will be created only in memory. Nothing 
*  will be saved on the disk.  In order for the Collection to be saved in 
*  disk, the function CDM_Sync must be called after the new Collection is 
*  created in memory.  This is usually done after all of the desired 
*  Documents have been added to the newly created Collection. 
*
*/
OCDM_Collection::OCDM_Collection(const char* name,
                       const OCDM_AttributeSet &AttrSet, const char* encoding) {
  storedObj         = NULL;
  storedDocumentObj = NULL;
  Collection = CDM_CreateCollection((char *) name, AttrSet.getObject(),
                                    (char *) encoding);
} /* OCDM_Collection::OCDM_Collection */

/**
* \brief A constructor that maps a CDM_Collection to an OCDM_Collection object.
*/
OCDM_Collection::OCDM_Collection(const CDM_Collection obj) {
  storedObj         = NULL;
  storedDocumentObj = NULL;
  Collection = obj;
  if (Collection) {
    Collection->RefCount++; /* We will close it on the destructor... */
  }
} /* OCDM_Collection::OCDM_Collection */

/*
 * ---------------------------------------------------------------------------
 *                           O p e r a t o r s 
 * ---------------------------------------------------------------------------
 */

/**
 * \brief This is the default Assignment operator.
 */
class OCDM_Collection& OCDM_Collection::operator= (
                                             const class OCDM_Collection& obj) {
  releaseStoredObject();
  /* We will close it on the destructor... */
  if (obj.Collection) obj.Collection->RefCount++;
  if (Collection) CDM_Close(Collection);
  Collection = obj.Collection;
  return *this;
} /* OCDM_Collection::operator = */

/*
 * ---------------------------------------------------------------------------
 *                         D e s t r u c t o r s 
 * ---------------------------------------------------------------------------
 */

/**
*  \brief This is the destructor of the OCDM_Collection Class. 
*  \par Description
*  It closes the given Collection and frees all the memory occupied by the 
*  Collection and its Documents.  All the deleted objects will also be 
*  unregistered from the current Tcl interpreter (CDM_Interp).
*
*/
OCDM_Collection::~OCDM_Collection() {
#ifdef LOG_DESTRUCTOR_CALLS
  CDM_Log("~OCDM_Collection(%p); ", Collection);
#endif /* LOG_DESTRUCTOR_CALLS */
  if (Collection) CDM_Close(Collection);
  Collection = NULL;
  releaseStoredObject();
#ifdef LOG_DESTRUCTOR_CALLS
  CDM_Log("~OCDM_Collection();\n");
#endif /* LOG_DESTRUCTOR_CALLS */
} /* OCDM_Collection::~OCDM_Collection */

/*
 * ---------------------------------------------------------------------------
 *               P u b l i c   H e l p e r   f u n c t i o n s 
 * ---------------------------------------------------------------------------
 */

/**
*  \brief This function will return true if an Attribute with the specified
*  name exists in the Collection object.
*  \par Arguments:
*  \arg \c name: The Attribute name to be found.
*    
*  \note In case of an Error an Exception of type OCDM_Exception
*  will be thrown. 
*
*/
OCDM_BOOL OCDM_Collection::AttributeExists(const char *name) const {
  ensureValidity();
  if (CDM_GetAttribute(Collection, (char *) name) != NULL) return true;
  return false;
} /* OCDM_Collection::AttributeExists */

/**
*  \brief This function creates a new Document object in an existing Collection 
*  object. 
*  \par Description
*  This function creates a new Document object in an existing Collection 
*  object.
*  \par Arguments
*  It accepts four arguments: 
*  \arg \c XID: The External Id of the Document 
*  (which represents the full path in disk of the original file that contains 
*  the Document's text)
*  \arg \c RawData: the text (RawData or ByteSequence) of the new Document, 
*  \arg \c Annotations: An initial Annotation set,
*  \arg \c Attributes: An  initial Attribute set 
*  \arg \c Encoding: The encoding of the Document.
*
*
*  The External Id must be a valid (absolute) path.  A local copy of this 
*  argument will be created for use by this function.  The existance of 
*  this path will not be checked.  This path value will be 
*  converted automatically to a valid platform dependant path.  As a result, 
*  for this value can be safely used the notation used for paths under the 
*  unix operating system.  That means that the path: 
*   
*  /Users/petasis/Collections 
*  will be converted internally to the following path, assuming windows as 
*  operating system : 
*   
*
*   C:\\Users\\petasis\\Collections
*
* 
*  The RawData parameter must contain the desired for the new Document 
*  text, in UTF-8 format.  A local copy of this variable will be also 
*  created.  If the text is not in UTF-8 (i.e.  contains text just red from 
*  a file using the standart C library routines and is in the ISO 8859-7 
*  encoding) then must be converted to UTF using CDM_ExternalToUtf. If 
*  the origin of this string is Tcl, then it is already in UTF format, as 
*  Tcl uses ONLY UTF for encoding strings internally.
*  The Annotations parameter must be of type OCDM_AnnotationSet.Note that 
*  a reference to this object will also be kept.
*  The Attributes parameter must hold a valid Attribute set created,
*  A reference to this object will also be kept.
*  Finally, an encoding can be specified.  A local copy of this 
*  string will be created.  The value of this parameter must be a standart 
*  Tcl encoding value (like iso8859-7 or cp1253).  For all available Tcl 
*  encodings please refer to the Tcl manuals.  
*  If this parameter is ommitted, then a default value will be used. 
*  This value will be inherited from the parent Collection object.  
*  It is important for the encoding to correctly describe the text. 
*  If the given UTF string cannot be converted to the 
*  Document's encoding, then the text will be filled with the character "?" 
*  in places where the conversion will fail.
*
*
*/
void OCDM_Collection::CreateDocument(const char *XID,
                    const OCDM_ByteSequence &RawData,
                    const OCDM_AnnotationSet &Annotations,
                    const OCDM_AttributeSet  &Attributes,
                    const char *encoding) {
  CDM_CreateDocument(Collection, (char *) XID,
                            RawData.getObject(),
                            Annotations.getObject(),
                            Attributes.getObject(), (char *) encoding);
} /* OCDM_Collection::CreateDocument */

/**
*  
*  \ brief This is an overloaded version of the above one. The only difference is that 
*  it does not hold encoding information. 
*
*/
void OCDM_Collection::CreateDocument(const char *XID,
                    const OCDM_ByteSequence &RawData,
                    const OCDM_AnnotationSet &Annotations,
                    const OCDM_AttributeSet &Attributes) {
  CDM_CreateDocument(Collection, (char *) XID,
                            RawData.getObject(),
                            Annotations.getObject(),
                            Attributes.getObject());
} /* OCDM_Collection::CreateDocument */

/**
*  \brief This function will return the Length (the number of all the Documents) 
*  of the specified Collection.
*
*/
int OCDM_Collection::Length(void) const {
  ensureValidity();
  int len = CDM_Length(Collection);
  if (len < 0) throw OCDM_Exception();
  return len;
} /* OCDM_Collection::Length */

/**
* \par Description
* This function will search all the Attributes that exist in the Collection
* object in use
* \return It will return an OCDM_Attribute object containing all the
* attributes having
* the same name as the "Name" parameter.
* If an Attribute with the specified name does not exist,
* an Exception of type OCDM_exception will be thrown.
* \par Arguments
* \arg \c name: The name of the attribute to be returned
*/
OCDM_REF(OCDM_Attribute) OCDM_Collection::GetAttribute(const char *name) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_Attribute>(*this,
                            CDM_GetAttribute(Collection, (char *) name));
} /* OCDM_Collection::GetAttribute */

/**
*  \par Description
*  This function returns a new object (of type OCDM_AttributeSet) that 
*  contains all the Attributes that the specified Collection currently 
*  has. 
*
*  \note This function will always create and return a new Attribute set object, 
*  even when the Collection has no Attributes defined.  (In this case an 
*  empty Attribute set will be returned.)
*
*/
OCDM_REF(OCDM_AttributeSet) OCDM_Collection::GetAttributes(void) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_AttributeSet>(*this,
                            CDM_GetAttributes(Collection));
} /* OCDM_Collection::GetAttributes */

/**
*  \par Description
*  This function will add a given Attribute to the specified Collection 
*  object. If an Attribute with the same name already exists in the Collection,
*  then it will be overwritten by the new Attribute.  Else, the 
*  new Attribute will be appended to the existing Attribute set of the 
*  specified Collection.
*  \par Arguments
*  \arg \c Attr: The name of the attribute to be inserted in the collection.
*
*/
int OCDM_Collection::PutAttribute(const OCDM_Attribute &Attr) {
  ensureValidity();
  return CheckStatus(CDM_PutAttribute(Collection, Attr.getObject()));
} /* OCDM_Collection::PutAttribute */

/**
*  \brief This function will return the Name of the specified Collection.
*  \par Description
*  \return This function will return the Name of the specified Collection.
*  This value will be the (absolute) full path to the directory that contains 
*  the disk representation of the specified Collection.  The returned value 
*  will be encoded using the UTF-8 encoding (thus enabling the 
*  existance of non Latin characters in the value).
*  
*/
const char *OCDM_Collection::GetName(void) const {
  ensureValidity();
  return ReturnString(CDM_GetName(Collection));
} /* OCDM_Collection::GetName */

/**
*  \par Description
*  This function will remove the Attribute named exactly as the "Name" 
*  parameter. If the requested Attribute does not exist, an Exception of type 
*  OCDM_exception will be thrown
*  \par Description
*  \arg \c name: The name of the attribute to be removed
*
*/
int OCDM_Collection::RemoveAttribute(const char *name) {
  ensureValidity();
  return CheckStatus(CDM_RemoveAttribute(Collection, (char *) name));
} /* OCDM_Collection::RemoveAttribute */

/**
*  \par Description
*  This function will return the encoding of the specified Collection. The 
*  return value will be of type char* and will be owned by CDM. Its value 
*  will be a standard Tcl encoding value (like iso8859-7 or cp1253).  For 
*  all available Tcl encodings please refer to the Tcl manuals.
*/
const char *OCDM_Collection::GetEncoding(void) const {
  ensureValidity();
  return ReturnString(CDM_GetEncoding(Collection));
} /* OCDM_Collection::GetEncoding */

/**
*  \par Description
*  This function will change the encoding of a Collection object to the 
*  encoding specified by the "Encoding" parameter.
*  The function will return a pointer (of type char*) to the string 
*  buffer that holds the new encoding value. This pointer will be at a 
*  different memory location than the given parameter value, as the CDM 
*  will create and manipulate a local copy.
*  In case of an error, an exception of type OCDM_Exception will be thrown
*  Note that the returned pointer is property of the CDM and should never 
*  be freed or modified in any way by the caller.
*
*/
const char *OCDM_Collection::SetEncoding(const char *encoding) {
  ensureValidity();
  return ReturnString(CDM_SetEncoding(Collection, (char *) encoding));
} /* OCDM_Collection::SetEncoding */

/**
*  \par Description
*  This function will return information about the specified Collection.
*  This information will be stored as text in a newly created OCDM_ByteSequence
*  object.
*  
*  \note In case of an error, an exception of type OCDM_Exception will be thrown.
*  
*/
OCDM_REF(OCDM_ByteSequence) OCDM_Collection::Status(void) const {
  ensureValidity();
  return OCDM_GetReference<OCDM_ByteSequence>(*this, CDM_Status(Collection));
} /* OCDM_Collection::Status */

/**
*  \par Description
*  This function will save a Collection object in disk.  If the Collection 
*  object has never been saved again, then this function will create the 
*  directory that the Name of the Collection specifies.  A representation 
*  of the Collection will be saved in this directory. If the directory 
*  already exists it will be overwritten.  If the directory cannot be 
*  created an an exception of type OCDM_Exception will be thrown..
*  
*  \note The return value from this function will be a standard Tcl completion 
*  code (of type int) with one of the values TCL_OK and TCL_ERROR. If the 
*  requested Collection is successfully saved, TCL_OK will be returned. 
*  In case of an error, an exception of type OCDM_Exception will be thrown.
*
*/
int OCDM_Collection::Sync(void) const {
  ensureValidity();
  return CheckStatus(CDM_Sync(Collection));
} /* OCDM_Collection::Sync */

/**
*  This function will run the specified annotator over the given 
*  Collection...  [Not Implemented]
*
*/
int OCDM_Collection::AnnotateColection(void) {
  ensureValidity();
  return CheckStatus(CDM_AnnotateCollection(Collection));
} /* OCDM_Collection::AnnotateCollection */

/**
*  \par Description
*  This function will delete the disk representation of the Collection 
*  object in use. The Name parameter must be the (absolute) path to the directory 
*  that holds the Collection.  Having a Collection object, this information 
*  can be obtained through the use of the function OCDM_GetName.
*  \par Description
*  \arg \c name: The name of the collection to be destroyed
*  
*  \note  this function will only delete the current disk 
*  representation of a Collection.  If the Collection is loaded in memory 
*  and we save the loaded Collection (with the help of OCDM_Sync), then the 
*  representaion of the Collection in disk will be re-created.
*
*/
int OCDM_Collection::Destroy(const char *name) {
  ensureValidity();
  return CheckStatus(CDM_Destroy((char *) name));
} /* OCDM_Collection::Destroy */

/**
*  \brief This function will return a Document object representing the first 
*  \par Description
*  This function will return a Document object representing the first 
*  Document of the Collection object in use. It can be used in order 
*  to start an iteration over all Documents of the Collection.  This 
*  function will return the first Document in the given Collection and make 
*  the proper arrangements so as subsequent calls to OCDM_NextDocument 
*  return the following Documents.
*
*  \note If an error occurs, then an OCDM_Exception will be thrown 
*   
*/
OCDM_REF(OCDM_Document) OCDM_Collection::FirstDocument(void) const {
  ensureValidity();
  CDM_Document doc = CDM_FirstDocument(Collection);
  if (doc == NULL) throw OCDM_Exception();
  return OCDM_GetReference<OCDM_Document>(*this, doc);
} /* OCDM_Collection::FirstDocument */

/**
*  \par Description
*  This function will return the Document object that has the same External 
*  Id as the value of the "XID" parameter. The value of the "XID" parameter 
*  is assumed to be encoded using the UTF-8 encoding (thus enabling the 
*  existance of non Latin characters in the value).  
*  If more than one Documents having the same External Id are found in the 
*  specified Collection, then the Document that was created first will be returned.
* 
* \note If an error occurs, then an OCDM_Exception will be thrown 
*/
OCDM_REF(OCDM_Document) OCDM_Collection::GetByExternalId(const char *XID)
                                                                         const {
  ensureValidity();
  CDM_Document doc = CDM_GetByExternalId(Collection, (char *) XID);
  if (doc == NULL) throw OCDM_Exception();
  return OCDM_GetReference<OCDM_Document>(*this, doc);
} /* OCDM_Collection::GetByExternalId */

/**
*  \par Description
*  This function will return the Document object that has the same Id as 
*  the value of the "Id" parameter. The value of the "Id" parameter 
*  is assumed to be encoded using the UTF-8 encoding (thus enabling the 
*  existance of non Latin characters in the value).
*
*  As usual, if an error occurs, then an OCDM_Exception will be thrown 
*
*/
OCDM_REF(OCDM_Document) OCDM_Collection::GetDocument(const char *ID) const {
  ensureValidity();
  CDM_Document doc = CDM_GetDocument(Collection, (char *) ID);
  if (doc == NULL) throw OCDM_Exception();
  return OCDM_GetReference<OCDM_Document>(*this, doc);
} /* OCDM_Collection::GetDocument */

/**
*  \par Description
*  This function will return the next available Document object. where "Next" 
*  means the Document that follows the one previously returned by calls to 
*  the following functions: CDM_FirstDocument, CDM_NextDocument, CDM_GetDocument, 
*  CDM_GetByExternalId.
*  If previous calls to this function have returned all available 
*  Documents, then an OCDM_Exception will be thrown  will be returned and the
*  current active Tcl interpreter (CDM_Interp) will be reseted. 
*  
*  As a result, the caller must check the 
*  information stored at this Tcl interpreter in order to distinguish 
*  between the normal termination of the Document chain or an error 
*  condition.
*
*/
OCDM_REF(OCDM_Document) OCDM_Collection::NextDocument(void) const {
  ensureValidity();
  CDM_Document doc = CDM_NextDocument(Collection);
  if (doc == NULL) throw OCDM_Exception();
  return OCDM_GetReference<OCDM_Document>(*this, doc);
} /* OCDM_Collection::NextDocument */

/**
*  \par Description
*  This function will return the Owner of the specified Collection. Its 
*  value is not of great importance, and usually defaults to the value 
*  "CDM".  The returned value will be encoded using the UTF-8 encoding 
*  (thus enabling the existance of non Latin characters in the value).
*  
*  \return The return value will be of type char* and will be owned by CDM.
*
*/
const char *OCDM_Collection::GetOwner(void) const {
  ensureValidity();
  return ReturnString(CDM_GetOwner(Collection));
} /* OCDM_Collection::GetOwner */

/**
*  \par Description
*  This function will remove the Document object that has as Id the 
*  specified by the parameter "Id" value.  If a Document with the requested 
*  Id does not exist, then an Exception of type OCDM_Exception will be thrown
*  \par Arguments
*  \arg \c Id: The Id parameter that specifies the document to be removed
*
*/
int OCDM_Collection::RemoveDocument(const char *Id) {
  ensureValidity();
  return CheckStatus(CDM_RemoveDocument(Collection, (char *) Id));
} /* OCDM_Collection::RemoveDocument */

/**
*  \par Description
*  This function will change the Name of the Collection object in use. The 
*  Name of the Collection represents the (absolute) path of the 
*  Collection's disk representation.  Changing the Name of a Collection 
*  will result into a new Collection representation in disk if the 
*  Collection is saved (through the use of the function CDM_Sync). 
*  Note that this function will not perform any checks on the value of the 
*  "Name" parameter.
*  \par Arguments
*  \arg \c Owner; the parameter of the new name
*
*  
*  \note If an invalid path is given, or a path that corresponds to an existing file, 
*  no error will be returned by this function.  (Although OCDM_Sync will 
*  return an error if the caller tries to save a Collection with an 
*  invalid Name.) 
*  The value of the "Name" parameter will not modified in 
*  any way by CDM, as a local copy will be created and modified internally 
*  by the CDM.
*
*/
const char *OCDM_Collection::SetName(const char *Name) {
  ensureValidity();
  return ReturnString(CDM_SetName(Collection, (char *) Name));
} /* OCDM_Collection::SetName */

/**
*  \par Description
*  This function will change the Owner of the Collection object in use.  
*  The value of the "Owner" parameter will not modified in any way by CDM,
*  as a local copy will be created and modified internally by the CDM.
*  \par Arguments
*  \arg \c Owner; the parameter of the new owner
*
*  \note This function will return a pointer (of type char*) to the string buffer 
*  that holds the new Collection Owner.  This pointer will be at a different 
*  memory location than the given parameter value, as the CDM will create 
*  and manipulate a local copy.
*   
*/
const char *OCDM_Collection::SetOwner(const char *Owner) {
  ensureValidity();
  return ReturnString(CDM_SetOwner(Collection, (char *) Owner));
} /* OCDM_Collection::SetOwner */

/**
*  \par Description
*  This function sets the provided Tcl object (of type Tcl_Obj*) as 
*  associated information to the  Collection object we are currently use. This 
*  "associated information" (Tcl object) can contain anything that the 
*  user wants to store ("associate") with this Collection. CDM does not 
*  modify or use this information in any way.
*
*  If an error occurs an exception of type OCDM_Exceptio will be thrown
*
*/
int OCDM_Collection::SetAssociatedInfo(const OCDM_ByteSequence& Info) {
  ensureValidity();
  return CheckStatus(CDM_SetAssociatedInfo(Collection, Info.getObject()));
} /* OCDM_Collection::SetAssociatedInfo */

/**
* \brief This method logs information. It is equivalent to
* OCDM_Utilities::Log()
*/
void OCDM_Collection::Log(const char *str, ...) const {
  CDM_Log(str);
}; /* OCDM_Collection::Log */

/**
*  \brief This function returns the number of Documents contained in the
*  Collection object.
*
*/
long OCDM_Collection::size(void) const {
  if (Collection == NULL) return 0;
  return CDM_Length(Collection);
} /* OCDM_Collection::size */

/**
*  \par Description
*  As the name implies this function checks for a valid Collection.
*  In other words, if a Document exists it returns a true OCDM_BOOLean variable.
*  Otherwise it returns false
*
*/
OCDM_BOOL OCDM_Collection::Valid(void) const {
  if (Collection) return true;
  return false;
} /* OCDM_Collection::Valid */

/**
*  \par Description
*  Return object as a formatted string.
*
*/
const char *OCDM_Collection::toString(void) const {
  if (Collection)
     return CDMTCL_GetCommandName(Collection->Interp, Collection->Tcl_Cmd);
  return "";
} /* OCDM_Collection::toString */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_Tcl                                             *******
 ******************************************************************************
 *****************************************************************************/

/**
*  \brief The constructor.
*
*/
OCDM_Tcl::OCDM_Tcl() {
  resultObj = NULL;
};

/**
*  \brief The destructor.
*
*/
OCDM_Tcl::~OCDM_Tcl() {
  if (resultObj) CDM_Free(resultObj);
};

/**
*  \brief Evaluates Tcl code contained in a string.
*  \par Description:
*  This method can be used to execute Tcl code. This function calls
*  the Tcl_Eval function from the Tcl C library.
*  The string should be a proper UTF-8 string. The string is parsed and
*  executed directly instead of compiling it and executing the bytecodes.
* \par Arguments:
* \arg \c tcl_code: A NULL terminated UTF-8 string.
* \return The return value from this method is a Tcl completion code
* (with one of the values OCDM_OK, OCDM_ERROR, OCDM_RETURN, OCDM_BREAK, or
* OCDM_CONTINUE, or possibly some other integer value originating in an
*  extension. In case of an error, an OCDM_Exception is thrown.
*
*/
int OCDM_Tcl::eval(const char *tcl_code) const {
  int status = CDMTCL_Eval(CDM_GetActiveInterpreter(), tcl_code);
  if (status == TCL_ERROR) throw OCDM_Exception();
  return status;
}; /* OCDM_Tcl::eval */

/**
*  \brief Executes an operating system command by using the Tcl "exec" command.
*
*/
int OCDM_Tcl::exec(const char *command) const {
  Tcl_DString ds;
  Tcl_DStringInit(&ds);
  Tcl_DStringAppend(&ds, "exec ", -1);
  Tcl_DStringAppend(&ds, command, -1);
  int status = CDMTCL_Eval(CDM_GetActiveInterpreter(), Tcl_DStringValue(&ds));
  Tcl_DStringFree(&ds);
  if (status == TCL_ERROR) throw OCDM_Exception();
  return status;
}; /* OCDM_Tcl::exec */

const char *OCDM_Tcl::GetResult(void) const {
  if (resultObj) CDM_Free(resultObj);
  resultObj = CDMTCL_GetObjResult(CDM_GetActiveInterpreter());
  if (resultObj == NULL) return "";
  CDMTCL_IncrRefCount(resultObj);
  return CDMTCL_GetString(resultObj);
}; /* OCDM_Tcl::GetResult */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_Utilities                                       *******
 ******************************************************************************
 *****************************************************************************/

/**
*  \brief The constructor.
*
*/
OCDM_Utilities::OCDM_Utilities() {
  resultObj = NULL;
};

/**
*  \brief The destructor.
*
*/
OCDM_Utilities::~OCDM_Utilities() {
  if (resultObj) CDM_Free(resultObj);
};

int OCDM_Utilities::DeleteFile(const char *path) const {
  return CDM_DeleteFile((char *) path);
}; /* OCDM_Utilities::DeleteFile */

int OCDM_Utilities::FileExists(const char *path) const {
  return CDM_FileExists((char *) path);
}; /* OCDM_Utilities::FileExists */

int OCDM_Utilities::MakeDirectory(const char *directory) const {
  return CDM_Mkdir((char *) directory);
}; /* OCDM_Utilities::MakeDirectory */

const char *OCDM_Utilities::GetIP(void) const {
  return CDM_GetIP();
}; /* OCDM_Utilities::GetIP */

const char *OCDM_Utilities::GetTempDir(void) const {
  return CDM_GetTempDir();
}; /* OCDM_Utilities::GetTempDir */

void OCDM_Utilities::Panic(const char *error) const {
  CDM_Panic((char *) error);
}; /* OCDM_Utilities::Panic */

void OCDM_Utilities::Log(const char *fmt) const {
  CDM_Log(fmt);
}; /* OCDM_Utilities::Log */

const char *OCDM_Utilities::GetResult(void) const {
  if (resultObj) CDM_Free(resultObj);
  resultObj = CDMTCL_GetObjResult(CDM_GetActiveInterpreter());
  if (resultObj == NULL) return "";
  CDMTCL_IncrRefCount(resultObj);
  return CDMTCL_GetString(resultObj);
}; /* OCDM_Utilities::GetResult */

/******************************************************************************
 ******************************************************************************
 *******   Class:  OCDM_Macros                                          *******
 ******************************************************************************
 *****************************************************************************/

/**
*  \brief The constructor.
*
*/
OCDM_Macros::OCDM_Macros() {
  system              = NULL;
  token               = NULL;
  storedDocumentObj   = NULL;
  storedCollectionObj = NULL;
};

OCDM_Macros::OCDM_Macros(const OCDM_Collection &col) {
  system              = CDM_MACRO_GetContainerSystem(col.Collection);
  token               = NULL;
  storedDocumentObj   = NULL;
  storedCollectionObj = NULL;
};

/**
*  \brief The destructor.
*
*/
OCDM_Macros::~OCDM_Macros() {
  if (system) CDM_Free(system);
  if (token)  CDM_Free(token);
  releaseStoredObject();
};

void  OCDM_Macros::releaseStoredObject(void) const {
  if (storedDocumentObj   != NULL) delete storedDocumentObj;
  if (storedCollectionObj != NULL) delete storedCollectionObj;
  storedDocumentObj   = NULL;
  storedCollectionObj = NULL;
}; /* OCDM_Macros::releaseStoredObject */

void  OCDM_Macros::storeObject(const class OCDM_Document *objPtr) const {
  if (storedDocumentObj != NULL) delete storedDocumentObj;
  storedDocumentObj = objPtr;
}; /* OCDM_Macros::storeObject */

void  OCDM_Macros::storeObject(const class OCDM_Collection *objPtr) const {
  if (storedCollectionObj != NULL) delete storedCollectionObj;
  storedCollectionObj = objPtr;
}; /* OCDM_Macros::storeObject */

void OCDM_Macros::OpenSystem(const char *name, const char *folder) const {
  if (system) CDM_Free(system);
  system = CDM_MACRO_OpenSystem(name, folder);
}; /* OCDM_Macros::OpenSystem */

void OCDM_Macros::EditSystem(const char *name, const char *folder) const {
  if (system) CDM_Free(system);
  system = CDM_MACRO_EditSystem(name, folder);
  if (system == NULL) throw OCDM_Exception();
}; /* OCDM_Macros::EditSystem */

void OCDM_Macros::OpenInSystem(const char *collection_name,
                               const char *document_id) const {
  if (system == NULL) throw OCDM_Exception("unknown system: please use "
                            "the OpenSystem method to specify a system!");
  if (token) CDM_Free(token);
  token = CDM_MACRO_OpenInSystem(system, collection_name, document_id);
  if (token == NULL) throw OCDM_Exception();
}; /* OCDM_Macros::OpenInSystem */

void OCDM_Macros::OpenInSystem(const char *collection_name) const {
  if (system == NULL) throw OCDM_Exception("unknown system: please use "
                            "the OpenSystem method to specify a system!");
  if (token) CDM_Free(token);
  token = CDM_MACRO_OpenInSystem(system, collection_name, "");
  if (token == NULL) throw OCDM_Exception();
}; /* OCDM_Macros::OpenInSystem */

OCDM_REF(OCDM_Collection) OCDM_Macros::GetOpenedCollection(void) const {
  if (token == NULL)  throw OCDM_Exception("unknown Collection/Document: "
      "please use the OpenInSystem method to specify a Collection/Document!");
  CDM_Collection col = CDM_MACRO_GetOpenedCollection(token);
  if (col == NULL) throw OCDM_Exception();
  return OCDM_GetReference<OCDM_Collection>(*this, col);
}; /* OCDM_Macros::GetOpenedCollection */

OCDM_REF(OCDM_Document) OCDM_Macros::GetOpenedDocument(void) const {
  if (token == NULL)  throw OCDM_Exception("unknown Collection/Document: "
      "please use the OpenInSystem method to specify a Collection/Document!");
  CDM_Document doc = CDM_MACRO_GetOpenedDocument(token);
  if (doc == NULL) throw OCDM_Exception();
  return OCDM_GetReference<OCDM_Document>(*this, doc);
}; /* OCDM_Macros::GetOpenedDocument */

int OCDM_Macros::RunSystemComponent(const char *Annotator) const {
  return RunSystemComponent(Annotator, false);
}; /* OCDM_Macros::RunSystemComponent */

int OCDM_Macros::RunSystemComponent(const char *Annotator,
                                    const bool force) const {
  if (system == NULL) throw OCDM_Exception("unknown system: please use "
                            "the OpenSystem method to specify a system!");
  if (token == NULL)  throw OCDM_Exception("unknown Collection/Document: "
      "please use the OpenInSystem method to specify a Collection/Document!");
  return CDM_MACRO_RunSystemComponent(system, token, Annotator, force);
}; /* OCDM_Macros::RunSystemComponent */

int OCDM_Macros::RunSystemTool(const char *Annotator) const {
  return RunSystemTool(Annotator, false);
}; /* OCDM_Macros::RunSystemTool */

int OCDM_Macros::RunSystemTool(const char *Annotator, const bool force) const {
  if (system == NULL) throw OCDM_Exception("unknown system: please use "
                            "the OpenSystem method to specify a system!");
  if (token == NULL)  throw OCDM_Exception("unknown Collection/Document: "
      "please use the OpenInSystem method to specify a Collection/Document!");
  return CDM_MACRO_RunSystemTool(system, token, Annotator, force);
}; /* OCDM_Macros::RunSystemTool */

int OCDM_Macros::DisplayViewer(const char *viewer_name,
                               const OCDM_Document &doc,
                               const char *parameter1,
                               const char *parameter2) const {
  if (system == NULL) throw OCDM_Exception("unknown system: please use "
                            "the OpenSystem method to specify a system!");
  return CDM_MACRO_DisplayViewer(system, viewer_name, doc.Document, parameter1,
  parameter2);
}; /* OCDM_Macros::DisplayViewer */

int OCDM_Macros::NFoldExecution(OCDM_Collection col,
                                const int folds,
                                OCDM_ByteSequenceSet TrainingCommands,
                                OCDM_ByteSequenceSet EvaluationCommands) const {
  return CDM_MACRO_NFoldExecution(col.Collection, folds,
                                  TrainingCommands.getObject(),
                                  EvaluationCommands.getObject());
}; /* OCDM_Macros::NFoldExecution */

int OCDM_Macros::NFoldExecution(OCDM_Collection col,
                                const int folds,
                                OCDM_RawDataSet TrainingCommands,
                                OCDM_RawDataSet EvaluationCommands) const {
  return CDM_MACRO_NFoldExecution(col.Collection, folds,
                                  TrainingCommands.getObject(),
                                  EvaluationCommands.getObject());
}; /* OCDM_Macros::NFoldExecution */

int OCDM_Macros::NFoldExecution(OCDM_Collection col, const int folds,
                 const std::vector<std::string>& TrainingCommands,
                 const std::vector<std::string>& EvaluationCommands) const {
  CDM_Object train = OCDM_STL_ContainerToCDM_Object(TrainingCommands);
  CDM_Object eval  = OCDM_STL_ContainerToCDM_Object(EvaluationCommands);
  return CDM_MACRO_NFoldExecution(col.Collection, folds, train, eval);
}; /* OCDM_Macros::NFoldExecution */

/*
 * End of File
 */
