##############################################################################
##  Python.tcl:                                                             ##
## -------------                                                            ##
##  This file implements various Python related procedures.                 ##
##                                                                          ##
##  Copyright 2002-2008 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

namespace eval ::CDM_Python {
  variable PythonLoaded 0
  variable PythonVersion 2.5.2
}

## ::CDM_Python::InitialisePython
#  Use this function to create a Python interpreter inside Ellogon.
proc ::CDM_Python::InitialisePython {} {
  variable PythonLoaded
  switch $::tcl_platform(platform) {
    windows {
      InitialisePython_Windows
    }
    unix {
      InitialisePython_Unix
    }
    default {
    }
  }
  global HomeDir ApplicationInfo GUI_Options ApplicationName SplashPercent
  set error "Python support is disabled: package tclpython has not been\
             instaled!"
  set shared_lib_ext [info sharedlibextension]
  set bin_dir $ApplicationInfo(LoadBinaryFolderName)
  foreach dir [lsort -decreasing \
                 [glob -nocomplain -type d $HomeDir/lib/tclpython*]] {
    ## Try to load the python bindings for all folders found in this directory.
    foreach dll [lsort -decreasing [glob -nocomplain -type f \
                       -directory $dir/$bin_dir libtclpython*$shared_lib_ext]] {
      if {[catch {CDM_load $dll Tclpython} error]} {
        continue
      }
      set PythonLoaded 1
      set version [lindex [split [file rootname [file tail $dll]] -] end]
      break
    }
    if {$PythonLoaded} {break}
  }
  if {!$PythonLoaded} {
    catch {puts stderr "Cannot initialise Python Bindings:\n\
                        The error was: $error"}
    ## The Python package failed to initialise. We will make another attempt
    ## before we report an error...
    if {$ApplicationInfo(EnableGUI)} {
      set VisScript [bind .splashscr <Visibility>]
      bind .splashscr <Visibility> {}
      lower .splashscr
      tk_messageBox -title Error... -type ok -default ok -icon error -message\
         "Cannot initialise Python Bindings:\n\
          The error was: $error" -parent .splashscr
      ## Maybe the user wants to disable Python support?
      set ans [gut_question "Do you want to disable Python support for the next\
        time $ApplicationName will start?" Question... 1 .splashscr]
      if {[string equal $ans yes]} {
        set GUI_Options(EnablePython) 0
      }
      unset ans
      bind .splashscr <Visibility> $VisScript
      raise .splashscr
      unset VisScript
    }
  }
  if {$PythonLoaded} {
    ## Initialise the python interpreter we will be using for executing python
    ## code...
    variable PythonInterp
    set CWD [pwd]
    cd $HomeDir/CDM/Python/$ApplicationInfo(LoadBinaryFolderName)
    if {[catch {
      set PythonInterp [::python::interp new]
      $PythonInterp exec {print("Python interpreter succesfully initialised!")}
      ## Load the "sys" package
      $PythonInterp exec {import sys}
      ## Ensure that the "sys.path" contains the '' entry...
      $PythonInterp exec {sys.path = sys.path + ['']}
      ## The tclpython was succesfully loaded. Load the CDM Python bindings
      ## (PythonCDM)
      LDM_SetSplashWindowStatus {Loading CDM Python bindings (PythonCDM)...} \
                                                             $SplashPercent 205
      } error]} {
      # We failed in creating a Python interpreter... Report the error!
      catch {puts stderr "Cannot initialise Python Bindings:\n\
                        The error was: $error"}
      if {$ApplicationInfo(EnableGUI)} {
        set VisScript [bind .splashscr <Visibility>]
        bind .splashscr <Visibility> {}
        lower .splashscr
        tk_messageBox -title Error... -type ok -default ok -icon error -message\
           "Cannot initialise Python Bindings:\n\
            The error was: $error" -parent .splashscr
        ## Maybe the user wants to disable Python support?
        set ans [gut_question "Do you want to disable Python support\
          for the next time $ApplicationName will start?" \
          Question... 1 .splashscr]
        if {[string equal $ans yes]} {
          set GUI_Options(EnablePython) 0
        }
        unset ans
        bind .splashscr <Visibility> $VisScript
        raise .splashscr
        unset VisScript
      }
    }
    if {[catch {$PythonInterp exec {import PythonCDM}} error]} {
      set PythonLoaded 0
      catch {puts stderr "Cannot Load PythonCDM:\nCurrent Directory: '[pwd]'\n\
                          The error was: $error"}
      if {$ApplicationInfo(EnableGUI)} {
        set VisScript [bind .splashscr <Visibility>]
        bind .splashscr <Visibility> {}
        lower .splashscr
        tk_messageBox -title Error... -type ok -default ok -icon error -message\
           "Cannot Load PythonCDM:\nCurrent Directory: \"[pwd]\"\n\
            The error was: $error" -parent .splashscr
        ## Maybe the user wants to disable Python support?
        set ans [gut_question "Do you want to disable Python support for the\
          next time $ApplicationName will start?" Question... 1 .splashscr]
        if {[string equal $ans yes]} {
          set GUI_Options(EnablePython) 0
        }
        unset ans
        bind .splashscr <Visibility> $VisScript
        raise .splashscr
        unset VisScript
      }
    } else {
      ## We have loaded Python!
      catch {LDM_SplashWindowNotify language python}
    }
    cd $CWD
  }
};# ::CDM_Python::InitialisePython

## ::CDM_Python::InitialisePython_Windows
#  Initialise Python under Windows
proc ::CDM_Python::InitialisePython_Windows {} {
};# ::CDM_Python::InitialisePython_Windows

## ::CDM_Python::InitialisePython_Unix
#  Initialise Python under Unix
proc ::CDM_Python::InitialisePython_Unix {} {
  ## Try to receive Python verion, by executing "python -V"...
  variable PythonVersion
  catch {exec python -V} version
  switch -glob -- $version {
    {Python 2.*} {set PythonVersion [lindex $version end]}
    default {}
  }
  # set ::env(PYTHONHOME) /usr/lib/python$PythonVersion
};# ::CDM_Python::InitialisePython_Unix

## ::CDM_Python::LoadPythonComponent
#  Load a Component defined as Python functions and make them available as Tcl
#  functions...
proc ::CDM_Python::LoadPythonComponent {module_path module} {
  if {!$::CDM_Python::PythonLoaded} {error "python bindings disabled"}
  variable PythonInterp
  ## Retrieve additional directories to be added to the current sys.path. A
  ## Component can store additional paths in the "sys.path" entry in
  ## creole_config.tcl
  if {![catch {PUL_plugin_module_config $module sys.path} \
          sys.path]} {
    ## The separator among different paths can be either ";" or ":"...
    if {[llength ${sys.path}]} {
      set paths [split ${sys.path} $::ApplicationInfo(PathSplitCharacters)]
      ## TclBlend seems to want path in Unix notation...
      set AddedPaths {}
      foreach path $paths {
        if {[catch {string map {\\ /} [uplevel #0 [list subst $path]]} path]} {
          puts "  Error: invalid value found in sys.path:\n  $path"
          continue
        }
        lappend AddedPaths $path
        $PythonInterp exec "sys.path = sys.path + \['$path'\]"
      }
      puts "  ++ sys.path: \"[join $AddedPaths {;}]\""
    }
  }
  ## source the python code stored in $module_path/$module.pl
  set file [open $module_path/$module.py]
  set python_code [read $file]
  close $file
  catch {
  $PythonInterp exec "creole_${module}_home = \'[file normalize $module_path]\'"
  }
  $PythonInterp exec $python_code
  ## Mirror the python subs into Tcl procs...
  set ProcBody {
    if {[llength $args]} {
      %PYTHONINTERP% exec "%PROC%('[join $args {', '}]')"
    } else {
      %PYTHONINTERP% exec {%PROC%()}
    }
  }
  foreach proc [list creole_$module creole_${module}_Initialize \
                     creole_${module}_Finish] {
    proc ::$proc {args} \
      [string map "%PYTHONINTERP% $PythonInterp %PROC% $proc" $ProcBody]
  }
  catch {
    $PythonInterp exec \
      "creole_$module_SetComponentHome(\'[file normalize $module_path]\')"
  }
};# ::CDM_Python::LoadPythonComponent
