/* 
 * This is a simple Prolog parser
 * It takes as input a whole prolog statement, and returns a list having the
 * following elements:
 *    { predicate_name arg1 arg2 ...}
 * If an argument is a complex prolog element, then the corresponding list
 * element will have the proper sub-list elements.
 * Typical Example:
 *    am_(_,5,[[5,6,7,8,9,10],'T_._Marshall,\)Hahn_Jr_.'],'T_._ahn_Jr_.',
 *        a,'proper_noun_person',invariable).
 *   {am_ _ 5 {{5 6 7 8 9 10} {T_._Marshall,\)Hahn_Jr_.}} T_._ahn_Jr_.
 *        a proper_noun_person invariable}
 *        
 *  Author: Georgios Petasis, petasis@iit.demokritos.gr
 *  Last Updated: 28/7/1999
 */

#include <tcl.h>
Tcl_Obj *CDM_ParsePrologStatement(char *line);
char *CDM_PrologGetSubList(char *line);

#ifdef STANDALONE_PROGRAM
main(int argc, char *argv[]) {
  char line[256];
  Tcl_Obj *List;

  while (!feof(stdin)) {
    fgets(line, 256, stdin);
    if (feof(stdin) || ferror(stdin)) break;
    printf("%s\n", line);
    List = CDM_ParsePrologStatement(line);
    printf("%s\n", Tcl_GetString(List));
  }
  exit(0);
}
#endif /* STANDALONE_PROGRAM */

#ifdef DYNAMIC_LIBRARY

#ifndef DLLEXPORT
#define DLLEXPORT
#endif

extern "C" int CDM_ParsePrologStatementObjCmd(
  ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

extern "C" DLLEXPORT int Parseprolog_Init(Tcl_Interp *Interp) {
  if(Tcl_CreateObjCommand(Interp, "prolog",
                           CDM_ParsePrologStatementObjCmd,
                           (ClientData) NULL,
                           (Tcl_CmdDeleteProc *) NULL)  == NULL)
       return TCL_ERROR;       
  return TCL_OK;
}

extern "C" int CDM_ParsePrologStatementObjCmd(
  ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) {
  Tcl_Obj *List;
  char *string;
  
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "prolog-statement");
    return TCL_ERROR;
  }
  string = Tcl_GetString(objv[1]);
  if (string == NULL) return TCL_OK;
  List = CDM_ParsePrologStatement(string);
  Tcl_SetObjResult(interp, List);
  return TCL_OK;
}
#endif /* DYNAMIC_LIBRARY */

Tcl_Obj *CDM_ParsePrologStatement(char *line) {
  Tcl_Obj *List, *Word;
  char *c = line, *word = line;
  int inside_atom = 0;

  List = Tcl_NewListObj(0, NULL);
  if (List == NULL) return NULL;

  while (*c != '\0') {
    if (*c == '\\') {
      c+=2; continue; /* Skip \? sequences... */
    }
    if (*c == '\'') {
       if (inside_atom == 0) {
	 inside_atom = 1;
	 word = c + 1;
       } else {
         *c = '\0';
	 if (strlen(word) > 0) {
           Word = Tcl_NewStringObj(word, -1);
           Tcl_ListObjAppendElement(NULL, List, Word);
	 }
         word = c + 1;
	 inside_atom = 0;
       }
    }
    if (!inside_atom && *c == '[') {
	word = CDM_PrologGetSubList(c);
	c++;
	Word = CDM_ParsePrologStatement(c);
	Tcl_ListObjAppendElement(NULL, List, Word);
	c = word;
    }
    if (!inside_atom && (*c == ')' || *c == '(' || *c == ',' || *c == '\'') ) {
       *c = '\0';
       if (strlen(word) > 0) {
         Word = Tcl_NewStringObj(word, -1);
         Tcl_ListObjAppendElement(NULL, List, Word);
       }
       word = c + 1;
    }
    if (!inside_atom && *c == '.') return List;
    c ++;
  }
  if (strlen(word) > 0) {
    Word = Tcl_NewStringObj(word, -1);
    Tcl_ListObjAppendElement(NULL, List, Word);
  }
  return List;
}

char *CDM_PrologGetSubList(char *line) {
  char *c = line;
  int level = 0, inside_atom = 0;

  if (*c != '[') return NULL;
  while (*c != '\0') {
     if (*c == '\\') {
      c+=2; continue; /* Skip \? sequences... */
    }
    if (*c == '\'') {
       if (inside_atom == 0) {
	 inside_atom = 1;
       } else {
	 inside_atom = 0;
       }
    }
    if (!inside_atom && *c == '[') {
	level++;
    }
    if (!inside_atom && *c == ']') {
        level--;
      if (!level) {
        *c = '\0';
        return c + 1;
      }
    }
    c++;
  }
  return NULL;
}
