/*****************************************************************************
 *  XMLIO.cpp:                                                               *
 * --------------------                                                      *
 *  This file implements various utilities used internally by the            *
 *  Collection & Document Manager to handle an XML-based storage format.     *
 *                                                                           *
 *  Copyright 1998-2010 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licensing information, please refer to the license conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/

#include "CDM.h"
#include <string.h>
#ifndef WINDOWS
#include <strings.h>
#endif /* WINDOWS */
#include <ctype.h>
#include "CDM_GlobalObjects.h"
extern Tcl_ThreadDataKey dataKey;
 
/*
 * Prototypes...
 */
static int CDM_XML_SaveCollection(CDM_Collection Collection);
static int CDM_XML_LoadCollection(CDM_Collection Collection);
static int CDM_XML_SaveDocument(CDM_Document Document, Tcl_Channel File);
static int CDM_XML_LoadDocument(CDM_Document Document, Tcl_Channel File,
                                char **Id, char **ExternalId);
static Tcl_Obj         *CDM_XML_CollectionXIdsPtr(char *dir);
static Tcl_Obj         *CDM_XML_CollectionIdsPtr(char *dir);
static CDM_AttributeSet CDM_XML_DocumentAttrsPtr(Tcl_Channel File);
static int              CDM_XML_WriteStrings TCL_VARARGS_DEF(Tcl_Channel, arg1);
int                     CDM_XMLCollectionTypeInit(Tcl_Interp *interp);

/*
 * --- CDM_XMLCollectionTypeInit:
 *  This function initialises and registers our XML Collection type.
 */
int CDM_XMLCollectionTypeInit(Tcl_Interp *interp) {
  return CDM_RegisterCollectionType("XML",
                                    CDM_XML_LoadCollection,
                                    CDM_XML_SaveCollection,
                                    CDM_XML_LoadDocument,
                                    CDM_XML_SaveDocument,
                                    CDM_XML_CollectionXIdsPtr,
                                    CDM_XML_CollectionIdsPtr,
                                    CDM_XML_DocumentAttrsPtr);
} /* CDM_XMLCollectionTypeInit */

/* 
 * --- CDM_XML_SaveCollection:
 *  This function saves a Collection in XML format...
 */
static int CDM_XML_SaveCollection(CDM_Collection Collection) {
  Tcl_Obj *Obj, *XObj, **Elem, **XElem;
  Tcl_Interp *interp;
  Tcl_Channel File;
  char *FileName, temp[64];
  int elements, i;
  CDM_THREAD_SPECIFIC_DATA;
  
  /*
   * Open the file that will hold our Collection description...
   */
  FileName = CDM_JoinPaths(Collection->Name, "CDM_CollectionIndex.xml", NULL);
  if (FileName == NULL) return TCL_ERROR;
  interp     = Collection->Interp;
  CDM_Interp = interp;
  File = Tcl_OpenFileChannel(interp, FileName, "w", 0666);
  if (File == NULL) {
    /*
     * Create the Collection's Host directory...
     */
    if (CDM_Mkdir(Collection->Name) != TCL_OK) {
      CDM_ReportError("CDM_XML_SaveCollection: Cannot create directory "
                      "\"%s\"", Collection->Name);
      return TCL_ERROR;
    }
    File = Tcl_OpenFileChannel(interp, FileName, "w", 0666);
  }
  if (File == NULL) {
    /*
     * Place an error message on the Collection's interpreter...
     */
    CDM_ReportError("CDM_XML_SaveCollection: Cannot open Collection "
                    "Index file: \"%s\"", FileName);
    CDM_Free(FileName);
    return TCL_ERROR;
  }
  CDM_Free(FileName);

  /*
   * Set the file encoding to utf-8...
   */
  if (Tcl_SetChannelOption(interp, File, "-encoding", "utf-8") != TCL_OK) {
    return CDM_ReportError("CDM_XML_SaveCollection: Cannot change the encoding"
                            "of the Collection Index file to \"utf-8\"");
  }

  /*
   * Write Header...
   */
  if (CDM_XML_WriteStrings(File,
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n",
    "<Collection>\n",
    "  <Encoding>", Collection->Encoding, "</Encoding>\n",
    "  <Name>",     Collection->Name,     "</Name>\n",
    "  <Owner>",    Collection->Owner,    "</Owner>\n",
    NULL) < 0) {
    Tcl_Close(interp, File);
    return TCL_ERROR;
  }

  /*
   * Write Document X/Ids...
   */
  Obj = CDM_ListGetObject(Collection->Documents, CDM_DOCUMENT_ID);
  if (Tcl_ListObjGetElements(NULL, Obj, &elements, &Elem) != TCL_OK) {
    CDM_ReportError("Invalid Document ID list");
    Tcl_Close(interp, File);
    Tcl_DecrRefCount(Obj);
    return TCL_ERROR;
  }
  XObj = CDM_ListGetObject(Collection->Documents, CDM_DOCUMENT_XID);
  if (Tcl_ListObjGetElements(NULL, XObj, &elements, &XElem) != TCL_OK) {
    CDM_ReportError("Invalid Document XID list");
    Tcl_Close(interp, File);
    Tcl_DecrRefCount(Obj);
    Tcl_DecrRefCount(XObj);
    return TCL_ERROR;
  }
  
  snprintf(temp, 62, "%d\">\n", elements);
  if (CDM_XML_WriteStrings(File,
    "  <DocumentXIDSet items=\"", temp, NULL) < 0) {
    Tcl_Close(interp, File);
    Tcl_DecrRefCount(Obj);
    Tcl_DecrRefCount(XObj);
    return TCL_ERROR;
  }
  for (i=0; i<elements; i++) {
    if (CDM_XML_WriteStrings(File,
      "    <DocumentID>",  Tcl_GetString(Elem[i]),  "</DocumentID>\n",
      "    <DocumentXID>", Tcl_GetString(XElem[i]), "</DocumentXID>\n",
      NULL) < 0) {
      Tcl_Close(interp, File);
      Tcl_DecrRefCount(Obj);
      Tcl_DecrRefCount(XObj);
      return TCL_ERROR;
    }
  }
  Tcl_DecrRefCount(Obj);
  Tcl_DecrRefCount(XObj);
  if (CDM_XML_WriteStrings(File, "  </DocumentXIDSet>\n", NULL) < 0) {
    Tcl_Close(interp, File);
    return TCL_ERROR;
  }

  /*
   * Write Attributes...
   */
  Obj = CDM_GetAttributes(Collection);
  if (Obj == NULL) {
    Tcl_Close(interp, File);
    return TCL_ERROR;
  }
  if (Tcl_ListObjGetElements(NULL, Obj, &elements, &Elem) != TCL_OK) {
     CDM_ReportError("Invalid Attribute list");
     Tcl_Close(interp, File);
     Tcl_DecrRefCount(Obj);
     return TCL_ERROR;
  }
  snprintf(temp, 62, "%d\">\n", elements);
  if (CDM_XML_WriteStrings(File,
    "  <AttributeSet items=\"", temp, NULL) < 0) {
    Tcl_Close(interp, File);
    Tcl_DecrRefCount(Obj);
    return TCL_ERROR;
  }
  for (i=0; i<elements; i++) {
    if (CDM_XML_WriteStrings(File,
      "    <Attribute>",  Tcl_GetString(Elem[i]),  "</Attribute>\n",
      NULL) < 0) {
      Tcl_Close(interp, File);
      Tcl_DecrRefCount(Obj);
      return TCL_ERROR;
    }
  }
  Tcl_DecrRefCount(Obj);
  if (CDM_XML_WriteStrings(File, "  </AttributeSet>\n", NULL) < 0) {
    Tcl_Close(interp, File);
    return TCL_ERROR;
  }

  /*
   * Write Comment...
   */
  Obj = CDM_GetComment(Collection);
  if (Obj) {
    snprintf(temp, 62, "  <Comment length=\"%d\">\n", 
                       Tcl_NumUtfChars(Tcl_GetString(Obj), -1)); 
    Tcl_WriteChars(File, temp, -1);
    Tcl_WriteObj(File, Obj);
    Tcl_WriteChars(File, "\n  </Comment>\n", -1);
  }
  
  /*
   * Write Footer...
   */
  Tcl_WriteChars(File, "</Collection>\n", -1);

  /*
   * Close file...
   */
  if (Tcl_Close(interp, File) != TCL_OK) return TCL_ERROR;
  Tcl_ResetResult(interp);
  return TCL_OK;
} /* CDM_XML_SaveCollection */

/*
 * --- CDM_XML_LoadCollection:
 *  This function loads a Collection in XML format...
 */
static int CDM_XML_LoadCollection(CDM_Collection Collection) {
  Tcl_Obj *LineObj = NULL;
  Tcl_Interp *interp;
  Tcl_Channel File;
  char *FileName, *Line, *tmp;
  int i, items;
  register CDM_ListNode *ptr, *lastNode;
  
  /*
   * Open the file that will hold our Collection description...
   */
  FileName = CDM_JoinPaths(Collection->Name, "CDM_CollectionIndex.xml", NULL);
  if (FileName == NULL) return TCL_ERROR;
  interp = Collection->Interp;
  File = Tcl_OpenFileChannel(interp, FileName, "r", 0666);
  CDM_Free(FileName);
  if (File == NULL) return TCL_ERROR;

  /*
   * Set the file encoding to utf-8...
   */
  if (Tcl_SetChannelOption(interp, File, "-encoding", "utf-8") != TCL_OK) {
     CDM_ReportError("cannot change encoding to \"utf-8\"");
     goto error;
  }
#ifdef WINDOWS
  /* Under windows, we have to disable ^Z (\x1a) as the EOF character,
     as our files can contain binary data... */
  Tcl_SetChannelOption(interp, File, "-eofchar", "");
#endif /* WINDOWS */

  /*
   * Read Header...
   */
  LineObj = Tcl_NewStringObj(NULL, -1);
  if (LineObj == NULL) {
    CDM_ReportError("CDM_XML_LoadCollection: Tcl_NewStringObj returned NULL -"
                    " out of memory?");
    goto error;
  }
  if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
  if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
  
  /* Read Encoding */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    tmp = strstr(Line, "</Encoding>"); if (tmp) *tmp = '\0';
    Collection->Encoding = CDM_Strdup(&Line[12]);
    if (tmp) *tmp = '<';
    
  /* Read Name */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    tmp = strstr(Line, "</Name>"); if (tmp) *tmp = '\0';
    Collection->Name = CDM_Strdup(&Line[8]);
    if (tmp) *tmp = '<';

  /* Read Owner */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    tmp = strstr(Line, "</Owner>"); if (tmp) *tmp = '\0';
    Collection->Owner = CDM_Strdup(&Line[9]);
    if (tmp) *tmp = '<';
  
  /* Read Document X/Ids */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    if (sscanf(&Line[2], "<DocumentXIDSet items=\"%d\">", &items) < 1) {
      CDM_ReportError("Cannot read items from DocumentXIDSet");
      goto error;
    }
    if (items > 0) {
      Tcl_SetObjLength(LineObj, 0);
      if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
      Line = Tcl_GetString(LineObj);
      ptr = CDM_NewListNode();
      tmp = strstr(Line, "</DocumentID>"); if (tmp) *tmp = '\0';
      ptr->ptr[CDM_DOCUMENT_ID]  = CDM_Strdup(&Line[16]);
      if (tmp) *tmp = '<';
      Tcl_SetObjLength(LineObj, 0);
      if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
      Line = Tcl_GetString(LineObj);
      tmp = strstr(Line, "</DocumentXID>"); if (tmp) *tmp = '\0';
      ptr->ptr[CDM_DOCUMENT_XID] = CDM_Strdup(&Line[17]);
      if (tmp) *tmp = '<';
      Collection->Documents = lastNode = CDM_AppendListNode(NULL, ptr);
      if (items > 1) {
        for (i=0; i<items; i++) {
          Tcl_SetObjLength(LineObj, 0);
          if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
          Line = Tcl_GetString(LineObj);
          ptr = CDM_NewListNode();
          tmp = strstr(Line, "</DocumentID>"); if (tmp) *tmp = '\0';
          ptr->ptr[CDM_DOCUMENT_ID]  = CDM_Strdup(&Line[16]);
          if (tmp) *tmp = '<';
          Tcl_SetObjLength(LineObj, 0);
          if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
          Line = Tcl_GetString(LineObj);
          tmp = strstr(Line, "</DocumentXID>"); if (tmp) *tmp = '\0';
          ptr->ptr[CDM_DOCUMENT_XID] = CDM_Strdup(&Line[17]);
          if (tmp) *tmp = '<';
          lastNode = CDM_AppendListNode(lastNode, ptr);
        }
      } /* if (items > 1) */
    } /* if (items > 0) */
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;

  /* Read Attributes */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    if (sscanf(&Line[2], "<AttributeSet items=\"%d\">", &items) < 1) {
      CDM_ReportError("Cannot read items from AttributeSet");
      goto error;
    }
    for (i=0; i<items; i++) {
      Tcl_SetObjLength(LineObj, 0);
      if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
      Line = Tcl_GetString(LineObj);
      tmp = strstr(Line, "</Attribute>"); if (tmp) *tmp = '\0';
      CDM_PutAttribute(Collection, Tcl_NewStringObj(&Line[15], -1));
      if (tmp) *tmp = '<';
    }
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;

  /* Read Comment */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    /* There is the possibility the comment to not exist... */
    if (strlen(Line) && strcmp("</Collection>", Line) != 0) {
      if (sscanf(&Line[2], "<Comment length=\"%d>", &items) < 1) {
        CDM_ReportError("Cannot read Comment Length!");
        goto error;
      }
      if (items) {
        Tcl_ReadChars(File, LineObj, items, 0);
        CDM_SetComment(Collection, Tcl_DuplicateObj(LineObj));
        Collection->Modified = 0;
      }
    }
  
  Tcl_DecrRefCount(LineObj);
  Tcl_Close(interp, File);
  return TCL_OK;
error:
  if (LineObj) Tcl_DecrRefCount(LineObj);
  Tcl_Close(interp, File);
  return TCL_ERROR;
} /* CDM_XML_LoadCollection */

/*
 * --- CDM_XML_SaveDocument:
 *  This function saves a Document in XML format...
 */
static int CDM_XML_SaveDocument(CDM_Document Document, Tcl_Channel File) {
  Tcl_Obj *Obj, **Elem;
  char temp[128];
  int elements, i;
  
  if (Document == NULL || File == NULL) return TCL_ERROR;
  /*
   * Write Header...
   */
  Tcl_WriteChars(File, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n", -1);
  Tcl_WriteChars(File, "<Document>\n", -1);
  Tcl_WriteChars(File, "  <Encoding>", -1);
  Tcl_WriteChars(File, Document->Encoding, -1);
  Tcl_WriteChars(File, "</Encoding>\n", -1);
  Tcl_WriteChars(File, "  <ExternalID>", -1);
  Tcl_WriteChars(File, Document->DocumentPtr->ptr[CDM_DOCUMENT_XID], -1);
  Tcl_WriteChars(File, "</ExternalID>\n", -1);
  Tcl_WriteChars(File, "  <ID>", -1);
  Tcl_WriteChars(File, Document->DocumentPtr->ptr[CDM_DOCUMENT_ID], -1);
  Tcl_WriteChars(File, "</ID>\n", -1);
  Tcl_WriteChars(File, "  <AnnotationMaxID>", -1);
  snprintf(temp, 126, "%ld", Document->AnnotationMaxID);
  Tcl_WriteChars(File, temp, -1);
  Tcl_WriteChars(File, "</AnnotationMaxID>\n", -1);

  /*
   * Write Attributes...
   */
  Obj = CDM_GetAttributes(Document);
  if (Obj == NULL) {
    return TCL_ERROR;
  }
  if (Tcl_ListObjGetElements(NULL, Obj, &elements, &Elem) != TCL_OK) {
     CDM_ReportError("Invalid Attribute list");
     Tcl_DecrRefCount(Obj);
     return TCL_ERROR;
  }
  Tcl_WriteChars(File, "  <AttributeSet items=\"", -1);
  snprintf(temp, 126, "%d\">\n", elements);
  Tcl_WriteChars(File, temp, -1);
  for (i=0; i<elements; i++) {
    Tcl_WriteChars(File, "    <Attribute>", -1);
    Tcl_WriteChars(File, Tcl_GetString(Elem[i]), -1);
    Tcl_InvalidateStringRep(Elem[i]);
    Tcl_WriteChars(File, "</Attribute>\n", -1);
  }
  Tcl_WriteChars(File, "  </AttributeSet>\n", -1);
  Tcl_DecrRefCount(Obj);

  /*
   * Write Annotations...
   */
  Obj = CDM_GetAnnotations(Document);
  if (Obj == NULL) {
    return TCL_ERROR;
  }
  if (Tcl_ListObjGetElements(NULL, Obj, &elements, &Elem) != TCL_OK) {
     CDM_ReportError("Invalid Annotation list");
     Tcl_DecrRefCount(Obj);
     return TCL_ERROR;
  }
  Tcl_WriteChars(File, "  <AnnotationSet items=\"", -1);
  snprintf(temp, 126, "%d\">\n", elements);
  Tcl_WriteChars(File, temp, -1);
  for (i=0; i<elements; i++) {
    Tcl_WriteChars(File, "    <Annotation>", -1);
    Tcl_WriteChars(File, Tcl_GetString(Elem[i]), -1);
    Tcl_InvalidateStringRep(Elem[i]);
    Tcl_WriteChars(File, "</Annotation>\n", -1);
  }
  Tcl_WriteChars(File, "  </AnnotationSet>\n", -1);
  Tcl_DecrRefCount(Obj);

  /*
   * Write Text...
   */
  Tcl_WriteChars(File, "  <Text length=", -1);
  snprintf(temp, 126, "%d>\n", 
                     Tcl_NumUtfChars(Tcl_GetString(Document->RawData), -1));
  Tcl_WriteChars(File, temp, -1);
  Tcl_WriteObj(File, Document->RawData);
  Tcl_WriteChars(File, "  </Text>\n", -1);
  
  /*
   * Write Footer...
   */
  Tcl_WriteChars(File, "</Document>\n", -1);
  return TCL_OK;
} /* CDM_XML_SaveDocument */

/*
 * --- CDM_XML_LoadDocument:
 *  This function loads a Document in XML format...
 */
static int CDM_XML_LoadDocument(CDM_Document Document, Tcl_Channel File,
                                char **Id, char **ExternalId) {
  Tcl_Obj *LineObj;
  char *Line, *tmp;
  long maxId;
  int items, chars, i;
  
  /*
   * Read Header...
   */
  LineObj = Tcl_NewStringObj(NULL, -1);
  if (LineObj == NULL) {
    CDM_ReportError("CDM_XML_LoadCollection: Tcl_NewStringObj returned NULL -"
                    " out of memory?");
    goto error;
  }
  if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
  if (Tcl_GetsObj(File, LineObj) <= 0) goto error;

  /* Read Encoding */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    tmp = strstr(Line, "</Encoding>"); if (tmp) *tmp = '\0';
    Document->Encoding = CDM_Strdup(&Line[12]);
    if (tmp) *tmp = '<';

  /* Read ExternalID */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    tmp = strstr(Line, "</ExternalID>"); if (tmp) *tmp = '\0';
    *ExternalId = CDM_Strdup(&Line[14]);
    if (tmp) *tmp = '<';

  /* Read ID */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    tmp = strstr(Line, "</ID>"); if (tmp) *tmp = '\0';
    *Id = CDM_Strdup(&Line[6]);
    if (tmp) *tmp = '<';
  
  /* Read AnnotationMaxID */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    if (sscanf(&Line[2], "<AnnotationMaxID>%ld", &maxId) < 1) {
      CDM_ReportError("Cannot read AnnotationMaxID");
      goto error;
    }
    Document->AnnotationMaxID = maxId;

  /* Read Attributes */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    if (sscanf(&Line[2], "<AttributeSet items=\"%d\">", &items) < 1) {
      CDM_ReportError("Cannot read items from AttributeSet");
      goto error;
    }
    for (i=0; i<items; i++) {
      Tcl_SetObjLength(LineObj, 0);
      if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
      Line = Tcl_GetString(LineObj);
      tmp = strstr(Line, "</Attribute>"); if (tmp) *tmp = '\0';
      CDM_PutAttribute(Document, Tcl_NewStringObj(&Line[15], -1));
      if (tmp) *tmp = '<';
    }
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;

  /* Read Annotations */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    if (sscanf(&Line[2], "<AnnotationSet items=\"%d\">", &items) < 1) {
      CDM_ReportError("Cannot read items from AnnotationSet");
      goto error;
    }
    for (i=0; i<items; i++) {
      Tcl_SetObjLength(LineObj, 0);
      if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
      Line = Tcl_GetString(LineObj);
      tmp = strstr(Line, "</Annotation>"); if (tmp) *tmp = '\0';
      CDM_AddAnnotation(Document, Tcl_NewStringObj(&Line[16], -1));
      if (tmp) *tmp = '<';
    }
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    
  /* Read Text */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    if (sscanf(&Line[2], "<Text length=%d>", &items) < 1) {
      CDM_ReportError("Cannot read text");
      goto error;
    }
    Tcl_SetObjLength(LineObj, 0);
    chars = Tcl_ReadChars(File, LineObj, items, 0);
    if (chars != items) goto error;
    Document->RawData = Tcl_NewStringObj(Tcl_GetString(LineObj), -1);
    Tcl_IncrRefCount(Document->RawData);
    
  return TCL_OK;
error:
  Tcl_DecrRefCount(LineObj);
  return TCL_ERROR;
} /* CDM_XML_LoadDocument */

static Tcl_Obj *CDM_XML_CollectionXIdsPtr(char *dir) {
  Tcl_Channel File;
  Tcl_Obj *LineObj = NULL, *Result;
  char *FileName, *Line, *tmp;
  int i, items;
  CDM_THREAD_SPECIFIC_DATA;
  
  /*
   * Open the file that will hold our Collection description...
   */
  FileName = CDM_JoinPaths(dir, "CDM_CollectionIndex.xml", NULL);
  if (FileName == NULL) return NULL;
  File = Tcl_OpenFileChannel(CDM_Interp, FileName, "r", 0666);
  CDM_Free(FileName);
  if (File == NULL) return NULL;

  /*
   * Set the file encoding to utf-8...
   */
  if (Tcl_SetChannelOption(CDM_Interp, File, "-encoding", "utf-8") != TCL_OK) {
     CDM_ReportError("cannot change encoding to \"utf-8\"");
     goto error;
  }
#ifdef WINDOWS
  /* Under windows, we have to disable ^Z (\x1a) as the EOF character,
     as our files can contain binary data... */
  Tcl_SetChannelOption(CDM_Interp, File, "-eofchar", "");
#endif /* WINDOWS */

  /*
   * Read Header...
   */
  LineObj = Tcl_NewStringObj(NULL, -1);
  if (LineObj == NULL) {
    CDM_ReportError("CDM_XML_LoadCollection: Tcl_NewStringObj returned NULL -"
                    " out of memory?");
    goto error;
  }
  if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
  if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
  
  /* Read Encoding */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    
  /* Read Name */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;

  /* Read Owner */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
  
  /* Read Document X/Ids */
    Result = Tcl_NewListObj(0, NULL);
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    if (sscanf(&Line[2], "<DocumentXIDSet items=\"%d\">", &items) < 1) {
      CDM_ReportError("Cannot read items from DocumentXIDSet");
      goto error;
    }
    for (i=0; i<items; i++) {
      Tcl_SetObjLength(LineObj, 0);
      if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
      Tcl_SetObjLength(LineObj, 0);
      if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
      Line = Tcl_GetString(LineObj);
      tmp = strstr(Line, "</DocumentXID>"); if (tmp) *tmp = '\0';
      Tcl_ListObjAppendElement(NULL, Result, Tcl_NewStringObj(&Line[17], -1));
      if (tmp) *tmp = '<';
    }
  
  Tcl_DecrRefCount(LineObj);
  Tcl_Close(CDM_Interp, File);
  return Result;
error:
  if (LineObj )Tcl_DecrRefCount(LineObj);
  Tcl_Close(CDM_Interp, File);
  return NULL;
} /* CDM_XML_CollectionXIdsPtr */

static Tcl_Obj *CDM_XML_CollectionIdsPtr(char *dir) {
  Tcl_Channel File;
  Tcl_Obj *LineObj = NULL, *Result;
  char *FileName, *Line, *tmp;
  int i, items;
  CDM_THREAD_SPECIFIC_DATA;
  
  /*
   * Open the file that will hold our Collection description...
   */
  FileName = CDM_JoinPaths(dir, "CDM_CollectionIndex.xml", NULL);
  if (FileName == NULL) return NULL;
  File = Tcl_OpenFileChannel(CDM_Interp, FileName, "r", 0666);
  CDM_Free(FileName);
  if (File == NULL) return NULL;

  /*
   * Set the file encoding to utf-8...
   */
  if (Tcl_SetChannelOption(CDM_Interp, File, "-encoding", "utf-8") != TCL_OK) {
     CDM_ReportError("cannot change encoding to \"utf-8\"");
     goto error;
  }
#ifdef WINDOWS
  /* Under windows, we have to disable ^Z (\x1a) as the EOF character,
     as our files can contain binary data... */
  Tcl_SetChannelOption(CDM_Interp, File, "-eofchar", "");
#endif /* WINDOWS */

  /*
   * Read Header...
   */
  LineObj = Tcl_NewStringObj(NULL, -1);
  if (LineObj == NULL) {
    CDM_ReportError("CDM_XML_LoadCollection: Tcl_NewStringObj returned NULL -"
                    " out of memory?");
    goto error;
  }
  if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
  if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
  
  /* Read Encoding */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    
  /* Read Name */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;

  /* Read Owner */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
  
  /* Read Document X/Ids */
    Result = Tcl_NewListObj(0, NULL);
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    Line = Tcl_GetString(LineObj);
    if (sscanf(&Line[2], "<DocumentXIDSet items=\"%d\">", &items) < 1) {
      CDM_ReportError("Cannot read items from DocumentXIDSet");
      goto error;
    }
    for (i=0; i<items; i++) {
      Tcl_SetObjLength(LineObj, 0);
      if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
      Line = Tcl_GetString(LineObj);
      tmp = strstr(Line, "</DocumentID>"); if (tmp) *tmp = '\0';
      Tcl_ListObjAppendElement(NULL, Result, Tcl_NewStringObj(&Line[16], -1));
      if (tmp) *tmp = '<';
      Tcl_SetObjLength(LineObj, 0);
      if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
    }
  
  Tcl_DecrRefCount(LineObj);
  Tcl_Close(CDM_Interp, File);
  return Result;
error:
  if (LineObj) Tcl_DecrRefCount(LineObj);
  Tcl_Close(CDM_Interp, File);
  return NULL;
} /* CDM_XML_CollectionIdsPtr */

static CDM_AttributeSet CDM_XML_DocumentAttrsPtr(Tcl_Channel File) {
  Tcl_Obj *LineObj, *Result;
  char *Line, *tmp;
  int items, i;
  
  /*
   * Read Header...
   */
  LineObj = Tcl_NewStringObj(NULL, -1);
  if (LineObj == NULL) {
    CDM_ReportError("CDM_XML_LoadCollection: Tcl_NewStringObj returned NULL -"
                    " out of memory?");
    goto error;
  }
  if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
  if (Tcl_GetsObj(File, LineObj) <= 0) goto error;

  /* Read Encoding */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;

  /* Read ExternalID */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;

  /* Read ID */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;
  
  /* Read AnnotationMaxID */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) <= 0) goto error;

  /* Read Attributes */
    Tcl_SetObjLength(LineObj, 0);
    if (Tcl_GetsObj(File, LineObj) < 0) goto error;
    Line = Tcl_GetString(LineObj);
    if (sscanf(&Line[2], "<AttributeSet items=\"%d\">", &items) < 1) {
      CDM_ReportError("Cannot read items from AttributeSet");
      goto error;
    }
    Result = Tcl_NewListObj(0, NULL);
    for (i=0; i<items; i++) {
      Tcl_SetObjLength(LineObj, 0);
      if (Tcl_GetsObj(File, LineObj) < 0) goto error;
      Line = Tcl_GetString(LineObj);
      tmp = strstr(Line, "</Attribute>"); if (tmp) *tmp = '\0';
      Tcl_ListObjAppendElement(NULL, Result, Tcl_NewStringObj(&Line[15], -1));
      if (tmp) *tmp = '<';
    }
    
  Tcl_DecrRefCount(LineObj);
  return Result;
error:
  Tcl_DecrRefCount(LineObj);
  return NULL;
} /* CDM_XML_DocumentAttrsPtr */

/*
 * CDM_XML_WriteStrings:
 * This function will write a variable number of strings into the provided Tcl
 * file channel. If an error occurs, -1 will be returned an error will be left
 * at CDM_Interp...
 */
static int CDM_XML_WriteStrings TCL_VARARGS_DEF(Tcl_Channel, arg1) {
  Tcl_Channel File;
  va_list argList;
  int bytes, total = 0;
  char *string;

  File = TCL_VARARGS_START(Tcl_Channel, arg1, argList);
  if (File == NULL) {
    CDM_ReportError("CDM_XML_WriteStrings: cannot get parameters!");
    return -1;
  }

  string = va_arg(argList, char *);
  while (string != NULL) {
    bytes = Tcl_WriteChars(File, string, -1);
    if (bytes == -1) {
      CDM_ReportError((char *) Tcl_ErrnoMsg(Tcl_GetErrno()));
      va_end(argList);
      return bytes;
    }
    total += bytes;
    string = va_arg(argList, char *);
  }
  va_end(argList);
  return total;
} /* CDM_XML_WriteStrings */
/*
 * XMLIO.cpp - EOF -
 */
