###########################################################################
# Document Section ########################################################
###########################################################################

##############################################################################
##  tip_Annotate.tcl:                                                       ##
## -------------------                                                      ##
##  This file implements various functions related to initialising, running ##
##  and finalising Annotators (Linguistic Processing Components).           ##
##                                                                          ##
##  Copyright 1998-2006 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

# tip_Annotate 
#
proc tip_Annotate {document annotator args} {
  global GUI_Text errorInfo CDM_CurrentStatus \
         CDM_FullGateCompatibility ::CDM::ComponentMode
  if {![llength [info procs    $annotator]] && \
      ![llength [info commands $annotator]]} {
    return \
        [list Error "$GUI_Text(AnnotatorNotLoaded)\n($annotator)" {}]
  }

  ## If the module has registered a Compatibility Mode, put CDM in this mode
  if {[info exists ::CDM::ComponentMode($annotator)]} {
    set GateCompatibility $CDM_FullGateCompatibility
    set CDM_FullGateCompatibility $::CDM::ComponentMode($annotator)
  }

  ## Important: Now, all strings are in UNICODE. We have already red the
  ## document files of collection in the proper encoding. Usually, from an
  ## annotator will open files and dump the document text. But we don't
  ## expect the user to know about encodings, nor we want to use specialised
  ## functions for doing the opening of the destination file. So, before
  ## annotators are called, we RENAME the open command, to one that opens
  ## and fconfigures the opened file for the proper encoding. After the
  ## annotator has run, open gets restored...
  if {![info exists CDM_CurrentStatus(ExecRenamed)]} {
    set CDM_CurrentStatus(ExecRenamed) 0
  }
  if {!$CDM_CurrentStatus(ExecRenamed)} {
    catch {rename open CDM_OriginalOpen}
    catch {
      rename CDM_OpenWithCorrectEncoding open
      set CDM_CurrentStatus(ExecRenamed) 1
      set Renamed 1
    }
    ## Make GDM to save collections into GATE format, while
    ## the annotator runs...
    set CDM_CurrentStatus(NowAnnotating) 1
    if {![info exists CDM_CurrentStatus(CurrentCollectionEncoding)]} {
      set CDM_CurrentStatus(CurrentCollectionEncoding) \
          [tip_GetEncoding $document]
    }
  }
  
  append CDM_CurrentStatus(TimeLog) "Plug-In Module $annotator:\n"
  if {[catch {time "$annotator $document $args"} result]} {
    append CDM_CurrentStatus(TimeLog) "Error: \n\t$result\n\n"
    set result [list Error $result $errorInfo]
  } else {
    append CDM_CurrentStatus(TimeLog) [expr {[lindex $result 0] / 1000000.0}]
    append CDM_CurrentStatus(TimeLog) \
        " seconds for processing document [tip_GetId $document]\n" \
        "\t([tip_GetExternalId $document]).\n\n"
    set result {Annotate Ok}
  }
  ## Restore Original tcl's open command
  if {[info exists Renamed] && $Renamed == 1} {
    catch {rename open CDM_OpenWithCorrectEncoding}
    catch {rename CDM_OriginalOpen open}
    set CDM_CurrentStatus(ExecRenamed) 0
  }
  set CDM_CurrentStatus(NowAnnotating) 0
  ## If the module has registered a Compatibility Mode, restore the default
  ## value, requested by the user...
  if {[info exists ::CDM::ComponentMode($annotator)]} {
    set CDM_FullGateCompatibility $GateCompatibility
  }

  return $result
}

## tip_RunTool
#  Runs a registered tool...
proc tip_RunTool {col doc tool} {
  global CDM_CurrentStatus GUI_Text errorInfo \
         ::CDM::ComponentMode CDM_FullGateCompatibility
  ## If the module has registered a Compatibility Mode, put CDM in this mode
  if {[info exists ::CDM::ComponentMode($tool)]} {
    set GateCompatibility $CDM_FullGateCompatibility
    set CDM_FullGateCompatibility $::CDM::ComponentMode($tool)
  }
  if {[string length $col]} {
    set enc [tip_GetEncoding $col]
    set CDM_CurrentStatus(CurrentCollectionEncoding) $enc
    ## See tip_Annotate...
    if {![info exists CDM_CurrentStatus(ExecRenamed)]} {
      set CDM_CurrentStatus(ExecRenamed) 0
    }
    if {!$CDM_CurrentStatus(ExecRenamed)} {
      catch {rename open CDM_OriginalOpen}
      catch {rename CDM_OpenWithCorrectEncoding open}
      set CDM_CurrentStatus(ExecRenamed) 1
      set Renamed 1
      ## Make CDM to save collections into GATE format, while
      ## the annotator runs...
      set CDM_CurrentStatus(NowAnnotating) 1
    }
  }
  set result [catch {creole_$tool $doc $col} error]
  ## If the module has registered a Compatibility Mode, restore the default
  ## value, requested by the user...
  if {[info exists ::CDM::ComponentMode($tool)]} {
    set CDM_FullGateCompatibility $GateCompatibility
  }
  if {[string length $col]} {
    ## Restore Original tcl's open command
    if {[info exists Renamed] && $Renamed == 1} {
      catch {rename open CDM_OpenWithCorrectEncoding}
      catch {rename CDM_OriginalOpen open}
      set CDM_CurrentStatus(ExecRenamed) 0
      set CDM_CurrentStatus(NowAnnotating) 0
    }
  }
  if {$result} {
    error [format $GUI_Text(CannotExecuteTool) $tool $error] $errorInfo
  }
}

## CDM_OpenWithCorrectEncoding
#    Replacement for open while an annotator is running.
proc CDM_OpenWithCorrectEncoding {name {access {}} {permissions {}}} {
  if {[string length $permissions]} {
    set Cmd1 [list CDM_OriginalOpen $name $access $permissions]
    set Cmd2 [list open $name $access $permissions]
  } elseif {[string length $access]} {
    set Cmd1 [list CDM_OriginalOpen $name $access]
    set Cmd2 [list open $name $access]
  } else {
    set Cmd1 [list CDM_OriginalOpen $name]
    set Cmd2 [list open $name]
  }

  ## In order to avoid a recursive call, we check the name of this procedure.
  ## We have to avoid calling a procedure with the same name...
  switch [lindex [info level 0] 0] {
    open                        {set Cmd $Cmd1}
    CDM_OpenWithCorrectEncoding {set Cmd $Cmd2}
    default {
      ## A name we have never seen before...
      if {[llength [info commands CDM_OriginalOpen]]} {
        set Cmd $Cmd1
      } else {
        set Cmd $Cmd2
      }
    }
  }
    
  if {[catch $Cmd File]} {
    gut_error $File
  }
  fconfigure $File -encoding $::CDM_CurrentStatus(CurrentCollectionEncoding)
  return $File
};# CDM_OpenWithCorrectEncoding

## tip_RunAnnotator
#   This function will accept an Annotator, it will get the Annotator's
#   default arguments and call tip_Annotate...
proc tip_RunAnnotator {document annotator {paramSpec {}} } {
  if {[catch {GUI_SystemWindowGetInfo [tip_GetParent $document] \
                                      CollectionOpenInSystem} inSystem]} {
    set inSystem 0
  }
  if {$inSystem} {
    if {[catch {GUI_SystemWindowGetInfo [tip_GetParent $document] \
                                     ModuleOptionValues $annotator} Val]} {
      foreach {Val Conf} [GUI_GetModuleOptions $annotator] {break}
    }
  } else {
    foreach {Val Conf} [GUI_GetModuleOptions $annotator] {break}
  }
  ## Replace any parameter values...
  foreach {pos value} $paramSpec {
    set Val [lreplace $Val $pos $pos $value]
  }
  return [eval tip_Annotate [list $document] [list creole_$annotator] $Val]
};# tip_RunAnnotator

## tip_InitialiseAnnotator
#   This function will accept an Annotator, it will get the Annotator's
#   default arguments and call its initialisation function...
proc tip_InitialiseAnnotator {document annotator {parent {}} {paramSpec {}} } {
  if {[llength [info commands creole_${annotator}_Initialize]]} {
    if {[catch {GUI_SystemWindowGetInfo [tip_GetParent $document] \
                                        CollectionOpenInSystem} inSystem]} {
      set inSystem 0
    }
    if {$inSystem} {
      if {[catch {GUI_SystemWindowGetInfo [tip_GetParent $document] \
                                     ModuleOptionValues $annotator} Val]} {
        foreach {Val Conf} [GUI_GetModuleOptions $annotator] {break}
      }
    } else {
      foreach {Val Conf} [GUI_GetModuleOptions $annotator] {break}
    }
    if {![string length $parent]} {
      if {[catch {tip_GetParent $document} parent]} {
        set parent {}
      }
    }
    ## Replace any parameter values...
    foreach {pos value} $paramSpec {
      set Val [lreplace $Val $pos $pos $value]
    }
    if {[catch {eval [list creole_${annotator}_Initialize] [list $parent] \
                     [list $document] $Val} InitError]} {
      global errorInfo
      return [list Error $InitError $errorInfo]
    }
  }
  return {Annotate Ok}
};# tip_InitialiseAnnotator

## tip_FinishAnnotator
#   This function will accept an Annotator, it will get the Annotator's
#   default arguments and call its finalisation function...
proc tip_FinishAnnotator {document annotator {parent {}} {paramSpec {}} } {
  if {[llength [info commands creole_${annotator}_Finish]]} {
    if {[catch {GUI_SystemWindowGetInfo [tip_GetParent $document] \
                                        CollectionOpenInSystem} inSystem]} {
      set inSystem 0
    }
    if {$inSystem} {
      if {[catch {GUI_SystemWindowGetInfo [tip_GetParent $document] \
                                     ModuleOptionValues $annotator} Val]} {
        foreach {Val Conf} [GUI_GetModuleOptions $annotator] {break}
      }
    } else {
      foreach {Val Conf} [GUI_GetModuleOptions $annotator] {break}
    }
    if {![string length $parent]} {
      if {[catch {tip_GetParent $document} parent]} {
        set parent {}
      }
    }
    ## Replace any parameter values...
    foreach {pos value} $paramSpec {
      set Val [lreplace $Val $pos $pos $value]
    }
    if {[catch {eval [list creole_${annotator}_Finish] [list $parent] \
                     [list $document] $Val} InitError]} {
      global errorInfo
      return [list Error $InitError $errorInfo]
    }
  }
  return {Annotate Ok}
};# tip_FinishAnnotator

## tip_ForeachDocumentInCollection
#  This function will apply "body" in each Document in the provided
#  Collection. During body execution, the current Document will be placed in
#  the "doc" variable...
proc tip_ForeachDocumentInCollection {col body {statusWin {}}} {
  global GUI_Text
  upvar doc doc
  if {[info exists doc]} {
    set Document $doc
  }
  ## Has the caller provided a status window? If not, create one...
  if {[string length $statusWin]} {
    set wwin $statusWin
  } else {
    set wwin [ggi_wait_message $GUI_Text(PleaseWait) 0.0]
  }
  set doc [tip_FirstDocument $col]
  set Length [tip_Length $col]
  set Step [expr {100.0/($Length+1)}]
  set Progress 0.0
  while {[string length $doc]} {
    ggi_wait_message "$GUI_Text(ProcessingDocument): [tip_GetId $doc]" \
                     $Progress 1 $wwin
    ## uplevel 1 "set doc $doc"
    ## Execute the provided body...
    if {[catch { uplevel 1 $body } result]} {
      return -errorinfo [CDM_ErrorInfoAsCaller uplevel \
                    tip_ForeachDocumentInCollection]  \
		    -errorcode $::errorCode -code error $result
    }
    set Progress [expr {$Progress+$Step}]
    set doc [tip_NextDocument $col]
  }
  if {[info exists Document]} {
    set doc $Document
  }
  if {![string length $statusWin]} {
    ggi_destroy_wait $wwin
  }
};# tip_ForeachDocumentInCollection

## Functions used internally by the CDM (i.e. its memory cache...)
proc ::CDM::DisplayProgressInfoDialog {message {percent 0}
                                       {title {Information...}}} {
  global ApplicationInfo
  ## If Tk is available, display our window...
  if {[info exists ApplicationInfo(EnableGUI)] && $ApplicationInfo(EnableGUI)} {
    set window .mw.cdm_cdm_DisplayProgressInfoDialog
    set win    $window; set index 0
    while {[winfo exists $win]} {
      set win ${window}_$index
      incr index
    }
    ## Create the window...
    frame $win

    ## Place a label with the message and the progress bar...
    label $win.message -text $message
    frame $win.prframe -relief sunken -bd 1
    Progress $win.prframe.progress
    SetProgress $win.prframe.progress $percent
    pack $win.message  -fill both -expand 1 -padx 20 -pady 10
    pack $win.prframe.progress -fill both -expand 1
    pack $win.prframe  -fill x    -expand 0 -padx 20 -pady 10
    update idletasks
    place $win -relx 0.5 -rely 0.5 -anchor c
    update
    return $win
  } else {
    ## Print the message into stdout...
    catch {puts $message}
    return {}
  }
  update
};# ::CDM::DisplayProgressInfoDialog

proc ::CDM::UpdateProgressInfoDialog {win progress} {
  SetProgress $win.prframe.progress $progress
  update
};# ::CDM::DestroyProgressInfoDialog

proc ::CDM::DestroyProgressInfoDialog {win} {
  destroy $win
  update
};# ::CDM::DestroyProgressInfoDialog
