##############################################################################
##  CollectionHandlers.tcl:                                                 ##
## -------------------------                                                ##
##  This file contains some useful Collection Handlers, like handlers for   ##
##  importing HTML documents along with their images.                       ##
##                                                                          ##
##  Copyright 1998-2004 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################
 
namespace eval ::CollectionHandlers {
  variable ConvertSGMLTags2Anns
  variable ConvertSGMLTags2Anns_AttributePattern \
    {[A-Za-z0-9_:.-]+(?:\s*\=\s*(?:\"[^"]+\"|\'[^']+\'))?}
}

## ::CollectionHandlers::Initialise
#  Use this function to set any initial parameters...
proc ::CollectionHandlers::Initialise {} {
  catch {ImportHTML&ImagesInit}
};# ::CollectionHandlers::Initialise

##############################################################################
##############################################################################
####                                                                      ####
####             H T M L   I m p o r t    H a n d l e r                   ####
####                                                                      ####
##############################################################################
##############################################################################

proc ::CollectionHandlers::ImportHTML&ImagesInit {{base64 1}} {
  if {[llength [package versions base64]]} {
    package require base64
    set ::CollectionHandlers::ImportHTML&Images(Base64) $base64
  } else {
    set ::CollectionHandlers::ImportHTML&Images(Base64) 0
  }
};# ::CollectionHandlers::ImportHTML&ImagesInit

## ::CollectionHandlers::ImportHTML&Images
#  A Collection handler that will try to import the images contained in a HTML
#  page, along with the HTML text...
#  The images will be added in the Document as Attributes: the Attribute name
#  will be the url of the image, where the Attribute value will be a binary
#  string containing the image encoded as gif, even if the original image
#  format was something different, like jpeg or png...
proc ::CollectionHandlers::ImportHTML&Images {col doc args} {
  ## Create an HTML widget and load the HTML into the widget. If such a widget
  ## cannot be created, simply return...
  set html .html_widget_ImportHTML&Images__$col
  catch {destroy $html}
  if {[catch {html $html -imagecommand \
     "::CollectionHandlers::ImportHTML&ImagesImageCallback $doc $html" \
     -bgimagecommand \
     "::CollectionHandlers::ImportHTML&ImagesImageCallback $doc $html"}]} {
    return
  }
  $html configure -base [tip_GetExternalId $doc]
  $html parse [tip_GetRawData $doc]
  destroy $html
};# ::CollectionHandlers::ImportHTML&Images

proc ::CollectionHandlers::ImportHTML&Images_OptionsDialog {win} {
  if {[winfo exists $win]} {return}
  toplevel $win
  wm withdraw $win
  wm transient $win [winfo parent $win]
  wm title $win {Import HTML Images Handler Options...}
  checkbutton $win.base64 -text {Store Images encoded in Base64 format} \
    -variable ::CollectionHandlers::ImportHTML&Images(Base64)
  if {[catch {package require base604}]} {
    ## Base64 package is not available!
    set ::CollectionHandlers::ImportHTML&Images(Base64) 0
    $win.base64 configure -state disabled -text {Package "base64" not present!}
  }
  pack $win.base64 -fill x -expand 1 -padx 4 -pady 10
  pack [button $win.close -textvariable ::GUI_Text(Close) -command \
               "destroy $win"] -padx 4 -pady 4
  gut_ModalDlg $win
  wm deiconify $win
};# ::CollectionHandlers::ImportHTML&Images_OptionsDialog

## ::CollectionHandlers::ImportHTML&ImagesImageCallback
#  The handler responsible for retrieving the images contained in an HTML
#  Document...
proc ::CollectionHandlers::ImportHTML&ImagesImageCallback {doc widget args} {
  ## The arguments are: src=, width=, height=, list of all other values...
  set Base64 [set ::CollectionHandlers::ImportHTML&Images(Base64)]
  foreach {w h} {1 1} {}
  foreach {fn w h spec} $args {break}
  #puts "Fetching image: \"$fn\"..."

  ## Is the location a local file we can open?
  if {[file exists $fn]} {
    #puts "Trying $fn..."
    set file [open $fn]
    fconfigure $file -translation binary
    set data [read $file [file size $fn]]
    close $file
    binary scan $data {}
    if {$Base64} {
      set data [::base64::encode -wrapchar {} $data]
      tip_PutAttribute $doc [list $fn [list CDM_BASE64_IMAGE $data]]
    } else {
      tip_PutAttribute $doc [list $fn [list CDM_IMAGE $data]]
    }
    return Blank
  }

  ## If it is a local file, then maybe it has an incomplete uri. Try to
  ## reconstruct the original uri...
  set base [$widget cget -base]
  switch -glob -- $base {
    http://* {set base [string range $base 7 end]}
    http:/*  {set base [string range $base 6 end]}
    file:*   {set base [string range $base 5 end]}
    default  {}
  }

  ## Is the location a local file we can open?
  if {[regexp {[A-Z]\:.*} $fn]} {
    foreach File [list [file dirname $base]/[string range $fn 2 end] \
                       $base/[string range $fn 2 end]] {
      if {[file exists $File]} {
        #puts "Trying $File..."
        set file [open $File]
        fconfigure $file -translation binary
        set data [read $file [file size $File]]
        close $file
        binary scan $data {}
        if {$Base64} {
          set data [::base64::encode -wrapchar {} $data]
          tip_PutAttribute $doc [list $fn [list CDM_BASE64_IMAGE $data]]
        } else {
          tip_PutAttribute $doc [list $fn [list CDM_IMAGE $data]]
        }
        return Blank
      }
    }
  }
  
  ## A local file?
  foreach File [list [file dirname $base]/$fn $base/$fn] {
    if {[file exists $File]} {
      #puts "Trying2 $File..."
      set file [open $File]
      fconfigure $file -translation binary
      set data [read $file [file size $File]]
      close $file
      binary scan $data {}
      if {$Base64} {
        set data [::base64::encode -wrapchar {} $data]
        tip_PutAttribute $doc [list $fn [list CDM_BASE64_IMAGE $data]]
      } else {
        tip_PutAttribute $doc [list $fn [list CDM_IMAGE $data]]
      }
      return Blank
    }
  }
  return Blank
  
  ## We failed to find the image locally. We have to download the image...
  ## Does our uri start from http or www.?
  if {[string match http* $fn] || [string match www.* $fn]} {
    set urls $fn
  } elseif {[string length [file extension $base]]} {
    set urls [list http://[file dirname $base]/$fn http://$base/$fn]
  } else {
    set urls [list http://$base/$fn http://[file dirname $base]/$fn]
  }
  
  ## We failed to load the image from disk. Then it might be located over
  ## Internet. DO NOT DOWNLOAD it now. Just create a blank image, and we
  ## will download the actual data later...
  if {[string length $w] && [string length $h]} {
    set img [image create photo -width $w -height $h]
  } else {
    set img [image create photo]
  }
};# ::CollectionHandlers::ImportHTML&ImagesImageCallback

## ::CollectionHandlers::ImportHTMLDocumentMetadataDirectory
#  This handler will copy the directory that accompanies an HTML document.
proc ::CollectionHandlers::ImportHTMLDocumentMetadataDirectory {col doc args} {
  set xid [tip_GetExternalId $doc]
  set metadata [file rootname $xid]_files
  if {[file isdirectory $metadata]} {
    set target_dir [tip_GetName $col]/Metadata/HtmlMetadata
    set target_name [string map [list \u00a0 { }] [file tail $metadata]]
    file mkdir $target_dir
    if {[file isdirectory $target_dir/$target_name]} {
      file copy -force $metadata $target_dir
    } else {
      file copy -force $metadata $target_dir/$target_name
    }
  }
};# ::CollectionHandlers::ImportHTMLDocumentMetadataDirectory


##############################################################################
##############################################################################
####                                                                      ####
#### S G M L   T o   A n n o t a t i o n s   I m p o r t    H a n d l e r ####
####                                                                      ####
##############################################################################
##############################################################################

## ::CollectionHandlers::ConvertSGMLTags2Anns
#  A Collection handler that will try to strip SGML/HTML/XML documents from
#  their SGML tags and convert them into Annotations...
package require struct 
package require cmdline
package require htmlparse
proc ::CollectionHandlers::ConvertSGMLTags2Anns {col doc args} {

  set Text [string trim [tip_GetRawData $doc]]
  unset -nocomplain ::CollectionHandlers::ConvertSGMLTags2AnnsTagCallback
  set ::CollectionHandlers::ConvertSGMLTags2Anns(Text) {}
  set ::CollectionHandlers::ConvertSGMLTags2Anns(Doc)  $doc
  set ::CollectionHandlers::ConvertSGMLTags2Anns(Col)  $col
  ::htmlparse::parse -cmd \
    ::CollectionHandlers::ConvertSGMLTags2AnnsTagCallback $Text

  ## Reset the Document Text...
  tip_SetRawData $doc $::CollectionHandlers::ConvertSGMLTags2Anns(Text)
  unset -nocomplain ::CollectionHandlers::ConvertSGMLTags2AnnsTagCallback \
                    ::CollectionHandlers::ConvertSGMLTags2Anns
};# ::CollectionHandlers::ConvertSGMLTags2Anns

## ::CollectionHandlers::ConvertSGMLTags2AnnsTagCallback
#  This procedure is invoked to handle each one of the tags found during the
#  SGML parsing...
proc ::CollectionHandlers::ConvertSGMLTags2AnnsTagCallback {
                                                        tag slash param text} {
  # puts "tag=\"$tag\", slash=\"$slash\" param=\"$param\" text=\"$text\""
  ## Is it the first/last time we are called?
  if {[string equal $tag hmstart]} {return}

  ## If there is text before the tag, append it to our striped text..
  set offset [string length $::CollectionHandlers::ConvertSGMLTags2Anns(Text)]
  if {[string length $text]} {
    append ::CollectionHandlers::ConvertSGMLTags2Anns(Text) $text
  }

  ## Is it a start or and end tag?
  if {[string length $slash]} {
    ## This is an end tag. Have we seen this tag before? If so, add an
    ## Annotation and forget the last tag of this type...
    if {[info exists \
      ::CollectionHandlers::ConvertSGMLTags2AnnsTagCallback($tag)]} {
      set spec [lindex \
              $::CollectionHandlers::ConvertSGMLTags2AnnsTagCallback($tag) end]
      set ::CollectionHandlers::ConvertSGMLTags2AnnsTagCallback($tag) [lrange \
         $::CollectionHandlers::ConvertSGMLTags2AnnsTagCallback($tag) 0 end-1]
      if {[llength $spec] != 2} {return}

      ## Create an Annotation...
      foreach {start attrs} $spec {break}
      set AttrSet {}
      foreach attr $attrs {
        foreach {name val} $attr {break}
        lappend AttrSet [tip_CreateAttribute $name \
                            [tip_CreateAttributeValue GDM_STRING $val]]
      }
      tip_AddAnnotation $::CollectionHandlers::ConvertSGMLTags2Anns(Doc) \
          [tip_CreateAnnotation $tag \
                                [tip_CreateSpanSet $start [expr {$offset-0}]] \
                                $AttrSet]
    } else {
      ## A tag of the form <??? />...
      ## Does it has any parameters?
      set Parameters {}
      foreach parameter [regexp -all -inline -- \
        $::CollectionHandlers::ConvertSGMLTags2Anns_AttributePattern $param] {
        set value {}
        foreach {name value} [split $parameter =] {break}
        lappend Parameters [list $name [string trim $value \"]]
      }
      ## Create an Annotation...
      set AttrSet {}
      foreach attr $Parameters {
        foreach {name val} $attr {break}
        lappend AttrSet [tip_CreateAttribute $name \
                            [tip_CreateAttributeValue GDM_STRING $val]]
      }
      tip_AddAnnotation $::CollectionHandlers::ConvertSGMLTags2Anns(Doc) \
          [tip_CreateAnnotation $tag \
                                [tip_CreateSpanSet $offset $offset] \
                                $AttrSet]
    }
  } else {
    ## This is a start tag. Does it has any parameters?
    set Parameters {}
    foreach parameter [regexp -all -inline -- \
      $::CollectionHandlers::ConvertSGMLTags2Anns_AttributePattern $param] {
      set value {}
      foreach {name value} [split $parameter =] {break}
      lappend Parameters [list $name [string trim $value \"]]
    }
    lappend ::CollectionHandlers::ConvertSGMLTags2AnnsTagCallback($tag) \
      [list $offset $Parameters]
  }
};# ::CollectionHandlers::ConvertSGMLTags2AnnsTagCallback
