##############################################################################
##  CompileComponent.tcl:                                                   ##
## -----------------------                                                  ##
##  This file implements utilities and an interface that can be used to:    ##
##  1) Create a Compilation configuration file in a component, if it does   ##
##     not already exists.                                                  ##
##  2) Compile the component and build its shared library.                  ##
##  The utilities in this file make use of the tBuild extension that        ##
##  handles all the low level details, like interacting with the compiler   ##
##  and setting the needed options in a cross-platform way...               ##
##                                                                          ##
##  Copyright 1998-2004 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

package require tBuild
GUI_LoadImage ISourceCpp
GUI_LoadImage IBinaryFile
GUI_LoadImage IDynamicComponent
GUI_LoadImage IAddArrowGreen

namespace eval ::CompileComponent {
}

## GenerateDefaultCompilationConfiguration
## This function will generate the default Compilation configuration file for
## a component, if it does not already exist.
proc ::CompileComponent::GenerateDefaultCompilationConfiguration {component
  {force 0} {PreCompileScript {}} {CppFiles {}} {Library {}}
  {PostCompileScript {}} {Debug 0} {Defines {}}} {
  ## A non-dynamic component does not need compilation...
  if {[catch {PUL_plugin_module_config $component coupling} coupling]} {
    error "Unknown Coupling of Component \"$component\"!"
  }
  if {![regexp {dynamic|c\+\+} [string tolower $coupling]]} {return}
  global creole_${component}_home
  set CompileConfFilename [file join [set creole_${component}_home] \
                                CompileComponent.cfg]
  if {![string length $Library]} {
    set Library lib${component}
  }
  set Library [file rootname $Library]
  if {![llength $CppFiles]} {
    set CppFiles $component.cpp
  }
  if {!$force} {
    if {[file exists $CompileConfFilename]} {return}
  }
  set file [open $CompileConfFilename w]
  puts $file "##########################################################"
  puts $file "## Component Compilation Information."
  puts $file "## Auto-generated by the $::ApplicationName platform."
  puts $file "##########################################################"
  puts $file ""
  puts $file "set PreCompileScript  \{\n$PreCompileScript\n\}"
  puts $file "set Defines           \"$Defines\""
  puts $file "set CppFiles          \"$CppFiles\""
  puts $file "set Library           \"$Library\""
  puts $file "set PostCompileScript \{\n$PostCompileScript\n\}"
  puts $file "set Debug             \"$Debug\""
  puts $file ""
  puts $file "##"
  puts $file "## End of file"
  puts $file "##"
  close $file
};# ::CompileComponent::GenerateDefaultCompilationConfiguration

## ::CompileComponent::SaveConfiguration
## The a customised configuration for a component...
proc ::CompileComponent::SaveConfiguration {component win} {
  global $win

  set MainArea $win.main_area
  set ${win}(PreCompileScript) [string trim \
    [$MainArea.parameters.pree.text get 1.0 end]]
  set ${win}(PostCompileScript) [string trim \
    [$MainArea.parameters.poste.text get 1.0 end]]
  
  GenerateDefaultCompilationConfiguration $component 1 \
    [set ${win}(PreCompileScript)] [set ${win}(CppFiles)] \
    [set ${win}(Library)] [set ${win}(PostCompileScript)] \
    [set ${win}(Debug)] [set ${win}(Defines)]
};# ::CompileComponent::SaveConfiguration

## ::CompileComponent::Compile
## This function will create a dialog through which the user can set various
## options and finally compile the component...
proc ::CompileComponent::Compile {component} {
  ## Has the component a default Compile Configuration file? If not, create
  ## one...
  GenerateDefaultCompilationConfiguration $component

  ## Load the existing configuration file...
  global creole_${component}_home GUI_Text GUI_Options
  set CompileConfFilename [file join [set creole_${component}_home] \
                                CompileComponent.cfg]
  set InitDir [set creole_${component}_home]
  if {[catch {source $CompileConfFilename} error]} {
    gut_error \
      "[format $GUI_Text(ErrorCompilationConfig) $CompileConfFilename]\n$error"
  }
  foreach var {PreCompileScript CppFiles Library PostCompileScript} {
    if {![info exists $var]} {
      gut_error "[format $GUI_Text(ErrorCompilationConfig) \
         $CompileConfFilename]\nVariable \"$var\" not defined!"
    }
  }
  if {![info exists Debug]}   {set Debug    0}
  if {![info exists Defines]} {set Defines {}}

  ## We are ready to create the Compilation window...
  ## 1) Create the window
  set win [ggi_unused_win .compileComponent]
  toplevel $win
  wm withdraw $win
  wm title $win "Compiling Component: \"$component\"..."

  ## 2) Create the window toolbar...
  set toolbar_frame $win.tf; 
  Toolbar::ToolbarFrame $toolbar_frame \
          -relief $GUI_Options(ToolbarFrameRelief) \
          -borderwidth $GUI_Options(ToolbarFrameReliefBorderWidth) \
          -decorate $GUI_Options(ToolbarDecorate)

    set toolbar $toolbar_frame.tb1
    Toolbar::create $toolbar -relief $GUI_Options(ToolbarRelief) \
        -borderwidth $GUI_Options(ToolbarReliefBorderWidth) -hide 0
    Toolbar::addbutton $toolbar -tooltipvar GUI_Text(Close) \
         -image ITExit -command "destroy $win"
    pack $toolbar -side left -fill y
  pack $toolbar_frame -side top -fill x

  ## 3) Our main area frame
  set MainArea $win.main_area
  frame $MainArea -relief flat
    ## Create a Header...
    frame $MainArea.header -relief flat -bg #8fbc8f
      label $MainArea.header.info -image ISourceCpp -text "$GUI_Text(Compile):"\
        -compound left -fg #008000 -bg #8fbc8f \
        -font {arial 14 bold italic} -anchor e
      label $MainArea.header.comp -font {arial 14 bold italic} \
        -anchor w -bg #8fbc8f -fg #6a5acd -text $component...
      grid $MainArea.header.info -sticky snew -padx 0 -pady 0
      grid $MainArea.header.comp -row 0 -column 1 -sticky snew -padx 2 -pady 0
      grid columnconfigure $MainArea.header 1 -weight 1
    pack $MainArea.header -fill x

    ## The parameters section...
    labelframe $MainArea.parameters -text $GUI_Text(Parameters):
      ## Pre-Compile script:
      label  $MainArea.parameters.pre -text {Pre-Compile Script:} -anchor w
      ::ScrolledWidgets::ScrolledText $MainArea.parameters.pree \
        -height 4 -width 40 -wrap none
      $MainArea.parameters.pree.text insert end [string trim $PreCompileScript]
      set ::${win}(PreCompileScript) [string trim $PreCompileScript]
      ## Defines:
      label  $MainArea.parameters.def -text {Defines:} -anchor w
      entry  $MainArea.parameters.defe -width 40 \
             -textvariable ${win}(Defines)
      $MainArea.parameters.defe insert end $Defines
      ## C++ Files:
      label  $MainArea.parameters.cpp -text {C++ Files:} -anchor w
      entry  $MainArea.parameters.cppe -width 40 \
             -textvariable ${win}(CppFiles)
      $MainArea.parameters.cppe insert end $CppFiles
      button $MainArea.parameters.cppb -compound left -image ITOpen \
        -textvariable GUI_Text(SelectFile) \
        -command "GUI_SelectFile $MainArea.parameters.cppe \
                  -initialdir $InitDir -relativedir $InitDir -multiple 1"
      GUI_MakeEntryFileDropTarget $MainArea.parameters.cppe
      ## Library name
      label  $MainArea.parameters.lib -text {Output Library Name:} -anchor w
      entry  $MainArea.parameters.libe -width 40 \
             -textvariable ${win}(Library)
      $MainArea.parameters.libe insert end $Library[info shared]
      ## Post-Compile script:
      label  $MainArea.parameters.post -text {Post-Compile Script:} -anchor w
      ::ScrolledWidgets::ScrolledText $MainArea.parameters.poste \
        -height 4 -width 40 -wrap none -maphorizontal 1 -mapvertical 1
      $MainArea.parameters.poste.text insert end \
        [string trim $PostCompileScript]
      set ::${win}(PostCompileScript) [string trim $PostCompileScript]
      ## Debug:
      set ::${win}(Debug) $Debug
      checkbutton $MainArea.parameters.dbg -text {Enable Debug} \
        -variable ${win}(Debug)

      grid $MainArea.parameters.pre -sticky snew -padx 1 -pady 1
      grid $MainArea.parameters.pree -row 0 -column 1 -columnspan 2 \
           -sticky snew -padx 1 -pady 1
      grid $MainArea.parameters.def $MainArea.parameters.defe \
           -sticky snew -padx 1 -pady 1
      grid $MainArea.parameters.cpp $MainArea.parameters.cppe \
           $MainArea.parameters.cppb -sticky snew -padx 1 -pady 1
      grid $MainArea.parameters.lib $MainArea.parameters.libe \
           -sticky snew -padx 1 -pady 1
      grid $MainArea.parameters.post -sticky snew -padx 1 -pady 1
      grid $MainArea.parameters.poste -row 4 -column 1 -columnspan 2 \
           -sticky snew -padx 1 -pady 1
      grid $MainArea.parameters.dbg -sticky snew -padx 1 -pady 1 -columnspan 3
      grid columnconfigure $MainArea.parameters 1 -weight 1
      grid rowconfigure $MainArea.parameters 0 -weight 1
      grid rowconfigure $MainArea.parameters 4 -weight 1
      
    pack $MainArea.parameters -fill both -expand 1

    ## Compilation Options...
    labelframe $MainArea.comppars -text $GUI_Text(CompilationParameters):
    set i 0
    foreach key {OS OSVersion Compiler CompileFlag CompileSharedFlag
      TclIncludeDirs ObjectSuffix OptimiseFlags
      WarningFlags LinkSharedLib LinkLibPaths LinkLibs} {
      label  $MainArea.comppars.l$i -text $key: -anchor w
      entry  $MainArea.comppars.e$i -width 40
      $MainArea.comppars.e$i insert end $::tBuild::tBuild($key)
      $MainArea.comppars.e$i configure -state readonly
      grid $MainArea.comppars.l$i $MainArea.comppars.e$i \
           -sticky snew -padx 1 -pady 1
      incr i
    }
    grid columnconfigure $MainArea.comppars 1 -weight 1
    pack $MainArea.comppars -fill x
    
  pack $MainArea -fill both -expand 1 -padx 2 -pady 2
  
  ## 4) The button section...
  set ButtonArea $win.buttons
  frame $ButtonArea -relief flat
    ## Add a horizontal separator:-)
    frame $ButtonArea.sep -relief sunken -bd 1 -height 2
    button $ButtonArea.compile -textvariable GUI_Text(Compile) \
      -image ITCompile -compound left -command \
      "::CompileComponent::ExecuteCompile $component $win"
    button $ButtonArea.save -textvariable GUI_Text(Save) \
      -image ITSave -compound left -command \
      "::CompileComponent::SaveConfiguration $component $win"
    button $ButtonArea.cancel -textvariable GUI_Text(Close) \
      -image ITCancel -compound left -command "destroy $win"
    bind $ButtonArea.cancel <Destroy> "unset -nocomplain $win"
    grid $ButtonArea.sep -sticky ew -padx 5 -columnspan 3
    grid $ButtonArea.compile -row 1 -column 0 -sticky e -padx 10 -pady 10
    grid $ButtonArea.save    -row 1 -column 1 -sticky w -padx 10 -pady 10
    grid $ButtonArea.cancel  -row 1 -column 2 -sticky w -padx 10 -pady 10
    grid columnconfigure $ButtonArea 0 -weight 1
    grid columnconfigure $ButtonArea 2 -weight 1
  pack $ButtonArea -fill x
  wm deiconify $win
};# ::CompileComponent::Compile

## ::CompileComponent::ExecuteCompile
## This function does the actual compilation...
proc ::CompileComponent::ExecuteCompile {component parent} {
  upvar #0 $parent spec

  set MainArea $parent.main_area
  set spec(PreCompileScript) [string trim \
    [$MainArea.parameters.pree.text get 1.0 end]]
  set spec(PostCompileScript) [string trim \
    [$MainArea.parameters.poste.text get 1.0 end]]
  
  ## Create a feedback dialog...
  set win $parent.feedback
  catch {destroy $win}
  toplevel $win
  wm withdraw $win
  wm group $win $parent
  wm transient $win $parent
  wm title $win "Compiling Component $component..."

  ## Create a vertical feedback with images...
  frame $win.vert -bg #8fbc8f -width 50
    label $win.vert.i1 -image ISourceCpp     -bg #8fbc8f \
      -width 48 -height 48 -bd 2 -relief flat
    label $win.vert.i2 -image IAddArrowGreen -bg #8fbc8f \
      -width 48 -height 48 -bd 2 -relief flat
    label $win.vert.i3 -image IBinaryFile    -bg #8fbc8f \
      -width 48 -height 48 -bd 2 -relief flat
    pack $win.vert.i3 $win.vert.i2 $win.vert.i1 -padx 10 -pady 15 -side bottom
  pack $win.vert -fill y -side left

  ## Our main area...
  set MainArea $win.main_area
  frame $MainArea
    label $MainArea.title -text $component... -font {arial 14 bold italic} \
      -anchor w -bg #8fbc8f -fg #6a5acd
    pack $MainArea.title -fill x
    ::ScrolledWidgets::ScrolledText $MainArea.feedback -height 12 -width 80
    pack $MainArea.feedback -fill both -expand 1
    Progress $MainArea.progress
    #$MainArea.progress.l.fill.l conf -image IBinaryFile -compound left
    #SetProgress $MainArea.progress 80
    pack $MainArea.progress -fill x -padx 45 -pady 15

    ## Finally a stop/close button
    button $MainArea.stop -textvariable GUI_Text(Close) -command "destroy $win"
    pack $MainArea.stop -pady 10
    
  pack $MainArea -fill both -expand 1

  gut_ModalDlg $win
  wm deiconify $win
  grab $win

  set Text $MainArea.feedback.text
  $Text tag configure exec -foreground #8fbc8f
  $Text tag configure msg  -foreground #6a5acd -font {arial 14 bold italic}
  $Text tag configure error -foreground red
  ## Register a callback in the tBuild package...
  tBuild::Reset
  ::tBuild::Configure -execallback "::CompileComponent::ExecCallback $win $Text"
  set CWD [pwd]
  cd [set ::creole_${component}_home]
  set CDM [file join $::HomeDir CDM]
  ## 1) Execute any possible pre-script...
  $Text insert end "Executing Pre-Compilation Script:\n" msg
  $Text insert end "[subst $spec(PreCompileScript)]\n" exec
  catch {eval $spec(PreCompileScript)} result
  $Text insert end "$result\n\n" exec

  ## 2) Compile our files...
  set Defines $spec(Defines)
  lappend Defines CDM_USE_STUBS
  set defines [join $Defines " -define "]
  foreach file $spec(CppFiles) {
    lappend Objs [eval tBuild::CompileSharedObject -debug $spec(Debug) \
       -include [list $CDM] -define $defines [list $file]]
  }
  SetProgress $MainArea.progress 50
  ## 3) Link our shared object...
  $win.vert.i1 configure -image IBinaryFile
  $win.vert.i3 configure -image IDynamicComponent
  eval tBuild::LinkSharedLibrary -debug $spec(Debug) \
                 -libpath "$CDM/$::ApplicationInfo(LoadBinaryFolderName)" \
                 -libpath "$CDM" -libpath kalo\
                 -lib libCDM.lib -output new_$spec(Library) $Objs
  
  ## 4) Execute any possible pre-script...
  $Text configure -state normal
  $Text insert end "Executing Post-Compilation Script:\n" msg
  $Text insert end "[subst $spec(PostCompileScript)]\n" exec
  catch {eval $spec(PostCompileScript)} result
  $Text insert end "$result\n\n" exec

  SetProgress $MainArea.progress 100
  $Text insert end \
    "Please rename file \"new_$spec(Library)\" to \"$spec(Library)\"" msg
  $Text configure -state disabled
  $Text see end
  cd $CWD
  grab release $win
  #$win.vert.i1 configure -image IDynamicComponent
  #$win.vert.i2 configure -image ITBlank
  #$win.vert.i3 configure -image ITBlank
  AnimateFeedback $win 11
};# ::CompileComponent::ExecuteCompile

proc ::CompileComponent::ExecCallback {win text args} {
  ## Insert the command to be executed in the text widget...
  set cmd [join $args { }]
  $text configure -state normal
  $text insert end "Executing Command:\n" msg
  $text insert end "$cmd\n" exec

  ## Execute our command...
  if {[catch {eval exec $args} error]} {
    $text insert end "$error\n\n" error
  } else {
    $text insert end "==> Completed without errors <==\n\n" msg
  }
  $text configure -state disabled
  $text see end
};# ::CompileComponent::ExecCallback

## ::CompileComponent::AnimateFeedback
## Simply for visual purposes. It animates the feedback window...
proc ::CompileComponent::AnimateFeedback {win {state 0} {T 400}} {
  if {![winfo exists $win]} {return}
  set l1 $win.vert.i1
  set l2 $win.vert.i2
  set l3 $win.vert.i3
  set i1 ISourceCpp
  set i2 IAddArrowGreen
  set i3 IBinaryFile
  set i4 IDynamicComponent
  set i5 ITBlank
  switch $state {
    0  {$l1 conf -image $i5;$l2 conf -image $i5;$l3 conf -image $i5}
    1  {$l1 conf -image $i1;$l2 conf -image $i5;$l3 conf -image $i5}
    2  {$l1 conf -image $i1;$l2 conf -image $i2;$l3 conf -image $i5}
    3  {$l1 conf -image $i1;$l2 conf -image $i2;$l3 conf -image $i3}
    4  {$l1 conf -image $i5;$l2 conf -image $i2;$l3 conf -image $i3}
    5  {$l1 conf -image $i5;$l2 conf -image $i5;$l3 conf -image $i3;set T 100}
    6  {$l1 conf -image $i5;$l2 conf -image $i3;$l3 conf -image $i5}
    7  {$l1 conf -image $i3;$l2 conf -image $i5;$l3 conf -image $i5;set T 400}
    8  {$l1 conf -image $i3;$l2 conf -image $i2;$l3 conf -image $i5}
    9  {$l1 conf -image $i3;$l2 conf -image $i2;$l3 conf -image $i4}
    10 {$l1 conf -image $i5;$l2 conf -image $i2;$l3 conf -image $i4}
    11 {$l1 conf -image $i5;$l2 conf -image $i5;$l3 conf -image $i4}
    12 {$l3 conf -relief groove ;#-bg #6a5acd}
    13 {$l3 conf -relief flat  ;#-bg #8fbc8f}
    14 {$l3 conf -relief groove ;#-bg #008000}
    15 {$l3 conf -relief flat  ;#-bg #8fbc8f}
  }
  incr state
  if {$state == 16} {set state 0}
  after $T "::CompileComponent::AnimateFeedback $win $state $T"
};# ::CompileComponent::AnimateFeedback
