##############################################################################
##  TextSearchDialogs.tcl:                                                  ##
## ------------------------                                                 ##
##  This file implements dialogs for finding text in Text & HTML widgets.   ##
##                                                                          ##
##  Copyright 2002-2004 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

namespace eval ::GUI_TextSearchFacilities {
  variable Windows
  variable Config
  variable SearchingText
  variable SearchingHtml

  set Config(Pattern) {}
  set Config(Direction)        0;  # Search Forward
  set Config(SearchType)       0;  # Don't use regexp-style searches.
  set Config(CaseSensitivity)  0;  # Don't require exact case match.
}

## ::GUI_TextSearchFacilities::TextFindDialog
#  Creates a Find dialog for searching a text widget.
proc ::GUI_TextSearchFacilities::TextFindDialog {text} {
  variable SearchingText
  variable Config
  variable Windows
  global GUI_Text

  set win $text.__finddialog__[pid]
  if {[winfo exists $win]} {raise $win; return}
  Clean
  set Windows($win) 1

  # Last place searched, line 1 column 0 (start).
  set SearchingText($win:LastFindIndex) 1.0
  toplevel $win 
  wm withdraw $win
  wm title $win Find...
  wm transient $win $text

  label $win.l_findwhat -textvariable GUI_Text(FindWhat) -anchor w
  entry $win.findwhat 
  $win.findwhat insert end $Config(Pattern) 
  bind $win.findwhat <Key-Return> \
       "::GUI_TextSearchFacilities::FindText $win $win.findwhat $text"

  grid config $win.l_findwhat -column 0 -row 0 -sticky w -pady 8
  grid config $win.findwhat   -column 1 -row 0 -sticky snew -pady 8

  set SearchingText($win:CaseSensitivity) $Config(CaseSensitivity)
  checkbutton $win.matchcase -textvariable GUI_Text(MatchCase) \
    -variable SearchingText($win:CaseSensitivity) -anchor w
  grid config $win.matchcase -column 0 -row 1 -columnspan 2 -sticky snew

  set SearchingText($win:SearchType) $Config(SearchType)
  checkbutton $win.regexp    -textvariable GUI_Text(RegexpSearch) \
    -variable SearchingText($win:SearchType) -anchor w
  grid config $win.regexp    -column 0 -row 2 -columnspan 2 -sticky snew

  set SearchingText($win:Direction) $Config(Direction)
  checkbutton $win.backwards -textvariable GUI_Text(SearchBackward) \
    -variable SearchingText($win:Direction) -anchor w
  grid config $win.backwards -column 0 -row 3 -columnspan 2 -sticky snew

  grid columnconfigure $win 1 -weight 1

  set fr $win.buttons
  frame $fr -bd 0
  button $fr.findnext -textvariable GUI_Text(FindNext) \
    -command "::GUI_TextSearchFacilities::FindText $win $win.findwhat $text"
  button $fr.findall  -textvariable GUI_Text(FindAll) \
    -command "::GUI_TextSearchFacilities::TextHighlightAll $win $text \
                                                           $win.findwhat"

  button $fr.cancel -textvariable GUI_Text(Close) \
        -command "destroy $win"
  pack $fr.findnext -side top -fill x
  pack $fr.findall  -side top -fill x -pady 2
  pack $fr.cancel   -side top -fill x -pady 6

  grid config $fr -column 2 -row 0 -rowspan 4 -padx 8 -pady 8
  gut_ModalDlg $win
  wm deiconify $win
};# ::GUI_TextSearchFacilities::TextFindDialog

## ::GUI_TextSearchFacilities::FindText
#  Called by Find dialog to search text.
proc ::GUI_TextSearchFacilities::FindText {win entrywidget text} {
  variable SearchingText
  variable Config
  global GUI_Text

  # Get pattern to search for.
  set pattern [$entrywidget get]
  set length [string length $pattern]

  if {$SearchingText($win:Direction) == 0} {
    set options "-forwards"
  } else {
    set options "-backwards"
  }
  if {$SearchingText($win:CaseSensitivity) == 0} {
    append options " -nocase"
  }
  if {$SearchingText($win:SearchType) == 1} {
    append options " -regexp"
  }
  set Config(Pattern)          $pattern
  set Config(Direction)        $SearchingText($win:Direction)
  set Config(SearchType)       $SearchingText($win:SearchType)
  set Config(CaseSensitivity)  $SearchingText($win:CaseSensitivity)

  set indx ""
  catch {set indx [eval $text search $options -- [list $pattern] \
                                     [list $SearchingText($win:LastFindIndex)]]}
  if {$indx != ""} {
    ## Remove old tag placement for any currently
    ## selected text.
    catch {$text tag remove sel sel.first sel.last}
    # Select text found.
    $text tag add sel $indx "$indx + $length chars"
    # Place insertion cursor at start of selected data.
    $text mark set insert $indx
    # Make sure text is visible.
    $text see $indx
    # Set focus back to text widget to show selected text.
    focus $text 

    # Keep index for next search.
    if {$SearchingText($win:Direction) == 0} {
      set SearchingText($win:LastFindIndex) "$indx + 1 char"
    } else {
      set SearchingText($win:LastFindIndex) "$indx - 1 char"
    }
  } else {
    # Clear Text Widget from previous selections...
    catch {$text tag delete __text_search_hilight__}
    # Display not found dialog.
    #tk_messageBox -type ok -icon info -parent $text \
    #    -title "Not Found" -message \
    #    [format $GUI_Text(TextNotFound) $pattern]
    gut_tk_dialog {Not Found...} [format $GUI_Text(TextNotFound) $pattern]\
        ok error
  }
};# ::GUI_TextSearchFacilities::FindText

## ::GUI_TextSearchFacilities::TextHighlightAll
#  Highlight all occurances of the pattern in a text widget.
proc ::GUI_TextSearchFacilities::TextHighlightAll {win widget {pattern {}}} {
  variable SearchingText
  global GUI_Options

  if {![string length $pattern]} {
    return [::GUI_TextSearchFacilities::TextFindDialog $widget]
  }
  catch {set pattern [$pattern get]}

  # Get length of pattern.
  set length [string length $pattern]
  set options "-forwards"
  if {$SearchingText($win:CaseSensitivity) == 0} {append options " -nocase"}
  if {$SearchingText($win:SearchType)      == 1} {append options " -regexp"}
  
  set state [$widget cget -state]
  $widget configure -state normal
  catch {$widget tag delete __text_search_hilight__}
  if {![string length $pattern]} {
    $widget configure -state $state
    return {}
  }
  set index {}
  catch {set index [eval $widget search $options -- [list $pattern] 1.0]}
  while {$index != {}} {
    $widget tag add __text_search_hilight__ $index ${index}+${length}chars
    set nindex {}
    catch {set nindex [eval $widget search $options -- [list $pattern] \
                                        ${index}+${length}chars end]}
    set index $nindex
  }
  $widget tag configure __text_search_hilight__ -background #f0e68c \
    -foreground #008080
  $widget configure -state $state
  update
};# ::GUI_TextSearchFacilities::TextHighlightAll

## ::GUI_TextSearchFacilities::HtmlFindDialog
#  Creates a Find dialog for searching an html widget.
proc ::GUI_TextSearchFacilities::HtmlFindDialog {html} {
  variable SearchingHtml
  variable Config
  variable Windows
  global GUI_Text

  set win $html.__finddialog__[pid]
  if {[winfo exists $win]} {raise $win; return}
  Clean
  set Windows($win) 1

  # Last place searched, line 1 column 0 (start).
  set SearchingHtml($win:LastFindIndex) 1.0
  toplevel $win 
  wm withdraw $win
  wm title $win Find...
  wm transient $win $html

  label $win.l_findwhat -textvariable GUI_Text(FindWhat) -anchor w
  entry $win.findwhat 
  $win.findwhat insert end $Config(Pattern) 
  bind $win.findwhat <Key-Return> \
       "::GUI_TextSearchFacilities::FindHtml $win $win.findwhat $html"

  grid config $win.l_findwhat -column 0 -row 0 -sticky w -pady 8
  grid config $win.findwhat   -column 1 -row 0 -sticky snew -pady 8

  set SearchingHtml($win:CaseSensitivity) $Config(CaseSensitivity)
  checkbutton $win.matchcase -textvariable GUI_Text(MatchCase) \
    -variable SearchingHtml($win:CaseSensitivity) -anchor w
  #grid config $win.matchcase -column 0 -row 1 -columnspan 2 -sticky snew

  set SearchingHtml($win:SearchType) $Config(SearchType)
  checkbutton $win.regexp    -textvariable GUI_Text(RegexpSearch) \
    -variable SearchingHtml($win:SearchType) -anchor w
  #grid config $win.regexp    -column 0 -row 2 -columnspan 2 -sticky snew

  set SearchingHtml($win:Direction) $Config(Direction)
  checkbutton $win.backwards -textvariable GUI_Text(SearchBackward) \
    -variable SearchingHtml($win:Direction) -anchor w
  #grid config $win.backwards -column 0 -row 3 -columnspan 2 -sticky snew

  grid columnconfigure $win 1 -weight 1

  set fr $win.buttons
  frame $fr -bd 0
  button $fr.findnext -textvariable GUI_Text(FindNext) \
    -command "::GUI_TextSearchFacilities::FindHtml $win $win.findwhat $html"
  button $fr.findall  -textvariable GUI_Text(FindAll) \
    -command "::GUI_TextSearchFacilities::TextHighlightAll $win $html \
                                                           $win.findwhat"

  button $fr.cancel -textvariable GUI_Text(Close) \
        -command "destroy $win"
  pack $fr.findnext -side top -fill x
  #pack $fr.findall  -side top -fill x -pady 2
  pack $fr.cancel   -side top -fill x -pady 6

  grid config $fr -column 2 -row 0 -rowspan 4 -padx 8 -pady 8
  gut_ModalDlg $win
  wm deiconify $win
};# ::GUI_TextSearchFacilities::HtmlFindDialog

## ::GUI_TextSearchFacilities::FindHtml
#
proc ::GUI_TextSearchFacilities::FindHtml {win entrywidget html} {
  variable SearchingHtml
  variable Config
  global GUI_Text

  # Get pattern to search for.
  set pattern [$entrywidget get]
  set length [string length $pattern]
  set Config(Pattern) $pattern
  set cmd [list $html text find $pattern]
  set last [$html index sel.last]
  if {$last != ""} {
    lappend cmd after $last
  }
  # puts "cmd $cmd"
  set found [eval $cmd]
  if {[llength $found] == 0} {
    $html selection clear
    gut_info "No (more) matches found for pattern \"$pattern\""
    return
  }
  foreach {begin end} $found break
  # Since text find returns shorter region, we should incr $end
  $html selection set $begin [HtmlIncrIndex $end]
  if {[catch {::GUI_TextSearchFacilities::HtmlSee $html.x $begin}]} {
    ## The function failed, meaning that the element is not visble. Repeat the
    ## search...
    return [FindHtml $win $entrywidget $html]
  }
};# ::GUI_TextSearchFacilities::FindHtml

proc ::GUI_TextSearchFacilities::HtmlSee {w ix} {
  foreach {x y} [$w coords $ix] {break}
  foreach {mx my} [$w coords] break
  $w yview moveto [expr {1.0*$y/$my}]
  $w xview moveto [expr {1.0*$x/$mx}]
};# ::GUI_TextSearchFacilities::HtmlSee

proc ::GUI_TextSearchFacilities::HtmlIncrIndex {ix {off 1}} {
  foreach {token char} [split $ix .] {break}
  return $token.[incr char $off]
};# ::GUI_TextSearchFacilities::HtmlIncrIndex

## ::GUI_TextSearchFacilities::Clean
#  This function cleans any unused entries in the state arrays...
proc ::GUI_TextSearchFacilities::Clean {} {
  variable SearchingText
  variable SearchingHtml
  variable Windows
  foreach win [array names Windows] {
    array unset SearchingText $win:*
    array unset SearchingHtml $win:*
  }
};# ::GUI_TextSearchFacilities::Clean
