##############################################################################
##  Wizard.tcl:                                                             ##
## -------------                                                            ##
##  This file implements a wizard dialog framework. The caller can create   ##
##  wizard dialogs with some of the predefined dialog elements offered by   ##
##  this implementaton...
##                                                                          ##
##  Copyright 1998-2004 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################
package provide WizardDialog 1.0

namespace eval ::WizardDialog {
  variable Wizard
  set Wizard(0|CancelCommand) {}
  set Wizard(0|BackCommand)   {}
  set Wizard(0|NextCommand)   {}
  set Wizard(0|FinishCommand) {}
}

## Initialise some needed images...
if {$::ApplicationInfo(EnableGUI)} {
  GUI_LoadImage TTik
  GUI_LoadImage ITPrev
  GUI_LoadImage ITNext
  GUI_LoadImage ITCancel
}

## ::WizardDialog::Dialog
#  Creates a wizard dialog. The dialog will contain some of the predefined
#  elements. These predefined elements are:
#    * an image showed at the left side of the dialog window.
#    * Four buttons placed at the lower side of the dialog. These buttons
#      default to the "Cancel", "Back", "Next" and "Finish" buttons.
#    * An area (frame) where the caller can place whatever he wants.
#  This function will return the area widget...
proc ::WizardDialog::Dialog {args} {
  global GUI_Text GUI_Options
  variable Wizard
  foreach one {Cancel Back Next Finish} {
    if {![info exists GUI_Text($one)]} {set GUI_Text($one) $one}
  }
  
  ## Default values...
  set Id     0
  set height 400
  set width  500
  set image  {}
  set title  {Wizard...}
  set header {Welcome}
  set win    .wdm_DefaultWizardDialogWindow
  set parent .
  set show   1
  set bWidth 100
  set visible {1 1 1 1}
  set state   {normal normal normal normal}
  set CancelLabel $GUI_Text(Cancel); set CancelImage ITCancel
  set BackLabel   $GUI_Text(Back)  ; set BackImage   ITPrev
  set NextLabel   $GUI_Text(Next)  ; set NextImage   ITNext
  set FinishLabel $GUI_Text(Finish); set FinishImage TTik
  ## Has the caller specified an Id?
  set index [lsearch -exact $args -id]
  if {$index != -1} {
    set Id [lindex $args [incr index]]
  }
  ## Has the caller specified a window?
  set index [lsearch -glob $args -win*]
  if {$index != -1} {
    set win [lindex $args [incr index]]
  }
  ## Find an unused window...
  if {[winfo exists $win]} {
    if {[string match .wdm_DefaultWizardDialogWindow* $win]} {
      ## Its another wizard window...
      set i 0
      while {1} {
        if {![winfo exists .wdm_DefaultWizardDialogWindow_$i]} {
          set win .wdm_DefaultWizardDialogWindow_$i; break
        }
        incr i
      }
    } else {
      ## If the window exists, we try to create a unique child of the window...
      if {[string equal $parent .]} {set parent $win}
      set i 0
      while {1} {
        if {![info exists $win.wdlg_$i]} {set win $win.wdlg_$i;break}
        incr i
      }
    }
  }
  
  if {[string length $Wizard($Id|CancelCommand)]} {
    set CancelCommand "destroy $win;$Wizard($Id|CancelCommand)"
  } else {set CancelCommand "destroy $win"}
  if {[string length $Wizard($Id|BackCommand)]} {
    set BackCommand "destroy $win;$Wizard($Id|BackCommand)"
  } else {set BackCommand   "destroy $win"}
  if {[string length $Wizard($Id|NextCommand)]} {
    set NextCommand "destroy $win;$Wizard($Id|NextCommand)"
  } else {set NextCommand   "destroy $win"}
  if {[string length $Wizard($Id|FinishCommand)]} {
    set FinishCommand "destroy $win;$Wizard($Id|FinishCommand)"
  } else {set FinishCommand "destroy $win"}

  ## Special pre-defined widgets...
  set message {}
  set html    {}
  set text    {}
  set fill    both
  set expand  1
  
  ## Handle arguments...
  foreach {op arg} $args {
    switch -glob -- $op {
      -buttonw* {set bWidth      $arg}
      -buttonv* {set visible     $arg}
      -buttons* {set state       $arg}
      -hei*     {set height      $arg}
      -ima*     {set image       $arg}
      -id       {set Id          $arg}
      -par*     {set parent      $arg}
      -sho*     {set show        $arg}
      -tit*     {set title       $arg}
      -wid*     {set width       $arg}
      -win*     {}
      -header   {set header      $arg}
      -canceli* {set CancelImage $arg}
      -backi*   {set BackImage   $arg}
      -nexti*   {set NextImage   $arg}
      -finishi* {set FinishImage $arg}
      -cancell* {set CancelLabel $arg}
      -backl*   {set BackLabel   $arg}
      -nextl*   {set NextLabel   $arg}
      -finishl* {set FinishLabel $arg}
      -cancelc* {set CancelCommand "$CancelCommand;$arg"}
      -backc*   {set BackCommand   "$BackCommand;$arg"}
      -nextc*   {set NextCommand   "$NextCommand;$arg"}
      -finishc* {set FinishCommand "$FinishCommand;$arg"}
      -mess*    {set message     $arg}
      -html     {set html        $arg}
      -text     {set text        $arg}
      -fill     {set fill        $arg}
      -expand   {set expand      $arg}
      default   {error "Unknown option \"$op\""}
    }
  }

  ## Create the dialog...
  toplevel $win -height $height -width $width
  wm withdraw $win
  wm transient $win $parent
  wm title $win $title
  wm protocol $win WM_DELETE_WINDOW $CancelCommand
  
  ## Create the Dialog's Layout
  foreach {Header Buttons UserArea} [list $win.header $win.buttons $win.user] {}
  if {[string equal ::tcl_platform(platform) unix]} {
    ## Is the arial font present?
    if {[lsearch [font families] arial] != -1} {
      label $Header -text $header -font {arial 24 bold italic}
    } else {
      label $Header -text $header -font {helvetica 20 bold italic}
    }
  } else {
    label $Header -text $header -font {helvetica 12 bold italic}
  }
  frame $Buttons -padx 10 -pady 10; frame $UserArea -padx 5 -pady 2
  foreach {Image} [list $win.image] {}
  label $Image -image $image -relief sunken -bd 1 -anchor c

  ## Are there any predefined widget elements specified?
  if {[string length $message]} {
    ## A message widget...
    pack [message $UserArea.message -width 400 -text $message] \
      -fill $fill -expand $expand
  }
  if {[string length $html]} {
    ## An HTML widget...
    ScrolledWidgets::ScrolledHtml $UserArea.html -width 400 -height 250 \
      -bg $GUI_Options(TextBg) -fg $GUI_Options(TextFg) \
      -borderwidth $GUI_Options(TextBorderWidth) -tablerelief raised \
      -relief $GUI_Options(TextRelief) -exportselection 1 \
      -unvisitedcolor blue -selectioncolor $GUI_Options(SelectBackground) \
      -fontcommand    "::HtmlInfoViewer::PickFont $UserArea.html.html"
    pack $UserArea.html -fill $fill -expand $expand
    $UserArea.html.html parse $html
    ::HtmlInfoViewer::LoadImagesCmd $UserArea.html.html
  }
  if {[string length $text]} {
    ## A text widget...
    ::ScrolledWidgets::ScrolledText $UserArea.text -width 60 -height 12
    pack $UserArea.text -fill $fill -expand $expand
    $UserArea.text.text insert end $text
    $UserArea.text.text configure -state disabled
  }

  foreach {cancel_state back_state next_state finish_state} $state {break}
  button $Buttons.cancel -text $CancelLabel -image $CancelImage -width $bWidth \
         -compound left  -padx 5 -command $CancelCommand -state $cancel_state
  button $Buttons.back   -text $BackLabel   -image $BackImage   -width $bWidth \
         -compound left  -padx 5 -command $BackCommand -state $back_state
  button $Buttons.next   -text $NextLabel   -image $NextImage   -width $bWidth \
         -compound right -padx 5 -command $NextCommand -state $next_state
  button $Buttons.finish -text $FinishLabel -image $FinishImage -width $bWidth \
         -compound left  -padx 5 -command $FinishCommand -state $finish_state
  if {[lindex $visible 0]} {pack $Buttons.cancel -side left  -padx 4}
  if {[lindex $visible 3]} {pack $Buttons.finish -side right -padx 4}
  if {[lindex $visible 2]} {pack $Buttons.next   -side right -padx 4}
  if {[lindex $visible 1]} {pack $Buttons.back   -side right -padx 4}

  grid $Header   -column 0 -row 0 -columnspan 2 -sticky snew -padx 15 -pady 10
  grid $Image    -column 0 -row 1 -sticky nw -pady 10 -padx 15
  grid $UserArea -column 1 -row 1 -sticky snew -padx 5
  frame $win.hr -relief sunken -bd 1 -height 2
  grid $win.hr   -column 0 -row 2 -columnspan 2 -sticky snew -pady 8 -padx 4
  grid $Buttons  -column 0 -row 3 -columnspan 2 -sticky snew
  grid columnconfigure $win 1 -weight 1
  grid rowconfigure    $win 1 -weight 1

  ## Place widget in the middle of the screen
  if {$show} {
    update
    set reqw [winfo reqwidth  $win]; set reqh [winfo reqheight $win]
    ## We also impose a minimum geometry, just for being nice...
    if {$reqw < 600} {set reqw 600}; if {$reqh < 400} {set reqh 400}
    set x [expr ([winfo screenwidth  .] - $reqw)/2]
    set y [expr ([winfo screenheight .] - $reqh)/2]
    wm geometry $win ${reqw}x$reqh+$x+$y
    wm deiconify $win
  }
  return $UserArea
};# ::WizardDialog::Dialog

## ::WizardDialog::State
#  Request/change the state of a button...
proc ::WizardDialog::State {area widget {state {}}} {
  set win [winfo toplevel $area]
  switch $widget {
    Cancel  {set w $win.buttons.cancel}
    Back    {set w $win.buttons.finish}
    Next    {set w $win.buttons.next}
    Finish  {set w $win.buttons.back}
    default {error "invalid widget \"$widget\""}
  }
  if {[string length $state]} {
    $w configure -state $state
  } else {
    return [$w cget -state]
  }
  return {}
};# ::WizardDialog::State

## ::WizardDialog::DialogShow
#  Show a not shown wizard dialog...
proc ::WizardDialog::DialogShow {area {width -1} {height -1}} {
  set win [winfo toplevel $area]
  update
  set reqw [winfo reqwidth  $win]; set reqh [winfo reqheight $win]
  ## We also impose a minimum geometry, just for being nice...
  if {$reqw < 600} {set reqw 600}; if {$reqh < 400} {set reqh 400}
  ## If a height/widgt is provided, honour it...
  if {$width  > 0 && $reqw < $width}  {set reqw $width}
  if {$height > 0 && $reqh < $height} {set reqh $height}
  set x [expr ([winfo screenwidth  .] - $reqw)/2]
  set y [expr ([winfo screenheight .] - $reqh)/2]
  wm geometry $win +$x+$y
  wm deiconify $win
};# ::WizardDialog::DialogShow
