##############################################################################
##  HandleComponent.tcl:                                                    ##
## ----------------------                                                   ##
##  This file contains some useful procedures that manipulate Components.   ##
##  With this functions, the caller can create, compile and edit the        ##
##  properties and documentation of components.                             ##
##                                                                          ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

namespace eval ::Component {
};# ::Component Namespace

## ::Component::Create
#  This function will present user a dialog, where the user can edit various
#  component properties. If the user accepts the action, a new component will
#  be created and saved on the specified location, according to the user
#  defined properties...
proc ::Component::Create {} {
  package require ELEP::WindowUtilities::DialogWindow
  global GUI_Text GUI_Options UserHomeDir GUI_CreateModuleStatus WMIcons

  set dlg [ELEP::WindowUtilities::DialogWindow create .mw_cnm]
  set win [$dlg getTopLevel]
  $dlg title [::ELEP::WindowUtilities::msg "Create New Component..."]
  set area [$dlg getframe]

  ## Create a Notebook Widget...
  set notebook $area.notebook
  ttk::notebook $notebook -width 680;# -height 250
  set pages  {Identity Preconditions Postconditions Viewers Parameters Advanced}
  set images {IModule IModulePreCons IModulePostCons IModuleViewers
              IModuleProps IModuleProps}
  foreach page {Identity Preconditions Postconditions
                Viewers Parameters Advanced} \
          img  {IModule IModulePreCons IModulePostCons IModuleViewers
                IModuleProps IModuleProps} {
    set frame [ttk::frame $notebook.t$page]
    $notebook add $frame -text [::ELEP::WindowUtilities::msg $page] \
                         -image $img -compound left
    CreateDialogPage $frame $page
  }
  pack $notebook -fill both -expand 1

  ## Button Section...
  $dlg addStdButton save [list ::Component::SaveNewComponent $win] ITSave
  $dlg addStdButton cancel default ITCancel

  $notebook select 0
  ttk::notebook::enableTraversal $notebook
  bind $win <Destroy> "+array unset ::GUI_CreateModuleStatus $win:*"
  catch {winico setwindow $win $WMIcons(NewModule)}
  $dlg show
};# ::Component::Create

## ::Component::CreateDialogPage
#    Create each one of the pages displayed when creating a new component...
proc ::Component::CreateDialogPage {area page} {
  global GUI_Text GUI_Options UserHomeDir GUI_CreateModuleStatus Locale
  set win [winfo toplevel $area]
  switch -exact $page {
    Identity {
      ttk::label $area.mn -textvariable GUI_Text(ModuleName) -foreground red \
            -anchor e
      gut_Ventry $area.mne -width 30 -permit {[a-zA-Z0-9_]} -keysyms \
        {Up Down} -textvariable GUI_CreateModuleStatus($win:Name)
      #$area.mne insert end NewModule
      set GUI_CreateModuleStatus($win:Coupling) {Loose (Tcl)}
      ttk::combobox $area.type -state readonly -values \
            [list {Loose (Tcl)} {Dynamic (C++)} {Java (Java)} \
                  {Perl (Loose)} {OO-Python (Loose)} {Python (Loose)}] \
            -textvariable GUI_CreateModuleStatus($win:Coupling)
      ## Try to select the previous selected language, if any...
      if {[info exists GUI_Options(NewComponentLanguage)]} {
        $area.type set $GUI_Options(NewComponentLanguage)
      }
      ttk::label $area.ml -textvariable GUI_Text(ModuleLocation) \
            -foreground red -anchor e
      if {[info exists GUI_Options(NewComponentLocation)]} {
        set GUI_CreateModuleStatus($win:Location) \
                $GUI_Options(NewComponentLocation)
      } else {
        set GUI_CreateModuleStatus($win:Location) \
            [file join $UserHomeDir modules]
      }
      gut_Ventry $area.mle -width 30 -permit {[a-zA-Z0-9_/\\:\"-]} \
            -textvariable GUI_CreateModuleStatus($win:Location)
      GUI_MakeEntryDirectoryDropTarget $area.mle
      ttk::button $area.mlb -image ITOpen -compound left -textvariable \
        GUI_Text(SelectDirectory) -command "GUI_SelectDirectory $area.mle"
      Tooltip::tooltip $area.mlb -textvariable GUI_Text(SelectDirectory)
      grid $area.mn $area.mne $area.type \
            -sticky snew -ipadx 3 -ipady 3 -padx 6 -pady 2
      grid ${area}.ml ${area}.mle ${area}.mlb \
            -sticky snew -ipadx 3 -ipady 3 -padx 6 -pady 2


      ## Non-Obligatory Data...
      ## Title
      ttk::label $area.mt -textvariable GUI_Text(ModuleTitle) -foreground \
         #2e8b57 -anchor e
      ttk::entry $area.mte -textvariable GUI_CreateModuleStatus($win:Title)
      #$area.mte insert end {Enter Module title!}
      set Title $area.mte
      grid $area.mt  -row 2 -sticky snew -ipadx 3 -ipady 3 -padx 6 -pady 2
      grid $area.mte -row 2 -sticky snew -ipadx 3 -ipady 3 -padx 6 -pady 2 \
           -column 1 -columnspan 2
  
      ## Description
      ttk::label $area.ds   -textvariable GUI_Text(Description) -foreground \
         #2e8b57 -anchor e
      ::ScrolledWidgets::ScrolledText $area.dse -width 40 -height 3
      set Description $area.dse.text
      set GUI_CreateModuleStatus($win:Description) $Description
      global tcl_platform
      $Description insert end "Author: $tcl_platform(user),\
         [clock format [clock seconds]]"
      grid ${area}.ds  -row 3 -sticky snew -ipadx 3 -ipady 3 -padx 6 -pady 2
      grid ${area}.dse -row 3 -sticky snew -ipadx 3 -ipady 3 -padx 6 -pady 2 \
           -column 1 -columnspan 2
      grid columnconfigure $area 1 -weight 1 
      # $notebook itemconfigure $Page -raisecmd "focus $area.mne"
    }
    Preconditions {
      ttk::label $area.pre  -textvariable GUI_Text(ModulePreconditions) \
        -foreground #b22222 -anchor w
      grid $area.pre -sticky snew -ipadx 3 -ipady 3 \
        -padx 6 -pady 2 -columnspan 2
    
      ttk::label $area.ca   -textvariable GUI_Text(CollectionAttributes) \
        -foreground #2e8b57 -anchor e
      ttk::entry $area.cae -textvariable GUI_CreateModuleStatus($win:ColPreCnd)
      set PreColAttr $area.cae
      grid $area.ca $area.cae -sticky snew -ipadx 3 -ipady 3 -padx 6 -pady 2

      ttk::label $area.da   -textvariable GUI_Text(DocumentAttributes) \
        -foreground #2e8b57 -anchor e
      ttk::entry $area.dae  -textvariable GUI_CreateModuleStatus($win:DocPreCnd)
      $area.dae insert end "language_[string tolower $Locale] "
      set PreDocAttr $area.dae
      grid $area.da $area.dae -sticky snew -ipadx 3 -ipady 3 -padx 6 -pady 2

      ttk::label $area.a    -textvariable GUI_Text(Annotations) \
        -foreground #2e8b57 -anchor e
      ttk::entry $area.ae   -textvariable GUI_CreateModuleStatus($win:AnnPreCnd)
      set PreAnns $area.ae
      grid $area.a $area.ae -sticky snew -ipadx 3 -ipady 3 -padx 6 -pady 2
      grid columnconfigure $area 1 -weight 1
      # $notebook itemconfigure $Page -raisecmd "focus $area.cae"
    }
    Postconditions {
      ttk::label $area.post -textvariable GUI_Text(ModulePostconditions) \
        -foreground #b22222 -anchor w
      grid $area.post -sticky snew -ipadx 3 -ipady 3 \
        -padx 6 -pady 2 -columnspan 2

      ttk::label $area.pca   -textvariable GUI_Text(CollectionAttributes) \
        -foreground #2e8b57 -anchor e
      ttk::entry $area.pcae  \
        -textvariable GUI_CreateModuleStatus($win:ColPosCnd)
      set PostColAttr $area.pcae
      grid $area.pca $area.pcae -sticky snew -ipadx 3 -ipady 3 \
        -padx 6 -pady 2

      ttk::label $area.pda   -textvariable GUI_Text(DocumentAttributes) \
        -foreground #2e8b57 -anchor e
      ttk::entry $area.pdae  \
        -textvariable GUI_CreateModuleStatus($win:DocPosCnd)
      set PostDocAttr $area.pdae
      grid $area.pda $area.pdae -sticky snew -ipadx 3 -ipady 3 \
        -padx 6 -pady 2
 
      ttk::label $area.pa    -textvariable GUI_Text(Annotations) \
        -foreground #2e8b57 -anchor e
      ttk::entry $area.pae   \
        -textvariable GUI_CreateModuleStatus($win:AnnPosCnd)
      set PostAnns $area.pae
      grid $area.pa $area.pae -sticky snew -ipadx 3 -ipady 3 -padx 6 -pady 2
      grid columnconfigure $area 1 -weight 1
      # $notebook itemconfigure $Page -raisecmd "focus $area.pcae"
    }
    Viewers {
      ttk::label $area.v     -textvariable GUI_Text(Viewers) \
        -foreground #2e8b57 -anchor e
      ScrolledWidgets::ScrolledListbox $area.listbox -widgetname ve \
        -bd $GUI_Options(BorderWidth) -height 5
      set Viewers $area.listbox.ve
      ## Add some entries in the listbox, if possible...
      $Viewers insert end {{} AnnotationExplorer {Explore Annotations...}}
      set GUI_CreateModuleStatus($win:Viewers) $Viewers
      ## Add the facilities for editing the viewers...
      ttk::frame $area.edit_controls
        ttk::label $area.edit_controls.tokenl -textvariable \
          GUI_Text(Annotation) -anchor e
        ttk::entry $area.edit_controls.token
        ttk::label $area.edit_controls.viewerl -textvariable \
          GUI_Text(Viewer) -anchor e
        ttk::combobox $area.edit_controls.viewer -values \
           [CDM_GetRegisteredViewerLabels] -state readonly
        ttk::label $area.edit_controls.labell -textvariable \
          GUI_Text(Label) -anchor e
        ttk::entry $area.edit_controls.label
        ttk::button $area.edit_controls.add -textvariable GUI_Text(Add) \
          -command "::Component::DialogCallback ViewerAdd $Viewers \
                    $area.edit_controls.token $area.edit_controls.viewer\
                    $area.edit_controls.label"
        ttk::button $area.edit_controls.delete -textvariable GUI_Text(Delete) \
          -command "::Component::DialogCallback ViewerDel $Viewers \
                    $area.edit_controls.token $area.edit_controls.viewer\
                    $area.edit_controls.label"
        grid $area.edit_controls.tokenl $area.edit_controls.token \
             -sticky snew -pady 2 -padx 1
        grid $area.edit_controls.viewerl $area.edit_controls.viewer \
             -sticky snew -pady 2 -padx 2
        grid $area.edit_controls.labell $area.edit_controls.label \
             -sticky snew -pady 2 -padx 1
        grid $area.edit_controls.delete $area.edit_controls.add \
             -sticky snw -pady 2 -padx 1
        grid $area.edit_controls.add -row 3 -column 1 \
             -sticky sne -pady 2 -padx 1
      grid $area.v -sticky w
      grid $area.listbox -row 1 -sticky snew -ipadx 2 -ipady 3 -pady 2
      grid $area.edit_controls -row 1 -column 2 -sticky ew \
         -padx 2 -pady 2
      grid columnconfigure $area 0 -weight 1
      grid rowconfigure    $area 1 -weight 1
      # $notebook itemconfigure $Page -raisecmd "focus $Viewers"
    }
    Parameters {
      ttk::label $area.p     -textvariable GUI_Text(Parameters) \
        -foreground #2e8b57 -anchor e
      ScrolledWidgets::ScrolledListbox $area.listbox -widgetname ve \
        -bd $GUI_Options(BorderWidth) -height 5
      set Parameters $area.listbox.ve
      set GUI_CreateModuleStatus($win:Parameters) $Parameters
      ## Add the facilities for editing the viewers...
      ttk::frame $area.edit_controls
        ttk::label $area.edit_controls.tokenl -textvariable \
          GUI_Text(Default) -anchor e
        ttk::entry $area.edit_controls.token
        ttk::label $area.edit_controls.viewerl -textvariable \
          GUI_Text(Type) -anchor e
        ttk::combobox $area.edit_controls.viewer -values \
           {FILE DIRECTORY NUMBER ENTRY COMBO ENCODING COMMAND BOOLEAN} \
           -state readonly
        ttk::label $area.edit_controls.labell -textvariable \
          GUI_Text(Label) -anchor e
        ttk::entry $area.edit_controls.label
        ttk::button $area.edit_controls.add -textvariable GUI_Text(Add) \
          -command "::Component::DialogCallback ParamAdd $Parameters \
                    $area.edit_controls.token $area.edit_controls.viewer\
                    $area.edit_controls.label"
        ttk::button $area.edit_controls.delete -textvariable GUI_Text(Delete) \
          -command "::Component::DialogCallback ParamDel $Parameters \
                    $area.edit_controls.token $area.edit_controls.viewer\
                    $area.edit_controls.label"
        grid $area.edit_controls.labell $area.edit_controls.label \
             -sticky snew -pady 2 -padx 1
        grid $area.edit_controls.viewerl $area.edit_controls.viewer \
             -sticky snew -pady 2 -padx 2
        grid $area.edit_controls.tokenl $area.edit_controls.token \
             -sticky snew -pady 2 -padx 1
        grid $area.edit_controls.delete $area.edit_controls.add \
             -sticky snw -pady 2 -padx 1
        grid $area.edit_controls.add -row 3 -column 1 \
             -sticky sne -pady 2 -padx 1
      grid $area.p -sticky w
      grid $area.listbox  -row 1 -sticky snew -ipadx 2 -ipady 3 -pady 2
      grid $area.edit_controls -row 1 -column 2 -sticky ew \
         -padx 2 -pady 2
      grid columnconfigure $area 0 -weight 1
      grid rowconfigure    $area 1 -weight 1
      # $notebook itemconfigure $Page -raisecmd "focus $Parameters"
    }
    Advanced {
      global GUI_CreateModuleStatus
      ttk::label $area.mn -textvariable GUI_Text(Type) -foreground \
            #2e8b57 -anchor e
      ttk::combobox $area.type -state readonly -values \
             [list module button combobox imageset] \
            -textvariable GUI_CreateModuleStatus($win:ModuleType)
      $area.type set module
      ## Select Encoding Choice
      set GUI_CreateModuleStatus($win:ModuleEncoding) [encoding system]
      ttk::label $area.ml -textvariable GUI_Text(Encoding) -foreground \
            #2e8b57 -anchor e
      set menu $area.mlb.mn
      menubutton $area.mlb -relief raised -indicatoron 1 -anchor c \
        -menu $menu -text $GUI_CreateModuleStatus($win:ModuleEncoding) \
        -textvariable GUI_CreateModuleStatus($win:ModuleEncoding) \
        -width 150 -image [::Encoding::GetEncodingImage \
        $GUI_CreateModuleStatus($win:ModuleEncoding)] -compound left -padx 10 \
        -bd $GUI_Options(BorderWidth) -fg #ffff00 -bg #009900
      menu $menu -tearoff 0
      GUI_AppendEncodingsToMenu $menu \
      "set GUI_CreateModuleStatus($win:ModuleEncoding) \$Encoding
       $area.mlb configure -image \$EncodingImage -text \$Encoding" 0
      grid $area.mn $area.type \
            -sticky snew -ipadx 3 -ipady 3 -padx 6 -pady 2
      grid $area.ml $area.mlb \
            -sticky snew -ipadx 3 -ipady 3 -padx 6 -pady 2
      grid columnconfigure $area 1 -weight 1
    }
    default return
  }
};# ::Component::CreateDialogPage

proc ::Component::DialogCallback {mode listbox value combo label} {
  switch $mode {
    ParamAdd {
      catch {
        set Default  [$value get]
        set Type     [$combo get]
        set Label    [$label get]
        set spec [list $Label %$Type% $Default]
        if {![string length $Label]} return
        $listbox insert end $spec
      }
    }
    ViewerAdd {
      catch {
        set Annotation  [$value get]
        set ViewerLabel [$combo get]
        set Viewer      [CDM_GetRegisteredViewerCommand $ViewerLabel]
        set Label       [$label get]
        if {![string length $Viewer]} return
        if {![string length $Label]} return
        set spec [list $Annotation [string map {ggi_view_ {}} $Viewer] $Label]
        $listbox insert end $spec
      }
    }
    ParamDel -
    ViewerDel {
      set sel [$listbox curselection]
      if {[llength $sel]} {
        $listbox delete $sel
      }
    }
    default {}
  }
};# ::Component::DialogCallback

## ::Component::SaveNewComponent
#  This function will save a newly created component in disk...
proc ::Component::SaveNewComponent {win} {
  global HomeDir GUI_Text ApplicationName ApplicationVersion creole_config \
         GUI_CreateModuleStatus KeyboardLocale

  set Coupling $GUI_CreateModuleStatus($win:Coupling)
  set ::GUI_Options(NewComponentLanguage) $Coupling
  switch $Coupling {
    {Loose (Tcl)}    {set File LooseTclModuleTemplate.tcl  ;set Extension .tcl}
    {Loose (C++)}    {set File LooseCModuleTemplate.cpp    ;set Extension .cpp}
    {Dynamic (C++)}  {set File DynamicCModuleTemplate.cpp  ;set Extension .cpp}
    {Static (C++)}   {set File StaticCModuleTemplate.cpp   ;set Extension .cpp}
    {Java (Java)}    {set File LooseJavaModuleTemplate.java;set Extension .java}
    {Perl (Loose)}   {set File LoosePerlModuleTemplate.pl  ;set Extension .pl}
    {Python (Loose)} {set File LoosePythonModuleTemplate.py;set Extension .py}
    {OO-Python (Loose)} {
                      set File LooseOOPythonModuleTemplate.py;set Extension .py}
  }
  set Template     [file join $HomeDir share Templates $File]
  set ModuleConfig [file join $HomeDir share Templates template_config.tcl]
  set Makefile     [file join $HomeDir share Templates MakefileTemplate]
  set VCMakefile   [file join $HomeDir share Templates Makefile.vc]
  set VCMakefile2  [file join $HomeDir share Templates MSVCTemplate.dsp]
  set VCMakefile3  [file join $HomeDir share Templates MSVCTemplate.vcproj]

  switch $Coupling {
    {Java (Java)}   {
      set Makefile   [file join $HomeDir share Templates JavaMakefileTemplate]
      set VCMakefile [file join $HomeDir share Templates \
                                                        JavaMakefileTemplate.vc]
      set VCMakefile2 {}
      set VCMakefile3 {}
    }
  }

  ## Ensure all variables are defined...
  foreach key {Description Title Viewers Parameters
               ColPreCnd DocPreCnd AnnPreCnd
               ColPosCnd DocPosCnd AnnPosCnd} {
    if {![info exists  GUI_CreateModuleStatus($win:$key)]} {
      set GUI_CreateModuleStatus($win:$key) {}
    }
  }

  set mname     $GUI_CreateModuleStatus($win:Name)
  set mlocation $GUI_CreateModuleStatus($win:Location)
  set ::GUI_Options(NewComponentLocation) $mlocation
  if {![string length $mname] || ![string length $mlocation]} {
    gut_ErrorDlg $GUI_Text(NewModuleMissingName/Location)
    return
  }

  if {[info exists creole_config($mname)]} {
    gut_ErrorDlg [format $GUI_Text(ModuleExists!) $mname]
    return
  }
  

  ## Get the Template Contents...
  if {[catch {open $Template r} file]} {
    gut_ErrorDlg [format $GUI_Text(CannotReadFile) $Template]
    return
  }
  set Data [read $file]
  close $file
  set GUI_CreateModuleStatus($win:Description) \
    [$GUI_CreateModuleStatus($win:Description) get 1.0 end-1c]

  set Date [clock format [clock seconds] -format {%A %B %d %H:%M:%S (%Z) %Y}]
  regsub -all {%%Date%%} $Data $Date Data
  regsub -all {%%ModuleName%%} $Data [string trim $mname] Data
  regsub -all {%%Program%%} $Data $ApplicationName Data
  regsub -all {%%Version%%} $Data $ApplicationVersion Data
  regsub -all {%%Coupling%%} $Data [string tolower [lindex $Coupling 0]] Data
  regsub -all {%%Title%%} $Data $GUI_CreateModuleStatus($win:Title) Data
  regsub -all {%%Description%%} $Data \
         \{$GUI_CreateModuleStatus($win:Description)\} Data
  regsub -all {%%PreColAttr%%} $Data \
         [string trim $GUI_CreateModuleStatus($win:ColPreCnd)] Data
  regsub -all {%%PreDocAttr%%} $Data \
         [string trim $GUI_CreateModuleStatus($win:DocPreCnd)] Data
  regsub -all {%%PreAnns%%} $Data \
         [string trim $GUI_CreateModuleStatus($win:AnnPreCnd)] Data
  regsub -all {%%PostColAttr%%} $Data \
         [string trim $GUI_CreateModuleStatus($win:ColPosCnd)] Data
  regsub -all {%%PostDocAttr%%} $Data \
         [string trim $GUI_CreateModuleStatus($win:DocPosCnd)] Data
  regsub -all {%%PostAnns%%} $Data \
         [string trim $GUI_CreateModuleStatus($win:AnnPosCnd)] Data
  set Viewers    $GUI_CreateModuleStatus($win:Viewers)
  set Parameters $GUI_CreateModuleStatus($win:Parameters)
  regsub -all {%%Viewers%%} $Data [join [$Viewers get 0 end] "\n      "] Data
  regsub -all {%%Parameters%%} $Data [join [$Parameters get 0 end] "\n      "] \
       Data
  regsub -all {%%Encoding%%} $Data $GUI_CreateModuleStatus($win:ModuleEncoding)\
       Data
  regsub -all {%%ModuleNameLowerCase%%} $Data [string tolower $mname] Data
  regsub -all {%%ModuleNameHome%%} $Data $mlocation Data
  

  set fpath [file join $mlocation $mname]
  set fname [file join $fpath ${mname}$Extension]
  switch $Coupling {
    {Java (Java)}   {
      set fname [file join $fpath ${mname}$Extension]
    }
  }
  ## Ensure that the path exists...
  file mkdir $fpath
  if {[catch {open $fname w} file]} {
    gut_ErrorDlg [format $GUI_Text(CannotWriteFile) $fname]
    return 
  }
  puts $file [string trim $Data]
  close $file

  ## Get the Plugin Module Config Template Contents...
  if {[catch {open $ModuleConfig r} file]} {
    gut_ErrorDlg [format $GUI_Text(CannotReadFile) $ModuleConfig]
    return 
  }
  set Data [read $file]
  close $file

  set Data [string map [list \
      %%Date%% $Date %%ModuleName%% $mname %%Program%% $ApplicationName \
      %%Version%%     $ApplicationVersion \
      %%Coupling%%    [string tolower [lindex $Coupling 0]] \
      %%Title%%       $GUI_CreateModuleStatus($win:Title) \
      %%PreColAttr%%  [string trim $GUI_CreateModuleStatus($win:ColPreCnd)] \
      %%PreDocAttr%%  [string trim $GUI_CreateModuleStatus($win:DocPreCnd)] \
      %%PreAnns%%     [string trim $GUI_CreateModuleStatus($win:AnnPreCnd)] \
      %%PostColAttr%% [string trim $GUI_CreateModuleStatus($win:ColPosCnd)] \
      %%PostDocAttr%% [string trim $GUI_CreateModuleStatus($win:DocPosCnd)] \
      %%PostAnns%%    [string trim $GUI_CreateModuleStatus($win:AnnPosCnd)] \
      %%Viewers%%     [join [$Viewers get 0 end] "\n      "] \
      %%Parameters%%  [join [$Parameters get 0 end] "\n      "] \
      %%Description%% \{$GUI_CreateModuleStatus($win:Description)\} \
      %%Encoding%%    $GUI_CreateModuleStatus($win:ModuleEncoding) \
      %%ModuleType%%  $GUI_CreateModuleStatus($win:ModuleType) \
      %%ApplicationHomeDir%% $HomeDir \
      %%TclHomeDir%% [file dirname [file dirname [info library]]] \
      %%ModuleNameLowerCase%% [string tolower $mname] \
      %%SharedLibExtension%% [info sharedlibextension] \
      %%ModuleNameHome%% $mlocation] $Data]

  set fpath [file join $mlocation $mname]
  set fname [file join $fpath creole_config.tcl]
  if {[catch {open $fname w} file]} {
    gut_ErrorDlg [format $GUI_Text(CannotWriteFile) $fname]
    return 
  }
  puts $file [string trim $Data]
  close $file

  ## Get the Makefile Contents...
  if {[catch {open $Makefile r} file]} {
    gut_ErrorDlg [format $GUI_Text(CannotReadFile) $Makefile]
    return
  }
  set Data [read $file]
  close $file

  switch $Coupling {
    {Dynamic (C++)} -
    {Static (C++)}  -
    {Java (Java)}   {}
    default {destroy $win; return}
  }
  set Date [clock format [clock seconds] -format {%A %B %d %H:%M:%S (%Z) %Y}]
  set Data [string map [list \
      %%Date%% $Date %%ModuleName%% $mname %%Program%% $ApplicationName \
      %%Version%% $ApplicationVersion %%ApplicationHomeDir%% $HomeDir \
      %%TclHomeDir%% [file dirname [file dirname [info library]]] \
      %%Coupling%% [string tolower [lindex $Coupling 0]] \
      %%Title%% $GUI_CreateModuleStatus($win:Title) \
      %%Description%% $GUI_CreateModuleStatus($win:Description) \
      %%Encoding%% $GUI_CreateModuleStatus($win:ModuleEncoding) \
      %%ModuleNameLowerCase%% [string tolower $mname] \
      %%SharedLibExtension%% [info sharedlibextension] \
      %%ModuleNameHome%% $mlocation] $Data]

  set fpath [file join $mlocation $mname]
  set fname [file join $fpath Makefile]
  ## Ensure that the path exists...
  file mkdir $fpath
  if {[catch {open $fname w} file]} {
    gut_ErrorDlg [format $GUI_Text(CannotWriteFile) $fname]
    return 
  }
  puts $file [string trim $Data]
  close $file

  ## Get the Visual C++ Makefile contents...
  if {[catch {open $VCMakefile r} file]} {
    gut_ErrorDlg [format $GUI_Text(CannotReadFile) $Makefile]
    return
  }
  set Data [read $file]
  close $file

  set Date [clock format [clock seconds] -format {%A %B %d %H:%M:%S (%Z) %Y}]
  set Data [string map [list \
      %%Date%% $Date %%ModuleName%% $mname %%Program%% $ApplicationName \
      %%Version%% $ApplicationVersion \
      %%ApplicationHomeDir%% [string map {/ \\} $HomeDir] \
      %%TclHomeDir%% [string map {/ \\} \
                             [file dirname [file dirname [info library]]]] \
      %%Coupling%% [string tolower [lindex $Coupling 0]] \
      %%Title%% $GUI_CreateModuleStatus($win:Title) \
      %%Description%% $GUI_CreateModuleStatus($win:Description) \
      %%Encoding%% $KeyboardLocale \
      %%ModuleNameLowerCase%% [string tolower $mname] \
      %%SharedLibExtension%% [info sharedlibextension] \
      %%ModuleNameHome%% [string map {/ \\} $mlocation]] $Data]

  set fpath [file join $mlocation $mname]
  switch $Coupling {
    {Dynamic (C++)} -
    {Static (C++)}  {set fname [file join $fpath Makefile.vc]}
    {Java (Java)}   {set fname [file join $fpath Makefile.vc]}
    default {}
  }
  ## Ensure that the path exists...
  file mkdir $fpath
  if {[catch {open $fname w} file]} {
    gut_ErrorDlg [format $GUI_Text(CannotWriteFile) $fname]
    return 
  }
  puts -nonewline $file [string trim $Data]
  close $file

  ## Get the Visual C++ Project file contents...
  if {[string length $VCMakefile2]} {
    if {[catch {open $VCMakefile2 r} file]} {
      gut_ErrorDlg [format $GUI_Text(CannotReadFile) $VCMakefile2]
      return
    }
    set Data [read $file]
    close $file

    set Date [clock format [clock seconds] -format {%A %B %d %H:%M:%S (%Z) %Y}]
    set Data [string map [list \
        %%Date%% $Date %%ModuleName%% $mname %%Program%% $ApplicationName \
        %%Version%% $ApplicationVersion \
        %%ApplicationHomeDir%% [string map {/ \\} $HomeDir] \
        %%TclHomeDir%% [string map {/ \\} \
                               [file dirname [file dirname [info library]]]] \
        %%Coupling%% [string tolower [lindex $Coupling 0]] \
        %%Title%% $GUI_CreateModuleStatus($win:Title) \
        %%Description%% $GUI_CreateModuleStatus($win:Description) \
        %%Encoding%% $KeyboardLocale \
        %%ModuleNameLowerCase%% [string tolower $mname] \
        %%SharedLibExtension%% [info sharedlibextension] \
        %%ModuleNameHome%% [string map {/ \\} $mlocation]] $Data]

    set fpath [file join $mlocation $mname]
    switch $Coupling {
      {Dynamic (C++)} -
      {Static (C++)}  {set fname [file join $fpath lib$mname.dsp]}
      default {}
    }
    ## Ensure that the path exists...
    file mkdir $fpath
    if {[catch {open $fname w} file]} {
      gut_ErrorDlg [format $GUI_Text(CannotWriteFile) $fname]
      return 
    }
    ## Some files must be written with specific translation...
    switch -glob [string tolower [file tail $fname]] {
      makefile.vc -
      *.dsp       -
      *.sln       -
      *.vcproj    {fconfigure $file -translation crlf}
      default     {# Use the platform's default translation...}
    }
    puts -nonewline $file [string trim $Data]
    close $file
  };# if {[string length $VCMakefile2]}

  ## Get the Visual C++ Project file contents...
  if {[string length $VCMakefile3]} {
    if {[catch {open $VCMakefile3 r} file]} {
      gut_ErrorDlg [format $GUI_Text(CannotReadFile) $VCMakefile3]
      return
    }
    set Data [read $file]
    close $file

    set Date [clock format [clock seconds] -format {%A %B %d %H:%M:%S (%Z) %Y}]
    set Data [string map [list \
        %%Date%% $Date %%ModuleName%% $mname %%Program%% $ApplicationName \
        %%Version%% $ApplicationVersion \
        %%ApplicationHomeDir%% [string map {/ \\} $HomeDir] \
        %%TclHomeDir%% [string map {/ \\} \
                               [file dirname [file dirname [info library]]]] \
        %%Coupling%% [string tolower [lindex $Coupling 0]] \
        %%Title%% $GUI_CreateModuleStatus($win:Title) \
        %%Description%% $GUI_CreateModuleStatus($win:Description) \
        %%Encoding%% $KeyboardLocale \
        %%ModuleNameLowerCase%% [string tolower $mname] \
        %%SharedLibExtension%% [info sharedlibextension] \
        %%ModuleNameHome%% [string map {/ \\} $mlocation]] $Data]

    set fpath [file join $mlocation $mname]
    switch $Coupling {
      {Dynamic (C++)} -
      {Static (C++)}  {set fname [file join $fpath lib$mname.vcproj]}
      default {}
    }
    ## Ensure that the path exists...
    file mkdir $fpath
    if {[catch {open $fname w} file]} {
      gut_ErrorDlg [format $GUI_Text(CannotWriteFile) $fname]
      return 
    }
    ## Some files must be written with specific translation...
    switch -glob [string tolower [file tail $fname]] {
      makefile.vc -
      *.dsp       -
      *.sln       -
      *.vcproj    {fconfigure $file -translation crlf}
      default     {# Use the platform's default translation...}
    }
    puts -nonewline $file [string trim $Data]
    close $file
  };# if {[string length $VCMakefile3]}
  destroy $win
};# ::Component::SaveNewComponent

## ::Component::RenameComponent
#  Rename a component.
proc ::Component::RenameComponent {path old_name new_name} {
  set old_name_lower [string tolower $old_name]
  set new_name_lower [string tolower $new_name]
  set map [list creole_${old_name}_ creole_${new_name}_ \
                creole_${old_name}: creole_${new_name}: \
                creole_${old_name}  creole_${new_name} \
                Creole_${old_name_lower}_Init Creole_${new_name_lower}_Init \
                ${old_name} ${new_name} ${old_name_lower} ${new_name_lower} \
                [string toupper $old_name] [string toupper $new_name]]
  ## Locate all the files in the current path...
  foreach filename [glob -nocomplain -directory $path -type f \
    *.tcl *.txt *.cpp *.c *.java *.pl *.perl *.pm *.py *.*htm* \
    *.dsp *.vcproj *.sln \
    *akefi*.* *akefi* *.cfg *.gif *.png *.xpm *.xml *.dtd *.lex] {
    switch -glob [string tolower [file tail $filename]] {
      *.gif       -
      *.png       {file rename -force -- $filename \
                     [file join $path [string map $map [file tail $filename]]]
      }
      default {
        ## Open each returned file and perform the needed substitutions...
        set file [open $filename]
        set data [read $file]
        close $file
        file delete -force $filename
        ## Perform substitutions...
        set data [string map $map $data]
        set file [open [file join $path [string map $map \
                                                [file tail $filename]]] w]
        ## Some files must be written with specific translation...
        switch -glob [string tolower [file tail $filename]] {
          makefile.vc -
          *.dsp       -
          *.sln       -
          *.vcproj    {fconfigure $file -translation crlf}
          default     {# Use the platform's default translation...}
        }
        puts $file [string trim $data]
        close $file
      }
    };# switch -glob [string tolower [file tail $filename]]
  }
  ## Recursivelly apply the rename into all directories...
  foreach dir [glob -nocomplain -directory $path -type d *] {
    RenameComponent [file normalize $dir] $old_name $new_name
  }
};# ::Component::RenameComponent

## ::Component::RenameComponentWizard
#  Display a wizard for renaming a module.
proc ::Component::RenameComponentWizard {} {
  package require WizardDialog
  GUI_LoadImage Install
  GUI_LoadImage IForward
  GUI_LoadImage IPartNoDownload
  global creole_config GUI_Text
  variable ComponentToRename
  variable ComponentNewPath
  variable ComponentNewName
  variable RenameComponentWizard
  set RenameComponentWizard {}

  ## Generate the dialog to select the module to load...
  set area [::WizardDialog::Dialog -image Install -buttonvisible {1 0 1 0} \
    -title {Rename Component Wizard!} \
    -header $GUI_Text(SpecifyingComponenttoRename...) \
    -message $GUI_Text(SelectComponent2Rename:) \
    -cancelcmd "set ::Component::RenameComponentWizard cancel"]
  ## Create a scrolled text and place some checkbuttons inside...
  ::ScrolledWidgets::ScrolledText $area.text -width 50 -height 12 \
    -highlightthickness 0
  pack $area.text -fill both -expand 1
  update
  set text $area.text.text
  set bg [$text cget -bg]
  ## Get all components...
  set i 1; set index 0
  set bg #ffebcd; set fg black
  set width [winfo width $text]
  incr width -40
  foreach module [lsort [array names creole_config]] {
    ## Add the language Indentification...
    incr i; set start $i.0
    set title [lindex $creole_config($module) \
        [expr {[lsearch $creole_config($module) title] +1}]]
    checkbutton $text.c$i -text "$module ($title)" -image IPartNoDownload \
      -selectimage IForward -indicatoron 0 -width $width \
      -offrelief flat -selectcolor {} -bd 0 -padx 10 \
      -variable ::Component::ComponentToRename \
      -compound left -anchor w -bg $bg -fg $fg -relief flat \
      -highlightthickness 0 -onvalue $module -offvalue {}
    bind $text.c$i <MouseWheel> \
      "$text yview scroll \[expr {- (%D / 120) * 4}\] units"
    bind $text.c$i <4> "$text yview scroll -5 units"
    bind $text.c$i <5> "$text yview scroll  5 units"
    $text window create $i.0 -window $text.c$i
    $text insert end "\n" $module
    $text tag configure $module -background $bg -foreground $fg
    incr i
  }
  $text configure -bg $bg
  $text delete $i.1 end
  $text configure -state disabled
  ::WizardDialog::DialogShow $area
  tkwait window $area
  if {$RenameComponentWizard == "cancel"} {return}
  if {![string length $ComponentToRename]} {
    gut_info $GUI_Text(PleaseSelectAComponent!)
    return
  }

  ## Generate the dialog to select the module destination directory...
  set area [::WizardDialog::Dialog -image Install -buttonvisible {1 0 0 1} \
    -title {Rename Component Wizard!} \
    -header {Specifying Output Directory...} \
    -message $GUI_Text(PleaseSelectDirectoryForRenamed:) \
    -cancelcmd "set ::Component::RenameComponentWizard cancel"]
  set home [set ::creole_${ComponentToRename}_home]
  frame $area.frame
    entry  $area.frame.entry -textvariable ::Component::ComponentNewPath
    set ComponentNewPath [file dirname $home]
    button $area.frame.button -image ITOpen -text ... -compound left -command \
    "GUI_SelectDirectory $area.frame.entry"
    label $area.frame.lname -text $GUI_Text(PleaseSpecifyComponentNewName:) \
      -anchor w
    entry $area.frame.name -textvariable ::Component::ComponentNewName
    set ComponentNewName $ComponentToRename
  grid $area.frame.entry $area.frame.button -sticky snew -padx 3 -pady 2
  grid $area.frame.lname -sticky snew -padx 3 -pady 2
  grid $area.frame.name -sticky snew -padx 3 -pady 2
  grid columnconfigure $area.frame 0 -weight 1
  grid rowconfigure $area.frame 3 -weight 1
  pack $area.frame -fill both -expand 1
  ::WizardDialog::DialogShow $area
  GUI_MakeEntryDirectoryDropTarget $area.frame.entry
  tkwait window $area
  if {$RenameComponentWizard == "cancel"} {return}
  if {$ComponentToRename == $ComponentNewName} {return}
  if {![string length $ComponentNewPath]} {
    gut_info $GUI_Text(PleaseSelectADestinationForRenamed:)
    return
  }
  if {![string length $ComponentNewName]} {
    gut_info $GUI_Text(PleaseSelectANameForRenamed:)
    return
  }
  ## Create the destination folder...
  if {[file exists $ComponentNewPath/$ComponentNewName]} {
    set a [gut_question [format $GUI_Text(OverwriteFolder?) \
        $ComponentNewPath/$ComponentNewName]]
    if {[string equal $a no]} {return}
    file delete -force $ComponentNewPath/$ComponentNewName
  }
  file mkdir $ComponentNewPath
  ## Copy the contents to be renamed...
  file copy -force $home $ComponentNewPath/$ComponentNewName
  ## Finally, perform the renaming...
  RenameComponent $ComponentNewPath/$ComponentNewName \
                  $ComponentToRename $ComponentNewName
};# ::Component::RenameComponentWizard
