##############################################################################
##  Macros.tcl:                                                             ##
## -------------                                                            ##
##  This file contains some useful procedures (macros) the user can call,   ##
##  in order to perform some common functions, that can otherwise be done   ##
##  only by mouse...                                                        ##
##                                                                          ##
##  Copyright 1998-2007 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

namespace eval ::Macro {
  variable NFoldExecution
  set NFoldExecution(Debug) 0
}

proc ::Macro::GetContainerSystem {col} {
  return [GUI_SystemWindowGetInfo $col SystemWindow]
};# ::Macro::GetContainerSystem

proc ::Macro::OpenSystem {{system {}} {folder /}} {
  if {[string length $system]} {
    set spec [list $system System [list $folder]]
  } else {
    set spec [GUI_MW_GetSelection]
  }
  GUI_MW_Open/EditSelectedSystem Open $spec
};# ::Macro::OpenSystem

proc ::Macro::EditSystem {{system {}} {folder /}} {
  if {[string length $system]} {
    set spec [list $system System [list $folder]]
  } else {
    set spec [GUI_MW_GetSelection]
  }
  GUI_MW_Open/EditSelectedSystem Edit $spec
};# ::Macro::EditSystem

proc ::Macro::OpenInSystem {System Collection {Document {}}} {
  global GUI_Current_Opened_Collection GUI_Current_Opened_Document

  if {![winfo exists $System]} {
    error "invalid system \"$System\""
  }
  set col [tip_OpenCollection $Collection]
  set GUI_Current_Opened_Collection $col
  if {[string length $Document]} {
    set doc [tip_GetDocument $col $Document]
    set GUI_Current_Opened_Document $doc
  } else {
    set doc {}
    set GUI_Current_Opened_Document {}
  }
  GUI_System_Get_Opened $System
  return [list $col $doc]
};# ::Macro::OpenInSystem

proc ::Macro::RunSystemComponent {System Annotators Collection {Document {}}
                                                               {force 0}} {
  foreach module $Annotators {
    catch {GUI_HandleModule $System.frame.canvas $module {}}
  }
  creole_Ext_RunComponentTool $Document $Collection $Annotators $force
};# ::Macro::RunSystemComponent

proc ::Macro::RunSystemTool {System Tool Collection {Document {}}} {
  tip_RunTool $Collection $Document $Tool
};# ::Macro::RunSystemTool

proc ::Macro::DisplayViewer {System Viewer Document AnnotationType Attribute} {
  set Col [tip_GetParent $Document]
  set Title "Viewer: $Viewer - Document: [tip_GetId $Document], From\
          Collection: [file tail [tip_GetName $Col]]"
  ggi_view_$Viewer $Document $AnnotationType $Attribute $Title
};# ::Macro::DisplayViewer

proc ::Macro::NFoldExecution {col folds
                              {TrainingCommands   {}}
                              {EvaluationCommands {}}
                              {FoldInitialiseCommands {}}
                              {FoldFinaliseCommands {}}
                              {FoldEvaluationInitialiseCommands {}}} {
  if {$folds < 2} {error "::Macro::NFoldExecution: Folds must be more than 1!"}
  catch {ggi_wait_message {Seperating Documents into folds...} 0.0} wwin
  NFoldExecutionLog {Seperating Documents into folds...}
  set Step [expr {100.0/([tip_Length $col]+1)}]
  set Progress 0.0
  ## Seperate the Documents into $folds sets...
  set fold 0
  foreach docId [$col GetFileIds] {
    NFoldExecutionLog "  + Fold $fold: Adding \"$docId\""
    lappend FoldDocs($fold) $docId
    incr fold
    if {$fold == $folds} {set fold 0}
    set Progress [expr {$Progress+$Step}]
    catch {ggi_wait_update {} $Progress $wwin}
  }
  NFoldExecutionLog {Documents separation into folds finished!}
  ## Ensure that we have enough documents to perform the experiment for all
  ## folds...
  for {set fold 0} {$fold < $folds} {incr fold} {
    NFoldExecutionLog "Examining fold: $fold..."
    catch {ggi_wait_update_message {} "Checking Fold: $fold..." $wwin}
    if {![info exists FoldDocs($fold)]} {
      ## Not enough documents!
      catch {ggi_destroy_wait $wwin}
      error "Not enough documents in collection for $folds folds.\
             Only $fold folds possible!"
    }
  }
  NFoldExecutionLog { ==> All folds seem fine <== }
  ## Perform all folds...
  set Step [expr {100.0/($folds+1)}]
  set Progress 0.0
  catch {ggi_wait_update {} $Progress $wwin}
  for {set fold 0} {$fold < $folds} {incr fold} {
    NFoldExecutionLog {******************************************************}
    NFoldExecutionLog "**                  F O L D   $fold"
    NFoldExecutionLog {******************************************************}
    ## Evaluate the "fold initialise" commands at the begining of each fold...
    ## Available variables: $col, $doc, $fold
    set doc {}
    NFoldExecutionLog {+ Step 1: Evaluating fold initialision commands...}
    foreach cmd $FoldInitialiseCommands {
      NFoldExecutionLog "  -> $cmd"
      eval $cmd
    }
    ## Evaluate the "training" commands for each Document...
    NFoldExecutionLog {+ Step 2: Evaluating training commands for each Document...}
    for {set i 0} {$i < $folds} {incr i} {
      if {$i == $fold} {continue}
      foreach docId $FoldDocs($i) {
        catch {ggi_wait_update_message {} \
               "Processing Fold: $fold...\nDocument: $docId" $wwin}
        ## Available variables: $col, $doc, $fold
        set doc [tip_GetDocument $col $docId]
        NFoldExecutionLog "  Fold: $fold, Document:  $docId"
        foreach cmd $TrainingCommands {
          NFoldExecutionLog "  -> $cmd"
          eval $cmd
        }
      }
    }
    ## Evaluate the "fold Evaluation initialise" commands, before evaluation
    ## starts...
    ## Available variables: $col, $doc, $fold
    set doc {}
    NFoldExecutionLog {+ Step 3: Evaluating fold evaluation initialisation commands...}
    foreach cmd $FoldEvaluationInitialiseCommands {
      NFoldExecutionLog "  -> $cmd"
      eval $cmd
    }
    ## Evaluate the "evaluation" commands for each Document...
    NFoldExecutionLog {+ Step 4: Evaluating fold evaluation commands for each Document...}
    foreach docId $FoldDocs($fold) {
      catch {ggi_wait_update_message {} \
               "Processing Fold: $fold...\nDocument: $docId" $wwin}
      ## Available variables: $col, $doc, $fold
      set doc [tip_GetDocument $col $docId]
      NFoldExecutionLog "  Fold: $fold, Document:  $docId"
      foreach cmd $EvaluationCommands {
        NFoldExecutionLog "  -> $cmd"
        eval $cmd
      }
    }
    ## Evaluate the "fold finalise" commands at the begining of each fold...
    ## Available variables: $col, $doc, $fold
    NFoldExecutionLog {+ Step 5: Evaluating fold finalisation commands...}
    foreach cmd $FoldFinaliseCommands {
      NFoldExecutionLog "  -> $cmd"
      eval $cmd
    }
    set Progress [expr {$Progress+$Step}]
    catch {ggi_wait_update {} $Progress $wwin}
  }
  catch {ggi_destroy_wait $wwin}
};# ::Macro::NFoldExecution

proc ::Macro::NFoldExecutionLog {msg} {
  variable NFoldExecution
  if {$NFoldExecution(Debug)} {
    puts $msg; update
  }
};# ::Macro::NFoldExecutionLog
