##############################################################################
##  TclProfiler.tcl:                                                        ##
## -----------------------------                                            ##
##  This file implements Profiler for Tcl procedures. A profiler is a tool  ##
##  that gathers information about the execution time of procedures. Its    ##
##  usage is simple: You start the profiler, you load the file containing   ##
##  the functions that are to be profiled, you run them and you see the     ##
##  profiling information...                                                ##
##                                                                          ##
##  Copyright 1998-2004 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

package require BWidget
package require profiler
## Initialise our namespace...
namespace eval ::CDM_Debug {
  namespace eval ::CDM_Debug::TclProfiler {
    variable Profiler
  }
}

## ::CDM_Debug::TclProfiler::Init
#  This function initialises the profiler and displays a graphical interface
#  for controlling profiling...
proc ::CDM_Debug::TclProfiler::Init {} {
  global GUI_Text GUI_Options
  variable Profiler
  set win .cdm_debug_TclProfiler
  if {![info exists Profiler(Initialised)]} {
    # Initiate profiling. All procedures created after this command is called
    # will be profiled. To profile an entire application, this command must be
    # called before any other commands...
    ::profiler::init
    set ::CDM_Debug::TclProfiler::Profiler(Initialised) 1
  } else {
    ## Profiler has been already initialised...
    if {[winfo exists $win]} {raise $win; return}
    ::profiler::resume
  }

  GUI_LoadImage IModuleProperties
  toplevel $win
  wm withdraw $win
  wm title $win {Tcl Procedure Profiler...}
  bind $win <Destroy> {::profiler::suspend}
  
  ## Create the window Toolbar...
  set toolbar_frame $win.tf; set toolbar $toolbar_frame.tb
  Toolbar::ToolbarFrame $toolbar_frame \
          -relief $GUI_Options(ToolbarFrameRelief) \
          -borderwidth $GUI_Options(ToolbarFrameReliefBorderWidth) \
          -decorate $GUI_Options(ToolbarDecorate)
    Toolbar::create $toolbar -relief $GUI_Options(ToolbarRelief) \
        -borderwidth $GUI_Options(ToolbarReliefBorderWidth) -hide 0
    Toolbar::addbutton $toolbar -tooltipvar GUI_Text(Reset) \
         -image ITCancel -command "::profiler::reset"
    Toolbar::addbutton $toolbar -tooltipvar GUI_Text(Exit) \
         -image ITExit -command "destroy $win"
    pack $toolbar -side left -fill y

  pack $toolbar_frame -side top -fill x

  ## Add a header title...
  label $win.title -textvariable GUI_Text(ProfilerForTclProcedures...) \
    -image IModuleProperties -compound left -anchor w
  pack $win.title -padx 2 -pady 2 -fill x

  ## The user must select a function from a combo box to see its profile...
  labelframe $win.proc -text $GUI_Text(ProcedureProfileInfo:)
    label $win.proc.psp -textvariable GUI_Text(PleaseSelectProc:)
    ComboBox $win.proc.combo -width 65 -postcommand \
      "::CDM_Debug::TclProfiler::FillProcCombo $win.proc.combo" \
      -editable 0 -modifycmd \
      "::CDM_Debug::TclProfiler::FillProcProfInfo $win.proc.combo $win.proc"
    grid $win.proc.psp $win.proc.combo -sticky snew -padx 2 -pady 10

    ## The area where the profile will be displayed...
    label $win.proc.totalCalls            -textvariable GUI_Text(TotalCalls) \
      -anchor w
    entry $win.proc.e_totalCalls
    label $win.proc.callerDist            -textvariable GUI_Text(CallerDist) \
      -anchor w
    ::ScrolledWidgets::ScrolledListbox $win.proc.e_callerDist -height 5
    label $win.proc.compileTime           -textvariable GUI_Text(CompileTime) \
      -anchor w
    entry $win.proc.e_compileTime
    label $win.proc.totalRuntime          -textvariable GUI_Text(TotalRunTime) \
      -anchor w
    entry $win.proc.e_totalRuntime
    label $win.proc.averageRuntime        -textvariable GUI_Text(AvgRunTime) \
      -anchor w
    entry $win.proc.e_averageRuntime
    label $win.proc.descendantTime        -textvariable GUI_Text(DescTime) \
      -anchor w
    entry $win.proc.e_descendantTime
    label $win.proc.averageDescendantTime -textvariable GUI_Text(AvgDescTime) \
      -anchor w
    entry $win.proc.e_averageDescendantTime
    grid $win.proc.totalCalls $win.proc.e_totalCalls \
      -sticky snew -padx 2 -pady 1
    grid $win.proc.callerDist $win.proc.e_callerDist \
      -sticky snew -padx 2 -pady 1
    grid $win.proc.compileTime $win.proc.e_compileTime \
      -sticky snew -padx 2 -pady 1
    grid $win.proc.totalRuntime $win.proc.e_totalRuntime \
      -sticky snew -padx 2 -pady 1
    grid $win.proc.averageRuntime $win.proc.e_averageRuntime \
      -sticky snew -padx 2 -pady 1
    grid $win.proc.descendantTime $win.proc.e_descendantTime \
      -sticky snew -padx 2 -pady 1
    grid $win.proc.averageDescendantTime $win.proc.e_averageDescendantTime \
      -sticky snew -padx 2 -pady 1
    button $win.proc.table -textvariable GUI_Text(ShowReport) -command \
      "::CDM_Debug::TclProfiler::ShowReport $win.proc.report \
       $win.proc.combo $win.proc"
    button $win.proc.refresh -textvariable GUI_Text(Update) -command \
      "::CDM_Debug::TclProfiler::FillProcProfInfo $win.proc.combo $win.proc"
    grid $win.proc.table $win.proc.refresh -padx 2 -pady 1
    grid columnconfigure $win.proc 1 -weight 1
  pack $win.proc -padx 2 -pady 2 -fill x -expand 1

  wm deiconify $win
};# ::CDM_Debug::TclProfiler::Init

## ::CDM_Debug::TclProfiler::FillProcCombo
#  Fill the combo box with all the names of the profiled functions, sorted
#  according to the average run time...
proc ::CDM_Debug::TclProfiler::FillProcCombo {combo} {
  set Procs {}
  foreach one [::profiler::sortFunctions totalRuntime] {
    foreach {proc val} $one {break}
    lappend Procs [string trim $proc :]
  }
  $combo configure -values [lsort $Procs]
};# ::CDM_Debug::TclProfiler::FillProcCombo

## ::CDM_Debug::TclProfiler::FillProcProfInfo
#  This function will fill the profiling information for the selected
#  procedure...
proc ::CDM_Debug::TclProfiler::FillProcProfInfo {combo frame} {
  $frame.e_totalCalls delete 0 end
  $frame.e_callerDist.listbox delete 0 end
  $frame.e_compileTime delete 0 end
  $frame.e_totalRuntime delete 0 end
  $frame.e_averageRuntime delete 0 end
  $frame.e_descendantTime delete 0 end
  $frame.e_averageDescendantTime delete 0 end
  
  set index [$combo getvalue]
  if {$index == -1} {return}
  set proc [lindex [$combo cget -values] $index]
  set profile [::profiler::dump ::$proc]
  foreach {name val} [lindex $profile 1] {
    switch $name {
      callCount             {$frame.e_totalCalls insert end $val}
      callerDist            {
        foreach {func time} $val {
          $frame.e_callerDist.listbox insert end "$func: $time clicks"
        }
      }
      compileTime           {
        $frame.e_compileTime insert end "$val clicks"
      }
      totalRuntime          {
        $frame.e_totalRuntime insert end "$val clicks"
      }
      averageRuntime        {
        $frame.e_averageRuntime insert end "$val clicks"
      }
      stddevRuntime         {}
      covpercentRuntime     {}
      descendantTime        {
        $frame.e_descendantTime insert end "$val clicks"
      }
      averageDescendantTime {
        $frame.e_averageDescendantTime insert end "$val clicks"
      }
      descendants           {}
      default               {}
    }
  }
};# ::CDM_Debug::TclProfiler::FillProcProfInfo

## ::CDM_Debug::TclProfiler::ShowReport
#  Display a separate window containing information about all procedures...
proc ::CDM_Debug::TclProfiler::ShowReport {win combo frame} {
  if {[winfo exists $win]} {raise $win; return}
  global GUI_Text GUI_Options
  toplevel $win -height 500
  wm withdraw $win
  wm title $win {Collective Profile Report for all Tcl Procedures...}

  ## Create the window Toolbar...
  set toolbar_frame $win.tf; set toolbar $toolbar_frame.tb
  set listbox $win.table.mclistbox
  Toolbar::ToolbarFrame $toolbar_frame \
          -relief $GUI_Options(ToolbarFrameRelief) \
          -borderwidth $GUI_Options(ToolbarFrameReliefBorderWidth) \
          -decorate $GUI_Options(ToolbarDecorate)
    Toolbar::create $toolbar -relief $GUI_Options(ToolbarRelief) \
        -borderwidth $GUI_Options(ToolbarReliefBorderWidth) -hide 0
    set ::CDM_Debug::TclProfiler::Profiler(Pattern) *
    Toolbar::addwidget $toolbar entry -width 20 \
      -bg $GUI_Options(InputBackground) -textvariable \
      ::CDM_Debug::TclProfiler::Profiler(Pattern)
    Toolbar::addbutton $toolbar -tooltipvar GUI_Text(ReloadPage) \
        -image ITRedraw -command \
        "::CDM_Debug::TclProfiler::ShowReportFillTable $listbox"
    Toolbar::addbutton $toolbar -tooltipvar GUI_Text(Reset) \
         -image ITCancel -command "::profiler::reset;\
         ::CDM_Debug::TclProfiler::ShowReportFillTable $listbox"
    Toolbar::addbutton $toolbar -tooltipvar GUI_Text(Exit) \
        -image ITExit -command "destroy $win"
    pack $toolbar -side left -fill y

  pack $toolbar_frame -side top -fill x

  ::ScrolledWidgets::ScrolledMclistbox $win.table -width 100 -height 19 \
     -highlightthickness 0 -selectbackground red \
     -selectforeground white -selectborderwidth 1 -columnrelief flat \
     -labelanchor w -columnborderwidth 0 -labelborderwidth 1 \
     -labelrelief raised -labelforeground navy -labelbackground bisque \
     -fillcolumn type -labelfont TextFont -bg $GUI_Options(InputBackground) \
     -selectcommand "::CDM_Debug::TclProfiler::ShowReportCallback $listbox $combo $frame"
  $listbox column add proc       -label Procedure -width 45
  $listbox column add calls      -label $GUI_Text(TotalCalls) \
    -width [string length $GUI_Text(TotalCalls)]
  $listbox column add compTime   -label $GUI_Text(CompileTime) \
    -width [string length $GUI_Text(CompileTime)]
  $listbox column add runtime    -label $GUI_Text(TotalRunTime) \
    -width [string length $GUI_Text(TotalRunTime)]
  $listbox column add avgRuntime -label $GUI_Text(AvgRunTime) \
    -width [string length $GUI_Text(AvgRunTime)]
  $listbox column add desTime    -label $GUI_Text(DescTime) \
    -width [string length $GUI_Text(DescTime)]
  $listbox column add avgDesTime -label $GUI_Text(AvgDescTime) \
    -width [string length $GUI_Text(AvgDescTime)]
  ## Sort Bindings...
  foreach id {proc calls compTime runtime avgRuntime desTime avgDesTime} {
    $listbox label bind $id <<MB_1>> \
      "::CDM_Debug::TclProfiler::ShowReportSortCallback %W $id"
  }
  bind $listbox <<MB_1>> {+focus %W}
  pack $win.table -padx 2 -pady 2 -fill both -expand 1

  ## Add a close button...
  pack [button $win.close -textvariable GUI_Text(Close) -command \
    "destroy $win"] -pady 10 -padx 10

  ## Fill our table listbox...
  ShowReportFillTable $listbox
  update
  wm deiconify $win
};# ::CDM_Debug::TclProfiler::ShowReport

## ::CDM_Debug::TclProfiler::ShowReportFillTable
#  Fill the listbox table with information...
proc ::CDM_Debug::TclProfiler::ShowReportFillTable {listbox} {
  variable Profiler
  $listbox delete 0 end
  ## Fill our table listbox...
  if {[catch {::profiler::dump ::[string trim $Profiler(Pattern) :]} procs]} {
    return
  }
  foreach {proc vals} $procs {
    set item [list [string trim $proc :] 0 0 0 0 0 0]
    foreach {name val} $vals {
      switch $name {
        callCount             {set item [lreplace $item 1 1 $val]}
        callerDist            {}
        compileTime           {set item [lreplace $item 2 2 $val]}
        totalRuntime          {set item [lreplace $item 3 3 $val]}
        averageRuntime        {set item [lreplace $item 4 4 $val]}
        stddevRuntime         {}
        covpercentRuntime     {}
        descendantTime        {set item [lreplace $item 5 5 $val]}
        averageDescendantTime {set item [lreplace $item 6 6 $val]}
        descendants           {}
        default               {}
      }
    }
    $listbox insert end $item
  }
};# ::CDM_Debug::TclProfiler::ShowReportFillTable

proc ::CDM_Debug::TclProfiler::ShowReportCallback {listbox combo frame {index -1} args} {
  if {$index == -1} {return}
  set proc [lindex [$listbox get $index] 0]
  FillProcCombo $combo
  set values [$combo cget -values]
  set index [lsearch $values $proc]
  if {$index == -1} {return}
  $combo setvalue @$index
  FillProcProfInfo $combo $frame
};# ::CDM_Debug::TclProfiler::ShowReportCallback

## ::CDM_Debug::TclProfiler::ShowReportSortCallback
#  Sorts our listbox table according to a column...
proc ::CDM_Debug::TclProfiler::ShowReportSortCallback {w id} {
  set data [$w get 0 end]
  set index [lsearch -exact [$w column names] $id]
  switch $id {
    proc {set result [lsort -index $index $data]}
    calls      -
    compTime   -
    runtime    -
    avgRuntime -
    desTime    -
    avgDesTime {set result [lsort -integer -index $index $data]}
    default    {set result [lsort -index $index $data]}
  }
  $w delete 0 end
  # ... and add our sorted data in
  eval $w insert end $result
};# ::CDM_Debug::TclProfiler::ShowReportSortCallback
