#!/usr/bin/wish
############################################ 
#
# CollapsableFrame.tcl
# ------------------------
# 
# Copyright (C) 2000 William J Giddings
# The Flying Facet Animation Company
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
# 
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA  02111-1307, USA.
# 
############################################ 
# 
# Description:
# -----------
# Provide a switchable open or closed frame to 
# allow for the vertical arrangement of widget
# controls. This is an alternative to notebook
# style tabbed widgets.
# 
# Args:
# ----
# CollapsableFrame::add base title height
# base    the full pathname of the megawidget.
# text    title of the frame widget
# height  maximum open height of the frame
# 
# Note:
# ----
# The images im_Open and im_Close are vital to
# the functioning of these widgets and should
# not be deleted.
#
# This is because a lable, rather than a button,
# is used to effect the switch control. A labels
# removes the up/down movement of the graphic.
# 
# As a label has no associated variable, the
# state of the frame, ie open/closed, is maintained
# by the name of the image displayed.
#   
############################################


package provide CollapsableFrame 1.0 


#-----------------------------------------------------------
# Add the widget to the parent.
#-----------------------------------------------------------

proc CollapsableFrame:add {base text height} {
# create button icons
   image create photo im_Open -data {
	R0lGODlhEAAQAKIAAP///9TQyICAgEBAQAAAAAAAAAAAAAAAACwAAAAAEAAQAAADNhi63BMg
	yinFAy0HC3Xj2EJoIEOM32WeaSeeqFK+say+2azUi+5ttx/QJeQIjshkcsBsOp/MBAA7
    }

   image create photo im_Close -data {
        R0lGODlhEAAQAKIAAP///9TQyICAgEBAQAAAAAAAAAAAAAAAACwAAAAAEAAQAAADMxi63BMg
        yinFAy0HC3XjmLeA4ngpRKoSZoeuDLmo38mwtVvKu93rIo5gSCwWB8ikcolMAAA7
    }
    
    frame $base  -height 47 -width 125 
    frame $base.fra52  -borderwidth 2 -height 16 -relief ridge -width 125 
    label $base.lab20  -borderwidth 0 -image im_Open -relief raised -text $height
    
    
    bind $base.lab20 <Button-1> {
	set a [%W cget -image]
	if { $a == "im_Open" } {
	    %W configure -image im_Close
	    [winfo parent %W].fra52 configure -height [%W cget -text]
	} else {
	    %W configure -image im_Open
	    [winfo parent %W].fra52 configure -height 16
	}
    }
    label $base.lab18  -anchor w -borderwidth 1 -text $text

    #
    # pack widgets
    #
    
    pack $base  -in [winfo parent $base] -anchor center -expand 0 -fill x -side top 
    pack $base.fra52  -in $base -anchor center -expand 1 -fill x -pady 7 -side left 
    place $base.lab20  -x 5 -y -1 -width 21 -height 21 -anchor nw -bordermode ignore 
    place $base.lab18  -x 23 -y 3  -height 12 -anchor nw -bordermode ignore
   
    # return path of the frame to receive child widgets
    return $base.fra52
}


############################################################
#
# Proceedures to provide programmable control of the widget.
# 
############################################################

#-----------------------------------------------------------
# Check the current widget state then reverse it.
#-----------------------------------------------------------
proc CollapsableFrame:toggle {w} {
    set i [winfo parent $w]
    set a [$i.lab20 cget -image]
    if { $a == "im_Open" } {
	$i.lab20 configure -image im_Close
	$i.fra52 configure -height [$i.lab20 cget -text]
    } else {
	$i.lab20 configure -image im_Open
	$i.fra52 configure -height 16
    }
}

#-----------------------------------------------------------
# Collapse the widget, display the 'can be opened' icon.
#-----------------------------------------------------------
proc CollapsableFrame:close {w} {
    set i [winfo parent $w]
    $i.lab20 configure -image im_Open
    $i.fra52 configure -height 16
}

#-----------------------------------------------------------
# Open the widget, display the 'can be closed' icon.
#-----------------------------------------------------------
proc CollapsableFrame:open {w} {
    set i [winfo parent $w]
    $i.lab20 configure -image im_Close
    $i.fra52 configure -height [$i.lab20 cget -text]
}



#-----------------------------------------------------------
# Build a scrolling canvas to hold a column of frames.
#-----------------------------------------------------------

proc ScrollingCanvas:add { base width height max} {
      set borderwidth 2
      set scrollbarwidth 15
      set width [expr $width - $scrollbarwidth - (4 * $borderwidth)]
    
      frame $base.fr1 \
	-relief ridge \
	-borderwidth 2 \
	-width $width \
	-height $height
    
      # note: scrollregion requires a list
      canvas $base.fr1.can -width $width -height $height \
	-scrollregion [list 0 0 0 $max]	\
	-yscrollcommand "$base.fr1.v_scroll set" 
	# -relief ridge -borderwidth 2     
      
      frame $base.fr1.can.f1 \
        -width $width \
	-height $height \
	-borderwidth 0

      scrollbar $base.fr1.v_scroll \
	-command "$base.fr1.can yview" \
	-width $scrollbarwidth

      place $base.fr1 -x 0 -y 0 ;# -width $width -height $height
      pack $base.fr1.v_scroll -in $base.fr1 -side right -fill y
      pack $base.fr1.can -in $base.fr1 -side left -expand 1 -fill both
      place $base.fr1.can.f1 -in $base.fr1.can -x 0 -y 0 -relwidth 1.0 -relheight 1.0
    
      # create the holding frame and base widgets
      $base.fr1.can create window 0 0 \
	-window $base.fr1.can.f1 -anchor nw -width [$base.fr1.can cget -width]

      
      # get changes to the size of the canvas window and update the scrollregion      
      bind $base.fr1.can.f1 <Configure> {
	  [winfo parent %W] configure -scrollregion [list 0 0 0 [winfo height %W]]
      }

      # return full path to the holding frame
      return $base.fr1.can.f1
  }


###########################################################################
#
# DEMONSTRATION CODE
#
###########################################################################


#add megawidget
#catch {
#  console show
#}

#set top .demo

#toplevel $top -width 250 -height 400
#set scrcan [ScrollingCanvas:add $top  250 390 550]

#---------
#set active [CollapsableFrame:add $scrcan.cf1 "Frame1 " 50]
#CollapsableFrame:toggle $active

#button $active.b1 -text "Close Frame 1!" 
#bind $active.b1 <Button-1> { 
#    CollapsableFrame:close [winfo parent %W]
#    puts %W
#}
#place $active.b1 -in $active -x 20 -y 15

#---------
#set active [CollapsableFrame:add $scrcan.cf2 "Frame2 " 60]
#button $active.b1 -text "Open Frame 1!"
#bind $active.b1 <Button-1> { 
#    CollapsableFrame:open .demo.fr1.can.f1.cf1.fra52
#}
#place $active.b1 -in $active -x 20 -y 15

#---------
#set active [CollapsableFrame:add $scrcan.cf3 "Frame3 " 70]
#button $active.b1 -text "Toggle Frames 1 and 2"
#bind $active.b1 <Button-1> { 
#    CollapsableFrame:toggle .demo.fr1.can.f1.cf1.fra52
#    CollapsableFrame:toggle .demo.fr1.can.f1.cf2.fra52
#}
#place $active.b1 -in $active -x 20 -y 15
#
#---------
#set active [CollapsableFrame:add $scrcan.cf4 "Frame4 " 500]
#button $active.b1 -text "Hello World!"
#place $active.b1 -in $active -x 20 -y 15



