##############################################################################
##  CompareCollections.tcl:                                                 ##
## -------------------------                                                ##
##  This file implements a utility for comparing Collections.               ##
##                                                                          ##
##  Copyright 1998-2009 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

#############################################################################
## A Small Utility for comparing collections.                              ##
##
## This utility can also be used in batch mode, i.e. without displaying
## windows or through another module. In order to use it in batch mode, some
## global variables must be defined:
##
##  CDMTools_CompareCollections(::KeyCollectionName::) :
##     Name of Key Collection (Full path name)
##  CDMTools_CompareCollections(::ResCollectionName::) :
##     Name of Responce Collection (Full path name)
##  CDMTools_CompareCollections(::KeyDocumentId::) :
##     The Id of a Document in the Key Collection. If empty ({}) all Documents
##     will be compared.
##  CDMTools_CompareCollections(::ResDocumentId::) :
##     The Id of a Document in the Res Collection. If empty ({}) all Documents
##     will be compared.
##  CDMTools_CompareCollections(KeyAnnType[0-24]) :
##     The Key Annotation types for comparison
##  CDMTools_CompareCollections(ResAnnType[0-24]) :
##     The Responce Annotation types for comparison
##  CDMTools_CompareCollections(::CheckAttributes::) :
##     Boolean value controlling Attribute comparison
##  CDMTools_CompareCollections(Attribute[0-24]) :
##     Attribute names to be compared (if Attribute comparison is enabled)
##  CDMTools_CompareCollections(::CompareValues::) :
##     Boolean value controlling whether Attribute values will be compared or
##     not. If false, only Attribute existence will be checked (if Attribute
##     comparison is enabled)
##  CDMTools_CompareCollections(::SpanOverlapPercent::) :
##     Double value that specifies the percentage two spans can have to be
##     considered equal. If 100.0, the two spans must be identical.
##  CDMTools_CompareCollections(::CValues[0-24]::) :
##     A list containing the Attribute values, like NNN or VB.
##     If non-empty, only Attributes having as values one of the list elements
##     will be included for comparison. Other Attributes will be ignored.
##     Note that each list of values is supposed to correspond to an
##     Attribute: The entry with key ::CValues0:: corresponds to Attribute0
##  CDMTools_CompareCollections(::CreateLog::) :
##     Binary value controlling whether the comparison window will be
##     displayed or not
##
##  Then the caller can simply call "CDMTools_CompareCollections 0" to perform
##  the comparison.
##
##  Simple example:
## set CDMTools_CompareCollections(::KeyCollectionName::) \
##     /disk2a/users/petasis/Collections/George1
## set CDMTools_CompareCollections(::ResCollectionName::) \
##     /disk2a/users/petasis/Collections/George2
## set CDMTools_CompareCollections(::KeyDocumentId::) B1-f2-w1-i0-mediascene-1.htm
## set CDMTools_CompareCollections(::ResDocumentId::) B1-f2-w1-i0-mediascene-1.htm
## set CDMTools_CompareCollections(::NumberAnnotations::) 1
## set CDMTools_CompareCollections(KeyAnnType0) NE
## set CDMTools_CompareCollections(ResAnnType0) NE
## set CDMTools_CompareCollections(::CheckAttributes::) 1
## set CDMTools_CompareCollections(::NumberAttributes::) 1
## set CDMTools_CompareCollections(Attribute0) Product_No
## #set CDMTools_CompareCollections(Attribute1) type
## set CDMTools_CompareCollections(::CompareValues::) 1
## set CDMTools_CompareCollections(::CValues0::) {}
## #set CDMTools_CompareCollections(::CValues1::) PUNC
## lappend IdleCommands "CDMTools_CompareCollections 0"

proc CDMTools_CompareCollections {{displayWindows 1} {selectDocuments 0}
                                  {printResults 1}} {
  global GUI_Text GUI_Options CDMTools_CompareCollections \
           HomeDir UserHomeDir CDMTools_CompareCollections

  set KeyDocId {}; set ResDocId {}
  if {![info exists CDMTools_CompareCollections(::NumberAnnotations::)]} {
    set CDMTools_CompareCollections(::NumberAnnotations::) 1
  }
  if {![info exists CDMTools_CompareCollections(::NumberAttributes::)]} {
    set CDMTools_CompareCollections(::NumberAttributes::)  1
  }
  if {![info exists CDMTools_CompareCollections(::CreateLog::)]} {
    set CDMTools_CompareCollections(::CreateLog::) 1
  }
  if {![info exists CDMTools_CompareCollections(::SpanOverlapPercent::)]} {
    set CDMTools_CompareCollections(::SpanOverlapPercent::) 100.0
  }
  if {![info exists CDMTools_CompareCollections(::CompareValues::)]} {
    set CDMTools_CompareCollections(::CompareValues::) 0
  }

  ###########################################################################
  ##  Graphical Dialog Code...
  ###########################################################################
  if {$displayWindows} {
    ::Annotation::Init
    ## Create Collection Selection Widnow...
    set win .__CompareAnnotationsTool
    if {[winfo exists $win]} {raise $win; return}

    toplevel $win
    wm withdraw $win
    if {$selectDocuments} {
      wm title $win {Compare Documents Tool...}
    } else {
      wm title $win {Compare Collections Tool...}
    }
    frame $win.select

    set key $win.select.key
    set res $win.select.res

    ## Key Collection
    frame $key -bd 1 -relief sunken
      if {$selectDocuments} {
        label $key.title -textvariable GUI_Text(KeyDocument) -width 40
      } else {
        label $key.title -textvariable GUI_Text(KeyCollection) -width 40
      }
      label $key.keycol -text {}
      if {$selectDocuments} {
        ::tk::IconList $key.icons -command \
          "GUITools_CompareAnnotationsSelectDocument $key.icons key"
      } else {
        ::tk::IconList $key.icons
      }
      upvar #0 ::tk::$key.icons data
      set data(font) ExplorerFont
      $data(canvas) configure -relief sunken -bd $GUI_Options(BorderWidth) \
        -bg $GUI_Options(ExplorerBg)
      bind $data(canvas) <FocusOut> {}
      bind $key.icons <<ListboxSelect>> \
        "+CDMTools_CompareCollectionsTrace %W $key.keycol"
      ::tk::IconList_DeleteAll $key.icons
      global $key.icons
      #set $key.icons(font) {helvetica 12}
      #set $key.icons(selectPath) {}
      grid $key.title -sticky snew
      grid $key.keycol -sticky snew -padx 8
      grid $key.icons -sticky snew
      grid columnconfigure $key 0 -weight 1
      grid rowconfigure $key 2 -weight 1
    #pack $key -side left -fill both -expand 1

    ## Response Collection
    frame $res -bd 1 -relief sunken
      if {$selectDocuments} {
        label $res.title -textvariable GUI_Text(ResponseDocument) -width 40
      } else {
        label $res.title -textvariable GUI_Text(ResponseCollection) -width 40
      }
      label $res.rescol -text {}
      if {$selectDocuments} {
        ::tk::IconList $res.icons -command \
          "GUITools_CompareAnnotationsSelectDocument $res.icons res"
      } else {
        ::tk::IconList $res.icons
      }
      upvar #0 ::tk::$res.icons data
      set data(font) ExplorerFont
      $data(canvas) configure -relief sunken -bd $GUI_Options(BorderWidth) \
        -bg $GUI_Options(ExplorerBg)
      bind $data(canvas) <FocusOut> {}
      bind $res.icons <<ListboxSelect>> \
        "+CDMTools_CompareCollectionsTrace %W $res.rescol"
      ::tk::IconList_DeleteAll $res.icons
      global $res.icons
      #set $res.icons(font) {helvetica 12}
      #set $res.icons(selectPath) {}
      grid $res.title -sticky snew
      grid $res.rescol -sticky snew -padx 8
      grid $res.icons -sticky snew
      grid columnconfigure $res 0 -weight 1
      grid rowconfigure $res 2 -weight 1
    #pack $res -side right -fill both -expand 1

    grid $key $res -sticky snew -padx 2
    grid rowconfigure $win.select 0 -weight 2
    grid columnconfigure $win.select 0 -weight 1
    grid columnconfigure $win.select 1 -weight 1
    pack $win.select -side top -fill both -expand 1

    ##
    ## The following code is for selecting the various comparison constraints
    ##
    frame $win.options
      frame $win.options.cells
        label   $win.options.cells.ann_nul  -text {Number of Annotations:}
        label   $win.options.cells.attr_nul -text {Number of Attributes:}
        spinbox $win.options.cells.ann_nu  -from 1 -to 25 -width 3 \
          -command [list CDMTools_CompareCollections_CreateCells \
           $win.options.anns $win.options.attrs \
           $win.options.cells.ann_nu $win.options.cells.attr_nu] \
           -textvariable CDMTools_CompareCollections(::NumberAnnotations::)
        spinbox $win.options.cells.attr_nu -from 1 -to 25 -width 3 \
          -command [list CDMTools_CompareCollections_CreateCells \
           $win.options.anns $win.options.attrs \
           $win.options.cells.ann_nu $win.options.cells.attr_nu] \
           -textvariable CDMTools_CompareCollections(::NumberAttributes::)
        checkbutton $win.options.cells.lg -textvariable GUI_Text(LogComparison)\
          -variable CDMTools_CompareCollections(::CreateLog::)
        grid $win.options.cells.ann_nul  $win.options.cells.ann_nu \
             $win.options.cells.lg \
             $win.options.cells.attr_nul $win.options.cells.attr_nu \
             -sticky snew -padx 0 -pady 4
      grid $win.options.cells -sticky snew -padx 0
      grid columnconfigure $win.options.cells 2 -weight 1
      ## Annotation Types selection area
      frame $win.options.anns
      label $win.options.anns.lann1 -textvariable GUI_Text(CompareAnn) -anchor w
      label $win.options.anns.lann2 -textvariable GUI_Text(CompareAnn) -anchor e
      frame $win.options.anns.overlap
        label $win.options.anns.overlap.msg \
          -textvariable GUI_Text(SpanOverlapPercent)
        label $win.options.anns.overlap.sep -text :
        label $win.options.anns.overlap.value -width 6 \
          -textvariable ::CDMTools_CompareCollections(::SpanOverlapPercent::)
        label $win.options.anns.overlap.percent -text %
        grid $win.options.anns.overlap.msg $win.options.anns.overlap.sep \
             $win.options.anns.overlap.value $win.options.anns.overlap.percent \
             -padx 0
      grid $win.options.anns.lann1 $win.options.anns.overlap \
           $win.options.anns.lann2 -sticky snew -pady 4
      grid $win.options.anns -sticky snew -padx 0
      grid columnconfigure $win.options.anns 1 -weight 1
      scale $win.options.anns.overlapScale -orient horizontal \
        -from 0.0 -to 100.0 -resolution 0.01 -highlightthickness 0\
        -showvalue 0 -digits 5 -tickinterval 0.0 \
	-variable ::CDMTools_CompareCollections(::SpanOverlapPercent::)
      ## Attribute Names selection area
      labelframe $win.options.attrs -labelwidget [checkbutton \
        $win.options.comp_attr -textvariable GUI_Text(CompareAttributes) \
        -variable CDMTools_CompareCollections(::CheckAttributes::) \
        -command "CDMTools_CompareCollections_TongleWidgetState \
        $win.options.attrs \$CDMTools_CompareCollections(::CheckAttributes::)"]
      grid $win.options.attrs -sticky snew -padx 0
      grid columnconfigure $win.options.attrs 0 -weight 1
    grid columnconfigure $win.options 0 -weight 1
    CDMTools_CompareCollections_CreateCells \
             $win.options.anns $win.options.attrs\
             $win.options.cells.ann_nu $win.options.cells.attr_nu
    CDMTools_CompareCollections_TongleWidgetState \
       $win.options.attrs $CDMTools_CompareCollections(::CheckAttributes::)
      
    pack $win.options -fill x -expand 1 -padx 30

    set CollectionNames [GUI_GetCollectionNames]
    foreach col $CollectionNames {
      set tail [file tail $col]
      set CDMTools_CompareCollections($tail) $col
      ::tk::IconList_Add $key.icons IFolderSmall [list $tail]
      ::tk::IconList_Add $res.icons IFolderSmall [list $tail]
    }

    frame $win.buttons -bd 0
      button $win.buttons.compare -textvariable GUI_Text(Compare) -command \
        "set CDMTools_CompareCollections(::Cancel::) 0; destroy $win"
      button $win.buttons.cancel -textvariable GUI_Text(Cancel) -command \
             "destroy $win"
      set CDMTools_CompareCollections(::Cancel::) 1
      pack $win.buttons.compare $win.buttons.cancel -padx 5 -side left
    pack $win.buttons -side bottom -pady 10
    ## Place window in screen center with geometry 580x300
    gut_ModalDlg $win 750 
    wm deiconify $win
    update
    ## Remember any previously requested Collections...
    if {[info exists CDMTools_CompareCollections(::KeyCollectionName::)]} {
      set index [lsearch $CollectionNames \
                         $CDMTools_CompareCollections(::KeyCollectionName::)]
      if {$index != -1} {
        ::tk::IconList_Selection $key.icons set $index
      }
    }
    if {[info exists CDMTools_CompareCollections(::ResCollectionName::)]} {
      set index [lsearch $CollectionNames \
                         $CDMTools_CompareCollections(::ResCollectionName::)]
      if {$index != -1} {
        ::tk::IconList_Selection $res.icons set $index
      }
    }
    tkwait window $win
  };# if $displayWindows
  ###########################################################################
  ##  End of Graphical Dialog Code...
  ###########################################################################

  ## Has the user canceled?
  if {[info exists CDMTools_CompareCollections(::Cancel::)] &&
      $CDMTools_CompareCollections(::Cancel::)} {
    ## User canceled operation...
    unset -nocomplain key.icons res.icons
    return
  }
  
  set Log $CDMTools_CompareCollections(::CreateLog::)

  if {$displayWindows} {
    set KeyColName [::tk::IconList_CurSelection $key.icons]
    set ResColName [::tk::IconList_CurSelection $res.icons]
  } else {
    set KeyColName $CDMTools_CompareCollections(::KeyCollectionName::)
    set ResColName $CDMTools_CompareCollections(::ResCollectionName::)
  }
  if {![string length $KeyColName] || ![string length $ResColName]} {
    if {$selectDocuments} {
      gut_info $GUI_Text(MustSelectDocuments)
    } else {
      gut_info $GUI_Text(MustSelectCollections)
    }
    unset -nocomplain key.icons res.icons
    if {$displayWindows} {
      after idle [list CDMTools_CompareCollections 1 $selectDocuments]
    }
    return
  }
  if {$displayWindows} {
    if {$selectDocuments} {
      if {![info exists CDMTools_CompareCollections(::KeyCollectionName::)] ||
          ![info exists CDMTools_CompareCollections(::ResCollectionName::)]} {
        gut_info $GUI_Text(MustSelectDocuments)
        unset -nocomplain key.icons res.icons
        return
      }
      set KeyColName $CDMTools_CompareCollections(::KeyCollectionName::)
      set ResColName $CDMTools_CompareCollections(::ResCollectionName::)
      ## If everything is ok, both lists must have documents selected.
      set KeyDocId [::tk::IconList_CurSelection $key.icons]
      set ResDocId [::tk::IconList_CurSelection $res.icons]
      if {![string length $KeyDocId] || ![string length $ResDocId]} {
        gut_info $GUI_Text(MustSelectDocuments)
        unset -nocomplain key.icons res.icons CDMTools_CompareCollections
        return
      }
      set KeyDocId   [::tk::IconList_Get $key.icons $KeyDocId]
      set ResDocId   [::tk::IconList_Get $res.icons $ResDocId]
    } else {
      set KeyColName [::tk::IconList_Get $key.icons $KeyColName]
      set ResColName [::tk::IconList_Get $res.icons $ResColName]
      set KeyColName $CDMTools_CompareCollections($KeyColName)
      set ResColName $CDMTools_CompareCollections($ResColName)
      set CDMTools_CompareCollections(::KeyCollectionName::) $KeyColName
      set CDMTools_CompareCollections(::ResCollectionName::) $ResColName
    }
  }

  if {!$displayWindows &&
      [info exists CDMTools_CompareCollections(::KeyDocumentId::)]} {
    if {[info exists CDMTools_CompareCollections(::ResDocumentId::)]} {
      set KeyDocId $CDMTools_CompareCollections(::KeyDocumentId::)
      set ResDocId $CDMTools_CompareCollections(::ResDocumentId::)
    }
  }

  ## Select the requested Annotations...
  set KeyAnnTypes {}; set ResAnnTypes {}
  for {set index 0} {$index <
             $CDMTools_CompareCollections(::NumberAnnotations::)} {incr index} {
    set type $CDMTools_CompareCollections(KeyAnnType$index)
    if {[string length $type]} {lappend KeyAnnTypes $type}
    set type $CDMTools_CompareCollections(ResAnnType$index)
    if {[string length $type]} {lappend ResAnnTypes $type}
  }

  ## Are we going to check Attributes?
  set AttributeNames {}; set AttributeValues {}
  if {$CDMTools_CompareCollections(::CheckAttributes::)} {
    for {set cell 0} \
        {$cell < $CDMTools_CompareCollections(::NumberAttributes::)} \
        {incr cell} {
      set attribute $CDMTools_CompareCollections(Attribute$cell)
      if {[string length $attribute]} {
        lappend AttributeNames $attribute
        if {[info exists CDMTools_CompareCollections(::CValues${cell}::)]} {
          lappend AttributeValues \
            $CDMTools_CompareCollections(::CValues${cell}::)
        } else {
          lappend AttributeValues {}
        }
      }
    }
    if {![llength $AttributeNames]} {
      gut_info $GUI_Text(MustSpecifyAttribute)
      return
    }
  }

  ## Now we have the Collections... Compare them...
  global CDM_CollectionFilename 

  ## Open Collections...
  set wwin [ggi_wait_message "Opening Key Collection $KeyColName..." 0]
  set KeyCol [tip_OpenCollection $KeyColName]
  ggi_wait_message "Opening Responce Collection $ResColName..." 0 1 $wwin
  set ResCol [tip_OpenCollection $ResColName]

  ## Ensure that the location of the Collections is correct...
  if {$displayWindows} {
    if {![string equal [tip_GetName $KeyCol] $KeyColName]} {
      GUI_FixCollectionName $KeyCol [tip_GetName $KeyCol] $KeyColName
    }
    if {![string equal [tip_GetName $ResCol] $ResColName]} {
      GUI_FixCollectionName $ResCol [tip_GetName $ResCol] $ResColName
    }
  }

  set totalKeyAnns 0
  set totalResponseAnns 0
  set totalCorrectAnns 0

  ## Get Collection Documents...
  if {[string length $KeyDocId] && [string length $ResDocId]} {
    set KeyColFiles [list $KeyDocId]
    set ResColFiles [list $ResDocId]
  } else {
    set KeyColFiles [lsort [$KeyCol GetFileIds]]
    set ResColFiles [lsort [$ResCol GetFileIds]]
    ## Have the Collections the same number of Documents?
    if {[llength $KeyColFiles] != [llength $ResColFiles]} {
      tip_Close $KeyCol
      tip_Close $ResCol
      ggi_destroy_wait $wwin
      gut_error "Collectios DO NOT have the same number of Documents..."
    }
    
  };# if {[string length $KeyDocId] && [string length $ResDocId]}

  ## If Logging of Comparison is desired, create the log window...
  if {$Log} {
    foreach {l1 l2 t1 t2 sc} [CDMTools_CompareCollectionsCreateLog \
         .__CompareCollectionToolLog $KeyColName $ResColName] {break}
    $l1 delete 0 end
    $l2 delete 0 end
  } else {
    catch {destroy .__CompareCollectionToolLog}
  }
  update
  raise $wwin
  update

  ## Iterate Over Documents...
  set Progress 0.0
  set step [expr {100.0/([llength $KeyColFiles]+1)}]
  foreach KeyDocName $KeyColFiles ResDocName $ResColFiles {

    if {$selectDocuments} {
      ggi_wait_message "$GUI_Text(ComparingDocuments):\
        \n$GUI_Text(KeyDocument):\t$KeyDocName\
        \n$GUI_Text(ResponseDocument):\t$ResDocName" [expr {int ($Progress)}]\
                      1 $wwin
    } else {
      ggi_wait_message "$GUI_Text(ComparingCollections):\
        \n$GUI_Text(KeyDocument):\t$KeyDocName\
        \n$GUI_Text(ResponseDocument):\t$ResDocName" [expr {int ($Progress)}]\
                      1 $wwin
    }
    set KeyDoc  [tip_GetDocument $KeyCol $KeyDocName]
    set ResDoc  [tip_GetDocument $ResCol $ResDocName]

    set KeyAnns {}; set ResAnns {}
    foreach type $KeyAnnTypes {
      set KeyAnns [concat $KeyAnns [tip_SelectAnnotations $KeyDoc $type]]
    }
    foreach type $ResAnnTypes {
      set ResAnns [concat $ResAnns [tip_SelectAnnotations $ResDoc $type]]
    }

    set log [tip_CompareAnnotationSets $KeyAnns $ResAnns \
                 $AttributeNames $AttributeValues \
                 $CDMTools_CompareCollections(::SpanOverlapPercent::) \
                 $CDMTools_CompareCollections(::CompareValues::) $Log]
    foreach {cCorrectAnns cKeyAnns cResponseAnns log} $log {break}
    incr totalCorrectAnns  $cCorrectAnns
    incr totalKeyAnns      $cKeyAnns
    incr totalResponseAnns $cResponseAnns

    ## Debug...
    #  puts "    Total Correct Annotations: $correct_anns"
    #  puts "    Total Annotations of type \"$AnnType\" in documents:"
    #  puts "         + Key Document:      $_key_max"
    #  puts "         + Response Document: $_res_max"

    ## Log information...
    if {$Log} {
      $l1 insert end {}
      $l1 itemconfigure end -background black
      $l1 insert end "           [tip_GetId $KeyDoc]"
      $l1 itemconfigure end -background black -foreground white
      $l1 insert end {}
      $l1 itemconfigure end -background black
      $l2 insert end {}
      $l2 itemconfigure end -background black
      $l2 insert end "           [tip_GetId $ResDoc]"
      $l2 itemconfigure end -background black -foreground white
      $l2 insert end {}
      $l2 itemconfigure end -background black
      foreach element $log {
        foreach {kAnn op rAnn} $element {break}
        switch -- $op {
          == {
            ## The two Annotations are equal
            $l1 insert end [lindex [CDM_PresentAnnotationForRawViewer $kAnn] 1]
            $l2 insert end [lindex [CDM_PresentAnnotationForRawViewer $rAnn] 1]
          }
          < {
            ## The key Annotation is missing
            $l1 insert end {}
            $l2 insert end [lindex [CDM_PresentAnnotationForRawViewer $rAnn] 1]
            $l2 itemconfigure end -background yellow
          }
          > {
            ## The responce Annotation is missing
            $l1 insert end [lindex [CDM_PresentAnnotationForRawViewer $kAnn] 1]
            $l1 itemconfigure end -background yellow
            $l2 insert end {}
          }
          != {
            ## The two Annotations differ in Attributes
            $l1 insert end [lindex [CDM_PresentAnnotationForRawViewer $kAnn] 1]
            $l2 insert end [lindex [CDM_PresentAnnotationForRawViewer $rAnn] 1]
            $l1 itemconfigure end -background red -foreground white
            $l2 itemconfigure end -background red -foreground white
          }
          ~ {
            ## The two Annotations differ in Attributes
            $l1 insert end [lindex [CDM_PresentAnnotationForRawViewer $kAnn] 1]
            $l2 insert end [lindex [CDM_PresentAnnotationForRawViewer $rAnn] 1]
            $l1 itemconfigure end -background #228b22 -foreground white
            $l2 itemconfigure end -background #228b22 -foreground white
          }
        }
      }
    }
    
    tip_Close $KeyDoc
    tip_Close $ResDoc
    set Progress [expr {$Progress + $step}]
    ggi_wait_update {} [expr {int($Progress)}] $wwin
  };# foreach
  tip_Close $KeyCol
  tip_Close $ResCol
  ggi_destroy_wait $wwin

  ## Calculate Recall-Precision...
  set totalCorrectAnns [expr double($totalCorrectAnns)]
  set totalKeyAnns [expr double($totalKeyAnns)]
  set totalResponseAnns [expr double($totalResponseAnns)]
  set anntypes [lsort -unique [join $KeyAnnTypes $ResAnnTypes]]
  set Message    "$GUI_Text(CompareAnn): [join $anntypes {, }]\n"
  append Message "$GUI_Text(CompareAttr): [join $AttributeNames {, }]\n"
  append Message "$GUI_Text(CompareOnlyValues) [join $AttributeValues {, }]\n\n"
  append Message "Total Correct Annotations: [expr {int($totalCorrectAnns)}]\n"
  append Message "Total Key Annotations: [expr {int($totalKeyAnns)}]\n"
  append Message "Total Response Annotations: [expr {int($totalResponseAnns)}]\n"

  if {($totalCorrectAnns == 0) && ($totalKeyAnns == 0)} {
    set recall 0.0
  } elseif {$totalKeyAnns == 0} {
    set recall 1.0
  } else {
    set recall [expr $totalCorrectAnns/$totalKeyAnns]
  }
  if {($totalCorrectAnns == 0) && ($totalResponseAnns == 0)} {
    set precision 0.0
  } elseif {$totalResponseAnns == 0} {
    set precision 1.0
  } else {
    set precision [expr $totalCorrectAnns/$totalResponseAnns]
  }

  if {($precision == 0) && ($recall == 0)} {
    set fMeasure 0.0
  } else {
    set fMeasure [expr (2*$recall*$precision)/($precision+$recall)]
  }
  append Message "\nPrecision:\t\t$precision\
                  \nRecall:\t\t$recall\
                  \nF-Measure:\t$fMeasure"
  if {$printResults} {puts $Message}

  ## Clean Up...
  foreach var [list CDM_CollectionFilename KeyIndex ResIndex KeyMax ResMax \
                KeyAnn ResAnn KeyAnns ResAnns KeyStart ResStart KeyEnd ResEnd] {
    catch {unset $var}
  }
  if {$Log} {
    CDMTools_CompareCollectionsCreateLog_DrawScrollbarBackground $sc $l1 $l2
  }
  if {$displayWindows} {
    gut_info $Message {Results...}
  }
  return [list $precision $recall $fMeasure \
               $totalCorrectAnns $totalKeyAnns $totalResponseAnns]
};# CDMTools_CompareCollections

proc CDMTools_CompareCollections_CreateCells {frame attrfr ann_spin attr_spin} {
  global GUI_Text GUI_Options CDMTools_CompareCollections
  set ann_cells  [$ann_spin get]
  set attr_cells [$attr_spin get]
  ## Unpack the overlap scale widget...
  catch {grid forget $frame.overlapScale}
  for {set cell 0} {$cell < $ann_cells} {incr cell} {
    if {[winfo exists $frame.${cell}kann]} {continue}
    ## Annotation Types selection area
    ComboBox $frame.${cell}kann -entrybg $GUI_Options(InputBackground) \
      -foreground $GUI_Options(InputForeground) -bd $GUI_Options(BorderWidth)\
      -highlightthickness 0 -editable 1 \
      -values [::Annotation::AnnotationTypes] \
      -textvariable CDMTools_CompareCollections(KeyAnnType$cell) -modifycmd \
      "CDMTools_CompareCollectionsFillCombo $frame $attrfr"
    ComboBox $frame.${cell}rann -entrybg $GUI_Options(InputBackground) \
      -foreground $GUI_Options(InputForeground) -bd $GUI_Options(BorderWidth)\
      -highlightthickness 0 -editable 1 \
      -values [::Annotation::AnnotationTypes] \
      -textvariable CDMTools_CompareCollections(ResAnnType$cell) -modifycmd \
      "CDMTools_CompareCollectionsFillCombo $frame $attrfr"
    grid $frame.${cell}kann  x $frame.${cell}rann  -sticky snew
  }
  ## Place again the overlap scale widget...
  grid $frame.overlapScale -column 1 -row 1 -padx 30 -sticky snew
  ## Destroy any unwanted widgets...
  for {set cell 25} {$cell >= $ann_cells} {incr cell -1} {
    if {![winfo exists $frame.${cell}kann]} {continue}
    destroy $frame.${cell}kann $frame.${cell}rann
    #unset -nocomplain CDMTools_CompareCollections(KeyAnnType$cell) \
    #                  CDMTools_CompareCollections(ResAnnType$cell)
  }
  ## Attribute Names selection area
  for {set cell 0} {$cell < $attr_cells} {incr cell} {
    if {[winfo exists $attrfr.$cell]} {continue}
    frame $attrfr.$cell
      label $attrfr.$cell.lattr -textvariable GUI_Text(CompareAttr) -anchor w
      ComboBox $attrfr.$cell.attr -entrybg $GUI_Options(InputBackground) \
      -foreground $GUI_Options(InputForeground) -bd $GUI_Options(BorderWidth) \
      -highlightthickness 0 -editable 1 \
      -textvariable CDMTools_CompareCollections(Attribute$cell)
      grid $attrfr.$cell.lattr $attrfr.$cell.attr -sticky snew -padx 10
      ## Compare Values?
      labelframe $attrfr.$cell.attrval -labelwidget [checkbutton \
        $attrfr.$cell.attrvalq -textvariable GUI_Text(CompareValues) \
        -variable CDMTools_CompareCollections(::CompareValues::)]
        label $attrfr.$cell.attrval.lval -textvariable \
              GUI_Text(CompareOnlyValues) -anchor w
        ComboBox $attrfr.$cell.attrval.val \
          -entrybg $GUI_Options(InputBackground) \
          -foreground $GUI_Options(InputForeground) \
          -bd $GUI_Options(BorderWidth) \
          -highlightthickness 0 -editable 1 \
          -textvariable CDMTools_CompareCollections(::CValues${cell}::)
        grid $attrfr.$cell.attrval.lval $attrfr.$cell.attrval.val \
             -sticky snew -padx 10 -pady 5
        grid columnconfigure $attrfr.$cell.attrval 1 -weight 1
      grid $attrfr.$cell.attrval -columnspan 2 -sticky snew -padx 20 
    grid columnconfigure $attrfr.$cell 1 -weight 1
    grid $attrfr.$cell -sticky snew -pady 10 -pady 10
  }
  ## Destroy any unwanted widgets...
  for {set cell 25} {$cell >= $attr_cells} {incr cell -1} {
    if {![winfo exists $attrfr.$cell]} {continue}
    destroy $attrfr.$cell
    #unset -nocomplain CDMTools_CompareCollections(::CompareValues::)
  }
  CDMTools_CompareCollectionsFillCombo $frame $attrfr
  CDMTools_CompareCollections_TongleWidgetState \
       $attrfr $CDMTools_CompareCollections(::CheckAttributes::)
  update
  set toplevel [winfo toplevel $frame]
  set reqw [winfo width  $toplevel]
  set reqh [winfo reqheight $toplevel]
  if {$reqh > 200} {
    wm geometry $toplevel ${reqw}x$reqh
  }
};# CDMTools_CompareCollections_CreateCells

proc GUITools_CompareAnnotationsSelectDocument {widget what args} {
##  CDMTools_CompareCollections(::KeyDocumentId::) :
##     The Id of a Document in the Key Collection. If empty ({}) all Documents
##     will be compared.
##  CDMTools_CompareCollections(::ResDocumentId::)
  global CDMTools_CompareCollections
  #global $widget
  #set ${widget}(selectPath) $what
  #$col configure -text $what
  set sel [::tk::IconList_CurSelection $widget]
  if {$sel < 0} {return}
  set col [::tk::IconList_Get $widget $sel]
  ## Is it a Collection?
  if {[info exists CDMTools_CompareCollections($col)]} {
    ::tk::IconList_DeleteAll $widget
    ::tk::IconList_Add $widget IUp [list ".. ($col)"]
    set col $CDMTools_CompareCollections($col)
    foreach doc [gut_GetCollectionDocuments $col] {
      ::tk::IconList_Add $widget IDocumentSmall [list $doc]
    }
    ::tk::IconList_Arrange $widget
    ## Save the Collection name as selected...
    switch $what {
      key {set CDMTools_CompareCollections(::KeyCollectionName::) $col}
      res {set CDMTools_CompareCollections(::ResCollectionName::) $col}
    }
  } else {
    if {[string match ".. (*" $col]} {
      ## We have to display all Collections again...
      ::tk::IconList_DeleteAll $widget
      foreach col [GUI_GetCollectionNames] {
        set tail [file tail $col]
        set CDMTools_CompareCollections($tail) $col
        ::tk::IconList_Add $widget IFolderSmall [list $tail]
      }
      ::tk::IconList_Arrange $widget
      ## Reset any saved Collection names...
      switch $what {
        key {
          set CDMTools_CompareCollections(::KeyCollectionName::) {}
          set CDMTools_CompareCollections(::KeyDocumentId::) {}
        }
        res {
          set CDMTools_CompareCollections(::ResCollectionName::) {}
          set CDMTools_CompareCollections(::ResDocumentId::) {}
        }
      }
    } else {
      ## Save the Document name...
      switch $what {
        key {set CDMTools_CompareCollections(::KeyDocumentId::) $col}
        res {set CDMTools_CompareCollections(::ResDocumentId::) $col}
      }
    }
  }
};# GUITools_CompareAnnotationsSelectDocument

proc CDMTools_CompareCollectionsFillCombo {ann_frame attr_frame} {
  global CDMTools_CompareCollections

  ## Check all selected Annotation Types...
  set AnnotationTypes {}
  for {set cell 0} {$cell <= 25} {incr cell} {
    if {[info exists CDMTools_CompareCollections(KeyAnnType$cell)]} {
      lappend AnnotationTypes $CDMTools_CompareCollections(KeyAnnType$cell)
    }
    if {[info exists CDMTools_CompareCollections(ResAnnType$cell)]} {
      lappend AnnotationTypes $CDMTools_CompareCollections(ResAnnType$cell)
    }
  }
  set AnnotationTypes [lsort -unique $AnnotationTypes]

  ## Iterate over all Annotation types and select all the associated
  ## Attributes...
  foreach type $AnnotationTypes {
    foreach attr [::Annotation::AnnotationAttributes $type] {
      set Values($attr) 0
    }
  }
  set values [lsort [array names Values]]
  if {![llength $values]} {
    #set CDMTools_CompareCollections(::CheckAttributes::) 0
    #for {set cell 0} {$cell <= 25} {incr cell} {
    #  if {[winfo exists $attr_frame.$cell.attr]} {
    #    $attr_frame.$cell.attr configure -values {} -text {}
    #  }
    #}
  } else {
    for {set cell 0} {$cell <= 25} {incr cell} {
      if {[winfo exists $attr_frame.$cell.attr]} {
        $attr_frame.$cell.attr configure -values $values
        ## Perhaps the current value can be reselected?
        if {[info exists CDMTools_CompareCollections(Attribute$cell)]} {
          set index [lsearch $values \
                             [set CDMTools_CompareCollections(Attribute$cell)]]
          if {$index != -1} {
            $attr_frame.$cell.attr setvalue @$index
          } else {
            $attr_frame.$cell.attr setvalue first
          }
        } else {
          $attr_frame.$cell.attr setvalue first
        }
      }
    }
  }
};# CDMTools_CompareCollectionsFillCombo

## This function is used in order to enable/disable the given widgets...
proc CDMTools_CompareCollections_TongleWidgetState {widget activate} {
  if {$activate} {
    foreach w [winfo children $widget] {
      catch {$w configure -state normal}
      CDMTools_CompareCollections_TongleWidgetState $w $activate
    }
  } else {
    foreach w [winfo children $widget] {
      catch {$w configure -state disabled}
      CDMTools_CompareCollections_TongleWidgetState $w $activate
    }
  }
}; # CDMTools_CompareCollections_TongleWidgetState

#############################################################################
##  Comparison Log Window...
#############################################################################
proc CDMTools_CompareCollectionsCreateLog {win KeyColName ResColName} {
  global GUI_Text GUI_Options PANE \
         CDMTools_CompareCollectionsCreateLog_SelectionChanged

  set f $win.f.fr; set l1 $f.list1; set l2 $f.list2
  set text_frame $win.f.fs; set t1 $text_frame.t1; set t2 $text_frame.t2

  array set CDMTools_CompareCollectionsCreateLog_SelectionChanged \
     [list $t1.Doc {} $t1.Text {} $t2.Doc {} $t2.Text {}]

  if {[winfo exists $win]} {
    wm deiconify $win
    raise $win
    bind $l1 <<ListboxSelect>> \
     [list CDMTools_CompareCollectionsCreateLog_SelectionChanged $l1 $l2 $t1 $KeyColName]
    bind $l2 <<ListboxSelect>> \
     [list CDMTools_CompareCollectionsCreateLog_SelectionChanged $l2 $l1 $t2 $KeyColName]
    return [list $l1 $l2 $t1 $t2 $f.vsc]
  }
  toplevel $win
  wm withdraw $win
  wm title $win {Collection Comparison Log}

  ## Toolbar...
  set toolbar_frame $win.tf;
  Toolbar::ToolbarFrame $toolbar_frame \
          -relief $GUI_Options(ToolbarFrameRelief) \
          -borderwidth $GUI_Options(ToolbarFrameReliefBorderWidth) \
          -decorate $GUI_Options(ToolbarDecorate)

    GUI_LoadImage ITTop
    GUI_LoadImage ITBottom
    GUI_LoadImage ITUp
    GUI_LoadImage ITDown
    set tw $win.f.fs.text
    set toolbar $toolbar_frame.tb1
    Toolbar::create $toolbar -relief $GUI_Options(ToolbarRelief) \
        -borderwidth $GUI_Options(ToolbarReliefBorderWidth) -hide 0
    Toolbar::addbutton $toolbar -tooltipvar GUI_Text(Save) \
        -image ITSave -command {};#"ggi_SaveTextWidgetContents $tw"
    Toolbar::addseparator $toolbar
    Toolbar::addbutton $toolbar -tooltipvar GUI_Text(Top) \
        -image ITTop -command "CDMTools_CompareCollectionsSeeError $l1 $l2 F"
    set w [Toolbar::addbutton $toolbar -tooltipvar GUI_Text(Previous) \
        -image ITUp -command "CDMTools_CompareCollectionsSeeError $l1 $l2 U"]
    $w configure -repeatdelay 1000 -repeatinterval 200
    set w [Toolbar::addbutton $toolbar -tooltipvar GUI_Text(Next) \
        -image ITDown -command "CDMTools_CompareCollectionsSeeError $l1 $l2 D"]
    $w configure -repeatdelay 1000 -repeatinterval 200
    Toolbar::addbutton $toolbar -tooltipvar GUI_Text(Bottom) \
        -image ITBottom -command "CDMTools_CompareCollectionsSeeError $l1 $l2 B"
    Toolbar::addseparator $toolbar
    Toolbar::addbutton $toolbar -tooltipvar GUI_Text(Exit) \
         -image ITExit -command "destroy $win"
    pack $toolbar -side left -fill y

  pack $toolbar_frame -side top -fill x

  ## Button
  frame $win.buttons
  pack  $win.buttons -side bottom
  button $win.buttons.dismiss -textvariable GUI_Text(Dismiss) \
    -command [list destroy $win]
  pack $win.buttons.dismiss -padx 2 -pady 2

  ## Main area
  frame $win.f -relief flat -width 650 -height 450

  ## ListBoxes...
  frame $f -relief flat
  label $f.key -text "Key Collection: [file tail $KeyColName]" -width 40
  label $f.res -text "Response Collection: [file tail $ResColName]" -width 40
  listbox $l1 -yscrollcommand "$f.vsc set" -xscrollcommand \
          "CDMTools_CompareCollectionsCreateLog_Scroll $f.x1 $f.x set" \
          -width 50 -height 18 -exportselection 0 -highlightthickness 0
  bind $l1 <Enter> {focus %W}
  bind $l1 <<ListboxSelect>> \
  [list CDMTools_CompareCollectionsCreateLog_SelectionChanged $l1 $l2 \
                                                              $t1 $KeyColName]
  listbox $l2 -yscrollcommand "$f.vsc set" -xscrollcommand "$f.x2 set" \
          -width 50 -height 18 -exportselection 0 -highlightthickness 0
  bind $l2 <Enter> {focus %W}
  bind $l2 <<ListboxSelect>> \
    [list CDMTools_CompareCollectionsCreateLog_SelectionChanged $l2 $l1 \
                                                                $t2 $ResColName]
  cscrollbar $f.vsc -orient vertical -slidercolor #008000 -command \
    "CDMTools_CompareCollectionsCreateLog_Scroll $l1 $l2 yview"
  bind $f.vsc.c <Configure> \
   "CDMTools_CompareCollectionsCreateLog_DrawScrollbarBackground $f.vsc $l1 $l2"
  bind $f.vsc.c <MouseWheel> [list eval [$f.vsc cget -command] scroll \
                       \[expr\ \{-(%D/120)*4\}\] units \; break]
  bind $f.vsc.c <4> [list eval [$f.vsc cget -command] scroll -5 units \; break]
  bind $f.vsc.c <5> [list eval [$f.vsc cget -command] scroll  5 units \; break]
  bind $f.vsc.c <Destroy> "+image delete $f.vsc.map"
  image create photo $f.vsc.map
  scrollbar $f.x1 -orient horizontal -command "$l1 xview"
  scrollbar $f.x2 -orient horizontal -command "$l2 xview"
  scrollbar $f.x  -orient horizontal -command \
    "CDMTools_CompareCollectionsCreateLog_Scroll $l1 $l2 xview"
  ## In order to make the two widgets behave as one, we have to also deliver
  ## some of the events received by one widget to the other. An alternative
  ## would have been to redifine the listbox commands and specially handle the
  ## x/yview subcommants. But would be very slow  for normal insertions, and
  ## we have many many of them...
  bind $l1 <Up>   "+$l2 activate \[$l1 index active\];tk::ListboxUpDown $l2 -1"
  bind $l1 <Down> "+$l2 activate \[$l1 index active\];tk::ListboxUpDown $l2  1"
  bind $l1 <Prior>      "+$l2 yview scroll -1 pages;$l2 activate @0,0"
  bind $l1 <Next>       "+$l2 yview scroll  1 pages;$l2 activate @0,0"
  bind $l1 <MouseWheel> "+$l2 yview scroll \[expr {- (%D / 120) * 4}\] units"
  bind $l1 <4>          "+if {!\$tk_strictMotif} \{$l2 yview scroll -5 units\}"
  bind $l1 <5>          "+if {!\$tk_strictMotif} \{$l2 yview scroll  5 units\}"
  bind $l2 <Up>   "+$l1 activate \[$l2 index active\];tk::ListboxUpDown $l1 -1"
  bind $l2 <Down> "+$l1 activate \[$l2 index active\];tk::ListboxUpDown $l1  1"
  bind $l2 <Prior>      "+$l1 yview scroll -1 pages;$l1 activate @0,0"
  bind $l2 <Next>       "+$l1 yview scroll  1 pages;$l1 activate @0,0"
  bind $l2 <MouseWheel> "+$l1 yview scroll \[expr {- (%D / 120) * 4}\] units"
  bind $l2 <4>          "+if {!\$tk_strictMotif} \{$l1 yview scroll -5 units\}"
  bind $l2 <5>          "+if {!\$tk_strictMotif} \{$l1 yview scroll  5 units\}"
  grid $f.key x $f.res    -sticky snew
  grid $l1 $f.vsc $l2     -sticky snew
  grid $f.x1 x $f.x2      -sticky snew
  grid $f.x -columnspan 3 -sticky snew
  grid rowconfigure $f 1 -weight 1
  grid columnconfigure $f 0 -weight 1
  grid columnconfigure $f 2 -weight 1

  ## Text Widgets...
  frame $text_frame -relief flat -bd 0
  ::ScrolledWidgets::ScrolledText $t1 -height 11
  ::ScrolledWidgets::ScrolledText $t2 -height 11
  bind $t1 <Destroy> \
    {unset -nocomplain CDMTools_CompareCollectionsCreateLog_SelectionChanged}
  button $text_frame.view_key_anns -text {View Annotations} -command \
    "CDMTools_CompareCollectionsCreateLog_ViewAnns $t1"
  button $text_frame.view_res_anns -text {View Annotations} -command \
    "CDMTools_CompareCollectionsCreateLog_ViewAnns $t2"

  grid $t1 $t2 -sticky snew
  grid $text_frame.view_key_anns $text_frame.view_res_anns
  grid columnconfigure $text_frame 0 -weight 1
  grid columnconfigure $text_frame 1 -weight 1
  grid rowconfigure $text_frame    0 -weight 1
  #$text configure -state disabled -exportselection 1

  ## Packing...
  pane $f $text_frame -orient vertical
  ## Add Pane Arrows...
  ArrowButton $win.f.__h1.b1 -type button -width 45 -height 7 \
    -dir top -bd 0 -highlightthickness 0 -bg $PANE(Color) -command \
    "Widget_PaneSet $win.f 1 0"
  ::$win.f.__h1.b1:cmd configure -cursor based_arrow_up
  button $win.f.__h1.b2 -image IUpDown -padx 0 -pady 0 -width 45 -height 5 \
    -bd 0 -cursor dot -relief flat -highlightthickness 0 \
    -activebackground [$win.f.__h1.b1 cget -activebackground] -anchor n \
    -command "Widget_PaneSet $win.f 1 0.80" -bg $PANE(Color)
  ArrowButton $win.f.__h1.b3 -type button -width 45 -height 7 \
    -dir bottom -bd 0 -highlightthickness 0 -bg $PANE(Color) -command \
    "Widget_PaneSet $win.f 1 100"
  ::$win.f.__h1.b3:cmd configure -cursor based_arrow_down
  place $win.f.__h1.b1 -relx 0.0
  place $win.f.__h1.b2 -relx 0.47
  place $win.f.__h1.b3 -relx 1.0 -anchor ne
  $win.f.__h1 configure -height [expr {[$win.f.__h1 cget -height]+2}]

  #pack $f -fill both -expand 1
  pack $win.f -fill both -expand 1
  Widget_PaneSet $win.f 1 0.80
  wm deiconify $win
  return [list $l1 $l2 $t1.text $t2.text $f.vsc]
}; #CDMTools_CompareCollectionsCreateLog

proc CDMTools_CompareCollectionsCreateLog_Scroll {w1 w2 what args} {
  ## Do not allow our double overflow... If it matches our pattern, we have
  ## at least 1.0e-1
  if {[regexp {e-\d+} [lindex $args end]]} {
    set args [lreplace $args end end 0.0]
  }
  eval $w1 $what $args \; $w2 $what $args
}; # CDMTools_CompareCollectionsCreateLog_Scroll

proc CDMTools_CompareCollectionsCreateLog_SelectionChanged {w ow t ColName} {
  global CDMTools_CompareCollectionsCreateLog_SelectionChanged
  set selection_index [lindex [$w curselection] 0]

  set selection [$w get $selection_index]

  ## Has the user clicked on a File ID Header?
  if {[string equal -length 2 "  " $selection]} {
    return
  }
  foreach {id type start end} $selection {break}

  ## Try to find the Document Id...
  for {set i $selection_index} {$i > 0} {incr i -1} {
    if {[string equal -length 2 "  " [$w get $i]]} {break}
  }

  set DocId [string trim [$w get $i]]

  ## Get Text Document
  if {[string equal [list $ColName $DocId] \
             $CDMTools_CompareCollectionsCreateLog_SelectionChanged($t.Doc)]} {
    set Text $CDMTools_CompareCollectionsCreateLog_SelectionChanged($t.Text)
  } else {
    set Text {}
    catch {
      set col  [tip_OpenCollection $ColName]
      set doc  [tip_GetDocument $col $DocId]
      set Text [tip_GetRawData $doc]
      tip_Close $doc
      tip_Close $col
    }
    set CDMTools_CompareCollectionsCreateLog_SelectionChanged($t.Doc)\
                    [list $ColName $DocId]
    set CDMTools_CompareCollectionsCreateLog_SelectionChanged($t.Text) $Text
  }

  $t.text configure -state normal
  $t.text delete 1.0 end
  $t.text insert end $Text
  if {![info exists end] || ![string length $end]} {
    set selection [$ow get $selection_index]
    foreach {id type start end} $selection {break}
  }
  catch {
    $t.text tag add annotation 1.0+${start}c 1.0+${end}c
    $t.text tag configure annotation -background blue -foreground white
    $t.text see annotation.first
  }
  $t.text configure -state disabled
}; # CDMTools_CompareCollectionsCreateLog_SelectionChanged

## CDMTools_CompareCollectionsCreateLog_ViewAnns
#    Display the Annotations of the selected document...
proc CDMTools_CompareCollectionsCreateLog_ViewAnns {t} {
  global CDMTools_CompareCollectionsCreateLog_SelectionChanged

  if {![info exists \
     CDMTools_CompareCollectionsCreateLog_SelectionChanged($t.Doc)]} {return}
  if {![string length \
    $CDMTools_CompareCollectionsCreateLog_SelectionChanged($t.Doc)]} {return}
  foreach {ColName DocId} \
     $CDMTools_CompareCollectionsCreateLog_SelectionChanged($t.Doc) {break}
  if {![string length $ColName]} {return}

  ## Display a menu allowing the user to select which viewer to invoke...
  set menu $t.viewer_menu
  catch {destroy $menu}
  menu $menu
  foreach label [CDM_GetRegisteredViewerLabels] {
    set viewer [CDM_GetRegisteredViewerCommand $label]
    set image  [CDM_GetRegisteredViewerImage $label]
    GUI_LoadImage $image
    $menu add command -image $image -label $label -compound \
      $::GUI_Options(MenuImageLocation) -command [list \
      CDMTools_CompareCollectionsCreateLog_ViewAnns_DisplayViewer $viewer \
      $ColName $DocId]
  }
  tk_popup $menu [winfo pointerx .] [winfo pointery .]
};# CDMTools_CompareCollectionsCreateLog_ViewAnns

proc CDMTools_CompareCollectionsCreateLog_ViewAnns_DisplayViewer {Viewer ColName
                                                                  DocId} {
  set col  [tip_OpenCollection $ColName]
  set doc  [tip_GetDocument $col $DocId]
  switch $Viewer {
    ggi_view_document {
      set win [$Viewer $col $doc]
    }
    default {
      set win [$Viewer $doc {} {} \
                            "Annotations of Document $DocId, $ColName"]
    }
  }
  tkwait window $win
  tip_Close $doc
  tip_Close $col
};# CDMTools_CompareCollectionsCreateLog_ViewAnns_DisplayViewer

proc CDMTools_CompareCollectionsTrace {w l} {
  set selection [::tk::IconList_CurSelection $w]
  if {[string length $selection]} {
    global CDMTools_CompareCollections
    set name [::tk::IconList_Get $w $selection]
    $l configure -text $name
    #puts $CDMTools_CompareCollections($name)
  }
};# CDMTools_CompareCollectionsTrace

## CDMTools_CompareCollectionsSeeError
#  This function will help the user to navigate among errors displayed in the
#  Error log...
proc CDMTools_CompareCollectionsSeeError {l1 l2 mode} {
  set size [$l1 size]
  switch $mode {
    F {
      ## The user wants to see the first error...
      for {set index 0} {$index < $size} {incr index} {
        set col1 [$l1 itemcget $index -background]
        set col2 [$l2 itemcget $index -background]
        ## Is this index an error?
        if {[regexp {red|yellow} "$col1$col2"]} {
          ## We have found one!
          $l1 see $index;            $l2 see $index
          $l1 selection clear 0 end; $l2 selection clear 0 end
          $l1 selection set $index;  $l2 selection set $index
          return
        }
      }
      $l1 see end;               $l2 see end
      $l1 selection clear 0 end; $l2 selection clear 0 end
      $l1 selection set end;     $l2 selection set end
      return
    }
    U {
      ## The user wants to see the previous error...
      set start1 [lindex [$l1 curselection] 0]
      set start2 [lindex [$l2 curselection] 0]
      if {[string length $start1]} {
        set start $start1
      } elseif {[string length $start2]} {
        set start $start2
      } else {
        set start [$l1 index @0,0]
      }
      incr start -1
      for {set index $start} {$index >= 0} {incr index -1} {
        set col1 [$l1 itemcget $index -background]
        set col2 [$l2 itemcget $index -background]
        ## Is this index an error?
        if {[regexp {red|yellow} "$col1$col2"]} {
          ## We have found one!
          $l1 see $index;            $l2 see $index
          $l1 selection clear 0 end; $l2 selection clear 0 end
          $l1 selection set $index;  $l2 selection set $index
          return
        }
      }
      $l1 see 0;                 $l2 see 0
      $l1 selection clear 0 end; $l2 selection clear 0 end
      $l1 selection set 0;       $l2 selection set 0
      return
    }
    D {
      ## The user wants to see the next error...
      set start1 [lindex [$l1 curselection] 0]
      set start2 [lindex [$l2 curselection] 0]
      if {[string length $start1]} {
        set start $start1
      } elseif {[string length $start2]} {
        set start $start2
      } else {
        set start [$l1 index @0,0]
      }
      incr start
      for {set index $start} {$index < $size} {incr index} {
        set col1 [$l1 itemcget $index -background]
        set col2 [$l2 itemcget $index -background]
        ## Is this index an error?
        if {[regexp {red|yellow} "$col1$col2"]} {
          ## We have found one!
          $l1 see $index;            $l2 see $index
          $l1 selection clear 0 end; $l2 selection clear 0 end
          $l1 selection set $index;  $l2 selection set $index
          return
        }
      }
      $l1 see end;               $l2 see end
      $l1 selection clear 0 end; $l2 selection clear 0 end
      $l1 selection set end;     $l2 selection set end
      return
    }
    B {
      ## The user wants to see the last error...
      for {set index [expr $size-1]} {$index >= 0} {incr index -1} {
        set col1 [$l1 itemcget $index -background]
        set col2 [$l2 itemcget $index -background]
        ## Is this index an error?
        if {[regexp {red|yellow} "$col1$col2"]} {
          ## We have found one!
          $l1 see $index;            $l2 see $index
          $l1 selection clear 0 end; $l2 selection clear 0 end
          $l1 selection set $index;  $l2 selection set $index
          return
        }
      }
      $l1 see 0;                 $l2 see 0
      $l1 selection clear 0 end; $l2 selection clear 0 end
      $l1 selection set 0;       $l2 selection set 0
      return
    }
  }
};# CDMTools_CompareCollectionsSeeError

## CDMTools_CompareCollectionsCreateLog_DrawScrollbarBackground
#  Draw a graphical diff in a canvas-based scrollbar widget.
proc CDMTools_CompareCollectionsCreateLog_DrawScrollbarBackground {w l1 l2} {
  update
  set canvas $w.c
  if {![winfo exists $canvas]} {return}
  set map $w.map
  set mapheight [winfo height $canvas]
  set mapwidth  [winfo width  $canvas]
  incr mapwidth
  set halfwidth [expr {int($mapwidth/2.0)}]
  set lines  [$l1 index end]
  if {$lines < [$l2 index end]} {set lines [$l2 index end]}
  set factor [expr {double($mapheight) / ($lines + 1)}]
  set lineheight [expr {int($factor+1)}]
  #$canvas configure -background white
  $canvas delete background thumbUL thumbLR thumbArea
  $map blank
  $map put \#000 -to 0 $mapheight $mapwidth $mapheight
  ## Iterate over all listbox items and draw our image...
  for {set i 0} {$i < $lines} {incr i} {
    set bg1 [$l1 itemcget $i -background]
    if {$bg1 != ""} {
      set y [expr {int($i*$factor)}]
      $map put $bg1 -to 0 $y $halfwidth [expr {$y + $lineheight}]
    }
    set bg2 [$l2 itemcget $i -background]
    if {$bg2 != ""} {
      set y [expr {int($i*$factor)}]
      $map put $bg2 -to $halfwidth $y $mapwidth [expr {$y + $lineheight}]
    }
  }
  $canvas create image 0 0 -image $map -anchor nw -tags background
  $canvas bind background <<MB_1>> \
    "cscrollbar::scrollPages $w vertical %x %y"
  $canvas lower background
  $canvas itemconfigure slider -fill [$canvas cget -background]
  ## Add a 3D bevel to the slider...
  foreach {x1 y1 x2 y2} [$canvas coords slider] {break}
  if {$y2 - $y1 < 8.0} {set y2 [expr {$y1+8.0}]}
  set dx1 [expr {$x1 + 1}]
  set dx2 [expr {$x2 - 1}]
  set dy1 [expr {$y1 + 1}]
  set dy2 [expr {$y2 - 1}]
  $canvas create rectangle $x1 $y1 $x2 $y2 -fill [$canvas cget -background] \
          -tags "thumbArea slider" -stipple gray50;# -fill #1e90ff -fill #2e8b57
  $canvas create line $x1 $y2 $x1 $y1 $x2 $y1 $dx2 $dy1 $dx1 $dy1 $dx1 $dy2 \
          -tags "thumbUL slider" -fill white
  $canvas create line $dx1 $y2 $x2 $y2 $x2 $dy1 $dx2 $dy1 $dx2 $dy2 $dx1 $dy2 \
          -tags "thumbLR slider" -fill black
  $canvas raise slider
};# CDMTools_CompareCollectionsCreateLog_DrawScrollbarBackground

proc CDMTools_CompareCollectionsCreateLog_FixScrollbarSliderBevel {canvas} {
  foreach {x1 y1 x2 y2} [$canvas coords slider] {break}
  if {$y2 - $y1 < 8.0} {set y2 [expr {$y1+8.0}]}
  set dx1 [expr {$x1 + 1}]
  set dx2 [expr {$x2 - 1}]
  set dy1 [expr {$y1 + 1}]
  set dy2 [expr {$y2 - 1}]
  $canvas coords thumbArea $x1 $y1 $x2 $y2
  $canvas coords thumbUL $x1 $y2 $x1 $y1 $x2 $y1 $dx2 $dy1 $dx1 $dy1 $dx1 $dy2
  $canvas coords thumbLR $dx1 $y2 $x2 $y2 $x2 $dy1 $dx2 $dy1 $dx2 $dy2 $dx1 $dy2
};# CDMTools_CompareCollectionsCreateLog_FixScrollbarSliderBevel

##
## Compatibility for Tk 8.4...
##
catch {::tk::IconList_CurSelection}
if {![llength [info commands ::tk::IconList_CurSelection]]} {
  namespace eval ::tk {
    proc IconList_CurSelection {w} {
      return [::tk::IconList_Curselection $w]
    };# IconList_CurSelection
  };# namespace ::tk
}
