##############################################################################
##  CreateStandAloneApplicationAppInit.tcl:                                 ##
## -----------------------------------------                                ##
##  This file implements a wizard for exporting a system or a collection of ##
##  out of Ellogon, in a stand-alone application. This application can have ##
##  a completelly custom interface or run in a batch mode, without requiring##
##  a graphical interface at all...                                         ##
##                                                                          ##
##  Copyright 1998-2008 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

## This is a stub proc, for sourcing the file...
proc ::CreateStandAloneApplication::LoadProcsToBeExported {} {}

## ::CreateStandAloneApplication::AppInitialisationProc
#  This procedure contains the initialisation code for a stand-alone
#  Application. This should be changed accordingly, when the main script of
#  the Application changes...
proc ::CreateStandAloneApplication::AppInitialisationProc {} {
## Initialisation procedure...
proc Init {} {
  global ApplicationName ApplicationVersion ApplicationVersionId \
         ApplicationInfo HomeDir UserHomeDir UserAppDir PluginModulePath \
         CollectionPath CDM_Embeded env auto_path tcl_platform argv argc
  ## Check Tcl version. We require at least 8.4...
  if {[catch {package require Tcl 8.4}]} {
    error "\n$ApplicationName requires Tcl/Tk version 8.4 (or greater)\
           \nin order to operate. Currently, the installed version\
           \nseems to be Tcl $::tcl_version\
           (exact version information: $::tcl_patchLevel)"
  }
  set ApplicationInfo(ARGV) $argv
  set ApplicationInfo(ServerMode) 0
  set ApplicationInfo(LoadHTMLRenderer) 1
  ## First of all, are we running under wish?
  if {[info exists ::tk_version]} {
    ## We are running under Tk...
    set ApplicationInfo(TkPresent)           1
    set ApplicationInfo(EnableGUI)           1
    set ApplicationInfo(DisplayMainWindow)   1
    set ApplicationInfo(DisplaySplashScreen) 1
    ## Iconify the main window. Actually, this is a bug under windows.
    ## Do not iconify that, but make it small and invisable...
    wm geometry . 1x1-1-1
    wm overrideredirect . 1
    wm transient .
  } else {
    ## We are running under plain Tcl...
    set ApplicationInfo(TkPresent)           0
    set ApplicationInfo(EnableGUI)           0
    set ApplicationInfo(DisplayMainWindow)   0
    set ApplicationInfo(DisplaySplashScreen) 0
  }

  ## Until tcl improoves its bindings for non-latin languages, do not use
  ## other locales...
  package require msgcat
  msgcat::mclocale en
  
  ## This may cause problems under LINUX...
  unset -nocomplain env(MALLOC_CHECK_)
  
  ## Program's Home Dir
  set HomeDir [file dirname [file normalize [info script]]]
  ## Is a symbolic link?
  set ScriptDir [info script]
  while {![catch {file readlink $ScriptDir} ScriptDir]} {
    set HomeDir [file nativename [file dirname $ScriptDir]]
  }
  if {[string equal $HomeDir .]} {set HomeDir [pwd]}
  
  ## Get User's home directory: We try to locate the proper path from a set of
  ## environmental variables...
  foreach var {HOME HOMEPATH USERPROFILE ALLUSERSPROFILE APPDATA} {
    if {[info exists env($var)]} {
      if {[file isdirectory $env($var)]} {
        set UserHomeDir $env($var)
        break
      }
    }
  }
  ## Under windows we have to also combine HOMEDRIVE & HOMEPATH...
  if {![info exists UserHomeDir] && 
       [string equal $tcl_platform(platform) windows] &&
       [info exist env(HOMEDRIVE)] && [info exist env(HOMEPATH)]} {
    if {[file isdirectory $env(HOMEDRIVE)$env(HOMEPATH)]} {
      set UserHomeDir $env(HOMEDRIVE)$env(HOMEPATH)
    }
  }
  ## Have we located the needed path?
  if {![info exists UserHomeDir]} {
    set UserHomeDir [pwd]
  }
  
  ## Application Preferences & tmp file Dir..
  set UserAppDir [file join $UserHomeDir [string tolower .$ApplicationName]]

  ## Handle user arguments...
  array set ApplicationInfo {InitCommands {} InitScripts {} ScriptArgs {}}
  set UserHandlersMissing 1
  if {$UserHandlersMissing && [file exists $HomeDir/UserHandlers.tcl]} {
   source $HomeDir/UserHandlers.tcl
   set UserHandlersMissing 0
  }
  foreach {arg val} $argv {
    switch -glob -- $arg {
      -comma*           {lappend ApplicationInfo(InitCommands) $val}
      -gui              {
        set ApplicationInfo(EnableGUI) [expr {$val}]
        if {$ApplicationInfo(EnableGUI)} {
          package require Tk
          set ApplicationInfo(TkPresent) 1
          set ApplicationInfo(DisplayMainWindow)   1
          set ApplicationInfo(DisplaySplashScreen) 1
          wm geometry . 1x1+0+0
          wm overrideredirect . 1
          wm transient .
        } else {
          set ApplicationInfo(TkPresent)           0
          set ApplicationInfo(DisplayMainWindow)   0
          set ApplicationInfo(DisplaySplashScreen) 0
        }
      }
      -encoding         {encoding system $val; set ::Encoding $val}
      -script           {lappend ApplicationInfo(InitScripts)  $val}
      -args             {lappend ApplicationInfo(ScriptArgs)   $val}
      -confdir          {set UserAppDir  $val}
      -tempdir          {set TempDir     $val}
      -userdir          {set UserHomeDir $val}
      -showMainWindow   {set ApplicationInfo(DisplayMainWindow) [expr {$val}]}
      -showSplashScreen {set ApplicationInfo(DisplaySplashScreen) [expr {$val}]}
      -printHomeDir     {puts $HomeDir; exit}
      -printVersion     {
        puts "$::ApplicationName $::ApplicationVersion"; exit
      }
      -printTclBlendVersion {puts "tclblend1.3.1"; exit}
      -printInfo        {
        puts [array get ApplicationInfo]; exit
      }
      -loadHTMLRenderer {set ApplicationInfo(LoadHTMLRenderer) [expr {$val}]}
      -inputdir         {set ::UserInputPaths  $val}
      -outputdir        {set ::UserOutputPaths $val}
      -inputfile        {lappend ::UserInputPaths  $val}
      -outputfile       {lappend ::UserOutputPaths $val}
      -server_mode      {set ApplicationInfo(ServerMode) $val}
      default           {
        ## We have encoutered an unrecognised option. Our last chance is to
        ## load the User Handlers and see if the argument handling procedure
        ## recognises the argument...
        if {![ApplicationArgumentHandler $arg $val]} {
          puts stderr "$::argv0: Error during Initialisation."
          puts stderr "\tUnknown parameter \"$arg\"!"
          puts stderr "  Valid Parameters are:"
          foreach spec [lsort {{gui <boolean>} {encoding <encoding>}
                              {command <string>} {script <file>} {args <string>}
                              {showMainWindow <boolean>}
                              {showSplashScreen <boolean>} {confdir <directory>}
                              {tempdir <directory>} {userdir <directory>}
                              {inputdir <directory>} {outputdir <directory>}
                              {inputfile <file>} {outputfile <file>}
                              {loadHTMLRenderer <boolean>}
                              {printHomeDir {}} {printInfo {}}
                              {printVersion {}} {printTclBlendVersion {}}
                              {server_mode <boolean>}}] {
            puts stderr "\t-[lindex $spec 0] [lindex $spec 1]"
          }
          foreach spec [ApplicationArgumentHandler {} {}] {
            puts stderr "\t-[lindex $spec 0] [lindex $spec 1]"
          }
          exit -1
        };# if {![ApplicationArgumentHandler $arg $val]}
      }
    };# switch -glob -- $arg
  };# foreach {arg val}
  set argv {}
  set argc 0

  if {![info exists TempDir]} {
    ## Try to locate a temporary directory...
    foreach var {ELLOGON_TEMP_DIR TEMP TMP} {
      if {[info exists env($var)]} {
        if {[file isdirectory $env($var)] && [file writable $env($var)]} {
          set TempDir $env($var)
          break
        }
      }
    }
    if {![info exists TempDir]} {
      foreach dir {/tmp C:/Temp C:/tmp} {
        if {[file isdirectory $dir] && [file writable $dir]} {
          set TempDir $dir
          break
        }
      }
    }
    if {![info exists TempDir]} {
      ## We have failed to locate a temporary directory. We place it
      ## inside the UserAppDir...
      set TempDir $UserAppDir
    }
    if {[info exists env(USER)]} {
      set TempDir $TempDir/[string tolower $ApplicationName]_temp_$env(USER)
    } elseif {[info exists env(USERNAME)]} {
      set TempDir $TempDir/[string tolower $ApplicationName]_temp_$env(USERNAME)
    } else {
      set TempDir $TempDir/[string tolower $ApplicationName]_temp
    }
  }

  ## Normalise the values. Especially under windows, retrieve the short path,
  ## i.e. the path without spaces...
  switch -glob $tcl_platform(machine) {
    *_64    {set bits x86_64}
    default {set bits x86_32}
  }
  switch $tcl_platform(platform) {
    unix    {
      set ApplicationInfo(LoadBinaryFolderName) $tcl_platform(os)/$bits
      set ApplicationInfo(PathSplitCharacters) "\;\:"
      set CDM_Embeded 0
    }
    windows -
    default {
      set ApplicationInfo(LoadBinaryFolderName) Windows/$bits
      set ApplicationInfo(PathSplitCharacters) "\;"
      foreach var {HomeDir UserHomeDir UserAppDir TempDir} {
        set $var [string map {\\ /} [set $var]]
        if {[string first { } [set $var]] != -1} {
          catch {file mkdir [set $var]
            set $var [file attributes [set $var] -shortname]
          }
        }
      }
      ## Under windows, there is always a possibility that we are running in an
      ## Active-X container. We may change our behaviour in such a case...
      switch -glob -- [string tolower [info nameofexecutable]] {
        *wish*  -
        *tclsh* {set CDM_Embeded 0}
        default {set CDM_Embeded 1}
      }
    }
  }

  ## Ensure that the various directories exist...
  file mkdir $TempDir
  file mkdir $UserAppDir
  
  ## Update auto_path, so the application's libraries can be found
  lappend auto_path [file join $HomeDir CDM] \
                    [file join $HomeDir lib] \
                    [file join $HomeDir GUI] \
                    [file join $HomeDir GUI Dialogs] \
                    [file join $HomeDir GUI Widgets] \
                    [file join $HomeDir GUI Widgets BWidget] \
                    [file join $HomeDir Viewers] \
                    [file join $HomeDir Utils] \
                    [file join $HomeDir lib tkdnd lib] \
                    [file join $HomeDir lib tclblend1.3.1] \
                    [file join $HomeDir lib tree-8.0.4.1 library] \
                    [file join $HomeDir ellogon2.0 lib_elep] \
                    [file join $HomeDir ellogon2.0 lib]
  
  ## Plug-in Module Path-a list of directories where plug-in modules may live
  set PluginModulePath [list $HomeDir/share/modules]
  
  ## Collection Path - a list of directories that contain collections
  set CollectionPath [list $UserHomeDir/Collections]

  #########################################################################
  ## Make the load command more clever, by trying to load shared objects in
  ## subdirectories of the specified path, according to the platform...
  namespace eval ::CDM {}
  rename load CDM_load
  proc load {filename {pkgName _@} {interp _@}} "
    set dir  \[file dirname \$filename\]
    set file \[file tail    \$filename\]
    set msg {}
    foreach fn \[list \$dir/$ApplicationInfo(LoadBinaryFolderName)/\$file \
                      \$filename \$dir/\$file\] {
      if {!\[string equal \$interp _@\]} {
        set code \[catch {uplevel 1 CDM_load \[list \$fn\]\
                                 \$pkgName \$interp} result\]
      } elseif {!\[string equal \$pkgName _@]} {
        set code \[catch {uplevel 1 CDM_load \[list \$fn\] \$pkgName} result\]
      } else {
        set code \[catch {uplevel 1 CDM_load \[list \$fn\]} result\]
      }
      if {!\$code} {return -code \$code \$result}
      append msg \"Loading Library: \$fn\nError: \$result\n\n\"
    }
    ## If we got here, nothing worked. Return an error...
    return -code \$code \$msg
  "
  catch {rename unload CDM_unload}
  proc unload {filename {pkgName _@} {interp _@}} "
    set dir  \[file dirname \$filename\]
    set file \[file tail    \$filename\]
    set msg {}
    foreach fn \[list \$dir/$ApplicationInfo(LoadBinaryFolderName)/\$file \
                      \$filename\] {
      if {!\[string equal \$interp _@\]} {
        set code \[catch {uplevel 1 CDM_unload \[list \$fn\]\
                                 \$pkgName \$interp} result\]
      } elseif {!\[string equal \$pkgName _@]} {
        set code \[catch {uplevel 1 CDM_unload \[list \$fn\] \$pkgName} result\]
      } else {
        set code \[catch {uplevel 1 CDM_unload \[list \$fn\]} result\]
      }
      if {!\$code} {return -code \$code \$result}
      append msg \"Loading Library: \$fn\nError: \$result\n\n\"
    }
    ## If we got here, nothing worked. Return an error...
    return -code \$code \$msg
  "

  ## Also rename exit, in order to cleanly exit the application...
  rename exit CDM_exit
  proc exit {{code 0}} {
    catch {
      ## This is to deal with a tk bug...
      foreach image [image names] {
        if {[string match *Over* $image]} {
          image delete $image
        }
      }
    }
    CDM_exit $code
  }

  ## Make some information public...
  foreach var {ApplicationName ApplicationVersion ApplicationVersionId
               HomeDir UserHomeDir UserAppDir PluginModulePath
               CollectionPath TempDir CDM_Embeded} {
    set ApplicationInfo($var) [set $var]
  }
  
  ## Try to load TclPro's ByteCode loader...
  catch {load $HomeDir/CDM/libTbcLoad[info sharedlibextension] TbcLoad}
  if {[string equal $tcl_platform(platform) windows]} {
    catch {load $HomeDir/CDM/libWinIco.dll}
  }
}
## Initialise Application...
Init
rename Init {}
set ConfigDir $HomeDir/share/Config
source $ConfigDir/Options.tcl
catch {source $HomeDir/Viewers/Colors.tcl}

## Load User Options...
set UserCommandsFile $UserAppDir/$ApplicationName.init
if {[file readable $UserCommandsFile]} {
  if {[catch {source $UserCommandsFile} error]} {
     catch {puts "Error in $UserCommandsFile:\n    $error"}
  }
}
catch {source $UserAppDir/UserPreferences.tcl}
if {![string equal [encoding system] $KeyboardLocale]} {
  if {[catch {encoding system $KeyboardLocale}]} {
    gut_info "Invalid Keyboard Encoding \"$KeyboardLocale\".\
              \nSetting Keyboard Encoding to \"[encoding system]\""
    set KeyboardLocale [encoding system]
  }
  catch {
    fconfigure stdout -encoding $KeyboardLocale
    fconfigure stderr -encoding $KeyboardLocale
  }
}
set GUI_Options(InputForeground) $GUI_Options(TextFg)
set GUI_Options(InputBackground) $GUI_Options(TextBg)
if {$ApplicationInfo(TkPresent)} {
  source $ConfigDir/OptionDatabase.tcl
}
unset ConfigDir

## Show the Start-Up Screen...
if {[string length $::GUI_MainWindow::SplashImage]} {
  if {![info exists ApplicationInfo(DisplaySplashScreen)]} {
    set ApplicationInfo(DisplaySplashScreen) 1
  }
  LDM_ShowSplashWindow $::GUI_MainWindow::SplashImage
} else {
  set ApplicationInfo(DisplaySplashScreen) 0
}
set SplashPercent 0
LDM_SetSplashWindowStatus {Loading Document Manager...} $SplashPercent

## Load the Collection/Document Manager (CDM). If an error occurs, give up :-)
if {[catch {load $HomeDir/CDM/libCDM[info sharedlibextension]} error]} {
  catch {bind .splashscr <Visibility> {}; lower .splashscr}
  catch \
    {puts stderr "The Core Library cannot be loaded. The error was:\n$error"}
  if {$ApplicationInfo(EnableGUI)} {
    if {[winfo exists .splashscr]} {
      tk_messageBox -title {Fatal Error: } -type ok -default ok -icon error \
        -message "The Core Library cannot be loaded. The error was:\n$error" \
        -parent .splashscr
    } else {
      tk_messageBox -title {Fatal Error: } -type ok -default ok -icon error \
        -message "The Core Library cannot be loaded. The error was:\n$error"
    }
  }
  exit -1
}
catch {LDM_SplashWindowNotify language cpp}
catch {
  set CDM_TempDir                 $ApplicationInfo(TempDir)
  set CDM_FullGateCompatibility   $GUI_Options(FullGateCompatibility)
  set CDM_EmulateGateGetValue     $GUI_Options(EmulateGateGetValue)
  set CDM_EnableMemoryCompression $GUI_Options(EnableMemoryCompression)
  set CDM_DefaultEncoding         $KeyboardLocale
}

if {$GUI_Options(OpenStartupConsole)} {
  ## If platform is windows, open a virtual console
  LDM_SetSplashWindowStatus {Opening Console Window...} $SplashPercent
  incr SplashPercent 5
  if {[catch {gut_Console} Console]} {
    catch {console show}
    catch {bind all <F2> {catch {console show}}}
  } else {
    gut_ConsoleConfigure $Console hide
  }
}

set CDM_DisplayErrorMessage 0
set CDM_Pid [pid]
incr SplashPercent 15

## Has the user enabled Java? See if we can load tclBlend...
if {[file exists $HomeDir/CDM/Java]} {set GUI_Options(EnableJava) 1}
if {$GUI_Options(EnableJava)} {
  incr SplashPercent  5
  LDM_SetSplashWindowStatus {Loading Java Bindings...} $SplashPercent 200
  ::CDM_Java::InitialiseJava
}

## Has the user enabled Perl? See if we can load tclperl...
if {$GUI_Options(EnablePerl)} {
  incr SplashPercent  5
  LDM_SetSplashWindowStatus {Loading Perl Bindings...} $SplashPercent 210
  ::CDM_Perl::InitialisePerl
}

## Has the user enabled Python? See if we can load tclpython...
if {$GUI_Options(EnablePython)} {
  incr SplashPercent  5
  LDM_SetSplashWindowStatus {Loading Python Bindings...} $SplashPercent 220
  ::CDM_Python::InitialisePython
}

LDM_SetSplashWindowStatus {Loading Plug-in Components...} $SplashPercent
source $HomeDir/Utils/Configure.tcl
if {$ApplicationInfo(EnableGUI)} {
  ## Load the Image package...
  if {$::tcl_platform(os) != "Darwin"} {
    catch {package require Img}
  }
  ## Load the Drag'n'Drop extension...
  catch {package require tkdnd}
}
source $HomeDir/Utils/LoadPlugins.tcl
LDM_SetSplashWindowStatus {Loading User Preferences...} $SplashPercent 280
set UserCommandsFile $UserAppDir/$ApplicationName.rc
if {[file readable $UserCommandsFile]} {
  if {[catch {source $UserCommandsFile} error]} {
    catch {puts "Error in $UserCommandsFile:\n    $error"}
  }
}
LDM_SetSplashWindowStatus {Loading Images...} 96 300
set tmp $ApplicationInfo(EnableGUI)
set ApplicationInfo(EnableGUI) 0
source $HomeDir/GUI/GUI.tcl          ;#Remove Line for wrapper
set ApplicationInfo(EnableGUI) $tmp
unset -nocomplain tmp

LDM_SetSplashWindowStatus {Registering Tools...} 98 260
GUI_RegisterTools

LDM_SetSplashWindowStatus "Loading $Locale Messages..." 99 310
## Start the Graphical User Interface...
if {[catch {GUI_ReadLocale}]} {
  set Locale English
  GUI_ReadLocale
}

if {!$CDM_Embeded && $ApplicationInfo(EnableGUI)} {
  LDM_SetSplashWindowStatus {Starting Graphical Interface...} 100 320
} else {
  catch {LDM_DeleteSplashWindow .splashscr}
}
if {[info exists ConsoleWin]} {
  gut_ConsoleConfigure $Console show; unset Console
}

unset -nocomplain file UserCommandsFile SplashPercent

after idle {source $HomeDir/GUI/Dialogs/BgError.tcl}
};# ::CreateStandAloneApplication::AppInitialisationProc

## ::CreateStandAloneApplication::CreateGUIAppWindow
#  This procedure will create and display the main window of a GUI
#  Application, generated by this wizard...
proc ::CreateStandAloneApplication::CreateGUIAppWindow {{win .}} {
  bind all <F2> {catch {console show}}
  set window .[string trim $win .]
  ## Set window title...
  catch {wm title $window $::GUI_MainWindow::Title}
  ## Remove the main window from the screen, so as our changes not to be
  ## visible...
  catch {wm withdraw $window}

  ## Create some images...
  image create photo ::GUI_MainWindow::ITOpen -data \
  {R0lGODlhEAAQAKEAAAAAAP//AISCAP///yH+Dk1hZGUgd2l0aCBHSU1QACH5
   BAEKAAMALAAAAAAQABAAAAIznI+pCL0JTowP0kYVHiCMgIXQR3peZnRliXJs
   CDcmINS2PX30fV8g35PsgAJUDFZJKhcFADs=}
  image create photo ::GUI_MainWindow::BasedOn -gamma 2.0 -data \
  {R0lGODlhyAAhAOcAAAgInGICjpJOro5CqoY2pn4qnnYemppWspJKroo+poIy
   ogIejp5etpZSso5GrIY6pn4unmZ6ul5utqJmuppatJ4CAlJmskperp5itoY+
   poIuov4qKP5WVj5WqjJKoiI6mopCqv5SUv56ev5ubv5eXjpSpjZOpjZKojpO
   pv5WUv5aVv5mZJI6soYmqnIelnIinnYqno42rpZGtjpSqiZGpm4alu7u9v7+
   /trG5sJCSvJWWv5eWipCnjpOuhY2ks7W5u7W1rZGTupWWv5iYgI+hjpSrho6
   lu7y9qqz2OLl8b5OVvJeXv5iXmqi5EaG3iJCosbS5tJWXn6u5nKm4kaK3kqK
   3hpSlsZ+WvKaYmKe4naq5ipmrt6OYk6O4lqW4mKa4nKCwPL2+taObEKG3laS
   4l6a4k6O3iJanip22jZ+3Spenvb2/t6Wdhpq1iZy2i563jpyuvLe1uaacjZq
   qvKeak7i3lLi3kK+utbu7jpQp1Zos/bu4uK6ev7WcjbS0mbi3lbi3ka2svLy
   +f769vbKbgqSjm7i4lri3lLOyk6+uvby7vLKegKGfmri3mLi3lbOyqbKWvr6
   8vr+9vb68ur21r7idv7Wbv7adl7i3rbmYq7WYrLWarbWbrracr7edrrmZv7W
   av7aejbe2kLe2rLiXrbiXyLa1ira2j7e2kbe3mrm4nLm4rbmZrrma77mcP7O
   VhrW1jre2nri3sLmeqbeQv7GOp5uAv7KSh7a1iba1lqGAq7iVqriTqbePqLe
   OJraKJ7eMJbaIpLaGorWCv4+Pv4uLv5CPv4yLv4iIv5OTP46Ov4aFv4eHv5K
   Rp4+Av6mZv6yev6udv6qav6eWv6WSv6OPP6qbv6iXv6aUP6GMP6SRP5+Iv6K
   Nv6CKP56F/6ucv7CJv7Sav7SWv7KRv7GNv7OSv7KOv7CKv6+Gv6+Fv//////
   ////////////////////////////////////////////////////////////
   /////////////////////yH5BAEKAP8ALAAAAADIACEAAAj+AP8JHEiwoMGD
   CBMqXMiwocOHECNKVBigYsWJGDNq3Mixo0eHAQQMIFDAQICPKFOqXMkSYoAD
   CBIoKHCypc2bOHMqXLCgYAAGDRw8gFBzYIUKCQEoBYCS6UenDplCXbpUoFOo
   KaUWpFq144IIEnr+qzhhAgWRBIjWrECsGFKDXLFulDvxKt2kVP/FtSr1Lse8
   BPc+9PvvqwQLF3hGwHDWQYIMCTQU2HCUg7FjyN4GnmoXa1e+W63CVTqQNOmt
   prVuHq0XNGq9nF+3di36NGzbq3PPbv25oGHEHTx8uKDggfEHCUAkGBBCxAgS
   yZRtWKZZdGm+Xbli79z79mnB2Lf+cydM+zps8eL7ZtfO0DPg1LcRLjh8oYQH
   EydQlDCeoH8CByGkoAIHK4RgTDHMVOcdbtbh9p1q55U2lWsMfkeheXi9B9qD
   EZ4XG4ftBWbee+QtcEFwJpiAwoostODCCzDEIEMKHKigAgkqNKOMgrY9uJ6P
   7HXY4VUYjmjXguQJGSGRG07YY1/lIfQkYEMmucB9KZYwAw012GDDDTbgoEAO
   OuxAwpnJKCgklGvCJ+GbcLpZ4ZEeJgnnahY2qWeDd+LVZ3hRGrQADyma0IMP
   P9yg6KI3ABGEEDsMMUQKaj5ZJ5vwpcZmnJjKBSKHDMqG55Gglmrdbgo5CKVp
   fxZERAT+WRZhxBE3IGHrrUgkYYMSSzAhqQjOoEYle8NOGSqJUVY14XYLxaXa
   lJv9SGeoB6kqJ50IEdGEE/o9AYWiPIXLExKNRjHECs9AE6xO7La0bLsOESGv
   FFM4QUUVViy6ALmMLnoFFs+kG40zBMNrcEdBHpwQEVk0ocUUWVBBxRa0gstv
   vzccwUXA0kgzDTXVrKvwyCSrREQXXnwxhRdV4PsluGDgenEYYjwjjTXXYFNN
   NiKX7PPPGBExRhVklGFGFV2cwai44TJaszXWYKNNNtv0DPTVWCtEBBppDN1y
   FWpUfMO+GCu6BhtQW6MNN914Y3VOCTdkJ0Rzp9rb3NRORET+G268kQbSVcDx
   rcVl3xCHHNZ88001bbsNL3hRPVW3lMQS1tlH8m6dhhlddDGHvhf3ywYdUO9M
   8Ns4Hdsmn7y5JyqrpLrul6U+SqgeoB7Ohq1BRKThRR123IEHuHnoYbzx5O7B
   Rx+hRMPzyHQJtpem0nLH7HjjUf5m9soCab2zeRPhxx922AFIIILUqgcS66+f
   xCB8ENLHJZeAg7pO08e+fadL8t8jhfyrFrHgAsAhFbB/CSmEIQ5hPkQkYngY
   UwQfFtGHCvbBFtCLnv5YV6q8TC9OI2oVevqXHSOF8IDkYUQjHAGIFgLiEZCI
   hCRuMAlKVCJ+lqBfKMKBQYWpbk/+ctqfqKqEwN1Q61j+m9YGEZi3fzACEJj4
   wyEAcYgqZkITm+BEJzzxCVDQj3mgEMc4emiwH+rudnk6Y4NW9SElxoc1r9FU
   EU11ISUNhBGiGIUdouiIQ2ACE6QoRSkyIchP0O8SfQDFK25BDnCQsV35o412
   rmU7zmDLjQvSXmjQ071KSvKABmGEKU6BilQcQhWrcIQgWVEKVrRCkK6gnyVe
   UQ5znAMdj8yaEcsYKF0ShBGwOEUsUsEIRsiiErOYhSsqUYpWkIIWtbCFLRgJ
   jnTkUpdxgxuVfHkQRuAiF4wQiC504QpWZCITu+BFL8YpEFucw5rcjKc8uxlO
   guginbQu8MUvdFEQac7znwA1iC58AYxf/CIY/AyoQhc6EF0cVBjCGEZCGUpR
   gI7zog4JCAA7}

  ##
  ## Window Title area
  ##
  frame ${win}title -relief flat
    ## We must be carefull with images...
    if {[catch {image create photo ::GUI_MainWindow::titleImage \
                             -data $::GUI_MainWindow::TitleImage}]} {
      ## Create an empty image...
      image create photo ::GUI_MainWindow::titleImage
    }
    label ${win}title.image -image ::GUI_MainWindow::titleImage;# \
      -background $::GUI_MainWindow::TitleTextBackground
    label ${win}title.text -text $::GUI_MainWindow::TitleText -anchor center \
      -foreground $::GUI_MainWindow::TitleTextForeground \
      -font       $::GUI_MainWindow::TitleTextFont ;#\
      -background $::GUI_MainWindow::TitleTextBackground
    grid ${win}title.image ${win}title.text -sticky snew
    grid columnconfigure ${win}title 1 -weight 1
  grid ${win}title -sticky snew -padx 2 -pady 2

  ##
  ## Input file/directory selection area
  ##
  set entry_width 60
  frame ${win}inout_select -relief flat
    label ${win}inout_select.ilabel -text $::GUI_MainWindow::InputMessage \
      -anchor e
    entry ${win}inout_select.ipaths -width $entry_width -textvariable \
      ::UserInputPaths -highlightthickness 0
    set ::GUI_MainWindow::InputPathWidget ${win}inout_select.ipaths
    button ${win}inout_select.ibrowse -text Browse \
       -image ::GUI_MainWindow::ITOpen -compound left \
       -padx 5 -highlightthickness 0
    ## Display the encoding selection menubutton, if specified...
    catch {
      label ${win}inout_select.iencLabel -anchor e \
        -text $::GUI_MainWindow::EncodingMessage
      set menu ${win}inout_select.ienc.mn
      menubutton ${win}inout_select.ienc -relief raised -indicatoron 1 \
        -anchor c -menu $menu -text $::Encoding -textvariable ::Encoding \
        -width 150 -image [::Encoding::GetEncodingImage $::Encoding] \
        -compound left -padx 10 -fg #ffff00 -bg #009900
      menu $menu -tearoff 0
      GUI_AppendEncodingsToMenu $menu "set ::Encoding \$Encoding
        ${win}inout_select.ienc configure -image \$EncodingImage \
        -text \$Encoding" 1
    }
    label ${win}inout_select.olabel -text $::GUI_MainWindow::OutputMessage \
      -anchor e
    entry ${win}inout_select.opaths -width $entry_width -textvariable \
      ::UserOutputPaths -highlightthickness 0
    set ::GUI_MainWindow::OutputPathWidget ${win}inout_select.opaths
    button ${win}inout_select.obrowse -text Browse \
       -image ::GUI_MainWindow::ITOpen -compound left \
       -padx 5 -highlightthickness 0
    switch $::GUI_MainWindow::FileSelectionMode {
      directory {
        GUI_MakeEntryDirectoryDropTarget ${win}inout_select.ipaths
        ${win}inout_select.ibrowse configure -command \
          "SelectDirectory ${win}inout_select.ipaths"
      }
      files {
        GUI_MakeEntryFileDropTarget ${win}inout_select.ipaths
        ${win}inout_select.ibrowse configure -command \
          "SelectFiles ${win}inout_select.ipaths"
      }
      default {
      }
    }
    switch $::GUI_MainWindow::OutputSelectionMode {
      directory {
        GUI_MakeEntryDirectoryDropTarget ${win}inout_select.opaths
        ${win}inout_select.obrowse configure -command \
          "SelectDirectory ${win}inout_select.opaths"
      }
      file {
        GUI_MakeEntrySaveFileDropTarget ${win}inout_select.opaths
        ${win}inout_select.obrowse configure -command \
          "SelectFiles ${win}inout_select.opaths 1"
      }
      default {}
    }
    label ${win}inout_select.based_on -image ::GUI_MainWindow::BasedOn
    
    grid ${win}inout_select.ilabel ${win}inout_select.ipaths \
         ${win}inout_select.ibrowse -sticky snew -padx 2 -pady 1
    if {$::GUI_MainWindow::EncodingSelectableByUser} {
      grid ${win}inout_select.iencLabel ${win}inout_select.ienc \
        -sticky snew -padx 2 -pady 1
    }
    if {![string equal $::GUI_MainWindow::OutputSelectionMode none]} {
      grid ${win}inout_select.olabel ${win}inout_select.opaths \
           ${win}inout_select.obrowse -sticky snew -padx 2 -pady 1
    }
    grid ${win}inout_select.based_on -columnspan 3 -sticky sne -padx 2 -pady 2
    grid columnconfigure ${win}inout_select 1 -weight 1
  grid ${win}inout_select -sticky snew -padx 2 -pady 2

  ##
  ## Buttons area
  ##
  frame ${win}buttons
    button ${win}buttons.exec -width 20 -text $::GUI_MainWindow::ExecMessage \
           -command Execute
    button ${win}buttons.exit -width 20 -text $::GUI_MainWindow::ExitMessage \
           -command Exit
    pack ${win}buttons.exec ${win}buttons.exit -side left -padx 20 -pady 10
  grid ${win}buttons -padx 2 -pady 2

  ## Call the user supplied window procedure. With this procedure the user can
  ## add/remove elements from the window, before it gets shown on the screen.
  if {[llength [info procs DisplayMainWindow_UserHandler]]} {
    DisplayMainWindow_UserHandler $window
  }

  ## Make our window visible again...
  grid columnconfigure $window 0 -weight 1
  catch {gut_ModalDlg $window}
  catch {wm deiconify $window}
  catch {LDM_DeleteSplashWindow .splashscr}
};# ::CreateStandAloneApplication::CreateGUIAppWindow

## ::CreateStandAloneApplication::AppExecutionProc
#  This procedure will process the specified files with the Annotators...
proc ::CreateStandAloneApplication::AppExecutionProc {} {
proc Execute {args} {
  global Encoding SyncCollection Components CollectionName
  if {[info exists ::GUI_MainWindow::InputPathWidget]} {
    set InputPathWidget  $::GUI_MainWindow::InputPathWidget
    set OutputPathWidget $::GUI_MainWindow::OutputPathWidget
  } else {
    set InputPathWidget  {}
    set OutputPathWidget {}
  }

  if {![llength $Components]} {
    gut_info "No Annotators specified!"
    return
  }

  if {![string equal $::GUI_MainWindow::OutputSelectionMode none]} {
    if {![string length $::UserOutputPaths]} {
      switch $::GUI_MainWindow::OutputSelectionMode {
        directory {gut_info "Please select an Output Directory!"}
        file      {gut_info "Please select an Output File!"}
        default   {}
      }
      GUI_FlashWidget $OutputPathWidget
      return
    }
  }
  
  ##
  ## Search for the files...
  ##
  set Files {}
  switch $::GUI_MainWindow::FileSelectionMode {
    directory {
      ## The user is expected to have selected a directory...
      if {![string length $::UserInputPaths]} {
        gut_info "Please select a Directory containing the Files\
                  to be processed!"
        GUI_FlashWidget $InputPathWidget
        return
      }
      ## Is it an existing Directory?
      if {![file exists $::UserInputPaths]} {
        gut_info "The specified Input Directory does not exist!"
        GUI_FlashWidget $InputPathWidget
        return
      }
      ## Is it a Directory?
      if {![file isdirectory $::UserInputPaths]} {
        gut_info "The specified Input Directory is not a Directory!"
        GUI_FlashWidget $InputPathWidget
        return
      }
      
      ## Search the directory to locate any files...
      foreach pattern $::GUI_MainWindow::DirectorySelectionExtensions {
        foreach filename [glob -nocomplain -directory $::UserInputPaths \
                               -type f -- $pattern] {
          if {![file isfile $filename]} {
            Log "  File $filename is not a normal file.\
                   Advancing to the next one..."
            continue
          }
          lappend Files $filename
        }
      }
    }
    files {
      ## The user is expected to have selected files...
      if {![llength $::UserInputPaths]} {
        gut_info "Please select some Files to be processed!"
        GUI_FlashWidget $InputPathWidget
        return
      }
      foreach filename $::UserInputPaths {
        if {![file isfile $filename]} {
          Log "  File $filename is not a normal file.\
                 Advancing to the next one..."
          continue
        }
        lappend Files $filename
      };# foreach filename $::UserInputPaths
    }
  };# switch $::GUI_MainWindow::FileSelectionMode
  set DocumentNumber [llength $Files]
  if {!$DocumentNumber} {
    gut_info "No Files for Processing found!"
    GUI_FlashWidget $InputPathWidget
    return
  }

  ## Ensure that the Output path has been specified...
  if {![string equal $::GUI_MainWindow::OutputSelectionMode none]} {
    if {![string length $::UserOutputPaths]} {
      switch $::GUI_MainWindow::OutputSelectionMode {
        directory {gut_info "Please select an Output Directory!"}
        file      {gut_info "Please select an Output File!"}
        default   {}
      }
      GUI_FlashWidget $OutputPathWidget
      return
    }
    ## If the Output path exists, must be of the same type as expected...
    if {[file exists $::UserOutputPaths]} {
      switch $::GUI_MainWindow::OutputSelectionMode {
        directory {
          ## Is it a directory?
          if {![file isdirectory $::UserOutputPaths]} {
            gut_info "The specified Output Directory exists,\
                      but it is not a Directory!"
            GUI_FlashWidget $OutputPathWidget
            return
          }
        }
        file      {
          ## Is it a file?
          if {![file isfile $::UserOutputPaths]} {
            gut_info "The specified Output File exists, but it is not a File!"
            GUI_FlashWidget $OutputPathWidget
            return
          }
        }
        default {}
      };# switch $::GUI_MainWindow::OutputSelectionMode
    };# if {[file exists $::UserOutputPaths]}
  };# if {![string equal $::GUI_MainWindow::OutputSelectionMode none]}

  ## The Input/Output paths seem fine. Rememeber them the next time this
  ## Application starts...
  set ::SavedState(::UserInputPaths)  $::UserInputPaths
  set ::SavedState(::UserOutputPaths) $::UserOutputPaths
  set ::SavedState(::Encoding)        $::Encoding

  ##
  ## Specify any special Annotator parameters:
  ## Each parameter has two elements:
  ##   1) Its position in the parameters specified in creole_config.tcl
  ##   2) Its value
  ## For example, if we want to change the first parameter to the value "val",
  ## we should specify the list {{0 val}}
  ## Remember, $::UserOutputPaths holds the selected by the user Output
  ## Directory/File(s), and $::UserInputPaths holds the selected by the user
  ## Input Directory/File(s)
  ##
@AnnotatorDefaultParameterList@

  set Progress 0.0
  set Step [expr {100.0/$DocumentNumber/[llength $Components]}]
  set wwin [ggi_wait_message {} $Progress]
  
  ##
  ## Create the Collection object...
  ##
  set InitTime [clock seconds]
  Log "Starting Execution: Time=[clock format $InitTime \
                                   -format {%H:%M:%S, %A %d %B %Y} -gmt 1] GMT"
  Log "Removing any previous contents of directory \"$CollectionName\"..."
  catch {file delete -force -- $CollectionName}
  Log "Creating Collection: \"$CollectionName\"..."
  set col [tip_CreateCollection $CollectionName {} $Encoding]
  Log "  Collection Tcl command: \"$col\", Collection encoding: \"$Encoding\""
  if {$SyncCollection} {
    Log "  Saving Collection..."
    tip_Sync $col
  }

  ##
  ## Add Files into the Collection...
  ##
  set index 0
  foreach filename $Files {
    incr index
    if {![file exists $filename]} {
      Log ">>>> Skipping non-existent file \"$filename\"..."
      Log "           (file $index out of $DocumentNumber)"
      continue
    }
    set file [open $filename]
    fconfigure $file -encoding $Encoding
    set data [read $file]
    close $file
    ## Create the Collection Document...
    Log ">>>> Adding file \"$filename\" into Collection \"$col\"..."
    Log "           (file $index out of $DocumentNumber)"
    set doc [tip_CreateDocument $col $filename $data {} {} $Encoding]
    Log "  Document Tcl command: \"$doc\", Document encoding: \"$Encoding\""
    ## Apply the Document Handler...
    if {[catch {DocumentCreationHandler $col $doc} status]} {
      ErrorLog $status
      continue
    }
    if {$SyncCollection} {
      Log "  Saving Document ($doc)..."
      tip_Sync $doc
    }

    ## The Document is in place, ready to be processed by the Annotators...
    catch {
      ggi_wait_message "Processing Document:\n   [tip_GetId $doc]\
                        ($index from $DocumentNumber)" $Progress 1 $wwin
    }

    if {$index == 1} {
      ## Initialise the Annotators (Components)...
      foreach Annotator $Components {
        ## Print the Parameters that will be used for each Annotator...
        set Values [lindex [GUI_GetModuleOptions $Annotator] 0]
        foreach {pos value} $AnnotatorParameters($Annotator) {
          set Values [lreplace $Values $pos $pos $value]
        }
        Log "@@ Parameters for Annotator \"$Annotator\":"
        set param_index 0
        foreach value $Values {
          Log "   Parameter [incr param_index]: \"$value\""
        }
        set time [time {
          set status [tip_InitialiseAnnotator $doc $Annotator $col \
                                              $AnnotatorParameters($Annotator)]
        }]
        Log "Initialising \"$Annotator\": $status ($time)"
      }
    }
  
    ## Execute the Annotators (Components)...
    foreach Annotator $Components {
      catch {
        ggi_wait_message "Processing Document:\n   [tip_GetId $doc]\
                          ($index from $DocumentNumber)" $Progress 1 $wwin
      }
      set Progress [expr {$Progress+$Step}]
      Log "*** Executing Annotator: \"$Annotator\"..."
      ## Put an Attribute showing the Component start time...
      # tip_PutAttribute $doc [tip_CreateAttribute InitTime_$Annotator \
      #    [tip_CreateAttributeValue GDM_STRING \
      #       "[clock format [clock seconds] -format \
      #                      {%H:%M:%S, %A %d %B %Y} -gmt 1] GMT"]]
      set time [time {
        set status [tip_RunAnnotator $doc $Annotator \
                                     $AnnotatorParameters($Annotator)]
      }]
      ## Put an Attribute showing the Component end time...
      # tip_PutAttribute $doc [tip_CreateAttribute EndTime_$Annotator \
      #    [tip_CreateAttributeValue GDM_STRING \
      #       "[clock format [clock seconds] -format \
      #                      {%H:%M:%S, %A %d %B %Y} -gmt 1] GMT"]]
      Log "    Status: $status ($time)"
      ## Handle Errors...
      if {![string equal $status {Annotate Ok}]} {
        ErrorLog "Advancing to the next Document..."
        break; # Do not process any further...
      }
      if {$SyncCollection} {
        Log "  Saving Document..."
        tip_Sync $doc
      }
    };# foreach Annotator $Components
  
    if {$SyncCollection} {
      Log "  Saving Document..."
      catch {tip_Sync $doc}
      catch {tip_Close $doc}
    } else {
      ## The user does not want the Collectio to be saved. Just descard the
      ## Document to free the associated memory...
      set id [tip_GetId $doc]
      # catch {tip_Close $doc}
      tip_RemoveDocument $col $id
    }
    CDM_ClearSavedValues;# Free the CDM internal cache...
    ## We have done processing this file. Advance tot the next one...
  };# foreach filename $Files

  ## Finalise Annotators...
  foreach Annotator $Components {
    set time [time {
      set status [tip_FinishAnnotator {} $Annotator $col \
                                         $AnnotatorParameters($Annotator)]
    }]
    Log "Finalising \"$Annotator\": $status ($time)"
  }
  ggi_destroy_wait $wwin

  ##
  ## Processing Done. Remove the Collection if desired...
  ##
  if {$SyncCollection} {
    Log "Saving Collection..."
    catch {tip_Sync $col}
  }
  tip_Close $col
  if {$SyncCollection && $::DeleteCollection && \
                           [file exists $CollectionName]} {
    Log "Deleting Collection \"$CollectionName\"..."
    file delete -force -- $CollectionName
  }
  Log "*** Processing Finished. End Time=[clock format [clock seconds] \
                                   -format {%H:%M:%S, %A %d %B %Y} -gmt 1] GMT"
};# Execute
};# ::CreateStandAloneApplication::AppExecutionProc

## ::CreateStandAloneApplication::UserProvidedFumctionsTemplate
#  The Template for the User provided Handlers...
proc ::CreateStandAloneApplication::UserProvidedFumctionsTemplate {} {
##############################################################################
##############################################################################
####                                                                      ####
####  User Provided Handlers                                              ####
####                                                                      ####
####  This file must always be in the same directory the the script file  ####
####  or the Init.tcl file of an Application.                             ####
####                                                                      ####
##############################################################################
##############################################################################

## ApplicationArgumentHandler:
## This function can be used in order to handle additional arguments.
## It will be called for each unknown argument encountered: this function
## should return true if it can recognise and handle the argument, false
## otherwise.
## Finally, this function should be also able to handle the special argument
## "{}", where the function is expected to return the list of the additional
## arguments it recognises.
proc ApplicationArgumentHandler {option value} {
  switch -exact -- $option {
    {} {return [list {help {}} ] }
    -help   {
      puts {This is an Ellogon based Application!}
      exit
    }
    default {}
  }
  return false
};# ApplicationArgumentHandler

## DisplayMainWindow_UserHandler:
## This procedure can be used in order to modify the window that will be
## shown to the user. It will be called after the window (stored in the "win"
## parameter) has been created and all default elements have been placed on
## it, but before it is displayed on the screen. This function can add new
## elements to the window or re-arrange/remove elements from it.
proc DisplayMainWindow_UserHandler {win} {
  if {!$::ApplicationInfo(EnableGUI)} {return}
};# DisplayMainWindow_UserHandler

## SpecifyComponentParameters:
##
## This function will be called in order to set a specific parameter
## on a Component (for component parameters that this parameter assignment
## method has been specified).
## The parameters of this function are:
##   1) The Component name (variable "component")
##   2) The position of this parameter in the Component parameter list,
##      starting from 0 (variable "index")
##   3) The format of the returned value (variable "spec")
##      (i.e. "-file %FILE%*")
##   4) The default value for the parameter, as found in the Component
##      configuration file (creole_config.tcl) (variable default)
##
## Note that this function will be called for all Components: the function is
## expected to return different argument lists for different Components and
## indexes.
##
## The return value should be the desired value for the parameter, in the
## format described by the "spec" variable. For example, if the "spec"
## variable contains the value "-d %DIRECTORY%", the return value should be
## "-d <path to directory>.
proc SpecifyComponentParameters {component index spec default} {
  switch -exact -- $component {
    default {
      switch $index {
        default {return [regsub -all {%[A-Z]+%} $spec \
                 [subst -nobackslashes -nocommands $default]]}
      }
    }
  }
  return [regsub -all {%[A-Z]+%} $spec \
                 [subst -nobackslashes -nocommands $default]]
};# SpecifyComponentParameters

## DocumentGetData:
## This function will be called in order tot retrieve the data that will
## placed in a Document. It will be called just before the Document will be
## created in the parent Collection.
##
## The parameter to this function is the url whose contents must be retrieved.
## The url can be either a local file name (like /<path to a file>,
## file://<path to a file>) or an Internet url (like http://<...>, ftp://<...>).
##
## The return value must be the data that should be placed in the Document.
proc DocumentGetData {url} {
  switch -glob -- $url {
    /*       -
    file://* {
      set file [open [string range $url 6 end]]
        set data [read $file]
      close $file
      return $data
    }
    default {
      return "URL \"$url\" not supported!"
    }
  }
  return {}
};# DocumentGetData

## DocumentCreationHandler
## This procedure will be called just after
## a new Document has been added to the Collection. If you want to perform
## something special on a newly created Document, you should modify this
## function.
proc DocumentCreationHandler {col doc} {
};# DocumentCreationHandler
};# ::CreateStandAloneApplication::UserProvidedFumctionsTemplate
