##############################################################################
##  ImportExportSGML.tcl:                                                   ##
## ---------------------                                                    ##
##  SGML Import - Export Utility:                                           ##
##   Converts a Document that contains SGML text defining some text ranges  ##
##   into a new Document that has SGML-stripped text, where the SGML tags   ##
##   have been converted into the corresponding Annotation Set...(Import)   ##
##   Converts a Ellogon Document into an SGML equivelant: it removes ALL    ##
##   Attributes-Annotations and inserts the corresponding SGML tags in      ##
##   Document's text...(Export)                                             ##
##  Initial version date: 17/01/2000                                        ##
##                                                                          ##
##  Copyright 1998-2004 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

## The Tcl_XML package is required...
package require xml

###############################################################################
###############################################################################
##
##           S G M L        I M P O R T      U T I L I T Y
##
###############################################################################
###############################################################################

## GUT_SGML_ConvertCollectionFromSGML
#   Strips the Document text from SGML and adds the corresponding Annotation
#   Set for all Documents in Collection...
proc GUT_SGML_ConvertCollectionFromSGML {{col {}}} {
  global GUI_Text
  if {![string length $col]} {
    global GUI_CollectionDlg GUT_SGML_ConvertionMap
    if {[string equal $GUI_CollectionDlg(Col2Open) {}]} return
    if {![string equal [GUI_IsCollectionOpened $col] {}]} {
      gut_warning $GUI_Text(CollectionIsOpen)
    }
    set col [tip_OpenCollection $GUI_CollectionDlg(Col2Open)]
  }

  ## Let the user modify the Convertion Arrays...
  GUT_SGML_Convert:TagMap $col $GUI_CollectionDlg(Window).sgml_ConvertTagMap \
    Import
  tkwait window $GUI_CollectionDlg(Window).sgml_ConvertTagMap
  ## Check if the user has canceled operation...
  if {![info exists GUT_SGML_ConvertionMap(AnnotationsRemain)]} return

  ## Process the Convertion Arrays...
  array unset GUT_SGML_ConvertionMap Map::*
  foreach tag $GUT_SGML_ConvertionMap(AnnotationsRemain) {
    set GUT_SGML_ConvertionMap(Map::$tag) $tag
  }
  foreach con $GUT_SGML_ConvertionMap(Annotations2Change) {
    set index [string first { -> } $con]
    set tag   [string range $con 0 [expr {$index-1}]]
    set alias [string range $con [expr {$index+4}] end]
    set GUT_SGML_ConvertionMap(Map::$tag) $alias
  }

  ## Itarate over Collection Documents...
  set files [$col GetFileIds]; set file_nu [llength $files]
  if {!$file_nu} {set file_nu 1}
  set progress 0.0
  set step [expr {100.0/$file_nu}]
  catch {ggi_wait_message $GUI_Text(PleaseWait) 0} wwin
  foreach file $files {
    set doc [tip_GetDocument $col $file]
    GUT_SGML_ConvertDocument $doc 0
    tip_Sync  $doc
    tip_Close $doc
    set progress [expr {$progress+$step}]
    catch {ggi_wait_update {} $progress $wwin}
  }
  catch {ggi_destroy_wait $wwin}
};# GUT_SGML_ConvertCollectionFromSGML


## SGML_ConvertDocument
#   Strips the Document text from SGML and adds the corresponding Annotation
#   Set...
proc GUT_SGML_ConvertDocument {doc {warnings 1}} {
  global GUT_SGML_ConvertDocument GUI_Text
  catch {array unset GUT_SGML_ConvertDocument}

  ## Initialise...
  array set GUT_SGML_ConvertDocument [list \
    Text {} Doc $doc ByteOffset 0]

  catch {ggi_wait_append "$GUI_Text(Document): [tip_GetId $doc]" 0} wwin
  set GUT_SGML_ConvertDocument(WaitWin) $wwin
  set Text "<SGML>[tip_GetRawData $doc]</SGML>"
  
  set ExistingAnnotations [tip_GetAnnotations $doc]
  if {[llength $ExistingAnnotations]} {
    if {$warnings} {
      puts "display warning about deleting Annotations..."
    }
    ## Remove ALL Annotations...
    tip_DeleteAnnotations $doc {} 
  }
  
  ## Create a parser object...
  set parser [::xml::parser]
  $parser configure -elementstartcommand  GUT_SGML_ConvertDocument:ElemStart \
                    -elementendcommand    GUT_SGML_ConvertDocument:ElemEnd \
                    -characterdatacommand GUT_SGML_ConvertDocument:ElemText
  $parser parse $Text
  $parser free
  $doc SetRawData $GUT_SGML_ConvertDocument(Text)
  array unset GUT_SGML_ConvertDocument
  catch {ggi_destroy_append $wwin}
};# GUT_SGML_ConvertDocument


###############################################################################
## The following routines are called by the xml parser...
##

## The start of a SGML tag was found. Setup things...
proc GUT_SGML_ConvertDocument:ElemStart {name attrs args} {
  global GUT_SGML_ConvertDocument GUT_SGML_ConvertionMap
  set name [string tolower $name]
  if {![info exists GUT_SGML_ConvertionMap(Map::$name)]} return

  lappend GUT_SGML_ConvertDocument($name) $GUT_SGML_ConvertDocument(ByteOffset)
  lappend GUT_SGML_ConvertDocument($name:Attrs) $attrs
};# GUT_SGML_ConvertDocument:ElemStart
## The end of a SGML tag was found. Add Annotation and clear things...
proc GUT_SGML_ConvertDocument:ElemEnd {name args} {
  global GUT_SGML_ConvertDocument GUT_SGML_ConvertionMap
  set name [string tolower $name]
  if {![info exists GUT_SGML_ConvertionMap(Map::$name)]} return

  set spans [list [tip_CreateSpan $GUT_SGML_ConvertDocument($name) \
                                  $GUT_SGML_ConvertDocument(ByteOffset)]]
  #puts $GUT_SGML_ConvertDocument($name:Attrs)
  set ann [tip_CreateAnnotation $GUT_SGML_ConvertionMap(Map::$name) $spans {}]
  if {![string equal $name SGML]} {
    tip_AddAnnotation $GUT_SGML_ConvertDocument(Doc) $ann
    #catch {ggi_wait_update {} 5 $GUT_SGML_ConvertDocument(WaitWin)}
  }

  ## Clean up this tag...
  set length [llength $GUT_SGML_ConvertDocument($name)]
  if {$length > 1} {
    set GUT_SGML_ConvertDocument($name) [lrange \
      $GUT_SGML_ConvertDocument($name) 0 [expr {$length -2}]]
  } else {
    set GUT_SGML_ConvertDocument($name) {}
  }
  set length [llength $GUT_SGML_ConvertDocument($name:Attrs)]
  if {$length > 1} {
    set GUT_SGML_ConvertDocument($name:Attrs) [lrange \
      $GUT_SGML_ConvertDocument($name:Attrs) 0 [expr {$length -2}]]
  } else {
    set GUT_SGML_ConvertDocument($name:Attrs) {}
  }
};# GUT_SGML_ConvertDocument:ElemEnd
## Processes the text among the Start/End SGML tags...
proc GUT_SGML_ConvertDocument:ElemText {data} {
  global GUT_SGML_ConvertDocument

  append GUT_SGML_ConvertDocument(Text) $data
  incr GUT_SGML_ConvertDocument(ByteOffset) [string length $data]
};# GUT_SGML_ConvertDocument:ElemText

##
## End of Callback routines...
###############################################################################


###############################################################################
###############################################################################
##
##           S G M L        E X P O R T      U T I L I T Y
##
###############################################################################
###############################################################################

## GUT_SGML_ConvertCollectionToSGML
#   Strips the Document text from SGML and adds the corresponding Annotation
#   Set for all Documents in Collection...
proc GUT_SGML_ConvertCollectionToSGML {{col {}}} {
  if {![string length $col]} {
    global GUI_CollectionDlg GUT_SGML_ConvertionMap GUI_Text
    if {[string equal $GUI_CollectionDlg(Col2Open) {}]} return
    if {![string equal [GUI_IsCollectionOpened $col] {}]} {
      gut_warning $GUI_Text(CollectionIsOpen)
    }
    set col [tip_OpenCollection $GUI_CollectionDlg(Col2Open)]
  }

  ## Let the user modify the Convertion Arrays...
  GUT_SGML_Convert:TagMap $col $GUI_CollectionDlg(Window).sgml_ConvertTagMap
  tkwait window $GUI_CollectionDlg(Window).sgml_ConvertTagMap
  ## Check if the user has canceled operation...
  if {![info exists GUT_SGML_ConvertionMap(AnnotationsRemain)]} return

  ## Process the Convertion Arrays...
  array unset GUT_SGML_ConvertionMap Map::*
  foreach tag $GUT_SGML_ConvertionMap(AnnotationsRemain) {
    set GUT_SGML_ConvertionMap(Map::$tag) $tag
  }
  foreach con $GUT_SGML_ConvertionMap(Annotations2Change) {
    set index [string first { -> } $con]
    set tag   [string range $con 0 [expr {$index-1}]]
    set alias [string range $con [expr {$index+4}] end]
    set GUT_SGML_ConvertionMap(Map::$tag) $alias
  }

  ## Itarate over Collection Documents...
  set files [$col GetFileIds]; set file_nu [llength $files]
  if {$file_nu == 0} {set file_nu 1}
  set progress 0.0
  set step [expr {100.0/$file_nu}]
  catch {ggi_wait_message $GUI_Text(PleaseWait) 0} wwin
  foreach file $files {
    set doc [tip_GetDocument $col $file]
    GUT_SGML_ConvertDocumentToSGML $doc
    tip_Sync  $doc
    tip_Close $doc
    set progress [expr {$progress+$step}]
    catch {ggi_wait_update {} $progress $wwin}
  }
  catch {ggi_destroy_wait $wwin}
};# GUT_SGML_ConvertCollectionToSGML

## GUT_SGML_ConvertDocumentToSGML
#  Removes ALL Annotations-Attributes from Document, and converts its text to
#  SGML, with all the information from Document Annotations...
proc GUT_SGML_ConvertDocumentToSGML {doc} {
  global GUT_SGML_ConvertionMap GUT_SGML_ConvertDocumentToSGML
  catch {array unset GUT_SGML_ConvertDocumentToSGML}

  ## Initialise...
  ## Create Document Header...
  set Header \
    "<SGML><DOC id=\"[tip_GetId $doc]\" xid=\"[tip_GetExternalId $doc]\""
  foreach attr [tip_GetAttributes $doc] {
    append Header " [tip_GetName $attr]=\"[tip_GetValueValue $attr]\""
    tip_RemoveAttribute $doc [tip_GetName $attr]
  }
  append Header ">"
  array set GUT_SGML_ConvertDocumentToSGML [list Text $Header Doc $doc]

  catch {ggi_wait_append "$GUI_Text(Document): [tip_GetId $doc]" 0} wwin
  set GUT_SGML_ConvertDocumentToSGML(WaitWin) $wwin

  ## Foreach User-selected Annotation type to be exported, add all the
  ## annotation ranges in a text widget...
  set text .gut_sgml_convertdocumenttosgml
  catch {text $text -setgrid 0}
  set GUT_SGML_ConvertDocumentToSGML(TextWidget) $text
  $text delete 1.0 end
  $text insert end [tip_GetByteSequence $doc]
  set ind 0
  foreach Ann [array names GUT_SGML_ConvertionMap Map::*] {
    set type [string range $Ann 5 end]
    foreach ann [tip_SelectAnnotations $doc $type] {
      foreach {Id Type Spans Attrs} $ann {break}
      foreach span $Spans {
        foreach {start end} $span {break}
        $text tag add Id:$Id 1.0+${start}chars 1.0+${end}chars
      }
      incr ind
      if {$ind > 10} {
        set ind 0
        catch {ggi_wait_update {} 10 $wwin}
      }
    }
  }
  $text dump -all -command GUT_SGML_ConvertDocumentToSGML-DUMP 1.0 end
  destroy $text
  tip_DeleteAnnotations $doc {}
  append GUT_SGML_ConvertDocumentToSGML(Text) "</DOC></SGML>"
  $doc SetRawData $GUT_SGML_ConvertDocumentToSGML(Text)
  array unset GUT_SGML_ConvertDocumentToSGML
  catch {ggi_destroy_wait $wwin}
};# GUT_SGML_ConvertDocumentToSGML

## GUT_SGML_RetrieveSGML-FromDocumentAnnotations
#  Returns information about Document Annotations (according to the user
#  defined mappings) in SGML format...
proc GUT_SGML_RetrieveSGML-FromDocumentAnnotations {doc {Header {}} {Footer {}}} {
  global GUT_SGML_ConvertionMap GUT_SGML_ConvertDocumentToSGML GUI_Text
  catch {array unset GUT_SGML_ConvertDocumentToSGML}

  ## Initialise...
  #set Header \
  #  "<SGML><DOC id=\"[tip_GetId $doc]\" xid=\"[tip_GetExternalId $doc]\">"
  array set GUT_SGML_ConvertDocumentToSGML [list Text $Header Doc $doc]

  catch {ggi_wait_append "$GUI_Text(Document): [tip_GetId $doc]" 0} wwin
  set GUT_SGML_ConvertDocumentToSGML(WaitWin) $wwin

  set Progress 0.0
  set text .gut_sgml_convertdocumenttosgml
  set ind 0
  set Annotations {}
  foreach type [array names GUT_SGML_ConvertionMap Map::*] {
    set Annotations [concat $Annotations \
            [tip_SelectAnnotations $doc [string range $type 5 end]]]
  }
  ## Is Tk available?
  if {0 && ![catch {text $text -setgrid 0}]} {
    set Annotations [tip_SortAnnotationSet $Annotations]
    ## Foreach User-selected Annotation type to be exported, add all the
    ## annotation ranges in a text widget...
    set GUT_SGML_ConvertDocumentToSGML(TextWidget) $text
    $text delete 1.0 end
    $text insert end [tip_GetByteSequence $doc]
    set Step [expr {50.0/([llength $Annotations]+1)}]
    set GUT_SGML_ConvertDocumentToSGML(ProgressStep) $Step
    foreach ann $Annotations {
      foreach {Id Type Spans} $ann {break}
      foreach span $Spans {
        tip_GetSpanOffsets $span start end
        $text tag add Id:$Id 1.0+${start}chars 1.0+${end}chars
      }
      incr ind
      if {$ind > 10} {
        set ind 0
        catch {ggi_wait_update {} $Progress $wwin}
        update
      }
      set Progress [expr {$Progress+$Step}]
    }
    unset -nocomplain Annotations
    set GUT_SGML_ConvertDocumentToSGML(Progress) 50
    $text dump -all -command GUT_SGML_ConvertDocumentToSGML-DUMP 1.0 end
    destroy $text
  } else {
    ## Tk is not available. Use tip_RawDataInsertString to insert the tags in
    ## a copy of the Document...
    set Annotations [tip_SortAnnotationSet $Annotations]
    ## Create a dummy Collection...
    set col  [tip_CreateCollection $::CDM_TempDir/TempCol {}]
    set tdoc [tip_CreateDocument $col TempDoc [tip_GetByteSequence $doc] \
                                 $Annotations {}]
    ## Iterate over all Annotations and select their ids...
    set ids {}
    foreach ann $Annotations {lappend ids [tip_GetId $ann]}
    unset -nocomplain Annotations
    set Step [expr {100.0/([llength $ids]+1)}]
    foreach id $ids {
      set ann      [tip_GetAnnotation $tdoc $id]
      set ann_type [tip_GetType $ann]
      if {[info exists GUT_SGML_ConvertionMap(Map::$ann_type)]} {
        set value $GUT_SGML_ConvertionMap(Map::$ann_type)
        set ann_type $value
      } else {
        set value $ann_type
      }
      append value " id=\"$id\""
      foreach attr [tip_GetAttributes $ann] {
        append value " [tip_GetName $attr]=\"[tip_GetValueValue $attr]\""
      }
      tip_GetSpanOffsets [tip_GetFirstSpan $ann] start end
      tip_ByteSequenceInsertString $tdoc $start "<$value>"
      ## We have just inserted a string before the Annotation. As a result,
      ## the end is invalid...
      set ann [tip_GetAnnotation $tdoc $id]
      tip_GetSpanOffsets [tip_GetFirstSpan $ann] start end
      tip_ByteSequenceInsertString $tdoc $end "</$ann_type>"
      incr ind
      if {$ind > 10} {
        set ind 0
        catch {ggi_wait_update {} $Progress $wwin}
        update
      }
      set Progress [expr {$Progress+$Step}]
    }
    set GUT_SGML_ConvertDocumentToSGML(Text) [tip_GetRawData $tdoc]
    tip_Close $tdoc
    tip_Close $col
  }
  append GUT_SGML_ConvertDocumentToSGML(Text) $Footer
  set SGML-Text $GUT_SGML_ConvertDocumentToSGML(Text)
  array unset GUT_SGML_ConvertDocumentToSGML
  catch {ggi_destroy_append $wwin}
  return ${SGML-Text}
};# GUT_SGML_RetrieveSGML-FromDocumentAnnotations

## GUT_SGML_ConvertDocumentToSGML-DUMP
#    For each tag in the text widget, write the appropriate SGML tag...
proc GUT_SGML_ConvertDocumentToSGML-DUMP {type tag index} {
  global GUT_SGML_ConvertDocumentToSGML GUT_SGML_ConvertionMap

  switch -exact -- $type {
    text   {append GUT_SGML_ConvertDocumentToSGML(Text) $tag}
    tagon  {
      set id [string range $tag 3 end]
      set ann [tip_GetAnnotation $GUT_SGML_ConvertDocumentToSGML(Doc) $id]
      set ann_type [tip_GetType $ann]
      if {[info exists GUT_SGML_ConvertionMap(Map::$ann_type)]} {
        set value $GUT_SGML_ConvertionMap(Map::$ann_type)
      } else {
        set value $ann_type
      }
      append value " id=\"$id\""
      foreach attr [tip_GetAttributes $ann] {
        append value " [tip_GetName $attr]=\"[tip_GetValueValue $attr]\""
      }
      append GUT_SGML_ConvertDocumentToSGML(Text) <$value>
    }
    tagoff {
      set id [string range $tag 3 end]
      set ann [tip_GetAnnotation $GUT_SGML_ConvertDocumentToSGML(Doc) $id]
      set ann_type [tip_GetType $ann]
      if {[info exists GUT_SGML_ConvertionMap(Map::$ann_type)]} {
        set value $GUT_SGML_ConvertionMap(Map::$ann_type)
      } else {
        set value $ann_type
      }
      append GUT_SGML_ConvertDocumentToSGML(Text) </$value>
      set GUT_SGML_ConvertDocumentToSGML(Progress) \
        [expr {$GUT_SGML_ConvertDocumentToSGML(Progress)+\
               $GUT_SGML_ConvertDocumentToSGML(ProgressStep)}]
      catch {ggi_wait_update {} $GUT_SGML_ConvertDocumentToSGML(Progress) \
                                $GUT_SGML_ConvertDocumentToSGML(WaitWin)}
    }
  }
}

###############################################################################
###############################################################################
##  GUT_SGML_Convert:TagMap
## -------------------------
##   Displays a Dialog that allows the user to redefine conversion rules for
##   SGML Import/Export...
###############################################################################
###############################################################################
proc GUT_SGML_Convert:TagMap {{col {}} {win .sgml_ConvertTagMap} \
                              {type Export}} {
  global GUI_Text GUI_Options GUT_SGML_ConvertionMap

  if {[winfo exists $win]} {
    wm deiconify $win
    raise $win
    return
  }

  toplevel $win
  wm withdraw $win
  wm title $win {SGML Tag Mappings...}

  set ann_frame $win.annotations
  frame $win.annotations -relief groove -bd 2
  switch -exact $type {
    Export {
      label $ann_frame.available -textvariable GUI_Text(AvailableAnnotations) \
        -fg navy
      label $ann_frame.dump -textvariable GUI_Text(Annotations2BeSaved) \
        -fg green
      label $ann_frame.not_change -textvariable \
        GUI_Text(SGML_NotChangeAnnotations)
      label $ann_frame.change -textvariable GUI_Text(SGML_ChangeAnnotations)
    }
    Import {
      label $ann_frame.available -textvariable GUI_Text(AvailableSGMLTags) \
        -fg navy
      label $ann_frame.dump -textvariable GUI_Text(SGMLTags2BeSaved) \
        -fg green
      label $ann_frame.not_change -textvariable \
        GUI_Text(SGML_NotChangeSGMLTags)
      label $ann_frame.change -textvariable GUI_Text(SGML_ChangeSGMLTags)
    }
  }

  frame $ann_frame.buttons -bd 0
    ArrowButton $ann_frame.buttons.left  -type button -width 40 -height 25 \
      -fg red   -dir left  -activeforeground red \
      -bd $GUI_Options(BorderWidth) -command "GUT_SGML_Convert:TagMap_Add \
      $ann_frame.l_not_change AnnotationsRemain AvailableAnnotations"
    ArrowButton $ann_frame.buttons.right -type button -width 40 -height 25 \
      -fg green -dir right -activeforeground green \
      -bd $GUI_Options(BorderWidth) -command "GUT_SGML_Convert:TagMap_Add \
      $ann_frame.l_not_change AnnotationsRemain Annotations2Change"
    pack $ann_frame.buttons.left  -fill x -side left -expand 1 -padx 5 -pady 4
    pack $ann_frame.buttons.right -fill x -side right -expand 1 -padx 5 -pady 4
  frame $ann_frame.buttons2 -bd 0
    ArrowButton $ann_frame.buttons2.left  -type button -width 40 -height 25 \
      -fg red   -dir left -activeforeground red \
      -bd $GUI_Options(BorderWidth) -command "GUT_SGML_Convert:TagMap_Add \
      $ann_frame.l_change Annotations2Change AnnotationsRemain"
    pack $ann_frame.buttons2.left  -fill x -side left -expand 1 -padx 5 -pady 4
  
  ArrowButton $ann_frame.right -type button -width 40 -height 25 \
      -fg green -dir right -activeforeground green \
      -bd $GUI_Options(BorderWidth) -command "GUT_SGML_Convert:TagMap_Add \
      $ann_frame.l_available AvailableAnnotations AnnotationsRemain"
 
  frame $ann_frame.add -relief flat -bd 0
    entry $ann_frame.add.entry -bd $GUI_Options(BorderWidth) -validate all\
      -vcmd "GUT_SGML_Convert:TagMap_HandleEntry $ann_frame.add.ok %P"
    ArrowButton $ann_frame.add.ok -dir bottom -bd $GUI_Options(BorderWidth) \
       -command \
       "lappend GUT_SGML_ConvertionMap(AvailableAnnotations) \
        \[string tolower \[$ann_frame.add.entry get\]\]
        $ann_frame.add.ok configure -state disabled" -state disabled
    bind $ann_frame.add.ok <Key-Return> {%W invoke}
    bind $ann_frame.add.entry <Key-Return> "$ann_frame.add.ok invoke"
    pack $ann_frame.add.entry $ann_frame.add.ok -side left -fill y

  ScrolledWidgets::ScrolledListbox $ann_frame.l_available -height 6 \
        -bd $GUI_Options(BorderWidth) -exportselection 0 -listvar \
    GUT_SGML_ConvertionMap(AvailableAnnotations)
  ScrolledWidgets::ScrolledListbox $ann_frame.l_not_change -height 6 \
        -bd $GUI_Options(BorderWidth) -exportselection 0 -listvar \
    GUT_SGML_ConvertionMap(AnnotationsRemain)
  ScrolledWidgets::ScrolledListbox $ann_frame.l_change -height 6 \
        -bd $GUI_Options(BorderWidth) -exportselection 0 -listvar \
    GUT_SGML_ConvertionMap(Annotations2Change)
  grid $ann_frame.available -sticky snew
  grid $ann_frame.add -row 1 -column 0 -sticky snew
  grid $ann_frame.l_available -row 2 -column 0 -rowspan 1 -sticky snew
  grid $ann_frame.right -row 3 -column 0
  grid $ann_frame.dump -row 0 -column 1 -columnspan 2 -sticky snew
  grid $ann_frame.not_change -row 1 -column 1 -sticky snew
  grid $ann_frame.l_not_change -row 2 -column 1 -sticky snew
  grid $ann_frame.buttons -row 3 -column 1 -padx 30
  grid $ann_frame.change -row 1 -column 2 -sticky snew
  grid $ann_frame.l_change -row 2 -column 2 -sticky snew
  grid $ann_frame.buttons2 -row 3 -column 2 -padx 30
  grid columnconfigure $ann_frame 0 -weight 1
  grid columnconfigure $ann_frame 1 -weight 1
  grid columnconfigure $ann_frame 2 -weight 1
  grid rowconfigure    $ann_frame 1 -weight 1

  pack $ann_frame -fill x -expand 1 -padx 4 -pady 4

  frame $win.okcancel
    ImageButton::Button $win.okcancel.ok -image ITik -textvariable GUI_Text(Ok)\
      -command "destroy $win"
    ImageButton::Button $win.okcancel.cancel -image ITCancel -textvariable \
      GUI_Text(Cancel) -command "destroy $win;unset GUT_SGML_ConvertionMap"
    pack $win.okcancel.ok $win.okcancel.cancel -pady 20 -padx 10 -side left
  pack $win.okcancel

  ## Iterate over Collection Documents to collect all Annotations-Attributes
  if {[string match Col* $col]} {
    catch {ggi_wait_message $GUI_Text(PleaseWait) 0} wwin
    set files [$col GetFileIds]; set file_nu [llength $files]
    if {$file_nu < 1} {set file_nu 1}
    set progress 0.0; set step [expr {100.0/$file_nu}]
    foreach file $files {
      set doc [tip_GetDocument $col $file]
      foreach ann [$doc GetAnnotationTypes] {
        CDM_LappendUnique GUT_SGML_ConvertionMap(AvailableAnnotations) $ann
      }
      tip_Close $doc
      set progress [expr {$progress+$step}]
      catch {ggi_wait_update {} $progress $wwin}
    }
    catch {ggi_destroy_wait $wwin}
  } elseif {[string match Doc* $col]} {
    foreach ann [$col GetAnnotationTypes] {
      CDM_LappendUnique GUT_SGML_ConvertionMap(AvailableAnnotations) $ann
    }
  } else {
    lappend GUT_SGML_ConvertionMap(AvailableAnnotations) token sentence
  }
  switch -exact $type {
    Export {
      CDM_LappendUnique GUT_SGML_ConvertionMap(AvailableAnnotations) \
      token sentence lookup syntax ne
    }
    Import {
      CDM_LappendUnique GUT_SGML_ConvertionMap(AvailableAnnotations) \
        doc s sentence token lookup ne syntax
    }
  }
  set GUT_SGML_ConvertionMap(AvailableAnnotations) \
    [lsort -unique $GUT_SGML_ConvertionMap(AvailableAnnotations)]
  wm deiconify $win
}

## GUT_SGML_Convert:TagMap_Add 
#   
proc GUT_SGML_Convert:TagMap_Add {source_listbox source_item target_item} {
  global GUT_SGML_ConvertionMap

  set index [lindex [$source_listbox.listbox curselection] 0]
  if {$index < 0} return
  set selection [lindex $GUT_SGML_ConvertionMap($source_item) $index]
  if {[string equal $target_item Annotations2Change]} {
    ## Get the new alias name...
    set alias [string trim [gut_EntryDialog -title {Enter Alias...} \
       -message {Enter new name} -initial $selection]]
    if {![string length $alias]} return
    set selection "$selection -> $alias"
  }
  set GUT_SGML_ConvertionMap($source_item) \
      [lreplace $GUT_SGML_ConvertionMap($source_item) $index $index]
  set target_list $GUT_SGML_ConvertionMap($target_item)
  if {[string equal $source_item Annotations2Change]} {
    set selection [string range $selection 0 \
                   [expr {[string first { -> } $selection] - 1}]]
  }
  lappend target_list $selection
  set GUT_SGML_ConvertionMap($target_item) [lsort -unique $target_list]
  $source_listbox.listbox selection clear 0 end
  $source_listbox.listbox selection set 0
  $source_listbox.listbox see 0
};# GUT_SGML_Convert:TagMap_Add

## GUT_SGML_Convert:TagMap_HandleEntry
proc GUT_SGML_Convert:TagMap_HandleEntry {button toadd args} {
  global GUT_SGML_ConvertionMap

  set toadd [string tolower $toadd]
  if {[string length $toadd] && [lsearch -exact \
        $GUT_SGML_ConvertionMap(AvailableAnnotations) $toadd] < 0} {
    $button configure -state normal
  } else {
    $button configure -state disabled
  }
  return 1
};# GUT_SGML_Convert:TagMap_HandleEntry
