##############################################################################
##  LoadPlugins.tcl:                                                        ##
## ---------------------                                                    ##
##  This file implements the Component loading subsystem.                   ##
##  Initial version date: 21/09/1998                                        ##
##                                                                          ##
##  Copyright 1998-2007 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

## PUL_LoadPlugins
#  Foreach Dir $PluginModulePath, load all plugins found
proc PUL_LoadPlugins {} {
  global PluginModulePath GUI_Text
  foreach Dir $PluginModulePath {
    set status [PUL_LoadPluginsInDir $Dir]
    if {[string length $status]} {lappend error $status}
  }
  if {[info exists error]} {
    if {![info exists GUI_Text(LoadingModulesErrors)]} {
      set GUI_Text(LoadingModulesErrors) \
        {The Following Errors Occured while Loading\n the Plug-in Modules...}
      set GUI_Text(Ok) Ok
    }
    after idle "gut_many_errors {Loading Modules Error...} \
       \$GUI_Text(LoadingModulesErrors) [list $error]"
  }
};# PUL_LoadPlugins

## PUL_LoadPluginsInDir
#  Loads all plugins found in Dir
proc PUL_LoadPluginsInDir {Dir} {
  if {![string length $Dir]} {return}
  foreach {error done_list} {{} {}} {break}
  set configs [glob -nocomplain [file join $Dir * creole_config.tcl]]
  # if not done already, load it
  foreach config $configs {
    set a_name [file tail [file dirname $config]]
    if {[lsearch $done_list $a_name] == -1} {
      set status [PUL_LoadPlugin [list [file dirname $config]]]
      if {[string length $status] > 2} {lappend error $status}
      lappend done_list $a_name
    }
  }
  return $error
};# PUL_LoadPluginsInDir

## PUL_LoadPlugin
#  Loads a plugin module
proc PUL_LoadPlugin {PathList} {
  global creole_config PUL_LoadPlugin SplashPercent HomeDir UserHomeDir \
         tcl_platform
  set current_encoding [encoding system]

  if {![llength $PathList]} {
    set PathList $PUL_LoadPlugin(Last\ Loaded\ Module)
  } else {
    set PUL_LoadPlugin(Last\ Loaded\ Module) $PathList
  }
  ## Function now can accept more than one path! (Petasis, 28/1/1998)
  foreach path $PathList {
    # the module name
    set module_name [file tail $path]
    if {[info exists creole_config($module_name)]} {
      set FirstTime 0
      puts -nonewline "Reloading $path"
    } else {
      set FirstTime 1
      puts -nonewline "Loading $path"
    }
    if {[string equal $tcl_platform(platform) windows]} {
      if {[string first { } $path] != -1} {
        set Path [file attributes $path -shortname]
        ## Ensure that the shortname is correct...
        if {![file exists $Path] || ![file isdirectory $Path]} {
          gut_info "Wrong shortname path: \"$Path\""
          set Path $path
        }
      } else {
        set Path $path
      }
    } else {
      set Path $path
    }
    # set up a global var called after the module whose value is its path
    global creole_${module_name}_home
    set creole_${module_name}_home $Path
    # Update Status Bar
    catch {
      if {$SplashPercent < 90} {incr SplashPercent}
      LDM_SetSplashWindowStatus "Loading Plug-in Module $module_name..." \
           $SplashPercent
    }
    if {[catch {source $path/creole_config.tcl} error]} {
      encoding system $current_encoding
      puts $error
      return "Can't add annotator at $path (can't load creole_config file):\n\
              $error"
    } else {
      set PUL_LoadPlugin(m::$module_name) $Path
    }
    ## Check for encoding entry... (Petasis, 1/2/1999)
    if {[catch {PUL_plugin_module_config $module_name module_encoding} \
          module_encoding]} {continue}
#   set module_encoding [PUL_plugin_module_config $module_name module_encoding]
    if {[string length $module_encoding]} {
       encoding system $module_encoding
       source $path/creole_config.tcl
    }
    
    # the coupling type
    if {[catch {PUL_plugin_module_config $module_name coupling} coupling]} {
      puts " \[$coupling\]"
      encoding system $current_encoding
      return $coupling
    }
    puts -nonewline " \[$coupling\] "
    puts "\([encoding system]\)"
    set PUL_LoadPlugin(c::$module_name) $coupling

    ## Check for an Image...
    catch {PUL_plugin_module_config $module_name image} image
    if {[string length $image]} {
      catch {
        image create photo ModImg_$module_name -file [file join $path $image]
      }
    }

    ## Load modules according to the coupling type...
    switch -exact -- $coupling {
      loose -
      tcl {
        if {[catch {source $path/$module_name.tcl}]} {
          ## We failed to load a .tcl file. Maybe a .tbc file is available?
          if {[file exists [file join $path $module_name.tbc]]} {
            if {[catch {source [file join $path $module_name.tbc]}]} {
              global errorInfo; set error $errorInfo
              catch {LDM_DeleteSplashWindow .splashscr}
              gut_tk_dialog Error $error Ok
              set result \
               "Can't add annotator at $path (can't load $module_name.tbc file)"
              append result "\nError was:\n$error\n\n"
              catch {encoding system $current_encoding}
              return $result
            }
          } else {
            global errorInfo; set error $errorInfo
            catch {LDM_DeleteSplashWindow .splashscr}
            gut_tk_dialog Error $error Ok
            set result \
             "Can't add annotator at $path (can't load $module_name.tcl file)"
            append result "\nError was:\n$error\n\n"
            catch {encoding system $current_encoding}
            return $result
          }
        }
      }
      dynamic -
      c++ {
        set libname [file join $path lib${module_name}[info sharedlibextension]]
        set pkgname Creole_$module_name
        if {!$FirstTime} {
          if {[catch {unload $libname $pkgname} error]} {
            puts " ** Warning: Could not Reload this Dynamic Component!"
            puts " ($error)\n"
          }
        }
        # Load the shared object...
        # puts "load $libname $pkgname"
        set creole_${module_name}_home $Path
        if {[catch {load $libname $pkgname} error]} {
          puts "Dynamic Load Failed for Module \"$module_name\": ignoring it..."
          puts "  ($error)"
          unset creole_config($module_name)
        }
      }
      java {
        set classpath [set creole_${module_name}_home]
	switch $::tcl_platform(platform) {
	  windows {
	    ## TclBlend seems to want path in Unix notation...
	    set classpath [string map {\\ /} "$classpath"]
	  }
	  default {
	  }
	}
        if {[catch {::CDM_Java::LoadJavaComponent $classpath $module_name} \
                                                                       error]} {
          puts "Load Failed for Java Module \"$module_name\": ignoring it..."
          puts "  ($error)"
          unset creole_config($module_name)
        }
      }
      perl {
        if {[catch {::CDM_Perl::LoadPerlComponent $path $module_name} error]} {
          puts "Loading Failed for Perl Module \"$module_name\": ignoring it..."
          puts "  ($error)"
          unset creole_config($module_name)
        }
      }
      oo-python -
      python {
        if {[catch {::CDM_Python::LoadPythonComponent \
                                                   $path $module_name} error]} {
          puts "Loading Failed for Python Module \"$module_name\":\
                ignoring it..."
          puts "  ($error)"
          unset creole_config($module_name)
        }
      }
      static -
      tight {
        if {$FirstTime} {
          set pkgname "Creole_$module_name"
          if {[catch {load {} $pkgname}]} {
            puts "Static Load Failed for Module \"$module_name\":\
                  ignoring it..."
            unset creole_config($module_name)
          }
          set creole_${module_name}_home $Path
        } else {
          puts " ** Warning: Could not Reload a Static Module!"
        }
      }
    };# coupling-specific code
    #set plugin_${module_name}_home $Path
  }
  encoding system $current_encoding
};# PUL_LoadPlugin

## PUL_plugin_module_config_list
#    Returns what PUL_plugin_module_config returns, but formatted as a proper
#    TCL List...
proc PUL_plugin_module_config_list {module {data  {}} {type {}}} {
  set result [PUL_plugin_module_config $module $data $type]
  regsub -all -- {[\n\t]}  $result { } result
  set Result {}
  eval lappend Result $result
  return $Result
};# PUL_plugin_module_config_list


## PUL_plugin_module_config
# required parameters are a module(/annotator) name + the name of an
# entry type from the config table.
# optional parameter is a subtype of the requested entry, currently only
# relevant to the pre- and post-conditions entries.
# returns the value of the specified entry from the config table.
proc PUL_plugin_module_config {module {data {}} {type {}}} {
    global creole_config GUI_Text
    if {![string length $data]} {return {}}

    # it is a fatal error if the module can't be found
    if {[catch {lsearch -glob $creole_config($module) "$data*"}]} {
      if {![info exists GUI_Text(LoadedModuleNotFound)]} {
        set GUI_Text(LoadedModuleNotFound) {Plug-in module "%s" was not found!}
      }
      error [format $GUI_Text(LoadedModuleNotFound) $module] {}
    }

    set pos [lsearch -glob $creole_config($module) "$data*"]
    if {$pos == -1} {
        return {}
    } else {
        return [PUL_plugin_module_conditions \
                [lindex $creole_config($module) [incr pos]] $type]
    }
} ;# PUL_plugin_module_config

# PUL_plugin_module_conditions
# internal function to search the pre- or post-conditions entry for a
# particular type of condition (using glob matching).
# returns just the condition, without its type.
proc PUL_plugin_module_conditions {conditions type} {
  # if no type specified, return list of all conditions
  if {![string length $type]} {
    return $conditions
  } else { # otherwise return list of conditions of specified type
    set pos [lsearch -glob $conditions "$type*"]
    if {$pos == -1} {
      return {}
    } else {
      return [lindex $conditions [incr pos]]
    }
  }
}

proc PUL_ToolConfig {tool data {class {}} {type {}}} {
  global GUI_ToolConfig

  # it is a fatal error if the module can't be found
  if {[catch {lsearch -glob $GUI_ToolConfig(${class}::$tool) "$data*"}]} {
    gut_error "Thought we had loaded module $tool, but not found!"
  }

  set pos [lsearch -glob $GUI_ToolConfig(${class}::$tool) "$data*"]
  if {$pos == -1} {
    return {}
  } else {
    return [PUL_plugin_module_conditions \
                [lindex $GUI_ToolConfig(${class}::$tool) [incr pos]] $type]
  }
} ;# PUL_ToolConfig
  
# Now, load the Plugin Modules...
PUL_LoadPlugins
