##############################################################################
##  SystemManagement.tcl:                                                   ##
## ---------------------                                                    ##
##  This file implements utilities for managing system specifications...    ##
##                                                                          ##
##  Copyright 1998-2004 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

if {![namespace exists ::SystemManagement]} {
  ## We use the tree and graph structures...
  package require struct
  namespace eval ::SystemManagement {
    variable ToplevelWindow
    variable SystemTree
  };# namespace ::SystemManagement
  
  namespace eval ::SystemManagement::Graph {
  };# namespace ::SystemManagement::Graph
}

## ::SystemManagement::DrawSystemGraph
#  This function draws a System graph on a canvas widget...
#  This function assumes that the following keys are defined in each graph
#  node/arc:
#  Nodes:
#    Type:         One of module, button
#    State:        The state of the node. One of 0, 1, 2, 3
#    Label:        The label of the node
#    Tooltip:      The Tooltip of the node
#    X:            The x coordinate of the node
#    Y:            The y coordinate of the node
#    Width:        The width of the node
#    Height:       The height of the node
#    BorderWidth:  The border width of the node
#    BorderColor:  The border color of the node
#    LabelWidth:   The label maximum width in characters
#    LabelPadX:    The padx of the label in pixels
#    LabelPadY:    The pady of the label in pixels
#    LabelJustify: The label justification
#    ImagePadX:    The padx of the image in pixels
#    ImagePadY:    The pady of the image in pixels
#  Arcs:
#    Source:       The arc source node
#    Target:       The arc target node
#    Points:       All the points coordinates
#    Color:        The arc color
#    ActiveColor:  The arc active color
#    Width:        The arc width in pixels
#    ActiveWidth:  The arc active width in pixels
#    Smooth:       The arc smooth option (boolean)
proc ::SystemManagement::DrawSystemGraph {graph canvas} {
  global GUI_Options
  #set graph ::SystemManagement::Graph::$graph
  array set STATE [list \
    FILL_0 $GUI_Options(ModuleInactive) FILL_1 $GUI_Options(ModuleEnabled) \
    FILL_2 $GUI_Options(ModuleActive)   FILL_3 $GUI_Options(ModuleRun) \
    Text_0 $GUI_Options(ModuleTextInactive) \
    Text_1 $GUI_Options(ModuleTextEnabled) \
    Text_2 $GUI_Options(ModuleTextActive) Text_3 $GUI_Options(ModuleTextRun) \
    Im_0 IModuleInactive Im_1 IModuleEnabled Im_2 IModuleActive Im_3 IModuleRun]
  $canvas delete all
  ## Draw the graph nodes...
  foreach node [$graph nodes] {
    set tags    [list $node ${node}_all]
    set type    [$graph node get $node -key Type]
    set state   [$graph node get $node -key State]
    set label   [$graph node get $node -key Label]
    set tooltip [$graph node get $node -key Tooltip]
    set x       [$graph node get $node -key X]
    set y       [$graph node get $node -key Y]
    set w       [$graph node get $node -key Width]
    set h       [$graph node get $node -key Height]
    set bd      [$graph node get $node -key BorderWidth]
    set bdColor [$graph node get $node -key BorderColor]
    set lwidth  [$graph node get $node -key LabelWidth]
    set lpadx   [$graph node get $node -key LabelPadX]
    set lpady   [$graph node get $node -key LabelPadY]
    set ljust   [$graph node get $node -key LabelJustify]
    set ipadx   [$graph node get $node -key ImagePadX]
    set ipady   [$graph node get $node -key ImagePadY]
    GUI_DrawModule $canvas $node $GUI_Options(ModuleShape) $type $state $x $y \
      $w $h $bd $STATE(FILL_$state) $bdColor $tags \
      $label $lwidth $lpadx $lpady  $STATE(Text_$state) \
      ModuleTextFont $ljust [list ${node}_text ${node}_all] $STATE(Im_$state) \
      $ipadx $ipady [list ${node}_img ${node}_all]
    ## Add bindings...
    $canvas bind ${node}_all <2> "GUI_SystemModulePopUp $canvas $node"
    if {[string equal $type button]} {
      $canvas._button$node configure -command \
        "GUI_HandleModule $canvas $node [list $label]"
      bind $canvas._button$node <<Motion_1>> \
        "GUI_DragModule $canvas $graph $node $type %x %y"
      if {[string length $tooltip]} {
        Tooltip::tooltip $canvas._button$node -text $tooltip
      }
    } else {
      $canvas bind ${node}_all <<MBR_1>> \
        "GUI_HandleModule $canvas $node [list $label]"
      $canvas bind ${node}_all <<MB_2>> \
        "GUI_ResetModule $canvas $node [list $label]"
      $canvas bind ${node}_all <<MB_M>> \
        "GUI_ResetModule $canvas $node [list $label]"
      $canvas bind ${node}_all <<Motion_1>> \
        "GUI_DragModule $canvas $graph $node $type %x %y"
      if {[string length $tooltip]} {
        Tooltip::canvas_tooltip $canvas ${node}_all $tooltip
      }
    }
  }
  ## Draw the graph arcs...
  foreach arc [$graph arcs] {
    set source [$graph arc get $arc -key Source]
    set target [$graph arc get $arc -key Target]
    set points [$graph arc get $arc -key Points]
    set color  [$graph arc get $arc -key Color]
    set acolor [$graph arc get $arc -key ActiveColor]
    set width  [$graph arc get $arc -key Width]
    set awidth [$graph arc get $arc -key ActiveWidth]
    set smooth [$graph arc get $arc -key Smooth]
    $canvas lower [GUI_DrawArc $canvas $points \
       $color $acolor $width $awidth $smooth \
       [list From_$source To_$target Line_$source Line_$target]]
  }
};# ::SystemManagement::DrawSystemGraph

## ::SystemManagement::SetDefaultSystemGraphKeys
#  This function will initialise the needed node and arc keys to their default
#  values, if they haven't already been initialised...
proc ::SystemManagement::SetDefaultSystemGraphKeys {graph} {
  foreach node [$graph nodes] {
    SetDefaultNodeKeys $graph $node
  }
  foreach arc [$graph arcs] {
    SetDefaultArcKeys $graph $arc
    CalculateArcPoints $graph $arc
  }
};# ::SystemManagement::SetDefaultSystemGraphKeys

## ::SystemManagement::SetDefaultNodeKeys
#  This function will check whether the essential keys exists in a System
#  graph node. If a key does not exist, it will be initialised to its default
#  value...
proc ::SystemManagement::SetDefaultNodeKeys {graph node} {
  global GUI_Options
  foreach {key value} [list Type module State 0 Label {} Tooltip {} X 0 Y 0 \
                            Width $GUI_Options(ModuleWidth) \
                            Height $GUI_Options(ModuleHeight) \
                            BorderWidth $GUI_Options(ModuleBorderWidth) \
                            BorderColor $GUI_Options(ModuleBorderColor) \
                            LabelWidth $GUI_Options(ModuleTextWidth) \
                            LabelPadX $GUI_Options(ModuleTextPadX) \
                            LabelPadY $GUI_Options(ModuleTextPadY) \
                            LabelJustify $GUI_Options(ModuleTextJustify) \
                            ImagePadX $GUI_Options(ModuleImagePadX) \
                            ImagePadY $GUI_Options(ModuleImagePadY)] {
    if {[catch {$graph node get $node -key $key}]} {
      $graph node set $node -key $key $value
    }
  }
};# ::SystemManagement::SetDefaultNodeKeys

## ::SystemManagement::SetDefaultArcKeys
#  This function will check whether the essential keys exists in a System
#  graph arc. If a key does not exist, it will be initialised to its default
#  value...
proc ::SystemManagement::SetDefaultArcKeys {graph arc} {
  global GUI_Options
  foreach {key value} [list Points {0 0 0 0} Color  $GUI_Options(ArcColor) \
                            ActiveColor $GUI_Options(ArcActiveColor) \
                            Width  $GUI_Options(ArcWidth) \
                            ActiveWidth $GUI_Options(ArcActiveWidth) \
                            Smooth $GUI_Options(ArcSmooth)] {
    if {[catch {$graph arc get $arc -key $key}]} {
      $graph arc set $arc -key $key $value
    }
  }
  if {[catch {$graph arc get $arc -key Source}]} {
    $graph arc set $arc -key Source [$graph arc source $arc]
  }
  if {[catch {$graph arc get $arc -key Target}]} {
    $graph arc set $arc -key Target [$graph arc target $arc]
  }
};# ::SystemManagement::SetDefaultArcKeys

## ::SystemManagement::CalculateArcPoints
#  This function will calculate the arc points based on the location of the
#  source/target nodes...
proc ::SystemManagement::CalculateArcPoints {graph arc} {
  set source  [$graph arc get $arc -key Source]
  set target  [$graph arc get $arc -key Target]
  set cpoints [$graph arc get $arc -key CurvePoints]
  ## Source node...
  set x [$graph node get $source -key X]
  set y [$graph node get $source -key Y]
  set w [$graph node get $source -key Width]
  set h [$graph node get $source -key Height]
  set sx1 [expr {int($x-$w/2)}]
  set sy1 [expr {int($y-$h/2)}]
  set sx2 [expr {int($x+$w/2)}]
  set sy2 [expr {int($y+$h/2)}]
  ## Target node...
  set x [$graph node get $target -key X]
  set y [$graph node get $target -key Y]
  set w [$graph node get $target -key Width]
  set h [$graph node get $target -key Height]
  set dx1 [expr {int($x-$w/2)}]
  set dy1 [expr {int($y-$h/2)}]
  set dx2 [expr {int($x+$w/2)}]
  set dy2 [expr {int($y+$h/2)}]
  ## Find modules centers...
  set scx [expr {($sx2 - $sx1)/2 + $sx1}]
  set scy [expr {($sy2 - $sy1)/2 + $sy1}]
  set dcx [expr {($dx2 - $dx1)/2 + $dx1}]
  set dcy [expr {($dy2 - $dy1)/2 + $dy1}]
  $graph arc set $arc -key Points [GUI_SpecifyLine [list \
    $scx $scy $cpoints $dcx $dcy] $sx1 $sy1 $sx2 $sy2 $dx1 $dy1 $dx2 $dy2]
};# ::SystemManagement::CalculateArcPoints

## ::SystemManagement::CalculateArcCurvePoints
#  This function will examine the source/target node locations and calculate
#  some "proper" curve points, if required...
proc ::SystemManagement::CalculateArcCurvePoints {graph arc} {
  set source [$graph arc get $arc -key Source]
  set target [$graph arc get $arc -key Target]
  set sx [$graph node get $source -key X]
  set tx [$graph node get $target -key X]
  if {$sx == $tx} {
    ## An vertical line
    $graph arc set $arc -key CurvePoints {}
    return
  }
  set sy [$graph node get $source -key Y]
  set ty [$graph node get $target -key Y]
  if {$sy == $ty} {
    ## An horizontal line
    $graph arc set $arc -key CurvePoints {}
    return
  }
  set slope [expr {double($ty-$sy)/double($tx-$sx)}]
  if {$slope == 1.0} {
    ## A line at 45 degrees...
    $graph arc set $arc -key CurvePoints {}
  } elseif {$slope < -1.0} {
    #puts "$arc: $source -> $target $slope $sx,$sy $tx,$sy $tx,$ty"
    $graph arc set $arc -key CurvePoints [list [list $tx $sy]]
  } elseif {$slope >  1.0} {
    #puts "$arc: $source -> $target $slope $sx,$sy $sx,$ty $tx,$ty"
    $graph arc set $arc -key CurvePoints [list [list [expr $sx+1] [expr $ty-1]]]
  } else {
    $graph arc set $arc -key CurvePoints {}
  }
};# ::SystemManagement::CalculateArcCurvePoints

## ::SystemManagement::LayoutGraph
#  This function will change the placement of all nodes and arcs in the
#  provided graph...
proc ::SystemManagement::LayoutGraph {graph} {
  set nodes [$graph nodes]
  ## Find the "initial" nodes, i.e. the nodes that have no ingoing arcs...
  foreach node $nodes {
    set NodeDegreeIn($node)  [$graph node degree -in $node]
    set NodeDegreeOut($node) [$graph node degree -out $node]
    if {!$NodeDegreeIn($node)} {
      lappend InitialNodesDegree($NodeDegreeOut($node)) $node
    }
  }
  set InitalNodes {}
  foreach degree [lsort -decreasing -integer [array names InitialNodesDegree]] {
    set InitalNodes [concat $InitalNodes $InitialNodesDegree($degree)]
  }
  $graph set -key _y 0
  ## Do a depth-first walk starting from the Initial nodes...
  foreach node $InitalNodes {
    $graph set -key _placedNodes 0
    $graph set -key _x 0
    $graph walk $node -order pre -type dfs -dir forward -command \
      "::SystemManagement::LayoutGraphCallback"
  }
  $graph unset -key _x
  $graph unset -key _y
  $graph unset -key _placedNodes
  
  ## Re-position all the arcs of the graph...
  foreach arc [$graph arcs] {
    CalculateArcCurvePoints $graph $arc
    CalculateArcPoints $graph $arc
  }
};# ::SystemManagement::LayoutGraph


proc ::SystemManagement::LayoutGraphCallback {mode graph node} {
  ## Get the currect x,y coordinates...
  set x [$graph get -key _x]
  set y [$graph get -key _y]
  $graph node set $node -key X $x
  $graph node set $node -key Y $y
  set nodes [$graph get -key _placedNodes]
  $graph set -key _placedNodes [incr nodes]
  ## Is this the end of the chain?
  if {$nodes < 5 && [$graph node degree -out $node]} {
    ## There are other nodes following...
    set width [$graph node get $node -key Width]
    $graph set -key _x [expr {$x + $width +$width/2.0}]
  } elseif {$nodes >= 5 && [$graph node degree -out $node]} {
    ## We have excided the maximum number of nodes in a chain...
    set height [$graph node get $node -key Height]
    $graph set -key _y [expr {$y + $height +$height/2.0}]
    $graph set -key _x 0
    $graph set -key _placedNodes 0
  } else {
    ## This is the end of the chain...
    set height [$graph node get $node -key Height]
    $graph set -key _y [expr {$y + $height +$height/2.0}]
    set width [$graph node get $node -key Width]
    $graph set -key _x [expr {$width +$width/2.0}]
    $graph set -key _placedNodes 0
  }
};# ::SystemManagement::LayoutGraphCallback
