##############################################################################
##  SystemStorage.tcl:                                                      ##
## ---------------------                                                    ##
##  This file implements utilities for reading/saving system specifications ##
##  in the configuration file...                                            ##
##                                                                          ##
##  Copyright 1998-2004 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

if {![namespace exists ::SystemManagement]} {
  ## We use the tree and graph structures...
  package require struct
  namespace eval ::SystemManagement {
    variable ToplevelWindow
    variable SystemTree
  };# namespace ::SystemManagement
  
  namespace eval ::SystemManagement::Graph {
  };# namespace ::SystemManagement::Graph
}

## ::SystemManagement::ReadConfigurationFile
#  Read the System's Information stored in the Configuration File...
proc ::SystemManagement::ReadConfigurationFile {File {destroy 1}} {
  variable ToplevelWindow
  variable SystemTree
  ## Remove any previously created tree...
  if {$destroy} {
    ## Destroy any previously created System Tree...
    catch {$SystemTree destroy}
    ## Remove any previously created System graphs...
    foreach graph [info commands ::SystemManagement::Graph::*] {
      $graph destroy
    }
    ## Create a new System Tree...
    set SystemTree [::struct::tree ::SystemManagement::SystemTree]
  }
  
  set file [open $File]
  ## Read the information regarding the top level window
  while {[gets $file line] != -1} {
    switch -glob -- [string trim $line] {
      __top_geometry=* {
        set ToplevelWindow(Geometry) [string range $line 15 end]
      }
      __top_pane_percent=* {
        set ToplevelWindow(PanePercent) [string range $line 19 end]
      }
      __start_graphs {break}
      default {}
    }
  }
  ## Read the System Names...
  while {[gets $file line] != -1} {
    switch -glob -- [string trim $line] {
      __end_graphs {}
      __END_TASKS  {break}
      default {
        foreach {id name alias children} [split $line \;] {break}
        ## Create a "tree" name. This name should not contain ' ' or ':'...
        set nodeName [string map "\\  \x01 : \x02" $name]
        ## Add the item as a node in the tree...
        $SystemTree insert root end $nodeName
        $SystemTree set $nodeName -key Name  $name
        $SystemTree set $nodeName -key Alias $alias
        if {[string length $children]} {
          ## This is a folder, not a system. Add the node into the tree...
          $SystemTree set $nodeName -key Type  Folder
          set Children {}
          foreach item [split $children =] {
            foreach {cAlias cName} [split $item |] {break}
            set cnodeName [string map "\\  \x01 : \x02" $cName]
            $SystemTree insert $nodeName end $cnodeName
            $SystemTree set $cnodeName -key Type  System
            $SystemTree set $cnodeName -key Name  $cName
            $SystemTree set $cnodeName -key Alias $cAlias
          }
        } else {
          ## This is a system (leaf in the tree)...
          $SystemTree set $nodeName -key Type System
        }
      }
    }
  }
  ## We have added all visible systems/folders in our tree.
  ## Its time to read the graphs...
  if {[catch {ReadSystemGraphFromConfigurationFile $file} graph]} {
    puts "Error processing graph: $graph"
  }
  while {[string length $graph]} {
    if {[catch {ReadSystemGraphFromConfigurationFile $file} graph]} {
      puts "Error processing graph: $graph"
    }
  }
  close $file
};# ::SystemManagement::ReadConfigurationFile

## ::SystemManagement::ReadSystemGraphFromConfigurationFile
#  Reads a system graph from the provided configuration file descriptor.
#  The system is represented by a newly created graph structure, which is
#  returned as the return value of this function...
proc ::SystemManagement::ReadSystemGraphFromConfigurationFile {file} {
  while {[gets $file line] != -1} {
    switch -glob -- [string trim $line] {
      __start_nodes {
        ## Read untill "__end_nodes" is found...
        while {[gets $file line] != -1} {
          set line [string trim $line]
          if {[string equal $line __end_nodes]} {break}
          foreach {name x y _ label} [split $line \;] {break}
          set name [lindex [split $name ,] 1]
          # puts "  Inserting Node: \"$name\""
          $graph node insert $name
          $graph node set $name -key X $x
          $graph node set $name -key Y $y
          $graph node set $name -key Label $label
        }
      }
      __start_arcs {
        ## Read untill "__end_arcs" is found...
        while {[gets $file line] != -1} {
          set line [string trim $line]
          if {[string equal $line __end_arcs]} {break}
          foreach {_ id _ source _ target cpoints} [split $line \;,] {break}
          # puts "  Inserting Arc: $source -> $target"
          set arc [$graph arc insert $source $target]
          $graph arc set $arc -key CurvePoints $cpoints
          $graph arc set $arc -key Source $source
          $graph arc set $arc -key Target $target
        }
      }
      __START_GRAPH=* {
        set name [string range $line 14 end]
        catch {::SystemManagement::Graph::$name destroy}
        set graph [::struct::graph ::SystemManagement::Graph::$name]
        $graph set -key Name $name
        #puts "*** Graph Found: \"$name\""
      }
      __geometry=* {
        catch {$graph set -key Geometry [string range $line 11 end]}
      }
      __timeLog=* {
        catch {$graph set -key TimeLog [string range $line 10 end]}
      }
      __grid=* {
        catch {$graph set -key Grid [string range $line 7 end]}
      }
      __END_GRAPH=* {
        ## We have finished!
	return $graph
      }
      __END_CONFIG {
        ## We have reached the end of configuration file...
        catch {$graph destroy}
        return {}
      }
      default {}
    }
  }
  return $graph
};# ::SystemManagement::ReadSystemGraphFromConfigurationFile


## ::SystemManagement::SaveConfigurationFile
#  Save the System related information in the specified file. The saved
#  information will include the System tree as well all System graphs...
proc ::SystemManagement::SaveConfigurationFile {File} {
  variable ToplevelWindow
  variable SystemTree

  set file [open $File w]
  ## Write Header...
  puts $file "__START_CONFIG\n__START_TASKS"
  puts $file "__top_geometry=$ToplevelWindow(Geometry)"
  puts $file "__top_pane_percent=$ToplevelWindow(PanePercent)\n__start_graphs"
  ## Write the System tree...
  set id 1; set Systems {}
  foreach node [lsort [$SystemTree children root]] {
    set alias [$SystemTree get $node -key Alias]
    puts -nonewline $file "$id;[$SystemTree get $node -key Name];$alias;"
    ## Is it a leaf or a folder?
    if {[$SystemTree isleaf $node]} {
      puts $file {}
      lappend Systems $alias
    } else {
      ## This is a folder. Get all its children...
      set ChildrenList {}
      foreach child [lsort [$SystemTree children $node]] {
	set alias [$SystemTree get $child -key Alias]
        lappend ChildrenList "$alias|[$SystemTree get $child -key Name]"
	lappend Systems $alias
      }
      puts $file [join $ChildrenList =]
    }
    incr id
  }
  puts $file "__end_graphs\n__END_TASKS"
  ## Save System graphs...
  set id 1
  foreach system $Systems {
    set graph ::SystemManagement::Graph::$system
    set name  [$graph get -key Name]
    ## Write Header...
    puts $file "__START_GRAPH=$name"
    puts $file "__geometry=[$graph get -key Geometry]"
    puts $file "__timeLog=[$graph get -key TimeLog]"
    puts $file "__grid=[$graph get -key Grid]"
    ## Write Nodes...
    puts $file "__start_nodes"
    foreach node [lsort [$graph nodes]] {
      puts $file "$name,$node;[$graph node get $node -key X];[$graph node get \
           $node -key Y];0;[$graph node get $node -key Label];;;unrunnable;"
    }
    puts $file "__end_nodes\n__start_arcs"
    ## Write Arcs...
    foreach arc [lsort [$graph arcs]] {
      puts $file "$name,$id;$name,[$graph arc source $arc];$name,[$graph arc \
           target $arc];[$graph arc get $arc -key CurvePoints];;unrunnable;"
      incr id
    }
    puts $file "__end_arcs\n__END_GRAPH=$name"
  }
  ## We have finished!
  puts $file "__END_CONFIG"
  close $file
};# ::SystemManagement::SaveConfigurationFile
