##############################################################################
## AttributeViewer.tcl:                                                     ##
## ---------------------                                                    ##
##  This file implements a viewer for viewing Attributes stored either in   ##
##  Documents or Collections.                                               ##
##                                                                          ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Software and Knowledge Engineering Laboratory,                          ##
##  Institute of Informatics and Telecommunications,                        ##
##  N.C.S.R. "Demokritos"                                                   ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Text Engineering Platform.                                    ##
##############################################################################

CDM_RegisterViewer {Attribute Viewer} ggi_view_Attributes ITViewDocAtt

proc ggi_view_Attributes {doc {name {}} {window 1} {title {}} {win {}}} {
  package require ELEP::Viewers::AttributesViewer
  global ggi_view_Attributes GUI_Text GUI_Options

  if {![string length $win]} {
    set win [::ELEP::WindowUtilities::UniqueWindowName .attribute_viewer]
  }

  if {![string length $title] && [string length $doc]} {
    set title "Attributes Viewer: [tip_GetId $doc]..."
  }

  if {![string length $window]} {set window 1}

  if {$window} {
    ## Create a toplevel window...
    set collection {}
    set document   $doc
    switch -glob $doc {
      Col* {set collection $doc; set document Doc}
      Doc* {set collection [tip_GetParent $document]}
    }
    set obj [ELEP::Viewers::AttributesViewer create $win $collection $document]
    $obj title $title
    $obj setStatusText $title
  } else {
    ## Create a widget...
     error "unsupported: window == 0"
  }

  ## If doc == {}, "name" has an Attribute set...
  if {![string length $doc]} {
    $obj onUpdateAttributes $name
  }
  $obj show
  return $obj

  ## Place Attributes in our text widget...
  set CreateWidgets 1
  if {[winfo exists $win.frame]} {set CreateWidgets 0}
  set max_width   120
  set max_height  120
  if {[string length $doc]} {
    set Attrs [tip_GetAttributes $doc]
  } else {
    set Attrs $name
    set doc   [list $name]
  }
  if {[catch {package require mkWidgets}]} {
    if {$CreateWidgets} {
      ::ScrolledWidgets::ScrolledText $win.frame -spacing1 10
      bind $tw <Destroy> "ggi_view_AttributesCleanImages $win"
    }
  
    foreach attr [lsort -dictionary $Attrs] {
      foreach {name Value} $attr  {break}
      foreach {type value} $Value {break}
      switch -exact $type {
        GDM_STRING {
          $tw insert end [format "%s=\"%s\" (%s)\n" $name $value $type]
        }
        GDM_STRING_SET {
          $tw insert end [format "%s=\[%s\] (%s)\n" $name [join $value ,] $type]
        }
        CDM_IMAGE {
          $tw insert end [format "%s=" $name]
          ## Retrieve the Image...
          binary scan $value {}
          if {[catch {image create photo -data $value} image]} {
            set image IFileBroken
          } else {
            ## Maybe we need to resize the image?
            if {[image width  $image] > $max_width || 
                [image height $image] > $max_height} { 
              set Image [GUI_ShrinkImage $image $max_width $max_height]
              image delete $image
              set image $Image
            }
            lappend ggi_view_Attributes($win:Images) $image
          }
          $tw image create end -image $image -align center 
          $tw insert end [format " (%s)\n" $type]
        }
        CDM_BASE64_IMAGE {
          $tw insert end [format "%s=" $name]
          ## Retrieve the Image...
          set value [::base64::decode $value]
          binary scan $value {}
          if {[catch {image create photo -data $value} image]} {
            set image IFileBroken
          } else {
            ## Maybe we need to resize the image?
            if {[image width  $image] > $max_width || 
                [image height $image] > $max_height} { 
              set Image [GUI_ShrinkImage $image $max_width $max_height]
              image delete $image
              set image $Image
            }
            lappend ggi_view_Attributes($win:Images) $image
          }
          $tw image create end -image $image -align center 
          $tw insert end [format " (%s)\n" $type]
        }
        default {
          $tw insert end [format "%s=\"%s\" (%s)\n" $name $value $type]
        }
      }
    }
    $tw configure -state disabled -exportselection 1
  } else {
    if {$CreateWidgets} {
      ## We have loaded the mkWidgets package!
      ## Fix some bindings...
      foreach event {<4> <5> <<MB_2>>} {
        bind Listcontroltext $event [bind Text $event]
      }
      ## In order to add automatic scrollbars in the ListControl widget, we make
      ## a trick: We create a text widget, and we substitute the text widget
      ## with the listcontrol widget...
      ##::ScrolledWidgets::ScrolledText $win.frame
      frame $win.frame -bd 0
      destroy $tw
      listcontrol $win.frame.text -cursor left_ptr
      set ::ScrolledWidgets::ScrolledWidgets_ScrollbarStatus($tw.scrh) 0
      set ::ScrolledWidgets::ScrolledWidgets_ScrollbarStatus($tw.scrh_args)\
        [eval list -sticky snew -column 0 -row 2]
      set ::ScrolledWidgets::ScrolledWidgets_ScrollbarStatus($tw.scrv)   0
      set ::ScrolledWidgets::ScrolledWidgets_ScrollbarStatus($tw.scrv_args) \
        [eval list -sticky snew -column 1 -row 0 -rowspan 2]
      $tw configure -xscrollcommand \
          [list ::ScrolledWidgets::ScrollbarSet $tw.scrh set] \
        -yscrollcommand \
          [list ::ScrolledWidgets::ScrollbarSet $tw.scrv set]
      grid forget $tw.scrh $tw.scrv
      grid rowconfigure    $win.frame 0 -weight 1
      grid columnconfigure $win.frame 0 -weight 1

      grid $win.frame.text -row 0 -column 0 -sticky snew 
      bind $tw <Destroy> "+ggi_view_AttributesCleanImages $win"
      ## We want 3 Columns...
      $tw column insert Name  end -text {Attribute Name} \
        -width 200 -minsize 130
      $tw column insert Value end -text {Attribute Value} \
        -width 150 -minsize 130
      $tw column insert Type  end -text {Value Type} \
        -width 145 -minsize 20
      $tw column bind   Name  <Double-1> "$tw column fit Name"
      $tw column bind   Value <Double-1> "$tw column fit Value"
      $tw column bind   Type  <Double-1> "$tw column fit Type"
    }

    $tw delete 0 end
    ggi_view_AttributesCleanImages $win
    set index 1
    foreach attr [lsort -dictionary $Attrs] {
      foreach {name Value} $attr  {break}
      foreach {type value} $Value {break}
      switch -exact $type {
        GDM_STRING {
          $tw insert end [list $name "\"$value\"" "($type)"]
        }
        GDM_STRING_SET {
          $tw insert end [list $name "\[[join $value ,]\]" "($type)"]
        }
        CDM_IMAGE {
          ## Retrieve the Image...
          binary scan $value {}
          if {[catch {image create photo -data $value} image]} {
            set image IFileBroken
          } else {
            ## Maybe we need to resize the image?
            if {[image width  $image] > $max_width || 
                [image height $image] > $max_height} { 
              set Image [GUI_ShrinkImage $image $max_width $max_height]
              image delete $image
              set image $Image
            }
              lappend ggi_view_Attributes($win:Images) $image
          }
          $tw insert end [list $name {} "($type)"]
          $tw image create Value.$index -image $image
          $tw tag add $image Value.$index
          $tw tag bind $image <Enter> "$tw configure -cursor hand2"
          $tw tag bind $image <<MB_1>> \
              "ggi_view_AttributeImage $doc [list $name] {} {} $win.vi"
          $tw tag bind $image <Leave> "$tw configure -cursor {}"
        }
        CDM_BASE64_IMAGE {
          ## Retrieve the Image...
          set value [::base64::decode $value]
          binary scan $value {}
          if {[catch {image create photo -data $value} image]} {
            set image IFileBroken
          } else {
            ## Maybe we need to resize the image?
            if {[image width  $image] > $max_width || 
                [image height $image] > $max_height} { 
              set Image [GUI_ShrinkImage $image $max_width $max_height]
              image delete $image
              set image $Image
            }
              lappend ggi_view_Attributes($win:Images) $image
          }
          $tw insert end [list $name {} "($type)"]
          $tw image create Value.$index -image $image
          $tw tag add $image Value.$index
          $tw tag bind $image <Enter> "$tw configure -cursor hand2"
          $tw tag bind $image <<MB_1>> \
              "ggi_view_AttributeImage $doc [list $name] {} {} $win.vi"
          $tw tag bind $image <Leave> "$tw configure -cursor {}"
        }
        default {
          $tw insert end [list $name "\"$value\"" "($type)"]
        }
      };# switch
      incr index
    };# foreach
    update
  }

  if {$CreateWidgets} {
    pack $win.frame -fill both -expand true
    update
    if {$window} {
      catch {winico setwindow $win $WMIcons(AnnView)}
      wm deiconify $win
      focus $tw
    }
  }
  catch {$tw column fit Value}
  catch {$tw column fit Type}
  return $tw
};# ggi_view_Attributes

proc ggi_view_AttributesCleanImages {win} {
  global ggi_view_Attributes
  if {![info exists ggi_view_Attributes($win:Images)]} {return}
  foreach image $ggi_view_Attributes($win:Images) {
    catch {image delete $image}
  }
  unset -nocomplain ggi_view_Attributes($win:Images)
};# ggi_view_AttributesCleanImages

proc ggi_view_AttributeImage {doc name {unused {}} {title {}}
                              {win .attribute_image_viewer}} {
  global GUI_Options GUI_Text
  set win [ggi_unused_win $win]
  if {![string length $title]} {
    set title "Attribute Image Viewer: $name..."
  }

  toplevel $win
  wm withdraw $win
  wm title $win $title

  ## Create the window
  ## Toolbar...
  set toolbar_frame $win.tf; set toolbar $toolbar_frame.tb
  Toolbar::ToolbarFrame $toolbar_frame \
          -relief $GUI_Options(ToolbarFrameRelief) \
          -borderwidth $GUI_Options(ToolbarFrameReliefBorderWidth) \
          -decorate $GUI_Options(ToolbarDecorate)

    set tw $win.frame.text
    set toolbar $toolbar_frame.tb1
    Toolbar::create $toolbar -relief $GUI_Options(ToolbarRelief) \
        -borderwidth $GUI_Options(ToolbarReliefBorderWidth) -hide 0
    Toolbar::addbutton $toolbar -tooltipvar GUI_Text(Exit) \
         -image ITExit -command "destroy $win"
    pack $toolbar -side left -fill y

  pack $toolbar_frame -side top -fill x

  ## Get the Image from the Attribute Value...
  set attr [tip_GetAttribute $doc $name]
  if {![llength $attr]} {
    ## Maybe doc holds the Attribute and is not a Document?
    set attr [lindex $doc 0]
  }
  foreach {name Value} $attr  {break}
  foreach {type value} $Value {break}
  switch -exact $type {
    CDM_IMAGE {
      ## Retrieve the Image...
      binary scan $value {}
      if {[catch {image create photo -data $value} image]} {
        set image IFileBroken
      }
    }
    CDM_BASE64_IMAGE {
      ## Retrieve the Image...
      set value [::base64::decode $value]
      binary scan $value {}
      if {[catch {image create photo -data $value} image]} {
        set image IFileBroken
      }
    }
    default {
      set image IFileBroken
    }
  }
  ## The Image display widget...
  pack [label $win.image -image $image -padx 10 -pady 10] \
        -padx 10 -pady 10 -fill both -expand 1
  bind $win.image <Destroy> "image delete $image"

  ## Various Image information
  frame $win.image_info
    label $win.image_info.width  -text "Width:  [image width  $image]" -anchor w
    label $win.image_info.height -text "Height: [image height $image]" -anchor w
    label $win.image_info.format -text "Format: [$image cget -format]" -anchor w
    pack $win.image_info.width  -side top -fill x
    pack $win.image_info.height -side top -fill x
    pack $win.image_info.format -side top -fill x
  pack $win.image_info -padx 10 -fill x

  frame $win.buttons
  pack  $win.buttons -side bottom
  button $win.buttons.dismiss -textvariable GUI_Text(Dismiss) \
    -command [list destroy $win]
  pack $win.buttons.dismiss -padx 2 -pady 2

  wm deiconify $win
  focus $win
};# ggi_view_AttributeImage
