#
# XML Document Editing interface
# 
# (c) Solagem Oy 2002
#

namespace eval xmledit {

package require BWidget
package require dom
package require xslt
package require http

# 
# Namespace variables
#
variable config
variable schema
variable schema_transformer1
variable schema_transformer2
variable gui_transformer

variable link

set config(debug) false
set config(saveEmpty) false

#
# Xmledit initialization. 
#
# An editing frame is initialized for specific XML schema.
# 
# Parameters:
#   w : xmledit root widget
#   s : document schema handle
#
proc init {w args} {
  variable link
  
  #
  # GUI generation
  #

  # top frame to hold all editor elements 
  PanedWindow $w

  # tree widget for XML document display
  set pane [$w add]
  ScrolledWindow $pane.sw
  pack $pane.sw -fill both -expand 1 -side left -anchor w
  set tree [Tree $pane.sw.tTree -selectcommand xmledit::node_selected -height 20]
  $pane.sw setwidget $tree
  
  # editing frame widget
  set pane [$w add -weight 2]
  set edit [frame $pane.fEdit -relief groove -borderwidth 3]
  pack $edit -fill both -expand 1 -side left -anchor w
  
  # create link array so that we can easily connect tree to edit frame etc.
  set link($w.TREE) $tree
  set link($tree.ROOT) $w  
  set link($w.EDIT) $edit
  set link($tree.EDIT) $edit
  set link($edit.ROOT) $w
  set link($edit.TREE) $tree

  # window global binds, these are effective with or without menubar
  bind ${tree}.c <Control-x> "xmledit::node_cut $tree"
  bind ${tree}.c <Control-X> "xmledit::node_cut $tree"
  bind ${tree}.c <Control-c> "xmledit::node_copy $tree"
  bind ${tree}.c <Control-C> "xmledit::node_copy $tree"
  bind ${tree}.c <Control-v> "xmledit::node_paste $tree"
  bind ${tree}.c <Control-V> "xmledit::node_paste $tree"
  bind ${tree}.c <Key-Delete> "xmledit::node_delete $tree"
  bind ${tree}.c <Key-Escape> "xmledit::node_reset $tree" 

  $tree bindText <ButtonPress-3> "xmledit::node_menu $tree %X %Y"
  $tree bindImage <ButtonPress-3> "xmledit::node_menu $tree %X %Y"
  $tree bindText <Double-ButtonPress-1> "xmledit::open_close_node $tree"
  $tree bindImage <Double-ButtonPress-1> "xmledit::open_close_node $tree"

  # process configuration options  
  foreach {name value} $args {
    configure $w $name $value
  }
  
  return $w
}

#
# A service procedure for opening/closing tree
#
proc open_close_node {w node} {
  if {[$w itemcget $node -open]} {
    $w closetree $node 0
  } else {
    $w opentree $node 0
  }
}

#
# Xmledit configuration. 
#
# Configuration options:
#
# -debug : debug on/off, display some debug information in console window
#
proc configure {w args} {
  variable config
  variable link
  variable schema
  variable schema_transformer1
  variable schema_transformer2
  variable gui_transformer

  foreach {name value} $args {
    switch -- $name {
      -debug {
        set config(debug) $value
        if {$config(debug)} {
          console show
        }
      }
      -saveEmpty {
        set config(saveEmpty) $value
      }
      -menubar {
        set menubar $value
        menu $menubar.edit -tearoff 0
        $menubar add cascade -label Edit -menu $menubar.edit -underline 0

        $menubar.edit add command -label Cut -underline 2 -accelerator Ctrl+X -command "xmledit::node_cut $link($w.TREE)"
        $menubar.edit add command -label Copy -underline 0 -accelerator Ctrl+C -command "xmledit::node_copy $link($w.TREE)"
        $menubar.edit add command -label Paste -underline 0 -accelerator Ctrl+V -command "xmledit::node_paste $link($w.TREE)"    

        $menubar.edit add separator
        $menubar.edit add command -label Delete -underline 0 -accelerator Del -command "xmledit::node_delete $link($w.TREE)"
      }
      -schema {
        set schema $value

        # first we do some preprocessing for the schema
        set schema [$schema_transformer1 transform $schema]
        set schema [$schema_transformer2 transform $schema]
        if {$config(debug)} {
          debug [serialize $schema]
        }

        # create GUI procedure that is used at run-time to draw element frames
        set doc_gui [$gui_transformer transform $schema proc 'GUI$link($w.EDIT)']
        set doc_gui_text [::dom::libxml2::serialize $doc_gui -method text]
        eval $doc_gui_text
        debug $doc_gui_text
        document_destroy $doc_gui
      }
      default {
        error "Invalid configuration option: $name"
      }
    }
  }
}


#
# Document editing interface
# 
# Parameters:
#   w : Editing window widget path
#   d : Document handle
#
proc edit {w d} {
  variable doc
  variable schema

  set doc $d
  
  # create document tree view
  document_display $w $d $schema
}

#
# Document create interface
# 
# Parameters:
#   w : Editing window widget path
#   d : Document handle
#
proc new {w args} {
  variable schema

  # process options
  foreach {name value} $args {
    switch -- $name {
      -noNamespaceSchemaLocation {
        set noNamespaceSchemaLocation $value
      }
      default {
        error "Unknown option: $name"
      }
    }
  }

  # select and create rootnode
  set root_elements [dom::node selectNode $schema /xs:schema/xs:element -namespaces {xs http://www.w3.org/2001/XMLSchema}]
  switch [llength $root_elements] {
    0 {
      error "Invalid schema, no root element defined"
    } 
    1 {
      # single root element, just use that
      set element [dom::element getAttribute [lindex $root_elements 0] name]
    } 
    default {
      # several root element possibilities, ask user
      Dialog $w.d -separator 1 -title "Select root element"
      set f [$w.d getframe]
      listbox $f.list -yscrollcommand "$f.scroll set"
      scrollbar $f.scroll -command "$f.list yview"
      foreach i $root_elements {
        $f.list insert end [dom::element getAttribute $i name]
      }
      pack $f.list -fill both -expand 1 -side left
      pack $f.scroll -fill y -expand 1 -side left
      $w.d add -command "$w.d enddialog \[$f.list get \[$f.list curselection\]\]" -text OK
      set element [$w.d draw]
    }
  }

  if {![string length $element]} {
    return
  }

  # create new document
  set doc [dom::libxml2::create]

  # create root node
  set root [dom::document createElement $doc $element]

  # set schema location
  if {[info exist noNamespaceSchemaLocation]} {
    dom::element setAttribute $root xmlns:xsi {http://www.w3.org/2001/XMLSchema-instance}
    dom::element setAttribute $root xsi:noNamespaceSchemaLocation $noNamespaceSchemaLocation
  }

  # create document tree view
  document_display $w $doc $schema

  return $doc
}

#
# Display whole document in tree widget
#
proc document_display {w node snode {root {}}} {
  variable link

  # when called for the first time, initialize recursion and display root node
  if {[string equal [dom::node cget $node -nodeType] document]} {
    # starting point of recursion
    set root [dom::document cget $node -documentElement]
    set tree $link($w.TREE)

    # clear old data
    foreach i [$tree nodes root] {
      $tree delete $i
    }
    
    # Current version if tcldomxml does not support xpath with namespaces.
    # This would be much easier with xpath.
    foreach i [dom::node children [dom::document cget $snode -documentElement]] {
      if {[string equal [dom::node cget $i -nodeName] element] && [string equal [dom::element getAttribute $i name] [dom::node cget $root -nodeName]]} {
        set sroot $i
        break
      }
    }
    if {![info exist sroot]} {
      error "Invalid schema for given document. Element \"[dom::node cget $node -nodeName]\" definition not found."
    }

    # enter recursion
    document_display $tree $root $sroot root

    # display initial frame and tree
    node_display $tree $root
  } else {
    # check, if node is considered as simple. If its i, do not add
    if {[is_simple $snode]} {
      return
    }

    # Insert the node to the tree. 
    # Node data relates the node to a schema node for GUI generation.
    $w insert end $root $node -image [Bitmap::get folder] -data $snode
    
    # label node
    node_label $w $node

    # find out the possible node children
    array set child_element [child_elements $snode]
  
    # add all node child elements for which there exist a schema node
    foreach child [::dom::node children $node] {
      if {[string equal [dom::node cget $child -nodeType] {element}]} {
        # child element path
        set name [dom::node cget $child -nodeName]
        if {[info exist child_element($name)]} {
          document_display $w $child $child_element($name) $node
        }
      }
    }
  }
}

#
# Return a list of valid children elements starting from given schema node
#
proc child_elements {snode {firstcall 1}} {
  set name [dom::node cget $snode -nodeName]
  
  set ref [dom::element getAttribute $snode ref]

  if {$firstcall && [string length $ref]} {
    # if schema node is just a reference, do search from referenced node
    return [child_elements [dom::node selectNode $snode /xs:schema/xs:element\[@name='$ref'\] -namespaces {xs http://www.w3.org/2001/XMLSchema}]]
  } elseif {$firstcall || ![string equal $name element]} {
    # recursion start point
    set r [list]
    foreach child [dom::node children $snode] {
      if {[string equal [dom::node cget $child -nodeType] element]} {
        foreach j [child_elements $child 0] {
          lappend r $j
        }
      }
    }
    debug $r
    return $r
  } else {
    # end of recursion. If element node contains ref, return link to the actual node
    if {[string length $ref]} {
      set name $ref
    } else {
      set name [dom::element getAttribute $snode name]
    }
    return [list $name $snode]
  }
}

#
# Heuristic to check from schema definition if a node is "simple" or not.
#
# Simple elements are shown as updateable fields in frame, while
# non-simple elements are shown as nodes in tree view.
#
proc is_simple {snode} {
  # if the node has id attribute set, it should have a GUI of its own, considered non-simple
  return [expr ![string length [dom::element getAttribute $snode id]]] 
}

#
# This command opens a editing frame for selected node
#
proc node_selected {wTree node} {
  variable link
  
  set edit $link($wTree.EDIT)

  frame_create $edit [$wTree itemcget $node -data]
  frame_display $edit $node
}

#
# This command posts node menu
#
proc node_menu {w x y node} {
  variable link
  $w selection set $node
  set element [dom::node cget $node -nodeName]
  set parent $node
  
  # check if there is already menu for given element, destroy if exist
  if {[winfo exist $w.m$element]} {
    destroy $w.m$element
  }
    menu $w.m$element -tearoff 0
    
    # Add commands for each possible non-simple child
    set snode [$w itemcget $node -data]
    set separator 0
    set sparent   {}
    set sparent_maxOccurs 1
    
    foreach {name node} [child_elements $snode] {
      # parse maxOccurs value from schema
      set maxOccurs [max_occurs $node]
      
      # additionally parse maxOccurs value of parent node if it's name is 'all', 'sequence' or 'choice'
      if {![string length $sparent]} {
        set sparent [dom::node cget $node -parentNode]
        set sparentname [dom::node cget $sparent -nodeName]
        if {[string equal $sparentname sequence] || [string equal $sparentname all] || [string equal $sparentname choice]} {
          set sparent_maxOccurs [max_occurs $sparent]
        }
      }

      if {$maxOccurs == -1 && $sparent_maxOccurs == -1} {	# both attributes are unbounded
        set maxOccurs 1
      } elseif {$maxOccurs == 0 || $sparent_maxOccurs == 0} {	# one of attributes is zero
        set maxOccurs 0
      } else {
        set result [expr $maxOccurs * $sparent_maxOccurs]
        if {$result < 0} {					# one of attributes is unbounded, another is positive
          set maxOccurs 1					
        } else {						# both attributes are positive
          set node_qty [llength [nodelist $parent $name]]	
          if {$node_qty >= $result} {
      	    set maxOccurs 0
          }
        }
      }
      
      if {![is_simple $node]} {
        # add item to the menu, add in disabled state if quota is exceeded
        if {$maxOccurs} {
          $w.m$element add command -label "Add $name" -command "xmledit::node_add $w $node"
        } else {
          $w.m$element add command -label "Add $name" -command "xmledit::node_add $w $node" -state disabled
        }
        set separator 1    
      }
    }
    
    if {$separator} {
      $w.m$element add separator
    }
    
    $w.m$element add command -label Cut -accelerator Ctrl+X -command "xmledit::node_cut $w"
    $w.m$element add command -label Copy -accelerator Ctrl+C -command "xmledit::node_copy $w"
    $w.m$element add command -label Paste -accelerator Ctrl+V -command "xmledit::node_paste $w"
    $w.m$element add separator
    
    # Delete command
    $w.m$element add command -label Delete -accelerator Del -command "xmledit::node_delete $w"

  $w.m$element post $x $y
}

#
# Get maxOccurs-attribute's value of the node
#
# parameters: 	node - schema node
# return:      -1 		if unbounded
#	   	1 		if no attribute defined
#  	   			parsed value otherwise
#
proc max_occurs {node} {
  set maxOccurs [dom::element getAttribute $node maxOccurs]
  if {![string length $maxOccurs]} {
    return 1
  } elseif {[string equal $maxOccurs unbounded]} {
    return -1
  }
  return $maxOccurs
}

#
# Check if paste allowed (check if child with the same name exists and maxOccures attribute's value)
#
# parameters:	node - node to be pasted
#		parent - target node (selected node or selected node's parent)
#
proc paste_allowed {w parent node} {

  # copying directly to the document root is not allowed
  if [string equal $parent root] {
    return 0
  }
    
  # get schema for parent
  set sparent [$w itemcget $parent -data]
    
  # find out the possible node children
  array set child_element [child_elements $sparent]
    
  # name of copied node
  set name [dom::node cget $node -nodeName]
    
  if {![info exist child_element($name)]} {
    return 0
  }
  
  #get schema nodes
  set snode [$w itemcget $node -data]
  
  # from schema document
  set sparentname [dom::node cget $sparent -nodeName]
  set maxOccurs [max_occurs $snode]
  
  set sparent_maxOccurs 1
  if {[string equal $sparentname sequence] || [string equal $sparentname all] || [string equal $sparentname choice]} {
    set sparent_maxOccurs [max_occurs $sparent]
  }
  
  if {$maxOccurs == -1 && $sparent_maxOccurs == -1} {		# both attributes are unbounded
    return 1
  } elseif {$maxOccurs == 0 || $sparent_maxOccurs == 0} {	# one of attributes is zero
    return 0
  } else {
    set result [expr $maxOccurs * $sparent_maxOccurs]
    if {$result < 0} {						# one of attributes is unbounded, 
      return 1;							# another is positive
    } else {							# both attributes are positive
      set node_qty [llength [nodelist $parent $name]]	
      if {$node_qty >= $result} {
        return 0
      } else {
        return 1
      }
    }
  }
}

#
# Cut node
#
proc node_cut {w} {
  variable clipboard
  
  # get selected item from tree widget
  set selection [$w selection get]
  
  # currently only single node cut'n'paste supported
  if {[llength $selection] != 1} {
    tk_messageBox -message "Invalid selection"
    return
  }
  
  set clipboard(data) $selection
  set clipboard(mode) cut  
}

#
# Copy node
#
proc node_copy {w} {
  variable clipboard
  
  # get selected item from tree widget
  set selection [$w selection get]

  # currently only single node cut'n'paste supported
  if {[llength $selection] != 1} {
    tk_messageBox -message "Invalid selection"
    return
  }

  set clipboard(data) $selection
  set clipboard(mode) copy
}

#
# Paste node
#
proc node_paste {w} {
  variable clipboard

  # test that there is data to paste
  if {![info exist clipboard]} {
    tk_messageBox -message "Select first data to paste"
    return
  }
  
  # get selected item from tree widget
  set selection [$w selection get]

  # we can have only one node selected for paste operation
  if {[llength $selection] != 1} {
    tk_messageBox -message "Invalid selection"
    return
  } 

  # depends on clipboard(mode), cut or copy
  switch $clipboard(mode) {
    cut {      
      # paste only if selected node and node to be copied differ
      if {![match_parent $w $selection $clipboard(data)] && [paste_allowed $w $selection $clipboard(data)]} {
        set parent [dom::node parent $clipboard(data)]
        set node   [dom::node removeChild $parent $clipboard(data)]
        set snode  [$w itemcget $node -data]
        $w delete $node
        
        dom::node appendChild $selection $node
        document_display $w $node $snode $selection
        $w opentree [$w parent $node] 0
        node_display $w $node
        unset clipboard
      } else {
        tk_messageBox -message "Cannot paste node \'[$w itemcget $clipboard(data) -text]\' here"
      }
    }
    copy {
      set node   [dom::node cloneNode $clipboard(data)]
      set snode  [$w itemcget $clipboard(data) -data]

      # paste under selected node if it is different then node to be copied,
      # otherwise paste node on the same level
      if [string compare $selection $clipboard(data)] {
        if {[paste_allowed $w $selection $clipboard(data)]} {
          dom::node appendChild $selection $node
	  document_display $w $node $snode $selection
	} elseif {[paste_allowed $w [$w parent $selection] $clipboard(data)]} {
	  dom::node appendChild [$w parent $selection] $node
	  document_display $w $node $snode [$w parent $selection]
	} else {
	  tk_messageBox -message "Cannot paste node \'[$w itemcget $clipboard(data) -text]\' here"
	  return
	}
      } elseif {[paste_allowed $w [$w parent $selection] $clipboard(data)]} {
        dom::node appendChild [$w parent $selection] $node
	document_display $w $node $snode [$w parent $selection]
      } else {
        tk_messageBox -message "Cannot paste node \'[$w itemcget $clipboard(data) -text]\' here"
        return
      }

      $w opentree [$w parent $node] 0
      node_display $w $node
    }
    default {
      "Not yet implemented: $clipboard(mode)"
    }
  }
}

#
# Delete node, ask for confirmation before
#
proc node_delete {w} {
  # check where is focus
  set selection [$w selection get]

  if {![llength $selection]} {
    tk_messageBox -message "No nodes selected"
    return
  }
  
  foreach i $selection {
    append text \n[$w itemcget $i -text]
  }
  set answer [tk_messageBox -message "Do you really want to delete nodes:$text" -type okcancel -icon question -title "Delete node?"]

  if {[string equal ok $answer]} {
    foreach i $selection {
      set parent [dom::node parent $i]
      if {[string equal [dom::node cget $parent -nodeType] document]} {
        tk_messageBox -message "Cannot delete root element"
      } else {
        node_reset $w
        set node   [dom::node removeChild $parent $i]
        dom::libxml2::destroy $node
        node_display $w $parent
        $w delete $node
      }
    }
  }
}

proc node_reset {w} {
  variable clipboard
  if {[info exist clipboard]} {
    unset clipboard
  }
}

#
# Add node 
#
proc node_add {w snode} {
  # get selected item from tree widget
  set selection [$w selection get]
  if {[llength $selection] != 1} {
    tk_messageBox -message "Invalid selection"
    return
  }

  # create new node and insert to tree
  set element [dom::element getAttribute $snode name]
  if {![string length $element]} {
    set element [dom::element getAttribute $snode ref]
  }

  set newnode [dom::document createElement $selection $element]
  $w insert end $selection $newnode -text $element -image [Bitmap::get folder] -data $snode

  # select and display created node
  node_display $w $newnode
}

#
# Generate node label.
#
# If a node is unique within its parent (1-1 relationship),
# node label is simply the element name.
#
# For non-unique nodes (1-N relationship) some identifier is needed
# in the label. Here we get simply the first attribute or first element value,
# and use it as the identifier (not quite general, but in most cases provides
# enough information.
#
proc node_label {w node} {
  # base label is always element name
  set label [dom::node cget $node -nodeName]

  set path [dom::element getAttribute [$w itemcget $node -data] nodelabel]
  if {[string length $path]} {
    set label "$label - [string range [string map {\n { }} [dom::node selectNode $node string($path)]] 0 25]"
  }

  # set node label in tree
  $w itemconfigure $node -text $label
}

#
# Display node on the tree and frame on the left side
#
proc node_display {w node {open 0} {opentreenode {}}} {
  if [string equal $opentreenode {}] {set opentreenode $node}
  $w opentree $opentreenode $open
  $w see $node
  $w selection set $node
  node_selected $w $node
}

#
# Create editing frame for given schema element. 
#
proc frame_create {w snode} {
  # clear old widgets in the frame
  foreach i [winfo children $w] {destroy $i}

  # Create element GUI based on schema node id
  GUI$w [dom::element getAttribute $snode id] $w

  # remember schema node if the frame has to be redrawn
  upvar #0 data$w framedata
  set framedata(SNODE) $snode
  
  pack $w
}


#
# Display XML node in a frame
#
proc frame_display {w {node {}}} {
  upvar #0 data$w framedata

  debug "frame_display"

  # if called with empty node, redraw frame and display current node
  if {![string length $node]} {
    set node $framedata(NODE)
    frame_create $w $framedata(SNODE)
  }

  # field xpath values all start with ".", just loop them and set frame variables
  foreach path [array names framedata .*] {
    set node2 [dom::node selectNode $node $path]
    if {[llength $node2]} {
      set value [dom::node stringValue $node2]
      set framedata($path) [dom::node stringValue $node2]
      debug "$path: $value"
    }
  }
  
  # node handle is saved to frame array
  set framedata(NODE) $node

  # for some reason BWidget frame occasionally fails to get focus
  # focus -force $w
}

#
# This procedure saves the frame data back to given XML document.
#
proc frame_save {w} {
  variable config
  variable link

  upvar #0 data$w framedata
  set root $framedata(NODE)

  debug "frame_save"

  # Go through field xpath list and assign framedata values to XML document
  foreach xpath [array names framedata .*] {
    if {[info exist framedata($xpath)]} {
      debug "$xpath: $framedata($xpath)"      
      # find node and assign node value, create if necessary
      node_save $root $xpath $framedata($xpath) $config(saveEmpty) 
    } 
  }

  # refresh label in tree widget
  node_label $link($w.TREE) $framedata(NODE)
}

#
# Utility procedures
#

#
# Helpers for managing stack for xpath in GUI procedure
#
proc push {stackname data} {
  upvar $stackname stack
  append stack /$data
}
proc pop {stackname} {
  upvar $stackname stack

  # the stack string delimiter is taken from the beginning of the string
  set pos [string last / $stack]
  set data [string range $stack $pos end]
  incr pos -1
  set stack [string range $stack 0 $pos]
  
  # return popped item
  return $data
}

# 
# Bind given XML field to frame widget. This is called from
# generated code in create_frame context.
# 
# Basic binding just sets textvariable configuration option,
# but special widget types may need other mechanism.
#
# After binding, all frame fields can be referenced through 
# frame array data.w(xpath)
#
proc xmlfield {w xpath} {
  upvar framedata framedata
  switch [winfo class $w] {
    Text {
      textvariable $w data$framedata(FRAME)($xpath)
      set framedata($xpath) {} ;# to make variable visible
    }
    Checkbutton - Radiobutton {
      $w configure -variable data$framedata(FRAME)($xpath)
    }
    default {
      $w configure -textvariable data$framedata(FRAME)($xpath)
    }
  }
  if {[lsearch $framedata(FIELDS) $xpath]} {
    lappend framedata(FIELDS) $xpath
  }
}

#
# Bind variable to text widget using trace. Have to do this
# because text widget does not have -textvariable configuration
# option.
#
proc textvariable {w varname} {
  upvar #0 $varname var
  set var {}
  trace variable var rw "xmledit::textvariable_trace $w"
}
proc textvariable_trace {w varname ix op} {
  upvar $varname var
  switch $op {
    r {
      set var($ix) [string trimright [$w get 1.0 end]]
    }
    w {
      $w delete 1.0 end
      $w insert end $var($ix)
    }
  }
}

#
# split text to lines
#
proc linesplit {s chars} {
  # remove carriage returns
  set s [string map {\r {}} $s]
  
  # cut to lines
  set s [split $s \n]
  set out {}
  foreach l $s {
    while {[set splitpos [string first { } $l $chars]] > -1} {
      append out [string range $l 0 $splitpos]\n
      set l [string range $l [incr splitpos] end]
    }
    append out $l\n
  }

  # trim result and return
  return [string trim $out]
}

#
# Bind helptext to widget
#
proc helptext {w s} {
  DynamicHelp::register $w balloon [linesplit $s 50]
}

#
# Bind variable to list widget using trace. 
#
proc listvariable {w varname} {
  upvar #0 $varname var
  trace variable var rw "xmledit::listvariable_trace $w"
}
proc listvariable_trace {w varname ix op} {
  upvar $varname var
  switch $op {
    w {
      foreach i $var($ix) { 
        $w insert end $i
      }
    }
  }
}

# Some XML Utility functions

#
# Create XML document object from input string
#
proc doc_from_string {data} {
  return [dom::libxml2::parse $data]
}

#
# Create XML document object from file
#
proc doc_from_file {xmlfile} {
  set f [open $xmlfile]
  fconfigure $f -encoding binary -translation binary
  set data [read $f]
  close $f

  set doc [dom::libxml2::parse $data];# -binary 1]
  return $doc
}

#
# Create XML document object from URL (http)
#
proc doc_from_url {url} {
  set token [::http::geturl $url -binary 1]
  set data [::http::data $token]
  ::http::cleanup $token

  set doc [dom::libxml2::parse $data -binary 1]
  return $doc
}

#
# Return node value, or empty, if not found
#
proc nodevalue {root {path {.}}} {
  if {[string compare $path .]} {
    set node [dom::node selectNode $root $path]
  } else {
    set node $root
  }
  
  if {[llength $node]} {
    return [dom::node stringValue [lindex $node 0]]
  } else {
    return ""
  }
}

# 
# Helper to locate schema file based on XML file name.
# Schema file is searched from:
#   XML file directory
#   Program home dir
#
proc locate_schema {doc xmlfile {home {.}}} {
  set xmlroot [file rootname $xmlfile]

  set schemafile [get_schemaname $doc]
  
  if {[file isfile [file join $xmlroot $schemafile]]} {
    return [file join $xmlroot $schemafile]
  }

  if {[file isfile [file join $home $schemafile]]} {
    return [file join $home $schemafile]
  }
  
  return {}
}

# 
# Helper to locate schema file based on XML file content.
#
proc get_schemaname {doc} {

  set docroot [dom::document cget $doc -documentElement]
  set schemafile [dom::element getAttribute $docroot noNamespaceSchemaLocation]
  
  return $schemafile
}

#
# Produce string data from XML document object
#
proc serialize {node {indent 0}} {
  return [dom::libxml2::serialize $node -indent $indent]
}

#
# Destroy dom object
#
proc document_destroy {doc} {
  ::dom::libxml2::destroy $doc
}

#
# Return org.w3c.dom.NodeList object resolved with xpath in a document
#
proc nodelist {node path} {
  return [dom::node selectNode $node $path]
}

#
# Return org.w3c.dom.Node object resolved with xpath in a document
# The node is the first matching node in a nodeset
#
proc node {node1 path} {
  set nodelist [nodelist $node1 $path]
  if {[llength $nodelist] == 0} {
    error "No such node: $path"
  } else {
    return [lindex $nodelist 0]  
  }
}

proc insert_element_todoc {curdoc node nodename value} {
	set elem [dom::document createElement $curdoc $nodename]
	dom::node appendChild $node $elem
	set txtelem [dom::document createTextNode $curdoc $value]
	dom::node appendChild $elem $txtelem
}

#
# node_save
#
# Traverse given xpath from some rootnode, and create needes elements.
# The last part of xpath can be attribute reference beginning with '@',
# whereas attribute node is created.
#
# Procedure returns true or false, depending outcome
#
proc node_save {node path value {save_always 0}} {
  foreach i [split $path /] {
    set match [dom::node selectNode $node $i]
    if {[llength $match]} {
      
      # remove attribute if value is empty
      if {[string match @* $i] && !($save_always || [string length $value])} {
        dom::element removeAttribute $node [string range $i 1 end]
        set node {}
        break
      } else {
        # partial match found, just continue matching
        set node [lindex $match 0]
      }
      
    } elseif {$save_always || [string length $value]} {
      # no match, create new node
      if {![string match @* $i]} {
        set node [dom::document createElement $node $i]
      } else {
        dom::element setAttribute $node [string range $i 1 end] "" 
        set node [dom::node selectNode $node $i]
      }
    } else {
      # no match, empty value not to save. Just quit
      set node {}
      break
    } 
  }
  
  if {[string length $node]} {
    dom::node configure $node -nodeValue [string map {{&} {&amp;}} $value]
    return 1
  } else {
    return 0
  }
}

#
# helper procedure, check does one of the node's parents has the same ID
#
proc match_parent {w selection node} {
  set next $selection
  while {![string equal $next {}]} {
    if {[string equal $node $next]} {
      return true
    } else {
      set next [$w parent $next]
    }
  }
  return false
}

#
# Local debug procedure
#
proc debug {args} {
  variable config
  if {$config(debug)} {
    foreach i $args {
      puts :$i
    }
  }
}

#
# Package initialization, load common stylesheets and compile them
#
set dir [file dirname [info script]]

set stylesheet [doc_from_file [file join $dir schema1.xsl]]
set schema_transformer1 [::xslt::compile $stylesheet]
document_destroy $stylesheet

set stylesheet [doc_from_file [file join $dir schema2.xsl]]
set schema_transformer2 [::xslt::compile $stylesheet]
document_destroy $stylesheet

set stylesheet [doc_from_file [file join $dir xmledit.xsl]]
set gui_transformer [::xslt::compile $stylesheet]
document_destroy $stylesheet

}

package provide xmledit 1.0
