##############################################################################
##  gzip.tcl:                                                               ##
## --------------------------------------------------                       ##
##  This class implements a tool for compressing/uncompressing files using  ##
##  gzip compression.                                                       ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require Trf

oo::class create ELEP::Archive::gzip {
  variable CRC

  constructor {args} {
  };# constructor

  ##########################################################################
  ## Method: gunzip
  ##########################################################################
  method gunzip {compressed output} {
    set f_in [open $compressed r]
    fconfigure $f_in -translation binary
    if {[catch {my attach_r $f_in} error]} {
      close $f_in
      error $error
    }
    set f_out [open $output w]
    fconfigure $f_out -translation binary
    fcopy $f_in $f_out
    close $f_out
    if {[catch {my detach_r $f_in} error]} {
      close $f_in
      error $error
    }
    close $f_in
  };# gunzip

  ##########################################################################
  ## Method: gzip
  ##########################################################################
  method gzip {input compressed} {
    set f_in [open $input r]
    fconfigure $f_in -translation binary
    set f_out [open $compressed w]
    fconfigure $f_out -translation binary
    my attach_w $f_out
    fcopy $f_in $f_out
    close $f_in
    my detach_w $f_out
    close $f_out
  };# gzip

  ##########################################################################
  ## Method: attach_w
  ##########################################################################
  # Attach to a channel for writing -- i.e., write .gz header,
  # enable compression
  method attach_w {f} {
    # Write header
    puts -nonewline $f [binary format "H*iH*" "1f8b0800" [clock seconds] "0003"]

    # set CRC [binary format x4]

    # Init/attach compression
    zip -attach $f -mode compress -nowrap 1
    fconfigure $f -translation binary -encoding binary

    # Init/attach CRC
    crc-zlib -attach $f -mode transparent \
      -write-destination [my varname CRC($f)] -write-type variable
    fconfigure $f -translation binary -encoding binary
    fconfigure $f -translation binary -encoding binary
    return $f
  };# attach_w

  ##########################################################################
  ## Method: detach_w
  ##########################################################################
  # # Detach from a channel for writing -- i.e., write .gz footer incl. CRC
  method detach_w {f} {
    set SIZE [tell $f]

    unstack $f   ; # CRC
    unstack $f   ; # gzip

    puts -nonewline $f $CRC($f)
    puts -nonewline $f [binary format "i" [expr $SIZE % 0x100000000]]
    unset CRC($f)
  };# detach_w

  ##########################################################################
  ## Method: attach_r
  ##########################################################################
  # # Attach to a channel for reading -- i.e., read and check .gz header,
  # enable decompression
  method attach_r {f} {
    # (using code from [http://wiki.tcl.tk/6175])
    set id [read $f 2]
    if { ![string equal $id \x1f\x8b] } {
      error "GZip channel $f: not a gzip file."
    }

    set cm [read $f 1]
    if { ![string equal $cm \x8] } {
      error "GZip channel $file: unknown compression method."
    }
    binary scan [read $f 1] b5 FLAGS
    # puts $FLAGS
    foreach {FTEXT FHCRC FEXTRA FNAME FCOMMENT} [split $FLAGS ""] {}
    binary scan [read $f 4] i MTIME
    set XFL [read $f 1]
    set OS [read $f 1]

    if { $FEXTRA } {
      binary scan [read $f 2] S XLEN
      set ExtraData [read $f $XLEN]
    }
    set name ""
    if { $FNAME } {
      set c [read $f 1]
      while { $c != "\x0" } {
        append name $c
        set c [read $f 1]
      }
    }
    set comment ""
    if { $FCOMMENT } {
      set c [read $f 1]
      while { $c != "\x0" } {
        append comment $c
        set c [read $f 1]
      }
    }
    set CRC16 ""
    if { $FHCRC } {
      set CRC16 [read $f 2]
    }

    # Init/attach decompression
    zip -attach $f -mode compress -nowrap 1
    fconfigure $f -translation binary -encoding binary

    # Init/attach CRC
    crc-zlib -attach $f -mode transparent \
      -read-destination [my varname CRC($f)] -read-type variable
    fconfigure $f -translation binary -encoding binary

    return $f
  };# attach_r

  ##########################################################################
  ## Method: detach_r
  ##########################################################################
  method detach_r {f} {
    set cmpSize [expr [tell $f] % 0x100000000]
    binary scan $CRC($f) i cmpCRC

    unstack $f   ; # CRC
    unstack $f   ; # gzip

    binary scan [read $f 4] i gzCRC
    binary scan [read $f 4] i gzSize
    binary scan [read $f 4] i gzSize
    if {$gzCRC != $cmpCRC} {
      error "GZip channel $f: CRC mismatch."
    }
    if {$gzSize != $cmpSize} {
      error "GZip channel $f: Size mismatch."
    }
    unset CRC($f)
  };# detach_r

  destructor {
  };# destructor

};# class ELEP::Archive::gzip

package provide ELEP::Archive::gzip 1.0
