##############################################################################
##                                                                          ##
##  Serialisable.tcl                                                        ##
## --------------------------------------------------                       ##
##  This file implements a class that can save itself into an XML file, and ##
##  restore itself from this XML file. Every class that iherits it, it can  ##
##  serialise/deserialise itself into/from XML.                             ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

##
## Load some needed packages.
##
package require TclOO    ;# TclOO adds object-oriented facilities to Tcl.

oo::class create ELEP::Base::serialisable {
  variable __serialisable_constructor_arguments \
           __serialisable_unserialisable_vars \
           __serialisable_execute_commands_after_loading

  constructor {args} {
    set __serialisable_unserialisable_vars {}
    set __serialisable_execute_commands_after_loading {}
    set __serialisable_constructor_arguments $args
  };# class constructor

  ## unserialisable_variables
  #  With this method the caller can declare which variables should not be
  #  serialised.
  method unserialisable_variables {args} {
    set __serialisable_unserialisable_vars $args
  };# unserialisable_variables

  ## execute_commands_after_object_deserialisation
  #  Register a list of commands to be executed just after the object is
  #  re-created from serialised data.
  method execute_commands_after_object_deserialisation {} {
  };# execute_commands_after_object_deserialisation

  method serialise {{format xml}} {
    switch $format {
      xml {return [my serialiseXML]}
      default {error "unsupported format \"$_format\""}
    }
  };# serialise

  ## deserialise:
  #  This function will try to reconstruct all objects described in "data".
  #  The contents of the "data" parameter must have been created with a call to
  #  "serialise", with the same format as the one specified to deserialse...
  method deserialise {data {format xml}} {
    switch $format {
      xml {return [my deserialiseXML $data]}
      default {error "unsupported format \"$_format\""}
    }
  };# deserialise

  method serialiseXML {} {
    set data "<object class=\"[[self] info class]\" name=\"$this\"\
      args=\"[string map {< &lt; > &gt; & &amp; \" &quot;} \
                         $__serialisable_constructor_arguments]\">"
    ## Process all variables and save their value.
    foreach {type name value_type value} [my RetrieveVariableInfo] {
      set value [string map {< &lt; > &gt; & &amp; \" &quot;} $value]
      append data "\n  <$type type=\"$value_type\"\
                              name=\"$name\">$value</variable>"
    }
    append data "\n</object>"
    return $data
  };# serialiseXML

  ## deserialiseXML:
  #  This function will try to reconstruct all objects described in "data".
  #  The contents of the "data" parameter must have been created with a call to
  #  "serialise", with the same format as the one specified to deserialse...
  method deserialiseXML {data} {
    set objs {}
    set index [string first <object $data]
    while {$index != -1} {
      set end_index [string first </object> $data $index]
      if {$end_index == -1} {
        error [::MSGS::msg PrematureEndOfSerialisedData]
      }
      set objData [string range $data $index [expr {$end_index-1}]]
      ## Parse the object class and object name:
      regexp {class=\"([^\"]+)\"} $objData _ objClass
      regexp {name=\"([^\"]+)\"}  $objData _ objName
      regexp {args=\"([^\"]+)\"}  $objData _ objArgs
      set index [string first <variable $objData]
      set objData [string range $objData $index end]
      ## Instead of parsing the XML data, convert it to a form Tcl can
      ## understand and then evaluate it :-)
      ## Note: the order in which substitutions are performed matters!
      set objData [string map {
        {<variable type="string" name="}    {set "@itcl [self] }
        {<common type="string" name="}      {set "@itcl [self] }
        {<variable type="array" name="}     {array set "@itcl [self] }
        {<variable type="undefined" name="} {set "undefined_}
        {<object class=} {set obj [}
        </variable> \}
        > \ \{ }                          $objData]
      foreach var {objClass objName objData objArgs} {
        set $var [string map {&lt; < &gt; > &amp; & &quot; \"} [set $var]]
      }
      ## Create the object:
      # Log "Deserialising: $objClass $objName $objArgs"
      ## 
      if {[catch {eval [list $objClass] [list $objName] $objArgs} obj]} {
        ## Perhaps the object already exists?
        Log "Warning: object $objName already exists?\n$obj"
        catch {itcl::delete object $objName}
        set obj [eval [list $objClass] [list $objName] $objArgs]
      }
      $obj deserialiseEvaluate $objData
      foreach cmd [set "@itcl $obj \
                              __serialisable_execute_commands_after_loading"] {
        $obj evaluate $cmd
      }
      lappend objs $obj
      # Log "Class: $objClass\nName: $objName\n$objData\n"
      set index [string first <object $data $end_index]
    }
    return $objs
  };# deserialiseXML

  method RetrieveVariableInfo {} {
    set info {}
    foreach var [[self] info variable] {
      switch -glob $var {
        *::this                                          -
        *::__serialisable_unserialisable_vars            -
        *::__serialisable_execute_commands_after_loading -
        *::__serialisable_constructor_arguments          {}
        default {
          ## Remove all namespace qualifications from the variable name.
          set index [string last :: $var]; incr index 2
          set stripped_var [string range $var $index end]
          ## Retrieve the variable value...
          set value [[self] info variable $var -value]
          set type  [[self] info variable $var -type]
          set value_type string
          
          ## If the value is the string <undefined>, then there are two
          ## possibilities: either the variable is really undefined, or it is
          ## something else, i.e. an array.
          if {[string equal $value <undefined>]} {
            #set scoped_varname "@itcl [self] [[self] info class]::$stripped_var"
            set scoped_varname "@itcl [self] $stripped_var"
            if {[array exists $scoped_varname]} {
              set value_type array
              set value [array get $scoped_varname]
            } else {
              set value_type undefined
              set value {}
            }
          }
          ## Finally, check if the variable must be serialised...
          if {[lsearch $__serialisable_unserialisable_vars $stripped_var]==-1} {
            lappend info $type $stripped_var $value_type $value
          }
        }
      };# switch -glob $var
    };# foreach var [[self] info variable]
    return $info
  };# RetrieveVariableInfo

  method deserialiseEvaluate {code} {
    ## Arrays are somewhat tricky, as values can be added by the contrsuctor
    ## that possible do not exist in the serialised object. Its better to
    ## unset arrays before we re-define their value...
    regsub -all {array set (\"@itcl \[self] [^\"]+\")} $code \
                {array unset \1; \0} code
    eval $code
  };# deserialiseEvaluate

  method evaluate {cmd} {
    return [eval $cmd]
  };# evaluate
};# class ELEP::Base::serialisable

package provide ELEP::Base::serialisable 1.0
