##############################################################################
##  Utilities.tcl:                                                          ##
## --------------------------------------------------                       ##
##  This file implements some basic utility procs/methods.                  ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require TclOO

##
## Define "classmethod"...
##
proc ::oo::define::classmethod {name {args ""} {body ""}} {
  # Create the method on the class if
  # the caller gave arguments and body
  # Code from: http://wiki.tcl.tk/21595#pagetoce30e53a1
  set argc [llength [info level 0]]
  if {$argc == 4} {
    uplevel 1 [list self method $name $args $body]
  } elseif {$argc == 3} {
    return -code error "wrong # args: should be \"[lindex [info level 0] 0]\
                                      name ?args body?\""
  }

  # Get the name of the current class
  set cls [lindex [info level -1] 1]

  # Get its private "my" command
  set my [info object namespace $cls]::my

  # Make the connection by forwarding
  tailcall forward $name $my $name
};# ::oo::define::classmethod

##
## Define the method "common" for all classes!
##
proc ::oo::define::common {varname args} {
  if {[llength $args] > 1} {
    return -code error "wrong # args: should be \"[lindex [info level 0] 0]\
                                      varName ?init?\""
  }
  # Get the name of the current class
  set cls [lindex [info level -1] 1]

  # Export method varname
  oo::define $cls self export varname

  # Initialise the variable
  if {[llength $args]} {
    set [$cls varname $varname] [lindex $args 0]
  }
};# ::oo::define::common

oo::define oo::object method common {args} {
  if {![llength $args]} return
  set callclass [lindex [self caller] 0]
  oo::define $callclass self export varname
  foreach vname $args {
    lappend pairs [$callclass varname $vname] $vname
  }
  uplevel 1 upvar {*}$pairs
};# common

##
## Define cget/configure for all classes!
##
oo::define oo::object method cget {_elep_oo_variable_name} {
  set _elep_oo_variable_name [string range $_elep_oo_variable_name 1 end]
  my variable $_elep_oo_variable_name
  return [set $_elep_oo_variable_name]
};# cget

oo::define oo::object method configure {args} {
  foreach {_elep_oo_variable_name _elep_oo_variable_value} $args {
    set _elep_oo_variable_name [string range $_elep_oo_variable_name 1 end]
    my variable $_elep_oo_variable_name
    set $_elep_oo_variable_name $_elep_oo_variable_value
  }
};# configure

##
## Define msg/msgVar/opt/optVar procedures...
##

## ::ELEP::Base::msg
#  Return the text that corresponds to the provided message code (key).
proc ::ELEP::Base::msg {code} {
  variable Text
  if {![info exists Text($code)]} {
    if {[info exists ::GUI_Text($code)]} {
      set Text($code) $::GUI_Text($code)
    } else {
      set Text($code) $code
    }
    Log "Warning: Message code \"$code\" undefined for language\
         [opt MsgLanguage]."
  }
  return $Text($code)
};# ::ELEP::Base::msg

## ::ELEP::Base::msgVar
#  Return the variable name that corresponds to the provided message code (key).
proc ::ELEP::Base::msgVar {code} {
  variable Text
  if {![info exists Text($code)]} {
    if {[info exists ::GUI_Text($code)]} {
      set Text($code) $::GUI_Text($code)
    } else {
      set Text($code) $code
    }
    Log "Warning: Message code \"$code\" undefined for language\
         [opt MsgLanguage]."
  }
  return ::ELEP::Base::Text($code)
};# ::ELEP::Base::msgVar

proc ::ELEP::Base::opt {code} {
  variable Option
  if {![info exists Option($code)]} {
    if {[info exists ::GUI_Options($code)]} {
      set Option($code) $::GUI_Options($code)
    } else {
      switch -exact $code {
        MsgLanguage {set Option($code) $::Locale}
        default {
          set Option($code) $code
        }
      }
    }
  }
  return $Option($code)
};# ::ELEP::Base::opt

proc ::ELEP::Base::optVar {code} {
  variable Option
  if {![info exists Option($code)]} {
    opt $code
  }
  return ::ELEP::Base::Option($code)
};# ::ELEP::Base::optVar

proc ::ELEP::Base::Log {msg args} {
  puts "Log: $msg"
};# ::ELEP::Base::Log

proc ::ELEP::Base::loadImage {image args} {
  return [GUI_LoadImage $image]
};# ::ELEP::Base::loadImage

##
## Class: ::ELEP::Base::Utilities
##
#  This class can be mixed-in, to get methods like msg, msgVar, opt, optVar,
#  Log, etc.

oo::class create ELEP::Base::Utilities {

  classmethod _dir {path}  {
    set path [file normalize $path]
    if {![file isdirectory $path]} {
      catch {file mkdir $path}
    }
    return  $path
  };# _dir

  classmethod loadImage       {image} {return [::ELEP::Base::loadImage $image]}
  classmethod msg             {code}  {return [::ELEP::Base::msg       $code]}
  classmethod msgVar          {code}  {return [::ELEP::Base::msgVar    $code]}
  classmethod opt             {code}  {return [::ELEP::Base::opt       $code]}
  classmethod optVar          {code}  {return [::ELEP::Base::optVar    $code]}
  classmethod userHomeDir          {} {return [my _dir $::UserHomeDir]}
  classmethod userAppDir           {} {return [my _dir $::UserAppDir]}
  classmethod configurationDir     {} {return [my _dir $::HomeDir/share/Config]}
  classmethod userConfigurationDir {} {return [my _dir \
                                                  $::UserAppDir/Config]}

};# class ELEP::Base::Utilities

package provide ELEP::Base::Utilities 1.0
