##############################################################################
##  CollectionSynchroniser.tcl:                                             ##
## --------------------------------------------------                       ##
##  This package ensures synchronisation of Collections with a              ##
##  synchronisation server (through a proper driver).                       ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities

## Set a variable that tracks whether dispatcher has been installed for this
## interpreter...
if {![info exists ELEP::CollectionUtilities::dispatcherInstalled]} {
  set ELEP::CollectionUtilities::dispatcherInstalled 0
  set ELEP::CollectionUtilities::dispatcherDrivers   {}
}

## The set of commands that will be dispatched (synchronised)...
set ELEP::CollectionUtilities::dispatchedCommands {
  tip_CreateCollection
  tip_CreateDocument
  tip_RemoveDocument
  tip_OpenCollection
  tip_Sync
  tip_Close
  tip_Destroy

  tip_GetDocument
  tip_FirstDocument
  tip_NextDocument

  GUI_GetCollectionNames
  GUI_GetCollectionDocuments
  GUI_IsCollectionOpened
  GUI_FixCollectionName
};# variable ELEP::CollectionUtilities::dispatchedCommands

## The command dispatcher...
proc ELEP::CollectionUtilities::cmdDispatcher {cmd args} {
  variable dispatcherDrivers
  if {[llength $dispatcherDrivers]} {
    foreach one $dispatcherDrivers {
      set result [$one dispatch $cmd {*}$args]
    }
    return $result
  }
  return [executeCommand $cmd {*}$args]
};# ELEP::CollectionUtilities::cmdDispatcher

proc ELEP::CollectionUtilities::executeCommand {cmd args} {
  return [ELEP::CollectionUtilities::$cmd {*}$args]
};# ELEP::CollectionUtilities::executeCommand

## Installs the command dispatcher in this interpreter...
proc ELEP::CollectionUtilities::installDispatcher {} {
  variable dispatcherInstalled
  if {$dispatcherInstalled} {return}
  set dispatcherInstalled 1

  variable dispatchedCommands
  variable DispatchedCommandTokens

  foreach cmd $dispatchedCommands {
    catch {::$cmd};# This causes the command to get loaded, if it is a proc
    rename ::$cmd ELEP::CollectionUtilities::$cmd
    set DispatchedCommandTokens($cmd) \
      [interp alias {} ::$cmd {} ELEP::CollectionUtilities::cmdDispatcher $cmd]
  }
};# ELEP::CollectionUtilities::installDispatcher

## Uninstalls the command dispatcher from this interpreter...
proc ELEP::CollectionUtilities::uninstallDispatcher {} {
  variable dispatcherInstalled
  if {!$dispatcherInstalled} {return}
  set dispatcherInstalled 0

  variable dispatchedCommands
  variable DispatchedCommandTokens

  foreach cmd $dispatchedCommands {
    ## Delete the alias...
    interp alias {} $DispatchedCommandTokens($cmd) {}
    rename ELEP::CollectionUtilities::$cmd ::$cmd
  }
};# ELEP::CollectionUtilities::uninstallDispatcher

## Register a driver object. The object must be an
## ELEP::CollectionUtilities::CollectionSynchroniser object.
proc ELEP::CollectionUtilities::registerDriver {object} {
  variable dispatcherDrivers
  lappend dispatcherDrivers $object
};# ELEP::CollectionUtilities::registerDriver

## Unregister a driver object. The object must be an
## ELEP::CollectionUtilities::CollectionSynchroniser object.
proc ELEP::CollectionUtilities::unregisterDriver {object} {
  variable dispatcherDrivers
  set index [lsearch -exact $dispatcherDrivers $object]
  if {$index == -1} {return $dispatcherDrivers}
  set dispatcherDrivers [lreplace $dispatcherDrivers $index $index]
};# ELEP::CollectionUtilities::unregisterDriver

ELEP::CollectionUtilities::installDispatcher

oo::class create ELEP::CollectionUtilities::CollectionSynchroniser {
  mixin ELEP::Base::Utilities

  ##########################################################################
  ## Method: constructor
  ##########################################################################
  constructor {args} {
    ELEP::CollectionUtilities::registerDriver [self]
  };# constructor

  ##########################################################################
  ## Method: dispatch
  ##########################################################################
  method dispatch {cmd args} {
    my $cmd {*}$args
  };# dispatch

  ##########################################################################
  ## Method: execute
  ##########################################################################
  method execute {cmd args} {
    return [ELEP::CollectionUtilities::executeCommand $cmd {*}$args]
  };# execute

  ##########################################################################
  ## Method: destructor
  ##########################################################################
  destructor {
    ELEP::CollectionUtilities::unregisterDriver [self]
  };# destructor

  ##########################################################################
  ## Method: showProgress
  ##########################################################################
  method showProgress {msg} {
    catch {
      package require ELEP::WindowUtilities::ProgressWidget
      my variable progress_object
      set focus_win [focus]
      set toplevel [winfo toplevel $focus_win]

      set progress_object [ELEP::WindowUtilities::ProgressWidget create \
        [ELEP::Base::auto ELEP::CollectionUtilities::Pr] $toplevel]
      $progress_object message $msg
      $progress_object placeInToplevelCenter
    }
  };# showProgress

  ##########################################################################
  ## Method: progress
  ##########################################################################
  method progress {value} {
    catch {
      my variable progress_object
      $progress_object progress $value
    }
  };# progress

  ##########################################################################
  ## Method: advance
  ##########################################################################
  method advance {{value 1}} {
    catch {
      my variable progress_object
      $progress_object advanceWithStep $value
    }
  };# advance

  ##########################################################################
  ## Method: hideProgress
  ##########################################################################
  method hideProgress {} {
    catch {
      my variable progress_object
      $progress_object progress 100
      $progress_object destroy
    }
  };# hideProgress

  ##########################################################################
  ## Method: tip_CreateCollection
  ##########################################################################
  method tip_CreateCollection {args} {
    return [my execute tip_CreateCollection {*}$args]
  };# tip_CreateCollection

  ##########################################################################
  ## Method: tip_CreateDocument
  ##########################################################################
  method tip_CreateDocument {args} {
    return [my execute tip_CreateDocument {*}$args]
  };# tip_CreateDocument

  ##########################################################################
  ## Method: tip_RemoveDocument
  ##########################################################################
  method tip_RemoveDocument {args} {
    return [my execute tip_RemoveDocument {*}$args]
  };# tip_RemoveDocument

  ##########################################################################
  ## Method: tip_OpenCollection
  ##########################################################################
  method tip_OpenCollection {args} {
    return [my execute tip_OpenCollection {*}$args]
  };# tip_OpenCollection

  ##########################################################################
  ## Method: tip_Sync
  ##########################################################################
  method tip_Sync {args} {
    return [my execute tip_Sync {*}$args]
  };# tip_Sync

  ##########################################################################
  ## Method: tip_Close
  ##########################################################################
  method tip_Close {args} {
    return [my execute tip_Close {*}$args]
  };# tip_Close

  ##########################################################################
  ## Method: tip_Destroy
  ##########################################################################
  method tip_Destroy {args} {
    return [my execute tip_Destroy {*}$args]
  };# tip_Destroy

  ##########################################################################
  ## Method: tip_GetDocument
  ##########################################################################
  method tip_GetDocument {args} {
    return [my execute tip_GetDocument {*}$args]
  };# tip_GetDocument

  ##########################################################################
  ## Method: tip_FirstDocument
  ##########################################################################
  method tip_FirstDocument {args} {
    return [my execute tip_FirstDocument {*}$args]
  };# tip_FirstDocument

  ##########################################################################
  ## Method: tip_NextDocument
  ##########################################################################
  method tip_NextDocument {args} {
    return [my execute tip_NextDocument {*}$args]
  };# tip_NextDocument

  ##########################################################################
  ## Method: GUI_GetCollectionNames
  ##########################################################################
  method GUI_GetCollectionNames {args} {
    return [my execute GUI_GetCollectionNames {*}$args]
  };# GUI_GetCollectionNames

  ##########################################################################
  ## Method: GUI_GetCollectionDocuments
  ##########################################################################
  method GUI_GetCollectionDocuments {args} {
    return [my execute GUI_GetCollectionDocuments {*}$args]
  };# GUI_GetCollectionDocuments

  ##########################################################################
  ## Method: GUI_IsCollectionOpened
  ##########################################################################
  method GUI_IsCollectionOpened {args} {
    return [my execute GUI_IsCollectionOpened {*}$args]
  };# GUI_IsCollectionOpened

  ##########################################################################
  ## Method: GUI_FixCollectionName
  ##########################################################################
  method GUI_FixCollectionName {args} {
    return [my execute GUI_FixCollectionName {*}$args]
  };# GUI_FixCollectionName

};# class ELEP::CollectionUtilities::CollectionSynchroniser

package provide ELEP::CollectionUtilities::CollectionSynchroniser 1.0
