##############################################################################
##  CollectionSynchroniserPostgreSQL.tcl:                                   ##
## --------------------------------------------------                       ##
##  This class implements a synchronisation driver for PostgreSQL database. ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::CollectionUtilities::CollectionSynchroniser
package require ELEP::Database::PostgreSQL

oo::class create ELEP::CollectionUtilities::CollectionSynchroniserPostgreSQL {
  superclass ELEP::CollectionUtilities::CollectionSynchroniser
  mixin      ELEP::Database::PostgreSQL

  ##########################################################################
  ## Method: constructor
  ##########################################################################
  constructor {args} {
    my init_database_object
    my configure {*}$args
    next
  };# constructor

  ##########################################################################
  ## Method: destructor
  ##########################################################################
  destructor {
    catch {my disconnect}
    next
  };# destructor

  ##########################################################################
  ## Method: tip_CreateCollection
  ##########################################################################
  method tip_CreateCollection {args} {
    return [my execute tip_CreateCollection {*}$args]
  };# tip_CreateCollection

  ##########################################################################
  ## Method: tip_CreateDocument
  ##########################################################################
  method tip_CreateDocument {args} {
    return [my execute tip_CreateDocument {*}$args]
  };# tip_CreateDocument

  ##########################################################################
  ## Method: tip_RemoveDocument
  ##########################################################################
  method tip_RemoveDocument {args} {
    return [my execute tip_RemoveDocument {*}$args]
  };# tip_RemoveDocument

  ##########################################################################
  ## Method: tip_OpenCollection
  ##########################################################################
  method tip_OpenCollection {args} {
    ## Update the whole collection from the database...
    my addCollectionFromDatabase [file tail [lindex $args 0]]
    return [my execute tip_OpenCollection {*}$args]
  };# tip_OpenCollection

  ##########################################################################
  ## Method: tip_Sync
  ##########################################################################
  method tip_Sync {args} {
    set result [my execute tip_Sync {*}$args]
    switch -glob [lindex $args 0] {
      Doc* {
        my showProgress [my msg {Saving Document into the database...}]
        ## Get the Collection id...
        set doc     [lindex $args 0]
        set id      [my getCollectionIdFromDoc $doc]
        set docname [tip_GetId   $doc]
        set colname [tip_GetName [tip_GetParent $doc]]
        my advance 30
        my sql "UPDATE documents SET\
           data='[my conninfo_quote [my readDocumentData $colname $docname]]'\
           WHERE collection=$id AND name='$docname'"
        my hideProgress
      }
      Col* {
        my showProgress [my msg {Saving Collection into the database...}]
        set col     [lindex $args 0]
        set colname [tip_GetName $col]
        my sql "UPDATE collections SET\
                  data='[my conninfo_quote [my readCollectionData $colname]]'\
                WHERE name='[file tail $colname]'"
        my hideProgress
      }
    }
    return $result
  };# tip_Sync

  ##########################################################################
  ## Method: tip_Close
  ##########################################################################
  method tip_Close {args} {
    switch -glob [lindex $args] {
      Doc* {
        my setDocumentOpened [lindex $args 0] FALSE
      }
      default {}
    }
    return [my execute tip_Close {*}$args]
  };# tip_Close

  ##########################################################################
  ## Method: tip_Destroy
  ##########################################################################
  method tip_Destroy {args} {
    return [my execute tip_Destroy {*}$args]
  };# tip_Destroy

  ##########################################################################
  ## Method: tip_GetDocument
  ##########################################################################
  method tip_GetDocument {args} {
    my showProgress [my msg {Retrieving Document from the the database...}]
    foreach {col doc_id} $args {break}
    set col_name [tip_GetName $col]
    set col_id   [my getCollectionId [file tail $col_name]]
    my saveDocumentData $col_name $doc_id $col_id 
    my progress 90
    set result [my execute tip_GetDocument {*}$args]
    my hideProgress
    my setDocumentOpened $result
    return $result
  };# tip_GetDocument

  ##########################################################################
  ## Method: tip_FirstDocument
  ##########################################################################
  method tip_FirstDocument {args} {
    set result [my execute tip_FirstDocument {*}$args]
    my setDocumentOpened $result
    return $result
  };# tip_FirstDocument

  ##########################################################################
  ## Method: tip_NextDocument
  ##########################################################################
  method tip_NextDocument {args} {
    set result [my execute tip_NextDocument {*}$args]
    my setDocumentOpened $result
    return $result
  };# tip_NextDocument

  method getCollectionIdFromDoc {doc} {
    set colname [tip_GetName [tip_GetParent $doc]]
    set coltail [file tail $colname]
    return [my getCollectionId $coltail]
  };# getCollectionIdFromDoc

  method getCollectionId {name} {
    set id [my sql "SELECT id FROM collections WHERE name='$name'"]
    return [lindex $id 0]
  };# getCollectionId

  method getDocumentOpened {doc} {
    if {$doc eq ""} {return 0}
    set id      [my getCollectionIdFromDoc $doc]
    if {$id  eq ""} {return 0}
    set docname [tip_GetId $doc]
    set value [my sql \
      "SELECT open FROM documents WHERE collection=$id AND name='$docname'"]
    return [lindex $value 0 0]
  };# getDocumentOpened

  method setDocumentOpened {doc {value TRUE}} {
    if {[my getDocumentOpened $doc] && $value} {
      error "[my msg {Document is already open in another Annotation Tool!}]\
             \n[tip_GetId $doc]"
    }
    if {$doc eq ""} {return}
    set id      [my getCollectionIdFromDoc $doc]
    if {$id  eq ""} {return}
    set docname [tip_GetId $doc]
    my sql "UPDATE documents SET open=$value WHERE\
            collection=$id AND name='$docname'"
  };# setDocumentOpened

  ##########################################################################
  ## Method: GUI_GetCollectionNames
  ##########################################################################
  method GUI_GetCollectionNames {args} {
    set collections_disk [my execute GUI_GetCollectionNames]
    set collections_db   [my sql "SELECT name FROM collections"]
    ## Compare the two lists...
    foreach one $collections_disk {
      set name [file tail    $one]
      set index [lsearch -exact -index 0 $collections_db $name]
      if {$index == -1} {
        ## The collection is not in the database. Add it...
        my addCollectionToDatabase $one
      } else {
        ## Ensure that the collection index is fresh...
        my saveCollectionData $one $name
      }
    }
    foreach one $collections_db {
      set name [lindex $one 0]
      set index [lsearch -glob $collections_disk */$name]
      if {$index == -1} {
        ## The collection is not on disk. Fetch it from the database...
        my addCollectionFromDatabase $name
      }
    }
    return [my execute GUI_GetCollectionNames {*}$args]
  };# GUI_GetCollectionNames

  method addCollectionToDatabase {name} {
    set collection_name [file tail $name]
    my showProgress [my msg {Saving Collection to the database...}]
    ## Add Collection...
    my sql "INSERT INTO collections(name, data)\
            VALUES('$collection_name',\
                   '[my conninfo_quote [my readCollectionData $name]]')"
    ## Get id...
    set id [my getCollectionId $collection_name]
    ## Add Collection Documents...
    set docs [my execute GUI_GetCollectionDocuments $name]
    set step [expr {100.0/([llength $docs]+1)}]
    foreach doc $docs {
      my advance $step
      my sql "INSERT INTO documents(collection, name, data)\
              VALUES($id, '$doc',\
              '[my conninfo_quote [my readDocumentData $name $doc]]')"
    }
    my hideProgress
  };# addCollectionToDatabase

  method readCollectionData {collection} {
    return [my readDocumentData $collection .CDM_CollectionIndex.txt]
  };# readCollectionData

  method readDocumentData {collection name} {
    set fd [open $collection/$name]
    fconfigure $fd -encoding utf-8
    set data [string trim [read $fd]]
    close $fd
    return $data
  };# readDocumentData

  method addCollectionFromDatabase {name} {
    my showProgress [my msg {Updating Collection from the database...}]
    set path [lindex $::CollectionPath 0]
    set collection_name [file join $path $name]
    file mkdir $collection_name
    ## Add Collection...
    set id [my saveCollectionData $collection_name $name]
    ## Add Collection Documents...
    set docs [my sql "SELECT name FROM documents WHERE collection=$id"]
    set step [expr {100.0/([llength $docs]+1)}]
    foreach one $docs {
      my advance $step
      set document_name [lindex $one 0]
      my saveDocumentData $collection_name $document_name $id
    }
    my hideProgress
  };# addCollectionToDatabase

  method saveCollectionData {path name} {
    set result  [my sql "SELECT id, data FROM collections WHERE name='$name'"]
    set id   [lindex $result 0 0]
    set data [lindex $result 0 1]
    set fd   [open $path/.CDM_CollectionIndex.txt w]
    fconfigure $fd -encoding utf-8
    puts $fd [string trim $data]
    close $fd
    return $id
  };# saveDocumentData

  method saveDocumentData {collection name col_id} {
    set data [my sql "SELECT data FROM documents WHERE collection=$col_id\
                      AND name='$name'"]
    set data [lindex $data 0 0]
    set fd [open $collection/$name w]
    fconfigure $fd -encoding utf-8
    puts $fd [string trim $data]
    close $fd
  };# saveDocumentData

  ##########################################################################
  ## Method: GUI_GetCollectionDocuments
  ##########################################################################
  method GUI_GetCollectionDocuments {args} {
    return [my execute GUI_GetCollectionDocuments {*}$args]
  };# GUI_GetCollectionDocuments

  ##########################################################################
  ## Method: GUI_IsCollectionOpened
  ##########################################################################
  method GUI_IsCollectionOpened {args} {
    return [my execute GUI_IsCollectionOpened {*}$args]
  };# GUI_IsCollectionOpened

  ##########################################################################
  ## Method: GUI_FixCollectionName
  ##########################################################################
  method GUI_FixCollectionName {args} {
    set result [my execute GUI_FixCollectionName {*}$args]
    my tip_Sync [lindex $args 0]
    return $result
  };# GUI_FixCollectionName

};# class ELEP::CollectionUtilities::CollectionSynchroniserPostgreSQL

package provide ELEP::CollectionUtilities::CollectionSynchroniserPostgreSQL 1.0
