##############################################################################
##  CollectionSynchroniserTDBC.tcl:                                         ##
## --------------------------------------------------                       ##
##  This class implements a synchronisation driver for TDBC database.       ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::CollectionUtilities::CollectionSynchroniser
package require ELEP::Database::TDBC
package require md5

oo::class create ELEP::CollectionUtilities::CollectionSynchroniserTDBC {
  superclass ELEP::CollectionUtilities::CollectionSynchroniser
  mixin      ELEP::Database::TDBC

  ##########################################################################
  ## Method: constructor
  ##########################################################################
  constructor {args} {
    my init_database_object
    my configure {*}$args
    next
  };# constructor

  ##########################################################################
  ## Method: destructor
  ##########################################################################
  destructor {
    catch {my disconnect}
    next
  };# destructor

  ##########################################################################
  ## Method: tip_CreateCollection
  ##########################################################################
  method tip_CreateCollection {args} {
    set col [my execute tip_CreateCollection {*}$args]
    set collection_name [file tail [tip_GetName $col]]
    set col_data {}
    my sql "INSERT INTO collections(name, data)\
            VALUES( :collection_name , :col_data )"
    return $col
  };# tip_CreateCollection

  ##########################################################################
  ## Method: tip_CreateDocument
  ##########################################################################
  method tip_CreateDocument {args} {
    return [my execute tip_CreateDocument {*}$args]
  };# tip_CreateDocument

  ##########################################################################
  ## Method: tip_RemoveDocument
  ##########################################################################
  method tip_RemoveDocument {args} {
    return [my execute tip_RemoveDocument {*}$args]
  };# tip_RemoveDocument

  ##########################################################################
  ## Method: tip_OpenCollection
  ##########################################################################
  method tip_OpenCollection {args} {
    ## Update the whole collection from the database...
    my addCollectionFromDatabase [file tail [lindex $args 0]]
    return [my execute tip_OpenCollection {*}$args]
  };# tip_OpenCollection

  ##########################################################################
  ## Method: tip_Sync
  ##########################################################################
  method tip_Sync {args} {
    set result [my execute tip_Sync {*}$args]
    switch -glob [lindex $args 0] {
      Doc* {
        my showProgress [my msg {Saving Document into the database...}]
        ## Get the Collection id...
        set doc     [lindex $args 0]
        set id      [my getCollectionIdFromDoc $doc]
        set docname [tip_GetId   $doc]
        set colname [tip_GetName [tip_GetParent $doc]]
        ## Is the document in database?
        my advance 10
        set exists [my sql "SELECT id FROM documents WHERE\
                            collection = :id AND name = :docname"]
        my advance 30
        set doc_data [my readDocumentData $colname $docname]
        set md5      [::md5::md5 -hex $doc_data]
        if {[llength $exists]} {
          my sql "UPDATE documents SET data = :doc_data , md5 = :md5 WHERE\
                         collection = :id AND name = :docname"
        } else {
          my sql "INSERT INTO documents(collection, name, data, md5)\
                  VALUES( :id , :docname , :doc_data , :md5 )"
        }
        my hideProgress
      }
      Col* {
        my showProgress [my msg {Saving Collection into the database...}]
        set col         [lindex $args 0]
        set colname     [tip_GetName $col]
        set col_data    [my readCollectionData $colname]
        set md5         [::md5::md5 -hex $col_data]
        set tail        [file tail $colname]
        set exists [my sql "SELECT id FROM collections WHERE name = :tail"]
        if {[llength $exists]} {
          my sql "UPDATE collections SET data = :col_data , md5 = :md5 WHERE\
                         name = :tail"
        } else {
          my sql "INSERT INTO collections(name, data, md5)\
                  VALUES( :tail , :col_data , :md5 )"
        }
        my hideProgress
      }
    }
    return $result
  };# tip_Sync

  ##########################################################################
  ## Method: tip_Close
  ##########################################################################
  method tip_Close {args} {
    switch -glob [lindex $args] {
      Doc* {
        my setDocumentOpened [lindex $args 0] FALSE
      }
      default {}
    }
    return [my execute tip_Close {*}$args]
  };# tip_Close

  ##########################################################################
  ## Method: tip_Destroy
  ##########################################################################
  method tip_Destroy {args} {
    set collection_name      [lindex $args 0]
    set collection_name_tail [file tail $collection_name]
    ## Flag the collection as deleted in the database...
    my sql "UPDATE collections SET deleted = true WHERE\
            name = :collection_name_tail"
    ## Remove all documents...
    set id [my getCollectionId $collection_name_tail]
    my sql {DELETE FROM documents WHERE collection = :id}
    return [my execute tip_Destroy {*}$args]
  };# tip_Destroy

  ##########################################################################
  ## Method: tip_GetDocument
  ##########################################################################
  method tip_GetDocument {args} {
    my showProgress [my msg {Retrieving Document from the the database...}]
    foreach {col doc_id} $args {break}
    set col_name [tip_GetName $col]
    set col_id   [my getCollectionId [file tail $col_name]]
    my saveDocumentData $col_name $doc_id $col_id 
    my progress 90
    set result [my execute tip_GetDocument {*}$args]
    my hideProgress
    my setDocumentOpened $result
    return $result
  };# tip_GetDocument

  ##########################################################################
  ## Method: tip_FirstDocument
  ##########################################################################
  method tip_FirstDocument {args} {
    set result [my execute tip_FirstDocument {*}$args]
    my setDocumentOpened $result
    return $result
  };# tip_FirstDocument

  ##########################################################################
  ## Method: tip_NextDocument
  ##########################################################################
  method tip_NextDocument {args} {
    set result [my execute tip_NextDocument {*}$args]
    my setDocumentOpened $result
    return $result
  };# tip_NextDocument

  method getCollectionIdFromDoc {doc} {
    set colname [tip_GetName [tip_GetParent $doc]]
    set coltail [file tail $colname]
    return [my getCollectionId $coltail]
  };# getCollectionIdFromDoc

  method getCollectionId {name} {
    set id [my sql {SELECT id FROM collections WHERE name = :name}]
    return [lindex $id 0 0]
  };# getCollectionId

  method getDocumentOpened {doc} {
    if {$doc eq ""} {return 0}
    set id      [my getCollectionIdFromDoc $doc]
    if {$id  eq ""} {return 0}
    set docname [tip_GetId $doc]
    set value [my sql \
      "SELECT open, \"user\" FROM documents WHERE\
              collection = :id AND name = :docname"]
    if {$value eq ""} {return 0}
    return [lindex $value 0]
  };# getDocumentOpened

  method getUser {} {
    global env
    foreach var {USER USERNAME user} {
      if {[info exists env($var)]} {return $env($var)}
    }
    return unknown_user
  };# getUser

  method setDocumentOpened {doc {value TRUE}} {
    lassign [my getDocumentOpened $doc] _open _user
    if {$_open && $value} {
      error \
       "[my msg {Document is already open in another Annotation Tool by user:}]\
                 $_user\n[my msg Document:] [tip_GetId $doc]"
    }
    if {$doc eq ""} {return}
    set id      [my getCollectionIdFromDoc $doc]
    if {$id  eq ""} {return}
    set docname [tip_GetId $doc]
    set user [my getUser]
    my sql "UPDATE documents SET open = :value , \"user\" = :user WHERE\
            collection = :id AND name = :docname"
  };# setDocumentOpened

  ##########################################################################
  ## Method: GUI_GetCollectionNames
  ##########################################################################
  method GUI_GetCollectionNames {args} {
    set collections_disk [my execute GUI_GetCollectionNames]
    set collections_db         {}
    set collections_db_deleted {}
    my foreach -as lists -- row {SELECT name, deleted FROM collections} {
      lassign $row name deleted
      lappend collections_db         $name
      lappend collections_db_deleted $deleted
    }
    ## Compare the two lists...
    foreach one $collections_disk {
      set name [file tail    $one]
      set index [lsearch -exact $collections_db $name]
      if {$index == -1} {
        ## The collection is not in the database. Add it...
        my addCollectionToDatabase $one
      } else {
        ## Is it a deleted collection?
        set deleted [lindex $collections_db_deleted $index]
        if {$deleted} {
          my execute tip_Destroy $one
        } else {
          ## Ensure that the collection index is fresh...
          my saveCollectionData $one $name
        }
      }
    }
    foreach name $collections_db deleted $collections_db_deleted {
      if {$deleted} {continue}
      set index [lsearch -glob $collections_disk */$name]
      if {$index == -1} {
        ## The collection is not on disk. Fetch it from the database...
        my addCollectionFromDatabase $name
      }
    }
    return [my execute GUI_GetCollectionNames {*}$args]
  };# GUI_GetCollectionNames

  method addCollectionToDatabase {name} {
    set collection_name [file tail $name]
    my showProgress [my msg {Saving Collection to the database...}]
    ## Add Collection...
    set col_data [my readCollectionData $name]
    set md5      [::md5::md5 -hex $col_data]
    my sql "INSERT INTO collections(name, data, md5)\
            VALUES( :collection_name , :col_data , :md5)"
    ## Get id...
    set id [my getCollectionId $collection_name]
    ## Add Collection Documents...
    set docs [my execute GUI_GetCollectionDocuments $name]
    set step [expr {100.0/([llength $docs]+1)}]
    foreach doc $docs {
      my advance $step
      set doc_data [my readDocumentData $name $doc]
      set md5      [::md5::md5 -hex $doc_data]
      my sql "INSERT INTO documents(collection, name, data, md5)\
              VALUES( :id , :doc , :doc_data , :md5 )"
    }
    my hideProgress
  };# addCollectionToDatabase

  method readCollectionData {collection} {
    return [my readDocumentData $collection .CDM_CollectionIndex.txt]
  };# readCollectionData

  method readDocumentData {collection name} {
    set fd [open $collection/$name]
    fconfigure $fd -encoding utf-8
    set data [string trim [read $fd]]
    close $fd
    return $data
  };# readDocumentData

  method addCollectionFromDatabase {name} {
    my showProgress [my msg {Updating Collection from the database...}]
    set path [lindex $::CollectionPath 0]
    set collection_name [file join $path $name]
    file mkdir $collection_name
    ## Add Collection...
    set id [my saveCollectionData $collection_name $name]
    ## Add Collection Documents...
    set docs [my sql "SELECT name FROM documents WHERE collection = :id"]
    set step [expr {100.0/([llength $docs]+1)}]
    foreach one $docs {
      my advance $step
      set document_name [lindex $one 0]
      my saveDocumentData $collection_name $document_name $id
    }
    my hideProgress
  };# addCollectionToDatabase

  method saveCollectionData {path name} {
    set result [my sql {SELECT id, md5 FROM collections WHERE name = :name}]
    lassign    [lindex $result 0] id md5

    set collection_index $path/.CDM_CollectionIndex.txt
    if {[file exists $collection_index]} {
      set data [my readCollectionData $path]
      if {$data ne ""} {
        set file_md5 [::md5::md5 -hex $data]
        if {$file_md5 eq $md5} {return $id}
      }
    }
    set result [my sql {SELECT data FROM collections WHERE id = :id}]
    set data   [string trim [lindex $result 0 0]]
    set fd     [open $collection_index w]
    fconfigure $fd -encoding utf-8
    puts $fd $data
    close $fd
    ## Ensure that the collection md5 is not empty!
    if {$md5 eq ""} {
      set md5 [::md5::md5 -hex $data]
      my sql {UPDATE collections SET md5 = :md5 WHERE id = :id}
    }
    return $id
  };# saveCollectionData

  method saveDocumentData {collection name col_id} {
    set result [my sql "SELECT id, md5 FROM documents WHERE\
                               collection = :col_id AND name = :name"]
    lassign    [lindex $result 0] id md5
    
    set document_file_name $collection/$name
    if {[file exists $document_file_name]} {
      set data [my readDocumentData $collection $name]
      if {$data ne ""} {
        set file_md5 [::md5::md5 -hex $data]
        if {$file_md5 eq $md5} {return $id}
      }
    }
    set result [my sql {SELECT data FROM documents WHERE id = :id}]
    set data   [string trim [lindex $result 0 0]]
    set fd     [open $document_file_name w]
    fconfigure $fd -encoding utf-8
    puts $fd $data
    close $fd
    ## Ensure that the document md5 is not empty!
    if {$md5 eq ""} {
      set md5 [::md5::md5 -hex $data]
      my sql {UPDATE documents SET md5 = :md5 WHERE id = :id}
    }
    return $id
  };# saveDocumentData

  ##########################################################################
  ## Method: GUI_GetCollectionDocuments
  ##########################################################################
  method GUI_GetCollectionDocuments {args} {
    return [my execute GUI_GetCollectionDocuments {*}$args]
  };# GUI_GetCollectionDocuments

  ##########################################################################
  ## Method: GUI_IsCollectionOpened
  ##########################################################################
  method GUI_IsCollectionOpened {args} {
    return [my execute GUI_IsCollectionOpened {*}$args]
  };# GUI_IsCollectionOpened

  ##########################################################################
  ## Method: GUI_FixCollectionName
  ##########################################################################
  method GUI_FixCollectionName {args} {
    set result [my execute GUI_FixCollectionName {*}$args]
    my tip_Sync [lindex $args 0]
    return $result
  };# GUI_FixCollectionName

};# class ELEP::CollectionUtilities::CollectionSynchroniserTDBC

package provide ELEP::CollectionUtilities::CollectionSynchroniserTDBC 1.0
