##############################################################################
##  CollectionsViewer.tcl:                                                  ##
## --------------------------------------------------                       ##
##  This class implements a tool for showing Collections/Documents, and     ##
##  allowing user to select a Collection/Document.                          ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities::WidgetBase
package require ELEP::WindowUtilities::ObjectListenerServer
package require tablelist_tile

GUI_LoadImage IFolder
GUI_LoadImage ICollection
GUI_LoadImage IDocumentSmall

oo::class create ELEP::CollectionUtilities::CollectionsViewer {
  superclass ELEP::WindowUtilities::WidgetBase \
             ELEP::WindowUtilities::ObjectListenerServer

  variable collection_names_widget \
           collection_documents_widget \
           selected_collection_name \
           selected_document_name

  constructor {parent_win {orientation horizontal}} {
    next ttk::frame [::ELEP::WindowUtilities::UniqueWindowName \
                           $parent_win.collectionsViewer]
    #next onCollectionDocumentSelectionChange
    my initListenerObject onCollectionDocumentSelectionChange
    set selected_collection_name {}
    set selected_document_name   {}
    my createWidgets $orientation
  };# constructor

  method createWidgets {orientation} {
    my variable widget
    ## Add a vertical pane widget.
    ttk::panedwindow $widget.pane -orient $orientation
    ## Add a widget for showing Collection names.
    $widget.pane add [my createCollectionNamesWidget $widget.pane]
    ## Add a widget for showing Collection Documents.
    $widget.pane add [my createCollectionDocumentsViewer $widget.pane]
    pack $widget.pane -fill both -expand 1 -padx 2 -pady 2
    my fillCollectionNamesWidget
  };# createWidgets

  method createCollectionNamesWidget {win} {
    ttk::frame $win.colNames
      ttk::label $win.colNames.label -anchor w -justify left \
        -textvariable GUI_Text(AvailableCollections)
      ScrolledWindow $win.colNames.sw
      set collection_names_widget [$win.colNames.sw getframe].t
      tablelist::tablelist $collection_names_widget -width 50 -height 20 \
        -stripebackground #e0e8f0 -background gray98 -stretch 1 \
        -columns {0 {} 0 {Collection Name} 4 (#) center} \
        -activestyle none -labelcommand {} -showlabels 1 -exportselection 0
      bind $collection_names_widget <<TablelistSelect>> \
        [list [self] onCollectionNameSelectionChange]
      $win.colNames.sw setwidget $collection_names_widget
      grid $win.colNames.label -sticky snew
      grid $win.colNames.sw -sticky snew
    grid columnconfigure $win.colNames 0 -weight 1
    grid rowconfigure    $win.colNames 1 -weight 1
    return $win.colNames
  };# createCollectionNamesWidget

  method createCollectionDocumentsViewer {win} {
    ttk::frame $win.docs
      ttk::label $win.docs.label -anchor w -justify left \
        -textvariable GUI_Text(Collection'sDocuments)
      ScrolledWindow $win.docs.sw
      set collection_documents_widget [$win.docs.sw getframe].t
      tablelist::tablelist $collection_documents_widget -height 10 \
        -stripebackground #e0e8f0 -background gray98 -stretch 0 \
        -columns {0 {Document Name}} \
        -activestyle none -labelcommand {} -showlabels 1 -exportselection 0
      bind $collection_documents_widget <<TablelistSelect>> \
        [list [self] onCollectionDocumentSelectionChange]
      $win.docs.sw setwidget $collection_documents_widget
      grid $win.docs.label -sticky snew
      grid $win.docs.sw -sticky snew
    grid columnconfigure $win.docs 0 -weight 1
    grid rowconfigure    $win.docs 1 -weight 1
    return $win.docs
  };# createCollectionDocumentsViewer

  method fillCollectionNamesWidget {} {
    $collection_names_widget     delete 0 end
    $collection_documents_widget delete 0 end
    set selected_collection_name {}
    set selected_document_name   {}
    foreach col [GUI_GetCollectionNames] {
      lappend path([file dirname $col]) $col
    }
    if {![info exists path]} {return}
    foreach p [lsort [array names path]] {
      $collection_names_widget insert end [list {} $p [llength $path($p)]]
      $collection_names_widget rowconfigure end -selectable 0 \
        -background #f0fff0; #-background #8fbc8f 
      $collection_names_widget cellconfigure end,0 -image IFolder
      foreach col $path($p) {
        if {[catch {GUI_GetCollectionDocuments $col} collection_documents]} {
          set len {invalid collection}
        } else {
          set len [llength $collection_documents]
        }
        $collection_names_widget insert end \
          [list {} [file tail $col] $len]
        $collection_names_widget cellconfigure end,1 -image ICollection
      }
    }
    my notifyListeners
  };# fillCollectionNamesWidget

  method fillCollectionDocumentsViewer {collection_name} {
    if {[catch {GUI_GetCollectionDocuments $collection_name} \
                                            collection_documents]} {
      $collection_documents_widget insert end [list {Invalid Collection!}]
    } else {
      foreach doc $collection_documents {
        $collection_documents_widget insert end [list $doc]
        $collection_documents_widget cellconfigure end,0 -image IDocumentSmall
      }
    }
  };# fillCollectionDocumentsViewer

  ##########################################################################
  ## Event: onCollectionNameSelectionChange
  ##########################################################################
  method onCollectionNameSelectionChange {} {
    set selected_collection_name {}
    set selected_document_name   {}
    $collection_documents_widget delete 0 end
    set sel [$collection_names_widget curselection]
    if {[llength $sel]} {
      set sel [lindex $sel 0]
      set col_name [lindex [$collection_names_widget get $sel] 1]
      ## Retrieve the path of the Collection
      for {} {$sel >=0} {incr sel -1} {
        if {![$collection_names_widget rowcget $sel -selectable]} {
          set col_path [lindex [$collection_names_widget get $sel] 1]
          break
        }
      }
      if {[info exists col_path]} {
        set selected_collection_name [file join $col_path $col_name]
        my fillCollectionDocumentsViewer $selected_collection_name
      }
    }
    my notifyListeners
  };# onCollectionNameSelectionChange

  ##########################################################################
  ## Event: onCollectionDocumentSelectionChange
  ##########################################################################
  method onCollectionDocumentSelectionChange {} {
    set selected_document_name {}
    set sel [$collection_documents_widget curselection]
    if {[llength $sel]} {
      set sel [lindex $sel 0]
      set selected_document_name \
        [lindex [$collection_documents_widget get $sel] 0]
    }
    my notifyListeners
  };# onCollectionDocumentSelectionChange

  ##########################################################################
  ## Method: getSelectedCollectionName
  ##########################################################################
  method getSelectedCollectionName {} {
    return $selected_collection_name
  };# getSelectedCollectionName

  ##########################################################################
  ## Method: getSelectedDocumentName
  ##########################################################################
  method getSelectedDocumentName {} {
   return $selected_document_name
  };# getSelectedDocumentName

  ##########################################################################
  ## Method: openSelectedCollection
  ##########################################################################
  method openSelectedCollection {} {
    if {![string length $selected_collection_name]} {return {}}
    set col [tip_OpenCollection $selected_collection_name]
    ## Is the Collection's location in accordance with it's name?
    set col_name [tip_GetName $col]
    if {![string equal $col_name $selected_collection_name]} {
      GUI_FixCollectionName $col $col_name $selected_collection_name
    }
    return $col
  };# openSelectedCollection

  ##########################################################################
  ## Method: update
  ##########################################################################
  method update {} {
   my fillCollectionNamesWidget
  };# update

  ##########################################################################
  ## Event: onUpdateRequestFromRegisteredListener
  ##########################################################################
  method onUpdateRequestFromRegisteredListener {obj} {
    update
  };# onUpdateRequestFromRegisteredListener

};# class ELEP::CollectionUtilities::CollectionsViewer

package provide ELEP::CollectionUtilities::CollectionsViewer 1.0
