##############################################################################
##  PostgreSQL.tcl:                                                         ##
## --------------------------------------------------                       ##
##  This class implements a class for accessing a PostgreSQL database.      ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require TclOO
package require pgintcl ;# Package to access a PostgreSQL server

oo::class create ELEP::Database::PostgreSQL {

  ##########################################################################
  ## Method: init_database_object
  ##########################################################################
  method init_database_object {} {
    my variable host port user password database db \
                offline error_on_offline

    set host             www.ellogon.org
    set host             localhost
    set port             5432
    set database         ellogon
    set user             ellogon
    set password         ellogon
    set db               {}
    set offline          0
    set error_on_offline 1
  };# init_database_object

  ##########################################################################
  ## Method: connect
  ##########################################################################
  method connect {} {
    my variable host port user password database db
    my disconnect
    set conninfo [list host=$host port=$port dbname=$database \
       user=$user password='[my conninfo_quote $password]']
    # puts $conninfo
    if {[catch {pg_connect -conninfo $conninfo} result]} {
      error "Failed to connect to the database: $result"
    }
    set db $result
  };# connect

  ##########################################################################
  ## Method: disconnect
  ##########################################################################
  method disconnect {} {
    my variable db
    if {[string length $db]} {
      if {[catch {pg_disconnect $db}]} {
        for {set i 0} {$i < 30} {incr i} {
          if {![catch {pg_disconnect $db}]} {
            break
          }
          my_sleep 5000
        };# for {set i 0} {$i < 30} {incr i}
      };# if {[catch {pg_disconnect $db}]}
    }
    set db {}
  };# disconnect

  method offline {} { my variable offline; return $offline }
  method online  {} { my variable offline; return [expr {!$offline}] }

  ##########################################################################
  ## Method: ensure_online
  ##########################################################################
  method ensure_online {} {
    my variable offline host port database
    if {$offline} {
      error "Database \"$host:$port/$database\" is offline!"
    }
  };# ensure_online

  ##########################################################################
  ## Method: conninfo_quote
  ##########################################################################
  method conninfo_quote {s} {
    regsub -all {\\} $s {\\} s
    regsub -all {'}  $s {\'} s
    return $s
  };# conninfo_quote

  ##########################################################################
  ## Method: sql
  ##########################################################################
  method sql {sql} {
    my variable error_on_offline offline db
    # catch {
    #   Log "host: $host, port: $port, db: $database"
    #   Log "  $sql"
    # }
    if {$error_on_offline} {
      my ensure_online
    }
    if {$offline} return {}
    catch {my connect}
    if {![string length $db]} {catch {my connect}}
    # Run the SQL, catch a backend or connection failure.
    if {[catch {pg_exec $db $sql} reply]} {
      catch {my connect}
      ## This query has failed. Try it again for a few times, and if it
      ## succeeds, return the succesful result...
      for {set i 0} {$i < 30} {incr i} {
        catch {my connect}
        if {![catch {pg_exec $db $sql} test_reply]} {
          ## This time the call succeeded!
          catch {my disconnect}
          return [my result $test_reply $sql]
        } else {
          ## We received an error from the database. Wait 60 seconds...
          catch {puts stderr "DB error: $test_reply"}
          catch {my connect}
          my my_sleep 60000
        }
      }
      error "Error while executing SQL: $reply ($database)"
    }
    set result [my result $reply $sql]
    # catch {Log "$result ($reply)"}
    catch {my disconnect}
    return $result
  };# sql

  ##########################################################################
  ## Method: result
  ##########################################################################
  method result {reply sql} {
    set result {}
    set status [pg_result $reply -status]
    switch $status {
      PGRES_COMMAND_OK {}
      PGRES_TUPLES_OK {
        # set ncol [pg_result $reply -numAttrs]
        set nrow [pg_result $reply -numTuples]
        for {set r 0} {$r < $nrow} {incr r} {
          lappend result [pg_result $reply -getTuple $r]
        }
      }
      default {
        set errmsg [pg_result $reply -error]
        pg_result $reply -clear
        error "Error ($status) while executing SQL:\n$sql\n$errmsg"
      }
    }
    pg_result $reply -clear
    return $result
  };# result

  ##########################################################################
  ## Method: my_sleep
  ##########################################################################
  method my_sleep { ms } {
    set uniq [ uniqkey ]
    set ::__sleep__tmp__$uniq 0
    after $ms set ::__sleep__tmp__$uniq 1
    vwait ::__sleep__tmp__$uniq
    unset ::__sleep__tmp__$uniq
  };# my_sleep

  ##########################################################################
  ## Method: uniqkey
  ##########################################################################
  method uniqkey { } {
    set key   [ expr { pow(2,31) + [ clock clicks ] } ]
    set key   [ string range $key end-8 end-3 ]
    set key   [ clock seconds ]$key
    return $key
  };# uniqkey

};# class ELEP::Database::PostgreSQL

package provide ELEP::Database::PostgreSQL 1.0
