##############################################################################
##  TDBC.tcl:                                                               ##
## --------------------------------------------------                       ##
##  This class implements a class for accessing a TDBC database.            ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::Base::Utilities
package require tdbc ;# Package to access a TDBC server

oo::class create ELEP::Database::TDBC {
  mixin ELEP::Base::Utilities

  ##########################################################################
  ## Method: init_database_object
  ##########################################################################
  method init_database_object {} {
    my variable driver database_filename host port user password database db \
                offline error_on_offline database error_dialog_obj timeout \
                show_error_dialog

    set database_filename [my userConfigurationDir]/database.sqlite3
    set driver            postgres ;# sqlite3, obdc, postgres, mysql
    set host              www.ellogon.org
    # set host              localhost
    set port              5432
    set database          ellogon
    set user              ellogon
    set password          ellogon
    set db                [self].db
    set offline           0
    set error_on_offline  1
    set error_dialog_obj  {}
    set show_error_dialog 1
    set timeout           60000 ;# 60 seconds
  };# init_database_object

  ##########################################################################
  ## Method: connect
  ##########################################################################
  method connect {} {
    my variable driver db database
    package require tdbc::$driver
    my disconnect
    switch $driver {
      sqlite3  {
        my variable database_filename
        tdbc::sqlite3::connection create $db $database_filename
      }
      postgres {
        my variable host port database user password
        tdbc::postgres::connection create $db \
          -host $host -port $port -database $database \
          -user $user -password $password
      }
      mysql    {
        my variable host port database user password
        tdbc::mysql::connection create $db \
          -host $host -port $port -database $database \
          -user $user -passwd $password
      }
      odbc     {}
      default  {error "unknown driver \"$driver\""}
    }
  };# connect

  ##########################################################################
  ## Method: disconnect
  ##########################################################################
  method disconnect {} {
    my variable db
    catch {$db close}
  };# disconnect

  method offline {} { my variable offline; return $offline }
  method online  {} { my variable offline; return [expr {!$offline}] }

  ##########################################################################
  ## Method: ensure_online
  ##########################################################################
  method ensure_online {} {
    my variable offline host port database
    if {$offline} {
      error "Database \"$host:$port/$database\" is offline!"
    }
  };# ensure_online

  ##########################################################################
  ## Method: first_element
  ##########################################################################
  method first_element {sql} {
    set result [uplevel 1 [list my sql $sql]]
    return [lindex $result 0 0]
  };# first_element

  ##########################################################################
  ## Method: foreach
  ##########################################################################
  method foreach {args} {
    my variable db
    my connect
    catch { uplevel 1 [list $db foreach {*}$args] } result options
    catch {my disconnect}
    return -options $options $result
  };# foreach

  ##########################################################################
  ## Method: sql
  ##########################################################################
  method sql {sql {tries 5}} {
    my variable error_on_offline offline db
    if {$error_on_offline} {
      my ensure_online
    }
    if {$offline} return {}
    ##
    ## Connect into the database...
    ##
    for {set i 1} {$i <= $tries} {incr i} {
      set status [catch {
        my connect
      } result options]
      if {$status != 1 || $i >= $tries} {break}
      my reportDatabaseFailure [my msg {Connection failed}] \
                               $result $options $i
    }
    if {$status == 1} {
      my reportDatabaseFailureDone
      return -options $options $result
    }

    ##
    ## Execute the SQL statement...
    ##
    for {set i 1} {$i <= $tries} {incr i} {
      set status [catch {
        uplevel 1 [list $db allrows -as lists -- $sql]
      } result options]
      if {$status != 1 || $i >= $tries} {break}
      set bug {}
      my reportDatabaseFailure [my msg {SQL query failed}] \
                               $result $options $i $sql
    }

    catch {my disconnect}
    my reportDatabaseFailureDone

    return -options $options $result
  };# sql

  ##########################################################################
  ## Method: sleep
  ##########################################################################
  method sleep { ms } {
    set uniq [ my uniqkey ]
    set ::__sleep__tmp__$uniq 0
    after $ms set ::__sleep__tmp__$uniq 1
    vwait ::__sleep__tmp__$uniq
    unset ::__sleep__tmp__$uniq
  };# sleep

  ##########################################################################
  ## Method: uniqkey
  ##########################################################################
  method uniqkey { } {
    set key   [ expr { pow(2,31) + [ clock clicks ] } ]
    set key   [ string range $key end-8 end-3 ]
    set key   [ clock seconds ]$key
    return $key
  };# uniqkey

  ##########################################################################
  ## Method: reportDatabaseFailure
  ##########################################################################
  method reportDatabaseFailure { msg result options try {sql {}} } {
    my variable timeout show_error_dialog
    if {$show_error_dialog} {
      return [my reportDatabaseFailureGUI $msg $result $options $try $sql]
    }
    catch {puts stderr "$msg (try $try): $result"}
    if {$sql ne ""} {catch {puts $sql}}
    my sleep $timeout
  };# reportDatabaseFailure

  ##########################################################################
  ## Method: reportDatabaseFailureDone
  ##########################################################################
  method reportDatabaseFailureDone { } {
    my variable error_dialog_obj
    catch {$error_dialog_obj destroy}
    set error_dialog_obj {}
  };# reportDatabaseFailureDone

  ##########################################################################
  ## Method: reportDatabaseFailureGUI
  ##########################################################################
  method reportDatabaseFailureGUI { msg result options try {sql {}} } {
    my variable error_dialog_obj
    catch {console show}; update
    catch {puts stderr "$msg (try $try): $result"}
    if {$sql ne ""} {catch {puts $sql}}
    if {$error_dialog_obj eq ""} {
      my createReportDatabaseFailureGUI $msg $result $options $try $sql
      $error_dialog_obj showWait
      set error_dialog_obj {}
    }
  };# reportDatabaseFailureGUI

  ##########################################################################
  ## Method: createReportDatabaseFailureGUI
  ##########################################################################
  method createReportDatabaseFailureGUI { msg result options try {sql {}} } {
    my variable error_dialog_obj
    catch {$error_dialog_obj destroy}
    package require ELEP::WindowUtilities::DialogWindow
    package require widget::scrolledwindow
    set focus_win [focus]
    if {$focus_win eq ""} {
      set toplevel {}
    } else {
      set toplevel [winfo toplevel $focus_win]
    }
    set error_dialog_obj [ELEP::WindowUtilities::DialogWindow create \
                                $toplevel.reportDatabaseFailureGUI]
    $error_dialog_obj title {Error during database interaction...}
    $error_dialog_obj addStdButton retry
    set win [$error_dialog_obj getframe]

      ttk::label $win.title -image [my loadImage IWarning] -compound left \
                 -text [my msg {Error during interaction with the database!}] \
                 -font {arial 18 bold} -foreground red
      widget::scrolledwindow $win.sw
      $win.sw setwidget [text $win.sw.text -borderwidth 0]
      $win.sw.text insert end "SQL code: \"$sql\"\n\n"
      $win.sw.text insert end "Error:\n----------------------------------\n"
      $win.sw.text insert end $result
      $win.sw.text configure -state disabled
    grid $win.title -sticky snew -padx 2 -pady 2
    grid $win.sw -sticky snew -padx 2 -pady 2
  };# createReportDatabaseFailureGUI

};# class ELEP::Database::TDBC

package provide ELEP::Database::TDBC 1.0
