##############################################################################
##  EncodingDetection.tcl:                                                  ##
## --------------------------------------------------                       ##
##  This class implements an automatic encoding detection facility.         ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

##
## Load some needed packages.
##
package require TclOO    ;# TclOO adds object-oriented facilities to Tcl.

oo::class create ELEP::DocumentUtilities::EncodingDetection {
  variable default_encoding

  constructor {{encoding utf-8}} {
    set default_encoding $encoding
  };# constructor

  ##########################################################################
  ## Method: encoding
  ##########################################################################
  method encoding {filename} {
    ## As a first step, check whether the file is already in utf-8/utf-16.
    # set encoding [encoding-utf $filename]
    # if {[string length $encoding]} {return $encoding}
    ## Not a utf-8/utf-16 file...
    switch -glob [string tolower [file extension $filename]] {
      *.htm*  {return [my encoding-html $filename]}
      default {return $default_encoding}
    }
  };# encoding

  ##########################################################################
  ## Method: encoding-html
  ##########################################################################
  method encoding-html {filename} {
    set file [open $filename]
    set data [read $file 2000]
    close $file
    return [my encoding-html-data $data iso8859-1]
  };# encoding-html

  ##########################################################################
  ## Method: encoding-html-data
  ##########################################################################
  method encoding-html-data {data {enc {}}} {
    if {![string length $enc]} {set enc $default_encoding}
    set pattern {charset=\s*["']?([^ '">;]+)}
    if {[regexp -nocase $pattern $data _ charset]} {
      if {![catch {my IANAEncoding2TclEncoding $charset} encoding]} {
        return $encoding
      }
    }
    return $enc
  };# encoding-html

  ##########################################################################
  ## Method: changeHTMLDataEncoding
  ##########################################################################
  method changeHTMLDataEncoding {data {initial_encoding {}}
                                      {target_encoding utf-8}} {
    if {![string length $initial_encoding]} {
      ## Auto-detect initial encoding...
      set initial_encoding [my encoding-html-data $data iso8859-1]
    }
    set data [::encoding convertfrom $initial_encoding $data]
    ## Replace the encoding specifier with the target encoding...
    set enc [my TclEncoding2IANAEncoding $target_encoding]
    set pattern  {charset=\s*[^ "'>;]+} ; #"
    set pattern2 {charset=\s*'[^ "'>;]+} ; #"
    set pattern3 {charset=\s*"[^ "'>;]+} ; #"
    regsub -nocase -all $pattern  $data "charset=$enc"     data
    regsub -nocase -all $pattern2 $data "charset='$enc'"   data
    regsub -nocase -all $pattern3 $data "charset=\"$enc\"" data
    return $data
  };# changeHTMLDataEncoding

  ##########################################################################
  ## Method: encoding-utf
  ##########################################################################
  method encoding-utf {filename} {
    set fd [open $filename]
    # The autodetection of the encoding follows
    # XML Recomendation, Appendix F
    fconfigure $fd -encoding binary
    if {![binary scan [read $fd 4] "H8" firstBytes]} {
      # very short (< 4 Bytes) file
      close $fd
      return utf-8
    }
    close $fd
    # First check for BOM
    switch [string range $firstBytes 0 3] {
      "feff" -
      "fffe" {
        # feff: UTF-16, big-endian BOM
        # ffef: UTF-16, little-endian BOM
        return identity
      }

      # 00 00 FE FF  	UTF-32, big-endian
      # FF FE 00 00 	UTF-32, little-endian
      # FE FF 	UTF-16, big-endian
      # FF FE 	UTF-16, little-endian
      # EF BB BF 	UTF-8
    }
    ## Not a utf-9/utf-16 file...
    return {}
  };# encoding-utf

  ##########################################################################
  ## Method: IANAEncoding2TclEncoding
  ## This method is an adaptation of code provided with the tDom Tcl library.
  ##########################################################################
  method IANAEncoding2TclEncoding {IANAName} {
    # First the most widespread encodings with there
    # preferred MIME name, to speed lookup in this
    # usual cases. Later the official names and the
    # aliases.
    #
    # For "official names for character sets that may be
    # used in the Internet" see
    # http://www.iana.org/assignments/character-sets
    # (that's the source for the encoding names below)
    #
    # Matching is case-insensitive

    switch [string tolower $IANAName] {
      "us-ascii"           {return ascii}
      "utf-8"              {return utf-8}
      "utf-16"             {return unicode; # not sure about this}
      "iso-8859-1"         {return iso8859-1}
      "iso-8859-2"         {return iso8859-2}
      "iso-8859-3"         {return iso8859-3}
      "iso-8859-4"         {return iso8859-4}
      "iso-8859-5"         {return iso8859-5}
      "iso-8859-6"         {return iso8859-6}
      "iso-8859-7"         {return iso8859-7}
      "iso-8859-8"         {return iso8859-8}
      "iso-8859-9"         {return iso8859-9}
      "iso-8859-10"        {return iso8859-10}
      "iso-8859-13"        {return iso8859-13}
      "iso-8859-14"        {return iso8859-14}
      "iso-8859-15"        {return iso8859-15}
      "iso-8859-16"        {return iso8859-16}
      "iso-2022-kr"        {return iso2022-kr}
      "euc-kr"             {return euc-kr}
      "iso-2022-jp"        {return iso2022-jp}
      "koi8-r"             {return koi8-r}
      "shift_jis"          {return shiftjis}
      "euc-jp"             {return euc-jp}
      "gb2312"             {return gb2312}
      "big5"               {return big5}
      "cp866"              {return cp866}
      "cp1250"             {return cp1250}
      "cp1253"             {return cp1253}
      "cp1254"             {return cp1254}
      "cp1255"             {return cp1255}
      "cp1256"             {return cp1256}
      "cp1257"             {return cp1257}

      "windows-1251"       -
      "cp1251"             {return cp1251}

      "windows-1252"       -
      "cp1252"             {return cp1252}

      "iso_8859-1:1987"    -
      "iso-ir-100"         -
      "iso_8859-1"         -
      "latin1"             -
      "l1"                 -
      "ibm819"             -
      "cp819"              -
      "csisolatin1"        {return iso8859-1}

      "iso_8859-2:1987"    -
      "iso-ir-101"         -
      "iso_8859-2"         -
      "iso-8859-2"         -
      "latin2"             -
      "l2"                 -
      "csisolatin2"        {return iso8859-2}

      "iso_8859-5:1988"    -
      "iso-ir-144"         -
      "iso_8859-5"         -
      "iso-8859-5"         -
      "cyrillic"           -
      "csisolatincyrillic" {return iso8859-5}

      "ms_kanji"           -
      "csshiftjis"         {return shiftjis}

      "csiso2022kr"        {return iso2022-kr}

      "ibm866"             -
      "csibm866"           {return cp866}

      default {
        # There are much more encoding names out there
        # It's only laziness, that let me stop here.
        error "Unrecognized encoding name '$IANAName'"
      }
    }
  };# IANAEncoding2TclEncoding

  ##########################################################################
  ## Method: TclEncoding2IANAEncoding
  ## This method is an adaptation of code provided with the tDom Tcl library.
  ##########################################################################
  method TclEncoding2IANAEncoding {IANAName} {
    # First the most widespread encodings with there
    # preferred MIME name, to speed lookup in this
    # usual cases. Later the official names and the
    # aliases.
    #
    # For "official names for character sets that may be
    # used in the Internet" see
    # http://www.iana.org/assignments/character-sets
    # (that's the source for the encoding names below)
    #
    # Matching is case-insensitive

    switch [string tolower $IANAName] {
      ascii                              {return us-ascii}
      utf-8                              {return utf-8}
      unicode                            {return utf-16}
      iso8859-1                          {return iso-8859-1}
      iso8859-2                          {return iso-8859-2}
      iso8859-3                          {return iso-8859-3}
      iso8859-4                          {return iso-8859-4}
      iso8859-5                          {return iso-8859-5}
      iso8859-6                          {return iso-8859-6}
      iso8859-7                          {return iso-8859-7}
      iso8859-8                          {return iso-8859-8}
      iso8859-9                          {return iso-8859-9}
      iso8859-10                         {return iso-8859-10}
      iso8859-13                         {return iso-8859-13}
      iso8859-14                         {return iso-8859-14}
      iso8859-15                         {return iso-8859-15}
      iso8859-16                         {return iso-8859-16}
      iso2022-kr                         {return iso-2022-kr}
      euc-kr                             {return euc-kr}
      iso2022-jp                         {return iso-2022-jp}
      koi8-r                             {return koi8-r}
      shiftjis                           {return shift_jis}
      euc-jp                             {return euc-jp}
      gb2312                             {return gb2312}
      big5                               {return big5}
      cp866                              {return cp866}
      cp1250                             {return cp1250}
      cp1253                             {return cp1253}
      cp1254                             {return cp1254}
      cp1255                             {return cp1255}
      cp1256                             {return cp1256}
      cp1257                             {return cp1257}
      cp1251                             {return windows-1251}
      cp1252                             {return windows-1252}

      default {
        # There are much more encoding names out there
        # It's only laziness, that let me stop here.
        error "Unrecognized encoding name '$IANAName'"
      }
    }
  };# TclEncoding2IANAEncoding

};# class ELEP::DocumentUtilities::EncodingDetection

package provide ELEP::DocumentUtilities::EncodingDetection 1.0
