##############################################################################
##  HTMLUtilities.tcl:                                                      ##
## --------------------------------------------------                       ##
##  This class implements an class that detects various HTML features.      ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

##
## Load some needed packages.
##
package require TclOO    ;# TclOO adds object-oriented facilities to Tcl.

oo::class create ELEP::DocumentUtilities::HTMLUtilities {
  variable text \
           css \
           html2text \
           html2links \
           html2imagetags \
           html2imagelinks \
           html2attributevalues \
           html2tags

  constructor {{t {}}} {
    my setHTML $t
  };# constructor

  ##########################################################################
  ## Method: setHTML
  ##########################################################################
  method setHTML {t} {
    set text [string trim $t]
    ## There is the possibility the file to start from a special signature,
    ## denoting a UTF-8 file...
    if {[string equal [string index $text 0] \uFEFF]} {
      set text [string range $text 1 end]
    }
  };# setHTML
  method setHTML_no_checks {t} {
    set text $t
  };# setHTML_no_checks

  ##########################################################################
  ## Method: getHTML
  ##########################################################################
  method getHTML {} {
    return $text
  };# getHTML

  ##########################################################################
  ## Method: setCSS
  ##########################################################################
  method setCSS {t} {
    set css [string trim $t]
    ## There is the possibility the file to start from a special signature,
    ## denoting a UTF-8 file...
    if {[string equal [string index $css 0] \uFEFF]} {
      set css [string range $css 1 end]
    }
  };# setCSS

  ##########################################################################
  ## Method: getCSS
  ##########################################################################
  method getCSS {} {
    return $css
  };# getCSS

  ##########################################################################
  ## Method: isHTML
  ##########################################################################
  method isHTML {} {
    set first_char [string index $text 0]
    if {![string equal < $first_char]} {return 0}
    return [my containsHTML]
  };# isHTML

  ##########################################################################
  ## Method: containsHTML
  ##########################################################################
  method containsHTML {} {
    foreach pattern [list {<html}  {<head}   {<body} {<!doctype html} \
                          {<table} {<a href} {<li}   {&nbsp;}] {
      if {[regexp -nocase -- $pattern $text]} {return 1}
    }
    return 0
  };# containsHTML

  ##########################################################################
  ## Method: isStyleSheet
  ##########################################################################
  method isStyleSheet {} {
    set first_char [string index $text 0]
    if {![string equal \{ $first_char]} {return 0}
    return [my containsStyleSheet]
  };# isStyleSheet

  ##########################################################################
  ## Method: containsStyleSheet
  ##########################################################################
  method containsStyleSheet {} {
    foreach pattern [list {margin-bottom:}     {margin-right:} \
                          {margin-top:}        {margin-left:} \
                          {padding-bottom:}    {padding-right:} \
                          {padding-top:}       {padding-left:} \
                          {background-image:}  {font-size:} \
                          {border-left:}       {border-right:} \
                          {text-decoration:}   {font-family:} \
                          {text-align:}        {border-bottom:} \
                          {font-weight:}       {font-style:} \
                          {background-repeat:} {background-position:} \
                          {border-collapse:}   {text-transform:} \
                          margin: padding: background: color: width: height: \
                          border: !important repeat-x white-space: nowrap \
                          no-repeat 0px; 100%; #ffffff #cccccc #FFCC00 url( ] {
      if {[regexp -nocase -- $pattern $text]} {return 1}
    }
    return 0
  };# containsStyleSheet

  ##########################################################################
  ## Method: getHead
  ##########################################################################
  method getHead {} {
    set t [regexp -inline -nocase -- {<head.*</head>} $text]
    if {[string length $t]} {return $t}
    return $text
  };# getHead

  ##########################################################################
  ## Method: getBody
  ##########################################################################
  method getBody {} {
    set t [regexp -inline -nocase -- {<body.*</body>} $text]
    if {[string length $t]} {return $t}
    return $text
  };# getBody

  ##########################################################################
  ## Method: getBase
  ##########################################################################
  method getBase {{base {}}} {
    return [my getAttribute [my getFirstTag base] href $base]
  };# getBase

  ##########################################################################
  ## Method: getFirstTag
  ##########################################################################
  method getFirstTag {tag {value {}}} {
    regexp -nocase "<$tag\[^>\]+>" $text value
    return $value
  };# getFirstTag

  ##########################################################################
  ## Procedure: getTagAttributes
  ##########################################################################
  proc getTagAttributes {tag} {
    set attrs {}
    foreach {_ attr} [regexp -all -inline -nocase {\s+(\w+)=} $tag] {
      lappend attrs $attr
    }
    return $attrs
  };# getTagAttributes

  ##########################################################################
  ## Procedure: getAttribute
  ##########################################################################
  proc getAttribute {tag attr {value {}}} {
    set patterns [list "$attr=\\s*\"(\[^\"\]*)\"" "$attr=\\s*'(\[^\'\]*)'" \
                       "$attr=\\s*(\[^ \]*)"]
    foreach one $patterns {
      if {[regexp -nocase $one $tag _ value]} {return $value}
    }
    return $value
  };# getAttribute

  ##########################################################################
  ## Method: removeTag
  ##########################################################################
  method removeTag {tag {value {}}} {
    regsub -all -nocase "<$tag\[^>\]*>" $text $value text
  };# removeTag

  ##########################################################################
  ## Method: getText
  ##########################################################################
  method getText {{strip_white_space 0}} {
    package require htmlparse
    set html2text {}
    if {$strip_white_space} {
      ::htmlparse::parse -cmd \
               [list [self] accumulateTextStripWhiteSpaceCallback] $text
    } else {
      ::htmlparse::parse -cmd [list [self] accumulateTextCallback] $text
    }
    return $html2text
  };# getText
  method accumulateTextCallback {tag slash param textBehindTheTag} {
    append html2text $textBehindTheTag
  };# accumulateTextCallback
  method accumulateTextStripWhiteSpaceCallback {tag slash param
                                                          textBehindTheTag} {
    append html2text [string trim $textBehindTheTag]
  };# accumulateTextStripWhiteSpaceCallback

  ##########################################################################
  ## Method: getHTMLLinks
  ##########################################################################
  method getHTMLLinks {} {
    package require htmlparse
    set html2links {}
    ::htmlparse::parse -cmd \
          [list [self] accumulateLinks] $text
    return $html2links
  };# getHTMLLinks
  method accumulateLinks {tag slash param textBehindTheTag} {
    if {[string length $slash]} {return}
    switch -exact -nocase -- $tag {
      a {
        ## Locate the href attribute...
        foreach href [regexp -all -inline -nocase {href=['"]?[^ '"]+} $param] {
          set url [string trim [string range $href 5 end] '\"]
          lappend html2links $url
        }
      }
      default {}
    }
  };# accumulateLinks

  ##########################################################################
  ## Method: getImageTags
  ##########################################################################
  method getImageTags {} {
    package require htmlparse
    set html2imagetags {}
    ::htmlparse::parse -cmd \
          [list [self] accumulateTags html2imagetags img] $text
    return $html2imagetags
  };# getImageTags
  method accumulateTags {varname tags tag slash param textBehindTheTag} {
    if {[string length $slash]} {return}
    if {[lsearch -exact -nocase $tags $tag] != -1} {
      lappend $varname "$tag $param"
    }
  };# accumulateTags

  ##########################################################################
  ## Method: getImageLinks
  ##########################################################################
  method getImageLinks {} {
    package require htmlparse
    set html2imagelinks {}
    ::htmlparse::parse -cmd \
          [list [self] accumulateImageLinks] $text
    return $html2imagelinks
  };# getImageLinks
  method accumulateImageLinks {tag slash param textBehindTheTag} {
    if {[string length $slash]} {return}
    switch -exact -nocase -- $tag {
      img {
        ## Locate the src attribute...
        foreach src [regexp -all -inline -nocase {src=['"]?[^ '"]+} $param] {
          set url [string trim [string range $src 4 end] '\"]
          lappend html2imagelinks $url
        }
      }
      default {}
    }
  };# accumulateImageLinks

  ##########################################################################
  ## Method: getTagAttributeValues
  ##########################################################################
  method getTagAttributeValues {tag attr} {
    package require htmlparse
    set html2attributevalues {}
    ::htmlparse::parse -cmd \
          [list [self] accumulateAttributeValues $tag $attr] $text
    return $html2attributevalues
  };# getTagAttributeValues
  method accumulateAttributeValues {tag_name attr
                                    tag slash param textBehindTheTag} {
    if {[string length $slash]} {return}
    if {![string equal -nocase $tag_name $tag]} {return}
    lappend html2attributevalues [getAttribute $param $attr]
  };# accumulateAttributeValues

  ##########################################################################
  ## Method: getTags
  ##########################################################################
  method getTags {tag} {
    package require htmlparse
    set html2tags {}
    ::htmlparse::parse -cmd \
          [list [self] accumulateHTags $tag] $text
    return $html2tags
  };# getTags
  method accumulateHTags {tag_name tag slash param textBehindTheTag} {
    if {[string length $slash]} {return}
    if {![string equal -nocase $tag_name $tag]} {return}
    lappend html2tags "<$tag $param>"
  };# accumulateHTags

  ##########################################################################
  ## Method: getCSSURLs
  ##########################################################################
  method getCSSURLs {} {
    return [_locate_stylesheet_urls $css]
  };# getCSSURLs

  ##########################################################################
  ## Method: getHTMLURLs
  ##########################################################################
  method getHTMLURLs {} {
    return [_locate_stylesheet_urls $text]
  };# getHTMLURLs

  proc _locate_stylesheet_urls {data} {
    set urls {}
    if {[string first url( $data] != -1} {
      foreach {_ url} [regexp -nocase -all -inline {url\(([^\)]+)\)} $data] {
        lappend urls [string trim $url \'\"]
      }
    }
    return $urls
  };# _locate_stylesheet_urls

  ##########################################################################
  ## Method: isImageURL
  ##########################################################################
  method isImageURL {url} {
    if {![string length $url]} {return false}
    switch -glob -nocase -- $url {
      *.jpeg - *.jpg -
      *.exif -
      *.tiff - *.tif -
      *.raw  -
      *.png  -
      *.gif  -
      *.bmp  -
      *.ppm  - *.pgm - *.pbm - *.pnm -
      *.ecw  -
      *.cgm  - *.svg -
      *.odg  -
      *.eps  - *.ps  - *.pdf - 
      *.swf  - *.wmf - *.emf -
      *.xps {
        return true
      }
      default {return false}
    }
    return false
  };# isImageURL

};# class ELEP::DocumentUtilities::HTMLUtilities

package provide ELEP::DocumentUtilities::HTMLUtilities 1.0
