##############################################################################
##  HashTableGazetteer.tcl:                                                 ##
## --------------------------------------------------                       ##
##  This class implements a simple gazetter. Strings to be found are stored ##
##  inside a hash table.                                                    ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

##
## Load some needed packages.
##
package require TclOO    ;# TclOO adds object-oriented facilities to Tcl.

oo::class create ELEP::DocumentUtilities::HashTableGazetteer {
  variable model \
           look_ahead \
           token_ann_type \
           token_ann_constraints \
           return_unique_word_lists \
           skip_single_character_patterns

  constructor {args} {
    set model                    {};# A model is a dict with the strings
                                    # to be matched.
    set look_ahead               1 ;# How many tokens should we look forward?
    set token_ann_type           token
    set token_ann_constraints    {}
    set return_unique_word_lists 0
    set skip_single_character_patterns 0
    if {[llength $args]} {my configure {*}$args}
  };# constructor

  method setModel     {m} {set model $m}
  method setLookAhead {l} {set look_ahead $l}

  ##
  ## addListFileToModel
  ##
  method addListFileToModel {filename type} {
    set words {}
    set fd [open $filename]
    fconfigure $fd -encoding utf-8 -eofchar {}
    while {[gets $fd line] >= 0} {
      set line [string toupper [string trim $line]]
      if {$skip_single_character_patterns && [string length $line] < 1} {
        continue
      }
      ## Compress excessive whitespace...
      regsub -all {\s+} $line { } line
      if {[::info exists seen($line)]} {
        if {!$return_unique_word_lists} {lappend words $line}
        continue
      }
      dict lappend model $line $type
      lappend words $line
      set seen($line) 1
      ## Count tokens...
      set len [llength [split $line]]
      if {$len > $look_ahead} {set look_ahead $len}
    }
    close $fd
    return $words
  };# addListFileToModel

  ##
  ## addListFileToModelWithTransormation:
  ##   Same as addListFileToModel, but with the ability to apply a user
  ##   provided transformation (args argument).
  ##
  method addListFileToModelWithTransormation {filename type args} {
    set words {}
    set fd [open $filename]
    fconfigure $fd -encoding utf-8 -eofchar {}
    while {[gets $fd line] >= 0} {
      ## Apply the user requested transformation...
      set line [string toupper [{*}$args [string trim $line]]]
      if {$skip_single_character_patterns && [string length $line] < 1} {
        continue
      }
      ## Compress excessive whitespace...
      regsub -all {\s+} $line { } line
      if {[::info exists seen($line)]} {
        if {!$return_unique_word_lists} {lappend words $line}
        continue
      }
      dict lappend model $line $type
      lappend words $line
      set seen($line) 1
      ## Count tokens...
      set len [llength [split $line]]
      if {$len > $look_ahead} {set look_ahead $len}
    }
    close $fd
    return $words
  };# addListFileToModelWithTransormation

  ##
  ## processDocument: process an Ellogon document. The document is assumed to
  ## contain tokens...
  ##
  method processDocument {doc {ann_type lookup}} {
    set tokens [tip_SelectAnnotationsSorted $doc $token_ann_type \
                       $token_ann_constraints]
    my processDocumentTokens $doc $tokens $ann_type
  };# processDocument

  method processDocumentTokens {doc {tokens {}} {ann_type lookup}} {
    catch {ggi_wait_update $doc 0}
    set text [string toupper [tip_GetRawData $doc]]
    ## Construct the list of tokens...
    set words {}
    foreach ann $tokens {
      lappend words [tip_GetFirstAnnotatedTextRange $text $ann]
    }
    catch {ggi_wait_update $doc 5}
    ## Perform the matching...
    set result [my processTokens $words]
    catch {ggi_wait_update $doc 80}
    ## Put results back as annotations...
    foreach type [dict keys $result] {
      set attrs [tip_CreateAttributeSet type \
                        [tip_CreateAttributeValue GDM_STRING $type]]
      foreach {word_start word_end} [dict get $result $type] {
        set ann_start [lindex $tokens $word_start]
        set ann_end   [lindex $tokens $word_end]
        tip_GetFirstSpanOffsets $ann_start start _
        tip_GetFirstSpanOffsets $ann_end _ end
        tip_AddAnnotation $doc [tip_CreateAnnotation $ann_type \
                 [list [list $start $end]] $attrs]
      }
    }
    catch {ggi_wait_update $doc 100}
  };# processDocumentTokens

  ##
  ## processTokens: process a set of tokens...
  ##
  method processTokens {tokens} {
    set result [dict create]
    set tokens_len [llength $tokens]
    ## Iterate over all tokens...
    for {set i 0} {$i < $tokens_len} {incr i} {
      set new_i $i
      array unset matches *
      set string  {}
      set j_end [expr {$i + $look_ahead}]
      ## Examine all tokens that look ahead suggests...
      for {set j $i} {$j < $j_end} {incr j} {
        append string [lindex $tokens $j]
        ## Is the string in the model?
        if {[dict exists $model $string]} {
          set matches([dict get $model $string]) [list $i $j]
          set new_i $j
        }
        append string { }; # add space for the next word...
      }
      foreach type [array names matches] {
        dict lappend result $type {*}$matches($type)
      }
      set i $new_i
    }
    return $result
  };# processTokens

  method configure {args} {
    foreach {varName varValue} $args {
      set [string range $varName 1 end] $varValue
    }
  };# configure

  method cget {varName} {
    return [set [string range $varName 1 end]]
  };# cget

};# class ELEP::DocumentUtilities::HashTableGazetteer

package provide ELEP::DocumentUtilities::HashTableGazetteer 1.0
