##############################################################################
##  XMLExporter.tcl:                                                        ##
## --------------------------------------------------                       ##
##  This class provides facilities for exporting Annotations to XML.        ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

##
## Load some needed packages.
##
package require TclOO    ;# TclOO adds object-oriented facilities to Tcl.

oo::class create ELEP::ExportConverters::XMLExporter {
  variable collection progress_max

  constructor {} {
    set progress_max 0
    ## Create a dummy Collection...
    set collection [tip_CreateCollection $::CDM_TempDir/TempCol {}]
  };# constructor

  ##########################################################################
  ## Method: convertAnnotationsToXML
  ##########################################################################
  method convertAnnotationsToXML {text annotations} {
    set anns_nu [llength $annotations]
    if {$anns_nu == 0} {return $text}
    my progressInitialise $anns_nu
    ## Create a dummy Document and add the provided Annotation into it. Adding
    ## Annotations to the Document will assign each one of them an id, in case
    ## an Annotation does not already have one...
    set tdoc [tip_CreateDocument $collection TempDoc $text $annotations {}]
    set annotations [tip_SortAnnotationSet [tip_GetAnnotations $tdoc]]
    ## Iterate over all Annotations starting from the last one, and select
    ## their ids. We start from the list end, as no Annotations come after,
    ## thus eliminating the need the CDM core to update their spans...
    set ids {}
    for {set i [incr anns_nu -1]} {$i > -1} {incr i -1} {
      lappend ids [tip_GetId [lindex $annotations $i]]
    }
    ## Now, iterate over all ids, and add the relevant xml tags in the tdoc:
    foreach id $ids {
      set ann [tip_GetAnnotation $tdoc $id]
      foreach {open_tag close_tag} [my annotationToXMLOpenCloseTags $ann] {}
      tip_RemoveAnnotation $tdoc $id
      tip_GetSpanOffsets [tip_GetFirstSpan $ann] start end
      ## We must first insert a string at the span end position. If we insert
      ## first at the start span position, then the end we have will be invalid.
      tip_ByteSequenceInsertString $tdoc $end   $close_tag
      tip_ByteSequenceInsertString $tdoc $start $open_tag
      my progressAdvance
    }
    my progressFinalise
    set xml [tip_GetRawData $tdoc]
    tip_Close $tdoc
    return $xml
  };# convertAnnotationsToXML

  ##########################################################################
  ## Method: convertAnnotationsToXML
  ##########################################################################
  method annotationToXMLOpenCloseTags {annotation} {
    set tag_name [tip_GetType $annotation]
    set open_tag  "<$tag_name id=[tip_GetId $annotation]"
    foreach attr [tip_GetAttributes $annotation] {
      append open_tag " [tip_GetName $attr]=\"[tip_GetValueValue $attr]\""
    }
    return [list $open_tag> </$tag_name>]
  };# annotationToXMLOpenCloseTags

  ##########################################################################
  ## Method: progressInitialise
  ##########################################################################
  method progressInitialise {max} {
    set progress_max $max
  };# progressInitialise

  ##########################################################################
  ## Method: progressAdvance
  ##########################################################################
  method progressAdvance {} {
  };# progressAdvance

  ##########################################################################
  ## Method: progressFinalise
  ##########################################################################
  method progressFinalise {} {
  };# progressFinalise

  destructor {
    tip_Close $collection
  };# destructor

};# class ELEP::ExportConverters::XMLExporter

package provide ELEP::ExportConverters::XMLExporter 1.0
