##############################################################################
##  Crawler.tcl:                                                            ##
## --------------------------------------------------                       ##
##  This class implements methods for crawling URIs.                        ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::FileUtilities::URI
package require ELEP::DocumentUtilities::EncodingDetection
package require ELEP::DocumentUtilities::HTMLUtilities
package require md5
package require fileutil::magic::mimetype
package require fileutil::magic::filetype
package require fileutil

oo::class create ELEP::FileUtilities::Crawler {
  public variable downloadThreads 4
  public variable storageDirectory $::CDM_TempDir/Crawler

  variable encoding_detector {}
  variable html_processor    {}

  constructor {args} {
    eval configure $args
    init
  };# constructor

  destructor {
    foreach object {encoding_detector html_processor} {
      if {[string length [set $object]]} {
        catch {itcl::delete object [set $object]}
      }
    }
  };# destructor

  method init {} {
    set encoding_detector [ELEP::DocumentUtilities::EncodingDetection \
        ELEP::FileUtilities::encoding_detector_#auto]
    set html_processor [ELEP::DocumentUtilities::HTMLUtilities \
        ELEP::FileUtilities::html_processor_#auto]
  };# init

  ##########################################################################
  ## Method: downloadURI
  ##########################################################################
  method downloadURI {uri args} {
    # log "downloadURI: $uri"
    ## Create a suitable name from the uri
    foreach {host url_md5} [fileSystemNameFromURI $uri] {break}
    set file_system_uri_name $host/$url_md5
    ## Download the url in binary mode...
    file mkdir $storageDirectory/$file_system_uri_name
    set local_resource_file_name $file_system_uri_name/uri.bin
    if {[file exists $storageDirectory/$local_resource_file_name]} {
      ## Do what?
    }
    ELEP::FileUtilities::URI::urlfile $uri \
          $storageDirectory/$local_resource_file_name binary
    set text_resource $local_resource_file_name
    set resourceIsText [resourceIsText $text_resource]
    if {$resourceIsText} {
      set text_resource [generateTextResource $text_resource]
    }
    ## Save download information into the indexes...
    saveResourceInGlobalIndex $uri $host $url_md5 $file_system_uri_name \
                                   $local_resource_file_name $text_resource
    saveResourceInformation   $uri $host $url_md5 $file_system_uri_name \
                                   $local_resource_file_name $text_resource
    ## If this is a text resource, download additional resources like images,
    ## etc...
    if {$resourceIsText} {
      set base [$html_processor getBase $uri]
      downloadURIResources $base $text_resource
    }
  };# downloadURI

  ##########################################################################
  ## Method: downloadURIResources
  ##########################################################################
  method downloadURIResources {base filename args} {
    if {![$html_processor isHTML]} {return {}}
    $html_processor removeTag base
    set data [$html_processor getHTML]
    set dir [file dirname $filename]/resources
    ## Make the resources subdirectory...
    file mkdir $storageDirectory/$dir
    ##
    ## Download images...
    ##
    file mkdir $storageDirectory/$dir/images
    foreach url [$html_processor getImageLinks] {
      if {![string length $url]} {continue}
      if {[catch {downloadResource $base $url $dir/images} resource]} {
        log $resource error
        continue
      }
      ## Substitute the link with the local file!
      set data [string map [list $url resources/images/$resource] $data]
    }
    ##
    ## Download other tags...
    ##
    foreach {tag attr} {link href script src} {
      file mkdir $storageDirectory/$dir/tag_$tag
      foreach url [$html_processor getTagAttributeValues $tag $attr] {
        if {![string length $url]} {continue}
        if {[catch {downloadResource $base $url $dir/tag_$tag} resource]} {
          log $resource error
          continue
        }
        ## Should the resource be processed further?
        set resource [examineResource $tag $url $dir/tag_$tag $resource]
        ## Substitute the link with the local file!
        set data [string map [list $url resources/tag_$tag/$resource] $data]
      }
    }
    ##
    ## Produce the local version of the resource!
    ##
    store $storageDirectory/[file rootname $filename]_local.html \
          [string trim $data]
  };# downloadURIResources

  ##########################################################################
  ## Method: downloadResource
  ##########################################################################
  method downloadResource {base resource_uri dir args} {
    set uri [uri::resolve $base $resource_uri]
    ## Get host...
    set host [ELEP::FileUtilities::URI::url_part $uri host]
    ## Get the resource file name from the uri...
    set local_resource_file_name [ELEP::FileUtilities::URI::url_part $uri path]
    set local_resource_file_name [file tail $local_resource_file_name]
    ELEP::FileUtilities::URI::urlfile $uri \
          $storageDirectory/$dir/$local_resource_file_name binary
    ## Save download information into the indexes...
    # saveResourceInGlobalIndex $uri $host $url_md5 $file_system_uri_name \
    #                                $local_resource_file_name $text_resource
    saveResourceInformation $uri $host $local_resource_file_name $dir \
                            $dir/$local_resource_file_name \
                            $dir/$local_resource_file_name \
                            [file rootname $local_resource_file_name]_info.xml
    return $local_resource_file_name
  };# downloadResource

  ##########################################################################
  ## Method: examineResource
  ##########################################################################
  method examineResource {tag uri dir resource args} {
    set filename $storageDirectory/$dir/$resource
    set extension [file extension $resource]
    switch -nocase -glob $extension {
      .css {return [processStyleSheet $tag $uri $dir $resource]}
      default {}
    }
    return $resource
  };# examineResource

  ##########################################################################
  ## Method: processStyleSheet
  ##########################################################################
  method processStyleSheet {tag css_uri dir resource args} {
    set modified 0
    set filename $storageDirectory/$dir/$resource
    set data [data $filename]
    $html_processor setCSS $data
    ## Get the urls from the stylesheet...
    foreach url [lsort -unique [$html_processor getCSSURLs]] {
      set modified 1
      file mkdir $storageDirectory/$dir/url
      set rc [downloadResource $css_uri $url $dir/url]
      ## Substitute the link with the local file!
      set data [string map [list $url url/$rc] $data]
    }
    ##
    ## Produce the local version of the resource!
    ##
    if {$modified} {
      store  [file rootname $filename]_local.css [string trim $data] binary
      return [file rootname $resource]_local.css
    }
    return $resource
  };# processStyleSheet

  ##########################################################################
  ## Method: resourceIsText
  ##########################################################################
  method resourceIsText {file args} {
    foreach type [::fileutil::fileType $storageDirectory/$file] {
      switch $type {
        text - xml - html {
          return 1
        }
      }
    }
    return 0
  };# resourceIsText

  ##########################################################################
  ## Method: generateTextResource
  ##########################################################################
  method generateTextResource {file args} {
    set types [::fileutil::fileType $storageDirectory/$file]
    set extension [lindex $types end]
    if {![string length $extension]} {
      set extension txt
    }
    ## Read the file as a binary one...
    set text_file [file rootname $file].$extension
    set data [data $storageDirectory/$file]
    ## Convert from original encoding to utf-8...
    set data [string trim [$encoding_detector changeHTMLDataEncoding $data]]
    ## Save the file as utf-8...
    store $storageDirectory/$text_file $data
    ## Update the HTML processor...
    $html_processor setHTML $data
    return $text_file
  };# generateTextResource

  ##########################################################################
  ## Method: fileSystemNameFromURI
  ##########################################################################
  method fileSystemNameFromURI {uri args} {
    ## Get the url scheme...
    array set uriparts [uri::split $uri]
    ## Create a safe name that includes the host and the rest of the url as an
    ## md5 hash...
    set host $uriparts(host)
    unset uriparts(host)
    set url [uri::join {*}[array get uriparts]]
    return [list $host [::md5::md5 -hex $url]]
  };# fileSystemNameFromURI

  ##########################################################################
  ## Method: saveResourceInGlobalIndex
  ##########################################################################
  method saveResourceInGlobalIndex {uri host url_md5 resource_dir
                                    resource_binary resource_text} {
    set fd [open $storageDirectory/$host/index.txt a]
    fconfigure $fd -encoding utf-8
    puts $fd "$url_md5\t$uri\t$resource_binary\t$resource_dir\t[clock \
                               format [clock seconds]]"
    close $fd
  };# saveResourceInGlobalIndex

  ##########################################################################
  ## Method: saveResourceInformation
  ##########################################################################
  method saveResourceInformation {uri host url_md5 resource_dir
                                  resource_binary resource_text
                                  {filename uri.xml}} {
    set fd [open $storageDirectory/$resource_dir/$filename w]
    fconfigure $fd -encoding utf-8
    puts $fd {<?xml version="1.0" encoding="UTF-8"?>}
    puts $fd \
       {<resource_info xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">}
    puts $fd "  <uri>$uri</uri>"
    puts $fd "  <host>$host</host>"
    puts $fd "  <id>$url_md5</id>"
    puts $fd "  <directory>$resource_dir</directory>"
    puts $fd "  <original>$resource_binary</original>"
    puts $fd "  <possible_types>[::fileutil::fileType\
                         $storageDirectory/$resource_binary]</possible_types>"
    puts $fd "  <mime_type>[::fileutil::magic::mimetype\
                              $storageDirectory/$resource_text]</mime_type>"
    puts $fd "  <file_type>[::fileutil::magic::filetype\
                              $storageDirectory/$resource_text]</file_type>"
    puts $fd "  <date>[clock format [clock seconds]]</date>"
    puts $fd {</resource_info>}
    close $fd
  };# saveResourceInformation

  proc data {filename {encoding binary}} {
    set fd [open $filename]
    fconfigure $fd -encoding $encoding
    set data [read $fd]
    close $fd
    return $data
  };# data

  proc store {filename data {encoding utf-8}} {
    set fd [open $filename w]
    fconfigure $fd -encoding $encoding
    puts $fd $data
    close $fd
  };# store

  method log {msg {type {}}} {
    switch $type {
      error   {puts stderr $msg}
      default {puts $msg}
    }
    update
  };# log

  method test {} {
    downloadURI http://www.google.gr
    downloadURI http://www.ellogon.org
  };# test

};# class ELEP::FileUtilities::Crawler

package provide ELEP::FileUtilities::Crawler 1.0

if {1} {
  package require ELEP::FileUtilities::Crawler
  set c [ELEP::FileUtilities::Crawler #auto]
  $c test
  itcl::delete object $c
}
