##############################################################################
##  HTMLPageDownloader.tcl:                                                 ##
## --------------------------------------------------                       ##
##  This class implements methods for downloading an HTML page.             ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::FileUtilities::URI
package require ELEP::DocumentUtilities::EncodingDetection
package require ELEP::DocumentUtilities::HTMLUtilities
package require md5
package require fileutil::magic::mimetype
package require fileutil::magic::filetype
package require fileutil
package require uri::urn
package require log

oo::class create ELEP::FileUtilities::HTMLPageDownloader {
  variable storageDirectory            \
           resourceIsText              \
           resourceIsHTML              \
           resourceBaseURI             \
           resourceTextFullPathName    \
           resourceBinaryFullPathName  \
           resourceURI                 \
           preprocessTextCmd           \
           downloadResources           \
           encoding_detector           \
           html_processor              \
           already_seen                \
           resource_counter

  constructor {args} {
    set storageDirectory $::CDM_TempDir/HTMLPageDownloader
    set resourceIsText              0
    set resourceIsHTML              0
    set resourceBaseURI            {}
    set resourceTextFullPathName   {}
    set resourceBinaryFullPathName {}
    set resourceURI                {}
    set preprocessTextCmd          {}
    set downloadResources        true

    set encoding_detector {}
    set html_processor    {}

    my configure {*}$args
    my init
  };# constructor

  destructor {
    foreach object {encoding_detector html_processor} {
      if {[string length [set $object]]} {
        catch {[set $object] destroy}
      }
    }
  };# destructor

  method init {} {
    set encoding_detector [ELEP::DocumentUtilities::EncodingDetection new]
    set html_processor    [ELEP::DocumentUtilities::HTMLUtilities new]
  };# init

  ##########################################################################
  ## Method: downloadURI
  ##########################################################################
  method downloadURI {uri args} {
    set resource_counter 0
    # log::logMsg "downloadURI: $uri"
    ## Create a suitable name from the uri
    foreach {host url_md5} [my fileSystemNameFromURI $uri] {break}
    set file_system_uri_name $host/$url_md5
    ## Download the url in binary mode...
    file mkdir $storageDirectory/$file_system_uri_name
    set local_resource_file_name $file_system_uri_name/uri.bin
    if {[file exists $storageDirectory/$local_resource_file_name]} {
      ## Do what?
    }
    ELEP::FileUtilities::URI urlfile $uri \
          $storageDirectory/$local_resource_file_name binary
    set text_resource $local_resource_file_name
    set resourceIsText [my resourceIsText $text_resource]
    set resourceIsHTML 0
    if {$resourceIsText} {
      set text_resource  [my generateTextResource $text_resource]
      set resourceIsHTML [$html_processor isHTML]
      if {$resourceIsHTML} {
        set html_resource [my generateHTMLResource $text_resource]
      }
    }
    set resourceTextFullPathName   $storageDirectory/$text_resource
    set resourceBinaryFullPathName $storageDirectory/$local_resource_file_name
    set resourceURI                $uri
    set resourceBaseURI            $uri
    ## Save download information into the indexes...
    my saveResourceInGlobalIndex $uri $host $url_md5 $file_system_uri_name \
                                      $local_resource_file_name $text_resource
    my saveResourceInformation   $uri $host $url_md5 $file_system_uri_name \
                                      $local_resource_file_name $text_resource
    ## If this is a text resource, download additional resources like images,
    ## etc...
    if {$resourceIsText && $downloadResources} {
      set resourceBaseURI [$html_processor getBase $uri]
      my downloadURIResources $resourceBaseURI $text_resource
    }
    return $text_resource
  };# downloadURI

  ##########################################################################
  ## Method: downloadURIResources
  ##########################################################################
  method downloadURIResources {base filename args} {
    if {!$resourceIsHTML} {return {}}
    $html_processor removeTag base
    set data [$html_processor getHTML]
    set dir [file dirname $filename]/res
    ## Make the resources subdirectory...
    file mkdir $storageDirectory/$dir
    array unset already_seen
    ##
    ## Download images...
    ##
    file mkdir $storageDirectory/$dir/img
    set time [lindex [time {
      if {[catch {
        foreach url [$html_processor getImageLinks] {
          if {![string length $url]} {continue}
          if {[catch {my downloadResource $base $url $dir/img} resource]} {
            log::logError "<img>: $resource\n$::errorInfo"
            continue
          }
          ## Substitute the link with the local file!
          set data [string map [list $url res/img/$resource] $data]
        }
      } error]} {
        log::logError "Malformed HTML? $error"
        log::logError $::errorInfo
      }
    }] 0]
    log::log debug "Images time: [expr {$time/1000000.0}] s"
    ##
    ## Download images in stylesheets...
    ##
    set time [lindex [time {
      if {[catch {
        foreach url [$html_processor getHTMLURLs] {
          if {![$html_processor isImageURL $url]} {continue}
          if {[catch {my downloadResource $base $url $dir/img} resource]} {
            log::logError "url(): $resource\n$::errorInfo"
            continue
          }
          ## Substitute the link with the local file!
          set data [string map [list $url res/img/$resource] $data]
        }
      } error]} {
        log::logError "Malformed HTML? $error"
        log::logError $::errorInfo
      }
    }] 0]
    log::log debug "Images (CSS) time: [expr {$time/1000000.0}] s"
    ##
    ## Download stylesheets and icons from link tags...
    ##
    set time [lindex [time {
      if {[catch {
        foreach tag [$html_processor getTags link] {
          set url [my downloadLink? $tag]
          if {![string length $url]} {continue}
          if {[catch {my downloadResource $base $url $dir/img} resource]} {
            log::logError "<link>: $resource\n$::errorInfo"
            continue
          }
          ## Substitute the link with the local file!
          set data [string map [list $url res/img/$resource] $data]
        }
      } error]} {
        log::logError "Malformed HTML? $error"
        log::logError $::errorInfo
      }
    }] 0]
    log::log debug "<link> time: [expr {$time/1000000.0}] s"
    ##
    ## Download other tags...
    ##
    foreach {tag attr} {script src} {
      file mkdir $storageDirectory/$dir/$tag
      set time [lindex [time {
        if {[catch {
          foreach url [$html_processor getTagAttributeValues $tag $attr] {
            if {![string length $url]} {continue}
            if {[catch {my downloadResource $base $url $dir/$tag} resource]} {
              log::logError "<$tag>: $resource\n$::errorInfo"
              continue
            }
            ## Should the resource be processed further?
            set resource [my examineResource $tag $url $dir/$tag $resource]
            ## Substitute the link with the local file!
            set data [string map [list $url res/$tag/$resource] $data]
          }
        } error]} {
          log::logError "Malformed HTML? $error"
          log::logError $::errorInfo
        }
      }] 0]
      log::log debug "<$tag> time: [expr {$time/1000000.0}] s"
    }
    ##
    ## Produce the local version of the resource!
    ##
    my store $storageDirectory/[file rootname $filename]_local.html \
             [string trim $data]
  };# downloadURIResources

  ##########################################################################
  ## Method: downloadResource
  ##########################################################################
  method downloadResource {base resource_uri dir args} {
    set uri [uri::resolve $base $resource_uri]
    if {[info exists already_seen($uri)]} {
      return $already_seen($uri)
    }
    ## Get host...
    set host [ELEP::FileUtilities::URI url_part $uri host]
    ## Get the resource file name from the uri...
    set local_resource_file_name [ELEP::FileUtilities::URI url_part $uri path]
    if {![string length $local_resource_file_name]} {
      ## There is no path. This is the front page of a site!
      set local_resource_file_name index-[incr resource_counter].html
    }
    set local_resource_file_name [uri::urn::unquote \
                                     [file tail $local_resource_file_name]]
    ## Escape eny illegal to the filesystem characters...
    set local_resource_file_name [string map \
      {: _ * _ ? _ \" _ < _ > _ | _ \\ _ / _} $local_resource_file_name]
    ## Perhaps we have exhausted the maximum file path?
    if {[string length $storageDirectory/$dir/$local_resource_file_name] >
                                                                      247} {
      set local_resource_file_name \
         R-[incr resource_counter][file extension $local_resource_file_name]
    }

    ELEP::FileUtilities::URI urlfile $uri \
          $storageDirectory/$dir/$local_resource_file_name binary
    ## Save download information into the indexes...
    # saveResourceInGlobalIndex $uri $host $url_md5 $file_system_uri_name \
    #                                $local_resource_file_name $text_resource
    my saveResourceInformation $uri $host $local_resource_file_name $dir \
                               $dir/$local_resource_file_name \
                               $dir/$local_resource_file_name \
                               [file rootname $local_resource_file_name]_inf.xml
    set already_seen($uri) $local_resource_file_name
    return $local_resource_file_name
  };# downloadResource

  ##########################################################################
  ## Method: examineResource
  ##########################################################################
  method examineResource {tag uri dir resource args} {
    set filename $storageDirectory/$dir/$resource
    set extension [file extension $resource]
    switch -nocase -glob $extension {
      .css {return [my processStyleSheet $tag $uri $dir $resource]}
      default {}
    }
    return $resource
  };# examineResource

  ##########################################################################
  ## Procedure: downloadLink?
  ##########################################################################
  classmethod downloadLink? {tag} {
    foreach one {rel href type} {
      set attr($one) [ELEP::DocumentUtilities::HTMLUtilities::getAttribute \
                                                              $tag $one]
    }
    if {![string length $attr(href)]} {return {}}

    switch -nocase -glob -- $attr(rel) {
      stylesheet -
      *icon {return $attr(href)}
    }
    # log::log debug "Skipping link: rel=\"$attr(rel)\" type=\"$attr(type)\"\
    #                                href=\"$attr(href)\""
    return {}
  };# downloadLink?

  ##########################################################################
  ## Method: processStyleSheet
  ##########################################################################
  method processStyleSheet {tag css_uri dir resource args} {
    set modified 0
    set filename $storageDirectory/$dir/$resource
    set data [my data $filename]
    $html_processor setCSS $data
    ## Get the urls from the stylesheet...
    foreach url [lsort -unique [$html_processor getCSSURLs]] {
      set modified 1
      file mkdir $storageDirectory/$dir/url
      set rc [my downloadResource $css_uri $url $dir/url]
      ## Substitute the link with the local file!
      set data [string map [list $url url/$rc] $data]
    }
    ##
    ## Produce the local version of the resource!
    ##
    if {$modified} {
      my store  [file rootname $filename]_local.css [string trim $data] binary
      return [file rootname $resource]_local.css
    }
    return $resource
  };# processStyleSheet

  ##########################################################################
  ## Method: resourceIsText
  ##########################################################################
  method resourceIsText {file args} {
    foreach type [::fileutil::fileType $storageDirectory/$file] {
      switch $type {
        text - xml - html {
          return 1
        }
      }
    }
    return 0
  };# resourceIsText

  ##########################################################################
  ## Method: generateTextResource
  ##########################################################################
  method generateTextResource {file args} {
    set types [::fileutil::fileType $storageDirectory/$file]
    set extension [lindex $types end]
    if {![string length $extension]} {
      set extension txt
    }
    ## Read the file as a binary one...
    set text_file [file rootname $file].$extension
    set data [my data $storageDirectory/$file]
    ## Convert from original encoding to utf-8...
    set data [string trim [$encoding_detector changeHTMLDataEncoding $data]]
    ## Save the file as utf-8...
    my store $storageDirectory/$text_file $data
    ## Update the HTML processor...
    $html_processor setHTML [my preProcessTextResource $data]
    return $text_file
  };# generateTextResource

  ##########################################################################
  ## Method: generateHTMLResource
  ##########################################################################
  method generateHTMLResource {filename} {
    set data [$html_processor getHTML]
    my store $storageDirectory/[file rootname $filename].html \
          [string trim $data]
  };# generateHTMLResource

  ##########################################################################
  ## Method: preProcessTextResource
  ##########################################################################
  method preProcessTextResource {data} {
    if {[llength $preprocessTextCmd]} {
      return [{*}$preprocessTextCmd $data]
    }
    return $data
  };# preProcessTextResource

  ##########################################################################
  ## Method: getHTMLLinks
  ##########################################################################
  method getHTMLLinks {} {
    if {!$resourceIsHTML} {return {}}
    ## Retrieve all links...
    set resourceLinks {}
    foreach url [$html_processor getHTMLLinks] {
      ## Get an absolute uri from the link...
      lappend resourceLinks [uri::resolve $resourceBaseURI $url]
    }
    return $resourceLinks
  };# getHTMLLinks

  ##########################################################################
  ## Method: getHTML
  ##########################################################################
  method getHTML {} {
    return [$html_processor getHTML]
  };# getHTML

  ##########################################################################
  ## Method: fileSystemNameFromURI
  ##########################################################################
  method fileSystemNameFromURI {uri args} {
    ## Get the url scheme...
    array set uriparts [uri::split $uri]
    ## Create a safe name that includes the host and the rest of the url as an
    ## md5 hash...
    set host $uriparts(host)
    unset uriparts(host)
    set url [uri::join {*}[array get uriparts]]
    return [list $host [::md5::md5 -hex $url]]
  };# fileSystemNameFromURI

  ##########################################################################
  ## Method: saveResourceInGlobalIndex
  ##########################################################################
  method saveResourceInGlobalIndex {uri host url_md5 resource_dir
                                    resource_binary resource_text} {
    set fd [open $storageDirectory/$host/index.txt a]
    fconfigure $fd -encoding utf-8
    puts $fd "$url_md5\t$uri\t$resource_binary\t$resource_dir\t[clock \
                               format [clock seconds]]"
    close $fd
  };# saveResourceInGlobalIndex

  ##########################################################################
  ## Method: saveResourceInformation
  ##########################################################################
  method saveResourceInformation {uri host url_md5 resource_dir
                                  resource_binary resource_text
                                  {filename uri.xml}} {
    set fd [open $storageDirectory/$resource_dir/$filename w]
    fconfigure $fd -encoding utf-8
    puts $fd {<?xml version="1.0" encoding="UTF-8"?>}
    puts $fd \
       {<resource_info xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">}
    puts $fd "  <uri>$uri</uri>"
    puts $fd "  <host>$host</host>"
    puts $fd "  <id>$url_md5</id>"
    puts $fd "  <directory>$resource_dir</directory>"
    puts $fd "  <original>$resource_binary</original>"
    puts $fd "  <possible_types>[::fileutil::fileType\
                         $storageDirectory/$resource_binary]</possible_types>"
    puts $fd "  <mime_type>[::fileutil::magic::mimetype\
                              $storageDirectory/$resource_text]</mime_type>"
    puts $fd "  <file_type>[::fileutil::magic::filetype\
                              $storageDirectory/$resource_text]</file_type>"
    puts $fd "  <date>[clock format [clock seconds]]</date>"
    puts $fd {</resource_info>}
    close $fd
  };# saveResourceInformation

  classmethod data {filename {encoding binary}} {
    set fd [open $filename]
    fconfigure $fd -encoding $encoding
    set data [read $fd]
    close $fd
    return $data
  };# data

  classmethod store {filename data {encoding utf-8}} {
    set fd [open $filename w]
    fconfigure $fd -encoding $encoding
    puts $fd $data
    close $fd
  };# store

};# class ELEP::FileUtilities::HTMLPageDownloader

package provide ELEP::FileUtilities::HTMLPageDownloader 1.0
