##############################################################################
##  URI.tcl:                                                                ##
## --------------------------------------------------                       ##
##  This class implements methods for downloading files from a URL.         ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require uri
package require http
package require ftp
## If package tls is available, we can also handle https://...
catch {
  package require tls
  ::http::register https 443 ::tls::socket
}

oo::class create ELEP::FileUtilities::URI {
  common scp_exe pscp
  common timeout 10000 ;# in milliseconds...
  common retry       3 ;# number of retries in case of timeout...
 
  constructor {args} {
    my configure {*}$args
  };# constructor

  ##########################################################################
  ## Procedure: urldata
  ##########################################################################
  classmethod urldata {url args} {
    ## Get the url scheme...
    array set urlparts [uri::split $url]
    ## Call uri::geturl...
    set result [uri::geturl $url]
    ## Handle results...
    return [my urldata_[string tolower $urlparts(scheme)] $result]
  };# urldata

  ##########################################################################
  ## Procedure: urldata_file
  ##########################################################################
  classmethod urldata_file {token args} {
    set data [set ${token}(data)]
    unset -nocomplain $token
    return $data
  };# urldata_file

  ##########################################################################
  ## Procedure: urldata_http
  ##########################################################################
  classmethod urldata_http {token args} {
    set data [::http::data $token]
    ::http::cleanup $token
    return $data
  };# urldata_http

  ##########################################################################
  ## Procedure: urldata_https
  ##########################################################################
  classmethod urldata_https {token args} {
    return [my urldata_http $token $args]
  };# urldata_https

  ##########################################################################
  ## Procedure: urldata_ftp
  ##########################################################################
  classmethod urldata_ftp {token args} {
    set data [set ${token}(data)]
    ::ftp::cleanup $token
    return $data
  };# urldata_ftp

  ##########################################################################
  ## Procedure: urlfile
  ##########################################################################
  classmethod urlfile {url filename {encoding utf-8} args} {
    ## Get the url scheme...
    array set urlparts [uri::split $url]
    ## Handle results...
    my urlfile_[string tolower $urlparts(scheme)] $url $filename $encoding
  };# urlfile

  ##########################################################################
  ## Procedure: urlfile_file
  ##########################################################################
  classmethod urlfile_file {url filename {encoding utf-8} args} {
    array set parts [uri::split $url]
    set path $parts(path)
    if {[info exists parts(host)] && [string length $parts(host)]} {
      switch -regexp $parts(host) {
        [A-Z] {
          ## This is a drive letter!
          set path $parts(host):$parts(path)
        }
        default {
          set path $parts(host)$parts(path)
        }
      }
    }
    file copy -force $path $filename
  };# urlfile_file

  ##########################################################################
  ## Procedure: urlfile_http
  ##########################################################################
  classmethod urlfile_http {url filename {encoding utf-8} args} {
    my common timeout retry
    # puts "urlfile_http: timeout: $timeout, retry: $retry, url: $url"
    set url [my url_fix $url]
    for {set try $retry} {$try} {incr try -1} {
      set fd [open $filename w]
      fconfigure $fd -encoding $encoding
      if {[catch {::http::geturl $url -binary 1 -channel $fd -strict 0 \
                                      -protocol 1.0 -timeout $timeout} token]} {
        close $fd
        set msg "url:\"$url\"\n$token"
        catch {::http::cleanup $token}
        ## Perhaps we can download as data?
        if {[catch {::http::geturl $url -binary 1 -strict 0 \
                                   -protocol 1.0 -timeout $timeout} token]} {
          catch {::http::cleanup $token}
          error $msg
        }
        set fd [open $filename w]
        fconfigure $fd -encoding $encoding
        puts -nonewline $fd [::http::data $token]
      }
      close $fd
      set status [::http::status $token]
      catch {::http::cleanup $token}
      ## Is it a timeout?
      if {$status ne "timeout"} {break}
    }
  };# urlfile_http

  ##########################################################################
  ## Procedure: urlfile_https
  ##########################################################################
  classmethod urlfile_https {url filename {encoding utf-8} args} {
    return [my urlfile_http $url $filename $encoding]
  };# urlfile_https

  ##########################################################################
  ## Procedure: urlfile_ftp
  ##########################################################################
  classmethod urlfile_ftp {url filename args} {
    array set parts [uri::split $url]
    if {![string length $parts(port)]} {set parts(port) 21}
    if {![string length $parts(user)]} {set parts(user) ftp}
    if {![string length $parts(pwd)]}  {set parts(pwd)  ftp@gmail.com}
    if {![string length $parts(type)]} {set parts(type) binary}
    set token [::ftp::Open $parts(host) $parts(user) $parts(pwd) \
                           -port $parts(port) -mode $parts(type)]
    if {$token == -1} {
      error "cannot connect/login at FTP server $parts(host)\
             at port $parts(port)"
    }
    if {[::ftp::Get $token $parts(path) $filename] == 0 } {
      error "cannot find the specified path in the FTP server:\
             \"$parts(path)\""
    }
    ::ftp::Close $token
  };# urlfile_ftp

  ##########################################################################
  ## Procedure: urlfile_sftp
  ##########################################################################
  classmethod urlfile_sftp {url filename {encoding utf-8} args} {
    my common scp_exe
    array set parts [uri::split $url]
    if {![string length $parts(port)]} {set parts(port) 21}
    if {![string length $parts(user)]} {set parts(user) sftp}
    if {![string length $parts(pwd)]}  {set parts(pwd)  sftp@gmail.com}
    if {![string length $parts(type)]} {set parts(type) binary}
    exec $scp_exe -q -P $parts(port) -pw $parts(pwd) \
         $parts(user)@${parts(host)}:$parts(path) $filename
  };# urlfile_sftp

  ##########################################################################
  ## Procedure: uploadfile
  ##########################################################################
  classmethod uploadfile {filename url args} {
    ## Get the url scheme...
    array set urlparts [uri::split $url]
    ## Handle results...
    my uploadfile_[string tolower $urlparts(scheme)] $filename $url
  };# uploadfile

  ##########################################################################
  ## Procedure: uploadfile_file
  ##########################################################################
  classmethod uploadfile_file {filename url args} {
    array set parts [uri::split $url]
    set path $parts(path)
    if {[info exists parts(host)] && [string length $parts(host)]} {
      switch -regexp $parts(host) {
        [A_Z] {
          ## This is a drive letter!
          set path $parts(host):$parts(path)
        }
        default {
          set path $parts(host)$parts(path)
        }
      }
    }
    file copy -force $filename $path
  };# uploadfile_file

  ##########################################################################
  ## Procedure: uploadfile_http
  ##########################################################################
  classmethod uploadfile_http {filename url args} {
    error "file uploading through HTTP not implemented!"
  };# uploadfile_http

  ##########################################################################
  ## Procedure: uploadfile_https
  ##########################################################################
  classmethod uploadfile_https {filename url args} {
    return [my uploadfile_http $url $filename]
  };# uploadfile_https

  ##########################################################################
  ## Procedure: uploadfile_ftp
  ##########################################################################
  classmethod uploadfile_ftp {filename url args} {
    array set parts [uri::split $url]
    if {![string length $parts(port)]} {set parts(port) 21}
    if {![string length $parts(user)]} {set parts(user) ftp}
    if {![string length $parts(pwd)]}  {set parts(pwd)  ftp@gmail.com}
    if {![string length $parts(type)]} {set parts(type) binary}
    set token [::ftp::Open $parts(host) $parts(user) $parts(pwd) \
                           -port $parts(port) -mode $parts(type)]
    if {$token == -1} {
      error "cannot connect/login at FTP server $parts(host)\
             at port $parts(port)"
    }
    if {[::ftp::Put $token $filename $parts(path)] == 0 } {
      error "cannot find the specified path in the FTP server:\
             \"$parts(path)\""
    }
    ::ftp::Close $token
  };# uploadfile_ftp

  ##########################################################################
  ## Procedure: uploadfile_sftp
  ##########################################################################
  classmethod uploadfile_sftp {filename url args} {
    my common scp_exe
    array set parts [uri::split $url]
    if {![string length $parts(port)]} {set parts(port) 21}
    if {![string length $parts(user)]} {set parts(user) sftp}
    if {![string length $parts(pwd)]}  {set parts(pwd)  sftp@gmail.com}
    if {![string length $parts(type)]} {set parts(type) binary}
    exec $scp_exe -q -P $parts(port) -pw $parts(pwd) \
           $filename $parts(user)@${parts(host)}:$parts(path)
  };# uploadfile_sftp

  ##########################################################################
  ## Procedure: urlexists
  ##########################################################################
  classmethod urlexists {url args} {
    ## Get the url scheme...
    array set urlparts [uri::split $url]
    ## Handle results...
    return [my urlexists_[string tolower $urlparts(scheme)] $url]
  };# urlexists

  ##########################################################################
  ## Procedure: urlexists_file
  ##########################################################################
  classmethod urlexists_file {url args} {
    array set parts [uri::split $url]
    set path $parts(path)
    if {[info exists parts(host)] && [string length $parts(host)]} {
      switch -regexp $parts(host) {
        [A-Z] {
          ## This is a drive letter!
          set path $parts(host):$parts(path)
        }
        default {
          set path $parts(host)$parts(path)
        }
      }
    }
    return [file exists $path]
  };# urlexists_file

  ##########################################################################
  ## Procedure: urlexists_http
  ##########################################################################
  classmethod urlexists_http {url args} {
    set result 0
    set token [::http::geturl $url -validate 1]
    switch [::http::ncode $token] {
      200     {set result 1}
      default {set result 0}
    }
    ::http::cleanup $token
    return $result
  };# urlexists_http

  ##########################################################################
  ## Procedure: urlexists_https
  ##########################################################################
  classmethod urlexists_https {url args} {
    return [my urlexists_http $url $args]
  };# urlexists_https

  ##########################################################################
  ## Procedure: urlexists_ftp
  ##########################################################################
  classmethod urlexists_ftp {url args} {
    error "file exists through FTP not implemented!"
  };# urlexists_ftp

  ##########################################################################
  ## Procedure: url_part
  ##########################################################################
  classmethod url_part {url part args} {
    set urlparts($part) {}
    array set urlparts [uri::split $url]
    return $urlparts($part)
  };# url_part

  ##########################################################################
  ## Procedure: url_fix
  ##########################################################################
  classmethod url_fix {url args} {
    set url [string map {[ %5B ] %5D | %7C} $url]
    ## Only one '?' is allowed!
    set index [string first ? $url]
    if {$index != -1} {
      set url [string replace $url $index $index { }]
      set url [string map {? %3F} $url]
      set url [string replace $url $index $index ?]
    }
    return $url
  };# url_fix

  destructor {
  };# destructor

};# class ELEP::FileUtilities::URI

package provide ELEP::FileUtilities::URI 1.0
