##############################################################################
##  Canvas.tcl:                                                             ##
## --------------------------------------------------                       ##
##  This class implements a canvas, which acts as a container for various   ##
##  flow control items.                                                     ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities::WidgetBase
package require widget::scrolledwindow
# package require tcanvas

oo::class create ELEP::FlowControl::Canvas {
  superclass ELEP::WindowUtilities::WidgetBase

  constructor {widget_type name args} {
    my variable canvas widget background_image \
                              background_image_id \
                              background_image_source \
                              background_rect_id
    set background_image_source {}
    set background_rect_id      {}
    next widget::scrolledwindow $name
    set canvas $widget.canvas
    $widget_type $canvas {*}$args
    $widget setwidget $canvas
    set background_image    [image create photo]
    set background_image_id [$canvas create image 0 0 -anchor nw \
        -image $background_image -tags TiledBackgroundImage]
    $canvas lower $background_image_id
    $canvas create line 0 0 0 0 -tags anchor00
    bind $canvas <Configure> [list [self] onConfigure]
  };# constructor

  ##########################################################################
  ## Method: encoding
  ##########################################################################
  method encoding {filename} {
  };# encoding

  ##########################################################################
  ## Method: update
  ##########################################################################
  method update {} {
    my variable canvas
    ::update idletasks
    catch {event generate $canvas <<ItemMoved>>}
    catch {
      $canvas configure -scrollregion [$canvas bbox {item||connector||anchor00}]
    }
  };# update

  ##########################################################################
  ## Method: createBackgroundRectangle
  ##########################################################################
  method createBackgroundRectangle {rect} {
    my variable background_rect_id canvas
    if {$background_rect_id eq ""} {
      ## Create the element that can be filled...
      set background_rect_id [$canvas create $rect 0 0 0 0 \
        -tags TiledBackgroundRectangle]
      $canvas lower $background_rect_id
    }
  };# createBackgroundRectangle

  ##########################################################################
  ## Method: fill
  ##########################################################################
  method fill {colour {rect prect}} {
    my variable canvas
    my createBackgroundRectangle $rect
    $canvas itemconfigure TiledBackgroundRectangle -fill $colour
  };# fill

  ##########################################################################
  ## Method: tile
  ##########################################################################
  method tile {image} {
    my variable canvas background_image background_image_source
    set background_image_source $image
    if {[string length $background_image_source]} {
      $background_image copy $background_image_source \
          -to 0 0 [winfo screenwidth $canvas] [winfo screenheight $canvas]
    } else {
      $background_image blank
      $background_image configure -width 0 -height 0
    }
  };# tile

  ##########################################################################
  ## Method: canvas
  ##########################################################################
  method canvas {} {
    my variable canvas
    return $canvas
  };# canvas

  ##########################################################################
  ## Method: see
  ##########################################################################
  method see { items } {
    my variable canvas
    set box [eval $canvas bbox $items]
    if {![llength $box]} { return }

    if {![llength [$canvas cget -scrollregion]]} {
      # People really should set -scrollregion you know...
      foreach {x y x1 y1} $box break

      set x [expr round(2.5 * ($x1+$x) / [winfo width $canvas])]
      set y [expr round(2.5 * ($y1+$y) / [winfo height $canvas])]

      $canvas xview moveto 0
      $canvas yview moveto 0
      $canvas xview scroll $x units
      $canvas yview scroll $y units
    } else {
      # If -scrollregion is set properly, use this
      foreach { x y x1 y1  } $box break
      foreach { top  btm   } [$canvas yview] break
      foreach { left right } [$canvas xview] break
      foreach { p q xmax ymax } [$canvas cget -scrollregion] break

      set xpos [expr (($x1+$x) / 2.0) / $xmax - ($right-$left) / 2.0]
      set ypos [expr (($y1+$y) / 2.0) / $ymax - ($btm-$top)    / 2.0]

      $canvas xview moveto $xpos
      $canvas yview moveto $ypos
    }
  };# see

  ##########################################################################
  ## Event: onConfigure
  ##########################################################################
  method onConfigure {} {
    my variable canvas
    set w [winfo width  $canvas]
    set h [winfo height $canvas]
    set bbox [$canvas bbox {item||connector||anchor00}]
    if {[llength $bbox] != 4} {
      incr w 10; incr h 10
      catch {$canvas coords TiledBackgroundRectangle -5 -5 $w $h}
      return
    }
    foreach {_ _ W H} $bbox {break}
    if {$w < $W} {set w $W}
    if {$h < $H} {set h $H}
    incr w 10; incr h 10
    catch {$canvas coords TiledBackgroundRectangle -5 -5 $w $h}
    after idle [list catch [list $canvas configure -scrollregion $bbox]]
  };# onConfigure

  destructor {
    my variable background_image
    catch {image delete $background_image}
    next
  };# destructor

};# class ELEP::FlowControl::Canvas

package provide ELEP::FlowControl::Canvas 1.0
