##############################################################################
##  ComponentItem.tcl:                                                      ##
## --------------------------------------------------                       ##
##  This class implements an item that is a wrapper for an Ellogon          ##
##  component.                                                              ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::FlowControl::GradientItem

proc ELEP::FlowControl::ComponentItem_StateIdToName {id} {
  switch $id {
    0 {return inactive}
    1 {return enabled}
    2 {return active}
    3 {return running}
    4 {return completed}
    5 {return error}
    default {error "unknown state id: $id"}
  }
};# ELEP::FlowControl::ComponentItem_StateIdToName
variable ELEP::FlowControl::ComponentItem_StateIdToName
array set ELEP::FlowControl::ComponentItem_StateIdToName \
{0 inactive 1 enabled 2 active 3 running 4 completed 5 error}

proc ELEP::FlowControl::ComponentItem_StateNameToId {name} {
  switch $name {
    inactive  {return 0}
    enabled   {return 1}
    active    {return 2}
    running   {return 3}
    completed {return 4}
    error     {return 5}
    default {error "unknown state name: $name"}
  }
};# ELEP::FlowControl::ComponentItem_StateNameToId
variable ELEP::FlowControl::ComponentItem_StateNameToId
array set ELEP::FlowControl::ComponentItem_StateNameToId \
{inactive 0 enabled 1 active 2 running 3 completed 4 error 5}

oo::class create ELEP::FlowControl::ComponentItem {
  superclass ELEP::FlowControl::GradientItem
  # superclass ELEP::FlowControl::Item

  constructor {c args} {
    my variable backgroundColour run reset enable_outputs col doc \
                component_state enabled_output_nodes viewer_labels \
                viewer_commands label component_description \
                parent_objects component system_component \
                system_component_obj
    # set backgroundColour;# This is an array

    set component             {}
    set component_description {}
    set run            {puts "run:               $label ($col->$doc)"}
    set reset          {puts "reset:             $label ($col->$doc)"}
    set enable_outputs {puts "enableOutputNodes: $label"; return 0}

    set col {};# TODO {setCollection}
    set doc {};# TODO {setDocument}
    # Valid states are: inactive enabled active running completed
    set component_state 0;#inactive
    set enabled_output_nodes {0}

    set viewer_labels   {}
    set viewer_commands {}

    set parent_objects  {}
    set system_component 0
    set system_component_obj {}

    next $c {*}$args
    my onComponentStateChange
  };# constructor

  ##########################################################################
  ## Method: registerViewer
  ##########################################################################
  method registerViewer {lbl cmd} {
    my variable viewer_labels viewer_commands
    lappend viewer_labels   $lbl
    lappend viewer_commands $cmd
  };# registerViewer

  ##########################################################################
  ## Method: run
  ##########################################################################
  method run {} {
    my variable run label col doc
    if {[string length $run]} {eval $run}
  };# run

  ##########################################################################
  ## Method: enableOutputNodes
  ##########################################################################
  method enableOutputNodes {} {
    my variable enable_outputs label col doc
    if {[string length $enable_outputs]} {return [eval $enable_outputs]}
  };# enableOutputNodes

  ##########################################################################
  ## Method: reset
  ##########################################################################
  method reset {} {
    my variable system_component
    if {$system_component} {
      ## The component is not in workflow mode. Notify the parent,
      ## and let the parent handle interaction...
      return [my notifyParent onComponentReset [self]]
    }
    my variable reset label col doc
    if {[string length $reset]} {eval $reset}
  };# reset

  ##########################################################################
  ## Method: setCollection
  ##########################################################################
  method setCollection {} {
    my variable col
    my notifyOutputComponents configure -col $col
  };# setCollection

  ##########################################################################
  ## Method: setDocument
  ##########################################################################
  method setDocument {} {
    my variable doc
    my notifyOutputComponents configure -doc $doc
  };# setDocument

  ##########################################################################
  ## Method: setState
  ##########################################################################
  method setState {state} {
    my variable component_state
    set component_state [ELEP::FlowControl::ComponentItem_StateNameToId $state]
    my onComponentStateChange
  };# setState

  ##########################################################################
  ## Method: getState
  ##########################################################################
  method getState {} {
    my variable component_state
    return [ELEP::FlowControl::ComponentItem_StateIdToName $component_state]
  };# getState

  ##########################################################################
  ## Method: nextState
  ##########################################################################
  method nextState {} {
    my variable component_state
    ## If component_state is not inactive, advance to the next state.
    if {$component_state > 
           $ELEP::FlowControl::ComponentItem_StateNameToId(inactive) &&
        $component_state < 
           $ELEP::FlowControl::ComponentItem_StateNameToId(completed)} {
      incr component_state
      my onComponentStateChange
    }
  };# nextState

  ##########################################################################
  ## Method: enableComponent
  ##########################################################################
  method enableComponent {} {
    my variable component_state
    if {$component_state == 0} {incr component_state; my onComponentStateChange}
  };# enableComponent

  ##########################################################################
  ## Method: disableComponent
  ##########################################################################
  method disableComponent {} {
    my variable component_state
    if {$component_state != 0} {
      set reset [expr {$component_state == 
                $ELEP::FlowControl::ComponentItem_StateNameToId(completed)}]
      set component_state 0
      my onComponentStateChange
      if {$reset} {my reset}
    }
  };# disableComponent

  ##########################################################################
  ## Event: onMouseClick
  ##########################################################################
  method onMouseClick {_x _y} {
    my variable component_state system_component
    ## If component_state is completed, show available viewers...
    if {$component_state ==
        $ELEP::FlowControl::ComponentItem_StateNameToId(completed)} {
      switch [my notifyParent onComponentShowViewers [self]] {
        show {
          my showViewersMenu $_x $_y
        }
      }
      return 
    }
    next $_x $_y
    ## If component_state is not inactive, advance to the next state.
    if {$system_component && $component_state <
        $ELEP::FlowControl::ComponentItem_StateNameToId(running)} {
      my nextState
    }
    my notifyParent onComponentClick [self]
  };# onMouseClick

  ##########################################################################
  ## Method: ensureStateCompleted
  ##########################################################################
  method ensureStateCompleted {} {
    my variable component_state
    if {$component_state == 
        $ELEP::FlowControl::ComponentItem_StateNameToId(active)} {
      incr component_state
      my onComponentStateChange
    } else {
      error "Component \"$label\" not in active state!"
    }
  };# ensureStateCompleted

  ##########################################################################
  ## Method: onRightMouseClick
  ##########################################################################
  method onRightMouseClick {} {
    next
    switch [my notifyParent onComponentRightClick [self]] {
      reset {
        my setState enabled
        my reset
      }
    }
  };# onRightMouseClick

  ##########################################################################
  ## Method: onComponentStateChange
  ##########################################################################
  method onComponentStateChange {} {
    my variable component_state system_component
    set state $ELEP::FlowControl::ComponentItem_StateIdToName($component_state)
    my drawWithState $state
    switch $state {
      inactive  {
        ## We have to notify all output connection items...
        my notifyOutputComponents disableComponent
      }
      enabled   {
        ## We have just been enabled. Ensure all outputs are disabled!
        my notifyOutputComponents disableComponent
      }
      active    {
        my variable outputNodeNumber
        ## We have been activated. If we have only a single output, enable it.
        if {$outputNodeNumber == 1} {
          my notifyOutputComponents enableComponent
        }
      }
      running   {
        ## We got in running state!
        if {$system_component} {
          ## The component is not in workflow mode. Notify the parent,
          ## and let the parent handle interaction...
          return [my notifyParent onComponentRun [self]]
        }
        ## Is our parent is completed state? If no
        if {[catch {my notifyInputComponents ensureStateCompleted} error]} {
          ## A parent component has not been completed!
          my setState active
          error $error
        }
        ## Run the component and change its state to completed!
        update
        if {[catch {my run} error]} {
          my setState active
          error $error
        }
        my variable outputNodeNumber enabled_output_nodes
        if {$outputNodeNumber > 1} {
          if {[catch {set enabled_output_nodes [my enableOutputNodes]} error]} {
            my setState active
            error $error
          }
        } else {
          set enabled_output_nodes {0}
        }
        my setState completed
      }
      completed {
        my variable enabled_output_nodes outputNodeToConnectionItemObject
        ## We have to notify the proper output connection item!
        foreach output_node $enabled_output_nodes {
          if {![info exists outputNodeToConnectionItemObject($output_node)]} {
            continue
          }
          foreach obj $outputNodeToConnectionItemObject($output_node) {
            catch {$obj notifyTarget enableComponent}
          }
        }
      }
      error {
      }
    }
  };# onComponentStateChange

  ##########################################################################
  ## Method: notifyInputComponents
  ##########################################################################
  method notifyInputComponents {method args} {
    my variable inputNodeToConnectionItemObject
    foreach input_node [lsort -integer \
                         [array names inputNodeToConnectionItemObject]] {
      foreach obj $inputNodeToConnectionItemObject($input_node) {
        $obj notifySource $method {*}$args
      }
    }
  };# notifyInputComponents

  ##########################################################################
  ## Method: notifyOutputComponents
  ##########################################################################
  method notifyOutputComponents {method args} {
    my variable outputNodeToConnectionItemObject
    foreach output_node [lsort -integer \
                          [array names outputNodeToConnectionItemObject]] {
      foreach obj $outputNodeToConnectionItemObject($output_node) {
        catch {$obj notifyTarget $method {*}$args}
      }
    }
  };# notifyOutputComponents

  ##########################################################################
  ## Method: showViewersMenu
  ##########################################################################
  method showViewersMenu {_x _y} {
    my variable viewer_commands
    if {![llength $viewer_commands]} {return}
    set menu $canvas.viewers_menu
    catch {destroy $menu}
    menu $menu -tearoff false
    foreach lbl $viewer_labels cmd $viewer_commands {
      $menu add command -label $lbl -command $cmd
    }
    tk_popup $menu [winfo pointerx .] [winfo pointery .]
  };# showViewersMenu

  ##########################################################################
  ## Method: init
  ##########################################################################
  method init {c} {
    my variable component label inputNodeNumber outputNodeNumber
    if {[dict get $c startschain]} {
      set inputNodeNumber 0
    }
    if {[dict get $c endschain]} {
      set outputNodeNumber 0
    }
    set component [dict get $c component]
    set label     [dict get $c label]
    my setSize    [dict get $c width] [dict get $c height]
    my place      [dict get $c x]     [dict get $c y]
    my locked     true
    my draw
  };# init

  ##########################################################################
  ## Method: notifyParent
  ##########################################################################
  method notifyParent {args} {
    my variable parent_objects
    foreach one $parent_objects {
      set result [$one {*}$args]
    }
    return $result
  };# notifyParent

  ##########################################################################
  ## Method: hasRun
  ##########################################################################
  method hasRun {doc} {
    my variable system_component_obj
    if {$doc eq "" || $system_component_obj eq ""} return
    if {[$system_component_obj hasRun $doc]} {
      my setState completed
    }
  };# hasRun

  ##########################################################################
  ## Method: updateStyle
  ##########################################################################
  method updateStyle {} {
    next
    [ELEP::Style::getStyle] initFlowControlComponentItem
  };# updateStyle

  ##########################################################################
  ## Method: createGradients
  ##########################################################################
  method createGradients {} {
    my variable bgGradient gradientOrientation gradientType \
                gradientColourB canvas \
                backgroundColour backgroundGradient
    if {!$::ELEP::FlowControl::FlowControl(tkpath)} {return}
    switch $gradientOrientation {
      diagonal-down {set lineartransition {0 0 1 1}}
      default       {set lineartransition {0 0 1 1}}
    }
    foreach state {inactive enabled active running completed error} {
      catch {$canvas gradient delete $backgroundGradient($state)}
      set gradientColourA $backgroundColour($state)
      set backgroundGradient($state) [$canvas gradient create $gradientType \
          -stops [list [list 0 $gradientColourA] [list 1 $gradientColourB]] \
          -lineartransition $lineartransition]
    }
  };# createGradients

  ##########################################################################
  ## Method: drawWithState
  ##########################################################################
  method drawWithState {state} {
    my variable backgroundColour backgroundGradient \
                bgColour bgGradient gradientColourA foregroundColour \
                activeForegroundColour labelColour labelActiveColour
    set bgColour          $backgroundColour($state)
    set bgGradient        $backgroundGradient($state)
    set labelColour       $foregroundColour($state)
    set labelActiveColour $activeForegroundColour($state)
    set gradientColourA $bgColour
    my onBackgroundColourChange
  };# drawWithState

  ##########################################################################
  ## Method: onBackgroundColourChange
  ##########################################################################
  method onBackgroundColourChange {} {
    my variable canvas bgShapeId labelTextId bgGradient \
                outlineColour labelColour labelActiveColour selected
    if {$::ELEP::FlowControl::FlowControl(tkpath)} {
      $canvas itemconfigure $bgShapeId -fill $bgGradient -stroke $outlineColour
    } else {
      next
    }
    if {$selected} {
      $canvas itemconfigure $labelTextId -fill $labelActiveColour
    } else {
      $canvas itemconfigure $labelTextId -fill $labelColour
    }
  };# onBackgroundColourChange

  destructor {
    my variable canvas bgGradient backgroundGradient
    set bgGradient {}
    foreach state {inactive enabled active running completed error} {
      catch {$canvas gradient delete $backgroundGradient($state)}
    }
    next
  };# destructor

};# class ELEP::FlowControl::ComponentItem

package provide ELEP::FlowControl::ComponentItem 1.0
