##############################################################################
##  ConnectionItem.tcl:                                                     ##
## --------------------------------------------------                       ##
##  This class implements an item that can appear in a flow control object. ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::Base::serialisable

namespace eval ::ELEP {
  namespace eval FlowControl {
    variable FlowControl
    if {!$FlowControl(tkpath) && ![catch {package require tkpath}]} {
      set FlowControl(tkpath) 1
      set ::tkpath::antialias 1
    }
  };# namespace  FlowControl
};# namespace  ::ELEP

oo::class create ELEP::FlowControl::ConnectionItem {
  superclass ELEP::Base::serialisable

  constructor {c _source _sourceIndex _target _targetIndex args} {
    my variable label labelColour labelBackgroundColour \
                labelBackgroundOutineColour \
                labelActiveColour labelDisabledColour \
                length width activeWidth colour activeColour arrowMode \
                smooth canvas source sourceIndex target targetIndex \
                arcPoints arcPointCoords arrowId sourceCoverId \
                labelTextId labelBackgroundId coordsList

    set label                       {}
    set labelColour                 black
    set labelBackgroundColour       yellow
    set labelBackgroundOutineColour black
    set labelActiveColour           #008000 ;# green
    set labelDisabledColour         #b0c4de ;# lightsteelblue

    set length 40         ;# The default "length" of the item. (used by
                          # graph layout algorithms)
    set width 1           ;# The line width in pixels.
    set activeWidth 5     ;# The line width in pixels when the line is
                          ;# active
    set colour blue       ;# The line colour.
    set activeColour blue ;# The line colour when the mouse is over.
    set arrowMode last    ;# The connection arrow mode. One of none,
                          ;# first, last, both.
    set smooth 1          ;# Whether the line should be smouthed using
                               ;# bezier interpolation.
    set canvas {}         ;# The canvas the connection can be drawn in
    set source {}         ;# The start item of the connection. It must be
                          ;# an object with a GetOutputNodeCoordinates
                          ;# method
    set sourceIndex {}    ;# The connected output node of the source block
    set target      {}    ;# The end item of the connection. It must be
                          ;# an object with a GetInputNodeCoordinates
                          ;# method
    set targetIndex    {} ;# The connected input node of the target block
    set arcPoints      {} ;# An array containing the arc points.
    set arcPointCoords {} ;# The coordinates of all arc points.

    ## Canvas specific information (various parts ids).
    set arrowId           {}
    set sourceCoverId     {}
    set labelTextId       {}
    set labelBackgroundId {}

    ## Drawing cache (automatically updated)
    set coordsList {}

    next $c $_source $_sourceIndex $_target $_targetIndex
    set canvas $c 
    set source [$_source name]
    set target [$_target name]
    set sourceIndex $_sourceIndex
    set targetIndex $_targetIndex
    bind $c <<ItemMoved>> [list +catch [list [self] updateCoordinates]]
    my configure {*}$args
    $source onOutputConnectionItemAttach $sourceIndex [self]
    $target onInputConnectionItemAttach  $targetIndex [self]
  };# constructor

  ##########################################################################
  ## Method: getLength
  ##########################################################################
  method getLength {} {
    my variable length
    return $length
  };# getLength

  ##########################################################################
  ## Method: notifySource
  ##########################################################################
  method notifySource {method args} {
    my variable source
    catch {$source $method {*}$args}
  };# notifySource

  ##########################################################################
  ## Method: notifyTarget
  ##########################################################################
  method notifyTarget {method args} {
    my variable target
    catch {eval $target $method {*}$args}
  };# notifyTarget

  ## source
  #  Get or set the source object.
  method source {{_source {}}} {
    my variable source
    if {[string length $_source]} {
      set source $_source
    } else {
      return $source
    }
  };# source
  method getFrom {} {my variable source; return $source}

  ## target
  #  Get or set the target object.
  method target {{_target {}}} {
    my variable target
    if {[string length $_target]} {
      set target $_target
    } else {
      return $target
    }
  };# target
  method getTo {} {my variable target; return $target}

  ## IsConnected
  #  This method returns true if the connector is connected with the provided
  #  object (i.e. the object is either the target or the source object). Else,
  #  false is returned.
  method isConnected {obj} {
    my variable source target
    return [expr {[string equal $obj $source] || [string equal $obj $target]}]
  };# isConnected

  ## coverOrigin
  #  Hide the connected source's node, so it cannot be selected.
  method coverOrigin {} {
    my variable sourceCoverId canvas source sourceIndex colour
    set sourceCoverId [$canvas create oval \
      [$source getOutputNodeBoundingBox $sourceIndex] \
      -fill $colour -tags [list [self] connector_cover]]
  };# coverOrigin

  ## uncoverOrigin
  #  Show the connected source's node, so it can be selected.
  method uncoverOrigin {} {
    my variable canvas sourceCoverId
    catch {$canvas delete $sourceCoverId}
  };# uncoverOrigin

  ## Draw the Connection on the host canvas.
  #  Returns 1 when the block was drawn, 0 otherwise.
  method draw {} {
    my variable canvas source target
    if {[string equal $canvas {}]} {return 0}
    catch {$canvas delete [self]}
    ## If both the source & target objects are not on the canvas, do not draw
    ## the connector...
    if {![$source isVisible] && ![$target isVisible]} {return 0}
    my drawArrow
    my drawLabel
  };# draw

  ##########################################################################
  ## Method: drawArrow
  ##########################################################################
  method drawArrow {} {
    my variable coordsList source sourceIndex arcPointCoords \
                target targetIndex canvas arrowMode smooth width activeWidth \
                colour activeColour arrowId
    set coordsList [concat [$source getOutputNodeCoordinates $sourceIndex] \
                            $arcPointCoords \
                           [$target getInputNodeCoordinates  $targetIndex]]
    set arrowId [$canvas create line $coordsList -arrow $arrowMode \
      -smooth $smooth -width $width -activewidth $activeWidth \
      -fill $colour -activefill $activeColour -tags [list [self] connector]]
    $canvas raise $arrowId
  };# drawArrow

  ##########################################################################
  ## Method: drawLabel
  ##########################################################################
  method drawLabel {} {
    my variable canvas label labelTextId labelBackgroundId coordsList \
                labelColour labelDisabledColour labelBackgroundColour \
                labelBackgroundOutineColour
    if {![string length $label]} {
      set labelTextId {}
      set labelBackgroundId {}
      return
    }
    set x1 [lindex $coordsList 0]
    set y1 [lindex $coordsList 1]
    set x2 [lindex $coordsList end-1]
    set y2 [lindex $coordsList end]
    set x [expr {$x1+($x2-$x1)/2.0}]
    set y [expr {$y1+($y2-$y1)/2.0}]
    set labelTextId [$canvas create text $x $y -anchor c\
      -text $label -fill $labelColour  -justify center \
      -disabledfill $labelDisabledColour -tags [list [self] connector]]
    set labelBackgroundId [$canvas create rectangle \
      [$canvas bbox $labelTextId] -fill $labelBackgroundColour \
      -outline $labelBackgroundOutineColour -tags [list [self] connector]]
    $canvas raise $labelTextId $labelBackgroundId
  };# drawLabel

  ## drawSelected
  #  Change the appearence of the connection to look as selected.
  method drawSelected {} {
    my variable canvas arrowId activeWidth activeColour
    $canvas itemconfigure $arrowId -width $activeWidth -fill $activeColour
  };# drawSelected

  ## drawDeselected
  #  Change the appearence of the connection to look as deselected.
  method drawDeselected {} {
    my variable canvas arrowId width colour
    $canvas itemconfigure $arrowId -width $width -fill $colour
  };# drawDeselected

  ## updateCoordinates
  #  Update the connector coordinates (i.e. because one of the source, target
  #  moved).
  method updateCoordinates {} {
    my variable source sourceIndex arcPointCoords target targetIndex \
                arrowId labelTextId labelBackgroundId canvas
    if {[catch {concat [$source getOutputNodeCoordinates $sourceIndex] \
                            $arcPointCoords \
                       [$target getInputNodeCoordinates  $targetIndex]} \
                           coordsList]} {
      ## There was an error in retrieving at least one node coordinates.
      ## Perhaps the node no longer exists. We should delete ourself...
      after idle "catch {::[self] destroy}"
      return
    }
    $canvas coords $arrowId $coordsList
    if {[string length $labelTextId]} {
      set x1 [lindex $coordsList 0]
      set y1 [lindex $coordsList 1]
      set x2 [lindex $coordsList end-1]
      set y2 [lindex $coordsList end]
      set x [expr {$x1+($x2-$x1)/2.0}]
      set y [expr {$y1+($y2-$y1)/2.0}]
      catch {
        $canvas coords $labelTextId $x $y
        $canvas coords $labelBackgroundId [$canvas bbox $labelTextId]
      }
    }
  };# UpdateCoordinates

  ## supportsPropertyPage
  #  This method returnes true if a specific property page is supported by the
  #  object.
  method supportsPropertyPage {section} {
    switch $section {
      Input      -
      Output     -
      Appearence -
      States     -
      Table      -
      Tree       {return 1}
      default {return 0}
    }
  };# supportsPropertyPage
  
  ## createPropertiesPage
  #  Create a dialog to edit various properties of the current object.
  method createPropertiesPage {section frame} {
    switch $section {
      default {return disabled}
    }
    return normal
  };# createPropertiesPage

  ## name
  #  Return the fully qualified name of an object.
  method name {} {
    return [self]
  };# name

  destructor {
    my variable canvas
    catch {$canvas delete [self]}
  };# destructor
};# class ELEP::FlowControl::ConnectionItem

package provide ELEP::FlowControl::ConnectionItem 1.0
