##############################################################################
##  GraphLayoutSpring.tcl:                                                  ##
## --------------------------------------------------                       ##
##  This class implements a graph layout algorithm, which models edges as   ##
##  springs. This algorithm is a refactoring of the graph layout algorithm  ##
##  originally implemented in the TouchGraph library.                       ##
##  (http://www.touchgraph.com/) Adapted from the JPowerGraph library.      ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

##
## Load some needed packages.
##
package require TclOO    ;# TclOO adds object-oriented facilities to Tcl.

oo::class create ELEP::FlowControl::GraphLayoutSpring {
  protected {
    variable TIME_NODE_REMAINS_CHANGED 200;# Specifies how long a node remains
                                           # in a special state after it is
                                           # inserted into the graph.
    variable m_graphManager {}            ;# The the graph manager, holding the                                            # graph being manipulated by
                                           # this strategy.
    variable m_damper 0.0                 ;# The damper value. A low damper
                                           # value causes the graph to move
                                           # slowly. A value of 1 means no
                                           # damping.
    variable m_maxMotion 0.0              ;# The maximum motion value after the
                                           # last move. Used to see when the
                                           # graph is stabilizing.
    variable m_motionRatio 0.0            ;# The motion ration after the last
                                           # move.
    variable m_damping 1                  ;# If the damping is true, the damper
                                           # value decreases.
    variable m_rigidity 1.0               ;# The rigidity has the same effect
                                           # as the damper, except that it is a
                                           # constant.
    variable m_nodeMovement_m_justChanged
    variable m_nodeMovement_m_timeWhenNodeBecomesNormal
    variable m_nodeMovement_m_dx
    variable m_nodeMovement_m_dy

    variable m_executing_step 0
  }

  constructor {GraphManager} {
    set m_graphManager $GraphManager
    graphContentsChanged
  };# constructor

  ##########################################################################
  ## Method: getGraph
  ##########################################################################
  method getGraphManager {} {
    return $m_graphManager
  };# getGraphManager

  ##########################################################################
  ## Method: elementsAdded
  ##########################################################################
  ## Notifies the strategy that elements were inserted into the graph.
  method elementsAdded {{nodes {}} {edges {}}} {
    $m_graphManager elementsAdded $nodes $edges
  };# elementsAdded

  ##########################################################################
  ## Method: elementsRemoved
  ##########################################################################
  ## Notifies the strategy that elements were removed from the graph.
  method elementsRemoved {{nodes {}} {edges {}}} {
    $m_graphManager elementsRemoved $nodes $edges
  };# elementsRemoved

  ##########################################################################
  ## Method: graphContentsChanged
  ##########################################################################
  ## Called when the graph has been completely changed.
  method graphContentsChanged {} {
    array unset m_nodeMovement_m_justChanged *
    array unset m_nodeMovement_m_timeWhenNodeBecomesNormal *
    array unset m_nodeMovement_m_dx *
    array unset m_nodeMovement_m_dy *
    foreach node [$m_graphManager getNodes] {
      set m_nodeMovement_m_justChanged($node) 0
      set m_nodeMovement_m_timeWhenNodeBecomesNormal($node) 0
      set m_nodeMovement_m_dx($node) 0
      set m_nodeMovement_m_dy($node) 0
    }
    notifyGraphLayoutUpdated
  };# graphContentsChanged

  ##########################################################################
  ## Method: notifyGraphLayoutUpdated
  ##########################################################################
  ## Notifies the strategy that the graph has changed due to an event that
  ## is not caused by the strategy itself.
  method notifyGraphLayoutUpdated {} {
    set m_damping 1
    set m_damper  1.0
  };# notifyGraphLayoutUpdated

  ##########################################################################
  ## Method: executeGraphLayoutStep
  ##########################################################################
  ## Executes one step in the layout.
  method executeGraphLayoutStep {} {
    set m_executing_step 1
    for {set i 0} {$i < 10} {incr i} {
      # puts "=========================================================="
      # puts " I: $i -> relaxEdges"
      # puts "=========================================================="
      relaxEdges
      # printNodeMovementInfo
      # puts "=========================================================="
      # puts " I: $i -> adjustNodePairs"
      # puts "=========================================================="
      adjustNodePairs
      # printNodeMovementInfo
      # puts "=========================================================="
      # puts " I: $i -> moveNodes"
      # puts "=========================================================="
      moveNodes
      # printNodeMovementInfo
      # puts "=========================================================="
      # puts " I: $i -> damp"
      # puts "=========================================================="
      damp
      # printNodeMovementInfo
    }
    set m_executing_step 0
  };# executeGraphLayoutStep

  ##########################################################################
  ## Method: shouldExecuteStep
  ##########################################################################
  ## Returns true if more steps in the layout should be executed.
  method shouldExecuteStep {} {
    return [expr {!($m_damper < 0.1 && $m_damping && $m_maxMotion < 0.001)}]
  };# shouldExecuteStep

  ##########################################################################
  ## Method: relaxEdges
  ##########################################################################
  method relaxEdges {} {
    foreach edge [$m_graphManager getVisibleEdges] {
      set from [$edge getFrom]
      set to   [$edge getTo]
      set deltaX [expr {[$to getX] - [$from getX]}]
      set deltaY [expr {[$to getY] - [$from getY]}]
      set currentLength [expr {sqrt($deltaX*$deltaX+$deltaY*$deltaY)}]
      set factor [expr {$m_rigidity*$currentLength/([$edge getLength]*100.0)}]
      set dx [expr {$deltaX*$factor}]
      set dy [expr {$deltaY*$factor}]
      ## Edges pull directly in proportion to the distance between the nodes.
      ## This is good, because we want the edges to be stretchy.
      ## The edges are ideal rubberbands. They don't become springs when they
      ## are too short.  That only causes the graph to oscillate.
      applyDelta $from $dx  $dy
      applyDelta $to  [expr {-$dx}] [expr {-$dy}]
    };# foreach edge
  };# relaxEdges

  ##########################################################################
  ## Method: adjustNodePairs
  ##########################################################################
  ## Adjusts the pairs of nodes.
  method adjustNodePairs {} {
    set nodes [$m_graphManager getNodes]
    foreach node1 $nodes {
      foreach node2 $nodes {
        if {$node1 != $node2} {
          foreach {dx dy} {0 0} {}
          set deltaX [expr {[$node1 getX] - [$node2 getX]}]
          set deltaY [expr {[$node1 getY] - [$node2 getY]}]
          set currentLengthSquared [expr {$deltaX*$deltaX+$deltaY*$deltaY}]
          if {$currentLengthSquared < 0.1} {
            ## If two nodes are right on top of each other, randomly separate
            ## them...
            set dx [expr {rand()}]
            set dy [expr {rand()}]
          } elseif {$currentLengthSquared < 600*600} {
            ## 600, because we don't want deleted nodes to fly too far away.
            ## If it was sqrt(len) then a single node surrounded by many
            ## others will always look like a circle.  This might look good at
            ## first, but I think it makes large graphs look ugly plus it
            ## contributes to oscillation.  A linear function does not fall
            ## off fast enough, so you get rough edges in the 'force field'
            set dx [expr {$deltaX/$currentLengthSquared}]
            set dy [expr {$deltaY/$currentLengthSquared}]
          }
          set factor [expr {100.0 * [$node1 getRepulsion] *
                                    [$node2 getRepulsion] * $m_rigidity}]
          set dx [expr {$dx * $factor}]
          set dy [expr {$dy * $factor}]
          applyDelta $node1  $dx  $dy
          applyDelta $node2 [expr {-$dx}] [expr {-$dy}]
        };# if {$node1 != $node2}
      }
    }
  };# adjustNodePairs

  ##########################################################################
  ## Method: moveNodes
  ##########################################################################
  ## Moves nodes for the computed delta.
  method moveNodes {} {
    set lastMaxMotion $m_maxMotion
    set m_maxMotion 0
    foreach node [$m_graphManager getVisibleNodes] {
      ## The damper slows things down.  It cuts down jiggling at the last
      ## moment, and optimizes layout.  As an experiment, get rid of the damper
      ## in these lines, and make a long straight line of nodes.  It wiggles
      ## too much and doesn't straighten out.
      set dx [expr {$m_nodeMovement_m_dx($node) * $m_damper}]
      set dy [expr {$m_nodeMovement_m_dy($node) * $m_damper}]
      ## Slow down, but don't stop.  Nodes in motion store momentum.
      ## This helps when the force on a node is very low, but you still want
      ## to get optimal layout.
      set m_nodeMovement_m_dx($node) [expr {$dx/2.0}]
      set m_nodeMovement_m_dy($node) [expr {$dy/2.0}]
      if {![$node isFixed]} {
        set distanceMoved [expr {sqrt($dx*$dx+$dy*$dy)}]
        if {$distanceMoved > $m_maxMotion} {set m_maxMotion $distanceMoved}
        ## Don't move faster then 30 units at a time. Important in order
        ## to prevent severed nodes from flying away.
        if {$dx >  30} {set dx  30} elseif {$dx < -30} {set dx -30}
        if {$dy >  30} {set dy  30} elseif {$dy < -30} {set dy -30}
        $node Move $dx $dy
      }
      if {$m_nodeMovement_m_justChanged($node)} {
        if {[clock clicks -milliseconds] > 
            $m_nodeMovement_m_timeWhenNodeBecomesNormal($node)} {
          set m_nodeMovement_m_justChanged($node) 0
        }
      }
    }
    if {$m_maxMotion > 0} {
      ## subtract 1 to make a positive value mean that things are moving faster
      set m_motionRatio [expr {$lastMaxMotion/$m_maxMotion-1}]
    } else {
      set m_motionRatio 0
    }
  };# moveNodes

  ##########################################################################
  ## Method: applyDelta
  ##########################################################################
  ## Applies a delta to a node.
  method applyDelta {node dx dy} {
    if {$m_nodeMovement_m_justChanged($node)} {
      set m_nodeMovement_m_dx($node) \
          [expr {$m_nodeMovement_m_dx($node)+$dx/10.0}]
      set m_nodeMovement_m_dy($node) \
          [expr {$m_nodeMovement_m_dy($node)+$dy/10.0}]
    } else {
      set m_nodeMovement_m_dx($node) [expr {$m_nodeMovement_m_dx($node)+$dx}]
      set m_nodeMovement_m_dy($node) [expr {$m_nodeMovement_m_dy($node)+$dy}]
    }
  };# applyDelta

  ##########################################################################
  ## Method: stopDamper
  ##########################################################################
  ## Turns damping off.
  method stopDamper {} {
    set m_damping 0
    set m_damper  1.0
  };# stopDamper
  
  ##########################################################################
  ## Method: stopMotion
  ##########################################################################
  ## Stabilizes the graph gently, by setting the damper to a low value.
  method stopMotion {} {
    set m_damping 1
    if {$m_damper > 0.3} {set m_damper 0.3} else {set m_damper 0}
  };# stopMotion

  ##########################################################################
  ## Method: damp
  ##########################################################################
  ## Applies the damping.
  method damp {} {
    if {$m_damping} {
      ## This is important.  Only damp when the graph starts to move faster
      ## When there is noise, you damp roughly half the time. (Which is a lot)
      ## If things are slowing down, then you can let them do so on their own,
      ## without damping.
      if {$m_motionRatio <= 0.001} {
        ## If max motion<0.2, damp away
        ## If by the time the damper has ticked down to 0.9, maxMotion is
        ##  still>1, damp away
        ## We never want the damper to be negative though
        if {($m_maxMotion < 0.2 || ($m_maxMotion > 1 && $m_damper < 0.9)) &&
             $m_damper > 0.01} {
          set m_damper [expr {$m_damper -0.01}]
        } elseif {$m_maxMotion < 0.4 && $m_damper > 0.003} {
          ## If we've slowed down significanly, damp more aggresively 
          ##  (then the line two below)
          set m_damper [expr {$m_damper -0.003}]
        } elseif {$m_damper > 0.0001} {
          ## If max motion is pretty high, and we just started damping,
          ##  then only damp slightly
          set m_damper [expr {$m_damper -0.0001}]
        }
      }
    }
    if {$m_maxMotion < 0.001 && $m_damping} {set m_damper 0}
  };# damp

  ##########################################################################
  ## Method: onItemMove
  ##########################################################################
  method onItemMove {} {
    notifyGraphLayoutUpdated
    if {!$m_executing_step} {runLayout}
  };# onItemMove

  ##########################################################################
  ## Method: runLayout
  ##########################################################################
  method runLayout {} {
    while {[shouldExecuteStep]} {
      update
      executeGraphLayoutStep
    }
  };# runLayout

  ##########################################################################
  ## Method: printNodeMovementInfo
  ##########################################################################
  method printNodeMovementInfo {} {
    foreach node [$m_graphManager getNodes] {
      puts "$node: dx=$m_nodeMovement_m_dx($node),\
                   dy=$m_nodeMovement_m_dy($node)"
    }
  };# printNodeMovementInfo

};# class ELEP::FlowControl::GraphLayoutSpring

package provide ELEP::FlowControl::GraphLayoutSpring 1.0
