##############################################################################
##  Item.tcl:                                                               ##
## --------------------------------------------------                       ##
##  This class implements an item that can appear in a flow control object. ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::Base::serialisable
package require ELEP::Style::Style

namespace eval ::ELEP {
  namespace eval FlowControl {
    variable FlowControl
    if {!$FlowControl(tkpath) &&
         [[ELEP::Style::getStyle] cget -tkpath_available]} {
      set FlowControl(tkpath) 1
      set ::tkpath::antialias 1
    }
  };# namespace  FlowControl
};# namespace  ::ELEP

oo::class create ELEP::FlowControl::Item {
  superclass ELEP::Base::serialisable

  constructor {c args} {
    my variable canvas visible unlocked
    set canvas $c     ;# The canvas the item can be drawn in

    ## Display settings
    set visible  0    ;# If true, the item is visible.
    set unlocked 1    ;# If true, the item is "unlocked": it can be
                      ;# dragged with mouse, or its properties can be
                      ;# changed.
    my variable bgShapeId bgImageId imageId image2Id \
                labelTextId labelBackgroundId lockIndicator resizeHandleId \
                inputNodeIds outputNodeIds afterId selected
    ## Canvas specific information (various parts ids).
    set bgShapeId                    {}
    set bgImageId                    {}
    set imageId                      {}
    set image2Id                     {}
    set labelTextId                  {}
    set labelBackgroundId            {}
    set lockIndicator                {}
    set resizeHandleId               {}
    set inputNodeIds                 {}
    set outputNodeIds                {}
    set afterId                      {}
    set selected                     0
  
    my variable mouseX mouseY partsDisabled \
                inputNodeCoords outputNodeCoords \
                inputNodeToConnectionItemObject \
                outputNodeToConnectionItemObject
    ## Drawing cache (automatically updated)
    set mouseX 0
    set mouseY 0
    set partsDisabled    {}
    # set inputNodeCoords ;# array
    # set outputNodeCoords;# array

    ## ConnectionItem information
    # set inputNodeToConnectionItemObject  ;# array
    # set outputNodeToConnectionItemObject ;# array

    my updateStyle

    next $c
    my configure {*}$args
  };# constructor

  ##########################################################################
  ## Method: updateStyle
  ##########################################################################
  method updateStyle {} {
    [ELEP::Style::getStyle] initFlowControlItem
  };# updateStyle

  ## Draw the BlockItem on the host canvas.
  #  Returns 1 when the block was drawn, 0 otherwise.
  method draw {} {
    my variable canvas image image2 visible
    if {[string equal $canvas {}]} {return 0}
    ## Delete any previous drawing of the block item...
    catch {$canvas delete [self]}
    ## Draw the various elements of the block item...
    my drawBackground
    if {[string length $image]}  {my drawImage}
    if {[string length $image2]} {my drawImage 2}
    my drawLockIndicator
    my drawResizeHandle
    my drawLabel
    my drawInputNodes
    my drawOutputNodes
    ## Place the proper bindings...
    catch {my addBindings}
    set visible 1
    return 1
  };# Draw

  ## unDraw
  #  Remove the drawing of this block from the canvas...
  method unDraw {} {
    my variable canvas resizeHandleId visible
    catch {$canvas delete [self]}
    catch {$canvas delete $resizeHandleId}
    set visible 0
    ### Request a redraw on all connectors, to show/hide theirselves.
    #foreach connector [concat $sourceConnectors $targetConnectors] {
    #  $connector Draw
    #}
  };# unDraw

  ## drawSelected
  #  Change the appearence of the block to look as selected.
  method drawSelected {} {
    my variable canvas bgShapeId lockIndicator labelTextId labelActiveColour \
                selected
    set selected 1
    $canvas itemconfigure $bgShapeId -width 2
    $canvas itemconfigure $lockIndicator -state normal
    $canvas itemconfigure $labelTextId -fill $labelActiveColour
  };# drawSelected

  ## drawDeselected
  #  Change the appearence of the block to look as deselected.
  method drawDeselected {} {
    my variable canvas bgShapeId lockIndicator labelTextId labelColour selected
    set selected 0
    $canvas itemconfigure $bgShapeId -width 1
    $canvas itemconfigure $lockIndicator -state hidden
    $canvas itemconfigure $labelTextId -fill $labelColour
  };# drawDeselected

  method drawBackground {} {
    my variable bgType x1 y1 x2 y2 width height bgRadius disabledColour \
                bgStipple canvas bgColour outlineColour outlineDisabledColour \
                bgShapeId
    switch $bgType {
      rectangle -
      oval {
        set tp $bgType
        set coords [list $x1 $y1 $x2 $y2]
        if {$::ELEP::FlowControl::FlowControl(tkpath)} {
          switch $tp {
            rectangle {set tp prect}
            oval      {set tp ellipse
                       set coords [list $x $y]
                       lappend args -rx [expr {$width/2.0}] \
                                    -ry [expr {$height/2.0}]}
            default   {}
          }
        }
      }
      triangle-right {
        set tp polygon
        set coords "$x1 $y1 $x2 [expr {$y1+$height/2.0}] $x1 $y2"
      }
      triangle-left {
        set tp polygon
        set coords "$x2 $y1 $x1 [expr {$y1+$height/2.0}] $x2 $y2"
      }
      rounded-rectangle {
        if {$::ELEP::FlowControl::FlowControl(tkpath)} {
          set tp prect
          set coords [list $x1 $y1 $x2 $y2]
          lappend args -rx $bgRadius
        } else {
          set tp polygon
          set coords [my roundPolygon $canvas [list \
                              $x1 $y1 $x2 $y1 $x2 $y2 $x1 $y2]\
                        [list $bgRadius $bgRadius $bgRadius $bgRadius]]
          lappend args -smooth 0 -splinesteps 20
        }
      }
      default {error "unknown item type: \"$bgType\""}
    }
    if {[string match *.xbm $bgStipple]} {
      set _filename [string range $bgStipple 1 end]
      if {![file exists $_filename]} {
        set bgStipple [::GUI_Utilities::GetBitmapFileName $_filename]
      }
    }
    switch $tp {
      prect -
      ellipse {
        set bgShapeId [$canvas create $tp \
          $coords -strokewidth 1 -fill $bgColour -stroke $outlineColour \
          -tags [list [self] item]]
      }
      default {
        set bgShapeId [$canvas create $tp \
          $coords -width 1 -fill $bgColour \
          -disabledfill $disabledColour -outline $outlineColour \
          -disabledoutline $outlineDisabledColour \
          -stipple $bgStipple -tags [list [self] item]]
      }
    }
    if {[info exists args]} {
      $canvas itemconfigure $bgShapeId {*}$args
    }
  };# drawBackground

  method drawImage { {id {}} } {
    my variable image${id} image${id}_coords image${id}Id canvas \
                x1 y1 width height image${id}_anchor \
                activeimage${id} disabledimage${id}
    ## Convert relative coordinates to absolute coordinates...
    foreach {relx rely} [set image${id}_coords] {break}
    set image${id}Id [$canvas create image \
      [expr {$x1+$width*$relx}] [expr {$y1+$height*$rely}] \
      -anchor [set image${id}_anchor] -image [set image${id}] \
      -activeimage [set activeimage${id}] \
      -disabledimage [set disabledimage${id}] \
      -tags [list [self] item]]
  };# drawImage

  method drawLockIndicator {} {
    return
    set lockIndicator [$canvas create image \
      [expr {$x2-0}] [expr {$y1+20}] -anchor se -tags [list [self] item] \
      -image ::GUI_Utilities::Unlocked]
    if {!$unlocked} {
      $canvas itemconfigure $lockIndicator -image ::GUI_Utilities::Locked
    }
  };# drawLockIndicator

  method drawResizeHandle {} {
    my variable canvas resizeHandleId x2 y2 outlineColour unlocked
    catch {$canvas delete $resizeHandleId}
    set resizeHandleId [$canvas create polygon \
      [expr {$x2-12}] $y2 $x2 $y2 $x2 [expr {$y2-12}] \
      -outline $outlineColour -width 2]
    $canvas bind $resizeHandleId <Enter> [list [self] onResizeHandleEnter]
    $canvas bind $resizeHandleId <Leave> [list [self] onResizeHandleLeave]
    $canvas bind $resizeHandleId <<MB_1>>     "[self] onMouseClick %x %y"
    $canvas bind $resizeHandleId <<Motion_1>> "[self] dragResize %x %y; break"
    if {!$unlocked || ![my actionAllowed Resize]} {
      $canvas itemconfigure $resizeHandleId -state hidden
    }
  };# drawResizeHandle

  method drawLabel {} {
    my variable canvas labelTextId x y label labelColour width \
                labelDisabledColour labelFont labelBackgroundId \
                labelBackgroundColour
    set labelTextId [$canvas create text $x $y -anchor c\
      -text $label -fill $labelColour -width [expr {$width-4}] -justify center \
      -disabledfill $labelDisabledColour -tags [list [self] item]]
    if {[string length $labelFont]} {
      $canvas itemconfigure $labelTextId -font $labelFont
    }
    set labelBackgroundId [$canvas create rectangle \
      [$canvas bbox $labelTextId] -fill $labelBackgroundColour \
      -outline {} -tags [list [self] item]]
    $canvas raise $labelTextId $labelBackgroundId
  };# drawLabel

  method drawNodes {side number drawer} {
    my variable width height x1 y1 x2 y2
    if {$number < 1} {return}
    switch $side {
      left {
        set step [expr {$height/double($number)}]
        set y [expr {$y1+$step/2.0}]
        for {set node 0} {$node < $number} {incr node} {
          my $drawer $x1 $y $node; set y [expr {$y+$step}]
        }
      }
      right {
        set step [expr {$height/double($number)}]
        set y [expr {$y1+$step/2.0}]
        for {set node 0} {$node < $number} {incr node} {
          my $drawer $x2 $y $node; set y [expr {$y+$step}]
        }
      }
      top {
        set step [expr {$width/double($number)}]
        set x    [expr {$x1+$step/2.0}]
        for {set node 0} {$node < $number} {incr node} {
          my $drawer $x $y1 $node; set x [expr {$x+$step}]; 
        }
      }
      bottom {
        set step [expr {$width/double($number)}]
        set x    [expr {$x1+$step/2.0}]
        for {set node 0} {$node < $number} {incr node} {
          my $drawer $x $y2 $node; set x [expr {$x+$step}]
        }
      }
      default {error "unsupported side: $side"}
    }
  };# drawNodes

  method drawInputNodes {} {
    my variable inputNodeCoords inputNodeIds inputNodesSide inputNodeNumber
    array unset inputNodeCoords
    set inputNodeIds {}
    my drawNodes $inputNodesSide $inputNodeNumber drawInputNode
  };# drawInputNodes

  method drawOutputNodes {} {
    my variable outputNodeCoords outputNodeIds outputNodesSide outputNodeNumber
    array unset outputNodeCoords
    set outputNodeIds {}
    my drawNodes $outputNodesSide $outputNodeNumber drawOutputNode
    return

    if {$outputNodeNumber} {
      if {$outputNodeNumber == 1} {
        set _x [expr {$x1+$width/2.0}]
        my drawOutputNode $_x $y2 0
        return
      }
      set step [expr {$width/double($outputNodeNumber)}]
      set _x [expr {$x1+$step/2.0}]
      set outputNodeIds {}
      for {set node 0} {$node < $outputNodeNumber} {incr node} {
        my drawOutputNode $_x $y2 $node
        set _x [expr {$_x+$step}]
      }
    }
  };# drawOutputNodes

  method drawInputNode {_x _y _index} {
    my variable canvas inputNodeCoords inputNodeIds inNodeRadius \
                inNodeBgColour inNodeActiveBgColour inNodeDisabledBgColour\
                inNodeFillColour inNodeActiveFillColour inNodeDisabledFillColour
    set inputNodeCoords($_index) [list $_x $_y]
    lappend inputNodeIds [$canvas create rectangle \
      [expr {$_x-$inNodeRadius}] [expr {$_y-$inNodeRadius}] \
      [expr {$_x+$inNodeRadius}] [expr {$_y+$inNodeRadius}] \
      -fill $inNodeBgColour -activefill $inNodeActiveBgColour \
      -disabledfill $inNodeDisabledBgColour \
      -outline $inNodeFillColour -activeoutline $inNodeActiveFillColour \
      -disabledoutline $inNodeDisabledFillColour \
      -tags [list [self] [self]_inputNode inputNode in_$_index item]]
  };# drawInputNode

  method drawOutputNode {_x _y _index} {
    my variable canvas outputNodeCoords outputNodeIds outNodeRadius \
                outNodeBgColour outNodeActiveBgColour outNodeDisabledBgColour \
                outNodeFillColour outNodeActiveFillColour \
                outNodeDisabledFillColour
    set outputNodeCoords($_index) [list $_x $_y]
    lappend outputNodeIds [$canvas create oval \
      [expr {$_x-$outNodeRadius}] [expr {$_y-$outNodeRadius}] \
      [expr {$_x+$outNodeRadius}] [expr {$_y+$outNodeRadius}] \
      -fill $outNodeBgColour -activefill $outNodeActiveBgColour \
      -disabledfill $outNodeDisabledBgColour \
      -outline $outNodeFillColour -activeoutline $outNodeActiveFillColour \
      -disabledoutline $outNodeDisabledFillColour \
      -tags [list [self] [self]_outputNode outputNode out_$_index item]]
  };# drawOutputNode

  ## addBindings:
  #  This method adds the proper event callbacks (bindings) in our
  #  drawn block...
  method addBindings {} {
    my addMoveBindings
    my variable canvas
    $canvas bind [self] <Enter>  [list [self] onEnter]
    $canvas bind [self] <Leave>  [list [self] onLeave]
    $canvas bind [self] <<MB_2>> [list [self] onRightMouseClick]
  };# addBindings
  
  ## addMoveBindings
  #  This method sets the proper binding for moving a block with the mouse.
  method addMoveBindings {} {
    my variable canvas lockIndicator
    $canvas bind [self] <<MB_1>>          [list [self] onMouseClick %x %y]
    $canvas bind [self] <<Motion_1>>      [list [self] drag %x %y]
    $canvas bind $lockIndicator <<MB_1>>  [list [self] toggleLockState]
  };# addMoveBindings

  ##########################################################################
  ## Method: onInputConnectionItemAttach
  ##########################################################################
  method onInputConnectionItemAttach {inputNodeIndex connectionItemObj} {
    my variable inputNodeToConnectionItemObject
    lappend inputNodeToConnectionItemObject($inputNodeIndex) \
            $connectionItemObj
  };# onInputConnectionItemAttach

  ##########################################################################
  ## Method: onInputConnectionItemDetach
  ##########################################################################
  method onInputConnectionItemDetach {inputNodeIndex connectionItemObj} {
    my variable inputNodeToConnectionItemObject
    set index [lsearch $inputNodeToConnectionItemObject($inputNodeIndex) \
                       $connectionItemObj]
    if {$index != -1} {
      set inputNodeToConnectionItemObject($inputNodeIndex) [lreplace \
          $inputNodeToConnectionItemObject($inputNodeIndex) $index $index]
    }
  };# onInputConnectionItemDetach

  ##########################################################################
  ## Method: onOutputConnectionItemAttach
  ##########################################################################
  method onOutputConnectionItemAttach {outputNodeIndex connectionItemObj} {
    my variable outputNodeToConnectionItemObject
    lappend outputNodeToConnectionItemObject($outputNodeIndex) \
            $connectionItemObj
  };# onOutputConnectionItemAttach

  ##########################################################################
  ## Method: onOutputConnectionItemDetach
  ##########################################################################
  method onOutputConnectionItemDetach {outputNodeIndex connectionItemObj} {
    my variable outputNodeToConnectionItemObject
    set index [lsearch $outputNodeToConnectionItemObject($outputNodeIndex) \
                       $connectionItemObj]
    if {$index != -1} {
      set outputNodeToConnectionItemObject($outputNodeIndex) [lreplace \
          $outputNodeToConnectionItemObject($outputNodeIndex) $index $index]
    }
  };# onOutputConnectionItemDetach

  ##########################################################################
  ## Method: onBackgroundColourChange
  ##########################################################################
  method onBackgroundColourChange {} {
    my variable canvas bgShapeId bgColour outlineColour disabledColour \
                outlineDisabledColour
    if {$::ELEP::FlowControl::FlowControl(tkpath)} {
      $canvas itemconfigure $bgShapeId -fill $bgColour -stroke $outlineColour
    } else {
      $canvas itemconfigure $bgShapeId -fill $bgColour \
        -disabledfill $disabledColour \
        -outline $outlineColour -disabledoutline $outlineDisabledColour
    }
  };# onBackgroundColourChange

  ##########################################################################
  ## Method: onEnter
  ##########################################################################
  method onEnter {} {
  };# onEnter

  ##########################################################################
  ## Method: onLeave
  ##########################################################################
  method onLeave {} {
  };# onLeave

  ##########################################################################
  ## Method: onResizeHandleEnter
  ##########################################################################
  method onResizeHandleEnter {} {
    my variable unlocked canvas resizeHandleId
    if {$unlocked && [my actionAllowed Resize]} {
      $canvas itemconfigure $resizeHandleId -state normal
      $canvas configure -cursor bottom_right_corner
    }
  };# onResizeHandleEnter

  ##########################################################################
  ## Method: onResizeHandleLeave
  ##########################################################################
  method onResizeHandleLeave {} {
    my variable canvas
    #$canvas itemconfigure $resizeHandleId -state hidden
    $canvas configure -cursor {}
  };# onResizeHandleLeave

  ## onMouseClick
  #  This method makes the needed arrangments in order to drag a block on
  #  a canvas with the mouse. It simply stores the current mouse coordinates
  #  in order to calculate later how many pixels the mouse has moved.
  method onMouseClick {_x _y} {
    my variable canvas mouseX mouseY resizeHandleId
    set mouseX [$canvas canvasx $_x]
    set mouseY [$canvas canvasy $_y]
    $canvas raise [self]
    $canvas raise $resizeHandleId [self]
    $canvas raise connector
  };# onMouseClick

  ##########################################################################
  ## Method: onRightMouseClick
  ##########################################################################
  method onRightMouseClick {} {
  };# onRightMouseClick

  ## drag
  #  This method moves a block into a new position. onMouseClick must have been
  #  called before this method is executed, as both "mouseX" & "mouseY"
  #  must have been initialised.
  method drag {_x _y} {
    my variable canvas unlocked partsDisabled mouseX mouseY x y
    if {!$unlocked || [llength $partsDisabled] || ![my actionAllowed Drag]} {
      return
    }
    set _x [$canvas canvasx $_x]
    set _y [$canvas canvasy $_y]
    set dx [expr {$_x-$mouseX}]
    set dy [expr {$_y-$mouseY}]
    set mouseX $_x
    set mouseY $_y
    my move $dx $dy
    my showTemporarily "(x:$x, y:$y)"
  };# drag

  ## move
  #  Move the block by ($dx, $dy) pixels.
  method move {{dx 0} {dy 0} {force 0}} {
    my variable x y x1 y1 x2 y2 width height inNodeRadius \
                inputNodeCoords outputNodeCoords canvas resizeHandleId
    if {!$force} {
      if {![my actionAllowed Move]} {return}
      set _x [expr {$x+$dx}]; set _y [expr {$y+$dy}]
      if {$_x-$width/2.0 < 0.0 || $_x+$width/2.0 > 3000.0} {return}
      if {$_y-$height/2.0-$inNodeRadius < 0.0 ||
          $_y+$height/2.0 > 3000.0} {return}
    }
    $canvas move [self] $dx $dy
    ## The $resizeHandleId is not part of [self], so we have to move it
    ## separetely...
    $canvas move $resizeHandleId $dx $dy
    ## Update the block's coordinates...
    set x  [expr {$x+$dx}]
    set y  [expr {$y+$dy}]
    set x1 [expr {$x-$width/2.0}]
    set y1 [expr {$y-$height/2.0}]
    set x2 [expr {$x+$width/2.0}]
    set y2 [expr {$y+$height/2.0}]
    ## Update the node coordinates...
    foreach index [array names inputNodeCoords] {
      foreach {_x _y} $inputNodeCoords($index) {}
      set inputNodeCoords($index) [list [expr {$_x+$dx}] [expr {$_y+$dy}]]
    }
    foreach index [array names outputNodeCoords] {
      foreach {_x _y} $outputNodeCoords($index) {}
      set outputNodeCoords($index) [list [expr {$_x+$dx}] [expr {$_y+$dy}]]
    }
    if {!$force} {
    event generate $canvas <<ItemMoved>>
    catch {$canvas configure -scrollregion [$canvas bbox !TiledBackgroundImage]}
    }
  };# move

  ## place
  #  Place the block into the provided coordinates...
  method place {_x _y {force 0}} {
    my variable x y
    my move [expr {$_x-$x}] [expr {$_y-$y}] $force
  };# place

  ##########################################################################
  ## Method: setSize
  ##########################################################################
  method setSize {w h} {
    my variable width height x y inNodeRadius
    set width  $w
    set height $h
    if {$x-$width/2.0 < 0.0} {Move [expr {abs($x-$width/2.0)}]}
    if {$y-$height/2.0-$inNodeRadius < 0.0} {
      my move 0 [expr {abs($y-$height/2.0-$inNodeRadius)}]
    }
    my resize 0 0
  };# setSize

  ## dragResize
  #  This method resizes a block using the mouse. onMouseClick must have been
  #  called before this method is executed, as both "mouseX" & "mouseY"
  #  must have been initialised.
  method dragResize {_x _y} {
    my variable unlocked partsDisabled mouseX mouseY width height
    if {!$unlocked || [llength $partsDisabled] ||
                     ![my actionAllowed DragResize]} {
      return
    }
    set _x [$canvas canvasx $_x]
    set _y [$canvas canvasy $_y]
    set dx [expr {$_x-$mouseX}]
    set dy [expr {$_y-$mouseY}]
    set mouseX $_x
    set mouseY $_y
    my resize $dx $dy
    my showTemporarily "(w:$width, h:$height)"
  };# dragResize
  
  ## resize
  #  This method resizes the block by $_x pixels in the horizontal and by $_y
  #  pixels in the vertical direction.
  method resize {dx dy} {
    my variable canvas width height x y x1 y1 x2 y2 bgType bgRadius \
                labelTextId labelBackgroundId lockIndicator resizeHandleId \
                bgShapeId
    if {![my actionAllowed Resize]} {return}
    ## Do not allow the block to become very small...
    if {($width < 80 && $dx < 0) || ($height < 50 && $dy < 0)} {return}
    set width  [expr {$width  + $dx}]
    set height [expr {$height + $dy}]
    $canvas move $resizeHandleId $dx $dy
    ## Update the block's coordinates...
    set x  [expr {$x+$dx/2.0}]
    set y  [expr {$y+$dy/2.0}]
    set x1 [expr {$x-$width/2.0}]
    set y1 [expr {$y-$height/2.0}]
    set x2 [expr {$x+$width/2.0}]
    set y2 [expr {$y+$height/2.0}]
    ## Avoid calling Draw, as this will destroy the object in $resizeHandleId,
    ## which will terminate any drag actions. Try to resize everything...
    switch $bgType {
      rectangle -
      oval      {$canvas coords $bgShapeId $x1 $y1 $x2 $y2}
      triangle-right {
        $canvas coords $bgShapeId $x1 $y1 $x2 [expr {$y1+$height/2.0}] $x1 $y2
      }
      triangle-left {
        $canvas coords $bgShapeId $x2 $y1 $x1 [expr {$y1+$height/2.0}] $x2 $y2
      }
      rounded-rectangle {
      $canvas coords $bgShapeId \
        [my roundPolygon $canvas [list $x1 $y1 $x2 $y1 $x2 $y2 $x1 $y2] \
                        [list $bgRadius $bgRadius $bgRadius $bgRadius]]
      }
    }
    $canvas coords $labelTextId $x $y
    $canvas coords $labelBackgroundId [$canvas bbox $labelTextId]
    $canvas coords $lockIndicator [expr {$x2-0}] [expr {$y1+20}]
    ## As the input/output nodes are complex to move, destroy and redraw all of
    ## them...
    $canvas delete [self]_inputNode [self]_outputNode
    my drawInputNodes
    my drawOutputNodes
    event generate $canvas <<ItemMoved>>
    catch {$canvas configure -scrollregion [$canvas bbox !TiledBackgroundImage]}
  };# resize

  ## getInputNodeCoordinates
  #  Return the coordinates (x,y) of the provided node index.
  method getInputNodeCoordinates {_index} {
    my variable visible graphObj inputNodeNumber inputNodeCoords
    if {!$visible && [string equal [$graphObj info class] ::GraphTreeWidget]} {
      set visibleParent [$graphObj getVisibleContainerBlock [name]]
      if {[string length $visibleParent]} {
        return [list [$visibleParent cget -x] [$visibleParent cget -y]]
      }
    }
    if {$_index >= $inputNodeNumber} {
      ## Request for a node we don't have. Expand our nodes to cover it...
      set inputNodeNumber [expr {$_index+1}]
      my drawInputNodes
    }
    return $inputNodeCoords($_index)
  };# getInputNodeCoordinates

  ## getOutputNodeCoordinates
  #  Return the coordinates (x,y) of the provided node index.
  method getOutputNodeCoordinates {_index} {
    my variable visible graphObj outputNodeNumber outputNodeCoords
    if {!$visible && [string equal [$graphObj info class] ::GraphTreeWidget]} {
      set visibleParent [$graphObj GetVisibleContainerBlock [name]]
      if {[string length $visibleParent]} {
        return [list [$visibleParent cget -x] [$visibleParent cget -y]]
      }
    }
    if {$_index >= $outputNodeNumber} {
      ## Request for a node we don't have. Expand our nodes to cover it...
      set outputNodeNumber [expr {$_index+1}]
      my drawOutputNodes
    }
    return $outputNodeCoords($_index)
  };# getOutputNodeCoordinates

  method showTemporarily {msg} {
    my variable afterId canvas labelTextId
    catch {after cancel $afterId}
    catch {
      $canvas itemconfigure $labelTextId -text $msg
      set afterId [after 1000 [list [self] restoreLabel]]
    }
  };# showTemporarily

  method restoreLabel {} {
    my variable canvas labelTextId label
    catch {$canvas itemconfigure $labelTextId -text $label}
  };# restoreLabel

  ## ActionAllowed
  #  This method returns true, if the provided action is permitted according
  #  to the current editing mode.
  method actionAllowed {action} {
    return 1
    switch -exact $::MainWindow::EditingMode {
      full {return 1}
      view {
        switch -glob -- $action {
          Drag -
          Move -
          DragResize -
          EditDescription -
          ChangeState -
          Resize  {return 0}
          Page_AnswerQuestions -
          Page_Probabilities {return 1}
          Page_*  {return 0}
          default {return 1} 
        }
      }
    }
  };# ActionAllowed

  method name         {} {return [self]}
  method getX         {} {my variable x;         return $x}
  method getY         {} {my variable y;         return $y}
  method getW         {} {my variable width;     return $width}
  method getH         {} {my variable height;    return $height}
  method getXY        {} {my variable x y;       return [list $x $y]}
  method getRepulsion {} {my variable repulsion; return $repulsion}
  method isVisible    {} {my variable visible;   return $visible}
  method isFixed      {} {return 0}

  ##########################################################################
  ## Method: locked
  ##########################################################################
  method locked {args} {
    my variable unlocked canvas resizeHandleId
    if {[llength $args]} {
      set unlocked [expr {![lindex $args 0]}]
      if {$unlocked} {
        if {[my actionAllowed Resize]} {
          $canvas itemconfigure $resizeHandleId -state normal
        }
      } else {
        $canvas itemconfigure $resizeHandleId -state hidden
      }
    } else {
      return [expr {!$unlocked}]
    }
  };# locked

  ##########################################################################
  ## Method: getDirectParents
  ##########################################################################
  method getDirectParents {} {
    my variable inputNodeToConnectionItemObject
    set parents {}
    foreach input_node [lsort -integer \
                         [array names inputNodeToConnectionItemObject]] {
      foreach obj $inputNodeToConnectionItemObject($input_node) {
        lappend parents [$obj Source]
      }
    }
    return $parents
  };# getDirectParents

  ##########################################################################
  ## Method: getDirectParentConnectors
  ##########################################################################
  method getDirectParentConnectors {} {
    my variable inputNodeToConnectionItemObject
    set parents {}
    foreach input_node [lsort -integer \
                         [array names inputNodeToConnectionItemObject]] {
      set parents [concat $parents \
                          $inputNodeToConnectionItemObject($input_node)]
    }
    return $parents
  };# getDirectParentConnectors

  ##########################################################################
  ## Method: getDirectChildren
  ##########################################################################
  method getDirectChildren {} {
    my variable outputNodeToConnectionItemObject
    set children {}
    foreach output_node [lsort -integer \
                          [array names outputNodeToConnectionItemObject]] {
      foreach obj $outputNodeToConnectionItemObject($output_node) {
        lappend children [$obj target]
      }
    }
    return $children
  };# getDirectChildren

  ##########################################################################
  ## Method: getDirectChildrenConnectors
  ##########################################################################
  method getDirectChildrenConnectors {} {
    my variable outputNodeToConnectionItemObject
    set children {}
    foreach output_node [lsort -integer \
                          [array names outputNodeToConnectionItemObject]] {
      set children [concat $children \
                           $outputNodeToConnectionItemObject($output_node)]
    }
    return $children
  };# getDirectChildrenConnectors

  destructor {
    my variable canvas resizeHandleId
    catch {$canvas delete [self]}
    catch {$canvas delete $resizeHandleId}
  };# destructor

  ##
  ## Code adapted from the wiki (http://wiki.tcl.tk/8590)
  ##

  #----------------------------------------------------------------------
  #
  # roundPolygon -- Draw a polygon with rounded corners in the canvas, based
  # off of ideas and code from "Drawing rounded rectangles"
  #
  # Parameters:
  #       w - Path name of the canvas
  #       xy - list of coordinates of the vertices of the polygon
  #       radii - list of radius of the bend each each vertex
  #       args - Other args suitable to a 'polygon' item on the canvas
  #
  # Results:
  #       Returns the canvas item number of the rounded polygon.
  #
  # Side effects:
  #       Creates a rounded polygon in the canvas.
  #
  #----------------------------------------------------------------------
  classmethod roundPolygon {w xy radii args} {
     set lenXY [llength $xy]
     set lenR [llength $radii]
     if {$lenXY != 2 * $lenR} {
         error "wrong number of vertices and radii"
     }
 
     # Walk down vertices keeping previous, current and next
     foreach {x0 y0} [lrange $xy end-1 end] break
     foreach {x1 y1} $xy break
     eval lappend xy [lrange $xy 0 1]
     set knots {}                                ;# These are the control points
 
     for {set i 0} {$i < $lenXY} {incr i 2} {
         set radius [lindex $radii [expr {$i/2}]]
         set r [winfo pixels $w $radius]
 
         foreach {x2 y2} [lrange $xy [expr {$i + 2}] [expr {$i + 3}]] break
         set z [my _RoundPolygon2 $x0 $y0 $x1 $y1 $x2 $y2 $r]
         lappend knots {*}$z
 
         foreach {x0 y0} [list $x1 $y1] break    ;# Current becomes previous
         foreach {x1 y1} [list $x2 $y2] break    ;# Next becomes current
     }
     # puts $knots
     return $knots
     set n [$w create polygon $knots -smooth 1 {*}$args]
     return $n
  }
  classmethod _RoundPolygon2 {x0 y0 x1 y1 x2 y2 radius} {
     set d [expr { 2 * $radius }]
     set maxr 0.75
 
     set v1x [expr {$x0 - $x1}]
     set v1y [expr {$y0 - $y1}]
     set v2x [expr {$x2 - $x1}]
     set v2y [expr {$y2 - $y1}]
 
     set vlen1 [expr {sqrt($v1x*$v1x + $v1y*$v1y)}]
     set vlen2 [expr {sqrt($v2x*$v2x + $v2y*$v2y)}]
     if {$d > $maxr * $vlen1} {
         set d [expr {$maxr * $vlen1}]
     }
     if {$d > $maxr * $vlen2} {
         set d [expr {$maxr * $vlen2}]
     }
 
     lappend xy [expr {$x1 + $d * $v1x/$vlen1}] [expr {$y1 + $d * $v1y/$vlen1}]
     lappend xy $x1 $y1
     lappend xy [expr {$x1 + $d * $v2x/$vlen2}] [expr {$y1 + $d * $v2y/$vlen2}]
 
     return $xy
  }

};# class ELEP::FlowControl::Item

package provide ELEP::FlowControl::Item 1.0
