##############################################################################
##  AbstractClusterer.tcl:                                                  ##
## --------------------------------------------------                       ##
##  This class implements an abstract clustering algorithm.                 ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require TclOO

oo::class create ELEP::MachineLearning::Clustering::AbstractClusterer {
  
  variable clusters_number K random_seed debug verbose instance2txt

  constructor {args} {
    set clusters_number   0; # The actual number of generated clusters
    set K                 0; # Number of clusters to generate
    set random_seed      10; # The seed we set to random
    set debug             0; # Log debug information
    set verbose           0; # Log useful information
    set instance2txt     {}; # A command prefix for converting an
                             # instance (data point) into a human readble
                             # string.
    my configure {*}$args
  };# constructor

  # buildClusterer --
  #    Generates a clusterer from a set of training data. The produced
  #    clusterer can be used to clasify new instances.
  #
  # Arguments:
  #    instances:    A list of instances serving as the training data
  # Result:
  #    -
  #
  method buildClusterer {instances} {
    expr {srand($random_seed)}
  };# buildClusterer

  # clusterInstance --
  #    Finds and returns the cluster the given instance belong to.
  #
  # Arguments:
  #    instance:     The instance whose cluster is wanted
  # Result:
  #    An integer specifying the cluster
  #
  method clusterInstance {instance} {
    set distribution [my distributionForInstance $instance]
    if {![llength $distribution]}       {error "empty distribution predicted"}
    if {[my pd_sum $distribution] <= 0} {error "unable to cluster instance"}
    return [my pd_maxIndex $distribution]
  };# clusterInstance

  # distributionForInstance --
  #    Finds and returns the cluster the given instance belong to.
  #
  # Arguments:
  #    instance:     The instance whose cluster is wanted
  # Result:
  #    An integer specifying the cluster
  #
  method distributionForInstance {instance} {
    set distribution {}
    set clusters_nu [my numberOfClusters]
    for {set i 0} {$i < $clusters_nu} {incr i} {
      lappend distribution 0.0
    }
    return [lset distribution [my clusterInstance $instance] 1.0]
  };# distributionForInstance

  # numberOfClusters --
  #    Returns the number of clusters.
  #
  # Arguments:
  #    -
  # Result:
  #    -
  #
  method numberOfClusters {} {
    return $clusters_number
  };# numberOfClusters

  #
  # Number of clusters estimator
  #

  # estimateClustersCV --
  #    Estimates the number of clusters, with the use of 10 fold cross
  #    validation.
  #
  # Arguments:
  #    -
  # Result:
  #    -
  #
  method estimateClustersCV {instances} {
    package require ELEP::MachineLearning::Clustering::NFoldCrossValidation
    set numClusters 1
    ELEP::Base::local \
      ELEP::MachineLearning::Clustering::NFoldCrossValidation nfcv \
      -instances $instances
    set max_numClusters [llength $instances]
    set numFolds [nfcv cget -N]
    set k_increased 1
    set peeking 0
    while {$k_increased} {
      set k_increased 0
      set sum 0.0
      for {set i 0} {$i < $numFolds} {incr i} {
        set cvTrain [nfcv getTrainInstances $i]
        if {$numClusters > [llength $cvTrain]} {break}
        set cvTest  [nfcv getTestInstances  $i]
        my buildClustererCV $numClusters $cvTrain
        set ev  [my evaluateClustererCV $cvTest]
        set sum [expr {$sum + $ev}]
        if {$verbose} {
          my Log "Clusters: K=$numClusters (real K=[my numberOfClusters]),\
                     Fold: $i, Estimation: $ev"
        }
      }
      set mean [expr {$sum/double($numFolds)}]
      if {$verbose} {
        my Log ===================================================================
        my Log "  Number of Clusters: $numClusters, Mean estimation: $mean"
        my Log ===================================================================
      }
      if {$numClusters > 1} {
        if {[my advanceClustererCV max_mean $mean]} {
          set peeking 0
          if {$numClusters < $max_numClusters-1} {
            set k_increased 1
            incr numClusters
          } else {
            if {$verbose} {
              my Log "  Maximum number of clusters (# of instances - 1) reached!"
            }
          }
        } else {
          if {$peeking < 1} {
            incr peeking
            set k_increased 1
            incr numClusters
          } else {
            # We must stop!
            if {$numClusters > 1} {incr numClusters -1}
          }
        }
      } else {
        set max_mean $mean
        set k_increased 1
        incr numClusters
      }
    }
    if {$verbose} {
      my Log "  ** Number of clusters: [expr {$numClusters-$peeking}]"
    }
    return [expr {$numClusters-$peeking}]
  };# estimateClustersCV

  method buildClustererCV {numClusters instances} {
    set old_K $K; set K $numClusters
    my buildClusterer $instances
    set K $old_K
  };# buildClustererCV

  # evaluateClustererCV --
  #    This method must return a double value (of a type that matches what
  #    advanceClustererCV expects), representing the "goodness" of the
  #    clustering model.
  #
  # Arguments:
  #    instances:    The set of test instances for a fold
  # Result:
  #    A value representing the goodness of the clusterer. 
  #
  method evaluateClustererCV {instances} {
    error "abstract method called"
  };# evaluateClustererCV

  method advanceClustererCV {best_value_var mean} {
    upvar $best_value_var max_mean
    if {$max_mean < $mean} {
      set max_mean $mean
      return true
    }
    return false
  };# advanceClustererCV

  #
  # Probability distributions functions
  #
  proc pd_sum {distribution} {
    set sum 0.0
    foreach {p} $distribution {
      set sum [expr {$sum+$p}]
    }
    return $sum
  };# pd_sum

  proc pd_maxIndex {distribution} {
    set max [lindex $distribution 0]
    set max_index 0
    set i 0
    foreach {p} $distribution {
      if {$p > $max} {set max $p; set max_index $i}
      incr i
    }
    return $max_index
  };# pd_maxIndex

  #
  # Logging facilities
  #
  method Log {msg {channel stdout}} {
    puts $channel $msg; update
  };# Log

  method instance2txt {instance} {
    if {[string length $instance2txt]} {
      return [uplevel #0 $instance2txt [list $instance]]
    }
    return $instance
  };# instance2txt

};# class ELEP::MachineLearning::Clustering::AbstractClusterer

package provide ELEP::MachineLearning::Clustering::AbstractClusterer 1.0
