##############################################################################
##  NFoldCrossValidation.tcl:                                               ##
## --------------------------------------------------                       ##
##  This class implements facilities for performing N fold cross            ##
##  validation.                                                             ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require TclOO

oo::class create ELEP::MachineLearning::Clustering::NFoldCrossValidation {
  
  variable fold N instances random_seed debug

  constructor {args} {
   #set fold                 ; # An array holding the instances of each fold
    set N                10; # The number of folds
    set instances        {}; # The instances to perform N-FCV on
    set random_seed      10; # The seed we set to random
    set debug             0
    my configure {*}$args
    if {[llength $instances]} {my partitionInstances}
  };# constructor

  # partitionInstances --
  #    Partition instances into the N folds.
  #
  # Arguments:
  #    instances_list:    A list of instances serving as the data (optional)
  # Result:
  #    -
  #
  method partitionInstances {{instances_list {}}} {
    if {[llength $instances_list]} {set instances $instances_list}
    expr {srand($random_seed)}
    #
    # Step 1: Randomise the set of instances...
    #
    my randomiseInstances
    #
    # Step 2: Clasify instances into the N folds...
    #
    array unset fold
    set i -1
    foreach instance $instances {
      lappend fold([incr i]) $instance
      if {$i >= $N} {set i -1}
    }
  };# partitionInstances

  # randomiseInstances --
  #    Randomise the list of instances.
  #
  # Arguments:
  #    -
  # Result:
  #    -
  #
  method randomiseInstances {} {
    set instances_nu [llength $instances]
    if {$N > $instances_nu} {set N $instances_nu}
    set new_instances {}
    while {$instances_nu} {
      set index [expr {int(rand()*$instances_nu)}]
      lappend new_instances [lindex $instances $index]
      set instances [lreplace $instances $index $index]
      set instances_nu [llength $instances]
    }
    set instances $new_instances
  };# randomiseInstances

  # getFoldInstances --
  #    Returns the list of instances classified in fold "fold_index"
  #
  # Arguments:
  #    fold_index:     An integer [0, N) specifying the fold
  # Result:
  #    A list of instances belonging to the fold_index fold
  #
  method getFoldInstances {fold_index} {
    if {$fold_index >= $N} {
      error "parameter \"fold_index\" ($fold_index) must be < \"N\" ($N)"
    }
    if {![info exists fold($fold_index)]} {
      my randomiseInstances
    }
    return $fold($fold_index)
  };# getFoldInstances

  # getTrainInstances --
  #    Returns the list of instances that can be used as training ones for
  #    fold "fold_index"
  #
  # Arguments:
  #    fold_index:     An integer [0, N) specifying the fold
  # Result:
  #    A list of instances belonging to the fold_index fold training set
  #
  method getTrainInstances {fold_index} {
    set train {}
    for {set i 0} {$i < $N} {incr i} {
      if {$i == $fold_index} {continue}
      set train [concat $train [my getFoldInstances $i]]
    }
    return $train
  };# getTrainInstances

  # getTestInstances --
  #    Returns the list of instances that can be used as testing ones for
  #    fold "fold_index"
  #
  # Arguments:
  #    fold_index:     An integer [0, N) specifying the fold
  # Result:
  #    A list of instances belonging to the fold_index fold testing set
  #
  method getTestInstances {fold_index} {
    return [my getFoldInstances $fold_index]
  };# getTestInstances

};# class ELEP::MachineLearning::Clustering::NFoldCrossValidation

package provide ELEP::MachineLearning::Clustering::NFoldCrossValidation 1.0
