##############################################################################
##  SimpleKMeans.tcl:                                                       ##
## --------------------------------------------------                       ##
##  This class implements a simple k-means clustering algorithm.            ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::MachineLearning::Clustering::AbstractClusterer

oo::class create ELEP::MachineLearning::Clustering::SimpleKMeans {
  superclass ELEP::MachineLearning::Clustering::AbstractClusterer
  
  variable clusterAssignments clusterCentroids clusterStdDev \
           iterations distance centroid {*}[info class variables \
           ELEP::MachineLearning::Clustering::AbstractClusterer]

  constructor {args} {
    set clusterAssignments {}; # A list holding the cluster for each instance
    set clusterCentroids   {}; # A list holding the cluster centroids
    set clusterStdDev      {}; # A list holding the cluster std
    set iterations          0; # Keep track of the number of iterations
                               # completed before convergence
    set distance           {}; # A command prefix for returning the distance
                               # between two instances
    set centroid           {}; # A command prefix for returning the cluster
                               # centroid of a cluster. If {}, a default
                               # implementation is used, which selects the 
                               # instance closer to the center of the
                               # cluster.
    next {*}$args
  };# constructor

  # buildClusterer --
  #    Generates a clusterer from a set of training data. The produced
  #    clusterer can be used to clasify new instances.
  #
  # Arguments:
  #    instances:    A list of instances serving as the training data
  # Result:
  #    -
  #
  method buildClusterer {instances} {
    set k $K
    expr {srand($random_seed)}
    set iterations 0
    #
    # Step 1: Select randomly k instances as the centroids of the k clusters
    #
    if {$debug} {puts "buildClusterer: K=$k"; update}
    set clusterCentroids {}
    my initialiseClusterCentroids $instances
    set k [llength $clusterCentroids]
    if {$debug} {
      puts "clusterCentroids:"
      for {set i 0} {$i < $k} {incr i} {
        puts "  k=$i, centroid: [my instance2txt [lindex $clusterCentroids $i]]"
      }
      update
    }
    #
    # Step 2: Run the k-means algorithm
    #
    set clusterAssignments {}
    foreach instance $instances {lappend clusterAssignments 0}
    set converged false
    while {!$converged} {
      incr iterations
      set converged true
      #
      # Step A: Find the cluster of each individual
      #
      if {$debug} {
        puts "Step A: Finding the cluster of each individual..."; update
      }
      array unset cluster_instances
      set i 0
      foreach instance $instances assignment $clusterAssignments {
        set new_assignment [my clusterInstance $instance]
        if {$new_assignment != $assignment} {
          lset clusterAssignments $i $new_assignment
          set converged false
        }
        lappend cluster_instances($new_assignment) $instance
        if {$debug} {
          puts "  Best cluster: $new_assignment (instance=[my instance2txt\
                  $instance], centroid: [my instance2txt \
                  [lindex $clusterCentroids $new_assignment]])"; update
        }
        incr i
      }
      set k [array size cluster_instances]
      #
      # Step B: Update the centroids
      #
      if {$debug} {
        puts "Step B: Updating the centroids (K = $k)..."; update
      }
      set empty_clusters 0
      set clusterCentroids {}
      foreach cluster [lsort -integer [array names cluster_instances]] {
        lappend clusterCentroids \
          [my clusterCentroid $cluster_instances($cluster)]
      }
      if {$debug} {
        set i -1
        foreach centre $clusterCentroids {
          puts "  Cluster [incr i] centroid: [my instance2txt $centre]"; update
        }
      }
    };# while {!$converged}
    set clusters_number $k
  };# buildClusterer

  # initialiseClusterCentroids --
  #    Initialise (randomly) the centroid of each cluster.
  #
  # Arguments:
  #    instances:     A list of instances serving as the training data
  # Result:
  #    -
  #
  method initialiseClusterCentroids {instances} {
    for {set i [expr {[llength $instances]-1}]} {$i >= 0} {incr i -1} {
      set index [expr {int(rand()*($i+1))}]
      set instance [lindex $instances $index]
      if {[lsearch -exact $clusterCentroids $instance] == -1} {
        lappend clusterCentroids $instance
      }
      lset instances $index [lindex $instances $i]
      lset instances $i     $instance
      if {[llength $clusterCentroids] == $K} {break}
    }
  };# initialiseClusterCentroids

  # clusterInstance --
  #    Finds and returns the cluster the given instance belong to.
  #
  # Arguments:
  #    instance:     The instance whose cluster is wanted
  # Result:
  #    An integer specifying the cluster
  #
  method clusterInstance {instance} {
    if {$debug > 2} {
      puts "  + clusterInstance: [my instance2txt $instance]"; update
    }
    set best_cluster 0
    if {$debug > 2} {puts "      Cluster: 0"; update}
    set min_distance [my distance $instance [lindex $clusterCentroids 0]]
    set k [llength $clusterCentroids]
    for {set i 1} {$i < $k} {incr i} {
      if {$debug > 2} {puts "      Cluster: $i"; update}
      set dist [my distance $instance [lindex $clusterCentroids $i]]
      if {$dist < $min_distance} {
        set best_cluster $i
        set min_distance $dist
      }
    }
    if {$debug > 2} {
      puts "  + clusterInstance: best cluster = $best_cluster"; update
    }
    return $best_cluster
  };# clusterInstance

  # clusterCentroid --
  #    Finds and returns the instance that is in the "center" of all given
  #    instances.
  #
  # Arguments:
  #    instances:    The list of instances, one of which should be returned
  # Result:
  #    An instance from the given instances that is the cluster centroid
  #
  method clusterCentroid {instances} {
    if {[string length $centroid]} {
      if {$debug > 3} {
        puts -nonewline \
          "        Calling \"$centroid {$instances}\": "; update
        set _centroid [uplevel #0 $centroid [list $instances]]
        puts "Centroid: [my instance2txt $_centroid]"; update
        return $_centroid
      } else {
        return [uplevel #0 $centroid [list $instances]]
      }
    }
    # Alternative implementation
    set instances_number [llength $instances]
    set _centroid [lindex $instances 0]
    if {$instances_number == 1} {return $_centroid}
    set min [expr {100*[my distance $_centroid [lindex $instances end]]}]
    for {set i 0} {$i < $instances_number} {incr i} {
      set instance [lindex $instances $i]
      set mean 0.0
      for {set j 0} {$j < $instances_number} {incr j} {
        if {$i == $j} {continue}
        set mean [expr {$mean + [my distance $instance [lindex $instances $j]]}]
      }
      set mean [expr {$mean/$instances_number}]
      if {$mean < $min} {
        set min $mean
        set _centroid $instance
      }
    }
    return $_centroid
  };# clusterCentroid

  # distance --
  #    Returns the distance between two instances.
  #
  # Arguments:
  #    instance1:    The first instance
  #    instance2:    The second instance
  # Result:
  #    A positive real number
  # Notes:
  #    The returned distance does not need to be in the range [0, 1]. The
  #    returned distance can be any positive real number
  #
  method distance {instance1 instance2} {
    if {$instance1 eq $instance2} {return 0.0}
    if {$debug > 3} {
      puts -nonewline \
        "        Calling \"$distance {$instance1} {$instance2}\": "; update
      set d [uplevel #0 $distance [list $instance1] [list $instance2]]
      puts "Distance: $d"; update
      return $d
    } else {
      return [uplevel #0 $distance [list $instance1] [list $instance2]]
    }
  };# distance

  # getClusterAssignments --
  #    Return all the instances, clasified into clusters.
  #
  # Arguments:
  #    -
  # Result:
  #    A list of size equal to the number of clusters. Each element is also a 
  #    list (the individuals of a cluster)
  #
  method getClusterAssignments {} {
    return $clusterAssignments
  };# getClusterAssignments

  # getClusterCentroids --
  #    Return all the centroids, clasified into clusters.
  #
  # Arguments:
  #    -
  # Result:
  #    A list of size equal to the number of clusters. Each element is also a 
  #    list (the centroid of a cluster)
  #
  method getClusterCentroids {} {
    return $clusterCentroids
  };# getClusterCentroids

  # evaluateClustererCV --
  #    This method calculates and returns the mean of all test instances from
  #    their cluster center.
  #
  # Arguments:
  #    instances:    The set of test instances for a fold
  # Result:
  #    A value representing the goodness of the clusterer. 
  #
  method evaluateClustererCV {instances} {
    # set k [llength $clusterCentroids]
    # set sum 0.0
    # foreach instance $instances {
    #   set min_distance [my distance $instance [lindex $clusterCentroids 0]]
    #   for {set i 1} {$i < $k} {incr i} {
    #     set dist [my distance $instance [lindex $clusterCentroids $i]]
    #     if {$dist < $min_distance} {
    #       set min_distance $dist
    #     }
    #   }
    #   set sum [expr {$sum+$min_distance}]
    # }
    # return [expr {$sum/double([llength $instances])}]
    set k [llength $clusterCentroids]
    set sum 0.0
    for {set i 0} {$i < $k} {incr i} {
      for {set j [expr {$i+1}]} {$j < $k} {incr j} {
        set sum [expr {$sum + [my distance [lindex $clusterCentroids $i] \
	               [lindex $clusterCentroids $j]]}]
      }
    }
    return $sum
  };# evaluateClustererCV

  method advanceClustererCV {best_value_var mean} {
    # upvar $best_value_var min_mean
    # if {$min_mean > $mean} {
    #   set min_mean $mean
    #   return true
    # }
    # return false
    upvar $best_value_var min_mean
    if {$min_mean < $mean} {
      set min_mean $mean
      return true
    }
    return false
  };# advanceClustererCV


};# class ELEP::MachineLearning::Clustering::SimpleKMeans

package provide ELEP::MachineLearning::Clustering::SimpleKMeans 1.0
