##############################################################################
##  ABoxViewer.tcl:                                                         ##
## --------------------------------------------------                       ##
##  This class implements a tool for viewing the ABox of an ontology.       ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities::ToplevelWindow
package require ELEP::Ontology::ABoxInstancesViewer
package require ELEP::Ontology::RacerPro
package require ELEP::FlowControl::Canvas
package require ELEP::FlowControl::GraphNode
package require ELEP::FlowControl::GraphEdge
package require ELEP::FlowControl::GraphManager
package require ELEP::FlowControl::GraphLayoutSpring
package require ELEP::FlowControl::GraphLayoutDot
package require ELEP::ViewerBase::HtmlWidgetDisplay

oo::class create ELEP::Ontology::ABoxViewer {
  superclass ELEP::WindowUtilities::ToplevelWindow

  common   reasoner_object_class          ELEP::Ontology::RacerPro
  common   graph_layout_object_class      ELEP::FlowControl::GraphLayoutDot
  common   graph_node_object_class        ELEP::FlowControl::GraphNode
  common   graph_edge_object_class        ELEP::FlowControl::GraphEdge
  variable reasoner_object                {}
  variable instances_viewer_object        {}
  variable instances_viewer_widget        {}
  variable instances_graph_canvas_object  {}
  variable instances_graph_canvas_widget  {}
  variable instances_graph_manager_object {}
  variable instances_graph_layout_object  {}
  variable document_viewer_object         {}

  public variable enable_document_viewer         1

  variable MLCi_colour #FFFFCC
  variable HLCi_colour #FFCC99
  variable Explanation_colour #CCFFCC

  protected variable I2N
  protected variable I2Offsets
  protected variable selected_individual_info {}

  constructor {args} {
    eval configure $args
    createWidgets
    createReasonerObject
    wm geometry $win 700x500
  };# constructor

  ##########################################################################
  ## Method: createWidgets
  ##########################################################################
  method createWidgets {} {
    ## Add a vertical pane.
    ttk::panedwindow $client_area.pane -orient horizontal
    ## On left, add a horizontal pane.
    ttk::panedwindow $client_area.pane.pane -orient vertical
    $client_area.pane add $client_area.pane.pane
    ## On the top of the horizontal pane, add an instances viewer.
    set instances_viewer_widget \
      [createInstancesDisplayWidget $client_area.pane.pane]
    $client_area.pane.pane add $instances_viewer_widget
    bind $instances_viewer_widget <<SelectionChanged>> \
      [list [self] onInstancesSelectionChange]
    ## On the bottom of the horizontal pane, add a listbox for showing details
    ## about the selected individual.
    ScrolledWindow $client_area.pane.pane.sw
    listbox $client_area.pane.pane.sw.listbox -listvar \
      [my varname selected_individual_info]
    $client_area.pane.pane.sw setwidget $client_area.pane.pane.sw.listbox
    $client_area.pane.pane add $client_area.pane.pane.sw
    ## On right, add a graph object.
    $client_area.pane add [createInstancesGraphDisplayWidget $client_area.pane]
    ## On the right, add also a document viewer. For now, use the HTML
    ## rendering: TODO: Fix this to be configurable!
    ## TODO TODO TODO TODO
    if {$enable_document_viewer} {
      $client_area.pane add [createDocumentDisplayWidget $client_area.pane]
    }
    ## Pack the pane widget.
    pack $client_area.pane -fill both -expand 1
    title {Ontology ABox Viewer...}
    $instances_graph_canvas_object update
  };# createWidgets

  ##########################################################################
  ## Method: createInstancesDisplayWidget
  ##########################################################################
  method createInstancesDisplayWidget {parent} {
    set instances_viewer_object [ELEP::Ontology::ABoxInstancesViewer \
        ELEP::Ontology::#auto [self] $parent]
    return [$instances_viewer_object getframe]
  };# createInstancesDisplayWidget

  ##########################################################################
  ## Method: createInstancesGraphDisplayWidget
  ##########################################################################
  method createInstancesGraphDisplayWidget {parent} {
    set instances_graph_canvas_object [ELEP::FlowControl::Canvas \
        ELEP::Ontology::cnvs#auto $parent -background white]
    set instances_graph_canvas_widget [$instances_graph_canvas_object canvas]
    set instances_graph_manager_object [ELEP::FlowControl::GraphManager \
        ELEP::Ontology::gm#auto]
    set instances_graph_layout_object [$graph_layout_object_class \
        ELEP::Ontology::gl#auto $instances_graph_manager_object]
    return [$instances_graph_canvas_object getframe]
  };# createInstancesDisplayWidget

  ##########################################################################
  ## Method: createInstancesGraphDisplayWidget
  ##########################################################################
  method createDocumentDisplayWidget {parent} {
    ## Create a scrollable HTML rendering widget...
    ScrolledWindow $parent.dvsw -borderwidth 0 -relief flat
    set display_widget [html $parent.dvsw.text]
    $parent.dvsw setwidget $display_widget
    switch [package require Tkhtml] {
      2.* {
        $display_widget configure -borderwidth 0 -relief flat
      }
      3.* {
      }
    }
    set document_viewer_object [ELEP::ViewerBase::HtmlWidgetDisplay \
        ELEP::Ontology::dv#auto [self] $display_widget]
    ## TODO: Fix this toolbar button added by the HTML rendering component
    pack forget $toolbar
    return $parent.dvsw
  };# createDocumentDisplayWidget

  ##########################################################################
  ## Method: createReasonerObject
  ##########################################################################
  method createReasonerObject {} {
    set reasoner_object [$reasoner_object_class ELEP::Ontology::rs#auto]
    $reasoner_object disconnect
  };# createReasonerObject

  ##########################################################################
  ## Method: onLoadFile
  ##########################################################################
  method onLoadFile {filename} {
    $reasoner_object connect
    $reasoner_object fullreset
    switch -glob -- [string tolower $filename] {
      *.owl   {$reasoner_object readowl $filename}
      *.racer -
      default {$reasoner_object load_abox $filename}
    }
    $instances_viewer_object updateView $reasoner_object
    updateInstancesGraph
    $reasoner_object disconnect
  };# onLoadFile

  ##########################################################################
  ## Method: onLoadDocument
  ##########################################################################
  method onLoadDocument {col doc} {
    if {$enable_document_viewer} {
      $document_viewer_object onDocumentChange $col $doc
    }
  };# onLoadDocument

  ##########################################################################
  ## Method: onInstancesSelectionChange
  ##########################################################################
  method onInstancesSelectionChange {} {
    ## Get the selected instance from the instances viewer...
    foreach instance [$instances_viewer_object cget -selected_instances] \
          info [$instances_viewer_object cget -selected_instances_description] {
      if {[info exists I2N($instance)]} {
        $instances_graph_manager_object select $I2N($instance)
        $instances_graph_canvas_object see [list $I2N($instance)]
      }
      updateInstanceDescription $info
    }
  };# onInstancesSelectionChange

  ##########################################################################
  ## Method: onInstanceClick
  ##########################################################################
  method onInstanceClick {instance} {
    if {$enable_document_viewer} {
      $document_viewer_object selectionClear
      if {[info exists I2Offsets($instance)]} {
        ## An MLC has been clicked!
        foreach {start end} $I2Offsets($instance) {break}
        $document_viewer_object setSelectionIndexes 1.0+${start}c 1.0+${end}c
      }
    }
  };# onInstanceClick

  ##########################################################################
  ## Method: updateInstanceDescription
  ##########################################################################
  method updateInstanceDescription {info} {
    set separator -------------------
    foreach {most_specific_concept properties_datatype
             properties_instance_is_source
             properties_instance_is_target} $info {break}
    set selected_individual_info \
        [list [simplify_name $most_specific_concept] {}]
    lappend selected_individual_info Properties $separator
    foreach {relation value} $properties_datatype {
      lappend selected_individual_info "[simplify_name $relation]=\"$value\""
    }
    lappend selected_individual_info {} Relations $separator
    foreach {relation value} $properties_instance_is_source {
      lappend selected_individual_info \
            "[simplify_name $relation] [simplify_name $value]"
    }
    lappend selected_individual_info {} {Related to} $separator
    foreach {relation value} $properties_instance_is_target {
      lappend selected_individual_info \
            "[simplify_name $value] [simplify_name $relation]"
    }
  };# updateInstanceDescription

  ##########################################################################
  ## Method: updateInstancesGraph
  ##########################################################################
  method updateInstancesGraph {} {
    $instances_graph_manager_object clear
    array unset I2N *
    set instances        [$instances_viewer_object cget -instances]
    set instances_simple [$instances_viewer_object cget -instances_simple_name]
    set instances_info   [$instances_viewer_object cget -instances_description]
    ## Add nodes...
    foreach instance $instances inst_simple $instances_simple \
            info     $instances_info {
      switch -glob $inst_simple {
        *character_* {
          foreach {relation value} [lindex $info 1] {
            switch -glob -- $relation {
              *hasOffset| {set Offset($instance) $value; break}
            }
          }
          continue
        }
        Phrase_* - TokenLocator_* {
          foreach {start end depicts} {0 0 {}} {break}
          foreach {relation value} [lindex $info 2] {
            switch -glob -- $value {
              *startcharacter_*| {catch {set start $Offset($value)}}
              *endcharacter_*|   {catch {set end   $Offset($value)}}
            }
            switch -glob -- $relation {
              *depicts|          {set depicts $value}
            }
          }
          set I2Offsets($depicts) [list $start $end]
          continue
        }
        Athlete_* -
        SportsCompetition_* -
        SportsRound_* -
        SportsTrial_* {set colour $HLCi_colour}
        SportsEvent_* {set colour $Explanation_colour}
        default       {
          ## This is an MLC.
          set colour $MLCi_colour
        }
      }
      ## Create a node:
      set label $inst_simple
      foreach {property value} [lindex $info 1] {
        append label "\n($value)"; break
      }
      set node [$graph_node_object_class ELEP::Ontology::n#auto \
        $instances_graph_canvas_widget -label $label \
        -gradientColourA $colour]
      $instances_graph_manager_object addNode $node
      set I2N($instance) [$node name]
      ## Add a binding for selecting the node...
      bind $instances_graph_canvas_widget <<onMouseClick:$I2N($instance)>> \
        [list +[self] onInstanceClick $instance]
    }
    ## Add edges...
    foreach role [$reasoner_object all_roles] {
      ## Get all individual pairs that are related through this role:
      switch -glob -- $role {
        *-value| {
        }
        default {
          foreach pair [$reasoner_object related_individuals $role] {
            foreach {i1 i2} $pair {break}
            if {![info exists I2N($i1)] || ![info exists I2N($i2)]} {continue}
            set edge [$graph_edge_object_class ELEP::Ontology::e#auto \
                $instances_graph_canvas_widget $I2N($i1) 0 $I2N($i2) 0]
            $instances_graph_manager_object addEdge $edge
          }
        }
      }
    }
    $instances_graph_layout_object  graphContentsChanged
    $instances_graph_layout_object  runLayout
    $instances_graph_manager_object Draw
    $instances_graph_canvas_object  update
  };# updateInstancesGraph

  proc simplify_name {name} {
    return [${reasoner_object_class}::simplify_name $name]
  };# simplify_name

  destructor {
    catch {itcl::delete object $reasoner_object}
    catch {itcl::delete object $instances_viewer_object}
    catch {itcl::delete object $instances_graph_canvas_object}
    catch {itcl::delete object $instances_graph_canvas_widget}
    catch {itcl::delete object $instances_graph_manager_object}
    catch {itcl::delete object $instances_graph_layout_object}
    next
  };# destructor

};# class ELEP::Ontology::ABoxViewer

package provide ELEP::Ontology::ABoxViewer 1.0
