##############################################################################
##  OWLParser.tcl:                                                          ##
## --------------------------------------------------                       ##
##  This class implements a generic OWL parsing interface.                  ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::Ontology::RDFParser
package require struct::tree

oo::class create ELEP::Ontology::OWLParser {
  variable rdf_parser_object   {}
  variable model_OWL_hierarchy {};# A tree structure that holds the class
                                  # hierarchy of the ontology.

  constructor {} {
    initParser
  };# constructor

  ##########################################################################
  ## Method: initParser
  ##########################################################################
  method initParser {} {
    set rdf_parser_object [ELEP::Ontology::RDFParser ELEP::Ontology::#auto]
  };# initParser

  ##########################################################################
  ## Method: parseLocalFile
  ##########################################################################
  method parseLocalFile {filename} {
    resetOWLModel
    ## Invoke the RDF parser to create a model out of the OWL file...
    $rdf_parser_object parseLocalFile $filename
    ## Parse the RDF statements to construct the OWL model...
    initOWLModel
  };# parseLocalFile

  ##########################################################################
  ## Method: modelTostring
  ##########################################################################
  method modelTostring {} {
    return [$rdf_parser_object modelTostring]
  };# modelTostring

  ##########################################################################
  ## Method: destroyOWLModel
  ##########################################################################
  method destroyOWLModel {} {
    if {[string length $model_OWL_hierarchy]} {
      catch {$model_OWL_hierarchy destroy}
    }
    set model_OWL_hierarchy {}
  };# destroyOWLModel

  ##########################################################################
  ## Method: resetOWLModel
  ##########################################################################
  method resetOWLModel {} {
    destroyOWLModel
    set model_OWL_hierarchy [::struct::tree]
    $rdf_parser_object resetModel
  };# resetOWLModel

  ##########################################################################
  ## Method: initOWLModel
  ##########################################################################
  method initOWLModel {} {
    #set iterator [$rdf_parser_object getStatementIterator]
    #set statement [$rdf_parser_object getStatementAndAdvanceIterator $iterator]
    #while {[string length $statement]} {
    #  set string [$rdf_parser_object statementToString $statement]
    #  regsub -all -- {http:[^#]+#} $string {} string
    #  puts $string
    #  set statement [$rdf_parser_object getStatementAndAdvanceIterator \
    #                                    $iterator]
    #}
    #$rdf_parser_object deleteStatementIterator $iterator
    

    ## Step 1: Identify the classes and properties that make up the ontology.
    set iterator [$rdf_parser_object getStatementIterator]
    set statement [$rdf_parser_object getStatementAndAdvanceIterator $iterator]
    while {[string length $statement]} {
      set string [$rdf_parser_object statementToString $statement]
      regsub -all -- {http:[^#]+#} $string {} string
      foreach {subject predicate object} \
         [$rdf_parser_object statementGetNodes $statement] {}
      switch -glob [$rdf_parser_object nodeGetURIString $predicate] {
        *#type {
          switch -glob [$rdf_parser_object nodeGetURIString $object] {
            *owl#Class {
              ## An OWL Class definition.
	      addClassInHierarchy $subject
            }
            default {
            }
          }
        }
        *#subClassOf {
          ## An OWL Sub-Class definition.
	  addClassInHierarchy $subject $object
        }
	*#equivalentClass {
	  ## Subject is equivalent to object. Add subject just below object in
	  ## the hierarchy.
	  puts ***$string
	}
        *#comment {
          set class [getClassNameFromNode $subject]
          set comment [$rdf_parser_object nodeGetLiteralValue $object]
          catch {$model_OWL_hierarchy append $class comment $comment}
        }
        default {
          # puts $string
        }
      }
      set statement [$rdf_parser_object getStatementAndAdvanceIterator \
                                        $iterator]
    }
    $rdf_parser_object deleteStatementIterator $iterator
  };# initOWLModel

  ##########################################################################
  ## Method: addClassInHierarchy
  ##########################################################################
  method addClassInHierarchy {node {parent root}} {
    set class [getClassNameFromNode $node]
    if {![$rdf_parser_object nodeIsBlank $node]} {
      if {[string equal $parent root]} {
        $model_OWL_hierarchy insert root end $class
        $model_OWL_hierarchy append $class label $class
      } elseif {![$rdf_parser_object nodeIsBlank $parent]} {
        $model_OWL_hierarchy insert [getClassNameFromNode $parent] end $class
        $model_OWL_hierarchy append $class label $class
      }
    }
    #lappend classes $class
  };# addClassInHierarchy

  ##########################################################################
  ## Method: examineNode
  ##########################################################################
  method examineNode {node} {
    set result {}
    foreach method {nodeGetType nodeGetURIString nodeGetBlankIdentifier
                    nodeGetLiteralValue
                    nodeGetLiteralValueDatatypeURIString
                    nodeGetLiteralValueLanguage} {
      lappend result "$method: \"[$rdf_parser_object $method $node]\""
    }
    return [join $result ", "]
  };# examineNode

  ##########################################################################
  ## Method: getClassNameFromNode
  ##########################################################################
  method getClassNameFromNode {node} {
    set uri [$rdf_parser_object nodeGetURIString $node]
    if {![string length $uri]} {
      return [$rdf_parser_object nodeGetBlankIdentifier $node]
    }
    set first_hash [string first # $uri]; incr first_hash
    return [string range $uri $first_hash end]
  };# getClassNameFromNode

  ##########################################################################
  ## Method: getHierarchyObject
  ##########################################################################
  method getHierarchyObject {} {
    return $model_OWL_hierarchy
  };# getHierarchyObject

  destructor {
    if {[string length $rdf_parser_object]} {
      itcl::delete object $rdf_parser_object
    }
    catch {$model_OWL_hierarchy destroy}
  };# destructor

};# class ELEP::Ontology::OWLParser

package provide ELEP::Ontology::OWLParser 1.0
