##############################################################################
##  RDFParser.tcl:                                                          ##
## --------------------------------------------------                       ##
##  This class implements a generic RDF parsing interface.                  ##
##                                                                          ##
##  Code in this file is based on an initial version written by             ##
##  Aris Theodorakos, artheo@iit.demokritos.gr                              ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

##
## Load some needed packages.
##
package require TclOO    ;# TclOO adds object-oriented facilities to Tcl.
package require redland

oo::class create ELEP::Ontology::RDFParser {
  variable world        NULL
  variable model        NULL
  variable storage      NULL
  variable rdf_filename {}

  constructor {} {
    initParser
  };# constructor

  ##########################################################################
  ## Method: initParser
  ##########################################################################
  method initParser {} {
    set world [librdf_new_world]
    librdf_world_open $world
    set storage [librdf_new_storage $world "memory" "test" NULL]
    if {$storage == "NULL"} {
      librdf_free_world $world
      error "Failed to create RDF storage"
    }
  };# initParser

  ##########################################################################
  ## Method: initModel
  ##########################################################################
  method initModel {} {
    set model [librdf_new_model $world $storage ""]
    if {$model == "NULL"} {
      error "Failed to create an RDF model"
    }
  };# initModel

  ##########################################################################
  ## Method: resetModel
  ##########################################################################
  method resetModel {} {
    if {![string equal $model NULL]} {librdf_free_model $model}
    initModel
  };# resetModel

  ##########################################################################
  ## Method: parseLocalFile
  ##########################################################################
  method parseLocalFile {filename} {
    if {[string equal $world NULL]} {error "uninitialised world!"}
    if {[string equal $model NULL]} {initModel}
    set uri [librdf_new_uri_from_filename $world $filename]
    if {$uri == "NULL"} {
      error "Cannot open URI: $filename"
    }
    set loadmodel [librdf_model_load $model $uri rdfxml \
                                     application/rdf+xml NULL]
    librdf_free_uri $uri
    if {$loadmodel != 0} {
      error "Cannot load model from URI: $filename"
    }
  };# parseLocalFile

  ##########################################################################
  ## Method: modelTostring
  ##########################################################################
  method modelTostring {} {
    if {[string equal $model NULL]} {error "uninitialised model!"}
    return [librdf_model_to_string $model NULL rdfxml application/rdf+xml NULL]
  };# modelTostring

  ##########################################################################
  ## Method: getStatementIterator
  ##########################################################################
  method getStatementIterator {} {
    if {[string equal $model NULL]} {error "uninitialised model!"}
    return [librdf_model_as_stream $model]
  };# getStatementIterator

  ##########################################################################
  ## Method: deleteStatementIterator
  ##########################################################################
  method deleteStatementIterator {iterator} {
    librdf_free_stream $iterator
  };# deleteStatementIterator

  ##########################################################################
  ## Method: getStatement
  ##########################################################################
  method getStatement {iterator} {
    if {[librdf_stream_end $iterator]} {return {}}
    return [librdf_stream_get_object $iterator]
  };# getStatement

  ##########################################################################
  ## Method: getStatementAndAdvanceIterator
  ##########################################################################
  method getStatementAndAdvanceIterator {iterator} {
    if {[librdf_stream_end $iterator]} {return {}}
    set statement [librdf_stream_get_object $iterator]
    librdf_stream_next $iterator
    return $statement
  };# getStatementAndAdvanceIterator

  ##########################################################################
  ## Method: statementIteratorNext
  ##########################################################################
  method statementIteratorNext {iterator} {
    if {[librdf_stream_end $iterator]} {return}
    return [librdf_stream_next $iterator]
  };# statementIteratorNext

  ##########################################################################
  ## Method: statementIteratorEnd
  ##########################################################################
  method statementIteratorEnd {iterator} {
    if {[librdf_stream_end $iterator]} {return 1}
    return 0
  };# statementIteratorEnd

  ##########################################################################
  ## Method: statementGetSubject
  ##########################################################################
  method statementGetSubject {statement} {
    return [librdf_statement_get_subject $statement]
  };# statementGetSubject

  ##########################################################################
  ## Method: statementGetPredicate
  ##########################################################################
  method statementGetPredicate {statement} {
    return [librdf_statement_get_predicate $statement]
  };# statementGetPredicate

  ##########################################################################
  ## Method: statementGetObject
  ##########################################################################
  method statementGetObject {statement} {
    return [librdf_statement_get_object $statement]
  };# statementGetObject

  ##########################################################################
  ## Method: statementGetNodes
  ##########################################################################
  method statementGetNodes {statement} {
    return [list [librdf_statement_get_subject $statement] \
                 [librdf_statement_get_predicate $statement] \
                 [librdf_statement_get_object $statement]]
  };# statementGetNodes

  ##########################################################################
  ## Method: statementToString
  ##########################################################################
  method statementToString {statement} {
    return [librdf_statement_to_string $statement]
  };# statementToString

  ##########################################################################
  ## Method: nodeGetType
  ##########################################################################
  # Valid types are:
  # LIBRDF_NODE_TYPE_UNKNOWN   = 0,  /* To catch uninitialised nodes */
  # LIBRDF_NODE_TYPE_RESOURCE  = 1,  /* rdf:Resource(&rdf:Property)-has a URI */
  # LIBRDF_NODE_TYPE_LITERAL   = 2,  /* rdf:Literal - has an XML string,
  #                                     language, XML space */
  # /* node type 3 is unused at present. Do not renumber this list. */
  # LIBRDF_NODE_TYPE_BLANK     = 4,  /* blank node has an identifier string */
  method nodeGetType {node} {
    switch -exact [librdf_node_get_type $node] {
      0 {return UNKNOWN}
      1 {return RESOURCE}
      2 {return LITERAL}
      4 {return BLANK}
    }
  };# nodeGetType

  ##########################################################################
  ## Method: nodeGetLiteralValue
  ##########################################################################
  method nodeGetLiteralValue {node} {
    return [librdf_node_get_literal_value $node]
  };# nodeGetLiteralValue

  ##########################################################################
  ## Method: nodeGetLiteralValueLanguage
  ##########################################################################
  method nodeGetLiteralValueLanguage {node} {
    return [librdf_node_get_literal_value_language $node]
  };# nodeGetLiteralValueLanguage
 
  ##########################################################################
  ## Method: nodeGetURI
  ##########################################################################
  method nodeGetURI {node} {
    return [librdf_node_get_uri $node]
  };# nodeGetURI
 
  ##########################################################################
  ## Method: nodeGetURIString
  ##########################################################################
  method nodeGetURIString {node} {
    return [librdf_uri_to_string [librdf_node_get_uri $node]]
  };# nodeGetURIString
 
  ##########################################################################
  ## Method: nodeGetLiteralValueDatatypeURI
  ##########################################################################
  method nodeGetLiteralValueDatatypeURI {node} {
    return [librdf_node_get_literal_value_datatype_uri $node]
  };# nodeGetLiteralValueDatatypeURI
 
  ##########################################################################
  ## Method: nodeGetLiteralValueDatatypeURIString
  ##########################################################################
  method nodeGetLiteralValueDatatypeURIString {node} {
    return [librdf_uri_to_string \
       [librdf_node_get_literal_value_datatype_uri $node]]
  };# nodeGetLiteralValueDatatypeURIString

  ##########################################################################
  ## Method: nodeGetBlankIdentifier
  ##########################################################################
  method nodeGetBlankIdentifier {node} {
    return [librdf_node_get_blank_identifier $node]
  };# nodeGetBlankIdentifier
 
  ##########################################################################
  ## Method: nodeToString
  ##########################################################################
  method nodeToString {node} {
    return [librdf_node_to_string $node]
  };# nodeToString

  ##########################################################################
  ## Method: nodeIsBlank
  ##########################################################################
  method nodeIsBlank {node} {
    return [librdf_node_is_blank $node]
  };# nodeIsBlank
 
#  ##########################################################################
#  ## Method: nodeTo
#  ##########################################################################
#  method nodeTo {node} {
#  };# nodeTo
# 
#  ##########################################################################
#  ## Method: nodeTo
#  ##########################################################################
#  method nodeTo {node} {
#  };# nodeTo

  destructor {
    if {![string equal NULL $model]}   {librdf_free_model   $model}
    if {![string equal NULL $storage]} {librdf_free_storage $storage}
    if {![string equal NULL $world]}   {librdf_free_world   $world}
  };# destructor

};# class ELEP::Ontology::RDFParser

package provide ELEP::Ontology::RDFParser 1.0
