##############################################################################
##  OWLParser.tcl:                                                          ##
## --------------------------------------------------                       ##
##  This class implements an interface around the Racer Pro reasoner.       ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

##
## Parts of the code in this file have been contributed by Aris Theodorakos,
## with funding from the BOEMIE EU-funded research project.
## (http://www.boemie.org)
##

##
## Load some needed packages.
##
package require TclOO
# package require struct::tree

oo::class create ELEP::Ontology::RacerPro {
  variable sockchan resultsParser rp_host rp_port debug

  classmethod log {msg} {puts stdout $msg; update}

  constructor { {host localhost} {port 8088} } {
    set resultsParser {}
    set rp_host   {}
    set rp_port   {}
    set debug     0
    # Creates a TCP connection with Racer Pro server.
    my connect $host $port
    set resultsParser \
      [[ELEP::Ontology::RacerProResultsParser new] name]
  };# constructor

  ###########################################################################
  ## Method: connect
  ###########################################################################
  # This method closes any opened connections and connects to a running Racer 
  # Pro instance.
  method connect { {host localhost} {port 8088} } {
    my disconnect
    set rp_host $host
    set rp_port $port
    # Creates a TCP connection with Racer Pro server.
    if {[catch {socket $rp_host $rp_port} sockchan]} {
      error "Racer Pro server is not found at $rp_host:$rp_port"
    }
    fconfigure $sockchan -buffering line -blocking 1 -encoding iso8859-1 \
               -buffersize 8192 -translation {auto crlf}
    # fileevent  $sockchan readable [list [self] onDataAvailable]
    if {[catch {puts $sockchan "(get-racer-version)"; flush $sockchan}]} {
      error "Cannot invoke Racer Pro commands: (get-racer-version) failed!\
             ($host:$port) ($host:$port)"
    }
    if {[catch {gets $sockchan response}]} {
      error "Cannot retrieve results from Racer Pro server ($host:$port)"
    }
    if { ![string match :answer* $response] } {
      error "Cannot parse results retrieved from Racer Pro server ($host:$port)"
    }
  };# connect

  ###########################################################################
  ## Method: disconnect
  ###########################################################################
  # Closes the connection with Racer Pro server.
  method disconnect {} {
    catch {close $sockchan}
  };# disconnect

  destructor {
    my disconnect
    # Deletes the allocated parser object.
    catch {$resultsParser destroy}
    catch {unset -nocomplain ::ELEP::Ontology::RacerProAnswer([self])}
  };# destructor

  ###########################################################################
  ## Method: onDataAvailable
  ###########################################################################
  # This method is called when data is available on the socket to read it...
  method onDataAvailable {} {
   # The channel is readable; try to read it.
    set status [catch { gets $sockchan line } result]
    my log "Data: \"$line\" $result $status"
    if {$status != 0} {
      error $result
    } elseif {$result >= 0} {
      # Successfully read the channel
      set ::ELEP::Ontology::RacerProAnswer($this) $line
      my log "onDataAvailable='$::ELEP::Ontology::RacerProAnswer($this)'"
    } elseif {[eof $sockchan] } {
      # End of file on the channel
      set ::ELEP::Ontology::RacerProAnswer($this) {}
      my log "onDataAvailable='$::ELEP::Ontology::RacerProAnswer($this)'"
    } elseif {[fblocked $f]} {
      # Read blocked.  Just return!
      return
    } else {
      # Something else
      set ::ELEP::Ontology::RacerProAnswer($this) {}
      my log "onDataAvailable='$::ELEP::Ontology::RacerProAnswer($this)'"
    }
  };# onDataAvailable

  ###########################################################################
  ## Proc: puts_error_check
  ###########################################################################
  # This procedure writes data to the provided file/socket while catches
  # errors. In case of an error, the socket is re-opened...
  method puts_error_check {sockchan data} {
    if {$debug} {my log "Q: $data"}
    # set ::ELEP::Ontology::RacerProAnswer($this) {}
    if {[catch {puts $sockchan $data} error]} {
      my connect $rp_host $rp_port
      error "Error while writting \"$data\": $error"
    }
  };# puts_error_check
  
  ###########################################################################
  ## Proc: getanswer
  ###########################################################################
  # Gets the response after a command is submitted, extracts the real
  # answer and returns it.
  # If the response is an error the return string has the format:
  # "Error: <error description>"
  method getanswer { sockchan } {
    gets $sockchan preanswer
    if {$debug} {my log $preanswer}
    # vwait ::ELEP::Ontology::RacerProAnswer($this)
    # set preanswer $::ELEP::Ontology::RacerProAnswer($this)
    # log "->$preanswer"
    # log "A: $preanswer"
    if { [regexp {:error\s*\d+\s*([^"]*)} $preanswer crap answer] } {
      return "Error: $answer\nReasoner response:\n$preanswer"
    }
    set pos [string first \" $preanswer]
    incr pos
    set stop 0
    set answer ""
    while { $stop == 0 } {
      set c [string index $preanswer $pos]
      if { $c eq "\\" } {
        incr pos
        set cc [string index $preanswer $pos]
        if { $cc eq "\"" } {
          append answer $cc
        } else {
          append answer $c $cc
        }
      } elseif { $c eq "\"" } {
        set stop 1
      } else {
        append answer $c
      }
      incr pos
    }
    return $answer
  };# getanswer

  ###########################################################################
  ## Proc: getanswer_error_check
  ###########################################################################
  # Gets the response after a command is submitted (as getanswer does), but
  # throws an error if an erroneous result is received.
  method getanswer_error_check {} {
    set answer [my getanswer $sockchan]
    if {[string first error [string tolower $answer]] != -1} {error $answer}
    return $answer
  };# getanswer_error_check

  ###########################################################################
  ## Proc: fixname
  ###########################################################################
  # Takes a concept name and returns:
  # a) its full name without the pipe characters ('|'), if mode="full"
  # b) its abbreviated name, if mode="abbrev"
  classmethod fixname { concept mode } {
    if { $concept eq "top" || $concept eq "TOP" } { return "root" }
    if { $mode ne "abbrev" } {
      if { [regexp {\|([^\|]+)\|} $concept crap fixed_concept] } {
        return $fixed_concept
      }
      else { return $concept }
    } elseif { [regexp {\|.*#(.*?)\|} $concept crap fixed_concept] } {
      return $fixed_concept
    } else {
      return $concept
    }
  };# fixname

  classmethod simplify_name {name} {
    set name  [string trim $name |]
    set index [string first # $name]
    if {$index != -1} {
      incr index
      set name [string range $name $index end]
    }
    return $name
  };# simplify_name
  
  ###########################################################################
  ## Method: version_racerpro
  ###########################################################################
  # Returns the Racer pro version
  method version_racerpro { } {
    my puts_error_check $sockchan "(get-racer-version)"
    return [my getanswer $sockchan]  
  };# version_racerpro
  
  ###########################################################################
  ## Method: fullreset
  ###########################################################################
  # Performs a full reset in the server. All data in the Tbox and Abox
  # are lost.
  method fullreset { } {
    my puts_error_check $sockchan "(full-reset)"
    my getanswer_error_check
  };# fullreset

  ###########################################################################
  ## Method: set_current_tbox
  ###########################################################################
  method set_current_tbox { filename } {
    my puts_error_check $sockchan "(set-current-tbox |$filename|)"
    my getanswer_error_check
  };# set_current_tbox

  ###########################################################################
  ## Method: current_tbox
  ###########################################################################
  method current_tbox { } {
    my puts_error_check $sockchan "(current-tbox)"
    my getanswer_error_check
  };# current_tbox

  ###########################################################################
  ## Method: set_current_abox
  ###########################################################################
  method set_current_abox { filename } {
    my puts_error_check $sockchan "(set-current-abox |$filename|)"
    my getanswer_error_check
  };# set_current_abox

  ###########################################################################
  ## Method: current_abox
  ###########################################################################
  method current_abox { } {
    my puts_error_check $sockchan "(current-abox)"
    my getanswer_error_check
  };# current_abox
  
  ###########################################################################
  ## Method: abox_valid
  ###########################################################################
  # Checks the consistency of the current A-box. If the A-box is consistent
  # it returns 1 otherwise it returns 0.
  method abox_valid {} {
    my puts_error_check $sockchan "(abox-consistent?)"
    set result [my getanswer $sockchan]
    if { $result eq "T" } { return 1 } \
    else { return 0 }
  };# abox_valid
  
  ###########################################################################
  ## Method: readowl
  ###########################################################################
  # Reads an OWL file given its full pathname and imports its data in the
  # knowledgeBase.
  method readowl { filename {set_as_current 1} } {
    set filename [string map {\\ /} $filename]
    my puts_error_check $sockchan "(owl-read-file \"$filename\")"
    set answer [my getanswer_error_check]
    if {$set_as_current} {
      my set_current_tbox $filename
      my set_current_abox $filename
    }
    return $answer
  };# readowl

  ###########################################################################
  ## Method: readowl_in_kb
  ###########################################################################
  # Reads an OWL file given its full pathname and imports its data in the
  # knowledgeBase.
  method readowl_in_kb { filename init kb {set_as_current 1} } {
    set filename [string map {\\ /} $filename]
    my puts_error_check $sockchan "(owl-read-file \"$filename\" :init $init\
                                   :kb-name $kb)"
    set answer [my getanswer_error_check]
    if {$set_as_current} {
      my set_current_tbox $kb
      my set_current_abox $kb
    }
    return $answer
  };# readowl_in_kb

  ###########################################################################
  ## Method: add_role_assertions_for_datatype_properties
  ###########################################################################
  method add_role_assertions_for_datatype_properties {} {
    my puts_error_check $sockchan "(add-role-assertions-for-datatype-properties)"
    set answer [my getanswer_error_check]
    return $answer
  };# add_role_assertions_for_datatype_properties

  ###########################################################################
  ## Method: save_kb
  ###########################################################################
  # Saves the current KnowledgeBase in OWL format in the given filename.
  method save_kb { filename } {
    set filename [string map {\\ /} $filename]
    my puts_error_check $sockchan "(save-kb \"$filename\" :syntax :owl)"
    return [my getanswer $sockchan]
  };# save_kb
  
  ###########################################################################
  ## Method: save_abox
  ###########################################################################
  # Saves the current A-box in Racer format in the given filename.
  method save_abox { filename } {
    set filename [string map {\\ /} $filename]
    my puts_error_check $sockchan "(save-abox \"$filename\" :syntax :race)"
    set answer [my getanswer $sockchan]
    if {[string first error [string tolower $answer]] != -1} {error $answer}
    return $answer
  };# save_abox

  ###########################################################################
  ## Method: save_abox_owl
  ###########################################################################
  # Saves the current A-box in Racer format in the given filename.
  method save_abox_owl { filename } {
    set filename [string map {\\ /} $filename]
    my puts_error_check $sockchan "(save-abox \"$filename\" :syntax :owl)"
    set answer [my getanswer $sockchan]
    if {[string first error [string tolower $answer]] != -1} {error $answer}
    return $answer
  };# save_abox_owl

  ###########################################################################
  ## Method: load_abox
  ###########################################################################
  # Loads an ABox in the Racer format and makes it the current ABox.
  method load_abox { filename } {
    set filename [string map {\\ /} $filename]
    my puts_error_check $sockchan "(racer-read-file \"$filename\")"
    my getanswer_error_check
    #set_current_abox DEFAULT
  };# load_abox

  ###########################################################################
  ## Method: import_abox
  ###########################################################################
  # Loads an ABox in the Racer format, appending it to the current ABox.
  method import_abox { filename } {
    set filename [string map {\\ /} $filename]
    my puts_error_check $sockchan "(include-kb \"$filename\")"
    return [my getanswer $sockchan]
  };# import_abox

  ###########################################################################
  ## Method: set_associated_tbox
  ###########################################################################
  # Associates the provided TBox to the current ABox.
  method set_associated_tbox { tbox } {
    set tbox [string map {\\ /} $tbox]
    my puts_error_check $sockchan "(set-associated-tbox |$tbox|)"
    return [my getanswer $sockchan]
  };# set_associated_tbox

  ###########################################################################
  ## Method: clone_abox
  ###########################################################################
  # Associates the provided TBox to the current ABox.
  method clone_abox { name1 name2 } {
    set name1 [string map {\\ /} $name1]
    set name2 [string map {\\ /} $name2]
    my puts_error_check $sockchan "(clone-abox |$name1| \"$name2\" t)"
    puts [my getanswer_error_check]
  };# clone_abox

  ###########################################################################
  ## Method: runcommand
  ###########################################################################
  # Runs any Racer Pro command and returns its result.
  method runcommand { command } {
    my puts_error_check $sockchan $command
    return [my getanswer $sockchan]
  };# runcommand

  ###########################################################################
  ## Method: children_concepts
  ###########################################################################
  # Returns a Tcl list of all the immediate children of the given concept.
  method children_concepts { concept } {
    my puts_error_check $sockchan "(concept-children $concept)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# children_concepts

  ###########################################################################
  ## Method: all_atomic_concepts
  ###########################################################################
  # Returns a Tcl list of all the instances in the current ABox.
  method all_atomic_concepts {} {
    my puts_error_check $sockchan "(all-atomic-concepts)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# all_atomic_concepts

  ###########################################################################
  ## Method: all_concept_properties
  ###########################################################################
  # Returns all the properties of the given concept along with
  # their filled values as a TCL list.
  method all_concept_properties { concept } {
    my puts_error_check $sockchan "(describe-concept $concept)"
    gets $sockchan answer
    ## Get the useful part out of answer...
    set answer [string trim $answer]
    set fp [string first ( $answer]
    set lp [string last  ) $answer]
    if {$fp == -1 || $lp == -1} {return {}}
    incr fp; incr lp -1
    set answer [string range $answer $fp $lp]
    set answer [string map {( \{ ) \} \\| | \\S \"} [string trim $answer]]
    # puts $answer
    set result {}
    foreach {tag value} [lrange $answer 1 end] {
      if {$value eq "NIL" || $value eq ":TO-BE-COMPUTED"} {continue}
      # puts "$tag -> $value"
      switch -glob -- [string tolower $tag] {
        :told-primitive-definition {
          switch -glob $value {
            *{*}* {
              return [my operator_parse_direct [string map {\{ [ \} ]} $value]]
            }
            default {return $value}
          }
        }
      }
    }
    return $result
  };# all_concept_properties

  ###########################################################################
  ## Method: all_individuals
  ###########################################################################
  # Returns a Tcl list of all the instances in the current ABox.
  method all_individuals {} {
    my puts_error_check $sockchan "(all-individuals)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# all_individuals

  ###########################################################################
  ## Method: related_individuals
  ###########################################################################
  # Returns a Tcl list of all instance pairs, that are related with the
  # specified role in the current ABox.
  method related_individuals {role} {
    my puts_error_check $sockchan "(related-individuals $role)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# related_individuals

  ###########################################################################
  ## Method: individuals_related?
  ###########################################################################
  # Returns true if the two individuals are directly related through the
  # specified role.
  method individuals_related? {i1 i2 role} {
    my puts_error_check $sockchan "(individuals-related? $i1 $i2 $role)"
    if {[my getanswer $sockchan] eq "T"} {return 1}
    return 0
  };# all_individuals
  
  ###########################################################################
  ## Method: concept_instances
  ###########################################################################
  # Returns a Tcl list of all the instances of the given concept.
  method concept_instances { concept } {
    my puts_error_check $sockchan "(concept-instances $concept)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# concept_instances

  ###########################################################################
  ## Method: most_specific_concept
  ###########################################################################
  # Returns a Tcl list of all the instances of the given concept.
  method most_specific_concept { instance } {
    my puts_error_check $sockchan "(most-specific-instantiators $instance)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# most_specific_concept

  ###########################################################################
  ## Method: instance_fillers
  ###########################################################################
  # Returns a Tcl list of all the fillers for the given role of the given
  # instance.
  method instance_fillers { instance role } {
    my puts_error_check $sockchan "(individual-fillers $instance $role)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# instance_fillers

  ###########################################################################
  ## Method: instance_datatype_filler
  ###########################################################################
  # Returns the string which is the filler for the given role of the given
  # instance.
  method instance_datatype_filler { instance role } {
    my puts_error_check $sockchan \
                    "(individual-told-datatype-fillers $instance $role)"
    set answer [my getanswer $sockchan]
    if { ![string match \(* $answer] } {
      ## Petasis: 20 Feb 2008: What if there are warnings (like cycles)?
      if {[regexp {\(\"[^\"]*\"\)} $answer match]} {
        return [string range $match 2 end-2]
      }
      return {}
    } elseif { [string first \" $answer] == 1 } {
      return [string range $answer 2 end-2]
    } else {
      return [string range $answer 1 end-1]
    }
  };# instance_datatype_filler

  ###########################################################################
  ## Method: instance_as_table
  ###########################################################################
  # This method returns the description of an individual as a table.
  # The returned table is a list of three items:
  #   a) The most specific concept for the provided individual
  #   b) A list of {role value} for data-type properties
  #   c) A list of {role filler-instances} for roles whose individual is
  #      a subject
  #   d) A list of {role filler-instances} for roles whose individual is
  #      an object
  method instance_as_table { instance } {
    set table [most_specific_concept $instance]
    ##
    ## Calculate the datatype properties...
    ##
    set properties_datatype {}
    foreach one [my all_concept_assertions_for_individual $instance] {
      foreach {i v} $one {break}
      switch [string tolower [lindex $v 0]] {
        some - d-filler {
          set v [lrange $v 1 end]
          foreach {role filler} $v {
            switch -glob -- $filler {
              *%* {
                lappend properties_datatype $role [lindex $filler end]
              }
            }
          }
        }
      }
    }
    lappend table $properties_datatype
    ##
    ## Calculate role fillers whose individual is a subject
    ##
    set properties_instance_is_source {}
    foreach one [my all_role_assertions_for_individual_in_domain $instance] {
      foreach {individuals role} $one {break}
      foreach {subject objects} $individuals {break}
      # puts "role_domain: $subject -> $role -> $objects"
      lappend properties_instance_is_source $role $objects
    }
    lappend table $properties_instance_is_source
    ##
    ## Calculate role fillers whose individual is an object
    ##
    set properties_instance_is_target {}
    foreach one [my all_role_assertions_for_individual_in_range $instance] {
      foreach {individuals role} $one {break}
      foreach {subject objects} $individuals {break}
      # puts "role_range: $subject -> $role -> $objects"
      lappend properties_instance_is_target $role $subject
    }
    lappend table $properties_instance_is_target
    return $table
  };# instance_as_table

  ###########################################################################
  ## Method: all_instance_properties
  ###########################################################################
  # Returns all the properties of the given instance along with
  # their filled values as a TCL list.
  method all_instance_properties { instance } {
    my puts_error_check $sockchan "(describe-individual $instance)"
    gets $sockchan answer
    ## Get the useful part out of answer...
    set answer [string trim $answer]
    set fp [string first ( $answer]
    set lp [string last  ) $answer]
    if {$fp == -1 || $lp == -1} {return {}}
    incr fp; incr lp -1
    set answer [string range $answer $fp $lp]
    set answer [string map {( \{ ) \} \\| | \\S \"} [string trim $answer]]
    # puts $answer
    set result {}
    foreach {tag value} [lrange $answer 1 end] {
      if {$value eq "NIL" || $value eq ":TO-BE-COMPUTED"} {continue}
      # puts "$tag -> $value"
      switch -glob -- [string tolower $tag] {
        *-fillers -
        :assertions {
          foreach {assertion} $value {
            foreach {key val} $assertion {
              # puts "key: $key, val: $val"
              if {[llength $val] == 1} {
                lappend result $key=$val
              } elseif {[llength $val] == 3 && [lindex $val 0] eq "SOME"} {
                lappend result [lindex $val 1]=\"[lindex $val end end]\"
              } else {
                ## Iterate over all elements and add everything that is an
                ## individual, or the last element if no individuals have been
                ## found...
                set individuals {}
                set role $key
                foreach {element} $val {
                  switch -glob $element {
                    |*#*| {lappend individuals $element}
                    default {}
                  }
                }
                if {![llength $individuals]} {
                  set role [lindex $val 1]
                  lappend individuals \"[lindex $val end end]\"
                }
                lappend result $role=[join $individuals]
              }
            }
          }
        }
        default {}
      }
    }
    return $result
  };# all_instance_properties

  ###########################################################################
  ## Method: all_instance_properties_simple_names
  ###########################################################################
  # Returns all the properties of the given instance along with
  # their filled values as a TCL list.
  method all_instance_properties_simple_names { instance } {
    my puts_error_check $sockchan "(describe-individual $instance)"
    gets $sockchan answer
    ## Get the useful part out of answer...
    set answer [string trim $answer]
    set fp [string first ( $answer]
    set lp [string last  ) $answer]
    if {$fp == -1 || $lp == -1} {return {}}
    incr fp; incr lp -1
    set answer [string range $answer $fp $lp]
    set answer [string map {( \{ ) \} \\| | \\S \"} [string trim $answer]]
    # puts $answer
    set result {}
    foreach {tag value} [lrange $answer 1 end] {
      if {$value eq "NIL" || $value eq ":TO-BE-COMPUTED"} {continue}
      # puts "$tag -> $value"
      switch -glob -- [string tolower $tag] {
        *-fillers -
        :assertions {
          foreach {assertion} $value {
            foreach {key val} $assertion {
              # puts "key: $key, val: $val"
              if {[llength $val] == 1} {
                lappend result [my simplify_name $key]=[my simplify_name $val]
              } elseif {[llength $val] == 3 && [lindex $val 0] eq "SOME"} {
                lappend result [my simplify_name [lindex $val 1]]=\"[my simplify_name\
                                [lindex $val end end]]\"
              } else {
                ## Iterate over all elements and add everything that is an
                ## individual, or the last element if no individuals have been
                ## found...
                set individuals {}
                set role [my simplify_name $key]
                foreach {element} $val {
                  switch -glob $element {
                    |*#*| {lappend individuals [my simplify_name $element]}
                    default {}
                  }
                }
                if {![llength $individuals]} {
                  set role [my simplify_name [lindex $val 1]]
                  lappend individuals \"[my simplify_name [lindex $val end end]]\"
                }
                lappend result $role=[join $individuals]
              }
            }
          }
        }
        default {}
      }
    }
    return $result
  };# all_instance_properties_simple_names

  ###########################################################################
  ## Method: instance_datatype_properties
  ###########################################################################
  # Returns the datatype properties of the given instance along with
  # their filled values as a TCL list.
  method instance_datatype_properties { instance } {
    my puts_error_check $sockchan "(describe-individual $instance)"
    gets $sockchan answer
    if {[regexp -nocase --\
            {:TOLD-DATATYPE-FILLERS(.+):ANNOTATION-DATATYPE-PROPERTY-FILLERS} \
            $answer crap dataprop] } {
      set dataprop [string trim $dataprop]
      set dataprop [string map {\\| | \\S \"} $dataprop]
      return [$resultsParser properties_parse $dataprop]
    } else {
      return "Error: Individual $instance does not exist!"
    }
  };# instance_datatype_properties

  ###########################################################################
  ## Method: all_concept_assertions_for_individual
  ###########################################################################
  method all_concept_assertions_for_individual { instance } {
    my puts_error_check $sockchan \
         "(all-concept-assertions-for-individual $instance)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# all_concept_assertions_for_individual

  ###########################################################################
  ## Method: retrieve_instances
  ###########################################################################
  # Retrieves instances according to a provided query.
  method retrieve_instances {query} {
    my puts_error_check $sockchan "(retrieve (?x) ($query))"
    set answer [my getanswer $sockchan]
    puts $answer ; update
    return [$resultsParser simple_parse $answer]
  };# retrieve_instances

  ###########################################################################
  ## Method: delete_instance
  ###########################################################################
  # Deletes the given instance of the given concept from the current ABox.
  method delete_instance { abox instance concept } {
    my puts_error_check $sockchan \
         "(forget (:abox |$abox|) (instance $instance $concept))"
    gets $sockchan answer
  }
  
  ###########################################################################
  ## Method: taxonomy
  ###########################################################################
  # Creates the Tcl tree of all concepts under the "parent" concept.
  # Nodes are named according to "mode". See above proc: fixname.
  method taxonomy { taxtree parent {mode abbrev}} {
    set top_concepts [my children_concepts $parent]
    foreach concept $top_concepts {
      if { $concept ne "*BOTTOM* BOTTOM" } {
        $taxtree insert [my fixname $parent $mode] \
            end [my fixname $concept $mode]
        taxonomy $taxtree $concept $mode
      }
    }
  };# taxonomy
  
  ###########################################################################
  ## Method: all_roles
  ###########################################################################
  # Returns a Tcl list of all role names of the Tbox.
  method all_roles {} {
    my puts_error_check $sockchan "(all-roles)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# all_roles
  
  ###########################################################################
  ## Method: getroledomain
  ###########################################################################
  # Returns a Tcl list of the domain concepts of the specified role.
  method getroledomain { role } {
    my puts_error_check $sockchan "(role-domain $role)"
    set answer [my getanswer $sockchan]
    return [$resultsParser operator_parse $answer]
  };# getroledomain
  
  ###########################################################################
  ## Method: getrolerange
  ###########################################################################
  # Returns a Tcl list of the concepts which are fillers for the "role", if
  # "role" is an objectproperty, or the datatype filler (integer, string,
  # etc.) if "role" is a datatype property.
  method getrolerange { role } {
    my puts_error_check $sockchan "(role-range $role)"
    set answer [my getanswer $sockchan]
    return [$resultsParser operator_parse $answer]
  };# getrolerange
  
  ###########################################################################
  ## Method: assignroles
  ###########################################################################
  # For the concepts that have at least one role (property), it creates a
  # key in the corresponding node of the concept tree named "properties"
  # and saves inside their roles.
  # Additionally, it creates a key in the root node for every role and
  # saves inside the role's range (see method: getrolerange).
  method assignroles { taxtree {mode abbrev} } {
    foreach role [my allroles] {
      set concepts [my getroledomain $role]
      set role_range [my getrolerange $role]
      foreach concept_inrange $role_range {
        $taxtree lappend root [my fixname $role $mode] \
              [my fixname $concept_inrange $mode]
      }
      foreach concept $concepts {
        if { $concept ne "TOP" } {
          $taxtree lappend [my fixname $concept $mode] properties \
              [my fixname $role $mode]
          foreach descendant [$taxtree descendants [my fixname $concept $mode]] {
            $taxtree lappend [my fixname $descendant $mode] properties \
                    [my fixname $role $mode]
          }
        }
      }
    }
  };# assignroles

  ###########################################################################
  ## Method: all_role_assertions_for_individual_in_domain
  ###########################################################################
  # Returns all role assertions for an individual from the current ABox
  # in which the individual is the role predecessor.
  method all_role_assertions_for_individual_in_domain {individual} {
    my puts_error_check $sockchan \
      "(all-role-assertions-for-individual-in-domain $individual)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# all_role_assertions_for_individual_in_domain

  ###########################################################################
  ## Method: all_role_assertions_for_individual_in_range
  ###########################################################################
  # Returns all role assertions for an individual from the current ABox
  # in which the individual is a role successor.
  method all_role_assertions_for_individual_in_range {individual} {
    my puts_error_check $sockchan \
      "(all-role-assertions-for-individual-in-range $individual)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# all_role_assertions_for_individual_in_range

  ##
  ## Triple-store management
  ##

  ###########################################################################
  ## Method: create_triple_store
  ###########################################################################
  # Creates a new tripple store inside a directory, overwritting any existing
  # triple-store found in the specified directory...
  method check_dir {dir} {
    set dir [string map {\\ /} $dir]
    # Ensure that the last character of the dir is a / or \! Racer has a bug
    # over this, and will complain if this character is missing!
    switch -glob $dir {
      */ - *\\ {}
      */*  {set dir $dir/}
      *\\* {set dir $dir\\}
      default {}
    }
    return $dir
  };# check_dir

  method create_triple_store {name dir {if_exists :supersede}} {
    my puts_error_check $sockchan \
      "(create-triple-store \"$name\" :directory \"[check_dir $dir]\"\
          :if-exists $if_exists)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# create_triple_store

  ###########################################################################
  ## Method: open_triple_store
  ###########################################################################
  # Opens a triple store from the specified directory...
  method open_triple_store {name dir} {
    my puts_error_check $sockchan \
      "(open-triple-store \"$name\" :directory \"[check_dir $dir]\")"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# open_triple_store

  ###########################################################################
  ## Method: triple_store_read_file
  ###########################################################################
  # Reads an Abox inside a triple store...
  method triple_store_read_file {file {init :nil} {ignore_import t}} {
    my puts_error_check $sockchan \
      "(triple-store-read-file \"$file\"\
           :init $init :ignore-import $ignore_import)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# triple_store_read_file

  ###########################################################################
  ## Method: materialize_inferences
  ###########################################################################
  # Materialises inferred knowledge...
  method materialize_inferences {kb_name db dir} {
    my puts_error_check $sockchan \
      "(materialize-inferences $kb_name :db $db \
          :directory \"[check_dir $dir]\"\
          :index-p t :abox t)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# materialize_inferences

  ###########################################################################
  ## Method: owl_read_document
  ###########################################################################
  # Loads an owl document...
  method owl_read_document {file kb_name {init t}} {
    my puts_error_check $sockchan \
      "(owl-read-document \"$file\" :kb-name $kb_name :init $init)"
    set answer [my getanswer $sockchan]
    return [$resultsParser simple_parse $answer]
  };# owl_read_document

  method send {cmd} {
    regsub -all {[\n\r]} $cmd {} cmd
    my puts_error_check $sockchan $cmd
    return [my getanswer $sockchan]
  };# send

  method simple_parse {answer} {
    return [$resultsParser simple_parse $answer]
  };# simple_parse

  method operator_parse {answer} {
    return [$resultsParser operator_parse $answer]
  };# operator_parse

  method operator_parse_direct {answer} {
    return [$resultsParser operator_parse_direct $answer]
  };# operator_parse_direct

};# class ELEP::Ontology::RacerPro

oo::class create ELEP::Ontology::RacerProResultsParser {
  
  constructor {} {}
  # constructor

  ###########################################################################
  ## Method: operator_parse
  ###########################################################################
  # Returns a Tcl list which contains the returned items (concepts, roles,
  # features...) only when the answer consists of logical operators on items.
  method operator_parse {str} {
    if { ![string match \(* $str] } { return [string trim $str] }
    set str [string range [string trim $str] 1 end-1]
    set cmd [string map {\n " " \( \[ \) \]} $str]
    return [eval $cmd]
  }; # operator_parse

  ###########################################################################
  ## Method: operator_parse_direct
  ###########################################################################
  # Returns a Tcl list which contains the returned items (concepts, roles,
  # features...) only when the answer consists of logical operators on items.
  method operator_parse_direct {str} {
    return [eval $str]
  }; # operator_parse_direct

  ###########################################################################
  ## Method: simple_parse
  ###########################################################################
  # Returns a Tcl list which contains the returned items (concepts, roles,
  # features...) only when the answer does NOT contain logical operators.
  # The first command drops the inverted roles.
  method simple_parse {str} {
    if { ![string match \(* $str] } { return {} }
    regsub -all {\(INV.+?\)} $str "" str
    set begin_pos [string first \( $str]
    set end_pos [string last \) $str]
    incr begin_pos; incr end_pos -1
    return [lindex [list [string map {\( \{ \) \}}\
        [string range $str $begin_pos $end_pos]]] 0]
  }; # simple_parse

  ###########################################################################
  ## Method: properties_parse
  ###########################################################################
  # Returns a TCL list which contains pairs of type "data property name" -
  # "property's value".
  method properties_parse {str} {
    if { ![string match \(* $str] } { return {} }
    set pos 1
    while {[regexp -start $pos {\((.+?)\s\((.+?)\)\)} $str match prop value]} {
      if { ![string match \"* $value] } {
        set value [split $value " "]
      }
      lappend result $prop $value
      incr pos [string length $match]
    }
    return $result
  }; # properties_parse
  
  ###########################################################################
  ## Method: AND
  ###########################################################################
  method AND {args} {
    return [eval concat $args]
  };# AND
  method and {args} {
    return [eval concat $args]
  };# and

  ###########################################################################
  ## Method: A
  ###########################################################################
  method A {args} {
    return [eval concat $args]
  };# A
  method a {args} {
    return [eval concat $args]
  };# a
  
  ###########################################################################
  ## Method: NOT
  ###########################################################################
  method NOT {args} {
    return {}
  };# NOT
  method not {args} {
    return {}
  };# not
  
  ###########################################################################
  ## Method: NO
  ###########################################################################
  method NO {args} {
    return {}
  };# NO
  method no {args} {
    return {}
  };# no
  
  ###########################################################################
  ## Method: INV
  ###########################################################################
  method INV {args} {
    return {}
  };# INV
  method inv {args} {
    return {}
  };# inv
  
  ###########################################################################
  ## Method: ALL
  ###########################################################################
  method ALL {args} {return ALL}
  method all {args} {return ALL}

  ###########################################################################
  ## Method: SOME
  ###########################################################################
  method SOME {args} {return [concat some $args]}
  method some {args} {return [concat some $args]}

  ###########################################################################
  ## Method: OR
  ###########################################################################
  method OR {args} {
    switch -exact [lindex $args 0] {
      ALL - SOME - all - some {return {}}
    }
    return [eval concat $args]
  };# OR
  method or {args} {
    switch -exact [lindex $args 0] {
      ALL - SOME - all - some {return {}}
    }
    return [eval concat $args]
  };# or

  method name {} {return [self]}
};# class ELEP::Ontology::RacerProResultsParser

package provide ELEP::Ontology::RacerPro 1.0
