##############################################################################
##  GrammarCompiler.tcl:                                                    ##
## --------------------------------------------------                       ##
##  This class uses YETI to parse and compile a grammar file.               ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

##
## Load some needed packages.
##
package require TclOO    ;# TclOO adds object-oriented facilities to Tcl.
package require ylex
package require yeti
package require ELEP::PatternEngine::EBNFScanner
package require ELEP::PatternEngine::EBNFParser

oo::class create ELEP::PatternEngine::GrammarCompiler {
  variable sg {};# The lexical scanner generator object.
  variable pg {};# The parser generator object.

  variable verbose 0
  variable verbout stderr

  constructor {} {
    if {$verbose} {
      set verbout [open log.txt w]
    }
    constractScannerGenerator
    constractParserGenerator
  };# constructor

  destructor {
    if {$verbout != "stderr"} {close $verbout}
    catch {itcl::delete object $sg}
    catch {itcl::delete object $pg}
  };# destructor

  ##########################################################################
  ## Method: constractScannerGenerator
  ##########################################################################
  method constractScannerGenerator {} {
    set sg [ELEP::PatternEngine::EBNFScanner ::ELEP::PatternEngine::temp::#auto]
    $sg configure -verbose $verbose
    $sg configure -verbout $verbout
  };# constractScannerGenerator

  ##########################################################################
  ## Method: constractParserGenerator
  ##########################################################################
  method constractParserGenerator {} {
    set pg [ELEP::PatternEngine::EBNFParser ::ELEP::PatternEngine::temp::#auto \
              -scanner $sg]
    $pg configure -verbose $verbose
    $pg configure -verbout $verbout
  };# constractParserGenerator

  ##########################################################################
  ## Method: compileGrammarFile
  ##########################################################################
  method compileGrammarFile {filename {grammar_name {}}} {
    if {![string length $grammar_name]} {
      set grammar_name ELEP::PatternEngine::grammars::[string map {{ } _} \
                            [file rootname [file tail $filename]]]
    }
    # Read the grammar from the file...
    set file [open $filename]
    fconfigure $file -encoding utf-8
    set data "[string trim [read $file]]\n "
    close $file
    # Process the grammar to add the match definition...
    #foreach rule [split $data -->
    #regsub -all -- {(Rule:\s+\w+\s+)(.+?)(?=-->)} $data {££\1 (\2)::match££} n
    #puts $n
    # Prepare the lexical scanner:
    $sg start $data
    # Prepare the parser:
    $pg reset
    $pg configure -name $grammar_name
    # Parse the grammar:
    $pg parse
    # Return the produced code:
    return [$pg dump]
  };# compileGrammarFile

  ##########################################################################
  ## Method: compileAndSaveGrammarFile
  ##########################################################################
  method compileAndSaveGrammarFile {in_filename {out_filename {}}
                                                {grammar_name {}}} {
    if {![string length $out_filename]} {
      set out_filename [file rootname $in_filename].tcl
    }
    set data [string trim [compileGrammarFile $in_filename $grammar_name]]
    set file [open $out_filename w]
    fconfigure $file -encoding utf-8
    puts $file {package require TclOO}
    puts $file $data
    close $file
  };# compileAndSaveGrammarFile

};# class ELEP::PatternEngine::GrammarCompiler

package provide ELEP::PatternEngine::GrammarCompiler 1.0
