##############################################################################
##  GrammarCompilerGenerator.tcl:                                           ##
## --------------------------------------------------                       ##
##  This class uses YETI to generate a YLEx scanner                         ##
##  (ELEP::PatternEngine::EBNFScanner) and a YETI parser that can compile   ##
##  a grammar file into an automaton. In other words, this class creates the##
##  grammar compiler classes.                                               ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

##
## Load some needed packages.
##
package require TclOO    ;# TclOO adds object-oriented facilities to Tcl.
package require ylex
package require yeti

oo::class create ELEP::PatternEngine::GrammarCompilerGenerator {
  variable sg {};# The lexical scanner generator object.
  variable pg {};# The parser generator object.

  variable binary_operators {match regexp}
  variable unary_operators  {isalnum isalpha isdigit isdouble isinteger
                             islower isupper ispunct isspace}


  constructor {} {
    constractScannerGenerator
    constractParserGenerator
  };# constructor

  ##########################################################################
  ## Method: constractScannerGenerator
  ##########################################################################
  method constractScannerGenerator {} {
    set sg [yeti::ylex ::ELEP::PatternEngine::#auto \
                             -name ELEP::PatternEngine::EBNFScanner]
    $sg add -state INITIAL {
      Rule:                  {return [list RULE               $yytext]}
      {\{}                   {return [list OPEN_BRACE         $yytext]}
      {\}}                   {return [list CLOSE_BRACE        $yytext]}
      {\(}                   {return [list OPEN_PAR           $yytext]}
      {\)}                   {return [list CLOSE_PAR          $yytext]}
      {\+}                   {return [list PLUS               $yytext]}
      {\*}                   {return [list STAR               $yytext]}
      {\?}                   {return [list QMARK              $yytext]}
      {\|}                   {return [list OR                 $yytext]}
      {\&}                   {return [list AND                $yytext]}
      {==}                   {return [list OP_EQ              $yytext]}
      {!=}                   {return [list OP_EQ_NOT          $yytext]}
    }

    # $sg add -state INITIAL {
    #   {-->.+?(?=<--|Rule:|\Z)} {
    #     return [list CODE [string trim [string range $yytext 3 end]]]
    #   }
    # }
    $sg add -state INITIAL {
      {-->\s*} {
        set yystate CODE
        return [list CODE_START $yystart]
      }
    }


    foreach op [concat $binary_operators $unary_operators] {
      $sg add -state INITIAL  $op \
              "return \[list OP_[string tolower $op] \$yytext\]"
      $sg add -state INITIAL !$op \
              "return \[list OP_[string tolower $op]_NOT \$yytext\]"
    }

    $sg add -state INITIAL {
      {\"([^\"]|\\\")*\"}    {return [list SYMBOL_STRING      $yytext]}
      {\[text\]}             {return [list SYMBOL_ANN_TEXT    $yytext]}
      {\[text\(\d+\)\]}      {return [list SYMBOL_ANN_TEXT_I  $yytext]}
      {\[start\]}            {return [list SYMBOL_ANN_START   $yytext]}
      {\[start\(\d+\)\]}     {return [list SYMBOL_ANN_START_I $yytext]}
      {\[end\]}              {return [list SYMBOL_ANN_END     $yytext]}
      {\[end\(\d+\)\]}       {return [list SYMBOL_ANN_END_I   $yytext]}
      {::match}              {return [list SYMBOL_NAME        match  ]}
      {:\w+}                 {return [list SYMBOL_NAME        $yytext]}
      {\w+}                  {return [list SYMBOL_WORD        $yytext]}
    }

    $sg add -state CODE {
      {\{}                   {return [list OPEN_BRACE         $yytext]}
      {\}}                   {return [list CLOSE_BRACE        $yytext]}
      {\=}                   {return [list OP_ASSIGN          $yytext]}
      {\;}                   {return [list SEMICOLON          $yytext]}
      {\.}                   {return [list PERIOD             $yytext]}
      {#[^\n]+\n}            {return [list COMMENT            $yytext]}
      {\"([^\"]|\\\")*\"}    {return [list SYMBOL_STRING      $yytext]}
      {:\w+}                 {return [list SYMBOL_NAME        $yytext]}
      {\w+}                  {return [list SYMBOL_WORD        $yytext]}
      {<--|\s(?=Rule:|\Z)}   {
        set yystate INITIAL
        return [list CODE_END $yystart]
      }
      {\n}                   {return [list SEMICOLON          $yytext]}
    }
    
  };# constractScannerGenerator

  ##########################################################################
  ## Method: constractParserGenerator
  ##########################################################################
  method constractParserGenerator {} {
    set pg [yeti::yeti ::ELEP::PatternEngine::#auto \
                             -name ELEP::PatternEngine::EBNFParser]

    $pg code public {
      variable pg {}          ;# The parser generator object.
      variable method_index 0
      variable symbol_methods ;# Maps from symbol constraints to method names.
      variable method_body    ;# Maps from method names to symbol constraints.
      variable method_vars    ;# Maps from method names to variables used.
      variable variables {}
      variable name      unnamed 

      # Operator counters
      variable op_counter_NAME 0
      variable op_counter_OR   0

      method constraintToMethod {constraint} {
        if {![info exists symbol_methods($constraint)]} {
          set method_name constraint_method_[incr method_index]
          set symbol_methods($constraint) $method_name
          set method_body($method_name) $constraint
          set method_vars($method_name) [lsort -unique $variables]
          set variables {}
        }
        return $symbol_methods($constraint)
      };# constraintToMethod


      method dump {} {
        $pg configure -name $name
        set data [$pg dump]
        ## Patch the code, with our step method...
        set start [string first "public method step {} \{" $data]
        set end   [string first "oops: unreachable code"   $data]
        incr start 23; incr end 23
        set code {
          if {$yyfinished} {
             error "step beyond end of parse"
          }
          
          if {$yyreadnext} {
              # Are there any symbol transitions?
              set yy_symbols [concat \
                [array names yytrans $yystate,S*] \
                [array names yyredux $yystate,S*]]
              if {[llength $yy_symbols]} {
                # Provide a hind to the lexer about the expected symbols...
                set yylookahead [$scanner next \
                  [lsort -unique $yy_symbols]]
                set yyreadnext 0
              }
          }
          
          set yyterm [lindex $yylookahead 0]
          set yydata [lindex $yylookahead 1]
          
          if {[info exists yytrans($yystate,$yyterm)]} {
              yyshift $yyterm $yydata
              set yyreadnext 1
              return [list $yyterm $yydata]
          } elseif {[info exists yyredux($yystate)] || \
                  [info exists yyredux($yystate,$yyterm)]} {
              if {[info exists yyredux($yystate)]} {
                  set yyruleno $yyredux($yystate)
              } else {
                  set yyruleno $yyredux($yystate,$yyterm)
              }
              
              set yyreduxlhs [lindex $yyrules($yyruleno) 0]
              set yyreduxdata [yyreduce $yyruleno]
              
              if {$yyreduxlhs == "__init__"} {
                  set yyfinished 1
              } else {
                  yyshift $yyreduxlhs $yyreduxdata
              }
              return [list $yyreduxlhs $yyreduxdata]
          } else {
              # global errorInfo
              # set yyerrmsg "parse error reading \"$yyterm\" in state $yystate"
              # set errorInfo $yyerrmsg
              # append errorInfo "\n    expecting one of:"
              # foreach yylaname [array names yytrans $yystate,*] {
              #     append errorInfo "\n        " [lindex [split $yylaname ,] 1]
              # }
              # yyerror $yyerrmsg
              # error $yyerrmsg $errorInfo

              # Errors are directed to the scanner!
              set yyfinished 1
              set yylookahead {{} {}}
              return [$scanner unexpected_symbol $yyterm]
          }
          
          error "oops: unreachable code"
        }
        set data [string replace $data $start $end "$code"]
        ## Patch the code, with our step method...
        set start [string first "public method parse {} \{"     $data]
        set end   [string first "return \[lindex \$yydata 1\]"   $data]
        incr start 24; incr end 25
        set code {
          if {$scanner == ""} {
            global errorInfo
            set yyerrmsg "cannot start parsing without a scanner"
            set errorInfo $yyerrmsg
            yyerror $yyerrmsg
            error $yyerrmsg $errorInfo
          }
          
          while {!$yyfinished} {
             set yydata [step]
             # puts "yydata: $yydata"
          }
          
          if {[lindex $yylookahead 0] != ""} {
            ## We have consumed a symbol. Put it back...
            #$scanner back
            # global errorInfo
            # set yyerrmsg "parser finished, but not at EOF (lookahead:\
            #               [lindex $yylookahead 0])"
            # set errorInfo $yyerrmsg
            # yyerror $yyerrmsg
            # error $yyerrmsg
          }
          
          return [lindex $yydata 1]
        }
        set data [string replace $data $start $end "$code"]
        return $data
      };# dump
    }

    $pg code constructor {{args} {
      eval [self] configure $args
      set pg [yeti::yeti ::ELEP::PatternEngine::temp::#auto]
    }
    }

    $pg code reset {
      set method_index    0
      set op_counter_NAME 0
      set op_counter_OR   0
      array unset symbol_methods *
      array unset method_body *
      array unset method_vars *
      set variables {}
      catch {itcl::delete object $pg}
      set pg [yeti::yeti ::ELEP::PatternEngine::temp::#auto]
    }

    $pg code destructor {{
      catch {itcl::delete object $pg}
    }
    }

    ##
    ## The following add block deals with symbols and their constraints.
    ## Example of symbols are:
    ##   {token [text] == "in" | [text] == "by"}
    ##   {token [text] regexp "19\\d\\d"}:date
    ##
    $pg add {
      symbol   SYMBOL_WORD             {lappend variables $1;return \$$1}
        |      SYMBOL_STRING           {return $1}
        |      SYMBOL_ANN_TEXT         {lappend variables $1;return \$text(0)}
        |      SYMBOL_ANN_TEXT_I       {lappend variables $1;return \$$1}
        |      SYMBOL_ANN_START        {lappend variables $1;return \$start(0)}
        |      SYMBOL_ANN_START_I      {lappend variables $1;return \$$1}
        |      SYMBOL_ANN_END          {lappend variables $1;return \$end(0)}
        |      SYMBOL_ANN_END_I        {lappend variables $1;return \$$1}

      token    {OPEN_PAR token CLOSE_PAR}    {return "($2)"}
        |      {token OR token}              {return "$1 || $3"}
        |      {token AND token}             {return "$1 && $3"}
        |      {symbol OP_EQ symbol}         {return "$1 == $3"}
        |      {symbol OP_EQ_NOT symbol}     {return "$1 != $3"}
    }
    foreach op $binary_operators {
      switch $op {
        match    {set cmd {\[string match $3 $1\]}}
        regexp   {set cmd {\[regexp -- $1 $3\]}}
        default  {error "unknown binary operator: $op"}
      }
      $pg add token [list symbol OP_[string tolower $op]     symbol] \
          "return \"$cmd\""
      $pg add token [list symbol OP_[string tolower $op]_NOT symbol] \
          "return \"!$cmd\""
    }
    foreach op $unary_operators {
      set cmd "string is [string range $op 2 end]"
      $pg add token [list symbol OP_[string tolower $op]    ] \
          "return \"\\\[$cmd \$1\\\]\""
      $pg add token [list symbol OP_[string tolower $op]_NOT] \
          "return \"!\\\[$cmd \$1\\\]\""
    }

    $pg add {
      expression {OPEN_BRACE SYMBOL_WORD token CLOSE_BRACE SYMBOL_NAME} {
        set rule __NAME_RULE_[incr op_counter_NAME]
        set symbol "{S $2 [constraintToMethod $3]}"
        $pg add $rule $symbol "define $5 \[list \$1\]"
        return $rule
      }

      expression {OPEN_BRACE SYMBOL_WORD token CLOSE_BRACE} {
        return "{S $2 [constraintToMethod $3]}"
      }

      expression {expression expression} {
        return "$1 $2"
      }

      expression {expression OR expression} {
        # Define a new rule, to be the OR of the two expressions.
        set rule __OR_RULE_[incr op_counter_OR]
        $pg add $rule $1
        $pg add $rule $3
        return "$rule"
      }

      expression {OPEN_PAR expression CLOSE_PAR} {
        return "$2"
      }

      expression {expression SYMBOL_NAME} {
        set rule __NAME_RULE_[incr op_counter_NAME]
        set code {
          set yy_ann_list [list $1]
          for {set yyi 2} {$yyi <= $yycount} {incr yyi} {
            lappend yy_ann_list [set $yyi]
          }
        }
            
        $pg add $rule $1 "$code; define $2 \$yy_ann_list"
        return $rule
      }

      expression {expression PLUS} {
        return "{+ $1}"
      }

      expression {expression STAR} {
        return "{* $1}"
      }

      expression {expression QMARK} {
        return "{? $1}"
      }

      code {CODE_START code_statements CODE_END} {
        return $2
      }

      code_statements {code_statements code_statements} {
        return "$1\n$2"
      }

      code_statements {code_statement} {
        return $1
      }

      code_statement {CODE} {
        return $1
      }

      code_statement {tcl_code_statement code_statement_end} {
        return $1
      }

      code_statement {SYMBOL_NAME PERIOD SYMBOL_WORD OP_ASSIGN 
                      OPEN_BRACE attributes CLOSE_BRACE code_statement_end} {
        return "createAnnotationFromDefinition $1 $3 {$6};"
      }

      code_statement {SYMBOL_WORD OP_ASSIGN 
                      OPEN_BRACE attributes CLOSE_BRACE code_statement_end} {
        return "createAnnotationFromMatch $1 {$4};"
      }

      code_statement_end {SEMICOLON} {
        return $1
      }

      code_statement_end {code_statement_end code_statement_end} {
        return $1
      }

      attributes {attributes attributes} {
        return "$1 $2"
      }

      attributes {attribute} {
        return $1
      }

      attribute {SYMBOL_WORD OP_ASSIGN SYMBOL_STRING} {
        return "$1 $3"
      }

      tcl_token {SYMBOL_WORD}   {return $1}
      tcl_token {SYMBOL_STRING} {return $1}
      tcl_code_statement {tcl_token} {return $1}
      tcl_code_statement {tcl_code_statement tcl_code_statement} {
        return "$1 $2"
      }
      tcl_code_statement {COMMENT} {
        return $1
      }


      rule {RULE SYMBOL_WORD expression code} {
        $pg add $2 $3 "\$scanner execute \{$4\}"
        $pg add start $2
      }

      rule {rule rule} {
      }

      start {rule} {
        set code {}
        foreach method [lsort [array names method_body]] {
          set c    "      method $method {doc ann} \{\n"
          foreach var $method_vars($method) {
            switch -glob -- $var {
              {\[text\]} {
                append c \
                  {set text(0) [tip_GetFirstAnnotatedTextRange $doc $ann]}
              }
              * {
                append c "set \{$var\} \[tip_GetAttribute \$ann \{$var\}\]\n"
                append c "if \{\[llength \$\{$var\}\]\} \{\n"
                append c "  set \{$var\} \[tip_GetValueValue \$\{$var\}\]\n"
                append c "\} else \{\n"
                append c "  set \{$var\} \{\}\n"
                append c "\}\n"
              }
            }
            append c \n
          }
          append c "if \{$method_body($method)\} \{return 1\}\n"
          append c "return 0"
          set c [string map [list \n "\n        "] $c]
          append c "\n      \}\;# $method\n"
          append code $c\n
        }
        $pg code public [string trimright $code]
        $pg code public {
           method define {definition list} {
             $scanner define $definition $list
           };# define
        }
      }
    }
  };# constractParserGenerator

  ##########################################################################
  ## Method: generateCode
  ##########################################################################
  method generateCode {} {
    return "[generateScannerCode]\n[generateParserCode]"
  };# generateCode

  ##########################################################################
  ## Method: generateScannerCode
  ##########################################################################
  method generateScannerCode {} {
    return [$sg dump]
  };# generateScannerCode

  ##########################################################################
  ## Method: generateParserCode
  ##########################################################################
  method generateParserCode {} {
    return [$pg dump]
  };# generateParserCode

  ##########################################################################
  ## Method: writePackages
  ##########################################################################
  method writePackages {} {
    set header \
{##############################################################################
##  %FILE%.tcl:                                                        ##
## --------------------------------------------------                       ##
##  This class was automatically generated by the                           ##
##  ELEP::PatternEngine::GrammarCompilerGenerator. Please DO NOT MODIFY     ##
##  this file.                                                              ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################
}
    set dir [file dirname [info script]]
    set file [open $dir/EBNFScanner.tcl w]
    puts $file [string map {%FILE% EBNFScanner} $header]
    puts $file [generateScannerCode]
    puts $file {package provide ELEP::PatternEngine::EBNFScanner 1.0}
    close $file
    set file [open $dir/EBNFParser.tcl w]
    puts $file [string map {%FILE% {EBNFParser }} $header]
    puts $file [generateParserCode]
    puts $file {package provide ELEP::PatternEngine::EBNFParser 1.0}
    close $file
  };# writePackages

  destructor {
    catch {itcl::delete object $sg}
    catch {itcl::delete object $pg}
  };# destructor

};# class ELEP::PatternEngine::GrammarCompilerGenerator

package provide ELEP::PatternEngine::GrammarCompilerGenerator 1.0

## Uncomment the following to generate the following packages:
## ELEP::PatternEngine::EBNFScanner (file: EBNFScanner.tcl)
## ELEP::PatternEngine::EBNFParser  (file: EBNFParser.tcl)
  set generator [ELEP::PatternEngine::GrammarCompilerGenerator #auto]
  $generator writePackages
  exit
