##############################################################################
##  GrammarExecutor.tcl:                                                    ##
## --------------------------------------------------                       ##
##  This class acts as a lexer, but drives the execution of a grammar.      ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

##
## Load some needed packages.
##
package require TclOO    ;# TclOO adds object-oriented facilities to Tcl.

oo::class create ELEP::PatternEngine::GrammarExecutor {
  variable document    {};# The document we operate on.
  variable grammar_obj {};# The object created from the grammar.

  constructor {{filename {}}} {
    if {[string length $filename]} {
      loadCompiledGrammar $filename
    }
  };# constructor

  ##########################################################################
  ## Method: loadCompiledGrammar
  ##########################################################################
  method loadCompiledGrammar {filename} {
    ## Open the file and read the first 512 characters. We need this header to
    ## identify the class name of the grammar.
    set file [open $filename]
    fconfigure $file -encoding utf-8
    set data [read $file 512]
    close $file

    if {![regexp {oo::class create\s+([^ ]+)\s+} $data _ grammar_class_name]} {
      error "ELEP::PatternEngine::GrammarExecutor::loadCompiledGrammar:\
             cannot identify the grammar class name. Perhaps not a compiled\
             grammar file? ($filename)"
    }
    unloadGrammar
    ## Ensure that older definitions of this class are removed...
    if {[llength [itcl::find classes $grammar_class_name]]} {
      if {[catch {itcl::delete class $grammar_class_name} error]} {
        error "ELEP::PatternEngine::GrammarExecutor::loadCompiledGrammar:\
               grammar \"filename\" has been loaded before and it is still in\
               use.\n($error)"
      }
    }
    uplevel #0 [list source $filename]
    ## Create a new object from the grammar...
    set grammar_obj [$grammar_class_name ELEP::PatternEngine::temp::#auto \
                             -scanner $this]
  };# loadCompiledGrammar

  ##########################################################################
  ## Method: parse
  ##########################################################################
  method parse {doc annotations} {
    start $annotations
    set document $doc
    while {!$yyfinished} {
      update
      $grammar_obj reset
      switch [$grammar_obj parse] {
        unexpected_symbol {
          # No match found.
        }
        default {
          if {$verbose} {
            puts $verbout "+-----------------------------------------------+"
            puts $verbout "|          M A T C H      F O U N D             |"
            puts $verbout "+-----------------------------------------------+"
            foreach ann $yymatch {
              puts $verbout "  [presentAnnotation $ann]"
            }
            puts $verbout "<===============================================>"
          }
        }
      }
      set yymatch {}
      array unset definitions *
    }
  };# parse

  ##########################################################################
  ## Method: unloadGrammar
  ##########################################################################
  method unloadGrammar {} {
    if {[string length $grammar_obj]} {
      itcl::delete object $grammar_obj
      set grammar_obj {}
    }
  };# unloadGrammar

  destructor {
    unloadGrammar
  };# destructor

  ##########################################################################
  ## YLEX scanner simulation:
  ##########################################################################
  public variable case 1
  public variable verbose 0
  public variable verbout stderr
  public variable yydata ""
  public variable yydata_len 0
  public variable yyindex 0
  public variable yystate "INITIAL"
  public variable yytext
  public variable yystart
  public variable yyend

  variable yymatch
  variable definitions

  variable yyfinished 0

  ##########################################################################
  ## Method: reset
  ##########################################################################
  public method reset {} {
    if {$verbose > 1} {
      puts $verbout "$this: reset, entering state INITIAL"
    }
    array unset definitions *
    set yyindex 0
    set yyfinished 0
    set yystate "INITIAL"

    set yymatch {}
  };# reset

  ##########################################################################
  ## Method: start
  ##########################################################################
  public method start {annotations} {
    reset
    set yydata $annotations
    set yydata_len [llength $yydata]
  };# start

  ##########################################################################
  ## Method: step
  ##########################################################################
  public method step {} {
  };# step

  ##########################################################################
  ## Method: next
  ##########################################################################
  public method next {{expected_symbols {}}} {
    # Have we reached the end of the buffer?
    if {$yyindex >= $yydata_len} {set yyfinished 1; return {}}
    set ann [lindex $yydata $yyindex]; incr yyindex
    if {$verbose > 1} {
      puts $verbout "next: [presentAnnotation $ann]"
    }
    # Are there any expected symbols?
    if {[llength $expected_symbols]} {
      foreach symbol $expected_symbols {
        foreach {_ annType method name} $symbol {break}
        lappend symbols($annType) [list $method $symbol]
        if {$verbose > 2} {
          puts $verbout "   expected symbol: $symbol"
        }
      }
      # Check whether the next annotation is of an expected type...
      set type  [tip_GetType $ann]
      if {[info exists symbols($type)]} {
        foreach item $symbols($type) {
          foreach {method symbol} $item {break}
          if {[$grammar_obj $method $document $ann]} {
            if {$verbose > 2} {
              puts $verbout "   -->> symbol matched: \"$symbol\"!\
                                     [presentAnnotation $ann]"
            }
            lappend yymatch $ann
            return [list [lreplace $symbol 0 0 S] $ann]
          }
        }
        # No matcing symbol has been found.
        if {$verbose > 2} {
          puts $verbout "   XXXX no symbol matched!\
                                     [presentAnnotation $ann]"
        }
        return [list $ann]
      }
    } else {
      if {$verbose > 2} {
        puts $verbout "   NO expected symbols!"
      }
    }
    return [list $ann]
  };# next

  ##########################################################################
  ## Method: back
  ##########################################################################
  public method back {} {
    incr yyindex -1
    if {$verbose > 1} {
      set ann [lindex $yydata $yyindex]
      puts $verbout "back: [presentAnnotation $ann]"
    }
  };# back

  ##########################################################################
  ## Method: unexpected_symbol
  ##########################################################################
  public method unexpected_symbol {symbol} {
    if {$verbose > 1} {
      if {$yyfinished} {
        puts $verbout "unexpected_symbol: (EOF)"
      } else {
        puts $verbout "unexpected_symbol: \"$symbol\""
      }
    }
    return {{} unexpected_symbol}
  };# unexpected_symbol

  ##########################################################################
  ## Method: run
  ##########################################################################
  public method run {} {
  };# run

  ##########################################################################
  ## Method: presentAnnotation
  ##########################################################################
  method presentAnnotation {ann} {
    return "([expr {$yyindex-1}])\
          \"[tip_GetFirstAnnotatedTextRange $document $ann]\" $ann"
  };# presentAnnotation

  ##########################################################################
  ## Code execution engine:
  ##########################################################################

  ##########################################################################
  ## Method: define
  ##########################################################################
  public method define {name ann_list} {
    set definitions($name) $ann_list
  };# define

  ##########################################################################
  ## Method: execute
  ##########################################################################
  public method execute {code} {
    eval $code
    foreach var [lsort [array names definitions]] {
      foreach ann $definitions($var) {
        puts "   $var =:= [presentAnnotation $ann]"
      }
    }
  };# execute

  ##########################################################################
  ## Method: createAnnotationFromDefinition
  ##########################################################################
  method createAnnotationFromDefinition {name ann_type attrs} {
    puts "Adding Annotation: $ann_type $attrs to definition $name"
    # Identify the first & last annotations...
    set first [lindex $definitions($name) 0]
    set last  [lindex $definitions($name) end]
    # Form a span from these two annotations...
    tip_GetFirstSpanOffsets $first start _
    tip_GetFirstSpanOffsets $last  _ end
    # Is there an existing annotation with the same span?
    set ann [tip_AnnotationsMatchingRange \
       [tip_SelectAnnotations $document $ann_type] $start $end]
    if {[llength $ann]} {
      if {[llength $ann] > 1} {set ann [lindex $ann 0]}
    } else {
      set ann [tip_CreateAnnotation $ann_type \
                  [tip_CreateSpanSet $start $end] {}]
    }
    # Add the attributes to the annotation:
    foreach {attr_name attr_val} $attrs {
      set ann [tip_PutAttribute $ann [tip_CreateAttribute $attr_name \
         [tip_CreateAttributeValue GDM_STRING $attr_val]]]
    }
    tip_AddAnnotation $document $ann
  };# createAnnotationFromDefinition

  ##########################################################################
  ## Method: createAnnotationFromMatch
  ##########################################################################
  method createAnnotationFromMatch {ann_type attrs} {
  };# createAnnotationFromMatch

};# class ELEP::PatternEngine::GrammarExecutor

package provide ELEP::PatternEngine::GrammarExecutor 1.0
