##############################################################################
##  ThreadPool.tcl:                                                         ##
## --------------------------------------------------                       ##
##  This class implements a class for managing a set of threads.            ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require TclOO
package require Thread

oo::class create ELEP::Process::ThreadPool {

  constructor {args} {
    my variable min_thread_number max_thread_number tpool \
                init_thread_cmd exit_thread_cmd idle_timeout pending_jobs

    set min_thread_number   1
    set max_thread_number   8
    set init_thread_cmd    {}
    set exit_thread_cmd    {}
    set idle_timeout      120 ;# Timeout (in seconds) before idle
                              ;# threads are destroyed.
    set pending_jobs       {}
    set tpool              {}

    my configure {*}$args

    set tpool [tpool::create -minworkers $min_thread_number \
                             -maxworkers $max_thread_number \
                             -idletime   $idle_timeout \
                             -initcmd    [my getProc onThreadInit] \
                             -exitcmd    [my getProc onThreadExit] ]
    tpool::preserve $tpool
  };# constructor

  ##########################################################################
  ## Method: getProc
  ##########################################################################
  method getProc {proc} {
    my variable {*}[info object vars [self object]]
    set class      [self class]
    set definition [info class definition $class $proc]
    set body       [lindex $definition 1]
    return [subst -nocommands -nobackslashes [string trim $body]]
  };# getProc

  ##########################################################################
  ## Method: onThreadInit
  ##########################################################################
  method onThreadInit {} {
    # puts "** onThreadInit"
    set ::auto_path [list $::auto_path]
    $init_thread_cmd
  };# onThreadInit

  ##########################################################################
  ## Method: onThreadExit
  ##########################################################################
  method onThreadExit {} {
    # puts "** onThreadExit"
    $exit_thread_cmd
  };# onThreadExit

  ##########################################################################
  ## Method: post
  ##########################################################################
  method post {script} {
    my variable tpool pending_jobs
    set id [tpool::post -nowait $tpool $script]
    lappend pending_jobs $id
    return $id
  };# post

  ##########################################################################
  ## Method: postAndForget
  ##########################################################################
  method postAndForget {script} {
    my variable tpool
    return [tpool::post -detached $tpool $script]
  };# postAndForget

  ##########################################################################
  ## Method: wait
  ##########################################################################
  method wait { {ids {}} } {
    my variable pending_jobs tpool
    if {![llength $ids]} {
      set ids $pending_jobs
    }
    return [tpool::wait $tpool $ids [my varname pending_jobs]]
  };# wait

  ##########################################################################
  ## Method: waitAll
  ##########################################################################
  method waitAll { {ids {}} } {
    my variable pending_jobs tpool
    if {![llength $ids]} {
      set ids $pending_jobs
    }
    foreach id $ids {set pending($id) 0}
    while {[array size pending]} {
      foreach completed [tpool::wait $tpool $ids [my varname pending_jobs]] {
        array unset pending $completed
      }
    }
    return $ids
  };# waitAll

  ##########################################################################
  ## Method: cancel
  ##########################################################################
  method cancel { {ids {}} } {
    my variable pending_jobs tpool
    if {![llength $ids]} {
      set ids $pending_jobs
    }
    return [tpool::cancel $tpool $ids]
  };# cancel

  ##########################################################################
  ## Method: result
  ##########################################################################
  method result { id } {
    my variable tpool
    return [tpool::get $tpool $id]
  };# result

  ##########################################################################
  ## Method: get
  ##########################################################################
  method get { id } {
    my variable tpool
    return [tpool::get $tpool $id]
  };# get

  destructor {
    my variable tpool
    catch {tpool::release $tpool}
    update idletasks
  };# destructor

};# class ELEP::Process::ThreadPool

package provide ELEP::Process::ThreadPool 1.0

# if {1} {
#   lappend auto_path [file dirname [file dirname [file normalize [info script]]]]
#   package require ELEP::Base::Utilities
#   set tp [ELEP::Process::ThreadPool new]
#   for {set i 1} {$i <= 10} {incr i} {
#     set id [$tp post "after [expr 1000*$i]; puts \{\tJob $i completed!\}
#                       set result $i"]
#     puts "Posted job: $id"; update
#   }
#   ## wait for all jobs...
#   set jobs [$tp waitAll]
#   puts "All finished!"
#   foreach one $jobs {
#     puts "-> Result $one: [$tp get $one]"
#   }
#   $tp destroy
#   exit
# }
