##############################################################################
##  CorpusStatistics.tcl:                                                   ##
## --------------------------------------------------                       ##
##  This class implements a library, that can be used to calcualte useful   ##
##  statistics about the tokens/sentences found in a corpus.                ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require TclOO

oo::class create ELEP::Statistics::CorpusStatistics {
  
  constructor {args} {
    my variable arrays
    set arrays {SentencesLength Count WordForms Lemmas Stems Pos}
    my reset
    my configure {*}$args
  };# constructor

  method reset {} {
    my variable arrays
    foreach one $arrays {
      my variable $one
      array unset $one
    }
  };# reset

  method processDocument {doc {sentence sentence} {constituents constituents}
                              {lemma lemma} {stem stem} {pos pos}} {
    my variable arrays
    my variable {*}$arrays
    ## Get all sentences, at least the ones that have constituents...
    set sentences [tip_SelectAnnotations $doc $sentence \
                                 "[llength ann::$constituents ]"]
    lappend Count(sentences) [llength $sentences]

    ## Iterate over sentences to process tokens...
    foreach one $sentences {
      set tokens [tip_GetValueValue [tip_GetAttribute $one $constituents]]
      set tokens_count [llength $tokens]
      incr DocumentCount(tokens)       $tokens_count
      lappend Count(TokensPerSentence) $tokens_count
      incr SentencesLength($tokens_count)
      array unset SentenceLemmas
      array unset SentenceStems
      array unset SentencePos
      foreach id $tokens {
        set token [tip_GetAnnotation $doc $id]
        ## Get word form...
        set form [tip_GetFirstAnnotatedTextRange $doc $token]
        incr WordForms($form)
        ## Process attributes...
        foreach attribute {lemma stem pos} array {Lemmas Stems Pos} {
          set attr [tip_GetAttribute $token [set $attribute]]
          if {![llength $attr]} {continue}
          set value [tip_GetValueValue $attr]
          incr Sentence${array}($value)
          if {$attribute eq "pos"} {
            set p [my genericPos $value]
            if {$p ne $value} {
              incr Sentence${array}($p)
            }
          }
        }
      }
      foreach array {Lemmas Stems Pos} {
        foreach one [array names Sentence$array] {
          lappend ${array}($one) [set Sentence${array}($one)]
        }
      }
    }
    lappend Count(tokens) $DocumentCount(tokens)
  };# processDocument

  method genericPos {pos} {
    switch -glob $pos {
      IDT -
      DDT  {return Determiner}
      VB*  {return Verb}
      NN*  {return Noun}
      JJ*  {return Adjective}
      ADJ* {return Adjective}
    }
    return $pos
  }

  method report {} {
    package require math
    package require math::statistics
    my variable arrays
    my variable {*}$arrays

    set msg    "Number of Sentences:  [math::sum {*}$Count(sentences)]\n"
    append msg "Number of Tokens:     [math::sum {*}$Count(tokens)]\n"
    if {[llength $Count(tokens)] > 1} {
      set  msg  [my basicStatistics $msg $Count(tokens)]
    }
    append msg "Token per Sentence:\n"
    set    msg  [my basicStatistics $msg $Count(TokensPerSentence)]
    append msg "Number of Word Forms: [array size WordForms]\n"
    append msg "Number of Lemmas:     [array size Lemmas]\n"
    append msg "Number of Stems:      [array size Stems]\n"
    append msg "Number of POSs:       [array size Pos]\n"
    foreach pos [lsort -dictionary [array names Pos]] {
      append msg "-> POS: $pos\n"
      set msg [my basicStatistics $msg $Pos($pos)]
      append msg =======================================\n
    }
    append msg "Sentences Length:\n"
    foreach len [lsort -integer [array names SentencesLength]] {
      append msg \
        "  Number of sentences with length $len: $SentencesLength($len)\n"
    }
    return $msg
    append msg "\n"
    append msg "\n"
    append msg "\n"
    append msg "\n"
    append msg "\n"
    append msg "\n"
  };# report

  method basicStatistics {msg values {tab {  }}} {
    foreach {mean min max count ssd sv psd pv} \
      [::math::statistics::basic-stats $values] {break}
    append msg "${tab}Sum:                           [::math::sum {*}$values]\n"
    append msg "${tab}Mean:                          $mean\n"
    append msg "${tab}Minimum value:                 $min\n"
    append msg "${tab}Maximum value:                 $max\n"
    append msg "${tab}Sample standard deviation:     $ssd\n"
    append msg "${tab}Sample variance:               $sv\n"
    append msg "${tab}Population standard deviation: $psd\n"
    append msg "${tab}Population variance:           $pv\n"
    append msg "${tab}Number of sentences:           $count\n"
    return $msg
  };# basicStatistics

};# class ELEP::Statistics::CorpusStatistics

package provide ELEP::Statistics::CorpusStatistics 1.0
