##############################################################################
##  TF-IDF-SQLite.tcl:                                                      ##
## --------------------------------------------------                       ##
##  This class implements a tf-idf calculation library, that uses an SQLite ##
##  database file as a data structure.                                      ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require TclOO
package require sqlite3

catch {ELEP::Statistics::TF-IDF-SQLite destroy}

oo::class create ELEP::Statistics::TF-IDF-SQLite {
  ## From:
  ## http://infomotions.com/blog/2009/04/tfidf-in-libraries-part-ii-of-iii-for-programmers/

  ## Term frequency (TF) is a percentage denoting the number of times a word
  ## appears in a document. It is mathematically expressed as C / T,
  ## where C is the number of times a word appears in a document and T
  ## is the total number of words in the same document.
  ## 
  ## Inverse document frequency (IDF) takes into acount that many words occur
  ## many times in many documents. Stop words and the word "human" in the
  ## MEDLINE database are very good examples. IDF is mathematically expressed
  ## as D / DF, where D is the total number of documents in a corpus and DF is
  ## the number of documents in which a particular word is found.
  ## As D / DF increases so does the significance of the given word.
  ## 
  ## Given these two factors, TFIDF is literally the product of TF and IDF:
  ## 
  ##     TFIDF = ( C / T ) * ( D / DF ) 
  ## 
  ## This is the basic form that has been used to denote relevance ranking
  ## for more than forty years, and please take note that it requires no
  ## advanced mathematical knowledge - basic arithmetic.
  ## 
  ## Like any good recipe or folk song, TFIDF has many variations.
  ## Google, for example, adds additional factors into their weighting scheme
  ## based on the popularity of documents. Other possibilities could include
  ## factors denoting the characteristics of the person using the texts.
  ## In order to accomodate for the wide variety of document sizes,
  ## the natural log of IDF will be employed throughout the balance of
  ## this demonstration. Therefore, for the purposes used here,
  ## TFIDF will be defined thus:
  ## 
  ##     TFIDF = ( C / T ) * log( D / DF )
  ##
  ##     C  = the number of times a word appears in a document (word frequency)
  ##     T  = the total number of words in the document
  ##     D  = the total number of documents in a corpus
  ##     DF = the number of documents in which a particular word is found

  ## Data structure:
  ##  We keep a single hash table for all terms in all documents (word_matrix).
  ##  The value stored is a dict. Dict keys are document indexes (indexing the
  ##  document "documents" list) and values are the word frequency in the
  ##  document (C).

  variable database_filename \
           database \
           current_document_id \
           word2word_id \
           cache_number_of_docs_containing_word

  constructor {{db_filename :memory:} args} {
    set database_filename $db_filename
    my init
    my configure {*}$args
  };# constructor

  destructor {
    catch {$database close}
  };# destructor

  method init {} {
    set database [self].db
    ## Does the file exists and has data?
    if {[file exists $database_filename] && [file size $database_filename]} {
      set init_db 0
    } else {
      set init_db 1
    }
    sqlite3 $database $database_filename
    if {$init_db} {my initdb}
  };# init

  method initdb {} {
    ## Table "words": Stores the list of all words found in the corpus.
    $database eval {
      CREATE TABLE words (
        id   INTEGER PRIMARY KEY AUTOINCREMENT,
        word TEXT    UNIQUE NOT NULL
      );
    }
    ## Table "documents": Stores the list of all documents in the corpus,
    ##                    along with the total number of words each document
    ##                    has.
    $database eval {
      CREATE TABLE documents (
        id    INTEGER PRIMARY KEY AUTOINCREMENT,
        doc   TEXT,
        words INTEGER NOT NULL
      );
    }
    ## Table "word_frequency": Stores the word frequencies per document.
    $database eval {
      CREATE TABLE word_frequency (
        word_id    INT     NOT NULL,
        doc_id     INT     NOT NULL,
        frequency  INTEGER NOT NULL,
        PRIMARY KEY (word_id, doc_id)
      );
    }
  };# initdb

  method resetdb {} {
    foreach table {word_frequency documents words} {
      $database eval "DROP TABLE $table"
    }
    array unset word2word_id
  };# resetdb

  method new_document {{id {}}} {
    $database eval {INSERT INTO documents(doc, words) VALUES($id, 0)}
    set current_document_id [$database last_insert_rowid]
  };# new_document

  method reset {} {
    my resetdb
    my initdb
  };# reset

  method set_current_document {doc} {
    set current_document_id $doc
  };# set_current_document

  method word2id {word} {
    if {[::info exists word2word_id($word)]} {return $word2word_id($word)}
    return [$database onecolumn {SELECT id FROM words WHERE word=$word}]
  };# word2id

  method add_tokens {tokens} {
    ## Update how many words this document has...
    set total_words [my total_words_in_document $current_document_id]
    incr total_words [llength $tokens]
    $database eval \
      {UPDATE documents SET words=$total_words WHERE id=$current_document_id}

    ## Iterate over all words (tokens) and update the per document
    ## frequences...
    foreach token $tokens {
      # set word_id [$database onecolumn \
      #    "SELECT id FROM words WHERE word='$token'"]
      if {[::info exists word2word_id($token)]} {
        set word_id $word2word_id($token)
        ## Increment the word frequency for this document...
        set freq [$database onecolumn {SELECT frequency FROM word_frequency\
             WHERE word_id=$word_id AND doc_id=$current_document_id}]
        if {$freq eq ""} {
          ## Add word frequency for this word & document...
          $database eval {INSERT INTO word_frequency\
                          VALUES($word_id, $current_document_id, 1)}
        } else {
          incr freq
          $database eval {UPDATE word_frequency SET frequency=$freq\
               WHERE word_id=$word_id AND doc_id=$current_document_id}
        }
      } else {
        ## The word is not in the database...
        ## Add the word to the words table...
        $database eval {INSERT INTO words(word) VALUES($token)}
        set word_id [$database last_insert_rowid]
        ## Add word frequency for this word & document...
        $database eval {INSERT INTO word_frequency\
                        VALUES($word_id, $current_document_id, 1)}
        set word2word_id($token) $word_id
      }
    }
  };# add_tokens

  method total_words_in_document {{doc -1}} {
    if {$doc < 0} {set doc $current_document_id}
    return [$database onecolumn {SELECT words FROM documents WHERE id=$doc}]
  };# total_words_in_document

  method document_number {} {
    return [$database onecolumn {SELECT COUNT(id) FROM documents}]
  };# document_number

  method document_id {} {
    return [$database onecolumn \
              {SELECT doc FROM documents WHERE id=$current_document_id}]
  };# documents

  method word_frequency_in_document {word doc} {
    if {[::info exists word2word_id($word)]} {
      set word_id $word2word_id($word)
    } else {
      set word_id \
        [$database onecolumn {SELECT id FROM words WHERE word=$word}]
      if {$word_id eq ""} {return 0}
    }
    set freq [$database onecolumn {SELECT frequency FROM word_frequency\
                                   WHERE word_id=$word_id AND doc_id=$doc}]
    if {$freq eq ""} {
      return 0
    }
    return $freq
  };# word_frequency_in_document

  method number_of_documents_containing_word {word} {
    if {[::info exists cache_number_of_docs_containing_word($word)]} {
      return $cache_number_of_docs_containing_word($word)
    }
    if {[::info exists word2word_id($word)]} {
      set word_id $word2word_id($word)
    } else {
      set word_id \
        [$database onecolumn {SELECT id FROM words WHERE word=$word}]
      set cache_number_of_docs_containing_word($word) 0
      if {$word_id eq ""} {return 0}
    }
    set cache_number_of_docs_containing_word($word) \
      [$database onecolumn \
            {SELECT COUNT(doc_id) FROM word_frequency WHERE word_id=$word_id}]
    return $cache_number_of_docs_containing_word($word)
  };# number_of_documents_containing_word

  method documents_containing_word {word} {
    if {[::info exists word2word_id($word)]} {
      set word_id $word2word_id($word)
    } else {
      set word_id \
        [$database onecolumn {SELECT id FROM words WHERE word=$word}]
      if {$word_id eq ""} {return {}}
    }
    return [$database eval \
              {SELECT doc_id FROM word_frequency WHERE word_id=$word_id}]
  };# documents_containing_word


  method tf {word doc} {
    set ttid [expr {double([my total_words_in_document $doc])}]
    if {$ttid == 0} {return 0.0}
    return [expr {[my word_frequency_in_document $word $doc] / $ttid}]
  };# tf

  method idf {word} {
    set dct [my number_of_documents_containing_word $word]
    if {$dct == 0} {return 1.0}
    return [expr {log ( [my document_number] / $dct)}]
  };# method

  method tf-idf {word {doc -1}} {
    if {$doc < 0} {set doc $current_document_id}
    return [expr {[tf $word $doc] * [idf $word]}]
  };# tf-idf

  method all_words {} {
    return [$database eval {SELECT word FROM words}]
  };# all_words

  method all_words_in_document {} {
    set word_ids [$database eval {SELECT word_id FROM word_frequency\
                                  WHERE doc_id=$current_document_id}]
    foreach word_id [lsort -integer -unique $word_ids] {
      lappend words [$database onecolumn {SELECT word FROM words\
                                          WHERE id=$word_id}]
    }
    return $words
  };# all_words_in_document

  method all_words_in_document_ranked {} {
    ## Get T (the total number of words in the document)...
    set T [expr {double([my total_words_in_document $current_document_id])}]
    if {$T < 1.0} {return {}}
    ## Get D (the total number of documents in a corpus)...
    set D [my document_number]

    set rank [dict create]
    foreach word [my all_words_in_document] {
      ## Get C (number of times the word appears in document (word frequency))
      set C  [my word_frequency_in_document $word $current_document_id]
      ## Get DF (the number of documents in which the word is found)
      set DF [my number_of_documents_containing_word $word]
      ## Calculate tf-idf...
      set tf_idf [expr {( $C / $T ) * log( $D / $DF )}]
      dict lappend rank $tf_idf $word
    }
    return $rank
  };# all_words_in_document_ranked

  method process_document_annotations {doc type {constraints {}}} {
    ## Get T (the total number of words in the document)...
    set T [expr {double([my total_words_in_document $current_document_id])}]
    if {$T < 1.0} {return {}}
    ## Get D (the total number of documents in a corpus)...
    set D [my document_number]

    foreach ann [tip_SelectAnnotations $doc $type $constraints] {
      set word [string tolower [tip_GetFirstAnnotatedTextRange $doc $ann]]
      ## Get C (number of times the word appears in document (word frequency))
      if {[::info exists cache_word_frequency_in_document($word)]} {
        set C $cache_word_frequency_in_document($word)
      } else {
        set C [my word_frequency_in_document $word $current_document_id]
        set cache_word_frequency_in_document($word) $C
      }
      ## Get DF (the number of documents in which the word is found)
      if {[::info exists cache_number_of_docs_containing_word($word)]} {
        set DF $cache_number_of_docs_containing_word($word)
      } else {
        set DF [my number_of_documents_containing_word $word]
      }
      ## Calculate tf-idf...
      set tf_idf [expr {( $C / $T ) * log( $D / $DF )}]
      ## Add attribute...
      tip_AddAnnotation $doc [tip_PutAttribute $ann \
          [tip_CreateAttribute tf-idf \
               [tip_CreateAttributeValue GDM_STRING $tf_idf]]]
    }
  };# process_document_annotations

  ##
  ## Iterations...
  ##

  method first_document {} {
    set current_document_id 1
    return [expr {$current_document_id < [my document_number]}]
  };# first_document

  method next_document {} {
    incr current_document_id
    return [expr {$current_document_id < [my document_number]}]
  };# next_document

  ##
  ## Serialisation...
  ##

  method save {filename} {
    $database backup [file normalize $filename]
  };# save

  method load {db_filename {reconstruct_cache 1}} {
    catch {$database close}
    set database_filename $db_filename
    sqlite3 $database $database_filename
    if {$reconstruct_cache} {my reconstruct_cache}
  };# load 

  method reconstruct_cache {} {
    array unset word2word_id
    $database eval {SELECT * FROM words ORDER BY id} {
      set word2word_id($word) $id
    }
  };# reconstruct_cache

  method add_indexes {} {
    $database eval {CREATE INDEX IF NOT EXISTS idx_word_frequency
                    ON word_frequency (word_id, doc_id)}
    $database eval {CREATE INDEX IF NOT EXISTS idx_words
                    ON words (word)}
  };# add_indexes

};# class ELEP::Statistics::TF-IDF-SQLite

package provide ELEP::Statistics::TF-IDF-SQLite 1.0

if {0} {
  ## Sample usage code
  set o [ELEP::Statistics::TF-IDF-SQLite new]
  $o reset

  for {set doc 0} {$doc < 5} {incr doc} {
    $o new_document
    $o add_tokens [list this is document $doc . is hello from here !]
  }
  
  for {set doc 1} {$doc <= 5} {incr doc} {
    $o set_current_document $doc
    puts "*********  Document: $doc"
    puts [$o all_words_in_document_ranked]
  }
  $o destroy
}
