##############################################################################
##  TF-IDF.tcl:                                                             ##
## --------------------------------------------------                       ##
##  This class implements a tf-idf calculation library.                     ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require TclOO

## catch {itcl::delete class ELEP::Statistics::TF-IDF}

oo::class create ELEP::Statistics::TF-IDF {
  ## From:
  ## http://infomotions.com/blog/2009/04/tfidf-in-libraries-part-ii-of-iii-for-programmers/

  ## Term frequency (TF) is a percentage denoting the number of times a word
  ## appears in a document. It is mathematically expressed as C / T,
  ## where C is the number of times a word appears in a document and T
  ## is the total number of words in the same document.
  ## 
  ## Inverse document frequency (IDF) takes into acount that many words occur
  ## many times in many documents. Stop words and the word "human" in the
  ## MEDLINE database are very good examples. IDF is mathematically expressed
  ## as D / DF, where D is the total number of documents in a corpus and DF is
  ## the number of documents in which a particular word is found.
  ## As D / DF increases so does the significance of the given word.
  ## 
  ## Given these two factors, TFIDF is literally the product of TF and IDF:
  ## 
  ##     TFIDF = ( C / T ) * ( D / DF ) 
  ## 
  ## This is the basic form that has been used to denote relevance ranking
  ## for more than forty years, and please take note that it requires no
  ## advanced mathematical knowledge - basic arithmatic.
  ## 
  ## Like any good recipe or folk song, TFIDF has many variations.
  ## Google, for example, adds additional factors into their weighting scheme
  ## based on the popularity of documents. Other possibilities could include
  ## factors denoting the characteristics of the person using the texts.
  ## In order to accomodate for the wide variety of document sizes,
  ## the natural log of IDF will be employed throughout the balance of
  ## this demonstration. Therefore, for the purposes used here,
  ## TFIDF will be defined thus:
  ## 
  ##     TFIDF = ( C / T ) * log( D / DF )
  ##
  ##     C  = the number of times a word appears in a document (word frequency)
  ##     T  = the total number of words in the document
  ##     D  = the total number of documents in a corpus
  ##     DF = the number of documents in which a particular word is found

  ## Data structure:
  ##  We keep a single hash table for all terms in all documents (word_matrix).
  ##  The value stored is a dict. Dict keys are document indexes (indexing the
  ##  document "documents" list) and values are the word frequency in the
  ##  document (C).

  variable word_matrix \
           documents \
           total_words_in_document \
           current_document_index \
           document_number
  
  constructor {args} {
    my init
    my configure {*}$args
  };# constructor

  method init {} {
    array unset word_matrix *
    set documents               {}
    set total_words_in_document {}
    set document_number          0
    set current_document_index  -1
  };# init

  method new_document {{id {}}} {
    lappend documents $id
    lappend total_words_in_document 0
    incr    current_document_index
    incr    document_number
  };# new_document

  method reset {} {
    my init
  };# reset

  method set_current_document {doc} {
    set current_document_index $doc
  };# set_current_document

  method add_tokens {tokens} {
    ## Update how many words this document has...
    set  total_words [lindex  $total_words_in_document $current_document_index]
    incr total_words [llength $tokens]
    lset total_words_in_document $current_document_index $total_words

    ## Iterate over all words (tokens) and update the per document
    ## frequences...
    foreach token $tokens {
      if {![::info exists word_matrix($token)]} {
        ## Initialise this term value with a dictionary...
        set word_matrix($token) [dict create]
      }
      ## Increment the word frequency for this document...
      dict incr word_matrix($token) $current_document_index
    }
  };# add_tokens

  method total_words_in_document {doc} {
    return [lindex $total_words_in_document $doc]
  };# total_words_in_document

  method word_frequency_in_document {word doc} {
    return [dict get $word_matrix($word) $doc]
  };# word_frequency_in_document

  method document_number {} {
    return $document_number
  };# document_number

  method document_id {} {
    return [lindex $documents $current_document_index]
  };# documents

  method documents_containing_word {word} {
    if {[::info exists word_matrix($word)]} {
      return [dict keys $word_matrix($word)]
    }
    return {}
  };# documents_containing_word

  method number_of_documents_containing_word {word} {
    if {[::info exists word_matrix($word)]} {
      return [dict size $word_matrix($word)]
    }
    return 0
  };# number_of_documents_containing_word

  method tf {word doc} {
    set ttid [expr {double([my total_words_in_document $doc])}]
    if {$ttid == 0} {return 0.0}
    return [expr {[my word_frequency_in_document $word $doc] / $ttid}]
  };# tf

  method idf {word} {
    set dct [my number_of_documents_containing_word $word]
    if {$dct == 0} {return 1.0}
    return [expr {log ( $document_number / $dct)}]
  };# method

  method tf-idf {word {doc -1}} {
    if {$doc < 0} {set doc $current_document_index}
    return [expr {[my tf $word $doc] * [idf $word]}]
  };# tf-idf

  method all_words {} {
    return [array names word_matrix]
  };# all_words

  method all_words_in_document {} {
    set words {}
    set searchId [array startsearch word_matrix]
    while {[string length [set word \
                                 [array nextelement word_matrix $searchId]]]} {
      if {[dict exists $word_matrix($word) $current_document_index]} {
        lappend words $word
      }
    }
    array donesearch word_matrix $searchId
    return $words
  };# all_words_in_document

  method all_words_ranked {} {
    ## Get T (the total number of words in the document)...
    set T [expr {double([my total_words_in_document $current_document_index])}]
    if {$T < 1.0} {return {}}
    ## Get D (the total number of documents in a corpus)...
    set D $document_number

    set rank [dict create]
    foreach word [my all_words_in_document] {
      ## Get C (number of times the word appears in document (word frequency))
      set C  [dict get $word_matrix($word) $current_document_index]
      ## Get DF (the number of documents in which the word is found)
      set DF [dict size $word_matrix($word)]
      ## Calculate tf-idf...
      set tf_idf [expr {( $C / $T ) * log( $D / $DF )}]
      dict lappend rank $tf_idf $word
    }
    return $rank
  };# all_words_ranked

  method process_document_annotations {doc type {constraints {}}} {
    ## Get T (the total number of words in the document)...
    set T [expr {double([my total_words_in_document $current_document_index])}]
    if {$T < 1.0} {return {}}
    ## Get D (the total number of documents in a corpus)...
    set D $document_number

    foreach ann [tip_SelectAnnotations $doc $type $constraints] {
      set word [string tolower [tip_GetFirstAnnotatedTextRange $doc $ann]]
      ## Get C (number of times the word appears in document (word frequency))
      set C  [dict get $word_matrix($word) $current_document_index]
      ## Get DF (the number of documents in which the word is found)
      set DF [dict size $word_matrix($word)]
      ## Calculate tf-idf...
      set tf_idf [expr {( $C / $T ) * log( $D / $DF )}]
      ## Add attribute...
      tip_AddAnnotation $doc [tip_PutAttribute $ann \
          [tip_CreateAttribute tf-idf \
               [tip_CreateAttributeValue GDM_STRING $tf_idf]]]
    }
  };# process_document_annotations

  ##
  ## Iterations...
  ##

  method first_document {} {
    set current_document_index 0
    return [expr {$current_document_index < $document_number}]
  };# first_document

  method next_document {} {
    incr current_document_index
    return [expr {$current_document_index < $document_number}]
  };# next_document

  ##
  ## Serialisation...
  ##

  method save {filename} {
    set fd [open $filename w]
    fconfigure $fd -encoding utf-8
    set searchId [array startsearch word_matrix]
    while {[string length [set word \
                                 [array nextelement word_matrix $searchId]]]} {
      ## Write directory...
      set dict $word_matrix($word)
      puts $fd "set dict \[dict create\]"
      dict for {key value} $dict {
        puts $fd "dict set dict \[cache $key\] \[cache $value\]"
      }

      puts $fd "set word \{[string map {\\ \\\\ \n \ } $word]\}"
      puts $fd "set word_matrix(\$word) \$dict"
    }
    array donesearch word_matrix $searchId

    foreach var {documents total_words_in_document current_document_index
                 document_number} {
      puts $fd "set $var \{[set $var]\}"
    }
    close $fd
  };# save

  method load {filename} {
    my init
    set fd [open $filename]
    fconfigure $fd -encoding utf-8
    set data [read $fd]
    close $fd
    eval $data
  };# load 

};# class ELEP::Statistics::TF-IDF

package provide ELEP::Statistics::TF-IDF 1.0

if {0} { 
  ## Sample usage code
  set o [ELEP::Statistics::TF-IDF new]
  
  for {set doc 0} {$doc < 5} {incr doc} {
    $o new_document
    $o add_tokens [list this is document $doc . is hello from here !]
  }
  
  for {set doc 0} {$doc < 5} {incr doc} {
    $o set_current_document $doc
    puts "*********  Document: $doc"
    puts [$o all_words_ranked]
  }
  $o destroy
}
