#=======================================================================
#
# Package: ASDOT.tcl
#
# Purpose: A tcl package to allow the reading and writing of graphs
# using the the Graphviz  format.
# See http://www.research.att.com/sw/tools/graphviz/ for details of the
# Graphviz tools

# Function Name  Description
# -------------  -------------------------------------------------------
# read_dot       Read and parse a .dot file.  Build a graph from the description
# write_dot      Create a .dot file from a Tcl ::struct::graph object

# History
# Version  Created By  Description
# -------  ----------  -------------------------------------------------
# 1.0      J McCann    Initial version

package provide ELEP::Structures::Graph::ASDOT::Exporter 1.0

#=======================================================================
# NAMESPACE ELEP::Structures::Graph::ASDOT::Exporter

namespace eval ELEP::Structures::Graph::ASDOT {}
namespace eval ELEP::Structures::Graph::ASDOT::Exporter {

   #===================================================================
   # Export the functions that the user should be calling
   namespace export write_dot

   # The graph package is needed for obvious reasons so we include
   # the struct package that contains it
   package require struct

   # Define the namespace variables
   variable n_graph

   # List of all the possible attributes for a node
   variable n_nodeAttr { bottomlabel color comment distortion \
                             fillcolor fixedsize fontcolor \
                             fontname fontsize group height \
                             label layer orientation peripheries \
                             regular shape shapefile sides skew \
                             style toplabel URL width z }

   # List of all possible attributes for an edge (arc)
   variable n_edgeAttr { arrowhead arrowsize arrowtail color comment \
                             constraint decorate dir fontcolor \
                             fontname fontsize headlabel headport \
                             headURL label labelangle labeldistance \
                             labelfloat labelfontcolor labelfontname \
                             labelfontsize layer lhead ltail minlen \
                             samehead sametail style taillabel \
                             tailport tailURL weight }

   # List of all possible attributes for a graph
   variable n_graphAttr { bgcolor center clusterrank color comment \
                              compound concentrate fillcolor \
                              fontcolor fontname fontpath fontsize \
                              label labeljust labelloc layers margin \
                              mclimit nodesep nslimit nslimit1 ordering \
                              orientation page pagedir quantum rank \
                              rankdir ranksep ratio remincross rotate \
                              samplepoints searchsize size style URL }

   #-------------------------------------------------------------------
   # writeGraphAttr
   #    Check to see if a graph level attribute exists.  If it does
   #    then write it into the output file
   # df - output file
   # g  - graph
   # pf - prefix this will be blank or one of node,|graph,|edge,
   # attr - attribute name
   proc writeGraphAttr { df g pf attr } {

       if { ![catch {
           set val [$g get "${pf}${attr}"]
       }] } {
           puts $df "   $attr = \"$val\""
       }
   }

   proc getNodeSymbolName {n} {
     return \"$n\"
   }

   #-------------------------------------------------------------------
   # getNodeAttr
   #    A node may have zero, one or many attributes defined for it.
   #    If it has none just return a semicolon.  If it has one or more
   #    the return them in the correct square bracket delimiters
   proc getNodeAttr {g n} {

       variable n_nodeAttr

       set attrCnt 0

       set result "   [getNodeSymbolName $n] \[\n"
       set attrs {}
       foreach attrName $n_nodeAttr {
           if { ![catch {
               set val [$g node get $n $attrName]
           }] } {
               lappend attrs "$attrName=\"$val\""
               incr attrCnt
           }
       }
       append result "      [join $attrs {;\n      }]\n"
       append result "   \] ;\n"

       if { $attrCnt == 0 } {
           set result "   [getNodeSymbolName $n] ;"
       }

       return $result
   }

   #-------------------------------------------------------------------
   # getEdgeAttr
   #    An edge may have zero, one or many attributes defined for it.
   #    If it has none just return a semicolon.  If it has one or more
   #    the return them in the correct square bracket delimiters
   proc getEdgeAttr {g e} {

       variable n_edgeAttr

       set srcNode [$g arc source $e]
       set destNode [$g arc target $e]

       set attrCnt 0

       set arcDef "   [getNodeSymbolName ${srcNode}]"
       # Check to see if a from node record has been specfied
       if { ![catch {
           set val [$g arc get $e "fromRecord"]
       }] } {
           append arcDef ":${val}"
       }

       append arcDef " -> "
       append arcDef "[getNodeSymbolName ${destNode}]"
       # Check to see if a to node record has been specfied
       if { ![catch {
           set val [$g arc get $e "toRecord"]
       }] } {
           append arcDef ":${val}"
       }

       set result "$arcDef \[\n"

        foreach attrName $n_edgeAttr {
           if { ![catch {
               set val [$g arc get $e $attrName]
           }] } {
               append result "      $attrName = \"$val\"\n"
               incr attrCnt
           }
       }
       append result "   \] ;\n"

       if { $attrCnt == 0 } {
           set result "$arcDef ;"
       }

       return $result
   }

   #-------------------------------------------------------------------
   # lreverse
   #    Reverse the order of the elements in a list
   # l - list to reverse
   proc lreverse { l } {

       set result {}

       set ll [expr [llength $l] -1]
       for {set i $ll} {$i >= 0} {incr i -1} {
           lappend result [lindex $l $i]
       }

       return $result
   }

   #-------------------------------------------------------------------
   # sg_sort
   #    Decide the order of two elements of the with_sg list
   # e1 - element 1
   # e2 - element 2
   proc sg_sort { e1 e2 } {
       set e1_sg [lindex $e1 0]
       set e2_sg [lindex $e2 0]
       return [string compare $e1_sg $e2_sg]
   }

   #===================================================================
   # write_dot
   #
   # file - Name of the file to write the graph format data out to
   # graph - The graph to be written to the file
   proc write_dot { file graph } {

       variable n_nodeAttr
       variable n_edgeAttr
       variable n_graphAttr

       # Open the file for output
       set df [open $file w]

       # Write the header information into the graph
       puts $df "digraph \"[set graph]\" \{\n"

       # Write out any graph level attributes
       foreach attrName $n_graphAttr {
           writeGraphAttr $df $graph "" $attrName
       }

       # Write out the graph section where graph attributes
       # can also be specified
       puts $df "   graph \["
       foreach attrName $n_edgeAttr {
           writeGraphAttr $df $graph "graph," $attrName
       }
       puts $df "   \];\n"

       # Write out the node section where attributes for all
       # nodes are specified
       puts $df "\n   node \["
       foreach attrName $n_nodeAttr {
           writeGraphAttr $df $graph "node," $attrName
       }
       puts $df "   \];\n"

       # Now write out the details of each node
       foreach node [lsort -dictionary [$graph nodes]] {
           puts $df [getNodeAttr $graph $node]
       }

       # Write out the edge section where attributes for all
       # edges are specified
       puts $df "   edge \["
       foreach attrName $n_edgeAttr {
           writeGraphAttr $df $graph "edge," $attrName
       }
       puts $df "   \];\n"
       
       # Now write out the details of each arc.  This process is complicated
       # by the need to cater for subgraphs
       set with_sg {}
       set no_sg {}
       foreach arc [$graph arcs] {

           # First partition the arcs into those that do and those that
           # do not have subgraphs
           if { [catch {
               set sg [$graph arc get $arc "subgraph"]
           } ] } {
               lappend no_sg $arc
           } else {
               lappend with_sg [list [lreverse $sg] $arc]
           }
       }

       # Write out the arcs that ar not involved in a subgraph
       foreach arc $no_sg {
           puts $df [getEdgeAttr $graph $arc]
       }

       if { [llength $with_sg] > 0 } {

           # Now write out the arcs that are involved in a subgraph
           set last_sg {}
           set last_level 0

           foreach sg_arc [lsort -command sg_sort $with_sg] {
               set sg [lindex $sg_arc 0]
               set arc [lindex $sg_arc 1]
               set level [llength $sg]

               #puts $df "$last_level --> $level"
               #puts $df "$last_sg --> $sg"

               if { ! [string equal $last_sg $sg] } {
                   if { $level > $last_level } {
                       puts $df "   subgraph [lindex $sg end] \{"
                   } elseif { $level == $last_level } {
                       puts $df "   subgraph [lindex $sg end] \{"
                       if { $level > 0 } {
                           puts $df "   \}"
                       }
                   } elseif { $level < $last_level } {

                       # Close of the difference in levels plus 1
                       set closeLevels [expr $last_level - $level + 1]
                       for {set i 0} {$i < $closeLevels} {incr i} {
                           puts $df "   \}"
                       }
                       # Start a new subgraph
                       puts $df "   subgraph [lindex $sg end] \{"
                   }
               }

               puts $df "[getEdgeAttr $graph $arc]"

               set last_sg $sg
               set last_level $level
           }
           # Close of the last remaining subgraphs
           set closeLevels [expr $last_level - $level + 1]
           for {set i 0} {$i < $closeLevels} {incr i} {
               puts $df "   \}"
           }
       }
       # Write the trailing bracket into the graph
       puts $df "\}"

       # Close the file
       close $df

   }
};# namespace ELEP::Structures::Graph::ASDOT::Exporter
