##############################################################################
##  ComponentOptionsEditor.tcl:                                             ##
## --------------------------------------------------                       ##
##  This file implements a class that can edit the options of a component.  ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities::DialogWindow

oo::class create ELEP::System::ComponentOptionsEditor {
  superclass ELEP::WindowUtilities::DialogWindow

  constructor {args} {
    next
    my variable name options spec applycmd
    set name     {}
    set options  {}
    set spec     {}
    set applycmd {}
    my configure {*}$args
    my createWidgets
    my onUpdate
    my title "[my msg {Component Parameters}]: $name"
  };# constructor

  method onUpdate {} {
  };# onUpdate

  method onDefault {} {
    my variable client_area value
    foreach index [array names value] {
      catch {$client_area.fr.def$index invoke}
    }
  };# onDefault

  method onApply {} {
    my variable options spec value applycmd
    set index 0
    set options {}
    foreach {label pattern default} $spec {
      set v $value($index)
      switch -glob -- $pattern {
        {*%FILE%*}       {set p %FILE%      }
        {*%DIRECTORY%*}  {set p %DIRECTORY% }
        {*%NUMBER%*}     {set p %NUMBER%    }
        {*%COMBO%*}      {set p %COMBO%;    }
        {*%ENCODING%*}   {set p %ENCODING%  }
        {*%ENTRY%*}      {set p %ENTRY%     }
        {*%COMMAND%*}    {set p %COMMAND%;  }
        {*%BOOLEAN%*}    {set p %BOOLEAN%   }
        default          {set p {}          }
      }
      if {$p ne "" && $v ne ""} {
        regsub -all -- $p $pattern $v v
      }
      lappend options $v
      incr index
    }
    if {$applycmd ne ""} {
      set cmd $applycmd
      lappend cmd [self]
      {*}$cmd
    }
  };# onApply

  method onOk {} {
    my onApply
    my destroy
  };# onOk

  ##########################################################################
  ## Method: createWidgets
  ##########################################################################
  method createWidgets {} {
    my variable client_area name options spec index padx pady value
    set index 0
    set padx 2
    set pady 2
    set msg "[my msg {Component Parameters}]: $name"
    ttk::labelframe $client_area.fr -text $msg
      my loadImage IModuleProperties
      set w $client_area.fr
      ttk::label $w.header -image [my loadImage IModuleProperties] -anchor w \
                           -text  $name -compound left -font {arial 14 bold}
      if {[catch {PUL_plugin_module_config $name description} description]} {
        set description [my msg {No description available}]
      }
      ttk::label $w.description -text $description -anchor center
      grid $w.header      - - - -padx 2 -pady 2 -sticky snew
      grid $w.description - - - -padx 2 -pady 2 -sticky snew
      grid columnconfigure $w 1 -weight 1
      foreach {label pattern default} $spec curr_val $options {
        set value($index) $curr_val
        ## Extract the actual value from the current value...
        regsub -- {%[A-Z]+%} $pattern {(.*)} mask 
        regexp -- $mask $curr_val _ value($index)
        my createWidgetsAddOption $w $label $pattern $default $curr_val
        incr index
      }
      if {[llength $options] == 0} {
        ttk::label $w.none -textvariable [my msgVar NoModuleParameters] \
          -anchor center
        grid $w.none - - - -padx 2 -pady 2 -sticky snew
      }
    grid $client_area.fr -padx 2 -pady 2 -sticky snew
    grid columnconfigure $client_area 0 -weight 1
    grid rowconfigure    $client_area 0 -weight 1
    ## Add buttons...
    my addStdButton def    [list [self] onDefault]
    my addStdButton apply  [list [self] onApply]
    my addStdButton ok     [list [self] onOk]
    my addStdButton cancel [list [self] destroy]
  };# createWidgets

  ##########################################################################
  ## Method: createWidgetsAddOption
  ##########################################################################
  method createWidgetsAddOption {w lbl pattern default curr_val} {
    switch -glob -- $pattern {
      {*%FILE%*}       {my widget_file      $w $lbl $pattern $default $curr_val}
      {*%DIRECTORY%*}  {my widget_directory $w $lbl $pattern $default $curr_val}
      {*%NUMBER%*}     {my widget_number    $w $lbl $pattern $default $curr_val}
      {*%COMBO%*}      {my widget_combo     $w $lbl $pattern $default $curr_val}
      {*%ENCODING%*}   {my widget_encoding  $w $lbl $pattern $default $curr_val}
      {*%ENTRY%*}      {my widget_entry     $w $lbl $pattern $default $curr_val}
      {*%COMMAND%*}    {my widget_command   $w $lbl $pattern $default $curr_val}
      {*%BOOLEAN%*}    {my widget_boolean   $w $lbl $pattern $default $curr_val}
      default          {my widget_default   $w $lbl $pattern $default $curr_val}
    }
  };# createWidgetsAddOption

  method default_value {type index} {
    my variable spec value
    set i [expr {$index*3+1}]
    set pattern [lindex $spec $i]
    set default [lindex $spec [incr i]]
    set default [uplevel #0 subst [list $default]]
    switch -glob -- $pattern {
      {*%COMBO%*}      {set value($index) [lindex $default 0]}
      {*%COMMAND%*}    {}
      default          {set value($index) $default}
    }
  };# default_value

  method default_button {w type} {
    my variable index padx pady
    upvar label label
    ttk::label $w.label_$index -text $label -anchor e
    ttk::button $w.def$index -image [my loadImage ITCancel] \
      -command [list [self] default_value $type $index]
    if {[winfo exists $w.button_$index]} {
      grid $w.label_$index $w.entry_$index $w.button_$index $w.def$index \
        -sticky snew -padx $padx -pady $pady
    } else {
      grid $w.label_$index $w.entry_$index - $w.def$index \
        -sticky snew -padx $padx -pady $pady
    }
  };# default_button

  method widget_file {w label pattern default curr_val} {
    my variable index
    set width [string length $curr_val]
    if {$width < 20} {set width 20}
    if {$width > 50} {set width 50}
    ttk::entry $w.entry_$index -width $width \
      -textvariable [my varname value]($index)
    $w.entry_$index xview moveto 1
    ttk::button $w.button_$index -compound left -image [my loadImage ITOpen] \
      -textvariable [my msgVar SelectFile] \
      -command [list GUI_SelectFile $w.entry_$index]
    GUI_MakeEntryFileDropTarget $w.entry_$index
    my default_button $w entry
  };# widget_file

  method widget_directory {w label pattern default curr_val} {
    my variable index
    set width [string length $curr_val]
    if {$width < 20} {set width 20}
    if {$width > 50} {set width 50}
    ttk::entry $w.entry_$index -width $width \
      -textvariable [my varname value]($index)
    $w.entry_$index xview moveto 1
    GUI_MakeEntryDirectoryDropTarget $w.entry_$index
    ttk::button $w.button_$index -compound left -image [my loadImage ITOpen] \
      -textvariable [my msgVar SelectFile] \
      -command [list GUI_SelectDirectory $w.entry_$index]
    my default_button $w entry
  };# widget_directory

  method widget_number {w label pattern default curr_val} {
    my widget_entry $w $label $pattern $default $curr_val
  };# widget_number

  method widget_combo {w label pattern default curr_val} {
    my variable index
    ttk::combobox $w.entry_$index -values $default \
      -textvariable [my varname value]($index) -state readonly
    # $w.entry_$index set $curr_val
    my default_button $w combobox
  };# widget_combo

  method widget_encoding {w label pattern default curr_val} {
    my variable index value
    menubutton $w.entry_$index -relief raised -indicatoron 1 -anchor c \
      -menu $w.entry_$index.menu \
      -textvariable [my varname value]($index) \
      -compound left -padx 10 -width 10 -bd [my opt BorderWidth]\
      -fg #ffff00 -bg #009900 \
      -image [::Encoding::GetEncodingImage $value($index)]
    menu $w.entry_$index.menu -tearoff 0
    GUI_AppendEncodingsToMenu $w.entry_$index.menu \
      "set [my varname value]($index) \$Encoding
       $w.entry_$index configure -image \$EncodingImage -text \$Encoding"
    my default_button $w menu
  };# widget_encoding

  method widget_entry {w label pattern default curr_val} {
    my variable index
    ttk::entry $w.entry_$index \
      -textvariable [my varname value]($index)
    my default_button $w entry
  };# widget_entry

  method widget_command {w label pattern default curr_val} {
    my variable index padx pady value
    set value($index) {}
    ttk::button $w.entry_$index -text $label {*}$default
    grid x $w.entry_$index - x -sticky snew -padx $padx -pady $pady
  };# widget_command

  method widget_boolean {w label pattern default curr_val} {
    my variable index
    ttk::checkbutton $w.entry_$index -text $label \
      -variable [my varname value]($index)
    set label {}
    my default_button $w checkbutton
  };# widget_boolean

  method widget_default {w label pattern default curr_val} {
    my variable index value
    set value($index) $curr_val
    my widget_boolean $w $label $pattern $default $curr_val
    $w.entry_$index configure -onvalue $pattern -offvalue $default \
                    -textvariable [my varname value]($index)
    $w.label_$index configure -text $label
  };# widget_default

};# class ELEP::System::ComponentOptionsEditor

package provide ELEP::System::ComponentOptionsEditor 1.0
