##############################################################################
##  ComponentRunner.tcl:                                                    ##
## --------------------------------------------------                       ##
##  This class implements facilities that can run components on             ##
##  Collections/Documents.                                                  ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::Process::ThreadPool
package require ELEP::System::Component

oo::class create ELEP::System::ComponentRunner {
  mixin ELEP::Base::Utilities
  # superclass ELEP::Process::ThreadPool

  constructor {args} {
    my variable progress
    set progress      {}
    my taskNew
    # next {*}$args
    my configure {*}$args
  };# constructor

  ##########################################################################
  ## Method: taskNew
  ##########################################################################
  method taskNew {} {
    my variable components collection documents \
                AutoSave IgnoreChanges LogTimings LogErrors

    set components    {}
    set collection    {}
    set documents     {}
    set AutoSave      0 
    set IgnoreChanges 0
    set LogTimings    0
    set LogErrors     0
  };# taskNew

  ##########################################################################
  ## Method: taskDefine
  ##########################################################################
  method taskDefine {col docs {comps {}}} {
    my taskDefineCollection $col
    my taskDefineDocuments  $docs
    foreach c $comps {
      my taskAppendComponent $c
    }
  };# taskDefine

  ##########################################################################
  ## Method: taskDefineCollection
  ##########################################################################
  method taskDefineCollection {col} {
    my variable collection
    set collection $col
  };# taskDefineCollection

  ##########################################################################
  ## Method: taskDefineDocuments
  ##########################################################################
  method taskDefineDocuments {docs} {
    my variable documents
    set documents $docs
  };# taskDefineDocuments

  ##########################################################################
  ## Method: taskAppendComponent
  ##########################################################################
  method taskAppendComponent {component} {
    my variable components
    lappend components $component
  };# taskAppendComponent

  ##########################################################################
  ## Method: checkResult
  ##########################################################################
  method checkResult {value} {
    foreach {status result} $value {break}
    switch $status {
      e {
        return -options [lindex $value 2] $result
      }
      default {}
    }
  };# checkResult

  ##########################################################################
  ## Method: taskExecuteAutoSaveDocument
  ##########################################################################
  method taskExecuteAutoSaveDocument {document} {
    my variable AutoSave
    if {$AutoSave} { tip_Sync $document }
  };# taskExecuteAutoSaveDocument

  ##########################################################################
  ## Method: taskExecuteAutoCloseDocument
  ##########################################################################
  method taskExecuteAutoCloseDocument {document} {
    my variable AutoSave IgnoreChanges
    if {$AutoSave} { tip_Sync $document }
    if {$AutoSave || $IgnoreChanges} {
      tip_Close $document
      CDM_ClearUnusedSavedValues
    }
  };# taskExecuteAutoCloseDocument

  ##########################################################################
  ## Method: taskExecute
  ##########################################################################
  method taskExecute {method {call run}} {
    my variable collection documents progress
    set count [llength $documents]
    switch -exact $count {
      1 {
        ## A single document. Process it...
        if {$progress ne ""} {$progress numberOfDocumentsToProcess 1}
        set doc [lindex $documents 0]
        my $method $doc $call
      }
      0 {
        ## All documents...
        if {$progress ne ""} {
          $progress numberOfDocumentsToProcess [tip_Length $collection]
        }
        set doc [tip_FirstDocument $collection]
        while {$doc ne ""} {
          my $method $doc $call
          set doc [tip_NextDocument $collection]
          my taskExecuteAutoCloseDocument $doc
        }
      }
      default {
        if {$progress ne ""} {$progress numberOfDocumentsToProcess $count}
        ## A list of Document Ids to process...
        foreach id $documents {
          set doc [tip_GetDocument $collection $id]
          my $method $doc $call
          my taskExecuteAutoCloseDocument $doc
        }
      }
    }
  };# taskExecute

  ##########################################################################
  ## Method: taskExecuteNoProgress
  ##########################################################################
  method taskExecuteNoProgress {document method} {
    my variable components collection
    foreach one $components {
      my checkResult [$one $method $collection $document]
      my taskExecuteAutoSaveDocument $document
    }
  };# taskExecuteNoProgress

  ##########################################################################
  ## Method: taskExecuteUpdateProgress
  ##########################################################################
  method taskExecuteUpdateProgress {document method} {
    my variable components collection progress
    $progress processingDocument $document
    foreach one $components {
      $progress processingComponent [$one component]
      my checkResult [$one $method $collection $document]
      $progress processingComponentFinished
      my taskExecuteAutoSaveDocument $document
    }
    $progress processingDocumentFinished $document
  };# taskExecuteUpdateProgress

  ##########################################################################
  ## Method: taskInit
  ##########################################################################
  method taskInit {} {
    my variable collection documents components progress
    switch -exact [llength $documents] {
      1       { set doc [lindex $documents 0] }
      0       { set doc [tip_FirstDocument $collection] }
      default { set doc [tip_GetDocument $collection [lindex $documents 0]] }
    }
    if {$progress eq ""} {
      foreach one $components {
        my checkResult [$one init $collection $doc]
      }
    } else {
      $progress numberOfDocumentsToProcess 1
      $progress processingDocument $doc
      $progress configure -document_name [my msg {Initialising Components...}]
      update
      foreach one $components {
        $progress processingComponent [$one component]
        my checkResult [$one init $collection $doc]
        $progress processingComponentFinished
      }
      $progress processingDocumentFinished $doc
    }
  };# taskInit

  ##########################################################################
  ## Method: taskRun
  ##########################################################################
  method taskRun {} {
    my variable progress
    if {$progress eq ""} {
      set method taskExecuteNoProgress
    } else {
      set method taskExecuteUpdateProgress
    }
    return [my taskExecute $method run]
  };# taskRun

  ##########################################################################
  ## Method: taskFinish
  ##########################################################################
  method taskFinish {} {
    my variable collection documents components progress
    switch -exact [llength $documents] {
      1       { set doc [lindex $documents 0] }
      0       { set doc {} }
      default { set doc {} }
    }
    if {$progress eq ""} {
      foreach one $components {
        my checkResult [$one finish $collection $doc]
      }
    } else {
      $progress numberOfDocumentsToProcess 1
      $progress processingDocument $doc
      $progress configure -document_name [my msg {Finalising Components...}]
      update
      foreach one $components {
        $progress processingComponent [$one component]
        my checkResult [$one finish $collection $doc]
        $progress processingComponentFinished
      }
      $progress processingDocumentFinished $doc
    }
  };# taskFinish

  ##########################################################################
  ## Method: taskReset
  ##########################################################################
  method taskReset {} {
    my variable progress
    if {$progress eq ""} {
      set method taskExecuteNoProgress
    } else {
      set method taskExecuteUpdateProgress
    }
    return [my taskExecute $method reset]
  };# taskReset

  ##########################################################################
  ## Method: onThreadInit
  ##########################################################################
  method onThreadInit {} {
    # puts "** onThreadInit"
    set ::auto_path [list $::auto_path]
    $init_thread_cmd
  };# onThreadInit


  ##########################################################################
  ## Method: onThreadInit
  ##########################################################################
  method onThreadInit {} {
    # puts "** onThreadInit"
    set ::auto_path [list $::auto_path]
    $init_thread_cmd
  };# onThreadInit

  ##########################################################################
  ## Method: initialiseThreads
  ##########################################################################
  method initialiseThreads {} {
    pack [my createToolbarFrame] -side top -fill x
  };# initialiseThreads

  ##########################################################################
  ## Method: createToolbarFrame
  ##########################################################################
  method createToolbarFrame {} {
    my variable toolbarFrame widget
    set toolbarFrame [ttk::frame $widget.toolbarFrame -class ToolbarFrame]
    my createToolbars
    return $toolbarFrame
  };# createToolbarFrame

  ##########################################################################
  ## Method: createToolbars
  ##########################################################################
  method createToolbars {} {
  };# createToolbars

  ##########################################################################
  ## Event: onDocumentChange
  ##########################################################################
  method onDocumentChange {col doc} {
  };# onDocumentChange

  ##########################################################################
  ## Event: onHide
  ##########################################################################
  method onHide {} {
  };# onHide

  ##########################################################################
  ## Event: onExpose
  ##########################################################################
  method onExpose {} {
    my variable toplevel_object
    $toplevel_object setStatusText 
  };# onExpose

};# class ELEP::System::ComponentRunner

package provide ELEP::System::ComponentRunner 1.0
