##############################################################################
##  SystemDeploy.tcl:                                                       ##
## --------------------------------------------------                       ##
##  This class implements an Ellogon System (Deploy).                       ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities::WidgetBase
package require ELEP::ViewerBase::DocumentSelector
package require ELEP::FlowControl::Canvas
package require ELEP::FlowControl::ComponentItem
package require ELEP::FlowControl::ConnectionItem
package require ELEP::Style::Style
package require ELEP::System::Component
package require ELEP::System::ComponentRunner
package require ELEP::System::Progress
package require widget::scrolledwindow

# namespace eval ELEP::System {
#   variable ToolbarColour #35496A
# 
#   ## Some nice gradient colours...
#   # #7ECD00 -> #46A300 (B: #3a8a00), #9ed83d -> #74b93d (B: #67a23a) (lighter)
# }
# 
# ttk::style configure System.TNotebook -background $ELEP::System::ToolbarColour
# ttk::style configure System.TPanedwindow -background $ELEP::System::ToolbarColour

oo::class create ELEP::System::SystemDeploy {
  superclass ELEP::WindowUtilities::WidgetBase

  constructor {name parent args} {
    my variable toplevel_object collection document manager selected_component \
                component_objects link_objects chain_start_objects \
                background_stops background_lineartransition style \
                component_execution_order component_runner
    set toplevel_object           $parent
    set collection                {}
    set document                  {}
    set manager                   {}
    set component_objects         {}
    set link_objects              {}
    set chain_start_objects       {}
    set selected_component        {}
    set style                     [ELEP::Style::getStyle]
    set component_execution_order {}
    set component_runner          {}

    next frame $name
    my configure {*}$args
    my createWidgets
    my createObjects
    my init
  };# constructor

  ##########################################################################
  ## Method: systemManager
  ##########################################################################
  method systemManager {{m {}}} {
    my variable manager
    catch {$manager destroy}
    set manager $m
    my reset
  };# systemManager

  ##########################################################################
  ## Method: reset
  ##########################################################################
  method reset {} {
    my variable component_objects link_objects
    foreach one $link_objects {
      $one destroy
    }
    foreach one $component_objects {
      catch {[$one cget -system_component_obj] destroy}
      $one destroy
    }
    my init
  };# reset

  ##########################################################################
  ## Method: resetComponentsState
  ##########################################################################
  method resetComponentsState {} {
    my variable chain_start_objects component_execution_order
    foreach one $chain_start_objects {
      $one setState inactive
    }
    set component_execution_order {}
  };# resetComponentsState

  ##########################################################################
  ## Method: init
  ##########################################################################
  method init {} {
    my variable manager canvas canvas_widget component_objects link_objects \
                chain_start_objects component_execution_order \
                component_options_runtime component_options_specification
    if {$manager eq ""} {return}
    set component_objects               {}
    set chain_start_objects             {}
    set component_execution_order       {}
    array unset component_options_runtime
    array unset component_options_specification
    set components [$manager cget -components]
    foreach component $components {
      set name [dict get $component component]
      ## Create a component...
      set c    [ELEP::System::Component create \
        [ELEP::Base::auto ELEP::System::CRC] $name]
      set item [ELEP::FlowControl::ComponentItem create \
        [ELEP::Base::auto ELEP::System::C] $canvas_widget \
          -parent_objects [list [self]] -system_component 1 \
          -system_component_obj $c]
      $item init $component
      $c setGraphicalObject $item
      ## Get component options...
      foreach {options spec} [GUI_GetModuleOptions $name silent] {break}
      if {$options eq "Error"} {
        lappend loading_component_errors $spec
        set component_options_runtime($name)       {}
        set component_options_specification($name) {}
      } else {
        set component_options_runtime($name)       $options
        set component_options_specification($name) $spec
      }
      set Components([dict get $component component]) $item
      lappend component_objects $item
      if {[dict get $component startschain]} {
        lappend chain_start_objects $item
      }
    }
    set link_objects {}
    set links [$manager cget -links]
    foreach link $links {
      set source [dict get $link source]
      set target [dict get $link target]
      set item [ELEP::FlowControl::ConnectionItem create \
        [ELEP::Base::auto ELEP::System::L] $canvas_widget \
        $Components($source) 0 $Components($target) 0]
      $item draw
      lappend link_objects $item
    }
    $canvas update
    my resetComponentsState
    if {[info exists loading_component_errors]} {
      gut_many_errors {Errors while opening System} \
        [my msg LoadingModulesErrors] [list $loading_component_errors]
    }
  };# init

  ##########################################################################
  ## Method: createObjects
  ##########################################################################
  method createObjects {} {
    my variable component_runner progress_widget
    if {$component_runner eq ""} {
      set component_runner [ELEP::System::ComponentRunner create \
          [ELEP::Base::auto ELEP::System::CR] -progress $progress_widget]
    }
  };# createObjects

  ##########################################################################
  ## Method: createWidgets
  ##########################################################################
  method createWidgets {} {
    my variable widget logs_widget clock_widget progress_widget \
                execution_time
    # grid [my createToolbarFrame] -sticky snew

    ## Create a vertical pane widget, for separating the system area from the
    ## log window...
    ttk::panedwindow $widget.vp -style System.TPanedwindow
    
    frame $widget.vp.fr
      ## Create a horizontal pane, for displaying info on the right side, such
      ## as component options or progress information...
      ttk::panedwindow $widget.vp.fr.hp -orient horizontal
      $widget.vp.fr.hp add [my createWidgetsCanvas $widget.vp.fr.hp] -weight 1
      set toolbox [my createWidgetsToolbox $widget.vp.fr]
      grid $widget.vp.fr.hp -sticky snew
      # grid $toolbox -row 0 -column 1 -sticky snew
      grid columnconfigure $widget.vp.fr 0 -weight 1
      grid rowconfigure    $widget.vp.fr 0 -weight 1

    $widget.vp add $widget.vp.fr -weight 1
    set logs_widget [my createWidgetsLogs $widget.vp]

    grid $widget.vp -sticky snew
    grid columnconfigure $widget 0 -weight 1
    grid rowconfigure    $widget 0 -weight 1

    ## Create a floating window for timming...
    my loadImage ICpu
    set bg #ffffaa
    set execution_time 00:00:00
    set clock_widget [frame $widget.clock -background $bg \
                           -highlightbackground black \
                           -highlightthickness 1 -highlightcolor black]
      label $clock_widget.img -relief flat -background $bg -image ICpu
      label $clock_widget.l   -relief flat -background $bg -foreground black \
                              -padx 2 -pady 0 -anchor w \
                              -textvariable [my varname execution_time]
      pack $clock_widget.img $clock_widget.l -side left -fill y
    ## Finally, create a floating progress widget...
    set progress_widget [ELEP::System::Progress create \
        [ELEP::Base::auto ELEP::System::Pr] $widget \
        -time_variable [my varname execution_time]]
  };# createWidgets

  ##########################################################################
  ## Method: createWidgetsLogs
  ##########################################################################
  method createWidgetsLogs {parent} {
    ttk::notebook $parent.nb -style System.TNotebook

    ## Output tab...
    set sw $parent.nb.out
    widget::scrolledwindow $sw
    text $sw.text -borderwidth 0 -relief flat -width 20 -height 4 -setgrid 0
    $sw setwidget $sw.text
    $parent.nb add $sw -text [my msg Output]

    ## Time Log tab...
    set sw $parent.nb.time
    widget::scrolledwindow $sw
    text $sw.text -borderwidth 0 -relief flat -width 20 -height 4 -setgrid 0
    $sw setwidget $sw.text
    $parent.nb add $sw -text [my msg {Time Log}]

    ## Error Log tab...
    set sw $parent.nb.error
    widget::scrolledwindow $sw
    text $sw.text -borderwidth 0 -relief flat -width 20 -height 4 -setgrid 0
    $sw setwidget $sw.text
    $parent.nb add $sw -text [my msg {Error Log}]

    return $parent.nb
  };# createWidgetsLogs

  ##########################################################################
  ## Method: createWidgetsToolbox
  ##########################################################################
  method createWidgetsToolbox {parent} {
    frame $parent.top -background #35496A -width 0
    # for {set i 0} {$i < 5} {incr i} {
    #   grid [ttk::button $parent.top.b$i -text $i]
    # }
    return $parent.top
  };# createWidgetsToolbox

  ##########################################################################
  ## Method: createWidgetsCanvas
  ##########################################################################
  method createWidgetsCanvas {parent} {
    my variable canvas canvas_widget style
    ## Create a Canvas...
    set canvas_type [$style cget -canvas_type]
    set canvas [ELEP::FlowControl::Canvas create \
        [ELEP::Base::auto ELEP::System] $canvas_type $parent.canvas \
        -background white -borderwidth 0 -highlightthickness 0]
    set canvas_widget [$canvas canvas]
    my background
    $canvas_widget bind TiledBackgroundRectangle <<MB_1>> \
      [list [self] onClick]
    $canvas_widget bind TiledBackgroundRectangle <<MB_2>> \
      [list [self] onClick]
    return $parent.canvas
  };# createWidgets

  ##########################################################################
  ## Method: background
  ##########################################################################
  method background {{mode default} args} {
    my variable canvas_widget canvas style
    
    set stops      [$style gradient]
    set transition [$style transition]

    switch $mode {
      image  {
        set image [lindex $args 0]
        $canvas tile $image
        return $image
      }
      default {
        set stops [$style gradient $mode]
      }
    }
    
    if {[$style cget -tkpath_available]} {
      set canvas_rect prect
      set canvas_fill [$canvas_widget gradient create linear \
         -stops $stops -lineartransition $transition]
      set result [list $stops $transition {}]
    } else {
      set canvas_rect rect
      # set canvas_fill #d6d6ff
      set canvas_fill [lindex $stops 0 1]
      set result [list {} {} $canvas_fill]
    }
    $canvas fill $canvas_fill $canvas_rect
    return $result
  };# background

  ##########################################################################
  ## Method: createToolbarFrame
  ##########################################################################
  method createToolbarFrame {} {
    my variable toolbarFrame widget
    set toolbarFrame [ttk::frame $widget.toolbarFrame -class ToolbarFrame]
    my createToolbars
    return $toolbarFrame
  };# createToolbarFrame

  ##########################################################################
  ## Method: createToolbars
  ##########################################################################
  method createToolbars {} {
    my createDocumentSelector
    return
    my variable toolbarFrame toolbar
    set toolbar [ttk::frame $toolbarFrame.tb -class Toolbar]
    $toolbar add button hello -text Hello!
    return $toolbarFrame
  };# createToolbars

  ##########################################################################
  ## Method: createDocumentSelector
  ##########################################################################
  method createDocumentSelector {} {
    my variable selector_object
    set selector_object [ELEP::ViewerBase::DocumentSelector create \
        [::ELEP::Base::auto ELEP::VisualAnnotators::] [self] [self]]
  };# createDocumentSelector

  ##########################################################################
  ## Method: getToolbarFrame
  ##########################################################################
  method getToolbarFrame {} {
    my variable toolbarFrame
    return $toolbarFrame
  };# getToolbarFrame

  ##########################################################################
  ## Method: component
  ##########################################################################
  method component {what component_object args} {
    set component [$component_object cget -component]
    switch $what {
      Options {my onComponentEditOptions $component_object}
      Folder  {GUI_ViewModuleDir $component}
      Reset   -
      Reload  {
      }
      default {error "unsupported option: \"$what\""}
    }
  };# component

  ##########################################################################
  ## Method: widget
  ##########################################################################
  method widget {what action} {
    my variable widget logs_widget
    switch $what {
      logs {
        my variable VisibleWidget
        if {![info exists VisibleWidget(Logs)]} {
          set VisibleWidget(Logs) 0
        }
        switch $action {
          show {
            if {$VisibleWidget(Logs)} {return}
            $widget.vp add $logs_widget -weight 0
            set VisibleWidget(Logs) 1
          }
          hide {
            if {!$VisibleWidget(Logs)} {return}
            $widget.vp forget $logs_widget
            set VisibleWidget(Logs) 0
          }
          toggle {
            if {$VisibleWidget(Logs)} {
              my widget logs hide
            } else {
              my widget logs show
            }
          }
          default {error "unsupported action: \"$action\""}
        }
      }
      execution {set w $$widget.vp; set c out;   set t 0}
      time      {set w $$widget.vp; set c time;  set t 1}
      error     {set w $$widget.vp; set c error; set t 2}
      options   {set w {}}
      default {error "unsupported option: \"$what\""}
    }
  };# widget

  ##########################################################################
  ## Method: filterEmptyElements
  ##########################################################################
  method filterEmptyElements {list} {
    set new {}
    foreach one $list {
      if {$one ne ""} {lappend new $one}
    }
    return $new
  };# filterEmptyElements

  ##########################################################################
  ## Event: onDocumentChange
  ##########################################################################
  method onDocumentChange {col doc} {
    my variable toplevel_object collection document chain_start_objects \
                component_execution_order component_objects
    set collection $col
    set document   $doc
    $toplevel_object updateStatusFromDocument $col $doc
    if {$collection eq ""} {
      set state disabled
    } else {
      set state enabled
    }
    foreach one $chain_start_objects {
      $one configure -col $collection -doc $document
      $one setState $state
    }
    set component_execution_order {}

    ## Check all components, to find which of them have run...
    if {$doc ne ""} {
      foreach one $component_objects {
        $one hasRun $doc
      }
    }
  };# onDocumentChange

  ##########################################################################
  ## Event: onComponentSelect
  ##########################################################################
  method onComponentSelect {component} {
    my variable selected_component toplevel_object
    if {$component eq $selected_component} {return}
    if {$selected_component ne ""} {$selected_component drawDeselected}
    set selected_component $component
    if {$selected_component ne ""} {$selected_component drawSelected}
    $toplevel_object onComponentSelect $selected_component
  };# onComponentSelect

  ##########################################################################
  ## Event: onClick
  ##########################################################################
  method onClick {} {
    my variable selected_component toplevel_object
    if {$selected_component ne ""} {$selected_component drawDeselected}
    set selected_component {}
    $toplevel_object onComponentSelect $selected_component
  };# onClick

  ##########################################################################
  ## Event: onComponentClick
  ##########################################################################
  method onComponentClick {component} {
    my variable selected_component toplevel_object component_execution_order
    my onComponentSelect $component
    ## If the state of the component is active, add the component to the list 
    ## of the objects to be executed.
    set state [$component getState]
    if {$state eq "active" && $component ni $component_execution_order} {
      lappend component_execution_order $component
    }
  };# onComponentClick

  ##########################################################################
  ## Event: onComponentRun
  ##########################################################################
  method onComponentRun {component} {
    my onClick
    my variable toplevel_object collection document \
                component_execution_order component_runner \
                component_options_runtime

    $component_runner taskNew
    ##
    ## Retrieve execution modes...
    ##
    foreach one {AutoSave IgnoreChanges LogTimings LogErrors} {
      set value [$toplevel_object cget -$one]
      $component_runner configure -$one $value
    }

    ##
    ## Declare the task in component runner...
    ##
    $component_runner taskDefine $collection $document
    ## Declare components...
    foreach one $component_execution_order {
      set sys_obj [$one cget -system_component_obj]
      set comp    [$sys_obj component]
      $sys_obj setOptions [my filterEmptyElements \
               $component_options_runtime($comp)]
      $component_runner taskAppendComponent $sys_obj
    }
    my onProgressShow
    ##
    ## Initialise Components...
    ##
    if {[catch {$component_runner taskInit} error err_options]} {
      my onProgressHide
      return -options $err_options $error
    }
    ##
    ## Run Components...
    ##
    if {[catch {$component_runner taskRun} error err_options]} {
      my onProgressHide
      return -options $err_options $error
    }
    ##
    ## Finalise Components...
    ##
    if {[catch {$component_runner taskFinish} error err_options]} {
      my onProgressHide
      return -options $err_options $error
    }
    my onProgressHide
    set component_execution_order {}
  };# onComponentRun

  ##########################################################################
  ## Event: onComponentReset
  ##########################################################################
  method onComponentReset {component} {
    my variable components_to_reset
    if {$component ni $components_to_reset} {
      lappend components_to_reset $component
    }
    return
    set sys_obj [$component cget -system_component_obj]
    set comp    [$sys_obj component]
  };# onComponentReset

  ##########################################################################
  ## Event: onComponentRightClick
  ##########################################################################
  method onComponentRightClick {component} {
    my variable component_execution_order
    my onComponentSelect $component
    ## If the state of the component is active, add the component to the list 
    ## of the objects to be executed.
    set state [$component getState]
    switch -exact $state {
      active {
        ## Set the component to the enabled state...
        $component setState enabled
        ## Remove components not in active state...
        set objs {}
        foreach one $component_execution_order {
          if {[$one getState] eq "active"} {lappend objs $one}
        }
        set component_execution_order $objs
      }
      completed {
        ## Ask the user...
        set label [$component cget -label]
        set msg "[my msg ResetModule] $label[my msg ?]"
        set ans [gut_question $msg {Reset Module ?} 1 [my getToplevel]]
        if {$ans ne "yes"} {return}
        my variable toplevel_object collection document \
                    component_runner components_to_reset
        set components_to_reset {}
        $component setState enabled
        $component reset

        $component_runner taskNew
        ## Retrieve execution modes...
        foreach one {AutoSave IgnoreChanges LogTimings LogErrors} {
          set value [$toplevel_object cget -$one]
          $component_runner configure -$one $value
        }
        ## Declare the task in component runner...
        $component_runner taskDefine $collection $document
        ## Declare components...
        foreach one $components_to_reset {
          set sys_obj [$one cget -system_component_obj]
          my declarePostConditions $sys_obj
          $component_runner taskAppendComponent $sys_obj
        }
        if {[llength $document] == 1} {
          $component_runner taskReset
        } else {
          my onProgressShow
          $component_runner taskReset
          my onProgressHide
        }
      }
    }
    return none
  };# onComponentRightClick

  ##########################################################################
  ## Method: declarePostConditions
  ##########################################################################
  method declarePostConditions {sys_obj} {
    set comp    [$sys_obj component]
    set info $::creole_config($comp)
    set post [lindex $info [expr {[lsearch $info post_conditions] +1}]]
    set post_anns [lindex $post [expr {[lsearch $post annotations] +1}]]
    set post_anns [lrange $post_anns 0 end]
    $sys_obj configure -post_annotations $post_anns \
                       -post_attributes  [list $comp]
  };# declarePostConditions

  ##########################################################################
  ## Event: onComponentShowViewers
  ##########################################################################
  method onComponentShowViewers {component} {
    my onComponentClick $component
    my variable toplevel_object collection canvas_widget
    ## Get a document...
    set doc [$toplevel_object getCurrentDocument]
    if {$doc eq ""} {return}
    ## Get the list of viewers from the component...
    set module [$component cget -component]
    if {$module eq ""} {return}
    set viewers       [PUL_plugin_module_config $module viewers]
    set viewers_count [llength $viewers]
    if {$viewers_count == 0} {return}
    ## Create a menu...
    set menu $canvas_widget.viewers_menu
    catch {destroy $menu}
    menu $menu -tearoff false
    $menu delete 0 end
    global creole_${module}_home
    for {set i 0} {$i < $viewers_count} {incr i 3} {
      set what [lindex $viewers $i]
      regsub -all {%D%} $what [tip_GetId $doc] what
      ## As the entry may contain variable references, ensure that the
      ## proper substitutions occur...
      catch {set what [uplevel #0 subst [list $what]]}
      set viewer [lindex $viewers [expr {$i+1}]]
      switch -- $viewer {
        single_span            {set image [my loadImage ITShowTags]}
        raw                    {set image [my loadImage ITViewAnns]}
        text_file              -
        document               {set image [my loadImage ITViewText]}
        AnnotationExplorer     {set image [my loadImage ITExplore]}
        HTMLAnnotationExplorer {set image [my loadImage ITExplore]}
        Attributes             {set image [my loadImage ITViewDocAtt]}
        parse                  {set image [my loadImage ITViewText]}
        morphology             {set image [my loadImage ITShowTags]}
        VideoFile              {set image [my loadImage ITViewText]}
        default                {
          set image [CDM_GetRegisteredViewerImage \
                        [CDM_GetRegisteredViewerLabel ggi_view_$viewer]]
        }
      }
      set text  [lindex $viewers [expr {$i+2}]]
      set title "\"Viewer: $viewer - Document: [tip_GetId $doc], From\
          Collection: [file tail [tip_GetName $collection]]\""
      if {[llength $what] == 2} {
        $menu add command -label $text -command \
              [list ggi_view_$viewer $doc\
                 [lindex $what 0] [lindex $what 1] $title] \
                 -image $image -compound [my opt MenuImageLocation]
      } else {
        $menu add command -label $text -command \
              [list ggi_view_$viewer $doc\
                 [lindex $what 0] {} $title] \
                 -image $image -compound [my opt MenuImageLocation]
      }
    }
    ## We are ready to show the menu!
    if {[string equal [$menu index end] none]} {
      destroy $menu
    } else {
      if {[my opt MenuImageLocation] == "none"} {
        ::GUI::RemoveAllImagesFromMenu $menu
      }
      tk_popup $menu [winfo pointerx .] [winfo pointery .]
    }
    return show
  };# onComponentShowViewers

  ##########################################################################
  ## Event: onComponentEditOptions
  ##########################################################################
  method onComponentEditOptions {component} {
    package require ELEP::System::ComponentOptionsEditor
    my variable widget component_options_runtime component_options_specification
    ## Create an editor object...
    set name [$component cget -component]
    set editor [::ELEP::WindowUtilities::UniqueWindowName $widget.ceo_$name]
    foreach {o s} [GUI_GetModuleOptions $name] {break}

    ## Set the current option values...
    if {[info exists component_options_runtime($name)]} {
      ## Ensure that the spec has not changed!
      if {$s ne $component_options_specification($name)} {
        ## Warn the user...
        gut_warning [my msg ComponentOptionSpecificationsChanged.Resetting] \
                    $widget
        set component_options_runtime($name)       $o
        set component_options_specification($name) $s
      }
    } else {
      ## It is strange to not have them, but anyway...
      set component_options_runtime($name)       $o
      set component_options_specification($name) $s
    }
    set editor [ELEP::System::ComponentOptionsEditor create $editor \
        -name $name -applycmd [list [self] onComponentOptionsChange $component]\
        -options $component_options_runtime($name) \
        -spec    $component_options_specification($name)]
    $editor show
  };# onComponentEditOptions

  ##########################################################################
  ## Event: onComponentOptionsChange
  ##########################################################################
  method onComponentOptionsChange {component editor} {
    my variable component_options_runtime component_options_specification
    set name                                   [$editor cget -name]
    set component_options_runtime($name)       [$editor cget -options]
    set component_options_specification($name) [$editor cget -spec]
  };# onComponentOptionsChange

  ##########################################################################
  ## Event: onHide
  ##########################################################################
  method onHide {} {
  };# onHide

  ##########################################################################
  ## Event: onExpose
  ##########################################################################
  method onExpose {} {
    my variable toplevel_object collection document
    $toplevel_object updateStatusFromDocument $collection $document
  };# onExpose

  ##########################################################################
  ## Event: onProgressShow
  ##########################################################################
  method onProgressShow {} {
    my onClockShow
    my variable progress_widget
    $progress_widget placeInToplevelCenter
  };# onProgressShow

  ##########################################################################
  ## Event: onProgressHide
  ##########################################################################
  method onProgressHide {} {
    my variable progress_widget clock_hide_timer
    $progress_widget hide
    set clock_hide_timer [after 120000 [list catch [list [self] onClockHide]]]
  };# onProgressHide

  ##########################################################################
  ## Event: onClockShow
  ##########################################################################
  method onClockShow {} {
    my variable progress_widget clock_widget canvas_widget clock_hide_timer
    catch {after cancel $clock_hide_timer}
    $progress_widget timerStart
    update
    place $clock_widget -in $canvas_widget \
      -x 2 -rely 1.0 -y -[expr {[image height ICpu] + 6}]
  };# onClockShow

  ##########################################################################
  ## Event: onClockHide
  ##########################################################################
  method onClockHide {} {
    my variable clock_widget
    place forget $clock_widget
  };# onClockHide

  destructor {
    my variable component_runner component_objects link_objects clock_hide_timer
    catch {after cancel $clock_hide_timer}
    catch {$component_runner destroy}
    foreach one $link_objects {
      catch {$one destroy}
    }
    foreach one $component_objects {
      catch {[$one cget -system_component_obj] destroy}
      catch {$one destroy}
    }
    next
  };# destructor

};# class ELEP::System::SystemDeploy

package provide ELEP::System::SystemDeploy 1.0
