##############################################################################
##  SystemManager.tcl:                                                      ##
## --------------------------------------------------                       ##
##  This class implements a class for managing an Ellogon System.           ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

oo::class create ELEP::System::SystemManager {
  mixin ELEP::Base::Utilities

  constructor {args} {
    my reset
    my configure {*}$args
  };# constructor

  ##########################################################################
  ## Method: reset
  ##########################################################################
  method reset {} {
    my variable legacy_system name geometry components links targets \
                chain_starts
    set name          [my msg None]
    set legacy_system {}
    set components    {}
    set links         {}
    set targets       {}
    set chain_starts  {}
  };# reset

  ##########################################################################
  ## Method: initFromEllogonMemoryLegacy
  ##########################################################################
  method initFromEllogonMemoryLegacy {system} {
    my variable legacy_system name geometry components links targets \
                chain_starts
    set legacy_system $system
    set info $::GUI_Graphs($legacy_system)
    foreach {name geometry _ _ cs ls} $info {break}
    ##
    ## Add Connectors...
    ##
    set links   {}
    set targets {}
    foreach one $ls {
      foreach {s t points} $one {break}
      set link [dict create]
      dict set link target $t
      dict set link source $s
      dict set link points $points
      my initLinkOptions link
      lappend links $link
      lappend targets $t
      set Target($t) 0
      set Source($s) 0
    }
    ##
    ## Add Components...
    ##
    set dx [my opt ModuleWidth]; set dy [my opt ModuleHeight]
    set dx [expr {$dx/2+10}];    set dy [expr {$dy/2+10}]
    set components {}
    foreach one $cs {
      foreach {c x y label} $one {break}
      incr x $dx; incr y $dy
      if {[catch {PUL_plugin_module_config $c type} type]} {
        set type module
      }
      ## Check for a module Description...
      if {[catch {PUL_plugin_module_config $m description} ToolTip]} {
        set ToolTip {}
      }
      ## Append the Location of the module to the Tooltip...
      if {[info exists ::creole_${c}_home]} {
        set ToolTip "[my msg ModuleName]: $c\n\
                   \n[my msg ModuleLocation]: [set ::creole_${c}_home]\n\
                   \n$ToolTip"
      }
      set component [dict create]
      dict set component label       [string map {\\n \n} $label]
      dict set component component   $c
      dict set component type        $type
      dict set component x           $x
      dict set component y           $y
      dict set component tooltip     $ToolTip
      my initComponentOptions component
      ## Is this component a source?
      if {[info exists Source($c)]} {
        dict set component endschain 0
      } else {
        dict set component endschain 1
      }
      ## Is this component a target?
      if {[info exists Target($c)]} {
        dict set component startschain 0
        lappend targets $component
      } else {
        dict set component startschain 1
        lappend chain_starts $component
      }
      
      lappend components $component
    }
  };# initFromEllogonMemoryLegacy

  ##########################################################################
  ## Method: initComponentOptions
  ##########################################################################
  method initComponentOptions {var_name} {
    upvar $var_name component
    foreach {option token} {shape         ModuleShape
                            width         ModuleWidth
                            height        ModuleHeight
                            borderwidth   ModuleBorderWidth
                            bordercolour  ModuleBorderColor
                            textwidth     ModuleTextWidth
                            textpadx      ModuleTextPadX
                            textpady      ModuleTextPadY
                            textjustify   ModuleTextJustify
                            textanchor    ModuleTextAnchor
                            textfont      ModuleTextFont
                            imageanchor   ModuleImageAnchor
                            imagepadx     ModuleImagePadX
                            imagepady     ModuleImagePadY
                            fillinactive  ModuleInactive
                            fillenabled   ModuleEnabled
                            fillactive    ModuleActive
                            fillrun       ModuleRun
                            textinactive  ModuleTextInactive
                            textenabled   ModuleTextEnabled
                            textactive    ModuleTextActive
                            textrun       ModuleTextRun
                            imageinactive IModuleInactive
                            imageenabled  IModuleEnabled
                            imageactive   IModuleActive
                            imagerun      IModuleRun
                            } {
      dict set component $option [my opt $token]
    }
  };# initComponentOptions

  ##########################################################################
  ## Method: initLinkOptions
  ##########################################################################
  method initLinkOptions {var_name} {
    upvar $var_name link
    foreach {option token} {width         ArcWidth
                            widthactive   ArcActiveWidth
                            fill          ArcColor
                            fillactive    ArcActiveColor
                           } {
      dict set link $option [my opt $token]
    }
  };# initLinkOptions

  ##########################################################################
  ## Method: identifyChainStarts
  ##########################################################################
  method identifyChainStarts {} {
    my variable components targets chain_starts
    foreach component 
  };# identifyChainStarts

  ##########################################################################
  ## Method: load
  ##########################################################################
  method load {folder} {
  };# load

  ##########################################################################
  ## Method: getSystemNameTails
  ##########################################################################
  method getSystemNameTails {} {
    my getSystemNames
    my variable collections collectionTails
    set collectionTails {}
    foreach col $collections {
      lappend collectionTails [file tail $col]
    }
    return $collectionTails
  };# getSystemNameTails

  ##########################################################################
  ## Method: getDocumentNames
  ##########################################################################
  method getDocumentNames {col} {
    my variable documents allowSystemOpen
    if {$allowSystemOpen} {
      set documents [list [::ELEP::WindowUtilities::msg All]]
    } else {
      set documents {}
    }
    set documents [concat $documents [gut_GetSystemDocuments $col]]
    return $documents
  };# getDocumentNames

  ##########################################################################
  ## Method: collectionAtIndex
  ##########################################################################
  method collectionAtIndex {index} {
    my variable collections
    return [lindex $collections $index]
  };# collectionAtIndex

  ##########################################################################
  ## Method: documentAtIndex
  ##########################################################################
  method documentAtIndex {index} {
    my variable documents
    return [lindex $documents $index]
  };# documentAtIndex

  ##########################################################################
  ## Method: selectSystem
  ##########################################################################
  method selectSystem {index} {
    my variable selectedSystem documents
    set documents {}
    set selectedSystem $index
    my selectDocument 0
  };# selectSystem

  ##########################################################################
  ## Method: getSelectedSystemIndex
  ##########################################################################
  method getSelectedSystemIndex {} {
    my variable selectedSystem
    return $selectedSystem
  };# getSelectedSystem

  ##########################################################################
  ## Method: getSelectedSystem
  ##########################################################################
  method getSelectedSystem {} {
    my variable selectedSystem collections
    if {$selectedSystem < 0} {return {}}
    return [lindex $collections $selectedSystem]
  };# getSelectedSystem

  ##########################################################################
  ## Method: getDocumentNamesOfSelectedSystem
  ##########################################################################
  method getDocumentNamesOfSelectedSystem {} {
    set col [my getSelectedSystem]
    if {$col eq ""} {return {}}
    return [my getDocumentNames $col]
  };# getDocumentNamesOfSelectedSystem

  ##########################################################################
  ## Method: selectDocument
  ##########################################################################
  method selectDocument {index} {
    my variable selectedDocument
    set selectedDocument $index
  };# selectDocument

  ##########################################################################
  ## Method: getSelectedDocumentIndex
  ##########################################################################
  method getSelectedDocumentIndex {} {
    my variable selectedDocument
    return $selectedDocument
  };# getSelectedDocument

  ##########################################################################
  ## Method: getSelectedDocument
  ##########################################################################
  method getSelectedDocument {} {
    my variable selectedDocument documents
    if {$selectedDocument < 1} {return {}}
    return [lindex $documents $selectedDocument]
  };# getSelectedDocument

  classmethod msg    {code}  {return [::ELEP::WindowUtilities::msg    $code]}
  classmethod msgVar {code}  {return [::ELEP::WindowUtilities::msgVar $code]}
  classmethod opt    {code}  {return [::ELEP::WindowUtilities::opt    $code]}
  classmethod optVar {code}  {return [::ELEP::WindowUtilities::optVar $code]}
};# class ELEP::System::SystemManager

package provide ELEP::System::SystemManager 1.0
