##############################################################################
##  1ClickSelector.tcl:                                                     ##
## --------------------------------------------------                       ##
##  This class implements a facility that modifiles the selection of display##
##  widgets, according to user defined regular expressions.                 ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::VisualAnnotators::VisualRegexp
package require ELEP::WindowUtilities::WidgetBase

oo::class create ELEP::ViewerBase::1ClickSelector {
  superclass ELEP::WindowUtilities::WidgetBase

  variable {*}[info class variables ELEP::WindowUtilities::WidgetBase] \
           window_obj \
           display_object \
           mouse_event \
           original_mouse_event_binding \
           regular_expression \
           pattern_list \
           edited_pattern \
           character_offset_halo \
           visual_regexp_object

  constructor {parent_win {disp_obj {}}} {
    set window_obj     {};# The object that holds our toplevel window.
    set display_object {};# The object that shows annotated information.
    set mouse_event    [event info <<MB_M>>];# The mouse event the selector
                                             # will respond.
    set original_mouse_event_binding {}
    set regular_expression           {}
    set pattern_list                 {{\w+} {(\d+:)?\d+\.\d+}}
    set edited_pattern               {}
    set character_offset_halo        100
    set visual_regexp_object         {}

    next ttk::frame [::ELEP::WindowUtilities::UniqueWindowName \
                           $parent_win.oneClickSelector]

    set display_object $disp_obj
    set client_area $widget
    pack [my createMouseButtonSelectionWidget] -fill x
    pack [my createPatternEditorWidget] -fill both -expand 1
    set visual_regexp_object [ELEP::VisualAnnotators::VisualRegexp new]
    my onMouseEventChange
    my updateRegularExpression
  };# constructor

  method createMouseButtonSelectionWidget {} {
    ttk::labelframe $client_area.mb -text { Mouse Buttons: }
      foreach {button vev} {1 <<MB_1>> 2 <<MB_M>> 3 <<MB_2>>} {
        set w {}
        foreach specifier {{} Control Shift Alt} {
          set event "<[string trim $specifier-$button -]>"
          set ev    "<[string trim $specifier-[string trim \
                              [event info $vev] <>] -]>"
          lappend w [ttk::radiobutton $client_area.mb.b$button$specifier \
            -text $event -value $ev -variable [my varname mouse_event] \
            -command [list [self] onMouseEventChange]]
        }
        eval grid $w -padx 4 -pady 2 -sticky snew
      }
      return $client_area.mb
  };# createMouseButtonSelectionWidget

  method createPatternEditorWidget {} {
    ttk::labelframe $client_area.pe -text { Pattern Editor: }
    ttk::button $client_area.pe.add -textvariable GUI_Text(Add) \
      -command [list [self] onPatternAdd]
    ttk::button $client_area.pe.delete -textvariable GUI_Text(Delete) \
      -command [list [self] onPatternDelete]
    ttk::button $client_area.pe.edit -textvariable GUI_Text(Edit) \
      -command [list [self] onPatternEdit]
    ttk::button $client_area.pe.save -textvariable GUI_Text(Save) \
      -command [list [self] onPatternsSave]
    ttk::button $client_area.pe.load -textvariable GUI_Text(Load) \
      -command [list [self] onPatternsLoad]
    ScrolledWindow $client_area.pe.sw
    listbox $client_area.pe.sw.listbox -listvariable [my varname pattern_list]
    $client_area.pe.sw setwidget $client_area.pe.sw.listbox

    grid $client_area.pe.sw     -rowspan 6 -padx 2 -pady 2 -sticky snew
    grid $client_area.pe.add    -column 1 -row 0 -padx 2 -pady 2 -sticky snew
    grid $client_area.pe.edit   -column 1 -row 1 -padx 2 -pady 2 -sticky snew
    grid $client_area.pe.delete -column 1 -row 2 -padx 2 -pady 2 -sticky new
    grid $client_area.pe.save   -column 1 -row 4 -sticky snew -padx 2 -pady 2
    grid $client_area.pe.load   -column 1 -row 5 -sticky snew -padx 2 -pady 2
    grid columnconfigure $client_area.pe 0 -weight 1
    grid rowconfigure    $client_area.pe 3 -weight 1
    return $client_area.pe
  };# createPatternEditorWidget

  #############################################################################
  ## Event: onMouseEventChange
  #############################################################################
  method onMouseEventChange {} {
    if {![string length $display_object]} {return}
    set ws [$display_object getWidgetForManagingBindings]
    foreach {event cmd} $original_mouse_event_binding {
      foreach w $ws {
        bind $w $event $cmd
      }
    }
    foreach w $ws {
      set original_mouse_event_binding \
        [list $mouse_event [bind $w $mouse_event]]
      bind $w $mouse_event [list +[self] onMouseClick %x %y %W]
    }
  };# onMouseEventChange

  #############################################################################
  ## Event: onMouseClick
  #############################################################################
  method onMouseClick {x y w} {
    if {![string length $regular_expression]} {return}
    if {![string length $display_object]} {return}
    set index [$display_object call getIndexAtxy $x $y]
    $display_object call mouseEventForWindow $w
    if {![string length $index]} {return}
    set offsetWRTDocument [$display_object call offsetWRTDocument]
    ## Get the text arround the clicked index:
    set offset [$display_object call index2CharacterOffset $index]
    set offset [expr {$offset + $offsetWRTDocument}]
    set start [expr {$offset-$character_offset_halo}]
    set end [expr {$offset+$character_offset_halo}]
    if {$start < 0} {set start 0}
    ## We should retrieve the text that corresponds to these offsets...
    set doc [$display_object call getDocument]
    if {![string length $doc]} {return}
    set text [string range [tip_GetByteSequence $doc] 0 $end]
    ## Apply the regural expression pattern, and find all matches:
    foreach index [regexp -all -indices -inline -start $start -- \
                          $regular_expression $text] {
      foreach {s e} $index {break}
      if {$s <= $offset && $offset <= $e} {set found 1; break}
    }
    if {![info exists found]} {
      ## The user has not clicked on a word. Clear selection...
      $display_object call selectionClear
      return
    }
    incr e

    set s [expr {$s - $offsetWRTDocument}]
    set e [expr {$e - $offsetWRTDocument}]

    ## [$s, $e] is the range that best matches the clicked offset.
    ## Retrieve the actual selection, and blend its offsets with the found
    ## ones...
    if {[$display_object call selectionExists]} {
      set start [$display_object call index2CharacterOffset \
                         [$display_object call getSelectionStartIndex]]
      set end   [$display_object call index2CharacterOffset \
                         [$display_object call getSelectionEndIndex]]
      if {$start < $s} {set s $start}
      if {$end   > $e} {set e $end}
    }

    $display_object call setSelectionIndexes \
      [$display_object call characterOffset2Index $s] \
      [$display_object call characterOffset2Index $e]
    return -code break
  };# onMouseClick

  #############################################################################
  ## Event: onPatternAdd
  #############################################################################
  method onPatternAdd {} {
    if {![string length $visual_regexp_object]} {return}
    $visual_regexp_object setPattern {}
    set edited_pattern [$visual_regexp_object show]
    if {![string length $edited_pattern]} {return}
    $client_area.pe.sw.listbox insert end $edited_pattern
    my updateRegularExpression
  };# onPatternAdd

  #############################################################################
  ## Event: onPatternDelete
  #############################################################################
  method onPatternDelete {} {
    set sel [$client_area.pe.sw.listbox curselection]
    if {![llength $sel]} {return}
    $client_area.pe.sw.listbox delete [lindex $sel 0]
    my updateRegularExpression
  };# onPatternDelete

  #############################################################################
  ## Event: onPatternEdit
  #############################################################################
  method onPatternEdit {} {
    if {![string length $visual_regexp_object]} {return}
    set sel [$client_area.pe.sw.listbox curselection]
    if {![llength $sel]} {return}
    set edited_pattern [$client_area.pe.sw.listbox get [lindex $sel 0]]
    $visual_regexp_object setPattern $edited_pattern
    set edited_pattern [$visual_regexp_object show]
    if {![string length $edited_pattern]} {return}
    $client_area.pe.sw.listbox delete [lindex $sel 0]
    $client_area.pe.sw.listbox insert [lindex $sel 0] $edited_pattern
    my updateRegularExpression
  };# onPatternEdit

  #############################################################################
  ## Event: onPatternsSave
  #############################################################################
  method onPatternsSave {} {
    set filename [GUI_SelectSaveFile {} -parent [my getToplevel]]
    if {![string length $filename]} {return}
    set file [open $filename w]
    fconfigure $file -encoding utf-8
    puts $file $pattern_list
    close $file
  };# onPatternsSave

  #############################################################################
  ## Event: onPatternsLoad
  #############################################################################
  method onPatternsLoad {} {
    set filename [GUI_SelectFile {} -parent [my getToplevel]]
    if {![string length $filename]} {return}
    set file [open $filename]
    fconfigure $file -encoding utf-8
    set pattern_list [read $file]
    close $file
  };# onPatternsLoad

  method updateRegularExpression {} {
    set regular_expression "([join $pattern_list )|(])"
  };# updateRegularExpression

  destructor {
    if {[string length $visual_regexp_object]} {
      $visual_regexp_object destroy
    }
  };# destructor

};# class ELEP::ViewerBase::1ClickSelector

package provide ELEP::ViewerBase::1ClickSelector 1.0
