##############################################################################
##  AnnotationHighlighter.tcl:                                              ##
## --------------------------------------------------                       ##
##  This class implements a utility that highlights text segments annotated ##
##  with specific Annotations.                                              ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities::WidgetBase
package require widget::scrolledwindow

oo::class create ELEP::ViewerBase::AnnotationHighlighter {
  superclass ELEP::WindowUtilities::WidgetBase

  variable {*}[info class variables ELEP::WindowUtilities::WidgetBase] \
           window_obj \
           display_object \
           annotation_types_tree_widget \
           document \
           highlight_colour_bg_widget \
           highlight_colour_fg_widget \
           highlight_apply_widget \
           SelectedAnnotationsInfo \
           _

  constructor {parent_win {disp_obj {}}} {
    next ttk::frame [::ELEP::WindowUtilities::UniqueWindowName \
                             $parent_win.annotationHighlighter]
    set window_obj     {};# The object that holds our toplevel window.
    set display_object {};# The object that shows annotated information.
    set document       {}
    set _           \u005;# field separator

    set display_object $disp_obj
    set client_area $widget
    pack [my createAnnotationTypesTree] -fill both -expand 1
    pack [my createHighlightOptions] -fill x -expand 0 -side bottom
  };# constructor

  method createAnnotationTypesTree {} {
    widget::scrolledwindow $client_area.sw -scrollbar vertical -auto vertical
    set annotation_types_tree_widget $client_area.sw.tv
    ttk::treeview $annotation_types_tree_widget -height 6 \
      -columns {values}
    $annotation_types_tree_widget heading #0 \
      -text {Annotation Types/Attributes}
    $annotation_types_tree_widget heading values \
      -text {Values}
    # $annotation_types_tree_widget heading shown \
    #  -text {Highlighted}
    $annotation_types_tree_widget column  #0     -width 200
    $annotation_types_tree_widget column  values -width 80
    # $annotation_types_tree_widget column  shown  -width 40
    $client_area.sw setwidget $annotation_types_tree_widget
    bind $annotation_types_tree_widget <<TreeviewSelect>> \
      [list [self] onAnnotationTypesTreeSelectionChange]
    return $client_area.sw
  };# createAnnotationTypesTree

  method createHighlightOptions {} {
    set f $client_area.options
    ttk::frame $f
      ttk::label $f.colour_l -text Colour:
      button $f.colour_bg -text {Background Colour} -background #adff2f \
        -command [list [self] onColourChange $f.colour_bg]
      set highlight_colour_bg_widget $f.colour_bg
      button $f.colour_fg -text {Foreground Colour} -background #008080 \
        -foreground white -command [list [self] onColourChange $f.colour_fg]
      set highlight_colour_fg_widget $f.colour_fg
      ttk::button $f.apply  -text Apply  -command [list [self] onHighlightApply]
      ttk::button $f.remove -text Remove -command \
                                         [list [self] onHighlightRemove]
      grid $f.colour_l $f.colour_bg $f.colour_fg $f.apply $f.remove \
        -sticky ew -padx 2
      grid columnconfigure $f 1 -weight 1
      grid columnconfigure $f 2 -weight 1
    return $f
  };# createHighlightOptions

  #############################################################################
  ## Event: onAnnotationTypesTreeSelectionChange
  #############################################################################
  method onAnnotationTypesTreeSelectionChange {} {
  };# onAnnotationTypesTreeSelectionChange

  #############################################################################
  ## Event: onColourChange
  #############################################################################
  method onColourChange {w} {
    set colour [SelectColor::menu $w.colour \
      [list below $w] \
      -color [$w cget -background]]
    if {[string length $colour]} {
      $w configure -background $colour
    }
  };# onColourChange

  method id2Annotations {id} {
    return [my selectAnnotations {*}[my id2AnnotationInfo $id]]
  };# id2Annotations

  method id2AnnotationInfo {id} {
    ## Extract the annotation type, attribute name & value, if available...
    foreach {type attr val} {{} {} {}} {break}
    foreach {type t_id attr a_id val v_id} [split $id $_] {break}
    return [list $type $attr $val]
  };# id2AnnotationInfo

  method id2AnnotationInfoPattern {id} {
    ## Extract the annotation type, attribute name & value, if available...
    foreach {type attr val} {* * *} {break}
    foreach {type t_id attr a_id val v_id} [split $id $_] {break}
    foreach one {type attr val} {
      if {![string length [set $one]]} {set $one *}
    }
    return [list $type $attr $val]
  };# id2AnnotationInfoPattern

  method selectAnnotations {type attr val} {
    if {[string length $val]} {
      ## The user selected type, attr & value...
      set Annotations [tip_SelectAnnotations $document $type \
        "\[string equal ann::$attr $val\]"]
    } elseif {[string length $attr]} {
      ## The user selected type & attr...
      set Annotations [tip_SelectAnnotations $document $type \
        "\[string length ann::$attr \]"]
    } elseif {[string length $type]} {
      ## The user selected only type...
      set Annotations [tip_SelectAnnotations $document $type]
    } else {
      return {}
    }
    return $Annotations
  };# selectAnnotations

  #############################################################################
  ## Event: onHighlightApply
  #############################################################################
  method onHighlightApply {} {
    set bg [$highlight_colour_bg_widget cget -background]
    set fg [$highlight_colour_fg_widget cget -background]
    foreach one [$annotation_types_tree_widget selection] {
      set SelectedAnnotationsInfo([my id2AnnotationInfo $one]) [list $bg $fg]
      my highlight $one $bg $fg
    }
  };# onHighlightApply

  method highlight {one bg fg} {
    set offsetWRTDocument [$display_object call offsetWRTDocument]
    my setSubtreeColours $one $bg $fg
    foreach ann [my id2Annotations $one] {
      set id [tip_GetId $ann]
      foreach span [tip_GetSpans $ann] {
        tip_GetSpanOffsets $span start end
        set start [expr {$start - $offsetWRTDocument}]
        set end   [expr {$end   - $offsetWRTDocument}]
        $display_object call addInactiveTag \
          [$display_object call characterOffset2Index $start] \
          [$display_object call characterOffset2Index $end] \
          $id:$start hl:$end $bg $fg
      }
    }
  };# highlight

  #############################################################################
  ## Event: onHighlightRemove
  #############################################################################
  method onHighlightRemove {} {
    set offsetWRTDocument [$display_object call offsetWRTDocument]
    foreach one [$annotation_types_tree_widget selection] {
      my setSubtreeColours $one
      array unset SelectedAnnotationsInfo [my id2AnnotationInfoPattern $one]
      foreach ann [my id2Annotations $one] {
        set id [tip_GetId $ann]
        foreach span [tip_GetSpans $ann] {
          tip_GetSpanOffsets $span start end
          set start [expr {$start - $offsetWRTDocument}]
          set end   [expr {$end   - $offsetWRTDocument}]
          $display_object call deleteInactiveTag $id:$start hl:$end
        }
      }
    }
  };# onHighlightRemove

  method setSubtreeColours {parent {bg {}} {fg {}}} {
    $annotation_types_tree_widget tag configure $parent \
        -background $bg -foreground $fg
    foreach one [$annotation_types_tree_widget children $parent] {
      my setSubtreeColours $one $bg $fg
    }
  };# setSubtreeColours

  ##########################################################################
  ## Event: onDocumentChange
  ##########################################################################
  method onDocumentChange {col doc} {
    set document $doc
    my onUpdateAnnotationTypes
  };# onDocumentChange

  #############################################################################
  ## Event: onUpdateAnnotationTypes
  #############################################################################
  method onUpdateAnnotationTypes {} {
    my updateAnnotationTypesTreeView
  };# onUpdateAnnotationTypes

  method clearAnnotationTypesTreeView {} {
    $annotation_types_tree_widget delete \
      [$annotation_types_tree_widget children {}]
  };# clearAnnotationView

  method updateAnnotationTypesTreeView {} {
    my clearAnnotationTypesTreeView
    set t_id 0
    set a_id 0
    set v_id  0
    foreach type [lsort [$document GetAnnotationTypes]] {
      $annotation_types_tree_widget insert {} end -id \
        $type$_$t_id -tags \
        $type$_$t_id -text $type -open 0
      foreach attribute [lsort [$document GetAnnotationAttributeNames $type]] {
        $annotation_types_tree_widget insert \
          $type$_$t_id end -id \
          $type$_$t_id$_$attribute$_$a_id -tags \
          $type$_$t_id$_$attribute$_$a_id -text $attribute -open 0
        foreach value [lsort [$document GetAnnotationAttributeValues \
                                        $type $attribute]] {
          $annotation_types_tree_widget insert \
            $type$_$t_id$_$attribute$_$a_id end -id \
            $type$_$t_id$_$attribute$_$a_id$_$value$_$v_id -tags \
            $type$_$t_id$_$attribute$_$a_id$_$value$_$v_id -text {} \
            -open 0 -values [list $value]
          my restoreHighlightedState \
            $type$_$t_id$_$attribute$_$a_id$_$value$_$v_id \
            $type $attribute $value
          incr v_id
        }
        my restoreHighlightedState $type$_$t_id$_$attribute$_$a_id \
                                   $type $attribute
        incr a_id
      }
      my restoreHighlightedState $type$_$t_id $type
      incr t_id
    }
  };# updateAnnotationTypesTreeView

  method restoreHighlightedState {one type {attr {}} {val {}}} {
    set key [list $type $attr $val]
    if {[info exists SelectedAnnotationsInfo($key)]} {
      foreach {bg fg} $SelectedAnnotationsInfo($key) {break}
      after idle [list [self] highlight $one $bg $fg]
    }
  };# restoreHighlightedState

  destructor {
  };# destructor

};# class ELEP::ViewerBase::AnnotationHighlighter

package provide ELEP::ViewerBase::AnnotationHighlighter 1.0
