##############################################################################
##  AnnotationSelector.tcl:                                                 ##
## --------------------------------------------------                       ##
##  This class implements a facility that modifiles the selection of display##
##  widgets, to existing annotations selected by the user.                  ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities::WidgetBase
package require widget::scrolledwindow

oo::class create ELEP::ViewerBase::AnnotationSelector {
  superclass ELEP::WindowUtilities::WidgetBase

  variable {*}[info class variables ELEP::WindowUtilities::WidgetBase] \
           window_obj \
           display_object \
           annotation_type \
           enable_tree_view \
           make_selected_annotation_visible \
           annotation_type_widget \
           annotation_view \
           document \
           attribute_names

  constructor {parent_win {disp_obj {}}} {
    set window_obj     {};# The object that holds our toplevel window.
    set display_object {};# The object that shows annotated information.
    set annotation_type {<Please select an Annotation Type>}
    set enable_tree_view 1
    set make_selected_annotation_visible 1
    set annotation_type_widget {}
    set annotation_view {}
    set document        {}
    set attribute_names {}
    
    next ttk::frame [::ELEP::WindowUtilities::UniqueWindowName \
                           $parent_win.annotationSelector]

    set display_object $disp_obj
    set client_area $widget
    grid [my createAnnotationTypeSelectionWidget] \
         [my createListTypeSelectionWidget] -padx 2 -pady 2 -sticky snew
    grid [my createAnnotationListWidget] \
         -columnspan 2 -padx 2 -pady 2 -sticky snew
    grid [my createAdditionalOptionsWidget] -columnspan 2 \
         -padx 2 -pady 2 -sticky snew
    grid columnconfigure $client_area 0 -weight 1
    grid rowconfigure    $client_area 1 -weight 1
  };# constructor

  method createAnnotationTypeSelectionWidget {} {
    set annotation_type_widget [ttk::combobox $client_area.annotationType \
      -state readonly -textvariable [my varname annotation_type]]
    bind $annotation_type_widget <<ComboboxSelected>> \
      [list [self] onAnnotationTypeChange]
    return $annotation_type_widget
  };# createAnnotationTypeSelectionWidget

  method createListTypeSelectionWidget {} {
    return [ttk::checkbutton $client_area.listType \
                -text {Tree view} -variable [my varname enable_tree_view]]
  };# createListTypeSelectionWidget

  method createAnnotationListWidget {} {
    widget::scrolledwindow $client_area.sw -scrollbar vertical -auto vertical
    set annotation_view $client_area.sw.tv
    ttk::treeview $annotation_view -height 6
    $annotation_view heading #0 -text ID
    $annotation_view column  #0 -width 50
    $client_area.sw setwidget $annotation_view
    bind $annotation_view <<TreeviewSelect>> \
      [list [self] onAnnotationChange]
    return $client_area.sw
  };# createAnnotationListWidget

  method createAdditionalOptionsWidget {} {
    return [ttk::checkbutton $client_area.autoSeeIndex \
                -text {Ensure selected Annotation is visible} \
                -variable [my varname make_selected_annotation_visible]]
  };# createAdditionalOptionsWidget

  ##########################################################################
  ## Event: onDocumentChange
  ##########################################################################
  method onDocumentChange {col doc} {
    set document $doc
    my onUpdateAnnotationTypes
    my onAnnotationTypeChange
  };# onDocumentChange

  #############################################################################
  ## Event: onUpdateAnnotationTypes
  #############################################################################
  method onUpdateAnnotationTypes {} {
    if {![string length $document]} {
      set annotation_type {<Please select an Annotation Type>}
      $annotation_type_widget configure -values {}
      return
    }
    $annotation_type_widget configure -values \
      [lsort [$document GetAnnotationTypes]]
  };# onUpdateAnnotationTypes

  #############################################################################
  ## Event: onAnnotationTypeChange
  #############################################################################
  method onAnnotationTypeChange {} {
    ## The annotation type has been changed. Update the Annotation list widget
    my clearAnnotationView
    if {$enable_tree_view} {
      my updateAnnotationTreeView
    } else {
      my updateAnnotationListView
    }
  };# onAnnotationTypeChange

  #############################################################################
  ## Event: onAnnotationChange
  #############################################################################
  method onAnnotationChange {} {
    $display_object call selectionClear
    foreach id [$annotation_view selection] {
      foreach span [tip_GetSpans [tip_GetAnnotation $document $id]] {
        tip_GetSpanOffsets $span s e
        catch {
          $display_object call addSelectionIndexes \
            [$display_object call characterOffset2Index $s] \
            [$display_object call characterOffset2Index $e]
        }
      }
    }
    if {$make_selected_annotation_visible &&
        [$display_object call selectionExists]} {
      $display_object call seeIndex \
        [$display_object call getSelectionStartIndex]
      $display_object call setFocus
    }
  };# onAnnotationChange

  method clearAnnotationView {} {
    $annotation_view delete [$annotation_view children {}]
    $annotation_view configure -columns {}
    $annotation_view see {}
  };# clearAnnotationView

  method setAnnotationViewColumns {} {
    if {![string length $document]} {return}
    set attribute_names [$document GetAnnotationAttributeNames $annotation_type]
    $annotation_view configure -columns $attribute_names
    foreach name $attribute_names {
      $annotation_view heading $name -text $name
      $annotation_view column  $name -width 60
    }
    if {$enable_tree_view} {
      $annotation_view column  #0 -width 120
    } else {
      $annotation_view column  #0 -width  60
    }
  };# setAnnotationViewColumns

  method annotation2ColumnValues {annotation} {
    set values {}
    foreach name $attribute_names {
      set attribute [tip_GetAttribute $annotation $name]
      if {[llength $attribute]} {
        lappend values [tip_GetValueValue $attribute]
      } else {
        lappend values {}
      }
    }
    return $values
  };# annotation2ColumnValues

  method updateAnnotationTreeView {} {
    if {![string length $document]} {return}
    my setAnnotationViewColumns
    set Annotations [tip_SelectAnnotations $document $annotation_type]
    foreach ann $Annotations {
      set id [tip_GetId $ann]
      $annotation_view insert {} end -id $id -text $id \
        -values [my annotation2ColumnValues $ann] -open 0
    }
    ## Arrange Annotations in a tree...
    while {[llength $Annotations]} {
      set ann [lindex $Annotations 0]
      my updateAnnotationTreeViewHelper {} $ann
    }
  };# updateAnnotationTreeView

  method updateAnnotationTreeViewHelper {parent ann} {
    upvar Annotations Annotations
    set id [tip_GetId $ann]
    
    ## Step 1: Move this node under the given parent...
    if {[string length $parent]} {
      $annotation_view move $id $parent end
    }

    ## Step 2: Remove the Annotation from the list...
    set index [lsearch -exact $Annotations $ann]
    if {$index != -1} {
      set Annotations [lreplace $Annotations $index $index]
    }

    ## Step 3: Find all annotations that ly within this annotation...
    set spans [lsort [tip_GetSpans $ann]]
    my find_min_start_&_max_end $ann min_start max_end
    set children [tip_AnnotationsInRange $Annotations $min_start $max_end]
    foreach child $children {
      set index [lsearch -exact $Annotations $child]
      if {$index != -1} {
        ## Ensure that the child lies completely inside this ann!
        my find_min_start_&_max_end $child c_min_start c_max_end
        if {$c_min_start == $min_start && $c_max_end == $max_end} {
          ## This child covers exactly the same region. Its not a child, but a
          ## sibling!
          if {[string length $parent]} {
            $annotation_view move [tip_GetId $child] $parent end
          }
          set Annotations [lreplace $Annotations $index $index]
        } elseif {$c_min_start >= $min_start && $c_max_end <= $max_end} {
          my updateAnnotationTreeViewHelper $id $child
        }
      }
    }
  };# updateAnnotationTreeViewHelper

  classmethod find_min_start_&_max_end {ann start_var end_var} {
    upvar $start_var min_start 
    upvar $end_var   max_end
    tip_GetFirstSpanOffsets $ann min_start max_end
    foreach span [lrange [tip_GetSpans $ann] 1 end] {
      tip_GetSpanOffsets $span start end
      if {$start < $min_start} {set min_start $start}
      if {$end   > $max_end  } {set max_end   $end}
    }
  };# find_min_start_&_max_end

  method updateAnnotationListView {} {
    if {![string length $document]} {return}
    my setAnnotationViewColumns
    foreach ann [tip_SelectAnnotations $document $annotation_type] {
      set id [tip_GetId $ann]
      $annotation_view insert {} end -id $id -text $id \
        -values [my annotation2ColumnValues $ann]
    }
  };# updateAnnotationListView

  destructor {
  };# destructor

};# class ELEP::ViewerBase::AnnotationSelector

package provide ELEP::ViewerBase::AnnotationSelector 1.0
