##############################################################################
##  AnnotationSpecificationSelector.tcl:                                    ##
## --------------------------------------------------                       ##
##  This class implements a widget that presents a set of options, that can ##
##  be used to select which Annotation types will be added to a Document.   ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require BWidget
package require ELEP::WindowUtilities
package require ELEP::ViewerBase::AnnotatorSpecificationReader

oo::class create ELEP::ViewerBase::AnnotationSpecificationSelector {
  variable reader \
           dialog_window \
           language \
           annotation \
           attribute \
           alternative \
           groups \
           values
  
  constructor {{init_dialog 1}} {
    set reader        {}
    set dialog_window {}
    set language      {}
    set annotation    {}
    set attribute     {}
    set alternative   {}
    set groups        {}
    set values        {}
    if {$init_dialog} {my init}
  };# constructor

  method init {} {
    my createReader
    my createDialog
    my populateDialogFrame [$dialog_window getframe]
    my restoreState [$dialog_window getframe]
  };# init

  method createReader {} {
    set reader [ELEP::ViewerBase::AnnotatorSpecificationReader new]
  };# createReader

  method createDialog {} {
    set dialog_window [Dialog [::ELEP::WindowUtilities::UniqueWindowName\
     .annotatorSpecificationSelector] -modal local -separator 1 \
     -title "Select Annotation Characteristics..." -default 0 -cancel 1]
    $dialog_window add -textvariable GUI_Text(Ok)
    $dialog_window add -textvariable GUI_Text(Cancel)
  };# createDialog

  method populateDialogFrame {frame} {
    ttk::label $frame.llang -anchor w -text {Language:}
    ttk::combobox $frame.lang -state readonly -exportselection 0 \
      -values [concat {{}} [$reader getLanguages]] \
      -textvariable [my varname language]
    bind $frame.lang <<ComboboxSelected>> "[self] onLanguageChange $frame"
    grid $frame.llang $frame.lang -sticky snew \
      -ipadx 3 -ipady 3 -padx 6 -pady 6
    ttk::label $frame.lann -anchor w -text {Annotation Type:}
    ttk::combobox $frame.ann -state readonly -exportselection 0 \
      -textvariable [my varname annotation]
    bind $frame.ann <<ComboboxSelected>> "[self] onAnnotationChange $frame"
    grid $frame.lann $frame.ann -sticky snew \
      -ipadx 3 -ipady 3 -padx 6 -pady 6
    ttk::label $frame.lattr -anchor w -text {Annotation Attribute:}
    ttk::combobox $frame.attr -state readonly -exportselection 0 \
      -textvariable [my varname attribute]
    bind $frame.attr <<ComboboxSelected>> "[self] onAttributeChange $frame"
    grid $frame.lattr $frame.attr -sticky snew \
      -ipadx 3 -ipady 3 -padx 6 -pady 6
    ttk::label $frame.lalternative -anchor w -text {Attribute Alternative:}
    ttk::combobox $frame.alternative -state readonly -exportselection 0 \
      -textvariable [my varname alternative]
    bind $frame.alternative <<ComboboxSelected>> "[self] onAlternativeChange"
    grid $frame.lalternative $frame.alternative -sticky snew \
      -ipadx 3 -ipady 3 -padx 6 -pady 6
    ttk::labelframe $frame.lf -text Values:
    ScrolledWindow $frame.lf.sc
    listbox $frame.lf.sc.values -listvariable [my varname values] \
      -height 5
    $frame.lf.sc setwidget $frame.lf.sc.values
    pack $frame.lf.sc -fill both -expand 1 -padx 2 -pady 2
    grid $frame.lf -row 0 -column 2 -rowspan 5 -sticky snew \
      -ipadx 3 -ipady 3 -padx 6 -pady 1
    grid rowconfigure $frame 4 -weight 1
    grid columnconfigure $frame 1 -weight 1
  };# populateDialogFrame

  method onLanguageChange {frame} {
    $frame.ann configure -values [$reader getAnnotationTypes $language]
    set annotation {}
    $frame.attr        configure -values {}; set attribute   {}
    $frame.alternative configure -values {}; set alternative {}
    set values {}
  };# onLanguageChange

  method onAnnotationChange {frame} {
    $frame.attr configure -values [$reader getAnnotationTypeAttributes \
       $language $annotation]
    set attribute {}
    $frame.alternative configure -values {}; set alternative {}
    set values {}
  };# onAnnotationChange

  method onAttributeChange {frame} {
    $frame.alternative configure -values [concat {{}} [$reader \
       getAnnotationAttributeAlternatives $language $annotation $attribute]]
    set alternative {}
    set values {}
  };# onAttributeChange

  method onAlternativeChange {} {
    set groups [$reader getAnnotationAttributeAlternativeGroups \
       $language $annotation $attribute $alternative]
    set values [$reader getAnnotationAttributeValues \
       $language $annotation $attribute $alternative]
  };# onAlternativeChange

  method getGroupValues {group} {
    return [$reader getAnnotationAttributeAlternativeGroupValues \
      $language $annotation $attribute $alternative $group]
  };# getGroupValues

  method getGroupButtonsNumber {group} {
    return [$reader getButtonsNumber \
      $language $annotation $attribute $alternative $group]
  };# getGroupButtonsNumber

  method getGroupSort {group} {
    return [$reader getSort \
      $language $annotation $attribute $alternative $group]
  };# getGroupSort

  method getMaxButtonsNumber {} {return [$reader getMaxButtonsNumber]}

  method getGroupDescription {group} {
    return [$reader getDescription \
      $language $annotation $attribute $alternative $group]
  };# getGroupDescription

  method getValueDescription {group value} {
    return [$reader getDescription \
      $language $annotation $attribute $alternative $group $value]
  };# getValueDescription

  method getValueBackgroundColour {group value} {
    return [$reader getBackgroundColour \
      $language $annotation $attribute $alternative $group $value]
  };# getValueBackgroundColour

  method getValueForegroundColour {group value} {
    return [$reader getForegroundColour \
      $language $annotation $attribute $alternative $group $value]
  };# getValueForegroundColour

  method getValueColumnSpan {group value} {
    return [$reader getAttribute columnspan 1 \
      $language $annotation $attribute $alternative $group $value]
  };# getValueColumnSpan

  method getValueEntryFields {group value} {
    return [$reader getAttribute entry {} \
      $language $annotation $attribute $alternative $group $value]
  };# getValueEntryFields

  method getValueEventFields {group value} {
    return [$reader getAttribute event {} \
      $language $annotation $attribute $alternative $group $value]
  };# getValueEventFields

  method show {} {
    if {[$dialog_window draw] == 0} {
      my saveState
      return ok
    } else {
      return cancel
    }
  };# show

  destructor {
    catch {$reader destroy}
  };# destructor

  method saveStateVariableName {} {return ::GUI_Options([self class])}

  method saveState {} {
    set [my saveStateVariableName] [list $language  $annotation \
                                         $attribute $alternative]
  };# saveState

  method restoreState {frame} {
    if {![info exists [my saveStateVariableName]]} {return}
    foreach {language _annotation _attribute _alternative} \
            [set [my saveStateVariableName]] {break}
    my onLanguageChange $frame
    set annotation  $_annotation;  my onAnnotationChange $frame
    set attribute   $_attribute;   my onAttributeChange  $frame
    set alternative $_alternative; my onAlternativeChange
  };# restoreState

};# class ELEP::ViewerBase::AnnotationSpecificationSelector

package provide ELEP::ViewerBase::AnnotationSpecificationSelector 1.0
