##############################################################################
##  AnnotatorSpecificationReader.tcl:                                       ##
## --------------------------------------------------                       ##
##  This class implements a widget that presents a set of buttons, that can ##
##  be used to add Annotations to a Document.                               ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require tdom
package require ELEP::Base::Utilities

oo::class create ELEP::ViewerBase::AnnotatorSpecificationReader {
  mixin ELEP::Base::Utilities
  common PATH [list $::UserHomeDir $::UserAppDir $::HomeDir/share/Config]
  variable path \
           specificationsFileNameList \
           doc \
           root \
           max_button_number

  constructor {{pathList {}} {fileNameList AnnotationsSpecifications.xml}} {
    set doc  {}
    set root {}
    set max_button_number 1
    set path $pathList
    set specificationsFileNameList $fileNameList

    my loadSpecificationsFile [my locateSpecificationsFile]
  };# constructor

  method xpath {{language {}} {annotation {}} {attribute {}} {alternative {}}
                {group {}} {value {}}} {
    set xpath {language}
    if {[string length $language]} {append xpath "\[@name='$language'\]"}
    foreach var {annotation attribute alternative group value} {
      if {[string length [set $var]]} {
        append xpath "//$var"
        if {![string equal [set $var] *!*]} {
          append xpath "\[@name='[set $var]'\]"
        }
      }
    }
    return $xpath
  };# xpath

  method xpathTemplate {{language {}} {template {}} {table {}} {field {}}} {
    set xpath {language}
    if {[string length $language]} {append xpath "\[@name='$language'\]"}
    foreach var {template table field} {
      if {[string length [set $var]]} {
        append xpath "//$var"
        if {![string equal [set $var] *!*]} {
          append xpath "\[@name='[set $var]'\]"
        }
      }
    }
    return $xpath
  };# xpathTemplate

  method xpathMultiAttributed {{language {}} {multi_attributed_annotation {}}
                               {alternative {}} {attribute {}} {value {}}} {
    set xpath {language}
    if {[string length $language]} {append xpath "\[@name='$language'\]"}
    foreach var {multi_attributed_annotation alternative attribute value} {
      if {[string length [set $var]]} {
        append xpath "//$var"
        if {![string equal [set $var] *!*]} {
          append xpath "\[@name='[set $var]'\]"
        }
      }
    }
    return $xpath
  };# xpathMultiAttributed

  method getNodes {{language {}} {annotation {}} {attribute {}} {alternative {}}
                   {group {}} {value {}}} {
    return [$root selectNodes [my xpath $language $annotation $attribute \
                                        $alternative $group $value]]
  };# getNodes

  method getTemplateNodes {{language {}} {template {}} {table {}} {field {}}} {
    return [$root selectNodes \
                  [my xpathTemplate $language $template $table $field]]
  };# getTemplateNodes

  method getMultiAttributedNodes {
     {language {}} {annotation {}} {alternative {}} {attribute {}} {value {}}} {
    return [$root selectNodes [my xpathMultiAttributed $language $annotation \
                                     $alternative $attribute $value]]
  };# getMultiAttributedNodes

  method getNames {{language {}} {annotation {}} {attribute {}} {alternative {}}
                   {group {}} {value {}}} {
    set list {}
    foreach node [my getNodes $language $annotation $attribute \
                                        $alternative $group $value] {
      lappend list [$node getAttribute name {}]
    }
    #return [lsort -unique $list]
    return $list
  };# getNames

  method getNamesTemplateUnsorted {{language {}} {template {}}
                                   {table {}} {field {}}} {
    set list {}
    foreach node [my getTemplateNodes $language $template $table $field] {
      lappend list [$node getAttribute name {}]
    }
    return $list
  };# getNamesTemplateUnsorted

  method getNamesTemplate {{language {}} {template {}} {table {}} {field {}}} {
    return [my getNamesTemplateUnsorted $language $template $table $field]
    #return [lsort -unique \
    #  [my getNamesTemplateUnsorted $language $template $table $field]]
  };# getNamesTemplate

  method getNamesMultiAttributed {{language {}} {annotation {}}
                                  {alternative {}} {attribute {}} {value {}} } {
    set list {}
    foreach node [my getMultiAttributedNodes $language $annotation \
                                             $alternative $attribute $value] {
      lappend list [$node getAttribute name {}]
    }
    #return [lsort -unique $list]
    return $list
  };# getNames

  method getLanguages {} {return [my getNames]}

  ##
  ## Annotation related methods.
  ##
  method getAnnotationTypes {{language {}}} {return [my getNames $language *!*]}
  method getAnnotationTypeAttributes {language annotation} {
    return [my getNames $language $annotation *!*]
  };# getAnnotationTypeAttributes
  method getAnnotationAttributeAlternatives {language annotation attribute} {
    return [my getNames $language $annotation $attribute *!*]
  };# getAnnotationAttributeAlternatives
  method getAnnotationAttributeAlternativeGroups {language annotation attribute
                                                  alternative} {
    return [my getNames $language $annotation $attribute $alternative *!*]
  };# getAnnotationAttributeAlternativeGroups
  method getAnnotationAttributeAlternativeGroupValues {language annotation
                               attribute alternative group} {
    return [my getNames $language $annotation $attribute\
                                  $alternative $group *!*]
  };# getAnnotationAttributeAlternativeGroups
  method getAnnotationAttributeValues {language annotation attribute
                                                           alternative} {
    return [my getAnnotationAttributeAlternativeGroupValues \
               $language $annotation $attribute $alternative {}]
  };# getAnnotationAttributeValues

  method getDescription {language {annotation {}} {attribute {}}
                         {alternative {}} {group {}} {value {}}} {
    foreach node [my getNodes $language $annotation $attribute \
                                     $alternative $group $value] {
      return [string trim [$node text]]
    }
    return {}
  };# getDescription

  ## getBackgroundColour
  #  Return the background colour (if defined)
  method getBackgroundColour {language {annotation {}} {attribute {}}
                              {alternative {}} {group {}} {value {}}} {
    foreach node [my getNodes $language $annotation $attribute \
                                        $alternative $group $value] {
      set colour [$node getAttribute background {}]
      if {[string length $colour]} {return $colour}
    }
    return {}
  };# getBackgroundColour

  ## getForegroundColour
  #  Return the foreground colour (if defined)
  method getForegroundColour {language {annotation {}} {attribute {}}
                              {alternative {}} {group {}} {value {}}} {
    foreach node [my getNodes $language $annotation $attribute \
                                        $alternative $group $value] {
      set colour [$node getAttribute foreground {}]
      if {[string length $colour]} {return $colour}
    }
    return {}
  };# getForegroundColour

  method getButtonsNumber {language {annotation {}} {attribute {}}
                           {alternative {}} {group {}} {value {}}} {
    set button_number $max_button_number
    foreach node [my getNodes $language $annotation $attribute \
                                        $alternative $group $value] {
      set button_number [$node getAttribute button_number $max_button_number]
      if {$max_button_number < $button_number} {
        set max_button_number $button_number
      }
    }
    return $button_number
  };# getButtonsNumber

  method getSort {language {annotation {}} {attribute {}}
                           {alternative {}} {group {}} {value {}}} {
    return [my getAttribute sort true $language $annotation $attribute \
                                        $alternative $group $value]
  };# getSort

  method getAttribute {name def_value language {annotation {}} {attribute {}}
                           {alternative {}} {group {}} {value {}}} {
    foreach node [my getNodes $language $annotation $attribute \
                                        $alternative $group $value] {
      return [$node getAttribute $name $def_value]
    }
    return $def_value
  };# getAttribute

  method getMaxButtonsNumber {} {return $max_button_number}

  ##
  ## Template related methods.
  ##
  method getTemplateNames {{language {}}} {
    return [my getNamesTemplate $language *!*]
  };# getTemplateNames

  method getTemplateTables {language template} {
    return [my getNamesTemplateUnsorted $language $template *!*]
  };# getTemplateTables

  method getTemplateTableFields {language template table} {
    return [my getNamesTemplateUnsorted $language $template $table *!*]
  };# getTemplateTableFields

  method getTemplateTableFieldValueType {language template table field} {
    foreach node [my getTemplateNodes $language $template $table $field] {
      return [$node getAttribute value_type annotation::id]
    }
    return annotation::id
  };# getTemplateTableFieldValueType

  method getTemplateTableFieldMaxValueNumber {language template table field} {
    foreach node [my getTemplateNodes $language $template $table $field] {
      return [$node getAttribute max_value_number 1]
    }
    return 1
  };# getTemplateTableFieldMaxValueNumber

  method getTemplateTableFieldPrimary {language template table field} {
    foreach node [my getTemplateNodes $language $template $table $field] {
      return [$node getAttribute primary 0]
    }
    return 0
  };# getTemplateTableFieldPrimary

  method getTemplateTableFieldValueConstraints {language template table field} {
    foreach node [my getTemplateNodes $language $template $table $field] {
      return [$node getAttribute value_constraints {}]
    }
    return {}
  };# getTemplateTableFieldValueConstraints

  method getTemplateDescription {language {template {}} {table {}} {field {}}} {
    foreach node [my getTemplateNodes $language $template $table $field] {
      return [string trim [$node text]]
    }
    return {}
  };# getTemplateDescription

  ##
  ## Multi-Attributed Annotations...
  ##
  method getMultiAttributedAnnotationTypes {{language {}}} {
    return [my getNamesMultiAttributed $language *!*]
  };# getMultiAttributedAnnotationTypes
  method getMultiAttributedAnnotationAlternatives {language annotation} {
    return [my getNamesMultiAttributed $language $annotation *!*]
  };# getMultiAttributedAnnotationAlternatives
  method getMultiAttributedAnnotationAttributes {language annotation
                                                 alternative} {
    return [my getNamesMultiAttributed $language $annotation $alternative *!*]
  };# getMultiAttributedAnnotationAttributes
  method getMultiAttributedAnnotationAttributeType {language annotation
                                                 alternative attribute} {
    foreach node [my getMultiAttributedNodes $language $annotation \
                                             $alternative $attribute] {
      return [$node getAttribute type {}]
    }
    return {}
  };# getMultiAttributedAnnotationAttributeType
  method getMultiAttributedAnnotationAttributeValues {language annotation
                                                 alternative attribute} {
    return [my getNamesMultiAttributed $language $annotation \
                                       $alternative $attribute *!*]
  };# getMultiAttributedAnnotationAttributeValues

  method getMultiAttributedDescription {language {annotation {}}
                                {alternative {}} {attribute {}} {value {}}} {
    foreach node [my getMultiAttributedNodes $language $annotation \
                                             $alternative $attribute $value] {
      return [string trim [$node text]]
    }
    return {}
  };# getMultiAttributedDescription

  method loadSpecificationsFile {filename} {
    ## Parse the configuration file...
    set fd  [tDOM::xmlOpenFile $filename]
    set doc [dom parse -channel $fd]
    close $fd
    set root [$doc documentElement]
  };# loadSpecificationsFile

  method locateSpecificationsFile {} {
    my common PATH
    foreach dir [concat $path $PATH] {
      foreach filename $specificationsFileNameList {
        if {[file exists $dir/$filename] && [file readable $dir/$filename]} {
          return $dir/$filename
        }
      }
    }
  };# locateSpecificationsFile

  destructor {
    catch {$doc delete}
  };# destructor

};# class ELEP::ViewerBase::AnnotatorSpecificationReader

package provide ELEP::ViewerBase::AnnotatorSpecificationReader 1.0
