##############################################################################
##  AutomaticTagger.tcl:                                                    ##
## --------------------------------------------------                       ##
##  This class implements a utility that automatically annotates Documents  ##
##  according to user-specified regular expressions.                        ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities::WidgetBase
package require ELEP::VisualAnnotators::VisualRegexp
package require tablelist_tile

oo::class create ELEP::ViewerBase::AutomaticTagger {
  superclass ELEP::WindowUtilities::WidgetBase

  variable {*}[info class variables ELEP::WindowUtilities::WidgetBase] \
           window_obj \
           display_object \
           annotation \
           attribute \
           attribute_values_object \
           visual_regexp_object \
           autoacq_frame \
           regexps_frame \
           autoapp_frame \
           patterns_frame \
           listbox \
           automatic_acquire_regexps \
           automatic_apply_on_doc_changes \
           automatic_acquire_type \
           regexp_list \
           char_map 

  constructor {parent_win {disp_obj {}}} {
    set window_obj     {};# The object that holds our toplevel window.
    set display_object {};# The object that shows annotated information.
    set annotation                     {}
    set attribute                      {}
    set attribute_values_object        {}
    set visual_regexp_object           {}
    set autoacq_frame                  {}
    set regexps_frame                  {}
    set autoapp_frame                  {}
    set patterns_frame                 {}
    set listbox                        {}
    set automatic_acquire_regexps      1
    set automatic_apply_on_doc_changes 0
    set regexp_list                    {}
    set char_map [list \{ \\\{ \} \\\} ( \( ) \) \\ \\\\ \
                            * {\*} + {\+} ? {\?} ^ {\^} $ {\$} \
                            . {\.} \[ {\[} \] {\]}]

    next ttk::frame [::ELEP::WindowUtilities::UniqueWindowName \
                           $parent_win.automaticAnnotator]

    set display_object $disp_obj
    set client_area $widget
    set visual_regexp_object [ELEP::VisualAnnotators::VisualRegexp new]
    ttk::notebook $widget.nb
    my createAutomaticAcquisitionPage
    $widget.nb add $autoacq_frame -text {Automatic Acquisition}
    my createRegularExpressionsPage
    $widget.nb add $regexps_frame -text {Regular Expressions}
    my createAutomaticApplicationPage
    $widget.nb add $autoapp_frame -text {Automatic Application}
    # createPatternsPage
    # $widget.nb add $patterns_frame -text {Patterns}
    $widget.nb select 1
    pack $widget.nb -fill both -expand 1
  };# constructor

  method createPatternsPage {} {
    set patterns_frame [ttk::frame $widget.nb.patterns]
  };# createPatternsPage

  method createAutomaticAcquisitionPage {} {
    set autoacq_frame [ttk::frame $widget.nb.autoacq]

    ttk::checkbutton $autoacq_frame.autoAcquire -text \
      {Automatically Acquire Regular Expressions} \
      -variable [my varname automatic_acquire_regexps]
    
    ttk::labelframe $autoacq_frame.ex -text {Exclude Values:}
    ScrolledWindow $autoacq_frame.ex.sw
    ScrollableFrame $autoacq_frame.ex.sw.client_area -constrainedwidth 1
    $autoacq_frame.ex.sw setwidget $autoacq_frame.ex.sw.client_area
    pack $autoacq_frame.ex.sw -fill both -expand 1 -padx 2 -pady 2
    pack $autoacq_frame.autoAcquire -fill x -expand 0 -padx 2 -pady 2
    pack $autoacq_frame.ex -fill both -expand 1 -padx 2 -pady 2
    my fillAutomaticAcquisitionPage
  };# createAutomaticAcquisitionPage

  method fillAutomaticAcquisitionPage {} {
    if {![string length $attribute_values_object]} {return}
    set frame [$autoacq_frame.ex.sw.client_area getframe]
    foreach value [lsort -unique [$attribute_values_object getValues]] {
      set w $frame.v$value
      if {[winfo exists $w]} {continue}
      ttk::checkbutton $w -text $value -variable \
        [my varname automatic_acquire_type($value)]
      pack $w -fill x -expand 0 -padx 6 -pady 1
    }
  };# fillAutomaticAcquisitionPage

  method createRegularExpressionsPage {} {
    set regexps_frame [ttk::frame $widget.nb.regexps]

    ScrolledWindow $regexps_frame.sw
    set listbox [tablelist::tablelist $regexps_frame.sw.tl \
      -columns {16 Value 0 {Regular Expression}} -selectmode extended \
      -stretch 1 -listvariable [my varname regexp_list] \
      -editstartcommand [list [self] onValueEdit]]
    $regexps_frame.sw setwidget $listbox
    $listbox columnconfigure 0 -editable yes -editwindow ttk::combobox \
      -sortmode dictionary

    ttk::button $regexps_frame.add -textvariable GUI_Text(Add) \
      -command [list [self] onPatternAdd]
    ttk::button $regexps_frame.delete -textvariable GUI_Text(Delete) \
      -command [list [self] onPatternDelete]
    ttk::button $regexps_frame.edit -textvariable GUI_Text(Edit) \
      -command [list [self] onPatternEdit]

    ttk::button $regexps_frame.save -textvariable GUI_Text(Save) \
      -command [list [self] onPatternsSave]
    ttk::button $regexps_frame.load -textvariable GUI_Text(Load) \
      -command [list [self] onPatternsLoad]

    grid $regexps_frame.sw     -rowspan 6 -sticky snew -padx 2 -pady 2
    grid $regexps_frame.add    -column 1 -row 0 -sticky snew -padx 2 -pady 2
    grid $regexps_frame.delete -column 1 -row 1 -sticky snew -padx 2 -pady 2
    grid $regexps_frame.edit   -column 1 -row 2 -sticky snew -padx 2 -pady 2
    grid $regexps_frame.save   -column 1 -row 4 -sticky snew -padx 2 -pady 2
    grid $regexps_frame.load   -column 1 -row 5 -sticky snew -padx 2 -pady 2
    grid columnconfigure $regexps_frame 0 -weight 1
    grid rowconfigure    $regexps_frame 3 -weight 1
  };# createRegularExpressionsPage

  method createAutomaticApplicationPage {} {
    set autoapp_frame [ttk::frame $widget.nb.autoapp]

    ttk::checkbutton $autoapp_frame.autoApplyOnDocLoad -text \
      {Automatically Apply on Document changes} \
      -variable [my varname automatic_apply_on_doc_changes]
    pack $autoapp_frame.autoApplyOnDocLoad -fill x -expand 0 -padx 2 -pady 2
  };# createAutomaticApplicationPage

  #############################################################################
  ## Event: addPatternFromAnnotationId
  #############################################################################
  method addPatternFromAnnotationId {id} {
    if {!$automatic_acquire_regexps} {return}
    if {![string length $display_object]} {return}
    set doc [$display_object call getDocument]
    set ann [tip_GetAnnotation $doc $id]
    set value [tip_GetValueValue [tip_GetAttribute $ann $attribute]]
    if {[info exists automatic_acquire_type($value)] &&
        $automatic_acquire_type($value)} {return}
    foreach text [tip_GetAnnotatedTextRanges $doc $ann] {
      my addPatternFromText $value [string trim $text]
    }
  };# addPatternFromAnnotationId

  #############################################################################
  ## Event: addPatternFromText
  #############################################################################
  method addPatternFromText {value text} {
    set pattern [my text2Pattern $text]
    ## Is pattern already in our pattern list?
    set end [llength $regexp_list]
    for {set i 0} {$i < $end} {incr i} {
      if {[string equal $pattern [lindex $regexp_list $i 1]]} {
        set index $i; break
      }
    }
    if {![info exists index]} {set index -1}
    if {$index == -1} {
      $listbox insert end [list $value $pattern]
    } else {
      set val [lindex [lindex $regexp_list $index] 0]
      if {![string equal $value $val]} {
        set regexp_list [lreplace $regexp_list $index $index \
                          [list $value $pattern]]
      }
    }
  };# addPatternFromText

  method text2Pattern {text} {
    set pattern [string map $char_map $text]
    set pattern [regsub -all -- {\s+} $pattern {\s+}]
    return "\\m$pattern\\M"
  };# text2Pattern

  #############################################################################
  ## Event: removePatternFromAnnotationId
  #############################################################################
  method removePatternFromAnnotationId {id} {
    if {!$automatic_acquire_regexps} {return}
    if {![string length $display_object]} {return}
    set doc [$display_object call getDocument]
    set ann [tip_GetAnnotation $doc $id]
    set value [tip_GetValueValue [tip_GetAttribute $ann $attribute]]
    foreach text [tip_GetAnnotatedTextRanges $doc $ann] {
      my removePatternFromText $value [string trim $text]
    }
  };# removePatternFromAnnotationId

  #############################################################################
  ## Event: removePatternFromText
  #############################################################################
  method removePatternFromText {value text} {
    set pattern [my text2Pattern $text]
    ## Is pattern already in our pattern list?
    set end [llength $regexp_list]
    for {set i 0} {$i < $end} {incr i} {
      if {[string equal $pattern [lindex $regexp_list $i 1]]} {
        set index $i; break
      }
    }
    if {![info exists index]} {return}
    set regexp_list [lreplace $regexp_list $index $index]
  };# removePatternFromText

  #############################################################################
  ## Event: onValueEdit
  #############################################################################
  method onValueEdit {tbl row col text} {
    if {![string length $attribute_values_object]} {return}
    set w [$tbl editwinpath]
    $w configure -values [lsort -unique [concat \
       [$attribute_values_object getValues] $text]]
    return $text
  };# onValueEdit

  #############################################################################
  ## Event: onPatternAdd
  #############################################################################
  method onPatternAdd {} {
    if {![string length $visual_regexp_object]} {return}
    $visual_regexp_object setPattern {}
    set edited_pattern [$visual_regexp_object show]
    if {![string length $edited_pattern]} {return}
    $listbox insert end [list {} $edited_pattern]
    my updateRegularExpression end
    $listbox editcell end,0
  };# onPatternAdd

  #############################################################################
  ## Event: onPatternDelete
  #############################################################################
  method onPatternDelete {} {
    set sel [$listbox curselection]
    if {![llength $sel]} {return}
    $listbox delete $sel
    my updateRegularExpression $sel
  };# onPatternDelete

  #############################################################################
  ## Event: onPatternEdit
  #############################################################################
  method onPatternEdit {} {
    if {![string length $visual_regexp_object]} {return}
    set sel [$listbox curselection]
    if {![llength $sel]} {return}
    set sel [lindex $sel 0]
    foreach {value edited_pattern} [$listbox get $sel] {break}
    $visual_regexp_object setPattern $edited_pattern
    set edited_pattern [$visual_regexp_object show]
    if {![string length $edited_pattern]} {return}
    $listbox delete $sel
    $listbox insert $sel [list $value $edited_pattern]
    my updateRegularExpression $sel
  };# onPatternEdit

  method updateRegularExpression {sel} {
  };# updateRegularExpression

  #############################################################################
  ## Event: onPatternsSave
  #############################################################################
  method onPatternsSave {} {
    set filename [GUI_SelectSaveFile {} -parent [getToplevel]]
    if {![string length $filename]} {return}
    set file [open $filename w]
    fconfigure $file -encoding utf-8
    puts $file $regexp_list
    puts $file [array get automatic_acquire_type]
    close $file
  };# onPatternsSave

  #############################################################################
  ## Event: onPatternsLoad
  #############################################################################
  method onPatternsLoad {} {
    set filename [GUI_SelectFile {} -parent [getToplevel]]
    if {![string length $filename]} {return}
    set file [open $filename]
    fconfigure $file -encoding utf-8
    gets $file regexp_list
    array set automatic_acquire_type [read $file]
    close $file
  };# onPatternsLoad

  ##########################################################################
  ## Event: onDocumentChange
  ##########################################################################
  method onDocumentChange {col doc} {
    my fillAutomaticAcquisitionPage
    if {!$automatic_apply_on_doc_changes} {return {}}
    return [my applyAutomaticTagger $col $doc]
  };# onDocumentChange

  ##########################################################################
  ## Event: applyAutomaticTagger
  ##########################################################################
  method applyAutomaticTagger {col doc} {
    set ids {}
    set text [tip_GetByteSequence $doc]
    set annotations [tip_SelectAnnotations $doc $annotation \
                       "\[string length ann::${attribute} \]"]
    ## Apply regular expressions...
    foreach item $regexp_list {
      foreach {value regexp} $item {break}
      foreach index [regexp -all -indices -inline -- $regexp $text] {
        foreach {start end} $index {break}
        incr end
        # Is there an annotation that covers this span?
        if {[llength [tip_AnnotationsContaining $annotations $start $end]]} {
          continue
        }
        set ann [tip_CreateAnnotation $annotation \
                     [tip_CreateSpanSet $start $end] \
                     [tip_CreateAttributeSet $attribute \
                          [tip_CreateAttributeValue GDM_STRING $value]]]
        lappend ids [tip_AddAnnotation $doc $ann]
        lappend annotations $ann
      }
    }
    return $ids
  };# applyAutomaticTagger

  destructor {
    if {[string length $visual_regexp_object]} {
      $visual_regexp_object destroy
    }
  };# destructor

};# class ELEP::ViewerBase::AutomaticTagger

package provide ELEP::ViewerBase::AutomaticTagger 1.0
