##############################################################################
##  ButtonAnnotator.tcl:                                                    ##
## --------------------------------------------------                       ##
##  This class implements a widget that presents a set of buttons, that can ##
##  be used to add Annotations to a Document.                               ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities::WidgetBase
package require ELEP::ViewerBase::ColourManager
package require ELEP::ViewerBase::AnnotationSpecificationSelector
package require ELEP::ViewerBase::AutomaticTagger
package require ELEP::ViewerBase::AnnotationHighlighter
package require ELEP::ViewerBase::1ClickSelector

oo::class create ELEP::ViewerBase::ButtonAnnotator {
  superclass ELEP::WindowUtilities::WidgetBase

  common FoundInCollectionGroup {Found in Collection}

  variable {*}[info class variables ELEP::WindowUtilities::WidgetBase] \
           window_obj \
           display_object \
           annotation \
           attribute \
           selector \
           colour_manager \
           annotation_highlighter_object \
           automatic_annotator_object \
           1_click_selector_object \
           window_index \
           images \
           row \
           column \
           button_number \
           accelerator_key

  constructor {win parent_win {disp_obj {}}} {
    my variable default_button_fg default_button_bg event_definer_object
    set window_obj     {};# The object that holds our toplevel window.
    set display_object {};# The object that shows annotated information.
    set annotation     {}
    set attribute      {}
    set selector       {}
    set colour_manager {}
    set annotation_highlighter_object {}
    set automatic_annotator_object    {}
    set 1_click_selector_object {};# An object that extends the way selection
                                   # is performed
    set event_definer_object {}
    set window_index 0
    set images       {}
    set default_button_fg black
    set default_button_bg white
    set row 0
    set column 0
    set button_number 0
    set accelerator_key {}

    next ttk::notebook [::ELEP::WindowUtilities::UniqueWindowName \
                           $parent_win.buttonAnnotator]

    set window_obj $win
    set display_object $disp_obj
    if {[string length $display_object]} {
      $display_object call registerAnnotatorObject [self]
    }
    ## Toolbar.
    my createToolbar

    ## Annotator Button area.
    ScrolledWindow $widget.buttonAnnotatorSW
    $widget add $widget.buttonAnnotatorSW -text {Button Annotator}
    ScrollableFrame $widget.buttonAnnotatorSW.client_area -constrainedwidth 1
    $widget.buttonAnnotatorSW setwidget $widget.buttonAnnotatorSW.client_area

    ## Automatic Annotator area.
    my createAutomaticAnnotatorObject
    $widget add [$automatic_annotator_object getframe] \
      -text {Automatic Annotator}

    ## Annotation Highlighter area.
    my createAnnotationHighlighterObject
    $widget add [$annotation_highlighter_object getframe] \
      -text {Annotation Highlighter}

    ## 1-Click Selector area.
    my create1ClickSelectorObject
    $widget add [$1_click_selector_object getframe] -text {1-Click Selector}
      
    set client_area [$widget.buttonAnnotatorSW.client_area getframe]
    ELEP::WindowUtilities::AutoFocusOnMouseEnter $client_area
    my addScrollBindings $widget.buttonAnnotatorSW.client_area $client_area
  };# constructor

  method createToolbar {} {
    global GUI_Text GUI_Options
    set toolbarFrame [$window_obj getToolbarFrame]
    set toolbar [ttk::frame $toolbarFrame.buttonAnnotator -class Toolbar]
    ttk::button $toolbar.automatic -text {Automatic Annotator} \
      -command "[self] invokeAutomaticAnnotator" -style Toolbutton
    grid $toolbar.automatic -sticky news -padx 2 -pady 2
    pack $toolbar -padx 2 -pady 2 -side left
  };# createToolbar
  
  method createAutomaticAnnotatorObject {} {
    set automatic_annotator_object [ELEP::ViewerBase::AutomaticTagger create \
            [::ELEP::Base::auto ::ELEP::ViewerBase::] $widget $display_object]
    $automatic_annotator_object configure -attribute_values_object [self]
  };# createAutomaticAnnotatorObject


  method createAnnotationHighlighterObject {} {
    set annotation_highlighter_object [ELEP::ViewerBase::AnnotationHighlighter \
      create [::ELEP::Base::auto ::ELEP::ViewerBase::] $widget $display_object]
  };# createAnnotationHighlighterObject

  method create1ClickSelectorObject {} {
    set 1_click_selector_object [ELEP::ViewerBase::1ClickSelector create \
          [::ELEP::Base::auto ::ELEP::ViewerBase::] $widget $display_object]
  };# create1ClickSelectorObject

  method createSpecificationSelectorObject {} {
    set selector [ELEP::ViewerBase::AnnotationSpecificationSelector new]
  };# createSpecificationSelectorObject

  method createColourManagerObject {} {
    set colour_manager [ELEP::ViewerBase::ColourManager new]
  };# createColourManagerObject

  method show {} {
    if {[string length $colour_manager]} {
      error "init can be called only once!"
    }
    my createSpecificationSelectorObject
    my createColourManagerObject
    set answer [$selector show]
    if {[string equal $answer ok]} {my init}
    $selector destroy
    return $answer
  };# show

  method init {} {
    my variable default_button_fg default_button_bg
    my common FoundInCollectionGroup
    set annotation  [$selector cget -annotation]
    set attribute   [$selector cget -attribute]
    set groups      [$selector cget -groups]
    if {![llength $groups]} {lappend groups {}};# No groups defined!
    foreach group $groups {
      my processGroup $group
    }
    my processGroup $FoundInCollectionGroup
    catch {
      set default_button_fg [$client_area.button2 cget -foreground]
      set default_button_bg [$client_area.button2 cget -background]
    }
    set button_number [$selector getMaxButtonsNumber]
    # Distribute evenly extra space among all columns...
    set columns [lindex [grid size $client_area] 0]
    for {set col 0} {$col < $columns} {incr col} {
      grid columnconfigure $client_area $col -weight 1
    }
    catch {$automatic_annotator_object fillAutomaticAcquisitionPage}
    focus $client_area
  };# init

  method processGroup {group} {
    set button_number     [$selector getGroupButtonsNumber $group]
    set max_button_number [$selector getMaxButtonsNumber]
    set sort              [$selector getGroupSort          $group]
    set raw_values        [$selector getGroupValues        $group]
    if {$sort} {
      set raw_values [lsort -unique $raw_values]
    }
    grid [my createGroupFrame $group] -sticky snew \
      -columnspan $max_button_number -padx 1 -pady 1
    set column -1; set row [lindex [grid size $client_area] 1]
    foreach value $raw_values {
      if {[incr column] >= $button_number} {incr row; set column 0}
      set columnspan [$selector getValueColumnSpan $group $value]
      ## Are there any entry fields?
      set additional_widgets 0
      set entries [split [$selector getValueEntryFields $group $value] \;]
      incr additional_widgets [llength $entries]

      set events [split [$selector getValueEventFields $group $value] \;]
      incr additional_widgets [llength $events]

      ## Destribute the columnspan to all widgets: if widgets do exist, then
      ## button allocates a single column, and the rest of the column span is
      ## distributed evenly among all other widgets. Else, the button occupies
      ## the whole columnspan...
      set button_columnspan $columnspan
      set widget_columnspan 0
      if {$additional_widgets} {
        set button_columnspan 1
        set widget_columnspan [expr {($columnspan-1)/$additional_widgets}]
      }

      ## Construct the button...
      set descr [$selector getValueDescription      $group $value]
      set bg    [$selector getValueBackgroundColour $group $value]
      set fg    [$selector getValueForegroundColour $group $value]
      grid [my createValueButton $client_area $group $value $descr $bg $fg] \
        -sticky snew -row $row -column $column -columnspan $button_columnspan \
        -padx 1 -pady 1
      incr column [expr {$button_columnspan - 1}]

      ## Construct event widgets...
      foreach one $events {
        incr column
        grid [my createValueButtonEvent $client_area $group $value \
                 $descr $bg $fg $one] \
        -sticky snew -row $row -column $column -columnspan $widget_columnspan \
        -padx 1 -pady 1
      }

      ## Construct entry widgets...
      foreach one $entries {
        incr column
        grid [my createValueButtonEntry $client_area $group $value \
                 $descr $bg $fg $one] \
        -sticky snew -row $row -column $column -columnspan $widget_columnspan \
        -padx 1 -pady 1
      }
    }
  };# processGroup

  method createGroupFrame {group} {
    set win [label $client_area.group[incr window_index] \
      -text "  \u25ba  $group" \
      -font {arial 10 bold} -foreground white -background #78B1F3 -anchor w]
    my addScrollBindings $widget.buttonAnnotatorSW.client_area $win; return $win
  };# createGroupFrame

  method createValueButton {parent group value {description {}}
                            {bg {}} {fg {}}} {
    my variable value2button value2attributes
    set win $client_area.button[incr window_index]
    if {![string length $bg] || ![string length $fg]} {
      foreach {bg fg} [$colour_manager ColourPairNext] {break}
    }
    lappend images [image create photo $win.image -width 12 -height 18]
    $win.image put $bg -to 0 0 8 18
    button $win -text $value -highlightthickness 0 \
      -activeforeground $fg -activebackground $bg -anchor w \
      -image $win.image -compound left -padx 2 -pady 0 \
      -command [list [self] onButtonPress $win $value $bg $fg]
    set value2button($value) $win
    set value2attributes($value) $attribute
    if {[string length $description]} {DynamicHelp::add $win -text $description}
    bind $win <<MB_2>> [list [self] onButtonAcceleratorChange $win]
    my addScrollBindings $widget.buttonAnnotatorSW.client_area $win; return $win
  };# createValueButton

  method createValueButtonEntry {parent group value {description {}}
                            {bg {}} {fg {}} entry_attribute} {
    my variable value2entry entry2attribute value2attributes
    set win $client_area.entry[incr window_index]
    ttk::entry $win -state disabled
    set value2entry($value) $win
    set entry2attribute($win) $entry_attribute
    lappend value2attributes($value) $entry_attribute
    my addScrollBindings $widget.buttonAnnotatorSW.client_area $win; return $win
  };# createValueButtonEntry

  method createValueButtonEvent {parent group value {description {}}
                            {bg {}} {fg {}} event_attribute} {
    my variable value2event event2attribute value2attributes
    set win $client_area.event[incr window_index]
    ttk::combobox $win -state disabled
    $win state readonly
    set value2event($value) $win
    set event2attribute($win) $event_attribute
    lappend value2attributes($value) $event_attribute
    my addScrollBindings $widget.buttonAnnotatorSW.client_area $win; return $win
  };# createValueButtonEntry

  method addValueButton {value} {
    my common FoundInCollectionGroup
    if {[incr column] >= $button_number} {incr row; set column 0}
    grid [my createValueButton $client_area $FoundInCollectionGroup $value] \
        -sticky snew -row $row -column $column
  };# addValueButton

  method getEvents {} {
    my variable event_definer_object
    if {$event_definer_object eq ""} {return {}}
    return [$event_definer_object getEventIdsAndTitles]
  };# getEvents

  method getCurrentEventId {} {
    my variable event_definer_object
    if {$event_definer_object eq ""} {return {}}
    return [$event_definer_object getCurrentEventId]
  };# getCurrentEventId

  method getAttributeNamesForValue {value} {
    my variable value2attributes
    if {[info exists value2attributes($value)]} {
      return $value2attributes($value)
    }
    return {}
  };# getAttributeNamesForValue

  method getAttributeNamesFromAnnotation {ann} {
    set attr [tip_GetAttribute $ann $attribute]
    if {[llength $attr] == 0} {return {}}
    set value [tip_GetValueValue $attr]
    return [my getAttributeNamesForValue $value]
  };# getAttributeNamesFromAnnotation

  ##########################################################################
  ## Event: onButtonPress
  ##########################################################################
  method onButtonPress {win value bg fg} {
    if {![string length $display_object]} {return}
    set id [$display_object call annotateSelection $annotation $attribute \
              $value $bg $fg]
    if {[string length $automatic_annotator_object] && [string length $id]} {
      $automatic_annotator_object addPatternFromAnnotationId $id
    }
  };# onButtonPress

  ##########################################################################
  ## Event: onEntryEdit
  ##########################################################################
  method onEntryEdit {win value attribute} {
    my variable display_object
    if {![string length $display_object]} {return false}
    return true
  };# onEntryEdit

  ##########################################################################
  ## Event: onEnableWidgets
  ##########################################################################
  method onEnableWidgets {value} {
    my variable value2entry entry2attribute value2event event2attribute \
                display_object
    if {![string length $display_object]} {return}

    set ids [$display_object call getSelectedAnnotationIds]

    if {[info exists value2entry($value)]} {
      set win $value2entry($value)
      $win state !disabled
      $win delete 0 end
      if {[llength $ids]} {
        $win insert end [my getAttributeValue [lindex $ids 0] \
                                              $entry2attribute($win)]
      }
    }

    if {[info exists value2event($value)]} {
      set win $value2event($value)
      $win state !disabled
      $win configure -values [my getEvents]
      $win set [my getCurrentEventId]
      if {[llength $ids]} {
        $win set [my getAttributeValue [lindex $ids 0] $event2attribute($win)]
      }
    }
  };# onEnableWidgets

  ##########################################################################
  ## Event: onDisableWidgets
  ##########################################################################
  method onDisableWidgets {value} {
    my variable value2entry entry2attribute value2event event2attribute \
                display_object event_definer_object
    if {![string length $display_object]} {return}
    set doc [$display_object call getDocument]

    if {[info exists value2entry($value)]} {
      set ids [$display_object call getSelectedAnnotationIds]
      set win $value2entry($value)
      if {[llength $ids]} {
        set v [string trim [$win get]]
        my setAttributeValue [lindex $ids 0] $entry2attribute($win) $v
      }

      $win delete 0 end
      $win state disabled
    }

    if {[info exists value2event($value)]} {
      set ids [$display_object call getSelectedAnnotationIds]
      set win $value2event($value)
      if {[llength $ids]} {
        set v [string trim [lindex [$win get] 0]]
        my setAttributeValue [lindex $ids 0] $event2attribute($win) $v
        catch {$event_definer_object onRelateDocumentToEvent \
                                     $doc $value.$event2attribute($win) $v}
      }
      $win configure -values {}
      $win set {}
      $win state disabled
    }
  };# onDisableWidgets

  method getAttributeValue {id attribute} {
    my variable display_object
    if {![string length $display_object]} {return}
    set doc [$display_object call getDocument]
    if {![string length $doc]} {return {}}
    set ann  [tip_GetAnnotation $doc $id]
    set attr [tip_GetAttribute $ann $attribute]
    if {![llength $attr]} {return {}}
    return [tip_GetValueValue $attr]
  };# getAttributeValue

  method setAttributeValue {id attribute value} {
    my variable display_object
    if {![string length $display_object]} {return}
    set doc [$display_object call getDocument]
    if {![string length $doc]} {return {}}
    set ann  [tip_GetAnnotation $doc $id]

    ## Ensure that the value is not the same, before changing it...
    set attr [tip_GetAttribute $ann $attribute]
    if {[llength $attr]} {
      set v [tip_GetValueValue $attr]
      if {$v eq $value} {return}
    }

    set attr [tip_CreateAttribute $attribute \
                  [tip_CreateAttributeValue GDM_STRING $value]]
    set ann  [tip_PutAttribute $ann $attr]
    tip_AddAnnotation $doc $ann
  };# setAttributeValue

  ##########################################################################
  ## Event: onButtonAcceleratorChange
  ##########################################################################
  method onButtonAcceleratorChange {win} {
    if {![string length $display_object]} {return}
    set dialog_window [Dialog [::ELEP::WindowUtilities::UniqueWindowName \
      [$window_obj getframe].editAcceleratorDlg] -modal local -separator 1 \
      -title "Setting Button Accelerator..." -default 0 -cancel 1]
    $dialog_window add -textvariable GUI_Text(Ok)
    $dialog_window add -textvariable GUI_Text(Cancel)
    ## Try to retrieve any previously defined accelarator key.
    set accelerator_key {}
    foreach event [bind $win] {
        switch -glob -- $event {
          ?       {set accelerator_key $event}
          <Key-*> {set accelerator_key [string range $event 5 end-1]}
          default {}
        }
      }
    set frame [$dialog_window getframe]
      message $frame.label -width 250 -text \
       "Please press the keyboard keys combination you want to assign as an\
       accelerator to the button with value \"[$win cget -text]\""
      ttk::label $frame.accl -text "Key:" -anchor e
      ttk::label $frame.acc -textvariable [my varname accelerator_key] \
           -anchor w
      bind [winfo toplevel $frame] <Key> \
           [list set [my varname accelerator_key] %K]
      grid $frame.label -sticky snew -columnspan 2
      grid $frame.accl $frame.acc -sticky snew
    if {[$dialog_window draw] == 0} {
      set toplevel [winfo toplevel $win]
      ## Remove any previously set accelerators...
      foreach event [bind $win] {
        switch -glob -- $event {
          ? - <Key-*> { bind $win $event {}; bind $toplevel $event {} }
          default {}
        }
      }
      ## Add the binding...
      if {[string length $accelerator_key]} {
        bind $win      <Key-$accelerator_key> "$win invoke; $win flash"
        bind $toplevel <Key-$accelerator_key> "$win invoke; $win flash"
      }
    }
  };# onButtonAcceleratorChange

  ##########################################################################
  ## Event: invokeAutomaticAnnotator
  ##########################################################################
  method invokeAutomaticAnnotator {} {
    if {![string length $automatic_annotator_object]} {return}
    if {![string length $display_object]} {return}
    set doc [$display_object call getDocument]
    if {![string length $doc]} {return}
    set col [$display_object call getCollection]
    foreach id [$automatic_annotator_object applyAutomaticTagger $col $doc] {
      lappend anns [tip_GetAnnotation $doc $id]
    }
    if {[info exists anns]} {
      $display_object call annotateAnnotationSet $anns $annotation $attribute
    }
  };# invokeAutomaticAnnotator

  ##########################################################################
  ## Event: select
  ##########################################################################
  method select {value {id {}} {tag {}}} {
    my variable value2button
    $value2button($value) configure -relief sunken \
      -foreground [$value2button($value) cget -activeforeground] \
      -background [$value2button($value) cget -activebackground]
    my onEnableWidgets $value
    catch {$widget.client_area see $value2button($value)}
  };# select

  ##########################################################################
  ## Event: unselect
  ##########################################################################
  method unselect {value {id {}} {tag {}}} {
    my variable value2button value2entry default_button_fg default_button_bg
    $value2button($value) configure -relief raised \
      -foreground $default_button_fg -background $default_button_bg
    my onDisableWidgets $value
  };# unselect

  ##########################################################################
  ## Event: getValueColourPair
  ##########################################################################
  method getValueColourPair {value} {
    my variable value2button
    if {![info exists value2button($value)]} {
      my addValueButton $value
      catch {$automatic_annotator_object fillAutomaticAcquisitionPage}
    }
    return [list [$value2button($value) cget -activebackground] \
                 [$value2button($value) cget -activeforeground]]
  };# getValueColourPair

  ##########################################################################
  ## Event: getValues
  ##########################################################################
  method getValues {} {
    my variable value2button
    return [lsort [array names value2button]]
  };# getValues

  ##########################################################################
  ## Event: onDocumentChange
  ##########################################################################
  method onDocumentChange {col doc} {
    if {![string length $automatic_annotator_object]} {return}
    $automatic_annotator_object configure -annotation $annotation \
      -attribute $attribute
    $automatic_annotator_object    onDocumentChange $col $doc
    $annotation_highlighter_object onDocumentChange $col $doc
  };# onDocumentChange

  ##########################################################################
  ## Event: annotationWillBeDeleted
  ##########################################################################
  method annotationWillBeDeleted {id} {
    if {![string length $automatic_annotator_object]} {return}
    $automatic_annotator_object removePatternFromAnnotationId $id
  };# annotationWillBeDeleted

  ##########################################################################
  ## Event: annotationAttributeWillBeDeleted
  ##########################################################################
  method annotationAttributeWillBeDeleted {id attr} {
    my annotationWillBeDeleted $id
  };# annotationAttributeWillBeDeleted

  ##########################################################################
  ## Event: deletionHandled
  ##########################################################################
  method deletionHandled {ann_document ann_id ann_attribute ann_tag} {
    return 0
  };# deletionHandled

  method addScrollBindings {sw win} {
    bind $win <MouseWheel> \
       "$sw yview scroll \[expr {- (%D / 120) * 4}\] units"
    bind $win <4> "$sw yview scroll -5 units"
    bind $win <5> "$sw yview scroll  5 units"
  };# addScrollBindings

  method addScrollBindingsToWidgetList {sw args} {
    foreach win $args {my addScrollBindings $sw $win}
  };# addScrollBindingsToWidgetList

  method getAnnotation {} {return $annotation}
  method getAttribute  {} {return $attribute}

  destructor {
    catch {$colour_manager destroy}
    catch {$1_click_selector_object} destroy
    catch {$automatic_annotator_object destroy}
    foreach image $images {catch {image delete $image}}
  };# destructor

};# class ELEP::ViewerBase::ButtonAnnotator

package provide ELEP::ViewerBase::ButtonAnnotator 1.0
