##############################################################################
##  DocumentSelector.tcl:                                                   ##
## --------------------------------------------------                       ##
##  This class provides facilities for opening a Document in a viewer.      ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require TclOO
foreach ELEP::ViewerBase::image {ITOpen ITPrev ITNext ITSave} {
  GUI_LoadImage $ELEP::ViewerBase::image
}
unset ELEP::ViewerBase::image

oo::class create ELEP::ViewerBase::DocumentSelector {
  variable window_obj \
           toolbar \
           collection_id \
           collection \
           document_id \
           document \
           have_specific_collection \
           have_specific_document \
           object_owns_collection \
           object_owns_document \
           allow_collection_open \
           show_save_toolbar_button \
           notify_objects

  constructor {win {ntfy_objs {}} {col {}} {doc {}} {allow_col_open 0}
                                                    {show_save_tb 1}} {
    set window_obj    {};# An object that holds our toplevel window.
    set toolbar       {};# Our toolbar, which will be created in the toolbar
                         # frame of the parent window.
  
    set collection_id {};# Holds a collection name
    set collection    {};# Holds a collection object (like Col1)
  
    set document_id   {};# Holds a document id (name)
    set document      {};# Holds a document object (like Doc1)
  
    set have_specific_collection 0 ;# True if a specific col was given
    set have_specific_document   0 ;# True if a specific doc was given
    set object_owns_collection   0 ;# True if we have opened the collection
    set object_owns_document     0 ;# True if we have opened the document
    set allow_collection_open    0 ;# True if we can open a collection as a
                                   ;# whole
  
    set notify_objects {};# A list of objects to be notified when the current
                               # Document changes. The object must provide a
                               # method named "onDocumentChange".

    set allow_collection_open    $allow_col_open
    set show_save_toolbar_button $show_save_tb
    set window_obj $win
    set notify_objects $ntfy_objs
    if {[string equal $doc $::GUI_Text(All)]} {set doc {}}
    if {[string length $col] && ![string length $doc]} {
      ## The caller has specified a Collection.
      set have_specific_collection 1
      set collection $col
      # set document [tip_FirstDocument $col]
      set object_owns_document     1
    } elseif {![string length $col] && ![string length $doc]} {
      ## The caller hasn't specifiec any Collection/Document.
      set object_owns_collection 1
      set object_owns_document   1
    } else {
      # The caller has specified both a Collection and a Document.
      set collection $col
      set document   $doc
      set have_specific_collection 1
      set have_specific_document   1
    }
    if {$have_specific_collection} {set collection_id [tip_GetName $col]}
    if {$have_specific_document}   {set document_id   [tip_GetId $doc]}
    my createToolBar
    bind $toolbar <Destroy> "+if {\[string equal $toolbar %W\]}\
                 {::ELEP::WindowUtilities::DeleteTclOOObject [self]}"
  };# constructor

  method save {} {
    if {[my validDocument]} {
      tip_Sync $document
    }
  };# save

  method loadDocument {doc_id} {
    set index [lsearch -exact [$toolbar.doc cget -values] $doc_id]
    if {$index == -1} {return 0}
    $toolbar.doc current $index; my load
    return 1
  };# loadDocument

  method load {{mode current}} {
    set col [$toolbar.col current]
    set doc [$toolbar.doc current]
    switch $mode {
      prev {incr doc -1; if {$doc < 0} {return}}
      next {incr doc; if {$doc>=[llength [$toolbar.doc cget -values]]} {return}}
      default {}
    }
    ## Is the opened Document modified?
    my saveModifiedDocument?
    
    $toolbar.doc current $doc
    if {$object_owns_collection} {
      ## The collection is owned by the object. If it has been changed, close
      ## the old one and open the new one...
      set col [lindex [GUI_GetCollectionNames] $col]
      if {![string equal $col $collection_id]} {
        set collection_id $col
        catch {tip_Close $collection}
        set collection [tip_OpenCollection $collection_id]
        ## Is the Collection's Location in accordance with it's Name?
        if {![string equal [tip_GetName $collection] $collection_id]} {
          GUI_FixCollectionName $collection [tip_GetName $collection] \
                                $collection_id
        }
      }
    }

    ## Now we have the correct Collection. Open the Document...
    catch {tip_Close $document}
    if {$allow_collection_open && $doc == 0} {
      ## This is a request to open a whole collection. So, don't open a
      ## document!
      set document {}
    } else {
      set document_id [lindex [$toolbar.doc cget -values] $doc]
      set document [tip_GetDocument $collection $document_id]
    }
    my notifyListeners
    my saveState
  };# load

  method validDocument {} {
    llength [info commands $document]
  };# validDocument

  method createToolBar {} {
    global GUI_Text GUI_Options
    set toolbarFrame [$window_obj getToolbarFrame]
    set toolbar [ttk::frame $toolbarFrame.opendocument -class Toolbar]
    if {$show_save_toolbar_button} {
      ttk::button $toolbar.save -image ITSave -textvariable GUI_Text(Save) \
        -command "[self] save" -style Toolbutton
    }
    if {!$have_specific_document} {
      ## Place these buttons only if we don't have a specific document.
      ttk::label $toolbar.col_l -textvariable GUI_Text(Collection)
      ttk::combobox $toolbar.col -state readonly -exportselection 0 \
        -width $GUI_Options(SystemExprToolbarCollectionWidth)
      ttk::label $toolbar.doc_l -textvariable GUI_Text(Document)
      ttk::combobox $toolbar.doc -state readonly -exportselection 0 \
        -width $GUI_Options(SystemExprToolbarDocumentWidth)
      ttk::button $toolbar.open -textvariable GUI_Text(Open) -image ITOpen \
        -command "[self] load" -style Toolbutton
      ttk::separator $toolbar.sep -orient vertical
      ttk::button $toolbar.prev -textvariable GUI_Text(Open) -image ITPrev \
        -command "[self] load prev" -style Toolbutton
      ttk::button $toolbar.next -textvariable GUI_Text(Open) -image ITNext \
        -command "[self] load next" -style Toolbutton
      ttk::separator $toolbar.sep2 -orient vertical
      if {$have_specific_collection} {
        $toolbar.col set [file tail [tip_GetName $collection]]
        $toolbar.col configure -state disabled
        set values [$collection GetFileIds]
        if {$allow_collection_open} {
          set values [linsert $values 0 [::ELEP::WindowUtilities::msg All]]
        }
        $toolbar.doc configure -values $values
        $toolbar.doc current 0
      } else {
        $toolbar.col configure -postcommand "[self] getCollectionNames"
        bind $toolbar.col <<ComboboxSelected>> "[self] updateDocumentCombobox"
        catch {my restoreState}
      }
      if {$show_save_toolbar_button} {
        grid $toolbar.col_l $toolbar.col $toolbar.doc_l $toolbar.doc \
             $toolbar.open $toolbar.sep $toolbar.prev $toolbar.next \
             $toolbar.sep2 $toolbar.save -sticky news -padx 2 -pady 2
      } else {
        grid $toolbar.col_l $toolbar.col $toolbar.doc_l $toolbar.doc \
             $toolbar.open $toolbar.sep $toolbar.prev $toolbar.next \
             -sticky news -padx 2 -pady 2
      }
    } else {
      if {$show_save_toolbar_button} {
        grid $toolbar.save -sticky news -padx 2 -pady 2
      }
    }

    pack $toolbar -side left
  };# createToolBar

  method getCollectionNames {} {
    $toolbar.col configure -values [GUI_GetCollectionNameTails]
  };# getCollectionNames

  method getDocumentNames {col} {
    set values [gut_GetCollectionDocuments $col]
    if {$allow_collection_open} {
      set values [concat [list [::ELEP::WindowUtilities::msg All]] $values]
    }
    return $values
  };# getDocumentNames

  method updateDocumentCombobox {} {
    set col [lindex [GUI_GetCollectionNames] [$toolbar.col current]]
    $toolbar.doc configure -values [my getDocumentNames $col]
    $toolbar.doc current 0
  };# updateDocumentCombobox

  method registerListener {obj} {
    lappend notify_objects $obj
  };# registerListener

  method notifyListeners {} {
    foreach obj $notify_objects {
      if {[catch {$obj onDocumentChange $collection $document} err]} {
        catch {puts "$obj [info object class $obj]"}
        puts stderr "DocumentSelector::notifyListeners: $err\n$::errorInfo"
      }
    }
  };# notifyListeners

  method saveModifiedDocument? {{mode yesnocancel}} {
    if {[my validDocument] && [$document Modified]} {
      set ans [gut_tk_dialog {Modified Document:} [format \
        $::GUI_Text(SaveModifiedDocument?) [tip_GetId $document]] $mode \
        question [$window_obj getTopLevel]]
      switch $ans {
        cancel  {return} no {} yes {my save}
        default {error "Unknown answer \"$ans\""}
      }
    }
  };# saveModifiedDocument?

  destructor {
    ## Is the opened Document modified?
    if {$show_save_toolbar_button} {
      my saveModifiedDocument? yesno
    }
    if {$object_owns_document}   {catch {tip_Close $document}}
    if {$object_owns_collection} {catch {tip_Close $collection}}
    if {[winfo exists $toolbar]} {destroy $toolbar}
  };# destructor

  method show {} {
    if {$have_specific_document} {my notifyListeners}
  };# show

  method saveState {} {
    if {![winfo exists $toolbar.col]} {return}
    set ::GUI_Options([self class]-$allow_collection_open) \
      [list [$toolbar.col get] [$toolbar.doc get]]
  };# saveState

  method restoreState {} {
    if {![info exists ::GUI_Options([self class]-$allow_collection_open)]} {
      return
    }
    foreach {col doc} $::GUI_Options([self class]-$allow_collection_open) {
      break
    }
    $toolbar.col set $col
    my getCollectionNames
    my updateDocumentCombobox
    $toolbar.doc set $doc
  };# restoreState
};# class ELEP::ViewerBase::DocumentSelector

package provide ELEP::ViewerBase::DocumentSelector 1.0
