##############################################################################
##  EventDefiner.tcl:                                                       ##
## --------------------------------------------------                       ##
##  This class implements a widget that can be used to define "events".     ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities::WidgetBase
package require ELEP::WindowUtilities::DialogWindow
package require widget::scrolledwindow
package require tablelist_tile
package require ELEP::Database::TDBC

oo::class create ELEP::ViewerBase::EventDefiner {
  superclass ELEP::WindowUtilities::WidgetBase
  mixin      ELEP::Database::TDBC

  ##########################################################################
  ## Method: constructor
  ##########################################################################
  constructor {parent_win args} {
    my init_database_object
    next ttk::frame [::ELEP::WindowUtilities::UniqueWindowName \
                             $parent_win.eventDefiner]
    my variable default_set default_set_id events_editor

    set default_set       sync3
    set default_set_id    {}
    set events_editor     {}

    my variable current_event_id current_event_title current_event_description \
                current_event_primary current_event_documents \
                current_event_iptc
    set current_event_id            {}
    set current_event_title         {}
    set current_event_iptc          {}
    set current_event_description   {}
    set current_event_primary        0
    set current_event_documents {}

    my variable events description_widget statusbar_text widget hourglass_window
    set events             {}
    set description_widget {}
    set statusbar_text     {}
    set hourglass_window   [winfo toplevel $widget].hourglass_window

    my configure {*}$args
    my init
    my createWidgets
    my updateEvents
    my onEventSelectionChanged
  };# constructor

  method init {} {
    my variable driver
    switch $driver {
      sqlite3 {catch {my initdb}}
    }
  };# init

  method initdb {} {
    my showProgress
    ## Table "sets": Stores the event sets in the database.
    my sql {
      CREATE TABLE sets (
        id            INTEGER PRIMARY KEY,
        description   TEXT    UNIQUE      NOT NULL,
        next_event_id INTEGER NOT NULL    DEFAULT '0'
      )
    } 1
    my setProgress 30
    ## Table "events": Stores the list of events.
    my sql {
      CREATE TABLE events (
        event_id       INTEGER NOT NULL,
        set_id         INTEGER NOT NULL,
        primary_event  INTEGER NOT NULL,
        title          TEXT,
        description    TEXT,
        iptc_code      TEXT,
        PRIMARY KEY (event_id, set_id)
      )
    } 1
    my setProgress 60
    ## Table "doc2event": Associates document with one or more events.
    my sql {
      CREATE TABLE doc2event (
        collection     TEXT,
        document       TEXT,
        attribute      TEXT,
        event_id       TEXT,
        set_id         INTEGER NOT NULL,
        PRIMARY KEY    (collection, document, attribute, set_id)
      )
    } 1
    my hideProgress
  };# initdb

  method resetdb {} {
    my showProgress
    foreach table {events doc2event} {
      my sql "DROP TABLE $table"
    }
    my hideProgress
  };# resetdb

  method reset {} {
    my resetdb
    my initdb
  };# reset

  method findSet {descr} {
    my showProgress
    set id [my first_element {SELECT id FROM sets WHERE description = :descr}]
    my setProgress 40
    if {$id eq ""} {
      ## Not found. Add it...
      my sql {INSERT INTO sets(description) VALUES( :descr )}
      my setProgress 70
      set id [my first_element \
                        {SELECT id FROM sets WHERE description = :descr}]
    }
    my hideProgress
    return $id
  };# findSet

  method updateEvents {} {
    my variable events database default_set_id
    set list [list]
    my showProgress
    my foreach -as lists -- row \
       {SELECT event_id, primary_event, description FROM events WHERE\
                           set_id = :default_set_id} {
      lappend list $row
    }
    my hideProgress
    set events [lsort -integer -index 0 $list]
  };# updateEvents

  method updateEventDocument {} {
    my variable current_event_id current_event_documents database default_set_id
    set list [list]
    my showProgress
    my foreach -as lists -- row \
                   {SELECT collection, document, attribute FROM doc2event\
                    WHERE set_id = :default_set_id AND\
                        event_id = :current_event_id} {
      lappend list $row
    }
    my hideProgress
    set current_event_documents $list
  };# updateEventDocument

  method getEventIds {} {
    my variable database default_set_id
    my showProgress
    set result [lsort -index 0 [my sql {SELECT event_id FROM events WHERE\
                                     set_id = :default_set_id}]]
    my hideProgress
    return $result
  };# getEventIds

  method getEventIdsAndTitles {} {
    my variable database default_set_id
    my showProgress
    set result \
        [lsort -index 0 [my sql {SELECT event_id, description FROM events WHERE\
                                     set_id = :default_set_id}]]
    my hideProgress
    return $result
  };# getEventIdsAndTitles

  method getCurrentEventId {} {
    my variable current_event_id
    return $current_event_id
  };# getCurrentEventId

  method getNextEventId {} {
    my variable database default_set_id
    my showProgress
    set id [my first_element {SELECT next_event_id FROM sets WHERE\
                                                   id = :default_set_id}]
    incr id
    my setProgress 50
    my sql {UPDATE sets SET next_event_id = :id WHERE id = :default_set_id}
    my hideProgress
    return $id
  };# getNextEventId

  method createWidgets {} {
    my variable widget default_set default_set_id

    ## Create a combobox, where the user can select, add or delete an event
    ## set...
    if {$default_set eq ""} {
      my createWidgets_setEditor $widget
    } else {
      set default_set_id [my findSet $default_set]
    }

    ## Create a listbox, where an event can be selected...
    my createWidgets_eventSelector $widget

    grid columnconfigure $widget 0 -weight 1
    grid rowconfigure    $widget 0 -weight 1
    grid rowconfigure    $widget 1 -weight 1
    grid rowconfigure    $widget 2 -weight 1
  };# createWidgets

  method createWidgets_setEditor {widget} {
  };# createWidgets_setEditor

  method createWidgets_eventSelector {widget} {
    my variable tablelist
    widget::scrolledwindow $widget.essw ;# -scrollbar vertical -auto vertical
    set tablelist $widget.essw.tbl
    tablelist::tablelist $tablelist -height 8 -width 33 \
        -columns {3 ID 3 P 30 Title} -stretch {2} \
        -stripebackground #e0e8f0 -background gray98 \
        -exportselection 0 -listvariable [my varname events]
    $tablelist columnconfigure 0 -sortmode dictionary
    $tablelist columnconfigure 1 -sortmode dictionary
    $tablelist columnconfigure 2 -sortmode dictionary
    bind $tablelist <<TablelistSelect>> [list [self] onEventSelectionChanged]
    $widget.essw setwidget $tablelist

    ttk::button $widget.addev  -image [my loadImage ITAdd] \
        -command [list [self] onAddEvent]
    ttk::button $widget.viewev -image [my loadImage ITEdit] \
        -command [list [self] onEditEvent]
    ttk::button $widget.delev  -image [my loadImage ITRemove] \
        -command [list [self] onDeleteEvent]
    grid $widget.essw -rowspan 3 -sticky snew
    grid $widget.addev -row 0 -column 1 -sticky snew
    grid $widget.viewev -row 1 -column 1 -sticky snew
    grid $widget.delev -row 2 -column 1 -sticky snew
  };# createWidgets_eventSelector

  method createEventEditor {} {
    my variable widget events_editor description_widget
    ## The editor object does not exist...
    set events_editor [ELEP::WindowUtilities::DialogWindow create \
                             $widget.events_editor_dlg]
    $events_editor title [my msg {Event Editor}]

    $events_editor addStdButton ok     [list [self] onOk]
    $events_editor addStdButton update [list [self] onUpdate]
    $events_editor addStdButton cancel [list [self] onCancel]

    ## Fill the client area...
    set frame [$events_editor getframe]
      pack [ttk::labelframe $frame.event -text [my msg {Event Details}]] \
        -fill x -expand 0 -padx 2 -pady 2
        grid [ttk::label $frame.event.id  -text [my msg {Id}]:] \
             [ttk::entry $frame.event.idv -textvariable \
                [my varname current_event_id] -state readonly] \
          -sticky snew -padx 2 -pady 2
        grid [ttk::label $frame.event.title  -text [my msg {Title}]:] \
             [ttk::entry $frame.event.titlev -textvariable \
                [my varname current_event_title]] \
          -sticky snew -padx 2 -pady 2
        grid [ttk::label $frame.event.iptc  -text [my msg {IPTC codes}]:] \
             [ttk::entry $frame.event.iptcv -textvariable \
                [my varname current_event_iptc]] \
          -sticky snew -padx 2 -pady 2
        grid [ttk::checkbutton $frame.event.primaryv \
          -text [my msg {Primary Event}] -onvalue 1 -offvalue 0 \
          -variable [my varname current_event_primary]] \
          -columnspan 2 -sticky snew -padx 2 -pady 2
        grid columnconfigure $frame.event 1 -weight 1
      pack [ttk::panedwindow $frame.p -orient vertical] \
        -fill both -expand 1 -padx 2 -pady 2

      ttk::labelframe $frame.p.d -text [my msg {Event Description}]
        # Add a scrolled window, which will host the display widget.
        set description_widget $frame.p.d.sw.text
        widget::scrolledwindow $frame.p.d.sw
        $frame.p.d.sw setwidget [text $description_widget -borderwidth 0 \
           -relief flat -height 10 -width 110]
        pack $frame.p.d.sw -fill both -expand 1 -padx 2 -pady 2

      ttk::labelframe $frame.p.docs -text [my msg {Documents related to event}]
        set event_documents_widget $frame.p.docs.sw.list
        widget::scrolledwindow $frame.p.docs.sw
        tablelist::tablelist $event_documents_widget -height 10 -width 110 \
            -columns {50 Collection 30 Document 20 Attribute} -stretch all \
            -stripebackground #e0e8f0 -background gray98 -exportselection 0\
            -listvariable [my varname current_event_documents]
        $event_documents_widget columnconfigure 0 -sortmode dictionary
        $event_documents_widget columnconfigure 1 -sortmode dictionary
        $event_documents_widget columnconfigure 2 -sortmode dictionary
        $frame.p.docs.sw setwidget $event_documents_widget
        pack $frame.p.docs.sw -fill both -expand 1 -padx 2 -pady 2
      $frame.p add $frame.p.d
      $frame.p add $frame.p.docs
  };# createEventEditor

  method showEditor {} {
    my variable events_editor
    if {[catch {$events_editor show}]} {
      my createEventEditor
    }
    $events_editor show
  };# showEditor

  #############################################################################
  ## Event: onEventSelectionChanged
  #############################################################################
  method onEventSelectionChanged {} {
    my variable tablelist default_set_id database description_widget \
                current_event_id current_event_title current_event_description \
                current_event_primary current_event_documents \
                current_event_iptc
    set selection [$tablelist curselection]
    if {![llength $selection]} {
      set current_event_id          {}
      set current_event_title       {}
      set current_event_iptc        {}
      set current_event_description {}
      set current_event_documents   {}
      set current_event_primary      0
      return
    }
    set row [lindex $selection 0]
    set row [lindex [$tablelist get $row $row] 0]
    foreach {current_event_id current_event_primary current_event_description} \
             $row {break}
    ## Update the description...
    my showProgress
    set result [my sql "SELECT title, iptc_code FROM events WHERE\
                event_id = :current_event_id AND set_id = :default_set_id"]
    lassign [lindex $result 0] current_event_title current_event_iptc
    my hideProgress
    if {[winfo exists $description_widget]} {
      $description_widget delete 1.0 end
      $description_widget insert end $current_event_description
    }
    my updateEventDocument
  };# onEventSelectionChanged

  #############################################################################
  ## Event: onAddEvent
  #############################################################################
  method onAddEvent {} {
    my variable current_event_id current_event_title current_event_primary \
                current_event_description description_widget \
                current_event_documents current_event_iptc
    set current_event_id            {}
    set current_event_title         {}
    set current_event_iptc          {}
    set current_event_description   {}
    set current_event_primary        1
    set current_event_documents {}
    if {[winfo exists $description_widget]} {
      $description_widget delete 1.0 end
      $description_widget insert end $current_event_description
    }
    my showEditor
  };# onAddEvent

  #############################################################################
  ## Event: onEditEvent
  #############################################################################
  method onEditEvent {} {
    my showEditor
    my onEventSelectionChanged
  };# onEditEvent

  #############################################################################
  ## Event: onDeleteEvent
  #############################################################################
  method onDeleteEvent {} {
    my variable current_event_id current_event_title current_event_description \
                description_widget default_set_id database
    if {[gut_question [my msg {Delete selected event?}]] ne "yes"} {return}
    my showProgress
    my sql "DELETE FROM events WHERE\
                   event_id = :current_event_id AND set_id = :default_set_id"
    my hideProgress
    my updateEvents
  };# onDeleteEvent

  #############################################################################
  ## Event: onOk
  #############################################################################
  method onOk {} {
    my onUpdate
    my onCancel
  };# onOk

  #############################################################################
  ## Event: onUpdate
  #############################################################################
  method onUpdate {} {
    my variable current_event_id current_event_title current_event_description \
                current_event_primary description_widget \
                current_event_documents default_set_id database \
                current_event_iptc
    if {![winfo exists $description_widget]} {return}
    set current_event_description \
       [string trim [$description_widget get 1.0 end]]
    set current_event_title [string trim $current_event_title]
    if {![string length $current_event_title]} {
      set current_event_title [string range $current_event_description 0 100]
    }
    ## Is this id already in the database?
    my showProgress
    set id {}
    if {$current_event_id ne ""} {
      set id [my first_element {SELECT event_id FROM events WHERE\
                event_id = :current_event_id AND set_id = :default_set_id}]
    }
    my setProgress 30
    if {![llength $id]} {
      set id [my getNextEventId]
      ## Not found. Add it...
      my sql {INSERT INTO events(event_id, set_id, primary_event,\
            title, description, iptc_code) VALUES( :id , :default_set_id ,\
            :current_event_primary , :current_event_title ,\
            :current_event_description , :current_event_iptc )}
    } else {
      ## Update the row...
      set id [lindex $id 0]
      my sql {UPDATE events SET title = :current_event_title ,\
        description = :current_event_description ,\
        iptc_code = :current_event_iptc ,\
        primary_event = :current_event_primary WHERE\
        event_id = :current_event_id AND set_id = :default_set_id}
    }
    my setProgress 60
    set current_event_id [lindex \
       [my first_element {SELECT event_id FROM events WHERE\
          set_id = :default_set_id AND\
          primary_event = :current_event_primary AND\
          title = :current_event_title AND\
          description = :current_event_description AND\
          iptc_code = :current_event_iptc}] 0]
    my hideProgress
    my updateEvents
  };# onUpdate

  #############################################################################
  ## Event: onCancel
  #############################################################################
  method onCancel {} {
    my variable events_editor
    $events_editor hide
  };# onCancel

  ##########################################################################
  ## Event: onDocumentChange
  ##########################################################################
  method onDocumentChange {col doc} {
    my updateEvents
  };# onDocumentChange

  ##########################################################################
  ## Event: onRelateDocumentToEvent
  ##########################################################################
  method onRelateDocumentToEvent {doc attribute event} {
    my variable default_set_id database
    if {$doc eq ""} {return}
    set col        [tip_GetParent $doc]
    set collection [tip_GetName   $col]
    set document   [tip_GetId     $doc]
    set event      [string trim $event]
    set collection [file tail $collection]

    my showProgress
    set exists [my sql {SELECT event_id, attribute FROM doc2event WHERE\
        collection = :collection AND document = :document AND\
        attribute = :attribute AND set_id = :default_set_id}]
    my setProgress 40
    if {$exists eq ""} {
      ## We must add the entry...
      my sql {INSERT INTO\
                doc2event(collection,  document,   attribute, event_id, set_id)\
                VALUES( :collection , :document , :attribute ,\
                        :event , :default_set_id)}
    } else {
      lassign [lindex $exists 0] db_event _
      if {$db_event eq $event} {
        my hideProgress
        return
      }
      my sql {UPDATE doc2event SET event_id = :event WHERE\
                     collection = :collection AND document = :document AND\
                     attribute  = :attribute  AND set_id = :default_set_id}
    }
    my hideProgress
  };# onRelateDocumentToEvent

  method showProgress  {} {
    my variable toplevel_object statusbar_text widget hourglass_window driver
    if {![info exists toplevel_object] || $toplevel_object eq ""} {
      set toplevel_object [my getToplevelObject]
    }

    switch $driver {
      sqlite3 {}
      default {
        if {![winfo exists $hourglass_window]} {
          label $hourglass_window -background white \
            -image [my loadImage HourglassWindow]
        }
        place $hourglass_window -anchor c -relx 0.5 -rely 0.5
      }
    }
    if {$toplevel_object eq ""} {return}
    set statusbar_text [$toplevel_object getStatusText]
    $toplevel_object setStatusText [my msg {Communicating with Database...}]
    $toplevel_object setProgress 0
    $toplevel_object showProgress
    update
  };# showProgress
  method hideProgress  {}    {
    my variable toplevel_object statusbar_text hourglass_window driver
    switch $driver {
      sqlite3 {}
      default {place forget $hourglass_window}
    }
    if {$toplevel_object eq ""} {return}
    $toplevel_object setProgress 100
    $toplevel_object setStatusText $statusbar_text
    $toplevel_object hideProgress
    update
  };# hideProgress
  method setProgress {val} {
    my variable toplevel_object
    if {$toplevel_object eq ""} {return}
    $toplevel_object setProgress $val
    update
  };# setProgress

  destructor {
    catch {my disconnect}
    next
  };# destructor

};# class ELEP::ViewerBase::EventDefiner

package provide ELEP::ViewerBase::EventDefiner 1.0
