##############################################################################
##  HTMLTagBase.tcl:                                                        ##
## --------------------------------------------------                       ##
##  This class provides facilities for adding tags in an HTML widget.       ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

oo::class create ELEP::ViewerBase::HTMLTagBase {
  variable html_widget \
           html_source \
           selection_state \
           selection_start

  constructor {widget} {
    set html_widget {};# The TkHtml widget we are linked with.
    set html_source {};# The HTML source currently rendered in the widget.
    set selection_state 0
    set selection_start {}

    set html_widget $widget
    bind $html_widget <Destroy> "+if {\[string equal $html_widget %W\]}\
                 {::ELEP::WindowUtilities::DeleteTclOOObject [self]}"
    bind $html_widget.x <Motion>          "+[self] onMotion %x %y"
    bind $html_widget.x <ButtonPress-1>   "+[self] onPress %x %y"
    bind $html_widget.x <ButtonRelease-1> "+[self] onRelease %x %y"
  };# constructor

  method setBase {path} {
    $html_widget configure -base $path
  };# setBase

  method getBase {} {
    return [$html_widget cget -base]
  };# getBase

  method index {args} {
    $html_widget index {*}$args 
  };# index

  method insert {index chars args} {
    $html_widget parse $chars
  };# insert

  method delete {args} {
    $html_widget clear
  };# delete

  method tag {option {tagname {}} args} {
    #puts "tag $option $tagname $args"
    switch -exact $option {
      add        {
        # puts "tag add $tagname [lindex $args 0] [lindex $args 1]"
        # puts "    [lindex $args 0] -> [tagNormaliseIndex [lindex $args 0]]"
        # puts "    [lindex $args 1] -> [tagNormaliseIndex [lindex $args 1]]"
        return [eval $html_widget tag add [list $tagname] \
                     [my tagNormaliseIndex [lindex $args 0]] \
                     [my tagNormaliseIndex [lindex $args 1]]]
      }
      bind      {return [eval $html_widget tag bind [list $tagname] $args]}
      cget      {return [eval $html_widget tag cget [list $tagname] $args]}
      configure {return [eval $html_widget tag configure [list $tagname] $args]}
      delete    {return [$html_widget tag delete $tagname]}
      lower     {catch {$html_widget tag lower $tagname}; return}
      names     {return [$html_widget tag names]}
      nextrange {return [eval $html_widget tag nextrange [list $tagname] $args]}
      prevrange {return [eval $html_widget tag prevrange [list $tagname] $args]}
      raise     {catch {$html_widget tag raise $tagname}; return}
      ranges    {return [$html_widget tag ranges $tagname]}
      remove    {return [eval $html_widget tag remove [list $tagname] $args]}
    }
    return {}
  };# tag

  method tagNormaliseIndex {index} {
    switch -glob $index {
      1.0+*c {return [$html_widget characterOffsetToIndex \
                                   [string range $index 4 end-1]]}
      default {return $index}
    }
  };# tagNormaliseIndex

  method index2CharacterOffset {index} {
    return [$html_widget indexToCharacterOffset $index]
  };#index2CharacterOffset

  method get {index1 {index2 {}}} {
    set from [$html_widget indexToCharacterOffset $index1]
    if {[string length $index2]} {
      set to [$html_widget indexToCharacterOffset $index2]
    } else {
      set to [expr {$from+1}]
    }
    return [string range $html_source $from $to]
  };# get

  method validTextWidget {} {return [winfo exists $html_widget]}

  method getWidgetForManagingBindings {} {return $html_widget.x}

  method getIndexAtxy {x y} {
    return [$html_widget index @$x,$y]
  };# getIndexAtxy

  method selectionExists {} {
    if {![catch {$html_widget index sel.first} first] &&
         [string length $first]} {
      return 1
    }
    return 0
  };# selectionExists

  method selectionClear {} {
    $html_widget selection clear
  };# selectionClear

  method getSelectionStartIndex {} {
    return [$html_widget index sel.first]
  };# getSelectionStartIndex

  method getSelectionEndIndex {} {
    return [$html_widget index sel.last]
  };# getSelectionEndIndex

  method setSelectionIndexes {first last} {
    $html_widget selection clear
    $html_widget selection set [my tagNormaliseIndex $first] \
                               [my tagNormaliseIndex $last]
  };# setSelectionIndexes

  method addSelectionIndexes {first last} {
    ## The Html widget does not support many selection indexes. So, we must
    ## keep the minimum first & the maximum last indexes...
    set n_first [my tagNormaliseIndex $first]
    set n_last  [my tagNormaliseIndex $last]
    if {[my selectionExists]} {
      set s_first [my getSelectionStartIndex]
      set s_last  [my getSelectionEndIndex]
      if {$s_first < $n_first} {set n_first $s_first}
      if {$s_last  > $n_last } {set n_last  $s_last}
    }
    $html_widget selection set $n_first $n_last
  };# addSelectionIndexes

  method seeIndex {index} {
    # puts "seeIndex: index=$index, [$html_widget coords $index]"
    foreach {x y}   [$html_widget coords $index] {break}
    foreach {mx my} [$html_widget coords]        {break}
    catch {$html_widget yview moveto [expr {1.0*$y/$my}]}
    catch {$html_widget xview moveto [expr {1.0*$x/$mx}]}
  };# seeIndex

  method setFocus {} {
    focus $html_widget
  };# setFocus

  ## Selection callbacks...
  method onPress {x y} {
    catch {
      $html_widget selection clear
      set selection_start [$html_widget index @$x,$y]
      set selection_state 1
    }
  };# press

  method onRelease {x y} {set selection_state 0}

  method onMotion {x y} {
    if {0 == $selection_state} {return}
    catch {
      $html_widget selection set $selection_start [$html_widget index @$x,$y]
      selection own $html_widget
    }
  };# motion

};# class ELEP::ViewerBase::HTMLTagBase

package provide ELEP::ViewerBase::HTMLTagBase 1.0
