##############################################################################
##  HTMLTagBase.tcl:                                                        ##
## --------------------------------------------------                       ##
##  This class provides facilities for adding tags in an HTML widget.       ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

oo::class create ELEP::ViewerBase::HTMLTagBase {
  variable html_widget {};# The TkHtml widget we are linked with.
  variable html_source {};# The HTML source currently rendered in the widget.

  variable Offset2Node {};# This is a sorted list, which maps from specific
                          # character offsets to character nodes.
  variable Offset2NodeLen;
  variable Node2Offset   ;# This is an array, which maps from nodes to
                          # character offsets.

  variable myState 0     ;# True when left-button is held down
  variable myFromNode
  variable myFromIdx

  variable myToNode
  variable myToIdx

  variable baseURL     {};# Used for keeping the base URL

  variable tagNames      ;# An array which holds all tags added to the widget

  common character_pattern        {\A(?:\s+)|(?:\&\w{2,8}\;)|(?:\S)}
  common character_entity_pattern {\&\w{2,8}\;}
  common character_space_pattern  {\s+}
  common character_pattern2       {(?:\s+)|(?:\&\w{2,8}\;)}

  constructor {widget} {
    set html_widget $widget
    bind $html_widget <Destroy> "+if {\[string equal $html_widget %W\]}\
                 {::ELEP::WindowUtilities::DeleteTclOOObject [self]}"
    ## Mouse bindings for selection handling...
    bind $html_widget <Motion>          "+[self] motion %x %y"
    bind $html_widget <ButtonPress-1>   "+[self] press %x %y"
    bind $html_widget <ButtonRelease-1> "+[self] release %x %y"
    ## Stylesheets...
    $html_widget handler script style  [list [self] onStyleTag]
    ## Scripts...
    $html_widget handler script script [list [self] onScriptTag]
    ## Text node parsing...
    $html_widget handler parse {} [list [self] onTextNodeParse]
  };# constructor

  ##########################################################################
  ## Method: setBase
  ##########################################################################
  method setBase {path} {
    set baseURL $path
  };# setBase

  ##########################################################################
  ## Method: getBase
  ##########################################################################
  method getBase {} {
    return $baseURL
  };# getBase

  ##########################################################################
  ## Method: index
  ##########################################################################
  method index {index args} {
    puts "index: $index $args"
    return $index
  };# index

  ##########################################################################
  ## Method: insert
  ##########################################################################
  method insert {index chars args} {
    setHTML $chars
  };# insert

  ##########################################################################
  ## Method: delete
  ##########################################################################
  method delete {args} {
    reset
  };# delete

  ##########################################################################
  ## Method: tag
  ##########################################################################
  method tag {option {tagname {}} args} {
    # puts "tag $option $tagname $args"
    switch -exact $option {
      add        {
        #  puts "tag add $tagname [lindex $args 0] [lindex $args 1]"
        #  puts "    [lindex $args 0] -> [tagNormaliseIndex [lindex $args 0]]"
        #  puts "    [lindex $args 1] -> [tagNormaliseIndex [lindex $args 1]]"
        #  set 1 [tagNormaliseIndex [lindex $args 0]]
        #  set 2 [tagNormaliseIndex [lindex $args 1]]
        #  if {[lindex $1 end] >= [lindex $2 end]} {
        #    puts  ==================================
        #  }
        #  puts "'[get $1 $2]'"
        return [eval $html_widget tag add [list $tagname] \
                     [tagNormaliseIndex [lindex $args 0]] \
                     [tagNormaliseIndex [lindex $args 1]]]
      }
      bind      {return [eval $html_widget tag bind [list $tagname] $args]}
      cget      {return [eval $html_widget tag cget [list $tagname] $args]}
      configure {return [eval $html_widget tag configure [list $tagname] $args]}
      delete    {return [$html_widget tag delete $tagname]}
      lower     {catch {$html_widget tag lower $tagname}; return}
      names     {return [array names tagNames]}
      nextrange {return [eval $html_widget tag nextrange [list $tagname] $args]}
      prevrange {return [eval $html_widget tag prevrange [list $tagname] $args]}
      raise     {catch {$html_widget tag raise $tagname}; return}
      ranges    {return [$html_widget tag ranges $tagname]}
      remove    {return [eval $html_widget tag remove [list $tagname] $args]}
    }
    return {}
  };# tag

  ##########################################################################
  ## Method: tagNormaliseIndex
  ##########################################################################
  method tagNormaliseIndex {index} {
    switch -glob $index {
      1.0+*c  {return [characterOffset2Index [string range $index 4 end-1]]}
      1.0     {return [characterOffset2Index 0]}
      end     {return [characterOffset2Index end]}
      default {return $index}
    }
  };# tagNormaliseIndex

  ##########################################################################
  ## Method: characterOffset2Index
  ##########################################################################
  method characterOffset2Index {offset} {
    if {$offset == 0 || [string equal $offset end]} {
      set node [lindex $Offset2Node $offset 1]
      return [list $node 0]
    }
    ## Locate the node that contains to the given offset:
    foreach {off node} [locateNode $offset] {break}
    if {$offset == $off} {return [list $node 0]}
    set text [string range $html_source $off $offset]
    ## Substitute any possible HTML entities and successive spaces...
    regsub -all $character_entity_pattern $text XX  text
    regsub -all $character_space_pattern  $text { } text
    #regsub -all $character_pattern2       $text { } text
    # puts "'[string range $html_source $off $offset]' -> '$text'"
    ## Since indexes start from 0, don't forget to remove 1 from length!
    return [list $node [expr {[string length $text]-1}]]
  };# characterOffset2Index

  ##########################################################################
  ## Method: index2CharacterOffset
  ##########################################################################
  method index2CharacterOffset {index} {
    foreach {node idx} $index {break}
    set offset [locateOffset $node]
    # Now that we have an index, we have to advance idx "characters": a
    # "character" can be a set of succesive white space characters, an html
    # entity {\&\w{2,8}\;} or a plain character. So, use a regular expression
    # for matching!
    for {} {$idx > 0} {incr idx -1} {
      incr offset [string length [lindex [regexp -start $offset -inline \
                                            $character_pattern $html_source] 0]]
    }
    return $offset
  };# index2CharacterOffset

  ##########################################################################
  ## Method: get
  ##########################################################################
  method get {index1 {index2 {}}} {
    set from [index2CharacterOffset $index1]
    if {[string length $index2]} {
      set to [index2CharacterOffset $index2]; incr to -1
    } else {
      set to [expr {$from+1}]
    }
    return [string range $html_source $from $to]
  };# get

  ##########################################################################
  ## Method: validTextWidget
  ##########################################################################
  method validTextWidget {} {return [winfo exists $html_widget]}

  ##########################################################################
  ## Method: getWidgetForManagingBindings
  ##########################################################################
  method getWidgetForManagingBindings {} {return $html_widget}

  ##########################################################################
  ## Method: getIndexAtxy
  ##########################################################################
  method getIndexAtxy {x y} {
    return [$html_widget node -index $x $y]
  };# getIndexAtxy

  ##########################################################################
  ## Method: selectionExists
  ##########################################################################
  method selectionExists {} {
  };# selectionExists

  ##########################################################################
  ## Method: selectionClear
  ##########################################################################
  method selectionClear {} {
  };# selectionClear

  ##########################################################################
  ## Method: getSelectionStartIndex
  ##########################################################################
  method getSelectionStartIndex {} {
  };# getSelectionStartIndex

  ##########################################################################
  ## Method: getSelectionEndIndex
  ##########################################################################
  method getSelectionEndIndex {} {
  };# getSelectionEndIndex

  ##########################################################################
  ## Method: setSelectionIndexes
  ##########################################################################
  method setSelectionIndexes {first last} {
  };# setSelectionIndexes

  ##########################################################################
  ## Method: press
  ##########################################################################
  method press {x y} {
    set myState 1
    set from [$html_widget node -index $x $y]
    if {[llength $from]==2} {
      foreach {node index} $from {}
      $html_widget tag delete selection
      $html_widget tag configure selection -foreground white \
                                           -background darkgrey
      set myFromNode $node
      set myFromIdx $index
      set myToNode $node
      set myToIdx $index
    } else {
      set myToNode ""
    }
  };# press

  method release {x y} {set myState 0}

  ##########################################################################
  ## Method: motion
  ##########################################################################
  method motion {x y} {
    if {0 == $myState} return
    if {$myToNode eq ""} {
      press $x $y
      return
    }
    set to [$html_widget node -index $x $y]
    if {[llength $to]==2} {
      foreach {node index} $to {}
      if {$myToNode ne $node || $index != $myToIdx} {
        $html_widget tag remove selection $myToNode $myToIdx $node $index
        set myToNode $node
        set myToIdx $index
        $html_widget tag add selection $myFromNode $myFromIdx $myToNode $myToIdx
        puts "selection: $myFromNode $myFromIdx -> $myToNode $myToIdx"
        puts [get [list $myFromNode $myFromIdx] [list $myToNode $myToIdx]]
      }
    }
    selection own $html_widget
  };# motion

  ##########################################################################
  ## Method: reset
  ##########################################################################
  method reset {} {
    $html_widget reset
    set html_source {}
    set Offset2Node {}
    set Offset2NodeLen 0
    array unset Node2Offset *
  };# reset

  ##########################################################################
  ## Method: locateNode
  ##########################################################################
  method locateNode {offset} {
    #if {$offset == 0}               {return [lindex $Offset2Node 0]}
    #if {[string equal $offset end]} {return [lindex $Offset2Node end]}
    #if {$Offset2NodeLen < 1} {return {}}
    ## This method performs a binary search to the order list of Offset2Node
    ## in order to locate the node that contains the given offset.
    set left 0; set right $Offset2NodeLen
    while {$left <= $right} {
      set mid [expr {($left+$right)/2}]
      set value [lindex $Offset2Node $mid 0]
      if {$offset > $value} {
        set left [expr {$mid+1}]
      } elseif {$offset < $value} {
        set right [expr {$mid-1}]
      } else {
        return [lindex $Offset2Node $mid]
      }
    }
    # An exact match was not found. Perform a linear search starting from $mid.
    set value [lindex $Offset2Node $mid 0]
    if {$offset < $value} {
      while {$offset < $value} {
        #if {$mid == 0} {break}
        incr mid -1
        set value [lindex $Offset2Node $mid 0]
      }
    } else {
      while {$offset > $value} {
        #if {$mid == $Offset2NodeLen} {break}
        incr mid
        set value [lindex $Offset2Node $mid 0]
      }
      incr mid -1
    }
    return [lindex $Offset2Node $mid]
  };# locateNode

  ##########################################################################
  ## Method: locateOffset
  ##########################################################################
  method locateOffset {node} {
    if {[info exists Node2Offset($node)]} {return $Node2Offset($node)}
    return -1
  };# locateOffset

  ##########################################################################
  ## Method: setHTML
  ##########################################################################
  method setHTML {html} {
    reset; set html_source $html
    $html_widget parse -final $html
    #set Offset2Node    [lsort -integer -index 0 $Offset2Node]
    set Offset2NodeLen [llength $Offset2Node]
    return
#   set character_offset 0
#   walkTree [$html_widget node] N {
#     set tag [$N tag]
#     if {[string length $tag]} {
#       set character_offset [string first <$tag $html_source $character_offset]
#       if {$character_offset > -1} {
#         lappend Offset2Node [list $character_offset $N 0]
#         set Node2Offset($N) $character_offset
#         set character_offset [string first > $html_source $character_offset]
#       }
#     } else {
#       # A text node. Locate its first element that is not white space.
#       set node_text { }; set node_offset 0
#       foreach token [$N text -tokens] {
#         foreach {type arg} $token {break}
#         switch $type {
#           text    {set node_text $arg; break}
#           space   {set node_offset 1;
#                    #append node_text [string repeat { } $arg]}
#           newline {set node_offset 1;#append node_text "\n"}
#         }
#       }
#       set character_offset [string first $node_text \
#                                          $html_source $character_offset]
#       if {$character_offset > -1} {
#         lappend Offset2Node [list $character_offset $N $node_offset]
#         set Node2Offset($N) $character_offset
#         set character_offset [string first < $html_source $character_offset]
#       }
#     }
#   };# walkTree
#   set Offset2Node [lsort -integer -index 0 $Offset2Node]
#   set Offset2NodeLen [llength $Offset2Node]
  };# setHTML

  method printMapInfo {{file stdout}} {
    foreach item $Offset2Node {
      foreach {offset node} $item {break}
      puts $file "$offset -> $node ([$node tag]) [$node text -tokens]\
                  '[string map {\n \ } [string range [$node text -pre] 0 190]]'"
    }
  };# printMapInfo

  ## walkTree
  #  This proc is used for depth first traversal of the document tree 
  #  headed by the argument node. 
  #  Example:
  #      ::hv3::walkTree [.html node] N {
  #        puts "Type of node: [$N tag]"
  #      }
  #  If the body of the loop executes a [continue], then the current iteration
  #  is terminated and the body is not executed for any of the current nodes
  #  children. i.e. [continue] prevents descent of the tree.
  method walkTree {N varname body} {
    set level "#[expr [info level] - 1]"
    walkTree2 $N $body $varname $level
  };# walkTree

  method walkTree2 {N body varname level} {
    uplevel $level [list set $varname $N]
    set rc [catch {uplevel $level $body} msg] 
    switch $rc {
      0 {foreach n [$N children] {walkTree2 $n $body $varname $level} ;# OK}
      1 {error $msg  ;# ERROR }
      2 {return $msg ;# RETURN}
      3 {error "break from within walkTree" ;# BREAK}
      4 {# Do nothing. Do not descend the tree. ;# CONTINUE}
    }
  };# walkTree2

  method nodeInfo {node} {
    foreach sc {attribute children parent 
                tag text} {
      puts "$node $sc: [$node $sc]"
    }
  };# nodeInfo

  ##########################################################################
  ## Method: onStyleTag
  ##########################################################################
  method onStyleTag {attr script} {
    $html_widget style $script
  };# onStyleTag

  ##########################################################################
  ## Method: onScriptTag
  ##########################################################################
  method onScriptTag {attr script} {
  };# onScriptTag

  ##########################################################################
  ## Method: onTextNodeParse
  ##########################################################################
  method onTextNodeParse {N character_offset} {
    lappend Offset2Node [list $character_offset $N]
    set Node2Offset($N) $character_offset
  };# onTextNodeParse

};# class ELEP::ViewerBase::HTMLTagBase

package provide ELEP::ViewerBase::HTMLTagBase 1.0
