##############################################################################
##  HtmlWidgetDisplay.tcl:                                                  ##
## --------------------------------------------------                       ##
##  This class provides facilities for adding tags in an HTML widget, that  ##
##  visualise text portions annotated by Document Annotations.              ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::ViewerBase::TextWidgetDisplay
package require ELEP::ViewerBase::HTMLTagBase

namespace eval ::HtmlInfoViewer {}

namespace eval ELEP::ViewerBase::HtmlImages {}

GUI_LoadImage Blank

oo::class create ELEP::ViewerBase::HtmlWidgetDisplay {
  superclass ELEP::ViewerBase::TextWidgetDisplay

  variable {*}[info class variables ELEP::ViewerBase::TextWidgetDisplay] \
           html_to_text_widget_wrapper_obj \
           image_command_index \
           image_commands_created \
           html_document_base

  constructor {tobj textw {doc {}}} {
    set html_to_text_widget_wrapper_obj {}
    set image_command_index             0
    set image_commands_created          {}
    set html_document_base              {}

    next $tobj $textw $doc

    # Variable text_widget holds our text widget. Wrap the given HTML widget in
    # a suitable object (to make it act like a text widget), and reuse the code!
    set html_to_text_widget_wrapper_obj [ELEP::ViewerBase::HTMLTagBase create \
      [::ELEP::Base::auto ::ELEP::ViewerBase::] $textw]
    set text_widget $html_to_text_widget_wrapper_obj
    catch {my configureHtmlWidget $textw}
  };# constructor

  method configureHtmlWidget {win} {
    global GUI_Options
    $win configure -bg $GUI_Options(HtmlBg) -fg $GUI_Options(HtmlFg) \
      -borderwidth $GUI_Options(TextBorderWidth) -exportselection 1 \
      -unvisitedcolor blue -selectioncolor $GUI_Options(SelectBackground) \
      -tablerelief raised -padx 0 -pady 0 -underlinehyperlinks 1 \
      -fontcommand    [list ::HtmlInfoViewer::PickFont $win] \
      -imagecommand   [list [self] retrieveImage $win] \
      -formcommand    [list ::HtmlInfoViewer::FormCmd $win] \
      -bgimagecommand [list [self] retrieveBackgroundImage $win]
    set ::HtmlInfoViewer::InfoViewer($win:showImages)   1
    set ::HtmlInfoViewer::InfoViewer($win:showBgImages) 1
    bind $win.x <<MB_2>> \
      [list ::HtmlInfoViewer::ContextualMenuHandler $win $document]
  };# configureHtmlWidget

  method retrieveImage {win args} {
    if {[info exists ::HtmlInfoViewer::InfoViewer($win:showImages)]} {
      if {!$::HtmlInfoViewer::InfoViewer($win:showImages)} {return {}}
    }
    set image_command \
        ELEP::ViewerBase::HtmlImages::img[incr image_command_index]
    foreach {fn w h spec} $args {break}
    ## Can we have a match in the metadata directory?
    set base [$html_to_text_widget_wrapper_obj getBase]
    if {[file isdirectory $base]} {
      set filename_tail [file tail $fn]
      if {[file exists $base/$filename_tail]} {
        if {![catch {image create photo $image_command \
                           -file $base/$filename_tail} image]} {
          lappend image_commands_created $image
          return $image
        }
      }
    }
    if {[file exists $fn]} {
      if {[catch {image create photo $image_command -file $fn} image]} {
        return Blank
      }
      lappend image_commands_created $image
      return $image
    }
    ## We haven't found an exact match. Try a fuzzy match...
    if {[file exists [string map {{ } *} $fn]]} {
      foreach file [glob -nocomplain -- [string map {{ } *} $filename]] {
        if {[catch {image create photo $image_command -file $file} image]} {
          return Blank
        }
        break
      }
      lappend image_commands_created $image
      return $image
    }
    ## Finally, try to fetch image from original files, if they exist...
    set xid [tip_GetExternalId $document]
    set metadata [file rootname $xid]_files
    set filename $metadata/[file tail $fn]
    foreach file [glob -nocomplain -- [string map {{ } *} $filename]] {
      if {[catch {image create photo $image_command -file $file} image]} {
        return Blank
      }
      break
    }
    if {![info exists image]} {return Blank}
    lappend image_commands_created $image
    return $image
  };# retrieveImage

  method retrieveBackgroundImage {win args} {
    if {[info exists ::HTML_InfoViewer($win:showBgImages)]} {
      if {!$::HTML_InfoViewer($win:showBgImages)} {return {}}
    }
    return [my retrieveImage $win $args]
  };# retrieveBackgroundImage

  method deleteCreatedImages {} {
    foreach image $image_commands_created {catch {image delete $image}}
    set image_commands_created {}
  };# deleteCreatedImages

  ##########################################################################
  ## Event: onDocumentChange
  ##########################################################################
  method onDocumentChange {col doc} {
    my deleteCreatedImages
    set html_document_base {}
    if {[string length $col] && [string length $doc]} {
      set metadata [tip_GetName $col]/Metadata/HtmlMetadata
      if {[file isdirectory $metadata]} {
        set html_document_base $metadata/[file tail [tip_GetExternalId $doc]]
      }
    }
    catch {$html_to_text_widget_wrapper_obj setBase $html_document_base}
    next $col $doc
  };# onDocumentChange

  destructor {
    next
    my deleteCreatedImages
    catch {$html_to_text_widget_wrapper_obj destroy}
  };# destructor

  method validTextWidget {} {
    return [$html_to_text_widget_wrapper_obj validTextWidget]
  };# validTextWidget

  method getWidgetForManagingBindings {} {
    #puts [info level 0]
    return [$html_to_text_widget_wrapper_obj getWidgetForManagingBindings]
  };# getWidgetForManagingBindings

  method getIndexAtxy {x y} {
    #puts [info level 0]
    return [$html_to_text_widget_wrapper_obj getIndexAtxy $x $y]
  };# getIndexAtxy

  method selectionExists {} {
    #puts [info level 0]
    return [$html_to_text_widget_wrapper_obj selectionExists]
  };# selectionExists

  method selectionClear         {} {
    #puts [info level 0]
    return [$html_to_text_widget_wrapper_obj selectionClear]
  };# selectionClear

  method getSelectionStartIndex {} {
    #puts [info level 0]
    return [$html_to_text_widget_wrapper_obj getSelectionStartIndex]
  };# getSelectionStartIndex

  method getSelectionEndIndex {} {
    #puts [info level 0]
    return [$html_to_text_widget_wrapper_obj getSelectionEndIndex]
  };# getSelectionEndIndex

  method setSelectionIndexes {first last} {
    #puts [info level 0]
    return [$html_to_text_widget_wrapper_obj setSelectionIndexes $first $last]
  };# setSelectionIndexes

  method addSelectionIndexes    {first last} {
    return [$html_to_text_widget_wrapper_obj addSelectionIndexes $first $last]
  };# addSelectionIndexes

  method textWidgetStateNormal   {} {}
  method textWidgetStateDisabled {} {}

  method index2CharacterOffset {index} {
    return [$html_to_text_widget_wrapper_obj index2CharacterOffset $index]
  };#index2CharacterOffset

  method examineAnnotationToBeAdded {ann} {
    ## HTML-related regular expressions...
    set HTML_SPACE             {[[:space:]\n]+}
    set HTML_ATTRIBUTE_NAME    {[A-Za-z0-9_:.-]+}
    set HTML_ATTRIBUTE_VALUE   {\"[^\"]+\"|\'[^']+\'|[^\> \t\n]+}
    set HTML_ATTRIBUTE         ${HTML_ATTRIBUTE_NAME}((${HTML_SPACE})*\=(${HTML_SPACE})*${HTML_ATTRIBUTE_VALUE})?
    set HTML_TAG_NAME          {[!]?[1-9A-Za-z:_]+[/]?}
    set HTML_TAG_ATTRS         (${HTML_SPACE})?(${HTML_SPACE}${HTML_ATTRIBUTE})*(${HTML_SPACE})*
    set HTML_TAG               \\<\\/?${HTML_TAG_NAME}${HTML_TAG_ATTRS}\[/\]?\\>
    
    set spanSet [tip_GetSpans $ann]
    set textRanges [tip_GetAnnotatedTextRanges $document $ann]
    set newSpanSet {}
    foreach span $spanSet text $textRanges {
      set text [string map {\n { }} $text]
      tip_GetSpanOffsets $span start end
      ## If the annotation spans only white-space, then leave it as it is...
      if {[string length [string trim $text]] == 0} {
        lappend newSpanSet [tip_CreateSpan $start $end]
        continue
      }
      ## Has the text any white space prefix/suffix?
      set trimmedLeft [string trimleft $text]
      if {$trimmedLeft != $text} {
        incr start [expr {[string length $text]-[string length $trimmedLeft]}]
        set text $trimmedLeft
      }
      set trimmedRight [string trimright $text]
      if {$trimmedRight != $text} {
        set text $trimmedRight
        set end [expr {$start + [string length $text]}]
      }
      ## Is it an HTML entity?
      if {$text == "&"} {
        ## As HTML entities are displayed as a single character, the selected
        ## text will always be "&". We should check & expand the span set...
        set entity [string range [tip_GetByteSequence $doc] \
                                 $start [expr {$start+10}]]
        if {[regexp -- {^\&\w+\;} $entity matched]} {
          set end [expr {$start + [string length $matched]}]
        }
      }
      ## Trim HTML elements from the begining of the string...
      if {[regsub -- ^(\\>(\\s+)*)?($HTML_TAG|\\s)+ $text {} htmlTrimmed]} {
        incr start [expr {[string length $text]-[string length $htmlTrimmed]}]
        set text $htmlTrimmed
      }
      ## Trim HTML elements from the end of the string...
      if {[regsub -- ($HTML_TAG|\\s)+$ $text {} htmlTrimmed]} {
        set text $htmlTrimmed
        set end [expr {$start + [string length $text]}]
      }
      lappend newSpanSet [tip_CreateSpan $start $end]
    }
    return [tip_CreateAnnotation [tip_GetType $ann] \
                                     $newSpanSet \
                                     [tip_GetAttributes $ann]]
  };# examineAnnotationToBeAdded

  method seeIndex {index} {
    set index [$html_to_text_widget_wrapper_obj tagNormaliseIndex $index]
    $html_to_text_widget_wrapper_obj seeIndex $index
  };# seeIndex

  method setFocus {} {$html_to_text_widget_wrapper_obj setFocus}

};# class ELEP::ViewerBase::HtmlWidgetDisplay

package provide ELEP::ViewerBase::HtmlWidgetDisplay 1.0
