##############################################################################
##  MultiAttributedAnnotationSpecificationSelector.tcl:                     ##
## --------------------------------------------------                       ##
##  This class implements a widget that presents a set of options, that can ##
##  be used to select which Annotation types will be added to a Document.   ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require BWidget
package require ELEP::WindowUtilities
package require ELEP::ViewerBase::AnnotatorSpecificationReader

oo::class create ELEP::ViewerBase::MultiAttributedAnnotationSpecificationSelector {
  variable reader \
           dialog_window \
           language \
           annotation \
           alternative \
           attributes \
           attributeValues \
           attributeValueDescriptions \
           attributeType
  
  constructor {} {
    my init
  };# constructor

  method getArray {var} {return [array get $var]}

  method init {} {
    my createReader
    my createDialog
    my populateDialogFrame [$dialog_window getframe]
    my restoreState [$dialog_window getframe]
  };# init

  method createReader {} {
    set reader [ELEP::ViewerBase::AnnotatorSpecificationReader create \
                    [::ELEP::Base::auto ::ELEP::ViewerBase::]]
  };# createReader

  method createDialog {} {
    set dialog_window [Dialog [::ELEP::WindowUtilities::UniqueWindowName \
     .annotatorSpecificationSelector] -modal local -separator 1 \
     -title "Select Annotation Characteristics..." -default 0 -cancel 1]
    $dialog_window add -textvariable GUI_Text(Ok)
    $dialog_window add -textvariable GUI_Text(Cancel)
  };# createDialog

  method populateDialogFrame {frame} {
    ## Language:
    ttk::label $frame.llang -anchor w -text {Language:}
    ttk::combobox $frame.lang -state readonly -exportselection 0 \
      -values [concat {{}} [$reader getLanguages]] \
      -textvariable [my varname language]
    bind $frame.lang <<ComboboxSelected>> "[self] onLanguageChange $frame"
    grid $frame.llang $frame.lang -sticky snew \
      -ipadx 3 -ipady 3 -padx 6 -pady 6
    
    ## Annotation Type:
    ttk::label $frame.lann -anchor w -text {Annotation Type:}
    ttk::combobox $frame.ann -state readonly -exportselection 0 \
      -textvariable [my varname annotation]
    bind $frame.ann <<ComboboxSelected>> "[self] onAnnotationChange $frame"
    grid $frame.lann $frame.ann -sticky snew \
      -ipadx 3 -ipady 3 -padx 6 -pady 6
    
    ## Alternative:
    ttk::label $frame.lalternative -anchor w -text {Attribute Alternative:}
    ttk::combobox $frame.alternative -state readonly -exportselection 0 \
      -textvariable [my varname alternative]
    bind $frame.alternative <<ComboboxSelected>> "[self] onAlternativeChange"
    grid $frame.lalternative $frame.alternative -sticky snew \
      -ipadx 3 -ipady 3 -padx 6 -pady 6

    ## Attributes:
    ttk::labelframe $frame.lf -text Attributes:
    ScrolledWindow $frame.lf.sc
    listbox $frame.lf.sc.values -listvariable [my varname attributes] \
      -height 5
    $frame.lf.sc setwidget $frame.lf.sc.values
    pack $frame.lf.sc -fill both -expand 1 -padx 2 -pady 2
    grid $frame.lf -row 0 -column 2 -rowspan 5 -sticky snew \
      -ipadx 3 -ipady 3 -padx 6 -pady 1
    grid rowconfigure $frame 4 -weight 1
    grid columnconfigure $frame 1 -weight 1
  };# populateDialogFrame

  method onLanguageChange {frame} {
    $frame.ann configure -values \
      [$reader getMultiAttributedAnnotationTypes $language]
    set annotation {}
    $frame.alternative configure -values {}; set alternative {}
    set attributes {}
  };# onLanguageChange

  method onAnnotationChange {frame} {
    $frame.alternative configure -values \
      [$reader getMultiAttributedAnnotationAlternatives \
       $language $annotation]
    set attributes {}
  };# onAnnotationChange

  method onAlternativeChange {} {
    set attributes {}
    foreach one [$reader getMultiAttributedAnnotationAttributes \
       $language $annotation $alternative] {
      if {[string length $one]} {lappend attributes $one}
    }
    ## For each attribute, try to get the list of values...
    foreach attr $attributes {
      set attributeValues($attr) \
        [$reader getMultiAttributedAnnotationAttributeValues \
                $language $annotation $alternative $attr]
      set type [string tolower \
        [$reader getMultiAttributedAnnotationAttributeType \
                $language $annotation $alternative $attr]]
      if {![string length $type] && [llength $attributeValues($attr)]} {
        set type combo
      }
      lappend attributeType $type
      set attributeValueTypes($attr) {}

      set attributeValueDescriptions($attr) {}
      foreach one $attributeValues($attr) {
        lappend attributeValueDescriptions($attr) \
          [$reader getMultiAttributedDescription \
            $language $annotation $alternative $attr $one]
      }
    }
  };# onAlternativeChange

  method getGroupValues {group} {
    return {}
  };# getGroupValues

  method getGroupButtonsNumber {group} {
    return 0
  };# getGroupButtonsNumber

  method getMaxButtonsNumber {} {return 0}

  method getGroupDescription {group} {
    return {}
  };# getGroupDescription

  method getAttributeDescription {attr} {
    return [$reader getMultiAttributedDescription \
      $language $annotation $alternative $attr]
  };# getAttributeDescription

  method getValueDescription {attr value} {
    return [$reader getMultiAttributedDescription \
      $language $annotation $alternative $attr $value]
  };# getValueDescription

  method getValueBackgroundColour {group value} {
    return white
  };# getValueBackgroundColour

  method getValueForegroundColour {group value} {
    return black
  };# getValueForegroundColour

  method show {} {
    if {[$dialog_window draw] == 0} {
      my saveState
      return ok
    } else {
      return cancel
    }
  };# show

  destructor {
    if {[string length $reader]} {catch {$reader destroy}}
    set reader {}
  };# destructor

  method saveStateVariableName {} {return ::GUI_Options([self class])}

  method saveState {} {
    set [my saveStateVariableName] [list $language    $annotation \
                                         $alternative $attributes]
  };# saveState

  method restoreState {frame} {
    if {![info exists [my saveStateVariableName]]} {return}
    foreach {language _annotation _alternative _attributes} \
            [set [my saveStateVariableName]] {break}
    my onLanguageChange $frame
    set annotation  $_annotation;  my onAnnotationChange $frame
    set alternative $_alternative; my onAlternativeChange
  };# restoreState

};# class ELEP::ViewerBase::MultiAttributedAnnotationSpecificationSelector

package provide ELEP::ViewerBase::MultiAttributedAnnotationSpecificationSelector 1.0
