##############################################################################
##  TemplateSpecificationSelector.tcl:                                      ##
## --------------------------------------------------                       ##
##  This class implements a widget that presents a set of options, that can ##
##  be used to select an Annotation template.                               ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require BWidget
package require ELEP::WindowUtilities
package require ELEP::ViewerBase::AnnotatorSpecificationReader

oo::class create ELEP::ViewerBase::TemplateSpecificationSelector {
  variable reader \
           dialog_window \
           frame \
           language \
           template \
           tables
  
  constructor {{init_dialog 1}} {
    set reader        {}
    set dialog_window {}
    set frame         {}
    set language      {}
    set template      {}
    set tables        {}
    if {$init_dialog} {my init}
  };# constructor

  method init {} {
    my createReader
    my createDialog
    my populateDialogFrame [$dialog_window getframe]
    my restoreState [$dialog_window getframe]
  };# init

  method createReader {} {
    set reader [ELEP::ViewerBase::AnnotatorSpecificationReader new]
  };# createReader

  method createDialog {} {
    set dialog_window [Dialog [::ELEP::WindowUtilities::UniqueWindowName \
     .annotatorSpecificationSelector] -modal local -separator 1 \
     -title "Select Template Characteristics..." -default 0 -cancel 1]
    $dialog_window add -textvariable GUI_Text(Ok)
    $dialog_window add -textvariable GUI_Text(Cancel)
    set frame [$dialog_window getframe]
  };# createDialog

  method populateDialogFrame {frame} {
    ttk::label $frame.llang -text {Language:}
    ttk::combobox $frame.lang -state readonly -exportselection 0 \
      -values [concat {{}} [$reader getLanguages]] \
      -textvariable [my varname language]
    bind $frame.lang <<ComboboxSelected>> "[self] onLanguageChange $frame"
    grid $frame.llang $frame.lang -sticky snew \
      -ipadx 3 -ipady 3 -padx 6 -pady 6
    ttk::label $frame.lann -text {Template:}
    ttk::combobox $frame.ann -state readonly -exportselection 0 \
      -textvariable [my varname template]
    bind $frame.ann <<ComboboxSelected>> "[self] onTemplateChange"
    grid $frame.lann $frame.ann -sticky snew \
      -ipadx 3 -ipady 3 -padx 6 -pady 6
    ttk::labelframe $frame.lf -text {Template Tables:}
    ScrolledWindow $frame.lf.sc
    listbox $frame.lf.sc.values -listvariable [my varname tables] \
      -height 5 -width 35
    $frame.lf.sc setwidget $frame.lf.sc.values
    pack $frame.lf.sc -fill both -expand 1 -padx 2 -pady 2
    grid $frame.lf -row 0 -column 2 -rowspan 5 -sticky snew \
      -ipadx 3 -ipady 3 -padx 6 -pady 1
    grid rowconfigure $frame 4 -weight 1
    grid columnconfigure $frame 1 -weight 1
  };# populateDialogFrame

  method onLanguageChange {frame} {
    $frame.ann configure -values [$reader getTemplateNames $language]
    set template {}
    set tables {}
  };# onLanguageChange

  method onTemplateChange {} {
    set tables {}
    foreach table [$reader getTemplateTables $language $template] {
      lappend tables " \u25ba  $table"
      foreach field [$reader getTemplateTableFields \
                             $language $template $table] {
        set max_values [$reader getTemplateTableFieldMaxValueNumber \
                             $language $template $table $field]
        switch -exact $max_values {
          0       {set val_descr "multivalued - unlimited"}
          1       {set val_descr "single valued"}
          default {set val_descr "multivalued - max: $max_values"}
        }
        lappend tables "     \u25ba  $field ($val_descr)"
      }
    }
  };# onTemplateChange

  method show {} {
    if {[$dialog_window draw] == 0} {
      my saveState
      return ok
    } else {
      return cancel
    }
  };# show

  method getTemplateTables {} {
    return [$reader getTemplateTables $language $template]
  };# getTemplateTables

  method getTemplateTableFields {table} {
    return [$reader getTemplateTableFields $language $template $table]
  };# getTemplateTableFields

  method getTemplateTableFieldValueType {table field} {
    return [$reader getTemplateTableFieldValueType \
                    $language $template $table $field]
  };# getTemplateTableFieldValueType

  method getTemplateTableFieldMaxValueNumber {table field} {
    return [$reader getTemplateTableFieldMaxValueNumber \
                    $language $template $table $field]
  };# getTemplateTableFieldMaxValueNumber

  method getTemplateTableFieldPrimary {table field} {
    return [$reader getTemplateTableFieldPrimary \
                    $language $template $table $field]
  };# getTemplateTableFieldPrimary

  method getTemplateTableFieldValueConstraints {table field} {
    return [$reader getTemplateTableFieldValueConstraints \
                    $language $template $table $field]
  };# getTemplateTableFieldValueConstraints

  method getTemplateDescription {{table {}} {field {}}} {
    return [$reader getTemplateDescription \
                    $language $template $table $field]
  };# getTemplateDescription

  destructor {
   catch {$reader destroy}
  };# destructor

  method saveStateVariableName {} {return ::GUI_Options([self class])}

  method saveState {} {
    set [my saveStateVariableName] [list $language $template]
  };# saveState

  method restoreState {frame} {
    if {![info exists [my saveStateVariableName]]} {return}
    foreach {language _template} [set [my saveStateVariableName]] {break}
    my onLanguageChange $frame
    set template $_template; my onTemplateChange
  };# restoreState

};# class ELEP::ViewerBase::TemplateSpecificationSelector

package provide ELEP::ViewerBase::TemplateSpecificationSelector 1.0
