##############################################################################
##  TextWidgetDisplay.tcl:                                                  ##
## --------------------------------------------------                       ##
##  This class provides facilities for adding tags in a text widget, that   ##
##  visualise text portions annotated by Document Annotations.              ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

## Class TextWidgetDisplay:
## Methods:
##   constructor
##   createToolbar
##   createToolbarWidgets
##   setText
##   clearText
##   textWidgetStateNormal
##   textWidgetStateDisabled
##   setToplevelObject
##   setStatusText
##   showProgress
##   hideProgress
##   setProgress
##   onDocumentChange
##   annotateDocument
##   filterDocumentAnnotations
##   annotateAnnotationSet
##   deleteSelection
##   annotateSelection
##   addAnnotation
##   addTag
##   index2CharacterOffset
##   characterOffset2Index
##   offsetWRTDocument
##   tag2Value
##   tag2Id
##   getSelectedAnnotationIds
##   getAnnotatedTextRanges
##   onTagPress
##   unselectSelectedTags
##   selectTag
##   raiseTag
##   onTagEnter
##   onTagLeave
##   registerAnnotatorObject
##   validTextWidget
##   getWidgetForManagingBindings
##   getDocument
##   getCollection
##   getIndexAtxy
##   selectionExists
##   selectionClear
##   getSelectionStartIndex
##   getSelectionEndIndex
##   setSelectionIndexes
##   restoreMainWidget
##   findActiveWidget
##   call
##   registerOnClickListener
##   mouseEventForWindow
##   onClick
##
package require ELEP::ViewerBase::ColourManager

oo::class create ELEP::ViewerBase::TextWidgetDisplay {
  variable text_widget \
           collection \
           document \
           text_widget_offset \
           toplevel_object \
           annotator_object \
           selected_tags \
           on_click_listeners

  constructor {tobj textw {doc {}}} {
    my variable always_remove_annotation
    set text_widget        {}
    set collection         {}
    set document           {}
    set text_widget_offset 0
    set toplevel_object    {}
    set annotator_object   {}
    set selected_tags      {}
    set on_click_listeners {}
    set always_remove_annotation 1

    set toplevel_object $tobj
    set text_widget $textw
    bind $text_widget <Destroy> "+if {\[string equal $text_widget %W\]}\
                 {::ELEP::WindowUtilities::DeleteTclOOObject [self]}"
    ELEP::WindowUtilities::AutoFocusOnMouseEnter $text_widget
    set document   $doc
    set collection {}
    if {[string length $doc]} {
      set collection [tip_GetParent $doc]
      my setText [tip_GetRawData $doc]
    }
    if {[string length $toplevel_object]} {
      $toplevel_object updateStatusFromDocument
    }
    my createToolbar
    my textWidgetStateDisabled
  };# constructor

  method restoreMainWidget {} {
  };# restoreMainWidget

  method findActiveWidget {} {
  };# findActiveWidget

  method createToolbar {} {
    set toolbarFrame [$toplevel_object getToolbarFrame]
    set toolbar $toolbarFrame.displayText
    if {![winfo exists $toolbar]} {
      ttk::frame $toolbar -class Toolbar
      my createToolbarWidgets $toolbar
      pack $toolbar -padx 2 -pady 2 -side left
    }
  };# createToolbar

  method createToolbarWidgets {toolbar} {
    ttk::button $toolbar.delete -textvariable [my msgVar {Delete Annotation}] \
      -command "[self] deleteSelection" -style Toolbutton
    grid $toolbar.delete -sticky news -padx 2 -pady 2
    bind $toolbar.delete <Key-Delete> "$toolbar.delete invoke"
    bind [winfo toplevel $toolbar.delete] <Key-Delete> \
      "$toolbar.delete invoke"
  };# createToolbarWidgets

  destructor {
    catch {my clearText}
  };# destructor

  method setText {val} {
    my restoreMainWidget
    if {![my validTextWidget]} {return}
    my textWidgetStateNormal
    $text_widget delete 1.0 end
    foreach tag [$text_widget tag names] {$text_widget tag delete $tag}
    $text_widget insert end $val
    $text_widget tag add back 1.0 end
    $text_widget tag bind back <<MB_1>> [list [self] onClick]
    $text_widget tag lower back
    $text_widget tag raise sel
    my textWidgetStateDisabled
  };# setText

  method clearText {} {
    my restoreMainWidget
    if {![my validTextWidget]} {return}
    my textWidgetStateNormal
    $text_widget delete 1.0 end
    foreach tag [$text_widget tag names] {$text_widget tag delete $tag}
    my textWidgetStateDisabled
  };# clearText

  method textWidgetStateNormal   {} {$text_widget configure -state normal}
  method textWidgetStateDisabled {} {$text_widget configure -state disabled}

  method setToplevelObject    {obj} {set toplevel_object $obj}

  method setStatusText {val} {
    if {[string length $toplevel_object]} {$toplevel_object setStatusText $val}
  };# setStatusText

  method showProgress {} {
    if {[string length $toplevel_object]} {$toplevel_object showProgress}
  };# showProgress

  method hideProgress {} {
    if {[string length $toplevel_object]} {$toplevel_object hideProgress}
  };# hideProgress

  method setProgress {val} {
    if {[string length $toplevel_object]} {
      $toplevel_object setProgress [expr {int($val)}]
      update
    }
  };# setProgress

  method filterDocumentAnnotations {annotations} {
    return [tip_SortAnnotationSet $annotations]
  };# filterDocumentAnnotations

  ##########################################################################
  ## Event: onDocumentChange
  ##########################################################################
  method onDocumentChange {col doc} {
    # puts "[self] onDocumentChange $col $doc $annotator_object"
    my unselectSelectedTags
    my clearText
    set collection $col
    set document   $doc
    ## Notify our annotator...
    if {[string length $annotator_object]} {
      $annotator_object onDocumentChange $col $doc
    }
    if {[string length $doc]} {
      if {![string length $collection]} {set collection [tip_GetParent $doc]}
      my setText [tip_GetRawData $doc]
      if {[string length $toplevel_object]} {
        $toplevel_object updateStatusFromDocument $col $doc
      }
      if {[string length $annotator_object]} {
        my annotateDocument [$annotator_object getAnnotation] \
                            [$annotator_object getAttribute]
      }
    } else {
      if {[string length $toplevel_object]} {
        $toplevel_object updateStatusFromDocument
      }
    }
  };# onDocumentChange

  #########################################################################
  ## Event: annotateDocument
  ##########################################################################
  method annotateDocument {annotation attribute} {
    # puts "[self] annotateDocument $annotation $attribute"
    if {[string length $attribute]} {
      set annotations [my selectAnnotations_Attribute $annotation $attribute]
    } else {
      set annotations [my selectAnnotations_NoAttribute $annotation]
    }
    my annotateAnnotationSet $annotations $annotation $attribute
  };# annotateDocument

  method selectAnnotations_Attribute {annotation attribute} {
    return [tip_SelectAnnotations $document $annotation \
                         "\[string length ann::${attribute} \]"]
  };# selectAnnotations_Attribute

  method selectAnnotations_NoAttribute {annotation} {
    # puts "[self] selectAnnotations_NoAttribute $annotation"
    return [tip_SelectAnnotations $document $annotation]
  };# selectAnnotations_NoAttribute

  #########################################################################
  ## Event: annotateAnnotationSet
  ##########################################################################
  method annotateAnnotationSet {annotations annotation attribute} {
    if {![my validTextWidget]} {return}
    if {![string length $annotator_object]} {return}
    set annotations [my filterDocumentAnnotations $annotations]
    set progress 0.0
    set progressStep [expr {100.0/([llength $annotations]+1)}]
    my setProgress $progress
    my showProgress
    if {[string length $attribute]} {
      set result [my annotateAnnotationSet_Attribute \
                             $annotations $attribute $progress $progressStep]
    } else {
      ## No attribute was specified! Use annotation id!
      set result [my annotateAnnotationSet_NoAttribute \
                             $annotations $progress $progressStep]
    }
    my hideProgress
    if {[llength $result]} {
      array set invalid $result
      set msg "Invalid Annotations Found:\
               \nAnnotations #   Value"
      foreach value [lsort [array names invalid]] {
        append msg "\n[llength $invalid($value)] \"$value\""
      }
      append msg "\nDo you want to remove these Annotations?"
      if {[string equal [gut_question $msg] yes]} {
        ## Remove the annotations from the document...
        set progress 0.0
        set progressStep [expr {100.0/([llength [array names invalid]]+1)}]
        my setProgress $progress
        my showProgress
        foreach value [array names invalid] {
          set progress [expr {$progress + $progressStep}]
          my setProgress $progress
          foreach id $invalid($value) {
            tip_RemoveAnnotation $document $id
          }
        }
        my hideProgress
      }
    }
  };# annotateAnnotationSet

  method annotateAnnotationSet_Attribute {annotations attribute
                                          progress progressStep} {
    foreach ann $annotations {
      set progress [expr {$progress + $progressStep}]
      my setProgress $progress
      set id    [tip_GetId $ann]
      set value [tip_GetValueValue [tip_GetAttribute $ann $attribute]]
      tip_GetFirstSpanOffsets $ann start end
      if {$end <= $start} {
        lappend invalid($value) $id
        continue
      }
      incr start -$text_widget_offset; incr end -$text_widget_offset
      if {![info exists value2pair($value)]} {
        set value2pair($value) [$annotator_object getValueColourPair $value]
      }
      foreach {bg fg} $value2pair($value) {break}
      my addTag 1.0+${start}c 1.0+${end}c $id $value $bg $fg
    }
    if {[info exists invalid]} {
      return [array get invalid]
    }
    return {}
  };# annotateAnnotationSet_Attribute

  method annotateAnnotationSet_NoAttribute {annotations progress progressStep} {
    foreach ann $annotations {
      set progress [expr {$progress + $progressStep}]
      my setProgress $progress
      set id [tip_GetId $ann]
      catch {$text_widget tag delete $id:$id}
      foreach span [tip_GetSpans $ann] {
        tip_GetSpanOffsets $span start end
        if {[my skipSpan $start $end]} {continue}
        if {$end <= $start} {
          lappend invalid($id) $id
          continue
        }
        incr start -$text_widget_offset; incr end -$text_widget_offset
        if {![info exists value2pair($id)]} {
          set value2pair($id) [$annotator_object getValueColourPair $id]
        }
        foreach {bg fg} $value2pair($id) {break}
        my addTag 1.0+${start}c 1.0+${end}c $id $id $bg $fg 0
      }
    }
    if {[info exists invalid]} {
      return [array get invalid]
    }
    return {}
  };# annotateAnnotationSet_NoAttribute

  method skipSpan {start end} {return 0}

  ##########################################################################
  ## Event: deleteSelection
  ##########################################################################
  method deleteSelection {} {
    if {![string length $annotator_object]} {return}
    my variable always_remove_annotation
    set ids {}
    set tags {}
    set annotation [$annotator_object getAnnotation]
    set attribute  [$annotator_object getAttribute]
    ## Is there selected text?
    if {[my selectionExists]} {
      set first [my getSelectionStartIndex]
      set last  [my getSelectionEndIndex]
      my selectionClear
      set start [my index2CharacterOffset $first]
      set end   [my index2CharacterOffset $last]
      incr start $text_widget_offset; incr end $text_widget_offset
      ## Retrieve all Annotations that lie between the two offsets...
      foreach ann [tip_AnnotationsInRange \
                    [tip_SelectAnnotations $document $annotation] $start $end] {
        set id    [tip_GetId $ann]
        set value [tip_GetValueValue [tip_GetAttribute $ann $attribute]]
        lappend ids  $id
        lappend tags $value:$id
      }
    } elseif {[llength $selected_tags]} {
      foreach tag $selected_tags {
        lappend ids [my tag2Id $tag]
      }
      set tags $selected_tags
      my unselectSelectedTags
    }
    foreach id $ids tag $tags {
      catch {$text_widget tag delete $tag}
      ## Maybe the annotator wants to handle removal?
      if {[$annotator_object deletionHandled $document $id $attribute $tag]} {
        continue
      }
      set ann [tip_GetAnnotation $document $id]
      ## Try to retrieve the attributes from the annotation...
      if {[catch {$annotator_object getAttributeNamesFromAnnotation $ann} \
                                                          attribute_names]} {
        ## Either the annotator does not support the method, or a real error
        ## occured...
        set attribute_names [list $attribute]
      } else {
        if {[llength $attribute_names] == 0} {
          set attribute_names [list $attribute]
        }
      }
      ## Iterate over all attribute names, and delete them...
      foreach attr $attribute_names {
        $annotator_object annotationAttributeWillBeDeleted $id $attr
        set ann [tip_RemoveAttribute $ann $attr]
        puts "Removing Attribute $attr: $ann"
      }

      ## If no attributes were left, remove the whole annotation...
      set attrs [tip_GetAttributes $ann]
      if {$always_remove_annotation || ([llength $attrs] < 1)} {
        ## Remove the whole Annotation
        puts "Removing Annotation: $ann"
        $annotator_object annotationWillBeDeleted $id
        tip_RemoveAnnotation $document $id
      } else {
        tip_AddAnnotation $document $ann
      }
    }
  };# deleteSelection

  ##########################################################################
  ## Event: annotateSelection
  ##########################################################################
  method annotateSelection {annotation attribute value bg fg} {
    if {![my validTextWidget]} {return}
    set id {}
    ## Is there selected text?
    if {[my selectionExists]} {
      set first [my getSelectionStartIndex]
      set last  [my getSelectionEndIndex]
      my selectionClear
      # $text_widget tag remove sel $first $last
      set start [my index2CharacterOffset $first]
      set end   [my index2CharacterOffset $last]
      set id    [my addAnnotation $annotation $attribute $start $end\
                                  $value $bg $fg]
      ## It is possible that the annotation that was added to the Document has a
      ## different span. In this case, we need to fix first & last.
      foreach one $id {
        tip_GetFirstSpanOffsets [tip_GetAnnotation $document $one] _start _end
        incr _start -$text_widget_offset; incr _end -$text_widget_offset
        if {$_start != $start} {set first 1.0+${_start}c}
        if {$_end   != $end}   {set last  1.0+${_end}c}
        my selectTag [my addTag $first $last $one $value $bg $fg]
      }
    } elseif {[llength $selected_tags]} {
      ## No text is selected. Is there a tag selected? If so, change its value.
      set tag [lindex $selected_tags 0]
      set id  [my tag2Id $tag]
      set ann [tip_GetAnnotation $document $id]
      tip_GetFirstSpanOffsets $ann start end
      incr start -$text_widget_offset; incr end -$text_widget_offset
      set first 1.0+${start}c; set last 1.0+${end}c
      set attr [tip_CreateAttribute $attribute \
                  [tip_CreateAttributeValue GDM_STRING $value]]
      set ann  [tip_PutAttribute $ann $attr]
      set id   [tip_AddAnnotation $document $ann]
      #$text_widget tag remove $tag $first $last
      $text_widget tag delete $tag
      my selectTag [my addTag $first $last $id $value $bg $fg]
    }
    return $id
  };# annotateSelection

  method addAnnotation {annotation attribute start end value bg fg} {
    incr start $text_widget_offset; incr end $text_widget_offset
    ## As a first step, examine existing Annotations, in case an Annotation with
    ## the same Span exists...
    set Annotations [tip_SelectAnnotations \
        [tip_AnnotationsMatchingRange $document $start $end] $annotation]
    set attr [tip_CreateAttribute $attribute \
                  [tip_CreateAttributeValue GDM_STRING $value]]
    if {[llength $Annotations]} {
      foreach ann $Annotations {
        ## The Selected Span matches existing annotation(s). We can re-use
        ## the annotation ONLY if the attribute we are annotating is not
        ## present...
        if {[llength [tip_GetAttribute $ann $attribute]]} {continue}
        catch {$text_widget tag delete [my id2Tag [tip_GetId $ann]]}
        lappend id [tip_AddAnnotation $document [tip_PutAttribute $ann $attr]]
        ## puts "Changing Annotation with ID $ID: $attr -> $value"
      }
    }
    if {![info exists id]} {
      ## A New Annotation must be created...
      set ann [my examineAnnotationToBeAdded [tip_CreateAnnotation $annotation \
                   [tip_CreateSpanSet $start $end] [list $attr]]]
      set id  [tip_AddAnnotation $document $ann]
      ## puts "Creating New $Annotation Annotation ($ID, $value)"
    }
    return $id
  };# addAnnotation

  method examineAnnotationToBeAdded {ann} {return $ann}

  method addTag {first last id value bg fg {delete_tag 1}} {
    # puts "[self] addTag $first $last $id $value $bg $fg"
    my variable id2Tag
    set tag $value:$id
    set id2Tag($id) $tag
    if {$delete_tag} {catch {$text_widget tag delete $tag}}
    $text_widget tag add $tag $first $last
    $text_widget tag configure $tag -foreground $fg -background $bg
    $text_widget tag bind $tag <<MB_1>> [list [self] onTagPress $id $value]
    $text_widget tag bind $tag <Enter>  [list [self] onTagEnter $id $value]
    $text_widget tag bind $tag <Leave>  [list [self] onTagLeave $id $value]
    return $tag
  };# addTag

  method addInactiveTag {first last id value bg fg} {
    set tag $value:$id
    catch {$text_widget tag delete $tag}
    $text_widget tag add $tag $first $last
    $text_widget tag configure $tag -foreground $fg -background $bg
    return $tag
  };# addInactiveTag

  method deleteInactiveTag {id value} {
    catch {$text_widget tag delete $value:$id}
  };# deleteInactiveTag

  method deleteTag {tag} {
    catch {$text_widget tag delete $tag}
  };# deleteTag

  method index2CharacterOffset {index} {
    return [string length [$text_widget get 1.0 $index]]
  };# index2CharacterOffset

  method characterOffset2Index {offset} {
    return 1.0+${offset}c
  };# characterOffset2Index

  method offsetWRTDocument {} {
    return $text_widget_offset
  };# offsetWRTDocument

  method tag2Value {tag} {
    set index [string last : $tag]
    return [string range $tag 0 [incr index -1]]
  };# tag2Value

  method tag2Id {tag} {
    set index [string last : $tag]
    return [string range $tag [incr index] end]
  };# tag2Id

  method id2Tag {id} {
    my variable id2Tag
    if {[info exists id2Tag($id)]} {return $id2Tag($id)}
    return {}
  };# id2Tag

  ##########################################################################
  ## Event: getSelectedAnnotationIds
  ##########################################################################
  method getSelectedAnnotationIds {} {
    set ids {}
    foreach tag $selected_tags {
      lappend ids [my tag2Id $tag]
    }
    return $ids
  };# getSelectedAnnotationIds

  method getAnnotatedTextRanges {id} {
    if {![string length $document]} {return {}}
    return [tip_GetAnnotatedTextRanges $document \
                [tip_GetAnnotation $document $id]]
  };# getAnnotatedTextRanges

  ##########################################################################
  ## Event: onTagPress
  ##########################################################################
  method onTagPress {id value} {
    # puts "[self] onTagPress: id:$id, value:$value"
    my selectTag $value:$id
  };# onTagPress

  ##########################################################################
  ## Event: onClick
  ##########################################################################
  method onClick {} {
    my unselectSelectedTags
    foreach one $on_click_listeners {
      $one onClick [self]
    }
  };# onClick

  ##########################################################################
  ## Event: mouseEventForWindow
  ##########################################################################
  method mouseEventForWindow {w} {
    foreach one $on_click_listeners {
      $one mouseEventForWindow $w
    }
  };# mouseEventForWindow

  method unselectSelectedTags {{notify_annotator 1}} {
    foreach tag $selected_tags {
      catch {$text_widget tag configure $tag -relief flat -borderwidth 0}
      if {$notify_annotator} {
        catch {$annotator_object unselect [my tag2Value $tag] \
                                          [my tag2Id $tag] $tag}
      }
    }
    set selected_tags {}
    catch {
      $toplevel_object setOverlappingSelectorItems {}
    }
  };# unselectSelectedTags

  method selectTag {tag {notify_annotator 1}} {
    if {![my validTextWidget]} {return}
    my unselectSelectedTags
    catch {my raiseTag $tag}
    $text_widget tag configure $tag -relief ridge -borderwidth 4
    set selected_tags [list $tag]
    set id [my tag2Id $tag]
    if {$notify_annotator} {
      catch {$annotator_object select [my tag2Value $tag] $id $tag}
    }
  };# selectTag

  ##########################################################################
  ## Method: raiseTag
  ##########################################################################
  ## Ensure that the tag is visible, by re-organising the order of all tags
  ## that overlap
  method raiseTag {tag} {
    set tag_id [my tag2Id $tag]
    set ann [tip_GetAnnotation $document $tag_id]
    tip_GetFirstSpanOffsets $ann start end
    ## Get overlapping annotations...
    set overlapping_annotations [my findOverlappingAnnotations $tag_id]
    set overlapping_annotations_sorted \
                                [tip_SortAnnotationSet $overlapping_annotations]
    set overlapping_descriptions [my getDescriptionFromOverlappingAnnotations \
                                        $overlapping_annotations]
    catch {
      $toplevel_object setOverlappingSelectorItems $overlapping_descriptions
    }
    ## The first annotation is always the one that spans the greater range...
    if {[llength $overlapping_annotations_sorted] > 1} {
      set one [lindex $overlapping_annotations_sorted 0]
      set id  [tip_GetId $one]
      set overlapping_annotations        [my findOverlappingAnnotations $id]
      set overlapping_annotations_sorted [tip_SortAnnotationSet \
                                              $overlapping_annotations]
    }
    ## Since we have the tags ordered, arrange their display priority...
    set prev_tag back
    foreach one $overlapping_annotations_sorted {
      ## Ensure that we are not covering exactly the same span as the original
      ## Annotation...
      set id    [tip_GetId $one]
      set ctag  [my id2Tag $id]
      if {$id ne $tag_id} {
        tip_GetFirstSpanOffsets $one s e
        if {$s == $start && $e == $end} {
          catch {$text_widget tag lower $ctag $prev_tag}
          continue
        }
      }
      catch {$text_widget tag raise $ctag $prev_tag}
      set prev_tag $ctag
    }
    #$text_widget tag raise $tag
    $text_widget tag raise sel
  };# raiseTag

  method selectOverlappingItem {item} {
    my selectTag [lindex $item 0]
  };# selectOverlappingItem

  method findOverlappingAnnotations {id} {
    if {![string length $annotator_object]} {return}
    set annotation  [$annotator_object getAnnotation]
    set Annotations [tip_SelectAnnotations $document $annotation]
    set Overlapping {}
    set ann [tip_GetAnnotation $document $id]
    foreach span [tip_GetSpans $ann] {
      tip_GetSpanOffsets $span start end
      set containing [lsort -unique [concat \
        [tip_AnnotationsContaining $Annotations $start] \
        [tip_AnnotationsContaining $Annotations $end] \
        [tip_AnnotationsInRange    $Annotations $start $end]]]
      set Overlapping [concat $Overlapping $containing]
    }
    return [lsort -unique $Overlapping]
  };# findOverlappingAnnotations

  method getDescriptionFromOverlappingAnnotations {Annotations} {
    if {![string length $annotator_object]} {return}
    set attribute [$annotator_object getAttribute]
    set description {}
    foreach ann $Annotations {
      set attr [tip_GetAttribute $ann $attribute]
      if {![llength $attr]} {continue}
      set id    [tip_GetId $ann]
      set value [tip_GetValueValue $attr]
      set text  [tip_GetFirstAnnotatedTextRange $document $ann]
      if {[string length $text] > 100} {
        set text "[string range $text 0 97]..."
      }
      lappend description "$value:$id $text"
    }
    return $description
  };# getDescriptionFromOverlappingAnnotations

  ##########################################################################
  ## Event: onTagEnter
  ##########################################################################
  method onTagEnter {id value} {
  };# onTagEnter

  ##########################################################################
  ## Event: onTagLeave
  ##########################################################################
  method onTagLeave {id value} {
  };# onTagLeave

  method registerAnnotatorObject {obj} {
    set annotator_object $obj
  };# registerAnnotatorObject

  method validTextWidget {} {return [winfo exists $text_widget]}

  method getWidgetForManagingBindings {} {return $text_widget}

  method getCollection {} {return $collection}
  method getDocument   {} {return $document}

  method getIndexAtxy {x y} {return [$text_widget index @$x,$y]}

  method selectionExists {} {
    return [llength [$text_widget tag ranges sel]]
    # if {![catch {$text_widget index sel.first} first] &&
    #      [string length $first]} {
    #   return 1
    # }
    # return 0
  };# selectionExists

  method selectionClear         {} {$text_widget tag remove sel 0.0 end}
  method getSelectionStartIndex {} {return [$text_widget index sel.first]}
  method getSelectionEndIndex   {} {return [$text_widget index sel.last]}
  method setSelectionIndexes    {first last} {
    $text_widget tag remove sel 0.0 end
    $text_widget tag add sel $first $last
  };# setSelectionIndexes
  method addSelectionIndexes    {first last} {
    $text_widget tag add sel $first $last
  };# addSelectionIndexes

  method seeIndex {index} {$text_widget see $index}

  method setFocus {} {focus $text_widget}

  method call {args} {
    return [my {*}$args]
  };# call

  method registerOnClickListener {args} {
    lappend on_click_listeners {*}$args
  };# registerOnClickListener

  classmethod loadImage {image} {return [GUI_LoadImage $image]}
  classmethod msg       {code}  {return [::ELEP::WindowUtilities::msg    $code]}
  classmethod msgVar    {code}  {return [::ELEP::WindowUtilities::msgVar $code]}
  
};# class ELEP::ViewerBase::TextWidgetDisplay

package provide ELEP::ViewerBase::TextWidgetDisplay 1.0
