##############################################################################
##  TextAnnotatorAlignment.tcl:                                             ##
## --------------------------------------------------                       ##
##  This class implements a tool for editing Annotations on an aligned      ##
##  corpus.                                                                 ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::VisualAnnotators::TextAnnotator
package require ELEP::ViewerBase::TextWidgetDisplayAlignment

##
## ELEP::VisualAnnotators::TextAnnotatorAlignment_DisplayObjectSwitcher
## This class holds two display objects, and switches them according to user
## input.
##
oo::class create \
ELEP::VisualAnnotators::TextAnnotatorAlignment_DisplayObjectSwitcher {
  variable display_objects \
           display_object \
           indicator_widgets \
           indicator_widget \
           indicator_colour

  constructor {args} {
    set display_objects   {}
    set display_object    {}
    set indicator_widgets {}
    set indicator_widget  {}
    set indicator_colour  green

    foreach {w i} $args {
      lappend display_objects   [string trim $w :]
      lappend indicator_widgets $i
    }
    set display_object   [lindex $display_objects   0]
    set indicator_widget [lindex $indicator_widgets 0]
    my indicatorOn
  };# constructor

  method call {method args} {
    switch -exact $method {
      registerAnnotatorObject -
      annotateAnnotationSet {
        foreach one $display_objects {
          $one $method {*}$args
        }
      }
      annotateDocument -
      default {
        $display_object call $method {*}$args
      }
    }
  };# call

  method callAll {method args} {
    foreach one $display_objects {
      $one $method {*}$args
    }
  };# call

  method indicatorOff {} {
    # $display_object unselectSelectedTags
    $indicator_widget configure -background \
      [[$display_object call getWidgetForManagingBindings] cget -background]
  };# indicatorOff

  method indicatorOn {} {
    $indicator_widget configure -background $indicator_colour
  };# indicatorOn

  method setActive {o} {
    if {$o eq $display_object} {return}
    my indicatorOff
    set index [lsearch -exact $display_objects $o]
    if {$index < 0} {return}
    set display_object   $o
    set indicator_widget [lindex $indicator_widgets $index]
    my indicatorOn
  };# setActiveWidget

  method onClick {o} {
    my setActive [string trim $o :]
  };# onClick

  ##########################################################################
  ## Event: mouseEventForWindow
  ##########################################################################
  method mouseEventForWindow {w} {
    foreach one $display_objects {
      if {[lsearch -exact [$one getWidgetForManagingBindings] $w] != -1} {
        my setActive $one; return
      }
    }
  };# mouseEventForWindow

  ##########################################################################
  ## Event: getWidgetForManagingBindings
  ##########################################################################
  method getWidgetForManagingBindings {} {
    set widgets {}
    foreach one $display_objects {
      lappend widgets [$one getWidgetForManagingBindings]
    }
    return $widgets
  };# getWidgetForManagingBindings

  ##########################################################################
  ## Event: activateNextWidget
  ##########################################################################
  method activateNextWidget {} {
    set index [lsearch -exact $display_objects $display_object]
    if {$index < 0} {return}
    incr index
    if {$index >= [llength $display_objects]} {set index 0}
    my setActive [lindex $display_objects $index]
  };# activateNextWidget

};# class ELEP::VisualAnnotators::TextAnnotatorAlignment_DisplayObjectSwitcher

oo::class create ELEP::VisualAnnotators::TextAnnotatorAlignment {
  superclass ELEP::VisualAnnotators::TextAnnotator

  variable {*}[info class variables ELEP::VisualAnnotators::TextAnnotator] \
           switcher_object \
           display_object1 \
           display_widget2 \
           display_object2 \
           scrollbar_widget \
           line_numbers_widget \
           display_indicator_widget \
           display_indicator_widget2 \
           text_background_colour \
           indicator_colour \
           indicator_height \
           switched

  constructor {{col {}} {doc {}}} {
    set switcher_object           {}
    set display_object1           {}
    set display_widget2           {}
    set display_object2           {}
    set scrollbar_widget          {}
    set line_numbers_widget       {}
    set display_indicator_widget  {}
    set display_indicator_widget2 {}
    set text_background_colour    {}
    set indicator_colour          green
    set indicator_height          4
    set switched                  0

    next $col $doc
  };# class constructor

  destructor {
    catch {$switcher_object destroy}
    next
  };# destructor

  method createLeftAreaWidgets {pane} {
    $pane add [my createScrolledDisplayWidget $pane] -weight 1
    $selector_object registerListener [self]
  };# createLeftAreaWidgets

  method createScrolledDisplayWidget {parent} {
    ttk::frame $parent.f
      ttk::frame $parent.f.fd
        # Add a frame with two text widgets, and a scrollbar
        ttk::frame $parent.f.fd.p
          my createDisplayWidget $parent.f.fd.p
          grid x $display_indicator_widget $display_indicator_widget2 \
              -sticky snew
          grid $line_numbers_widget $display_widget $display_widget2 \
              -sticky snew
          grid columnconfigure $parent.f.fd.p 1 -weight 1
          grid columnconfigure $parent.f.fd.p 2 -weight 1
          grid rowconfigure    $parent.f.fd.p 1 -weight 1

        set scrollbar_widget [ttk::scrollbar $parent.f.fd.s \
          -orient vertical -command [list [self] yview]]
        grid $parent.f.fd.p $parent.f.fd.s -sticky snew
        grid columnconfigure $parent.f.fd 0 -weight 1
        grid rowconfigure    $parent.f.fd 0 -weight 1
      grid $parent.f.fd -sticky snew
      # Add a widget for selecting overlapped annotated ranges.
      grid [my createOverlappingSelectorWidget $parent.f] -sticky snew
      grid columnconfigure $parent.f 0 -weight 1
      grid rowconfigure $parent.f 0 -weight 1
    return $parent.f
  };# createScrolledDisplayWidget

  method createDisplayWidget {parent} {
    ## Create the first text widget/object...
    set display_widget  [text $parent.text1 -borderwidth 0 -relief flat \
        -exportselection 0 -font FixedTextFont -setgrid 0 \
        -yscrollcommand [list [self] yset] -wrap char]
    set text_background_colour [$display_widget cget -background]
    set display_indicator_widget [frame $parent.ind1 -height $indicator_height \
        -bd 0 -background $text_background_colour]
    set display_object1  [ELEP::ViewerBase::TextWidgetDisplayAlignment create \
        [::ELEP::Base::auto ELEP::VisualAnnotators::] [self] \
        $display_widget $display_indicator_widget]
    
    ## Create the second text widget/object...
    set display_widget2 [text $parent.text2 -borderwidth 0 -relief flat \
        -exportselection 0 -font FixedTextFont -setgrid 0 \
        -yscrollcommand [list [self] yset] -wrap char]
    set display_indicator_widget2 [frame $parent.ind2 -height $indicator_height\
        -bd 0 -background $text_background_colour]
    set display_object2 [ELEP::ViewerBase::TextWidgetDisplayAlignment create \
        [::ELEP::Base::auto ELEP::VisualAnnotators::] [self] \
        $display_widget2 $display_indicator_widget2]
    ## Create the switcher object...
    set display_object \
      [ELEP::VisualAnnotators::TextAnnotatorAlignment_DisplayObjectSwitcher \
        create [::ELEP::Base::auto ELEP::VisualAnnotators::] \
                  $display_object1 $display_indicator_widget \
                  $display_object2 $display_indicator_widget2]
    ## Register the switcher to receive onClick events...
    $display_object1 registerOnClickListener $display_object
    $display_object2 registerOnClickListener $display_object

    set line_numbers_widget [canvas $parent.ln_canvas -width 26 \
        -highlightthickness 0 -background [$display_widget cget -background]]
    # $display_object registerAlternativeWidget \
        $display_widget2 $display_indicator_widget2
    return $display_widget
  };# createDisplayWidget

  method createAnnotatorObject {parent} {
    next $parent
    $display_object2 registerAnnotatorObject $annotator_object
  };# createAnnotatorObject

  method setTitle {} {
    my title {Aligned Text Visual Annotation Editor...}
  };# setTitle
  
  ##########################################################################
  ## Event: switchDisplayWidgets
  ##########################################################################
  method switchDisplayWidgets {} {
    grid forget $display_indicator_widget $display_indicator_widget2 \
                $display_widget $display_widget2
    if {$switched} {
      grid $display_indicator_widget  -column 1 -row 0 -sticky snew
      grid $display_indicator_widget2 -column 2 -row 0 -sticky snew
      grid $display_widget            -column 1 -row 1 -sticky snew
      grid $display_widget2           -column 2 -row 1 -sticky snew
    } else {
      grid $display_indicator_widget2 -column 1 -row 0 -sticky snew
      grid $display_indicator_widget  -column 2 -row 0 -sticky snew
      grid $display_widget2           -column 1 -row 1 -sticky snew
      grid $display_widget            -column 2 -row 1 -sticky snew

    }
    set switched [expr {!$switched}]
  };# switchDisplayWidgets
 
  ##########################################################################
  ## Event: switchActiveDisplayWidget
  ##########################################################################
  method switchActiveDisplayWidget {} {
    $display_object activateNextWidget
  };# switchActiveDisplayWidget

  ##########################################################################
  ## Event: onDocumentChange
  ##########################################################################
  method onDocumentChange {col doc} {
    foreach {a1 a2} [tip_SelectAnnotations $doc alignment-info] {
      foreach attr [tip_GetAttributes $a1] {
        set val [tip_GetValueValue $attr]
        if {[string length $val]} {
          $display_object1 setAlignmentSet $val ; break
        }
      }
      foreach attr [tip_GetAttributes $a2] {
        set val [tip_GetValueValue $attr]
        if {[string length $val]} {
          $display_object2 setAlignmentSet $val ; break
        }
      }
      break
    }
    $display_object1 onDocumentChange $col $doc
    $display_object2 onDocumentChange $col $doc
    my draw_line_numbers
  };# onDocumentChange

  ##########################################################################
  ## Event: yset (called by the text widgets)
  ##########################################################################
  method yset {args} {
    $scrollbar_widget set {*}$args
    my yview moveto [lindex [$scrollbar_widget get] 0]
    my draw_line_numbers
  };# yset

  ##########################################################################
  ## Event: yview (called by the scrollbar widget)
  ##########################################################################
  method yview {args} {
    $display_widget  yview {*}$args
    $display_widget2 yview {*}$args
    my draw_line_numbers
  };# yview

  ##########################################################################
  ## Event: line numbers
  ##########################################################################
  method draw_line_numbers {args} {
    # only redraw if args are null (meaning we were called by a binding)
    # or called by the trace and the command could potentially change
    # the size of a line.
    set benign {
      mark bbox cget compare count debug dlineinfo
      dump get index mark peer search
    }
    if {[llength $args] == 0 || [lindex $args 0 1] ni $benign} {
      $line_numbers_widget delete all
      set i [$display_widget index @0,0]
      while true {
        set string [$display_widget get $i "$i lineend"]
        set dline [$display_widget dlineinfo $i]
        if {[llength $dline] == 0} break
        set height [lindex $dline 3]
        set y [lindex $dline 1]
        set cy [expr {$y + int($height/2.0)}]
        set linenum [lindex [split $i .] 0]
        $line_numbers_widget create text 0 $y \
           -anchor nw -text $linenum
        set i [$display_widget index "$i + 1 line"]
      }
    }
  };# draw_line_numbers

  method createToolbar {} {
    set tbarFrame [my getToolbarFrame]
    set tbar $tbarFrame.textAnnotatorAlignment
    if {![winfo exists $tbar]} {
      ttk::frame $tbar -class Toolbar
      my createToolbarWidgets $tbar
      pack $tbar -padx 2 -pady 2 -side left
    }
  };# createToolbar

  method createToolbarWidgets {tbar} {
    ttk::button $tbar.switch -textvariable [my msgVar SwitchAnnotationObjects] \
      -command "[self] switchDisplayWidgets" -style Toolbutton
    ttk::button $tbar.activeswitch -textvariable \
      [my msgVar SwitchActiveAnnotationObject] \
      -command "[self] switchActiveDisplayWidget" -style Toolbutton
    grid $tbar.switch $tbar.activeswitch -sticky news -padx 2 -pady 2
  };# createToolbarWidgets

};# class ELEP::VisualAnnotators::TextAnnotatorAlignment

package provide ELEP::VisualAnnotators::TextAnnotatorAlignment 1.0
