##############################################################################
##  TextCoreferenceAnnotator.tcl:                                           ##
## --------------------------------------------------                       ##
##  This class implements a tool for editing Annotations on a text widget.  ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities::ToplevelWindow
package require ELEP::ViewerBase::TextWidgetDisplay
package require ELEP::ViewerBase::DocumentSelector
package require ELEP::ViewerBase::CoreferenceAnnotator
package require widget::scrolledwindow

oo::class create ELEP::VisualAnnotators::TextCoreferenceAnnotator {
  superclass ELEP::WindowUtilities::ToplevelWindow

  variable {*}[info class variables ELEP::WindowUtilities::ToplevelWindow] \
           display_widget \
           overlapping_selector_widget \
           overlapping_selector_itemnu_widget \
           display_object \
           annotator_object \
           selector_object

  constructor {{col {}} {doc {}}} {
    next
    my createWidgets $col $doc
  };# constructor

  method createWidgets {col doc} {
    ## Add a document selector object (toolbar).
    my createDocumentSelector $col $doc
    ## Add a horizontal pane.
    ttk::panedwindow $client_area.pane -orient horizontal
    ## Create the widgets on the left side of the horizontal pane widget.
    my createLeftAreaWidgets   $client_area.pane
    ## Create the widgets on the right side of the horizontal pane widget.
    my createRightAreaWidgets  $client_area.pane
    ## Pack the pane widget.
    pack $client_area.pane -fill both -expand 1
    ## Set the proper title to our toplevel window.
    my setTitle
  };# createWidgets

  method createDocumentSelector {col doc} {
    set selector_object [ELEP::ViewerBase::DocumentSelector create \
        [::ELEP::Base::auto ELEP::VisualAnnotators::] [self] {} $col $doc]
  };# createDocumentSelector

  method createLeftAreaWidgets {pane} {
    ## On left, add a display widget object...
    $pane add [my createScrolledDisplayWidget $pane] -weight 1
    $selector_object registerListener $display_object
  };# createLeftAreaWidgets

  method createRightAreaWidgets {pane} {
    ## On the rigth side of the pane widget, add a vertical pane widget, to
    ## separate two areas: right top & right bottom...
    ttk::panedwindow $pane.right -orient vertical
    ## Create the widgets on the right top side of the vertical pane widget.
    my createRightTopAreaWidgets    $pane.right
    ## Create the widgets on the right bottom side of the vertical pane widget.
    my createRightBottomAreaWidgets $pane.right
    $pane add $pane.right -weight 1
  };# createRightAreaWidgets

  method createRightTopAreaWidgets {pane} {
  };# createRightTopAreaWidgets

  method createRightBottomAreaWidgets {pane} {
    ## On right, add a CoreferenceAnnotator object.
    my createAnnotatorObject $pane
    $pane add [$annotator_object getframe] -weight 1
  };# createRightBottomAreaWidgets

  method createScrolledDisplayWidget {parent} {
    ttk::frame $parent.f
      # Add a scrolled window, which will host the display widget.
      widget::scrolledwindow $parent.f.sw
      $parent.f.sw setwidget [my createDisplayWidget $parent.f.sw]
      grid $parent.f.sw -sticky snew
      # Add a widget for selecting overlapped annotated ranges.
      grid [my createOverlappingSelectorWidget $parent.f] -sticky snew
      grid columnconfigure $parent.f 0 -weight 1
      grid rowconfigure $parent.f 0 -weight 1
    return $parent.f
  };# createScrolledDisplayWidget

  method createDisplayWidget {parent} {
    set display_widget [text $parent.text -borderwidth 0 -relief flat \
        -exportselection 0]
    set display_object [ELEP::ViewerBase::TextWidgetDisplay create \
        [::ELEP::Base::auto ELEP::VisualAnnotators::] [self] $display_widget]
    return $display_widget
  };# createDisplayWidget

  method createOverlappingSelectorWidget {parent} {
    ttk::frame $parent.fos
      set overlapping_selector_widget $parent.fos.selector
      set overlapping_selector_itemnu_widget $parent.fos.selitems
      ttk::label $parent.fos.lbl -text {Overlapping areas: }
      ttk::label $overlapping_selector_itemnu_widget -text 0
      ttk::combobox $overlapping_selector_widget -exportselection 0
      $overlapping_selector_widget state {disabled readonly}
      grid $parent.fos.lbl $overlapping_selector_itemnu_widget \
                         $overlapping_selector_widget -sticky snew
      grid columnconfigure $parent.fos 2 -weight 1
      bind $overlapping_selector_widget <<ComboboxSelected>> \
        [list [self] onOverlappingSelection]
    return $parent.fos
  };# createOverlappingSelectorWidget

  method createAnnotatorObject {parent} {
    set annotator_object [ELEP::ViewerBase::CoreferenceAnnotator create \
        [::ELEP::Base::auto ELEP::VisualAnnotators::] [self] \
        $parent $display_object]
  };# createAnnotatorObject

  method setTitle {} {
    my title {Text Visual Annotation Editor...}
  };# setTitle

  ##########################################################################
  ## Event: onOverlappingSelection
  ##########################################################################
  method onOverlappingSelection {} {
    set selected [$overlapping_selector_widget get]
    catch {
      $display_object call selectOverlappingItem $selected
    }
  };# onOverlappingSelection

  method setOverlappingSelectorItems {items} {
    $overlapping_selector_itemnu_widget configure -text [llength $items]
    $overlapping_selector_widget configure -values $items
    $overlapping_selector_widget set {}
    if {[llength $items]} {
      $overlapping_selector_widget set {Please select overlapping area...}
      $overlapping_selector_widget state !disabled
    } else {
      $overlapping_selector_widget state disabled
    }
  };# setOverlappingSelectorItems

  method show {} {
    update
    if {[string equal [$annotator_object show] cancel]} {
      after idle "destroy $win"; return
    }
    next
    $selector_object show
  };# show

  method loadDocument {doc_id} {
    $selector_object loadDocument $doc_id
  };# loadDocument

};# class ELEP::VisualAnnotators::TextCoreferenceAnnotator

package provide ELEP::VisualAnnotators::TextCoreferenceAnnotator 1.0
