##############################################################################
##  Google.tcl:                                                             ##
## --------------------------------------------------                       ##
##  This class implements methods for performing searches through the       ##
##  Google API.                                                             ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require TclOO
package require http
package require json
package require uri
package require log

oo::class create ELEP::WebSearch::Google {
  ##
  ## Documentation:
  ## http://code.google.com/intl/el-GR/apis/ajaxsearch/documentation/reference.html#_intro_fonje
  ##
  variable APIKey              \
           APIReferer          \
           APIURL              \
           APIVersion          \
           APILanguage         \
           URL                 \
           Headers             \
           SearchParametersKey \
           resultsFilter

  constructor {args} {
    set APIKey        ABQIAAAAZD-3d8UZPhlpVaqIgg54uBTqwDnIPXVuSbbl2nLbdZh9YSCAuhRjbygmqji7_lx0ZErAwVh7A9Cuaw
    set APIReferer    http://www.ellogon.org/
    set APIURL        http://ajax.googleapis.com/ajax/services/search
    set APIVersion    1.0
    set APILanguage   en
    set Headers {}
    set SearchParametersKey ELEP::WebSearch::Google
    set resultsFilter       {my keepResult?}

    my configure {*}$args
    my init
  };# constructor

  method init {} {
    foreach searcher {web local video blogs news books images patent} {
      set URL($searcher) $APIURL/${searcher}?
    }
    set Headers [list Referer $APIReferer]
  };# init

  method search { tokens {mode web}     {size large} {start 0}
                         {arguments {}} {headers {}} } {
    set url $URL($mode)
    log::log debug "(?) Keywords: \"[join $tokens]\""
    append url [http::formatQuery q [join $tokens { }] v $APIVersion \
                                  rsz $size hl $APILanguage \
                                  key $APIKey start $start]
    if {[llength $arguments]} {
      append url [http::formatQuery {*}$arguments]
    }
    set query_dict [dict create tokens    $tokens    mode    $mode size $size \
                                start     $start \
                                arguments $arguments headers $headers]
    set jd [my post $url $headers]
    dict set jd $SearchParametersKey $query_dict
    return $jd
  };# search

  method getResults {jd} {
    return [dict get [dict get $jd responseData] results]
  };# getResults

  method getCursor {jd} {
    return [dict get [dict get $jd responseData] cursor]
  };# getCursor

  method getSearchParameters {jd} {
    return [dict get $jd $SearchParametersKey]
  };# getSearchParameters

  method nextPage {jd} {
    set nstart [my getNextPageIndex $jd]
    if {$nstart == -1} {
      ## No more pages!
      return {}
    }
    ## Restore the parameters...
    dict for {key val} [my getSearchParameters $jd] { set $key $val }
    return [my search $tokens $mode $size $nstart $arguments $headers]
  };# nextPage

  method getFromCursor {jd field} {
    set cursor [my getCursor $jd]
    if {![dict exists $cursor $field]} {
      log::log warning "Field \"$field\" not in cursor!"
      log::log warning $jd
      return {}
    }
    return [dict get $cursor $field]
  };# getFromCursor

  method currentPageIndex {jd} {
    set index [my getFromCursor $jd currentPageIndex]
    if {$index eq ""} {return 0}
    return $index
  };# currentPageIndex

  method pages {jd} {
    return [my getFromCursor $jd pages]
  };# pages

  method estimatedResultCount {jd} {
    set count [my getFromCursor $jd estimatedResultCount]
    if {$count eq ""} {return 0}
    return $count
  };# estimatedResultCount

  method getNextPageIndex {jd} {
    set pages [my pages $jd]
    if {[llength $pages]} {return -1}
    set currentPageIndex [my currentPageIndex $jd]
    incr currentPageIndex
    if {$currentPageIndex >= [llength $pages]} {return -1}
    set page [lindex [my pages $jd] $currentPageIndex]
    return [dict get $page start]
  };# getNextPageIndex

  method hasMorePages {jd} {
    return [expr {[my getNextPageIndex $jd] != -1}]
  };# hasMorePages

  method post {url {headers {}}} {
    # set url [string map {%2f / %3a :} $url]
    log::log debug "@@@ Quering Google: $url"
    set results {}
    set token [http::geturl $url -headers [concat $Headers $headers]]
    if { [http::ncode $token] == 200 } {
      set d [http::data $token]
      set results [json::json2dict $d]
      # log::log debug $results
      log::log debug \
               "@@@ estimatedResultCount: [my estimatedResultCount $results],\
                currentPage: [my currentPageIndex $results],\
                pages: [llength [my pages $results]],\
                hasMorePages: [my hasMorePages $results]"
    } else {
      log::log error "Error=[http::error $t]"
      log::log error "Status=[http::status $t]"
      log::log error "Code=[http::code $t]"
      log::log error "Ncode=[http::ncode $t]"
      log::log error "Data=[http::data $t]"
    }
    http::cleanup $token
    # puts [my getCursor $results]
    return $results
  };# post

  method getFirstNResults {tokens {N 10} {mode web} {size large} {start 0}
                                  {arguments {}} {headers {}} } {
    set results {}; set n 0
    set jd [my search $tokens $mode $size $start $arguments $headers]
    foreach one [my getResults $jd] {
      if {[{*}$resultsFilter $one]} {
        lappend results $one; incr n
        if {$n >= $N} {return $results}
      }
    }
    while {[my hasMorePages $jd]} {
      set jd [my nextPage $jd]
      foreach one [my getResults $jd] {
        if {[{*}$resultsFilter $one]} {
          lappend results $one; incr n
          if {$n >= $N} {return $results}
        }
      }
    }
    return $results
  };# getFirstNResults

  method keepResult? {result} {return true}

  method itemField {item args} {
    foreach one $args {
      if {[dict exists $item $one]} {
        return [dict get $item $one]
      }
    }
    return {}
  };# itemField

  method print {result} {
    foreach one $result {
      my print_item $one
    }
  };# print

  method print_item {item} {
    log::logMsg \
      ====================================================================
    foreach key {unescapedUrl blogUrl postUrl url visibleUrl cacheUrl title
                 titleNoFormatting author publishedDate content} {
      if {![dict exists $item $key]} {continue}
      log::logMsg "$key: \"[dict get $item $key]\""
    }
  };# print_item

  method configure {args} {
    foreach {varName varValue} $args {
      set _vname [string range $varName 1 end]
      my variable $_vname
      set $_vname $varValue
    }
  };# configure

  method cget {varName} {
    set _vname [string range $varName 1 end]
    my variable $_vname
    return [set $_vname]
  };# cget

};# class ELEP::WebSearch::Google

package provide ELEP::WebSearch::Google 1.0

# catch {console show}
# set o [ELEP::WebSearch::Google #auto]
# $o print [$o getFirstNResults {George Petasis} 20 blogs]
