##############################################################################
##  Technorati.tcl:                                                         ##
## --------------------------------------------------                       ##
##  This class implements methods for performing searches through the       ##
##  Technorati API.                                                         ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require TclOO
package require http
package require struct::tree
package require htmlparse
package require uri
package require log

oo::class create ELEP::WebSearch::Technorati {
  ##
  ## Documentation:
  ## http://technorati.com/developers/api/search.html
  ##
  variable APIKey              \
           APIReferer          \
           APIURL              \
           APIVersion          \
           APILanguage         \
           URL                 \
           Headers             \
           SearchParametersKey \
           resultsFilter       \
           _xml2dict_stack     \
           _xml2dict_jd

  constructor {args} {
    set APIKey        963637bd073e28e639da772b29dd2683
    set APIReferer    http://www.ellogon.org/
    set APIURL        http://api.technorati.com
    set APIVersion    1.0
    set APILanguage   en
    set Headers {}
    set SearchParametersKey ELEP::WebSearch::Technorati
    set resultsFilter       {my keepResult?}

    my configure {*}$args
    my init
  };# constructor

  method init {} {
    foreach searcher {search} {
      set URL($searcher) $APIURL/${searcher}?
    }
    foreach one {web local video blogs news books images patent} {
      set URL($one) $URL(search)
    }
    set Headers [list Referer $APIReferer]
  };# init

  method search { tokens {mode search}  {size 20} {start 1} {authority n}
                         {arguments {}} {headers {}} } {
    set url $URL($mode)
    log::log debug "(?) Keywords: \"[join $tokens]\""
    append url [http::formatQuery \
                      query [join $tokens { }] key       $APIKey \
                      limit  $size           language  $APILanguage \
                      format xml             authority $authority \
                      claim   1              start     $start]
    if {[llength $arguments]} {
      append url [http::formatQuery {*}$arguments]
    }
    set query_dict [dict create tokens    $tokens    mode    $mode size $size \
                                start     $start     claim   1 \
                                authority $authority \
                                arguments $arguments headers $headers]
    return [my post $url $headers $query_dict]
  };# search

  method getResults {jd} {
    return [dict get $jd tapi document item]
  };# getResults

  method getCursor {jd} {
    return [dict get $jd cursor]
  };# getCursor

  method getSearchParameters {jd} {
    return [dict get $jd $SearchParametersKey]
  };# getSearchParameters

  method nextPage {jd} {
    set nstart [my getNextPageIndex $jd]
    if {$nstart == -1} {
      ## No more pages!
      return {}
    }
    ## Restore the parameters...
    dict for {key val} [my getSearchParameters $jd] { set $key $val }
    return [my search $tokens $mode $size $nstart $arguments $headers]
  };# nextPage

  method getFromCursor {jd field} {
    set cursor [my getCursor $jd]
    if {![dict exists $cursor $field]} {
      return {}
    }
    return [dict get $cursor $field]
  };# getFromCursor

  method currentPageIndex {jd} {
    set index [my getFromCursor $jd currentPageIndex]
    if {$index eq ""} {return 0}
    return $index
  };# currentPageIndex

  method pages {jd} {
    return [my getFromCursor $jd pages]
  };# pages

  method estimatedResultCount {jd} {
    set count [my getFromCursor $jd estimatedResultCount]
    if {$count eq ""} {return 0}
    return $count
  };# estimatedResultCount

  method getNextPageIndex {jd} {
    set currentPageIndex [my currentPageIndex $jd]
    incr currentPageIndex
    set pages [my pages $jd]
    if {$currentPageIndex >= [llength $pages]} {return -1}
    set page [lindex [my pages $jd] $currentPageIndex]
    return [dict get $page start]
  };# getNextPageIndex

  method hasMorePages {jd} {
    return [expr {[my getNextPageIndex $jd] != -1}]
  };# hasMorePages

  method post {url {headers {}} {query_dict {}}} {
    log::log debug "@@@ Quering Technorati: $url"
    set results {}
    set token [http::geturl $url -headers [concat $Headers $headers]]
    if { [http::ncode $token] == 200 } {
      set d [http::data $token]
      # puts $d
      set results [my xml2dict $d]
      ## Always ensure that the keys tapi document item exists!
      if {![dict exists $results tapi]} {dict set results tapi {}}
      if {![dict exists $results tapi document]} {
        dict set results tapi document {}
      }
      if {![dict exists $results tapi document item]} {
        dict set results tapi document item {}
      }
      if {![dict exists $results tapi document result]} {
        dict set results tapi document result {}
      }
      dict set results $SearchParametersKey $query_dict
      set results [my fakeCursor $results]
      log::log debug \
               "@@@ estimatedResultCount: [my estimatedResultCount $results],\
                currentPage: [my currentPageIndex $results],\
                pages: [llength [my pages $results]]"
    } else {
      log::log error "Error=[http::error $token]"
      log::log error "Status=[http::status $token]"
      log::log error "Code=[http::code $token]"
      log::log error "Ncode=[http::ncode $token]"
      log::log error "Data=[http::data $token]"
    }
    http::cleanup $token
    # puts [my getCursor $results]
    return $results
  };# post

  method getFirstNResults {tokens {N 10} {mode search} {size 20} {start 1}
                                  {authority n} {arguments {}} {headers {}} } {
    set results {}; set n 0
    set jd [my search $tokens $mode $size $start $authority $arguments $headers]
    # puts $jd
    foreach one [my getResults $jd] {
      if {[{*}$resultsFilter $one]} {
        lappend results $one; incr n
        if {$n >= $N} {return $results}
      }
    }
    while {[my hasMorePages $jd]} {
      set jd [my nextPage $jd]
      foreach one [my getResults $jd] {
        if {[{*}$resultsFilter $one]} {
          lappend results $one; incr n
          if {$n >= $N} {return $results}
        }
      }
    }
    return $results
  };# getFirstNResults

  method fakeCursor {jd} {
    set max_pages 4
    set sp [my getSearchParameters $jd]
    set start [dict get $sp start]
    set size  [dict get $sp size]
    set pages {}
    set label 0
    for {set s $start} {$s <= $max_pages*$size+1} {incr s $size} {
      lappend pages [dict create start $s label [incr label]]
    }
    set cursor [dict create pages $pages \
                            estimatedResultCount [expr {$max_pages*$size}] \
                            currentPageIndex 0 moreResultsUrl {}]
    return [dict set jd cursor $cursor]
  };# fakeCursor

  method xml2dict {xml} {
    set tree [::struct::tree]
    ::htmlparse::2tree $xml $tree
    # ## Perform a walk over the tree to convert the tree into a dict...
    # $tree walkproc root -order pre -type bfs [list [self] xml2dict_cb]
    set hmstart [lindex [$tree children root] 0]
    set jd [my xml2dict_cb $tree $hmstart]
    $tree destroy
    return [dict get $jd hmstart]
  };# xml2dict

  method xml2dict_cb {tree node} {
    set tag      [$tree get $node type]
    set children [$tree children $node]
    set nch      [llength $children]
    if {$nch == 0} {
      return [$tree get $node data]
    } elseif {$nch == 1} {
      set child [lindex $children 0]
      return [list $tag [my xml2dict_cb $tree $child]]
    } else {
      set result [dict create]
      foreach one $children {
        lassign [my xml2dict_cb $tree $one] t v
        dict lappend result $t $v
      }
      ## Filter values that are a single element!
      dict for {t v} $result {
        if {[llength $v] == 1} {dict set result $t [lindex $v 0]}
      }
      return [list $tag $result]
    }
    return {}
  };# xml2dict_cb

  method keepResult? {result} {return true}

  method itemField {item args} {
    foreach one $args {
      switch -nocase $one {
        postUrl {set one permalink}
      }
      if {[dict exists $item $one]} {
        return [dict get $item $one]
      } elseif {[dict exists $item weblog $one]} {
        return [dict get $item weblog $one]
      }
    }
    return {}
  };# itemField

  method print {result} {
    foreach one $result {
      my print_item $one
    }
  };# print

  method print_item {item} {
    log::logMsg \
      ====================================================================
    foreach key {name url rssurl atomurl inboundblogs inboundlinks lastupdate
                 author title excerpt created permalink} {
      if {[dict exists $item $key]} {
        log::logMsg "$key: \"[dict get $item $key]\""
      } elseif {[dict exists $item weblog $key]} {
        log::logMsg "$key: \"[dict get $item weblog $key]\""
      }
    }
  };# print_item

  method configure {args} {
    foreach {varName varValue} $args {
      set _vname [string range $varName 1 end]
      my variable $_vname
      set $_vname $varValue
    }
  };# configure

  method cget {varName} {
    set _vname [string range $varName 1 end]
    my variable $_vname
    return [set $_vname]
  };# cget

};# class ELEP::WebSearch::Technorati

package provide ELEP::WebSearch::Technorati 1.0

# catch {console show}
# set o [ELEP::WebSearch::Technorati #auto]
# $o print [$o getFirstNResults {Obama} ]
