##############################################################################
##  Yahoo.tcl:                                                             ##
## --------------------------------------------------                       ##
##  This class implements methods for performing searches through the       ##
##  Yahoo API.                                                             ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require TclOO
package require http
package require json
package require uri
package require log

oo::class create ELEP::WebSearch::Yahoo {
  ##
  ## Documentation:
  ## http://code.google.com/intl/el-GR/apis/ajaxsearch/documentation/reference.html#_intro_fonje
  ##
  public variable APIKey        N8k5J1rV34HapN.K1c3qtR0Za2gDzProrQKvAX77Uc.GDY4eCaiBLo7XTqZFKGo-
  public variable APIReferer    http://www.ellogon.org/
  public variable APIURL        http://search.yahooapis.com/SiteExplorerService
  public variable APIVersion    1
  public variable APILanguage   en

  public variable URL
  public variable Headers {}
  public variable SearchParametersKey ELEP::WebSearch::Yahoo

  public variable resultsFilter       keepResult?

  constructor {args} {
    configure {*}$args
    init
  };# constructor

  method init {} {
    foreach searcher {inlinkData} {
      set URL($searcher) $APIURL/V${APIVersion}/${searcher}?
    }
  };# init

  method search { tokens {mode inlinkData} {results 100} {start 1}
                         {arguments {}} {headers {}} } {
    set url $URL($mode)
    log::log debug "(?) Keywords: \"[join $tokens]\""
    append url [http::formatQuery query [join $tokens { }] \
                                  results $results start $start \
                                  output json appid $APIKey]
    if {[llength $arguments]} {
      append url [http::formatQuery {*}$arguments]
    }
    set query_dict [dict create tokens    $tokens    mode    $mode \
                                results   $results \
                                start     $start \
                                arguments $arguments headers $headers]
    set jd [post $url $headers]
    dict set jd $SearchParametersKey $query_dict
    return $jd
  };# search

  method getResults {jd} {
    set count [getFromCursor $jd totalResultsReturned]
    if {$count eq "" || $count == 0} {return {}}
    set value [dict get [dict get $jd ResultSet] Result]
    if {$count == 1} {
      return [list $value]
    }
    return $value
  };# getResults

  method getCursor {jd} {
    return [dict get $jd ResultSet]
  };# getCursor

  method getSearchParameters {jd} {
    return [dict get $jd $SearchParametersKey]
  };# getSearchParameters

  method nextPage {jd} {
    set nstart [getNextPageIndex $jd]
    if {$nstart == -1} {
      ## No more pages!
      return {}
    }
    ## Restore the parameters...
    dict for {key val} [getSearchParameters $jd] { set $key $val }
    return [search $tokens $mode $size $nstart $arguments $headers]
  };# nextPage

  method getFromCursor {jd field} {
    set cursor [getCursor $jd]
    if {![dict exists $cursor $field]} {
      log::log warning "Field \"$field\" not in cursor!"
      log::log warning $jd
      return {}
    }
    return [dict get $cursor $field]
  };# getFromCursor

  method currentPageIndex {jd} {
    # TODO: Fix this
    return 1
    set index [getFromCursor $jd currentPageIndex]
    if {$index eq ""} {return 0}
    return $index
  };# currentPageIndex

  method pages {jd} {
    # TODO: Fix this
    return 1
    return [getFromCursor $jd pages]
  };# pages

  method estimatedResultCount {jd} {
    set count [getFromCursor $jd totalResultsAvailable]
    if {$count eq ""} {return 0}
    return $count
  };# estimatedResultCount

  method getNextPageIndex {jd} {
    set pages [pages $jd]
    if {[llength $pages]} {return -1}
    set currentPageIndex [currentPageIndex $jd]
    incr currentPageIndex
    if {$currentPageIndex >= [llength $pages]} {return -1}
    set page [lindex [pages $jd] $currentPageIndex]
    return [dict get $page start]
  };# getNextPageIndex

  method hasMorePages {jd} {
    return [expr {[getNextPageIndex $jd] != -1}]
  };# hasMorePages

  method post {url {headers {}}} {
    # set url [string map {%2f / %3a :} $url]
    log::log debug "@@@ Quering Yahoo: $url"
    set results {}
    set token [http::geturl $url -headers [concat $Headers $headers]]
    if { [http::ncode $token] == 200 } {
      set d [http::data $token]
      # log::log debug $d
      set results [json::json2dict $d]
      log::log debug $results
      log::log debug \
               "@@@ estimatedResultCount: [estimatedResultCount $results],\
                currentPage: [currentPageIndex $results],\
                pages: [llength [pages $results]],\
                hasMorePages: [hasMorePages $results]"
    } else {
      log::log error "Error=[http::error $t]"
      log::log error "Status=[http::status $t]"
      log::log error "Code=[http::code $t]"
      log::log error "Ncode=[http::ncode $t]"
      log::log error "Data=[http::data $t]"
    }
    http::cleanup $token
    # puts [getCursor $results]
    return $results
  };# post

  method getFirstNResults {tokens {N 10} {mode inlinkData} {size 100} {start 1}
                                  {arguments {}} {headers {}} } {
    set results [list]; set n 0
    set jd [search $tokens $mode $size $start $arguments $headers]
    foreach one [getResults $jd] {
      if {[$resultsFilter $one]} {
        lappend results $one; incr n
        if {$n >= $N} {return $results}
      }
    }
    while {[hasMorePages $jd]} {
      set jd [nextPage $jd]
      foreach one [getResults $jd] {
        if {[$resultsFilter $one]} {
          lappend results $one; incr n
          if {$n >= $N} {return $results}
        }
      }
    }
    return $results
  };# getFirstNResults

  proc keepResult? {result} {return true}

  proc itemField {item args} {
    foreach one $args {
      set one [itemFieldNative $one]
      if {[dict exists $item $one]} {
        return [dict get $item $one]
      }
    }
    return {}
  };# itemField

  proc itemFieldNative {field} {
    switch -nocase $field {
      postUrl {return ClickUrl}
    }
    return Url
  };# itemFieldNative

  method print {result} {
    foreach one $result {
      print_item $one
    }
  };# print

  proc print_item {item} {
    log::logMsg \
      ====================================================================
    foreach key {unescapedUrl blogUrl postUrl url visibleUrl cacheUrl title
                 titleNoFormatting author publishedDate content} {
      if {![dict exists $item $key]} {continue}
      log::logMsg "$key: \"[dict get $item $key]\""
    }
  };# print_item

};# class ELEP::WebSearch::Yahoo

package provide ELEP::WebSearch::Yahoo 1.0

# catch {console show}
# set o [ELEP::WebSearch::Yahoo #auto]
# $o print [$o getFirstNResults {George Petasis} 20 blogs]
