# ------------------------------------------------------------------------------
#  ButtonNotebook.tcl
#     This widget is similar to the notebook widget, but has a vertical layout.
#     Tabs are simple buttons, and pages are placed among the buttons.
#     The layout of this widget resembles the button bar of Outlook.
#  
# Copyright (C) 2000-2006 George P. Petasis
#
# $Revision: 1.0 $
#
# ---------------------------------------------------------------------------
#  Index of commands:
#     - ButtonNoteBook::create
#     - ButtonNoteBook::configure
#     - ButtonNoteBook::cget
#     - ButtonNoteBook::compute_size
#     - ButtonNoteBook::insert
#     - ButtonNoteBook::delete
#     - ButtonNoteBook::itemconfigure
#     - ButtonNoteBook::itemcget
#     - ButtonNoteBook::bindtabs
#     - ButtonNoteBook::raise
#     - ButtonNoteBook::see
#     - ButtonNoteBook::page
#     - ButtonNoteBook::pages
#     - ButtonNoteBook::index
#     - ButtonNoteBook::getframe
#     - ButtonNoteBook::create_page
#   Private commands (event bindings)
#     - ButtonNoteBook::_test_page
#     - ButtonNoteBook::_itemconfigure
#     - ButtonNoteBook::_compute_width
#     - ButtonNoteBook::_get_x_page
#     - ButtonNoteBook::_xview
#     - ButtonNoteBook::_highlight
#     - ButtonNoteBook::_select
#     - ButtonNoteBook::_draw_page
#     - ButtonNoteBook::_resize
# ---------------------------------------------------------------------------

namespace eval ButtonNoteBook {
    Widget::define ButtonNoteBook notebook DynamicHelp

    namespace eval Page {
        Widget::declare ButtonNoteBook::Page {
            {-state      Enum       normal 0 {normal disabled}}
            {-relief     Enum      raised  0 {raised sunken flat ridge solid groove link}}
            {-createcmd  String     ""     0}
            {-raisecmd   String     ""     0}
            {-leavecmd   String     ""     0}
            {-image      TkResource ""     0 label}
            {-text       String     ""     0}
            {-foreground         String     ""     0}
            {-background         String     ""     0}
            {-activeforeground   String     ""     0}
            {-activebackground   String     ""     0}
            {-disabledforeground String     ""     0}
            {-borderwidth        Int 1   0 "%d >= 0 && %d <= 10"}
        }
    }

    DynamicHelp::include ButtonNoteBook::Page balloon

    Widget::declare ButtonNoteBook {
	{-foreground		TkResource "" 0 button}
        {-background		TkResource "" 0 button}
        {-activebackground	TkResource "" 0 button}
        {-activeforeground	TkResource "" 0 button}
        {-disabledforeground	TkResource "" 0 button}
        {-font			TkResource "" 0 button}
        {-side			Enum       top 0 {top bottom}}
        {-homogeneous		Boolean 0   0}
        {-borderwidth		Int 1   0 "%d >= 0 && %d <= 10"}
 	{-internalborderwidth	Int 10  0 "%d >= 0"}
        {-width			Int 0   0 "%d >= 0"}
        {-height		Int 0   0 "%d >= 0"}
        {-tilewidgets           Boolean 0   0}

        {-fg                 Synonym -foreground}
        {-bg                 Synonym -background}
        {-bd                 Synonym -borderwidth}
        {-ibd                Synonym -internalborderwidth}

	{-arcradius          Int     2     0 "%d >= 0 && %d <= 8"}
	{-tabbevelsize       Int     0     0 "%d >= 0 && %d <= 8"}
        {-tabpady            Padding {0 6} 0 "%d >= 0"}
    }

    Widget::addmap ButtonNoteBook "" :cmd {
      -background {} -borderwidth {}
    }

    Widget::addmap ButtonNoteBook "" .middle {
      -background {} -width {} -height {}
    }

    bind ButtonNoteBook <Destroy>   [list ButtonNoteBook::_destroy %W]
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::create
# ---------------------------------------------------------------------------
proc ButtonNoteBook::create { path args } {
    #ButtonNoteBook::_puts "ButtonNoteBook::create $path $args"
    variable $path
    upvar 0  $path data

    Widget::init ButtonNoteBook $path $args

    set data(select)   ""
    set data(pages)    {}
    set data(realized) 0
    set data(tilewidgets) [Widget::cget $path -tilewidgets]

    eval [list frame $path -class ButtonNoteBook -width 100 -height 100 \
      -borderwidth 0 -highlightthickness 0 -relief flat] \
      [Widget::subcget $path :cmd]
    set bg [$path cget -background]
    # We will add 3 areas: A top frame holding the top buttons, a middle area
    # (where the page will be shown) and a bottom frame for the lower buttons.
    frame $path.top    -relief flat -bd 0 -bg $bg
    PagesManager $path.middle -background $bg
    frame $path.bottom -relief flat -bd 0 -bg $bg
    grid $path.top    -sticky snew -padx 2 -pady 1
    grid $path.middle -sticky snew -padx 2 -pady 1
    grid $path.bottom -sticky snew -padx 2 -pady 1
    grid columnconfigure $path 0 -weight 1
    grid rowconfigure    $path 0 -weight 0
    grid rowconfigure    $path 1 -weight 1
    grid rowconfigure    $path 2 -weight 0
    grid columnconfigure $path.top    0 -weight 1
    grid columnconfigure $path.bottom 0 -weight 1

    Widget::create ButtonNoteBook $path

    return $path
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::configure
# ---------------------------------------------------------------------------
proc ButtonNoteBook::configure { path args } {
    #ButtonNoteBook::_puts "ButtonNoteBook::configure $path $args"
    set res [Widget::configure $path $args]
    if {[Widget::hasChanged $path -background bg]} {
      variable $path
      upvar 0  $path data
      $path.middle configure -background $bg
      foreach page $data(pages) {
        [$path.middle getframe $page] configure -background $bg
      }
    }
    return $res
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::cget
# ---------------------------------------------------------------------------
proc ButtonNoteBook::cget { path option } {
    #ButtonNoteBook::_puts "ButtonNoteBook::cget $path $option"
    return [Widget::cget $path $option]
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::compute_size
# ---------------------------------------------------------------------------
proc ButtonNoteBook::compute_size { path } {
    #ButtonNoteBook::_puts "ButtonNoteBook::compute_size $path"
    $path.middle compute_size
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::insert
# ---------------------------------------------------------------------------
proc ButtonNoteBook::insert { path index page args } {
    #ButtonNoteBook::_puts "ButtonNoteBook::insert $path $index $page $args"
    variable $path
    upvar 0  $path data

    if { [lsearch -exact $data(pages) $page] != -1 } {
        return -code error "page \"$page\" already exists"
    }

    set f $path.top.b$page
    Widget::init ButtonNoteBook::Page $f $args

    set data(pages) [linsert $data(pages) $index $page]
    $path.middle add $page

    foreach option {state image text foreground background borderwidth font
                    activeforeground activebackground disabledforeground
                    relief} {
      set $option [_getoption $path $page -$option]
    }
    # If the page doesn't exist, create it; if it does reset its bg and ibd
    if { ![winfo exists $f] } {
      if {$data(tilewidgets)} {
        ttk::button $f -compound left \
          -image $image -text $text -state $state \
          -command "ButtonNoteBook::raise $path $page"
        ttk::button $path.bottom.b$page -compound left \
          -image $image -text $text -state $state \
          -command "ButtonNoteBook::raise $path $page"
      } else {
        button $f -compound left -relief $relief \
          -state $state -image $image -text $text -borderwidth $borderwidth \
          -foreground $foreground -background $background -font $font \
          -activeforeground $activeforeground \
          -activebackground $activebackground \
          -disabledforeground $disabledforeground \
          -anchor w -padx 10 -pady 5 -highlightthickness 0 \
          -command "ButtonNoteBook::raise $path $page"
        button $path.bottom.b$page -compound left -relief $relief \
          -state $state -image $image -text $text -borderwidth $borderwidth \
          -foreground $foreground -background $background \
          -activeforeground $activeforeground \
          -activebackground $activebackground \
          -disabledforeground $disabledforeground \
          -anchor w -padx 10 -pady 5 -highlightthickness 0 \
          -command "ButtonNoteBook::raise $path $page"
      }
      set data($page,realized) 0
    } else {
      $f configure -relief $relief \
        -state $state -image $image -text $text -borderwidth $borderwidth \
        -foreground $foreground -background $background -font $font \
        -activeforeground $activeforeground -activebackground $activebackground\
        -disabledforeground $disabledforeground \
        -anchor w -padx 10 -pady 5 -highlightthickness 0 \
        -command "ButtonNoteBook::raise $path $page"
      $path.bottom.b$page configure -relief $relief \
        -state $state -image $image -text $text -borderwidth $borderwidth \
        -foreground $foreground -background $background \
        -activeforeground $activeforeground -activebackground $activebackground\
        -disabledforeground $disabledforeground \
        -anchor w -padx 10 -pady 5 -highlightthickness 0 \
        -command "ButtonNoteBook::raise $path $page"
    }
    _set_help $path $page
    return $f
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::delete
# ---------------------------------------------------------------------------
proc ButtonNoteBook::delete { path page {destroyframe 1} } {
    #ButtonNoteBook::_puts "ButtonNoteBook::delete $path $page $destroyframe"
    variable $path
    upvar 0  $path data

    set pos [_test_page $path $page]
    set data(pages) [lreplace $data(pages) $pos $pos]
    $path.middle delete $page
    if { $data(select) == $page } {
        set data(select) ""
    }
    if { $destroyframe } {
        destroy $path.top.b$page
        destroy $path.bottom.b$page
    }
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::itemconfigure
# ---------------------------------------------------------------------------
proc ButtonNoteBook::itemconfigure { path page args } {
    #ButtonNoteBook::_puts "ButtonNoteBook::itemconfigure $path $page $args"
    _test_page $path $page
    return [_itemconfigure $path $page $args]
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::itemcget
# ---------------------------------------------------------------------------
proc ButtonNoteBook::itemcget { path page option } {
    #ButtonNoteBook::_puts "ButtonNoteBook::itemcget $path $page $option"
    _test_page $path $page
    return [Widget::cget $path.top.b$page $option]
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::bindtabs
# ---------------------------------------------------------------------------
proc ButtonNoteBook::bindtabs { path event script } {
    #ButtonNoteBook::_puts "ButtonNoteBook::bindtabs $path $event $script"
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::move
# ---------------------------------------------------------------------------
proc ButtonNoteBook::move { path page index } {
    #ButtonNoteBook::_puts "ButtonNoteBook::move $path $page $index"
    variable $path
    upvar 0  $path data

    set pos [_test_page $path $page]
    set data(pages) [linsert [lreplace $data(pages) $pos $pos] $index $page]
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::raise
# ---------------------------------------------------------------------------
proc ButtonNoteBook::raise { path {page ""} } {
    #ButtonNoteBook::_puts "ButtonNoteBook::raise $path $page"
    variable $path
    upvar 0  $path data

    if { $page != "" } {
        _test_page $path $page
        _select $path $page
    }
    return $data(select)
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::see
# ---------------------------------------------------------------------------
proc ButtonNoteBook::see { path page } {
   #ButtonNoteBook::_puts "ButtonNoteBook::see $path $page"
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::page
# ---------------------------------------------------------------------------
proc ButtonNoteBook::page { path first {last ""} } {
    #ButtonNoteBook::_puts "ButtonNoteBook::page $path $first $last"
    variable $path
    upvar 0  $path data

    if { $last == "" } {
        return [lindex $data(pages) $first]
    } else {
        return [lrange $data(pages) $first $last]
    }
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::pages
# ---------------------------------------------------------------------------
proc ButtonNoteBook::pages { path {first ""} {last ""}} {
    #ButtonNoteBook::_puts "ButtonNoteBook::pages $path $first $last"
    variable $path
    upvar 0  $path data

    if { ![string length $first] } {
	return $data(pages)
    }

    if { ![string length $last] } {
        return [lindex $data(pages) $first]
    } else {
        return [lrange $data(pages) $first $last]
    }
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::index
# ---------------------------------------------------------------------------
proc ButtonNoteBook::index { path page } {
    #ButtonNoteBook::_puts "ButtonNoteBook::index $path $page"
    variable $path
    upvar 0  $path data

    return [lsearch -exact $data(pages) $page]
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::_destroy
# ---------------------------------------------------------------------------
proc ButtonNoteBook::_destroy { path } {
    #ButtonNoteBook::_puts "ButtonNoteBook::_destroy $path"
    variable $path
    upvar 0  $path data

    Widget::destroy $path
    unset data
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::getframe
# ---------------------------------------------------------------------------
proc ButtonNoteBook::getframe { path page } {
    #ButtonNoteBook::_puts "ButtonNoteBook::getframe $path $page"
    # return $path.f$page
    return [$path.middle getframe $page]
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::create_page
# ---------------------------------------------------------------------------
proc ButtonNoteBook::create_page { path page } {
    #ButtonNoteBook::_puts "ButtonNoteBook::create_page $path $page"
    variable $path
    upvar 0  $path data

    if { ![string equal $page ""] } {
        _test_page $path $page
	if { !$data($page,realized) } {
	    set data($page,realized) 1
	    set cmd [Widget::cget $path.top.b$page -createcmd]
	    if { ![string equal $cmd ""] } {
		uplevel \#0 $cmd
	    }
	}
    }
}


# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::_test_page
# ---------------------------------------------------------------------------
proc ButtonNoteBook::_test_page { path page } {
    #ButtonNoteBook::_puts "ButtonNoteBook::_test_page $path $page"
    variable $path
    upvar 0  $path data

    if { [set pos [lsearch -exact $data(pages) $page]] == -1 } {
        return -code error "page \"$page\" does not exists"
    }
    return $pos
}

proc ButtonNoteBook::_getoption { path page option } {
    #ButtonNoteBook::_puts "ButtonNoteBook::_getoption $path $page $option"
    if {[catch {Widget::cget $path.top.b$page $option} value]} {set value {}}
    if {![string length $value]} {
      if {[catch {Widget::cget $path $option} value]} {set value {}}
    }
    return $value
}

# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::_itemconfigure
# ---------------------------------------------------------------------------
proc ButtonNoteBook::_itemconfigure { path page lres } {
    #ButtonNoteBook::_puts "ButtonNoteBook::_itemconfigure $path $page $lres"
    variable $path
    upvar 0  $path data

    set res [Widget::configure $path.top.b$page $lres]
    # Configure buttons...
    foreach option {state image text foreground background borderwidth font
                    activeforeground activebackground disabledforeground
                    relief} {
      set $option [_getoption $path $page -$option]
    }
    if {$data(tilewidgets)} {
      $path.top.b$page configure -image $image -text $text -state $state
      $path.bottom.b$page configure -image $image -text $text -state $state
    } else {
      $path.top.b$page configure -relief $relief \
        -state $state -image $image -text $text -borderwidth $borderwidth \
        -foreground $foreground -background $background -font $font \
        -activeforeground $activeforeground -activebackground $activebackground\
        -disabledforeground $disabledforeground
      $path.bottom.b$page configure -relief $relief  \
        -state $state -image $image -text $text -borderwidth $borderwidth \
        -foreground $foreground -background $background \
        -activeforeground $activeforeground -activebackground $activebackground\
        -disabledforeground $disabledforeground
    }
    if { [Widget::hasChanged $path.top.b$page -state state] &&
         $state == "disabled" && $data(select) == $page } {
        set data(select) ""
    }
    return $res
}

# ---------------------------------------------------------------------------
#  Command ButtonNoteBook::_select
# ---------------------------------------------------------------------------
proc ButtonNoteBook::_select { path page } {
    #ButtonNoteBook::_puts "ButtonNoteBook::_select $path $page"
    variable $path
    upvar 0  $path data

    if {![string equal [Widget::cget $path.top.b$page -state] "normal"]} {return}

    set oldsel $data(select)

    if {[string equal $page $oldsel]} { return }

    if { ![string equal $oldsel ""] } {
	set cmd [Widget::cget $path.top.b$oldsel -leavecmd]
	if { ![string equal $cmd ""] } {
	    set code [catch {uplevel \#0 $cmd} res]
	    if { $code == 1 || $res == 0 } {
		return -code $code $res
	    }
	}
	set data(select) ""
    }

    set data(select) $page
    if { ![string equal $page ""] } {
	if { !$data($page,realized) } {
	    set data($page,realized) 1
	    set cmd [Widget::cget $path.top.b$page -createcmd]
	    if { ![string equal $cmd ""] } {
		uplevel \#0 $cmd
	    }
	}
	set cmd [Widget::cget $path.top.b$page -raisecmd]
	if { ![string equal $cmd ""] } {
	    uplevel \#0 $cmd
	}
	_draw_page $path $page
    }

}


# ----------------------------------------------------------------------------
#  Command ButtonNoteBook::_draw_page
# ----------------------------------------------------------------------------
proc ButtonNoteBook::_draw_page { path page } {
    #ButtonNoteBook::_puts "ButtonNoteBook::_draw_page $path $page"
    variable $path
    upvar 0  $path data

    $path.middle raise $page
    set pos [lsearch -exact $data(pages) $page]
    set last [llength $data(pages)]
    if {$pos == $last-1} {$path.bottom configure -height 1}
    for {set i 0} {$i < $last} {incr i} {
      set page [lindex $data(pages) $i]
      if {$i <= $pos} {
        grid $path.top.b$page -row $i -sticky snew
        grid forget $path.bottom.b$page
      } else {
        grid forget $path.top.b$page
        grid $path.bottom.b$page -row $i -sticky snew
      }
    }
}


# -----------------------------------------------------------------------------
#  Command ButtonNoteBook::_resize
# -----------------------------------------------------------------------------
proc ButtonNoteBook::_resize { path } {
    #ButtonNoteBook::_puts "ButtonNoteBook::_resize $path"
    variable $path
    upvar 0  $path data

    if {!$data(realized)} {
	if { [set width  [Widget::cget $path -width]]  == 0 ||
	     [set height [Widget::cget $path -height]] == 0 } {
	    compute_size $path
	}
	set data(realized) 1
    }
}


# Tree::_set_help --
#
#	Register dynamic help for a node in the tree.
#
# Arguments:
#	path		Tree to query
#	node		Node in the tree
#       force		Optional argument to force a reset of the help
#
# Results:
#	none
# Tree::_set_help --
#
#	Register dynamic help for a node in the tree.
#
# Arguments:
#	path		Tree to query
#	node		Node in the tree
#       force		Optional argument to force a reset of the help
#
# Results:
#	none
proc ButtonNoteBook::_set_help { path page } {
    Widget::getVariable $path help

    set item $path.top.b$page
    set opts [list -helptype -helptext -helpvar]
    foreach {cty ctx cv} [eval [list Widget::hasChangedX $item] $opts] break
    set text [Widget::getoption $item -helptext]

    ## If we've never set help for this item before, and text is not blank,
    ## we need to setup help.  We also need to reset help if any of the
    ## options have changed.
    if { (![info exists help($page)] && $text != "") || $cty || $ctx || $cv } {
	set help($page) 1
	set type [Widget::getoption $item -helptype]
        switch $type {
            balloon {
		DynamicHelp::register $path.c balloon p:$page $text
            }
            variable {
		set var [Widget::getoption $item -helpvar]
		DynamicHelp::register $path.c variable p:$page $var $text
            }
        }
    }
}

#proc ButtonNoteBook::_puts {msg} {
#  #puts $msg
#}
