##############################################################################
##  Dialog.tcl:                                                             ##
## --------------------------------------------------                       ##
##  This file implements a class that holds a Dialog Ellogon window.        ##
##  A Dialog window is a plain window (i.e. no menubar, toolbar, etc)       ##
##  centered in screen.                                                     ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities

oo::class create ELEP::WindowUtilities::DialogWindow {
  mixin ELEP::Base::Utilities

  constructor {} {
    my variable win client_area dialog_result
    set dialog_result {}
    set win [toplevel [namespace tail [self]]_DialogWindow \
                -class ELEP_DialogWindow]
    wm withdraw $win
    bind itk-delete-$win <Destroy> "[self] destroy"
    set tags [bindtags $win]
    bindtags $win [linsert $tags 0 itk-delete-$win]
    my createComponents
    my createComponentLayout
    my fillComponents
    my addBindings
  };# constructor

  method addBindings {} {
    my variable win
    bind $win <Return> {
      if {[winfo class %W] in "Button TButton"} {
        %W invoke
      }
    }
  };# addBindings

  ##
  ## Component Creation
  ##
  method createComponents {} {
    my createMenuBar
    my createToolBar
    my createClientArea
    my createButtonBar
    my createStatusBar
  };# createComponents

  method createMenuBar {} {
  };# createMenuBar

  method createToolBar {} {
  };# createToolBar

  method createClientArea {} {
    my variable win client_area
    set client_area [ttk::frame $win.client_area -class ClientArea]
  };# createClientArea

  method createButtonBar {} {
    my variable win buttonbar
    set buttonbar [ttk::frame $win.button_bar -class ButtonBar]
    grid anchor $buttonbar center
  };# createButtonBar

  method createStatusBar {} {
  };# createStatusBar

  ##
  ## Component Layout
  ##
  method createComponentLayout {} {
    my variable client_area buttonbar
    pack $buttonbar -side bottom -fill x
    pack $client_area -fill both -expand 1
  };# createComponentLayout

  ##
  ## Component Filling
  ##
  method fillComponents {} {
  };# fillComponents

  ##
  ## Buttons facilities...
  ##
  method addButton {label cmd {image {}}} {
    my variable buttonbar buttonbar_button_count
    incr buttonbar_button_count
    set b $buttonbar.b$buttonbar_button_count

    ttk::button $b -text $label -command $cmd -default normal
    if {$image ne ""} {
      $b configure -image $image -compound left
    }
    grid $b -row 0 -column $buttonbar_button_count -padx 3m -pady 3m -sticky ew
    grid columnconfigure $buttonbar $buttonbar_button_count -uniform buttons
    return $b
  };# addButton

  method addStdButton {type {cmd {default}} {image {}}} {

    if {$cmd eq "default"} {
      set cmd [list [self] setResultAndDestroy $type]
    }
    switch $type {
      ok     {set label Ok}
      cancel {set label Cancel}
      yes    {set label Yes}
      no     {set label No}
      abort  {set label Abort}
      retry  {set label Retry}
      ignore {set label Ignore}
      apply  {set label Apply}
      def    {set label Default}
      update {set label Update}
      close  {set label Close}
      save   {set label Save}
      default {
        error "unknown type \"$type\""
      }
    }
    my addButton [my msg $label] $cmd $image
  };# addStdButton

  ##
  ## Window Facilities
  ##
  method getTopLevel {}     {my variable win;         return $win}
  method getframe {}        {my variable client_area; return $client_area}
  method title    {title}   {my variable win;         wm title $win $title}
  method show     {}        {my variable win; update; wm deiconify $win}
  method hide     {}        {my variable win; update; wm withdraw $win}
  method wait     {}        {
    my variable win dialog_result; tkwait window $win; return $dialog_result
  };# wait
  method showWait {}        {my show; update; return [my wait]}
  method setResultAndDestroy {result} {
    my variable dialog_result
    set dialog_result $result
    my destroyWindow
  };# setResultAndDestroy

  method onDestroyDialog {} {my variable win; destroy $win}
  method destroyWindow   {} {my variable win; destroy $win}

  destructor {
    my variable win
    if {[winfo exists $win]} {
      set tags [bindtags $win]
      set i [lsearch $tags itk-delete-$win]
      if {$i >= 0} {
        bindtags $win [lreplace $tags $i $i]
      }
      destroy $win
    }
    catch {bind itk-delete-$win <Destroy> {}}
  };# destructor

};# class ELEP::WindowUtilities::DialogWindow

package provide ELEP::WindowUtilities::DialogWindow 1.0
