##############################################################################
##  ObjectListenerServer.tcl:                                               ##
## --------------------------------------------------                       ##
##  This class implements a class for supporting object listeners. Such an  ##
##  object keeps a list of listener objects and their methods, that should  ##
##  be called, when an event must be delivered to the objects.              ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

##
## Load some needed packages.
##
package require TclOO    ;# TclOO adds object-oriented facilities to Tcl.

oo::class create ELEP::WindowUtilities::ObjectListenerServer {

  variable object_listener_listener_objects \
           object_listener_listener_methods \
           object_listener_default_method

  constructor {{default_method notify}} {
    my initListenerObject $default_method
  };# constructor

  ##########################################################################
  ## Method: initListenerObject
  ##########################################################################
  method initListenerObject {{default_method notify}} {
    set object_listener_listener_objects {}
    set object_listener_listener_methods {}
    set object_listener_default_method $default_method
  };# initListenerObject

  ##########################################################################
  ## Method: registerListener
  ##########################################################################
  method registerListener {obj {method {}}} {
    if {![string length $method]} {set method $object_listener_default_method}
    if {[lsearch -exact $object_listener_listener_objects $obj] == -1} {
      lappend object_listener_listener_objects $obj
      lappend object_listener_listener_methods $method
      ## Notify object about this registration:
      catch {$obj registeredAsListener [self] $method}
    }
  };# registerListener

  ##########################################################################
  ## Method: unregisterListener
  ##########################################################################
  method unregisterListener {obj} {
    set index [lsearch -exact $object_listener_listener_objects $obj]
    if {$index != -1} {
      set method [lindex $object_listener_listener_methods $index]
      catch {$obj unregisteredAsListener [self] $method}
      set object_listener_listener_objects \
        [lreplace $object_listener_listener_objects $index $index]
      set object_listener_listener_methods \
        [lreplace $object_listener_listener_methods $index $index]
    }
  };# unregisterListener

  ##########################################################################
  ## Method: notifyListeners
  ##########################################################################
  method notifyListeners {} {
    foreach obj    $object_listener_listener_objects \
            method $object_listener_listener_methods {
      $obj $method [self]
    }
  };# notifyListeners

  ##########################################################################
  ## Event: onUpdateRequestFromRegisteredListener
  ##########################################################################
  method onUpdateRequestFromRegisteredListener {obj} {
  };# onUpdateRequestFromRegisteredListener

  destructor {
    if {![info exists object_listener_listener_objects]} {return}
    foreach obj    $object_listener_listener_objects \
            method $object_listener_listener_methods {
      catch {$obj unregisteredAsListener [self] $method}
    }
  };# destructor

};# class ELEP::WindowUtilities::ObjectListenerServer

package provide ELEP::WindowUtilities::ObjectListenerServer 1.0
