##############################################################################
##  ProgressWidget.tcl:                                                     ##
## --------------------------------------------------                       ##
##  This class implements a widget that can show a set of messages/progress ##
##  bar pairs.                                                              ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities::WidgetBase
package require tablelist_tile

oo::class create ELEP::WindowUtilities::ProgressWidget {
  superclass ELEP::WindowUtilities::WidgetBase

  constructor {parent_win {max 100}} {
    my variable message_wrap_length default_maximum_value relative_width \
                show_buttons
    next ttk::frame [::ELEP::WindowUtilities::UniqueWindowName \
         $parent_win.progressWidget] -relief raised -borderwidth 1
    set message_wrap_length   200
    set relative_width        0.9
    set default_maximum_value $max
    set show_buttons          0
    my createWidgets
  };# constructor

  ##########################################################################
  ## Method: createWidgets
  ##########################################################################
  method createWidgets {} {
    my variable widget show_buttons
    ## Add a client area, where pairs of message/progress bar widgets will be
    ## placed.
    ttk::frame $widget.client_area
    grid columnconfigure $widget.client_area 0 -weight 1
    ## Add some buttons, for pausing/stopping process...
    ttk::frame $widget.buttons
      ttk::button $widget.buttons.pause  -textvariable [my msgVar Pause]
      ttk::button $widget.buttons.cancel -textvariable [my msgVar Cancel]
      pack $widget.buttons.cancel $widget.buttons.pause -side right \
        -padx 2 -pady 2
    grid $widget.client_area -sticky snew -padx 2 -pady 2
    if {$show_buttons} {
      grid $widget.buttons -sticky snew -padx 2 -pady 2
    }
    grid columnconfigure $widget 0 -weight 1
    grid rowconfigure $widget 0 -weight 1
    my addGroupWidgets 0
  };# createWidgets

  ##########################################################################
  ## Method: addGroupWidgets
  ##########################################################################
  method addGroupWidgets {group {max {}}} {
    my variable current_value current_message widget \
                default_maximum_value message_wrap_length
    if {![string length $max]} {set max $default_maximum_value}
    set group [string tolower $group]
    set current_value($group)   0
    set current_message($group) [my msg PleaseWait]
    ttk::label $widget.client_area.lbl_$group \
      -textvariable [my varname current_message]($group)\
      -wraplength $message_wrap_length -justify left -anchor nw
    ttk::progressbar $widget.client_area.pgr_$group -orient horizontal \
      -maximum $max -mode determinate \
      -variable [my varname current_value]($group)
    my addCustomGroupWidgets $group
    my packGroupWidgets $group
    return $group
  };# addGroupWidgets

  ##########################################################################
  ## Method: addCustomGroupWidgets
  ##########################################################################
  method addCustomGroupWidgets {group} {
  };# addCustomGroupWidgets

  ##########################################################################
  ## Method: packGroupWidgets
  ##########################################################################
  method packGroupWidgets {group} {
    my variable widget
    grid $widget.client_area.lbl_$group -sticky snew -padx 2 -pady 2
    grid $widget.client_area.pgr_$group -sticky snew -padx 2 -pady 2
  };# packGroupWidgets

  ##########################################################################
  ## Method: message
  ##########################################################################
  method message {message {group 0}} {
    my variable current_message
    set current_message($group) $message
    update
  };# message

  ##########################################################################
  ## Method: progress
  ##########################################################################
  method progress {value {group 0}} {
    my variable current_value
    set current_value($group) $value
    update
  };# progress

  ##########################################################################
  ## Method: advance
  ##########################################################################
  method advance {{group 0}} {
    my variable current_value
    incr current_value($group)
    update
  };# advance

  ##########################################################################
  ## Method: advanceWithStep
  ##########################################################################
  method advanceWithStep {step {group 0}} {
    my variable current_value
    set current_value($group) [expr {int ($current_value($group) + $step)}]
    update
  };# advanceWithStep

  ##########################################################################
  ## Method: setMessage&Advance
  ##########################################################################
  method setMessage&Advance {msg {group 0}} {
    my variable current_value current_message widget
    set  current_message($group) $msg
    incr current_value($group)
    update
  };# setMessage&Advance

  ##########################################################################
  ## Method: placeInToplevelCenter
  ##########################################################################
  method placeInToplevelCenter {} {
    my variable widget relative_width
    # set toplevel [winfo toplevel $widget]
    # puts -1>$widget
    # puts -2>$toplevel
    place $widget -relwidth $relative_width -anchor c -relx 0.5 -rely 0.5
    my updateMessageWrapLength
  };# placeInToplevelCenter

  method updateMessageWrapLength {} {
    my variable widget current_value message_wrap_length
    update
    set message_wrap_length [winfo width $widget.client_area.lbl_0]
    incr message_wrap_length -20
    foreach group [array names current_value] {
      $widget.client_area.lbl_$group configure -wraplength $message_wrap_length
    }
  };# updateMessageWrapLength

  ##########################################################################
  ## Method: hide
  ##########################################################################
  method hide {} {
    my variable widget
    place forget $widget
    update
  };# hide

};# class ELEP::WindowUtilities::ProgressWidget

package provide ELEP::WindowUtilities::ProgressWidget 1.0
