##############################################################################
##  Toplevel.tcl:                                                           ##
## --------------------------------------------------                       ##
##  This file implements a class that holds a Toplevel Ellogon window.      ##
##  A Toplevel window is a window that has a menubar, a toolbar, a main     ##
##  and a status bar.                                                       ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

package require ELEP::WindowUtilities
catch {package require ELEP::WindowUtilities::ConsoleWindow}

oo::class create ELEP::WindowUtilities::ToplevelWindow {
  mixin ELEP::Base::Utilities
  variable win \
           menubar \
           toolbar \
           client_area \
           statusbar \
           statusbar_text \
           statusbar_progress \
           statusbar_progress_pady \
           menu_file_exit

  constructor {{allow_exit 0}} {
    my variable win menu_file_exit
    set menu_file_exit $allow_exit
    set win [toplevel [namespace tail [self]]_ToplevelWindow \
               -class [namespace tail [self class]]]
    wm withdraw $win
    bind itk-delete-$win <Destroy> "[self] destroy"
    set tags [bindtags $win]
    bindtags $win [linsert $tags 0 itk-delete-$win]
    # bind $win <Destroy> "if {\[string equal $win %W\]}\
    #           {::ELEP::WindowUtilities::DeleteTclOOObject [self]}"
    my createComponents
    my createComponentLayout
    my fillComponents
  };# constructor

  ##
  ## Component Creation
  ##
  method createComponents {} {
    my createMenuBar
    my createToolBar
    my createClientArea
    my createStatusBar
  };# createComponents

  method createMenuBar {} {
    my variable menubar
    set menubar [menu $win.menubar]
  };# createMenuBar

  method createToolBar {} {
    my variable win toolbar
    set toolbar [ttk::frame $win.toolbarFrame -class ToolbarFrame]
  };# createToolBar

  method createClientArea {} {
    my variable client_area
    set client_area [ttk::frame $win.client_area -class ClientArea]
  };# createClientArea

  method createStatusBar {} {
    my variable statusbar statusbar_text statusbar_progress \
                statusbar_progress_pady
    set statusbar [ttk::frame $win.status_bar -class StatusBar]
    ttk::label $statusbar.label -textvariable [my varname statusbar_text] \
      -anchor w -justify left
    ttk::progressbar $statusbar.progress -orient horizontal -maximum 100 \
      -variable [my varname statusbar_progress]
    ttk::sizegrip $statusbar.szg

    # Calculate the amount of vertical padding that should be set.
    set hmax 0; update idletasks
    foreach w [winfo children $statusbar] {
      set h [winfo reqheight $w]; set hmax [expr {$h>$hmax ? $h : $hmax}]
    }
    incr hmax 2;# Add 2 pixels of padding.

    # -pady for label...
    set diff [expr {$hmax-[winfo reqheight $statusbar.label]}]
    set upper_pady [expr {$diff/2}]
    grid $statusbar.label -column 0 -row 0 -sticky snew -padx 2 \
      -pady [list $upper_pady [expr {$diff-$upper_pady}]]
    grid $statusbar.szg -column 2 -row 0 -sticky snew -padx 1
    grid columnconfigure $statusbar 0 -weight 1
    # -pady for progress bar...
    set diff [expr {$hmax-[winfo reqheight $statusbar.progress]}]
    set upper_pady [expr {$diff/2}]
    set statusbar_progress_pady [list $upper_pady [expr {$diff-$upper_pady}]]
  };# createStatusBar

  ##
  ## Component Layout
  ##
  method createComponentLayout {} {
    my variable toolbar statusbar client_area
    my createComponentLayoutSetMenubar
    pack $toolbar -side top -fill x
    pack $statusbar -side bottom -fill x
    pack $client_area -fill both -expand 1
  };# createComponentLayout

  method createComponentLayoutSetMenubar {} {
    my variable win menubar
    $win configure -menu $menubar
  };# createComponentLayoutSetMenubar

  ##
  ## Component Filling
  ##
  method fillComponents {} {
    my fillMenuBar
    my fillToolBar
    my fillStatusBar
    my fillClientArea
  };# fillComponents

  method fillMenuBar {} {
    my variable menubar
    $menubar delete 0 end
    my fillMenu_File
    #my fillMenu_Edit
    my fillMenu_Other
    my fillMenu_Window
    #my fillMenu_Help
  };# fillMenuBar

  ##########################################################################
  ## Menu: File
  ##########################################################################
  method fillMenu_File {} {
    my variable menubar menu_file_exit
    global GUI_Text GUI_Underline GUI_Options
    set file [menu $menubar.file]
    $menubar add cascade -label $GUI_Text(File) -menu $file \
              -underline $GUI_Underline(File)
    $file add command -label $GUI_Text(OpenConsole) -image ITConsole \
              -underline $GUI_Underline(OpenConsole) -command {tkcon show} \
              -compound $GUI_Options(MenuImageLocation)
    $file add separator
    $file add command -label $GUI_Text(Options)... \
              -underline $GUI_Underline(Options) -command GUI_EditPreferences\
              -image ITBlank -compound $GUI_Options(MenuImageLocation)
    $file add separator
    if {$menu_file_exit} {
      wm protocol $win WM_DELETE_WINDOW "GUI_CloseWindow .mw"
      $file add command -label $GUI_Text(Exit) \
                -underline $GUI_Underline(Exit) \
                -command "GUI_CloseWindow .mw" -image ITExit \
                -compound $GUI_Options(MenuImageLocation)
    } else {
      wm protocol $win WM_DELETE_WINDOW "GUI_CloseWindow $win"
      $file add command -label $GUI_Text(Close) \
                -underline $GUI_Underline(Close) \
                -command "GUI_CloseWindow $win" -image ITExit \
                -compound $GUI_Options(MenuImageLocation)
    }
  };# fillMenu_File

  ##########################################################################
  ## Menu: Edit
  ##########################################################################
  method fillMenu_Edit {} {
    my variable menubar
    global GUI_Text GUI_Underline GUI_Options
    set edit [menu $menubar.edit]
    $menubar add cascade -label $GUI_Text(Edit) -menu $edit \
              -underline $GUI_Underline(Edit)
  };# fillMenu_Edit

  ##########################################################################
  ## Menu: Other
  ##########################################################################
  method fillMenu_Other {} {
  };# fillMenu_Other

  ##########################################################################
  ## Menu: Window
  ##########################################################################
  method fillMenu_Window {} {
    my variable menubar
    global GUI_Text GUI_Underline GUI_Options
    set window [menu $menubar.window]
    set theme  [menu $menubar.window.theme]
    $menubar add cascade -label $GUI_Text(Window) -menu $window \
              -underline $GUI_Underline(Window)
    $window add cascade -label $GUI_Text(WidgetTheme) -image ITBlank \
              -underline $GUI_Underline(WidgetTheme) -menu $theme \
              -compound $GUI_Options(MenuImageLocation)
    ## Fill the widget theme cascade menu...
    foreach name [lsort [ttk::themes]] {
      $theme add radiobutton -label [string totitle $name] \
        -variable ::GUI_Options(TileWidgetTheme) -value $name \
        -command [list ttk::setTheme $name]
    }
  };# fillMenu_Window

  ##########################################################################
  ## Menu: Help
  ##########################################################################
  method fillMenu_Help {} {
    my variable menubar
    global GUI_Text GUI_Underline GUI_Options
    set help [menu $menubar.help]
    $menubar add cascade -label $GUI_Text(Help) -menu $help \
              -underline $GUI_Underline(Help)
    GUI_CreateHelpMenuEntries $help $win
  };# fillMenu_Help

  method fillToolBar {} {
  };# fillToolBar

  method fillStatusBar {} {
  };# fillStatusBar

  method fillClientArea {} {
  };# fillClientArea

  ##
  ## Progress Bar Facilities
  ##
  method showProgress  {} {
    my variable statusbar
    grid $statusbar.progress -column 1 -row 0 -sticky snew -padx 1 \
         -pady $statusbar_progress_pady
  };# showProgress
  method hideProgress  {}    {
    my variable statusbar
    grid forget $statusbar.progress
  };# hideProgress
  method setProgress   {val} {
    my variable statusbar_progress
    set statusbar_progress $val
  };# setProgress

  ##
  ## Status Bar Facilities
  ##
  method getStatusText {} {
    my variable statusbar_text
    return $statusbar_text
  };# getStatusText

  method setStatusText {{val {}}} {
    my variable statusbar_text
    set statusbar_text $val
  };# setStatusText

  method statusMessageFromDocument {{col {}} {doc {}}} {
    set c [my msg none]; set d $c
    if {[string length $col]} {
      if {[catch {tip_GetName $col} c]} {set c $col}; set d [my msg All]
    }
    if {[string length $doc]} {
      if {[catch {tip_GetId $doc} d]} {set d $doc}
      if {$d eq ""} {set d $doc}
    }
    return "[my msg Collection]:($c), [my msg Document]:($d)"
  };# statusMessageFromDocument

  method updateStatusFromDocument {{col {}} {doc {}}} {
    my setStatusText [my statusMessageFromDocument $col $doc]
  };# updateStatusFromDocument

  ##
  ## Window Facilities
  ##
  method getframe {}        {my variable client_area; return $client_area}
  method getTopLevel {}     {my variable win;         return $win}
  method getToolbarFrame {} {my variable toolbar;     return $toolbar}
  method title    {title}   {my variable win;         wm title $win $title}
  method show     {}        {my variable win; update; wm deiconify $win}
  method wait     {}        {my variable win;         tkwait window $win}
  method showWait {}        {my show; update; my wait}

  method onDestroyDialog {} {my variable win; destroy $win}

  destructor {
    my variable win
    if {[winfo exists $win]} {
      set tags [bindtags $win]
      set i [lsearch $tags itk-delete-$win]
      if {$i >= 0} {
        bindtags $win [lreplace $tags $i $i]
      }
      destroy $win
    }
    catch {bind itk-delete-$win <Destroy> {}}
  }

};# class ELEP::WindowUtilities::ToplevelWindow

package provide ELEP::WindowUtilities::ToplevelWindow 1.0
