##############################################################################
##  WindowUtilities.tcl:                                                    ##
## --------------------------------------------------                       ##
##  This files implements some useful functions for managing windows.       ##
##                                                                          ##
##  This file is part of the Ellogon Language Engineering Platform.         ##
##  Copyright 1998-2010 by:                                                 ##
##  Georgios Petasis,                                                       ##
##  Athens, Greece.                                                         ##
##  E-mail: petasis@iit.demokritos.gr                                       ##
##          petasisg@yahoo.gr                                               ##
##          petasisg@gmail.com                                              ##
##                                                                          ##
##  For licensing information, please refer to the license conditions of    ##
##  "Ellogon" Language Engineering Platform.                                ##
##############################################################################

##
## Load some needed packages.
##
package require Ttk     ;# Tile widgets (Tk 8.6+)
# package require BWidget ;# BWidget offers additional widgets.
package require TclOO   ;# TclOO adds object-oriented facilities to Tcl.
package require ELEP::Base::Utilities
catch {ttk::setTheme $::GUI_Options(TileWidgetTheme)}

## ::ELEP::WindowUtilities::FindToplevelObject
#  Accepts a window name and returns a toplevel object (of type
#  ELEP::WindowUtilities::ToplevelWindow), if such an object exists.
proc ::ELEP::WindowUtilities::FindToplevelObject {win} {
  set obj {}
  switch -glob $win {
    *_ToplevelWindow {
      set obj [lindex [bind itk-delete-$win <Destroy>] 0]
    }
    default {}
  }
  return $obj
};# ::ELEP::WindowUtilities::FindToplevelObject

## ::ELEP::WindowUtilities::UniqueWindowName
#  Accepts a window name and returns a variation of the name, so as name does
#  not represent an existing window.
proc ::ELEP::WindowUtilities::UniqueWindowName {win} {
  if {[UniqueWindowName_Exists $win]} {
    set win_id 0
    while {[incr win_id]} {
      set w ${win}_$win_id
      if {![UniqueWindowName_Exists $w]} {return $w}
    }
  }
  return $win
};# ::ELEP::WindowUtilities::UniqueWindowName

proc ::ELEP::WindowUtilities::UniqueWindowName_Exists {w} {
  if { [winfo exists $w] ||
       [winfo exists ${w}_ToplevelWindow] ||
       [winfo exists ${w}_DialogWindow] } {return 1}
  return [llength [info commands $w]]
};# ::ELEP::WindowUtilities::UniqueWindowName_Exists

## ::ELEP::WindowUtilities::DeleteTclOOObject
#  Checks the existance of an TclOO object, and deletes it if it exists.
proc ::ELEP::WindowUtilities::DeleteTclOOObject {object} {
  if {[llength [info commands $object]]} {
    ## The object exists. Delete it...
    $object destroy
  }
};# ::ELEP::WindowUtilities::DeleteTclOOObject

## ::ELEP::WindowUtilities::AutoFocusOnMouseEnter
#  Arranges the widget to receive focus, when the mouse enters it.
proc ::ELEP::WindowUtilities::AutoFocusOnMouseEnter {widget} {
  bind $widget <Enter> [list \
    +::ELEP::WindowUtilities::AutoFocusOnMouseEnterCallback $widget]
};# ::ELEP::WindowUtilities::AutoFocusOnMouseEnter

proc ::ELEP::WindowUtilities::AutoFocusOnMouseEnterCallback {widget} {
  if {![winfo exists $widget]} {return}
  set focus [focus]
  ## Has application lost the focus?
  if {![string length $focus]} {return}
  ## Is the focus still in the same toplevel?
  if {![string equal [winfo toplevel $widget] [winfo toplevel $focus]]} {return}
  focus $widget
};# ::ELEP::WindowUtilities::AutoFocusOnMouseEnterCallback

## ::ELEP::WindowUtilities::msg
#  Return the text that corresponds to the provided message code (key).
proc ::ELEP::WindowUtilities::msg {code} {
  return [::ELEP::Base::msg $code]
};# ::ELEP::WindowUtilities::msg

## ::ELEP::WindowUtilities::msgVar
#  Return the variable name that corresponds to the provided message code (key).
proc ::ELEP::WindowUtilities::msgVar {code} {
  return [::ELEP::Base::msgVar $code]
};# ::ELEP::WindowUtilities::msgVar

proc ::ELEP::WindowUtilities::opt {code} {
  return [::ELEP::Base::opt $code]
};# ::ELEP::WindowUtilities::opt

proc ::ELEP::WindowUtilities::optVar {code} {
  return [::ELEP::Base::optVar $code]
};# ::ELEP::WindowUtilities::optVar

proc ::ELEP::WindowUtilities::Log {args} {
  return [::ELEP::Base::Log $code]
};# ::ELEP::WindowUtilities::Log

package provide ELEP::WindowUtilities 1.0
