/*****************************************************************************
 *  EmbedAPI.c:                                                              *
 * ----------------------                                                    *
 *  This file implements an API for embeding Ellogon in a C application.     *
 *                                                                           *
 *  Copyright 1998-2009 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licencing information, please refer to the licence conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/

#include "EmbedAPI.h"
#include <stdio.h>

#ifdef WIN32
#ifndef snprintf
#define snprintf  _snprintf
#endif
#endif

/* 
 * Access to the list of threads and to the thread send results
 * (defined below) is guarded by this mutex. 
 */
TCL_DECLARE_MUTEX(ellogonEmbedMutex)

/*
 * Ellogon_Init:
 *
 * This procedure will initialise the embding API. It should be called
 * only once in the lifetime of an application.
 * argc & argv come from the main() of the application.
 */
Tcl_Interp *Ellogon_Init(int argc,   /* Number of arguments. */
                         char **argv /* Array of argument strings. */) {
  Tcl_Interp *interp;
  Tcl_MutexLock(&ellogonEmbedMutex);
  interp = Tcl_CreateInterp();
  Tcl_FindExecutable(argv[0]);
  Tcl_Preserve(interp);
  if (Tcl_Init(interp) == TCL_ERROR) {
    fprintf(stderr, "Error in Ellogon_Init: %s\n",
            Tcl_GetStringResult(interp));
  }
  /* Load the thread package... */
  if (Tcl_Eval(interp, "package require Thread") == TCL_ERROR) {
    fprintf(stderr, "Error in Ellogon_Init: %s\n",
            Tcl_GetStringResult(interp));
  }
  if (Tcl_Eval(interp, "set instances -1") == TCL_ERROR) {
    fprintf(stderr, "Error in Ellogon_Init: %s\n",
            Tcl_GetStringResult(interp));
  }
  Tcl_MutexUnlock(&ellogonEmbedMutex);
  return interp;
}; /* Ellogon_Init */

/*
 * Ellogon_CreateInstance:
 *
 * This procedure will create an instance of Ellogon, in its own thread.
 * The return value will be the instance id, or -1 in case of error...
 */
int Ellogon_CreateInstance(Tcl_Interp *interp,
                           const char *path_to_ellogon,
                           int argc,   /* Number of arguments. */
                           char **argv /* Array of argument strings. */) {
  char script[512];
  int i, instance = -1;
  Tcl_Obj *obj;
  Tcl_MutexLock(&ellogonEmbedMutex);
  /* Create the new instance thread... */
  if (Tcl_Eval(interp, "set thread [thread::create]; puts [info nameofexe]; "
                       "set E([incr instances]) $thread; "
                       "thread::send $thread {set argv {}}")
                                                  != TCL_OK) goto error;
  /* Declare the arguments... */
  for (i = 0; i < argc; ++i) {
    snprintf(script, 512, "thread::send $thread {lappend argv {%s}}", argv[i]);
    if (Tcl_Eval(interp, script) != TCL_OK) goto error;
  }
  if (Tcl_Eval(interp, "thread::send $thread {set argc [llength $argv]}")
                                                          != TCL_OK) goto error;
  /* Finally source the Ellogon init script... */
  snprintf(script, 512, "thread::send $thread {source {%s}}", path_to_ellogon);
  if (Tcl_Eval(interp, script) != TCL_OK) goto error;
  /* Get the instance id... */
  obj = Tcl_GetVar2Ex(interp, "instances", NULL, TCL_GLOBAL_ONLY);
  if (obj == NULL) goto error;
  if (Tcl_GetIntFromObj(interp, obj, &instance) != TCL_OK) goto error;
  Tcl_MutexUnlock(&ellogonEmbedMutex);
  return(instance);
error:
  fprintf(stderr, "Error in Ellogon_CreateInstance: %s\n",
            Tcl_GetStringResult(interp));
  Tcl_MutexUnlock(&ellogonEmbedMutex);
  return -1;
}; /* Ellogon_CreateInstance */

/*
 * Ellogon_CreateInstance:
 *
 * This procedure will create an instance of Ellogon, in its own thread.
 * The return value will be the instance id, or -1 in case of error...
 */
int Ellogon_InstanceEval(Tcl_Interp *interp, int instance, const char *cmd) {
  char script[512];
  Tcl_MutexLock(&ellogonEmbedMutex);
  snprintf(script, 512, "set thread $E(%d)", instance);
  if (Tcl_Eval(interp, script) != TCL_OK) goto error;
  /* Execute the command... */
  if (Tcl_VarEval(interp, "thread::send $thread {", cmd, "}", NULL) != TCL_OK)
                  goto error;
  Tcl_MutexUnlock(&ellogonEmbedMutex);
  return TCL_OK;
error:
  fprintf(stderr, "Error in Ellogon_InstanceEval: %s\n",
            Tcl_GetStringResult(interp));
  Tcl_MutexUnlock(&ellogonEmbedMutex);
  return TCL_ERROR;
}; /* Ellogon_InstanceEval */

