/*****************************************************************************
 *  EmbedAPI.cpp:                                                            *
 * ----------------------                                                    *
 *  This file implements an API for embeding Ellogon in a C++ application.   *
 *                                                                           *
 *  Copyright 1998-2009 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licencing information, please refer to the licence conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/

#include "EmbedAPI.hpp"
#include <iostream>

using namespace std;
using namespace ELLOGON;

/* 
 * Access to the list of threads and to the thread send results
 * (defined below) is guarded by this mutex. 
 */
TCL_DECLARE_MUTEX(ellogonEmbedMutex)

static bool TclInitialized = false;
static bool debugFlag      = false;

#define COUT cout

/*
 * InitializeTcl:
 *
 * This method initialises Tcl (once).
 */
Tcl_Interp *Ellogon::InitializeTcl(Tcl_Interp *ip) {
  Tcl_MutexLock(&ellogonEmbedMutex);
  if (!TclInitialized) {
    if (debugFlag) {
      COUT << "  + Initialising the Tcl library... [thread: "
           << GetThreadId() << "]" << endl << flush;
    }
    /* This is the first time we have been called. Initialise Tcl... */
    if (ip != NULL) {
      /* If a Tcl interpeter has been provided, initialise stubs... */
#if defined(USE_TCL_STUBS) || defined(CDM_USE_STUBS)
      if (Tcl_InitStubs(ip, "8.4", 0) == NULL) {
        Tcl_MutexUnlock(&ellogonEmbedMutex);
        return NULL;
      }
#endif /* USE_TCL_STUBS || CDM_USE_STUBS */
    } else {
      /* No interpreter has been provided. Create one... */
      ip = Tcl_CreateInterp();
    }
    Tcl_FindExecutable(NULL);
    if (Tcl_Init(ip) == TCL_ERROR) {
      cerr << "Error in Ellogon_Init: " << Tcl_GetStringResult(ip) << endl;
      Tcl_MutexUnlock(&ellogonEmbedMutex);
      return NULL;
    }
    TclInitialized = true;
    if (debugFlag) {
      COUT << "  + Tcl library initialised! [thread: "
           << GetThreadId() << "]" << endl << flush;
    }
  } else {
    /* Already initialised... */
  }
  Tcl_MutexUnlock(&ellogonEmbedMutex);
  return ip;
}; /* Ellogon::InitializeTcl */

/*
 * InitializeInterpreter:
 *
 * This method creates a Tcl interpreter.
 */
int Ellogon::InitializeInterpreter(Tcl_Interp *ip) {
  interp = InitializeTcl(ip);
  if (interp == NULL) {
    Tcl_MutexLock(&ellogonEmbedMutex);
    interp = Tcl_CreateInterp();
    Tcl_Init(interp);
    Tcl_MutexUnlock(&ellogonEmbedMutex);
  }
  if (interp) {
    if (debugFlag) {
      COUT << "  + Initialising a Tcl interpreter... [thread: "
           << GetThreadId() << "]" << endl << flush;
      COUT << "    -> auto_path: \"" << variable("auto_path")
           << "\" [thread: " << GetThreadId() << "]" << endl << flush;
    }
    Tcl_Preserve(interp);
    return TCL_OK;
  }
  return TCL_ERROR;
}; /* Ellogon::InitializeInterpreter */

/*
 * Ellogon:
 *
 * The empty constructor. init() must be called to an Ellogon instance.
 */
Ellogon::Ellogon(void) {
  interp = NULL;
}; /* Ellogon::Ellogon */

/*
 * Ellogon:
 *
 * The default constructor. Starts the Ellogon instance.
 */
Ellogon::Ellogon(const char *path_to_ellogon,
                 int argc, const char **argv) {
  interp = NULL;
  init(path_to_ellogon, argc, argv);
}; /* Ellogon::Ellogon */

/*
 * init:
 *
 * The default constructor. Starts the init instance.
 */
int Ellogon::init(const char *path_to_ellogon,
              int argc, const char **argv) {
  int status;
  if (debugFlag) {
    COUT << "Ellogon::init() called on thread: "
         << GetThreadId() << endl << flush;
  }
  /* If there is another interpreter, kill it... */
  if (interp) destroy();
  /* Create a Tcl interpreter... */
  InitializeInterpreter();
  if (interp == NULL) return TCL_ERROR;
  eval("set ::argv {}");
  /* Set any provided arguments... */
  for (int i = 0; i < argc; ++i) {
    Tcl_VarEval(interp, "lappend ::argv {", argv[i], "}", NULL);
  }
  eval("set ::argc [llength $::argv]");
  /* Initialise Ellogon... */
  if (path_to_ellogon) {
    if (debugFlag) {
      COUT << "  + Starting Ellogon... [thread: "
           << GetThreadId() << "]" << endl << flush;
      COUT << "    -> " << path_to_ellogon << endl << flush;
    }
    status = Tcl_VarEval(interp, "source {", path_to_ellogon, "}", NULL);
    if (debugFlag) {
      COUT << "    -> status = " << status << "(TCL_OK=" << TCL_OK
           << ") [thread: " << GetThreadId() << "]" << endl << flush;
      if (status != TCL_OK) {
        COUT << "    -> ERROR: " << errorInfo()
             << " [thread: " << GetThreadId() << "]" << endl << flush;
      }
    }
  }
  if (debugFlag) {
    COUT << "Ellogon::init() finished on thread: "
         << GetThreadId() << endl << flush;
  }
  return status;
}; /* init::init */

void Ellogon::destroy(void) {
  if (debugFlag) {
    COUT << "Ellogon::destroy() called on thread: "
         << GetThreadId() << endl << flush;
  }
  if (interp) {
    if (debugFlag) {
      COUT << "  + Killing previous instance... [thread: "
           << GetThreadId() << "]" << endl << flush;
    }
    Tcl_Release(interp);
    Tcl_DeleteInterp(interp);
    interp = NULL;
  }
  if (debugFlag) {
    COUT << "Ellogon::destroy() finished on thread: "
         << GetThreadId() << endl << flush;
  }
}; /* Ellogon::destroy */

Ellogon::~Ellogon(void) {
  destroy();
}; /* Ellogon::~Ellogon */

/*
 * Ellogon::eval:
 *
 * Evaluates an Ellogon or Tcl command/script.
 */
int Ellogon::eval(const char *cmd) const {
  if (!interp) return TCL_ERROR;
  return Tcl_Eval(interp, cmd);
}; /* Ellogon::eval */

/*
 * Ellogon::result:
 *
 * Returns the result of the previous call to eval as a string.
 * The returned pointer will be valid until the next call to eval, or until
 * the object is destroyed.
 */
const char *Ellogon::result(void) const {
  if (!interp) return NULL;
  return Tcl_GetStringResult(interp);
}; /* Ellogon::result */

/*
 * Ellogon::errorInfo:
 *
 * Returns the errorInfo of the previous call to eval as a string.
 * The returned pointer will be valid until the next call to eval, or until
 * the object is destroyed.
 */
const char *Ellogon::errorInfo(void) const {
  Tcl_Obj *obj;
  if (!interp) return NULL;
  obj = Tcl_GetVar2Ex(interp, "::errorInfo", NULL, TCL_GLOBAL_ONLY);
  if (obj == NULL) return NULL;
  return Tcl_GetString(obj);
}; /* Ellogon::errorInfo */

/*
 * Ellogon::variable:
 *
 * Returns the value of the specified variable.
 */
const char *Ellogon::variable(const char *name, const char *key) const {
  if (!interp) return NULL;
  Tcl_Obj *obj = Tcl_GetVar2Ex(interp, name, key, TCL_GLOBAL_ONLY);
  if (!obj) return NULL;
  return Tcl_GetString(obj);
}; /* Ellogon::variable */

/*
 * Ellogon::GetThreadId:
 *
 * Returns the id of the current thread.
 */
Tcl_ThreadId Ellogon::GetThreadId(void) {
  return Tcl_GetCurrentThread();
}; /* Ellogon::GetThreadId */

void Ellogon::debug(bool enable) {
  Tcl_MutexLock(&ellogonEmbedMutex);
  debugFlag = enable;
  Tcl_MutexUnlock(&ellogonEmbedMutex);
}; /* Ellogon::debug */
