######################################################################
## This is the implementation of the tBuild Tcl extension.
## tBuild tries to provide a Tcl interface that can be used to
## build Tcl extensions in a portable way. tBuild will try to 
## identify how compilation and linkage of dynamic libraries can
## be done under a specific operating system.
##
## tBuild can be configured in two ways: It can operate by executing
## OS commands that build the shared object, or writes a makefile that
## can be used to build the extension.
##
## tBuild is copyrighted by:
## George Petasis, National Centre for Scientific Research "Demokritos",
## Aghia Paraskevi, Athens, Greece.
## e-mail: petasis@iit.demokritos.gr
##
## The following terms apply to all files associated
## with the software unless explicitly disclaimed in individual files.
##
## The authors hereby grant permission to use, copy, modify, distribute,
## and license this software and its documentation for any purpose, provided
## that existing copyright notices are retained in all copies and that this
## notice is included verbatim in any distributions. No written agreement,
## license, or royalty fee is required for any of the authorized uses.
## Modifications to this software may be copyrighted by their authors
## and need not follow the licensing terms described here, provided that
## the new terms are clearly indicated on the first page of each file where
## they apply.
## 
## IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
## FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
## ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
## DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
## POSSIBILITY OF SUCH DAMAGE.
## 
## THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
## INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
## FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
## IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
## NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
## MODIFICATIONS.
##

#     -------------------------------------------------------------------
#              I n i t i a l i s a t i o n      S e c t i o n
#     -------------------------------------------------------------------

namespace eval ::tBuild {
  variable tBuild
  variable tBuildMake
  variable TclConfig
  variable TkConfig
};# namespace tBuild

## ::tBuild::Reset
## Reset the tBuild package into the initial state.
proc ::tBuild::Reset {} {
  variable tBuild
  variable tBuildMake
  variable Tools
  unset -nocomplain tBuild
  unset -nocomplain tBuildMake
  unset -nocomplain Tools
  IdentifyPlatform
};# ::tBuild::Reset

## This function tries to identify what OS we are running on and set some
## platform specific parameters.
proc ::tBuild::IdentifyPlatform {} {
  global   tcl_platform env
  variable tBuild

  set tBuild(Platform)      $tcl_platform(platform)
  set tBuild(OS)            $tcl_platform(os)
  set tBuild(OSVersion)     $tcl_platform(osVersion)
  set tBuild(Debug)         0
  set tBuild(UseStubs)      1
  set tBuild(OutputMode)    exec
  set tBuild(ExecCallback)  {}
  set tBuild(UseTclDefines) 0
  if {![info exists tBuild(tBuildScript)]} {
    ## Set the home only once. Subsequent calls to IdentifyPlatform will
    ## place the empty script, as [info script] returns nothing...
    set tBuild(tBuildScript)  [file normalize [info script]]
    set tBuild(tBuildHome)    [file normalize [file dirname [info script]]]
  }
  
  ## Do platform specific initialisation...
  switch $tBuild(Platform) {
    unix    {::tBuild::InitialisePlatformUnix}
    windows {::tBuild::InitialisePlatformWindows}
    default {error "platform $tBuild(Platform) is not yet supported by tBuild"}
  }
  ## Load a platform-depended configuration file, if available...
  set DefaultConf [file join $tBuild(tBuildHome) platform $tBuild(OS).cfg]
  if {[file exists $DefaultConf]} {
    if {[catch {source $DefaultConf} error]} {
      ## Do what? Under windows we cannot use puts...
    }
  }
  ## Load a user specific platform-depended configuration file, if available...
  if {[info exists env(HOME)]} {
    set UserHome $env(HOME)
  } else {
    set UserHome [pwd]
  }
  set UserConf [file join $UserHome tBuild platform $tBuild(OS).cfg]
  if {[file exists $UserConf]} {
    if {[catch {source $UserConf} error]} {
      ## Do what? Under windows we cannot use puts...
    }
  }
  MakefileInit
  package provide tBuild 1.0
};# ::tBuild::IdentifyPlatform

## Initialisation for Unix platforms.
proc ::tBuild::InitialisePlatformUnix {} {
  variable tBuild
  variable Tools
  variable TclConfig
  variable TkConfig

  LoadTclConfigFiles
  ## Do OS specific initialisation
  switch $tBuild(OS) {
    Linux   {}
    Solaris {}
    default {}
  }
  foreach key {TK_XINCLUDES TK_LIB_SPEC TK_STUB_LIB_SPEC TK_XLIBSW} {
    if {![info exists TkConfig($key)]} {
      set TkConfig($key) {}
    }
  }

  set tBuild(Compiler)          $TclConfig(TCL_CC)
  set tBuild(CompileFlag)       -c
  set tBuild(IncludeFlag)       -I
  set tBuild(DefineFlag)        -D
  set tBuild(CompileDebugFlag)  $TclConfig(TCL_CFLAGS_DEBUG)
  set tBuild(CompileSharedFlag) $TclConfig(TCL_SHLIB_CFLAGS)
  set tBuild(WarningFlags)      $TclConfig(TCL_CFLAGS_WARNING)
  set tBuild(OptimiseFlags)     $TclConfig(TCL_CFLAGS_OPTIMIZE)
  set tBuild(CompileFlags)      {}
  set tBuild(CompileDefines)    $TclConfig(TCL_DEFS)
  set tBuild(TclIncludeDirs)    "$TclConfig(TCL_INCLUDE_SPEC)\
                                 $TkConfig(TK_XINCLUDES)"
  set tBuild(ObjectSuffix)      .o

  set tBuild(LinkSharedLib)     $TclConfig(TCL_SHLIB_LD)
  set tBuild(LinkDebugFlag)     {}
  set tBuild(LinkNoDebugFlag)   {}
  set tBuild(LinkOutputFlag)    -o
  set tBuild(LinkLibPaths)      "$TclConfig(TCL_LIB_SPEC)\
                                 $TclConfig(TCL_STUB_LIB_SPEC)\
                                 $TkConfig(TK_LIB_SPEC)\
                                 $TkConfig(TK_STUB_LIB_SPEC)\
                                 $TkConfig(TK_XLIBSW)"
  set tBuild(LinkLibs)          $TclConfig(TCL_LIBS)
  set tBuild(LinkLibFlag)       -l
  set tBuild(LinkLibPathFlag)   -L

  ## Locate various tools...
  if {[catch {exec which flex} Tools(flex)]} {set Tools(flex) {}}
  if {[catch {exec which swig} Tools(swig)]} {set Tools(swig) {}}
};# ::tBuild::InitialisePlatformUnix

## Initialise for Windows platforms.
proc ::tBuild::InitialisePlatformWindows {} {
  global   env
  variable tBuild
  variable Tools

  ## We have to find where Tcl is hidden:-). Probably the interpreter we are
  ## running in knows something...
  set LibDir  [info library]
  set TclDir  [file native [file dirname [file dirname $LibDir]]]
  set Version [string map {. {}} [info tclversion]]

  ## We should locate a compiler. For now, only MS Visual C++ v. 5.x/6.x/NET is
  ## supported...
  set CC_DIR {}
  
  if {[info exists env(MSVCDIR)]} {
    set CC_DIR $env(MSVCDIR)
    set CC     [file join $CC_DIR Bin CL.EXE]
    set LINK   [file join $CC_DIR Bin LINK.EXE]
  } else {
    package require registry
    ## The first key we find in the registry wins:-)
    set root "HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\DevStudio\\"
    set path1 "${root}6.0\\Products\\Microsoft Visual C++"
    set path2 "${root}5.0\\Products\\Microsoft Visual C++"
    set path3 "HKEY_LOCAL_MACHINE\\SOFTWARE\\Classes\\dswfile\\shell\\&Open with MSDev\\command"
    set path4 "HKEY_LOCAL_MACHINE\\SOFTWARE\\Classes\\dswfile\\shell\\Open\\command"
    set path5 "HKEY_CLASSES_ROOT\\Applications\\devenv.exe\\shell\\open\\command"
    foreach key [list $path1 $path2] {
      if {![catch {registry get $key ProductDir} CC_DIR]} {break}
      set CC_DIR {}
    }
    if {[string length $CC_DIR]} {
      set CC   [file join $CC_DIR Bin CL.EXE]
      set LINK [file join $CC_DIR Bin LINK.EXE]
    } else {
      if {![catch {registry get $path3 {}} MSDEV_DIR]} {
        set MSDEV_DIR [string trim $MSDEV_DIR {\"%1 }]
        set MSDEV_DIR [file dir [file dir [file dir [file dir $MSDEV_DIR]]]]
        set CC_DIR [file join $MSDEV_DIR VC98]
        set CC     [file join $CC_DIR Bin CL.EXE]
        set LINK   [file join $CC_DIR Bin LINK.EXE]
      } elseif {![catch {registry get $path4 {}} MSDEV_DIR] ||
                ![catch {registry get $path5 {}} MSDEV_DIR]} {
        ## Here we found the .NET version...
        set MSDEV_DIR [string trim $MSDEV_DIR {\"%1 }]
        set MSDEV_DIR [file dir [file dir [file dir $MSDEV_DIR]]]
        set CC_DIR [file join $MSDEV_DIR VC7]
        set CC     [file join $CC_DIR Bin CL.EXE]
        set LINK   [file join $CC_DIR Bin LINK.EXE]
        ## Set some environmental variables, in the same way running the
        ## VSVARS32.bat would have done...
        set env(VSINSTALLDIR) [file native [file join $MSDEV_DIR Common7 IDE]]
        set env(VCINSTALLDIR) [file native $MSDEV_DIR]
        set env(DevEnvDir)    $env(VSINSTALLDIR)
        set env(MSVCDir)      [file native $CC_DIR]
        set env(FrameworkDir) [file native [file join $MSDEV_DIR Framework]]
        set env(FrameworkSDKDir) \
                              [file native [file join $MSDEV_DIR FrameworkSDK]]
        if {[info exists env(INCLUDE)]} {
          set INCLUDE $env(INCLUDE)
        } else {set INCLUDE {}}
        if {[info exists env(LIB)]} {
          set LIB $env(LIB)
        } else {set LIB {}}
        if {[info exists env(PATH)]} {
          set PATH $env(PATH)
        } else {set PATH {}}
        set env(INCLUDE) [join [list \
             [file native [file join $CC_DIR ATLMFC INCLUDE]] \
             [file native [file join $CC_DIR INCLUDE]] \
             [file native [file join $CC_DIR PlatformSDK include prerelease]] \
             [file native [file join $CC_DIR PlatformSDK include]] \
             [file native [file join $MSDEV_DIR FrameworkSDK include]] \
             $INCLUDE] \;]
        set env(LIB) [join [list \
             [file native [file join $CC_DIR ATLMFC LIB]] \
             [file native [file join $CC_DIR LIB]] \
             [file native [file join $CC_DIR PlatformSDK lib prerelease]] \
             [file native [file join $CC_DIR PlatformSDK lib]] \
             [file native [file join $MSDEV_DIR FrameworkSDK lib]] \
             $LIB] \;]
        set env(PATH) [join [list $env(DevEnvDir) \
             [file native [file join $env(MSVCDir) BIN]] \
             [file native [file join $env(VCINSTALLDIR) Common7 Tools]] \
             [file native [file join $env(VCINSTALLDIR) Common7 Tools \
                                                        bin prerelease]] \
             [file native [file join $env(VCINSTALLDIR) Common7 Tools bin]] \
             [file native [file join $env(FrameworkSDKDir) bin]] \
             $PATH] \;]
      }
    }
  }
  if {![string length $CC_DIR]} {
    error "cannot locate MS Visual C++ (supported versions: 5.x/6.x/.NET)!"
  }
  if {![file exists $CC]} {
    error "Non-existent C++ compiler path \"$CC\" (from registry)"
  }
  if {![file exists $LINK]} {
    error "Non-existent C++ linker path \"$LINK\" (from registry)"
  }

  set tBuild(Compiler)          [list $CC]
  set tBuild(CompileFlag)       -c
  set tBuild(IncludeFlag)       -I
  set tBuild(DefineFlag)        -D
  set tBuild(CompileDebugFlag)  {-Z7 -Od -WX -debug:full -debugtype:cv}
  set tBuild(CompileSharedFlag) {};# -GD is deprecated and removed in .NET...
  set tBuild(WarningFlags)      {-W3}
  set tBuild(OptimiseFlags)     {-O2}
  set tBuild(CompileFlags)      {-nologo -Gs -MD}
  set tBuild(CompileDefines)    {}
  set tBuild(TclIncludeDirs)    [list -I$TclDir\\include]
  set tBuild(ObjectSuffix)      .obj

  set tBuild(LinkSharedLib)     "[list $LINK] -dll"
  set tBuild(LinkDebugFlag)     "-debug -nologo"
  set tBuild(LinkNoDebugFlag)   "-release -nologo"
  set tBuild(LinkOutputFlag)    -out:
  set tBuild(LinkLibPaths)      [list -libpath:$TclDir\\lib]
  set tBuild(LinkLibs)          "tcl$Version.lib tk$Version.lib\
                                 tclstub$Version.lib tkstub$Version.lib"
  set tBuild(LinkLibFlag)       {}
  set tBuild(LinkLibPathFlag)   -libpath:
  
  ## Locate various tools...
  set Tools(flex) flex.exe
};# ::tBuild::InitialisePlatformWindows

## Load the Tcl & Tk configuration files, in order to "steal" how Tcl/Tk was
## configured & build on a Unix platform.
proc ::tBuild::LoadTclConfigFiles {} {
  variable TclConfig
  variable TkConfig
  set LibDir [info library]
  set ConfDir [file dirname $LibDir]
  set TclConfFileName [file join $ConfDir tclConfig.sh]
  set TkConfFileName  [file join $ConfDir tkConfig.sh]

  ## Initialise some variables by trying to locate the proper executables...
  foreach {prg var} [list ar AR {-ldl -lm} LIBS $LibDir LIB_RUNTIME_DIR \
                          [info tclversion] VERSION g DBGX] {
    ## Is the variable already defined? Perhaps the caller has already set
    ## this variable, because we have to use a specific executable...
    if {[info exists TclConfig($var)]} {
      set $var [set TclConfig($var)]
      continue
    }
    ## Try to locate the executable...
    if {[catch {exec which $prg} $var]} {
      set $var $prg
    }
    set TclConfig($var) [set $var]
  }
  ## Parse the two configuration files...
  foreach filename [list $TclConfFileName $TkConfFileName] \
          Array    {TclConfig TkConfig} {
    ## Does the file exists? If not, this is a problem:-)
    if {![file exists $filename]} {
      switch [string tolower $::tcl_platform(os)] {
        linux {set msg "File \"$filename\" is missing from your Tcl/Tk\
          installation. If you have compiled Tcl/Tk from sources, please\
          re-install it. If you have installed a binary distribution (like an\
          RPM package), perhaps this file is contained in a development package\
          you forgot to install?"
        }
        default {
          set msg "File \"$filename\" is missing from your Tcl/Tk\
          installation. If you have compiled Tcl/Tk from sources, please\
          re-install it. If you have installed a binary distribution,\
          perhaps this file is contained in a development package\
          you forgot to install?"
        }
      }
      error $msg
    }
    set file [open $filename]
      while {[gets $file line] != -1} {
        set line [string trim $line]
        ## Skip empty lines...
        if {![string length $line]} {continue}
        ## Skip Comments...
        if {[string equal \# [string range $line 0 0]]} {continue}
        ## If we got here, a variable is defined...
        set varspec [split $line =]
        ## Has 'varspec' 2 elements? We are lucky:-)
        if {[llength $varspec] == 2} {
          foreach {varname varvalue} $varspec {break}
          ## Define the variable globally, in the proper array...
        } else {
          ## Here we suppose that the variable name does not contain "=":-)
          set varname  [lindex $varspec 0]
          set varvalue [join [lrange $varspec 1 end] =]
        }
        set varvalue [string trim $varvalue ']
        set $varname $varvalue
        ## Eval is used here in order to solve variable references in
        ## variable values...
        eval set ${Array}($varname) \[list $varvalue\]
      }
    close $file
  }
};# ::tBuild::LoadTclConfigFiles

#     -------------------------------------------------------------------
#                C o n f i g u r a t i o n      S e c t i o n
#     -------------------------------------------------------------------

## Change tBuild configuration options
proc ::tBuild::Configure {args} {
  variable tBuild
  foreach {opt val} $args {
    switch -glob -- $opt {
      -deb*  {set tBuild(Debug) $val}
      -exe*  {set tBuild(ExecCallback) $val}
      -mod*  {
        switch $val {
          exec* {set tBuild(OutputMode) exec}
          mak*  {set tBuild(OutputMode) make}
          default {error "invalid value $val for option $opt"}
        }
      }
      -use*  -
      -stu*  {set tBuild(UseStubs) $val}
      -tcld* {set tBuild(UseTclDefines) $val}
      default {error "unknown option $opt"}
    }
  }
  return {}
};# ::tBuild::Configure

## Request tBuild configuration options
proc ::tBuild::Cget {opt} {
  variable tBuild
  switch -glob -- $opt {
    -deb*  {return $tBuild(Debug)}
    -exe*  {return $tBuild(ExecCallback)}
    -mod*  {return $tBuild(OutputMode)}
    -use*  -
    -stu*  {return $tBuild(UseStubs)}
    -tcld* {return $tBuild(UseTclDefines)}
    default {error "unknown option $opt"}
  }
  return {}
};# ::tBuild::Cget

## Request/Set tBuild platform-specific information
proc ::tBuild::option {{opt keys} {value @@}} {
  variable tBuild
  variable Tools
  if {[info exists tBuild($opt)]} {
    if {![string equal $value @@]} {set tBuild($opt) $value}
    return $tBuild($opt)
  } elseif {[info exists Tools($opt)]} {
    if {![string equal $value @@]} {set Tools($opt) $value}
    return $Tools($opt)
  } else {
    if {[string equal $opt keys]} {
      ## A request to deliver all the recognised keys...
      return [lsort [concat [array names tBuild] [array names Tools]]]
    } else {
      error "unknown option \"$opt\". Valid options are:\
            \n[lsort [concat [array names tBuild] [array names Tools]]]"
    }
  }
};# ::tBuild::option

#     -------------------------------------------------------------------
#                     M a k e f i l e      S e c t i o n
#     -------------------------------------------------------------------

## Initialise a Makefile session
proc ::tBuild::MakefileInit {} {
  variable tBuildMake
  catch {array unset tBuildMake}
};# ::tBuild::MakefileInit

## Reguest a tBuild Makefile. The makefile is returned as the result of this
## procedure...
proc ::tBuild::Makefile {} {
  variable tBuild
  variable tBuildMake
  set Makefile \
"########################################################################
## This is an auto-generated Makefile by the tBuild Tcl extension.
## This makefile is probably specific to the platform (or even the
## machine) it was generated. Modify with care...
########################################################################\n\n\n"
  switch $tBuild(Platform) {
    windows -
    unix {
      ## Write variable definitions...
      append Makefile \
      "############################################################\n" \
      "## Definitions\n" \
      "############################################################\n"
      foreach key [lsort [array names tBuildMake -*]] {
        set key [string range $key 1 end]
        append Makefile "[string toupper $key]=" \
               $tBuild($key) \n
      }
      append Makefile \n
      append Makefile \
      "############################################################\n" \
      "## Commands\n" \
      "############################################################\n"
      foreach cmd $tBuildMake(Commands) {
        append Makefile [join $cmd { }] \n\n
      }
      append Makefile \
      "############################################################\n" \
      "## End of Makefile\n" \
      "############################################################\n"
    }
  }
};# ::tBuild::Makefile

#     -------------------------------------------------------------------
#                  C o m p i l a t i o n      S e c t i o n
#     -------------------------------------------------------------------

## This function will compile the given source file for inclusion in a shared
## library...
proc ::tBuild::CompileSharedObject {args} {
  variable tBuild

  set IncludeDirs $tBuild(IncludeFlag).
  set Defines {}
  set Debug $tBuild(Debug)
  set Stubs $tBuild(UseStubs)
  set UserFlags {}
  set Files {}
  set TclDefs $tBuild(UseTclDefines)
  
  set argc [llength $args]
  for {set i 0} {$i < $argc} {incr i} {
    set opt [lindex $args $i]
    switch -glob -- $opt {
      -deb*   {incr i;set     Debug       [lindex $args $i]}
      -def*   {incr i;lappend Defines     $tBuild(DefineFlag)[lindex $args $i]}
      -fla*   {incr i;lappend UserFlags   [lindex $args $i]}
      -inc*   {incr i;lappend IncludeDirs \
                   $tBuild(IncludeFlag)[file nativename [lindex $args $i]]}
      -stu*   {incr i;set     Stubs       [lindex $args $i]}
      -tcld*  {incr i;set     TclDefs     [lindex $args $i]}
      default {lappend Files $opt}
    }
  }
  if {![llength $Files]} {return}
  if {$Stubs} {lappend Defines $tBuild(DefineFlag)USE_TCL_STUBS \
                               $tBuild(DefineFlag)USE_TK_STUBS}

  foreach file $Files {
    if {$Debug} {
      if {$TclDefs} {
        Execute tBuild(Compiler) tBuild(CompileFlag) tBuild(CompileSharedFlag)\
          tBuild(CompileDebugFlag) tBuild(WarningFlags) tBuild(OptimiseFlags) \
          tBuild(CompileFlags) UserFlags \
          tBuild(TclIncludeDirs) IncludeDirs \
          tBuild(CompileDefines) Defines file
      } else {
        Execute tBuild(Compiler) tBuild(CompileFlag) tBuild(CompileSharedFlag)\
          tBuild(CompileDebugFlag) tBuild(WarningFlags) tBuild(OptimiseFlags) \
          tBuild(CompileFlags) UserFlags \
          tBuild(TclIncludeDirs) IncludeDirs \
          Defines file
      }
    } else {
      if {$TclDefs} {
        Execute tBuild(Compiler) tBuild(CompileFlag) tBuild(CompileSharedFlag)\
          tBuild(WarningFlags) tBuild(OptimiseFlags) \
          tBuild(CompileFlags) UserFlags \
          tBuild(TclIncludeDirs) IncludeDirs \
          tBuild(CompileDefines) Defines file
      } else {
        Execute tBuild(Compiler) tBuild(CompileFlag) tBuild(CompileSharedFlag)\
          tBuild(WarningFlags) tBuild(OptimiseFlags) \
          tBuild(CompileFlags) UserFlags \
          tBuild(TclIncludeDirs) IncludeDirs \
          Defines file
      }
    }
  }

  ## Return the object filenames...
  set Objs {}
  foreach file $Files {
    lappend Objs [file tail [file rootname $file]]$tBuild(ObjectSuffix)
  }
  return $Objs
};# ::tBuild::CompileSharedObject

#     -------------------------------------------------------------------
#                    L i n k a g e      S e c t i o n
#     -------------------------------------------------------------------

##
proc ::tBuild::LinkSharedLibrary {args} {
  variable tBuild

  set Debug $tBuild(Debug)
  set LibPaths {}
  set Libs {}
  set UserFlags {}
  set Files {}
  set Output libtBuild[info sharedlibextension]
  
  set argc [llength $args]
  for {set i 0} {$i < $argc} {incr i} {
    set opt [lindex $args $i]
    switch -glob -- $opt {
      -deb*   {incr i;set     Debug       [lindex $args $i]}
      -fla*   {incr i;lappend UserFlags   [lindex $args $i]}
      -lib    {incr i;
        ## We place a requirement that the full name of the library should be
        ## given. But under unix we have to strip the prefix "lib" and the
        ## suffix...
        switch $tBuild(Platform) {
          unix {
            set _lib [lindex $args $i]
            if {![string match lib* $_lib]} {
              error "value of -lib parameter should be a complete\
                     library name, like \"libm.a\" or \"libtcl84.so\""
            }
            set _lib [file root [string range $_lib 3 end]]
            lappend Libs $tBuild(LinkLibFlag)$_lib
          }
          windows {
            lappend Libs $tBuild(LinkLibFlag)[lindex $args $i]
          }
          default {
            lappend Libs $tBuild(LinkLibFlag)[lindex $args $i]
          }
        }
      }
      -libp*  {incr i;lappend LibPaths    \
                    $tBuild(LinkLibPathFlag)[file nativename [lindex $args $i]]}
      -o*     {incr i;set     Output      [lindex $args $i]}
      default {lappend Files $opt}
    }
  }
  if {![llength $Files]} {return}

  ## Make sure that the output file can be created...
  set outputDir [file dirname $Output]
  if {[string length $outputDir]} {
    catch {file mkdir $outputDir}
  }

  if {$Debug} {
    Execute tBuild(LinkSharedLib) tBuild(LinkDebugFlag) \
      tBuild(LinkOutputFlag) Output UserFlags \
      Files tBuild(LinkLibPaths) LibPaths Libs tBuild(LinkLibs)
  } else {
    Execute tBuild(LinkSharedLib) tBuild(LinkNoDebugFlag) \
      tBuild(LinkOutputFlag) Output UserFlags \
      Files tBuild(LinkLibPaths) LibPaths Libs tBuild(LinkLibs)
  }

  ## Return the library filenames...
  return $Output
};# ::tBuild::LinkSharedLibrary

#     -------------------------------------------------------------------
#                    A c t i o n s      S e c t i o n
#     -------------------------------------------------------------------

## This command executes the directives from the tBuild commands.
## According to the output mode this function will either execute a system
## command or append to a file...
proc ::tBuild::Execute {args} {
  variable tBuild
  variable tBuildMake
  set Cmd {}

  ## Our arguments are variable names. We have to collect the values and
  ## split them.
  if {[string equal $tBuild(OutputMode) exec]} {
    set Args {}
    set PreStr {}
    foreach arg $args {
      set val [uplevel 1 set $arg]
      if {![string length $val]} {continue}
      ## Handle some parameters differently...
      switch -- $val {
        -libpath: -
        -out:     {set PreStr $val; continue}
        default   {}
      }
      if {[string length $PreStr]} {
        lappend Args $PreStr$val
        set PreStr {}
      } else {
        lappend Args $val
      }
    }
    foreach arg $Args {
      foreach element $arg {
        lappend Cmd $element
      }
    }
    set tBuild(Executing) $Cmd
    if {[string length $tBuild(ExecCallback)]} {
      uplevel #0 $tBuild(ExecCallback) $Cmd
    } else {
      eval exec $Cmd
    }
  }
  ## We add the Makefile information anyway. Perhaps the user may want to also
  ## save a Makefile :-)
  set Cmd {}
  foreach arg $args {
    ## Is the variable name a tBuild reference? This reference should be kept.
    if {[regexp {tBuild\((.*)\)} $arg tmp key]} {
      set tBuildMake(-$key) 1
      lappend Cmd \$\{$key\}
    } else {
      foreach item [uplevel 1 set $arg] {
        lappend Cmd $item
      }
    }
  }
  lappend tBuildMake(Commands) $Cmd
};# ::tBuild::Execute

##############################################################################
##############################################################################
## Initialise the tBuild extension.
::tBuild::IdentifyPlatform

## Test the package :-)
#set objs [tBuild::CompileSharedObject -debug 1 \
#                            -define PETASIS=8 -define ADD \
#                            -define TCL_SHLIB_EXT=\"[info shared]\" \
#                            -flag -KALO -flag -kako \
#                            -include "Program Files/Include" \
#                            -stubs 1 test.c test2.c]
#eval tBuild::LinkSharedLibrary -debug 1 -flag -kala \
#                            -output libtest[info shared] $objs
