######################################################################
## This file is part of the tBuild Tcl package. It provides an interface
## for the easy creation of TEA compliant extensions, that can be also
## built with the help of tBuild.
##
## tBuild is copyrighted by:
## George Petasis, National Centre for Scientific Research "Demokritos",
## Aghia Paraskevi, Athens, Greece.
## e-mail: petasis@iit.demokritos.gr
##
## The following terms apply to all files associated
## with the software unless explicitly disclaimed in individual files.
##
## The authors hereby grant permission to use, copy, modify, distribute,
## and license this software and its documentation for any purpose, provided
## that existing copyright notices are retained in all copies and that this
## notice is included verbatim in any distributions. No written agreement,
## license, or royalty fee is required for any of the authorized uses.
## Modifications to this software may be copyrighted by their authors
## and need not follow the licensing terms described here, provided that
## the new terms are clearly indicated on the first page of each file where
## they apply.
## 
## IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
## FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
## ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
## DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
## POSSIBILITY OF SUCH DAMAGE.
## 
## THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
## INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
## FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
## IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
## NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
## MODIFICATIONS.
##

package provide tBuild::Extension 1.0

namespace eval ::tBuild {
  namespace eval ::tBuild::tExtension {
    variable TemplatesDir
    set TemplatesDir [file dirname [file normalize [info script]]]/templates
  };# namespace ::tBuild::tExtension
};# namespace ::tBuild

#     -------------------------------------------------------------------
#                      P u b l i c     I n t e r f a c e
#     -------------------------------------------------------------------

## ::tBuild::tExtension::CreateTclExtension
#  This function creates a new TEA compliant Tcl extension.
proc ::tBuild::tExtension::CreateTclExtension {args} {
  set pkgName tclExtension
  set pkgDir  [pwd]
  set author  "\"tBuild package extension builder\""
  set commands {}

  ## Process arguments...
  foreach {param value} $args {
    switch -glob -- $param {
      -auth*  {set author  $value}
      -pkgN*  {set pkgName [string map {{ } _} $value]}
      -pkgD*  {set pkgDir  $value}
      -cmdS*  {
        if {[llength $value] !=2} {
          error "-cmdSpec requires a list of two items (tcl-cmd_name c-name)"
        }
        lappend commands $value
      }
      default {error "unknown option \"$param\""}
    }
  };# foreach {param value} $args

  ## Create the directory the extension will be hosted.
  set homeDir [file join [file normalize $pkgDir] $pkgName]
  file mkdir $homeDir
  ## Create the various directories.
  foreach subdir {generic lib} {
    file mkdir $homeDir/$subdir
  }
  foreach subdir {unix win doc tests lib} {
    #file mkdir $homeDir/$subdir
  }

  ## Emmit the needed code that is common for all platforms.
  foreach {fileName template} [list \
           generic/tcl$pkgName.c pkgInit.c \
           generic/$pkgName.h    pkg.h \
           generic/$pkgName.c    pkg.c \
           build.tcl             build.tcl \
           lib/pkgIndex.tcl      pkgIndex.tcl \
           README                README \
           lisence.terms         lisence.terms \
           ] {
    set file [open $homeDir/$fileName w]
    CreateTclExtension_WriteTemplate $file $template $pkgName $author $commands
    close $file
  }
};# ::tBuild::tExtension::CreateTclExtension

#     -------------------------------------------------------------------
#                      P r i v a t e     I n t e r f a c e
#     -------------------------------------------------------------------

## ::tBuild::tExtension::CreateTclExtension_WriteTemplate
#  Internal function for expanding templates.
proc ::tBuild::tExtension::CreateTclExtension_WriteTemplate {file template
                                              pkgName author commands} {
  variable TemplatesDir
  ## Process the given commands.
  set headerCode           {}
  set headerCallbackCode   {}
  set functionCode         {}
  set functionCallbackCode {}
  set registrationCode     {}
  foreach command $commands {
    foreach {tcl_name c_name} $command {
      lappend headerCallbackCode \
      "static int ${pkgName}_${tcl_name}_ObjCmd _ANSI_ARGS_((\n   ClientData\
       clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv\[\]));"
      lappend headerCode \
      "void ${c_name} _ANSI_ARGS_((void));"
      lappend functionCallbackCode \
"/*
 *----------------------------------------------------------------------
 *
 * ${pkgName}_${tcl_name}_ObjCmd --
 *
 *      Implements the new Tcl \"$tcl_name\" command.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      The function \"${c_name}\" is executed.
 *
 *----------------------------------------------------------------------
 */
static int
${pkgName}_${tcl_name}_ObjCmd(ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv\[\]) {
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, \"\");
    return TCL_ERROR;
  }
  /* Please modify the function call and place the proper result */
  ${c_name}();
  Tcl_ResetResult(interp); /* This call can be ommited if you set a result */
  return TCL_OK;
}; /* ${pkgName}_${tcl_name}_ObjCmd */"
      lappend functionCode \
"/*
 *----------------------------------------------------------------------
 *
 * ${c_name} --
 *
 *      Implements the new Tcl \"$tcl_name\" command.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The function \"${c_name}\" is executed.
 *
 *----------------------------------------------------------------------
 */
void
${c_name}(void) {
}; /* ${c_name} */"
      lappend registrationCode \
       "if(Tcl_CreateObjCommand(interp, \"${tcl_name}\",
                      (Tcl_ObjCmdProc*) ${pkgName}_${tcl_name}_ObjCmd,
                      (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL)  == NULL)
         return TCL_ERROR;"
    }
  }
  set keywordValues [list \
    %pkgName%         $pkgName \
    %pkgNameTitle%    [string totitle $pkgName] \
    %pkgNameUpper%    [string toupper $pkgName] \
    %year%            [clock format [clock seconds] -format %Y] \
    %author%          $author \
    %functions%       [join $functionCode         ] \
    %callbacks%       [join $functionCallbackCode ] \
    %prototypes%      [join $headerCode           \n] \
    %callbacksproto%  [join $headerCallbackCode   \n] \
    %registration%    [join $registrationCode     \n] \
  ]
  set initFile [open $TemplatesDir/$template]
  set code [string map $keywordValues [read $initFile]]
  close $initFile
  puts $file [string trimright $code]
};# ::tBuild::tExtension::CreateTclExtension_WriteTemplate
