#!/bin/sh
# The next line is executed by /bin/sh, but not tcl \
exec tclsh "$0" ${1+"$@"}
set CWD [pwd]
cd [file dirname [file normalize [info script]]]
## Tcl Script to Build the Main Gazetteer from individual list files...
set MODULE_NAME HGazetteer
if {$argc != 0} {
  set MODULE_NAME [lindex $argv 0]
}

proc WriteLexHeader File {
  global MODULE_NAME
  puts $File \
"%\{
 /*
  * This is an autogenerated Lexer. 
  * Author: George Petasis, petasis@iit.demokritos.gr
  *       9/10/1999
  *  Modified at 19 September 2000
  *  Updated: 27 November 2006: to flex 2.5.4 (Fedora Core 5)
  */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <CDM.h>

#ifdef __WIN32__
#include <io.h>
#define snprintf _snprintf
#define isatty   _isatty
#endif /* __WIN32__ */

#define ${MODULE_NAME}_Type(type, tag) \\
        HGazetteer_ProcessMatch((unsigned char *) yytext, type, tag)

extern /*\"C\"*/
void ${MODULE_NAME}_ProcessMatch(unsigned char *, const char *, const char *);
%\}

SPAN        \\,\[0-9\]+\\+\[0-9\]+>
SPAN_ONLY   \\,\[0-9\]+\\+\[0-9\]+
NUMBER      \\ \[0-9\]+
TIME        \\ \[0-9\]?\[0-9\]:\[0-9\]\[0-9\]?(-\[0-9\]?\[0-9\]:\[0-9\]\[0-9\]?)?
DATE        \\ \[0-9\]?\[0-9\]\[-/\]\[0-9\]?\[0-9\](\[-/\]\[0-9\]?\[0-9\])?
YEAR        (1\[8-9\]\[0-9\]\[0-9\]|20\[0-9\]\[0-9\])
YEARS       ({YEAR}\"s\"|{YEAR}{SPAN}?-{SPAN}?{YEAR})
YEAR2       \\ \[0-9\]\[0-9\]
YEARS2      {YEAR2}\"s\"
SSTARTT     (^|>)

%%

 /*
  * Rule Section...
  */
"}

proc Date/TimeRules File {
  global MODULE_NAME
  puts $File \
"
(\\ {YEAR}|\\ {YEARS}|\\ {YEARS2}){SPAN_ONLY}         {
                      ${MODULE_NAME}_Type(\"year\", \"year\");}
{TIME}{SPAN_ONLY}                                     {
                      ${MODULE_NAME}_Type(\"time\", \"time\");}
{DATE}{SPAN_ONLY}                                     {
                      ${MODULE_NAME}_Type(\"date\", \"date\");}
{NUMBER}{SPAN}?(\[-/\]{NUMBER})?{SPAN_ONLY}           {
                      ${MODULE_NAME}_Type(\"number\", \"number\");}
{NUMBER}{SPAN}?(\[-/\]{NUMBER})?{SPAN}?\\%{SPAN_ONLY} {
                      ${MODULE_NAME}_Type(\"percent\", \"percent\");}

"
}

proc WriteFooter File {
  global MODULE_NAME LEXER FUNCTIONS
  set lexer [string map {- _ { } _} $LEXER]
  set FUNCTIONS($lexer) 0
  puts $File ". {}\n"
  puts $File "%%\n"
  puts $File "int yywrap() {return 1;}\n"
#  puts $File \
#"int ${MODULE_NAME}_Type(char *type, char *subtype) {
#  ${MODULE_NAME}_ProcessMatch(yytext, type, subtype);
#  return 0;
#} 
puts $File \
"int ${MODULE_NAME}Start_${lexer}(char *Text) {
  YY_BUFFER_STATE flex_buffer;
  int status;
  
  /*printf(\"${MODULE_NAME}Start_${lexer} Called!\\n\");*/
  /*flex_buffer = yy_scan_string((yyconst YY_CHAR *) Text);*/
  flex_buffer = yy_scan_string((yyconst char *) Text);
  status = yylex();
  yy_delete_buffer(flex_buffer);
  return status;
}
"
}

## Guther Results...
set DestDir [file join .. Lexers]
puts "   **** Creating Lexer Directory \"$DestDir\"..."
file delete -force $DestDir
file mkdir  $DestDir
puts "   **** Processing List Files..."
# fconfigure stdout -encoding 
foreach one [glob -nocomplain *.lst] {
  set Items 0
  puts "     + Processing List File: \"$one\"..."
  set type [file rootname [file tail $one]]
  set file [open $one r]
  fconfigure $file -encoding iso8859-7
  set LEXER [string totitle [file rootname $one]]
  set FILES($LEXER) 0
  set dest [open [file join $DestDir $LEXER.lex] w]
  fconfigure $dest -encoding iso8859-7
  WriteLexHeader $dest
  array unset Hash
  array unset Index

  while {1} {
    set line [gets $file]
    if {[eof $file]} {break}
    if {![string length $line]} {continue}
    ## See if the entry has a subtype...
    set index [string last :=: $line]
    if {$index != -1} {
      set NE  [string trim [string range $line 0 [expr {$index-1}]]]
      set Tag [string trim [string range $line [expr {$index+3}] end]]
    } else {
      set NE  [string trim $line]
      set Tag {}
    }
    if {![info exists Hash($NE)]} {
      lappend Index([string length $NE]) [list $NE $type $Tag]
      set Hash($NE) 1
      incr Items
    }
  }

  ## Print Results
  puts "       + Saving Lexer \"[file join $DestDir $LEXER.lex]\"..."
  catch {
    set remove [llength $Index(0)]
    incr Items -$remove
    unset Index(0)
  }
  set line 0; set lexer_index 1
  foreach index [lsort -integer -decreasing [array names Index]] {
    foreach val $Index($index) {
      incr line
      if {$line >= 300} {
        puts "       *** Lexer too long: splitting ($lexer_index)..."
        WriteFooter $dest
        close $dest
        set index_ [string last _ $LEXER]
        if {$index_ >0} {set LEXER [string range $LEXER 0 [expr {$index_-1}]]}
        set LEXER ${LEXER}_$lexer_index
        set FILES($LEXER) 0
        set dest [open [file join $DestDir $LEXER.lex] w]
        fconfigure $dest -encoding iso8859-7
        WriteLexHeader $dest
        set line 0
        incr lexer_index
      }
      foreach {text type tag} $val {break}
      set text [split $text]
      puts -nonewline $dest "{SSTARTT}"
      set list_len [llength $text]
      for {set i 0} {$i < $list_len} {incr i} {
        set token [string map \
        {/ \\/ * \\* . \\. \" \\\" \; \\\; \{ \\\{ \} \\\} \( \\\( \) \\\) \[ \\\[ \] \\\] + \\+ , \\, ' \\' | \\| # \\#}\
          [lindex $text $i]]
        if {[string match {[0-9]*} $token]} {puts -nonewline $dest {\ }}
        if {$i == $list_len -1} {
          puts -nonewline $dest "$token{SPAN_ONLY}"
        } else {
          puts -nonewline $dest "$token{SPAN}"
        }
      }
      puts $dest " {${MODULE_NAME}_Type(\"$type\", \"$tag\");}"
    }
  }
  close $file
  WriteFooter $dest
  close $dest
}

## Print Results
puts "   **** Saving Main Lexer ([file join .. Gazetteer.cpp])..."
set Lex  [open [file join .. Gazetteer.cpp] w]
fconfigure $Lex  -encoding iso8859-7
puts $Lex { /*
  * This is an autogenerated Lexer. 
  * Author: George Petasis, petasis@iit.demokritos.gr
  *       9/10/1999
  *  Modified at 10 April 2000
  *  Modified at 27 June  2008
  */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <CDM.h>

#ifdef __WIN32__
#include <io.h>
#define snprintf _snprintf
#define isatty   _isatty
#endif /* __WIN32__ */
}

foreach lexer [lsort -dictionary [array names FUNCTIONS]] {
  puts $Lex "extern int ${MODULE_NAME}Start_${lexer}(char *Text);"
  # puts $Lex "#include \"${lexer}.cpp\""
}

puts $Lex "
/*
 * This function will call all the created lexers in order...
 */
int ${MODULE_NAME}Startlex(char *sentence) \{"

#puts $Lex "  printf(\"Parsing sentence:\\n%s\\n\", sentence);"

foreach lexer [lsort -dictionary [array names FUNCTIONS]] {
#  puts $Lex "  printf(\"->>%s<<-\\n\", sentence);"
#  puts $Lex "  printf(\"Calling Lexer: ${MODULE_NAME}Start_${lexer}\\n\");"
  puts $Lex "  ${MODULE_NAME}Start_${lexer}(sentence);"
}

puts $Lex "  return 0;\n\} /* ${MODULE_NAME}Startlex */"
puts $Lex "\n/*\n * EOF\n */"
close $Lex

puts "   **** Creating Lexers/Makefile..."
set make [open [file join $DestDir Makefile] w]
puts -nonewline $make "
#############################################################################
###                                                                       ###
###   Makefile for Auto-Generated Lexers                                  ###
###                                                                       ###
#############################################################################

##
##  Modify the Following values according to your system...
##
CPP = g++
LEX = flex -8 -t
AR  = ar -cr
RM  = /bin/rm -rf

# Location of Tcl. Under TCL_DIR Directory, include and lib must exist...
TCL_DIR = \$(shell echo \"puts \[file dirname \[file dirname \[info library]]]\" \
                  | tclsh)
TCL_VERSION = \$(shell echo \"puts \[info tclversion]\" | tclsh)

# Needed Directories for tcl.h and CDM.h
INCLUDES = -I../../../../CDM -I/disk2b/local/include\
           -I\$(TCL_DIR)/include\
           -I\$(TCL_DIR)/include/tcl\${TCL_VERSION}

# Needed Directories for Tcl Library
LIBPATH = -L/disk2b/local/lib
LIBS = -ltcl -lm

COMPILE_OPTIONS = -c -O3 -fPIC -DYY_NEVER_INTERACTIVE -DUSE_TCL_STUBS -DCDM_USE_STUBS

OBJS = "
foreach file [lsort [array names FILES]] {
  puts -nonewline $make "$file.o "
}

puts $make "

##
## DO NOT modify below this level...
##
COMPILE = \$(CPP) \$(COMPILE_OPTIONS)
all: \$(OBJS)
	\$(AR) lexers.a \$(OBJS)
	\$(RM) *.cpp *.o
	@echo \"  ***** Lexers Made Successfully! ***\"
	@echo \"\"
"
foreach file [lsort [array names FILES]] lexer [lsort [array names FUNCTIONS]] {
  puts $make "$file.o:\n\t\$(LEX) -P${MODULE_NAME}${lexer} \$*.lex > \$*.cpp"
  puts $make "\t\$(COMPILE) \$(INCLUDES) \$*.cpp -o \$*.o\n"
}

puts $make "
## 
## End of Makefile
##"
close $make

set make [open [file join $DestDir Makefile.tcl] w]
puts $make \
{#############################################################################
###                                                                       ###
###   Tcl Makefile for Auto-Generated Lexers                              ###
###                                                                       ###
#############################################################################

cd [file dirname [file normalize [info script]]]
set LEX [lindex $argv 0]}

foreach file [lsort [array names FILES]] {
  puts $make "puts \"Generating lexer: $file.cpp...\""
  puts $make "exec \$LEX -PHGazetteer${file} -o$file.cpp $file.lex"
}

puts $make "
## 
## End of Makefile
##"
close $make
cd $CWD
