/*****************************************************************************
 *  htmlTags.c:                                                              *
 * ----------------------                                                    *
 *  This file extends the Tk HTML widget with support for tags, like the     *
 *  ones found in the html widget.                                           *
 *  This extension was designed and implemented by Georgios Petasis,         *
 *  e-mail: petasis@iit.demokritos.gr                                        *
 *  URL:    http://www.iit.demokritos.gr/~petasis                            *
 *                                                                           *
 *  Copyright 1998-2003 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Html Engineering Platform.            *
 *  For licencing information, please refer to the licence conditions of     *
 *  "Ellogon" Html Engineering Platform.                                     *
 *****************************************************************************/

#ifdef ENABLE_TAG_SUPPORT
#include <stdlib.h>
#include <string.h>
#include "tk.h"
#include "htmlTags.h"

#define BUTTON_DOWN		8
#ifdef __WIN32__
# define DEF_BUTTON_FG             "SystemButtonText"
#else
# define DEF_BUTTON_FG             "Black"
#endif

/*
 * Forward declarations for procedures defined later in this file:
 */

static void             TkHtmlAddTag _ANSI_ARGS_((HtmlWidget *htmlPtr,
                            TkHtmlTag *tagPtr,
                            HtmlIndex *index1, HtmlIndex *index2, int add));
static void		ChangeTagPriority _ANSI_ARGS_((HtmlWidget *htmlPtr,
			    TkHtmlTag *tagPtr, int prio));
static TkHtmlTag *	FindTag _ANSI_ARGS_((Tcl_Interp *interp,
			    HtmlWidget *htmlPtr, CONST char *tagName));
static void		SortTags _ANSI_ARGS_((int numTags,
			    TkHtmlTag **tagArrayPtr));
static int		TagSortProc _ANSI_ARGS_((CONST VOID *first,
			    CONST VOID *second));
static void             TkHtmlSortTagRanges _ANSI_ARGS_((HtmlWidget *htmlPtr));
static void             TkHtmlTagBindProc _ANSI_ARGS_((ClientData clientData,
                            XEvent *eventPtr));
void                    TkHtmlClearTags _ANSI_ARGS_((HtmlWidget *htmlPtr));
static int              HtmlIndexCmp _ANSI_ARGS_((HtmlIndex *index1,
                                                  HtmlIndex *index2));
static void             TkHtmlMoveIndex _ANSI_ARGS_((HtmlWidget *htmlPtr,
                            HtmlIndex *index, int offset, int forward));

static Tk_ConfigSpec tagConfigSpecs[] = {
    {TK_CONFIG_BORDER, "-background", (char *) NULL, (char *) NULL,
	(char *) NULL, Tk_Offset(TkHtmlTag, border), TK_CONFIG_NULL_OK},
    {TK_CONFIG_BITMAP, "-bgstipple", (char *) NULL, (char *) NULL,
	(char *) NULL, Tk_Offset(TkHtmlTag, bgStipple), TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-bd", (char *) NULL, (char *) NULL,
	"0", Tk_Offset(TkHtmlTag, bdString),
	TK_CONFIG_DONT_SET_DEFAULT|TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-borderwidth", (char *) NULL, (char *) NULL,
	"0", Tk_Offset(TkHtmlTag, bdString),
	TK_CONFIG_DONT_SET_DEFAULT|TK_CONFIG_NULL_OK},
    {TK_CONFIG_BITMAP, "-fgstipple", (char *) NULL, (char *) NULL,
	(char *) NULL, Tk_Offset(TkHtmlTag, fgStipple), TK_CONFIG_NULL_OK},
    {TK_CONFIG_COLOR, "-foreground", (char *) NULL, (char *) NULL,
	(char *) NULL, Tk_Offset(TkHtmlTag, fgColor), TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-overstrike", (char *) NULL, (char *) NULL,
	(char *) NULL, Tk_Offset(TkHtmlTag, overstrikeString),
	TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-relief", (char *) NULL, (char *) NULL,
	(char *) NULL, Tk_Offset(TkHtmlTag, reliefString), TK_CONFIG_NULL_OK},
    {TK_CONFIG_STRING, "-underline", (char *) NULL, (char *) NULL,
	(char *) NULL, Tk_Offset(TkHtmlTag, underlineString),
	TK_CONFIG_NULL_OK},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlTagsInit --
 *
 *	Init the data structures utilised by the tags engine. Must be
 *	called for each newly created HTML widget.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The tag-related fields of the htmlPtr structure are initialised.
 *
 *----------------------------------------------------------------------
 */

void
TkHtmlTagsInit(HtmlWidget *htmlPtr)
    /* HtmlWidget *htmlPtr;	 Widget to be initialised. */
{
  if (!htmlPtr) return;
  Tcl_InitHashTable(&htmlPtr->tagTable, TCL_STRING_KEYS);
  htmlPtr->tagRanges    = NULL;
  htmlPtr->currentRange = NULL;
  htmlPtr->lastRange    = NULL;
  htmlPtr->numTags      = 0;
  htmlPtr->numRanges    = 0;
  Tk_CreateEventHandler(htmlPtr->clipwin, KeyPressMask|KeyReleaseMask
	    |ButtonPressMask|ButtonReleaseMask|EnterWindowMask
	    |LeaveWindowMask|PointerMotionMask|VirtualEventMask,
	    TkHtmlTagBindProc, (ClientData) htmlPtr);
} /* TkHtmlTagsInit */

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlTagsFinilise --
 *
 *	Finilise the data structures utilised by the tags engine. Must be
 *	called before destroying an HTML widget.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Tag-related data structured are freed.
 *
 *----------------------------------------------------------------------
 */

void
TkHtmlTagsFinilise(HtmlWidget *htmlPtr)
    /* HtmlWidget *htmlPtr;	 Widget to be initialised. */
{
  if (!htmlPtr) return;
  TkHtmlClearTags(htmlPtr);
  Tcl_DeleteHashTable(&htmlPtr->tagTable);
  if (htmlPtr->bindingTable != NULL) {
    Tk_DeleteBindingTable(htmlPtr->bindingTable);
  }
} /* TkHtmlTagsFinilise */

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlCreateTag --
 *
 *	Find the record describing a tag within a given html widget,
 *	creating a new record if one doesn't already exist.
 *
 * Results:
 *	The return value is a pointer to the TkHtmlTag record for tagName.
 *
 * Side effects:
 *	A new tag record is created if there isn't one already defined
 *	for tagName.
 *
 *----------------------------------------------------------------------
 */

TkHtmlTag *
TkHtmlCreateTag(HtmlWidget *htmlPtr, CONST char *tagName)
    /* HtmlWidget *htmlPtr;	 Widget in which tag is being used. */
    /* CONST char *tagName;	 Name of desired tag. */
{
    register TkHtmlTag *tagPtr;
    Tcl_HashEntry *hPtr;
    int new;

    hPtr = Tcl_CreateHashEntry(&htmlPtr->tagTable, tagName, &new);
    if (!new) {
	return (TkHtmlTag *) Tcl_GetHashValue(hPtr);
    }

    /*
     * No existing entry.  Create a new one, initialize it, and add a
     * pointer to it to the hash table entry.
     */

    tagPtr = (TkHtmlTag *) Tcl_Alloc(sizeof(TkHtmlTag));
    memset(tagPtr, 0, sizeof(TkHtmlTag));
    tagPtr->name = Tcl_GetHashKey(&htmlPtr->tagTable, hPtr);
    tagPtr->priority = htmlPtr->numTags;
    tagPtr->border = NULL;
    tagPtr->bdString = NULL;
    tagPtr->borderWidth = 0;
    tagPtr->reliefString = NULL;
    tagPtr->relief = TK_RELIEF_FLAT;
    tagPtr->bgStipple = None;
    tagPtr->fgColor = NULL;
    tagPtr->tkfont = NULL;
    tagPtr->fgStipple = None;
    tagPtr->overstrikeString = NULL;
    tagPtr->overstrike = 0;
    tagPtr->underlineString = NULL;
    tagPtr->underline = 0;
    tagPtr->affectsDisplay = 0;
    htmlPtr->numTags++;
    Tcl_SetHashValue(hPtr, tagPtr);
    return tagPtr;
}

/*
 *--------------------------------------------------------------
 *
 * TkHtmlTagCmd --
 *
 *	This procedure is invoked to process the "tag" options of
 *	the widget command for html widgets. See the user documentation
 *	for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
TkHtmlTagCmd(HtmlWidget *htmlPtr, Tcl_Interp *interp,
                int argc, CONST char **argv)
/*  register HtmlWidget *htmlPtr;  Information about html widget. */
/*  Tcl_Interp *interp;	           Current interpreter. */
/*  int argc;			   Number of arguments. */
/*  CONST char **argv;		   Argument strings.  Someone else has already
				 * parsed this command enough to know that
				 * argv[1] is "tag". */
{
    int c, i, addTag;
    size_t length;
    char *fullOption;
    register TkHtmlTag *tagPtr;
    HtmlIndex index1, index2;

    Tcl_ResetResult(interp);
    if (argc < 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " tag option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    c = argv[2][0];
    length = strlen(argv[2]);
    if ((c == 'a') && (strncmp(argv[2], "add", length) == 0)) {
	fullOption = "add";
	addTag = 1;

	addAndRemove:
	if (argc < 5) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " tag ", fullOption,
		    " tagName index1 ?index2 index1 index2 ...?\"",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	tagPtr = TkHtmlCreateTag(htmlPtr, argv[3]);
	for (i = 4; i < argc; i += 2) {
            if (HtmlGetIndex(htmlPtr, argv[i],
                                      &index1.p, &index1.i) != 0) {
              Tcl_AppendResult(interp, "malformed index: \"", argv[i], "\"", 0);
	      return TCL_ERROR;
	    }
	    if (argc > (i+1)) {
		if (HtmlGetIndex(htmlPtr, argv[i+1],
                                          &index2.p, &index2.i) !=0) {
                    Tcl_AppendResult(interp, "malformed index: \"", argv[i],
                                             "\"", 0);
		    return TCL_ERROR;
		}
		if (HtmlIndexCmp(&index1, &index2) >= 0) {
		    return TCL_OK;
		}
                TkHtmlMoveIndex(htmlPtr, &index2, 1, 0);
	    } else {
                index2.p = index1.p;
                index2.i = index1.i;
	    }
            /* Add the tag to the HtmlElements */
            /* printf("TkHtmlAddTag: %s -> %d.%d, %s -> %d.%d\n",
                                     argv[i], index1.p->base.id, index1.i,
                                     argv[i+1], index2.p->base.id, index2.i); */
            TkHtmlAddTag(htmlPtr, tagPtr, &index1, &index2, addTag);
	}
    } else if ((c == 'b') && (strncmp(argv[2], "bind", length) == 0)) {
	if ((argc < 4) || (argc > 6)) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " tag bind tagName ?sequence? ?command?\"",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	tagPtr = TkHtmlCreateTag(htmlPtr, argv[3]);

	/*
	 * Make a binding table if the widget doesn't already have
	 * one.
	 */

	if (htmlPtr->bindingTable == NULL) {
	    htmlPtr->bindingTable = Tk_CreateBindingTable(interp);
	}

	if (argc == 6) {
	    int append = 0;
	    unsigned long mask;

	    if (argv[5][0] == 0) {
		return Tk_DeleteBinding(interp, htmlPtr->bindingTable,
			(ClientData) tagPtr, (char *) argv[4]);
	    }
	    if (argv[5][0] == '+') {
		argv[5]++;
		append = 1;
	    }
	    mask = Tk_CreateBinding(interp, htmlPtr->bindingTable,
		    (ClientData) tagPtr, (char *) argv[4], (char *) argv[5],
                    append);
	    if (mask == 0) {
		return TCL_ERROR;
	    }
	    if (mask & (unsigned) ~(ButtonMotionMask|Button1MotionMask
		    |Button2MotionMask|Button3MotionMask|Button4MotionMask
		    |Button5MotionMask|ButtonPressMask|ButtonReleaseMask
		    |EnterWindowMask|LeaveWindowMask|KeyPressMask
		    |KeyReleaseMask|PointerMotionMask|VirtualEventMask)) {
		Tk_DeleteBinding(interp, htmlPtr->bindingTable,
			(ClientData) tagPtr, (char *) argv[4]);
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "requested illegal events; ",
			"only key, button, motion, enter, leave, and virtual ",
			"events may be used", (char *) NULL);
		return TCL_ERROR;
	    }
	} else if (argc == 5) {
	    CONST char *command;
    
	    command = Tk_GetBinding(interp, htmlPtr->bindingTable,
		    (ClientData) tagPtr, (char *) argv[4]);
	    if (command == NULL) {
		CONST char *string = Tcl_GetStringResult(interp); 

		/*
		 * Ignore missing binding errors.  This is a special hack
		 * that relies on the error message returned by FindSequence
		 * in tkBind.c.
		 */

		if (string[0] != '\0') {
		    return TCL_ERROR;
		} else {
		    Tcl_ResetResult(interp);
		}
	    } else {
		Tcl_SetResult(interp, (char *) command, TCL_STATIC);
	    }
	} else {
	    Tk_GetAllBindings(interp, htmlPtr->bindingTable,
		    (ClientData) tagPtr);
	}
    } else if ((c == 'c') && (strncmp(argv[2], "cget", length) == 0)
	    && (length >= 2)) {
	if (argc != 5) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " tag cget tagName option\"",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	tagPtr = FindTag(interp, htmlPtr, argv[3]);
	if (tagPtr == NULL) {
	    return TCL_ERROR;
	}
	return Tk_ConfigureValue(interp, htmlPtr->tkwin, tagConfigSpecs,
		(char *) tagPtr, (char *) argv[4], 0);
    } else if ((c == 'c') && (strncmp(argv[2], "configure", length) == 0)
	    && (length >= 2)) {
	if (argc < 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " tag configure tagName ?option? ?value? ",
		    "?option value ...?\"", (char *) NULL);
	    return TCL_ERROR;
	}
	tagPtr = TkHtmlCreateTag(htmlPtr, argv[3]);
	if (argc == 4) {
	    return Tk_ConfigureInfo(interp, htmlPtr->tkwin, tagConfigSpecs,
		    (char *) tagPtr, (char *) NULL, 0);
	} else if (argc == 5) {
	    return Tk_ConfigureInfo(interp, htmlPtr->tkwin, tagConfigSpecs,
		    (char *) tagPtr, (char *) argv[4], 0);
	} else {
	    int result;

	    result = Tk_ConfigureWidget(interp, htmlPtr->tkwin, tagConfigSpecs,
		    argc-4,  argv+4, (char *) tagPtr, 0);
	    /*
	     * Some of the configuration options, like -underline
	     * and -justify, require additional translation (this is
	     * needed because we need to distinguish a particular value
	     * of an option from "unspecified").
	     */

	    if (tagPtr->bdString != NULL) {
		if (Tk_GetPixels(interp, htmlPtr->tkwin, tagPtr->bdString,
			&tagPtr->borderWidth) != TCL_OK) {
		    return TCL_ERROR;
		}
		if (tagPtr->borderWidth < 0) {
		    tagPtr->borderWidth = 0;
		}
	    }
	    if (tagPtr->reliefString != NULL) {
		if (Tk_GetRelief(interp, tagPtr->reliefString,
			&tagPtr->relief) != TCL_OK) {
		    return TCL_ERROR;
		}
	    }
	    if (tagPtr->overstrikeString != NULL) {
		if (Tcl_GetBoolean(interp, tagPtr->overstrikeString,
			&tagPtr->overstrike) != TCL_OK) {
		    return TCL_ERROR;
		}
	    }
	    if (tagPtr->underlineString != NULL) {
		if (Tcl_GetBoolean(interp, tagPtr->underlineString,
			&tagPtr->underline) != TCL_OK) {
		    return TCL_ERROR;
		}
	    }

	    tagPtr->affectsDisplay = 0;
	    if ((tagPtr->border != NULL)
		    || (tagPtr->bdString != NULL)
		    || (tagPtr->reliefString != NULL)
		    || (tagPtr->bgStipple != None)
		    || (tagPtr->fgColor != NULL)
		    || (tagPtr->fgStipple != None)
		    || (tagPtr->overstrikeString != NULL)
		    || (tagPtr->underlineString != NULL)) {
		tagPtr->affectsDisplay = 1;
	    }
            /* Redraw all text in the clipping window */
            TkHtmlScheduleTagsRedraw(htmlPtr);
	    return result;
	}
    } else if ((c == 'd') && (strncmp(argv[2], "delete", length) == 0)) {
	Tcl_HashEntry *hPtr;

	if (argc < 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " tag delete tagName tagName ...\"",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	for (i = 3; i < argc; i++) {
	    hPtr = Tcl_FindHashEntry(&htmlPtr->tagTable, argv[i]);
	    if (hPtr == NULL) {
		continue;
	    }
	    tagPtr = (TkHtmlTag *) Tcl_GetHashValue(hPtr);
	    Tcl_DeleteHashEntry(hPtr);
	    if (htmlPtr->bindingTable != NULL) {
		Tk_DeleteAllBindings(htmlPtr->bindingTable,
			(ClientData) tagPtr);
	    }
	
	    /*
	     * Update the tag priorities to reflect the deletion of this tag.
	     */

	    ChangeTagPriority(htmlPtr, tagPtr, htmlPtr->numTags-1);
	    htmlPtr->numTags -= 1;
	    TkHtmlFreeTag(htmlPtr, tagPtr);
            /* Redraw all text in the clipping window */
            TkHtmlScheduleTagsRedraw(htmlPtr);
	}
    } else if ((c == 'l') && (strncmp(argv[2], "lower", length) == 0)) {
	TkHtmlTag *tagPtr2;
	int prio;

	if ((argc != 4) && (argc != 5)) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " tag lower tagName ?belowThis?\"",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	tagPtr = FindTag(interp, htmlPtr, argv[3]);
	if (tagPtr == NULL) {
	    return TCL_ERROR;
	}
	if (argc == 5) {
	    tagPtr2 = FindTag(interp, htmlPtr, argv[4]);
	    if (tagPtr2 == NULL) {
		return TCL_ERROR;
	    }
	    if (tagPtr->priority < tagPtr2->priority) {
		prio = tagPtr2->priority - 1;
	    } else {
		prio = tagPtr2->priority;
	    }
	} else {
	    prio = 0;
	}
	ChangeTagPriority(htmlPtr, tagPtr, prio);
        /* Redraw all text in the clipping window */
        TkHtmlScheduleTagsRedraw(htmlPtr);
    } else if ((c == 'n') && (strncmp(argv[2], "names", length) == 0)
            && (length >= 2)) {
        TkHtmlTag **arrayPtr;
        int arraySize;

        if ((argc != 3) && (argc != 4)) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                    argv[0], " tag names ?index?\"",
                    (char *) NULL);
            return TCL_ERROR;
        }
        arrayPtr = (TkHtmlTag **) Tcl_Alloc((unsigned)
                (htmlPtr->numTags * sizeof(TkHtmlTag *)));
        if (argc == 3) {
            Tcl_HashSearch search;
            Tcl_HashEntry *hPtr;

            for (i = 0, hPtr = Tcl_FirstHashEntry(&htmlPtr->tagTable, &search);
                    hPtr != NULL; i++, hPtr = Tcl_NextHashEntry(&search)) {
                arrayPtr[i] = (TkHtmlTag *) Tcl_GetHashValue(hPtr);
            }
            arraySize = htmlPtr->numTags;
        } else {
            TkHtmlTagRange *pList;
            int offset, found;
            if (HtmlGetIndex(htmlPtr, argv[3], &index1.p, &index1.i) != 0) {
              Tcl_Free((char *) arrayPtr);
              return TCL_ERROR;
            }
            arraySize = 0;
            /* Get all tags that cover the given index... */
            offset = index1.p->base.offs + index1.i;
            /* printf("Tag range index: %s -> %d\n", argv[3], offset); */
            for (pList = htmlPtr->tagRanges; pList; pList = pList->pNext) {
              if (((pList->startElement->base.offs + pList->charStart)
                                                                 <= offset) &&
                  ((pList->endElement->base.offs +
                                   pList->charEnd_plus_one_char) >= offset)) {
                /* If the tag is already in the array, don't put it again */
                found = 0;
                for (i = 0; i <arraySize; i++) {
                  if (arrayPtr[i] == pList->tag) {
                    found = 1;
                    break;
                  }
                }
                if (!found) {
                  /* printf("Adding tag %s (%d)\n",
                            pList->tag->name, arraySize); */
                  arrayPtr[arraySize++] = pList->tag;
                }
              }
            }
        }
        if (arraySize) {
          SortTags(arraySize, arrayPtr);
          for (i = 0; i < arraySize; i++) {
              tagPtr = arrayPtr[i];
              Tcl_AppendElement(interp, tagPtr->name);
          }
        }
        Tcl_Free((char *) arrayPtr);
    } else if ((c == 'n') && (strncmp(argv[2], "nextrange", length) == 0)
	    && (length >= 2)) {
#if 0
	TkHtmlSearch tSearch;
	char position[TK_POS_CHARS];

	if ((argc != 5) && (argc != 6)) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " tag nextrange tagName index1 ?index2?\"",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	tagPtr = FindTag((Tcl_Interp *) NULL, htmlPtr, argv[3]);
	if (tagPtr == NULL) {
	    return TCL_OK;
	}
	if (HtmlGetIndex(htmlPtr, argv[4], &index1.p, &index1.i) != 0) {
	    return TCL_ERROR;
	}
	TkHtmlMakeByteIndex(htmlPtr->tree, TkBTreeNumLines(htmlPtr->tree),
		0, &last);
	if (argc == 5) {
	    index2 = last;
	} else if (HtmlGetIndex(htmlPtr, argv[5], &index2.p, &index2.i)
		!= 0) {
	    return TCL_ERROR;
	}

	/*
	 * The search below is a bit tricky.  Rather than use the B-tree
	 * facilities to stop the search at index2, let it search up
	 * until the end of the file but check for a position past index2
	 * ourselves.  The reason for doing it this way is that we only
	 * care whether the *start* of the range is before index2;  once
	 * we find the start, we don't want TkBTreeNextTag to abort the
	 * search because the end of the range is after index2.
	 */

	TkBTreeStartSearch(&index1, &last, tagPtr, &tSearch);
	if (TkBTreeCharTagged(&index1, tagPtr)) {
	    TkHtmlSegment *segPtr;
	    int offset;

	    /*
	     * The first character is tagged.  See if there is an
	     * on-toggle just before the character.  If not, then
	     * skip to the end of this tagged range.
	     */

	    for (segPtr = index1.linePtr->segPtr, offset = index1.byteIndex; 
		    offset >= 0;
		    offset -= segPtr->size, segPtr = segPtr->nextPtr) {
		if ((offset == 0) && (segPtr->typePtr == &tkHtmlToggleOnType)
			&& (segPtr->body.toggle.tagPtr == tagPtr)) {
		    goto gotStart;
		}
	    }
	    if (!TkBTreeNextTag(&tSearch)) {
		 return TCL_OK;
	    }
	}

	/*
	 * Find the start of the tagged range.
	 */

	if (!TkBTreeNextTag(&tSearch)) {
	    return TCL_OK;
	}
	gotStart:
	if (HtmlIndexCmp(&tSearch.curIndex, &index2) >= 0) {
	    return TCL_OK;
	}
	TkHtmlPrintIndex(&tSearch.curIndex, position);
	Tcl_AppendElement(interp, position);
	TkBTreeNextTag(&tSearch);
	TkHtmlPrintIndex(&tSearch.curIndex, position);
	Tcl_AppendElement(interp, position);
#endif
    } else if ((c == 'p') && (strncmp(argv[2], "prevrange", length) == 0)
	    && (length >= 2)) {
#if 0
	TkHtmlSearch tSearch;
	char position1[TK_POS_CHARS];
	char position2[TK_POS_CHARS];

	if ((argc != 5) && (argc != 6)) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " tag prevrange tagName index1 ?index2?\"",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	tagPtr = FindTag((Tcl_Interp *) NULL, htmlPtr, argv[3]);
	if (tagPtr == NULL) {
	    return TCL_OK;
	}
	if (HtmlGetIndex(htmlPtr, argv[4], &index1.p, &index1.i) != 0) {
	    return TCL_ERROR;
	}
	if (argc == 5) {
	    TkHtmlMakeByteIndex(htmlPtr->tree, 0, 0, &index2);
	} else if (HtmlGetIndex(htmlPtr, argv[5], &index2.p, &index2.i)
		!= 0) {
	    return TCL_ERROR;
	}

	/*
	 * The search below is a bit weird.  The previous toggle can be
	 * either an on or off toggle. If it is an on toggle, then we
	 * need to turn around and search forward for the end toggle.
	 * Otherwise we keep searching backwards.
	 */

	TkBTreeStartSearchBack(&index1, &index2, tagPtr, &tSearch);

	if (!TkBTreePrevTag(&tSearch)) {
	    return TCL_OK;
	}
	if (tSearch.segPtr->typePtr == &tkHtmlToggleOnType) {
	    TkHtmlPrintIndex(&tSearch.curIndex, position1);
	    TkHtmlMakeByteIndex(htmlPtr->tree, TkBTreeNumLines(htmlPtr->tree),
		    0, &last);
	    TkBTreeStartSearch(&tSearch.curIndex, &last, tagPtr, &tSearch);
	    TkBTreeNextTag(&tSearch);
	    TkHtmlPrintIndex(&tSearch.curIndex, position2);
	} else {
	    TkHtmlPrintIndex(&tSearch.curIndex, position2);
	    TkBTreePrevTag(&tSearch);
	    if (HtmlIndexCmp(&tSearch.curIndex, &index2) < 0) {
		return TCL_OK;
	    }
	    TkHtmlPrintIndex(&tSearch.curIndex, position1);
	}
	Tcl_AppendElement(interp, position1);
	Tcl_AppendElement(interp, position2);
#endif
    } else if ((c == 'r') && (strncmp(argv[2], "raise", length) == 0)
	    && (length >= 3)) {
	TkHtmlTag *tagPtr2;
	int prio;

	if ((argc != 4) && (argc != 5)) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " tag raise tagName ?aboveThis?\"",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	tagPtr = FindTag(interp, htmlPtr, argv[3]);
	if (tagPtr == NULL) {
	    return TCL_ERROR;
	}
	if (argc == 5) {
	    tagPtr2 = FindTag(interp, htmlPtr, argv[4]);
	    if (tagPtr2 == NULL) {
		return TCL_ERROR;
	    }
	    if (tagPtr->priority <= tagPtr2->priority) {
		prio = tagPtr2->priority;
	    } else {
		prio = tagPtr2->priority + 1;
	    }
	} else {
	    prio = htmlPtr->numTags-1;
	}
	ChangeTagPriority(htmlPtr, tagPtr, prio);
        TkHtmlScheduleTagsRedraw(htmlPtr);
	//TkHtmlRedrawTag(htmlPtr, (HtmlIndex *) NULL, (HtmlIndex *) NULL,
	//	tagPtr, 1);
    } else if ((c == 'r') && (strncmp(argv[2], "ranges", length) == 0)
	    && (length >= 3)) {
        TkHtmlTagRange *pRange;
        char tmp[128];

	if (argc != 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " tag ranges tagName\"", (char *) NULL);
	    return TCL_ERROR;
	}
	tagPtr = FindTag((Tcl_Interp *) NULL, htmlPtr, argv[3]);
	if (tagPtr == NULL) {
	    return TCL_OK;
	}
        for (pRange = htmlPtr->tagRanges; pRange; pRange = pRange->pNext) {
          if (pRange->tag == tagPtr) {
            sprintf(tmp, "%d.%d %d.%d",
                    HtmlTokenNumber(pRange->startElement), pRange->charStart,
                    HtmlTokenNumber(pRange->endElement),
                                              pRange->charEnd_plus_one_char);
            Tcl_AppendElement(interp, tmp);
          }
        }
    } else if ((c == 'r') && (strncmp(argv[2], "remove", length) == 0)
	    && (length >= 2)) {
	fullOption = "remove";
	addTag = 0;
	goto addAndRemove;
    } else {
	Tcl_AppendResult(interp, "bad tag option \"", argv[2],
		"\": must be add, bind, cget, configure, delete, lower, ",
		"names, nextrange, raise, ranges, or remove",
		(char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlAddTag --
 *
 *	Annotates all HtmlElements within the two provided indexes with
 *	the specified tag.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The element->base.tags pointer will be modified.
 *
 *----------------------------------------------------------------------
 */
static void
TkHtmlAddTag(HtmlWidget *htmlPtr, TkHtmlTag *tagPtr,
             HtmlIndex *start, HtmlIndex *end, int add)
/*  HtmlWidget *htmlPtr;	Widget in which tag is being added. */
/*  TkHtmlTag *tagPtr;  	Pointer to tag information. */
/*  HtmlIndex *start;           The start index of the tag */
/*  HtmlIndex *end;             The end index of the tag */
{
  register TkHtmlTagRange *tagItemPtr;
  if (start == NULL || end == NULL) return;
  /*
  char *tmp[512];
  memset(tmp, 0 ,512);
  memcpy(tmp, &htmlPtr->zText[start->p->base.offs+start->i],
         end->p->base.offs+end->i - start->p->base.offs-start->i+1);
  printf("TkHtmlAddTag: %d->%d \"%s\"\n",
         start->p->base.offs+start->i, end->p->base.offs+end->i, tmp); */
  if (add) {
    Tcl_UniChar ch;
    /* Allocate the structure that holds this tag range */
    tagItemPtr = (TkHtmlTagRange *) Tcl_Alloc(sizeof(TkHtmlTagRange));
    if (tagItemPtr == NULL) return;
    memset(tagItemPtr, 0, sizeof(TkHtmlTagRange));
    tagItemPtr->tag = tagPtr;
    tagItemPtr->startElement = start->p;
    tagItemPtr->endElement   = end->p;
    tagItemPtr->charStart    = start->i;
    tagItemPtr->charEnd      = end->i;
    tagItemPtr->charEnd_plus_one_char = tagItemPtr->charEnd + 
              Tcl_UtfToUniChar(&htmlPtr->zText[end->p->base.offs+end->i], &ch);
    if (htmlPtr->tagRanges == NULL) {
      htmlPtr->tagRanges = tagItemPtr;
    } else {
      htmlPtr->lastRange->pNext = tagItemPtr;
      tagItemPtr->pPrev = htmlPtr->lastRange;
    }
    htmlPtr->lastRange = tagItemPtr;
    htmlPtr->numRanges++;
  } else {
    // TODO: remove the range

  }
} /* TkHtmlAddTag */

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlGetNextTaggedElement --
 *
 *	Return the first range found tagged with the specified tag.
 *
 * Results:
 *	The range pointer if an element has been found, NULL otherwise.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */
static TkHtmlTagRange *
TkHtmlGetNextTaggedElement(HtmlWidget *htmlPtr, char *tagName, int first) {
  register TkHtmlTagRange *pList;
  TkHtmlTagRange *result = NULL;
  int offset;

  if (first) {
    offset = htmlPtr->nText + 1;
    for (pList = htmlPtr->tagRanges; pList; pList = pList->pNext) {
      if (strcmp(tagName, pList->tag->name) == 0) {
        if (pList->startElement->base.offs < offset) {
          offset = pList->startElement->base.offs;
          result = pList;
        }
      }
    }
  } else {
    offset = 0;
    for (pList = htmlPtr->lastRange; pList; pList = pList->pPrev) {
      if (strcmp(tagName, pList->tag->name) == 0) {
        if (pList->startElement->base.offs > offset) {
          offset = pList->startElement->base.offs;
          result = pList;
        }
      }
    }
  }
  return result;
}; /* TkHtmlGetNextTaggedElement */

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlDecodeIndex --
 *
 *	Convert a tag index to an Html index.
 *
 * Results:
 *	1 if the index was found, 0 otherwise.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */
int
TkHtmlDecodeIndex(HtmlWidget *htmlPtr, CONST char *index,
                  HtmlElement **ppElement, int *pIndex) {
  char *tagName, *period = strchr(index, '.');
  TkHtmlTagRange *pList = NULL;
  
  if (period == NULL) return 0;
  tagName = Tcl_Alloc(strlen(index)+1);
  strcpy(tagName, index);
  tagName[period-index] = '\0';
  *ppElement = NULL;
  *pIndex = -1;
  if (strcmp(period, ".first") == 0) {
    /* Return the first tagged character with the tag... */
    pList = TkHtmlGetNextTaggedElement(htmlPtr, tagName, 1);
    if (pList) {
      *ppElement = pList->startElement;
      *pIndex    = pList->charStart;
    }
  } else if (strcmp(period, ".last") == 0) {
    /* Return the last tagged character with the tag + 1... */
    pList = TkHtmlGetNextTaggedElement(htmlPtr, tagName, 0);
    if (pList) {
      *ppElement = pList->endElement;
      *pIndex    = pList->charEnd_plus_one_char;
    }
  } else if (strcmp(period, ".end") == 0) {
    /* Return the last tagged character with the tag... */
    pList = TkHtmlGetNextTaggedElement(htmlPtr, tagName, 0);
    if (pList) {
      *ppElement = pList->endElement;
      *pIndex    = pList->charEnd;
    }
  }
  Tcl_Free(tagName);
  if (pList) return 1;
  return 0;
} /* TkHtmlDecodeIndex */

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlDrawTagRange --
 *
 *	Draws a tag range on a drawable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Drawing does occur.
 *
 *----------------------------------------------------------------------
 */
static void
TkHtmlDrawTagRange(HtmlWidget *htmlPtr, TkHtmlTagRange *pRange,
                        Drawable drawable, int left, int top)
/*  HtmlWidget *htmlPtr;	Widget in which tag is being used. */
/*  TkHtmlTagRange *pRange;     The tag range to be drawn */
/*  Drawable drawable;          Draw on this drawable */
/*  int left; int top;          Virtual coords of top-left of drawable */
{
  register HtmlElement *pElement;
  GC gc;
  XGCValues gcValues;
  Tk_Font elementFont;
  int x, y, width, height, ascent, descent, len;
  XRectangle xrec;
  char *text;
  
  if (pRange == NULL) return;
  pRange->xrec.height = 1;
  /*
   * printf("--TkHtmlDrawTagRange: start=%p, end=%p\n",
   *                               pRange->startElement, pRange->endElement);
   * printf("\tstartOff=%d, endOff=%d -> %d\n",
   *            pRange->startElement->base.offs, pRange->endElement->base.offs,
   *            pRange->endElement->base.offs-pRange->startElement->base.offs);
   */
  /* Iterate over all elements within the range and draw them... */
  for (pElement = pRange->startElement; pElement; pElement = pElement->pNext) {
     /* printf("TkHtmlDrawTagRange: tag=%s, element=%p\n",
      *         pRange->tag->name, pElement); */
    /* If the element is not visible, don't bother drawing it */
    if (((pElement->base.flags & HTML_Visible) != 0) &&
        (pElement->base.type == Html_Text)) {
      /* printf("TkHtmlDrawTagRange: \"%s\"\n", pElement->text.zText);
       * printf("TkHtmlDrawTagRange: charStart=%d, charEnd=%d, "
       *        "charEnd_plus_one_char=%d\n",
       *     pRange->charStart, pRange->charEnd, pRange->charEnd_plus_one_char);
       */
      ascent      = pElement->text.ascent;
      descent     = pElement->text.descent;
      x           = pElement->text.x - left;
      y           = pElement->text.y - top;
      width       = pElement->text.w;
      height      = pElement->text.ascent + pElement->text.descent;
      elementFont = HtmlGetFont(htmlPtr, pElement->base.style.font);
      text        = pElement->text.zText;
      len         = strlen(text);
      if ((pElement == pRange->startElement) && (pRange->charStart > 0)) {
        x += Tk_TextWidth(elementFont, text, pRange->charStart);
        text = &text[pRange->charStart];
        len -= pRange->charStart - 1;
      }
      /* If text has been consumed, there is nothing to draw. Ommit the tag */
      if (len >= 0 && *text != '\0') {
        if ((pElement == pRange->endElement) && (pRange->charEnd != -1)) {
          len = pRange->charEnd_plus_one_char;
          /* If the same element is also the tag start, we should exclude any
           * possible start... */
          if ((pElement == pRange->startElement) && (pRange->charStart > 0)) {
            len -= pRange->charStart;
          }
          width = Tk_TextWidth(elementFont, text, len);
        }
        xrec.x = x - pRange->tag->borderWidth;
        xrec.y = y - pRange->tag->borderWidth - ascent;
        xrec.width  = width  + 2 * pRange->tag->borderWidth;
        xrec.height = height + 2 * pRange->tag->borderWidth;
        if (pRange->xrec.height < xrec.height) {
          /* This is the first time we draw this tag range... */
          pRange->xrec.x      = xrec.x; 
          pRange->xrec.y      = xrec.y; 
          pRange->xrec.width  = xrec.width;
          pRange->xrec.height = xrec.height;
        } else {
          /* Enlarge the rect if necessary... */
          if (xrec.x < pRange->xrec.x) pRange->xrec.x = xrec.x;
          if (xrec.y < pRange->xrec.y) pRange->xrec.y = xrec.y;
          if (xrec.width >pRange->xrec.width)  pRange->xrec.width =xrec.width;
          if (xrec.height>pRange->xrec.height) pRange->xrec.height=xrec.height;
        }
        /* Draw the background */
        //XFillRectangle(htmlPtr->display, drawable, xrec.x, xrec.y,
        //                                         xrec.width, xrec.height);
        if (pRange->tag->border) {
          /* Draw the tag... */
          Tk_Fill3DRectangle(htmlPtr->tkwin, drawable, pRange->tag->border,
                             xrec.x, xrec.y, xrec.width, xrec.height,
                             pRange->tag->borderWidth, pRange->tag->relief);
        }
        /* Draw the text */
        if (pRange->tag->fgColor) {
          gcValues.foreground = pRange->tag->fgColor->pixel;
        } else {
          gcValues.foreground = 
                htmlPtr->apColor[pRange->startElement->base.style.color]->pixel;
        }
        gcValues.graphics_exposures = True;
        gcValues.font               = Tk_FontId(elementFont);
        gc = Tk_GetGC(htmlPtr->tkwin, GCForeground | GCGraphicsExposures |
                                      GCFont, &gcValues);
        // TODO: Fix
        if (len > strlen(text)) len = strlen(text);
        Tk_DrawChars(htmlPtr->display, drawable, gc,
                                       elementFont, text, len, x, y);
        if (pRange->tag->underline) {
          Tk_UnderlineChars(htmlPtr->display, drawable, gc, elementFont,
                   text, x, y, 0, len);
        }
        if (pRange->tag->overstrike) {
          XRectangle xrec;
          xrec.x = x;
          xrec.y = y - ascent + width / 2;
          xrec.width = width;
          xrec.height = 1 + (width > 15);
          XFillRectangles(htmlPtr->display, drawable, gc, &xrec, 1);
        }
        Tk_FreeGC(htmlPtr->display, gc);
      } /* if (*text != '\0') */
    } /* if (((pElement->base.flags & HTML_Visible) != 0) &&
              (pElement->base.type == Html_Text)) */
    if (pElement == pRange->endElement) break;
  }
  pRange->xrec.x += left;
  pRange->xrec.y += top;
   /* Finally, draw the border */
   //if (pRange->tag->border) {
   //  Tk_Draw3DRectangle(htmlPtr->tkwin, drawable, pRange->tag->border,
   //                     pRange->xrec.x, pRange->xrec.y,
   //                     pRange->xrec.width, pRange->xrec.height,
   //                     pRange->tag->borderWidth, pRange->tag->relief);
   //}
} /* TkHtmlDrawTagRange */

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlDrawTaggedBlock --
 *
 *	Draws an Html block (that contains tagged elements) on a drawable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Drawing does occur.
 *
 *----------------------------------------------------------------------
 */
void
TkHtmlDrawTaggedBlock(HtmlWidget *htmlPtr, HtmlBlock *blockPtr,
                        Drawable drawable, int x, int y)
/*  HtmlWidget *htmlPtr;	Widget in which tag is being used. */
/*  HtmlBlock *blockPtr;        The block to be drawn */
/*  Drawable drawable;          Draw on this drawable */
/*  int x; int y;               Virtual coords of top-left of drawable */
{
  register TkHtmlTagRange *pRange;
  int offsetStart = 0, offsetEnd = htmlPtr->nText;
  if (!blockPtr) return;

  offsetStart = blockPtr->base.pNext->base.offs;
  /*
   * We have been called to draw a block. This means that we have to draw all
   * elements that are within the block. But since we don't know where the
   * block ends, we draw all elements untill the first element contained in
   * the next block...
   */
  if (blockPtr->pNext) {
    offsetEnd = blockPtr->pNext->base.pNext->base.offs - 1;
  }
  /* printf("TkHtmlDrawTaggedBlock: %d - %d (%d)\n",
                    offsetStart, offsetEnd, offsetEnd - offsetStart); */
  for (pRange = htmlPtr->tagRanges; pRange; pRange = pRange->pNext) {
    if ((pRange->startElement->base.offs >= offsetStart) &&
        (pRange->endElement->base.offs <= offsetEnd)) {
      /* printf("  Tag: %s\n", pRange->tag->name); */
      TkHtmlDrawTagRange(htmlPtr, pRange, drawable, x, y);
    }
  }
} /* TkHtmlDrawTaggedBlock */

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlDrawTags --
 *
 *	Draws all tags on a drawable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Drawing does occur.
 *
 *----------------------------------------------------------------------
 */
void
TkHtmlDrawTags(HtmlWidget *htmlPtr, Drawable drawable, int x, int y,
                                                       int w, int h)
/*  HtmlWidget *htmlPtr;	Widget in which tag is being used. */
/*  Drawable drawable;          Draw on this drawable */
/*  int x; int y;               Virtual coords of top-left of drawable */
{
  register TkHtmlTagRange *pRange;
  /* printf("TkHtmlDrawTags: x=%d, y=%d, w=%d, h=%d\n", x, y, w, h); */
  for (pRange = htmlPtr->tagRanges; pRange; pRange = pRange->pNext) {
    if (pRange->xrec.height) {
      if ((pRange->xrec.y >= y - 30) && (pRange->xrec.y <= y + h + 30)) {
        /* printf("  Tag: %s\n", pRange->tag->name); */
        TkHtmlDrawTagRange(htmlPtr, pRange, drawable, x, y);
      }
    } else if (pRange->tag->affectsDisplay) {
      /* printf(" +Tag: %s\n", pRange->tag->name); */
      TkHtmlDrawTagRange(htmlPtr, pRange, drawable, x, y);
    }
    /* TkHtmlDrawTagRange(htmlPtr, pRange, drawable, x, y); */
  }
} /* TkHtmlDrawTags */

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlScheduleTagsRedraw --
 *
 *	Schedules a redraw for all visible tags.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Drawing does occur.
 *
 *----------------------------------------------------------------------
 */
void
TkHtmlScheduleTagsRedraw(HtmlWidget *htmlPtr)
/*  HtmlWidget *htmlPtr;	Widget to redtaw. */
{
  htmlPtr->flags |= REDRAW_TEXT;
  //HtmlRefresh(htmlPtr,0);
  HtmlScheduleRedraw(htmlPtr);
}; /* TkHtmlScheduleTagsRedraw */

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlClearTags --
 *
 *	Remove all tag related information.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	All allocated structures for the tags will be freed.
 *
 *----------------------------------------------------------------------
 */
void
TkHtmlClearTags(HtmlWidget *htmlPtr)
/*  HtmlWidget *htmlPtr;	Widget to clear. */
{
  register TkHtmlTagRange *pList, *pNext;
  Tcl_HashSearch search;
  Tcl_HashEntry *hPtr;
  register TkHtmlTag *tagPtr;

  if (!htmlPtr) return;
  for (pList = htmlPtr->tagRanges; pList; pList = pNext) {
    pNext = pList->pNext;
    Tcl_Free((char *) pList);
  }
  htmlPtr->tagRanges    = NULL;
  htmlPtr->currentRange = NULL;
  htmlPtr->lastRange    = NULL;
  htmlPtr->numTags      = 0;
  htmlPtr->numRanges    = 0;
  htmlPtr->tagFlags     = 0;
  
  for (hPtr = Tcl_FirstHashEntry(&htmlPtr->tagTable, &search);
       hPtr != NULL; hPtr = Tcl_NextHashEntry(&search)) {
    tagPtr = (TkHtmlTag *) Tcl_GetHashValue(hPtr);
    Tcl_Free((char *) tagPtr);
  }
  Tcl_DeleteHashTable(&htmlPtr->tagTable);
  Tcl_InitHashTable(&htmlPtr->tagTable, TCL_STRING_KEYS);
}; /* TkHtmlClearTags */

/*
 *----------------------------------------------------------------------
 *
 * FindTag --
 *
 *	See if tag is defined for a given widget.
 *
 * Results:
 *	If tagName is defined in htmlPtr, a pointer to its TkHtmlTag
 *	structure is returned.  Otherwise NULL is returned and an
 *	error message is recorded in the interp's result unless interp
 *	is NULL.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static TkHtmlTag *
FindTag(Tcl_Interp *interp, HtmlWidget *htmlPtr, CONST char *tagName)
/*  Tcl_Interp *interp;		Interpreter to use for error message;
			      * if NULL, then don't record an error
			      * message. */
/*  HtmlWidget *htmlPtr;	Widget in which tag is being used. */
/*  CONST char *tagName;	Name of desired tag. */
{
    Tcl_HashEntry *hPtr;

    hPtr = Tcl_FindHashEntry(&htmlPtr->tagTable, tagName);
    if (hPtr != NULL) {
	return (TkHtmlTag *) Tcl_GetHashValue(hPtr);
    }
    if (interp != NULL) {
	Tcl_AppendResult(interp, "tag \"", tagName,
		"\" isn't defined in html widget", (char *) NULL);
    }
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlFreeTag --
 *
 *	This procedure is called when a tag is deleted to free up the
 *	memory and other resources associated with the tag.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Memory and other resources are freed.
 *
 *----------------------------------------------------------------------
 */

void
TkHtmlFreeTag(HtmlWidget *htmlPtr, TkHtmlTag *tagPtr)
/*    HtmlWidget *htmlPtr;		Info about overall widget. */
/*    register TkHtmlTag *tagPtr;	Tag being deleted. */
{
    TkHtmlTagRange *pRange, *pNext;
    /*
     * Remove all the ranges of this tag...
     */
    for (pRange = htmlPtr->tagRanges; pRange; pRange = pNext) {
      pNext = pRange->pNext;
      if (pRange->tag == tagPtr) {
        if (pRange == htmlPtr->tagRanges) htmlPtr->tagRanges = pRange->pNext;
        if (pRange->pPrev) {
          pRange->pPrev->pNext = pRange->pNext;
        }
        if (pRange->pNext) {
          pRange->pNext->pPrev = pRange->pPrev;
        }
        Tcl_Free((char *) pRange);
        htmlPtr->numRanges--;
      } else {
        /* This range stays... */
        htmlPtr->lastRange = pRange;
      }
    }
    
    if (tagPtr->border != None) {
	Tk_Free3DBorder(tagPtr->border);
    }
    if (tagPtr->bdString != NULL) {
	Tcl_Free(tagPtr->bdString);
    }
    if (tagPtr->reliefString != NULL) {
	Tcl_Free(tagPtr->reliefString);
    }
    if (tagPtr->bgStipple != None) {
	Tk_FreeBitmap(htmlPtr->display, tagPtr->bgStipple);
    }
    if (tagPtr->fgColor != None) {
	Tk_FreeColor(tagPtr->fgColor);
    }
    Tk_FreeFont(tagPtr->tkfont);
    if (tagPtr->fgStipple != None) {
	Tk_FreeBitmap(htmlPtr->display, tagPtr->fgStipple);
    }
    if (tagPtr->overstrikeString != NULL) {
	Tcl_Free(tagPtr->overstrikeString);
    }
    if (tagPtr->underlineString != NULL) {
	Tcl_Free(tagPtr->underlineString);
    }
    Tcl_Free((char *) tagPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * SortTags --
 *
 *	This procedure sorts an array of tag pointers in increasing
 *	order of priority, optimizing for the common case where the
 *	array is small.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static void
SortTags(int numTags, TkHtmlTag **tagArrayPtr)
/*    int numTags;		Number of tag pointers at *tagArrayPtr. */
/*    TkHtmlTag **tagArrayPtr;	Pointer to array of pointers. */
{
    int i, j, prio;
    register TkHtmlTag **tagPtrPtr;
    TkHtmlTag **maxPtrPtr, *tmp;

    if (numTags < 2) {
	return;
    }
    if (numTags < 20) {
	for (i = numTags-1; i > 0; i--, tagArrayPtr++) {
	    maxPtrPtr = tagPtrPtr = tagArrayPtr;
	    prio = tagPtrPtr[0]->priority;
	    for (j = i, tagPtrPtr++; j > 0; j--, tagPtrPtr++) {
		if (tagPtrPtr[0]->priority < prio) {
		    prio = tagPtrPtr[0]->priority;
		    maxPtrPtr = tagPtrPtr;
		}
	    }
	    tmp = *maxPtrPtr;
	    *maxPtrPtr = *tagArrayPtr;
	    *tagArrayPtr = tmp;
	}
    } else {
	qsort((VOID *) tagArrayPtr, (unsigned) numTags, sizeof (TkHtmlTag *),
		    TagSortProc);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TagSortProc --
 *
 *	This procedure is called by qsort when sorting an array of
 *	tags in priority order.
 *
 * Results:
 *	The return value is -1 if the first argument should be before
 *	the second element (i.e. it has lower priority), 0 if it's
 *	equivalent (this should never happen!), and 1 if it should be
 *	after the second element.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
TagSortProc(CONST VOID *first, CONST VOID *second)
/*    CONST VOID *first, *second;	Elements to be compared. */
{
    TkHtmlTag *tagPtr1, *tagPtr2;

    tagPtr1 = * (TkHtmlTag **) first;
    tagPtr2 = * (TkHtmlTag **) second;
    return tagPtr1->priority - tagPtr2->priority;
}

/*
 *----------------------------------------------------------------------
 *
 * ChangeTagPriority --
 *
 *	This procedure changes the priority of a tag by modifying
 *	its priority and the priorities of other tags that are affected
 *	by the change.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Priorities may be changed for some or all of the tags in
 *	htmlPtr.  The tags will be arranged so that there is exactly
 *	one tag at each priority level between 0 and htmlPtr->numTags-1,
 *	with tagPtr at priority "prio".
 *
 *----------------------------------------------------------------------
 */

static void
ChangeTagPriority(HtmlWidget *htmlPtr, TkHtmlTag *tagPtr, int prio)
/*    HtmlWidget *htmlPtr;		Information about html widget. */
/*    TkHtmlTag *tagPtr;		Tag whose priority is to be
					changed. */
/*    int prio;				New priority for tag. */
{
    int low, high, delta;
    register TkHtmlTag *tagPtr2;
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;

    if (prio < 0) {
	prio = 0;
    }
    if (prio >= htmlPtr->numTags) {
	prio = htmlPtr->numTags-1;
    }
    if (prio == tagPtr->priority) {
	return;
    } else if (prio < tagPtr->priority) {
	low = prio;
	high = tagPtr->priority-1;
	delta = 1;
    } else {
	low = tagPtr->priority+1;
	high = prio;
	delta = -1;
    }
    for (hPtr = Tcl_FirstHashEntry(&htmlPtr->tagTable, &search);
	    hPtr != NULL; hPtr = Tcl_NextHashEntry(&search)) {
	tagPtr2 = (TkHtmlTag *) Tcl_GetHashValue(hPtr);
	if ((tagPtr2->priority >= low) && (tagPtr2->priority <= high)) {
	    tagPtr2->priority += delta;
	}
    }
    tagPtr->priority = prio;
    TkHtmlSortTagRanges(htmlPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * TagRangesSortProc --
 *
 *	This procedure is called by qsort when sorting an array of
 *	tags in priority order.
 *
 * Results:
 *	The return value is -1 if the first argument should be before
 *	the second element (i.e. it has lower priority), 0 if it's
 *	equivalent (this should never happen!), and 1 if it should be
 *	after the second element.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
TagRangesSortProc(CONST VOID *first, CONST VOID *second)
/*    CONST VOID *first, *second;	Elements to be compared. */
{
    TkHtmlTagRange *rangePtr1, *rangePtr2;

    rangePtr1 = * (TkHtmlTagRange **) first;
    rangePtr2 = * (TkHtmlTagRange **) second;
    return rangePtr1->tag->priority - rangePtr2->tag->priority;
}

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlSortTagRanges --
 *
 *	This procedure sorts the tag ranges list so as the last elements to
 *	have higher priorities than the first elements.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Internal tag-related structures are re-arranged.
 *
 *----------------------------------------------------------------------
 */

static void
TkHtmlSortTagRanges(HtmlWidget *htmlPtr)
/*    HtmlWidget *htmlPtr;		Information about html widget. */
{
  TkHtmlTagRange **arrayPtr, *pRange;
  int i = 0;
  if (htmlPtr->numRanges < 2) return;
  arrayPtr = (TkHtmlTagRange **) Tcl_Alloc((unsigned)
                (htmlPtr->numRanges * sizeof(TkHtmlTagRange *)));
  memset(arrayPtr, 0, (unsigned)
                (htmlPtr->numRanges * sizeof(TkHtmlTagRange *)));
  for (pRange = htmlPtr->tagRanges; pRange; pRange = pRange->pNext) {
    arrayPtr[i++] = pRange;
  }
  qsort((VOID *) arrayPtr, (unsigned) htmlPtr->numRanges,
        sizeof(TkHtmlTagRange *), TagRangesSortProc);
  htmlPtr->tagRanges = arrayPtr[0];
  htmlPtr->tagRanges->pNext = arrayPtr[1];
  htmlPtr->tagRanges->pPrev = NULL;
  for (i = 1; i < htmlPtr->numRanges - 1; i++) {
    arrayPtr[i]->pPrev = arrayPtr[i-1];
    arrayPtr[i]->pNext = arrayPtr[i+1];
  }
  htmlPtr->lastRange = arrayPtr[htmlPtr->numRanges - 1];
  htmlPtr->lastRange->pNext = NULL;
  htmlPtr->lastRange->pPrev = arrayPtr[htmlPtr->numRanges - 2];
  Tcl_Free((char *) arrayPtr);
}; /* TkHtmlSortTagRanges */

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlFindTagRangeUnderMouse --
 *
 *	This procedure finds the tag range that is under the given
 *	coordinates. If more than one tag ranges are found, the one with
 *	the higher priority is returned. If no range is found, NULL is
 *	returned.
 *
 * Results:
 *	The tag range under the mouse or NULL if none under the mouse.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */
TkHtmlTagRange *TkHtmlFindTagRangeUnderMouse(HtmlWidget *htmlPtr, int x,int y) {
  register TkHtmlTagRange *pRange;
#if 0
  HtmlElement *pElement;
  char index[128];
  
  /* Locate the element & offset under the given coordinates */
  snprintf(index, 64, "@%d,%d", x, y);
  /* if (DecodeBaseIndex(htmlPtr, index, &pElement, &offset)) return NULL; */
  if (HtmlGetIndex(htmlPtr, index, &pElement, &offset)) return NULL;
  if (pElement == NULL) return NULL;
  offset += pElement->base.offs;
  /* printf("TkHtmlFindTagRangeUnderMouse: x=%d, y=%d\n", x, y); */
  for (pRange = htmlPtr->lastRange; pRange; pRange = pRange->pPrev) {
    if (((pRange->startElement->base.offs + pRange->charStart) <= offset) &&
        ((pRange->endElement->base.offs + pRange->charEnd) >= offset)) {
      return pRange;
    }
  }
#endif
  Tk_Font font;
  HtmlBlock *pBlock;
  HtmlElement *pElement;
  int width, bytes, offset;

  if (htmlPtr->tagRanges == NULL) return NULL;
  
  x += htmlPtr->xOffset;
  y += htmlPtr->yOffset;
  /* printf("TkHtmlFindElementUnderMouse: x=%d, y=%d\n", x, y); */
  for (pBlock = htmlPtr->firstBlock; pBlock; pBlock = pBlock->pNext) {
    if (pBlock->top  > y || pBlock->bottom < y ||
        pBlock->left > x || pBlock->right  < x) {continue;}
    for (pElement = pBlock->base.pNext; pElement; pElement=pElement->pNext) {
      if (pBlock->pNext && pElement == pBlock->pNext->base.pNext) break;
      switch (pElement->base.type) {
        case Html_Text: {
          if ((x >= pElement->text.x) &&
              (y >= pElement->text.y - pElement->text.ascent) &&
              (x <= pElement->text.x + pElement->text.w) &&
              (y <= pElement->text.y + pElement->text.descent)) {
            /* The coordinates are within this text element. Try to find
             * the offset the mouse is over... */
            HtmlLock(htmlPtr);
            font = HtmlGetFont(htmlPtr, pElement->base.style.font);
            if (HtmlUnlock(htmlPtr) || font == 0) {continue;}
            bytes = Tk_MeasureChars(font, pElement->text.zText,
                                    strlen(pElement->text.zText)+1,
                                    x - pElement->text.x,
                                    TK_PARTIAL_OK, &width);
            offset = pElement->base.offs + bytes;
            /* Find the tag range with greater priority that covers the offset*/
            for (pRange = htmlPtr->lastRange; pRange; pRange = pRange->pPrev) {
              if (((pRange->startElement->base.offs + pRange->charStart)
                                                                 <= offset) &&
                  ((pRange->endElement->base.offs +
                                  pRange->charEnd_plus_one_char) >= offset)) {
                 return pRange;
              }
            }
          }
        }
        case Html_Space: {
        }
        default: {
        }
      }
    }
  }

  /*
   * If we have reached here, this probably means that the coordinates are not
   * over text, as no block matched or not tag range is defined to cover the
   * coordinates. Here, we will try to find a range that its rectangular
   * covers the coordinates. Note, that when the tag is not a rectangle, this
   * may be slight wrong :-)
   */
  for (pRange = htmlPtr->lastRange; pRange; pRange = pRange->pPrev) {
    if ((x >= pRange->xrec.x) && (y >= pRange->xrec.y) &&
        (x <= pRange->xrec.x + pRange->xrec.width) &&
        (y <= pRange->xrec.y + pRange->xrec.height)) {
      return pRange;
    }
  }
  return NULL;
#if 0
  HtmlBlock *pBlock;
  HtmlIndex pIndex;

  pIndex.p = NULL; pIndex.i = 0;
  x += htmlPtr->xOffset; y += htmlPtr->yOffset;
  for (pBlock = htmlPtr->firstBlock; pBlock; pBlock = pBlock->pNext) {
    if (pBlock->top > y || pBlock->bottom < y ||
        pBlock->left > x || pBlock->right < x) continue;
    printf("block=%p\n", pBlock);
    pIndex.p = NULL; pIndex.i = 0;
    printf("TkHtmlFindTagRangeUnderMouse: top=%d, bottom=%d, left=%d, right=%d\n", pBlock->top, pBlock->bottom, pBlock->left, pBlock->right);
    printf("TkHtmlFindTagRangeUnderMouse: x=%d, x_block=%d\n",
                     x, x - pBlock->left);
    FindIndexInBlock(htmlPtr, pBlock, x - pBlock->left, &pIndex.p, &pIndex.i);
    if (pIndex.p) break;
    //for (pElement = pBlock->base.pNext; pElement; pElement=pElement->pNext) {
    //  if (pBlock->pNext && pElement==pBlock->pNext->base.pNext) break;
    //}
  }
  if (pIndex.p == NULL) return NULL;
  offset = pIndex.p->base.offs + pIndex.i;
  printf("TkHtmlFindTagRangeUnderMouse: offset=%d\n", offset);
  for (pRange = htmlPtr->lastRange; pRange; pRange = pRange->pPrev) {
    if (((pRange->startElement->base.offs + pRange->charStart) <= offset) &&
        ((pRange->endElement->base.offs + pRange->charEnd) >= offset)) {
      return pRange;
    }
  }
  /* printf("TkHtmlFindTagRangeUnderMouse -> %p\n", result); */
  return NULL;
#endif
} /* TkHtmlFindTagRangeUnderMouse */

/*
 *--------------------------------------------------------------
 *
 * TkHtmlTagBindProc --
 *
 *	This procedure is invoked by the Tk dispatcher to handle
 *	events associated with bindings on items.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Depends on the command invoked as part of the binding
 *	(if there was any).
 *
 *--------------------------------------------------------------
 */
static void
TkHtmlTagBindProc(ClientData clientData, XEvent *eventPtr)
/*    ClientData clientData;		 Pointer to canvas structure. */
/*    XEvent *eventPtr;			 Pointer to X event that just
					 happened. */
{
    HtmlWidget *htmlPtr = (HtmlWidget *) clientData;
    TkHtmlTagRange *pRange = NULL;
    TkHtmlTag *array[1];
    int x, y;

# define AnyButtonMask (Button1Mask|Button2Mask|Button3Mask\
                        |Button4Mask|Button5Mask)

    Tcl_Preserve((ClientData) htmlPtr);

    if ((eventPtr->type == ButtonPress) ||
        (eventPtr->type == ButtonRelease)) {
        x = eventPtr->xbutton.x;
        y = eventPtr->xbutton.y;
        pRange = TkHtmlFindTagRangeUnderMouse(htmlPtr, x, y);
    } else if ((eventPtr->type == EnterNotify) ||
               (eventPtr->type == LeaveNotify) ||
               (eventPtr->type == MotionNotify)) {
        x = eventPtr->xmotion.x;
        y = eventPtr->xmotion.y;
        pRange = TkHtmlFindTagRangeUnderMouse(htmlPtr, x, y);
        if (htmlPtr->currentRange != pRange) {
          if (htmlPtr->currentRange) {
            /* Send a Leave event... */
            XEvent event = *eventPtr;
            event.type = LeaveNotify;
            event.xcrossing.detail = NotifyAncestor;
            array[0] = htmlPtr->currentRange->tag;
            /*printf("Sending <Leave> to %p\n", htmlPtr->currentRange);*/
            Tk_BindEvent(htmlPtr->bindingTable, &event, htmlPtr->tkwin,
                                                1, (ClientData *) array);
          }
          if (pRange && eventPtr->type != LeaveNotify) {
            XEvent event = *eventPtr;
            event.type = EnterNotify;
            event.xcrossing.detail = NotifyAncestor;
            array[0] = pRange->tag;
            /*printf("Sending <Enter> to %p\n", pElement);*/
            Tk_BindEvent(htmlPtr->bindingTable, &event, htmlPtr->tkwin,
                                                1, (ClientData *) array);
          }
          htmlPtr->currentRange = pRange;
        }
    }
    if (pRange && (htmlPtr->bindingTable != NULL) && (htmlPtr->tkwin != NULL)) {
        /* printf("pElement=%p (%d)\n", pElement, objects); */
      array[0] = pRange->tag;
      Tk_BindEvent(htmlPtr->bindingTable, eventPtr, htmlPtr->tkwin,
                                          1, (ClientData *) array);
    }
    Tcl_Release((ClientData) htmlPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * TagRangesSortProc --
 *
 *	Compare two indices to see which one is earlier in the text.
 *
 * Results:
 *	The return value is 0 if index1Ptr and index2Ptr refer to the same
 *	position in the file, -1 if index1Ptr refers to an earlier position
 *	than index2Ptr, and 1 otherwise.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */
static int
HtmlIndexCmp(HtmlIndex *index1, HtmlIndex *index2) {
  int off1, off2;

  off1 = index1->p->base.offs + index1->i;
  off2 = index2->p->base.offs + index2->i;
  if (off1 == off2) return 0;
  if (off1 <  off2) return -1;
  return 1;
}; /* HtmlIndexCmp */

/*
 *----------------------------------------------------------------------
 *
 * TkHtmlMoveIndex --
 *
 *	Move an index the provided characters in the given direction.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The provided index will be modified.
 *
 *----------------------------------------------------------------------
 */
static void
TkHtmlMoveIndex(HtmlWidget *htmlPtr, HtmlIndex *index, int offset, int forward)
{
  CONST char *ptr;
  int charOffset;
  
  if (index == NULL) return;
  if (index->p == NULL) return;
  charOffset = index->p->base.offs + index->i;
  ptr = &htmlPtr->zText[charOffset];
  /* Move the Offset offset UTF-8 characters... */
  for (; offset; offset --) {
    if (forward) {
      ptr = Tcl_UtfNext(ptr);
    } else {
      ptr = Tcl_UtfPrev(ptr, htmlPtr->zText);
    }
  }
  charOffset = ptr - htmlPtr->zText;
  /* Find the proper element to house the new offset */
  while (index->p) {
    if (forward) {
      
    } else {
      if (index->p->base.offs <= charOffset) break;
      if (index->p->base.pPrev == NULL) break;
      index->p = index->p->base.pPrev;
    }
  }
  index->i = charOffset - index->p->base.offs;
}; /* TkHtmlMoveIndex */
#endif
