/*****************************************************************************
 *  CDM.cpp:                                                                 *
 * ----------------------                                                    *
 *  This file contains the implementation of all CDM Public API.             *
 *  CDM is the Collection and Document Manager and the core component of the *
 *  Ellogon Text Engineering Platform.                                       *
 *  CDM was designed and implemented by Georgios Petasis,                    *
 *  e-mail: petasis@iit.demokritos.gr                                        *
 *  URL:    http://www.iit.demokritos.gr/~petasis                            *
 *                                                                           *
 *  Copyright 1998-2010 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licencing information, please refer to the licence conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/

#include "CDM.h"
#include "CDM_ObjectTypes.h"
#include <string.h>
#include <stdlib.h>
#include "CDM_GlobalObjects.h"

Tcl_ThreadDataKey dataKey;
TCL_DECLARE_MUTEX(ellogonCDMMutex);

extern "C" DLLEXPORT int Cdm_Init(Tcl_Interp *Interp);
void           CDM_PointerObj_UpdateStringProc(Tcl_Obj *obj);
int            CDM_PointerObj_SetPointerFromAnyProc(Tcl_Interp *interp,
                                                    Tcl_Obj *obj);
int            CDM_SaveCollection(CDM_Collection Collection);
CDM_Collection CDM_LoadCollection(CDM_CONST char *Path);
int            CDM_SaveDocument(CDM_Document Document, char *Directory);
CDM_Document   CDM_LoadDocument(CDM_Collection Collection, char *Path);
int            CDM_GetAttributeTypeFromObject(Tcl_Obj *Type);
Tcl_ObjType CDM_TclPointerObjType = {(char *) "CDM_TclPointerObj",
       (Tcl_FreeInternalRepProc *) NULL,  (Tcl_DupInternalRepProc *) NULL,
       CDM_PointerObj_UpdateStringProc,   CDM_PointerObj_SetPointerFromAnyProc};

/*
 * Various static functions used locally in this file...
 */
static
CDM_Annotation CDM_InternalDisplaceAnnotation(CDM_Annotation Annotation,
                                              long offset, long displacement);
void
CDM_DocumentInternalAddAnnotation(CDM_Document Document, CDM_Annotation Ann);

#define CDM_MAX_NUMBER_OF_DISPLACEMENTS_BEFORE_CLEAN 15000

/*
 * From generic/tclParse.c --
 *
 * The following table provides parsing information about each possible 8-bit
 * character. The table is designed to be referenced with either signed or
 * unsigned characters, so it has 384 entries. The first 128 entries
 * correspond to negative character values, the next 256 correspond to
 * positive character values. The last 128 entries are identical to the first
 * 128. The table is always indexed with a 128-byte offset (the 128th entry
 * corresponds to a character value of 0).
 *
 * The macro CHAR_TYPE is used to index into the table and return information
 * about its character argument. The following return values are defined.
 *
 * TYPE_NORMAL -        All characters that don't have special significance to
 *                      the Tcl parser.
 * TYPE_SPACE -         The character is a whitespace character other than
 *                      newline.
 * TYPE_COMMAND_END -   Character is newline or semicolon.
 * TYPE_SUBS -          Character begins a substitution or has other special
 *                      meaning in ParseTokens: backslash, dollar sign, or
 *                      open bracket.
 * TYPE_QUOTE -         Character is a double quote.
 * TYPE_CLOSE_PAREN -   Character is a right parenthesis.
 * TYPE_CLOSE_BRACK -   Character is a right square bracket.
 * TYPE_BRACE -         Character is a curly brace (either left or right).
 */

#define TYPE_NORMAL             0
#define TYPE_SPACE              0x1
#define TYPE_COMMAND_END        0x2
#define TYPE_SUBS               0x4
#define TYPE_QUOTE              0x8
#define TYPE_CLOSE_PAREN        0x10
#define TYPE_CLOSE_BRACK        0x20
#define TYPE_BRACE              0x40

#define CHAR_TYPE(c) (charTypeTable+128)[(int)(c)]

static const char charTypeTable[] = {
    /*
     * Negative character values, from -128 to -1:
     */

    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,

    /*
     * Positive character values, from 0-127:
     */

    TYPE_SUBS,        TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_SPACE,       TYPE_COMMAND_END, TYPE_SPACE,
    TYPE_SPACE,       TYPE_SPACE,       TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_SPACE,       TYPE_NORMAL,      TYPE_QUOTE,       TYPE_NORMAL,
    TYPE_SUBS,        TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_CLOSE_PAREN, TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_COMMAND_END,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_SUBS,
    TYPE_SUBS,        TYPE_CLOSE_BRACK, TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_BRACE,
    TYPE_NORMAL,      TYPE_BRACE,       TYPE_NORMAL,      TYPE_NORMAL,

    /*
     * Large unsigned character values, from 128-255:
     */

    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
    TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,      TYPE_NORMAL,
};


/*
 * This file contains the CDM API Function Implementations...
 */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     C O L L E C T I O N            S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 * ---- CDM_CollectionCallback:
 *  This function is called for all CDM Collections Tcl Commands. Having
 *  Collections structure's pointer in ClientData, calls the corresponding API
 *  functions...
 */
extern "C" int CDM_CollectionCallback(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection = (CDM_Collection ) clientData;
  Tcl_Obj *ObjPtr, *item_obj;
  CDM_Document CDM_Ptr;
  CDM_ListNode *ptr;
  char *CPtr;
  int index, status;

  static const char *Methods[] = {
    "AnnotateCollection",   "Close",            "CreateDocument",
    "FirstDocument",        "Format",
    "GetAssociatedInfo",    "GetAttribute",     "GetAttributes",
    "GetByExternalId",
    "GetComment",           "GetDocument",
    "GetEncoding",          "GetFileIds",       "GetFileXIds",
    "GetName",              "GetOpenedIds",     "GetOpenedDocs",
    "GetOwner",             "Length",           "Modified",
    "NextDocument",         "PutAttribute",     "RemoveAttribute",
    "RemoveDocument",
    "SetAssociatedInfo",    "SetComment",
    "SetEncoding",          "SetName",
    "SetOwner",             "Sync",             "Status",
    (char *) NULL
  };
  enum methods {
    ANNOTATECOLLECTION,      CLOSE,              CREATEDOCUMENT,
    FIRSTDOCUMENT,           FORMAT,
    GETASSOCIATEDINFO,       GETATTRIBUTE,       GETATTRIBUTES,
    GETBYEXTERNALID,
    GETCOMMENT,              GETDOCUMENT,
    GETENCODING,             GETFILEIDS,         GETFILEXIDS,
    GETNAME,                 GETOPENEDIDS,       GETOPENEDDOCS,
    GETOWNER,                LENGTH,             MODIFIED,
    NEXTDOCUMENT,            PUTATTRIBUTE,       REMOVEATTRIBUTE,
    REMOVEDOCUMENT,
    SETASSOCIATEDINFO,       SETCOMMENT,
    SETENCODING,             SETNAME,
    SETOWNER,                SYNC,               STATUS
  };

  CDM_THREAD_SPECIFIC_DATA;
  CDM_Interp = interp;
  Tcl_ResetResult(interp);

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    return CDM_ReportError("CDM_CollectionCallback: Aborting...");
  }

  if (Collection->Type != CDM_TYPE_COLLECTION)
    return CDM_ReportError("CDM_CollectionCallback: Invalid Collection Type");
  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
    return TCL_ERROR;
  }
  if (Tcl_GetIndexFromObj(interp, objv[1], (const char **) Methods,
                          "method", 0, &index) != TCL_OK) return TCL_ERROR;

  /* Handle Arguments... */
  switch ((enum methods) index) {
    case ANNOTATECOLLECTION: {
      if (objc != 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "destination annotator");
        return TCL_ERROR;
      }
      //return CDM_AnnotateCollection(Collection,
      return CDM_ReportError("not ready yet!");
    }
    case CLOSE: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
        return CDM_Close(Collection);
    }
    case CREATEDOCUMENT: {
      if (objc != 6 && objc != 7) {
        Tcl_WrongNumArgs(interp, 2, objv,
                    "XID RawData Annotations Attributes ?Encoding?");
            return TCL_ERROR;
      }
      if (objc == 7) {
        CDM_Ptr = CDM_CreateDocument(Collection, Tcl_GetString(objv[2]),
                   objv[3], objv[4], objv[5], Tcl_GetString(objv[6]));
      } else {
        CDM_Ptr = CDM_CreateDocument(Collection, Tcl_GetString(objv[2]),
                   objv[3], objv[4], objv[5]);
      }
      if (CDM_Ptr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, CDM_NewStringObj(
          Tcl_GetCommandName(CDM_Ptr->Interp, CDM_Ptr->Tcl_Cmd)));
      return TCL_OK;
    }
    case FIRSTDOCUMENT: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      CDM_Ptr = CDM_FirstDocument(Collection);
      if (CDM_Ptr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, CDM_NewStringObj(
          Tcl_GetCommandName(CDM_Ptr->Interp, CDM_Ptr->Tcl_Cmd)));
      return TCL_OK;
    }
    case FORMAT: {
      if (objc < 2 || objc > 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "?native/gate/gate2/registered?");
        return TCL_ERROR;}
      if (objc == 3) {
        if (CDM_Strcmp(Tcl_GetString(objv[2]), "gate2") == 0) {
          Collection->FileFormat = CDM_FORMAT_GATE_2x;
        } else if (CDM_Strcmp(Tcl_GetString(objv[2]), "gate") == 0) {
          Collection->FileFormat = CDM_FORMAT_GATE_1x;
        } else if (CDM_Strcmp(Tcl_GetString(objv[2]), "native") == 0) {
          Collection->FileFormat = CDM_FORMAT_NATIVE2;
        } else {
          /*
           * A registered format?
           */
          int format;
          Collection->FileFormat = CDM_FORMAT_NATIVE2;
          for (format=0; format<CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER;
               format++) {
            if (CDM_AdditionalCollectionTypes[format].CollectionType != NULL &&
               CDM_Strcmp(CDM_AdditionalCollectionTypes[format].CollectionType,
                           Tcl_GetString(objv[2])) == 0) {
              Collection->FileFormat = format + CDM_FORMAT_REGISTERED;
              break;
            }
          }
        }
      }
      switch (Collection->FileFormat) {
        case CDM_FORMAT_NATIVE:
        case CDM_FORMAT_NATIVE2: {
          ObjPtr = CDM_NewStringObj("native"); break;
        }
        case CDM_FORMAT_GATE_1x: {
          ObjPtr = CDM_NewStringObj("gate");   break;
        }
        case CDM_FORMAT_GATE_2x: {
          ObjPtr = CDM_NewStringObj("gate2");   break;
        }
        default: {
          /*
           * A registered format?
           */
          int format = Collection->FileFormat - CDM_FORMAT_REGISTERED;
          if (format >= 0 && format < CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER &&
              CDM_AdditionalCollectionTypes[format].CollectionType != NULL) {
            ObjPtr = CDM_NewStringObj(
              CDM_AdditionalCollectionTypes[format].CollectionType);
          } else {
            ObjPtr = CDM_NewStringObj("native");
          }
          break;
        }
      }
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case GETASSOCIATEDINFO: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      ObjPtr = CDM_GetAssociatedInfo(Collection);
      if (ObjPtr == NULL) ObjPtr = &CDM_Obj_NULL;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case GETATTRIBUTE: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "AttributeName"); return TCL_ERROR;}
      ObjPtr = CDM_GetAttribute(Collection, Tcl_GetString(objv[2]));
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case GETATTRIBUTES: {
      if (objc != 2) {Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      ObjPtr = CDM_GetAttributes(Collection);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case GETBYEXTERNALID: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "XID"); return TCL_ERROR;}
      CDM_Ptr = CDM_GetByExternalId(Collection, Tcl_GetString(objv[2]));
      if (CDM_Ptr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, CDM_NewStringObj(
          Tcl_GetCommandName(CDM_Ptr->Interp, CDM_Ptr->Tcl_Cmd)));
      return TCL_OK;
    }
    case GETCOMMENT: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      ObjPtr = CDM_GetComment(Collection);
      if (ObjPtr == NULL) ObjPtr = &CDM_Obj_NULL;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case GETDOCUMENT: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Id"); return TCL_ERROR;}
      CDM_Ptr = CDM_GetDocument(Collection, Tcl_GetString(objv[2]));
      if (CDM_Ptr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, CDM_NewStringObj(
          Tcl_GetCommandName(CDM_Ptr->Interp, CDM_Ptr->Tcl_Cmd)));
      return TCL_OK;
    }
    case GETENCODING: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      CPtr = CDM_GetEncoding(Collection);
      if (CPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, CDM_NewStringObj(CPtr));
      return TCL_OK;
    }
    case GETFILEIDS: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      ObjPtr = Tcl_NewListObj(0, NULL);
      if (ObjPtr == NULL) return TCL_ERROR;
      ptr = Collection->Documents;
      while (ptr != NULL) {
        CPtr = ptr->ptr[CDM_DOCUMENT_ID];
        item_obj = Tcl_NewStringObj(CPtr, -1);
        if (item_obj == NULL) return TCL_ERROR;
        Tcl_ListObjAppendElement(interp, ObjPtr, item_obj);
        ptr = ptr->next;
      }
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case GETFILEXIDS: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      ObjPtr = Tcl_NewListObj(0, NULL);
      if (ObjPtr == NULL) return TCL_ERROR;
      ptr = Collection->Documents;
      while (ptr != NULL) {
        CPtr = (char *) ptr->ptr[CDM_DOCUMENT_XID];
        item_obj = Tcl_NewStringObj(CPtr, -1);
        if (item_obj == NULL) return TCL_ERROR;
        Tcl_ListObjAppendElement(interp, ObjPtr, item_obj);
        ptr = ptr->next;
      }
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case GETNAME: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      CPtr = CDM_GetName(Collection);
      if (CPtr == NULL) return TCL_ERROR;
      Tcl_SetResult(interp, CPtr, TCL_VOLATILE);
      return TCL_OK;
    }
    case GETOPENEDIDS: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      ObjPtr = Tcl_NewListObj(0, NULL);
      if (ObjPtr == NULL) return TCL_ERROR;
      ptr = Collection->Documents;
      while (ptr != NULL) {
        if (ptr->ptr[CDM_DOCUMENT_OPENED] != NULL) {
          item_obj = Tcl_NewStringObj(ptr->ptr[CDM_DOCUMENT_ID], -1);
          if (item_obj == NULL) return TCL_ERROR;
          Tcl_ListObjAppendElement(interp, ObjPtr, item_obj);
        }
        ptr = ptr->next;
      }
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case GETOPENEDDOCS: {
      CDM_Document Doc;
      CDM_CONST char *Doc_Cmd;

      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      ObjPtr = Tcl_NewListObj(0, NULL);
      if (ObjPtr == NULL) return TCL_ERROR;
      ptr = Collection->Documents;
      while (ptr != NULL) {
        if (ptr->ptr[CDM_DOCUMENT_OPENED] != NULL) {
          Doc = (CDM_Document) ptr->ptr[CDM_DOCUMENT_POINTER];
          if (Doc != NULL && Doc->Type == CDM_TYPE_DOCUMENT) {
            Doc_Cmd = (char *) Tcl_GetCommandName(CDM_Interp, Doc->Tcl_Cmd);
            if (Doc_Cmd == NULL) Doc_Cmd = "(NULL)";
          } else {
            Doc_Cmd = "(ERROR)";
          }
          item_obj = Tcl_NewStringObj(Doc_Cmd, -1);
          if (item_obj == NULL) return TCL_ERROR;
          Tcl_ListObjAppendElement(interp, ObjPtr, item_obj);
        }
        ptr = ptr->next;
      }
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case GETOWNER: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      CPtr = CDM_GetOwner(Collection);
      if (CPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, CDM_NewStringObj(CPtr));
      return TCL_OK;
    }
    case LENGTH: {
      int Length;
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      Length = CDM_Length(Collection);
      if (Length < 0) return TCL_ERROR;
      Tcl_SetObjResult(interp, CDM_NewLongObj((long) Length));
      return TCL_OK;
    }
    case MODIFIED: {
      if (objc < 2 || objc > 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "?boolean?"); return TCL_ERROR;}
      if (objc == 3) {
        if (Tcl_GetBoolean(interp, Tcl_GetString(objv[2]), &status) != TCL_OK)
          return TCL_ERROR;
        Collection->Modified=Collection->ModifiedDocuments = (CDM_UCHAR) status;
      }
      status = 0;
      if (Collection->Modified) {
        status = 1;
      } else if (Collection->ModifiedDocuments) {
        CDM_Document Doc;

        ptr = Collection->Documents;
        while (ptr != NULL) {
          if (ptr->ptr[CDM_DOCUMENT_OPENED] != NULL) {
            Doc = (CDM_Document) ptr->ptr[CDM_DOCUMENT_POINTER];
            if (Doc != NULL && Doc->Modified) {status = 1; break;}
          }
          ptr = ptr->next;
        }
      }
      Tcl_SetObjResult(interp, CDM_NewLongObj((long) status));
      return TCL_OK;
    }
    case NEXTDOCUMENT: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      CDM_Ptr = CDM_NextDocument(Collection);
      if (CDM_Ptr == NULL) {
        /*
         * We have to check whether it an error, or simply we have already open
         * all the Documents. In case of an error, the interpreter will have a
         * string, describing the error...
         */
        if (strlen(Tcl_GetString(Tcl_GetObjResult(interp)))) {
          return TCL_ERROR;
        }
        /*
         * Else, the empty string is already in the interpreter, as we have
         * resetted it...
         */
      } else {
        Tcl_SetResult(interp, (char *)
            Tcl_GetCommandName(CDM_Ptr->Interp, CDM_Ptr->Tcl_Cmd),TCL_VOLATILE);
      }
      return TCL_OK;
    }
    case PUTATTRIBUTE: {
      if (objc != 3 && objc != 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "Attribute ?args?");return TCL_ERROR;}
      return CDM_PutAttribute(Collection, objv[2]);
    }
    case REMOVEATTRIBUTE: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "AttributeName"); return TCL_ERROR;}
      return CDM_RemoveAttribute(Collection, Tcl_GetString(objv[2]));
    }
    case REMOVEDOCUMENT: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Id"); return TCL_ERROR;}
      return CDM_RemoveDocument(Collection, Tcl_GetString(objv[2]));
    }
    case SETASSOCIATEDINFO: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Info"); return TCL_ERROR;}
      return CDM_SetAssociatedInfo(Collection, objv[2]);
    }
    case SETCOMMENT: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Comment"); return TCL_ERROR;}
      return CDM_SetComment(Collection, objv[2]);
    }
    case SETENCODING: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Encoding"); return TCL_ERROR;}
      CPtr = CDM_SetEncoding(Collection, Tcl_GetString(objv[2]));
      if (CPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, CDM_NewStringObj(CPtr));
      return TCL_OK;
    }
    case SETNAME: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Name"); return TCL_ERROR;}
      CPtr = CDM_SetName(Collection, Tcl_GetString(objv[2]));
      if (CPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, Tcl_NewStringObj(CPtr, -1));
      return TCL_OK;
    }
    case SETOWNER: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Owner"); return TCL_ERROR;}
      CPtr = CDM_SetOwner(Collection, Tcl_GetString(objv[2]));
      if (CPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, CDM_NewStringObj(CPtr));
      return TCL_OK;
    }
    case SYNC: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      return CDM_Sync(Collection);
    }
    case STATUS: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      Tcl_SetObjResult(interp, CDM_Status(Collection));
      return TCL_OK;
    }
  }
  return TCL_OK;
} /* CDM_CollectionCallback */

/*
 * ---- CDM_AnnotateCollection:
 *  DESCRIPTION:
 *   <English:
 *     This function will run the specified annotator over the given
 *     Collection... [Not Implemented]
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  Collection Object. (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotateCollection (Tcl API).
 *     This function is not currently implemented.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *     error occurs, the return value will be TCL_ERROR and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp). If all annotators run without errors, TCL_OK will be
 *     returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_AnnotateCollection
 *   >:&:
 */
CDM_EXPORT
int CDM_AnnotateCollection(CDM_Collection Collection) {
  if (Collection == NULL) {
    return CDM_ReportError("CDM_AnnotateCollection: NULL Collection Pointer");
  }
  CDM_NotReady();
  return TCL_OK;
} /* CDM_AnnotateCollection */

/*
 * ---- CDM_Close:
 *  DESCRIPTION:
 *   <English:
 *     Closes the given Collection and frees all the memory occupied by the
 *     Collection and its Documents. All the deleted objects will also be
 *     unregistered from the current Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  Collection Object to be closed. (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_Close (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int). As this function ignores any possible errors, the
 *     return value will always be TCL_OK.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Close
 *   >:&:
 */
CDM_EXPORT
int CDM_Close(CDM_Collection Collection) {
  CDM_Document Doc;
  CDM_ListNode *ptr;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch searchPtr;
  Tcl_Obj *ObjPtr;
  CDM_THREAD_SPECIFIC_DATA;

  if (Collection == NULL) {
    return CDM_ReportError("CDM_Close: NULL Collection Pointer");
  }
  if (Collection->Type != CDM_TYPE_COLLECTION)
    return CDM_ReportError("CDM_Close: Invalid Collection Type");

  /*
   * Check Document's reference count. If > 0, we cannot close the document,
   * as it is still referenced by some process...
   */
  Collection->RefCount--;
  if (Collection->RefCount > 0) return TCL_OK;

  if (Collection->AssociatedInfo != NULL) CDM_Free(Collection->AssociatedInfo);
  if (Collection->CommentObj     != NULL) CDM_Free(Collection->CommentObj);
  if (Collection->Encoding != NULL) Tcl_Free(Collection->Encoding);
  if (Collection->Name     != NULL) Tcl_Free(Collection->Name);
  if (Collection->Owner    != NULL) Tcl_Free(Collection->Owner);
  if (Collection->Tcl_Cmd  != NULL)
      Tcl_DeleteCommandFromToken(Collection->Interp,Collection->Tcl_Cmd);

  CDM_DEBUG("Freeing Attribute Hash Table...\n");

  entryPtr = Tcl_FirstHashEntry(&(Collection->Attributes_HashTable),&searchPtr);
  while (entryPtr != NULL) {
    ObjPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    Tcl_DecrRefCount(ObjPtr);
    entryPtr = Tcl_NextHashEntry(&searchPtr);
  }
  Tcl_DeleteHashTable(&(Collection->Attributes_HashTable));

  /*
   * Free memory occupied by opened Documents...
   */
  ptr = Collection->Documents;
  while (ptr != NULL) {
    if (ptr->ptr[CDM_DOCUMENT_OPENED] != NULL) {
      Doc = (CDM_Document) ptr->ptr[CDM_DOCUMENT_POINTER];
      CDM_DEBUG("Freeing Memory for Document: %s...\n",
                 ptr->ptr[CDM_DOCUMENT_ID]);
      if (Doc != NULL) CDM_Free(Doc);
    }
    ptr = ptr->next;
  }
  CDM_DeleteList(Collection->Documents);
  if (Collection->RemoveDiskDocuments != NULL)
      Tcl_DecrRefCount(Collection->RemoveDiskDocuments);
  Tcl_Free((char *)Collection);
#ifdef FREE_HASHED_OBJS_ON_CLOSE
  ObjectCache.Clear();
#else
  ObjectCache.RemoveUnusedObjects();
#endif /* FREE_HASHED_OBJS_ON_CLOSE */
  return TCL_OK;
} /* CDM_Close */

/*
 * ---- CDM_CreateCollection:
 *  DESCRIPTION:
 *   <English:
 *     This function creates a new Collection object. It accepts three
 *     arguments: the Collection's name which must be the absolute path of
 *     the Collection on disk, a set of Attributes (created with
 *     CDM_CreateAtttributeSet) that will be inserted into
 *     the new Collection and the encoding of the Collection. The Collection's
 *     encoding will be the default encoding that will be used when a new
 *     Document will be created in this Collection. Note that each Document can
 *     have a different encoding than the encoding of the container
 *     Collection. The only place that this encoding will be used, is in the
 *     case where a new Document of unspecified encoding will be created.
 *
 *     If the Collection encoding argument is ommited, it defaults to
 *     "CDM_DefaultEncoding". "CDM_DefaultEncoding" is a global variable (of
 *     type char *), that has as initial value the system encoding.
 *     This variable is also
 *     exported at the Tcl level under the same name and as a result its value
 *     can easily be changed by the end user.
 *
 *     Note that the new Collection will be created only in memory. Nothing
 *     will be saved on the disk. In order for the Collection to be saved in
 *     disk, the function CDM_Sync must be called after the new Collection is
 *     created in memory. This is usually done after all of the desired
 *     Documents have been added to the newly created Collection.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Name:        The Collection's desired disk location (Name) (char *)
 *      Attributes:  A set of Attributes (CDM_AttributeSet)
 *      Encoding:    The Collection's Encoding (Optional) (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_CreateCollection (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Collection object (of type CDM_Collection) that
 *     can be used as a reference to all functions that require a Collection
 *     argument as input. This object will be valid until the Collection that
 *     is referred by this object is closed (through CDM_Close) and the
 *     Collection removed from memory. If this object is used after the
 *     referred Collection has been closed, an undefined behaviour will occur.
 *
 *     If an error occurs, then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Close, CDM_CreateDocument, CDM_CreateAttribute,
 *     CDM_CreateAttributeSet, CDM_Sync, tip_CreateCollection
 *
 *     CDM_DefaultEncoding
 *   >:&:
 */
CDM_EXPORT
CDM_Collection CDM_CreateCollection(
      char *Name, CDM_AttributeSet Attributes, char *Encoding) {
  CDM_Collection NewCollection;
  Tcl_Obj **elemPtrs;
  char CmdName[SHORT_TEMP_STRING];
  int Obj_ID, listLen, result, i;
  CDM_THREAD_SPECIFIC_DATA;

  NewCollection = CDM_NewCollectionObj();

  NewCollection->Interp = CDM_Interp;
  if (Encoding == NULL) Encoding = CDM_DefaultEncoding;
  NewCollection->Encoding = CDM_Strdup(Encoding);
  /*NewCollection->*/Obj_ID = ID.CollectionUniqID();
  if (snprintf(CmdName, SHORT_TEMP_STRING, "Col%d",
                       /*NewCollection->*/Obj_ID) < 1) {
    CDM_Panic("CDM_CreateCollection: snprintf failed!\n");
  }

  NewCollection->Name                = CDM_Strdup(Name);
  NewCollection->Owner               = CDM_Strdup("CDM");
  NewCollection->Documents           = NULL;
  NewCollection->RemoveDiskDocuments = Tcl_NewListObj(0, NULL);
  if (NewCollection->RemoveDiskDocuments == NULL) {
      CDM_Panic("CDM_CreateCollection: Tcl_NewListObj returned NULL - "
                "out of memory?\n");
  }
  Tcl_IncrRefCount(NewCollection->RemoveDiskDocuments);
  NewCollection->CurrentDocumentIndex = NULL;

  if (Attributes != NULL) {
    CDM_DEBUG("Adding Attributes to Collection...\n");
    // Add Attributes (if any) to Collection
    result=Tcl_ListObjGetElements(CDM_Interp, Attributes, &listLen, &elemPtrs);
    if (result != TCL_OK) goto error;
    for (i = 0; i < listLen; i++) {
      if (!CDM_ValidAttribute(elemPtrs[i])) goto error;
      if (CDM_PutAttribute(NewCollection, elemPtrs[i]) != TCL_OK) goto error;
    }
  }
  NewCollection->Modified = 1;
  NewCollection->RefCount = 1;
  CDM_DEBUG("Adding Attributes to Collection DONE!\n");

  // Register the new Collection's command...
  NewCollection->Tcl_Cmd = Tcl_CreateObjCommand(CDM_Interp, CmdName,
                           /*(Tcl_ObjCmdProc*)*/ CDM_CollectionCallback,
                           (ClientData) NewCollection,
                           (Tcl_CmdDeleteProc *) NULL);
  if (NewCollection->Tcl_Cmd == NULL) {
     CDM_ReportError("Cannot create Collection's command \"%s\"", CmdName);
     goto error;
  }
  CDM_DEBUG("Command '%s' registered...\n", CmdName);

  // Everything was ok, return...
  return NewCollection;

error: // Free Allocated Resources...
  ID.CancelCollectionUniqID(/*NewCollection->*/Obj_ID);
  CDM_Close(NewCollection);
  return NULL;
} /* CDM_CreateCollection */

CDM_Collection CDM_CreateCollection(char *Name, CDM_AttributeSet Attributes) {
  CDM_THREAD_SPECIFIC_DATA;
  return CDM_CreateCollection(Name, Attributes, CDM_DefaultEncoding);
} /* CDM_CreateCollection */

/*
 * ---- CDM_CreateDocument:
 *  DESCRIPTION:
 *   <English:
 *     This function creates a new Document object in an existing Collection
 *     object. It accepts five (or six) arguments: the Collection object in
 *     which the new Document is going to be created (Document's Parent), the
 *     External Id of the Document (which represents the full path in disk of
 *     the original file that contains the Document's text), the text (RawData
 *     or ByteSequence) of the new Document, an initial Annotation set, an
 *     initial Attribute set and (optionally) the encoding of the Document.
 *
 *     The Collection object must be a valid object (corresponding to an open
 *     Collection) returned by either CDM_CreateCollection or
 *     CDM_OpenCollection.
 *
 *     The External Id must be a valid (absolute) path. A local copy of this
 *     argument will be created for use by this function. The existance of
 *     this path will not be checked. This path value will be
 *     converted automatically to a valid platform dependant path. As a result,
 *     for this value can be safely used the notation used for paths under the
 *     unix operating system. That means that the path:
 *     [[br]][[code]][[center]]/Users/petasis/Collections[[/center]][[/code]]
 *     will be converted interanlly to the following path, assuming windows as
 *     operating system :
 *     [[br]][[code]][[center]]
 *     C:\Users\petasis\Collections[[/center]][[/code]][[br]]
 *
 *     The RawData parameter must contain the desired for the new Document
 *     text, in UTF-8 format. A local copy of this variable will be also
 *     created. If the text is not in UTF-8 (i.e. contains text just red from
 *     a file using the standart C library routines and is in the ISO 8859-7
 *     encoding) then must be converted to UTF using CDM_ExternalToUtf. If
 *     the origin of this string is Tcl, then it is already in UTF format, as
 *     Tcl uses ONLY UTF for encoding strings internally.
 *
 *     The Annotations parameter must contain an Annotation set created with
 *     CDM_CreateAnnotationSet. A reference to this object will be kept.
 *
 *     The Attributes parameter must hold a valid Attribute set created with
 *     CDM_CreateAttributeSet. A reference to this object will also be kept.
 *
 *     Finally, an optional encoding can be specified. A local copy of this
 *     string will be created. The value of this parameter must be a standart
 *     Tcl encoding value (like iso8859-7 or cp1253). For all available Tcl
 *     encodings please refer to the
 *     [[A HREF="../TclTk/TclLib/Encoding.htm"]]Tcl manuals[[/A]].
 *     If this parameter is ommitted,
 *     then a default value will be used. This value will be inherited from the
 *     parent Collection object. It is important for the encoding to correctly
 *     describe the text. If the given UTF string cannot be converted to the
 *     Document's encoding, then the text will be filled with the character "?"
 *     in places where the conversion will fail.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  Collection Object.             (CDM_Collection)
 *      ExternalId:  Document's External Id (disk location of original text file).     (char *)
 *      RawData:     Document's Text.               (CDM_ByteSequence)
 *      Annotations: Document's Annotations         (CDM_AnnotationSet)
 *      Attributes:  Document's Attributes          (CDM_AttributeSet)
 *      Encoding:    Document's Encoding (Optional) (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_CreateDocument (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Document object (of type CDM_Document) that can
 *     be used as a reference to all functions that require a Document object
 *     as input. This object will be valid until the Collection that holds this
 *     Document object is closed (through CDM_Close) and the parent Collection
 *     removed from memory. If this object is used after its parent Collection
 *     has been closed, an undefined behaviour will occur.
 *
 *     If an error occurs, then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_CreateAnnotationSet, CDM_CreateAttributeSet, CDM_CreateCollection,
 *     CDM_ExternalToUtf, CDM_OpenCollection, tip_CreateDocument,
 *     [[A HREF="../TclTk/TclLib/Encoding.htm"]]The Tcl Encoding manual[[/A]].
 *   >:&:
 */
CDM_EXPORT
CDM_Document CDM_CreateDocument(CDM_Collection Collection, char *XID,
    CDM_ByteSequence RawData, CDM_AnnotationSet Annotations,
    CDM_AttributeSet Attributes, char *Encoding) {
    CDM_Document NewDocument;
  Tcl_Obj **elemPtrs;
  CDM_ListNode *ptr;
  char *ExternalId, *Id, CmdName[SHORT_TEMP_STRING];
  int Obj_ID, listLen, result, i;
#ifdef RECONSTRUCT_ATTRIBUTES
  Tcl_Obj *AttrObj;
#endif /* RECONSTRUCT_ATTRIBUTES */
#ifdef RECONSTRUCT_ANNOTATIONS
  CDM_Annotation Ann;
#endif /* RECONSTRUCT_ANNOTATIONS */
#ifdef WINDOWS
  char *cptr;
#endif /* WINDOWS */
  CDM_THREAD_SPECIFIC_DATA;

  if (Collection == NULL) {
    CDM_ReportError("CDM_CreateDocument: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_CreateDocument: Invalid Collection Type");
    return NULL;
  }

  /* Get Native File Name */
  ExternalId = CDM_GetNativeFileName(XID);
  if (ExternalId == NULL) {
    // Let CDM_GetNativeFileName define an error message for us...
    return NULL;
  }

#ifdef WINDOWS
  /* Under Windows, we don't want to have "\\" in our paths, as it can create
   * problems. We turn them back to the "/", which can be used globally where
   * Tcl runs... */
  cptr = ExternalId;
  while (*cptr != '\0') {
    if (*cptr == '\\') *cptr = '/';
    cptr++;
  }
#endif /* WINDOWS */

  Obj_ID = ID.DocumentUniqID();
  snprintf(CmdName, SHORT_TEMP_STRING, "Doc%d", Obj_ID);
  NewDocument = CDM_NewDocumentObj();
  if (Encoding == NULL) Encoding = CDM_DefaultEncoding;
  NewDocument->Encoding = CDM_Strdup(Encoding);
  NewDocument->ParentPtr = (void *) Collection;
  NewDocument->RawData = RawData;
  if (RawData) Tcl_IncrRefCount(RawData);
  NewDocument->AnnotationMaxID = 0;
  NewDocument->FileFormat = Collection->FileFormat;

  Id = CDM_FileNameTail(ExternalId);
  if (Id == NULL) goto error;
  //NewDocument->ExternalID = ExternalId;
  /* Get tail of External Id */
  //NewDocument->ID = CDM_FileNameTail(ExternalId);
  //if (NewDocument->ID == NULL) goto error;

  /*
   * Is there another document with the same Id?
   * If this is the case, append "_number" until we create a unique Id...
   */
  ptr = CDM_ListFind(Collection->Documents, Id, CDM_DOCUMENT_ID);
  if (ptr != NULL) {
     char *new_id;
     register int index = 0, new_id_len = strlen(Id)+30;

     new_id = Tcl_Alloc(new_id_len);
     if (new_id == NULL) goto error;
     while (ptr != NULL) {
       index++;
#ifdef WINDOWS
       _snprintf_s(new_id, new_id_len, new_id_len-2, "%s_%d", Id, index);
#else
        snprintf(new_id, new_id_len, "%s_%d", Id, index);
#endif /* WINDOWS */
       ptr = CDM_ListFind(Collection->Documents, new_id, CDM_DOCUMENT_ID);
     }
     Tcl_Free(Id);
     Id = new_id;
  }

  /*
   * Add Attributes (if any) to Document...
   */
  if (Attributes) {
    result=Tcl_ListObjGetElements(CDM_Interp, Attributes, &listLen, &elemPtrs);
    if (result != TCL_OK) goto error;
    for (i = 0; i < listLen; i++) {
#ifdef RECONSTRUCT_ATTRIBUTES
      AttrObj = CDM_ReconstructAttribute(elemPtrs[i]);
      if (AttrObj == NULL) goto error;
      CDM_PutAttribute(NewDocument, AttrObj);
#else  /* RECONSTRUCT_ATTRIBUTES */
      if (!CDM_ValidAttribute(elemPtrs[i])) goto error;
      if (CDM_PutAttribute(NewDocument, elemPtrs[i]) != TCL_OK) goto error;
#endif /* RECONSTRUCT_ATTRIBUTES */
    }
  }

  /*
   * Add Annotations (if any) to Document
   */
  if (Annotations) {
    result=Tcl_ListObjGetElements(CDM_Interp, Annotations, &listLen, &elemPtrs);
    if (result != TCL_OK) goto error;
    for (i = 0; i < listLen; i++) {
#ifdef RECONSTRUCT_ANNOTATIONS
      Ann = CDM_ReconstructAnnotation(elemPtrs[i]);
      if (Ann == NULL) goto error;
      CDM_AddAnnotation(NewDocument, Ann);
#else  /* RECONSTRUCT_ANNOTATIONS */
      if (!CDM_ValidAnnotation(elemPtrs[i])) goto error;
      if (CDM_AddAnnotation(NewDocument, elemPtrs[i]) < 0) goto error;
#endif /* RECONSTRUCT_ANNOTATIONS */
    }
  }

  // Update Parent Collection...
  // Add the new document:
  ptr = CDM_NewListNode();
  ptr->ptr[CDM_DOCUMENT_ID]      = Id;
  ptr->ptr[CDM_DOCUMENT_XID]     = ExternalId;
  ptr->ptr[CDM_DOCUMENT_OPENED]  = (char *) 1;
  ptr->ptr[CDM_DOCUMENT_POINTER] = (char *) NewDocument;
  if (Collection->Documents) {
    CDM_AppendListNode(Collection->Documents, ptr);
  } else {
    Collection->Documents = CDM_AppendListNode(NULL, ptr);
  }
  NewDocument->DocumentPtr = ptr;

  // Register the new Document's command...
  NewDocument->Tcl_Cmd = Tcl_CreateObjCommand(CDM_Interp, CmdName,
                           /*(Tcl_ObjCmdProc*)*/ CDM_DocumentCallback,
                           (ClientData) NewDocument,
                           (Tcl_CmdDeleteProc *) NULL);
  if (NewDocument->Tcl_Cmd == NULL) {
     CDM_ReportError("Cannot create Document's command \"%s\"", CmdName);
     goto error;
  }
  Collection->Modified = Collection->ModifiedDocuments = 1;
  NewDocument->Modified = 1;
  NewDocument->RefCount++;

  /*
   * If there are Documents to be removed, ensure that the new Document is not
   * one of them (in case the user has removed a Document and added a new one
   * with the same id).
   */
  if (Collection->RemoveDiskDocuments != NULL) {
    Tcl_DString ds;
    char *argv[2];
    Tcl_Obj *tmp_obj;
    int obj_to_delete_pos;

    Tcl_DStringInit(&ds);
    argv[0] = Collection->Name; argv[1] = Id;
    Tcl_JoinPath(2, argv, &ds);
    tmp_obj = Tcl_NewStringObj(Tcl_DStringValue(&ds), -1);
    Tcl_IncrRefCount(tmp_obj);
    Tcl_DStringFree(&ds);
    while (CDM_Lsearch(Collection->RemoveDiskDocuments, tmp_obj,
                    &obj_to_delete_pos, CDM_LSEARCH_EXACT) != NULL) {
      /* We have found an entry! Delete it! */
      if (Tcl_ListObjReplace(Collection->Interp,
                             Collection->RemoveDiskDocuments,
                             obj_to_delete_pos, 1, 0, NULL) != TCL_OK) break;
    }
    Tcl_DecrRefCount(tmp_obj);
  }
  return NewDocument;
error: // Free Allocated Resources...
  ID.CancelDocumentUniqID(/*NewDocument->*/Obj_ID);
  CDM_Free(NewDocument);
  return NULL;
} /* CDM_CreateDocument */

CDM_EXPORT
CDM_Document CDM_CreateDocument(CDM_Collection  Collection, char *ExternalId,
    CDM_ByteSequence RawData, CDM_AnnotationSet Annotations,
    CDM_AttributeSet Attributes) {
  char *Encoding;
  CDM_THREAD_SPECIFIC_DATA;

  Encoding = CDM_GetEncoding(Collection);
  if (Encoding == NULL) Encoding = CDM_DefaultEncoding;
  return CDM_CreateDocument(Collection, ExternalId, RawData,
                     Annotations, Attributes, Encoding);
} /* CDM_CreateDocument */

CDM_EXPORT
int CDM_Free(CDM_Document Document) {
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch searchPtr;
  Tcl_Obj *ObjPtr;

  if (Document == NULL) {
    return CDM_ReportError("CDM_Free: NULL Document Pointer");
  }
  if (Document->Type != CDM_TYPE_DOCUMENT)
    return CDM_ReportError("CDM_Free: Invalid Document Type");
  if (Document->Encoding != NULL)        Tcl_Free(Document->Encoding);
  if (Document->Tcl_Cmd != NULL)
      Tcl_DeleteCommandFromToken(Document->Interp, Document->Tcl_Cmd);
  CDM_DEBUG("Freeing Attribute Hash Table...\n");
  entryPtr = Tcl_FirstHashEntry(&(Document->Attributes_HashTable), &searchPtr);
  while (entryPtr != NULL) {
    ObjPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    Tcl_DecrRefCount(ObjPtr);
    entryPtr = Tcl_NextHashEntry(&searchPtr);
  }
  Tcl_DeleteHashTable(&(Document->Attributes_HashTable));
  CDM_DEBUG("Freeing Annotation Hash Table...\n");
#ifdef DONT_HASH_DOC_ANNOTATIONS
  delete Document->Annotations_List;
#else  /* DONT_HASH_DOC_ANNOTATIONS */
  entryPtr = Tcl_FirstHashEntry(&(Document->Annotations_HashTable), &searchPtr);
  while (entryPtr != NULL) {
    ObjPtr = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    Tcl_DecrRefCount(ObjPtr);
    entryPtr = Tcl_NextHashEntry(&searchPtr);
  }
  Tcl_DeleteHashTable(&(Document->Annotations_HashTable));
#endif /* DONT_HASH_DOC_ANNOTATIONS */

  if (Document->RawData != NULL) Tcl_DecrRefCount(Document->RawData);
  Tcl_Free((char *)Document);
  return TCL_OK;
} /* CDM_Free */

/*
 * ---- CDM_Destroy:
 *  DESCRIPTION:
 *   <English:
 *     This function will delete the disk representation of a Collection
 *     object. The Name parameter must be the (absolute) path to the directory
 *     that holds the Collection. Having a Collection object, this information
 *     can be obtained through CDM_GetName.
 *
 *     Note that this function will only delete the current disk
 *     representation of a Collection. If the Collection is loaded in memory
 *     and we save the loaded Collection (with the help of CDM_Sync), then the
 *     representaion of the Collection in disk will be re-created.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Name:  The Collection's Name                         (char *)
 *   >
 *  NOTES:
 *   <English:
 *     Note that this function will delete the whole directory that holds the
 *     Collection, along with ALL the files in that directory. If the user has
 *     placed files inside the Collection's directory, these files will be
 *     also deleted.
 *
 *     This function is equivelant to tip_Destroy (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value of this function will be a standart Tcl completion code
 *     (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     Collection's directory cannot be removed, the return value will be
 *     TCL_ERROR and an error message describing the error will be left at the
 *     current active Tcl interpreter (CDM_Interp). If the deletion was done
 *     successfully, TCL_OK will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetName, tip_Destroy
 *   >:&:
 */
CDM_EXPORT
int CDM_Destroy(char *Name) {
  return CDM_DeleteFile(Name);
} /* CDM_Destroy */

/*
 * ---- CDM_FirstDocument:
 *  DESCRIPTION:
 *   <English:
 *    This function will return a Document object representing the first
 *    Document of the specified Collection. It can be used in order
 *    to start an iteration over all Documents of the Collection. This
 *    function will return the first Document in the given Collection and make
 *    the proper arrangements so as subsequent calls to CDM_NextDocument
 *    return the following Documents.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_FirstDocument (Tcl API).
 *
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Document object (of type CDM_Document) that can
 *     be used as a reference to all functions that require a Document object
 *     as input. This object will be valid until the Collection that holds this
 *     Document object is closed (through CDM_Close) and the parent Collection
 *     removed from memory. If this object is used after its parent Collection
 *     has been closed, an undefined behaviour will occur.
 *
 *     If an error occurs, then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Close, CDM_NextDocument, tip_FirstDocument
 *   >:&:
 */
CDM_EXPORT
CDM_Document CDM_FirstDocument(CDM_Collection Collection) {
  Tcl_DString ds;
  CDM_Document Doc;
  CDM_ListNode *ptr;
  char *argv[2];

  if (Collection == NULL) {
    CDM_ReportError("CDM_FirstDocument: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_FirstDocument: Invalid Collection Type");
    return NULL;
  }

  ptr = Collection->Documents;
  if (ptr == NULL) {
    CDM_ReportError("Collection has no Documents!");
    return NULL;
  }
  Collection->CurrentDocumentIndex = ptr;

  // Is Document already open?
  CDM_DEBUG(" ** CDM_FirstDocument Called...\n");
  CDM_DEBUG(" ** Loading Document \"%s\"\n", ptr->ptr[CDM_DOCUMENT_ID]);

  if (ptr->ptr[CDM_DOCUMENT_OPENED] != NULL) { // Document is already open...
    // Get Document's Pointer
    Doc = (CDM_Document) ptr->ptr[CDM_DOCUMENT_POINTER];
    if (Doc == NULL) {
      CDM_ReportError("Cannot locate Opened document with Id=\"%s\" in "
                      "Collection %s!", ptr->ptr[CDM_DOCUMENT_ID],
                   Tcl_GetCommandName(Collection->Interp, Collection->Tcl_Cmd));
      return NULL;
    }
    CDM_DEBUG(" ** This Document is already open: Re-using Structure...\n");
    Doc->RefCount++;
    return Doc;
  }

  CDM_DEBUG(" ** This Document is not Open: Loading from disk...\n");
  // Get Document Path...
  Tcl_DStringInit(&ds);
  argv[0] = Collection->Name; argv[1] = ptr->ptr[CDM_DOCUMENT_ID];
  Tcl_JoinPath(2, argv, &ds);
  Doc = CDM_LoadDocument(Collection, Tcl_DStringValue(&ds));
  Tcl_DStringFree(&ds);
  if (Doc != NULL) Doc->RefCount++;
  return Doc;
} /* CDM_FirstDocument */

/*
 * ---- CDM_GetAssociatedInfo:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the Tcl object (of type Tcl_Obj*) associated by
 *     the user to this Collection object.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     Note that this information is not saved when a Collection is saved. It
 *     exists only while the Collection is open, and is different for each
 *     Collection instance, even if more than one Collection instances have
 *     been created from a single Collection image on the disk.
 *
 *     This function is equivelant to tip_GetAssociatedInfo (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Tcl object (of type Tcl_Obj*) if a Tcl object
 *     has already been associated with this Collection object. Otherwise,
 *     NULL is returned.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or directly modified in any way by the caller. Subsequent calls
 *     of either the C++ or Tcl API can modify the returned object. If the
 *     caller wants to keep a reference to the current state of the returned
 *     object, he must place such a
 *     request by using Tcl_IncrRefCount. When the caller wants to release the
 *     object (and return it to CDM), the functions Tcl_DecrRefCount of
 *     CDM_Free must be used.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_SetAssociatedInfo, CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], Tcl_IncrRefCount
 *   >:&:
 */
CDM_EXPORT
Tcl_Obj *CDM_GetAssociatedInfo(CDM_Collection Collection) {
  if (Collection == NULL) {
    CDM_ReportError("CDM_GetAssociatedInfo: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_GetAssociatedInfo: Invalid Collection Type");
    return NULL;
  }

  return Collection->AssociatedInfo;
} /* CDM_GetAssociatedInfo */

/*
 * ---- CDM_GetAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function will search all the Attributes that the specified
 *     Collection has, for an Attribute that has the same name as the "Name"
 *     parameter. If such an Attribute is found, the Attribute object is
 *     return. If an Attribute with the specified name does not exist, NULL is
 *     returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *      Name:        Attribute name.      (const char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetAttribute (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Attribute object (of type CDM_Attribute) if an
 *     Attribute named as specified is found. In case that the specified
 *     Attribute does not exist, NULL is returned.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or directly modified in any way by the caller. Subsequent calls
 *     of either the C++ or Tcl API can modify the returned object. If the
 *     caller wants to keep a reference to the current state of the returned
 *     object, he must place such a
 *     request by using Tcl_IncrRefCount. When the caller wants to release the
 *     object (and return it to CDM), the functions Tcl_DecrRefCount of
 *     CDM_Free must be used.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]],
 *     tip_GetAttribute
 *   >:&:
 */
CDM_EXPORT
CDM_Attribute CDM_GetAttribute(CDM_Collection Collection, const char *Name) {
  Tcl_HashEntry *entryPtr;

  if (Collection == NULL) {
    CDM_ReportError("CDM_GetAttribute: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_GetAttribute: Invalid Collection Type");
    return NULL;
  }

#ifdef CDM_OBJ_HASH_KEYS
  /*
   * We don't add a reference to this object. If it is an old one, it's
   * refCount is left intact. If it is a new one, it will have a refCount of
   * 1, as it is referenced by the CDM hash table. It will be freed when the
   * table is freed...
   */
  Tcl_Obj *NameObj;
  NameObj = CDM_NewStringObj(Name);
  entryPtr = Tcl_FindHashEntry(&(Collection->Attributes_HashTable),
                                (char *) NameObj);
#else /* CDM_OBJ_HASH_KEYS */
  entryPtr = Tcl_FindHashEntry(&(Collection->Attributes_HashTable), Name);
#endif /* CDM_OBJ_HASH_KEYS */
  if (entryPtr == NULL) return NULL;
  return (CDM_Attribute) Tcl_GetHashValue(entryPtr);
} /* CDM_GetAttribute */

/*
 * ---- CDM_GetAttributes:
 *  DESCRIPTION:
 *   <English:
 *     This function returns a new object (of type CDM_AttributeSet) that
 *     contains all the Attributes that the specified Collection currently
 *     has. CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *
 *     This function will always create and return a new Attribute set object,
 *     even when the Collection has no Attributes defined. (In this case an
 *     empty Attribute set will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetAttributes (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Attribute set object (of type CDM_AttributeSet).
 *     The caller is responsible to free this object through Tcl_DecrRefCount
 *     or CDM_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_GetAttributes
 *   >:&:
 */
CDM_EXPORT
CDM_AttributeSet CDM_GetAttributes(CDM_Collection Collection) {
  CDM_AttributeSet Attributes;
  Tcl_Interp *Interp = Collection->Interp;
  Tcl_Obj    *AttrValue;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch searchPtr;

  if (Collection == NULL) {
    CDM_ReportError("CDM_GetAttributes: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_GetAttributes: Invalid Collection Type");
    return NULL;
  }
  Attributes = Tcl_NewListObj(0, NULL);
  if (Attributes == NULL) {
    CDM_Panic("CDM_GetAttributes: Tcl_NewListObj returned NULL - "
              "out of memory?\n");
  }
  /* Search Attributes Hash Table and Collect All Attributes... */
  entryPtr = Tcl_FirstHashEntry(&(Collection->Attributes_HashTable),&searchPtr);
  while (entryPtr != NULL) {
    AttrValue = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    if (AttrValue == NULL) continue;
    if (Tcl_ListObjAppendElement(Interp, Attributes, AttrValue) != TCL_OK)
          goto error;
    entryPtr = Tcl_NextHashEntry(&searchPtr);
  }
  return Attributes;
error:
  Tcl_DecrRefCount(Attributes);
  return NULL;
} /* CDM_GetAttributes */

/*
 * ---- CDM_GetByExternalId:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Document object that has the same External
 *     Id as the value of the "XID" parameter. The value of the "XID" parameter
 *     is assumed to be encoded using the UTF-8 encoding (thus enabling the
 *     existance of non Latin characters in the value).
 *     If more than one Documents
 *     having the same External Id are found in the specified Collection, then
 *     the Document that was created first will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object.                     (CDM_Collection)
 *      XID:         The External Id of the desired Document.  (char *)
 *   >
 *  NOTES:
 *   <English:
 *     Note that a Collection can have more than one Documents with the same
 *     External Id. These Documents will all have the same External Id but
 *     they will have different Document Ids.
 *
 *     This function will make the proper arrangements so as a call to
 *     CDM_NextDocument (or tip_NextDocument) will return the Document that
 *     was created after the one returned by this function.
 *
 *     This function is equivelant to tip_GetByExternalId (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Document object (of type CDM_Document) that can
 *     be used as a reference to all functions that require a Document object
 *     as input. This object will be valid until the Collection that holds this
 *     Document object is closed (through CDM_Close) and the parent Collection
 *     removed from memory. If this object is used after its parent Collection
 *     has been closed, an undefined behaviour will occur.
 *
 *     If a Document with the requested external Id does not exist, or an
 *     error occurs, then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Close, CDM_GetDocument, CDM_NextDocument, tip_GetByExternalId
 *   >:&:
 */
CDM_EXPORT
CDM_Document CDM_GetByExternalId(CDM_Collection Collection, char *XID) {
  Tcl_DString ds;
  CDM_Document Doc;
  CDM_ListNode *ptr;
  char *argv[2];

  if (Collection == NULL) {
    CDM_ReportError("CDM_GetByExternalId: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_GetByExternalD: Invalid Collection Type");
    return NULL;
  }
  CDM_DEBUG(" ** CDM_GetByExternalId Called...\n");
  CDM_DEBUG(" ** Loading Document \"%s\"\n", XID);
  ptr = CDM_ListFind(Collection->Documents, XID, CDM_DOCUMENT_XID);
  if (ptr == NULL) {
    CDM_ReportError("Collection %s does not contain document with XID=\"%s\"",
              Tcl_GetCommandName(Collection->Interp, Collection->Tcl_Cmd), XID);
    return NULL;
  }
  Collection->CurrentDocumentIndex = ptr; // CDM_NextDocument will return the
                                          // next document...
  // Is Document already open?
  if (ptr->ptr[CDM_DOCUMENT_OPENED] != NULL) { // Document is already open...
    // Get Document's Pointer
    Doc = (CDM_Document) ptr->ptr[CDM_DOCUMENT_POINTER];
    if (Doc == NULL) {
      CDM_ReportError("Cannot locate Opened document with XID=\"%s\" in "
                      "Collection %s!", XID,
                   Tcl_GetCommandName(Collection->Interp, Collection->Tcl_Cmd));
      return NULL;
    }
    CDM_DEBUG(" ** This Document is already open: Re-using Structure...\n");
    Doc->RefCount++;
    return Doc;
  }

  CDM_DEBUG(" ** This Document is not Open: Loading from disk...\n");
  // Get Document Path...
  Tcl_DStringInit(&ds);
  argv[0] = Collection->Name; argv[1] = ptr->ptr[CDM_DOCUMENT_ID];
  Tcl_JoinPath(2, argv, &ds);
  Doc = CDM_LoadDocument(Collection, Tcl_DStringValue(&ds));
  Tcl_DStringFree(&ds);
  if (Doc != NULL) Doc->RefCount++;
  return Doc;
} /* CDM_GetByExternalId */

/*
 * ---- CDM_GetComment:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the Tcl object (of type Tcl_Obj*) holding a
 *     comment associated to this this Collection object.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetComment (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Tcl object (of type Tcl_Obj*) if a comment
 *     has already been associated with this Collection object. Otherwise,
 *     NULL is returned.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or directly modified in any way by the caller. Subsequent calls
 *     of either the C++ or Tcl API can modify the returned object. If the
 *     caller wants to keep a reference to the current state of the returned
 *     object, he must place such a
 *     request by using Tcl_IncrRefCount. When the caller wants to release the
 *     object (and return it to CDM), the functions Tcl_DecrRefCount of
 *     CDM_Free must be used.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_SetComment, CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], Tcl_IncrRefCount
 *   >:&:
 */
CDM_EXPORT
Tcl_Obj *CDM_GetComment(CDM_Collection Collection) {
  if (Collection == NULL) {
    CDM_ReportError("CDM_GetComment: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_GetComment: Invalid Collection Type");
    return NULL;
  }

  return Collection->CommentObj;
} /* CDM_GetComment */

/*
 * ---- CDM_GetDocument:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Document object that has the same Id as
 *     the value of the "Id" parameter. The value of the "Id" parameter
 *     is assumed to be encoded using the UTF-8 encoding (thus enabling the
 *     existance of non Latin characters in the value).
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object.                     (CDM_Collection)
 *      Id:          The Id of the desired Document.          (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function will make the proper arrangements so as a call to
 *     CDM_NextDocument (or tip_NextDocument) will return the Document that
 *     was created after the one returned by this function.
 *
 *     This function is equivelant to tip_GetDocument (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Document object (of type CDM_Document) that can
 *     be used as a reference to all functions that require a Document object
 *     as input. This object will be valid until the Collection that holds this
 *     Document object is closed (through CDM_Close) and the parent Collection
 *     removed from memory. If this object is used after its parent Collection
 *     has been closed, an undefined behaviour will occur.
 *
 *     If a Document with the requested Id does not exist, or an error occurs,
 *     then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Close, CDM_GetByExternalId, tip_GetDocument
 *   >:&:
 */
CDM_EXPORT
CDM_Document CDM_GetDocument(CDM_Collection Collection, char *Id) {
  Tcl_DString ds;
  CDM_Document Doc;
  CDM_ListNode *ptr;
  char *argv[2];

  if (Collection == NULL) {
    CDM_ReportError("CDM_GetDocument: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_GetDocument: Invalid Collection Type");
    return NULL;
  }

  CDM_DEBUG(" ** CDM_GetDocument Called...\n");
  CDM_DEBUG(" ** Loading Document \"%s\"\n", Id);
  /*
   * Georgios Petasis, 02 Feb 2003:
   * Usually, the caller calls this function while:
   * 1) Has got the Document Ids as a list
   * 2) Iterates over this list, by asking sequentially the Documents.
   * To take advantage of this situation, we should start searching from the
   * current Document, and if that fails, we search the entire list. This
   * simple optimisation may have a large gain on very large Collections...
   */
  if (Collection->CurrentDocumentIndex != NULL) {
    ptr = CDM_ListFind(Collection->CurrentDocumentIndex, Id, CDM_DOCUMENT_ID);
    /* If the Document was not found, check the entire Document list... */
    if (ptr == NULL) {
      ptr = CDM_ListFind(Collection->Documents, Id, CDM_DOCUMENT_ID);
    }
  } else {
    ptr = CDM_ListFind(Collection->Documents, Id, CDM_DOCUMENT_ID);
  }
  if (ptr == NULL) {
    CDM_ReportError("Collection %s does not contain document with Id=\"%s\"",
               Tcl_GetCommandName(Collection->Interp, Collection->Tcl_Cmd), Id);
    return NULL;
  }
  Collection->CurrentDocumentIndex = ptr;

  // Is Document already open?
  if (ptr->ptr[CDM_DOCUMENT_OPENED] != NULL) { // Document is already open...
    // Get Document's Pointer
    Doc = (CDM_Document) ptr->ptr[CDM_DOCUMENT_POINTER];
    if (Doc == NULL) {
      CDM_ReportError("Cannot locate Opened document with Id=\"%s\" in "
                      "Collection %s!", Id,
                   Tcl_GetCommandName(Collection->Interp, Collection->Tcl_Cmd));
      return NULL;
    }
    CDM_DEBUG(" ** This Document is already open: Re-using Structure...\n");
    Doc->RefCount++;
    return Doc;
  }

  CDM_DEBUG(" ** This Document is not Open: Loading from disk...\n");
  // Get Document Path...
  Tcl_DStringInit(&ds);
  argv[0] = Collection->Name; argv[1] = Id;
  Tcl_JoinPath(2, argv, &ds);
  Doc = CDM_LoadDocument(Collection, Tcl_DStringValue(&ds));
  Tcl_DStringFree(&ds);
  if (Doc != NULL) Doc->RefCount++;
  return Doc;
} /* CDM_GetDocument */

/*
 * ---- CDM_GetEncoding:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the encoding of the specified Collection. The
 *     return value will be of type char* and will be owned by CDM. Its value
 *     will be a standart Tcl encoding value (like iso8859-7 or cp1253). For
 *     all available Tcl encodings please refer to the
 *     [[A HREF="../TclTk/TclLib/Encoding.htm"]]Tcl manuals[[/A]].
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object.           (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetEncoding (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a UTF-8 string that represents the encoding of
 *     the specified Collection.
 *
 *     If an error occurs, then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_SetEncoding, tip_GetEncoding,
 *     [[A HREF="../TclTk/TclLib/Encoding.htm"]]The Tcl Encoding manual[[/A]].
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetEncoding(CDM_Collection Collection) {
  if (Collection == NULL) {
    CDM_ReportError("CDM_GetEncoding: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_GetEncoding: Invalid Collection Type");
    return NULL;
  }
  return Collection->Encoding;
} /* CDM_GetEncoding */

/*
 * ---- CDM_GetName:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Name of the specified Collection. This
 *     value will be the (absolute) full path to the directory that contains
 *     the disk representation of the specified Collection. The returned value
 *     will be encoded using the UTF-8 encoding (thus enabling the
 *     existance of non Latin characters in the value).
 *
 *     The return value will be of type char* and will be owned by CDM.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object.           (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetName (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a UTF-8 string that represents the Name of the
 *     specified Collection (the disk (absolute) path of the Collection in
 *     disk).
 *
 *     If an error occurs, then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never
 *     be freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_SetName, tip_GetName
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetName(CDM_Collection Collection) {
  if (Collection == NULL) {
    CDM_ReportError("CDM_GetName: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_GetName: Invalid Collection Type");
    return NULL;
  }
  return Collection->Name;
} /* CDM_GetName */

/*
 * ---- CDM_GetOwner:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Owner of the specified Collection. Its
 *     value is not of great importance, and usually defaults to the value
 *     "CDM". The returned value will be encoded using the UTF-8 encoding
 *     (thus enabling the existance of non Latin characters in the value).
 *
 *     The return value will be of type char* and will be owned by CDM.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.           (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetOwner (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a UTF-8 string that represents the Owner of the
 *     specified Collection.
 *
 *     If an error occurs, then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_SetOwner, tip_GetOwner
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetOwner(CDM_Collection Collection) {
  if (Collection == NULL) {
    CDM_ReportError("CDM_GetOwner: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_GetOwner: Invalid Collection Type");
    return NULL;
  }
  return Collection->Owner;
} /* CDM_GetOwner */

/*
 * ---- CDM_Length:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Length (the number of all the Documents)
 *     of the specified Collection.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.           (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetLength (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return an integer value that will represent the
 *     total number of the Documents that the specified Collection currently
 *     holds.
 *
 *     If an error occurs, then a negative value (-1) will be return and an
 *     error message will be left at the current active Tcl interpreter
 *     (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetLength
 *   >:&:
 */
CDM_EXPORT
int CDM_Length(CDM_Collection Collection) {
  if (Collection == NULL) {
    CDM_ReportError("CDM_Length: NULL Collection Pointer");
    return -1;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_Length: Invalid Collection Type");
    return -1;
  }
  return CDM_ListLength(Collection->Documents);
} /* CDM_Length */

/*
 * ---- CDM_NextDocument:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the next available Document object. "Next"
 *     means the Document that follows the one previously returned by calls to
 *     the following functions:
 *     CDM_FirstDocument, CDM_NextDocument, CDM_GetDocument,
 *     CDM_GetByExternalId.
 *
 *     If previous calls to this function have returned all available
 *     Documents, then NULL will be returned and the current active Tcl
 *     interpreter (CDM_Interp) will be reseted. If an error occurs, then NULL
 *     will also be return and an error message will be left at the current
 *     active Tcl interpreter. As a result, the caller must check the
 *     information stored at this Tcl interpreter in order to distinguish
 *     between the normal termination of the Document chain or an error
 *     condition.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.           (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     Note that if this function is called with a Collection object that has
 *     just be created or opened, an undefined behaviour will occur.
 *
 *     This function is equivelant to tip_NextDocument (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Document object (of type CDM_Document) that can
 *     be used as a reference to all functions that require a Document object
 *     as input. This object will be valid until the Collection that holds this
 *     Document object is closed (through CDM_Close) and the parent Collection
 *     removed from memory. If this object is used after its parent Collection
 *     has been closed, an undefined behaviour will occur.
 *
 *     If an error occurs, then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp). NULL will
 *     also be returned if the last Document of the specified Collection has
 *     already been returned by any of the following functions:
 *     CDM_FirstDocument, CDM_NextDocument, CDM_GetDocument,
 *     CDM_GetByExternalId.
 *
 *     Also note that the returned object is property of the CDM and should
 *     never be freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Close, CDM_FirstDocument, CDM_NextDocument, CDM_GetDocument,
 *     CDM_GetByExternalId, tip_NextDocument
 *   >:&:
 */
CDM_EXPORT
CDM_Document CDM_NextDocument(CDM_Collection Collection) {
  Tcl_DString ds;
  CDM_Document Doc;
  CDM_ListNode *ptr;
  char *argv[2];

  if (Collection == NULL) {
    CDM_ReportError("CDM_NextDocument: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_NextDocument: Invalid Collection Type");
    return NULL;
  }
  ptr = Collection->CurrentDocumentIndex;
  if (ptr == NULL) {
    Collection->CurrentDocumentIndex = NULL;
    return NULL;
  }
  /*
   * Advance to the next Document...
   */
  ptr = ptr->next;
  Collection->CurrentDocumentIndex = ptr;
  if (ptr == NULL) {
    /*
     * No more Documents...
     */
    return NULL;
  }

  // Is Document already open?
  CDM_DEBUG(" ** CDM_NextDocument Called...\n");
  CDM_DEBUG(" ** Loading Document \"%s\"\n", ptr->ptr[CDM_DOCUMENT_ID]);
  if (ptr->ptr[CDM_DOCUMENT_OPENED] != NULL) { // Document is already open...
    // Get Document's Pointer
    Doc = (CDM_Document) ptr->ptr[CDM_DOCUMENT_POINTER];
    if (Doc == NULL) {
      CDM_ReportError("Cannot locate Opened document with Id=\"%s\" in "
                      "Collection %s!", ptr->ptr[CDM_DOCUMENT_ID],
               Tcl_GetCommandName(Collection->Interp, Collection->Tcl_Cmd));
      return NULL;
    }
    CDM_DEBUG(" ** This Document is already open: Re-using Structure...\n");
    Doc->RefCount++;
    return Doc;
  }

  CDM_DEBUG(" ** This Document is not Open: Loading from disk...\n");
  // Get Document Path...
  Tcl_DStringInit(&ds);
  argv[0] = Collection->Name;
  argv[1] = ptr->ptr[CDM_DOCUMENT_ID];
  Tcl_JoinPath(2, argv, &ds);
  Doc = CDM_LoadDocument(Collection, Tcl_DStringValue(&ds));
  Tcl_DStringFree(&ds);
  if (Doc != NULL) Doc->RefCount++;
  return Doc;
} /* CDM_NextDocument */

/*
 * ---- CDM_OpenCollection:
 *  DESCRIPTION:
 *   <English:
 *     This function will load the Collection from the directory specified by
 *     the "Name" argument. If the target that "Name" points is a directory
 *     that is readable by the current user and contains a valid Collection
 *     specification (either in native or in GATE format), then this
 *     Collection will be loaded from disk to the memory. A new Collection
 *     object (of type CDM_Collection) will be created and returned to the
 *     caller.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Name:  The Name of the Collection (absolute path).   (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_OpenCollection (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will construct and return a Collection object (of type
 *     CDM_Collection) that will represent the Collection stored at the
 *     directory specified by the "Name" parameter. The caller must use this
 *     object in order to refer to the newly opened Collection and must use
 *     the function CDM_Close in order to close the Collection and free the
 *     occupied memory if the Collection is not needed any more.
 *
 *     In case of an error NULL will be returned and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never
 *     be freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Close, tip_OpenCollection
 *   >:&:
 */
CDM_EXPORT
CDM_Collection CDM_OpenCollection(CDM_CONST char *Name) {
  CDM_Collection Col;

  if (Name == NULL || *Name == '\0') {
    CDM_ReportError("CDM_OpenCollection: NULL or empty Collection Name!");
    return NULL;
  }
  Col = CDM_LoadCollection(Name);
  if (Col != NULL) Col->RefCount = 1;
  return Col;
} /* CDM_OpenCollection */

/*
 * ---- CDM_PutAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function will add a given Attribute to the specified Collection
 *     object. If an Attribute with the same name already exists in the
 *     Collection, then it will be overwritten by the new Attribute. Else, the
 *     new Attribute will be appended to the existing Attribute set of the
 *     specified Collection.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object.       (CDM_Collection)
 *      Attribute:   The Attribute to be added. (CDM_Atrribute)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_PutAttribute (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *     error occurs, the return value will be TCL_ERROR and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp). Else, TCL_OK will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_PutAttribute
 *   >:&:
 */
CDM_EXPORT
int CDM_PutAttribute(CDM_Collection Collection, CDM_Attribute Attr) {
  register Tcl_HashEntry *entryPtr;
  Tcl_Obj **elemPtrs, *Obj;
  int items;

  if (Collection == NULL) {
    return CDM_ReportError("CDM_PutAttribute: NULL Collection Pointer");
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    return CDM_ReportError("CDM_PutAttribute: Invalid Collection Type");
  }
  if (!CDM_ValidAttribute(Attr)) return TCL_ERROR;
  Tcl_ListObjGetElements(Collection->Interp, Attr, &items, &elemPtrs);
#ifdef CDM_OBJ_HASH_KEYS
  entryPtr = Tcl_CreateHashEntry(&(Collection->Attributes_HashTable),
                                 (char *) elemPtrs[0], &items);
#else /* CDM_OBJ_HASH_KEYS */
  entryPtr = Tcl_CreateHashEntry(&(Collection->Attributes_HashTable),
                                 Tcl_GetString(elemPtrs[0]), &items);
#endif /* CDM_OBJ_HASH_KEYS */
  if (entryPtr == NULL) return TCL_ERROR;
  Tcl_IncrRefCount(Attr);
  if (items == 0) {
    Obj = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    Tcl_DecrRefCount(Obj);
  }
  Tcl_SetHashValue(entryPtr, (ClientData) Attr);
  Collection->Modified = 1;
  return TCL_OK;
} /* CDM_PutAttribute */

/*
 * ---- CDM_RegisterCollectionType:
 *  DESCRIPTION:
 *   <English:
 *   >
 *  ARGUMENTS:
 *   <English:
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Attribute exists and is successfully removed, then TCL_OK
 *     will be returned. In case of an error, the return value will be
 *     TCL_ERROR and an error message describing the error will be left at the
 *     current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
int CDM_RegisterCollectionType(const char *type,
                               CDM_LoadCollectionProc     *loadCollectionPtr,
                               CDM_SaveCollectionProc     *saveCollectionPtr,
                               CDM_LoadDocumentProc       *loadDocumentPtr,
                               CDM_SaveDocumentProc       *saveDocumentPtr,
                               CDM_CollectionXIdsProc     *CollectionXIdsPtr,
                               CDM_CollectionIdsProc      *CollectionIdsPtr,
                               CDM_DocumentAttributesProc *DocumentAttrsPtr) {
  int i;
  CDM_THREAD_SPECIFIC_DATA;

  for (i=0; i<CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER; i++) {
    if (CDM_AdditionalCollectionTypes[i].CollectionType != NULL) continue;
    CDM_AdditionalCollectionTypes[i].CollectionType    = CDM_Strdup(type);
    CDM_AdditionalCollectionTypes[i].loadCollectionPtr = loadCollectionPtr;
    CDM_AdditionalCollectionTypes[i].saveCollectionPtr = saveCollectionPtr;
    CDM_AdditionalCollectionTypes[i].loadDocumentPtr   = loadDocumentPtr;
    CDM_AdditionalCollectionTypes[i].saveDocumentPtr   = saveDocumentPtr;
    CDM_AdditionalCollectionTypes[i].CollectionXIdsPtr = CollectionXIdsPtr;
    CDM_AdditionalCollectionTypes[i].CollectionIdsPtr  = CollectionIdsPtr;
    CDM_AdditionalCollectionTypes[i].DocumentAttrsPtr  = DocumentAttrsPtr;
    return TCL_OK;
  }
  return TCL_ERROR;
} /* CDM_RegisterCollectionType */

/*
 * ---- CDM_RemoveAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function will remove the Attribute named exactly as the "Name"
 *     parameter. If the requested Attribute does not exist, an error
 *     condition will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.           (CDM_Collection)
 *     Name:        The Name of the Attribute to be removed. (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_RemoveAttribute (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Attribute exists and is successfully removed, then TCL_OK
 *     will be returned. In case of an error, the return value will be
 *     TCL_ERROR and an error message describing the error will be left at the
 *     current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_RemoveAttribute
 *   >:&:
 */
CDM_EXPORT
int CDM_RemoveAttribute(CDM_Collection Collection, char *Name) {
  Tcl_HashEntry *entryPtr;
  Tcl_Obj *Obj;

  if (Collection == NULL) {
    return CDM_ReportError("CDM_RemoveAttribute: NULL Collection Pointer");
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    return CDM_ReportError("CDM_RemoveAttribute: Invalid Collection Type");
  }

#ifdef CDM_OBJ_HASH_KEYS
  /*
   * We don't add a reference to this object. If it is an old one, it's
   * refCount is left intact. If it is a new one, it will have a refCount of
   * 1, as it is referenced by the CDM hash table. It will be freed when the
   * table is freed...
   */
  Tcl_Obj *NameObj;
  NameObj = CDM_NewStringObj(Name);
  entryPtr = Tcl_FindHashEntry(&(Collection->Attributes_HashTable),
                                (char *) NameObj);
#else /* CDM_OBJ_HASH_KEYS */
  entryPtr = Tcl_FindHashEntry(&(Collection->Attributes_HashTable), Name);
#endif /* CDM_OBJ_HASH_KEYS */
  if (entryPtr == NULL) {
    return CDM_ReportError("Undefined Attribute Name: \"%s\"", Name);
  }
  Obj = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
  if (Obj != NULL) Tcl_DecrRefCount(Obj);
  Tcl_DeleteHashEntry(entryPtr);
  Collection->Modified = 1;
  return TCL_OK;
} /* CDM_RemoveAttribute */

/*
 * ---- CDM_RemoveDocument:
 *  DESCRIPTION:
 *   <English:
 *     This function will remove the Document object that has as Id the
 *     specified by the parameter "Id" value. If a Document with the requested
 *     Id does not exist, then an error will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.                  (CDM_Collection)
 *      Id:         The Id of the Document to be removed. (char *)
 *   >
 *  NOTES:
 *   <English:
 *     If the removed Document object has been previously opened (through
 *     CDM_GetByExternalId, CDM_GetDocument, CDM_FirstDocument or
 *     CDM_NextDocument) then the returned by these functions Document object
 *     will be invalidated. The caller should not use such a pointer in any
 *     way.
 *
 *     This function is equivelant to tip_RemoveDocument (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Document exists and is successfully removed, then TCL_OK will
 *     be returned. In case of an error, the return value will be TCL_ERROR and
 *     an error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetByExternalId, CDM_GetDocument, CDM_FirstDocument,
 *     CDM_NextDocument, tip_RemoveDocument
 *   >:&:
 */
CDM_EXPORT
int CDM_RemoveDocument(CDM_Collection Collection, char *Id) {
  Tcl_DString ds;
  CDM_Document Doc;
  CDM_ListNode *ptr;
  char *argv[2];
  int deleting_current_document = 0;

  if (Collection == NULL) {
    return CDM_ReportError("CDM_RemoveDocument: NULL Collection Pointer");
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    return CDM_ReportError("CDM_RemoveDocument: Invalid Collection Type");
  }
  if (Id == NULL || *Id == '\0') {
    return CDM_ReportError("Invalid Id \"%s\"", Id);
  }
  ptr = CDM_ListFind(Collection->Documents, Id, CDM_DOCUMENT_ID);
  if (ptr == NULL) {
    return CDM_ReportError("Collection %s does not contain document with "
                           "Id=\"%s\"",
               Tcl_GetCommandName(Collection->Interp, Collection->Tcl_Cmd), Id);
  }
  if (Collection->CurrentDocumentIndex == ptr) {
    deleting_current_document = 1;
    Collection->CurrentDocumentIndex = Collection->Documents;
  }
  // Is Document Opened? That means that occupies memory.
  // So, if Document is open, free memory...
  if (ptr->ptr[CDM_DOCUMENT_OPENED] != NULL) {
    Doc = (CDM_Document) ptr->ptr[CDM_DOCUMENT_POINTER];
    if (Doc == NULL) {
      return CDM_ReportError("Cannot locate Opened document with Id=\"%s\" in "
                             "Collection %s!", Id,
                   Tcl_GetCommandName(Collection->Interp, Collection->Tcl_Cmd));
    }
    // Remove Document from memory...
    CDM_DEBUG("Removing Document \"%s\" from memory...\n", Id);
    CDM_Free(Doc);
  }
  // Remove Document from Collection
  Collection->Documents = CDM_DeleteListNode(Collection->Documents, ptr);

  /*
   * The Collection->Documents may have changed if we deleted the first
   * document. So, update the index, if needed...
   */
  if (deleting_current_document) {
    Collection->CurrentDocumentIndex = Collection->Documents;
  }

  /*
   * Remove Document from disk: Any modifications made to Collection will not
   * be made in disk until the Collection gets saved. So, just keep somewhere
   * the filename of the file to be removed and not remove it now...
   */
    // Get Document Path...
    Tcl_DStringInit(&ds);
    argv[0] = Collection->Name; argv[1] = Id;
    Tcl_JoinPath(2, argv, &ds);
    Tcl_ListObjAppendElement(Collection->Interp,
                             Collection->RemoveDiskDocuments,
                             Tcl_NewStringObj(Tcl_DStringValue(&ds), -1));
    Tcl_DStringFree(&ds);

  Collection->Modified = Collection->ModifiedDocuments = 1;
  return TCL_OK;
} /* CDM_RemoveDocument */

/*
 * ---- CDM_SetAssociatedInfo:
 *  DESCRIPTION:
 *   <English:
 *     This function sets the provided Tcl object (of type Tcl_Obj*) as
 *     associated information to the specified Collection object. This
 *     "associated information" (Tcl object) can contain anything that the
 *     user wants to store ("associate") with this Collection. CDM does not
 *     modify or use this information in any way.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *      Info:        A Tcl object holding the information that is to be associated with the specified Collection (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetAssociatedInfo (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Collection exists and the information Tcl object is
 *     successfully associated with the Collection, then TCL_OK will
 *     be returned. In case of an error, the return value will be TCL_ERROR and
 *     an error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetAssociatedInfo, CDM_SetComment, CDM_GetComment, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], Tcl_IncrRefCount
 *   >:&:
 */
CDM_EXPORT
int CDM_SetAssociatedInfo(CDM_Collection Collection, Tcl_Obj *Info) {
  if (Collection == NULL) {
    CDM_ReportError("CDM_SetAssociatedInfo: NULL Collection Pointer");
    return TCL_ERROR;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_SetAssociatedInfo: Invalid Collection Type");
    return TCL_ERROR;
  }

  if (Collection->AssociatedInfo != NULL) CDM_Free(Collection->AssociatedInfo);
  Collection->AssociatedInfo = Info;
  if (Info != NULL) Tcl_IncrRefCount(Info);
  return TCL_OK;
} /* CDM_SetAssociatedInfo */

/*
 * ---- CDM_SetComment:
 *  DESCRIPTION:
 *   <English:
 *     This function sets the provided Tcl object (of type Tcl_Obj*) as
 *     comment to the specified Collection object. This "comment" (Tcl object)
 *     can contain anything that the user wants to store with this Collection.
 *     CDM does not modify or use this information in any way. This comment
 *     object will be saved allong the Collection, if the Collection is saved.
 *     This is the primary difference of comment towards associated
 *     information that can be inserted with
 *     CDM_GetAssociatedInfo/CDM_SetAssociatedInfo.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *      Info:        A Tcl object holding the information that is to be associated as comment with the specified Collection (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetComment (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Collection exists and the comment Tcl object is
 *     successfully associated with the Collection, then TCL_OK will
 *     be returned. In case of an error, the return value will be TCL_ERROR and
 *     an error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetComment, CDM_GetAssociatedInfo, CDM_SetAssociatedInfo, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], Tcl_IncrRefCount
 *   >:&:
 */
CDM_EXPORT
int CDM_SetComment(CDM_Collection Collection, Tcl_Obj *Info) {
  if (Collection == NULL) {
    CDM_ReportError("CDM_SetComment: NULL Collection Pointer");
    return TCL_ERROR;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_SetComment: Invalid Collection Type");
    return TCL_ERROR;
  }

  if (Collection->CommentObj != NULL) CDM_Free(Collection->CommentObj);
  Collection->CommentObj = Info;
  if (Info != NULL) Tcl_IncrRefCount(Info);
  Collection->Modified = 1;
  return TCL_OK;
} /* CDM_SetComment */

/*
 * ---- CDM_SetEncoding:
 *  DESCRIPTION:
 *   <English:
 *     This function will change the encoding of a Collection object to the
 *     encoding specified by the "Encoding" parameter.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.                        (CDM_Collection)
 *     Encoding:    The desired new encoding of the Collection. (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SetEncoding (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a pointer (of type char*) to the string
 *     buffer that holds the new encoding value. This pointer will be at a
 *     different memory location than the given parameter value, as the CDM
 *     will create and manipulate a local copy.
 *
 *     In case of an error, NULL will be returned and an error message
 *     describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *
 *     Note that the returned pointer is property of the CDM and should never
 *     be freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetEncoding, tip_SetEncoding
 *   >:&:
 */
CDM_EXPORT
char *CDM_SetEncoding(CDM_Collection Collection, char *Encoding) {
  CDM_THREAD_SPECIFIC_DATA;
  if (Collection == NULL) {
    CDM_ReportError("CDM_SetEncoding: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_SetEncoding: Invalid Collection Type");
    return NULL;
  }
  if (Collection->Encoding != NULL) Tcl_Free(Collection->Encoding);
  if (Encoding == NULL || *Encoding == '\0') Encoding = CDM_DefaultEncoding;
  Collection->Encoding = CDM_Strdup(Encoding);
  Collection->Modified = 1;
  return Collection->Encoding;
} /* CDM_SetEncoding */

/*
 * ---- CDM_SetName:
 *  DESCRIPTION:
 *   <English:
 *     This function will change the Name of the specified Collection. The
 *     Name of the Collection represents the (absolute) path of the
 *     Collection's disk representation. Changing the Name of a Collection
 *     will result into a new Collection representation in disk if the
 *     Collection is saved (through CDM_Sync). Note that this function will
 *     not perform any checks on the value of the "Name" parameter. If an
 *     invalid path is given, or a path that corresponds to an existing file,
 *     no error will be returned by this function. (Although CDM_Sync will
 *     return an error if the caller tries to save a Collection with an
 *     invalid Name.)
 *
 *     The value of the "Name" parameter will not modified in
 *     any way by CDM, as a local copy will be created and modified internally
 *     by the CDM.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.                (CDM_Collection)
 *     Name:    The desired new Name of the Collection. (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SetName (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a pointer (of type char*) to the string buffer
 *     that holds the new Collection Name. This pointer will be at a different
 *     memory location than the given parameter value, as the CDM will create
 *     and manipulate a local copy.
 *
 *     In case of an error, NULL will be returned and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp).
 *
 *     Note that the returned pointer is property of the CDM and should never
 *     be freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Sync, tip_SetName
 *   >:&:
 */
CDM_EXPORT
char *CDM_SetName(CDM_Collection Collection, char *Name) {
  if (Collection == NULL) {
    CDM_ReportError("CDM_SetName: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_SetName: Invalid Collection Type");
    return NULL;
  }
  if (Collection->Name != NULL) Tcl_Free(Collection->Name);
  Collection->Name = CDM_Strdup(Name);
  Collection->Modified = 1;
  return Collection->Name;
} /* CDM_SetName */

/*
 * ---- CDM_SetOwner:
 *  DESCRIPTION:
 *   <English:
 *     This function will change the Owner of the specified Collection.
 *     The value of the "Owner" parameter will not modified in
 *     any way by CDM, as a local copy will be created and modified internally
 *     by the CDM.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.                     (CDM_Collection)
 *     Owner:       The desired new Owner of the Collection. (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SetOwner (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a pointer (of type char*) to the string buffer
 *     that holds the new Collection Owner. This pointer will be at a different
 *     memory location than the given parameter value, as the CDM will create
 *     and manipulate a local copy.
 *
 *     In case of an error, NULL will be returned and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp).
 *
 *     Note that the returned pointer is property of the CDM and should never
 *     be freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_SetOwner
 *   >:&:
 */
CDM_EXPORT
char *CDM_SetOwner(CDM_Collection Collection, char *Owner) {
  if (Collection == NULL) {
    CDM_ReportError("CDM_SetOwner: NULL Collection Pointer");
    return NULL;
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    CDM_ReportError("CDM_SetOwner: Invalid Collection Type");
    return NULL;
  }
  if (Collection->Owner != NULL) Tcl_Free(Collection->Owner);
  Collection->Owner = CDM_Strdup(Owner);
  Collection->Modified = 1;
  return Collection->Owner;
} /* CDM_SetOwner */

/*
 * ---- CDM_Sync:
 *  DESCRIPTION:
 *   <English:
 *     This function will save a Collection object in disk. If the Collection
 *     object has never been saved again, then this function will create the
 *     directory that the Name of the Collection specifies. A representation
 *     of the Collection will be saved in this directory. If the directory
 *     already exists it will be overwritten. If the directory cannot be
 *     created an error condition will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.           (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     The user is not expected to modify directories that hold Collections in
 *     any way. Modifing any of the files under the Collection directory may
 *     result in a corrupted, unreadble Collection.
 *
 *     This function is equivelant to tip_Sync (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Collection is successfully saved, TCL_OK will
 *     be returned. In case of an error, the return value will be TCL_ERROR and
 *     an error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Sync
 *   >:&:
 */
CDM_EXPORT
int CDM_Sync(CDM_Collection Collection) {
  if (Collection == NULL) {
    return CDM_ReportError("CDM_Sync: NULL Collection Pointer");
  }
  if (Collection->Type != CDM_TYPE_COLLECTION) {
    return CDM_ReportError("CDM_Sync: Invalid Collection Type");
  }
  return CDM_SaveCollection(Collection);
} /* CDM_Sync */

/*
 * ---- CDM_Status:
 *  DESCRIPTION:
 *   <English:
 *    This function will return information about the specified Collection.
 *    This information will be stored as text in a newly created Tcl object.
 *    CDM does not own the returned object: the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *    associated with it.
 *
 *    This function will always create and return a new Tcl string object.
 *    In case of an error, NULL will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.           (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a Tcl string object (of type Tcl_Obj *),
 *    containing the information of the specified Collection.
 *    The caller is responsible to free this object through Tcl_DecrRefCount
 *    or CDM_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Free, Tcl_DecrRefCount
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequence CDM_Status(CDM_Collection  Collection) {
  CDM_Document Doc;
  CDM_ListNode *ptr;
  Tcl_Obj *Info, *DocumentIDs, *DocumentXIDs, *OpenedDocumentIDs, *strObj,
          *Attrs;
  const char *format, *Doc_Cmd;
  int i;
  CDM_THREAD_SPECIFIC_DATA;

  if (Collection == NULL) return NULL;
  if (Collection->Type != CDM_TYPE_COLLECTION) return NULL;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_Status: Aborting...");
    return NULL;
  }

  Info = Tcl_NewStringObj(NULL, 0);
  switch (Collection->FileFormat) {
    case CDM_FORMAT_NATIVE:
    case CDM_FORMAT_NATIVE2: {format = "native"; break;}
    case CDM_FORMAT_GATE_1x: {format = "gate";   break;}
    case CDM_FORMAT_GATE_2x: {format = "gate2";  break;}
    default:                 {
      /*
       * A registered format?
       */
      format = "?";
      i = Collection->FileFormat - CDM_FORMAT_REGISTERED;
      if (i >= 0 && i < CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER &&
          CDM_AdditionalCollectionTypes[i].CollectionType != NULL) {
        format = CDM_AdditionalCollectionTypes[i].CollectionType;
      }
      break;
    }
  }

  /*
   * Create the required objects from the Document list...
   */
  ptr               = Collection->Documents;
  DocumentIDs       = CDM_ListGetObject(ptr, CDM_DOCUMENT_ID);
  DocumentXIDs      = CDM_ListGetObject(ptr, CDM_DOCUMENT_XID);
  OpenedDocumentIDs = Tcl_NewListObj(0, NULL);
  while (ptr != NULL) {
    if (ptr->ptr[CDM_DOCUMENT_OPENED] != NULL) {
      /*
       * First, add the Document Id...
       */
      strObj = Tcl_NewStringObj(ptr->ptr[CDM_DOCUMENT_ID], -1);
      if (strObj == NULL) break;
      /*
       * And then append the corresponding Tcl command...
       */
      Doc = (CDM_Document) ptr->ptr[CDM_DOCUMENT_POINTER];
      if (Doc != NULL && Doc->Type == CDM_TYPE_DOCUMENT) {
        Doc_Cmd = Tcl_GetCommandName(CDM_Interp, Doc->Tcl_Cmd);
        if (Doc_Cmd == NULL) Doc_Cmd = "(NULL)";
      } else {
        Doc_Cmd = "(ERROR)";
      }
      Tcl_AppendToObj(strObj, " ", 1);
      Tcl_AppendToObj(strObj, Doc_Cmd, -1);
      Tcl_ListObjAppendElement(NULL, OpenedDocumentIDs, strObj);
    }
    ptr = ptr->next;
  }

  Attrs = CDM_GetAttributes(Collection);
  Tcl_AppendStringsToObj(Info,
      "Name: ",              Collection->Name,
      "\nEncoding: ",        Collection->Encoding,
      "\nTcl Command: ",
      Tcl_GetCommandName(Collection->Interp, Collection->Tcl_Cmd),
      "\nOwner: ",           Collection->Owner,
      "\nFormat: ",          format,
      "\nComment: ",         (Collection->CommentObj) ?
                             Tcl_GetString(Collection->CommentObj) : "",
      "\nDocumentIds:\n  ",  Tcl_GetString(DocumentIDs),
      "\nDocumentXIDs:\n  ", Tcl_GetString(DocumentXIDs),
      "\nModified: ",
      ((Collection->Modified+Collection->ModifiedDocuments)? "(Yes)" : "(No)"),
      "\nAttributes:\n  ",   (Attrs) ? Tcl_GetString(Attrs) : "",
      "\nOpened Document Ids:\n  ",
                             Tcl_GetString(OpenedDocumentIDs),
      (char *) NULL);
  if (Attrs)             Tcl_DecrRefCount(Attrs);
  if (DocumentIDs)       Tcl_DecrRefCount(DocumentIDs);
  if (DocumentXIDs)      Tcl_DecrRefCount(DocumentXIDs);
  if (OpenedDocumentIDs) Tcl_DecrRefCount(OpenedDocumentIDs);
  return Info;
} /* CDM_Status */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |  E N D   O F    C O L L E C T I O N    S E C T I O N  |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     D O C U M E N T                S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

extern "C" int CDM_DocumentCallback(CDM_TCL_STD_ARGS) {
  CDM_Document Document = (CDM_Document ) clientData;
  Tcl_Obj *ObjPtr;
  char    *CPtr;
  int index, Integer;
  long Long, Long2;

  static const char *Methods[] = {
    "AddAnnotation",           "AnnotationsAt",
    "AnnotationsContaining",   "AnnotationsInRange",
    "AnnotationsMatchingRange",
    "ByteSequenceInsertString","ByteSequenceReplace",
    "ByteSequenceReplaceCharacters",
    "Close",
    "DeleteAnnotations",       "DisplaceAnnotations",
    "FindMaxUsedAnnotationId", "FirstAnnotationContaining",
    "Format",
    "GetAnnotatedTextRanges",  "GetAnnotation",
    "GetAnnotations",          "GetAnnotationAttributeNames",
    "GetAnnotationAttributeValues",
    "GetAnnotationTypes",      "GetAttribute",
    "GetAttributes",           "GetByteSequence",         "GetEncoding",
    "GetExternalId",           "GetFirstAnnotatedTextRange",
    "GetFirstAnnotation",      "GetId",
    "GetNextAnnotation",       "GetParent",
    "GetRawData",              "Modified",                "NextAnnotations",
    "PutAttribute",
    "RawDataInsertString",     "RawDataReplace",
    "RawDataReplaceCharacters",
    "ReferenceCount",          "RemoveAnnotation",
    "RemoveAttribute",
    "SelectAnnotations",       "SelectAnnotationsSorted", "SetByteSequence",
    "SetEncoding",             "SetExternalId",           "SetRawData",
    "Status",                  "Sync",                    (char *) NULL
  };
  enum methods {
     ADDANNOTATION,             ANNOTATIONSAT,
     ANNOTATIONSCONTAINING,     ANNOTATIONSINRANGE,
     ANNOTATIONSMATCHINGRANGE,
     BYTESEQUENCEINSERTSTRING,  BYTESEQUENCEREPLACE,
     BYTESEQUENCEREPLACECHARACTERS,
     CLOSE,
     DELETEANNOTATIONS,         DISPLACEANNOTATIONS,
     FINDMAXUSEDANNOTATIONID,   FIRSTANNOTATIONCONTAINING,
     FORMAT,
     GETANNOTATEDTEXTRANGES,    GETANNOTATION,
     GETANNOTATIONS,            GETANNOTATIONATTRIBUTENAMES,
     GETANNOTATIONATTRIBUTEVALUES,
     GETANNOTATIONTYPES,        GETATTRIBUTE,
     GETATTRIBUTES,             GETBYTESEQUENCE,          GETENCODING,
     GETEXTERNALID,             GETFIRSTANNOTATEDTEXTRANGE,
     GETFIRSTANNOTATION,        GETID,
     GETNEXTANNOTATION,         GETPARENT,
     GETRAWDATA,                MODIFIED,                 NEXTANNOTATIONS,
     PUTATTRIBUTE,
     RAWDATAINSERTSTRING,       RAWDATAREPLACE,
     RAWDATAREPLACECHARACTERS,
     REFERENCECOUNT,            REMOVEANNOTATION,
     REMOVEATTRIBUTE,
     SELECTANNOTATIONS,         SELECTANNOTATIONSSORTED,  SETBYTESEQUENCE,
     SETENCODING,               SETEXTERNALID,            SETRAWDATA,
     STATUS,                    SYNC
  };
  CDM_THREAD_SPECIFIC_DATA;

  CDM_Interp = interp;
  Tcl_ResetResult(interp);

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    return CDM_ReportError("CDM_CollectionCallback: Aborting...");
  }

  if (Document->Type != CDM_TYPE_DOCUMENT)
    return CDM_ReportError("CDM_DocumentCallback: Invalid Document Type");
  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
    return TCL_ERROR;
  }
  if (Tcl_GetIndexFromObj(interp, objv[1], (const char **) Methods,
                          "method",0,&index) != TCL_OK) return TCL_ERROR;


  /* Handle Arguments... */
  switch ((enum methods) index) {
    case ADDANNOTATION: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Annotation"); return TCL_ERROR;}
      Long = CDM_AddAnnotation(Document, objv[2]);
      if (Long < 0) return TCL_ERROR;
      ObjPtr = ObjectCache.FindLong(Long);
      if (ObjPtr == NULL)
          CDM_Panic("CDM_DocumentCallback: ObjectCache::FindLong returned NULL"
                    " - out of memory?\n");
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case ANNOTATIONSAT: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Position"); return TCL_ERROR;}
      if (Tcl_GetLongFromObj(interp, objv[2], &Long) != TCL_OK)
        return TCL_ERROR;
      ObjPtr = CDM_AnnotationsAt(Document, Long);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case ANNOTATIONSCONTAINING: {
      if (objc != 3 && objc != 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "position ?position2?");
        return TCL_ERROR;}
      if (Tcl_GetLongFromObj(interp, objv[2], &Long) != TCL_OK)
        return TCL_ERROR;
      if (objc == 4) {
        if (Tcl_GetLongFromObj(interp, objv[3], &Long2) != TCL_OK)
        return TCL_ERROR;
        ObjPtr = CDM_AnnotationsContaining(Document, Long, Long2);
      } else {
        ObjPtr = CDM_AnnotationsContaining(Document, Long);
      }
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case ANNOTATIONSINRANGE: {
      long start, end;
      if (objc != 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "range-start range-end");
        return TCL_ERROR;
      }
      if (Tcl_GetLongFromObj(interp, objv[2], &start) != TCL_OK)
        return TCL_ERROR;
      if (Tcl_GetLongFromObj(interp, objv[2], &end)   != TCL_OK)
        return TCL_ERROR;
      ObjPtr = CDM_AnnotationsInRange(Document, start, end);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case ANNOTATIONSMATCHINGRANGE: {
      long start, end;
      if (objc != 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "range-start range-end");
        return TCL_ERROR;
      }
      if (Tcl_GetLongFromObj(interp, objv[2], &start) != TCL_OK)
        return TCL_ERROR;
      if (Tcl_GetLongFromObj(interp, objv[2], &end)   != TCL_OK)
        return TCL_ERROR;
      ObjPtr = CDM_AnnotationsMatchingRange(Document, start, end);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case BYTESEQUENCEINSERTSTRING:
    case RAWDATAINSERTSTRING: {
      if (objc != 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "position string");
        return TCL_ERROR;}
      if (Tcl_GetLongFromObj(interp, objv[2], &Long) != TCL_OK)
        return TCL_ERROR;
      ObjPtr = CDM_ByteSequenceInsertString(Document, Long,
                                            Tcl_GetString(objv[3]));
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    } /* BYTESEQUENCEINSERTSTRING, RAWDATAINSERTSTRING */
    case BYTESEQUENCEREPLACE:
    case RAWDATAREPLACE: {
      if (objc != 4 && objc != 5) {
        Tcl_WrongNumArgs(interp, 2, objv, "first last ?newstring?");
        return TCL_ERROR;}
      if (Tcl_GetLongFromObj(interp, objv[2], &Long)  != TCL_OK)
        return TCL_ERROR;
      if (Tcl_GetLongFromObj(interp, objv[3], &Long2) != TCL_OK)
        return TCL_ERROR;
      if (objc == 5) CPtr = Tcl_GetString(objv[4]);
      else           CPtr = NULL;
      ObjPtr = CDM_ByteSequenceReplace(Document, Long, Long2, CPtr);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    } /* BYTESEQUENCEREPLACE, RAWDATAREPLACE */
    case BYTESEQUENCEREPLACECHARACTERS:
    case RAWDATAREPLACECHARACTERS: {
      if (objc != 5) {
        Tcl_WrongNumArgs(interp, 2, objv, "position number_of_chars_to_replace"
                                          " replacement_string");
        return TCL_ERROR;}
      if (Tcl_GetLongFromObj(interp, objv[2], &Long) != TCL_OK)
        return TCL_ERROR;
      if (Tcl_GetIntFromObj(interp, objv[3], &Integer) != TCL_OK)
        return TCL_ERROR;
      ObjPtr = CDM_ByteSequenceReplaceCharacters(Document, Long, Integer,
                                            Tcl_GetString(objv[4]));
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    } /* BYTESEQUENCEREPLACECHARACTERS, RAWDATAREPLACECHARACTERS */
    case CLOSE: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
        return CDM_Close(Document);
    }
    case DELETEANNOTATIONS: {
      if (objc < 3 || objc > 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "Type ?Constraints?");
        return TCL_ERROR;}
      if (objc == 3) {
        Integer = CDM_DeleteAnnotations(Document, Tcl_GetString(objv[2]));
      } else {
        Integer = CDM_DeleteAnnotations(Document, Tcl_GetString(objv[2]),
                                                  objv[3]);
      }
      if (Integer < 0) return TCL_ERROR;
      ObjPtr = CDM_NewLongObj((long) Integer);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case DISPLACEANNOTATIONS: {
      long offset, displacement;
      int status;
      if (objc != 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "offset displacement");
        return TCL_ERROR;}
      status = Tcl_GetLongFromObj(interp, objv[2], &offset);
      if (status != TCL_OK) return status;
      status = Tcl_GetLongFromObj(interp, objv[3], &displacement);
      if (status != TCL_OK) return status;
      status = CDM_DisplaceAnnotations(Document, offset, displacement);
      return status;
    } /* DISPLACEANNOTATIONS */
    case FINDMAXUSEDANNOTATIONID: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "Document");
        return TCL_ERROR;}
      Long = CDM_FindMaxUsedAnnotationId(Document);
      if (Long < 0) return TCL_ERROR;
      Tcl_SetObjResult(interp, CDM_NewLongObj(Long));
      return TCL_OK;
    } /* FINDMAXUSEDANNOTATIONID */
    case FIRSTANNOTATIONCONTAINING: {
      long position1, position2;
      int status;
      if (objc < 3 || objc > 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "position ?position2?");
        return TCL_ERROR;}
      status = Tcl_GetLongFromObj(interp, objv[2], &position1);
      if (status != TCL_OK) return status;
      if (objc == 4) {
        status = Tcl_GetLongFromObj(interp, objv[3], &position2);
        if (status != TCL_OK) return status;
        ObjPtr = CDM_FirstAnnotationContaining(Document, position1, position2);
      } else {
        ObjPtr = CDM_FirstAnnotationContaining(Document, position1);
      }
      if (ObjPtr == NULL) {
        /* We should report an error only if we have an error message */
        if (!CDM_ActiveInterpreterResultEmpty()) return TCL_ERROR;
      } else {
        Tcl_SetObjResult(interp, ObjPtr);
      }
      return TCL_OK;
    } /* FIRSTANNOTATIONCONTAINING */
    case FORMAT: {
      if (objc < 2 || objc > 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "?native/gate/gate2/registered?");
        return TCL_ERROR;}
      if (objc == 3) {
        if (CDM_Strcmp(Tcl_GetString(objv[2]), "gate2") == 0) {
          Document->FileFormat = CDM_FORMAT_GATE_2x;
        } else if (CDM_Strcmp(Tcl_GetString(objv[2]), "gate") == 0) {
          Document->FileFormat = CDM_FORMAT_GATE_1x;
        } else if (CDM_Strcmp(Tcl_GetString(objv[2]), "native") == 0) {
          Document->FileFormat = CDM_FORMAT_NATIVE2;
        } else {
          /*
           * A registered format?
           */
          int format;
          Document->FileFormat = CDM_FORMAT_NATIVE2;
          for (format=0; format<CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER;
               format++) {
            if (CDM_AdditionalCollectionTypes[format].CollectionType != NULL &&
                CDM_Strcmp(CDM_AdditionalCollectionTypes[format].CollectionType,
                           Tcl_GetString(objv[2])) == 0) {
              Document->FileFormat = format + CDM_FORMAT_REGISTERED;
              break;
            }
          }
        }
      }
      switch (Document->FileFormat) {
        case CDM_FORMAT_NATIVE:
        case CDM_FORMAT_NATIVE2: {
          ObjPtr = CDM_NewStringObj("native"); break;
        }
        case CDM_FORMAT_GATE_1x: {
          ObjPtr = CDM_NewStringObj("gate");   break;
        }
        case CDM_FORMAT_GATE_2x: {
          ObjPtr = CDM_NewStringObj("gate2");   break;
        }
        default: {
          /*
           * A registered format?
           */
          int format = Document->FileFormat - CDM_FORMAT_REGISTERED;
          if (format >= 0 && format < CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER &&
              CDM_AdditionalCollectionTypes[format].CollectionType != NULL) {
            ObjPtr = CDM_NewStringObj(
               CDM_AdditionalCollectionTypes[format].CollectionType);
          } else {
            ObjPtr = CDM_NewStringObj("native");
          }
          break;
        }
      }
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    } /* FORMAT */
    case GETANNOTATEDTEXTRANGES: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Annotation"); return TCL_ERROR;}
      ObjPtr = CDM_GetAnnotatedTextRanges(Document, objv[2]);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    } /* GETANNOTATEDTEXTRANGES */
    case GETANNOTATION: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Annotation-Id"); return TCL_ERROR;}
      if (Tcl_GetLongFromObj(interp, objv[2], &Long) != TCL_OK)
        return TCL_ERROR;
      ObjPtr = CDM_GetAnnotation(Document, Long);
      if (ObjPtr == NULL) return TCL_ERROR;
      if (CDM_FullGateCompatibility) {
        /* Gate wraps the annotation inside a list. Emulate that bug... */
        Tcl_Obj *obj[1];
        obj[0] = ObjPtr;
        ObjPtr = Tcl_NewListObj(1, obj);
      }
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    } /* GETANNOTATION */
    case GETANNOTATIONS: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      ObjPtr = CDM_GetAnnotations(Document);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    } /* GETANNOTATIONS */
    case GETANNOTATIONATTRIBUTENAMES: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "annotation-type"); return TCL_ERROR;}
      ObjPtr = CDM_GetAnnotationAttributeNames(Document, objv[2]);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    } /* GETANNOTATIONATTRIBUTENAMES */
    case GETANNOTATIONATTRIBUTEVALUES: {
      if (objc != 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "annotation-type attribute-name");
        return TCL_ERROR;}
      ObjPtr = CDM_GetAnnotationAttributeValues(Document, objv[2], objv[3]);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case GETANNOTATIONTYPES: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      ObjPtr = CDM_GetAnnotationTypes(Document);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    } /* GETANNOTATIONTYPES */
    case GETATTRIBUTE: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Attribute-Name"); return TCL_ERROR;}
      ObjPtr = CDM_GetAttribute(Document, Tcl_GetString(objv[2]));
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    } /* GETATTRIBUTE */
    case GETATTRIBUTES: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      ObjPtr = CDM_GetAttributes(Document);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    } /* GETATTRIBUTES */
    case GETBYTESEQUENCE: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      ObjPtr = CDM_GetByteSequence(Document);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    } /* GETBYTESEQUENCE */
    case GETENCODING: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      CPtr = CDM_GetEncoding(Document);
      if (CPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, CDM_NewStringObj(CPtr));
      return TCL_OK;
    } /* GETENCODING */
    case GETEXTERNALID: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      CPtr = CDM_GetExternalId(Document);
      if (CPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, Tcl_NewStringObj(CPtr, -1));
      return TCL_OK;
    } /* GETEXTERNALID */
    case GETFIRSTANNOTATEDTEXTRANGE: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Annotation"); return TCL_ERROR;}
      ObjPtr = CDM_GetFirstAnnotatedTextRange(Document, objv[2]);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    } /* GETFIRSTANNOTATEDTEXTRANGE */
    case GETFIRSTANNOTATION: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Annotation-type"); return TCL_ERROR;}
      ObjPtr = CDM_GetFirstAnnotation(Document, Tcl_GetString(objv[2]));
      if (ObjPtr == NULL) return TCL_ERROR;
      if (CDM_FullGateCompatibility) {
        /* Gate wraps the annotation inside a list. Emulate that bug... */
        Tcl_Obj *tobjv[1];
        tobjv[0] = ObjPtr;
        ObjPtr = Tcl_NewListObj(1, tobjv);
      }
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case GETID: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      CPtr = CDM_GetId(Document);
      if (CPtr == NULL) return TCL_ERROR;
      Tcl_SetResult(interp, CPtr, TCL_VOLATILE);
      return TCL_OK;
    }
    case GETNEXTANNOTATION: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      ObjPtr = CDM_GetNextAnnotation(Document);
      if (ObjPtr == NULL) return TCL_ERROR;
      if (CDM_FullGateCompatibility) {
        /* Gate wraps the annotation inside a list. Emulate that bug... */
        Tcl_Obj *tobjv[1];
        tobjv[0] = ObjPtr;
        ObjPtr = Tcl_NewListObj(1, tobjv);
      }
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case GETPARENT: {
      CDM_Collection Parent;
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      Parent = CDM_GetParent(Document);
      if (Parent == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, CDM_NewStringObj(
                    Tcl_GetCommandName(interp, Parent->Tcl_Cmd)));
      return TCL_OK;
    }
    case GETRAWDATA: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      ObjPtr = CDM_GetRawData(Document);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case MODIFIED: {
      if (objc < 2 || objc > 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "?boolean?");
        return TCL_ERROR;}
      if (objc == 3) {
        if (Tcl_GetBoolean(interp, Tcl_GetString(objv[2]), &Integer) != TCL_OK)
          return TCL_ERROR;
        Document->Modified = (CDM_UCHAR) Integer;
      }
      if (Document->Modified)
         ((CDM_Collection) Document->ParentPtr)->ModifiedDocuments = 1;
      ObjPtr = CDM_NewLongObj((long) Document->Modified);
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case NEXTANNOTATIONS: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Position"); return TCL_ERROR;}
      if (Tcl_GetLongFromObj(interp, objv[2], &Long) != TCL_OK)
        return TCL_ERROR;
      ObjPtr = CDM_NextAnnotations(Document, Long);
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case PUTATTRIBUTE: {
      if (objc < 3 || objc > 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "Attribute ?Undefined?");
        return TCL_ERROR;}
      return CDM_PutAttribute(Document, objv[2]);
    }
    case REFERENCECOUNT: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      Tcl_SetObjResult(interp, Tcl_NewIntObj((int) Document->RefCount));
      return TCL_OK;
    }
    case REMOVEANNOTATION: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Annotation-Id"); return TCL_ERROR;}
      if (Tcl_GetLongFromObj(interp, objv[2], &Long) != TCL_OK)
        return TCL_ERROR;
      return CDM_RemoveAnnotation(Document, Long);
    }
    case REMOVEATTRIBUTE: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Attribute-Name"); return TCL_ERROR;}
      return CDM_RemoveAttribute(Document, Tcl_GetString(objv[2]));
    }
    case SELECTANNOTATIONS: {
      if (objc < 3 || objc > 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "Type ?Constraints?");
        return TCL_ERROR;}
      if (objc == 3) {
        ObjPtr = CDM_SelectAnnotations(Document, Tcl_GetString(objv[2]));
      } else {
        ObjPtr = CDM_SelectAnnotations(Document, Tcl_GetString(objv[2]),
                                                 Tcl_GetString(objv[3]));
      }
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case SELECTANNOTATIONSSORTED: {
      if (objc < 3 || objc > 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "Type ?Constraints?");
        return TCL_ERROR;}
      if (objc == 3) {
        ObjPtr = CDM_SelectAnnotationsSorted(Document, Tcl_GetString(objv[2]));
      } else {
        ObjPtr = CDM_SelectAnnotationsSorted(Document, Tcl_GetString(objv[2]),
                                                       Tcl_GetString(objv[3]));
      }
      if (ObjPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, ObjPtr);
      return TCL_OK;
    }
    case SETBYTESEQUENCE: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Text"); return TCL_ERROR;}
      return CDM_SetRawData(Document, objv[2]);
    }
    case SETRAWDATA: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Text"); return TCL_ERROR;}
      return CDM_SetByteSequence(Document, objv[2]);
    }
    case SETENCODING: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "Encoding"); return TCL_ERROR;}
      CPtr = CDM_SetEncoding(Document, Tcl_GetString(objv[2]));
      if (CPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, CDM_NewStringObj(CPtr));
      return TCL_OK;
    }
    case SETEXTERNALID: {
      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "External-Id"); return TCL_ERROR;}
      CPtr = CDM_SetExternalId(Document, Tcl_GetString(objv[2]));
      if (CPtr == NULL) return TCL_ERROR;
      Tcl_SetObjResult(interp, Tcl_NewStringObj(CPtr, -1));
      return TCL_OK;
    }
    case STATUS: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      Tcl_SetObjResult(interp, CDM_Status(Document));
      return TCL_OK;
    }
    case SYNC: {
      if (objc != 2) {
        Tcl_WrongNumArgs(interp, 2, objv, ""); return TCL_ERROR;}
      return CDM_Sync(Document);
    }
  }
  return TCL_OK;
} /* CDM_DocumentCallback */

/*
 * ---- CDM_AddAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will add the given Annotation (specified by the "Ann"
 *     parameter) into the Document specified by the "Document" parameter.
 *     If the Annotation does not already have an Id, a new unique Id will be
 *     assigned to that Annotation and the Annotation will be added to the
 *     current Annotation set of the Document.
 *
 *     If the Annotation already has an Id, then this Id will left intact.
 *     If an Annotation with the same Id already exists in the current
 *     Annotation set, then the new Annotation will replace the old one. If
 *     an Annotation with the same Id does not exist, the given Annotation will
 *     be appended to the existing Annotation set.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Ann:         The Annotation to be added.   (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AddAnnotation (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns the Id of the Annotation that was added to the
 *     specified Document object. Whether the Annotation had or had not an Id
 *     makes no difference to the returned value. Note that after using this
 *     function, CDM will keep a reference to the given Annotation. This will
 *     transform the given Annotation into a [[FONT COLOR=red]]shared[[/FONT]]
 *     object, that cannot be manipulated any more by the caller. Always
 *     remember that an attempt to modify a shared object will abort the
 *     execution of the whole platform. If the caller wants to modify a shared
 *     object, the following steps must be followed:
 *     [[LI]] A check should be made in order to find if an object is shared
 *     or not, by using the Tcl function Tcl_IsShared.[[/LI]]
 *     [[LI]] If the object is not shared (i.e. Tcl_IsShared returns an
 *     integer value < 2), the object can be safely modified.[[/LI]]
 *     [[LI]] If the object is shared, a new copy of the object must be
 *     created by using the Tcl function Tcl_DuplicateObj. The new object will
 *     be an exact copy of the original one but it will not be shared. The
 *     returned object can be safely modified.[[/LI]]
 *
 *     In case of an error the return value will be -1 and
 *     an error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     Tcl_IsShared,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DuplicateObj[[/A]],
 *     tip_AddAnnotation
 *   >:&:
 */
CDM_EXPORT
long CDM_AddAnnotation(CDM_Document Document, CDM_Annotation Ann) {
#ifndef DONT_HASH_DOC_ANNOTATIONS
  register Tcl_HashEntry *entryPtr;
  register Tcl_Obj *Obj;
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  Tcl_Obj *value, *attributes;
  register long Id;
  int items, incr_ref_count = 1;
  CDM_THREAD_SPECIFIC_DATA;

#ifdef DEBUG
  CDM_DEBUG("CDM_AddAnnotation: Adding Annotation \"%s\"\n",
             Tcl_GetString(Ann));
#endif /* DEBUG */

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AddAnnotation: Aborting...");
    return -1;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_AddAnnotation: NULL Document Pointer");
    return -1;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_AddAnnotation: Invalid Document Type");
    return -1;
  }

  if (!CDM_ValidAnnotation(Ann)) {
    return -1;
  }

  /*
   * We increament the reference count of the provided Annotation. When we
   * finish with the Annotation, we will decreament the reference count. This
   * will free the Annotation object, if the caller has not kept it...
   */
  Tcl_IncrRefCount(Ann);

  /*
   * Ckeck for Annotation's Id field...
   */
  Id = CDM_GetId(Ann);

  /*
   * Does the Annotation already has an ID?
   */
  if (Id == -1) {
    Id = Document->AnnotationMaxID;
    Document->AnnotationMaxID++;
    if (Tcl_IsShared(Ann)) {
      #ifdef DUPLICATE_DEBUG
      CDM_Debug("AddAnnotation: Duplicating (%d): %s\n", Ann->refCount,
             Tcl_GetString(Ann));
      #endif /* DUPLICATE_DEBUG */
      Tcl_DecrRefCount(Ann);
      Ann = Tcl_DuplicateObj(Ann);
      Tcl_IncrRefCount(Ann);
    }

    /*
     * Now the Annotation kept in Ann is not shared. We can freely modify
     * it...
     */
#ifdef ENABLE_ANNOTATION_OBJECT
    if (ANNOTATION_OBJ(Ann)) {
      CDMAnnObj_SetId(Ann, Id);
    } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
      value = ObjectCache.FindLong(Id);
      if (Tcl_ListObjReplace(CDM_Interp, Ann, 0, 1, 1, &value) != TCL_OK) {
        CDM_ReportError("CDM_AddAnnotation: Cannot Update Annotation ID (%s)",
                        Tcl_GetString(Ann));
        Tcl_DecrRefCount(Ann);
        return -1;
      }
#ifdef ENABLE_ANNOTATION_OBJECT
    }
#endif /* ENABLE_ANNOTATION_OBJECT */
  } else {
    if (Id >= Document->AnnotationMaxID) Document->AnnotationMaxID = Id+1;
  }

  /*
   * If memory compression is enabled, we have to check whether the same
   * attribute set is used by some other annotation in order to share the
   * previous one and free this...
   */
  if (CDM_EnableMemoryCompression) {
    attributes = CDM_GetAttributes(Ann);
    value = CDM_SearchForAttributeSet(attributes);
    if (value && value != attributes) {
      /*
       * Bug found by Petasis on 14/06/2002: If the Annotation already had an
       * Id, we have not checked it whether it was shared or not...
       */
      if (Tcl_IsShared(Ann)) {
        #ifdef DUPLICATE_DEBUG
        CDM_Debug("AddAnnotation: Duplicating (%d): %s\n", Ann->refCount,
               Tcl_GetString(Ann));
        #endif /* DUPLICATE_DEBUG */
        Tcl_DecrRefCount(Ann);
        Ann = Tcl_DuplicateObj(Ann);
        Tcl_IncrRefCount(Ann);
      }
#ifdef ENABLE_ANNOTATION_OBJECT
      if (ANNOTATION_OBJ(Ann)) {
        CDMAnnObj_SetAttributes(Ann, value);
      } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
        Tcl_ListObjReplace(CDM_Interp, Ann, 3, 1, 1, &value);
#ifdef ENABLE_ANNOTATION_OBJECT
      }
#endif /* ENABLE_ANNOTATION_OBJECT */
    }
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  if (!ANNOTATION_OBJ(Ann)) {
    Tcl_ConvertToType(CDM_Interp, Ann, &cdmAnnotationType);
  }
  CDMAnnObj_FreeKeptSpanSet(Ann);
#endif /* ENABLE_ANNOTATION_OBJECT */


#ifdef DONT_HASH_DOC_ANNOTATIONS
  (Document->Annotations_List)->Set(Ann, Id);
  Tcl_IncrRefCount(Ann);
#else  /* DONT_HASH_DOC_ANNOTATIONS */
  #ifdef CDM_ONE_WORD_KEYS
  entryPtr = Tcl_CreateHashEntry(&(Document->Annotations_HashTable), (char *)Id,
                                 &items);
  #else /* CDM_ONE_WORD_KEYS */
  snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", Id);
  entryPtr = Tcl_CreateHashEntry(&(Document->Annotations_HashTable), CDM_Temp,
                                 &items);
  #endif /* CDM_ONE_WORD_KEYS */
  if (entryPtr == NULL) {
    Tcl_DecrRefCount(Ann);
    return -1;
  }
  if (items == 0) {
    Obj = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    // Petasis, 31/12/1999: If we have changed a non shared annotation, we may
    // free (!) the changed annotation and then add the invalid pointer as a
    // valid one...
    // This was a quite impressive bug, that even "purify" failed to
    // identify. It was very strange that CDM didn't crash even at very
    // difficult situation, as the one that Ion has (where memory usage was
    // very large, more than 1.2 GB, and OS was very memory stressed)
    // The bug was revealed when all "strcpy"s was substituted by "CDM_Strcpy"
    // and "CDM_Strcpy"'s behaviour was changed to be fully compatible with
    // "strcpy" (SVID 3, POSIX, BSD 4.3, ISO 9899). (Under tcl 8.3b1)
    if (Obj != Ann) {
    //^^^^^^^^^^^^^^
      Tcl_DecrRefCount(Obj);
    } else {
      incr_ref_count = 0;
    }
  }
  Tcl_SetHashValue(entryPtr, (ClientData) Ann);
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  if (incr_ref_count) Tcl_IncrRefCount(Ann);
  Document->Modified = 1;
  ((CDM_Collection) Document->ParentPtr)->ModifiedDocuments = 1;
  Tcl_DecrRefCount(Ann);
  return Id;
} /* CDM_AddAnnotation */

void
CDM_DocumentInternalAddAnnotation(CDM_Document Document, CDM_Annotation Ann) {
#ifndef DONT_HASH_DOC_ANNOTATIONS
  register Tcl_HashEntry *entryPtr;
  register Tcl_Obj *Obj;
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  Tcl_Obj **elements;
  long Id;
  int items;
  CDM_THREAD_SPECIFIC_DATA;

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Ann)) {
    Id = CDMAnnObj_GetId(Ann);
    if (Id < 0) {
      CDM_ReportError("Malformed Annotation: Invalid Id in (%s)\nAborting!",
                      Tcl_GetString(Ann));
      CDM_Panic(CDM_Temp);
    }
  } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
    Tcl_ListObjGetElements(NULL, Ann, &items, &elements);
    if (Tcl_GetCharLength(elements[0]) == 0) {
      CDM_ReportError("CDM_DocumentInternalAddAnnotation: Invalid Annotation "
                      "\"%s\"\nAborting!", Tcl_GetString(Ann));
      CDM_Panic(CDM_Temp);
    }
    if (Tcl_GetLongFromObj(NULL, elements[0], &Id) != TCL_OK) {
      CDM_ReportError("Malformed Annotation: Invalid Id \"%s\" (%s)\nAborting!"
                      , Tcl_GetString(elements[0]), Tcl_GetString(Ann));
      CDM_Panic(CDM_Temp);
    }
#ifdef ENABLE_ANNOTATION_OBJECT
    Tcl_ConvertToType(CDM_Interp, Ann, &cdmAnnotationType);
  }
  CDMAnnObj_FreeKeptSpanSet(Ann);
#endif /* ENABLE_ANNOTATION_OBJECT */
  if (Id >= Document->AnnotationMaxID) Document->AnnotationMaxID = Id+1;

#ifdef DONT_HASH_DOC_ANNOTATIONS
  (Document->Annotations_List)->Set(Ann, Id);
  Tcl_IncrRefCount(Ann);
#else  /* DONT_HASH_DOC_ANNOTATIONS */
  #ifdef CDM_ONE_WORD_KEYS
  entryPtr = Tcl_CreateHashEntry(&(Document->Annotations_HashTable), (char *)Id,
                                 &items);
  #else /* CDM_ONE_WORD_KEYS */
  snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", Id);
  entryPtr = Tcl_CreateHashEntry(&(Document->Annotations_HashTable), CDM_Temp,
                                 &items);
  #endif /* CDM_ONE_WORD_KEYS */
  if (entryPtr == NULL) {
    return;
  }
  if (items == 0) {
    Obj = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    if (Obj != Ann) {
      Tcl_DecrRefCount(Obj);
      Tcl_IncrRefCount(Ann);
    }
  } else {
    /* This is a new entry. Keep the object... */
    Tcl_IncrRefCount(Ann);
  }
  Tcl_SetHashValue(entryPtr, (ClientData) Ann);
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  Document->Modified = 1;
  ((CDM_Collection) Document->ParentPtr)->ModifiedDocuments = 1;
} /* CDM_DocumentInternalAddAnnotation */
/*
 * ---- CDM_AnnotationsAt:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that have the specified text position (as given
 *     by the parameter "Position") as a start position in any of their spans.
 *     CDM does not own the returned
 *     object: the caller is responsible to use Tcl_DecrRefCount or CDM_Free
 *     in order to free the object and the memory associated with it.
 *
 *     This function will always create and return a new Annotation set object,
 *     even when the Document has no Annotations defined or when no Annotation
 *     is found that includes the specified text position into its span set.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Position:    Position in text.             (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationsAt (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_AnnotationsContaining, CDM_AnnotationsInRange,
 *     CDM_AnnotationsMatchingRange, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_AnnotationsAt
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_AnnotationsAt(CDM_Document Document, long Position) {
  CDM_AnnotationSet Anns, Result;
  CDM_SpanSet       Spans;
  Tcl_Obj **Ann, **Span;
  register long start;
  register int i,j;
  int items, spans;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationsAt: Aborting...");
    return NULL;
  }

  if (Position < 0) {
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_AnnotationsAt: Invalid Document Type");
    return NULL;
  }

  Anns = CDM_GetAnnotations(Document);
  if (Anns == NULL) return NULL;

  i = Tcl_ListObjGetElements(CDM_Interp, Anns, &items, &Ann);
  if (i != TCL_OK) {
    CDM_ReportError("Cannot Get Document's Annotations");
    return NULL;
  }
  /* If there are no Annotations, return an empty (list) object... */
  if (items < 1) {
    Tcl_DecrRefCount(Anns);
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }

  Result = Tcl_NewListObj(0, NULL);
  if (Result == NULL) CDM_Panic("CDM_AnnotationsAt: Tcl_NewListObj returned "
                                "NULL - out of memory?\n");

  for (i=0; i<items; i++) {
    Spans = CDM_GetSpans(Ann[i]);
    if (Spans == NULL) {
      CDM_ReportError("CDM_AnnotationsAt: Document has Annotation without "
                      "Span?\n");
      CDM_AppendError(Tcl_GetString(Ann[i]));
      Tcl_DecrRefCount(Result);
      Tcl_DecrRefCount(Anns);
      return NULL;
    }
    /* Check All Spans... */
    j = Tcl_ListObjGetElements(CDM_Interp, Spans, &spans, &Span);
    if (j != TCL_OK) {
      CDM_ReportError("CDM_AnnotationsAt: Error Getting Spans From Annotation");
      CDM_AppendError(Tcl_GetString(Ann[i]));
      Tcl_DecrRefCount(Result);
      Tcl_DecrRefCount(Anns);
      return NULL;
    }
    for (j=0; j<spans; j++) {
      start = CDM_GetStart(Span[j]);
      if (start == Position) {
        Tcl_ListObjAppendElement(CDM_Interp, Result, Ann[i]);
        break; /* We have found at least one... */
      }
    }
  }
  Tcl_DecrRefCount(Anns);
  return Result;
} /* CDM_AnnotationsAt */

/*
 * ---- CDM_AnnotationsContaining:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that include the specified text position (as given
 *     by the parameter "Position") into their spans. CDM does not own the
 *     returned object: the caller is responsible to use Tcl_DecrRefCount or
 *     CDM_Free in order to free the object and the memory associated with it.
 *
 *     This function will always create and return a new Annotation set object,
 *     even when the Document has no Annotations defined or when no Annotation
 *     is found that includes the specified text position into its span.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Position:    Position in text.             (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationsContaining (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_AnnotationsAt, CDM_AnnotationsInRange, CDM_AnnotationsMatchingRange,
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_AnnotationsContaining
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_AnnotationsContaining(CDM_Document Document, long Position) {
  CDM_AnnotationSet Result;
  register CDM_Annotation Ann;
  register int i, status;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationsContaining: Aborting...");
    return NULL;
  }

  if (Position < 0) {
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_AnnotationsContaining: Invalid Document Type");
    return NULL;
  }

  Result = Tcl_NewListObj(0, NULL);
  if (Result == NULL) CDM_Panic("CDM_AnnotationsContaining: Tcl_NewListObj "
                                "returned NULL - out of memory?\n");

  for (i=0; i <= Document->AnnotationMaxID; i++) {
    Ann = CDM_GetAnnotation(Document, i);
    if (Ann == NULL) continue;
    status = CDM_AnnotationContainsPosition(Ann, Position);
    if (status == -1) {
      Tcl_DecrRefCount(Result);
      return NULL;
    }
    if (status) {
      Tcl_ListObjAppendElement(CDM_Interp, Result, Ann);
    }
  }
  return Result;
} /* CDM_AnnotationsContaining */

/*
 * ---- CDM_AnnotationsContaining:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that include the two specified text positions
 *     (as given by the parameters "Position1" and "Position2")
 *     into their spans. CDM does not own the
 *     returned object: the caller is responsible to use Tcl_DecrRefCount or
 *     CDM_Free in order to free the object and the memory associated with it.
 *
 *     This function will always create and return a new Annotation set object,
 *     even when the Document has no Annotations defined or when no Annotation
 *     is found that includes the specified text positions into its span.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Position1:   Position in text.             (long)
 *     Position2:   Position in text.             (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationsContaining (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_AnnotationsAt, CDM_AnnotationsInRange, CDM_AnnotationsMatchingRange,
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_AnnotationsContaining
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_AnnotationsContaining(CDM_Document Document,
                                            long Position1, long Position2) {
  CDM_AnnotationSet Anns, Result;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationsContaining: Aborting...");
    return NULL;
  }

  /*
   * The algorithm is simple: We select all the Annotations that contain the
   * first position, and then we filter them to the ones that include the
   * second position...
   */
  Anns = CDM_AnnotationsContaining(Document, Position1);
  if (Anns == NULL) return NULL;
  Result = CDM_AnnotationsContaining(Anns, Position2);
  Tcl_DecrRefCount(Anns);
  return Result;
} /* CDM_AnnotationsContaining */

/*
 * ---- CDM_AnnotationsInRange:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that have at least one span within the specified
 *     range (["Start", "End"]). CDM will not own the returned
 *     object: the caller is responsible to use Tcl_DecrRefCount or CDM_Free
 *     in order to free the object and the memory associated with it.
 *
 *     This function will always create and return a new AnnotationSet object,
 *     even when the Document has no Annotations defined or when no Annotation
 *     is found that resides inside the specified range.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     Start:       The start of the desired range.  (long)
 *     End:         The end of the desired range.    (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationsInRange (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_AnnotationsAt, CDM_AnnotationsContaining,
 *     CDM_AnnotationsMatchingRange, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_AnnotationsInRange
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_AnnotationsInRange(CDM_Document Document,
                                         long Start, long End) {
  CDM_AnnotationSet Anns, Result;
  CDM_SpanSet       Spans;
  Tcl_Obj **Ann, **Span;
  long start, end;
  register int i,j;
  int items, spans;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationsInRange: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_AnnotationsInRange: Invalid Document Type");
    return NULL;
  }

  Anns = CDM_GetAnnotations(Document);
  if (Anns == NULL) return NULL;

  i = Tcl_ListObjGetElements(CDM_Interp, Anns, &items, &Ann);
  if (i != TCL_OK) {
    CDM_ReportError("Cannot Get Document's Annotations");
    return NULL;
  }
  /* If there are no Annotations, return an empty (list) object... */
  if (items == 0) {
    Tcl_DecrRefCount(Anns);
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }

  Result = Tcl_NewListObj(0, NULL);
  if (Result == NULL)
      CDM_Panic("CDM_AnnotationsInRange: Tcl_NewListObj returned "
                                "NULL - out of memory?\n");

  for (i=0; i<items; i++) {
    Spans = CDM_GetSpans(Ann[i]);
    if (Spans == NULL) {
      CDM_ReportError("CDM_AnnotationsInRange: Document has Annotation without "
                      "Span?\n");
      CDM_AppendError(Tcl_GetString(Ann[i]));
      Tcl_DecrRefCount(Result);
      Tcl_DecrRefCount(Anns);
      return NULL;
    }
    /* Check All Spans... */
    j = Tcl_ListObjGetElements(CDM_Interp, Spans, &spans, &Span);
    if (j != TCL_OK) {
      CDM_ReportError("CDM_AnnotationsInRange: Error Getting Spans From "
                      "Annotation");
      CDM_AppendError(Tcl_GetString(Ann[i]));
      Tcl_DecrRefCount(Result);
      Tcl_DecrRefCount(Anns);
      return NULL;
    }
    for (j=0; j<spans; j++) {
      if (CDM_GetSpanOffsets(Span[j], &start, &end) != TCL_OK) {
        CDM_ReportError("CDM_AnnotationsInRange: Error Getting Span %d From "
                        "Annotation", j);
        CDM_AppendError(Tcl_GetString(Ann[i]));
        Tcl_DecrRefCount(Result);
        Tcl_DecrRefCount(Anns);
        return NULL;
      }
      if (start >= Start && end <= End) {
        Tcl_ListObjAppendElement(CDM_Interp, Result, Ann[i]);
        break; /* We have found at least one... */
      }
    }
  }
  Tcl_DecrRefCount(Anns);
  return Result;
} /* CDM_AnnotationsInRange */

/*
 * ---- CDM_AnnotationsMatchingRange:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that have at least one span that
 *     exactly matches the specified
 *     range (["Start", "End"]). CDM will not own the returned
 *     object: the caller is responsible to use Tcl_DecrRefCount or CDM_Free
 *     in order to free the object and the memory associated with it.
 *
 *     This function will always create and return a new AnnotationSet object,
 *     even when the Document has no Annotations defined or when no Annotation
 *     is found that resides inside the specified range.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     Start:       The start of the desired range.  (long)
 *     End:         The end of the desired range.    (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationsMatchingRange (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_AnnotationsAt, CDM_AnnotationsContaining,
 *     CDM_AnnotationsInRange, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_AnnotationsMatchingRange
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_AnnotationsMatchingRange(CDM_Document Document,
                                         long Start, long End) {
  CDM_AnnotationSet Result;
  register CDM_Annotation Ann;
  register int i, status;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationsMatchingRange: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_AnnotationsMatchingRange: Invalid Document Type");
    return NULL;
  }

  Result = Tcl_NewListObj(0, NULL);
  if (Result == NULL)
      CDM_Panic("CDM_AnnotationsMatchingRange: Tcl_NewListObj returned "
                                "NULL - out of memory?\n");

  for (i=0; i <= Document->AnnotationMaxID; i++) {
    Ann = CDM_GetAnnotation(Document, i);
    if (Ann == NULL) continue;
    status = CDM_AnnotationMatchesRange(Ann, Start, End);
    if (status == -1) {
      Tcl_DecrRefCount(Result);
      return NULL;
    }
    if (status) {
      Tcl_ListObjAppendElement(CDM_Interp, Result, Ann);
    }
  }
  return Result;
} /* CDM_AnnotationsMatchingRange */

/*
 * ---- CDM_ByteSequenceInsertString:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the text contained in the specified Document,
 *     by inserting a string at the positions specified by the "pos"
 *     parameter. This position "pos" is the number of characters that have to
 *     be skipped before the string is inserted, where the first character of
 *     the Document text has a position equal to 0. If position is lower than
 *     zero, it is treaded as the [[i]]last[[/i]] character of the Document
 *     text.
 *
 *     For example, if the Document contains as text the
 *     string "This is a test", the following code:[[br]][[code]][[center]]
 *            CDM_ByteSequenceInsertString(Doc, 10, "simple ");
 *     [[/center]][[/code]][[br]] will modify the text of the Document to
 *     "This is a simple test".
 *
 *     The most important aspect of this function is that it will make the
 *     proper arrangements so as the Annotations contained inside the Document
 *     to be displaced by a proper amound of characters so as Annotations
 *     spanning text ranges after the insertion point to span the same text
 *     portions, after the string has been inserted.
 *
 *     This function will return the new Document text (as an object of type
 *     CDM_ByteSequence) if the insertion was succesful.
 *     Note that the returned object is owned by the CDM. The caller is not
 *     supposed to free or modify the returned object.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     pos:         The character offset (starting from 0) where the string will be inserted. The string will be inserted at the character pointed by this value, displacing any existing text at this position after the inserted text. (long)
 *     string:      The string to be inserted.       (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_ByteSequenceInsertString (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the new Document text (as an object of type
 *     CDM_ByteSequence) if the insertion was succesful.
 *     The returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_RawDataInsertString, CDM_ByteSequenceReplaceCharacters,
 *     CDM_DisplaceAnnotations,
 *     CDM_GetByteSequence, CDM_SetByteSequence,
 *     tip_ByteSequenceInsertString
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequence CDM_ByteSequenceInsertString(CDM_Document Document,
                                               long pos, char *string) {
  CDM_ByteSequence Result, Text;
  int charLength, status;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_ByteSequenceInsertString: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_ByteSequenceInsertString: Invalid Document Type");
    return NULL;
  }

  Text = CDM_GetByteSequence(Document);
  if (Text == NULL) return NULL;

  if (string == NULL || *string == '\0') return Text;

  /*
   * Handle some special cases more efficiently...
   */
  charLength = Tcl_NumUtfChars(string, -1);
  if (pos == 0) {
    Result     = Tcl_NewStringObj(string, -1);
    Tcl_AppendObjToObj(Result, Text);
  } else if (pos <= -1) {
    pos        = Tcl_GetCharLength(Text);
    Result     = Tcl_DuplicateObj(Text);
    Tcl_AppendToObj(Result, string, -1);
  } else {
    Tcl_Obj *Obj;
    Result     = Tcl_GetRange(Text, 0, pos - 1);
    Tcl_AppendToObj(Result, string, -1);
    Obj        = Tcl_GetRange(Text, pos, -1);
    Tcl_AppendObjToObj(Result, Obj);
    Tcl_DecrRefCount(Obj);
  }

  /* Displace all Annotations... */
  status = CDM_DisplaceAnnotations(Document, pos, charLength);
  if (status != TCL_OK) {
    CDM_Free(Result);
    return NULL;
  }
  /* Make Result the Document's new text... */
  CDM_SetByteSequence(Document, Result);
  return Result;
} /* CDM_ByteSequenceInsertString */

/*
 * ---- CDM_ByteSequenceReplace:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the text contained in the specified Document,
 *     by removing a range of consecutive characters from the Document Text,
 *     starting with the character whose index is the value of the parameter
 *     "first" and ending with the character whose index is the value of the
 *     parameter "last". An index of 0 refers to the first character of the
 *     Document Text, while a negative value (i.e. -1) refers to the last
 *     character of the Document Text. If the value of the parameter "newstring"
 *     is not NULL, then it is placed in the removed character range.
 *     If "first" is less than zero then it is treated as if it was zero, and
 *     if "last" is greater than the length of the Document Text
 *     then it is treated as if it was equal to the length of the Document Text.
 *     It is an error if "first" is greater than "last".
 *
 *     The most important aspect of this function is that it will make the
 *     proper arrangements so as the Annotations contained inside the Document
 *     to be displaced by a proper amound of characters so as Annotations
 *     spanning text ranges after the replacement point to span the same text
 *     portions, after the string has replaced the desired characters.
 *
 *     This function will return the new Document text (as an object of type
 *     CDM_ByteSequence) if the replacement was succesful.
 *     Note that the returned object is owned by the CDM. The caller is not
 *     supposed to free or modify the returned object.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     first:       The character offset (starting from 0) of the first character that will be replaced. (long)
 *     last:        The character offset (starting from 0) of the last character that will be replaced. (long)
 *     newstring:   The string to be placed in the removed character range. (char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function can be also used in order to [[b]]delete[[/b]]
 *    characters from the text contained in the specified Document, if the
 *    value of the "newstring" parameter is NULL or an empty string (like "").
 *    In this case, all Annotations objects contained in the specified
 *    Document are searched: if any of them has a Span object in its Span
 *    set whoose offsets span portion of the charactes to be deleted,
 *    then this Span is eliminated from the Span set of the Annotation
 *    object. Note that no Annotation objects are deleted from the Document,
 *    although character deletion can result in Annotations with an empty
 *    Span set.
 *
 *    This function is equivelant to tip_ByteSequenceReplace (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the new Document text (as an object of type
 *     CDM_ByteSequence) if the replacement was succesful.
 *     The returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_RawDataReplace,
 *     CDM_ByteSequenceReplaceCharacters, CDM_RawDataReplaceCharacters,
 *     CDM_ByteSequenceInsertString, CDM_RawDataInsertString,
 *     CDM_DisplaceAnnotations,
 *     CDM_GetByteSequence, CDM_SetByteSequence,
 *     CDM_GetRawData, CDM_SetRawData,
 *     tip_ByteSequenceReplace
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequence CDM_ByteSequenceReplace(CDM_Document Document,
                                       long first, long last, char *newstring) {
  CDM_ByteSequence Text;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_ByteSequenceReplace: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_ByteSequenceReplace: Invalid Document Type");
    return NULL;
  }

  Text = CDM_GetByteSequence(Document);
  if (Text == NULL) return NULL;
  if (first < 0) first = 0;
  if (last < 0 || last > Tcl_GetCharLength(Text)) {
    last = Tcl_GetCharLength(Text);
  }
  if (last < first) {
    CDM_ReportError("CDM_ByteSequenceReplace: Parameter \"last\" cannot be "
                    "less than parameter \"first\"");
    return NULL;
  }

  return CDM_ByteSequenceReplaceCharacters(Document, first, last-first+1,
                                           newstring);
} /* CDM_ByteSequenceReplace */

/*
 * ---- CDM_ByteSequenceReplaceCharacters:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the text contained in the specified Document,
 *     by replacing the specified number of characters at the position
 *     specified by the "pos" parameter by the provided string.
 *     This position "pos" is the number of characters that have to
 *     be skipped before the characters are replaced by the string,
 *     where the first character of
 *     the Document text has a position equal to 0. If position is lower than
 *     zero, it is treaded as the [[i]]last[[/i]] character of the Document
 *     text. In the case the last character is specified as position, then
 *     the last "characters" characters are replaced.
 *
 *     For example, if the Document contains as text the
 *     string "This is a test", the following code:[[br]][[code]][[center]]
 *            CDM_ByteSequenceReplaceCharacters(Doc, 10, 4, "simple test");
 *     [[/center]][[/code]][[br]] will modify the text of the Document to
 *     "This is a simple test".
 *
 *     The most important aspect of this function is that it will make the
 *     proper arrangements so as the Annotations contained inside the Document
 *     to be displaced by a proper amound of characters so as Annotations
 *     spanning text ranges after the replacement point to span the same text
 *     portions, after the string has replaced the desired characters.
 *
 *     This function will return the new Document text (as an object of type
 *     CDM_ByteSequence) if the replacement was succesful.
 *     Note that the returned object is owned by the CDM. The caller is not
 *     supposed to free or modify the returned object.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     pos:         The character offset (starting from 0) where the replacement will happen. The string will be inserted at the character pointed by this value, replacing the specified number of characters. (long)
 *     characters:  An integer specifying the number of characters that will be replaced. (int)
 *     string:      The string to be inserted.       (char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function can be also used in order to [[b]]delete[[/b]]
 *    characters from the text contained in the specified Document, if the
 *    value of the "string" parameter is NULL or an empty string (like "").
 *    In this case, all Annotations objects contained in the specified
 *    Document are searched: if any of them has a Span object in its Span
 *    set whoose offsets span portion of the charactes to be deleted,
 *    then this Span is eliminated from the Span set of the Annotation
 *    object. Note that no Annotation objects are deleted from the Document,
 *    although character deletion can result in Annotations with an empty
 *    Span set.
 *
 *    This function is equivelant to tip_ByteSequenceReplaceCharacters (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the new Document text (as an object of type
 *     CDM_ByteSequence) if the replacement was succesful.
 *     The returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_RawDataReplaceCharacters, CDM_ByteSequenceInsertString,
 *     CDM_DisplaceAnnotations,
 *     CDM_GetByteSequence, CDM_SetByteSequence,
 *     tip_ByteSequenceReplaceCharacters
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequence CDM_ByteSequenceReplaceCharacters(CDM_Document Document,
                                 long pos, int characters, const char *string) {
  CDM_ByteSequence Result, Text, Obj;
  int charLength, displacement, status;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_ByteSequenceReplaceCharacters: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_ByteSequenceReplaceCharacters: Invalid Document Type");
    return NULL;
  }

  if (characters < 0) {
    CDM_ReportError("CDM_ByteSequenceReplaceCharacters: the number of "
                    "characters to be replaced must be greater or equal to 0!");
    return NULL;
  }

  Text = CDM_GetByteSequence(Document);
  if (Text == NULL) return NULL;
  /*
   * Do not remove this. Bug reported by Sigletos:
   *   CDM_ByteSequenceReplaceCharacters(Doc, a, 0, "string");
   * was not equivalent to:
   *   CDM_ByteSequenceInsertString(Doc, a+1, "string");
   */
  /* pos++; */
  /* Petasis, Updated 06/07/2002: Actually this fix was also bug...
   * Now, the semantics of the related functions have changed.
   */

  if (string == NULL || *string == '\0') {
    CDM_AnnotationSet Annotations;
    CDM_Annotation   *Annotation, Ann;
    int i, items;
    /*
     * The caller wants to delete the characters instead of replacing them
     * (wants to replace characters with the empty string...)
     */

    /*
     * As a first step, we have to remove all Annotations that are contained
     * inside the text range to be deleted:
     */
    Annotations = CDM_AnnotationsInRange(Document, (long) pos,
                                                   (long) pos + characters);
    if (Annotations == NULL) return NULL;
    /*
     * Each of these Annotations will have at least one span in the range to
     * be deleted. If all Annotation spans are whithin this range, we have to
     * delete the Annotation. Else, we simply remove from the Annotation the
     * span that is within the range to be deleted...
     */
    Tcl_ListObjGetElements(CDM_Interp, Annotations, &items, &Annotation);
    for (i=0; i<items; i++) {
      Ann = CDM_RemoveSpan(Annotation[i], (long) pos, (long) pos + characters);
      /* Was any Span object removed from the Annotation? */
      if (Ann != Annotation[i]) {
        /*
         * Avoid to use CDM_AddAnnotation, as this function does various
         * optimisations regarding cached memory. We don't need all this fuss,
         * as the Annotation was already in the Document...
         * CDM_AddAnnotation(Document, Ann);
         */
        CDM_DocumentInternalAddAnnotation(Document, Ann);
      }
    }
    CDM_Free(Annotations);
    if (string == NULL) string = "";
  }

  /*
   * Handle some special cases more efficiently...
   */
  charLength = Tcl_NumUtfChars(string, -1);
  displacement = charLength - characters;
  if (pos == 0) {
    Result     = Tcl_NewStringObj(string, -1);
    Obj        = Tcl_GetRange(Text, characters, -1);
    Tcl_AppendObjToObj(Result, Obj);
    Tcl_DecrRefCount(Obj);
  } else if (pos <= -1) {
    pos        = Tcl_GetCharLength(Text) - characters;
    Result     = Tcl_GetRange(Text, 0, pos);
    Tcl_AppendToObj(Result, string, -1);
  } else {
    Result     = Tcl_GetRange(Text, 0, pos - 1);
    Tcl_AppendToObj(Result, string, -1);
    Obj        = Tcl_GetRange(Text, pos + characters, -1);
    Tcl_AppendObjToObj(Result, Obj);
    Tcl_DecrRefCount(Obj);
  }

  /* Displace all Annotations... */
  if (displacement != 0) {
    status = CDM_DisplaceAnnotations(Document, pos+1, displacement);
    if (status != TCL_OK) {
      CDM_Free(Result);
      return NULL;
    }
  }
  /* Make Result the Document's new text... */
  CDM_SetByteSequence(Document, Result);
  return Result;
} /* CDM_ByteSequenceReplaceCharacters */

/*
 * Helper Functions and definitions for CDM_CreateVectors.
 */
#define CODE_NONE               0
#define CODE_A                  1
#define CODE_ANN                2
#define CODE_ANNID              3
#define CODE_B                  4
#define CODE_CLASS              5
#define CODE_CLASS_ANN_TEXT     6
#define CODE_MINUS_A            7
#define CODE_MINUS_B            8
#define CODE_ROOT_ANN_TEXT      9
#define CODE_ROOT_ANN_ATTR     10
static int CDM_CreateVectorsGetCode(char *str, int len, int *value,
                                               char **name, char **val) {
  int start = 0, i, code = CODE_NONE;
  *name = *val = NULL;
  CDM_THREAD_SPECIFIC_DATA;

  if (len < 2) return CODE_NONE;
  switch (str[0]) {
    case '-': {
      /* One of -A* or -B* */
      if (str[1] != 'A' && str[1] != 'B') return CODE_NONE;
      /* Check whether the rest of the code is an integer... */
      if (str[1] == 'A') code = CODE_MINUS_A;
      else code = CODE_MINUS_B;
      start = 2;
      goto get_integer;
    }
    case 'A': {
      /* One of A* or ANN or ANNID */
      if (strcmp(str, "ANNID") == 0) return CODE_ANNID;
      if (strcmp(str, "ANN")   == 0) return CODE_ANN;
      /* Check whether the rest of the code is an integer... */
      start = 1; code = CODE_A;
      goto get_integer;
    }
    case 'B': {
      /* One of B* */
      /* Check whether the rest of the code is an integer... */
      start = 1; code = CODE_B;
      goto get_integer;
    }
    case 'C': {
      /* One of CLASS or CLASS_ANN_TEXT */
      if (strcmp(str, "CLASS_ANN_TEXT") == 0) return CODE_CLASS_ANN_TEXT;
      if (strcmp(str, "CLASS")          == 0) return CODE_CLASS;
      return CODE_NONE;
    }
    case 'R': {
      /* One of ROOT_ANN_TEXT */
      if (strcmp(str, "ROOT_ANN_TEXT") == 0) return CODE_ROOT_ANN_TEXT;
      if (CDM_Strncmp(str, "ROOT_ANN_ATTR::", 15) == 0) {
        /* The expected format is CODE_ROOT_ANN_ATTR::type::def_val */
        if (len < 17) return CODE_NONE;
        /* Locate the default value start... */
        for (i = len-2; i > 15; --i) {
          if (str[i] == ':' && str[i-1] == ':') {
            /* We have located the value start! */
            *val = CDM_Strdup(&str[i+1]);
            break;
          }
        }
        *name = CDM_Strdup(&str[15]);
        if (*name == NULL) return CODE_NONE;
        (*name)[i-16] = '\0';
        return CODE_ROOT_ANN_ATTR;
      }
      return CODE_NONE;
    }
    default: {return CODE_NONE;}
  }
get_integer:
  for (i = start; i < len; i++) {
    if (!CDM_Isdigit(str[i])) return CODE_NONE;
  }
  /* The rest of the string is an integer. Get its value... */
  i = Tcl_GetInt(CDM_Interp, &str[start], value);
  if (i != TCL_OK) return CODE_NONE;
  return code;
}; /* CDM_CreateVectorsGetCode */

static int CDM_CreateVectorsProcessAnnotation(CDM_Document Document,
                                     CDM_AnnotationSet *CellAnns,
                                     const int CellNumber,
                                     const CDM_CreateVectorsCellSpec CellSpec[],
                                     Tcl_Obj *Vector, CDM_Annotation ann,
                                     Tcl_Obj *QMark) {
  Tcl_Obj **objElements, *tmpObj, *tmp2Obj, *value;
  long start, end;
  int status, i = CellNumber, j, k, elements;
  CDM_Attribute attr;
  CDM_THREAD_SPECIFIC_DATA;

  status = CDM_GetFirstSpanOffsets(ann, &start, &end);
  if (status != TCL_OK || QMark == NULL) {
    CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
    goto error;
  }
  /* Add the cell attributes... */
  for (i = 0; i < CellNumber; i++) {
    CellSpec[i].tmpObj = NULL;
    if (CellSpec[i].CellAnnotation[0] == '\0') continue;
    if (CellSpec[i].AnnotationText) {
      CDM_ByteSequenceSet strings = CDM_GetAnnotatedTextRanges(Document, ann);
      if (strings == NULL) {
        CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
        goto error;
      }
      /* Eliminate completely the white space from the returned set */
      status = Tcl_ListObjGetElements(CDM_Interp, strings,
                                          &elements, &objElements);
      if (status != TCL_OK) {
        Tcl_DecrRefCount(strings);
        CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
        goto error;
      }
      if (elements == 1) {
        status = Tcl_ListObjAppendElement(CDM_Interp, Vector, objElements[0]);
        Tcl_DecrRefCount(strings);
        if (status != TCL_OK) {
          CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
          goto error;
        }
      } else {
        tmpObj = Tcl_ConcatObj(elements, objElements);
        Tcl_DecrRefCount(strings);
        if (tmpObj == NULL) {
          CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
          goto error;
        }
        strings = CDM_SplitString(Tcl_GetString(tmpObj), NULL, 0);
        Tcl_DecrRefCount(tmpObj);
        if (strings == NULL) {
          CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
          goto error;
        }
        Tcl_IncrRefCount(strings);
        status = Tcl_ListObjAppendElement(CDM_Interp, Vector, strings);
        Tcl_DecrRefCount(strings);
        if (status != TCL_OK) {
          CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
          goto error;
        }
      }
    } else {
      CDM_AnnotationSet CAnns = NULL;
      int cached = 0;
      /*
       * In order to speed things up, we examine the possibility of caching
       * CAnns: if the same CellAnnotation type has been already examined, we
       * can reuse its cached results...
       */
      for (j = 0; j < i; j++) {
        if (CellSpec[i].Contained == CellSpec[j].Contained &&
            CDM_Strcmp((char *) CellSpec[i].CellAnnotation,
                       (char *) CellSpec[j].CellAnnotation) == 0) {
          /* Reuse the CAnns! */
          CAnns = CellSpec[j].tmpObj;
          if (CAnns) cached = 1;
          break;
        }
      }
      if (cached == 0) {
        if (CellSpec[i].Contained) {
          /* If our cell can be contained inside a larger span range, select
           * all the Annotations that contain our span start... */
          /* TODO: why not also use end along with start? */
          CAnns = CDM_AnnotationsContaining(CellAnns[i], start);
        } else {
          /* Else, select only Annotations that have as span exactly our
           * span... */
          CAnns = CDM_AnnotationsMatchingRange(CellAnns[i], start, end);
        }
      }
      if (CAnns == NULL) {
        CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
        goto error;
      }
      if (CDM_Length(CAnns) == 0) {
        /* No Annotations found... */
        Tcl_DecrRefCount(CAnns);
        if (CellSpec[i].Expand) {
          for (int expand = 0; expand < CellSpec[i].Expand; expand++) {
            status = Tcl_ListObjAppendElement(CDM_Interp, Vector,
                                                  CellSpec[i].CellFailObj);
            if (status != TCL_OK) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
          }
        } else {
          status = Tcl_ListObjAppendElement(CDM_Interp, Vector,
                                                CellSpec[i].CellFailObj);
          if (status != TCL_OK) {
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
        }
      } else { /* if (CDM_Length(CAnns) == 0) */
        Tcl_IncrRefCount(CAnns);
        status = Tcl_ListObjGetElements(CDM_Interp, CAnns,
                                            &elements, &objElements);
        if (status != TCL_OK) {
          Tcl_DecrRefCount(CAnns);
          CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
          goto error;
        }
        value = Tcl_NewListObj(0, NULL);
        if (value == NULL) {
          Tcl_DecrRefCount(CAnns);
          CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
          goto error;
        }
        Tcl_IncrRefCount(value);
        for (j = 0; j < elements; j++) {
          attr = CDM_GetAttribute(objElements[j],
                                  (char *) CellSpec[i].CellAttribute);
          if (attr == NULL) {
            if (CellSpec[i].CellFailObj == NULL) {
              Tcl_DecrRefCount(CAnns);
              Tcl_DecrRefCount(value);
              CDM_ReportError("CDM_CreateVectorsProcessAnnotation: Annotation "
                "with Id=%ld does not have any attribute named \"%s\", "
                "and no failure value has ben specified\n(%s)",
                CDM_GetId(ann), CellSpec[i].CellAttribute, Tcl_GetString(ann));
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            status = Tcl_ListObjAppendElement(CDM_Interp, value,
                                              CellSpec[i].CellFailObj);
            if (status != TCL_OK) {
              Tcl_DecrRefCount(CAnns);
              Tcl_DecrRefCount(value);
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
          } else {
            attr = CDM_GetValueValue(attr);
            if (attr == NULL) {
              Tcl_DecrRefCount(CAnns);
              Tcl_DecrRefCount(value);
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            status = Tcl_ListObjAppendElement(CDM_Interp, value, attr);
            if (status != TCL_OK) {
              Tcl_DecrRefCount(CAnns);
              Tcl_DecrRefCount(value);
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
          }
        } /* for (j = 0; j < elements; j++) */
        if (cached) {
          Tcl_DecrRefCount(CAnns);
        } else {
          /* Cache the CAnns... */
          CellSpec[i].tmpObj = CAnns; /* CAnns has a refCount = 1; */
        }
        /* If more than one element in our value, we have to exec lsort
         * -unique... */
        if (elements > 1 && CDM_Interp) {
          /* Execute Tcl Command "lsort -unique" */
          Tcl_Obj *objv[3];
          objv[0] = Tcl_NewStringObj("lsort", 5);
          objv[1] = Tcl_NewStringObj("-unique", 7);
          objv[2] = value;
          for (k=0; k<3; k++) Tcl_IncrRefCount(objv[k]);
          status = Tcl_EvalObjv(CDM_Interp, 3, objv, TCL_EVAL_GLOBAL);
          for (k=0; k<3; k++) Tcl_DecrRefCount(objv[k]);
          if (status == TCL_OK) {
            /* Replave value with the sorted one... */
            Tcl_DecrRefCount(value);
            value = Tcl_DuplicateObj(Tcl_GetObjResult(CDM_Interp));
            if (value == NULL) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            Tcl_IncrRefCount(value);
          }
        } /* if (elements > 0 && CDM_Interp) */
        if (CellSpec[i].Expand) {
          /* If the caller wants to expand the first N values as features, do
           * the expansion and remove them from the value list... */
           Tcl_Obj **valueElem;
           int expand, valueElements;
           status = Tcl_ListObjGetElements(CDM_Interp, value,
                                            &valueElements, &valueElem);
           if (status != TCL_OK) {
             Tcl_DecrRefCount(value);
             CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
             goto error;
           }
           if (CellSpec[i].Expand == valueElements) {
             /* The positions can be filled exactly... */
             for (expand = 0; expand < valueElements; expand++) {
               status = Tcl_ListObjAppendElement(CDM_Interp, Vector,
                                                             valueElem[expand]);
               if (status != TCL_OK) {
                Tcl_DecrRefCount(value);
                CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
                goto error;
               }
             }
           } else if (CellSpec[i].Expand < valueElements) {
             /* The last element will be filled with the remaining elements,
              * concatenated win '-'... */
             for (expand = 0; expand < CellSpec[i].Expand - 1; expand++) {
               status = Tcl_ListObjAppendElement(CDM_Interp, Vector,
                                                             valueElem[expand]);
               if (status != TCL_OK) {
                Tcl_DecrRefCount(value);
                CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
                goto error;
               }
             }
             tmpObj = Tcl_NewListObj(valueElements - CellSpec[i].Expand + 1,
                                     &valueElem[CellSpec[i].Expand - 1]);
             if (tmpObj == NULL) {
               Tcl_DecrRefCount(value);
               CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
               goto error;
             }
             tmp2Obj = CDM_JoinSetElements(tmpObj, "-");
             Tcl_DecrRefCount(tmpObj); tmpObj = NULL;
             if (tmp2Obj == NULL) {
               Tcl_DecrRefCount(value);
               CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
               goto error;
             }
             Tcl_IncrRefCount(tmp2Obj);
             status = Tcl_ListObjAppendElement(CDM_Interp, Vector, tmp2Obj);
             Tcl_DecrRefCount(tmp2Obj);
             if (status != TCL_OK) {
               Tcl_DecrRefCount(value);
               CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
               goto error;
             }
           } else {
             /* Empty positions will remain, which have to be filled with '?' */
             for (expand = 0; expand < valueElements; expand++) {
               status = Tcl_ListObjAppendElement(CDM_Interp, Vector,
                                                             valueElem[expand]);
               if (status != TCL_OK) {
                Tcl_DecrRefCount(value);
                CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
                goto error;
               }
             }
             /* Fill the remaining positions with '?'s... */
             for (expand=valueElements; expand<CellSpec[i].Expand; expand++) {
               status = Tcl_ListObjAppendElement(CDM_Interp, Vector,
                                                     CellSpec[i].CellFailObj);
               if (status != TCL_OK) {
                Tcl_DecrRefCount(value);
                CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
                goto error;
               }
             } /* for (expand=valueElements, expand<CellSpec[i].Expand; ... */
           } /* Contitions over CellSpec[i].Expand & valueElements */
           Tcl_DecrRefCount(value);
        } else {
          if (elements == 1) {
            status = Tcl_ListObjAppendElement(CDM_Interp, Vector, value);
            Tcl_DecrRefCount(value);
          } else {
            /* Join the values with "-"... */
            tmpObj = CDM_JoinSetElements(value, "-");
            Tcl_DecrRefCount(value);
            if (tmpObj == NULL) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            Tcl_IncrRefCount(tmpObj);
            status = Tcl_ListObjAppendElement(CDM_Interp, Vector, tmpObj);
            Tcl_DecrRefCount(tmpObj);
          } /* if (elements == 1) */
          if (status != TCL_OK) {
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
        } /* if (CellSpec[i].Expand) */
      } /* if (CDM_Length(CAnns) == 0) */
    } /* if (CellSpec[i].AnnotationText) */
  } /* for (i = 0; i < CellNumber; i++) */
  return TCL_OK;
error:
  for (j = 0; j <= i && j < CellNumber; j++) {
    if (CellSpec[j].tmpObj) Tcl_DecrRefCount(CellSpec[j].tmpObj);
    CellSpec[j].tmpObj = NULL;
  }
  return TCL_ERROR;
} /* CDM_CreateVectorsProcessAnnotation */

/*
 * ---- CDM_CreateVectors:
 *  DESCRIPTION:
 *   <English:
 *    This function will create a set of vectors (for uses like instrumenting
 *    a machine learning algorithm for example) based on a wide range of
 *    information provided by the various parameters.
 *
 *    The return value of this function will be NULL if an error has
 *    occured, or a newly created object (of type CDM_ByteSequenceSet) which
 *    will contain the generated vectors if everything went as expected. Note
 *    that CDM does not owns the returned object: the caller is responsible
 *    to use CDM_Free of Tcl_DecrRefCount on the object if it is not needed
 *    any more.
 *
 *    The vectors in the returned object will be stored as a set of sets of
 *    text objects (of type CDM_ByteSequence or equivalent CDM_RawData). The
 *    caller must use CDM_SetGetElements on the restured object
 *    in order to get the individual vectors (each one encoded as a set of
 *    text objects of type CDM_ByteSequenceSet or CDM_RawDataSet or CDM_Set).
 *    Applying CDM_SetGetElements on a vector object, the individual text
 *    elements of the vector can be accessed.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Document:    A Document object.            (CDM_Document)
 *    VectorAnnotation: The Annotation from which vectors will be created. (const char *)
 *    VectorAnnConstraints: The Annotation from which vectors will be created. (const char *)
 *    VectorFormat: The Annotation from which vectors will be created. (const char *)
 *    ClassAnnotation: The Annotation from which vectors will be created. (const char *)
 *    ClassAttribute: The Annotation from which vectors will be created. (const char *)
 *    ClassAttributeAbsenceValue: The value to be used if the Class Attribute is not found (char *)
 *    ClassFail: The Annotation from which vectors will be created. (const char *)
 *    ClassCanContain: The Annotation from which vectors will be created. (const char *)
 *    ClassPrefixStartMode: The Annotation from which vectors will be created. (const char *)
 *    ClassPrefixEndMode: The Annotation from which vectors will be created. (const char *)
 *    ClassPrefixStart: The Annotation from which vectors will be created. (const char *)
 *    ClassPrefixEnd: The Annotation from which vectors will be created. (const char *)
 *    CellNumber: The Annotation from which vectors will be created. (const char *)
 *    CellSpec: The Annotation from which vectors will be created. (const char *)
 *    StatusWindow: The Annotation from which vectors will be created. (const char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_CreateVectors (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a set of vectors, or NULL
 *     if an error has been occured. In case of an error,  an
 *     error message describing the error will be left at the current active Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_SetGetElements, tip_CreateVectors
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequenceSet CDM_CreateVectors(CDM_Document Document,
                              const char *VectorAnnotation,
                              const char *VectorAnnConstraints,
                              const char *VectorFormat,
                              const char *ClassAnnotation,
                              const char *ClassAttribute,
                              const char *ClassAttributeAbsenceValue,
                              const char *ClassFail,
                                    int   ClassCanContain,
                                    int   ClassPrefixStartMode,
                                    int   ClassPrefixEndMode,
                              const char *ClassPrefixStart,
                              const char *ClassPrefixEnd,
                                    int   CellNumber,
                              const CDM_CreateVectorsCellSpec CellSpec[],
                              const char *StatusWindow) {
  Tcl_Obj *wwin = NULL, *tmpObj = NULL, *tmpObj2 = NULL, **objElements,
          *ggi_wait_update = NULL, *progressObj;
  Tcl_Obj *Vectors = NULL, *VectorFormatObj = NULL, **cell,
          *ColId = NULL, **preElem, **inElem, **postElem;
  Tcl_Obj *objv[4], *QMark = NULL;
  CDM_AnnotationSet *CellAnns = NULL, Sentences = NULL, VAnns = NULL,
                    CAnns = NULL, Pre = NULL, In = NULL, Post = NULL,
                    Vector = NULL;
  CDM_Annotation *vann, *sentence, *con, ann;
  CDM_Attribute attr;
  char *cellStr, *tmpStr, *attrName = NULL, *defaultValue = NULL;
  double Progress = 0.0, lastProgress = 0.0, Step = 0.0;
  long vstart, vend, start, end;
  int status, elements, VAnnsElements, attrElements, cellElements,
      sentenceElements, currentSentence = 0, code, cellStrLen, cellIndex,
      preElements, inElements, postElements, i, j ,k;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_CreateVectors: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_CreateVectors: NULL Document Pointer");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_CreateVectors: Invalid Document Type");
    return NULL;
  }

  /*
   * Status dialog: If one is provided, use it, else create a new one...
   */
  if (StatusWindow) {
    wwin = Tcl_NewStringObj(StatusWindow, -1);
  } else {
    snprintf(CDM_Temp, MAX_TEMP_STRING, "Creating Vectors from Document:\n%s",
             CDM_GetId(Document));
    status = CDM_Eval(CDM_Interp, "ggi_wait_append", CDM_Temp, "0.0", NULL);
    if (status == TCL_OK) wwin = Tcl_GetObjResult(CDM_Interp);
  }
  if (wwin) {
    Tcl_IncrRefCount(wwin);
    ggi_wait_update = CDM_NewStringObj("ggi_wait_update");
    if (ggi_wait_update) Tcl_IncrRefCount(ggi_wait_update);
  }
  if (wwin == NULL || ggi_wait_update == NULL || CDM_Interp == NULL) {
    /* There is no point in keeping these objects arround, as we won't be able
       to issue the command... */
    if (wwin)            Tcl_DecrRefCount(wwin);
    if (ggi_wait_update) Tcl_DecrRefCount(ggi_wait_update);
    wwin = ggi_wait_update = NULL;
  }

  QMark = CDM_NewStringObj("?");
  if (QMark == NULL) {
    CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
    goto error;
  }
  Tcl_IncrRefCount(QMark);

  /*
   * Step 1: Get All needed information from Document...
   */

  /*
   * Allocate the space for CellAnns, and store the requested Annotation
   * sets...
   */
  if (CellNumber) {
    CellAnns =
        (CDM_AnnotationSet *) Tcl_Alloc(CellNumber*sizeof(CDM_AnnotationSet));
    if (CellAnns == NULL) {
      CDM_ReportError("CDM_CreateVectors: out of memory");
      CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
      goto error;
    }
    for (i = 0; i < CellNumber; i++) {
      /* Cache the fail value for all cell entries, if not already cached... */
      if (CellSpec[i].CellFailObj == NULL) {
        CellSpec[i].CellFailObj = CDM_NewStringObj(CellSpec[i].CellFail);
      }
      if (CellSpec[i].CellFailObj) Tcl_IncrRefCount(CellSpec[i].CellFailObj);
      CellAnns[i] = CDM_SelectAnnotationsSorted(Document,
                                           (char *) CellSpec[i].CellAnnotation);
      if (CellAnns[i] == NULL) {
        CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
        goto error;
      }
    }

    /* Get the sentence Annotations... */
    Sentences = CDM_SelectAnnotationsSorted(Document, "sentence",
                                                "[llength ann::constituents ]");
    if (Sentences == NULL) {
      CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
      goto error;
    }
    Tcl_IncrRefCount(Sentences);
    status = Tcl_ListObjGetElements(CDM_Interp, Sentences,
                                    &sentenceElements, &sentence);
    if (status != TCL_OK) {
      CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
      goto error;
    }
  } /* if (CellNumber) */
  /* Get the Annotations for creating the vectors... */
  VAnns = CDM_SelectAnnotationsSorted(Document, (char *) VectorAnnotation,
                                                (char *) VectorAnnConstraints);
  if (VAnns == NULL) {
    CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
    goto error;
  }
  Tcl_IncrRefCount(VAnns);
  status = Tcl_ListObjGetElements(CDM_Interp, VAnns, &VAnnsElements, &vann);
  if (status != TCL_OK) {
    CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
    goto error;
  }
  /* Get the Class Annotations. Remember that ClassAnnotation can contain a
   * set of Annotation types... */
  tmpObj = Tcl_NewStringObj(ClassAnnotation, -1);
  if (tmpObj == NULL) {
    CDM_ReportError("CDM_CreateVectors: out of memory");
    CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
    goto error;
  }
  status = Tcl_ListObjGetElements(CDM_Interp, tmpObj, &elements, &objElements);
  if (status != TCL_OK) {
    CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
    goto error;
  }
  CAnns = Tcl_NewListObj(0, NULL);
  if (CAnns == NULL) {
    CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
    goto error;
  }
  Tcl_IncrRefCount(CAnns);
  for (i = 0; i < elements; i++) {
    tmpObj2 = CDM_SelectAnnotations(Document, Tcl_GetString(objElements[i]));
    if (tmpObj2 == NULL) {
      CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
      goto error;
    }
    status = Tcl_ListObjAppendList(CDM_Interp, CAnns, tmpObj2);
    Tcl_DecrRefCount(tmpObj2); tmpObj2 = NULL; // They get freed on errors...
    if (status != TCL_OK) {
      CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
      goto error;
    }
  }
  Tcl_DecrRefCount(tmpObj);
  tmpObj = NULL; // They get freed on errors...
  status = Tcl_ListObjGetElements(CDM_Interp, VAnns, &VAnnsElements, &vann);
  if (status != TCL_OK) {
    CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
    goto error;
  }

  VectorFormatObj = Tcl_NewStringObj(VectorFormat, -1);
  if (VectorFormatObj == NULL) {
    CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
    goto error;
  }
  status = Tcl_ListObjGetElements(CDM_Interp, VectorFormatObj, &cellElements,
                                  &cell);
  if (status != TCL_OK) {
    CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
    goto error;
  }

  Vectors = Tcl_NewListObj(0, NULL);
  if (Vectors == NULL) {
    CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
    goto error;
  }

  /*
   *  Step 2: Iterate ovel Annotatons that are to exported as vectors...
   */
  Step = 100.0 / (VAnnsElements+1);
  for (i = 0; i < VAnnsElements; i++) {
    status = CDM_GetFirstSpanOffsets(vann[i], &vstart, &vend);
    if (status != TCL_OK) {
      CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
      goto error;
    }
    if (CellNumber) {
      /* Get the sentence Annotation covering this Annotation.
       * The initial algorithm was to call
       * CDM_FirstAnnotationContaining(Sentences, vstart), but this tends to be
       * slow on Documents with many sentences. As a result, now we are sorting
       * our sentences, and we make our search in the sorted set...*/
      /*
      sentence = CDM_FirstAnnotationContaining(Sentences, vstart);
      if (sentence == NULL) {
        if (CDM_ActiveInterpreterResultEmpty()) {
          CDM_Debug("CDM_CreateVectors: Skipping Annotation %ld: "
                 "Not covered by any sentence Annotation!", CDM_GetId(vann[i]));
          continue;
        } else {
          CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
          goto error;
        }
      }*/
      while (!CDM_AnnotationContainsPosition(
                                    sentence[currentSentence], vstart)) {
        ++currentSentence;
        if (currentSentence == sentenceElements) {
          /* We have reached the end of sentences. This should never happen! */
          CDM_ReportError("CDM_CreateVectors: "
                          "error in sentence search process!");
          CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
          goto error;
        }
      }
      /* Get the constituents from the sentence Annotation... */
      attr = CDM_GetAttribute(sentence[currentSentence], "constituents");
      if (attr == NULL) {
        CDM_Debug("CDM_CreateVectors: Skipping Annotation %ld: "
                  "Covered by a sentence Annotation without constituents!",
                  CDM_GetId(vann[i]));
        continue;
      }
      attr = CDM_GetValueValue(attr);
      if (attr == NULL) {
        CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
        goto error;
      }
      /* Iterate over all constituents */
      status = Tcl_ListObjGetElements(CDM_Interp, attr, &attrElements, &con);
      if (status != TCL_OK) {
        CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
        goto error;
      }
      Pre    = Tcl_NewListObj(0, NULL);
      if (Pre    == NULL) {
        CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
        goto error;
      }
      In     = Tcl_NewListObj(0, NULL);
      if (In     == NULL) {
        CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
        goto error;
      }
      Post   = Tcl_NewListObj(0, NULL);
      if (Post   == NULL) {
        CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
        goto error;
      }
      Tcl_IncrRefCount(Pre); Tcl_IncrRefCount(In); Tcl_IncrRefCount(Post);
      for (j = 0; j < attrElements; j++) {
        ann = CDM_GetAnnotation(Document, con[j]);
        if (ann == NULL) {
          CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
          goto error;
        }
        status = CDM_GetFirstSpanOffsets(ann, &start, &end);
        if (status != TCL_OK) {
          CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
          goto error;
        }
        /* Categorise the Annotation according to its span... */
        if (start < vstart) {
          status = Tcl_ListObjAppendElement(CDM_Interp, Pre, ann);
        } else if (start >= vstart && end <= vend) {
          status = Tcl_ListObjAppendElement(CDM_Interp, In, ann);
        } else if (end > vend) {
          status = Tcl_ListObjAppendElement(CDM_Interp, Post, ann);
        } else {
          if (StatusWindow == NULL && wwin) {
            CDM_Eval(CDM_Interp, "ggi_destroy_append",
                                 Tcl_GetString(wwin), NULL);
          }
          CDM_ReportError("CDM_CreateVectors: Document %s: "
             "Token Annotation with"
             " id=\"%ld\" not properly contained inside Annotation \"%s\"?",
             CDM_GetId(Document), CDM_GetId(ann), Tcl_GetString(ann));
          CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
          goto error;
        }
      } /* Iteration over sentence constituents */

      status = Tcl_ListObjGetElements(CDM_Interp, Pre, &preElements,  &preElem);
      if (status != TCL_OK) {
        CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
        goto error;
      }
      status = Tcl_ListObjGetElements(CDM_Interp, In,  &inElements,   &inElem);
      if (status != TCL_OK) {
        CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
        goto error;
      }
      status = Tcl_ListObjGetElements(CDM_Interp, Post,&postElements,&postElem);
      if (status != TCL_OK) {
        CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
        goto error;
      }
    } /* if (CellNumber) */
    Vector = Tcl_NewListObj(0, NULL);
    if (Vector == NULL) {
      CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
      goto error;
    }
    Tcl_IncrRefCount(Vector);

    /*
     * Now, we have the ordered sets and we are ready to place cells into
     * right positions...
     */
    for (j = 0; j < cellElements; j++) {
      cellStr = Tcl_GetStringFromObj(cell[j], &cellStrLen);
      /* All vector format codes have at least 2 characters... */
      if (cellStrLen == 0) {
        CDM_ReportError("CDM_CreateVectors: NULL in vector format?");
        CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
        goto error;
      }
      /* Get the code stored in the string */
      code = CDM_CreateVectorsGetCode(cellStr, cellStrLen, &cellIndex,
                                               &attrName, &defaultValue);
      /* Take the proper action according to the returned code: */
      switch (code) {
        case CODE_A: {
          if (CellNumber == 0) {
            CDM_ReportError("CDM_CreateVectors: code \"%s\" specified but no "
                            "Cells have been defined!", cellStrLen);
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          if (cellIndex < 1) {
            CDM_ReportError("CDM_CreateVectors: code \"%s\" out of bounds",
                            cellStrLen);
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          if (cellIndex > inElements) {
            /* The Annotation is missing. Add a range of "?"... */
            for (k = 0; k < CellNumber; k++) {
              if (CellSpec[k].Expand) {
                for (int expand = 0; expand < CellSpec[k].Expand; expand++) {
                  status = Tcl_ListObjAppendElement(CDM_Interp, Vector, QMark);
                  if (status != TCL_OK) {
                    CDM_AppendError("\n(File \"%s\", line=%d)",
                                    __FILE__, __LINE__);
                    goto error;
                  }
                }
              } else {
                status = Tcl_ListObjAppendElement(CDM_Interp, Vector, QMark);
                if (status != TCL_OK) {
                  CDM_AppendError("\n(File \"%s\", line=%d)",
                                  __FILE__, __LINE__);
                  goto error;
                }
              }
            }
          } else {
            ann = inElem[cellIndex-1];
            status = CDM_CreateVectorsProcessAnnotation(Document, CellAnns,
                                      CellNumber, CellSpec, Vector, ann, QMark);
            if (status != TCL_OK) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
          }
          break;
        } /* CODE_A */
        case CODE_B: {
          if (CellNumber == 0) {
            CDM_ReportError("CDM_CreateVectors: code \"%s\" specified but no "
                            "Cells have been defined!", cellStrLen);
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          if (cellIndex < 1) {
            CDM_ReportError("CDM_CreateVectors: code \"%s\" out of bounds",
                            cellStrLen);
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          if (cellIndex > postElements) {
            /* The Annotation is missing. Add a range of "?"... */
            for (k = 0; k < CellNumber; k++) {
              if (CellSpec[k].Expand) {
                for (int expand = 0; expand < CellSpec[k].Expand; expand++) {
                  status = Tcl_ListObjAppendElement(CDM_Interp, Vector, QMark);
                  if (status != TCL_OK) {
                    CDM_AppendError("\n(File \"%s\", line=%d)",
                                    __FILE__, __LINE__);
                    goto error;
                  }
                }
              } else {
                status = Tcl_ListObjAppendElement(CDM_Interp, Vector, QMark);
                if (status != TCL_OK) {
                  CDM_AppendError("\n(File \"%s\", line=%d)",
                                  __FILE__, __LINE__);
                  goto error;
                }
              }
            }
          } else {
            ann = postElem[cellIndex-1];
            status = CDM_CreateVectorsProcessAnnotation(Document, CellAnns,
                                      CellNumber, CellSpec, Vector, ann, QMark);
            if (status != TCL_OK) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
          }
          break;
        } /* CODE_B */
        case CODE_MINUS_A: {
          if (CellNumber == 0) {
            CDM_ReportError("CDM_CreateVectors: code \"%s\" specified but no "
                            "Cells have been defined!", cellStrLen);
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          if (cellIndex < 1) {
            CDM_ReportError("CDM_CreateVectors: code \"%s\" out of bounds",
                            cellStrLen);
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          if (cellIndex > preElements) {
            /* The Annotation is missing. Add a range of "?"... */
            for (k = 0; k < CellNumber; k++) {
              if (CellSpec[k].Expand) {
                for (int expand = 0; expand < CellSpec[k].Expand; expand++) {
                  status = Tcl_ListObjAppendElement(CDM_Interp, Vector, QMark);
                  if (status != TCL_OK) {
                    CDM_AppendError("\n(File \"%s\", line=%d)",
                                    __FILE__, __LINE__);
                    goto error;
                  }
                }
              } else {
                status = Tcl_ListObjAppendElement(CDM_Interp, Vector, QMark);
                if (status != TCL_OK) {
                  CDM_AppendError("\n(File \"%s\", line=%d)",
                                  __FILE__, __LINE__);
                  goto error;
                }
              }
            }
          } else {
            ann = preElem[preElements-cellIndex];
            status = CDM_CreateVectorsProcessAnnotation(Document, CellAnns,
                                      CellNumber, CellSpec, Vector, ann, QMark);
            if (status != TCL_OK) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
          }
          break;
        } /* CODE_MINUS_A */
        case CODE_MINUS_B: {
          if (CellNumber == 0) {
            CDM_ReportError("CDM_CreateVectors: code \"%s\" specified but no "
                            "Cells have been defined!", cellStrLen);
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          if (cellIndex < 1) {
            CDM_ReportError("CDM_CreateVectors: code \"%s\" out of bounds",
                            cellStrLen);
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          if (cellIndex > inElements) {
            /* The Annotation is missing. Add a range of "?"... */
            for (k = 0; k < CellNumber; k++) {
              if (CellSpec[k].Expand) {
                for (int expand = 0; expand < CellSpec[k].Expand; expand++) {
                  status = Tcl_ListObjAppendElement(CDM_Interp, Vector, QMark);
                  if (status != TCL_OK) {
                    CDM_AppendError("\n(File \"%s\", line=%d)",
                                    __FILE__, __LINE__);
                    goto error;
                  }
                }
              } else {
                status = Tcl_ListObjAppendElement(CDM_Interp, Vector, QMark);
                if (status != TCL_OK) {
                  CDM_AppendError("\n(File \"%s\", line=%d)",
                                  __FILE__, __LINE__);
                  goto error;
                }
              }
            }
          } else {
            ann = inElem[inElements-cellIndex];
            status = CDM_CreateVectorsProcessAnnotation(Document, CellAnns,
                                      CellNumber, CellSpec, Vector, ann, QMark);
            if (status != TCL_OK) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
          }
          break;
        } /* CODE_MINUS_B */
        case CODE_ANN: {
          /* ANN is a special case: we process the whole annotation at once,
           * with full length... */
          for (k = 0; k < inElements; k++) {
            status = CDM_CreateVectorsProcessAnnotation(Document, CellAnns,
                                CellNumber, CellSpec, Vector, inElem[k], QMark);
            if (status != TCL_OK) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
          }
          break;
        } /* CODE_ANN */
        case CODE_ANNID: {
          /* Initialise the CollectionId, if not yet initialised... */
          if (ColId == NULL) {
            CDM_Collection col = CDM_GetParent(Document);
            if (col == NULL) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            char *parent = CDM_GetName(col);
            if (parent == NULL) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            parent = CDM_FileNameTail(parent);
            if (parent == NULL) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            ColId = Tcl_NewStringObj(parent, -1);
            CDM_Free(parent);
            if (ColId == NULL) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            Tcl_AppendToObj(ColId, "|", 1);
            parent = CDM_GetId(Document);
            if (parent == NULL) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            Tcl_AppendToObj(ColId, parent, -1);
          }
          /* Generate the Annotation Id... */
          tmpObj = Tcl_DuplicateObj(ColId);
          tmpStr = CDM_GetType(vann[i]);
          if (tmpStr == NULL) {
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          Tcl_AppendToObj(tmpObj, tmpStr, -1);
          Tcl_AppendToObj(tmpObj, "\\|", 2);
          Tcl_AppendObjToObj(tmpObj, CDM_NewLongObj(CDM_GetId(vann[i])));
          status = Tcl_ListObjAppendElement(CDM_Interp, Vector, tmpObj);
          tmpObj = NULL;
          if (status != TCL_OK) {
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          break;
        } /* CODE_ANNID */
        case CODE_CLASS: {
          /* Add the Class... */
          CDM_AnnotationSet ClassAnns;
          long pref_start, pref_end;
          int appendStart = 0, appendEnd = 0;
          if (ClassCanContain) {
            ClassAnns = CDM_AnnotationsContaining(CAnns, vstart, vend);
          } else {
            ClassAnns = CDM_AnnotationsMatchingRange(CAnns, vstart, vend);
          }
          if (ClassAnns == NULL) {
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          if (CDM_Length(ClassAnns) == 0) {
            /* No Annotations found... */
            Tcl_DecrRefCount(ClassAnns);
            status = Tcl_ListObjAppendElement(CDM_Interp, Vector,
                                            CDM_NewStringObj(ClassFail));
            if (status != TCL_OK) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
          } else { /* if (ClassAnns == NULL) */
            Tcl_Obj *value;
            status = Tcl_ListObjGetElements(CDM_Interp, ClassAnns,
                                                 &elements, &objElements);
            if (status != TCL_OK) {
              Tcl_DecrRefCount(ClassAnns);
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            value = Tcl_NewListObj(0, NULL);
            if (value == NULL) {
              Tcl_DecrRefCount(ClassAnns);
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            Tcl_IncrRefCount(value);
            elements--;
            for (k = 0; k <= elements; k++) {
              attr = CDM_GetAttribute(objElements[k], (char *)ClassAttribute);
              if (attr == NULL) {
                /* The attribute was not found. Use the absence value... */
                if (ClassAttributeAbsenceValue == NULL) {
                  Tcl_DecrRefCount(ClassAnns);
                  Tcl_DecrRefCount(value);
                  CDM_AppendError("\n(File \"%s\", line=%d)",
                                  __FILE__, __LINE__);
                  goto error;
                }
                attr = CDM_NewStringObj(ClassAttributeAbsenceValue);
              } else {
                attr = CDM_GetValueValue(attr);
                if (attr == NULL) {
                  Tcl_DecrRefCount(ClassAnns);
                  Tcl_DecrRefCount(value);
                  CDM_AppendError("\n(File \"%s\", line=%d)",
                                  __FILE__, __LINE__);
                  goto error;
                }
              }
              status = Tcl_ListObjAppendElement(CDM_Interp, value, attr);
              if (status != TCL_OK) {
                Tcl_DecrRefCount(ClassAnns);
                Tcl_DecrRefCount(value);
                CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
                goto error;
              }
              /* Check for a start prefix... */
              if (ClassCanContain && k == 0 && ClassPrefixStartMode) {
                status = CDM_GetFirstSpanOffsets(objElements[k],
                                                 &pref_start, &pref_end);
                if (status != TCL_OK) {
                  Tcl_DecrRefCount(ClassAnns);
                  Tcl_DecrRefCount(value);
                  CDM_AppendError("\n(File \"%s\", line=%d)",__FILE__,__LINE__);
                  goto error;
                }
                if (vstart == pref_start) appendStart = 1;
              }
              /* Check for an end prefix... */
              if (ClassCanContain && k == elements && ClassPrefixEndMode) {
                status = CDM_GetFirstSpanOffsets(objElements[k],
                                                 &pref_start, &pref_end);
                if (status != TCL_OK) {
                  Tcl_DecrRefCount(ClassAnns);
                  Tcl_DecrRefCount(value);
                  CDM_AppendError("\n(File \"%s\", line=%d)",__FILE__,__LINE__);
                  goto error;
                }
                if (vend == pref_end) appendEnd = 1;
              }
            }
            Tcl_DecrRefCount(ClassAnns);
            /* If more than one element in our value, we have to exec lsort
             * -unique... */
            if (elements > 0 && CDM_Interp) {
              /* Execute Tcl Command "lsort -unique" */
              objv[0] = Tcl_NewStringObj("lsort", 5);
              objv[1] = Tcl_NewStringObj("-unique", 7);
              objv[2] = value;
              for (k=0; k<3; k++) Tcl_IncrRefCount(objv[k]);
              status = Tcl_EvalObjv(CDM_Interp, 3, objv, TCL_EVAL_GLOBAL);
              for (k=0; k<3; k++) Tcl_DecrRefCount(objv[k]);
              if (status == TCL_OK) {
                /* Replace value with the sorted one... */
                Tcl_DecrRefCount(value);
                value = Tcl_DuplicateObj(Tcl_GetObjResult(CDM_Interp));
                if (value == NULL) {
                  CDM_AppendError("\n(File \"%s\", line=%d)",__FILE__,__LINE__);
                  goto error;
                }
                Tcl_IncrRefCount(value);
              }
            }
            /* Join the values with "-"... */
            tmpObj = CDM_JoinSetElements(value, "-");
            Tcl_DecrRefCount(value);
            if (tmpObj == NULL) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            Tcl_IncrRefCount(tmpObj); value = tmpObj; tmpObj = NULL;
            /* Now, "value" holds our result string... */
            if (ClassCanContain) {
              if (appendStart) {
                tmpObj = Tcl_NewStringObj(ClassPrefixStart, -1);
                if (tmpObj == NULL) {
                  Tcl_DecrRefCount(value);
                  CDM_AppendError("\n(File \"%s\", line=%d)",__FILE__,__LINE__);
                  goto error;
                }
                Tcl_AppendObjToObj(tmpObj, value);
                Tcl_DecrRefCount(value); Tcl_IncrRefCount(tmpObj);
                value = tmpObj; tmpObj = NULL;
              }
              if (appendEnd) {
                tmpObj = Tcl_NewStringObj(ClassPrefixEnd, -1);
                if (tmpObj == NULL) {
                  Tcl_DecrRefCount(value);
                  CDM_AppendError("\n(File \"%s\", line=%d)",__FILE__,__LINE__);
                  goto error;
                }
                Tcl_AppendObjToObj(tmpObj, value);
                Tcl_DecrRefCount(value); Tcl_IncrRefCount(tmpObj);
                value = tmpObj; tmpObj = NULL;
              }
            } /* if (ClassCanContain) */
            status = Tcl_ListObjAppendElement(CDM_Interp, Vector, value);
            Tcl_DecrRefCount(value);
            if (status != TCL_OK) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
          } /* if (ClassAnns == NULL) */
          break;
        } /* CODE_CLASS */
        case CODE_CLASS_ANN_TEXT: {
          /* Add the Class spanned text... */
          CDM_AnnotationSet ClassAnns =
              CDM_AnnotationsMatchingRange(CAnns, vstart, vend);
          if (ClassAnns == NULL) {
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          if (CDM_Length(ClassAnns) == 0) {
            Tcl_DecrRefCount(ClassAnns);
            status = Tcl_ListObjAppendElement(CDM_Interp, Vector,
                                              CDM_NewStringObj("NO_TEXT"));
            if (status != TCL_OK) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
          } else {
            CDM_ByteSequence value;
            CDM_ByteSequenceSet strings = Tcl_NewListObj(0, NULL);
            if (strings == NULL) {
              Tcl_DecrRefCount(ClassAnns);
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            status = Tcl_ListObjGetElements(CDM_Interp, ClassAnns,
                                            &elements, &objElements);
            if (status != TCL_OK) {
              Tcl_DecrRefCount(ClassAnns);
              Tcl_DecrRefCount(strings);
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            for (k = 0; k < elements; k++) {
              tmpObj =  CDM_GetAnnotatedTextRanges(Document, objElements[k]);
              if (tmpObj == NULL) {
                Tcl_DecrRefCount(ClassAnns);
                Tcl_DecrRefCount(strings);
                CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
                goto error;
              }
              status = Tcl_ListObjAppendList(CDM_Interp, strings, tmpObj);
              Tcl_DecrRefCount(tmpObj); tmpObj = NULL;
            }
            Tcl_DecrRefCount(ClassAnns);
            /* The rest of the code is similar to the CODE_ROOT_ANN_TEXT. Just
             * copied from there... */
            /* Eliminate completely the white space from the returned set */
            status = Tcl_ListObjGetElements(CDM_Interp, strings,
                                            &elements, &objElements);
            if (status != TCL_OK) {
              Tcl_DecrRefCount(strings);
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            tmpObj = Tcl_ConcatObj(elements, objElements);
            Tcl_DecrRefCount(strings);
            if (tmpObj == NULL) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            tmpObj2 = CDM_SplitString(Tcl_GetString(tmpObj), NULL, 0);
            Tcl_DecrRefCount(tmpObj); tmpObj = NULL;
            if (tmpObj2 == NULL) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            value = CDM_JoinSetElements(tmpObj2, "-");
            Tcl_DecrRefCount(tmpObj2); tmpObj2 = NULL;
            if (value == NULL) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
            Tcl_IncrRefCount(value);
            status = Tcl_ListObjAppendElement(CDM_Interp, Vector, value);
            Tcl_DecrRefCount(value);
            if (status != TCL_OK) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
          }
          break;
        } /* CODE_CLASS_ANN_TEXT */
        case CODE_ROOT_ANN_TEXT: {
          /* Append to the vector the text spanned by the root ann */
          CDM_ByteSequence value;
          CDM_ByteSequenceSet strings =
              CDM_GetAnnotatedTextRanges(Document, vann[i]);
          if (strings == NULL) {
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          /* Eliminate completely the white space from the returned set */
          status = Tcl_ListObjGetElements(CDM_Interp, strings,
                                          &elements, &objElements);
          if (status != TCL_OK) {
            Tcl_DecrRefCount(strings);
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          tmpObj = Tcl_ConcatObj(elements, objElements);
          Tcl_DecrRefCount(strings);
          if (tmpObj == NULL) {
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          tmpObj2 = CDM_SplitString(Tcl_GetString(tmpObj), NULL, 0);
          Tcl_DecrRefCount(tmpObj); tmpObj = NULL;
          if (tmpObj2 == NULL) {
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          value = CDM_JoinSetElements(tmpObj2, "-");
          Tcl_DecrRefCount(tmpObj2); tmpObj2 = NULL;
          if (value == NULL) {
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          Tcl_IncrRefCount(value);
          status = Tcl_ListObjAppendElement(CDM_Interp, Vector, value);
          Tcl_DecrRefCount(value);
          if (status != TCL_OK) {
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          break;
        } /* CODE_ROOT_ANN_TEXT */
        case CODE_ROOT_ANN_ATTR: {
          CDM_AttributeValue value;
          attr = CDM_GetAttribute(vann[i], attrName);
          if (attr == NULL) {
            value = CDM_NewStringObj(defaultValue);
          } else {
            value = CDM_GetValueValue(attr);
            if (value == NULL) {
              CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
              goto error;
            }
          }
          Tcl_IncrRefCount(value);
          status = Tcl_ListObjAppendElement(CDM_Interp, Vector, value);
          Tcl_DecrRefCount(value);
          if (status != TCL_OK) {
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          break;
        } /* CODE_ROOT_ANN_ATTR */
        case CODE_NONE: {
          /* This is a static string. Append it directly into our vector */
          status = Tcl_ListObjAppendElement(CDM_Interp, Vector, cell[j]);
          if (status != TCL_OK) {
            CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
            goto error;
          }
          break;
        } /* CODE_NONE */
      } /* switch (code) */
      if (attrName     != NULL) CDM_Free(attrName);
      if (defaultValue != NULL) CDM_Free(defaultValue);
      attrName = defaultValue = NULL;
    } /* Iteration over all format cells */

    status = Tcl_ListObjAppendElement(CDM_Interp, Vectors, Vector);
    if (status != TCL_OK) {
      CDM_AppendError("\n(File \"%s\", line=%d)", __FILE__, __LINE__);
      goto error;
    }
    if (Pre)    Tcl_DecrRefCount(Pre);
    if (In)     Tcl_DecrRefCount(In);
    if (Post)   Tcl_DecrRefCount(Post);
    if (Vector) Tcl_DecrRefCount(Vector);
    Pre = In = Post = Vector = NULL;

    /*
     * Update the status window...
     */
    if (wwin) {
      Progress = Progress + Step;
      if (Progress - lastProgress >= 1.0) {
        progressObj = Tcl_NewDoubleObj(Progress);
        if (progressObj) {
          Tcl_IncrRefCount(progressObj);
          objv[0] = ggi_wait_update;
          objv[1] = &CDM_Obj_NULL;
          objv[2] = progressObj;
          objv[3] = wwin;
          Tcl_SetDoubleObj(progressObj, Progress);
          Tcl_EvalObjv(CDM_Interp, 4, objv, TCL_EVAL_GLOBAL);
          lastProgress = Progress;
          Tcl_DecrRefCount(progressObj);
        }
      }
    } /* if (wwin) */
  } /* Iterate over all Vector Annotations... */

  /*
   * Clean up memory
   */
  if (CellAnns) {
    for (i = 0; i < CellNumber; i++) {
      if (CellAnns[i]) Tcl_DecrRefCount(CellAnns[i]);
      if (CellSpec[i].CellFailObj) Tcl_DecrRefCount(CellSpec[i].CellFailObj);
    }
    Tcl_Free((char *) CellAnns);
  }
  if (wwin)            Tcl_DecrRefCount(wwin);
  if (ggi_wait_update) Tcl_DecrRefCount(ggi_wait_update);
  if (QMark)           Tcl_DecrRefCount(QMark);
  if (Sentences)       Tcl_DecrRefCount(Sentences);
  if (VAnns)           Tcl_DecrRefCount(VAnns);
  if (CAnns)           Tcl_DecrRefCount(CAnns);
  if (tmpObj)          Tcl_DecrRefCount(tmpObj);
  if (tmpObj2)         Tcl_DecrRefCount(tmpObj2);
  if (Pre)             Tcl_DecrRefCount(Pre);
  if (In)              Tcl_DecrRefCount(In);
  if (Post)            Tcl_DecrRefCount(Post);
  if (Vector)          Tcl_DecrRefCount(Vector);
  if (VectorFormatObj) Tcl_DecrRefCount(VectorFormatObj);
  if (ColId)           Tcl_DecrRefCount(ColId);
  return Vectors;

error:
  if (CellAnns) {
    for (i = 0; i < CellNumber; i++) {
      if (CellAnns[i]) Tcl_DecrRefCount(CellAnns[i]);
      if (CellSpec[i].CellFailObj) Tcl_DecrRefCount(CellSpec[i].CellFailObj);
    }
    Tcl_Free((char *) CellAnns);
  }
  if (attrName)        CDM_Free(attrName);
  if (defaultValue)    CDM_Free(defaultValue);
  if (wwin)            Tcl_DecrRefCount(wwin);
  if (ggi_wait_update) Tcl_DecrRefCount(ggi_wait_update);
  if (QMark)           Tcl_DecrRefCount(QMark);
  if (Vectors)         Tcl_DecrRefCount(Vectors);
  if (Sentences)       Tcl_DecrRefCount(Sentences);
  if (VAnns)           Tcl_DecrRefCount(VAnns);
  if (CAnns)           Tcl_DecrRefCount(CAnns);
  if (tmpObj)          Tcl_DecrRefCount(tmpObj);
  if (tmpObj2)         Tcl_DecrRefCount(tmpObj2);
  if (Pre)             Tcl_DecrRefCount(Pre);
  if (In)              Tcl_DecrRefCount(In);
  if (Post)            Tcl_DecrRefCount(Post);
  if (Vector)          Tcl_DecrRefCount(Vector);
  if (VectorFormatObj) Tcl_DecrRefCount(VectorFormatObj);
  if (ColId)           Tcl_DecrRefCount(ColId);
  return NULL;
}; /* CDM_CreateVectors */

/*
 * ---- CDM_Close:
 *  DESCRIPTION:
 *   <English:
 *     This function will close the given (opened) Document object. All the
 *     objects that the Document to be closed owns, as well the Document
 *     object itself, will be deleted from memory and will be unregistered
 *     from the current Tcl interpreter (CDM_Interp). The Document object must
 *     be a valid CDM Document (of type CDM_Document) and must have been
 *     returned by one of the functions CDM_GetByExternalId, CDM_GetDocument,
 *     CDM_FirstDocument, CDM_NextDocument or CDM directly (as a parameter to
 *     the main function of a module).
 *
 *     Note that this function will not save the specified Document. In order
 *     for the Document to be saved, the caller must previously call CDM_Sync.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function will not save the specified Document before deleting it
 *     from memory.
 *
 *     This function is equivelant to tip_Close (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR.
 *     If the requested Document is successfully closed, TCL_OK will be
 *     returned. In case of an error, the return value will be TCL_ERROR and an
 *     error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetByExternalId, CDM_GetDocument, CDM_FirstDocument,
 *     CDM_NextDocument, CDM_Sync, tip_Close
 *   >:&:
 */
CDM_EXPORT
int CDM_Close(CDM_Document Document) {
  CDM_ListNode *ptr;

  if (Document == NULL)
    return CDM_ReportError("CDM_Close: NULL Document Pointer");
  if (Document->Type != CDM_TYPE_DOCUMENT)
    return CDM_ReportError("CDM_Close: Invalid Document Type");

  Document->RefCount--;
  /*
   * Check Document's reference count. If > 0, we cannot close the document,
   * as it is still referenced by some process...
   */
  if (Document->RefCount > 0) return TCL_OK;

  /*
   * Remove Document from Opened Documents...
   */
  ptr = Document->DocumentPtr;
  if (ptr == NULL) return CDM_ReportError("Document is not opened");
  ptr->ptr[CDM_DOCUMENT_OPENED]  = NULL;
  ptr->ptr[CDM_DOCUMENT_POINTER] = NULL;

  return CDM_Free(Document);
} /* CDM_Close */

/*
 * ---- CDM_DeleteAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function accepts two arguments: a Document object and a UTF-8 string
 *     that defines the Annotations to be deleted. According to the value
 *     specified by the "Type" parameter, this function will delete
 *     Annotations or Annotation Attributes from the specified Document.
 *
 *     If the value of the "Type" parameter is an empty string or NULL, then
 *     ALL the Annotations that the specified Document currently has will
 *     be deleted
 *     from this Document object. The used Annotation Ids will be also reseted.
 *     This means that the first Annotation that will be added to this
 *     Document will take the value "0" as Annotation Id, if it does not
 *     already have an Id. If the
 *     deletion is done successfully, the total number of the deleted
 *     Annotations will be returned.
 *
 *     If Annotations having as type the value of the "Type" parameter exist,
 *     then all these Annotations will be deleted from the Document. If the
 *     deletion is done successfully, the total number of the deleted
 *     Annotations will be returned.
 *
 *     If all Annotations have been searched and no Annotation having as type
 *     the value of the "Type" parameter was found, the value of the "Type"
 *     parameter is examined whether it can be splitted in two words.  If this
 *     is the case (the value of the "Type" parameter contains at least one
 *     space character), then the value is splitted in two parts on the first
 *     space character that is found. The first word will be used as an
 *     Annotation type and the rest of the value (second word)
 *     will be used as an Attribute
 *     name. All Annotations will be seached. If Annotations having as type
 *     the first word are found, they are examined whether they contain an
 *     Attribute named as the second word. If such an Attribute is found in an
 *     Annotation, it will be deleted from the Annotation. The total number of
 *     deleted Attributes will be returned as the return value of this
 *     function.
 *
 *     If this function fails to delete any Annotation or any Annotation
 *     Attribute, then the value "0" will be deleted. In case of an error, the
 *     value "-1" will be returned, and an error message describing the error
 *     will be left at the current Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                          (CDM_Document)
 *     Type:        The type of the Annotations to be deleted.  (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_DeleteAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the number of the deleted Annotations or
 *     Annotation Attributes. If nothing was deleted, this function will
 *     return "0". In case of an error, he value "-1" will be returned, and an
 *     error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_FindMaxUsedAnnotationId, tip_DeleteAnnotations
 *   >:&:
 */
CDM_EXPORT
int CDM_DeleteAnnotations(CDM_Document Document, char *Type) {
#ifndef DONT_HASH_DOC_ANNOTATIONS
  Tcl_HashEntry *entryPtr;
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  Tcl_Obj *result = NULL;
  long i;
  int Deleted = 0;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_DeleteAnnotations: Aborting...");
    return -1;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_DeleteAnnotations: NULL Document Pointer");
    return -1;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_DeleteAnnotations: Invalid Document Type");
    return -1;
  }
  CDM_DEBUG("CDM_DeleteAnnotations: Removing Annotation: %s\n", Type);
  if (Type == NULL || CDM_Strlen(Type) == 0) {
    // If Type == {}, Remove all Annotations...
    for (i=0; i <= Document->AnnotationMaxID; i++) {
      if (CDM_RemoveAnnotation(Document, i) == TCL_OK) Deleted ++;
    }
    Document->AnnotationMaxID = 0;
  } else {
    for (i=0; i <= Document->AnnotationMaxID; i++) {
#ifdef DONT_HASH_DOC_ANNOTATIONS
      result = (Document->Annotations_List)->Ann(i);
      if (result != NULL) {
#else  /* DONT_HASH_DOC_ANNOTATIONS */
#ifdef CDM_ONE_WORD_KEYS
      entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),
                                   (char *) i);
#else
      snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", i);
      entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),CDM_Temp);
#endif /* CDM_ONE_WORD_KEYS */
      if (entryPtr != NULL) {
        result = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
#endif /* DONT_HASH_DOC_ANNOTATIONS */
        if (CDM_Strcmp(CDM_GetType(result), Type) == 0) {
          CDM_RemoveAnnotation(Document, i);
          Deleted ++;
        }
      }
    }
  }

  // If we have deleted nothing, and the Type has 2 items, then we should
  // have removed an ATTRIBUTE...
  if (Deleted == 0) {
    char *attr, *type, *ann_type;
    CDM_Annotation ann;

    type = CDM_Strdup(Type);
    attr = strstr(type, " ");
    if (attr == NULL) {
      CDM_Free(type);
      return 0;
    }
    while (*attr == ' ' && *attr != '\0') {
      *attr = '\0'; attr++;
    }
    if (*attr == '\0') {
      CDM_Free(type);
      return 0;
    }

    for (i=0; i <= Document->AnnotationMaxID; i++) {
#ifdef DONT_HASH_DOC_ANNOTATIONS
      ann = (Document->Annotations_List)->Ann(i);
      {
#else  /* DONT_HASH_DOC_ANNOTATIONS */
#ifdef CDM_ONE_WORD_KEYS
      entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),
                                   (char *) i);
#else
      snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", i);
      entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),CDM_Temp);
#endif /* CDM_ONE_WORD_KEYS */
      if (entryPtr != NULL) {
        ann = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
        if (ann == NULL) printf("NULL Annotation????\n");
#endif /* DONT_HASH_DOC_ANNOTATIONS */
        if (ann != NULL) {
          ann_type = CDM_GetType(ann);
          /*if (CDM_Strncmp(ann_type, Type, -1) == 0) { Petasis, 03/02/2000*/
          if (CDM_Strcmp(ann_type, type) == 0) {
            if (CDM_GetAttribute(ann, attr) != NULL) {
              result = CDM_RemoveAttribute(ann, attr);
              if (result == NULL) {
                CDM_ReportError("Cannot remove Attribute \"%s\" from "
                                "Annotation \"%s\"", attr, Tcl_GetString(ann));
                CDM_Free(type);
                return -1;
              }
              if (CDM_AddAnnotation(Document, result) < 0) {
                CDM_Free(type);
                return -1;
              }
              Deleted ++;
            }
          }
        }
      }
    }
    CDM_Free(type);
  }
  if (Deleted) {
    Document->Modified = 1;
    ((CDM_Collection) Document->ParentPtr)->ModifiedDocuments = 1;
    CDM_FindMaxUsedAnnotationId(Document);
  }
  return Deleted;
} /* CDM_DeleteAnnotations */

/*
 * ---- CDM_DisplaceAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function displases or "moves" the Annotations of the specified
 *     Document by "displacement" characters. This function will iterate over
 *     all Annotations contained in the Document. For each Annotation, all
 *     spans contained in the span set will be examined: for each offset in
 *     the span (either start or end), if it is equal or greater than the value
 *     provided by the "offset" parameter, then the value of the
 *     "displacement" parameter will be added.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:  A Document Object                               (CDM_Document)
 *     offset:    A character offset after which all offsets are modified (long)
 *     displacement: The number of characters that offsets will be moved  (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_DisplaceAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int). As this function ignores any possible errors, the
 *     return value will always be TCL_OK.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_DisplaceAnnotation, tip_DisplaceAnnotations
 *   >:&:
 */
CDM_EXPORT
int CDM_DisplaceAnnotations(CDM_Document Document, long offset,
                                            long displacement) {
  CDM_AnnotationSet Annotations;
  CDM_Annotation ann;
  Tcl_Obj **Ann;
  int status, items, i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_DisplaceAnnotations: Aborting...");
    return TCL_ERROR;
  }

  if (Document == NULL)
    return CDM_ReportError("CDM_DisplaceAnnotations: NULL Document Pointer");
  if (Document->Type != CDM_TYPE_DOCUMENT)
    return CDM_ReportError("CDM_DisplaceAnnotations: Invalid Document Type");

  /* If displacement is 0, there is nothing to do... */
  if (displacement == 0) return TCL_OK;

  /*
   * Get all Annotations from the Document...
   */
  Annotations = CDM_GetAnnotations(Document);
  if (Annotations == NULL) return TCL_ERROR;

  /*
   * Iterate over all Annotations...
   */
  status = Tcl_ListObjGetElements(Document->Interp, Annotations, &items, &Ann);
  if (status != TCL_OK) {
    CDM_Free(Annotations);
    return TCL_ERROR;
  }

  for (i=0; i<items; i++) {
    /*
     * Avoid to use CDM_DisplaceAnnotation, simply to save time for checking
     * the validity of Annotations. Annotations are already in Document, so
     * they must be valid...
     */
    ann = CDM_InternalDisplaceAnnotation(Ann[i], offset, displacement);
    if (ann == NULL) {
      CDM_Free(Annotations);
      return TCL_ERROR;
    }
    /* CDM_DisplaceAnnotation will return the given Annotation if a
     * displacement was not needed... */
    if (ann != Ann[i]) {
      /*
       * Avoid to use CDM_AddAnnotation, as this function does various
       * optimisations regarding cached memory. We don't need all this fuss,
       * as the Annotation was already in the Document...
       * CDM_AddAnnotation(Document, ann);
       */
      CDM_DocumentInternalAddAnnotation(Document, ann);
      NumberOfDisplacedAnnotations++;
    }
  }

  CDM_Free(Annotations);
  /*
   * We have displaced our Annotations, an action that may have generated too
   * many unneeded entries in our memory cache for long integers. Try to
   * remove excess (unused) entries...
   * This may take some more time, but we save large amounts of memory.
   * Imagine making a few thousants displacements without clearing this
   * cache...
   */
  if (NumberOfDisplacedAnnotations >
      CDM_MAX_NUMBER_OF_DISPLACEMENTS_BEFORE_CLEAN) {
    ObjectCache.RemoveUnusedLongs();
    NumberOfDisplacedAnnotations = 0;
  }

  return TCL_OK;
} /* CDM_DisplaceAnnotations */

/*
 * ---- CDM_DeleteAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function accepts three arguments: a Document object, a UTF-8 string
 *     that defines the Annotations to be deleted and a boolean expression
 *     that must be true in order for an Annotation to be deleted.
 *     According to the value specified by the "Type" parameter, this function
 *     will delete Annotations or Annotation Attributes from the specified
 *     Document.
 *
 *     If all Annotations have been searched and no Annotation having as type
 *     the value of the "Type" parameter was found, the value of the "Type"
 *     parameter is examined whether it can be splitted in two words.  If this
 *     is the case (the value of the "Type" parameter contains at least one
 *     space character), then the value is splitted in two parts on the first
 *     space character that is found. The first word will be used as an
 *     Annotation type and the rest of the value (second word)
 *     will be used as an Attribute
 *     name. All Annotations will be seached. If Annotations having as type
 *     the first word are found, they are examined whether they contain an
 *     Attribute named as the second word. If such an Attribute is found in an
 *     Annotation and the boolean expression is true for this Annotation,
 *     it will be deleted from the Annotation. The total number of
 *     deleted Attributes will be returned as the return value of this
 *     function.
 *
 *     The boolean expression that can be specified through the "Constraints"
 *     parameter will be evaluated before an Annotation or an Annotation
 *     Attribute is deleted. If the expression is true, then the deletion will
 *     be done. Else, this Annotation or Annotation Attribute will not be
 *     deleted. This expression can be any valid Tcl boolean expression (i.e. a
 *     boolean expression that will be accepted by the "expr" Tcl command),
 *     with the following excepion:
 *
 *     [[LI]]The boolean expression can contain references to the values of
 *     Atttributes of the Annotation, by utilising the notation
 *     "ann::%%%lt;Attribute Name%%%gt;". The only limitation is that
 *     "%%%lt;Attribute Name%%%gt;" cannot exceed 120 characters. For example,
 *     the following code will delete
 *     all Annotations from the specified Document that their type is token,
 *     they have a "type" Attribute which has as value "EFW" and also have a
 *     "pos" Attribute that has a value other than "NN":[[/LI]]
 *     [[br]][[code]][[center]]
 *     CDM_DeleteAnnotations(Doc, "token", "ann::type == \"EFW\" %%%amp;%%%amp; ann::pos != \"NN\"");
 *     [[/center]][[/code]][[br]]
 *
 *     If the value of the "Type" parameter is an empty string or NULL, then
 *     all Annotations will be searched for deletion, regardless their type.
 *     If the value of the "Constraints" parameter is an empty string or NULL,
 *     then the expression is ignored and the deletions will be based only on
 *     the value of the "Type" parameter.
 *     Note that if all Annotations are deleted from the specified Document,
 *     then the set of utilised Ids will also be reseted. As a result,
 *     the first Annotation that will be added to this Document after
 *     the deletion will take the value "0" as Annotation Id, if it does not
 *     already have an Id.
 *
 *     If any deletion are done successfully, the total number of the deleted
 *     (or modified) Annotations will be returned.
 *
 *     If this function fails to delete any Annotation or any Annotation
 *     Attribute, then the value "0" will be deleted. In case of an error, the
 *     value "-1" will be returned, and an error message describing the error
 *     will be left at the current Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                          (CDM_Document)
 *     Type:        The type of the Annotations to be deleted.  (char *)
 *     Constraints: A boolean expression.                       (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_DeleteAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the number of the deleted Annotations or
 *     Annotation Attributes. If nothing was deleted, this function will
 *     return "0". In case of an error, he value "-1" will be returned, and an
 *     error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_FindMaxUsedAnnotationId, tip_DeleteAnnotations,
 *     [[A HREF="../TclTk/TclCmd/expr.htm"]]The Tcl "expr" manual[[/A]].
 *   >:&:
 */
CDM_EXPORT
int CDM_DeleteAnnotations(CDM_Document Document, char *Type,
                          char *Constraints) {
#ifndef DONT_HASH_DOC_ANNOTATIONS
  Tcl_HashEntry *entryPtr;
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  Tcl_Obj *result = NULL, *AttributeNames, *Expr, *Boolean, **Name;
  Tcl_Parse parsePtr;
  Tcl_Token *token;
  char *ptr, *qptr, *start, name[128], TclSafeName[256];
  long i, RejectedAnnPrefix;
  int Deleted = 0, BoolVal, status, NameNu, j;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_DeleteAnnotations: Aborting...");
    return -1;
  }

  /* No need for complicated things if no expression... */
  if (Constraints == NULL || strlen(Constraints) == 0 ||
      strcmp(Constraints, "{}") == 0) {
    return CDM_DeleteAnnotations(Document, Type);
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_DeleteAnnotations: NULL Document Pointer");
    return -1;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_DeleteAnnotations: Invalid Document Type");
    return -1;
  }

  /*
   * Since we want to offer Attribute tokens for the Attribute Annotations, we
   * have to preprocess the boolean expresion. This is done in the following
   * steps:
   * 1) First, make sure the array we use for storing temporary values is
   *    clean. Ignore any errors...
   */
  Tcl_UnsetVar2(CDM_Interp, "::CDM::CDM_DeleteAnnotationsArray", NULL, 0);
  /*
   * 2) Foreach Annotation Attribute name (denoted with ann::<name>), create
   * an entry in the array, and keep its name for further reference...
   */
  start          = Constraints;
  ptr            = strstr(Constraints, "ann::");
  if (ptr != NULL) {
    /* Try to find the previous non-white character... */
    for (qptr=ptr-1; (*qptr==' ' || *qptr=='\t') && qptr>=Constraints; qptr--);
    if (*qptr == '\"' || *qptr == '{') ptr = qptr;
  }
  AttributeNames = Tcl_NewListObj(0, NULL);
  Expr           = Tcl_NewStringObj(NULL, -1);
  // Tcl_IncrRefCount(Expr);
  // printf("Original: {%s}\n", Constraints);
  while (ptr != NULL) {
    /* Copy unchanged bytes... */
    Tcl_AppendToObj(Expr, start, ptr - start);
    RejectedAnnPrefix = 0;
    /* Try to get the ann::<attribute name> substring. This is not an easy
     * task, as it may be enclosed in "" or {}. Use Tcl suitable routines to
     * parse such cases...
     */
    if (*ptr == '{' || *ptr == '\"' || *(ptr+5) == '{' || *(ptr+5) == '\"') {
      if (*ptr == '{') {
        status = Tcl_ParseBraces(CDM_Interp, ptr, -1, &parsePtr, 0,
                                 (const char **) &qptr);
      } else if (*ptr == '\"') {
        status = Tcl_ParseQuotedString(CDM_Interp, ptr, -1, &parsePtr, 0,
                                       (const char **) &qptr);
      } else if (*(ptr+5) == '{') {
        status = Tcl_ParseBraces(CDM_Interp, ptr+5, -1, &parsePtr, 0,
                                 (const char **) &qptr);
        RejectedAnnPrefix = 1;
      } else {
        status = Tcl_ParseQuotedString(CDM_Interp, ptr+5, -1, &parsePtr, 0,
                                       (const char **) &qptr);
        RejectedAnnPrefix = 1;
      }
      if (status != TCL_OK) {
        CDM_AppendError("\nCDM_DeleteAnnotationsArray: while evaluating "
                        "expression \n  (%s)", Constraints);
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        return -1;
      }
      if (parsePtr.numTokens != 1) {
        CDM_ReportError("CDM_DeleteAnnotationsArray: invalid "
                        "ann::<attribute name> "
                        "reference in constraints expression:\n  (%s)", ptr);
        Tcl_FreeParse(&parsePtr);
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        return -1;
      }
      token = parsePtr.tokenPtr;
      if (token->type != TCL_TOKEN_TEXT) {
        CDM_ReportError("CDM_DeleteAnnotationsArray: illegal "
                        "ann::<attribute name> "
                        "reference in constraints expression:\n  (%s)", ptr);
        Tcl_FreeParse(&parsePtr);
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        return -1;
      }
      if (token->size > 127) {
        CDM_ReportError("CDM_DeleteAnnotationsArray: too long "
                        "ann::<attribute name> "
                        "reference. <attribute name> must be < 128 characters");
        Tcl_FreeParse(&parsePtr);
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        return -1;
      }
      if (RejectedAnnPrefix) {
        strncpy(name, token->start, token->size);
        name[token->size] = '\0';
      } else {
        strncpy(name, token->start+5, token->size-5);
        name[token->size-5] = '\0';
      }
      Tcl_FreeParse(&parsePtr);
      start = qptr;
    } else {
      ptr += 5; status = 0;
      while (!CHAR_TYPE(*ptr) && status < 128) {
        name[status++] = *ptr; ++ptr;
      }
      name[status] = '\0';
      start = ptr;
    }
    /*
     * Given the ann::<attribute name> reference, try to construct a safe for
     * Tcl name that can be used in the expression so as to correctly refer to
     * the proper array element...
     */
    for (qptr = name, status = 0; *qptr != '\0'; qptr++) {
      if (*qptr == ' ' || *qptr == '\t' || *qptr == ';' || *qptr == '\"' ||
          *qptr == '{' || *qptr == '}' || *qptr == '[' || *qptr == ']' ||
          *qptr == '\\') {
        TclSafeName[status++] = '\\';
        TclSafeName[status++] = *qptr;
      } else {
        TclSafeName[status++] = *qptr;
      }
    }
    TclSafeName[status] = *qptr;

    /* Is the variable already defined? */
    if (Tcl_GetVar2Ex(CDM_Interp, "::CDM::CDM_DeleteAnnotationsArray",
        name, 0) == NULL) {
      /* The first time we meet this name... */
      Tcl_ListObjAppendElement(Document->Interp, AttributeNames,
                               CDM_NewStringObj(name));
      Tcl_SetVar2Ex(Document->Interp, "::CDM::CDM_DeleteAnnotationsArray",
                               name, &CDM_Obj_NULL, 0);
    }
    /* Substitute ann::<name> with the proper array element... */
    Tcl_AppendStringsToObj(Expr, "$::CDM::CDM_DeleteAnnotationsArray(",
                                 TclSafeName, ")", (char *) NULL);
    ptr   = strstr(start, "ann::");
    if (ptr != NULL) {
      /* Try to find the previous non-white character... */
      for (qptr=ptr-1; (*qptr==' ' || *qptr=='\t')&& qptr>=Constraints; qptr--);
      if (*qptr == '\"' || *qptr == '{') ptr = qptr;
    }
  }
  /* Finally, add the rest of the expression... */
  Tcl_AppendToObj(Expr, start, -1);
  //printf("Compiled: {%s}\n", Tcl_GetString(Expr));
  /* Place the various Attribute name objects in an array... */
  Tcl_ListObjGetElements(NULL, AttributeNames, &NameNu, &Name);

  CDM_DEBUG("CDM_DeleteAnnotations: Removing Annotation: %s\n", Type);
  if (Type == NULL || CDM_Strlen(Type) == 0) {
    // If Type == {}, Remove all Annotations...
    for (i=0; i <= Document->AnnotationMaxID; i++) {
      if (CDM_RemoveAnnotation(Document, i) == TCL_OK) Deleted ++;
    }
    Document->AnnotationMaxID = 0;
  } else {
    for (i=0; i <= Document->AnnotationMaxID; i++) {
#ifdef DONT_HASH_DOC_ANNOTATIONS
      result = (Document->Annotations_List)->Ann(i);
      if (result != NULL) {
#else  /* DONT_HASH_DOC_ANNOTATIONS */
#ifdef CDM_ONE_WORD_KEYS
      entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),
                                   (char *) i);
#else
      snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", i);
      entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),CDM_Temp);
#endif /* CDM_ONE_WORD_KEYS */
      if (entryPtr != NULL) {
        result = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
#endif /* DONT_HASH_DOC_ANNOTATIONS */
        if (CDM_Strcmp(CDM_GetType(result), Type) == 0) {
          /*
           * Now, we have to fill the array with the correct values. In case an
           * attribute does not exist, we place an empty value...
           */
          for (j=0; j<NameNu; j++) {
            Boolean = CDM_GetAttribute(result, Tcl_GetString(Name[j]));
            if (Boolean == NULL) {
              /* Set the empty value... */
              Tcl_ObjSetVar2(Document->Interp,
                        CDM_NewStringObj("::CDM::CDM_DeleteAnnotationsArray"),
                        Name[j], &CDM_Obj_NULL, 0);
            } else {
              Boolean = CDM_GetValueValue(Boolean);
              Tcl_ObjSetVar2(Document->Interp,
                        CDM_NewStringObj("::CDM::CDM_DeleteAnnotationsArray"),
                        Name[j], Boolean, 0);
            }
          }
          if (Tcl_ExprObj(Document->Interp, Expr, &Boolean) != TCL_OK) {
            CDM_Free(AttributeNames);
            CDM_Free(Expr);
            CDM_AppendError("\nCDM_DeleteAnnotations: invalid boolean "
                        "expression\n  (type=\"%s\", constraints=\"%s\")",
                        Type, Constraints);
            return -1;
          }
          /* Was it true? */
          status = Tcl_GetBooleanFromObj(Document->Interp, Boolean, &BoolVal);
          CDM_Free(Boolean);
          if (status != TCL_OK) {
            CDM_Free(AttributeNames);
            CDM_Free(Expr);
            CDM_AppendError("\nCDM_DeleteAnnotations: invalid boolean "
                        "expression result\n"
                        "(type=\"%s\", constraints=\"%s\")",
                        Type, Constraints);
            return -1;
          }
          if (BoolVal) {
            /* We succeded! Expression was true... */
            CDM_RemoveAnnotation(Document, i);
            Deleted ++;
          }
        }
      }
    }
  }

  // If we have deleted nothing, and the Type has 2 items, then we should
  // have removed an ATTRIBUTE...
  if (Deleted == 0) {
    char *attr, *type, *ann_type;
    CDM_Annotation ann;

    type = CDM_Strdup(Type);
    attr = strstr(type, " ");
    if (attr == NULL) {
      CDM_Free(type);
      CDM_Free(AttributeNames);
      CDM_Free(Expr);
      return 0;
    }
    while (*attr == ' ' && *attr != '\0') {
      *attr = '\0'; attr++;
    }
    if (*attr == '\0') {
      CDM_Free(type);
      CDM_Free(AttributeNames);
      CDM_Free(Expr);
      return 0;
    }

    for (i=0; i <= Document->AnnotationMaxID; i++) {
#ifdef DONT_HASH_DOC_ANNOTATIONS
      ann = (Document->Annotations_List)->Ann(i);
      {
#else  /* DONT_HASH_DOC_ANNOTATIONS */
#ifdef CDM_ONE_WORD_KEYS
      entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),
                                   (char *) i);
#else
      snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", i);
      entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),CDM_Temp);
#endif /* CDM_ONE_WORD_KEYS */
      if (entryPtr != NULL) {
        ann = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
        if (ann == NULL) printf("NULL Annotation????\n");
#endif /* DONT_HASH_DOC_ANNOTATIONS */
        if (ann != NULL) {
          ann_type = CDM_GetType(ann);
          /*if (CDM_Strncmp(ann_type, Type, -1) == 0) { Petasis, 03/02/2000*/
          if (CDM_Strcmp(ann_type, type) == 0) {
            if (CDM_GetAttribute(ann, attr) != NULL) {
              /*
               * Now, we have to fill the array with the correct values.
               * In case an attribute does not exist, we place an empty value...
               */
              for (j=0; j<NameNu; j++) {
                Boolean = CDM_GetAttribute(ann, Tcl_GetString(Name[j]));
                if (Boolean == NULL) {
                  /* Set the empty value... */
                  Tcl_ObjSetVar2(Document->Interp,
                      CDM_NewStringObj("::CDM::CDM_DeleteAnnotationsArray"),
                      Name[j], &CDM_Obj_NULL, 0);
                } else {
                  Boolean = CDM_GetValueValue(Boolean);
                  Tcl_ObjSetVar2(Document->Interp,
                      CDM_NewStringObj("::CDM::CDM_DeleteAnnotationsArray"),
                      Name[j], Boolean, 0);
                }
              }
              if (Tcl_ExprObj(Document->Interp, Expr, &Boolean) != TCL_OK) {
                CDM_Free(AttributeNames);
                CDM_Free(Expr);
                CDM_Free(type);
                CDM_AppendError("\nCDM_DeleteAnnotations: invalid boolean "
                        "expression\n  (type=\"%s\", constraints=\"%s\")",
                        Type, Constraints);
                return -1;
              }
              /* Was it true? */
              status =
                     Tcl_GetBooleanFromObj(Document->Interp, Boolean, &BoolVal);
              CDM_Free(Boolean);
              if (status != TCL_OK) {
                CDM_Free(AttributeNames);
                CDM_Free(Expr);
                CDM_Free(type);
                CDM_AppendError("\nCDM_DeleteAnnotations: invalid boolean "
                        "expression result\n"
                        "(type=\"%s\", constraints=\"%s\")",
                        Type, Constraints);
                return -1;
              }
              if (BoolVal) {
                /* We succeded! Expression was true... */
                result = CDM_RemoveAttribute(ann, attr);
                if (result == NULL) {
                  CDM_ReportError("Cannot remove Attribute \"%s\" from "
                                 "Annotation \"%s\"", attr, Tcl_GetString(ann));
                  CDM_Free(AttributeNames);
                  CDM_Free(Expr);
                  CDM_Free(type);
                  return -1;
                }
                if (CDM_AddAnnotation(Document, result) < 0) {
                  CDM_Free(AttributeNames);
                  CDM_Free(Expr);
                  CDM_Free(type);
                  return -1;
                }
                Deleted ++;
              }
            }
          }
        }
      }
    }
    CDM_Free(type);
  }
  CDM_Free(AttributeNames);
  CDM_Free(Expr);
  if (Deleted) {
    Document->Modified = 1;
    ((CDM_Collection) Document->ParentPtr)->ModifiedDocuments = 1;
    CDM_FindMaxUsedAnnotationId(Document);
  }
  return Deleted;
} /* CDM_DeleteAnnotations */

CDM_EXPORT
int CDM_DeleteAnnotations(CDM_Document Document, char *Type,
                          Tcl_Obj *Constraints) {
  return CDM_DeleteAnnotations(Document, Type, Tcl_GetString(Constraints));
}; /* CDM_DeleteAnnotations */

/*
 * ---- CDM_GetAnnotatedTextRanges:
 *  DESCRIPTION:
 *   <English:
 *    This function will return a set of strings from the text of the
 *    specified Document object
 *    that are annotated by the provided Annotation. In simple words, this
 *    function will examine the span set contained in the provided Annotation
 *    and return in a list all text ranges (from the text of the
 *    specified Document object) indexed by each one of the spans.
 *
 *    CDM does not own the returned object: the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *    associated with it.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. (In this case an empty
 *    object will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Document:   A Document object.                           (CDM_Document)
 *    Annotation: The Annotation that indexes the text ranges. (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to tip_GetAnnotatedTextRanges (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a set of ByteSequence/RawData objects (of type
 *    CDM_ByteSequence or CDM_RawData) containing the text ranges (or
 *    portions) from the text contained in the "Document" object indexed by
 *    the "Annotation" spans.
 *    In order to retrieve the individual text ranges, the user can call
 *    CDM_Nth as usual with set objects.
 *
 *    Note that CDM does not own the returned object:
 *    the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *    associated with it.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. (In this case an empty
 *    object will be returned.)
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_GetByteSequence, CDM_GetRawData, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], CDM_Free,
 *    CDM_GetFirstAnnotatedTextRange, tip_GetAnnotatedTextRanges
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequenceSet CDM_GetAnnotatedTextRanges(CDM_Document Document,
                                            CDM_Annotation Annotation) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetAnnotatedTextRanges: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_GetAnnotatedTextRanges: NULL Document Type");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetAnnotatedTextRanges: Invalid Document Type");
    return NULL;
  }
  return CDM_GetAnnotatedTextRanges(CDM_GetRawData(Document), Annotation);
} /* CDM_GetAnnotatedTextRanges */

/*
 * ---- CDM_GetFirstAnnotatedTextRange:
 *  DESCRIPTION:
 *   <English:
 *    This function will return a the text range from the text of the
 *    specified Document object that is annotated by the first span of
 *    the provided Annotation.
 *    In simple words, this function will retrieve the first span from the
 *    provided Annotation and return the text range (from the text of the
 *    specified Document object) that is indexed by this span.
 *
 *    CDM does not own the returned object: the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *    associated with it.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. (In this case an empty
 *    object will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:   A Document object.                          (CDM_Document)
 *     Annotation: The Annotation that indexes the text range. (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to tip_GetFirstAnnotatedTextRange (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a ByteSequence/RawData object (of type
 *    CDM_ByteSequence or CDM_RawData) containing the text range (or
 *    portion) from the text contained in the "Document" object indexed
 *    by the first span of the "Annotation" parameter.
 *
 *    Note that CDM does not own the returned object:
 *    the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *    associated with it.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. In this case an empty
 *    object (i.e. one containing the string "") will be returned.)
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_GetByteSequence, CDM_GetRawData, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], CDM_Free,
 *    CDM_GetAnnotatedTextRanges, tip_GetFirstAnnotatedTextRange
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequence CDM_GetFirstAnnotatedTextRange(CDM_Document Document,
                                            CDM_Annotation Annotation) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetFirstAnnotatedTextRange: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_GetFirstAnnotatedTextRange: NULL Document Type");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetFirstAnnotatedTextRange: Invalid Document Type");
    return NULL;
  }
  return CDM_GetFirstAnnotatedTextRange(CDM_GetRawData(Document), Annotation);
} /* CDM_GetFirstAnnotatedTextRange */

/*
 * ---- CDM_FindMaxUsedAnnotationId:
 *  DESCRIPTION:
 *   <English:
 *     This function will examine the Annotations stored in the specified
 *     Document to find the maximum used Annotation Id. This function will
 *     make the proper arrangements so as the next Annotation that will be
 *     added to this Document to receive as Id the following Id after the
 *     identified maximum Ids.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                     (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function does not have an equivelant in the Tcl API.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the Id that will be received by the next
 *     Annotation that will be added to this Document.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_DeleteAnnotations
 *   >:&:
 */
CDM_EXPORT
long CDM_FindMaxUsedAnnotationId(CDM_Document Document) {
  CDM_Annotation Annotation;
  long i, newId;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_FindMaxUsedAnnotationId: Aborting...");
    return -1;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_FindMaxUsedAnnotationId: NULL Document Type");
    return -1;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_FindMaxUsedAnnotationId: Invalid Document Type");
    return -1;
  }

  /*
   * Iterate over all possible Annotations...
   */
  newId = -1; //Document->AnnotationMaxID - 1;

  for (i=Document->AnnotationMaxID; i >= 0; i--) {
    Annotation = CDM_GetAnnotation(Document, i);
    if (Annotation != NULL) {
      newId = i;
      break;
    }
  }
  Document->AnnotationMaxID = ++newId;
  return Document->AnnotationMaxID;
} /* CDM_FindMaxUsedAnnotationId */

/*
 * ---- CDM_FirstAnnotationContaining:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation object (of type
 *     CDM_Annotation) that will be the first Annotation (with Annotations
 *     sorted by their Id) of the
 *     specified Document that includes the specified text position (as given
 *     by the parameter "Position") into its spans. If no Annotation is found
 *     that includes the specified position, NULL will be returned and
 *     the empty object will be placed as a result in the current active
 *     Tcl interpreter (CDM_Interp). The caller can use the function
 *     CDM_ActiveInterpreterResultEmpty to check whether the
 *     current active Tcl interpreter (CDM_Interp) has as result the empty
 *     object.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Position:    Position in text.             (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_FirstAnnotationContaining (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns an Annotation object (of type
 *    CDM_Annotation). The returned object is owned by CDM.
 *
 *    If an error occurs, then NULL will be returned and an error message
 *    will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_FirstAnnotationAt, CDM_FirstAnnotationInRange, CDM_FirstAnnotationMatchingRange,
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_FirstAnnotationContaining
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_FirstAnnotationContaining(CDM_Document Document,
                                                          const long Position) {
  CDM_Annotation Ann;
  CDM_SpanSet    Spans;
  Tcl_Obj        **Span;
  long start, end;
  int items, i, j, status;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_FirstAnnotationContaining: Aborting...");
    return NULL;
  }

  if (Position < 0) {
    CDM_ReportError("CDM_FirstAnnotationContaining: invalid position %d",
                    Position);
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_FirstAnnotationContaining: Invalid Document Type");
    return NULL;
  }

  for (i=0; i <= Document->AnnotationMaxID; i++) {
    if (CDM_AbortCurrentProcess) {
      if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
      CDM_ReportError("CDM_GetAnnotation: Aborting...");
      return NULL;
    }
    Ann = CDM_GetAnnotation(Document, i);
    if (Ann == NULL) continue;
#ifdef ENABLE_ANNOTATION_OBJECT
    if (ANNOTATION_OBJ(Ann) && CDMAnnObj_GetStart(Ann) != -1) {
      /* This Annotation has a single span. Handle it directly... */
      start = CDMAnnObj_GetStart(Ann);
      end   = CDMAnnObj_GetEnd(Ann);
      if (start <= Position && Position <= end) {
        return Ann;
      }
    } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
      /* Either the Annotation object type is disabled, or the Annotation has
       * more than one Span... */
      Spans = CDM_GetSpans(Ann);
      if (Spans == NULL) {
        CDM_ReportError("CDM_FirstAnnotationContaining: Document has "
                        "Annotation without Span?\n");
        CDM_AppendError(Tcl_GetString(Ann));
        return NULL;
      }
      status = Tcl_ListObjGetElements(CDM_Interp, Spans, &items, &Span);
      if (status != TCL_OK) {
        CDM_ReportError("CDM_FirstAnnotationContaining: Error Getting Spans "
                        "From Annotation ");
        CDM_AppendError(Tcl_GetString(Ann));
        return NULL;
      }
      for (j=0; j<items; j++) {
        if (CDM_GetSpanOffsets(Span[j], &start, &end) != TCL_OK) {
          CDM_ReportError("CDM_FirstAnnotationContaining: Error Getting Span "
                          "%d From Annotation", j);
          CDM_AppendError(Tcl_GetString(Ann));
          return NULL;
        }
        if (start <= Position && Position <= end) {
          return Ann;
        }
      }
#ifdef ENABLE_ANNOTATION_OBJECT
    }
#endif /* ENABLE_ANNOTATION_OBJECT */
  } /* for (i=0; i <= Document->AnnotationMaxID; i++) */
  return NULL;
} /* CDM_FirstAnnotationContaining */

/*
 * ---- CDM_FirstAnnotationContaining:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation object (of type
 *     CDM_Annotation) that will be the first Annotation (with Annotations
 *     sorted by their Id) of the
 *     specified Document that include the two specified text positions
 *     (as given by the parameters "Position1" and "Position2")
 *     into its spans. If no Annotation is found
 *     that includes the specified positions, NULL will be returned and
 *     the empty object will be placed as a result in the current active
 *     Tcl interpreter (CDM_Interp). The caller can use the function
 *     CDM_ActiveInterpreterResultEmpty to check whether the
 *     current active Tcl interpreter (CDM_Interp) has as result the empty
 *     object.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Position1:   Position in text.             (long)
 *     Position2:   Position in text.             (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_FirstAnnotationContaining (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation object (of type
 *     CDM_Annotation). The returned object is owned by CDM.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_FirstAnnotationAt, CDM_FirstAnnotationInRange, CDM_FirstAnnotationMatchingRange,
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_FirstAnnotationContaining
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_FirstAnnotationContaining(CDM_Document Document,
                                   const long Position1, const long Position2) {
  CDM_Annotation Ann;
  CDM_SpanSet    Spans;
  Tcl_Obj        **Span;
  long start, end;
  int items, i, j, status;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_FirstAnnotationContaining: Aborting...");
    return NULL;
  }

  if (Position1 < 0) {
    CDM_ReportError("CDM_FirstAnnotationContaining: invalid position %d",
                    Position1);
    return NULL;
  }
  if (Position2 < 0) {
    CDM_ReportError("CDM_FirstAnnotationContaining: invalid position %d",
                    Position2);
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_FirstAnnotationContaining: Invalid Document Type");
    return NULL;
  }

  for (i=0; i <= Document->AnnotationMaxID; i++) {
    if (CDM_AbortCurrentProcess) {
      if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
      CDM_ReportError("CDM_GetAnnotation: Aborting...");
      return NULL;
    }
    Ann = CDM_GetAnnotation(Document, i);
    if (Ann == NULL) continue;
#ifdef ENABLE_ANNOTATION_OBJECT
    if (ANNOTATION_OBJ(Ann) && CDMAnnObj_GetStart(Ann) != -1) {
      /* This Annotation has a single span. Handle it directly... */
      start = CDMAnnObj_GetStart(Ann);
      end   = CDMAnnObj_GetEnd(Ann);
      if (start <= Position1 && Position1 <= end &&
          start <= Position2 && Position2 <= end) {
        return Ann;
      }
    } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
      /* Either the Annotation object type is disabled, or the Annotation has
       * more than one Span... */
      Spans = CDM_GetSpans(Ann);
      if (Spans == NULL) {
        CDM_ReportError("CDM_FirstAnnotationContaining: Document has "
                        "Annotation without Span?\n");
        CDM_AppendError(Tcl_GetString(Ann));
        return NULL;
      }
      status = Tcl_ListObjGetElements(CDM_Interp, Spans, &items, &Span);
      if (status != TCL_OK) {
        CDM_ReportError("CDM_FirstAnnotationContaining: Error Getting Spans "
                        "From Annotation ");
        CDM_AppendError(Tcl_GetString(Ann));
        return NULL;
      }
      for (j=0; j<items; j++) {
        if (CDM_GetSpanOffsets(Span[i], &start, &end) != TCL_OK) {
          CDM_ReportError("CDM_FirstAnnotationContaining: Error Getting Span "
                          "%d From Annotation", j);
          CDM_AppendError(Tcl_GetString(Ann));
          return NULL;
        }
        if (start <= Position1 && Position1 <= end &&
            start <= Position2 && Position2 <= end) {
          return Ann;
        }
      }
#ifdef ENABLE_ANNOTATION_OBJECT
    }
#endif /* ENABLE_ANNOTATION_OBJECT */
  } /* for (i=0; i <= Document->AnnotationMaxID; i++) */
  return NULL;
} /* CDM_FirstAnnotationContaining */

/*
 * ---- CDM_GetAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Annotation object that has as Id the value
 *     specified by the "Id" parameter. If the requested Annotation does not
 *     exist, an error will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                     (CDM_Document)
 *     Id:          The Id of the desired Annotation.      (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetAnnotation (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation object (of type CDM_Annotation) if
 *     an Annotation having as Id the value of the "Id" parameter is found.
 *     In case that the specified Annotation does not exist, NULL is returned.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or directly modified in any way by the caller. Subsequent calls of
 *     either the C++ or Tcl API can modify the returned object. If the caller
 *     wants to keep a reference to the current state of the returned object,
 *     he must place such a request by using Tcl_IncrRefCount. When the caller
 *     wants to release the object (and return it to CDM) after he has placed
 *     a request with [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]], the functions
 *     Tcl_DecrRefCount or CDM_Free must be used.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]], [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_GetAnnotation
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_GetAnnotation(CDM_Document Document, long Id) {
#ifndef DONT_HASH_DOC_ANNOTATIONS
  Tcl_HashEntry *entryPtr;
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  Tcl_Obj *result;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetAnnotation: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_GetAnnotation: NULL Document Type");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetAnnotation: Invalid Document Type");
    return NULL;
  }
#ifdef DONT_HASH_DOC_ANNOTATIONS
  result = (Document->Annotations_List)->Ann(Id);
#else  /* DONT_HASH_DOC_ANNOTATIONS */
#ifdef CDM_ONE_WORD_KEYS
  entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),(char *)Id);
#else
  snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", Id);
  entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable), CDM_Temp);
#endif /* CDM_ONE_WORD_KEYS */
  if (entryPtr == NULL) return NULL;
  result = (CDM_Annotation) Tcl_GetHashValue(entryPtr);
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  if (result == NULL) {
    CDM_ReportError("Annotation with Id=%ld Not Found!", Id);
    return NULL;
  }
  return result;
} /* CDM_GetAnnotation */

/*
 * ---- CDM_GetAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Annotation object that has as Id the value
 *     specified by the "Id" parameter. If the requested Annotation does not
 *     exist, an error will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                     (CDM_Document)
 *     Id: A Tcl object containing the Id of the desired Annotation. (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetAnnotation (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation object (of type CDM_Annotation) if
 *     an Annotation having as Id the value of the "Id" parameter is found.
 *     In case that the specified Annotation does not exist, or the Tcl object
 *     (specified by the "Id" parameter) does not contain a long integer
 *     value, NULL is returned.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or directly modified in any way by the caller. Subsequent calls of
 *     either the C++ or Tcl API can modify the returned object. If the caller
 *     wants to keep a reference to the current state of the returned object,
 *     he must place such a request by using Tcl_IncrRefCount. When the caller
 *     wants to release the object (and return it to CDM) after he has placed
 *     a request with [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]], the functions
 *     Tcl_DecrRefCount or CDM_Free must be used.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]], [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_GetAnnotation
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_GetAnnotation(CDM_Document Document, Tcl_Obj *Id) {
  long AnnotationId;
  int status;
  CDM_THREAD_SPECIFIC_DATA;

  status = Tcl_GetLongFromObj(Document->Interp, Id, &AnnotationId);
  if (status != TCL_OK) {
    CDM_ReportError("CDM_GetAnnotation: \"%s\" is not a valid Annotation Id!",
                    Tcl_GetString(Id));
    return NULL;
  }
  return CDM_GetAnnotation(Document, AnnotationId);
} /* CDM_GetAnnotation */

/*
 * ---- CDM_GetAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new object
 *     (of type CDM_AnnotationSet) that
 *     contains all the Annotations that the specified Document currently has.
 *     CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *
 *     Note that this function will always create and return a new Annotation
 *     set object, even when the Document has no Annotations defined.
 *     (In this case an empty Annotation set will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_GetAnnotations
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_GetAnnotations(CDM_Document Document) {
#ifndef DONT_HASH_DOC_ANNOTATIONS
  Tcl_HashEntry *entryPtr;
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  Tcl_Obj *result;
  CDM_AnnotationSet Set;
  long i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetAnnotations: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetAnnotations: Invalid Document Type");
    return NULL;
  }

  /*
   * Initilise the new list object...
   */
  Set = Tcl_NewListObj(0, NULL);

  /*
   * Iterate over all possible Annotation IDs...
   */
  for (i=0; i <= Document->AnnotationMaxID; i++) {
#ifdef DONT_HASH_DOC_ANNOTATIONS
    result = (Document->Annotations_List)->Ann(i);
    {
#else  /* DONT_HASH_DOC_ANNOTATIONS */
#ifdef CDM_ONE_WORD_KEYS
    entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),(char *) i);
#else
    snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", i);
    entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable), CDM_Temp);
#endif /* CDM_ONE_WORD_KEYS */
    if (entryPtr != NULL) {
      result = (CDM_Annotation) Tcl_GetHashValue(entryPtr);
#endif /* DONT_HASH_DOC_ANNOTATIONS */
      if (result != NULL) {
        Tcl_ListObjAppendElement(Document->Interp, Set, result);
      }
    }
  }
  return Set;
} /* CDM_GetAnnotations */

/*
 * ---- CDM_GetAnnotationAttributeNames:
 *  DESCRIPTION:
 *   <English:
 *     This function will search all Annotations of the specified Document,
 *     that their type corresponds to the value of the "Type" parameter, and
 *     will return a new object (of type Tcl_Obj *) that will contain all the
 *     Attribute names found in the Annotation Attribute sets.
 *
 *     CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *
 *     Note that this function will always create and return a new object, even
 *     when the Document has no Annotations defined. (In this case an empty
 *     object will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document: A Document object.                           (CDM_Document)
 *     Type:     The type of the Annotations to be searched.  (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *     This function does not have an equivelant in the Tcl API. In order for
 *     this function to be accessed from the Tcl API, the user must call the
 *     method "GetAnnotationAttributeNames" of a Document object. For example,
 *     if "Doc0" is a Tcl Document object returned by the CDM to the user,
 *     the user must use the following code:
 *     [[p]][[center]][[code]]Doc0 GetAnnotationAttributeNames[[/code]][[/center]][[/p]]
 *     Note that this functionality does not exist under the GATE NLP
 *     Platform.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Tcl list object (of type Tcl_Obj *). The caller
 *     is responsible to free this object through Tcl_DecrRefCount or CDM_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, Tcl_DecrRefCount
 *   >:&:
 */
CDM_EXPORT
Tcl_Obj *CDM_GetAnnotationAttributeNames(CDM_Document Document,
                                                  Tcl_Obj *Type) {
  CDM_AnnotationSet Anns;
  Tcl_Obj **Ann, **Attr, *NameList, *CurAttrs, *CurAttrName;
  int AnnNu, AttrNu;
  register int i, j;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetAnnotationAttributeNames: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_GetAnnotationAttributeNames: NULL Document Type");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetAnnotationAttributeNames: Invalid Document Type");
    return NULL;
  }

  Anns = CDM_SelectAnnotations(Document, Tcl_GetString(Type));
  if (Anns == NULL) return NULL;

  NameList = Tcl_NewListObj(0, NULL);
  if (NameList == NULL) {
    CDM_ReportError("CDM_GetAnnotationAttributeNames: Out of Memory!");
    return NULL;
  }

  Tcl_ListObjGetElements(CDM_Interp, Anns, &AnnNu, &Ann);
  for (i=0; i<AnnNu; i++) {
    CurAttrs = CDM_GetAttributes(Ann[i]);
    Tcl_ListObjGetElements(CDM_Interp, CurAttrs, &AttrNu, &Attr);
    for (j=0; j<AttrNu; j++) {
      CurAttrName = CDM_GetNameObj(Attr[j]);
      if (CDM_Lsearch(NameList, CurAttrName, CDM_LSEARCH_EXACT) == NULL) {
        Tcl_ListObjAppendElement(CDM_Interp, NameList, CurAttrName);
      }
    }
  }
  Tcl_DecrRefCount(Anns);
  return NameList;
} /* CDM_GetAnnotationAttributeNames */

/*
 * ---- CDM_GetAnnotationAttributeValues:
 *  DESCRIPTION:
 *   <English:
 *     This function will search all Annotations of the specified Document,
 *     that their type corresponds to the value of the "Type" parameter, and
 *     will return a new object (of type Tcl_Obj *) that will contain all the
 *     Attribute values found in the Annotation Attribute sets whoose name is
 *     the value of the "Name" parameter.
 *
 *     CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *
 *     Note that this function will always create and return a new object, even
 *     when the Document has no Annotations defined. (In this case an empty
 *     object will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document: A Document object.                           (CDM_Document)
 *     Type:     The type of the Annotations to be searched.  (Tcl_Obj *)
 *     Name:     The name of the Attribute whoose values will be collected. (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *     This function does not have an equivelant in the Tcl API. In order for
 *     this function to be accessed from the Tcl API, the user must call the
 *     method "GetAnnotationAttributeValues" of a Document object. For example,
 *     if "Doc0" is a Tcl Document object returned by the CDM to the user,
 *     the user must use the following code:
 *     [[p]][[center]][[code]]Doc0 GetAnnotationAttributeValues type name[[/code]][[/center]][[/p]]
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Tcl list object (of type Tcl_Obj *). The caller
 *     is responsible to free this object through Tcl_DecrRefCount or CDM_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, Tcl_DecrRefCount
 *   >:&:
 */
CDM_EXPORT
Tcl_Obj *CDM_GetAnnotationAttributeValues(CDM_Document Document,
                                          Tcl_Obj *Type, Tcl_Obj *Name) {
  CDM_AnnotationSet Anns;
  CDM_Annotation *Ann;
  CDM_Attribute Attr;
  CDM_AttributeValue AttrVal;
  CDM_Object ValueList;
  char *name = Tcl_GetString(Name);
  int AnnNu, i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetAnnotationAttributeValues: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_GetAnnotationAttributeValues: NULL Document Type");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetAnnotationAttributeValues: Invalid Document Type");
    return NULL;
  }

  Anns = CDM_SelectAnnotations(Document, Tcl_GetString(Type));
  if (Anns == NULL) return NULL;

  ValueList = Tcl_NewListObj(0, NULL);
  if (ValueList == NULL) {
    CDM_ReportError("CDM_GetAnnotationAttributeValues: Out of Memory!");
    return NULL;
  }

  Tcl_ListObjGetElements(CDM_Interp, Anns, &AnnNu, &Ann);
  for (i=0; i<AnnNu; i++) {
    Attr = CDM_GetAttribute(Ann[i], name);
    if (Attr != NULL) {
      AttrVal = CDM_GetValueValue(Attr);
      if (AttrVal != NULL) {
        if (CDM_Lsearch(ValueList, AttrVal, CDM_LSEARCH_EXACT) == NULL) {
          Tcl_ListObjAppendElement(CDM_Interp, ValueList, AttrVal);
        }
      }
    }
  }
  Tcl_DecrRefCount(Anns);
  return ValueList;
} /* CDM_GetAnnotationAttributeValues */

/*
 * ---- CDM_GetAnnotationTypes:
 *  DESCRIPTION:
 *   <English:
 *     This function will search all Annotations of the specified Document, and
 *     will return a new object (of type Tcl_Obj *) that will contain all the
 *     Annotation types found in the Annotation set of the Document.
 *
 *     CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *
 *     Note that this function will always create and return a new object, even
 *     when the Document has no Annotations defined. (In this case an empty
 *     object will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document: A Document object. (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function does not have an equivelant in the Tcl API. In order for
 *     this function to be accessed from the Tcl API, the user must call the
 *     method "GetAnnotationTypes" of a Document object. For example,
 *     if "Doc0" is a Tcl Document object returned by the CDM to the user,
 *     the user must use the following code:
 *     [[p]][[center]][[code]]Doc0 GetAnnotationTypes[[/code]][[/center]][[/p]]
 *     Note that this functionality does not exist under the GATE NLP
 *     Platform.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Tcl list object (of type Tcl_Obj *). The caller
 *     is responsible to free this object through Tcl_DecrRefCount or CDM_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, Tcl_DecrRefCount
 *   >:&:
 */
CDM_EXPORT
Tcl_Obj *CDM_GetAnnotationTypes(CDM_Document Document) {
  CDM_AnnotationSet Anns;
  Tcl_Obj **Ann, *TypeList, *CurType;
  int AnnNu;
  register int i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetAnnotationTypes: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_GetAnnotationTypes: NULL Document Type");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetAnnotationTypes: Invalid Document Type");
    return NULL;
  }

  Anns = CDM_GetAnnotations(Document);
  if (Anns == NULL) return NULL;

  TypeList = Tcl_NewListObj(0, NULL);
  if (TypeList == NULL) {
    CDM_ReportError("CDM_GetAnnotationTypes: Out of Memory!");
    return NULL;
  }

  Tcl_ListObjGetElements(CDM_Interp, Anns, &AnnNu, &Ann);
  for (i=0; i<AnnNu; i++) {
    CurType = CDM_GetTypeObj(Ann[i]);
    if (CDM_Lsearch(TypeList, CurType, CDM_LSEARCH_EXACT) == NULL) {
      Tcl_ListObjAppendElement(CDM_Interp, TypeList, CurType);
    }
  }
  Tcl_DecrRefCount(Anns);
  return TypeList;
} /* CDM_GetAnnotationTypes */

/*
 * ---- CDM_GetAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function will search all the Attributes that the specified
 *     Document has, for an Attribute that has the same name as the "Name"
 *     parameter. If such an Attribute is found, the Attribute object is
 *     return. If an Attribute with the specified name does not exist, NULL is
 *     returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:  A Document object. (CDM_Document)
 *      Name:      Attribute name (const char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetAttribute (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Attribute object (of type CDM_Attribute) if an
 *     Attribute named as specified is found. In case that the specified
 *     Attribute does not exist, NULL is returned.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or directly modified in any way by the caller. Subsequent calls
 *     of either the C++ or Tcl API can modify the returned object. If the
 *     caller wants to keep a reference to the current state of the returned
 *     object, he must place such a
 *     request by using Tcl_IncrRefCount. When the caller wants to release the
 *     object (and return it to CDM), the functions Tcl_DecrRefCount of
 *     CDM_Free must be used.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]],
 *     tip_GetAttribute
 *   >:&:
 */
CDM_EXPORT
CDM_Attribute CDM_GetAttribute(CDM_Document Document, const char *Name) {
  Tcl_HashEntry *entryPtr;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetAttribute: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_GetAttribute: NULL Document Type");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetAttribute: Invalid Document Type");
    return NULL;
  }

#ifdef CDM_OBJ_HASH_KEYS
  /*
   * We don't add a reference to this object. If it is an old one, it's
   * refCount is left intact. If it is a new one, it will have a refCount of
   * 1, as it is referenced by the CDM hash table. It will be freed when the
   * table is freed...
   */
  Tcl_Obj *NameObj;
  NameObj = CDM_NewStringObj(Name);
  entryPtr = Tcl_FindHashEntry(&(Document->Attributes_HashTable),
                                (char *) NameObj);
#else /* CDM_OBJ_HASH_KEYS */
  entryPtr = Tcl_FindHashEntry(&(Document->Attributes_HashTable), Name);
#endif /* CDM_OBJ_HASH_KEYS */
  if (entryPtr == NULL) return NULL;
  return (CDM_Attribute) Tcl_GetHashValue(entryPtr);
} /* CDM_GetAttribute */

/*
 * ---- CDM_GetAttributes:
 *  DESCRIPTION:
 *   <English:
 *     This function returns a new object (of type CDM_AttributeSet) that
 *     contains all the Attributes that the specified Document currently
 *     has. CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *
 *     This function will always create and return a new Attribute set object,
 *     even when the Document has no Attributes defined. (In this case an
 *     empty Attribute set will be returned...)
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:  A Document object. (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetAttributes (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Attribute set object (of type CDM_AttributeSet).
 *     The caller is responsible to free this object through Tcl_DecrRefCount
 *     or CDM_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_GetAttributes
 *   >:&:
 */
CDM_EXPORT
CDM_AttributeSet CDM_GetAttributes(CDM_Document Document) {
  CDM_AttributeSet Attributes;
  Tcl_Interp *Interp = Document->Interp;
  Tcl_Obj    *AttrValue;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch searchPtr;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetAttributes: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_GetAttributes: NULL Document Type");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetAttributes: Invalid Document Type");
    return NULL;
  }
  Attributes = Tcl_NewListObj(0, NULL);
  if (Attributes == NULL) {
    CDM_Panic("CDM_GetAttributes: Tcl_NewListObj returned NULL - "
              "out of memory?\n");
  }

  /* Search Attributes Hash Table and Collect All Attributes... */
  entryPtr = Tcl_FirstHashEntry(&(Document->Attributes_HashTable),&searchPtr);
  while (entryPtr != NULL) {
    AttrValue = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    if (AttrValue == NULL) continue;
    if (Tcl_ListObjAppendElement(Interp, Attributes, AttrValue) != TCL_OK)
          goto error;
    entryPtr = Tcl_NextHashEntry(&searchPtr);
  }
  return Attributes;
error:
  Tcl_DecrRefCount(Attributes);
  return NULL;
} /* CDM_GetAttributes */

/*
 * ---- CDM_GetByteSequence:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Text of the specified Document. It will
 *     return an object (of type CDM_ByteSequence) that will contain the text
 *     of the Document. The caller can use the Tcl function Tcl_GetString in
 *     order to get the text stored in the returned object as a C pointer (of
 *     type char*). The text is stored in the UTF-8 encoding.
 *     In order to convert
 *     the returned C pointer into a specific encoding (like iso8859-1 or
 *     iso8859-7), the function CDM_UtfToExternal can be used.
 *
 *     Note that the returned object is owned by the CDM. The caller is not
 *     supposed to free or modify the returned object.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:  A Document object. (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetByteSequence (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a Text object (of type CDM_ByteSequence) that
 *     will contain the text (or byte sequence or raw data) of the specified
 *     Document. The returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetRawData, CDM_UtfToExternal, tip_GetByteSequence
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequence CDM_GetByteSequence(CDM_Document Document) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetByteSequence: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetByteSequence: Invalid Document Type");
    return NULL;
  }
  return Document->RawData;
} /* CDM_GetByteSequence */

/*
 * ---- CDM_GetByteSequence:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Text of the specified Document. It will
 *     return a C pointer (of type char*) that will contain the text
 *     of the Document. The text will be stored in a newly allocated memory
 *     space, using the requested by the value of the "Encoding" parameter
 *     encoding. The pointer to that new memory segment will be returned to
 *     the caller.
 *
 *     Note that the returned object is not owned by the CDM. The caller is
 *     responsible to use CDM_Free (or Tcl_Free) in order to free the string and
 *     the memory associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:  A Document object. (CDM_Document)
 *      Encoding:  The desired encoding that the text of the Document is to be converted to. (const char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function does not have an equivelant function (Tcl API). The Tcl
 *     core uses UTF-8 as its native encoding, so the convertion to a specific
 *     encoding is not needed.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a C pointer (of type char*) that
 *     will contain the text (or byte sequence or raw data) of the specified
 *     Document encoded in the requested encoding.  CDM does not own the
 *     returned memory segment: the caller is responsible to use CDM_Free (or
 *     Tcl_Free) in order to free the string and the memory associated with it.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, CDM_GetRawData, Tcl_Free
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetByteSequence(CDM_Document Document, const char *Encoding) {
  CDM_ByteSequence RawData;
  Tcl_DString Text;
  Tcl_Encoding encoding;
  char *Value;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetByteSequence: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetByteSequence: Invalid Document Type");
    return NULL;
  }

  /* Get Document's Text... */
  RawData = CDM_GetByteSequence(Document);
  if (RawData == NULL) {
    CDM_ReportError("Document has no Text?");
    return NULL;
  }
  /* Work arround a bug in Tcl_GetEncoding, where it modifies Interpreter's
   * result without checking whether it is shared or not! */
  Tcl_ResetResult(CDM_Interp);
  encoding = Tcl_GetEncoding(CDM_Interp, Encoding);
  if (encoding == NULL) return NULL; // Tcl_GetEncoding has left an error...

  /* Convert Text in desired encoding... */
  Tcl_DStringInit(&Text);
  Tcl_UtfToExternalDString(encoding, Tcl_GetString(RawData), -1, &Text);
  Tcl_FreeEncoding(encoding);

  Value = CDM_Strdup(Tcl_DStringValue(&Text));
  Tcl_DStringFree(&Text);
  return Value;
} /* CDM_GetByteSequence */

/*
 * ---- CDM_GetByteSequence:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Text of the specified Document. It will
 *     return a C pointer (of type char*) that will contain the text
 *     of the Document. The text will be stored in a newly allocated memory
 *     space, using the requested by the value of the "Encoding" parameter
 *     encoding. The pointer to that new memory segment will be returned to
 *     the caller.
 *
 *     Note that the returned object is not owned by the CDM. The caller is
 *     responsible to use CDM_Free (or Tcl_Free) in order to free the string and
 *     the memory associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:  A Document object. (CDM_Document)
 *      Encoding:  The desired encoding that the text of the Document is to be converted to. (Tcl_Encoding)
 *   >
 *  NOTES:
 *   <English:
 *     This function does not have an equivelant function (Tcl API). The Tcl
 *     core uses UTF-8 as its native encoding, so the convertion to a specific
 *     encoding is not needed.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a C pointer (of type char*) that
 *     will contain the text (or byte sequence or raw data) of the specified
 *     Document encoded in the requested encoding.  CDM does not own the
 *     returned memory segment: the caller is responsible to use CDM_Free (or
 *     Tcl_Free) in order to free the string and the memory associated with it.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, CDM_GetRawData, Tcl_GetEncoding, Tcl_Free
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetByteSequence(CDM_Document Document, const Tcl_Encoding encoding) {
  CDM_ByteSequence RawData;
  Tcl_DString Text;
  char *Value;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetByteSequence: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetByteSequence: Invalid Document Type");
    return NULL;
  }

  if (encoding == NULL) {
    CDM_ReportError("CDM_GetByteSequence: Invalid Encoding Parameter (NULL)");
    return NULL;
  }

  /* Get Document's Text... */
  RawData = CDM_GetByteSequence(Document);
  if (RawData == NULL) {
    CDM_ReportError("Document has no Text?");
    return NULL;
  }

  /* Convert Text in desired encoding... */
  Tcl_DStringInit(&Text);
  Tcl_UtfToExternalDString(encoding, Tcl_GetString(RawData), -1, &Text);

  Value = CDM_Strdup(Tcl_DStringValue(&Text));
  Tcl_DStringFree(&Text);
  return Value;
} /* CDM_GetByteSequence */

/*
 * ---- CDM_GetEncoding:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the encoding of the specified Document. The
 *     return value will be of type char* and will be owned by CDM. Its value
 *     will be a standart Tcl encoding value (like iso8859-7 or cp1253). For
 *     all available Tcl encodings please refer to the
 *     [[A HREF="../TclTk/TclLib/Encoding.htm"]]Tcl manuals[[/A]].
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetEncoding (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a UTF-8 string that represents the encoding of
 *     the specified Document.
 *
 *     If an error occurs, then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned memory segment is property of the CDM and should
 *     never be freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_SetEncoding, tip_GetEncoding,
 *     [[A HREF="../TclTk/TclLib/Encoding.htm"]]The Tcl Encoding manual[[/A]].
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetEncoding(CDM_Document Document) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetEncoding: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetEncoding: Invalid Document Type");
    return NULL;
  }
  return Document->Encoding;
} /* CDM_GetEncoding */

/*
 * ---- CDM_GetExternalId:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the external Id (XID) of the specified
 *     Document. The external Id is the (absolute) path of the initial file
 *     that contained the text which was used in order to create the specified
 *     Document. The returned value will be encoded using the UTF-8 encoding
 *     (thus enabling the existance of non Latin characters in the value).
 *
 *     The return value will be of type char* and will be owned by CDM.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:    A Document object.            (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetExternalId (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a UTF-8 string that represents the external Id of
 *     the specified Document.
 *
 *     If an error occurs, then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetExternalId
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetExternalId(CDM_Document Document) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetExternalId: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetExternalId: Invalid Document Type");
    return NULL;
  }
  return Document->DocumentPtr->ptr[CDM_DOCUMENT_XID];
} /* CDM_GetExternalId */

/*
 * ---- CDM_GetFirstAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will start an iteration over the Annotations of the
 *     specified type that the specified Document has. This function will
 *     return the first Annotation object that has as type the value
 *     specified by the "Type" parameter. If the requested Annotation does not
 *     exist, an error will be returned. Subsequent calls to
 *     CDM_GetNextAnnotation will return the rest of the Annotations of the
 *     same type, sorted according to their Id.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                     (CDM_Document)
 *     Type:        The type of the desired Annotations.   (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetFirstAnnotation (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation object (of type CDM_Annotation) if
 *     an Annotation having as type the value of the "Type" parameter is found.
 *     In case that the specified Annotation does not exist, NULL will be
 *     returned.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or directly modified in any way by the caller. Subsequent calls of
 *     either the C++ or Tcl API can modify the returned object. If the caller
 *     wants to keep a reference to the current state of the returned object,
 *     he must place such a request by using Tcl_IncrRefCount. When the caller
 *     wants to release the object (and return it to CDM), the functions
 *     Tcl_DecrRefCount or CDM_Free must be used.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetNextAnnotation, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]], [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_GetFirstAnnotation
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_GetFirstAnnotation(CDM_Document Document, char *Type) {
#ifndef DONT_HASH_DOC_ANNOTATIONS
  Tcl_HashEntry *entryPtr;
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  Tcl_Obj *result;
  long i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Reset the value of the current Annotation Id...
   */
  Document->CurrentAnnID = -1;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetFirstAnnotation: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_GetFirstAnnotation: NULL Document Type");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetFirstAnnotation: Invalid Document Type");
    return NULL;
  }

  for (i=0; i <= Document->AnnotationMaxID; i++) {
#ifdef DONT_HASH_DOC_ANNOTATIONS
    result = (Document->Annotations_List)->Ann(i);
#else  /* DONT_HASH_DOC_ANNOTATIONS */
#ifdef CDM_ONE_WORD_KEYS
    entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),(char *) i);
#else
    snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", i);
    entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable), CDM_Temp);
#endif /* CDM_ONE_WORD_KEYS */
    if (entryPtr == NULL) continue;
    result = (CDM_Annotation) Tcl_GetHashValue(entryPtr);
#endif /* DONT_HASH_DOC_ANNOTATIONS */
    if (result != NULL && strcmp(CDM_GetType(result), Type) == 0) {
      Document->CurrentAnnID = i;
      return result;
    }
  }
  return NULL;
} /* CDM_GetFirstAnnotation */

/*
 * ---- CDM_GetId:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Id of the specified
 *     Document. The Id is the file name of the initial file
 *     that contained the text which was used in order to create the specified
 *     Document. The returned value will be encoded using the UTF-8 encoding
 *     (thus enabling the existance of non Latin characters in the value).
 *
 *     The return value will be of type char* and will be owned by CDM.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:    A Document object.            (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetId (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a UTF-8 string that represents the Id of
 *     the specified Document.
 *
 *     If an error occurs, then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetId
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetId(CDM_Document Document) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetId: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetId: Invalid Document Type");
    return NULL;
  }
  return Document->DocumentPtr->ptr[CDM_DOCUMENT_ID];
} /* CDM_GetId */

/*
 * ---- CDM_GetNextAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the next Annotation of the iteration started
 *     with a call to CDM_GetFirstAnnotation, or of the Annotation returned by
 *     a previous call to CDM_GetNextAnnotation. If no more Annotations of the
 *     same type (which was specified through a call to CDM_GetFirstAnnotation)
 *     exist in the specified Document, NULL will be returned. NULL will
 *     also be returned if an error occurs.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                     (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetNextAnnotation (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation object (of type CDM_Annotation) if
 *     an Annotation having as type the value specified by a prevous call to
 *     CDM_GetFirstAnnotation exists in the specified Document.
 *     In case that such an Annotation does not exist, NULL will be returned.
 *     If NULL is returned, this function cannot be called unless
 *     CDM_GetFirstDocument is called again, in order to initialise a new
 *     iteration over Annotations of a specific type.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or directly modified in any way by the caller. Subsequent calls of
 *     either the C++ or Tcl API can modify the returned object. If the caller
 *     wants to keep a reference to the current state of the returned object,
 *     he must place such a request by using Tcl_IncrRefCount. When the caller
 *     wants to release the object (and return it to CDM), the functions
 *     Tcl_DecrRefCount or CDM_Free must be used.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetFirstAnnotation, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]], [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_GetNextAnnotation
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_GetNextAnnotation(CDM_Document Document) {
#ifndef DONT_HASH_DOC_ANNOTATIONS
  Tcl_HashEntry *entryPtr;
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  Tcl_Obj *result;
  char *Type;
  long i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetNextAnnotation: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_GetNextAnnotation: NULL Document Type");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetNextAnnotation: Invalid Document Type");
    return NULL;
  }

  if (Document->CurrentAnnID == -1) {
    CDM_ReportError("CDM_GetNextAnnotation: Please call CDM_GetFirstAnnotation"
                    " before calling CDM_GetNextAnnotation");
    return NULL;
  }

  /*
   * Get the current Annotation (the one returned by last call to
   * CDM_Get{First|Next}Annotation...
   */
  result = CDM_GetAnnotation(Document, Document->CurrentAnnID);
  if (result == NULL) {
    CDM_ReportError("CDM_GetNextAnnotation: Please call CDM_GetFirstAnnotation"
                    " before calling CDM_GetNextAnnotation");
    return NULL;
  }
  /*
   * Get the type of the last current Annotation...
   */
  Type = CDM_GetType(result);

  for (i=Document->CurrentAnnID+1; i <= Document->AnnotationMaxID; i++) {
#ifdef DONT_HASH_DOC_ANNOTATIONS
    result = (Document->Annotations_List)->Ann(i);
#else  /* DONT_HASH_DOC_ANNOTATIONS */
#ifdef CDM_ONE_WORD_KEYS
    entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),(char *) i);
#else
    snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", i);
    entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable), CDM_Temp);
#endif /* CDM_ONE_WORD_KEYS */
    if (entryPtr == NULL) continue;
    result = (CDM_Annotation) Tcl_GetHashValue(entryPtr);
#endif /* DONT_HASH_DOC_ANNOTATIONS */
    if (result != NULL && strcmp(CDM_GetType(result), Type) == 0) {
      Document->CurrentAnnID = i;
      return result;
    }
  }

  /*
   * No more Annotations of the desired type found. We are done :-)
   */
  Document->CurrentAnnID = -1;
  return NULL;
} /* CDM_GetNextAnnotation */

/*
 * ---- CDM_GetParent:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the parent of the specified
 *     Document. The parent of a Document is the Collection object that holds
 *     the specified Document. The return value will be the parent Collection
 *     object (of type CDM_Collection) and will be owned by CDM.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetParent (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the Collection object that contains the
 *     specified Document.
 *
 *     In case of an error NULL will be returned and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never
 *     be freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetParent
 *   >:&:
 */
CDM_EXPORT
CDM_Collection CDM_GetParent(CDM_Document Document) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetParent: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetParent: Invalid Document Type");
    return NULL;
  }
  return (CDM_Collection) Document->ParentPtr;
} /* CDM_GetParent */

/*
 * ---- CDM_GetRawData:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Text of the specified Document. It will
 *     return an object (of type CDM_ByteSequence) that will contain the text
 *     of the Document. The caller can use the Tcl function Tcl_GetString in
 *     order to get the text stored in the returned object as a C pointer (of
 *     type char*). The text will be stored using the UTF-8 encoding.
 *     In order to convert
 *     the returned C pointer into a specific encoding (like iso8859-1 or
 *     iso8859-7), the function CDM_UtfToExternal can be used.
 *
 *     Note that the returned object is owned by the CDM. The caller is not
 *     supposed to free or modify the returned object.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:  A Document object. (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetByteSequence (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a Text object (of type CDM_ByteSequence) that
 *     will contain the text (or byte sequence or raw data) of the specified
 *     Document. The returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetByteSequence, CDM_UtfToExternal, tip_GetByteSequence
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequence CDM_GetRawData(CDM_Document Document) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetRawData: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetRawData: Invalid Document Type");
    return NULL;
  }
  return Document->RawData;
} /* CDM_GetRawData */

/*
 * ---- CDM_GetRawData:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Text of the specified Document. It will
 *     return a C pointer (of type char*) that will contain the text
 *     of the Document. The text will be stored in a newly allocated memory
 *     space, using the requested by the value of the "Encoding" parameter
 *     encoding. The pointer to that new memory segment will be returned to
 *     the caller.
 *
 *     Note that the returned object is not owned by the CDM. The caller is
 *     responsible to use CDM_Free (or Tcl_Free) in order to free the string and
 *     the memory associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:  A Document object. (CDM_Document)
 *      Encoding:  The desired encoding that the text of the Document is to be converted to. (const char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function does not have an equivelant function (Tcl API). The Tcl
 *     core uses UTF-8 as its native encoding, so the convertion to a specific
 *     encoding is not needed.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a C pointer (of type char*) that
 *     will contain the text (or byte sequence or raw data) of the specified
 *     Document encoded in the requested encoding.  CDM does not own the
 *     returned memory segment: the caller is responsible to use CDM_Free (or
 *     Tcl_Free) in order to free the string and the memory associated with it.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, CDM_GetByteSequence, Tcl_Free
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetRawData(CDM_Document Document, const char *Encoding) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetRawData: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetRawData: Invalid Document Type");
    return NULL;
  }
  return CDM_GetByteSequence(Document, Encoding);
} /* CDM_GetRawData */

/*
 * ---- CDM_GetRawData:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Text of the specified Document. It will
 *     return a C pointer (of type char*) that will contain the text
 *     of the Document. The text will be stored in a newly allocated memory
 *     space, using the requested by the value of the "Encoding" parameter
 *     encoding. The pointer to that new memory segment will be returned to
 *     the caller.
 *
 *     Note that the returned object is not owned by the CDM. The caller is
 *     responsible to use CDM_Free (or Tcl_Free) in order to free the string and
 *     the memory associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:  A Document object. (CDM_Document)
 *      Encoding:  The desired encoding that the text of the Document is to be converted to. (Tcl_Encoding)
 *   >
 *  NOTES:
 *   <English:
 *     This function does not have an equivelant function (Tcl API). The Tcl
 *     core uses UTF-8 as its native encoding, so the convertion to a specific
 *     encoding is not needed.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a C pointer (of type char*) that
 *     will contain the text (or byte sequence or raw data) of the specified
 *     Document encoded in the requested encoding.  CDM does not own the
 *     returned memory segment: the caller is responsible to use CDM_Free (or
 *     Tcl_Free) in order to free the string and the memory associated with it.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, CDM_GetByteSequence, Tcl_GetEncoding, Tcl_Free
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetRawData(CDM_Document Document, const Tcl_Encoding encoding) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetRawData: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_GetRawData: Invalid Document Type");
    return NULL;
  }
  return CDM_GetByteSequence(Document, encoding);
} /* CDM_GetRawData */

/*
 * ---- CDM_NextAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that their spans ranges are after
 *     the specified text position (as given by the parameter "Position").
 *     CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *
 *     This function will always create and return a new Annotation set object,
 *     even when the Document has no Annotations defined or no Annotations
 *     exist that all their spans are beneath the specified text position.
 *     In this case an empty Annotation set will be returned.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Position:    Position in text.             (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_NextAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_NextAnnotations
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_NextAnnotations(CDM_Document Document, long Position) {
  CDM_AnnotationSet Anns;
  CDM_SpanSet       Spans;
  Tcl_Obj **Ann, **Span;
  register long start, min = -1;
  register int i,j;
  int items, spans;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_NextAnnotations: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_NextAnnotations: Invalid Document Type");
    return NULL;
  }

  Anns = CDM_GetAnnotations(Document);
  if (Anns == NULL) return NULL;

  i = Tcl_ListObjGetElements(CDM_Interp, Anns, &items, &Ann);
  if (i != TCL_OK) {
    CDM_ReportError("Cannot Get Document's Annotations");
    return NULL;
  }
  /* If there are no Annotations, return an empty (list) object... */
  if (items < 1) {
    Tcl_DecrRefCount(Anns);
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }

  for (i=0; i<items; i++) {
    Spans = CDM_GetSpans(Ann[i]);
    if (Spans == NULL) {
      CDM_ReportError("CDM_NextAnnotations: Document has Annotation without "
                      "Span?\n");
      CDM_AppendError(Tcl_GetString(Ann[i]));
      Tcl_DecrRefCount(Anns);
      return NULL;
    }
    /* Check All Spans... */
    j = Tcl_ListObjGetElements(CDM_Interp, Spans, &spans, &Span);
    if (j != TCL_OK) {
      CDM_ReportError("CDM_NextAnnotations: Error Getting Spans From "
                      "Annotation");
      CDM_AppendError(Tcl_GetString(Ann[i]));
      Tcl_DecrRefCount(Anns);
      return NULL;
    }
    for (j=0; j<spans; j++) {
      start = CDM_GetStart(Span[j]);
      if (start >= Position) {
        if (start < min || min == -1) min = start;
      }
    }
  }
  Tcl_DecrRefCount(Anns);
  if (min == -1) {
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }
  return CDM_AnnotationsAt(Document, min);
} /* CDM_NextAnnotations */

/*
 * ---- CDM_PutAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function will add a given Attribute to the specified Document
 *     object. If an Attribute with the same name already exists in the
 *     Document, then it will be overwritten by the new Attribute. Else, the
 *     new Attribute will be appended to the existing Attribute set of the
 *     specified Document.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:    A Document object.            (CDM_Document)
 *      Attribute:   The Attribute to be added. (CDM_Atrribute)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_PutAttribute (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *     error occurs, the return value will be TCL_ERROR and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp). Else, TCL_OK will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_PutAttribute
 *   >:&:
 */
CDM_EXPORT
int CDM_PutAttribute(CDM_Document Document, CDM_Attribute Attr) {
  Tcl_HashEntry *entryPtr;
  Tcl_Obj **elemPtrs, *Obj;
  int result, items, incr_ref_count = 1;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    return CDM_ReportError("CDM_PutAttribute: Aborting...");
  }

  if (Document == NULL) {
    return CDM_ReportError("CDM_PutAttribute: NULL Document Type");
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    return CDM_ReportError("CDM_PutAttribute: Invalid Document Type");
  }
  if (!CDM_ValidAttribute(Attr)) return TCL_ERROR;
  result = Tcl_ListObjGetElements(CDM_Interp, Attr, &items, &elemPtrs);
#ifdef CDM_OBJ_HASH_KEYS
  entryPtr = Tcl_CreateHashEntry(&(Document->Attributes_HashTable),
                                 (char *) elemPtrs[0], &items);
#else /* CDM_OBJ_HASH_KEYS */
  entryPtr = Tcl_CreateHashEntry(&(Document->Attributes_HashTable),
                                 Tcl_GetString(elemPtrs[0]), &items);
#endif /* CDM_OBJ_HASH_KEYS */
  if (entryPtr == NULL) return TCL_ERROR;
  if (items == 0) {
    Obj = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
    if (Obj != Attr) {
      Tcl_DecrRefCount(Obj);
    } else {
      incr_ref_count = 0;
    }
  }
  Tcl_SetHashValue(entryPtr, (ClientData) Attr);
  if (incr_ref_count) Tcl_IncrRefCount(Attr);
  Document->Modified = 1;
  ((CDM_Collection) Document->ParentPtr)->ModifiedDocuments = 1;
  return TCL_OK;
} /* CDM_PutAttribute */

/*
 * ---- CDM_RawDataInsertString:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the text contained in the specified Document,
 *     by inserting a string at the positions specified by the "pos"
 *     parameter. This position "pos" is the number of characters that have to
 *     be skipped before the string is inserted, where the first character of
 *     the Document text has a position equal to 0. If position is lower than
 *     zero, it is treaded as the [[i]]last[[/i]] character of the Document
 *     text.
 *
 *     For example, if the Document contains as text the
 *     string "This is a test", the following code:[[br]][[code]][[center]]
 *            CDM_RawDataInsertString(Doc, 10, "simple ");
 *     [[/center]][[/code]][[br]] will modify the text of the Document to
 *     "This is a simple test".
 *
 *     The most important aspect of this function is that it will make the
 *     proper arrangements so as the Annotations contained inside the Document
 *     to be displaced by a proper amound of characters so as Annotations
 *     spanning text ranges after the insertion point to span the same text
 *     portions, after the string has been inserted.
 *
 *     This function will return the new Document text (as an object of type
 *     CDM_ByteSequence) if the insertion was succesful.
 *     Note that the returned object is owned by the CDM. The caller is not
 *     supposed to free or modify the returned object.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     pos:         The character offset (starting from 0) where the string will be inserted. The string will be inserted at the character pointed by this value, displacing any existing text at this position after the inserted text. (long)
 *     string:      The string to be inserted.       (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_RawDataInsertString (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the new Document text (as an object of type
 *     CDM_ByteSequence) if the insertion was succesful.
 *     The returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_ByteSequenceInsertString, CDM_RawDataReplaceCharacters,
 *     CDM_DisplaceAnnotations,
 *     CDM_GetRawData, CDM_SetRawData,
 *     tip_RawDataInsertString
 *   >:&:
 */
CDM_EXPORT
CDM_RawData CDM_RawDataInsertString(CDM_Document Document,
                                          long pos, char *string) {
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_RawDataInsertString: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_RawDataInsertString: Invalid Document Type");
    return NULL;
  }

  return CDM_ByteSequenceInsertString(Document, pos, string);
} /* CDM_RawDataInsertString */

/*
 * ---- CDM_RawDataReplace:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the text contained in the specified Document,
 *     by removing a range of consecutive characters from the Document Text,
 *     starting with the character whose index is the value of the parameter
 *     "first" and ending with the character whose index is the value of the
 *     parameter "last". An index of 0 refers to the first character of the
 *     Document Text, while a negative value (i.e. -1) refers to the last
 *     character of the Document Text. If the value of the parameter "newstring"
 *     is not NULL, then it is placed in the removed character range.
 *     If "first" is less than zero then it is treated as if it was zero, and
 *     if "last" is greater than the length of the Document Text
 *     then it is treated as if it was equal to the length of the Document Text.
 *     It is an error if "first" is greater than "last".
 *
 *     The most important aspect of this function is that it will make the
 *     proper arrangements so as the Annotations contained inside the Document
 *     to be displaced by a proper amound of characters so as Annotations
 *     spanning text ranges after the replacement point to span the same text
 *     portions, after the string has replaced the desired characters.
 *
 *     This function will return the new Document text (as an object of type
 *     CDM_RawData) if the replacement was succesful.
 *     Note that the returned object is owned by the CDM. The caller is not
 *     supposed to free or modify the returned object.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     first:       The character offset (starting from 0) of the first character that will be replaced. (long)
 *     last:        The character offset (starting from 0) of the last character that will be replaced. (long)
 *     newstring:   The string to be placed in the removed character range. (char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function can be also used in order to [[b]]delete[[/b]]
 *    characters from the text contained in the specified Document, if the
 *    value of the "newstring" parameter is NULL or an empty string (like "").
 *    In this case, all Annotations objects contained in the specified
 *    Document are searched: if any of them has a Span object in its Span
 *    set whoose offsets span portion of the charactes to be deleted,
 *    then this Span is eliminated from the Span set of the Annotation
 *    object. Note that no Annotation objects are deleted from the Document,
 *    although character deletion can result in Annotations with an empty
 *    Span set.
 *
 *    This function is equivelant to tip_RawDataReplace (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the new Document text (as an object of type
 *     CDM_RawData) if the replacement was succesful.
 *     The returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_ByteSequenceReplace,
 *     CDM_RawDataReplaceCharacters, CDM_ByteSequenceReplaceCharacters,
 *     CDM_RawDataInsertString, CDM_ByteSequenceInsertString,
 *     CDM_DisplaceAnnotations,
 *     CDM_GetRawData, CDM_SetRawData,
 *     CDM_GetByteSequence, CDM_SetByteSequence,
 *     tip_RawDataReplace
 *   >:&:
 */
CDM_EXPORT
CDM_RawData CDM_RawDataReplace(CDM_Document Document,
                                       long first, long last, char *newstring) {
  CDM_RawData Text;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_RawDataReplace: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_RawDataReplace: Invalid Document Type");
    return NULL;
  }

  Text = CDM_GetRawData(Document);
  if (Text == NULL) return NULL;
  if (first < 0) first = 0;
  if (last < 0 || last > Tcl_GetCharLength(Text)) {
    last = Tcl_GetCharLength(Text);
  }
  if (last < first) {
    CDM_ReportError("CDM_RawDataReplace: Parameter \"last\" cannot be "
                    "less than parameter \"first\"");
    return NULL;
  }

  return CDM_RawDataReplaceCharacters(Document, first, last-first+1, newstring);
} /* CDM_RawDataReplace */

/*
 * ---- CDM_RawDataReplaceCharacters:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the text contained in the specified Document,
 *     by replacing the specified number of characters at the position
 *     specified by the "pos" parameter by the provided string.
 *     This position "pos" is the number of characters that have to
 *     be skipped before the characters are replaced by the string,
 *     where the first character of
 *     the Document text has a position equal to 0. If position is lower than
 *     zero, it is treaded as the [[i]]last[[/i]] character of the Document
 *     text. In the case the last character is specified as position, then
 *     the last "characters" characters are replaced.
 *
 *     For example, if the Document contains as text the
 *     string "This is a test", the following code:[[br]][[code]][[center]]
 *            CDM_RawDataReplaceCharacters(Doc, 10, 4, "simple test");
 *     [[/center]][[/code]][[br]] will modify the text of the Document to
 *     "This is a simple test".
 *
 *     The most important aspect of this function is that it will make the
 *     proper arrangements so as the Annotations contained inside the Document
 *     to be displaced by a proper amound of characters so as Annotations
 *     spanning text ranges after the replacement point to span the same text
 *     portions, after the string has replaced the desired characters.
 *
 *     This function will return the new Document text (as an object of type
 *     CDM_RawData) if the replacement was succesful.
 *     Note that the returned object is owned by the CDM. The caller is not
 *     supposed to free or modify the returned object.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     pos:         The character offset (starting from 0) where the replacement will happen. The string will be inserted at the character pointed by this value, replacing the specified number of characters. (long)
 *     characters:  An integer specifying the number of characters that will be replaced. (int)
 *     string:      The string to be inserted.       (char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function can be also used in order to [[b]]delete[[/b]]
 *    characters from the text contained in the specified Document, if the
 *    value of the "string" parameter is NULL or an empty string (like "").
 *    In this case, all Annotations objects contained in the specified
 *    Document are searched: if any of them has a Span object in its Span
 *    set whoose offsets span portion of the charactes to be deleted,
 *    then this Span is eliminated from the Span set of the Annotation
 *    object. Note that no Annotation objects are deleted from the Document,
 *    although character deletion can result in Annotations with an empty
 *    Span set.
 *
 *    This function is equivelant to tip_RawDataReplaceCharacters (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the new Document text (as an object of type
 *     CDM_RawData) if the replacement was succesful.
 *     The returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_ByteSequenceReplaceCharacters, CDM_RawDataInsertString,
 *     CDM_DisplaceAnnotations,
 *     CDM_GetRawData, CDM_SetRawData,
 *     tip_RawDataReplaceCharacters
 *   >:&:
 */
CDM_EXPORT
CDM_RawData CDM_RawDataReplaceCharacters(CDM_Document Document,
                                       long pos, int characters, char *string) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_RawDataReplaceCharacters: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_RawDataReplaceCharacters: Invalid Document Type");
    return NULL;
  }

  if (characters < 0) {
    CDM_ReportError("CDM_RawDataReplaceCharacters: the number of "
                    "characters to be replaced must be greater or equal to 0!");
    return NULL;
  }

  return CDM_ByteSequenceReplaceCharacters(Document, pos, characters, string);
} /* CDM_RawDataReplaceCharacters */

/*
 * ---- CDM_RemoveAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will remove the Annotation object that has as Id the value
 *     specified by the "Id" parameter. If the requested Annotation does not
 *     exist, an error will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:  A Document object.                         (CDM_Document)
 *      Id:        The Id of the Annotation to be deleted.    (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_RemoveAnnotation (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Annotation exists and is successfully removed, then TCL_OK
 *     will be returned. If the requested Annotation does not exist, or in
 *     case of an error, the return value will be
 *     TCL_ERROR and an error message describing the error will be left at the
 *     current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_RemoveAnnotation
 *   >:&:
 */
CDM_EXPORT
int CDM_RemoveAnnotation(CDM_Document Document, long Id) {
#ifndef DONT_HASH_DOC_ANNOTATIONS
  Tcl_HashEntry *entryPtr;
  Tcl_Obj *Obj;
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    return CDM_ReportError("CDM_RemoveAnnotation: Aborting...");
  }

  if (Document == NULL) {
    return CDM_ReportError("CDM_RemoveAnnotation: NULL Document Type");
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    return CDM_ReportError("CDM_RemoveAnnotation: Invalid Document Type");
  }

#ifdef DONT_HASH_DOC_ANNOTATIONS
  (Document->Annotations_List)->Delete(Id);
#else  /* DONT_HASH_DOC_ANNOTATIONS */
#ifdef CDM_ONE_WORD_KEYS
  entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable), (char *) Id);
#else
  snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", Id);
  entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable), CDM_Temp);
#endif /* CDM_ONE_WORD_KEYS */
  if (entryPtr == NULL) {
    return CDM_ReportError("CDM_RemoveAnnotation: Annotation with "
                           "Id=%ld not found!", Id);
  }
  Obj = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
  if (Obj != NULL) Tcl_DecrRefCount(Obj);
  Tcl_DeleteHashEntry(entryPtr);
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  Document->Modified = 1;
  ((CDM_Collection) Document->ParentPtr)->ModifiedDocuments = 1;
  return TCL_OK;
} /* CDM_RemoveAnnotation */

/*
 * ---- CDM_RemoveAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function will remove the Attribute named exactly as the "Name"
 *     parameter. If the requested Attribute does not exist, an error will be
 *     returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:  A Document object.           (CDM_Document)
 *     Name:      The Name of the Attribute to be removed. (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_RemoveAttribute (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Attribute exists and is successfully removed, then TCL_OK
 *     will be returned. In case of an error, the return value will be
 *     TCL_ERROR and an error message describing the error will be left at the
 *     current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_RemoveAttribute
 *   >:&:
 */
CDM_EXPORT
int CDM_RemoveAttribute(CDM_Document Document, char *Name) {
  Tcl_HashEntry *entryPtr;
  Tcl_Obj *Obj;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    return CDM_ReportError("CDM_RemoveAttribute: Aborting...");
  }

  if (Document == NULL) {
    return CDM_ReportError("CDM_RemoveAttribute: NULL Document Type");
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    return CDM_ReportError("CDM_RemoveAttribute: Invalid Document Type");
  }

#ifdef CDM_OBJ_HASH_KEYS
  /*
   * We don't add a reference to this object. If it is an old one, it's
   * refCount is left intact. If it is a new one, it will have a refCount of
   * 1, as it is referenced by the CDM hash table. It will be freed when the
   * table is freed...
   */
  Tcl_Obj *NameObj;
  NameObj = CDM_NewStringObj(Name);
  entryPtr = Tcl_FindHashEntry(&(Document->Attributes_HashTable),
                                (char *) NameObj);
#else /* CDM_OBJ_HASH_KEYS */
  entryPtr = Tcl_FindHashEntry(&(Document->Attributes_HashTable), Name);
#endif /* CDM_OBJ_HASH_KEYS */
  if (entryPtr == NULL) {
    return CDM_ReportError("Undefined Attribute Name: \"%s\"", Name);
  }
  Obj = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
  if (Obj != NULL) Tcl_DecrRefCount(Obj);
  Tcl_DeleteHashEntry(entryPtr);
  Document->Modified = 1;
  ((CDM_Collection) Document->ParentPtr)->ModifiedDocuments = 1;
  return TCL_OK;
} /* CDM_RemoveAttribute */

/*
 * ---- CDM_SelectAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that their type is the same as the value of the
 *     "Type" parameter. This function will always create and return a new
 *     object, even when no Annotations of the requested type are found or
 *     when the Document has no Annotations defined. In this case an empty
 *     object will be returned.
 *
 *     Note that CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Type:        The type of the Annotations to be selected.  (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SelectAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return an AnnotationSet object, that will
 *     contain all the Annotations of the specified Document that have as type
 *     the value of the "Type" parameter.
 *
 *     Note that the caller is responsible to free this object through
 *     Tcl_DecrRefCount or CDM_Free.
 *
 *     In case of an error, NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_SelectAnnotations
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_SelectAnnotations(CDM_Document Document,
                                        const char *Type) {
#ifndef DONT_HASH_DOC_ANNOTATIONS
  Tcl_HashEntry *entryPtr;
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  Tcl_Obj *result;
  CDM_AnnotationSet Set;
  long i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_SelectAnnotations: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_SelectAnnotations: NULL Document Type");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_SelectAnnotations: Invalid Document Type");
    return NULL;
  }

  Set = Tcl_NewListObj(0, NULL);
  for (i=0; i <= Document->AnnotationMaxID; i++) {
#ifdef DONT_HASH_DOC_ANNOTATIONS
    result = (Document->Annotations_List)->Ann(i);
#else  /* DONT_HASH_DOC_ANNOTATIONS */
#ifdef CDM_ONE_WORD_KEYS
    entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),(char *) i);
#else
    snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", i);
    entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable), CDM_Temp);
#endif /* CDM_ONE_WORD_KEYS */
    if (entryPtr == NULL) continue;
    result = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
#endif /* DONT_HASH_DOC_ANNOTATIONS */
    if (result != NULL && strcmp(CDM_GetType(result), Type) == 0) {
      Tcl_ListObjAppendElement(Document->Interp, Set, result);
    }
  }
  return Set;
} /* CDM_SelectAnnotations */

/*
 * ---- CDM_SelectAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that their type is the same as the value of the
 *     "Type" parameter and the provided boolean expression (specified through
 *     the "Constraints" parameter) is true for every one of these Annotations.
 *     This function will always create and return a new
 *     object, even when no Annotations of the requested type are found or
 *     when the Document has no Annotations defined. In this case an empty
 *     object will be returned.
 *
 *     The boolean expression that can be specified through the "Constraints"
 *     parameter can be any valid Tcl boolean expression (i.e. a boolean
 *     expression that will be accepted by the "expr" Tcl command), with the
 *     following excepion:
 *
 *     [[LI]]The boolean expression can contain references to the values of
 *     Atttributes of the Annotation, by utilising the notation
 *     "ann::%%%lt;Attribute Name%%%gt;". The only limitation is that
 *     "%%%lt;Attribute Name%%%gt;" cannot exceed 120 characters. For example,
 *     the following code will create a new Annotation set that will contain
 *     all Annotations from the specified Document that their type is token,
 *     they have a "type" Attribute which has as value "EFW" and also have a
 *     "pos" Attribute that has a value other than "NN":[[/LI]]
 *     [[br]][[code]][[center]]
 *     CDM_SelectAnnotations(Doc, "token", "ann::type == \"EFW\" %%%amp;%%%amp; ann::pos != \"NN\"");
 *     [[/center]][[/code]][[br]]
 *
 *     Note that CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                           (CDM_Document)
 *     Type:        The type of the Annotations to be selected.  (char *)
 *     Constraints: A boolean expression.                        (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SelectAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return an AnnotationSet object, that will
 *     contain all the Annotations of the specified Document that have as type
 *     the value of the "Type" parameter and the provided by the "Constraints"
 *     parameter boolean expression was true for these Annotations.
 *
 *     Note that the caller is responsible to free this object through
 *     Tcl_DecrRefCount or CDM_Free.
 *
 *     In case of an error, NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_SelectAnnotations,
 *     [[A HREF="../TclTk/TclCmd/expr.htm"]]The Tcl "expr" manual[[/A]].
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_SelectAnnotations(CDM_Document Document, const char *Type,
                                        const char *Constraints) {
#ifndef DONT_HASH_DOC_ANNOTATIONS
  Tcl_HashEntry *entryPtr;
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  Tcl_Obj *result, *AttributeNames, *Expr, *Boolean, **Name;
  Tcl_Parse parsePtr;
  Tcl_Token *token;
  CDM_AnnotationSet Set;
  const char *ptr, *qptr, *start;
  char name[128], TclSafeName[256];
  int BoolVal, status, NameNu, j, RejectedAnnPrefix;
  long i;
  CDM_THREAD_SPECIFIC_DATA;

  /* No need for complicated things if no expression... */
  if (Constraints == NULL || strlen(Constraints) == 0 ||
      strcmp(Constraints, "{}") == 0) {
    return CDM_SelectAnnotations(Document, Type);
  }

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_SelectAnnotations: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_SelectAnnotations: NULL Document Type");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_SelectAnnotations: Invalid Document Type");
    return NULL;
  }

  /*
   * Since we want to offer Attribute tokens for the Attribute Annotations, we
   * have to preprocess the boolean expresion. This is done in the following
   * steps:
   * 1) First, make sure the array we use for storing temporary values is
   *    clean. Ignore any errors...
   */
  Tcl_UnsetVar2(CDM_Interp, "::CDM::CDM_SelectAnnotationsArray", NULL, 0);
  /*
   * 2) Foreach Annotation Attribute name (denoted with ann::<name>), create
   * an entry in the array, and keep its name for further reference...
   */
  start   = Constraints;
  ptr     = strstr(Constraints, "ann::");
  if (ptr != NULL) {
    /* Try to find the previous non-white character... */
    for (qptr=ptr-1; (*qptr==' ' || *qptr=='\t') && qptr>=Constraints; qptr--);
    if (*qptr == '\"' || *qptr == '{') ptr = qptr;
  }
  AttributeNames        = Tcl_NewListObj(0, NULL);
  Expr                  = Tcl_NewStringObj(NULL, -1);
  // Tcl_IncrRefCount(Expr);
  // printf("Original: ->%s<-\n", Constraints);
  while (ptr != NULL) {
    /* Copy unchanged bytes... */
    Tcl_AppendToObj(Expr, start, ptr - start);
    RejectedAnnPrefix = 0;
    /* Try to get the ann::<attribute name> substring. This is not an easy
     * task, as it may be enclosed in "" or {}. Use Tcl suitable routines to
     * parse such cases...
     */
    if (*ptr == '{' || *ptr == '\"' || *(ptr+5) == '{' || *(ptr+5) == '\"') {
      if (*ptr == '{') {
        status = Tcl_ParseBraces(CDM_Interp, ptr, -1, &parsePtr, 0,
                                 (const char **) &qptr);
      } else if (*ptr == '\"') {
        status = Tcl_ParseQuotedString(CDM_Interp, ptr, -1, &parsePtr, 0,
                                       (const char **) &qptr);
      } else if (*(ptr+5) == '{') {
        status = Tcl_ParseBraces(CDM_Interp, ptr+5, -1, &parsePtr, 0,
                                 (const char **) &qptr);
        RejectedAnnPrefix = 1;
      } else {
        status = Tcl_ParseQuotedString(CDM_Interp, ptr+5, -1, &parsePtr, 0,
                                       (const char **) &qptr);
        RejectedAnnPrefix = 1;
      }
      if (status != TCL_OK) {
        CDM_AppendError("\nCDM_SelectAnnotations: while evaluating expression "
                        "\n  (%s)", Constraints);
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        return NULL;
      }
      if (parsePtr.numTokens != 1) {
        CDM_ReportError("CDM_SelectAnnotations: invalid ann::<attribute name> "
                        "reference in constraints expression:\n  (%s)", ptr);
        Tcl_FreeParse(&parsePtr);
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        return NULL;
      }
      token = parsePtr.tokenPtr;
      if (token->type != TCL_TOKEN_TEXT) {
        CDM_ReportError("CDM_SelectAnnotations: illegal ann::<attribute name> "
                        "reference in constraints expression:\n  (%s)", ptr);
        Tcl_FreeParse(&parsePtr);
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        return NULL;
      }
      if (token->size > 127) {
        CDM_ReportError("CDM_SelectAnnotations: too long ann::<attribute name> "
                        "reference. <attribute name> must be < 128 characters");
        Tcl_FreeParse(&parsePtr);
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        return NULL;
      }
      if (RejectedAnnPrefix) {
        strncpy(name, token->start, token->size);
        name[token->size] = '\0';
      } else {
        strncpy(name, token->start+5, token->size-5);
        name[token->size-5] = '\0';
      }
      Tcl_FreeParse(&parsePtr);
      start = qptr;
    } else {
      ptr += 5; status = 0;
      while (!CHAR_TYPE(*ptr) && status < 128) {
        name[status++] = *ptr; ++ptr;
      }
      name[status] = '\0';
      start = ptr;
    }
    /*
     * Given the ann::<attribute name> reference, try to construct a safe for
     * Tcl name that can be used in the expression so as to correctly refer to
     * the proper array element...
     */
    for (qptr = name, status = 0; *qptr != '\0'; qptr++) {
      if (*qptr == ' ' || *qptr == '\t' || *qptr == ';' || *qptr == '\"' ||
          *qptr == '{' || *qptr == '}' || *qptr == '[' || *qptr == ']' ||
          *qptr == '\\') {
        TclSafeName[status++] = '\\';
        TclSafeName[status++] = *qptr;
      } else {
        TclSafeName[status++] = *qptr;
      }
    }
    TclSafeName[status] = *qptr;

    /* Is the variable already defined? */
    if (Tcl_GetVar2Ex(CDM_Interp, "::CDM::CDM_SelectAnnotationsArray",
        name, 0) == NULL) {
      /* The first time we meet this name... */
      Tcl_ListObjAppendElement(Document->Interp, AttributeNames,
                               CDM_NewStringObj(name));
      Tcl_SetVar2Ex(Document->Interp, "::CDM::CDM_SelectAnnotationsArray",
                               name, &CDM_Obj_NULL, 0);
    }
    /* Substitute ann::<name> with the proper array element... */
    Tcl_AppendStringsToObj(Expr, "$::CDM::CDM_SelectAnnotationsArray(",
                                 TclSafeName, ")", (char *) NULL);
    ptr   = strstr(start, "ann::");
    if (ptr != NULL) {
      /* Try to find the previous non-white character... */
      for (qptr=ptr-1; (*qptr==' ' || *qptr=='\t')&& qptr>=Constraints; qptr--);
      if (*qptr == '\"' || *qptr == '{') ptr = qptr;
    }
  }
  /* Finally, add the rest of the expression... */
  Tcl_AppendToObj(Expr, start, -1);
  // printf("Compiled: ->%s<-\n", Tcl_GetString(Expr));

  /* Create the new Annotation set... */
  Set = Tcl_NewListObj(0, NULL);
  /* Place the various Attribute name objects in an array... */
  Tcl_ListObjGetElements(NULL, AttributeNames, &NameNu, &Name);
  for (i=0; i <= Document->AnnotationMaxID; i++) {
#ifdef DONT_HASH_DOC_ANNOTATIONS
    result = (Document->Annotations_List)->Ann(i);
#else  /* DONT_HASH_DOC_ANNOTATIONS */
#ifdef CDM_ONE_WORD_KEYS
    entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable),(char *) i);
#else
    snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", i);
    entryPtr = Tcl_FindHashEntry(&(Document->Annotations_HashTable), CDM_Temp);
#endif /* CDM_ONE_WORD_KEYS */
    if (entryPtr == NULL) continue;
    result = (Tcl_Obj *) Tcl_GetHashValue(entryPtr);
#endif /* DONT_HASH_DOC_ANNOTATIONS */
    if (result != NULL && strcmp(CDM_GetType(result), Type) == 0) {
      /*
       * The Annotation has the requested type. It's time to apply the boolean
       * expession: If evaluates to true, we append the Annotation to the
       * Annotation set...
       */

      /*
       * Now, we have to fill the array with the correct values. In case an
       * attribute does not exist, we place an empty value...
       */
      for (j=0; j<NameNu; j++) {
        Boolean = CDM_GetAttribute(result, Tcl_GetString(Name[j]));
        if (Boolean == NULL) {
          /* Set the empty value... */
          Tcl_ObjSetVar2(Document->Interp,
                         CDM_NewStringObj("::CDM::CDM_SelectAnnotationsArray"),
                         Name[j], &CDM_Obj_NULL, 0);
        } else {
          Boolean = CDM_GetValueValue(Boolean);
          Tcl_ObjSetVar2(Document->Interp,
                         CDM_NewStringObj("::CDM::CDM_SelectAnnotationsArray"),
                         Name[j], Boolean, 0);
        }
      }
      if (Tcl_ExprObj(Document->Interp, Expr, &Boolean) != TCL_OK) {
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        CDM_Free(Set);
        CDM_AppendError("\nCDM_SelectAnnotations: invalid boolean expression\n"
                        "  (type=\"%s\", constraints=\"%s\")",
                        Type, Constraints);
        return NULL;
      }
      /* Was it true? */
      status = Tcl_GetBooleanFromObj(Document->Interp, Boolean, &BoolVal);
      CDM_Free(Boolean);
      if (status != TCL_OK) {
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        CDM_Free(Set);
        CDM_AppendError("CDM_SelectAnnotations: invalid boolean expression "
                        "result\n  (type=\"%s\", constraints=\"%s\")",
                        Type, Constraints);
        return NULL;
      }
      if (BoolVal) {
        /* We succeded! Expression was true... */
        Tcl_ListObjAppendElement(Document->Interp, Set, result);
      }
    }
  }
  Tcl_UnsetVar2(CDM_Interp, "::CDM::CDM_SelectAnnotationsArray", NULL, 0);
  CDM_Free(AttributeNames);
  CDM_Free(Expr);
  return Set;
} /* CDM_SelectAnnotations */

/*
 * ---- CDM_SelectAnnotationsSorted:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that their type is the same as the value of the
 *     "Type" parameter. The Annotations will be sorted according to their
 *     first span range, in ascending order.
 *
 *     This function will always create and return a new
 *     object, even when no Annotations of the requested type are found or
 *     when the Document has no Annotations defined. In this case an empty
 *     object will be returned.
 *
 *     Note that CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Type:        The type of the Annotations to be selected.  (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SelectAnnotationsSorted (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return an AnnotationSet object, that will
 *     contain all the Annotations of the specified Document that have as type
 *     the value of the "Type" parameter. The Annotations in that object will
 *     be sorted according to their first span in ascending order.
 *
 *     Note that the caller is responsible to free this object through
 *     Tcl_DecrRefCount or CDM_Free.
 *
 *     In case of an error, NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_SelectAnnotations
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_SelectAnnotationsSorted(CDM_Document Document,
                                              const char *Type){
  CDM_AnnotationSet Annotations;
  Tcl_Obj *objv[4];
  int status, i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_SelectAnnotationsSorted: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_SelectAnnotationsSorted: Invalid Document Type");
    return NULL;
  }
  Annotations = CDM_SelectAnnotations(Document, Type);
  /* Execute Tcl Command "lsort -command tip_CompareAnnotations $Annotations" */
  objv[0] = Tcl_NewStringObj("lsort", 5);
  objv[1] = Tcl_NewStringObj("-command", 8);
  objv[2] = Tcl_NewStringObj("tip_CompareAnnotations", 22);
  objv[3] = Annotations;
  for (i=0; i<4; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(CDM_Interp, 4, objv, TCL_EVAL_GLOBAL);
  for (i=0; i<4; i++) Tcl_DecrRefCount(objv[i]);
  if (status != TCL_OK) return NULL;
  return Tcl_DuplicateObj(Tcl_GetObjResult(CDM_Interp));
} /* CDM_SelectAnnotationsSorted */

/*
 * ---- CDM_SelectAnnotationsSorted:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that their type is the same as the value of the
 *     "Type" parameter and the provided boolean expression (specified through
 *     the "Constrains" parameter) is true for every one of these Annotations.
 *     The Annotations will be sorted according to their
 *     first span range, in ascending order.
 *
 *     This function will always create and return a new
 *     object, even when no Annotations of the requested type are found or
 *     when the Document has no Annotations defined. In this case an empty
 *     object will be returned.
 *
 *     The boolean expression that can be specified through the "Constrains"
 *     parameter can be any valid Tcl boolean expression (i.e. a boolean
 *     expression that will be accepted by the "expr" Tcl command), with the
 *     following excepion:
 *
 *     [[LI]]The boolean expression can contain references to the values of
 *     Atttributes of the Annotation, by utilising the notation
 *     "ann::%%%lt;Attribute Name%%%gt;". The only limitation is that
 *     "%%%lt;Attribute Name%%%gt;" cannot exceed 120 characters. For example,
 *     the following code will create a new Annotation set that will contain
 *     all Annotations from the specified Document that their type is token,
 *     they have a "type" Attribute which has as value "EFW" and also have a
 *     "pos" Attribute that has a value other than "NN":[[/LI]]
 *     [[br]][[code]][[center]]
 *     CDM_SelectAnnotationsSorted(Doc, "token", "ann::type == \"EFW\" %%%amp;%%%amp; ann::pos != \"NN\"");
 *     [[/center]][[/code]][[br]]
 *
 *     Note that CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                           (CDM_Document)
 *     Type:        The type of the Annotations to be selected.  (char *)
 *     Constraints: A boolean expression.                        (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SelectAnnotationsSorted (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return an AnnotationSet object, that will
 *     contain all the Annotations of the specified Document that have as type
 *     the value of the "Type" parameter and the provided by the "Constrains"
 *     parameter boolean expression was true for these Annotations.
 *     The Annotations in that object will
 *     be sorted according to their first span in ascending order.
 *
 *     Note that the caller is responsible to free this object through
 *     Tcl_DecrRefCount or CDM_Free.
 *
 *     In case of an error, NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_SelectAnnotations, [[A HREF="../TclTk/TclCmd/expr.htm"]]The Tcl
 *     "expr" manual[[/A]].
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_SelectAnnotationsSorted(CDM_Document Document,
                                              const char *Type,
                                              const char *Constraints){
  CDM_AnnotationSet Annotations;
  Tcl_Obj *objv[4];
  int status, i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_SelectAnnotationsSorted: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_SelectAnnotationsSorted: Invalid Document Type");
    return NULL;
  }
  Annotations = CDM_SelectAnnotations(Document, Type, Constraints);
  /* Execute Tcl Command "lsort -command tip_CompareAnnotations $Annotations" */
  objv[0] = Tcl_NewStringObj("lsort", 5);
  objv[1] = Tcl_NewStringObj("-command", 8);
  objv[2] = Tcl_NewStringObj("tip_CompareAnnotations", 22);
  objv[3] = Annotations;
  for (i=0; i<4; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(CDM_Interp, 4, objv, TCL_EVAL_GLOBAL);
  for (i=0; i<4; i++) Tcl_DecrRefCount(objv[i]);
  if (status != TCL_OK) return NULL;
  return Tcl_DuplicateObj(Tcl_GetObjResult(CDM_Interp));
} /* CDM_SelectAnnotationsSorted */

/*
 * ---- CDM_SetByteSequence:
 *  DESCRIPTION:
 *   <English:
 *     This function will change the text of a Document object to the text
 *     specified by the "Text" parameter. The "Text" parameter must be an
 *     object of type CDM_ByteSequence. Such an object can be created with
 *     Tcl_NewStringObj. The text that this function accepts, is asumend to be
 *     encoding according to the UTF-8 encoding. If the text is stored using a
 *     different encoding, the caller must use CDM_ExternalToUtf in order to
 *     convert the text to UTF. CDM will internally keep a reference to this
 *     object.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Text:        The desired new text object.  (CDM_ByteSequence)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SetByteSequence (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *     error occurs, the return value will be TCL_ERROR and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_ExternalToUtf, CDM_SetRawData, Tcl_NewStringObj, CDM_ExternalToUtf,
 *     tip_SetByteSequence
 *   >:&:
 */
CDM_EXPORT
int CDM_SetByteSequence(CDM_Document Document, CDM_ByteSequence Text) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    return CDM_ReportError("CDM_SetByteSequence: Aborting...");
  }

  if (Document == NULL)
    return CDM_ReportError("CDM_SetByteSequence: NULL Document Pointer");
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_SetByteSequence: Invalid Document Type");
    return TCL_ERROR;
  }
  if (Document->RawData != NULL) Tcl_DecrRefCount(Document->RawData);
  Document->RawData = Text;
  Tcl_IncrRefCount(Document->RawData);
  return TCL_OK;
} /* CDM_SetByteSequence */

/*
 * ---- CDM_SetEncoding:
 *  DESCRIPTION:
 *   <English:
 *     This function will change the encoding of a Document object to the
 *     encoding specified by the "Encoding" parameter.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                        (CDM_Document)
 *     Encoding:    The desired new encoding of the Document. (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SetEncoding (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a pointer (of type char*) to the string
 *     buffer that holds the new encoding value. This pointer will be at a
 *     different memory location than the given parameter value, as the CDM
 *     will create and manipulate a local copy.
 *
 *     In case of an error, NULL will be returned and an error message
 *     describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *
 *     Note that the returned pointer is property of the CDM and should never
 *     be freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetEncoding, tip_SetEncoding
 *   >:&:
 */
CDM_EXPORT
char *CDM_SetEncoding(CDM_Document Document, char *Encoding) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_SetEncoding: Aborting...");
    return NULL;
  }

  if (Document == NULL) {
    CDM_ReportError("CDM_SetEncoding: NULL Document Pointer");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_SetEncoding: Invalid Document Type");
    return NULL;
  }
  if (Document->Encoding != NULL) Tcl_Free(Document->Encoding);
  if (*Encoding == '\0') Encoding = CDM_DefaultEncoding;
  Document->Encoding = CDM_Strdup(Encoding);
  Document->Modified = 1;
  ((CDM_Collection) Document->ParentPtr)->ModifiedDocuments = 1;
  return Document->Encoding;
} /* CDM_SetEncoding */

/*
 * ---- CDM_SetExternalId:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the external Id (XID) of the specified
 *     Document. The external Id is the (absolute) path of the initial file
 *     that contained the text which was used in order to create the specified
 *     Document. The value is expected to be encoded using the UTF-8 encoding
 *     (thus enabling the existance of non Latin characters in the value). The
 *     caller can use the function CDM_ExternalToUft in order to convert a
 *     string from any supported encoding to UTF.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                            (CDM_Document)
 *     ExternalId:  The desired new external Id of the Document.  (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SetExternalId (Tcl API).
 *     The Document Id is not changed, as a result of calling this function.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     If the external Id of the specified Document is successfully updated,
 *     TCL_OK will be returned. In case of an error, the return value will be
 *     TCL_ERROR and an error message describing the error will be left at the
 *     current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetExternalId, CDM_SetExternalId, tip_SetExternalId
 *   >:&:
 */
CDM_EXPORT
char *CDM_SetExternalId(CDM_Document Document, char *ExternalId) {
  CDM_ListNode *ptr;
  Tcl_DString ds;
  char *XID;
#if 0
  /* Petasis, 23 November 2006 */
  char **PathParts;
  int items;
#endif
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_SetExternalId: Aborting...");
    return NULL;
  }

  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_SetExternalId: Invalid Document Type");
    return NULL;
  }
  ptr = Document->DocumentPtr;
  if (ptr == NULL) {
    CDM_ReportError("CDM_SetExternalId: Invalid Document");
    return NULL;
  }
  ExternalId = (char *)Tcl_TranslateFileName(Document->Interp, ExternalId, &ds);
  if (ExternalId == NULL) {
    Tcl_DStringFree(&ds);
    CDM_ReportError("Invalid Document External Id: (NULL)");
    return NULL;
  }
  XID = CDM_Strdup(ExternalId);
  Tcl_DStringFree(&ds);

#if 0
  /* Petasis, 23 November 2006:
     Updating the unique ID with an id generated by the arguments given to
     CDM_SetExternalId is not desired.
   */
  /* Get tail of External Id */
  Tcl_SplitPath(XID, &items, (const char ***) &PathParts);
  if (items > 0) {
    if ((items > 1) || (Tcl_GetPathType(PathParts[0]) == TCL_PATH_RELATIVE)) {
      if (ptr->ptr[CDM_DOCUMENT_ID] != NULL)
         Tcl_Free(ptr->ptr[CDM_DOCUMENT_ID]);
      ptr->ptr[CDM_DOCUMENT_ID] = CDM_Strdup(PathParts[items - 1]);
    } else goto error;
  } else {
    goto error;
  }
#endif
  if (ptr->ptr[CDM_DOCUMENT_XID] != NULL) Tcl_Free(ptr->ptr[CDM_DOCUMENT_XID]);
  ptr->ptr[CDM_DOCUMENT_XID] = XID;
  Document->Modified = 1;
  ((CDM_Collection) Document->ParentPtr)->ModifiedDocuments = 1;
  return XID;
/*
error:
  return NULL;
 */
} /* CDM_SetExternalId */

/*
 * ---- CDM_SetRawData:
 *  DESCRIPTION:
 *   <English:
 *     This function will change the text of a Document object to the text
 *     specified by the "Text" parameter. The new text is asumend to be
 *     encoding according to the UTF-8 encoding. If the text is stored using a
 *     different encoding, the caller must use CDM_ExternalToUtf in order to
 *     convert the text to UTF. CDM will internally alocate a new memory area
 *     to store the given text.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Text:        The desired new text.         (CDM_RawData)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SetRawData (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *     error occurs, the return value will be TCL_ERROR and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_ExternalToUtf, CDM_SetByteSequence, tip_SetRawData
 *   >:&:
 */
CDM_EXPORT
int CDM_SetRawData(CDM_Document Document, CDM_RawData Text) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    return CDM_ReportError("CDM_SetRawData: Aborting...");
  }

  if (Document == NULL)
    return CDM_ReportError("CDM_SetRawData: NULL Document Pointer");
  if (Document->Type != CDM_TYPE_DOCUMENT)
    return CDM_ReportError("CDM_SetRawData: Invalid Document Type");
  if (Document->RawData != NULL) Tcl_DecrRefCount(Document->RawData);
  Document->RawData = Text;
  Tcl_IncrRefCount(Document->RawData);
  return TCL_OK;
} /* CDM_SetRawData */

/*
 * ---- CDM_Status:
 *  DESCRIPTION:
 *   <English:
 *     This function will return information about the specified Document.
 *     This information will be stored as text in a newly created Tcl object.
 *     CDM does not own the returned object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *
 *     This function will always create and return a new Tcl string object.
 *     In case of an error, NULL will be returned.
 *   >
 *   <Greek:
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document: A Document object.       (CDM_Document)
 *   >
 *   <Greek:
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a Tcl string object (of type Tcl_Obj *),
 *    containing the information of the specified Document.
 *    The caller is responsible to free this object through Tcl_DecrRefCount
 *    or CDM_Free.
 *   >
 *   <Greek:
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Free, Tcl_DecrRefCount
 *   >
 *   <Greek:
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequence CDM_Status(CDM_Document Document) {
  Tcl_Obj *Info, *Attrs, *Anns;
  const char *format;
  int i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_Status: Aborting...");
    return NULL;
  }

  if (Document == NULL) return NULL;
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_Status: Invalid Document Type");
    return NULL;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) return NULL;
  Info = Tcl_NewStringObj(NULL, 0);
  snprintf(CDM_Temp, MAX_TEMP_STRING, "%ld", Document->AnnotationMaxID);
  switch (Document->FileFormat) {
    case CDM_FORMAT_NATIVE:
    case CDM_FORMAT_NATIVE2: {format = "native"; break;}
    case CDM_FORMAT_GATE_1x: {format = "gate";   break;}
    case CDM_FORMAT_GATE_2x: {format = "gate2";  break;}
    default:                 {
      /*
       * A registered format?
       */
      format = "?";
      i = Document->FileFormat - CDM_FORMAT_REGISTERED;
      if (i >= 0 && i < CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER &&
          CDM_AdditionalCollectionTypes[i].CollectionType != NULL) {
        format = CDM_AdditionalCollectionTypes[i].CollectionType;
      }
    }
  }
  Attrs = CDM_GetAttributes(Document);
  Anns  = CDM_GetAnnotations(Document);
  Tcl_AppendStringsToObj(Info,
      "Id: ",                Document->DocumentPtr->ptr[CDM_DOCUMENT_ID],
      "\nXID: ",             Document->DocumentPtr->ptr[CDM_DOCUMENT_XID],
      "\nEncoding: ",        Document->Encoding,
      "\nTcl Command: ",
                   Tcl_GetCommandName(Document->Interp, Document->Tcl_Cmd),
      "\nFormat: ",          format,
      "\nModified: ",        (Document->Modified ? "(Yes)" : "(No)"),
      "\nText:\n  ",         Tcl_GetString(Document->RawData),
      "\nNumber of Annotations: ", CDM_Temp,
      "\nAttributes:\n  ",   (Attrs) ? Tcl_GetString(Attrs) : "",
      "\nAnnotations:\n  ",  (Anns)  ? Tcl_GetString(Anns) : "",
      (char *) NULL);
  if (Attrs) Tcl_DecrRefCount(Attrs);
  if (Anns)  Tcl_DecrRefCount(Anns);
  return Info;
} /* CDM_Status */

/*
 * ---- CDM_Sync:
 *  DESCRIPTION:
 *   <English:
 *     This function will save a Document object in disk. If the Document
 *     object has never been saved again, then this function will create a
 *     file according to the Id the specified Document. A representation
 *     of the Document will be saved in this file. If the file
 *     already exists it will be overwritten. If the file cannot be
 *     created an error condition will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     The user is not expected to modify file that hold Documents in
 *     any way. Modifing any Document file may
 *     result in a corrupted, unreadble Document.
 *
 *     This function is equivelant to tip_Sync (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Document is successfully saved, TCL_OK will
 *     be returned. In case of an error, the return value will be TCL_ERROR and
 *     an error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Sync
 *   >:&:
 */
CDM_EXPORT
int CDM_Sync(CDM_Document Document) {
  CDM_Collection Parent;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    return CDM_ReportError("CDM_Sync: Aborting...");
  }

  if (Document == NULL)
      return CDM_ReportError("CDM_Sync: NULL Document Pointer");
  if (Document->Type != CDM_TYPE_DOCUMENT)
    return CDM_ReportError("CDM_Sync: Invalid Document Type");

  Parent = CDM_GetParent(Document);
  if (Parent == NULL)
    return CDM_ReportError("CDM_Sync: Cannot Locate Parent Collection");
  return CDM_SaveDocument(Document, Parent->Name);
} /* CDM_Sync */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     E N D   O F   D O C U M E N T  S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     A T T R I B U T E              S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 * ---- CDM_AddAttribute:
 *  DESCRIPTION:
 *   <English:
 *    This function will append the given Attribute (the value of the
 *    "Attribute" parameter) to the specified Attribute set (by the value of
 *    the "AttributeSet" parameter). A reference to the Attribute will be
 *    added by the AttributeSet, so after this call the Attribute object may
 *    be converted to a shared object.
 *
 *    The Attribute set specified by the "AttributeSet" parameter will be
 *    modified only if it is not a shared object. If it is a shared object, a
 *    new (duplicate) object will be created and the new Attribute will be
 *    appended to this new object. If a new Attribute set is created, the
 *    initial one will be not modified in any way. In all cases, a reference
 *    to the Attribute set that additionally contains the new Attribute will
 *    be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     AttributeSet: Attribute set where the Attribute will be appended. (CDM_AttributeSet)
 *     Attribute:    Attribute to be appended         (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AddAttribute (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return an Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all Annotations (including the new
 *     one). This Annotation set may not be the original Annotation set that
 *     was specified through the "AnnotationSet" parameter.
 *
 *     In case of an error, NULL will be returned and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     Tcl_IsShared, tip_AddAttribute
 *   >:&:
 */
CDM_EXPORT
CDM_AttributeSet CDM_AddAttribute(CDM_AttributeSet AttributeSet,
                                  CDM_Attribute Attribute) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AddAttribute: Aborting...");
    return NULL;
  }

  if (AttributeSet == NULL || Attribute == NULL) {
    CDM_ReportError("CDM_AddAttribute: Invalid AttributeSet or "
                    "Attribute Pointer");
    return NULL;
  }
  if (Tcl_IsShared(AttributeSet)) {
#ifdef DUPLICATE_DEBUG
    CDM_Debug("AddAttribute: Duplicating (%d): %s\n", AttributeSet->refCount,
            Tcl_GetString(AttributeSet));
#endif /* DUPLICATE_DEBUG */
    AttributeSet = Tcl_DuplicateObj(AttributeSet);
  }
  if (AttributeSet == NULL) CDM_Panic("CDM_AddAttribute: Tcl_DuplicateObj"
                                      "returned NULL - out of memory?\n");
  if (Tcl_ListObjAppendElement(CDM_Interp, AttributeSet, Attribute) != TCL_OK) {
    CDM_ReportError("CDM_AddAttribute: Cannot append to given Attribute set!");
    return NULL;
  }
  return AttributeSet;
} /* CDM_AddAttribute */

/*
 * ---- CDM_AttributeSetGetElements:
 *  DESCRIPTION:
 *   <English:
 *    This function returns a count and a pointer to an array of the elements
 *    (Attributes) in an Attribute set object (of type CDM_AttributeSet).
 *    It returns the count by storing it in the address pointed by the "objcPtr"
 *    parameter (of type int *). Similarly, it returns the array pointer
 *    by storing it in the address pointed by the "objvPtr" parameter (of type
 *    CDM_Attribute **). The memory pointed to is managed by CDM and should
 *    not be freed by the caller.
 *
 *    In case of an error, TCL_ERROR will be returned and an error message will
 *    be left in the current active Tcl interpreter (CDM_Interp).
 *    Otherwise TCL_OK will be returned after storing the count and
 *    array pointer.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     AttributeSet: The Attribute set whose elements will be placed in the array. (CDM_AttributeSet)
 *     objcPtr:      A pointer to an integer, where the number of Attribute objects placed in the array will be saved. (int *)
 *     objvPtr:      A location where the pointer to the array of Attribute objects will be stored. (CDM_Attribute **)
 *   >
 *  NOTES:
 *   <English:
 *    Usage example:[[br]][[code]][[pre]]
 *    int objc;
 *    CDM_Attribute **objv;
 *    CDM_AttributeSet set = CDM_GetAttributes(Doc);
 *    int status = CDM_AttributeSetGetElements(set, objc, objv);
 *    if (status != TCL_OK) {
 *      // An error occured. What is the error message?
 *      Tcl_interp *interp = CDM_GetActiveInterpreter();
 *      if (interp) {
 *        const char* msg = Tcl_GetStringResult(interp);
 *      }
 *    }
 *    // Make an iteration over all Attributes...
 *    for (int i = 0; i < objc, i++) {
 *      // Now, objv[i] is an Attribute...
 *    }
 *    [[/pre]][[/code]]
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a count and a pointer to an array of the elements
 *    (Attributes) in an Attribute set object (of type CDM_AttributeSet).
 *    It returns the count by storing it in the address pointed by the "objcPtr"
 *    parameter (of type int *). Similarly, it returns the array pointer
 *    by storing it in the address pointed by the "objvPtr" parameter (of type
 *    CDM_Attribute **). The memory pointed to is managed by CDM and should
 *    not be freed by the caller.
 *
 *    In case of an error, TCL_ERROR will be returned and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). Otherwise TCL_OK will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_AnnotationSetGetElements, CDM_SpanSetGetElements
 *   >:&:
 */
CDM_EXPORT
int CDM_AttributeSetGetElements(CDM_AttributeSet AttributeSet,
                                int *objcPtr, CDM_Attribute **objvPtr) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    return CDM_ReportError("CDM_AttributeSetGetElements: Aborting...");
  }

  if (AttributeSet == NULL) {
    return CDM_ReportError("CDM_AttributeSetGetElements: Invalid AttributeSet "
                           "or Attribute Pointer");
  }

  return Tcl_ListObjGetElements(CDM_Interp, AttributeSet, objcPtr, objvPtr);
} /* CDM_AttributeSetGetElements */

/*
 * ---- CDM_CreateAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return an Attribute object
 *     (of type CDM_Attribute). In order to reduce memory requirements, this
 *     function will try to locate if an Attribute with exactly the same
 *     characteristics (same Name and Attribute value of the same type and
 *     value) already exists. In such a case, this (possibly shared) Attribute
 *     object will be returned. Else, a new Attribute object will be created
 *     and returned.
 *
 *     Special attention should be given to the Attribute value object
 *     (specified by the "Value" parameter) so as not to leak memory. This
 *     function internally will always increase and then decrease the
 *     reference count of this object, meaning that this object may be freed
 *     if it has a reference count of 0. If the caller wants to preserve a
 *     reference to this object, so as to free it at a later stage through
 *     CDM_Free or [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], Tcl_IncrRefCount should be called be
 *     called before CDM_CreateAttribute is called.
 *     The value of the "Name" parameter will not be modified
 *     or referenced in any way, as CDM will create and manipulate a local
 *     copy of this value.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Name:   The name of the Attribute.      (const char *)
 *     Value:  The value of the Attribute.     (CDM_AttributeValue)
 *   >
 *  NOTES:
 *   <English:
 *     Note that the object specified by the "Value" parameter will be freed,
 *     if it has a reference count equal to 0. If the caller wants to ensure
 *     that this object won't be freed by this function, Tcl_IncrRefCount must
 *     be used. Then the caller is responsible to free the object later with a
 *     call to either CDM_Free or Tcl_DecrRefCount. This behaviour has been
 *     chosen in order to avoid memory leakage when using code like the
 *     following example:
 *     [[br]][[code]][[center]]CDM_CreateAttribute("name",
 *     CDM_CreateAttributeValue(CDM_STRING, "value"));[[/center]][[/code]][[br]]
 *
 *     This function is equivelant to tip_CreateAttribute (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return an Attribute object (of type of
 *     CDM_Attribute). This object may be an old, already in use (shared)
 *     object or a newly created object, according to the specified parameters.
 *
 *     In case of an error, NULL will be returned and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp). In such a case, the object described by the "Value"
 *     parameter won't be altered in any way, leaving the responsibility to
 *     free it to the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_CreateAttribute
 *   >:&:
 */
CDM_EXPORT
CDM_Attribute CDM_CreateAttribute(const char *Name, CDM_AttributeValue Value) {
  CDM_Attribute Attr;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_CreateAttribute: Aborting...");
    return NULL;
  }

  if (Name == NULL) {
    CDM_ReportError("CDM_CreateAttribute: NULL \"Name\" parameter (char *)");
    return NULL;
  }
  if (Value == NULL) {
    CDM_ReportError("CDM_CreateAttribute: NULL \"Value\" parameter "
                    "(CDM_AttributeValue)");
    return NULL;
  }

  Tcl_IncrRefCount(Value);
  Attr = ObjectCache.FindAttributeObj(Name, Value);
  Tcl_DecrRefCount(Value);
  if (Attr == NULL) {
    /*
     * Add information to CDM_Interp trace...
     */
    Tcl_AppendResult(CDM_Interp, "\nwhile executing: CDM_CreateAttribute",NULL);
  }
#ifdef CDM_OBJ_HASH_KEYS
  /*
   * The following code is illegal in case we use ObjHashTables, as we have
   * already added the object in a list (Tcl_IncrRefCount). If the whole
   * Attribute object is hashed, the list would have been removed
   * (Tcl_DecrRefCount) => object has been freed!!!
   */
#else /* CDM_OBJ_HASH_KEYS */
  //if (Value->refCount == 0) Tcl_DecrRefCount(Value);
#endif /* CDM_OBJ_HASH_KEYS */
  return Attr;
} /* CDM_CreateAttribute */

/*
 * ---- CDM_CreateAttributeSet:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return an empty Attribute set object
 *     (of type CDM_AttributeSet). CDM does not own the returned
 *     object: the caller is responsible to use Tcl_DecrRefCount or CDM_Free
 *     in order to free the object and the memory associated with it.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None: (void)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_CreateAttributeSet (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an empty Attribute set object (of type
 *     CDM_AttributeSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_CreateAttributeSet
 *   >:&:
 */
CDM_EXPORT
CDM_AttributeSet CDM_CreateAttributeSet(void) {
  CDM_AttributeSet Set;
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_CreateAttributeSet: Aborting...");
    return NULL;
  }

  Set = Tcl_NewListObj(0, NULL);

  if (Set == NULL) {
    CDM_Panic("CDM_CreateAttributeSet: Tcl_NewListObj returned NULL "
              "- out of memory?\n");
  }
  return Set;
} /* CDM_CreateAttributeSet */

/*
 * ---- CDM_CreateAttributeSet:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new Attribute set object
 *     (of type CDM_AttributeSet) that will contain a newly created Attribute
 *     object, created according to the values specified as parameters. Note
 *     that this function is equivelant to creating an empty Attribute set
 *     object with CDM_CreateAttributeSet (void), creating a new Attribute
 *     object with CDM_CreateAttribute (char *Name, CDM_AttributeValue Value)
 *     and adding this Attribute object to the newly created Annotation set.
 *     Also note that the object specified through the "Value" parameter may
 *     be freed. For more details, please refer to the CDM_CreateAttribute
 *     documentation.
 *
 *     CDM does not own the returned object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Name:   Attribute's Name      (const char *)
 *      Value:  Attribute's Value     (CDM_AttributeValue)
 *   >
 *  NOTES:
 *   <English:
 *     Note that the object specified by the "Value" parameter will be freed,
 *     if it has a reference count equal to 0. For more details, please refer
 *     to the CDM_CreateAttribute documentation.
 *
 *     This function is equivelant to tip_CreateAttributeSet (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a new Attribute set object (of type
 *     CDM_AttributeSet) that will contain a newly created Attribute
 *     object, created according to the values specified as parameters.
 *     The caller is responsible to free this object through Tcl_DecrRefCount
 *     or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], CDM_CreateAttribute, tip_CreateAttributeSet
 *   >:&:
 */
CDM_EXPORT
CDM_AttributeSet CDM_CreateAttributeSet(const char *Name,
                                        CDM_AttributeValue Value) {
  CDM_AttributeSet Set;
  Tcl_Obj *argv[1];
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_CreateAttributeSet: Aborting...");
    return NULL;
  }

  argv[0] = CDM_CreateAttribute(Name, Value);
  if (argv[0] == NULL) return NULL;

  Set = Tcl_NewListObj(1, argv);
  if (Set == NULL)
    CDM_Panic("CDM_CreateAttributeSet: Tcl_NewListObj returned NULL "
              "- out of memory?\n");
  return Set;
} /* CDM_CreateAttributeSet */

/*
 * ---- CDM_CreateAttributeValue:
 *  DESCRIPTION:
 *   <English:
 *    Creates a new Attribute Value...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Type:   Attribute's Type      (int)
 *      Value:  Attribute's Value     (const char *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
CDM_AttributeValue CDM_CreateAttributeValue(int Type, const char *Value) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_CreateAttributeValue: Aborting...");
    return NULL;
  }

  switch (Type) {
    case CDM_STRING: {
      return ObjectCache.FindAttributeValueObj(CDM_STRING, Value);
    }
    case CDM_STRING_SET: {
      return ObjectCache.FindAttributeValueObj(CDM_STRING_SET, Value);
    }
    case CDM_IMAGE: {
      /* Actually, this function cannot be used to store binary data. As
       * binary data can contain NULL characters, there is no way we can
       * safely find the length of the binary data string.
       * Just issue an error... */
      CDM_ReportError("CDM_CreateAttributeValue: This function cannot be used "
      "to store binary data. Please convert the binary data into an object "
      "(with Tcl_NewByteArrayObj) and call CDM_CreateAttributeValue with this "
      "object instead of the binary data stored as (char *)");
      return NULL;
    }
    case CDM_BASE64_IMAGE: {
      Tcl_Obj *List[2];
      /* We don't re-use objects holding images in base64... */
      List[0] = &CDM_Obj_CDM_BASE64_IMAGE;
      List[1] = Tcl_NewStringObj(Value, -1);
      return Tcl_NewListObj(2, List);
    }
    default: {
       CDM_ReportError("Unkown Attribute value type \"%d\"", Type);
    }
  }
  return NULL;
} /* CDM_CreateAttributeValue */

/*
 * ---- CDM_CreateAttributeValue:
 *  DESCRIPTION:
 *   <English:
 *    Creates a new Attribute Value...
 *    Use this function when the type is "CDM_STING_SET" to reuse the given
 *    object!
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Type:   Attribute's Type      (int)
 *      Value:  Attribute's Value     (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
CDM_AttributeValue CDM_CreateAttributeValue(int Type, Tcl_Obj *Value) {
  CDM_AttributeValue Attr;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_CreateAttributeValue: Aborting...");
    return NULL;
  }
  if (Value == NULL) {
    CDM_ReportError("CDM_CreateAttributeValue: NULL \"Value\" parameter "
                    "(Tcl_Obj *)");
    return NULL;
  }

  switch (Type) {
    case CDM_STRING: {
      Tcl_IncrRefCount(Value);
      Attr = ObjectCache.FindAttributeValueObj(CDM_STRING, Value);
      Tcl_DecrRefCount(Value);
      break;
    }
    case CDM_STRING_SET: {
      Tcl_IncrRefCount(Value);
      Attr = ObjectCache.FindAttributeValueObj(CDM_STRING_SET, Value);
      Tcl_DecrRefCount(Value);
      break;
    }
    case CDM_IMAGE: {
      Tcl_Obj *List[2];
      /* We don't re-use objects holding images as binary data... */
      List[0] = &CDM_Obj_CDM_IMAGE;
      List[1] = Value;
      return Tcl_NewListObj(2, List);
    }
    case CDM_BASE64_IMAGE: {
      Tcl_Obj *List[2];
      /* We don't re-use objects holding images in base64... */
      List[0] = &CDM_Obj_CDM_BASE64_IMAGE;
      List[1] = Value;
      return Tcl_NewListObj(2, List);
    }
    default: {
       CDM_ReportError("Unkown Attribute value type \"%d\"", Type);
       return NULL;
    }
  }
  return Attr;
} /* CDM_CreateAttributeValue */

/*
 * ---- CDM_GetName:
 *  DESCRIPTION:
 *   <English:
 *    Returns Attribute's Name...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Attribute:   Attribute      (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetName(CDM_Attribute Attribute) {
  Tcl_Obj **elements;
  int result, items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetName: Aborting...");
    return NULL;
  }

  result = Tcl_ListObjGetElements(CDM_Interp, Attribute, &items, &elements);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("Malformed Attribute \"%s\"", Tcl_GetString(Attribute));
    return NULL;
  }
  return Tcl_GetString(elements[0]);
} /* CDM_GetName */

/*
 * ---- CDM_GetNameObj:
 *  DESCRIPTION:
 *   <English:
 *    Returns Attribute's Name...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Attribute:   Attribute      (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
CDM_AttributeName CDM_GetNameObj(CDM_Attribute Attribute) {
  Tcl_Obj **elements;
  int result, items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetNameObj: Aborting...");
    return NULL;
  }

  result = Tcl_ListObjGetElements(CDM_Interp, Attribute, &items, &elements);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("Malformed Attribute \"%s\"", Tcl_GetString(Attribute));
    return NULL;
  }
  return elements[0];
} /* CDM_GetNameObj */

/*
 * ---- CDM_GetValue:
 *  DESCRIPTION:
 *   <English:
 *    Returns Attribute's Value...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Attribute:   Attribute      (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
CDM_AttributeValue CDM_GetValue(CDM_Attribute Attribute) {
  Tcl_Obj **elements, **ValueElements;
  int result, items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetValue: Aborting...");
    return NULL;
  }

  result =
      Tcl_ListObjGetElements(CDM_Interp, Attribute, &items, &elements);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("Malformed Attribute \"%s\"", Tcl_GetString(Attribute));
    goto error;
  }
  result =
      Tcl_ListObjGetElements(CDM_Interp, elements[1], &items, &ValueElements);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("Malformed Attribute Value\"%s\"",
                    Tcl_GetString(elements[1]));
    goto error;
  }
  return elements[1];
error:
  return NULL;
} /* CDM_GetValue */

/*
 * ---- CDM_GetValueType:
 *  DESCRIPTION:
 *   <English:
 *    Returns Attribute's Type...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Attribute:   Attribute      (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
int CDM_GetValueType(CDM_Attribute Attribute) {
  Tcl_Obj **elements, **ValueElements;
  int result, items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetValueType: Aborting...");
    return -1;
  }

  result =
      Tcl_ListObjGetElements(CDM_Interp, Attribute, &items, &elements);
  if (result != TCL_OK || items != 2) {
    snprintf(CDM_Temp, MAX_TEMP_STRING, "Malformed Attribute \"%s\"",
                   Tcl_GetString(Attribute));
    goto error;
  }
  result =
      Tcl_ListObjGetElements(CDM_Interp, elements[1], &items, &ValueElements);
  if (result != TCL_OK || items != 2) {
    snprintf(CDM_Temp, MAX_TEMP_STRING, "Malformed Attribute Value\"%s\"",
                   Tcl_GetString(elements[1]));
    goto error;
  }
  result = CDM_GetAttributeTypeFromObject(ValueElements[0]);
  if (result < CDM_STRING) {
    snprintf(CDM_Temp, MAX_TEMP_STRING, "Invalid type found (\"%s\") while "
           "processing Attribute \"%s\"",
           Tcl_GetString(elements[1]), Tcl_GetString(Attribute));
    goto error;
  }
  return result;
error:
  CDM_ReportError(CDM_Temp);
  return -1;
} /* CDM_GetValueType */

/*
 * ---- CDM_GetValueString:
 *  DESCRIPTION:
 *   <English:
 *    Returns Attribute's Value String...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Attribute:   Attribute      (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetValueString(CDM_Attribute Attribute) {
  Tcl_Obj **elements, **ValueElements;
  int result, items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetValueString: Aborting...");
    return NULL;
  }

  result =
      Tcl_ListObjGetElements(CDM_Interp, Attribute, &items, &elements);
  if (result != TCL_OK || items != 2) {
    snprintf(CDM_Temp, MAX_TEMP_STRING, "Malformed Attribute \"%s\"",
                   Tcl_GetString(Attribute));
    goto error;
  }
  result =
      Tcl_ListObjGetElements(CDM_Interp, elements[1], &items, &ValueElements);
  if (result != TCL_OK || items != 2) {
    snprintf(CDM_Temp, MAX_TEMP_STRING, "Malformed Attribute Value\"%s\"",
                   Tcl_GetString(elements[1]));
    goto error;
  }
  return Tcl_GetString(ValueElements[1]);
error:
  CDM_ReportError(CDM_Temp);
  return NULL;
} /* CDM_GetValueString */

/*
 * ---- CDM_GetValueValue:
 *  DESCRIPTION:
 *   <English:
 *    This function will return the value of the Attribute value of the
 *    specified Attribute. An Attribute value consists of two elements: the
 *    type of the information stored as Attribute value (for example
 *    GDM_STRING or GDM_STRING_SET) and the information that is stored as a
 *    value of the Attribute. This function returns the latter object.
 *    This function is different from CDM_GetValue, as CDM_GetValue returns a
 *    list object that holds two elements, the value type and the value. On
 *    the other hand, CDM_GetValueValue returns directly the value, by
 *    ommiting the object specifying the type. If the caller wants to also get
 *    the type of the value, he can either call CDM_GetValue or
 *    CDM_GetValueType.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Attribute: The Attribute object whose value is wanted. (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetValueValue (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns the object that holds the value stored in an
 *     Attribute object. The returned object is property of the Attribute and
 *     should not be modified or freed. If the caller wants to keep a reference
 *     to the returned object, he must place such a
 *     request by using
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]].
 *     When the returned object is not needed any more and the caller wants to
 *     release the object, the functions
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *     CDM_Free must be used.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetValue, CDM_GetValueValue, CDM_Free, tip_GetValueValue,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]],
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]]
 *   >:&:
 */
CDM_EXPORT
CDM_AttributeValue CDM_GetValueValue(CDM_Attribute Attribute) {
  Tcl_Obj **elements, **ValueElements;
  int result, items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetValueValue: Aborting...");
    return NULL;
  }

  result =
      Tcl_ListObjGetElements(CDM_Interp, Attribute, &items, &elements);
  if (result != TCL_OK || items != 2) {
    snprintf(CDM_Temp, MAX_TEMP_STRING, "Malformed Attribute \"%s\"",
                   Tcl_GetString(Attribute));
    goto error;
  }
  result =
      Tcl_ListObjGetElements(CDM_Interp, elements[1], &items, &ValueElements);
  if (result != TCL_OK || items != 2) {
    snprintf(CDM_Temp, MAX_TEMP_STRING, "Malformed Attribute Value\"%s\"",
                   Tcl_GetString(elements[1]));
    goto error;
  }
  return ValueElements[1];
error:
  CDM_ReportError(CDM_Temp);
  return NULL;
} /* CDM_GetValueValue */

/*
 * ---- CDM_ReconstructAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function accepts as argument a valid Attribute object (of type
 *     CDM_Attribute). This function will construct and return a new Attribute
 *     object that will be an exact copy of the provided one. The only
 *     difference will be that the returned Attribute object will utilise
 *     elements from the object cache that is managed internally by CDM. As a
 *     result the returned object will normally require less memory to be
 *     stored than the provided one, as it will re-use elements already in use
 *     by other CDM objects.
 *
 *     This function should be used on Attribute objects that originate from
 *     an unknown source. For example, if an Attribute object is stored in a
 *     file and then red back as a string, it will not use cached by the CDM
 *     objects. It will be a good idea if the caller created a new object by
 *     calling this function and delete the object red from the file by either
 *     calling [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *     CDM_Free.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Attribute: The Attribute object that is to be reconstucted. (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_ReconstructAttribute (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a new Attribute object (of type CDM_Attribute)
 *     that is an exact copy of the porvided Attribute object, but requires
 *     less memory due to re-using cached objects.
 *
 *     In case of an error (i.e. the specified object is not a valid Attribute
 *     object) NULL will be returned and an error message will be left at the
 *     current active interpreter (CDM_Interp).
 *
 *     The returned object is not property of CDM and the caller is
 *     responsible to free this object when it is no longer needed by either
 *     calling [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *     CDM_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_ReconstructAnnotation, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_ReconstructAttribute
 *   >:&:
 */
CDM_EXPORT
CDM_Attribute CDM_ReconstructAttribute(CDM_Attribute Attribute) {
  Tcl_Obj **elements, **ValueElements, *Name, *Value;
  int result, items, Type;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_ReconstructAttribute: Aborting...");
    return NULL;
  }

  /*
   * Get the Attribute name and value objects...
   */
  result = Tcl_ListObjGetElements(NULL, Attribute, &items, &elements);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("Malformed Attribute \"%s\"", Tcl_GetString(Attribute));
    return NULL;
  }

  Name  = elements[0];
  Value = elements[1];

  /*
   * Get value type and value from the Attribute value...
   */
  result = Tcl_ListObjGetElements(NULL, Value, &items, &ValueElements);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("Malformed Attribute Value\"%s\"", Tcl_GetString(Value));
    return NULL;
  }

  Type = CDM_GetAttributeTypeFromObject(ValueElements[0]);
  if (Type < CDM_STRING) return NULL;

  /*
   * Now we have all required information to reconstruct the Attribute...
   */
  Value = CDM_CreateAttributeValue(Type, ValueElements[1]);
  if (Value == NULL) return NULL;
  return CDM_CreateAttribute(Tcl_GetString(Name), Value);
} /* CDM_ReconstructAttribute */

/*
 * ---- CDM_ReconstructAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function accepts as argument a valid Attribute object (of type
 *     CDM_Attribute). This function will construct and return a new Attribute
 *     object that will be an exact copy of the provided one. The only
 *     difference will be that the returned Attribute object will utilise
 *     elements from the object cache that is managed internally by CDM. As a
 *     result the returned object will normally require less memory to be
 *     stored than the provided one, as it will re-use elements already in use
 *     by other CDM objects.
 *
 *     This function should be used on Attribute objects that originate from
 *     an unknown source. For example, if an Attribute object is stored in a
 *     file and then red back as a string, it will not use cached by the CDM
 *     objects. It will be a good idea if the caller created a new object by
 *     calling this function and delete the object red from the file by either
 *     calling [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *     CDM_Free.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Attribute: The Attribute object that is to be reconstucted. (CDM_Attribute)
 *     type: A pointer to an integer, where the type of the Attribute value will be stored. (int *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_ReconstructAttribute (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a new Attribute object (of type CDM_Attribute)
 *     that is an exact copy of the porvided Attribute object, but requires
 *     less memory due to re-using cached objects.
 *
 *     In the provided integer pointer, the Type of the Attribute value will
 *     be stored, if not NULL.
 *
 *     In case of an error (i.e. the specified object is not a valid Attribute
 *     object) NULL will be returned and an error message will be left at the
 *     current active interpreter (CDM_Interp).
 *
 *     The returned object is not property of CDM and the caller is
 *     responsible to free this object when it is no longer needed by either
 *     calling [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *     CDM_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_ReconstructAnnotation, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_ReconstructAttribute
 *   >:&:
 */
CDM_EXPORT
CDM_Attribute CDM_ReconstructAttribute(CDM_Attribute Attribute, int *type) {
  Tcl_Obj **elements, **ValueElements, *Name, *Value;
  int result, items, Type;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_ReconstructAttribute: Aborting...");
    return NULL;
  }

  /*
   * Get the Attribute name and value objects...
   */
  result = Tcl_ListObjGetElements(NULL, Attribute, &items, &elements);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("Malformed Attribute \"%s\"", Tcl_GetString(Attribute));
    return NULL;
  }

  Name  = elements[0];
  Value = elements[1];

  /*
   * Get value type and value from the Attribute value...
   */
  result = Tcl_ListObjGetElements(NULL, Value, &items, &ValueElements);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("Malformed Attribute Value\"%s\"", Tcl_GetString(Value));
    return NULL;
  }

  Type = CDM_GetAttributeTypeFromObject(ValueElements[0]);
  if (Type < CDM_STRING) return NULL;
  if (type != NULL) *type = Type;

  /*
   * Now we have all required information to reconstruct the Attribute...
   */
  Value = CDM_CreateAttributeValue(Type, ValueElements[1]);
  if (Value == NULL) return NULL;
  return CDM_CreateAttribute(Tcl_GetString(Name), Value);
} /* CDM_ReconstructAttribute */

/*
 * CDM_GetAttributeTypeFromObject:
 * This internal function is used to retrieve the Annotation type stored in
 * the specified object as string.
 */
int CDM_GetAttributeTypeFromObject(Tcl_Obj *Type) {
  int index;
  static const char *Types[] = {
    "GDM_STRING",     "CDM_STRING",
    "GDM_STRING_SET", "CDM_STRING_SET",
    "CDM_IMAGE",      "CDM_BASE64_IMAGE",
    (char *) NULL
  };
  enum types {
    STRING,     STRING2,
    STRING_SET, STRING_SET2,
    IMAGE,      BASE64_IMAGE
  };
  CDM_THREAD_SPECIFIC_DATA;

  if (Tcl_GetIndexFromObj(CDM_Interp, Type, (const char **) Types,
                          "Attribute Value Type", 0, &index) != TCL_OK) {
    return -1;
  }

  switch ((enum types) index) {
    case STRING:
    case STRING2:      {return CDM_STRING;}
    case STRING_SET:
    case STRING_SET2:  {return CDM_STRING_SET;}
    case IMAGE:        {return CDM_IMAGE;}
    case BASE64_IMAGE: {return CDM_BASE64_IMAGE;}
  }
  return -1;
} /* CDM_GetAttributeTypeFromObject */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |   E N D   O F   A T T R I B U T E   S E C T I O N     |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/


/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     A N N O T A T I O N            S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 * ---- CDM_AnnotationContainsAttributeMatchingValue:
 *  DESCRIPTION:
 *   <English:
 *    This function will check the provided Annotation (specified through the
 *    "Annotation" parameter) to locate an Attribute that has the specified
 *    name (through the "AttributeName" parameter). If such an Attribute is
 *    not found in the Annotation, 0 will be returned. If a proper Attribute
 *    is found, its value will be matched against the provided (though the
 *    "ValuePattern" parameter) pattern: if a match is found 1 will be
 *    returned, else this function will return 0.
 *
 *    The pattern specified through the "ValuePattern" parameter must follow
 *    the same rules as the patterns used with the
 *    %%%quot;[[A HREF="../TclTk/TclCmd/string.htm#M35"
 *    ]]string match[[/A]]%%%quot; Tcl command.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Ann:           The Annotation to be examined             (CDM_Annotation)
 *     AttributeName: The name of the desired Attribute            (char *)
 *     ValuePattern:  The pattern the Attribute Value should match (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationContainsAttributeMatchingValue (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return 1 (true) if an Attribute with the provided
 *    name exists in the specified Annotation, and its value matches the
 *    provided pattern. Else, 0 (false) will be returned.
 *
 *    In case of an error, -1 will be returned and an error message will be
 *    left at the current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_AnnotationContainsAttributeMatchingValues,
 *     tip_AnnotationContainsAttributeMatchingValue,
 *     [[A HREF="../TclTk/TclCmd/string.htm#M35"]]string match[[/A]]
 *   >:&:
 */
CDM_EXPORT
int CDM_AnnotationContainsAttributeMatchingValue(CDM_Annotation Ann,
                                     char *AttributeName, char *ValuePattern) {
  CDM_Attribute Attr;
  char *value;
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationContainsAttributeMatchingValue: "
                    "Aborting...");
    return -1;
  }

  if (!CDM_ValidAnnotation(Ann)) return -1;

  Attr = CDM_GetAttribute(Ann, AttributeName);
  /* The Attribute was not found... */
  if (Attr == NULL) return 0;
  value = CDM_GetValueString(Attr);
  if (value == NULL) return -1;

  /* We check whether the value and the pattern match: */
  return Tcl_StringMatch(value, ValuePattern);
} /* CDM_AnnotationContainsAttributeMatchingValue */

/*
 * ---- CDM_AnnotationContainsAttributeMatchingValues:
 *  DESCRIPTION:
 *   <English:
 *    This function will check the provided Annotation (specified through the
 *    "Annotation" parameter) to locate an Attribute that has the specified
 *    name (through the "AttributeName" parameter). If such an Attribute is
 *    not found in the Annotation, 0 will be returned. If a proper Attribute
 *    is found, its value will be matched against the set of provided
 *    (though the "ValuePatternsObj" parameter) patterns:
 *    if a match is found with any of the patterns 1 will be
 *    returned, else this function will return 0.
 *
 *    The patterns specified through the "ValuePatternsObj" parameter must
 *    follow the same rules as the patterns used with the
 *    %%%quot;[[A HREF="../TclTk/TclCmd/string.htm#M35"
 *    ]]string match[[/A]]%%%quot; Tcl command.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Ann:              The Annotation to be examined           (CDM_Annotation)
 *    AttributeName:    The name of the desired Attribute            (char *)
 *    ValuePatternsObj: The pattern the Attribute Value should match (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationContainsAttributeMatchingValues (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return 1 (true) if an Attribute with the provided
 *    name exists in the specified Annotation, and its value matches any of the
 *    provided patterns. Else, 0 (false) will be returned.
 *
 *    In case of an error, -1 will be returned and an error message will be
 *    left at the current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_AnnotationContainsAttributeMatchingValue,
 *     tip_AnnotationContainsAttributeMatchingValues,
 *     [[A HREF="../TclTk/TclCmd/string.htm#M35"]]string match[[/A]]
 *   >:&:
 */
CDM_EXPORT
int CDM_AnnotationContainsAttributeMatchingValues(CDM_Annotation Ann,
                               char *AttributeName, Tcl_Obj *ValuePatternsObj) {
  CDM_Attribute Attr;
  Tcl_Obj **pattern;
  char *value;
  int items, i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationContainsAttributeMatchingValues: Aborting...");
    return -1;
  }

  if (!CDM_ValidAnnotation(Ann)) return -1;

  Attr = CDM_GetAttribute(Ann, AttributeName);
  /* The Attribute was not found... */
  if (Attr == NULL) return 0;
  value = CDM_GetValueString(Attr);
  if (value == NULL) return -1;

  if (Tcl_ListObjGetElements(CDM_Interp, ValuePatternsObj, &items, &pattern)
                                         != TCL_OK) return -1;
  if (items == 0) return 1;
  for (i=0; i<items; i++) {
    /* We check whether the value and the pattern match: */
    if (Tcl_StringMatch(value, Tcl_GetString(pattern[i]))) return 1;
  }
  return 0;
} /* CDM_AnnotationContainsAttributeMatchingValues */

/*
 * ---- CDM_AnnotationContainsPosition:
 *  DESCRIPTION:
 *   <English:
 *    This function will check if the provided Annotation (specified through the
 *    "Annotation" parameter) contains in any of its spans the specified
 *    position.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Ann:      The Annotation to be examined         (CDM_Annotation)
 *    Position: The position to me checked            (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationContainsPosition (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return 1 (true) if the Annotation contains the
 *    specified position in any of its spans. Else, 0 (false)
 *    will be returned.
 *
 *    In case of an error, -1 will be returned and an error message will be
 *    left at the current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_AnnotationMatchesRange, tip_AnnotationContainsPosition
 *   >:&:
 */
CDM_EXPORT
int CDM_AnnotationContainsPosition(const CDM_Annotation Ann,
                                   const long Position) {
  long start, end;
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationContainsPosition: Aborting...");
    return -1;
  }

  if (!CDM_ValidAnnotation(Ann)) return -1;
#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Ann) && CDMAnnObj_GetStart(Ann) != -1) {
    /* This Annotation has a single span. Handle it directly... */
    start = CDMAnnObj_GetStart(Ann);
    end   = CDMAnnObj_GetEnd(Ann);
    if (start <= Position && Position <= end) return 1;
    return 0;
  } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
     /* Either the Annotation object type is disabled, or the Annotation has
     * more than one Span... */
    CDM_SpanSet Spans = CDM_GetSpans(Ann);
    Tcl_Obj **Span;
    int status, items, i;
    if (Spans == NULL) {
      CDM_ReportError("CDM_AnnotationContainsPosition: Document has "
                      "Annotation without Span?\n");
      CDM_AppendError(Tcl_GetString(Ann));
      return -1;
    }
    status = Tcl_ListObjGetElements(CDM_Interp, Spans, &items, &Span);
    if (status != TCL_OK) {
      CDM_ReportError("CDM_AnnotationContainsPosition: Error Getting Spans "
                      "From Annotation ");
      CDM_AppendError(Tcl_GetString(Ann));
      return -1;
    }
    for (i = 0; i < items; i++) {
      if (CDM_GetSpanOffsets(Span[i], &start, &end) != TCL_OK) {
        CDM_ReportError("CDM_AnnotationContainsPosition: Error Getting Span "
                        "%d From Annotation", i);
        CDM_AppendError(Tcl_GetString(Ann));
        return -1;
      }
      if (start <= Position && Position <= end) return 1;
    }
#ifdef ENABLE_ANNOTATION_OBJECT
  }
#endif /* ENABLE_ANNOTATION_OBJECT */
  return 0;
} /* CDM_AnnotationContainsPosition */

/*
 * ---- CDM_AnnotationMatchesRange:
 *  DESCRIPTION:
 *   <English:
 *    This function will check if the provided Annotation (specified through the
 *    "Annotation" parameter) contains a span that has the specified start and
 *    end offsets.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Ann:    The Annotation to be examined           (CDM_Annotation)
 *    Start:  The span start to me matched            (long)
 *    End:    The span end to me matched              (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationMatchesRange (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return 1 (true) if the Annotation contains a span
 *    that matches the specified start and end offsets. Else, 0 (false)
 *    will be returned.
 *
 *    In case of an error, -1 will be returned and an error message will be
 *    left at the current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_AnnotationContainsPosition, tip_AnnotationMatchesRange
 *   >:&:
 */
CDM_EXPORT
int CDM_AnnotationMatchesRange(const CDM_Annotation Ann,
                               const long Start, const long End) {
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationMatchesRange: Aborting...");
    return -1;
  }

  if (!CDM_ValidAnnotation(Ann)) return -1;
#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Ann) && CDMAnnObj_GetStart(Ann) != -1) {
    /* This Annotation has a single span. Handle it directly... */
    if (Start != CDMAnnObj_GetStart(Ann)) return 0;
    if (End   != CDMAnnObj_GetEnd(Ann)) return 0;
    return 1;
  } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
     /* Either the Annotation object type is disabled, or the Annotation has
     * more than one Span... */
    CDM_SpanSet Spans = CDM_GetSpans(Ann);
    Tcl_Obj **Span;
    int status, items, i;
    long start, end;
    if (Spans == NULL) {
      CDM_ReportError("CDM_AnnotationMatchesRange: Document has "
                      "Annotation without Span?\n");
      CDM_AppendError(Tcl_GetString(Ann));
      return -1;
    }
    status = Tcl_ListObjGetElements(CDM_Interp, Spans, &items, &Span);
    if (status != TCL_OK) {
      CDM_ReportError("CDM_AnnotationMatchesRange: Error Getting Spans "
                      "From Annotation ");
      CDM_AppendError(Tcl_GetString(Ann));
      return -1;
    }
    for (i = 0; i < items; i++) {
      if (CDM_GetSpanOffsets(Span[i], &start, &end) != TCL_OK) {
        CDM_ReportError("CDM_AnnotationMatchesRange: Error Getting Span "
                        "%d From Annotation", i);
        CDM_AppendError(Tcl_GetString(Ann));
        return -1;
      }
      if (start == Start && end == End ) return 1;
    }
#ifdef ENABLE_ANNOTATION_OBJECT
  }
#endif /* ENABLE_ANNOTATION_OBJECT */
  return 0;
} /* CDM_AnnotationMatchesRange */

/*
 * ---- CDM_CompareAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will compare two Annotations, Ann1 and Ann2, according to
 *     their Span sets.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Ann1:        The first Annotation                (CDM_Annotation)
 *     Ann2:        The second Annotation               (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_CompareAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns -1 if Ann1 is lower than
 *    Ann2, i.e. Ann1 spans a text range that starts earlier than all the
 *    text ranges spanned by the Ann2 Annotation. If the two Annotations span
 *    the same text ranges, zero is returned. Finally, 1
 *    is returned if Ann1 is greater than Ann2, i.e. when Ann2 spans a
 *    text range starts earlier than all the text ranges spanned by the Ann1
 *    Annotation. In case of an error, a value equal or greater than 10 will
 *    be returned and an error message will be left at the current Tcl
 *    interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_SelectAnnotationsSorted, tip_CompareAnnotations,
 *     [[A HREF="../TclTk/TclCmd/lsort.htm"]]lsort[[/A]]
 *   >:&:
 */
CDM_EXPORT
int CDM_CompareAnnotations(CDM_Annotation Ann1, CDM_Annotation Ann2) {
  CDM_Span SpanSet1, SpanSet2;
  Tcl_Obj **Span1, **Span2;
  long start, end, minstart1 = -1, minend1 = -1, minstart2 = -1, minend2 = -1;
  int status, items1, items2, i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_CompareAnnotations: Aborting...");
    return 20;
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Ann1) && ANNOTATION_OBJ(Ann2) &&
      CDMAnnObj_GetStart(Ann1) != -1 && CDMAnnObj_GetStart(Ann2) != -1) {
    minstart1 = CDMAnnObj_GetStart(Ann1);
    minend1   = CDMAnnObj_GetEnd(Ann1);
    minstart2 = CDMAnnObj_GetStart(Ann2);
    minend2   = CDMAnnObj_GetEnd(Ann2);
  } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
    /*
     * Is our input valid?
     */
    if (!CDM_ValidAnnotation(Ann1)) return 30;
    if (!CDM_ValidAnnotation(Ann2)) return 40;

    /*
     * Now its time to compare our Annotations...
     */
    SpanSet1 = CDM_GetSpans(Ann1);
    if (SpanSet1 == NULL) return 30;
    SpanSet2 = CDM_GetSpans(Ann2);
    if (SpanSet2 == NULL) return 40;

    /*
     * Get the various spans out of the span set...
     */
    status = Tcl_ListObjGetElements(NULL, SpanSet1, &items1, &Span1);
    if (status != TCL_OK) return 31;
    status = Tcl_ListObjGetElements(NULL, SpanSet2, &items2, &Span2);
    if (status != TCL_OK) return 41;

    /*
     * For each Annotation, get the earlier span range...
     */
    for (i=0; i<items1 && i<items2; i++) {
      status = CDM_GetSpanOffsets(Span1[i], &start, &end);
      if (status != TCL_OK) return 32;
      if (minstart1 < 0 || minstart1 > start) minstart1 = start;
      if (minend1   < 0 || minend1   > end)   minend1   = end;
      status = CDM_GetSpanOffsets(Span2[i], &start, &end);
      if (status != TCL_OK) return 42;
      if (minstart2 < 0 || minstart2 > start) minstart2 = start;
      if (minend2   < 0 || minend2   > end)   minend2   = end;
    }
#ifdef ENABLE_ANNOTATION_OBJECT
  }
#endif /* ENABLE_ANNOTATION_OBJECT */

  /*
   * ... and compare them...
   */
  if (minstart1  > minstart2) return  1;
  if (minstart1  < minstart2) return -1;
  if (minstart1 == minstart2) {
    /* We consider the Annotation having the longest span as greater... */
    if (minend1  > minend2) return  1;
    if (minend1  < minend2) return -1;
    if (minend1 == minend2) {
      /* We consider the Annotation having the more spans as greater... */
      if (items1 == items2) return  0; /* Missing. Bug reported by Sigletos */
      if (items1 >  items2) return  1;
      if (items1 <  items2) return -1;
    }
  }
  return -1;
} /* CDM_CompareAnnotations */

/*
 * ---- CDM_AnnotationSetGetElements:
 *  DESCRIPTION:
 *   <English:
 *    This function returns a count and a pointer to an array of the elements
 *    (Annotations) in an Annotation set object (of type CDM_AnnotationSet).
 *    It returns the count by storing it in the address pointed by the "objcPtr"
 *    parameter (of type int *). Similarly, it returns the array pointer
 *    by storing it in the address pointed by the "objvPtr" parameter (of type
 *    CDM_Annotation **). The memory pointed to is managed by CDM and should
 *    not be freed by the caller.
 *
 *    In case of an error, TCL_ERROR will be returned and an error message will
 *    be left in the current active Tcl interpreter (CDM_Interp).
 *    Otherwise TCL_OK will be returned after storing the count and
 *    array pointer.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     AnnotationSet: The Annotation set whose elements will be placed in the array. (CDM_AnnotationSet)
 *     objcPtr:      A pointer to an integer, where the number of Annotation objects placed in the array will be saved. (int *)
 *     objvPtr:      A location where the pointer to the array of Annotation objects will be stored. (CDM_Annotation **)
 *   >
 *  NOTES:
 *   <English:
 *    Usage example:[[br]][[code]][[pre]]
 *    int objc;
 *    CDM_Annotation **objv;
 *    CDM_AnnotationSet set = CDM_GetAnnotations(Doc);
 *    int status = CDM_AnnotationSetGetElements(set, objc, objv);
 *    if (status != TCL_OK) {
 *      // An error occured. What is the error message?
 *      Tcl_interp *interp = CDM_GetActiveInterpreter();
 *      if (interp) {
 *        const char* msg = Tcl_GetStringResult(interp);
 *      }
 *    }
 *    // Make an iteration over all Annotations...
 *    for (int i = 0; i < objc, i++) {
 *      // Now, objv[i] is an Annotation...
 *    }
 *    [[/pre]][[/code]]
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a count and a pointer to an array of the elements
 *    (Annotations) in an Annotation set object (of type CDM_AnnotationSet).
 *    It returns the count by storing it in the address pointed by the "objcPtr"
 *    parameter (of type int *). Similarly, it returns the array pointer
 *    by storing it in the address pointed by the "objvPtr" parameter (of type
 *    CDM_Annotation **). The memory pointed to is managed by CDM and should
 *    not be freed by the caller.
 *
 *    In case of an error, TCL_ERROR will be returned and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). Otherwise TCL_OK will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_AttributeSetGetElements, CDM_SpanSetGetElements
 *   >:&:
 */
CDM_EXPORT
int CDM_AnnotationSetGetElements(CDM_AnnotationSet AnnotationSet,
                                int *objcPtr, CDM_Annotation **objvPtr) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    return CDM_ReportError("CDM_AnnotationSetGetElements: Aborting...");
  }

  if (AnnotationSet == NULL) {
    return CDM_ReportError("CDM_AnnotationSetGetElements: Invalid AnnotationSet"
                           " or Annotation Pointer");
  }

  return Tcl_ListObjGetElements(CDM_Interp, AnnotationSet, objcPtr, objvPtr);
} /* CDM_AnnotationSetGetElements */

/*
 * ---- CDM_CreateAnnotation:
 *  DESCRIPTION:
 *   <English:
 *    This function will create and return a new Annotation object. The newly
 *    created Annotation object will have an empty Annotation Id and the
 *    Annotation type, Span set and Annotation set will be set to the values
 *    provided by the caller through the corresponding parameters.
 *
 *    This function may or may not use
 *    the provided Span and Attribute set object: this function will internally
 *    increment/decrement the reference count of the provided set
 *    objects. If the reference count of any of the provided set objects
 *    equals 0 (i.e. the Span set object has been created with a call to
 *    CDM_CreateSpanSet or the Attribute set object has been created with
 *    a call to CDM_CreateAttribute), this function may free the object.
 *    As a result, if
 *    the caller wants to ensure that any of the provided set objects is
 *    valid after calling this function, he should increment its reference
 *    count with [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]].
 *
 *    Note that the returned object is not property of the CDM: the caller is
 *    responsible for using
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *    CDM_Free to free the object
 *    and the memory associated with it when the object is not needed anymore.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Type:        The Annotation's Type                     (const char *)
 *      Spans:       A Set of Spans that the Annotation Covers (CDM_SpanSet)
 *      Attributes:  A Set of Attributes                  (CDM_AttributeSet)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to tip_CreateAnnotation (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will create and return a new Annotation object (of type
 *    CDM_Annotation). In case of an error, NULL will be returned and an error
 *    message describing the error will be left at the current active
 *    Tcl interp (CDM_Interp).
 *
 *    Note that the returned object is not property of the CDM: the caller is
 *    responsible for using
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *    CDM_Free to free the object
 *    and the memory associated with it when the object is not needed anymore.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]],
 *    tip_CreateAnnotation
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation
CDM_CreateAnnotation(const char *Type, CDM_SpanSet Spans,
                     CDM_AttributeSet Attributes) {
  CDM_Annotation Ann;
  Tcl_Obj *argv[4], **Span, **Attr;
  long start, end;
  int i, status, spans, attrs;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_CreateAnnotation: Aborting...");
    return NULL;
  }

  /*
   * Did we get a valid Span set?
   */
  if (Spans == NULL) {
    Spans = &CDM_Obj_NULL;
    spans = 0;
  } else {
    if (Tcl_ListObjGetElements(NULL, Spans, &spans, &Span) != TCL_OK) {
      CDM_ReportError("CDM_CreateAnnotation: Invalid Span Set object: ");
      CDM_AppendError(Tcl_GetString(Spans));
      return NULL;
    }
    for (i=0; i<spans; i++) {
      status = CDM_GetSpanOffsets(Span[i], &start, &end);
      if (status != TCL_OK) {
        CDM_AppendError(": mallformed Span in position %d in Span Set \"%s\"",
                        i, Tcl_GetString(Spans));
        return NULL;
      }
    }
  }

  /*
   * Did we get a valid Attribute set?
   */
  if (Attributes == NULL) {
    Attributes = &CDM_Obj_NULL;
  } else {
    if (Tcl_ListObjGetElements(NULL, Attributes, &attrs, &Attr) != TCL_OK) {
      CDM_ReportError("CDM_CreateAnnotation: Invalid Attribute Set object: ");
      CDM_AppendError(Tcl_GetString(Attributes));
      return NULL;
    }
    for (i=0; i<attrs; i++) {
      if (!CDM_ValidAttribute(Attr[i])) {
        return NULL;
      }
    }
  }

  // Add the (Empty) Id
  argv[0] = &CDM_Obj_NULL;


  // Add type
  argv[1] = ObjectCache.FindAnnotationType(Type);
  if (argv[1] == NULL) CDM_Panic("CDM_CreateAnnotation: "
    "ObjectCache::FindAnnotationType returned NULL - out of memory?\n");

  // Add Spans
  argv[2] = Spans;

  // Add Attributes
  /*
   * If memory compression is enabled, we have to check whether the same
   * attribute set is used by some other annotation in order to share the
   * previous one and free this...
   */
  if (CDM_EnableMemoryCompression) {
    Tcl_IncrRefCount(Attributes);
    argv[3] = CDM_SearchForAttributeSet(Attributes);
    Tcl_DecrRefCount(Attributes);
  } else {
    argv[3] = Attributes;
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  if (spans == 1) {
    Tcl_IncrRefCount(argv[2]); Tcl_DecrRefCount(argv[2]);
    Ann = CDM_NewAnnotationObj(-1, argv[1], start, end, argv[3]);
  } else {
    Ann = CDM_NewAnnotationObj(-1, argv[1], argv[2], argv[3]);
  }
#else /* ENABLE_ANNOTATION_OBJECT */
  Ann = Tcl_NewListObj(4, argv);
  if (Ann == NULL)
    CDM_Panic("CDM_CreateAnnotation: Tcl_NewListObj returned NULL "
              "- out of memory?\n");
#endif /* ENABLE_ANNOTATION_OBJECT */
  return Ann;
} /* CDM_CreateAnnotation */


/*
 * ---- CDM_CreateAnnotation:
 *  DESCRIPTION:
 *   <English:
 *    This function will create and return a new Annotation object. The newly
 *    created Annotation object will have an empty Annotation Id and the
 *    Annotation type, Span set and Annotation set will be set to the values
 *    provided by the caller through the corresponding parameters.
 *
 *    This function may or may not use
 *    the provided Span and Attribute set object: this function will internally
 *    increment/decrement the reference count of the provided set
 *    objects. If the reference count of any of the provided set objects
 *    equals 0 (i.e. the Span set object has been created with a call to
 *    CDM_CreateSpanSet or the Attribute set object has been created with
 *    a call to CDM_CreateAttribute), this function may free the object.
 *    As a result, if
 *    the caller wants to ensure that any of the provided set objects is
 *    valid after calling this function, he should increment its reference
 *    count with [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]].
 *
 *    Note that the returned object is not property of the CDM: the caller is
 *    responsible for using
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *    CDM_Free to free the object
 *    and the memory associated with it when the object is not needed anymore.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Type:        The Annotation's Type        (const char *)
 *      start:       The start character offset to be stored in the only Span object that will be created for the Annotation      (long)
 *      end:         The end character offset to be stored in the only Span object that will be created for the Annotation        (long)
 *      Attributes:  A Set of Attributes          (CDM_AttributeSet)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to tip_CreateAnnotation (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will create and return a new Annotation object (of type
 *    CDM_Annotation). In case of an error, NULL will be returned and an error
 *    message describing the error will be left at the current active
 *    Tcl interp (CDM_Interp).
 *
 *    Note that the returned object is not property of the CDM: the caller is
 *    responsible for using
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *    CDM_Free to free the object
 *    and the memory associated with it when the object is not needed anymore.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]],
 *    tip_CreateAnnotation
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation
CDM_CreateAnnotation(const char *Type, const long start, const long end,
                     CDM_AttributeSet Attributes) {
  CDM_Annotation Ann;
  Tcl_Obj *argv[4], **Attr;
  int i, attrs;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_CreateAnnotation: Aborting...");
    return NULL;
  }

  /*
   * Did we get a valid Attribute set?
   */
  if (Attributes == NULL) {
    Attributes = &CDM_Obj_NULL;
  } else {
    if (Tcl_ListObjGetElements(NULL, Attributes, &attrs, &Attr) != TCL_OK) {
      CDM_ReportError("CDM_CreateAnnotation: Invalid Attribute Set object: ");
      CDM_AppendError(Tcl_GetString(Attributes));
      return NULL;
    }
    for (i=0; i<attrs; i++) {
      if (!CDM_ValidAttribute(Attr[i])) {
        return NULL;
      }
    }
  }

  // Add the (Empty) Id
  argv[0] = &CDM_Obj_NULL;

  // Add type
  argv[1] = ObjectCache.FindAnnotationType((char *) Type);
  if (argv[1] == NULL) CDM_Panic("CDM_CreateAnnotation: "
    "ObjectCache::FindAnnotationType returned NULL - out of memory?\n");


  // Add Attributes
  /*
   * If memory compression is enabled, we have to check whether the same
   * attribute set is used by some other annotation in order to share the
   * previous one and free this...
   */
  if (CDM_EnableMemoryCompression) {
    Tcl_IncrRefCount(Attributes);
    argv[3] = CDM_SearchForAttributeSet(Attributes);
    Tcl_DecrRefCount(Attributes);
  } else {
    argv[3] = Attributes;
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  Ann = CDM_NewAnnotationObj(-1, argv[1], start, end, argv[3]);
#else /* ENABLE_ANNOTATION_OBJECT */
  /* Add Span... */
  argv[2] = CDM_CreateSpanSet(start, end);
  if (argv[2] == NULL) return NULL;
  Ann = Tcl_NewListObj(4, argv);
  if (Ann == NULL)
    CDM_Panic("CDM_CreateAnnotation: Tcl_NewListObj returned NULL "
              "- out of memory?\n");
#endif /* ENABLE_ANNOTATION_OBJECT */
  return Ann;
} /* CDM_CreateAnnotation */

/*
 * ---- CDM_DisplaceAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function displases or "moves" the specified Annotation by
 *     "displacement" characters. This function will iterate over
 *     all spans contained in the Annotation. For each span all the offsets
 *     contained in it (either start or end) will be examined: if any of this
 *     offsets is equal or greater than the value specified through the "offset"
 *     parameter, then the value of the
 *     "displacement" parameter will be added to it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Annotation:   The Annotation to be displaced             (CDM_Annotation)
 *     offset:    A character offset after which all offsets are modified (long)
 *     displacement: The number of characters that offsets will be moved  (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_DisplaceAnnotation (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a new Annotation object, if
 *     any displacement happened. In this case, the user is responsible to
 *     free the returned Annotation (through Tcl_DecrRefCount of CDM_Free)
 *     when it is not needed any more.
 *
 *     If no displacement occured, then the specified Annotation through the
 *     "Annotation" parameter will be returned.
 *
 *     Finally, in case of an error, NULL will be returned and a message will
 *     be left at the current interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_DisplaceAnnotations, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_DisplaceAnnotation
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_DisplaceAnnotation(CDM_Annotation Annotation, long offset,
                                      long displacement) {
  CDM_THREAD_SPECIFIC_DATA;
   /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_DisplaceAnnotation: Aborting...");
    return NULL;
  }

  /*
   * Is our input valid?
   */
  if (!CDM_ValidAnnotation(Annotation)) return NULL;

  /*
   * If displacement is 0, there is nothing to do...
   */
  if (displacement == 0) return Annotation;

  return CDM_InternalDisplaceAnnotation(Annotation, offset, displacement);
}

static
CDM_Annotation CDM_InternalDisplaceAnnotation(CDM_Annotation Annotation,
                                              long offset, long displacement) {
  CDM_SpanSet Spans, NewSpans;
  Tcl_Obj **Span, *Obj;
  long start, end;
  int status, items, i, displaced = 0;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * If displacement is 0, there is nothing to do...
   */
  if (displacement == 0) return Annotation;

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Annotation) && CDMAnnObj_GetStart(Annotation) != -1) {
    start = CDMAnnObj_GetStart(Annotation);
    end   = CDMAnnObj_GetEnd(Annotation);
    if (start < offset && end < offset) return Annotation;
    /* Do the displacement... */
    if (start >= offset) start += displacement;
    if ((end > offset) ||
        ((start == end) && (end >= offset))) end += displacement;
    /* Do not allow offsets to get out of bounds. Unfortunatelly, we cannot
     * check validity against the maximum length of the Document text... */
    if (start < 0 || end < 0) {
      return NULL;
    }
    /* Create and append a new span... */
    Annotation = Tcl_DuplicateObj(Annotation);
    CDMAnnObj_SetStart(Annotation, start);
    CDMAnnObj_SetEnd(Annotation, end);
    return Annotation;
  }
#endif /* ENABLE_ANNOTATION_OBJECT */

  /*
   * Get the Spans from the provided Annotation...
   */
  Spans = CDM_GetSpans(Annotation);
  if (Spans == NULL) return NULL;

  /*
   * Iterate over all spans...
   */
  status = Tcl_ListObjGetElements(CDM_Interp, Spans, &items, &Span);
  if (status != TCL_OK) {
    CDM_Free(Spans);
    return NULL;
  }

  NewSpans = CDM_CreateSpanSet();
  if (NewSpans == NULL) return NULL;
  for (i=0; i<items; i++) {
    status = CDM_GetSpanOffsets(Span[i], &start, &end);
    if (status != TCL_OK) {
      CDM_Free(NewSpans);
      return NULL;
    }
    if (start < offset && end < offset) {
      Tcl_ListObjAppendElement(CDM_Interp, NewSpans, Span[i]);
    } else {
      /* Do the displacement... */
      if (start >= offset) start += displacement;
      if ((end > offset) ||
          ((start == end) && (end >= offset))) end += displacement;
      /* Do not allow offsets to get out of bounds. Unfortunatelly, we cannot
       * check validity against the maximum length of the Document text... */
      if (start < 0 || end < 0) {
        CDM_Free(NewSpans);
        return NULL;
      }
      /* Create and append a new span... */
      Obj = CDM_CreateSpan(start, end);
      if (Obj == NULL) {
        CDM_Free(NewSpans);
        return NULL;
      }
      Tcl_ListObjAppendElement(CDM_Interp, NewSpans, Obj);
      displaced = 1;
    }
  }

  /*
   * We return a new Annotation, if a displacement has occured...
   */
  if (displaced) {
    Annotation = Tcl_DuplicateObj(Annotation);
    if (Annotation == NULL) {
      CDM_Free(NewSpans);
      return NULL;
    }
#ifdef ENABLE_ANNOTATION_OBJECT
    if (ANNOTATION_OBJ(Annotation)) {
      CDMAnnObj_SetSpans(Annotation, NewSpans);
    } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
      status = Tcl_ListObjReplace(CDM_Interp, Annotation, 2, 1, 1, &NewSpans);
      if (status != TCL_OK) {
        CDM_Free(NewSpans);
        CDM_Free(Annotation);
        return NULL;
      }
#ifdef ENABLE_ANNOTATION_OBJECT
      Tcl_ConvertToType(CDM_Interp, Annotation, &cdmAnnotationType);
    }
#endif /* ENABLE_ANNOTATION_OBJECT */
  } else {
    Tcl_DecrRefCount(NewSpans);
  }
  return Annotation;
} /* CDM_InternalDisplaceAnnotation */

/*
 * ---- CDM_GetAnnotatedTextRanges:
 *  DESCRIPTION:
 *   <English:
 *    This function will return a set of strings from the provided Text object
 *    that are annotated by the provided Annotation. In simple words, this
 *    function will examine the span set contained in the provided Annotation
 *    and return in a list all text ranges (from the specified text object)
 *    indexed by each one of the spans.
 *
 *    CDM does not own the returned object: the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *    associated with it.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. (In this case an empty
 *    object will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Text:       The text object indexed by the Annotation. (CDM_ByteSequence)
 *    Annotation: The Annotation that indexes the text ranges. (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *    Note that usually the provided text object is the text of a Document,
 *    retrieved through calls to either CDM_GetByteSequence or CDM_GetRawData.
 *    The text object is expected to contain text in Unicode or UTF-8
 *    representation.
 *
 *    This function is equivelant to tip_GetAnnotatedTextRanges (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a set of ByteSequence/RawData objects (of type
 *    CDM_ByteSequence or CDM_RawData) containing the text ranges (or
 *    portions) from the "Text" object indexed by the "Annotation" spans.
 *    In order to retrieve the individual text ranges, the user can call
 *    CDM_Nth as usual with set objects.
 *
 *    Note that CDM does not own the returned object:
 *    the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *    associated with it.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. (In this case an empty
 *    object will be returned.)
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_GetByteSequence, CDM_GetRawData, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], CDM_Free,
 *    CDM_GetFirstAnnotatedTextRange, tip_GetAnnotatedTextRanges
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequenceSet CDM_GetAnnotatedTextRanges(CDM_ByteSequence Text,
                                            CDM_Annotation Annotation) {
  CDM_SpanSet SpanSet;
  CDM_ByteSequenceSet TextSet;
  CDM_ByteSequence range;
  Tcl_Obj **elements;
  long start, end;
  int result, items, i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetAnnotatedTextRanges: Aborting...");
    return NULL;
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Annotation) && CDMAnnObj_GetStart(Annotation) != -1) {
    Tcl_Obj *obj;
    start = CDMAnnObj_GetStart(Annotation);
    end   = CDMAnnObj_GetEnd(Annotation);
    obj   = Tcl_GetRange(Text, (int) start, (int) end - 1);
    return Tcl_NewListObj(1, &obj);
  }
#endif /* ENABLE_ANNOTATION_OBJECT */

  /*
   * Get the SpanSet from the Annotation...
   */
  SpanSet = CDM_GetSpans(Annotation);
  if (SpanSet == NULL) return NULL;

  /*
   * Iterate over spans...
   */
  result = Tcl_ListObjGetElements(CDM_Interp, SpanSet, &items, &elements);
  if (result != TCL_OK) {
    CDM_ReportError("Malformed SpanSet \"%s\"", Tcl_GetString(SpanSet));
    return NULL;
  }

  TextSet = Tcl_NewListObj(0, NULL);
  for (i=0; i<items; i++) {
    if (CDM_GetSpanOffsets(elements[i], &start, &end) != TCL_OK) {
      CDM_ReportError("Malformed SpanSet \"%s\"", Tcl_GetString(SpanSet));
      return NULL;
    }
    range = Tcl_GetRange(Text, (int) start, (int) end - 1);
    if (range == NULL ||
        Tcl_ListObjAppendElement(CDM_Interp, TextSet, range) != TCL_OK) {
      CDM_Free(TextSet);
      return NULL;
    }
  }
  return TextSet;
} /* CDM_GetAnnotatedTextRanges */

/*
 * ---- CDM_GetFirstAnnotatedTextRange:
 *  DESCRIPTION:
 *   <English:
 *    This function will return a the text range from the provided Text object
 *    that is annotated by the first span of the provided Annotation.
 *    In simple words, this function will retrieve the first span from the
 *    provided Annotation and return the text range (from the specified text
 *    object) that is indexed by this span.
 *
 *    CDM does not own the returned object: the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *    associated with it.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. (In this case an empty
 *    object will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Text:       The text object indexed by the Annotation. (CDM_ByteSequence)
 *     Annotation: The Annotation that indexes the text range. (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *    Note that usually the provided text object is the text of a Document,
 *    retrieved through calls to either CDM_GetByteSequence or CDM_GetRawData.
 *    The text object is expected to contain text in Unicode or UTF-8
 *    representation.
 *
 *    This function is equivelant to tip_GetFirstAnnotatedTextRange (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a ByteSequence/RawData object (of type
 *    CDM_ByteSequence or CDM_RawData) containing the text range (or
 *    portion) from the "Text" object indexed by the first span of the
 *    "Annotation" parameter.
 *
 *    Note that CDM does not own the returned object:
 *    the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *    associated with it.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. In this case an empty
 *    object (i.e. one containing the string "") will be returned.)
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_GetByteSequence, CDM_GetRawData, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], CDM_Free,
 *    CDM_GetAnnotatedTextRanges, tip_GetFirstAnnotatedTextRange
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequence CDM_GetFirstAnnotatedTextRange(CDM_ByteSequence Text,
                                            CDM_Annotation Annotation) {
  CDM_SpanSet SpanSet;
  Tcl_Obj **elements;
  long start, end;
  int result, items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetFirstAnnotatedTextRange: Aborting...");
    return NULL;
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Annotation) && CDMAnnObj_GetStart(Annotation) != -1) {
    start = CDMAnnObj_GetStart(Annotation);
    end   = CDMAnnObj_GetEnd(Annotation);
    return Tcl_GetRange(Text, (int) start, (int) end - 1);
  }
#endif /* ENABLE_ANNOTATION_OBJECT */

  /*
   * Get the SpanSet from the Annotation...
   */
  SpanSet = CDM_GetSpans(Annotation);
  if (SpanSet == NULL) return NULL;

  /*
   * Iterate over spans...
   */
  result = Tcl_ListObjGetElements(CDM_Interp, SpanSet, &items, &elements);
  if (result != TCL_OK) {
    CDM_ReportError("Malformed SpanSet \"%s\"", Tcl_GetString(SpanSet));
    return NULL;
  }
  if (items == 0) {
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }

  if (CDM_GetSpanOffsets(elements[0], &start, &end) != TCL_OK) return NULL;
  return Tcl_GetRange(Text, (int) start, (int) end - 1);
} /* CDM_GetFirstAnnotatedTextRange */

/*
 * ---- CDM_GetAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function will search all the Attributes that the specified
 *     Annotation object has, for an Attribute that has the same name as the
 *    "Name" parameter. If such an Attribute is found, the Attribute object is
 *     return. If an Attribute with the specified name does not exist, NULL is
 *     returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Annotation:   The Annotation to be searched  (CDM_Annotation)
 *     Name:         The Attribute's Name           (const char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetAttribute (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Attribute object (of type CDM_Attribute) if an
 *     Attribute named as specified is found. In case that the specified
 *     Attribute does not exist, NULL is returned.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or directly modified in any way by the caller. Subsequent calls
 *     of either the C++ or Tcl API can modify the returned object. If the
 *     caller wants to keep a reference to the current state of the returned
 *     object, he must place such a
 *     request by using Tcl_IncrRefCount. When the caller wants to release the
 *     object (and return it to CDM), the functions Tcl_DecrRefCount of
 *     CDM_Free must be used.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]],
 *     tip_GetAttribute
 *   >:&:
 */
CDM_EXPORT
CDM_Attribute CDM_GetAttribute(CDM_Annotation Annotation, const char *Name) {
  Tcl_Obj *obj, **Attributes, **elements;
  int result, items, attr_items, i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetAttribute: Aborting...");
    return NULL;
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Annotation)) {
    obj = CDMAnnObj_GetAttributes(Annotation);
  } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
    result = Tcl_ListObjGetElements(CDM_Interp, Annotation, &items, &elements);
    if (result != TCL_OK || items != 4) {
      CDM_ReportError("CDM_GetAttribute: Malformed Annotation \"%s\"",
                      Tcl_GetString(Annotation));
      return NULL;
    }
    obj = elements[3];
#ifdef ENABLE_ANNOTATION_OBJECT
  }
#endif /* ENABLE_ANNOTATION_OBJECT */

  /* Get Attributes from Annotation */
  result = Tcl_ListObjGetElements(CDM_Interp, obj, &items, &Attributes);
  if (result != TCL_OK) {
    CDM_ReportError("CDM_GetAttribute: Malformed Attribute Set \"%s\"",
                    Tcl_GetString(obj));
    return NULL;
  }
  for (i=0; i<items; i++) {
    result = Tcl_ListObjGetElements(CDM_Interp, Attributes[i], &attr_items,
                                    &elements);
    if (result != TCL_OK || attr_items != 2) {
      CDM_ReportError("CDM_GetAttribute: Malformed Attribute \"%s\"",
                      Tcl_GetString(Attributes[i]));
      return NULL;
    }
    if (CDM_Strcmp(Name, Tcl_GetString(elements[0])) == 0) return Attributes[i];
  }
  return NULL;
} /* CDM_GetAttribute */

/*
 * ---- CDM_GetAttributes:
 *  DESCRIPTION:
 *   <English:
 *     This function returns a new object (of type CDM_AttributeSet) that
 *     contains all the Attributes that the specified Annotation currently
 *     has.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or directly modified in any way by the caller. Subsequent calls
 *     of either the C++ or Tcl API can modify the returned object. If the
 *     caller wants to keep a reference to the current state of the returned
 *     object, he must place such a
 *     request by using Tcl_IncrRefCount. When the caller wants to release the
 *     object (and return it to CDM), the functions Tcl_DecrRefCount of
 *     CDM_Free must be used.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Annotation:   An Annotation object         (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetAttributes (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Attribute set object (of type CDM_AttributeSet),
 *     that contains all the Attribute objects held by the provided Annotation
 *     object.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or directly modified in any way by the caller. Subsequent calls
 *     of either the C++ or Tcl API can modify the returned object. If the
 *     caller wants to keep a reference to the current state of the returned
 *     object, he must place such a
 *     request by using Tcl_IncrRefCount. When the caller wants to release the
 *     object (and return it to CDM), the functions Tcl_DecrRefCount of
 *     CDM_Free must be used.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]],
 *     tip_GetAttributes
 *   >:&:
 */
CDM_EXPORT
CDM_AttributeSet CDM_GetAttributes(CDM_Annotation Annotation) {
  Tcl_Obj **Attributes;
  int result, items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetAttributes: Aborting...");
    return NULL;
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Annotation)) {
    return CDMAnnObj_GetAttributes(Annotation);
  }
#endif /* ENABLE_ANNOTATION_OBJECT */

  result = Tcl_ListObjGetElements(CDM_Interp, Annotation, &items, &Attributes);
  if (result != TCL_OK || items != 4) {
    CDM_ReportError("Malformed Annotation \"%s\"", Tcl_GetString(Annotation));
    return NULL;
  }
  return Attributes[3];
} /* CDM_GetAttribute */

/*
 * ---- CDM_GetFirstSpan:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the first span range of the span list
 *     contained in the specified Annotation.
 *     Note that the returned object is owned by the CDM. The caller is not
 *     supposed to free or modify the returned object.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Annotation:  The Annotation object.              (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to the following C++ code:
 *     [[br]][[code]][[center]]CDM_Nth(CDM_GetSpans(CDM_Annotation ann), 0);
 *     [[/center]][[/code]]
 *     This function is also equivelant to tip_GetFirstSpan (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Span object (of type CDM_Span).
 *
 *     If an error occurs, then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetSpans, CDM_Nth, tip_GetFirstSpan
 *   >:&:
 */
CDM_EXPORT
CDM_Span CDM_GetFirstSpan(CDM_Annotation Annotation) {
  CDM_SpanSet Spans;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetFirstSpan: Aborting...");
    return NULL;
  }

  if (Annotation == NULL) {
    CDM_ReportError("CDM_GetFirstSpan: invalid Annotation object (NULL)");
    return NULL;
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Annotation)) {
    long start = CDMAnnObj_GetStart(Annotation);
    if (start != -1) {
      /* Is a proper span set already stored? */
      CDM_SpanSet spans = CDMAnnObj_GetSpans(Annotation);
      if (spans) {
        return CDM_Nth(spans, 0);
      } else {
        long end = CDMAnnObj_GetEnd(Annotation);
        spans = CDM_CreateSpanSet(start, end);
        if (spans == NULL) return NULL;
        CDMAnnObj_KeepSpanSet(Annotation, spans);
      }
    }
    return CDM_Nth(CDMAnnObj_GetSpans(Annotation), 0);
  }
#endif /* ENABLE_ANNOTATION_OBJECT */

  Spans = CDM_GetSpans(Annotation);
  if (Spans == NULL) return NULL;
  return CDM_Nth(Spans, 0);
} /* CDM_GetFirstSpan */

/*
 * ---- CDM_GetFirstSpanOffsets:
 *  DESCRIPTION:
 *   <English:
 *    This function will place the values of both span start and end of
 *    the Annotation's first span in the two specified long integer pointers.
 *    Each span object is a set of exactly two integers, named as
 *    "start" and "end".
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Annotation:  An Annotation object  (CDM_Annotation)
 *      start:  A pointer pointing to a long, where the start value of the specified span will be placed. (long *)
 *      end:    A pointer pointing to a long, where the end value of the specified span will be placed. (long *)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to the following C++ code:
 *    [[br]][[code]][[center]]
 *    span = CDM_GetFirstSpan(Annotation);[[br]]
 *    *start = CDM_GetStart(span);[[br]]
 *    *end   = CDM_GetEnd(Span);
 *    [[/center]][[/code]]
 *    This function is equivelant to tip_GetFirstSpanOffsets (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will place the two Offsets contained in the first span
 *    object of the provided Annotation (the start and end of the span)
 *    into the two long variables
 *    pointed by the two provided pointers.
 *
 *    The return value from this function will be a standart Tcl completion
 *    code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *    error occurs, the return value will be TCL_ERROR and an error message
 *    describing the error will be placed at the current Tcl interpreter
 *    (CDM_Interp). Else, TCL_OK will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_GetFirstSpan, CDM_GetStart, CDM_GetEnd, tip_GetFirstSpanOffsets
 *   >:&:
 */
CDM_EXPORT
int CDM_GetFirstSpanOffsets(CDM_Annotation Annotation, long *start, long *end) {
  CDM_Span Span;
  Tcl_Obj **elements;
  int result, items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetFirstSpanOffsets: Aborting...");
    return TCL_ERROR;
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Annotation) && CDMAnnObj_GetStart(Annotation) != -1) {
    *start = CDMAnnObj_GetStart(Annotation);
    *end   = CDMAnnObj_GetEnd(Annotation);
    if (*end >= 0) return TCL_OK;
  }
#endif /* ENABLE_ANNOTATION_OBJECT */
  Span = CDM_GetFirstSpan(Annotation);
  if (Span ==NULL) return TCL_ERROR;
  result = Tcl_ListObjGetElements(CDM_Interp, Span, &items, &elements);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("CDM_GetFirstSpanOffsets: Malformed Span \"%s\"",
                    Tcl_GetString(Span));
    return TCL_ERROR;
  }

  if (start != NULL) {
    result = Tcl_GetLongFromObj(CDM_Interp, elements[0], start);
    if (result != TCL_OK) return TCL_ERROR;
    if (*start < 0) {
      return CDM_ReportError("Negative value %ld detected "
             "in start position of Span \"%s\"", *start, Tcl_GetString(Span));
    }
  }
  if (end != NULL) {
    result = Tcl_GetLongFromObj(CDM_Interp, elements[1], end);
    if (result != TCL_OK) return TCL_ERROR;
    if (*end < 0) {
      return CDM_ReportError("Negative value %ld detected "
             "in end position of Span \"%s\"", *end, Tcl_GetString(Span));
    }
  }
  return TCL_OK;
} /* CDM_GetFirstSpanOffsets */

/*
 * ---- CDM_GetSpans:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the span list
 *     contained in the specified Annotation.
 *     Note that the returned object is owned by the CDM. The caller is not
 *     supposed to free or modify the returned object.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Annnotation:  The Annotation object.            (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetSpans (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Span Set object (of type CDM_SpanSet).
 *
 *     If an error occurs, then NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetFirstSpan, tip_GetSpans
 *   >:&:
 */
CDM_EXPORT
CDM_SpanSet CDM_GetSpans(CDM_Annotation Annotation) {
  Tcl_Obj **elements;
  int result, items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetSpans: Aborting...");
    return NULL;
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Annotation)) {
    long start = CDMAnnObj_GetStart(Annotation);
    if (start != -1) {
      /* Is a proper span set already stored? */
      CDM_SpanSet spans = CDMAnnObj_GetSpans(Annotation);
      if (spans) {
        return spans;
      } else {
        long end = CDMAnnObj_GetEnd(Annotation);
        spans = CDM_CreateSpanSet(start, end);
        if (spans == NULL) return NULL;
        CDMAnnObj_KeepSpanSet(Annotation, spans);
      }
    }
    return CDMAnnObj_GetSpans(Annotation);
  }
#endif /* ENABLE_ANNOTATION_OBJECT */

  result = Tcl_ListObjGetElements(CDM_Interp, Annotation, &items, &elements);
  if (result != TCL_OK || items != 4) {
    CDM_ReportError("Malformed Annotation \"%s\"", Tcl_GetString(Annotation));
    return NULL;
  }
  return elements[2];
} /* CDM_GetSpans */

/*
 * ---- CDM_GetType:
 *  DESCRIPTION:
 *   <English:
 *    This function returns the type of the specified Annotation object.
 *    The returned value will be encoded using the UTF-8 encoding.
 *
 *    The return value will be of type char* and will be owned by CDM.
 *    As a result the user must not free or modify the returned value.
 *
 *    Note that the return value will be invalidated if the user either
 *    destroy the Annotation or change its type.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Annotation: The Annotation whose type is to be returned. (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *    Note that the return value will be invalidated if the user either
 *    destroy the Annotation or change its type. If this value has to be kept
 *    for prolonged period of time, the caller should create a copy of the
 *    returned value with CDM_StrCpy or equivalent.
 *
 *    This function is equivelant to tip_GetType (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value of this function will be a string (of type char *),
 *    representing the type of the specified Annotation object.
 *    In case of an error, NULL will be returned and an error will be left at
 *    the current active interpreter (CDM_Interp).
 *
 *    Note that the returned value is owned by CDM:
 *    the user must not modify the contents or free the returned value.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_GetTypeObj, CDM_StrCpy, tip_GetType
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetType(CDM_Annotation Annotation) {
  Tcl_Obj **elements;
  int items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetType: Aborting...");
    return NULL;
  }

  if (Annotation == NULL) return NULL;

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Annotation)) {
    return Tcl_GetString(CDMAnnObj_GetType(Annotation));
  }
#endif /* ENABLE_ANNOTATION_OBJECT */

  if (!CDM_ValidAnnotation(Annotation)) return NULL;
  Tcl_ListObjGetElements(CDM_Interp, Annotation, &items, &elements);
  return Tcl_GetString(elements[1]);
} /* CDM_GetType */

/*
 * ---- CDM_GetTypeObj:
 *  DESCRIPTION:
 *   <English:
 *    This function returns the type of the specified Annotation object.
 *    The returned value will be encoded using the UTF-8 encoding, as a valid
 *    object of type CDM_ByteSequence.
 *
 *    The return value will be of type CDM_ByteSequence and will be owned by
 *    CDM.
 *    As a result the user must not free or modify the returned object.
 *
 *    Note that the return value will be invalidated if the user either
 *    destroy the Annotation or change its type.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Annotation: The Annotation whose type is to be returned. (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *    Note that the return value will be invalidated if the user either
 *    destroy the Annotation or change its type. If this value has to be kept
 *    for prolonged period of time, the caller should create a copy of the
 *    returned value with Tcl_DuplicateObj.
 *
 *    This function is equivelant to tip_GetTypeObj (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value of this function will be a text object
 *    (of type CDM_ByteSequence or equivalently CDM_RawData),
 *    representing the type of the specified Annotation object.
 *
 *    In case of an error, NULL will be returned and an error will be left at
 *    the current active interpreter (CDM_Interp).
 *
 *    Note that the returned value is owned by CDM:
 *    the user must not modify the contents or free the returned object.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_GetType,
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DuplicateObj[[/A]],
 *    tip_GetTypeObj
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequence CDM_GetTypeObj(CDM_Annotation Annotation) {
  Tcl_Obj **elements;
  int items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetTypeObj: Aborting...");
    return NULL;
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Annotation)) {
    return CDMAnnObj_GetType(Annotation);
  }
#endif /* ENABLE_ANNOTATION_OBJECT */

  if (!CDM_ValidAnnotation(Annotation)) return NULL;
  Tcl_ListObjGetElements(CDM_Interp, Annotation, &items, &elements);
  return elements[1];
} /* CDM_GetTypeObj */

/*
 * ---- CDM_GetId:
 *  DESCRIPTION:
 *   <English:
 *    This function will return the Id of the specified Annotation. The
 *    Id of an Annotation is a unique integer (of type long) that
 *    unambigiously specifies an Annotation object inside a Document. CDM
 *    ensures that each Annotation of the Annotations contained in a Document
 *    will have a unique Id.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Ann: The Annotation whose Id is to be returned (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to tip_GetId (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns an integer (of type long) that is the Id of the
 *    specified Annotation object.
 *
 *    In case of an error, -1 will be returned and an error will be left at
 *    the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_GetType, tip_GetId
 *   >:&:
 */
CDM_EXPORT
long CDM_GetId(CDM_Annotation Annotation) {
  Tcl_Obj **elements;
  int items;
  long Id;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetId: Aborting...");
    return -1;
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Annotation)) {
    return CDMAnnObj_GetId(Annotation);
  }
#endif /* ENABLE_ANNOTATION_OBJECT */

  if (!CDM_ValidAnnotation(Annotation)) return -1;
  Tcl_ListObjGetElements(CDM_Interp, Annotation, &items, &elements);
  if (Tcl_GetCharLength(elements[0]) == 0) return -1;
  if (Tcl_GetLongFromObj(CDM_Interp, elements[0], &Id) != TCL_OK) {
    CDM_ReportError("Malformed Annotation: Invalid Id \"%s\"",
                    Tcl_GetString(elements[0]));
    return -1;
  }
  return Id;
}

/*
 * ---- CDM_PutAttribute:
 *  DESCRIPTION:
 *   <English:
 *    This function will add a given Attribute to the specified Annotation
 *    object. If an Attribute with the same name already exists in the
 *    Annotation, then it will be overwritten by the new Attribute. Else, the
 *    new Attribute will be appended to the existing Attribute set of the
 *    specified Annotation.
 *
 *    The returned value of this function may be a new Annotation
 *    object or the provided one. What the result will be depends on whether
 *    the provided Annotation is a shared object or not. If the provided
 *    Annotation is a shared object (i.e.
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IsShared[[/A]] returns true)
 *    a new Annotation object will be created and returned, since a shared
 *    object cannot be modified. If the provided Annotation is not a shared
 *    object, then the required changes will be done in place, and the same
 *    Annotation object will be returned.
 *
 *    As a result, the caller should be always check the returned value, if he
 *    is the owner but does not want to keep the initial Annotation object.
 *    A typical use is shown in the following code:
 *
 *    [[pre]][[code]]
 *    ann = CDM_CreateAnnotation("token", spanset, attrset);
 *    new_ann = CDM_PutAttribute(ann, Attr);
 *    if (new_ann != ann) {
 *      CDM_Free(ann);
 *      ann = new_ann;
 *    }
 *    CDM_AddAnnotation(Doc, ann); // Now CDM is the owner, we can forget about it...
 *    [[/code]][[/pre]]
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Ann:         Annotation Pointer.            (CDM_Annotation)
 *      Attribute:   Attribute to be added          (CDM_Atrribute)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_PutAttribute (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return an Annotation object (of type CDM_Annotation).
 *    The returned value of this function may be a new Annotation
 *    object or the provided one. What the result will be depends on whether
 *    the provided Annotation is a shared object or not. If the provided
 *    Annotation is a shared object (i.e.
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IsShared[[/A]] returns true)
 *    a new Annotation object will be created and returned, since a shared
 *    object cannot be modified. If the provided Annotation is not a shared
 *    object, then the required changes will be done in place, and the same
 *    Annotation object will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_GetAttribute, CDM_GetAttributes, tip_PutAttribute
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IsShared[[/A]],
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]]
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_PutAttribute(CDM_Annotation Ann, CDM_Attribute Attr) {
  Tcl_Obj *AttributeSet, *NewAttributeSet, **Attributes;
  char *AttrName, *Name;
  int result, items, i, replaced = 0;
  CDM_THREAD_SPECIFIC_DATA;

  CDM_DEBUG("CDM_PutAttribute - Adding Attribute to Annotation...\n");
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_PutAttribute: Aborting...");
    return NULL;
  }

  if (!CDM_ValidAnnotation(Ann)) return NULL;
  if (!CDM_ValidAttribute(Attr)) return NULL;
  if (Tcl_IsShared(Ann)) {
#ifdef DUPLICATE_DEBUG
    CDM_Debug("PutAttribute: Duplicating (%d): %s\n", Ann->refCount,
            Tcl_GetString(Ann));
#endif /* DUPLICATE_DEBUG */
    Ann = Tcl_DuplicateObj(Ann);
  }
  if (Ann == NULL) CDM_Panic("CDM_PutAttribute: Tcl_DuplicateObj returned NULL "
                             "- out of memory?\n");
  AttrName = CDM_GetName(Attr);

  /*
   * In order to put the new attribute in the annotation, we should check
   * whether an attribute with the same name already exists. If we found such
   * a case, change its value. Else, just add the new attribute...
   */

  /*
   * Get Attribute Set from Annotation...
   */
  AttributeSet = CDM_GetAttributes(Ann);
  if (AttributeSet == NULL) {
    CDM_ReportError("CDM_PutAttribute: Malformed Annotation \"%s\"",
                    Tcl_GetString(Ann));
    return NULL;
  }
  /* If we try to modify a shared object, we loose! */
  if (Tcl_IsShared(AttributeSet)) {
#ifdef DUPLICATE_DEBUG
    CDM_Debug("PutAttribute: Duplicating (%d): %s\n", AttributeSet->refCount,
           Tcl_GetString(AttributeSet));
#endif /* DUPLICATE_DEBUG */
    AttributeSet = Tcl_DuplicateObj(AttributeSet);
  }
  if (AttributeSet == NULL) {
    CDM_Panic("CDM_PutAttribute: Tcl_DuplicateObj returned NULL - "
              "out of memory?\n");
  }

  /*
   *  Get Attributes from Attribute Set...
   */
  result = Tcl_ListObjGetElements(CDM_Interp, AttributeSet,
                                     &items, &Attributes);
  if (result != TCL_OK ) {
    CDM_ReportError("CDM_PutAttribute: Malformed Attribute Set \"%s\"",
                    Tcl_GetString(AttributeSet));
    return NULL;
  }

  /*
   * Search all Attributes for one with the same name. If we found one, then
   * we should replace its value...
   */
  for (i = 0; i < items; i++) {
    Name = CDM_GetName(Attributes[i]);
    if (Name == NULL) {
      CDM_ReportError("Malformed Attribute \"%s\"",
                      Tcl_GetString(Attributes[i]));
      return NULL;
    }
    if (CDM_Strcmp(Name, AttrName) == 0) {
      /*
       * We have found an existing attribute with the same name. Replace it...
       */
      CDM_DEBUG("Replaced older value...\n");
      result = Tcl_ListObjReplace(CDM_Interp, AttributeSet, i, 1, 1, &Attr);
      if (result != TCL_OK) return NULL;
      replaced = 1;
      break;
    }
  }

  if (!replaced) {
    /*
     * There is no attribute with same name as the attribute to be added. Just
     * append the new attribute to the end of annotation's attribute set...
     */
    if (Tcl_ListObjAppendElement(CDM_Interp, AttributeSet, Attr) != TCL_OK) {
      return NULL;
    }
  }

  /*
   * Now, the updated attribute set is in the variable "AttributeSet".
   * If memory compression is enabled, we have to check whether the same
   * attribute set is used by some other annotation in order to share the
   * previous one and free this...
   */
  if (CDM_EnableMemoryCompression) {
    NewAttributeSet = CDM_SearchForAttributeSet(AttributeSet);
    if (NewAttributeSet && NewAttributeSet != AttributeSet) {
      Tcl_DecrRefCount(AttributeSet);
      AttributeSet = NewAttributeSet;
      Tcl_IncrRefCount(AttributeSet);
    }
  }

  /*
   *    SEVERE BUG: by Petasis George, 28/01/2000 (missed by purify...)
   *  -------------------------------------------------------------------
   *  If the given annotation is not shared, then we append to the existing
   *  Attribute set. This set has a reference count of 0 or 1. So when it
   *  comes to replace the attribute set, the Attribute set is first freed and
   *  then appended to the Annotation...
   *
   *    Fix:
   *  -------------------------------------------------------------------
   *  Just a simple Tcl_Incr/DecrRefCount(AttributeSet) will do the trick...
   */
   Tcl_IncrRefCount(AttributeSet);
#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Ann)) {
    CDMAnnObj_SetAttributes(Ann, AttributeSet);
  } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
    if (Tcl_ListObjReplace(CDM_Interp, Ann, 3, 1, 1, &AttributeSet) != TCL_OK) {
      return NULL;
    }
#ifdef ENABLE_ANNOTATION_OBJECT
  }
#endif /* ENABLE_ANNOTATION_OBJECT */
  Tcl_DecrRefCount(AttributeSet);
  CDM_DEBUG("CDM_PutAttribute Done...\n");
  return Ann;
} /* CDM_PutAttribute */

/*
 * ---- CDM_RemoveAttribute:
 *  DESCRIPTION:
 *   <English:
 *    This function will remove the Attribute having as name the value of the
 *    "Name" parameter from the specified Annotation, if such an Attribute
 *    exists. The returned value of this function may be a new Annotation
 *    object or the provided one. What the result will be depends on whether
 *    the provided Annotation is a shared object or not. If the provided
 *    Annotation is a shared object (i.e.
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IsShared[[/A]] returns true)
 *    a new Annotation object will be created and returned, since a shared
 *    object cannot be modified. If the provided Annotation is not a shared
 *    object, then the required changes will be done in place, and the same
 *    Annotation object will be returned.
 *
 *    As a result, the caller should be always check the returned value, if he
 *    is the owner but does not want to keep the initial Annotation object.
 *    A typical use is shown in the following code:
 *
 *    [[pre]][[code]]
 *    ann = CDM_CreateAnnotation("token", spanset, attrset);
 *    new_ann = CDM_RemoveAttribute(ann, "pos");
 *    if (new_ann != ann) {
 *      CDM_Free(ann);
 *      ann = new_ann;
 *    }
 *    CDM_AddAnnotation(Doc, ann); // Now CDM is the owner, we can forget about it...
 *    [[/code]][[/pre]]
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Ann:    Annotation's Pointer                   (CDM_Annotation)
 *     Name:   Attribute's name to be removed         (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_RemoveAttribute (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return an Annotation object (of type CDM_Annotation).
 *    The returned value of this function may be a new Annotation
 *    object or the provided one. What the result will be depends on whether
 *    the provided Annotation is a shared object or not. If the provided
 *    Annotation is a shared object (i.e.
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IsShared[[/A]] returns true)
 *    a new Annotation object will be created and returned, since a shared
 *    object cannot be modified. If the provided Annotation is not a shared
 *    object, then the required changes will be done in place, and the same
 *    Annotation object will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Free, tip_RemoveAttribute,
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IsShared[[/A]],
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]]
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_RemoveAttribute(CDM_Annotation Annotation, char *Name) {
  Tcl_Obj **Attributes, **elements;
  CDM_Annotation Ann = Annotation;
  CDM_AttributeSet AttrSet;
  int result, items, attr_items, i;
  CDM_THREAD_SPECIFIC_DATA;

  CDM_DEBUG("CDM_RemoveAttribute - Removing Attribute \"%s\"...\n", Name);
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_RemoveAttribute: Aborting...");
    return NULL;
  }
  if (!CDM_ValidAnnotation(Ann)) return NULL;
  if (Tcl_IsShared(Ann)) {
#ifdef DUPLICATE_DEBUG
    CDM_Debug("RemoveAttribute: Duplicating (%d): %s\n", Ann->refCount,
            Tcl_GetString(Ann));
#endif /* DUPLICATE_DEBUG */
    Ann = Tcl_DuplicateObj(Ann);
  }
  if (Ann == NULL) {
    CDM_Panic("CDM_RemoveAttribute: Tcl_DuplicateObj returned NULL - "
              "out of memory?\n");
  }
  AttrSet = CDM_GetAttributes(Ann);
  if (AttrSet == NULL) {
    CDM_ReportError("CDM_RemoveAttribute: Malformed Annotation \"%s\"",
                    Tcl_GetString(Ann));
    return NULL;
  }
  // Get Attributes from Attribute Set...
  result = Tcl_ListObjGetElements(CDM_Interp, AttrSet, &items, &Attributes);
  if (result != TCL_OK) {
    CDM_ReportError("CDM_RemoveAttribute: Malformed Attribute Set \"%s\"",
                    Tcl_GetString(AttrSet));
    return NULL;
  }
  CDM_DEBUG(" ## \"%s\"\n", Tcl_GetString(AttrSet));
  for (i=0; i<items; i++) {
    result = Tcl_ListObjGetElements(CDM_Interp, Attributes[i], &attr_items,
                                    &elements);
    if (result != TCL_OK || attr_items != 2) {
      CDM_ReportError("CDM_RemoveAttribute: Malformed Attribute \"%s\"",
                      Tcl_GetString(Attributes[i]));
      return NULL;
    }
    if (CDM_Strcmp(Name, Tcl_GetString(elements[0])) == 0) {// We have found it!
      // If we give Tcl_ListObjReplace a shared object, we will die!
      if (Tcl_IsShared(AttrSet)) {
        Tcl_Obj *objv[1];
#ifdef DUPLICATE_DEBUG
        CDM_Debug("RemoveAttribute: Duplicating (%d): %s\n",
               AnnotationParts[3]->refCount, Tcl_GetString(AttrSet));
#endif /* DUPLICATE_DEBUG */
        AttrSet = Tcl_DuplicateObj(AttrSet);
#ifdef ENABLE_ANNOTATION_OBJECT
        if (ANNOTATION_OBJ(Ann)) {
          CDMAnnObj_SetAttributes(Ann, AttrSet);
        } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
          objv[0] = AttrSet;
          Tcl_ListObjReplace(CDM_Interp, Ann, 3, 1, 1, &AttrSet);
#ifdef ENABLE_ANNOTATION_OBJECT
        }
#endif /* ENABLE_ANNOTATION_OBJECT */
      }
      if (Tcl_ListObjReplace(CDM_Interp, AttrSet, i, 1, 0, NULL)
                       != TCL_OK) {
        CDM_ReportError("Cannot Remove Attribute \"%s\" from the Attribute "
                        "Set \"%s\"", Name, Tcl_GetString(AttrSet));
        return NULL;
      }
      CDM_DEBUG("Removed Attribute \"%s\"\n", Name);
      CDM_DEBUG(" ## \"%s\"\n", Tcl_GetString(AttrSet));
      Tcl_InvalidateStringRep(Ann);
      return Ann;
    }
  }
  CDM_DEBUG("Attribute \"%s\" NOT FOUND!\n", Name);
  return Ann;
} /* CDM_RemoveAttribute */

/*
 * ---- CDM_ReconstructAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function accepts as argument a valid Annotation object (of type
 *     CDM_Annotation). This function will construct and return a new Annotation
 *     object that will be an exact copy of the provided one. The only
 *     difference will be that the returned Annotation object will utilise
 *     elements from the object cache that is managed internally by CDM. As a
 *     result the returned object will normally require less memory to be
 *     stored than the provided one, as it will re-use elements already in use
 *     by other CDM objects.
 *
 *     This function should be used on Annotation objects that originate from
 *     an unknown source. For example, if an Annotation object is stored in a
 *     file and then red back as a string, it will not use cached by the CDM
 *     objects. It will be a good idea if the caller created a new object by
 *     calling this function and delete the object red from the file by either
 *     calling [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *     CDM_Free.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Annotation: The Annotation object that is to be reconstucted. (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_ReconstructAnnotation (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a new Annotation object (of type CDM_Annotation)
 *     that is an exact copy of the porvided Annotation object, but requires
 *     less memory due to re-using cached objects.
 *
 *     In case of an error (i.e. the specified object is not a valid Annotation
 *     object) NULL will be returned and an error message will be left at the
 *     current active interpreter (CDM_Interp).
 *
 *     The returned object is not property of CDM and the caller is
 *     responsible to free this object when it is no longer needed by either
 *     calling [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *     CDM_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_ReconstructAttribute, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_ReconstructAnnotation
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_ReconstructAnnotation(CDM_Annotation Annotation) {
  Tcl_Obj **AnnElem, **Elem, *argv[4], *objv[4], *obj;
  long id, start, end;
  int status, items, i, cacheAttrSet = 1;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_ReconstructAnnotation: Aborting...");
    return NULL;
  }

  /*
   * Get the Annotation Elements...
   */
#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Annotation)) {
    id = CDMAnnObj_GetId(Annotation);
    /*
     * Reconstruct the Annotation Type...
     */
    argv[1] = ObjectCache.FindAnnotationType(
                           Tcl_GetString(CDMAnnObj_GetType(Annotation)));

    if (CDMAnnObj_GetStart(Annotation) == -1) {
      objv[2] = CDMAnnObj_GetSpans(Annotation);
      start = end = -1;
    } else {
      objv[2] = NULL;
      start = CDMAnnObj_GetStart(Annotation);
      end   = CDMAnnObj_GetEnd(Annotation);
    }
    objv[3] = CDMAnnObj_GetAttributes(Annotation);
  } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
    status = Tcl_ListObjGetElements(NULL, Annotation, &items, &AnnElem);
    if (status != TCL_OK || items != 4) {
      CDM_ReportError("Malformed Annotation \"%s\"",Tcl_GetString(Annotation));
      return NULL;
    }

    /*
     * Has the Annotation an Id?
     */
    if (Tcl_GetLongFromObj(NULL, AnnElem[0], &id) != TCL_OK) {
#ifdef ENABLE_ANNOTATION_OBJECT
      id = -1;
#else /* ENABLE_ANNOTATION_OBJECT */
      /*
       * The Id was not a long integer. Remove it and place the empty string...
       */
      argv[0] = &CDM_Obj_NULL;
    } else {
      argv[0] = CDM_NewLongObj(id);
#endif /* ENABLE_ANNOTATION_OBJECT */
    }

    /*
     * Reconstruct the Annotation Type...
     */
    argv[1] = ObjectCache.FindAnnotationType(Tcl_GetString(AnnElem[1]));
    objv[2] = AnnElem[2];
    objv[3] = AnnElem[3];
#ifdef ENABLE_ANNOTATION_OBJECT
  }
#endif /* ENABLE_ANNOTATION_OBJECT */

  /*
   * Reconstruct the Attribute set...
   */
  status = Tcl_ListObjGetElements(NULL, objv[3], &items, &Elem);
  if (status != TCL_OK) {
    CDM_ReportError("Malformed Attribute Set of Annotation \"%s\"",
                     Tcl_GetString(Annotation));
    return NULL;
  }
  if (items) {
    argv[3] = Tcl_NewListObj(0, NULL);
    if (argv[3] == NULL) return NULL;
    for (i=0; i<items; i++) {
      obj = CDM_ReconstructAttribute(Elem[i], &status);
      if (obj == NULL) {
        CDM_ReportError("Malformed Attribute \"%s\" in Annotation \"%s\"",
                       Tcl_GetString(Elem[i]), Tcl_GetString(Annotation));
        CDM_Free(argv[3]);
        return NULL;
      }
      Tcl_ListObjAppendElement(NULL, argv[3], obj);
      /*
       * Is the Attribute value type a set? We should not cache such Attribute
       * sets...
       */
      if (cacheAttrSet && status == CDM_STRING_SET) {
        cacheAttrSet = 0;
      }
    }
  } else {
    argv[3] = &CDM_Obj_NULL;
  }

  /*
   * If memory compression is enabled, we have to check whether the same
   * attribute set is used by some other annotation in order to share the
   * previous one and free this...
   */
  /* Now this variable is true if argv[3] should be freed... */
  items = 1;
  if (cacheAttrSet && argv[3] != &CDM_Obj_NULL && CDM_EnableMemoryCompression) {
    obj = CDM_SearchForAttributeSet(argv[3]);
    if (obj && obj != argv[3]) {
      if (argv[3] != &CDM_Obj_NULL) Tcl_DecrRefCount(argv[3]);
      argv[3] = obj;
      items = 0; /* Now this variable is true if argv[3] should be freed... */
    }
  }
  /* Now this variable is true if argv[3] should be freed... */
  cacheAttrSet = items;

  /*
   * Reconstruct the Span set...
   */
  if (objv[2] != NULL) {
    status = Tcl_ListObjGetElements(NULL, objv[2], &items, &Elem);
    if (status != TCL_OK) {
      CDM_ReportError("Malformed Span Set of Annotation \"%s\"",
                       Tcl_GetString(Annotation));
      /*
       * Id_Obj & Type_Obj should not be freed, as they are managed by cache...
       */
      if (cacheAttrSet && argv[3] != &CDM_Obj_NULL) Tcl_DecrRefCount(argv[3]);
      return NULL;
    }
  } else {
    items = 0;
  }
  if (items) {
    argv[2] = Tcl_NewListObj(0, NULL);
    if (argv[2] == NULL) {
      if (cacheAttrSet && argv[3] != &CDM_Obj_NULL) Tcl_DecrRefCount(argv[3]);
      return NULL;
    }
    for (i=0; i<items; i++) {
      status = CDM_GetSpanOffsets(Elem[i], &start, &end);
      if (status != TCL_OK) {
        CDM_ReportError("Malformed Span \"%s\" in Annotation \"%s\"",
                       Tcl_GetString(Elem[i]), Tcl_GetString(Annotation));
        CDM_Free(argv[2]);
        if (cacheAttrSet && argv[3] != &CDM_Obj_NULL) Tcl_DecrRefCount(argv[3]);
        return NULL;
      }
      obj = CDM_CreateSpan(start, end);
      if (obj == NULL) {
        CDM_Free(argv[2]);
        if (cacheAttrSet && argv[3] != &CDM_Obj_NULL) Tcl_DecrRefCount(argv[3]);
        return NULL; /* We are out of memory here... */
      }
      Tcl_ListObjAppendElement(NULL, argv[2], obj);
    }
  } else {
    argv[2] = &CDM_Obj_NULL;
  }

  /*
   * We have finished. Just create and return the new Annotation...
   */
#ifdef ENABLE_ANNOTATION_OBJECT
  if (objv[2] != NULL) {
    if (items == 1) {
      Tcl_DecrRefCount(argv[2]);
      return CDM_NewAnnotationObj(id, argv[1], start, end, argv[3]);
    } else {
      return CDM_NewAnnotationObj(id, argv[1], argv[2], argv[3]);
    }
  } else {
    return CDM_NewAnnotationObj(id, argv[1], start, end, argv[3]);
  }
#else
  return Tcl_NewListObj(4, argv);
#endif /* ENABLE_ANNOTATION_OBJECT */
} /* CDM_ReconstructAnnotation */

/*
 * ---- CDM_AddAnnotationSpan:
 *  DESCRIPTION:
 *   <English:
 *    This function will add the provided Span object (of type CDM_Span) in the
 *    Span Set object held internally by the Annotation object.
 *    The returned value of this function will always be a new Annotation
 *    object with the provided Span object added.
 *    In the case of an error
 *    NULL will be returned and an error will be left at the current active
 *    Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Ann:    An Annotation object.                            (CDM_Annotation)
 *     span:   The Span object that should be added in the set of spans held by the specifiied Annotation object. (CDM_Span)
 *     end:    The end byte offset of the Span to be deleted.   (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AddSpan (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a new Annotation object (of type CDM_Annotation).
 *    In case of an error
 *    NULL will be returned and an error will be left at the current active
 *    Tcl interpreter (CDM_Interp).
 *
 *    Note that the returned Annotation object
 *    (if not NULL and not the same as the
 *    original Annotation object) is not property of CDM and
 *    the caller is responsible to free this object when it is no longer needed
 *    by either calling
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or CDM_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *    CDM_RemoveSpan,
 *    tip_AddSpan
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_AddAnnotationSpan(CDM_Annotation Annotation, CDM_Span span) {
  CDM_Annotation ResultAnnotation;
  CDM_SpanSet    SpanSet, NewSpanSet;
  CDM_Span      *Span;
  int id, items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  // printf("CDM_AddAnnotationSpan\n"); fflush(0);
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AddAnnotationSpan: Aborting...");
    return NULL;
  }
  if (!CDM_ValidAnnotation(Annotation)) return NULL;

//#ifdef ENABLE_ANNOTATION_OBJECT
//  if (ANNOTATION_OBJ(Annotation) && CDMAnnObj_GetStart(Annotation) != -1) {
//    start = CDMAnnObj_GetStart(Annotation);
//    end   = CDMAnnObj_GetEnd(Annotation);
//    return Tcl_GetRange(Text, (int) start, (int) end - 1);
//  }
//#endif /* ENABLE_ANNOTATION_OBJECT */

  SpanSet = CDM_GetSpans(Annotation);
  // TODO: Fix crash when adding in annotation with empty span set...
  //printf("SpanSet=%p (%p)\n", SpanSet, &CDM_Obj_NULL); fflush(0);
  //printf("Spans: %p %d\n", SpanSet, SpanSet->refCount);
  //printf(" => %s\n", Tcl_GetString(SpanSet)); fflush(0);
  Tcl_ListObjGetElements(CDM_GetActiveInterpreter(), SpanSet, &items, &Span);
  NewSpanSet = Tcl_NewListObj(items, Span); Tcl_IncrRefCount(NewSpanSet);
  Tcl_ListObjAppendElement(NULL, NewSpanSet, span);

  ResultAnnotation = CDM_CreateAnnotation(CDM_GetType(Annotation),
                                          NewSpanSet,
                                          CDM_GetAttributes(Annotation));
  if (ResultAnnotation == NULL) {
    Tcl_DecrRefCount(NewSpanSet);
    return NULL;
  }
  /* Has the Annotation an Id? */
  id = CDM_GetId(Annotation);
  if (id != -1) {
    /* We have to add the same Annotation Id... */
#ifdef ENABLE_ANNOTATION_OBJECT
    if (ANNOTATION_OBJ(ResultAnnotation)) {
      CDMAnnObj_SetId(ResultAnnotation, id);
    } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
      Tcl_Obj *objv[1];
      objv[0] = CDM_NewLongObj(id);
      Tcl_ListObjReplace(NULL, ResultAnnotation, 0, 1, 1, objv);
#ifdef ENABLE_ANNOTATION_OBJECT
    }
#endif /* ENABLE_ANNOTATION_OBJECT */
  }
  Tcl_DecrRefCount(NewSpanSet);
  return ResultAnnotation;
} /* CDM_AddAnnotationSpan */

/*
 * ---- CDM_RemoveSpan:
 *  DESCRIPTION:
 *   <English:
 *    This function will remove the Span having as start and end offsets
 *    the values specified through the "start" and "end" parameters
 *    from the specified Annotation, if such a Span object exists.
 *    The returned value of this function will always be a new Annotation
 *    object if a proper Span object was found and removed.
 *    If a Span object with the
 *    proper start and end offsets was not found, the same Annotation object
 *    as the provided one will be returned. Finally, in the case of an error
 *    NULL will be returned and an error will be left at the current active
 *    Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Ann:    An Annotation object.                            (CDM_Annotation)
 *     start:  The start byte offset of the Span to be deleted. (long)
 *     end:    The end byte offset of the Span to be deleted.   (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_RemoveSpan (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a new Annotation object (of type CDM_Annotation) if
 *    a Span object with the same offsets as the provided ones is found and
 *    removed from the provided Annotation.
 *    In the case that a proper Span object is not found, the provided
 *    Annotation object (specified through the "Annotation" parameter)
 *    is returned. Finally, in the case of an error
 *    NULL will be returned and an error will be left at the current active
 *    Tcl interpreter (CDM_Interp).
 *
 *    Note that the returned Annotation object
 *    (if not NULL and not the same as the
 *    original Annotation object) is not property of CDM and
 *    the caller is responsible to free this object when it is no longer needed
 *    by either calling
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or CDM_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *    tip_RemoveSpan
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_RemoveSpan(CDM_Annotation Annotation, long start, long end) {
  CDM_Annotation ResultAnnotation = Annotation;
  CDM_SpanSet    SpanSet, NewSpanSet;
  CDM_Span      *Span;
  long offset1, offset2;
  int i, items, removed = 0;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_RemoveSpan: Aborting...");
    return NULL;
  }
  if (!CDM_ValidAnnotation(Annotation)) return NULL;

  SpanSet    = CDM_GetSpans(Annotation);
  NewSpanSet = Tcl_NewListObj(0, NULL); Tcl_IncrRefCount(NewSpanSet);
  Tcl_ListObjGetElements(NULL, SpanSet, &items, &Span);
  for (i=0; i<items; i++) {
    CDM_GetSpanOffsets(Span[i], &offset1, &offset2);
    if (offset1 == start && offset2 == end) {
      removed++;
    } else {
      Tcl_ListObjAppendElement(NULL, NewSpanSet, Span[i]);
    }
  }

  /*
   * Do we have to remove any Span objects?
   */
  if (removed) {
    ResultAnnotation = CDM_CreateAnnotation(CDM_GetType(Annotation),
                                            NewSpanSet,
                                            CDM_GetAttributes(Annotation));
    if (ResultAnnotation == NULL) {
      Tcl_DecrRefCount(NewSpanSet);
      return NULL;
    }
    /* Has the Annotation an Id? */
    offset1 = CDM_GetId(Annotation);
    if (offset1 != -1) {
      /* We have to add the same Annotation Id... */
#ifdef ENABLE_ANNOTATION_OBJECT
      if (ANNOTATION_OBJ(ResultAnnotation)) {
        CDMAnnObj_SetId(ResultAnnotation, offset1);
      } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
        Tcl_Obj *objv[1];
        objv[0] = CDM_NewLongObj(offset1);
        Tcl_ListObjReplace(NULL, ResultAnnotation, 0, 1, 1, objv);
#ifdef ENABLE_ANNOTATION_OBJECT
      }
#endif /* ENABLE_ANNOTATION_OBJECT */
    }
  }
  Tcl_DecrRefCount(NewSpanSet);
  return ResultAnnotation;
} /* CDM_RemoveSpan */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |    E N D  O F  A N N O T A T I O N   S E C T I O N    |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/


/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     A N N O T A T I O N    S E T   S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 * ---- CDM_CreateAnnotationSet:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new (and empty) Annotation Set
 *     object (of type CDM_AnnotationSet). This new Annotation Set can be used
 *     in order to place Annotation objects (of type CDM_Annotation) into it.
 *
 *     CDM does not own the returned
 *     object: the caller is responsible to use Tcl_DecrRefCount or CDM_Free
 *     in order to free the object and the memory associated with it.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *    None: (void)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to tip_CreateAnnotationSet (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns an empty Annotation set object (of type
 *    CDM_AnnotationSet). The caller is responsible to free this object
 *    through Tcl_DecrRefCount or CDM_Free.
 *
 *    If an error occurs, then NULL will be returned and an error message
 *    will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_CreateAnnotationSet
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_CreateAnnotationSet(void) {
  CDM_AnnotationSet Set = Tcl_NewListObj(0, NULL);
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_CreateAnnotationSet: Aborting...");
    return NULL;
  }

  if (Set == NULL) {
    CDM_Panic("CDM_CreateAnnotationSet: Tcl_NewListObj returned NULL "
              "- out of memory?\n");
  }
  return Set;
} /* CDM_CreateAnnotationSet */

/*
 * Utility function used by CDM_CompareAnnotationSets to re-arrange an
 * AnnotationSet...
 */
static int CDM_SortAnnotationsWithRespectToAttributeValues(Tcl_Obj *set,
                                         Tcl_Obj *Attrs, long start, long end) {
  CDM_Attribute attr;
  Tcl_Obj **element, **Attribute, **tmp, *tmpObj, *elemObj;
  Tcl_Obj *objv[2];
  int i, j, items, status;
  CDM_THREAD_SPECIFIC_DATA;

  if (Tcl_ListObjGetElements(CDM_Interp, set,   &items, &element) != TCL_OK)
    return TCL_ERROR;
  if (Tcl_ListObjGetElements(CDM_Interp, Attrs, &items, &Attribute) != TCL_OK)
    return TCL_ERROR;

  tmpObj = Tcl_NewListObj(0, NULL);
  Tcl_IncrRefCount(tmpObj);
  for (i=start; i<end; i++) {
    elemObj = Tcl_NewListObj(0, NULL);
    for (j=0; j<items; j++) {
      /* Does the Attribute exists? */
      attr = CDM_GetAttribute(element[i], Tcl_GetString(Attribute[j]));
      if (attr == NULL) continue;
      Tcl_ListObjAppendElement(NULL, elemObj, CDM_GetValueValue(attr));
    }
    /* Append the position so as to know what to exchange after sorting... */
    Tcl_ListObjAppendElement(NULL, elemObj, CDM_NewLongObj(i));
    /* Append the object into the list to be sorted... */
    Tcl_ListObjAppendElement(NULL, tmpObj, elemObj);
  }

  /*
   * Sort the object, using regular Tcl command "lsort"...
   */
  objv[0] = Tcl_NewStringObj("lsort", 5);
  objv[1] = tmpObj;
  Tcl_IncrRefCount(objv[0]);
  status = Tcl_EvalObjv(CDM_Interp, 2, objv, TCL_EVAL_GLOBAL);
  Tcl_DecrRefCount(objv[0]);
  Tcl_DecrRefCount(tmpObj);
  if (status != TCL_OK) return status;
  tmpObj = Tcl_GetObjResult(CDM_Interp);
  Tcl_IncrRefCount(tmpObj);

  if (Tcl_ListObjGetElements(NULL, tmpObj, &items, &tmp) != TCL_OK) {
    Tcl_DecrRefCount(tmpObj); return TCL_ERROR;
  }
  /* Place the Annotations with the sorted order in a new object... */
  elemObj = Tcl_NewListObj(0, NULL);
  Tcl_IncrRefCount(elemObj);
  for (i=0; i<items; i++) {
    /* Get the integer stored as last element... */
    if (Tcl_ListObjGetElements(NULL, tmp[i], &j, &Attribute) != TCL_OK) {
      Tcl_DecrRefCount(tmpObj);
      Tcl_DecrRefCount(elemObj);
      return TCL_ERROR;
    }
    if (Tcl_GetIntFromObj(NULL, Attribute[j-1], &j) != TCL_OK) {
      Tcl_DecrRefCount(tmpObj);
      Tcl_DecrRefCount(elemObj);
      return TCL_ERROR;
    }
    /* Append the proper Annotation into the object... */
    Tcl_ListObjAppendElement(NULL, elemObj, element[j]);
  }
  Tcl_DecrRefCount(tmpObj);

  /* Finally, replace the elements with the sorted ones... */
  if (Tcl_ListObjGetElements(CDM_Interp, elemObj, &items, &element) != TCL_OK) {
    Tcl_DecrRefCount(elemObj);
    return TCL_ERROR;
  }
  if (Tcl_ListObjReplace(CDM_Interp, set, start, items, items, element)
                                                                    != TCL_OK) {
    Tcl_DecrRefCount(elemObj);
    return TCL_ERROR;
  }
  Tcl_DecrRefCount(elemObj);
  return TCL_OK;
} /* CDM_SortAnnotationsWithRespectToAttributeValues */

/*
 * ---- CDM_CompareAnnotationSets:
 *  DESCRIPTION:
 *   <English:
 *    This function can be used to compare two Annotation set objects, and
 *    calculate the number of Annoatations that are equal, according to the
 *    specified constraints.
 *
 *    As a first step, the two Annotation sets will be sorted internally based
 *    on their Spans (by using CDM_SortAnnotationSet). Then, if parameter
 *    "checkValues" is true (thus enabling Attribute Value checking) and the
 *    Tcl objects "AttributesList" and "ValuesList" are not empty, the
 *    Annotations that have Attributes with names from "AttributesList" but
 *    not values from the corresponding entries in "ValuesList".
 *
 *    When the initial filtering stage is completed, the two Annotation sets
 *    will be compared. Each Annotation from the key set will be compared with
 *    each Annotation from the responce set, based on their spans. If the two
 *    Annotations share the same spans, then their Attributes (with names from
 *    the "AttributesList" list) will be compared, and if Attribute Value
 *    checking is enabled the comparison will be also performed on the
 *    Attribute values.
 *
 *    The return value from this function will be a standart Tcl completion
 *    code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *    error occurs, the return value will be TCL_ERROR and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). If all annotators run without errors, TCL_OK will be
 *    returned.
 *
 *    The results of the Tcl code execution will be placed as the result in
 *    the current Tcl interpreter (CDM_Interp), if the command has been
 *    succesfully executed. The caller can use
 *    [[A HREF="../TclTk/TclLib/SetResult.htm"]]Tcl_GetObjResult[[/A]](CDM_GetActiveInterpreter()) to get the stored result
 *    Tcl object.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    KeySet: The set holding the Key Annotations      (CDM_AnnotationSet)
 *    ResSet: The set holding the Responce Annotations (CDM_AnnotationSet)
 *    AttributesList: A Tcl object containing the Attribute names to be examined (Tcl_Obj *)
 *    ValuesList: A Tcl object containing the values of Attributes to be examined (Tcl_Obj *)
 *    spanOverlapPercent: A double value specifying the overlap the Annotations' spans must have to be considered as equal. If this value is 100.0, then the spans must be identical. (double)
 *    checkValues: A boolean value specifying whether to enable Attribute Value checking (int)
 *    LogInfo: A boolean value specifying whether to generate logging information [LogInfo != 0] or not [LogInfo = 0] (int)
 *   >
 *  NOTES:
 *   <English:
 *    Note that the parameter "checkValues" will be ignored if the parameter
 *    "AttributesList" is an empty Tcl list object (i.e. has no elements).
 *    Also note that the object "ValuesList" must have exactly the same number
 *    of elements with "AttributesList": there must be a correspondance
 *    between the two objects, as the N-th element of "ValuesList" should be a
 *    list object, holding the allowable value patterns for the Attribute name
 *    stored in the N-th element of the "AttributesList" object.
 *
 *    An important limitation of this function is that it may produce incorect
 *    results, if any of the Annotations contained in the two sets has more
 *    than one Span elements in its Span set.
 *
 *    This function is equivelant to tip_CompareAnnotationSets (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value from this function will be a standart Tcl completion
 *    code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *    error occurs, the return value will be TCL_ERROR and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). If all annotators run without errors, TCL_OK will be
 *    returned.
 *
 *    The result of the comparison will be placed as the result object in the
 *    current active Tcl interpreter (CDM_Interp). This object will be a list
 *    object with four elements. The first element will be the number of
 *    correct Annotations (i.e. the Annotations that matched), the second
 *    element will be the number of the key Annotations and the third element
 *    will be the number of responce Annotations.
 *
 *    The fourth (last) element is the logging info that will be generated by
 *    this function, if the parameter "LogInfo" is true (!= 0). If "LogInfo"
 *    is false (0), then this element will be an empty object. If logging
 *    information is enabled, this element will hold a list object, where each
 *    list element will have three elements: The first and third element will
 *    either be an Annotation object or the empty object. The second element
 *    will be a flag signaling the relation between the two Annotations. Valid
 *    flags are "<" (denoting that the first element is an empty object), "=="
 *    (denoting that the two Annotations are equal), "!=" (denoting that the
 *    two Annotations span the same regions but differ in Attributes) and ">"
 *    (denoting that the third element is an empty object).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_CompareAnnotations, tip_CompareAnnotationSets
 *   >:&:
 */
CDM_EXPORT
int CDM_CompareAnnotationSets(CDM_AnnotationSet KeySet,
                              CDM_AnnotationSet ResSet,
                              Tcl_Obj *AttributesList, Tcl_Obj *ValuesList,
                              double spanOverlapPercent,
                              int checkValues, int LogInfo) {

#define _CDM_LOG_REPORT_UNEQUAL \
    if (LogInfo) { \
      objv[0] = key[keyIndex]; \
      objv[1] = unequal; \
      objv[2] = res[resIndex]; \
      Tcl_ListObjAppendElement(CDM_Interp, argv[3], Tcl_NewListObj(3, objv)); \
    } /* if (LogInfo) */
#define _CDM_LOG_REPORT_EQUAL \
    if (LogInfo) { \
      objv[0] = key[keyIndex]; \
      if (overlapValid == 2) objv[1] = fuzzyEq;\
      else objv[1] = equal; \
      objv[2] = res[resIndex]; \
      Tcl_ListObjAppendElement(CDM_Interp, argv[3], Tcl_NewListObj(3, objv)); \
    } /* if (LogInfo) */
#define _CDM_LOG_REPORT_LOWER \
    if (LogInfo) { \
      objv[0] = &CDM_Obj_NULL; \
      objv[1] = lower; \
      objv[2] = res[resIndex]; \
      Tcl_ListObjAppendElement(CDM_Interp, argv[3], Tcl_NewListObj(3, objv)); \
    } /* if (LogInfo) */
#define _CDM_LOG_REPORT_GREATER \
    if (LogInfo) { \
      objv[0] = key[keyIndex]; \
      objv[1] = greater; \
      objv[2] = &CDM_Obj_NULL; \
      Tcl_ListObjAppendElement(CDM_Interp, argv[3], Tcl_NewListObj(3, objv)); \
    } /* if (LogInfo) */

  CDM_Attribute keyAttr, resAttr;
  CDM_AnnotationSet sortedKeySet = NULL, sortedResSet = NULL;
  Tcl_Obj *argv[4], *objv[3];
  Tcl_Obj *lower   = CDM_NewStringObj("<");
  Tcl_Obj *equal   = CDM_NewStringObj("==");
  Tcl_Obj *fuzzyEq = CDM_NewStringObj("~");
  Tcl_Obj *unequal = CDM_NewStringObj("!=");
  Tcl_Obj *greater = CDM_NewStringObj(">");
  Tcl_Obj **element, **key, **res, **Attribute, **Value;
  char *keyAttrValue, *resAttrValue;
  long keyLen, resLen, Correct = 0, totalKeyAnns, totalResponseAnns;
  long keyStart, keyEnd, resStart, resEnd;
  long nextkeyStart, nextkeyEnd;
  int i, j, items, keyIndex = 0, resIndex = 0, attributeNumber, valueNumber;
  int attributesLen = CDM_Length(AttributesList), checkAttributes = 0;
  int valuesLen     = CDM_Length(ValuesList), accept;
  int overlapValid, maxStart, minEnd;
  CDM_THREAD_SPECIFIC_DATA;

  if (attributesLen) checkAttributes = 1;
  Tcl_IncrRefCount(lower);
  Tcl_IncrRefCount(equal);
  Tcl_IncrRefCount(fuzzyEq);
  Tcl_IncrRefCount(unequal);
  Tcl_IncrRefCount(greater);

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    Tcl_DecrRefCount(lower);
    Tcl_DecrRefCount(equal);
    Tcl_DecrRefCount(fuzzyEq);
    Tcl_DecrRefCount(unequal);
    Tcl_DecrRefCount(greater);
    return CDM_ReportError("CDM_CompareAnnotationSets: Aborting...");
  }

  if (checkAttributes) {
    if (Tcl_ListObjGetElements(CDM_Interp, AttributesList, &attributeNumber,
                               &Attribute) != TCL_OK) goto error;
    if (valuesLen) {
      if (Tcl_ListObjGetElements(CDM_Interp, ValuesList, &valueNumber,
                                 &Value) != TCL_OK) goto error;
      /* If values are provided, their number must match the Attribute names */
      if (attributeNumber != valueNumber) {
        Tcl_DecrRefCount(lower);
        Tcl_DecrRefCount(equal);
        Tcl_DecrRefCount(fuzzyEq);
        Tcl_DecrRefCount(unequal);
        Tcl_DecrRefCount(greater);
        return CDM_ReportError("CDM_CompareAnnotationSets: parameters "
        "\"AttributesList\" and \"ValuesList\" must have the same number "
        "of items");
      }
      /* There is the possibility the provided value array to have all its
       * elements empty. We should catch such a case, as we will loose
       * execution time in useless checking and copying of Annotations... */
      j = 1;
      for (i=0; i<valueNumber; i++) {
        if (CDM_Length(Value[i])) {j = 0; break;}
      }
      if (j) valuesLen = 0; /*All elements in ValueList are the empty element*/
    }
  } else {
    checkValues = 0;
    valuesLen   = 0;
  }

  //printf("AttributesList = %s\n", Tcl_GetString(AttributesList));
  //printf("ValuesList = %s\n", Tcl_GetString(ValuesList));
  //printf("checkAttributes = %d\n", checkAttributes);
  //printf("checkValues = %d\n", checkValues);
  //printf("valuesLen = %d\n", valuesLen);
  //printf("LogInfo = %d\n", LogInfo);

  /*
   * Sort the two sets according to their spans and filter the Annotations
   * that possibly have unwanted values in their Attributes...
   */
  if (KeySet) {
    Tcl_IncrRefCount(KeySet);
    sortedKeySet = CDM_SortAnnotationSet(KeySet);
    Tcl_IncrRefCount(sortedKeySet);
    Tcl_DecrRefCount(KeySet);
    if (checkAttributes && valuesLen) {
      /* We have to remove the Annotations whoose Attribute values are not
       * desirable... */
      CDM_AnnotationSet tmpSet = Tcl_NewListObj(0, NULL);
      Tcl_IncrRefCount(tmpSet);
      if (Tcl_ListObjGetElements(CDM_Interp, sortedKeySet, &items, &key)
                                                        != TCL_OK) goto error;
      /* Iterate over all Annotations... */
      for (i=0; i<items; i++) {
        /* Iterate over all Attributes... */
        accept = 1;
        for (j=0; j<attributeNumber; j++) {
          if (!CDM_AnnotationContainsAttributeMatchingValues(key[i],
                                     Tcl_GetString(Attribute[j]), Value[j])) {
            accept = 0; break;
          }
        } /* Iterate over all Attributes... */
        if (accept) Tcl_ListObjAppendElement(NULL, tmpSet, key[i]);
      } /* Iterate over all Annotations... */
      Tcl_DecrRefCount(sortedKeySet);
      sortedKeySet = tmpSet;
    }
  }
  if (ResSet) {
    Tcl_IncrRefCount(ResSet);
    sortedResSet = CDM_SortAnnotationSet(ResSet);
    Tcl_IncrRefCount(sortedResSet);
    Tcl_DecrRefCount(ResSet);
    if (checkAttributes && valuesLen) {
      /* We have to remove the Annotations whoose Attribute values are not
       * desirable... */
      CDM_AnnotationSet tmpSet = Tcl_NewListObj(0, NULL);
      Tcl_IncrRefCount(tmpSet);
      if (Tcl_ListObjGetElements(CDM_Interp, sortedResSet, &items, &key)
                                                        != TCL_OK) goto error;
      /* Iterate over all Annotations... */
      for (i=0; i<items; i++) {
        /* Iterate over all Attributes... */
        accept = 1;
        for (j=0; j<attributeNumber; j++) {
          if (!CDM_AnnotationContainsAttributeMatchingValues(key[i],
                                     Tcl_GetString(Attribute[j]), Value[j])) {
            accept = 0; break;
          }
        } /* Iterate over all Attributes... */
        if (accept) Tcl_ListObjAppendElement(NULL, tmpSet, key[i]);
      } /* Iterate over all Annotations... */
      Tcl_DecrRefCount(sortedResSet);
      sortedResSet = tmpSet;
    } /* if (checkAttributes && valuesLen) */
  }

  /*
   * In order to find differences correctly, we have to search for Annotations
   * that share common spans, and re-arrange them according to their Attribute
   * Values. This will make our tool working on a case like the following:
   *
   * (a,b) (type=1)     (a,b) (type=2)
   * (a,b) (type=2)     (a,b) (type=1)
   *
   * which will be characterised as 2 mismatches (instead of 2 correct
   * matches...)
   */
  if (KeySet) {
    long lastStart = -1, lastEnd = -1, startIndex = -1;
    if (Tcl_ListObjGetElements(CDM_Interp, sortedKeySet, &items, &key)
                                                        != TCL_OK) goto error;
    /* Iterate over all Annotations... */
    for (i=0; i<items; i++) {
      if (CDM_GetFirstSpanOffsets(key[i], &keyStart, &keyEnd) != TCL_OK) {
        goto error;
      }
      if (lastStart != keyStart || lastEnd != keyEnd) {
        if (i == startIndex + 1) {
          lastStart = keyStart; lastEnd = keyEnd;
          startIndex = i; continue;
        } else {
          /* We have located Annotations that should be sorted... */
          if (CDM_SortAnnotationsWithRespectToAttributeValues(sortedKeySet,
              AttributesList, startIndex, i) != TCL_OK) goto error;
          if (Tcl_ListObjGetElements(CDM_Interp, sortedKeySet, &items, &key)
                                                        != TCL_OK) goto error;
          lastStart = keyStart; lastEnd = keyEnd;
          startIndex = i;
        }
      }
      /* This Annotation has the same span as the previous one. Simply do
       * nothing... */
    } /* for (i=0; i<items; i++) */
    if (items != startIndex +1) {
      /* We have located Annotations that should be sorted... */
      if (CDM_SortAnnotationsWithRespectToAttributeValues(sortedKeySet,
          AttributesList, startIndex, i) != TCL_OK) goto error;
      if (Tcl_ListObjGetElements(CDM_Interp, sortedKeySet, &items, &key)
                                                    != TCL_OK) goto error;
    }
  } /* if (KeySet) */
  if (ResSet) {
    long lastStart = -1, lastEnd = -1, startIndex = -1;
    if (Tcl_ListObjGetElements(CDM_Interp, sortedResSet, &items, &key)
                                                        != TCL_OK) goto error;
    /* Iterate over all Annotations... */
    for (i=0; i<items; i++) {
      if (CDM_GetFirstSpanOffsets(key[i], &keyStart, &keyEnd) != TCL_OK) {
        goto error;
      }
      if (lastStart != keyStart || lastEnd != keyEnd) {
        if (i == startIndex + 1) {
          lastStart = keyStart; lastEnd = keyEnd;
          startIndex = i; continue;
        } else {
          /* We have located Annotations that should be sorted... */
          if (CDM_SortAnnotationsWithRespectToAttributeValues(sortedResSet,
              AttributesList, startIndex, i) != TCL_OK) goto error;
          if (Tcl_ListObjGetElements(CDM_Interp, sortedResSet, &items, &key)
                                                        != TCL_OK) goto error;
          lastStart = keyStart; lastEnd = keyEnd;
          startIndex = i;
        }
      }
      /* This Annotation has the same span as the previous one. Simply do
       * nothing... */
    } /* for (i=0; i<items; i++) */
    if (items != startIndex +1) {
      /* We have located Annotations that should be sorted... */
      if (CDM_SortAnnotationsWithRespectToAttributeValues(sortedResSet,
          AttributesList, startIndex, i) != TCL_OK) goto error;
      if (Tcl_ListObjGetElements(CDM_Interp, sortedResSet, &items, &key)
                                                    != TCL_OK) goto error;
    }
  } /* if (ResSet) */

  /*
   * Now, in the two variables sortedKeySet & sortedResSet the two provided
   * sets are stored. These sets have been sorted according to their spans,
   * and the Annotations who contain Attribute values not wanted by the caller
   * have been removed. All Annotations in the two sets have allowed Attribute
   * values...
   */

  keyLen = CDM_Length(sortedKeySet);
  resLen = CDM_Length(sortedResSet);
  totalKeyAnns      = keyLen;
  totalResponseAnns = resLen;
  argv[0] = NULL;  /* Number of Correct Annotations */
  argv[1] = NULL;  /* Number of Key Annotations */
  argv[2] = NULL;  /* Number of Responce Annotations */
  argv[3] = Tcl_NewListObj(0, NULL); /* The object holding the log */

  /*
   * Consider the special case where at least one of the Annotation sets is
   * empty...
   */
  if (keyLen == 0 || resLen == 0) {
    /* Create the result value object: */
    argv[0] = Tcl_NewLongObj(0);       /* Number of Correct Annotations */
    argv[1] = Tcl_NewLongObj(keyLen);  /* Number of Key Annotations */
    argv[2] = Tcl_NewLongObj(resLen);  /* Number of Responce Annotations */
    if (LogInfo) {
      if (keyLen == 0) {
        /* No Key Annotations... */
        if (Tcl_ListObjGetElements(CDM_Interp, sortedResSet, &items, &element)
                                                          != TCL_OK) goto error;
        for (i = 0; i < items; i++) {
          objv[0] = &CDM_Obj_NULL;
          objv[1] = lower;
          objv[2] = element[i];
          Tcl_ListObjAppendElement(CDM_Interp, argv[3],Tcl_NewListObj(3, objv));
        }
      } else if (resLen == 0) {
        /* No Responce Annotations... */
        if (Tcl_ListObjGetElements(CDM_Interp, sortedKeySet, &items, &element)
                                                          != TCL_OK) goto error;
        for (i = 0; i < items; i++) {
          objv[0] = element[i];
          objv[1] = greater;
          objv[2] = &CDM_Obj_NULL;
          Tcl_ListObjAppendElement(CDM_Interp, argv[3],Tcl_NewListObj(3, objv));
        }
      }
    }
    Tcl_DecrRefCount(lower);
    Tcl_DecrRefCount(equal);
    Tcl_DecrRefCount(fuzzyEq);
    Tcl_DecrRefCount(unequal);
    Tcl_DecrRefCount(greater);
    if (sortedKeySet) Tcl_DecrRefCount(sortedKeySet);
    if (sortedResSet) Tcl_DecrRefCount(sortedResSet);
    Tcl_SetObjResult(CDM_Interp, Tcl_NewListObj(4, argv));
    return TCL_OK;
  } /* if (keyLen == 0 || resLen == 0) */

  /*
   * Difference Calculation
   */
  if (Tcl_ListObjGetElements(CDM_Interp, sortedKeySet, &items, &key) != TCL_OK)
     goto error;
  if (Tcl_ListObjGetElements(CDM_Interp, sortedResSet, &items, &res) != TCL_OK)
     goto error;

  while (keyIndex < keyLen || resIndex < resLen) {
    if ((keyIndex < keyLen) && (resIndex < resLen)) {
      /* Both Annotations exist */
      if (CDM_GetFirstSpanOffsets(key[keyIndex], &keyStart, &keyEnd) !=TCL_OK) {
        goto error;
      }
      if (CDM_GetFirstSpanOffsets(res[resIndex], &resStart, &resEnd) !=TCL_OK) {
        goto error;
      }

      /* Calculate the overlap of the two spans... */
      if (spanOverlapPercent < 100.0) {
        if ((keyStart == resStart) && (keyEnd == resEnd)) {
          overlapValid = 1;
        } else if (resStart >= keyStart && resStart <= keyEnd) {
          /* The two spans differ, but they overlap with at least one
           * character... */
          maxStart = keyStart;
          if (resStart > maxStart) maxStart = resStart;
          minEnd = keyEnd;
          if (resEnd   < minEnd)   minEnd = resEnd;
          /* Calculate span overlap percentage */
          if ((double(minEnd-maxStart)/double(keyEnd-keyStart)*100.0)
                                                     >= spanOverlapPercent) {
            overlapValid = 2; /* This signals *fuzzy* matching */
            /*printf("keyStart=%d, keyEnd=%d, resStart=%d, resEnd=%d, "
             *       "percent=%f, spanOverlapPercent=%f\n",
             *       keyStart, keyEnd, resStart, resEnd,
             *       double(minEnd-maxStart)/double(keyEnd-keyStart)*100.0,
             *       spanOverlapPercent); */
          } else {
            overlapValid = 0;
          }
        } else {
          overlapValid = 0;
        }
      } else {
        if ((keyStart == resStart) && (keyEnd == resEnd)) {
          overlapValid = 1;
        } else {
          overlapValid = 0;
        }
      }
      /* Annotations are categorised according to their first span */
      /* if ((keyStart == resStart) && (keyEnd == resEnd)) { */
      if (overlapValid) {
        /* The two Annotations span the same text range... */
        if (checkAttributes) {
          int annotationMatch = 1;
          /* Iterate over all Attribute names, to check at least existence */
          for (i = 0; i < attributeNumber; i++) {
            keyAttr = CDM_GetAttribute(key[keyIndex],
                                       Tcl_GetString(Attribute[i]));
            resAttr = CDM_GetAttribute(res[resIndex],
                                       Tcl_GetString(Attribute[i]));
            if ((keyAttr == NULL) && (resAttr == NULL)) {
              /* The Attribute is missing from both Annotations. */
              totalKeyAnns--; totalResponseAnns--; keyIndex++; resIndex++;
              break; /* It doesn't matter what the status of other Attributes
                        is... */
            } else if (keyAttr == NULL || resAttr == NULL) {
              /* The Attribute is missing from one Attribute.
               * This is a mismatch! But before we report an unequality, we
               * must check whether the next res Annotation has the same
               * span...
               */
              annotationMatch = 0;
              if (resIndex+1 < resLen) {
                if (CDM_GetFirstSpanOffsets(res[resIndex+1], &nextkeyStart,
                                            &nextkeyEnd) != TCL_OK) goto error;
                if (nextkeyStart == keyStart && nextkeyEnd == keyEnd) {
                  _CDM_LOG_REPORT_LOWER
                  resIndex++; break;
                }
              }
              /* If we reach here, there is no other responce Annotation with
               * the same span. Perhaps there is another key Annotation? */
              if (keyIndex+1 < keyLen) {
                if (CDM_GetFirstSpanOffsets(key[keyIndex+1], &nextkeyStart,
                                            &nextkeyEnd) != TCL_OK) goto error;
                if (nextkeyStart == keyStart && nextkeyEnd == keyEnd) {
                  _CDM_LOG_REPORT_GREATER
                  keyIndex++; break;
                }
              }
              _CDM_LOG_REPORT_UNEQUAL
              keyIndex++; resIndex++; break;
            } else {
              /* Atrribute exists in both Annotations... */
              if (checkValues) {
                keyAttrValue = CDM_GetValueString(keyAttr);
                resAttrValue = CDM_GetValueString(resAttr);
                /* We have to check the values of the Attributes:
                 * 1) If they match the values given by the caller
                 * 2) and of course are the same! */
                /* Are the two Attribute values the same? */
                if (CDM_Strcmp(keyAttrValue, resAttrValue) != 0) {
                  annotationMatch = 0;
                  if (resIndex+1 < resLen) {
                    if (CDM_GetFirstSpanOffsets(res[resIndex+1], &nextkeyStart,
                                             &nextkeyEnd) != TCL_OK) goto error;
                    if (nextkeyStart == keyStart && nextkeyEnd == keyEnd) {
                      _CDM_LOG_REPORT_LOWER
                      resIndex++; break;
                    }
                  }
                  /* If we reach here, there is no other responce Annotation
                   * with the same span. Perhaps there is another key
                   * Annotation? */
                  if (keyIndex+1 < keyLen) {
                    if (CDM_GetFirstSpanOffsets(key[keyIndex+1], &nextkeyStart,
                                             &nextkeyEnd) != TCL_OK) goto error;
                    if (nextkeyStart == keyStart && nextkeyEnd == keyEnd) {
                      _CDM_LOG_REPORT_GREATER
                      keyIndex++; break;
                    }
                  }
                  _CDM_LOG_REPORT_UNEQUAL
                  keyIndex++; resIndex++; break;
                } else {
                  /* The values are the same. Do nothing, as we report only
                   * mismatches... */
                } /* if (CDM_Strcmp(keyAttrValue, resAttrValue) != 0) */
              } else { /* if (checkValues) */
                /* The user only wants to check Attribute existence. The
                 * Attribute exists in both Annotations. We don't have to do
                 * anything, as we report only mismatches... */
              } /* if (checkValues) */
            } /* ((keyAttr == NULL) && (resAttr == NULL)) */
          } /* for (i = 0; i < attributeNumber; i++) */
          if (annotationMatch && (keyIndex < keyLen) && (resIndex < resLen)) {
            /* We have checked the Attributes and everything is ok... */
            _CDM_LOG_REPORT_EQUAL
            Correct++; keyIndex++; resIndex++;
          } /* if (annotationMatch) */
        } else {
          /* Do not compare Attributes... */
          _CDM_LOG_REPORT_EQUAL
          Correct++; keyIndex++; resIndex++;
        } /* if (checkAttributes) */
      } else if ((keyStart <= resStart) && (keyEnd >= resEnd)) {
        /* Test to see if the next key Annotation has a start inside this key
         * Annotation. If this is the case, get rid of this key Annotation as
         * we may loose alignment... */
        if (keyIndex+1 < keyLen) {
          if (CDM_GetFirstSpanOffsets(key[keyIndex+1], &nextkeyStart,
                                      &nextkeyEnd) !=TCL_OK) goto error;
          if (nextkeyStart < keyEnd && nextkeyStart <= resStart) {
            _CDM_LOG_REPORT_GREATER
            keyIndex++;
            continue;
          }
        }
        /* The key Annotation subsumes the response */
        _CDM_LOG_REPORT_LOWER
        resIndex++;
        /* Test to break out of subsumation */
        if (resIndex < resLen) {
          if (CDM_GetFirstSpanOffsets(res[resIndex], &resStart, &resEnd)
                                                           !=TCL_OK) goto error;
          if (keyEnd < resEnd) {
            _CDM_LOG_REPORT_GREATER
            keyIndex++;
          }
        } else {
          _CDM_LOG_REPORT_GREATER
          keyIndex++;
        }
      } else if ((resStart <= keyStart) && (resEnd >= keyEnd)) {
        /*
         * The response Annotation subsumes the key not at end of key span.
         *
         * Here we should simply skip the key annotation. But what if there
         * is the possibility that we get an exact match in the following
         * responce annotations? We should check that, and advance to the
         * next key annotation only if we are sure...
         */
        CDM_AnnotationSet matchingRange =
          CDM_AnnotationsMatchingRange(sortedResSet, keyStart, resStart);
        if (matchingRange == NULL) goto error;
        Tcl_IncrRefCount(matchingRange);
        if (CDM_Length(matchingRange)) {
          resIndex++;
        } else { /* if (CDM_Length(matchingRange)) */
          _CDM_LOG_REPORT_GREATER
          keyIndex++;
          /* Test to break out of subsumation */
          if (keyIndex < keyLen) {
            if (CDM_GetFirstSpanOffsets(key[keyIndex], &keyStart, &keyEnd)
                                                          != TCL_OK) goto error;
            if (resEnd < keyEnd) {
              _CDM_LOG_REPORT_LOWER
              resIndex++;
            }
          } else {
            _CDM_LOG_REPORT_LOWER
            resIndex++;
          }
        } /* if (CDM_Length(matchingRange)) */
        Tcl_DecrRefCount(matchingRange);
      } else if (keyStart < resStart) {
        /* case where the key start span is less than the response start span
         * test for subsumation is previous to this so end span of response
         * must be less than end span of key */
        _CDM_LOG_REPORT_GREATER
        keyIndex++;
      } else if (resStart < keyStart) {
        /* case where the response start span is less than the key start span
         * test for subsumation is previous to this so end span of key
         * must be less than end span of response */
        _CDM_LOG_REPORT_LOWER
        resIndex++;
      } else {
        /* We should never end here! */
        CDM_ReportError("CDM_CompareAnnotationSets: Internal Error!");
        goto error;
      }
    } else if ((keyIndex < keyLen) && !(resIndex < resLen)) {
      /* Only Key Annotation */
      _CDM_LOG_REPORT_GREATER
      keyIndex++;
    } else if (!(keyIndex < keyLen) && (resIndex < resLen)) {
      /* Only Responce Annotation */
      _CDM_LOG_REPORT_LOWER
      resIndex++;
    }
  }
  argv[0]=Tcl_NewLongObj(Correct);          /* Number of Correct Annotations */
  argv[1]=Tcl_NewLongObj(totalKeyAnns);     /* Number of Key Annotations */
  argv[2]=Tcl_NewLongObj(totalResponseAnns);/* Number of Responce Annotations */
  if (sortedKeySet) Tcl_IncrRefCount(sortedKeySet);
  if (sortedResSet) Tcl_IncrRefCount(sortedResSet);
  Tcl_DecrRefCount(lower);
  Tcl_DecrRefCount(equal);
  Tcl_DecrRefCount(fuzzyEq);
  Tcl_DecrRefCount(unequal);
  Tcl_DecrRefCount(greater);
  Tcl_SetObjResult(CDM_Interp, Tcl_NewListObj(4, argv));
  return TCL_OK;
error:
  if (argv[0]) Tcl_DecrRefCount(argv[0]);  /* Number of Correct Annotations */
  if (argv[1]) Tcl_DecrRefCount(argv[1]);  /* Number of Key Annotations */
  if (argv[2]) Tcl_DecrRefCount(argv[2]);  /* Number of Responce Annotations */
  if (argv[3]) Tcl_DecrRefCount(argv[3]);
  if (sortedKeySet) Tcl_DecrRefCount(sortedKeySet);
  if (sortedResSet) Tcl_DecrRefCount(sortedResSet);
  Tcl_DecrRefCount(lower);
  Tcl_DecrRefCount(equal);
  Tcl_DecrRefCount(fuzzyEq);
  Tcl_DecrRefCount(unequal);
  Tcl_DecrRefCount(greater);
  return TCL_ERROR;
#undef _CDM_LOG_REPORT_UNEQUAL
#undef _CDM_LOG_REPORT_EQUAL
#undef _CDM_LOG_REPORT_LOWER
#undef _CDM_LOG_REPORT_GREATER
} /* CDM_CompareAnnotationSets */

/*
 * ---- CDM_AddAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will add the given Annotation (specified by the "Ann"
 *     parameter) into the Annotation set specified by the "Set" parameter.
 *     This function (unlike CDM_AddAnnotation for Documents) will not
 *     modify in any way the Id of the provided Annotation: it is the
 *     responsibility of the caller to set a proper Id to the provided
 *     Annotation. The Annotation object will be appended to the Annotation
 *     set even if it does not have an Id (Id = -1).
 *
 *     This function will simply append the provided Annotation object to the
 *     provided Annotation set. It will not check if an Annotation with the
 *     same Id already exists in order to replace it, as CDM_AddAnnotation
 *     will do when it is applied on a Document object.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set:         Annotation Set.                (CDM_AnnotationSet)
 *     Ann:         Annotation.                    (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AddAnnotation (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return a new AnnotationSet object (of
 *     type CDM_AnnotationSet) that will contain all elements of the provided
 *     through the "Set" parameter set (with exactly the same order) plus the
 *     provided Annotation object.
 *
 *     CDM does not own the returned
 *     object: the caller is responsible to use Tcl_DecrRefCount or CDM_Free
 *     in order to free the object and the memory associated with it.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_AddAnnotation
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_AddAnnotation(CDM_AnnotationSet Set, CDM_Annotation Ann) {
  CDM_AnnotationSet NewSet = Set;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AddAnnotation: Aborting...");
    return NULL;
  }

  if (!CDM_ValidAnnotation(Ann)) return NULL;
  if (Tcl_IsShared(NewSet)) {
#ifdef DUPLICATE_DEBUG
    CDM_Debug("AddAnnotation: Duplicating (%d): %s\n",
            Set->refCount, Tcl_GetString(Set));
#endif /* DUPLICATE_DEBUG */
    NewSet = Tcl_DuplicateObj(Set);
  }
  if (NewSet == NULL) return NULL;
  if (Tcl_ListObjAppendElement(CDM_Interp, NewSet, Ann) != TCL_OK) return NULL;
  return NewSet;
} /* CDM_AddAnnotation */

/*
 * ---- CDM_AnnotationsAt:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Annotation set that have the specified text position (as given
 *     by the parameter "Position") as a start position in any of their spans.
 *     CDM does not own the returned
 *     object: the caller is responsible to use Tcl_DecrRefCount or CDM_Free
 *     in order to free the object and the memory associated with it.
 *
 *     This function will always create and return a new Annotation set object,
 *     even when the Annotation set has no Annotations defined or
 *     when no Annotation
 *     is found that includes the specified text position into its span set.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Set:         Annotation Set object.         (CDM_AnnotationSet)
 *      Position:    Position in text.              (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationsAt (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_AnnotationsContaining, CDM_AnnotationsInRange,
 *     CDM_AnnotationsMatchingRange, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_AnnotationsAt
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_AnnotationsAt(CDM_AnnotationSet Set, long Position) {
  CDM_AnnotationSet Result;
  CDM_SpanSet       Spans;
  Tcl_Obj **Ann, **Span;
  register long start;
  register int i,j;
  int items, spans;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationsAt: Aborting...");
    return NULL;
  }

  if (Position < 0) {
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }
  if (Set == NULL) {
    CDM_ReportError("CDM_AnnotationsAt: Invalid Annotation Set");
    return NULL;
  }

  i = Tcl_ListObjGetElements(CDM_Interp, Set, &items, &Ann);
  if (i != TCL_OK) {
    CDM_ReportError("Cannot Get Annotations from Annotation Set");
    return NULL;
  }
  /* If there are no Annotations, return an empty (list) object... */
  if (items < 1) {
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }

  Result = Tcl_NewListObj(0, NULL);
  if (Result == NULL) CDM_Panic("CDM_AnnotationsAt: Tcl_NewListObj returned "
                                "NULL - out of memory?\n");

  for (i=0; i<items; i++) {
    Spans = CDM_GetSpans(Ann[i]);
    if (Spans == NULL) {
      Tcl_DecrRefCount(Result);
      CDM_ReportError("CDM_AnnotationsAt: Annotation Set has Annotation "
                      "without Span?\n");
      CDM_AppendError(Tcl_GetString(Ann[i]));
      return NULL;
    }
    /* Check All Spans... */
    j = Tcl_ListObjGetElements(CDM_Interp, Spans, &spans, &Span);
    if (j != TCL_OK) {
      Tcl_DecrRefCount(Result);
      CDM_ReportError("CDM_AnnotationsAt: Error Getting Spans From Annotation");
      CDM_AppendError(Tcl_GetString(Ann[i]));
      return NULL;
    }
    for (j=0; j<spans; j++) {
      start = CDM_GetStart(Span[j]);
      if (start == Position) {
        Tcl_ListObjAppendElement(CDM_Interp, Result, Ann[i]);
        break; /* We have found at least one... */
      }
    }
  }

  return Result;
} /* CDM_AnnotationsAt */

/*
 * ---- CDM_AnnotationsContaining:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Annotation set that include the specified text position
 *     (as given by the parameter "Position") into their spans.
 *     CDM does not own the returned
 *     object: the caller is responsible to use Tcl_DecrRefCount or CDM_Free
 *     in order to free the object and the memory associated with it.
 *
 *     This function will always create and return a new Annotation set object,
 *     even when the original Annotations set did not have any Annotations
 *     defined, or when no Annotation
 *     is found that includes the specified text position into its span.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Set:         Annotation Set object.         (CDM_AnnotationSet)
 *      Position:    Position in text.              (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationsContaining (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_AnnotationsAt, CDM_AnnotationsInRange,
 *     CDM_AnnotationsMatchingRange, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_AnnotationsContaining
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_AnnotationsContaining(CDM_AnnotationSet Set,
                                            long Position) {
  CDM_AnnotationSet Result;
  Tcl_Obj **Ann;
  register int i, status;
  int items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationsContaining: Aborting...");
    return NULL;
  }

  if (Position < 0) {
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }

  if (Set == NULL) {
    CDM_ReportError("CDM_AnnotationsContaining: Invalid Annotation Set");
    return NULL;
  }

  i = Tcl_ListObjGetElements(CDM_Interp, Set, &items, &Ann);
  if (i != TCL_OK) {
    CDM_ReportError("Cannot Get Annotations from Annotation Set");
    return NULL;
  }
  /* If there are no Annotations, return an empty (list) object... */
  if (items < 1) {
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }

  Result = Tcl_NewListObj(0, NULL);
  if (Result == NULL) CDM_Panic("CDM_AnnotationsContaining: Tcl_NewListObj "
                                "returned NULL - out of memory?\n");

  for (i=0; i<items; i++) {
    status = CDM_AnnotationContainsPosition(Ann[i], Position);
    if (status == -1) {
      Tcl_DecrRefCount(Result);
      return NULL;
    }
    if (status) {
      Tcl_ListObjAppendElement(CDM_Interp, Result, Ann[i]);
    }
  }
  return Result;
} /* CDM_AnnotationsContaining */

/*
 * ---- CDM_AnnotationsContaining:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Annotation set that include the specified texts positions
 *     (as given by the parameters "Position1" and "Position2")
 *     into their spans.
 *     CDM does not own the returned
 *     object: the caller is responsible to use Tcl_DecrRefCount or CDM_Free
 *     in order to free the object and the memory associated with it.
 *
 *     This function will always create and return a new Annotation set object,
 *     even when the original Annotations set did not have any Annotations
 *     defined, or when no Annotation
 *     is found that includes the specified text position into its span.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Set:         Annotation Set object.         (CDM_AnnotationSet)
 *      Position1:   Position in text.              (long)
 *      Position2:   Position in text.              (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationsContaining (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_AnnotationsAt, CDM_AnnotationsInRange,
 *     CDM_AnnotationsMatchingRange, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_AnnotationsContaining
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_AnnotationsContaining(CDM_AnnotationSet Set,
                                            long Position1, long Position2) {
  CDM_AnnotationSet Result;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationsContaining: Aborting...");
    return NULL;
  }

  /*
   * The algorithm is simple: We select all the Annotations that contain the
   * first position, and then we filter them to the ones that include the
   * second position...
   */
  Set = CDM_AnnotationsContaining(Set, Position1);
  if (Set == NULL) return NULL;
  Result = CDM_AnnotationsContaining(Set, Position2);
  Tcl_DecrRefCount(Set);
  return Result;
} /* CDM_AnnotationsContaining */

/*
 * ---- CDM_AnnotationsInRange:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Annotation Set that have at least one span within the specified
 *     range (["Start", "End"]). CDM will not own the returned
 *     object: the caller is responsible to use Tcl_DecrRefCount or CDM_Free
 *     in order to free the object and the memory associated with it.
 *
 *     This function will always create and return a new AnnotationSet object,
 *     even when the original Annotations set did not have any Annotations
 *     defined, or when no Annotation from the given Annotation set
 *     is found to reside inside the specified range.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set:           Annotation Set object.           (CDM_AnnotationSet)
 *     Start:         The start of the desired range.  (long)
 *     End:           The end of the desired range.    (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationsInRange (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_AnnotationsAt, CDM_AnnotationsContaining,
 *     CDM_AnnotationsMatchingRange, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_AnnotationsInRange
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_AnnotationsInRange(CDM_AnnotationSet Set,
                                         long Start, long End) {
  CDM_AnnotationSet Result;
  CDM_SpanSet       Spans;
  Tcl_Obj **Ann, **Span;
  long start, end;
  register int i,j;
  int items, spans;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationsInRange: Aborting...");
    return NULL;
  }

  if (Set == NULL) {
    CDM_ReportError("CDM_AnnotationsInRange: Invalid Annotation Set");
    return NULL;
  }

  i = Tcl_ListObjGetElements(CDM_Interp, Set, &items, &Ann);
  if (i != TCL_OK) {
    CDM_ReportError("Cannot Get Annotations from Annotation Set");
    return NULL;
  }
  /* If there are no Annotations, return an empty (list) object... */
  if (items < 1) {
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }

  Result = Tcl_NewListObj(0, NULL);
  if (Result == NULL) CDM_Panic("CDM_AnnotationsInRange: Tcl_NewListObj "
                                "returned NULL - out of memory?\n");

  for (i=0; i<items; i++) {
    Spans = CDM_GetSpans(Ann[i]);
    if (Spans == NULL) {
      Tcl_DecrRefCount(Result);
      CDM_ReportError("CDM_AnnotationsContaining: Annotation Set has "
                      "Annotation without Span?\n");
      CDM_AppendError(Tcl_GetString(Ann[i]));
      return NULL;
    }
    /* Check All Spans... */
    j = Tcl_ListObjGetElements(CDM_Interp, Spans, &spans, &Span);
    if (j != TCL_OK) {
      Tcl_DecrRefCount(Result);
      CDM_ReportError("CDM_AnnotationsInRange: Error Getting Spans From "
                      "Annotation");
      CDM_AppendError(Tcl_GetString(Ann[i]));
      return NULL;
    }
    for (j=0; j<spans; j++) {
      if (CDM_GetSpanOffsets(Span[j], &start, &end) != TCL_OK) {
        Tcl_DecrRefCount(Result);
        CDM_ReportError("CDM_AnnotationsInRange: Error Getting Span %d From "
                        "Annotation", j);
        CDM_AppendError(Tcl_GetString(Ann[i]));
        return NULL;
      }
      if (start >= Start && end <= End) {
        Tcl_ListObjAppendElement(CDM_Interp, Result, Ann[i]);
        break; /* We have found at least one... */
      }
    }
  }
  return Result;
} /* CDM_AnnotationsInRange */

/*
 * ---- CDM_AnnotationsMatchingRange:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Annotation Set that have at least one span that exactly
 *     matches the specified range (["Start", "End"]).
 *     CDM will not own the returned
 *     object: the caller is responsible to use Tcl_DecrRefCount or CDM_Free
 *     in order to free the object and the memory associated with it.
 *
 *     This function will always create and return a new AnnotationSet object,
 *     even when the original Annotations set did not have any Annotations
 *     defined, or when no Annotation from the given Annotation set
 *     is found to reside inside the specified range.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set:           Annotation Set object.           (CDM_AnnotationSet)
 *     Start:         The start of the desired range.  (long)
 *     End:           The end of the desired range.    (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_AnnotationsMatchingRange (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_AnnotationsAt, CDM_AnnotationsContaining, CDM_AnnotationsInRange,
 *     CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_AnnotationsMatchingRange
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_AnnotationsMatchingRange(CDM_AnnotationSet Set,
                                         long Start, long End) {
  CDM_AnnotationSet Result;
  Tcl_Obj **Ann;
  register int i, status;
  int items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AnnotationsMatchingRange: Aborting...");
    return NULL;
  }

  if (Set == NULL) {
    CDM_ReportError("CDM_AnnotationsMatchingRange: Invalid Annotation Set");
    return NULL;
  }

  i = Tcl_ListObjGetElements(CDM_Interp, Set, &items, &Ann);
  if (i != TCL_OK) {
    CDM_ReportError("Cannot Get Annotations from Annotation Set");
    return NULL;
  }
  /* If there are no Annotations, return an empty (list) object... */
  if (items < 1) {
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }

  Result = Tcl_NewListObj(0, NULL);
  if (Result == NULL) CDM_Panic("CDM_AnnotationsMatchingRange: Tcl_NewListObj "
                                "returned NULL - out of memory?\n");

  for (i=0; i<items; i++) {
    status = CDM_AnnotationMatchesRange(Ann[i], Start, End);
    if (status == -1) {
      Tcl_DecrRefCount(Result);
      return NULL;
    }
    if (status) {
      Tcl_ListObjAppendElement(CDM_Interp, Result, Ann[i]);
    }
  }
  return Result;
} /* CDM_AnnotationsMatchingRange */

/*
 * ---- CDM_DisplaceAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function displases or "moves" the Annotations of the specified
 *     Annotation set by "displacement" characters.
 *     This function will iterate over
 *     all Annotations contained in the Annotation set. For each Annotation, all
 *     spans contained in the span set will be examined: for each offset in
 *     the span (either start or end), if it is equal or greater than the value
 *     provided by the "offset" parameter, then the value of the
 *     "displacement" parameter will be added.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set:          Annotation Set object                   (CDM_AnnotationSet)
 *     offset:    A character offset after which all offsets are modified (long)
 *     displacement: The number of characters that offsets will be moved  (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_DisplaceAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet), that contains all the Annotations contained in the
 *     original set, where some of them may be displaced.
 *     The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free when it is not needed any more.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_DisplaceAnnotation, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_DisplaceAnnotations
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_DisplaceAnnotations(CDM_AnnotationSet Set, long offset,
                                          long displacement) {
  CDM_AnnotationSet NewSet;
  CDM_Annotation ann;
  Tcl_Obj **Ann;
  int status, items, i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_DisplaceAnnotations: Aborting...");
    return NULL;
  }

  if (Set == NULL) {
    CDM_ReportError("CDM_DisplaceAnnotations: Invalid Annotation Set");
    return NULL;
  }

  /*
   * If displacement is 0, there is nothing to do...
   */
  if (displacement == 0) return Tcl_DuplicateObj(Set);

  /*
   * Iterate over all Annotations...
   */
  status = Tcl_ListObjGetElements(CDM_Interp, Set, &items, &Ann);
  if (status != TCL_OK) return NULL;

  NewSet = Tcl_NewListObj(0, NULL);
  if (NewSet == NULL) return NULL;

  for (i=0; i<items; i++) {
    ann = CDM_DisplaceAnnotation(Ann[i], offset, displacement);
    if (ann == NULL) return NULL;
    /* CDM_DisplaceAnnotation will return the given Annotation if a
     * displacement was not needed... */
    if (ann != Ann[i]) NumberOfDisplacedAnnotations++;
    if (Tcl_ListObjAppendElement(CDM_Interp, NewSet, ann) != TCL_OK) {
      CDM_Free(NewSet);
      return NULL;
    }
  }

  /*
   * We have displaced our Annotations, an action that may have generated to
   * many unneeded entries in our memory cache for long integers. Try to
   * remove excess (unused) entries...
   * This may take some more time, but we save large amounts of memory.
   * Imagine making a few thousants displacements without clearing this
   * cache...
   */
  if (NumberOfDisplacedAnnotations >
      CDM_MAX_NUMBER_OF_DISPLACEMENTS_BEFORE_CLEAN) {
    ObjectCache.RemoveUnusedLongs();
    NumberOfDisplacedAnnotations = 0;
  }

  return NewSet;
} /* CDM_DisplaceAnnotations */

/*
 * ---- CDM_FirstAnnotationContaining:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation object (of type
 *     CDM_Annotation) that will be the first Annotation (with Annotations
 *     sorted by their Id) of the
 *     specified Annotation set that includes the specified text position
 *     (as given by the parameter "Position") into its spans.
 *     If no Annotation is found
 *     that includes the specified position, NULL will be returned and
 *     the empty object will be placed as a result in the current active
 *     Tcl interpreter (CDM_Interp). The caller can use the function
 *     CDM_ActiveInterpreterResultEmpty to check whether the
 *     current active Tcl interpreter (CDM_Interp) has as result the empty
 *     object.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set:         Annotation Set object.        (CDM_AnnotationSet)
 *     Position:    Position in text.             (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_FirstAnnotationContaining (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns an Annotation object (of type
 *    CDM_Annotation). The returned object is owned by CDM.
 *
 *    If an error occurs, then NULL will be returned and an error message
 *    will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_FirstAnnotationAt, CDM_FirstAnnotationInRange, CDM_FirstAnnotationMatchingRange,
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_FirstAnnotationContaining
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_FirstAnnotationContaining(CDM_AnnotationSet Set,
                                                          const long Position) {
  CDM_SpanSet Spans;
  Tcl_Obj     **Ann, **Span;
  long start, end;
  int items, ann_items, i, j, status;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_FirstAnnotationContaining: Aborting...");
    return NULL;
  }

  if (Position < 0) {
    CDM_ReportError("CDM_FirstAnnotationContaining: invalid position %d",
                    Position);
    return NULL;
  }

  if (Set == NULL) {
    CDM_ReportError("CDM_FirstAnnotationContaining: Invalid Annotation Set");
    return NULL;
  }

  status = Tcl_ListObjGetElements(CDM_Interp, Set, &ann_items, &Ann);
  if (status != TCL_OK) return NULL;

  for (i=0; i < ann_items; i++) {
    if (Ann[i] == NULL) continue;
#ifdef ENABLE_ANNOTATION_OBJECT
    if (ANNOTATION_OBJ(Ann[i]) && CDMAnnObj_GetStart(Ann[i]) != -1) {
      /* This Annotation has a single span. Handle it directly... */
      start = CDMAnnObj_GetStart(Ann[i]);
      end   = CDMAnnObj_GetEnd(Ann[i]);
      if (start <= Position && Position <= end) {
        return Ann[i];
      }
    } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
      /* Either the Annotation object type is disabled, or the Annotation has
       * more than one Span... */
      Spans = CDM_GetSpans(Ann[i]);
      if (Spans == NULL) {
        CDM_ReportError("CDM_FirstAnnotationContaining: Annotation Set has "
                        "Annotation without Span?\n");
        CDM_AppendError(Tcl_GetString(Ann[i]));
        return NULL;
      }
      status = Tcl_ListObjGetElements(CDM_Interp, Spans, &items, &Span);
      if (status != TCL_OK) {
        CDM_ReportError("CDM_FirstAnnotationContaining: Error Getting Spans "
                        "From Annotation ");
        CDM_AppendError(Tcl_GetString(Ann[i]));
        return NULL;
      }
      for (j=0; j<items; j++) {
        if (CDM_GetSpanOffsets(Span[j], &start, &end) != TCL_OK) {
          CDM_ReportError("CDM_FirstAnnotationContaining: Error Getting Span "
                          "%d From Annotation", j);
          CDM_AppendError(Tcl_GetString(Ann[i]));
          return NULL;
        }
        if (start <= Position && Position <= end) {
          return Ann[i];
        }
      }
#ifdef ENABLE_ANNOTATION_OBJECT
    }
#endif /* ENABLE_ANNOTATION_OBJECT */
  } /* for (i=0; i <= Document->AnnotationMaxID; i++) */
  return NULL;
} /* CDM_FirstAnnotationContaining */

/*
 * ---- CDM_FirstAnnotationContaining:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation object (of type
 *     CDM_Annotation) that will be the first Annotation (with Annotations
 *     sorted by their Id) of the
 *     specified Document that include the two specified text positions
 *     (as given by the parameters "Position1" and "Position2")
 *     into ts spans. If no Annotation is found
 *     that includes the specified positions, NULL will be returned and
 *     the empty object will be placed as a result in the current active
 *     Tcl interpreter (CDM_Interp). The caller can use the function
 *     CDM_ActiveInterpreterResultEmpty to check whether the
 *     current active Tcl interpreter (CDM_Interp) has as result the empty
 *     object.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set:         Annotation Set object.        (CDM_AnnotationSet)
 *     Position1:   Position in text.             (long)
 *     Position2:   Position in text.             (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_FirstAnnotationContaining (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation object (of type
 *     CDM_Annotation). The returned object is owned by CDM.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_FirstAnnotationAt, CDM_FirstAnnotationInRange, CDM_FirstAnnotationMatchingRange,
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_FirstAnnotationContaining
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_FirstAnnotationContaining(CDM_AnnotationSet Set,
                                   const long Position1, const long Position2) {
  CDM_SpanSet Spans;
  Tcl_Obj     **Ann, **Span;
  long start, end;
  int items, ann_items, i, j, status;
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_FirstAnnotationContaining: Aborting...");
    return NULL;
  }

  if (Position1 < 0) {
    CDM_ReportError("CDM_FirstAnnotationContaining: invalid position %d",
                    Position1);
    return NULL;
  }
  if (Position2 < 0) {
    CDM_ReportError("CDM_FirstAnnotationContaining: invalid position %d",
                    Position2);
    return NULL;
  }

  if (Set == NULL) {
    CDM_ReportError("CDM_FirstAnnotationContaining: Invalid Annotation Set");
    return NULL;
  }

  status = Tcl_ListObjGetElements(CDM_Interp, Set, &ann_items, &Ann);
  if (status != TCL_OK) return NULL;

  for (i=0; i < ann_items; i++) {
    if (Ann[i] == NULL) continue;
#ifdef ENABLE_ANNOTATION_OBJECT
    if (ANNOTATION_OBJ(Ann[i]) && CDMAnnObj_GetStart(Ann[i]) != -1) {
      /* This Annotation has a single span. Handle it directly... */
      start = CDMAnnObj_GetStart(Ann[i]);
      end   = CDMAnnObj_GetEnd(Ann[i]);
      if (start <= Position1 && Position1 <= end &&
          start <= Position2 && Position2 <= end) {
        return Ann[i];
      }
    } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
      /* Either the Annotation object type is disabled, or the Annotation has
       * more than one Span... */
      Spans = CDM_GetSpans(Ann[i]);
      if (Spans == NULL) {
        CDM_ReportError("CDM_FirstAnnotationContaining: Annotation Set has "
                        "Annotation without Span?\n");
        CDM_AppendError(Tcl_GetString(Ann[i]));
        return NULL;
      }
      status = Tcl_ListObjGetElements(CDM_Interp, Spans, &items, &Span);
      if (status != TCL_OK) {
        CDM_ReportError("CDM_FirstAnnotationContaining: Error Getting Spans "
                        "From Annotation ");
        CDM_AppendError(Tcl_GetString(Ann[i]));
        return NULL;
      }
      for (j=0; j<items; j++) {
        if (CDM_GetSpanOffsets(Span[j], &start, &end) != TCL_OK) {
          CDM_ReportError("CDM_FirstAnnotationContaining: Error Getting Span "
                          "%d From Annotation", j);
          CDM_AppendError(Tcl_GetString(Ann[i]));
          return NULL;
        }
        if (start <= Position1 && Position1 <= end &&
            start <= Position2 && Position2 <= end) {
          return Ann[i];
        }
      }
#ifdef ENABLE_ANNOTATION_OBJECT
    }
#endif /* ENABLE_ANNOTATION_OBJECT */
  } /* for (i=0; i < ann_items; i++) */
  return NULL;
} /* CDM_FirstAnnotationContaining */

/*
 * ---- CDM_GetAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Annotation object that has as Id the value
 *     specified by the "Id" parameter. If the requested Annotation does not
 *     exist, an error will be returned.
 *
 *     Note that this function will also return an error if an Annotation
 *     object is encoutered that does not have an Id (as Annotation objects
 *     with empty Ids are permitted inside Annotation set objects).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set:         Annotation Set.                (CDM_AnnotationSet)
 *     Id:          Annotation's Id.               (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_GetAnnotation (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation object (of type CDM_Annotation) if
 *     an Annotation having as Id the value of the "Id" parameter is found.
 *     In case that the specified Annotation does not exist or an Annotation
 *     with an empty Id is found, NULL is returned.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or directly modified in any way by the caller. Subsequent calls of
 *     either the C++ or Tcl API can modify the returned object. If the caller
 *     wants to keep a reference to the current state of the returned object,
 *     he must place such a request by using Tcl_IncrRefCount. When the caller
 *     wants to release the object (and return it to CDM) after he has placed
 *     a request with [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]], the functions
 *     Tcl_DecrRefCount or CDM_Free must be used.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]], [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_GetAnnotation
 *   >:&:
 */
CDM_EXPORT
CDM_Annotation CDM_GetAnnotation(CDM_AnnotationSet Set, long Id) {
  Tcl_Obj **Anns;
  long id;
  int i, items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetAnnotation: Aborting...");
    return NULL;
  }

  if (Set == NULL) {
    CDM_ReportError("CDM_GetAnnotation: NULL Set Pointer!");
    return NULL;
  }

  /* Get Set's List Elements... */
  i = Tcl_ListObjGetElements(CDM_Interp, Set, &items, &Anns);
  if (i != TCL_OK) {
    CDM_ReportError("Invalid Annotation Set");
    return NULL;
  }

  /* Search Annotations... */
  for (i=0; i<items; i++) {
    id = CDM_GetId(Anns[i]);
    if (id < 0) return NULL;
    if (id == Id) return Anns[i];
  }
  return NULL;
} /* CDM_GetAnnotation */


/*
 * ---- CDM_NextAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Annotation set that their spans ranges are after
 *     the specified text position (as given by the parameter "Position").
 *     CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *
 *     This function will always create and return a new Annotation set object,
 *     even when the Annotation set has no Annotations defined or no Annotations
 *     exist that all their spans are beneath the specified text position.
 *     In this case an empty Annotation set will be returned.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Set:         Annotation Set.                (CDM_AnnotationSet)
 *      Position:    Requested position             (int)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_NextAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object
 *     through Tcl_DecrRefCount or CDM_Free.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_NextAnnotations
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_NextAnnotations(CDM_AnnotationSet Set, long Position) {
  CDM_SpanSet Spans;
  Tcl_Obj **Ann, **Span;
  register long start, min = -1;
  register int i,j;
  int items, spans;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_NextAnnotations: Aborting...");
    return NULL;
  }

  if (Set == NULL) {
    CDM_ReportError("CDM_NextAnnotations: Invalid Annotation Set");
    return NULL;
  }

  i = Tcl_ListObjGetElements(CDM_Interp, Set, &items, &Ann);
  if (i != TCL_OK) {
    CDM_ReportError("Cannot Get Annotations from Annotation Set");
    return NULL;
  }
  /* If there are no Annotations, return an empty (list) object... */
  if (items < 1) {
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }

  for (i=0; i<items; i++) {
    Spans = CDM_GetSpans(Ann[i]);
    if (Spans == NULL) {
      CDM_ReportError("CDM_NextAnnotations: Annotation Set has Annotation "
                      "without Span?\n");
      CDM_AppendError(Tcl_GetString(Ann[i]));
      return NULL;
    }
    /* Check All Spans... */
    j = Tcl_ListObjGetElements(CDM_Interp, Spans, &spans, &Span);
    if (j != TCL_OK) {
      CDM_ReportError("CDM_NextAnnotationsAt: Error Getting Spans From "
                      "Annotation");
      CDM_AppendError(Tcl_GetString(Ann[i]));
      return NULL;
    }
    for (j=0; j<spans; j++) {
      start = CDM_GetStart(Span[j]);
      if (start >= Position) {
        if (start < min || min == -1) min = start;
      }
    }
  }

  if (min == -1) {
    Tcl_IncrRefCount(&CDM_Obj_NULL); /* The user will free it... */
    return &CDM_Obj_NULL;
  }
  return CDM_AnnotationsAt(Set, min);
} /* CDM_NextAnnotations */

/*
 * ---- CDM_Length:
 *  DESCRIPTION:
 *   <English:
 *    This function will return the number of elements contained in a set
 *    of objects. This set can be any CDM set object (for example any object
 *    of type CDM_AnnotationSet, CDM_AttributeSet or CDM_SpanSet) or a Tcl
 *    list object (of type Tcl_Obj *),
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Set: The set object.  (CDM_AnnotationSet or CDM_AttributeSet or CDM_SpanSet or Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to the "llength" Tcl command.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an integer representing the number of objects
 *     contained in the provided set object. If the provided set is empty, 0
 *     will be returned. In case of an error (i.e. the provided set object is
 *     not a valid set object), -1 is returned and an error will be reported
 *     in the current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetLength
 *   >:&:
 */
CDM_EXPORT
int CDM_Length(Tcl_Obj *Set) {
  int length, result;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_Length: Aborting...");
    return -1;
  }

  if (Set == NULL) return 0;
  result = Tcl_ListObjLength(CDM_Interp, Set, &length);
  if (result != TCL_OK) {
    CDM_ReportError("Malformed Set:\"%s\"", Tcl_GetString(Set));
    return -1;
  }
  return length;
} /* CDM_Length */

/*
 * ---- CDM_Nth:
 *  DESCRIPTION:
 *   <English:
 *    Return Nth element of a set...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Set:    Set Pointer      (Tcl_Obj *)
 *      n:      Item number      (long)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
Tcl_Obj *CDM_Nth(Tcl_Obj *Set, long n) {
  Tcl_Obj *ptr = NULL;
  int result;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_Nth: Aborting...");
    return NULL;
  }
  if (Set == NULL) return NULL;

  if (n < 0) n = 0;
  result = Tcl_ListObjIndex(CDM_Interp, Set, (int) n, &ptr);
  if (result != TCL_OK) {
    CDM_ReportError("Malformed Set \"%s\" or index out borders!",
                   Tcl_GetString(Set));
    return NULL;
  }
  if (ptr == NULL) {
    result = -1;
    Tcl_ListObjLength(CDM_Interp, Set, &result);
    if (result > 0) result = Tcl_ListObjIndex(CDM_Interp, Set, --result, &ptr);
    if (result != TCL_OK || ptr == NULL) return NULL;
  }
  return ptr;
} /* CDM_Nth */

/*
 * ---- CDM_MergeAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function accepts two Annotation set objects (both of type
 *     CDM_AnnotationSet) and returns a new Annotation set object that is the
 *     concatenation of the two provided sets. In case of an error, NULL will
 *     be returned and an error will be left at the current active interpeter
 *     (CDM_Interp).
 *
 *     Note that CDM does not own the returned object:
 *     the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Set1: First Set  (CDM_AnnotationSet)
 *      Set2: Second Set (CDM_AnnotationSet)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_MergeAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return a new Annotation set object (of
 *     type CDM_AnnotationSet) that will contain the merge of the two provided
 *     Annotation set objects.
 *
 *     If an error occurs, then NULL will be returned and an error message
 *     will be left at the current active Tcl interpreter (CDM_Interp).
 *
 *     The caller is responsible to free this object through Tcl_DecrRefCount
 *     or CDM_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_MergeAnnotations
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_MergeAnnotations(
    CDM_AnnotationSet Set1, CDM_AnnotationSet Set2) {
  CDM_AnnotationSet Set;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_MergeAnnotations: Aborting...");
    return NULL;
  }

  Set = Tcl_NewListObj(0, NULL);
  if (Set == NULL)
    CDM_Panic("CDM_MergeAnnotations: Tcl_NewListObj returned NULL "
              "- out of memory?\n");
  if (Tcl_ListObjAppendList(CDM_Interp, Set, Set1) != TCL_OK) goto error;
  if (Tcl_ListObjAppendList(CDM_Interp, Set, Set2) != TCL_OK) goto error;
  return Set;
error:
  Tcl_DecrRefCount(Set);
  CDM_ReportError("Cannot Merge Annotations: Malformed Sets?");
  return NULL;
} /* CDM_MergeAnnotations */


/*
 * ---- CDM_RemoveAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will remove the Annotation object that has as Id the value
 *     specified by the "Id" parameter from the provided Annotation set object.
 *     This function will construct and return a new Annotation set object
 *     containing all Annotations of the original Annotation set object,
 *     except the Annotation having as Id the specified by the parameter "Id"
 *     value.
 *     If the provided Annotation set does not contain an Annotation with the
 *     specified Id, then NULL will be returned signing an error.
 *
 *     Note that CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Set:  A set of Annotations.                       (CDM_AnnotationSet)
 *      Id:   The Id of the Annotation to be deleted.     (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_RemoveAnnotation (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return an Annotation set object (of type
 *     CDM_AnnotationSet), that will contain all the Annotations of the
 *     provided by the parameter "Set" Annotation set object, except the
 *     Annotation having as Id the specified value of the "Id" parameter.
 *     It is an error for the original Annotation set object to not
 *     contain an Annotation with Id the specified Id value.
 *
 *     Note that the caller is responsible to free this object through
 *     Tcl_DecrRefCount or CDM_Free.
 *
 *     In case of an error, NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_RemoveAnnotation
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_RemoveAnnotation(CDM_AnnotationSet Set, long Id) {
  Tcl_Obj **Ann, *NewSet;
  int result, elements, i, ann_found = 0;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_RemoveAnnotation: Aborting...");
    return NULL;
  }

  /*
   * Get the Annotation Set elements...
   */
  result = Tcl_ListObjGetElements(CDM_Interp, Set, &elements, &Ann);
  if (result != TCL_OK) return NULL;

  NewSet = Tcl_NewListObj(0, NULL);
  if (NewSet == NULL) CDM_Panic("CDM_RemoveAnnotation: Tcl_NewListObj returned "
                                "NULL - out of memory?\n");
  for (i=0; i<elements; i++) {
    if (CDM_GetId(Ann[i]) == Id) {
      if (ann_found) {
        CDM_ReportError("CDM_RemoveAnnotation: Annotation Set contains more "
                        "than one Annotation with Id=%ld?", Id);
        CDM_Free(NewSet);
        return NULL;
      }
      ann_found = 1;
      continue;
    }
    Tcl_ListObjAppendElement(CDM_Interp, NewSet, Ann[i]);
  }
  if (!ann_found) {
    CDM_ReportError("CDM_RemoveAnnotation: Annotation Set does not contain "
                    "Annotation with Id=%ld", Id);
    CDM_Free(NewSet);
    return NULL;
  }
  return NewSet;
} /* CDM_RemoveAnnotation */


/*
 * ---- CDM_SelectAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Annotation set that their type is the same as the value of the
 *     "Type" parameter. This function will always create and return a new
 *     object, even when no Annotations of the requested type are found or
 *     when the Annotation set is empty. In this case an empty
 *     object will be returned.
 *
 *     Note that CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set:   A set of Annotations.                        (CDM_AnnotationSet)
 *     Type:  The type of the Annotations to be selected.  (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SelectAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return an AnnotationSet object, that will
 *     contain all the Annotations of the specified Annotation set that have as
 *     type the value of the "Type" parameter.
 *
 *     Note that the caller is responsible to free this object through
 *     Tcl_DecrRefCount or CDM_Free.
 *
 *     In case of an error, NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_SelectAnnotations
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_SelectAnnotations(CDM_AnnotationSet Set, char *Type) {
  Tcl_Obj **Ann;
  CDM_AnnotationSet ResSet;
  int result, elements;
  long i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_SelectAnnotations: Aborting...");
    return NULL;
  }

  /*
   * Get the Annotation Set elements...
   */
  result = Tcl_ListObjGetElements(CDM_Interp, Set, &elements, &Ann);
  if (result != TCL_OK) return NULL;

  ResSet = Tcl_NewListObj(0, NULL);
  for (i=0; i < elements; i++) {
    if (!CDM_ValidAnnotation(Ann[i])) {
      CDM_Free(ResSet);
      return NULL;
    }
    if (strcmp(CDM_GetType(Ann[i]), Type) == 0) {
      Tcl_ListObjAppendElement(CDM_Interp, ResSet, Ann[i]);
    }
  }
  return ResSet;
} /* CDM_SelectAnnotations */


/*
 * ---- CDM_SelectAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Annotation set that their type is the same as the value of the
 *     "Type" parameter and the provided boolean expression (specified through
 *     the "Constraints" parameter) is true for every one of these Annotations.
 *     This function will always create and return a new
 *     object, even when no Annotations of the requested type are found or
 *     when the Annotation set is empty. In this case an empty
 *     object will be returned.
 *
 *     The boolean expression that can be specified through the "Constraints"
 *     parameter can be any valid Tcl boolean expression (i.e. a boolean
 *     expression that will be accepted by the "expr" Tcl command), with the
 *     following excepion:
 *
 *     [[LI]]The boolean expression can contain references to the values of
 *     Atttributes of the Annotation, by utilising the notation
 *     "ann::%%%lt;Attribute Name%%%gt;". The only limitation is that
 *     "%%%lt;Attribute Name%%%gt;" cannot exceed 120 characters. For example,
 *     the following code will create a new Annotation set that will contain
 *     all Annotations from the specified Annotation set that their type is
 *     token,
 *     they have a "type" Attribute which has as value "EFW" and also have a
 *     "pos" Attribute that has a value other than "NN":[[/LI]]
 *     [[br]][[code]][[center]]
 *     CDM_SelectAnnotations(Doc, "token", "ann::type == \"EFW\" %%%amp;%%%amp; ann::pos != \"NN\"");
 *     [[/center]][[/code]][[br]]
 *
 *     Note that CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set:          A set of Annotations.             (CDM_AnnotationSet)
 *     Type:         The type of the Annotations to be selected.  (char *)
 *     Constraints:  A boolean expression.                        (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SelectAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return an AnnotationSet object, that will
 *     contain all the Annotations of the specified Annotation set that have as
 *     type
 *     the value of the "Type" parameter and the provided by the "Constraints"
 *     parameter boolean expression was true for these Annotations.
 *
 *     Note that the caller is responsible to free this object through
 *     Tcl_DecrRefCount or CDM_Free.
 *
 *     In case of an error, NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_SelectAnnotations,
 *     [[A HREF="../TclTk/TclCmd/expr.htm"]]The Tcl "expr" manual[[/A]].
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_SelectAnnotations(CDM_AnnotationSet Set, char *Type,
                                        char *Constraints) {
  Tcl_Obj **Ann, *AttributeNames, *Expr, *Boolean, **Name;
  Tcl_Parse parsePtr;
  Tcl_Token *token;
  CDM_AnnotationSet ResSet;
  char *ptr, *qptr, *start, name[128], TclSafeName[256];
  int BoolVal, status, NameNu, j, elements, RejectedAnnPrefix;
  long i;
  CDM_THREAD_SPECIFIC_DATA;

  /* No need for complicated things if no expression... */
  if (Constraints == NULL || strlen(Constraints) == 0 ||
      strcmp(Constraints, "{}") == 0) {
    return CDM_SelectAnnotations(Set, Type);
  }

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_SelectAnnotations: Aborting...");
    return NULL;
  }

  /*
   * Get the Annotation Set elements...
   */
  BoolVal = Tcl_ListObjGetElements(CDM_Interp, Set, &elements, &Ann);
  if (BoolVal != TCL_OK) return NULL;

  /*
   * Since we want to offer Attribute tokens for the Attribute Annotations, we
   * have to preprocess the boolean expresion. This is done in the following
   * steps:
   * 1) First, make sure the array we use for storing temporary values is
   *    clean. Ignore any errors...
   */
  Tcl_UnsetVar2(CDM_Interp, "::CDM::CDM_SelectAnnotationsArray", NULL, 0);
  /*
   * 2) Foreach Annotation Attribute name (denoted with ann::<name>), create
   * an entry in the array, and keep its name for further reference...
   */
  start   = Constraints;
  ptr     = strstr(Constraints, "ann::");
  if (ptr != NULL) {
    /* Try to find the previous non-white character... */
    for (qptr=ptr-1; (*qptr==' ' || *qptr=='\t') && qptr>=Constraints; qptr--);
    if (*qptr == '\"' || *qptr == '{') ptr = qptr;
  }
  AttributeNames = Tcl_NewListObj(0, NULL);
  Expr    = Tcl_NewStringObj(NULL, -1);
  // Tcl_IncrRefCount(Expr);
  // printf("Original: {%s}\n", Constraints);
  while (ptr != NULL) {
    /* Copy unchanged bytes... */
    Tcl_AppendToObj(Expr, start, ptr - start);
    RejectedAnnPrefix = 0;
    /* Try to get the ann::<attribute name> substring. This is not an easy
     * task, as it may be enclosed in "" or {}. Use Tcl suitable routines to
     * parse such cases...
     */
    if (*ptr == '{' || *ptr == '\"' || *(ptr+5) == '{' || *(ptr+5) == '\"') {
      if (*ptr == '{') {
        status = Tcl_ParseBraces(CDM_Interp, ptr, -1, &parsePtr, 0,
                                 (const char **) &qptr);
      } else if (*ptr == '\"') {
        status = Tcl_ParseQuotedString(CDM_Interp, ptr, -1, &parsePtr, 0,
                                       (const char **) &qptr);
      } else if (*(ptr+5) == '{') {
        status = Tcl_ParseBraces(CDM_Interp, ptr+5, -1, &parsePtr, 0,
                                 (const char **) &qptr);
        RejectedAnnPrefix = 1;
      } else {
        status = Tcl_ParseQuotedString(CDM_Interp, ptr+5, -1, &parsePtr, 0,
                                       (const char **) &qptr);
        RejectedAnnPrefix = 1;
      }
      if (status != TCL_OK) {
        CDM_AppendError("\nCDM_SelectAnnotations: while evaluating expression "
                        "\n  (%s)", Constraints);
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        return NULL;
      }
      if (parsePtr.numTokens != 1) {
        CDM_ReportError("CDM_SelectAnnotations: invalid ann::<attribute name> "
                        "reference in constraints expression:\n  (%s)", ptr);
        Tcl_FreeParse(&parsePtr);
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        return NULL;
      }
      token = parsePtr.tokenPtr;
      if (token->type != TCL_TOKEN_TEXT) {
        CDM_ReportError("CDM_SelectAnnotations: illegal ann::<attribute name> "
                        "reference in constraints expression:\n  (%s)", ptr);
        Tcl_FreeParse(&parsePtr);
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        return NULL;
      }
      if (token->size > 127) {
        CDM_ReportError("CDM_SelectAnnotations: too long ann::<attribute name> "
                        "reference. <attribute name> must be < 128 characters");
        Tcl_FreeParse(&parsePtr);
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        return NULL;
      }
      if (RejectedAnnPrefix) {
        strncpy(name, token->start, token->size);
        name[token->size] = '\0';
      } else {
        strncpy(name, token->start+5, token->size-5);
        name[token->size-5] = '\0';
      }
      Tcl_FreeParse(&parsePtr);
      start = qptr;
    } else {
      ptr += 5; status = 0;
      while (!CHAR_TYPE(*ptr) && status < 128) {
        name[status++] = *ptr; ++ptr;
      }
      name[status] = '\0';
      start = ptr;
    }
    /*
     * Given the ann::<attribute name> reference, try to construct a safe for
     * Tcl name that can be used in the expression so as to correctly refer to
     * the proper array element...
     */
    for (qptr = name, status = 0; *qptr != '\0'; qptr++) {
      if (*qptr == ' ' || *qptr == '\t' || *qptr == ';' || *qptr == '\"' ||
          *qptr == '{' || *qptr == '}' || *qptr == '[' || *qptr == ']' ||
          *qptr == '\\') {
        TclSafeName[status++] = '\\';
        TclSafeName[status++] = *qptr;
      } else {
        TclSafeName[status++] = *qptr;
      }
    }
    TclSafeName[status] = *qptr;

    /* Is the variable already defined? */
    if (Tcl_GetVar2Ex(CDM_Interp, "::CDM::CDM_SelectAnnotationsArray",
        name, 0) == NULL) {
      /* The first time we meet this name... */
      Tcl_ListObjAppendElement(CDM_Interp, AttributeNames,
                               CDM_NewStringObj(name));
      Tcl_SetVar2Ex(CDM_Interp, "::CDM::CDM_SelectAnnotationsArray",
                               name, &CDM_Obj_NULL, 0);
    }
    /* Substitute ann::<name> with the proper array element... */
    Tcl_AppendStringsToObj(Expr, "$::CDM::CDM_SelectAnnotationsArray(",
                                 TclSafeName, ")", (char *) NULL);
    ptr   = strstr(start, "ann::");
    if (ptr != NULL) {
      /* Try to find the previous non-white character... */
      for (qptr=ptr-1; (*qptr==' ' || *qptr=='\t')&& qptr>=Constraints; qptr--);
      if (*qptr == '\"' || *qptr == '{') ptr = qptr;
    }
  }
  /* Finally, add the rest of the expression... */
  Tcl_AppendToObj(Expr, start, -1);
  //printf("Compiled: {%s}\n", Tcl_GetString(Expr));

  /* Create the new Annotation set... */
  ResSet = Tcl_NewListObj(0, NULL);
  /* Place the various Attribute name objects in an array... */
  Tcl_ListObjGetElements(NULL, AttributeNames, &NameNu, &Name);
  for (i=0; i <elements; i++) {
    if (!CDM_ValidAnnotation(Ann[i])) {
      CDM_Free(AttributeNames);
      CDM_Free(Expr);
      CDM_Free(ResSet);
      return NULL;
    }
    if (strcmp(CDM_GetType(Ann[i]), Type) == 0) {
      /*
       * The Annotation has the requested type. It's time to apply the boolean
       * expession: If evaluates to true, we append the Annotation to the
       * Annotation set...
       */

      /*
       * Now, we have to fill the array with the correct values. In case an
       * attribute does not exist, we place an empty value...
       */
      for (j=0; j<NameNu; j++) {
        Boolean = CDM_GetAttribute(Ann[i], Tcl_GetString(Name[j]));
        if (Boolean == NULL) {
          /* Set the empty value... */
          Tcl_ObjSetVar2(CDM_Interp,
                         CDM_NewStringObj("::CDM::CDM_SelectAnnotationsArray"),
                         Name[j], &CDM_Obj_NULL, 0);
        } else {
          Boolean = CDM_GetValueValue(Boolean);
          Tcl_ObjSetVar2(CDM_Interp,
                         CDM_NewStringObj("::CDM::CDM_SelectAnnotationsArray"),
                         Name[j], Boolean, 0);
        }
      }
      if (Tcl_ExprObj(CDM_Interp, Expr, &Boolean) != TCL_OK) {
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        CDM_Free(ResSet);
        CDM_AppendError("\nCDM_SelectAnnotations: invalid boolean expression\n"
                        "  (type=\"%s\", constraints=\"%s\")",
                        Type, Constraints);
        return NULL;
      }
      /* Was it true? */
      status = Tcl_GetBooleanFromObj(CDM_Interp, Boolean, &BoolVal);
      CDM_Free(Boolean);
      if (status != TCL_OK) {
        CDM_Free(AttributeNames);
        CDM_Free(Expr);
        CDM_Free(ResSet);
        CDM_AppendError("CDM_SelectAnnotations: invalid boolean expression "
                        "result\n  (type=\"%s\", constraints=\"%s\")",
                        Type, Constraints);
        return NULL;
      }
      if (BoolVal) {
        /* We succeded! Expression was true... */
        Tcl_ListObjAppendElement(CDM_Interp, ResSet, Ann[i]);
      }
    }
  }
  Tcl_UnsetVar2(CDM_Interp, "::CDM::CDM_SelectAnnotationsArray", NULL, 0);
  CDM_Free(AttributeNames);
  CDM_Free(Expr);
  return ResSet;
} /* CDM_SelectAnnotations */


/*
 * ---- CDM_SelectAnnotationsSorted:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Annotation set that their type is the same as the value of the
 *     "Type" parameter. The Annotations will be sorted according to their
 *     first span range, in ascending order.
 *
 *     This function will always create and return a new
 *     object, even when no Annotations of the requested type are found or
 *     when the Annotation set is empty. In this case an empty
 *     object will be returned.
 *
 *     Note that CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set:   A set of Annotations.                        (CDM_AnnotationSet)
 *     Type:  The type of the Annotations to be selected.  (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SelectAnnotationsSorted (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return an AnnotationSet object, that will
 *     contain all the Annotations of the specified Annotation set that have as
 *     type
 *     the value of the "Type" parameter. The Annotations in that object will
 *     be sorted according to their first span in ascending order.
 *
 *     Note that the caller is responsible to free this object through
 *     Tcl_DecrRefCount or CDM_Free.
 *
 *     In case of an error, NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_SelectAnnotations
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_SelectAnnotationsSorted(CDM_AnnotationSet Set,char *Type){
  CDM_AnnotationSet Annotations;
  Tcl_Obj *objv[4];
  int status, i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_SelectAnnotationsSorted: Aborting...");
    return NULL;
  }

  Annotations = CDM_SelectAnnotations(Set, Type);
  if (Annotations == NULL) return NULL;
  /* Execute Tcl Command "lsort -command tip_CompareAnnotations $Annotations" */
  objv[0] = Tcl_NewStringObj("lsort", 5);
  objv[1] = Tcl_NewStringObj("-command", 8);
  objv[2] = Tcl_NewStringObj("tip_CompareAnnotations", 22);
  objv[3] = Annotations;
  for (i=0; i<4; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(CDM_Interp, 4, objv, TCL_EVAL_GLOBAL);
  for (i=0; i<4; i++) Tcl_DecrRefCount(objv[i]);
  if (status != TCL_OK) return NULL;
  return Tcl_DuplicateObj(Tcl_GetObjResult(CDM_Interp));
} /* CDM_SelectAnnotationsSorted */


/*
 * ---- CDM_SelectAnnotationsSorted:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Annotation set that their type is the same as the value of the
 *     "Type" parameter and the provided boolean expression (specified through
 *     the "Constrains" parameter) is true for every one of these Annotations.
 *     The Annotations will be sorted according to their
 *     first span range, in ascending order.
 *
 *     This function will always create and return a new
 *     object, even when no Annotations of the requested type are found or
 *     when the Annotation set contains no Annotations. In this case an empty
 *     object will be returned.
 *
 *     The boolean expression that can be specified through the "Constrains"
 *     parameter can be any valid Tcl boolean expression (i.e. a boolean
 *     expression that will be accepted by the "expr" Tcl command), with the
 *     following excepion:
 *
 *     [[LI]]The boolean expression can contain references to the values of
 *     Atttributes of the Annotation, by utilising the notation
 *     "ann::%%%lt;Attribute Name%%%gt;". The only limitation is that
 *     "%%%lt;Attribute Name%%%gt;" cannot exceed 120 characters. For example,
 *     the following code will create a new Annotation set that will contain
 *     all Annotations from the specified Annotation set that their type is
 *     token,
 *     they have a "type" Attribute which has as value "EFW" and also have a
 *     "pos" Attribute that has a value other than "NN":[[/LI]]
 *     [[br]][[code]][[center]]
 *     CDM_SelectAnnotationsSorted(Set, "token", "ann::type == \"EFW\" %%%amp;%%%amp; ann::pos != \"NN\"");
 *     [[/center]][[/code]][[br]]
 *
 *     Note that CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set:         A set of Annotations.             (CDM_AnnotationSet)
 *     Type:        The type of the Annotations to be selected.  (char *)
 *     Constraints: A boolean expression.                        (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SelectAnnotationsSorted (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return an AnnotationSet object, that will
 *     contain all the Annotations of the specified Annotation set that have as
 *     type
 *     the value of the "Type" parameter and the provided by the "Constrains"
 *     parameter boolean expression was true for these Annotations.
 *     The Annotations in that object will
 *     be sorted according to their first span in ascending order.
 *
 *     Note that the caller is responsible to free this object through
 *     Tcl_DecrRefCount or CDM_Free.
 *
 *     In case of an error, NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_SelectAnnotations, [[A HREF="../TclTk/TclCmd/expr.htm"]]The Tcl
 *     "expr" manual[[/A]].
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_SelectAnnotationsSorted(CDM_AnnotationSet Set,
                                              char *Type, char *Constraints){
  CDM_AnnotationSet Annotations;
  Tcl_Obj *objv[4];
  int status, i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_SelectAnnotationsSorted: Aborting...");
    return NULL;
  }

  Annotations = CDM_SelectAnnotations(Set, Type, Constraints);
  if (Annotations == NULL) return NULL;
  /* Execute Tcl Command "lsort -command tip_CompareAnnotations $Annotations" */
  objv[0] = Tcl_NewStringObj("lsort", 5);
  objv[1] = Tcl_NewStringObj("-command", 8);
  objv[2] = Tcl_NewStringObj("tip_CompareAnnotations", 22);
  objv[3] = Annotations;
  for (i=0; i<4; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(CDM_Interp, 4, objv, TCL_EVAL_GLOBAL);
  for (i=0; i<4; i++) Tcl_DecrRefCount(objv[i]);
  if (status != TCL_OK) return NULL;
  return Tcl_DuplicateObj(Tcl_GetObjResult(CDM_Interp));
} /* CDM_SelectAnnotationsSorted */

/*
 * ---- CDM_SortAnnotationSet:
 *  DESCRIPTION:
 *   <English:
 *     This function will sort the Annotations contained in the provided
 *     Annotation set, according to their Span sets.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set:        The Annotation set to be sorted      (CDM_AnnotationSet)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_SortAnnotationSet (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return a new AnnotationSet object,
 *     that will contain all the Annotations contained in the provided by the
 *     "Set" parameter set, sorted by their span ranges. For more information
 *     about the sorting order, please refer to the CDM_CompareAnnotations
 *     function.
 *
 *     Note that CDM does not own this object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *
 *     In case of an error, NULL will be return and an error message will
 *     be left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_SelectAnnotationsSorted, CDM_CompareAnnotations,
 *     tip_SortAnnotationSet,
 *     [[A HREF="../TclTk/TclCmd/lsort.htm"]]lsort[[/A]]
 *   >:&:
 */
CDM_EXPORT
CDM_AnnotationSet CDM_SortAnnotationSet(CDM_AnnotationSet Set){
  Tcl_Obj *objv[4];
  int status, i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_SortAnnotationSet: Aborting...");
    return NULL;
  }

  /* Execute Tcl Command "lsort -command tip_CompareAnnotations $Set" */
  objv[0] = Tcl_NewStringObj("lsort", 5);
  objv[1] = Tcl_NewStringObj("-command", 8);
  objv[2] = Tcl_NewStringObj("tip_CompareAnnotations", 22);
  objv[3] = Set;
  for (i=0; i<4; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(CDM_Interp, 4, objv, TCL_EVAL_GLOBAL);
  for (i=0; i<4; i++) Tcl_DecrRefCount(objv[i]);
  if (status != TCL_OK) return NULL;
  return Tcl_DuplicateObj(Tcl_GetObjResult(CDM_Interp));
} /* CDM_SortAnnotationSet */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    | E N D  O F  A N N O T A T I O N  S E T  S E C T I O N |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     S P A N                        S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 * ---- CDM_AddSpan:
 *  DESCRIPTION:
 *   <English:
 *    Append given Span to the given Span set...
 *   >
 *  ARGUMENTS:
 *   <English:
 *       SpanSet:  Span set where the Span will be appended  (CDM_SpanSet)
 *       Span:     Span to be appended                       (CDM_Span)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
CDM_SpanSet CDM_AddSpan(CDM_SpanSet SpanSet, CDM_Span Span) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_AddSpan: Aborting...");
    return NULL;
  }

  if (Tcl_IsShared(SpanSet)) {
#ifdef DUPLICATE_DEBUG
    CDM_Debug("AddSpan: Duplicating (%d): %s\n", SpanSet->refCount,
            Tcl_GetString(SpanSet));
#endif /* DUPLICATE_DEBUG */
    SpanSet = Tcl_DuplicateObj(SpanSet);
  }
  if (SpanSet == NULL)
    CDM_Panic("CDM_AddSpan: Tcl_DuplicateObj returned NULL - out of memory?\n");
  if (Tcl_ListObjAppendElement(CDM_Interp, SpanSet, Span) != TCL_OK) {
    CDM_ReportError("CDM_AddSpan: Cannot append to given span set!");
    return NULL;
  }
  return SpanSet;
} /* CDM_AddSpan */

/*
 * ---- CDM_CreateSpanSet:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return an empty SpanSet object
 *     (of type CDM_SpanSet).
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is not property of the CDM: the caller is
 *     responsible for using
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]] in order
 *     to keep a reference to the object or
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] -
 *     CDM_Free to free the object
 *     and the memory associated with it when the object is not needed anymore.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None: (void)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_CreateSpanSet (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return an empty SpanSet object (of type of
 *     CDM_SpanSet).
 *
 *     In case of an error, NULL will be returned and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp).
 *
 *     Note that the returned object is not property of the CDM: the caller is
 *     responsible for using
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]] in order
 *     to keep a reference to the object or
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] -
 *     CDM_Free to free the object
 *     and the memory associated with it when the object is not needed anymore.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_CreateSpan, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]],
 *     tip_CreateSpanSet
 *   >:&:
 */
CDM_EXPORT
CDM_SpanSet CDM_CreateSpanSet(void) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_CreateSpanSet: Aborting...");
    return NULL;
  }

  return (CDM_SpanSet) Tcl_NewListObj(0, NULL);
} /* CDM_CreateSpanSet */

/*
 * ---- CDM_CreateSpanSet:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new SpanSet object
 *     (of type CDM_SpanSet) which will contain a single Span object (of type
 *     CDM_Span) with the specified start and end values.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is not property of the CDM: the caller is
 *     responsible for using
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]] in order
 *     to keep a reference to the object or
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] -
 *     CDM_Free to free the object
 *     and the memory associated with it when the object is not needed anymore.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      start:  Span Start     (long)
 *      end:    Span End       (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to the following C++ code:
 *    [[br]][[code]][[center]]CDM_Span span = CDM_CreateSpan(start, end);[[br]]
 *          CDM_SpanSet spanSet = CDM_CreateSpanSet();[[br]]
 *          CDM_AddSpan(spanSet, span);
 *     [[/center]][[/code]]
 *     This function is equivelant to tip_CreateSpanSet (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a new SpanSet object (of type of
 *     CDM_SpanSet). This SpanSet object will contain a single Span object,
 *     whoose start and end values will be initialised to the provided ones,
 *     through the "start" and "end" parameters.
 *
 *     In case of an error, NULL will be returned and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp).
 *
 *     Note that the returned object is not property of the CDM: the caller is
 *     responsible for using
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]] in order
 *     to keep a reference to the object or
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] -
 *     CDM_Free to free the object
 *     and the memory associated with it when the object is not needed anymore.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_CreateSpan, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]],
 *     tip_CreateSpanSet
 *   >:&:
 */
CDM_EXPORT
CDM_SpanSet CDM_CreateSpanSet(long start, long end) {
  register CDM_SpanSet SpanSet;
  Tcl_Obj *argv[1];
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_CreateSpanSet: Aborting...");
    return NULL;
  }

  argv[0] = CDM_CreateSpan(start, end);
  SpanSet = Tcl_NewListObj(1, argv);
  if (SpanSet == NULL)
    CDM_Panic("CDM_CreateSpanSet: Tcl_NewListObj returned NULL "
              "- out of memory?\n");
  return SpanSet;
} /* CDM_CreateSpanSet */

/*
 * ---- CDM_CreateSpan:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a Span object
 *     (of type CDM_Span). In order to reduce memory requirements, this
 *     function will try to locate if a Span with exactly the same
 *     characteristics (same start and end values) already exists.
 *     In such a case, this (possibly shared) Span
 *     object will be returned. Else, a new Span object will be created
 *     and returned.
 *
 *     In case of an error, NULL will be return and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *
 *     Note that the returned object is not property of the CDM: the caller is
 *     responsible for using
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *     CDM_Free to free the object
 *     and the memory associated with it when the object is not needed anymore.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      start:  Span Start     (long)
 *      end:    Span End       (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_CreateSpan (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a Span object (of type of
 *     CDM_Span). This object may be an old, already in use (shared)
 *     object or a newly created object, according to the specified parameters.
 *
 *     In case of an error, NULL will be returned and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp).
 *
 *     Note that the returned object is not property of the CDM: the caller is
 *     responsible for using
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]] in order
 *     to keep a reference to the object or
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] -
 *     CDM_Free to free the object
 *     and the memory associated with it when the object is not needed anymore.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_CreateSpanSet, CDM_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]],
 *     tip_CreateSpan
 *   >:&:
 */
CDM_EXPORT
CDM_Span CDM_CreateSpan(long start, long end) {
  CDM_Span Span;
  Tcl_Obj *argv[2];
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_CreateSpan: Aborting...");
    return NULL;
  }

  argv[0] = ObjectCache.FindLong(start);
  if (argv[0] == NULL) CDM_Panic("CDM_CreateSpan: ObjectCache.FindLong "
                                 "returned NULL - out of memory?\n");
  argv[1] = ObjectCache.FindLong(end);
  if (argv[1] == NULL) CDM_Panic("CDM_CreateSpan: ObjectCache.FindLong "
                                 "returned NULL - out of memory?\n");

  Span = Tcl_NewListObj(2, argv);
  if (Span == NULL)
    CDM_Panic("CDM_CreateSpan: Tcl_NewListObj returned NULL "
              "- out of memory?\n");
  return Span;
} /* CDM_CreateSpan */

/*
 * ---- CDM_GetStart:
 *  DESCRIPTION:
 *   <English:
 *    This function accepts a span (of type CDM_Span) and returns the start of
 *    this span. Each span object is a set of exactly two integers, named as
 *    "start" and "end".
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Span: Span   (CDM_Span)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to tip_GetStart (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns the start value of the provided span object. If
 *    the provided object is not a valid span object or some other error
 *    occurs, CDM_GetStart will return -1 and an error message describing the
 *    error will be placed at the current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_CreateSpan, CDM_GetEnd, tip_GetStart
 *   >:&:
 */
CDM_EXPORT
long CDM_GetStart(CDM_Span Span) {
  Tcl_Obj **elements;
  int result, items;
  long start;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetStart: Aborting...");
    return -1;
  }

  if (Span == NULL) return -1;
  result = Tcl_ListObjGetElements(CDM_Interp, Span, &items, &elements);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("Malformed Span \"%s\"", Tcl_GetString(Span));
    return -1;
  }
  result = Tcl_GetLongFromObj(CDM_Interp, elements[0], &start);
  if (result != TCL_OK) return -1;
  return start;
} /* CDM_GetStart */

/*
 * ---- CDM_GetEnd:
 *  DESCRIPTION:
 *   <English:
 *    This function accepts a span (of type CDM_Span) and returns the end of
 *    this span. Each span object is a set of exactly two integers, named as
 *    "start" and "end".
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Span: Span   (CDM_Span)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to tip_GetEnd (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns the end value of the provided span object. If
 *    the provided object is not a valid span object or some other error
 *    occurs, CDM_GetEnd will return -1 and an error message describing the
 *    error will be placed at the current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_CreateSpan, CDM_GetStart, tip_GetEnd
 *   >:&:
 */
CDM_EXPORT
long CDM_GetEnd(CDM_Span Span) {
  Tcl_Obj **elements;
  int result, items;
  long end;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetEnd: Aborting...");
    return -1;
  }

  if (Span ==NULL) return -1;
  result = Tcl_ListObjGetElements(CDM_Interp, Span, &items, &elements);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("Malformed Span \"%s\"", Tcl_GetString(Span));
    return -1;
  }
  result = Tcl_GetLongFromObj(CDM_Interp, elements[1], &end);
  if (result != TCL_OK) return -1;
  return end;
} /* CDM_GetEnd */

/*
 * ---- CDM_GetSpanOffsets:
 *  DESCRIPTION:
 *   <English:
 *    This function will place the values of the both span start and end of
 *    the provided span in the two specified long integer pointers.
 *    Each span object is a set of exactly two integers, named as
 *    "start" and "end".
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Span:   A Span object  (CDM_Span)
 *      start:  A pointer pointing to a long, where the start value of the specified span will be placed. (long *)
 *      end:    A pointer pointing to a long, where the end value of the specified span will be placed. (long *)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to the following C++ code:
 *    [[br]][[code]][[center]]*start = CDM_GetStart(Span);[[br]]
 *                            *end   = CDM_GetEnd(Span);
 *     [[/center]][[/code]]
 *    This function is equivelant to tip_GetSpanOffsets (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will place the two Offsets contained in the provided Span
 *    object (the start and end of the span) into the two long variables
 *    pointed by the two provided pointers.
 *
 *    The return value from this function will be a standart Tcl completion
 *    code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *    error occurs, the return value will be TCL_ERROR and an error message
 *    describing the error will be placed at the current Tcl interpreter
 *    (CDM_Interp). Else, TCL_OK will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_CreateSpan, CDM_GetStart, CDM_GetEnd
 *   >:&:
 */
CDM_EXPORT
int CDM_GetSpanOffsets(CDM_Span Span, long *start, long *end) {
  Tcl_Obj **elements;
  int result, items;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_GetSpanOffsets: Aborting...");
    return TCL_ERROR;
  }

  if (Span ==NULL) return TCL_ERROR;
  result = Tcl_ListObjGetElements(CDM_Interp, Span, &items, &elements);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("CDM_GetSpanOffsets: Malformed Span \"%s\"",
                    Tcl_GetString(Span));
    return TCL_ERROR;
  }

  if (start != NULL) {
    result = Tcl_GetLongFromObj(CDM_Interp, elements[0], start);
    if (result != TCL_OK) return TCL_ERROR;
    if (*start < 0) {
      return CDM_ReportError("Negative value %ld detected "
             "in start position of Span \"%s\"", *start, Tcl_GetString(Span));
    }
  }
  if (end != NULL) {
    result = Tcl_GetLongFromObj(CDM_Interp, elements[1], end);
    if (result != TCL_OK) return TCL_ERROR;
    if (*end < 0) {
      return CDM_ReportError("Negative value %ld detected "
             "in end position of Span \"%s\"", *end, Tcl_GetString(Span));
    }
  }
  return TCL_OK;
} /* CDM_GetSpanOffsets */

/*
 * ---- CDM_SpanSetGetElements:
 *  DESCRIPTION:
 *   <English:
 *    This function returns a count and a pointer to an array of the elements
 *    (Spans) in an Span set object (of type CDM_SpanSet).
 *    It returns the count by storing it in the address pointed by the "objcPtr"
 *    parameter (of type int *). Similarly, it returns the array pointer
 *    by storing it in the address pointed by the "objvPtr" parameter (of type
 *    CDM_Span **). The memory pointed to is managed by CDM and should
 *    not be freed by the caller.
 *
 *    In case of an error, TCL_ERROR will be returned and an error message will
 *    be left in the current active Tcl interpreter (CDM_Interp).
 *    Otherwise TCL_OK will be returned after storing the count and
 *    array pointer.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     SpanSet: The Span set whose elements will be placed in the array. (CDM_SpanSet)
 *     objcPtr:      A pointer to an integer, where the number of Span objects placed in the array will be saved. (int *)
 *     objvPtr:      A location where the pointer to the array of Span objects will be stored. (CDM_Span **)
 *   >
 *  NOTES:
 *   <English:
 *    Usage example:[[br]][[code]][[pre]]
 *    int objc;
 *    CDM_Span **objv;
 *    CDM_SpanSet set = CDM_GetSpans(Annotation);
 *    int status = CDM_SpanSetGetElements(set, objc, objv);
 *    if (status != TCL_OK) {
 *      // An error occured. What is the error message?
 *      Tcl_interp *interp = CDM_GetActiveInterpreter();
 *      if (interp) {
 *        const char* msg = Tcl_GetStringResult(interp);
 *      }
 *    }
 *    // Make an iteration over all Spans...
 *    for (int i = 0; i < objc, i++) {
 *      // Now, objv[i] is an Span...
 *    }
 *    [[/pre]][[/code]]
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a count and a pointer to an array of the elements
 *    (Spans) in an Span set object (of type CDM_SpanSet).
 *    It returns the count by storing it in the address pointed by the "objcPtr"
 *    parameter (of type int *). Similarly, it returns the array pointer
 *    by storing it in the address pointed by the "objvPtr" parameter (of type
 *    CDM_Span **). The memory pointed to is managed by CDM and should
 *    not be freed by the caller.
 *
 *    In case of an error, TCL_ERROR will be returned and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). Otherwise TCL_OK will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_AttributeSetGetElements, CDM_AnnotationSetGetElements
 *   >:&:
 */
CDM_EXPORT
int CDM_SpanSetGetElements(CDM_SpanSet SpanSet,
                                int *objcPtr, CDM_Span **objvPtr) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    return CDM_ReportError("CDM_SpanSetGetElements: Aborting...");
  }

  if (SpanSet == NULL) {
    return CDM_ReportError("CDM_SpanSetGetElements: Invalid SpanSet or "
                           "Span Pointer");
  }

  return Tcl_ListObjGetElements(CDM_Interp, SpanSet, objcPtr, objvPtr);
} /* CDM_SpanSetGetElements */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |    E N D      O F      S P A N       S E C T I O N    |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     V A L I D A T I O N           S E C T I O N       |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 * ---- CDM_ValidAnnotation:
 *  DESCRIPTION:
 *   <English:
 *    This function will thoroughly check the provided Annotation in order to
 *    see if it is valid or not. If the provided Annotation object (of type
 *    CDM_Annotation) is valid, "1" will be return. In the case where the
 *    Annotation is not valid (or malformed), "0" is returned and an error
 *    message is left at the current active interpeter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Ann:   The Annotation whoose validity is in question.  (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return true (1) if the provided Annotation object is
 *    valid. Else, false (0) will be returned and an error message describing
 *    the error will be left at the current active interpeter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_ValidAttribute
 *   >:&:
 */
CDM_EXPORT
int CDM_ValidAnnotation(CDM_Annotation Ann) {
  Tcl_Obj **elements, **Spans, **Attrs;
  int result, items, i;
  long start, end;
  CDM_THREAD_SPECIFIC_DATA;

  if (Ann == NULL) {
    CDM_ReportError("CDM_ValidAnnotation: NULL Annotation");
    return 0;
  }
#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(Ann)) return 1;
#endif /* ENABLE_ANNOTATION_OBJECT */

  result = Tcl_ListObjGetElements(NULL, Ann, &items, &elements);
  if (result != TCL_OK || items != 4) {
    CDM_ReportError("Malformed Annotation \"%s\"", Tcl_GetString(Ann));
    return 0;
  }
#ifdef DEBUGVALIDS
  CDM_Debug("CDM_ValidAnnotation: (%d) %s\n",Ann->refCount, Tcl_GetString(Ann));
  CDM_Debug("Refs: (%d) (%d) (%d) (%d)\n", elements[0]->refCount,
          elements[1]->refCount, elements[2]->refCount, elements[3]->refCount);
#endif

  // Check Spans...
  result = Tcl_ListObjGetElements(CDM_Interp, elements[2], &items, &Spans);
  if (result != TCL_OK || items < 0) {
    CDM_ReportError("Malformed SpanList in Annotation \"%s\"",
                    Tcl_GetString(Ann));
    return 0;
  }
  for (i=0; i<items; i++) {
    result = CDM_GetSpanOffsets(Spans[i], &start, &end);
    if (result != TCL_OK) {
      CDM_AppendError(": mallformed Span in position %d in Annotation \"%s\"",
                      i, Tcl_GetString(Ann));
      return 0;
    }
  }

  // Check Attributes
  result = Tcl_ListObjGetElements(CDM_Interp, elements[3], &items, &Attrs);
  if (result != TCL_OK || items < 0) {
    CDM_ReportError("Malformed Attribute Set in Annotation \"%s\"",
                    Tcl_GetString(Ann));
    return 0;
  }
  for (i=0; i<items; i++) {
    if (!CDM_ValidAttribute(Attrs[i])) {
      CDM_AppendError(": mallformed Attribute in position %d in Annotation "
                      "\"%s\"", i, Tcl_GetString(Ann));
      return 0;
    }
  }
  return 1;
} /* CDM_ValidAnnotation */

/*
 * ---- CDM_ValidAttribute:
 *  DESCRIPTION:
 *   <English:
 *    This function will thoroughly check the provided Attribute in order to
 *    see if it is valid or not. If the provided Attribute object (of type
 *    CDM_Attribute) is valid, "1" will be return. In the case where the
 *    Attribute is not valid (or malformed), "0" is returned and an error
 *    message is left at the current active interpeter (CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Ann:   The Attribute whoose validity is in question.  (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return true (1) if the provided Attribute object is
 *    valid. Else, false (0) will be returned and an error message describing
 *    the error will be left at the current active interpeter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_ValidAttribute
 *   >:&:
 */
CDM_EXPORT
int CDM_ValidAttribute(CDM_Attribute Attr) {
  Tcl_Obj *Value, **elemPtrs;
  int result, items;
  CDM_THREAD_SPECIFIC_DATA;

  if (Attr == NULL) {
    CDM_ReportError("CDM_ValidAttribute: NULL Attribute pass for validation");
    return 0;
  }

  // Every Attribute is a list: {AttrName {Type Value} }
  result = Tcl_ListObjGetElements(CDM_Interp, Attr, &items, &elemPtrs);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("Invalid Attribute \"%s\"", Tcl_GetString(Attr));
    return 0;
  }
  Value = elemPtrs[1];
  result = Tcl_ListObjGetElements(CDM_Interp, Value, &items, &elemPtrs);
  if (result != TCL_OK || items != 2) {
    CDM_ReportError("Invalid Attribute Value \"%s\"", Tcl_GetString(Value));
    return 0;
  }

  /* Is the attribute value a valid one? */
  if (CDM_GetAttributeTypeFromObject(elemPtrs[0]) < 0) {
    CDM_ReportError("Invalid Attribute Value Type \"%s\"",
    Tcl_GetString(elemPtrs[0]));
    return 0;
  }
  return 1;
} /*CDM_ValidAttribute */

/*
 * ---- CDM_ValidCollectionPointer:
 *  DESCRIPTION:
 *   <English:
 *    Checks whether a given pointer is a Collection Pointer...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Pointer: The pointer under question          (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
int CDM_ValidCollectionPointer(CDM_Collection Pointer) {
  if (Pointer == NULL) return 0;
  if (Pointer->Type != CDM_TYPE_COLLECTION) return 0;
  return 1;
} /* CDM_ValidCollectionPointer */

/*
 * ---- CDM_ValidDocumentPointer:
 *  DESCRIPTION:
 *   <English:
 *    Checks whether a given pointer is a Document Pointer...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Pointer: The pointer under question          (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
int CDM_ValidDocumentPointer(CDM_Document Pointer) {
  if (Pointer == NULL) return 0;
  if (Pointer->Type != CDM_TYPE_DOCUMENT) return 0;
  return 1;
} /* CDM_ValidDocumentPointer */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |  E N D   O F   V A L I D A T I O N   S E C T I O N    |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     U T I L I T I E S              S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 *  Unique ID Generator Class.
 *    This class generates unique ID for Collections/Documents etc.
 */
CDM_UniqIDs::CDM_UniqIDs() {
  Init();
}
void CDM_UniqIDs::Init(){
  CollectionMaxID = DocumentMaxID = MaxID = 0;
}
int CDM_UniqIDs::CollectionUniqID(void) {
   int Id; Id = CollectionMaxID++; return Id;
}
int CDM_UniqIDs::DocumentUniqID(void) {
   int Id; Id = DocumentMaxID++; return Id;
}
int CDM_UniqIDs::UniqID(void) {int Id; Id = MaxID++; return Id;}
void CDM_UniqIDs::CancelCollectionUniqID(int Id) {
   if (CollectionMaxID == Id + 1) CollectionMaxID--;
}
void CDM_UniqIDs::CancelDocumentUniqID(int Id) {
   if (DocumentMaxID == Id + 1) DocumentMaxID--;
}
void CDM_UniqIDs::CancelUniqID(int Id) {
   if (MaxID == Id + 1) MaxID--;
} /* Unique ID Generator Class */

#ifdef DONT_HASH_DOC_ANNOTATIONS
/*
 * Annotation Linked List Implementation...
 *  Petasis, 28/10/1999
 */
CDM_AnnotationList::CDM_AnnotationList(void) {
  Head = NULL; Last = NULL;
} /* Constructor */
CDM_AnnotationList::~CDM_AnnotationList(void) {
  register CDM_AnnotationListNode *ptr = Head, *next;

  while(ptr != NULL) {
    if (ptr->Ann != NULL) Tcl_DecrRefCount(ptr->Ann);
    next = ptr->next;
    Tcl_Free((char *)ptr);
    ptr = next;
  }
} /* Destructor */
CDM_AnnotationListNode *CDM_AnnotationList::FindPosition(long Id) {
  register CDM_AnnotationListNode *ptr, *nptr = NULL;
  /*
   * In "Last" there is the last accessed item. Take this as a starting
   * point (or Head if list has no elements)...
   */
  if (Last != NULL) ptr = Last; else ptr = Head;

  /* Is this the first element? */
  if (ptr == NULL) {
    nptr = (CDM_AnnotationListNode *) Tcl_Alloc(sizeof(CDM_AnnotationListNode));
    if (nptr == NULL) CDM_Panic("CDM_AnnotationList: out of memory?\n");
    nptr->Ann = NULL;
    Last = Head = nptr;
    Head->prev = Head->next = NULL;
    return nptr;
  }

  if (ptr->Id == Id) return ptr;
  if (ptr->Id < Id) {
    while ((ptr->next != NULL)) {
      if (ptr->Id == Id) return ptr;
      if (ptr->Id > Id) {ptr = ptr->prev; break;}
      ptr = ptr->next;
    }
    //if ((ptr->next)->Id == Id) return ptr->next;
    nptr = (CDM_AnnotationListNode *) Tcl_Alloc(sizeof(CDM_AnnotationListNode));
    if (nptr == NULL) CDM_Panic("CDM_AnnotationList: out of memory?\n");
    nptr->Ann = NULL;
    nptr->next = ptr->next;
    nptr->prev = ptr;
    ptr->next = nptr;
    if (nptr->next != NULL) (nptr->next)->prev = nptr;
  } else if (ptr->Id > Id) {
    while ((ptr->prev != NULL)) {
      if (ptr->Id == Id) return ptr;
      if (ptr->Id < Id) {ptr = ptr->next; break;}
      ptr = ptr->prev;
    }
    //if ((ptr->prev)->Id == Id) return ptr->prev;
    nptr = (CDM_AnnotationListNode *) Tcl_Alloc(sizeof(CDM_AnnotationListNode));
    if (nptr == NULL) CDM_Panic("CDM_AnnotationList: out of memory?\n");
    nptr->Ann = NULL;
    nptr->prev = ptr->prev;
    nptr->next = ptr;
    ptr->prev = nptr;
    if (nptr->prev != NULL) (nptr->prev)->next = nptr;
  }
  return nptr;
} /* Find */
CDM_Annotation CDM_AnnotationList::Ann(long Id) {
  register CDM_AnnotationListNode *ptr;

  ptr = Get(Id);
  if (ptr!= NULL) {
    Last = ptr;
    return ptr->Ann;
  }
  return NULL;
}
int CDM_AnnotationList::Set(CDM_Annotation Ann, long Id) {
  register CDM_AnnotationListNode *ptr;

  /* Is it a simple replace? */
  if (Last != NULL && Last->Id == Id) {
    Tcl_DecrRefCount(Last->Ann);
    Last->Ann = Ann; Last->Id = Id;
    return TCL_OK;
  }

  /* Create a new list node or find an existing one... */
  ptr = FindPosition(Id);
  if (ptr->Ann != NULL) Tcl_DecrRefCount(ptr->Ann);
  ptr->Ann = Ann; ptr->Id = Id;
  Last = ptr;
  return TCL_OK;
} /* Set */
CDM_AnnotationListNode *CDM_AnnotationList::Get(long Id) {
  register CDM_AnnotationListNode *ptr = Last;

  if (Head == NULL || ptr == NULL) return NULL;
  if (ptr->Id == Id) return ptr;
  if (ptr->Id < Id) {
    while ((ptr != NULL)) {
      if (ptr->Id == Id) {Last = ptr; return ptr;}
      if (ptr->Id > Id)  break;
      ptr = ptr->next;
    }
  } else if (ptr->Id > Id) {
    while ((ptr != NULL)) {
      if (ptr->Id == Id) {Last = ptr; return ptr;}
      if (ptr->Id < Id)  break;
      ptr = ptr->prev;
    }
  }
  return NULL;
} /* Get */
int CDM_AnnotationList::Delete(long Id) {
  register CDM_AnnotationListNode *ptr;

  ptr = Get(Id);
  if (ptr == NULL) return TCL_ERROR;
  Tcl_DecrRefCount(ptr->Ann);
  if (ptr->prev != NULL) (ptr->prev)->next = ptr->next;
  if (ptr->next != NULL) (ptr->next)->prev = ptr->prev;
  Last = ptr->prev;
  if (ptr->prev == NULL) {Head = Last = ptr->next;}
  Tcl_Free((char *) ptr);
  return TCL_OK;
} /* Delete */
#endif /* DONT_HASH_DOC_ANNOTATIONS */

/*
 * ---- CDM_NewCollectionObj:
 *    Creates a new Collection Object (in memory...)
 *    Arguments:
 *      -
 */
CDM_Collection CDM_NewCollectionObj(void) {
  CDM_Collection ObjPtr;
  CDM_THREAD_SPECIFIC_DATA;

  ObjPtr = (CDM_Collection) CDM_Malloc(sizeof(CDM_CollectionStr));
  ObjPtr->Type                 = CDM_TYPE_COLLECTION;
  ObjPtr->Encoding             = NULL;
  //ObjPtr->Obj_ID             = -1;
  ObjPtr->Tcl_Cmd              = NULL;
  ObjPtr->Interp               = CDM_Interp;
  ObjPtr->FileFormat           = CDM_FORMAT_NATIVE2;
  ObjPtr->Name                 = NULL;
  ObjPtr->Owner                = NULL;
  ObjPtr->Documents            = NULL;
  ObjPtr->CurrentDocumentIndex = NULL;
  ObjPtr->RemoveDiskDocuments  = NULL;
#ifdef CDM_OBJ_HASH_KEYS
  Tcl_InitObjHashTable(&(ObjPtr->Attributes_HashTable));
#else /* CDM_OBJ_HASH_KEYS */
  Tcl_InitHashTable(&(ObjPtr->Attributes_HashTable), TCL_STRING_KEYS);
#endif /* CDM_OBJ_HASH_KEYS */
  ObjPtr->RefCount             = 0;
  ObjPtr->Modified             = 0;
  ObjPtr->ModifiedDocuments    = 0;
  ObjPtr->AssociatedInfo       = NULL;
  ObjPtr->CommentObj           = NULL;
  return ObjPtr;
}

/*
 * ---- CDM_NewDocumentObj:
 *    Creates a new Document Object (in memory...)
 *    Arguments:
 *      -
 */
CDM_Document CDM_NewDocumentObj(void) {
  CDM_Document ObjPtr;
  CDM_THREAD_SPECIFIC_DATA;

  ObjPtr = (CDM_Document) CDM_Malloc(sizeof(CDM_DocumentStr));
  ObjPtr->Type            = CDM_TYPE_DOCUMENT;
  ObjPtr->Encoding        = NULL;
  ObjPtr->Tcl_Cmd         = NULL;
  ObjPtr->Interp          = CDM_Interp;
  ObjPtr->FileFormat      = CDM_FORMAT_NATIVE2;
  ObjPtr->ParentPtr       = NULL;
  ObjPtr->DocumentPtr     = NULL;
  ObjPtr->RawData         = NULL;
  ObjPtr->AnnotationMaxID = 0;
  ObjPtr->RefCount        = 0;
  ObjPtr->Modified        = 0;

#ifdef CDM_OBJ_HASH_KEYS
  Tcl_InitObjHashTable(&(ObjPtr->Attributes_HashTable));
#else /* CDM_OBJ_HASH_KEYS */
  Tcl_InitHashTable(&(ObjPtr->Attributes_HashTable), TCL_STRING_KEYS);
#endif /* CDM_OBJ_HASH_KEYS */
#ifdef DONT_HASH_DOC_ANNOTATIONS
  (ObjPtr->Annotations_List) = new CDM_AnnotationList;
#else  /* DONT_HASH_DOC_ANNOTATIONS */
#ifdef CDM_ONE_WORD_KEYS
  Tcl_InitHashTable(&(ObjPtr->Annotations_HashTable), TCL_ONE_WORD_KEYS);
#else
  Tcl_InitHashTable(&(ObjPtr->Annotations_HashTable), TCL_STRING_KEYS);
#endif /* CDM_ONE_WORD_KEYS */
#endif /* DONT_HASH_DOC_ANNOTATIONS */
  return ObjPtr;
} /* CDM_NewDocumentObj */

/*
 * ---- CDM_Eval:
 *  DESCRIPTION:
 *   <English:
 *    This function can be used to execute a Tcl statement/code from within
 *    C/C++.
 *    CDM_Eval accepts a variable number of strings (of type char*) and
 *    assumes that each such string corresponds to a token of the Tcl command.
 *    CDM_Eval evaluates then the command, by-passing variable substitution
 *    and byte-code compilation.
 *
 *    For example, in order to execute the Tcl statement %%%quot;puts
 *    %%%quot;This is a test%%%quot;%%%quot;, the following C code should be
 *    used:[[br]][[code]][[center]]
 *    CDM_Eval(%%%quot;puts%%%quot;, %%%quot;This is a test%%%quot;, NULL);
 *    [[/center]][[/code]][[br]]
 *
 *    The return value from this function will be a standart Tcl completion
 *    code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *    error occurs, the return value will be TCL_ERROR and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). If all annotators run without errors, TCL_OK will be
 *    returned.
 *
 *    The results of the Tcl code execution will be placed as the result in
 *    the current Tcl interpreter (CDM_Interp), if the command has been
 *    succesfully executed. The caller can use
 *    [[A HREF="../TclTk/TclLib/SetResult.htm"]]Tcl_GetObjResult[[/A]](CDM_GetActiveInterpreter()) to get the stored result
 *    Tcl object.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Interp: Interpreter Pointer (Tcl_Interp)
 *    arg1: A (NULL terminated) sequence of strings (char *)
 *   >
 *  NOTES:
 *   <English:
 *    Note that the last argument should always be NULL, signaling the end of
 *    arguments.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    The return value from this function will be a standart Tcl completion
 *    code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *    error occurs, the return value will be TCL_ERROR and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). If all annotators run without errors, TCL_OK will be
 *    returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Exec
 *   >:&:
 */
CDM_EXPORT
int CDM_Eval TCL_VARARGS_DEF(Tcl_Interp *, arg1) {
  Tcl_Interp *interp;
  Tcl_CmdInfo info;
  va_list argList;
  char *cmd, *arg, **argv;
  int argc, i, max, result;
  CDM_THREAD_SPECIFIC_DATA;

  interp = TCL_VARARGS_START(Tcl_Interp *, arg1, argList);
  if (interp == NULL) interp = CDM_Interp;
  Tcl_ResetResult(interp);

  /*
   * Map from the command to the C procedure...
   */
  cmd = va_arg(argList, char *);
  if (!Tcl_GetCommandInfo(interp, cmd, &info)) {
    /* The command does not seem to exist. Try to execute it anyway, as there is
     * the possibility that can be loaded by unknown... */
    Tcl_Eval(interp, cmd);
    if (!Tcl_GetCommandInfo(interp, cmd, &info)) {
      Tcl_AppendResult(interp, "unknown command \"", cmd, "\"", NULL);
      va_end(argList);
      return TCL_ERROR;
    }
  }

  max = 20; /* Initial size of argument vector */

  /*
   * Check whether the object interface is preferred for this command
   */
  if (info.isNativeObjectProc) {
    Tcl_Obj **objv;
    int objc;

    objv = (Tcl_Obj **) Tcl_Alloc(max * sizeof(Tcl_Obj *));
    objv[0] = Tcl_NewStringObj(cmd, strlen(cmd));
    Tcl_IncrRefCount(objv[0]);
    objc = 1;

    /*
     * Build a vector out of the rest of the arguments...
     */
    while (1) {
      arg = va_arg(argList, char *);
      if (arg == (char *) NULL) {
        objv[objc] = (Tcl_Obj *) NULL;
        break;
      }
      objv[objc] = Tcl_NewStringObj(arg, strlen(arg));
      Tcl_IncrRefCount(objv[objc]);
      objc++;
      if (objc >= max) {
        /* allocate a bigger vector and copy the old one */
        Tcl_Obj **oldv = objv;
        max *= 2;
        objv = (Tcl_Obj **) Tcl_Alloc(max * sizeof(Tcl_Obj *));
        for (i=0; i<objc; i++) {
          objv[i] = oldv[i];
        }
        Tcl_Free((char *) oldv);
      }
    }
    va_end(argList);

    /*
     * Invoke the C procedure
     */
    result = (*info.objProc) (info.objClientData, interp, objc, objv);

    /*
     * Make sure the string value of the result is valid and release our
     * references to the arguments...
     */
    (void) Tcl_GetStringResult(interp);
    for (i=0; i<objc; i++) {
      Tcl_DecrRefCount(objv[i]);
    }
    Tcl_Free((char *) objv);
    return result;
  } /* if(info.isNativeObjectProc) */

  /*
   * If we reach this point, then the requested command is build according to
   * the old string-based approach...
   */
  argv = (char **) Tcl_Alloc(max * sizeof(char *));
  argv[0] = cmd;
  argc = 1;

  /*
   * Build a vectror out of the rest of the arguments
   */
  while (1) {
    arg = va_arg(argList, char *);
    argv[argc] = arg;
    if (arg == (char *) NULL) {
      break;
    }
    argc++;
    if (argc >= max) {
      /* allocate a bigger vector and copy old one */
      char **oldv = argv;
      max *= 2;
      argv = (char **) Tcl_Alloc(max * sizeof(char *));
      for (i=0; i<argc; i++) {
        argv[i] = oldv[i];
      }
      Tcl_Free((char *) oldv);
    }
  }
  va_end(argList);

  /*
   * Invoke the C procedure
   */
  result = (*info.proc) (info.clientData, interp, argc, (const char **) argv);

  /*
   * Release the memory
   */
  Tcl_Free((char *) argv);
  return result;
} /* CDM_Eval */

/*
 * ---- CDM_Exec:
 *  DESCRIPTION:
 *   <English:
 *    This function will accept a variable number of strings that will try to
 *    execute as an operating system command. Plugin modules that want to
 *    execute an external (operating system) command are advised to use this
 *    function instead of the "system" ANSI C function. The main advantage of
 *    CDM_Exec over "system" is the use of the [[b]]Execution Server[[/b]].
 *    The Execution Server acts like a deamon whoose its main purpose is to
 *    execute operating system commands.
 *    Its use is especially encouraged, especially when processing
 *    memory intensive systems, as it eliminates the need for the whole
 *    platform to fork. During a fork, a copy of the parental running process
 *    is created in the memory, thus requiring twice as much memory as is
 *    currently occupied by the parent process. If this process is the
 *    Execution Server, the additional required memory will be a steady amount
 *    for every command that is executed, totally indepented from the amound
 *    of memory that the platform occupies at that particular time instance.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Interp: Interpreter Pointer (Tcl_Interp *)
 *        arg1: A (NULL terminated) sequence of strings (char *)
 *   >
 *  NOTES:
 *   <English:
 *     Note that every string given as an argument to this fuction will be
 *     passed as a single argument to the operating system, even if it
 *     contains spaces. For example, the following call:[[br]][[code]]
 *     [[center]] CDM_Exec(interp, %%%quot;vi%%%quot;, "file with spaces", NULL); [[/center]] [[br]]
 *     is equivelant to the following command (assuming written to a shell):
 *     [[br]][[center]]$%%%gt; vi "file with spaces"[[/center]][[/code]]
 *
 *     An error condition will be raised if the operating system returns an
 *     error condition [[b]]or when the executed command writes anything to
 *     its stderr[[/b]].
 *
 *     Also note that this call is equivelant to the "exec" Tcl command.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *     error occurs, the return value will be TCL_ERROR and an error message
 *     describing the error will be left at the specified Tcl interpreter.
 *     Note that if the executed command writes anything to its
 *     stderr an error will be also returned. If the command completes
 *     succesfully, TCL_OK will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *      CDM_Eval, exec
 *   >:&:
 */
CDM_EXPORT
int CDM_Exec TCL_VARARGS_DEF(Tcl_Interp *, arg1) {
  Tcl_Interp *interp;
  Tcl_CmdInfo info;
  va_list argList;
  const char *cmd;
  char *arg, **argv;
  int argc, i, max, result;
  CDM_THREAD_SPECIFIC_DATA;

  interp = TCL_VARARGS_START(Tcl_Interp *, arg1, argList);
  if (interp == NULL) interp = CDM_Interp;
  Tcl_ResetResult(interp);

  /*
   * Map from the command to the C procedure...
   */
  cmd = "exec";
  if (!Tcl_GetCommandInfo(interp, cmd, &info)) {
    Tcl_AppendResult(interp, "unknown command \"", cmd, "\"", NULL);
    va_end(argList);
    return TCL_ERROR;
  }

  max = 20; /* Initial size of argument vector */

  /*
   * Check whether the object interface is preferred for this command
   */
  if (info.isNativeObjectProc) {
    Tcl_Obj **objv;
    int objc;

    objv = (Tcl_Obj **) Tcl_Alloc(max * sizeof(Tcl_Obj *));
    objv[0] = Tcl_NewStringObj(cmd, strlen(cmd));
    Tcl_IncrRefCount(objv[0]);
    objc = 1;

    /*
     * Build a vector out of the rest of the arguments...
     */
    while (1) {
      arg = va_arg(argList, char *);
      if (arg == (char *) NULL) {
        objv[objc] = (Tcl_Obj *) NULL;
        break;
      }
      objv[objc] = Tcl_NewStringObj(arg, strlen(arg));
      Tcl_IncrRefCount(objv[objc]);
      objc++;
      if (objc >= max) {
        /* allocate a bigger vector and copy the old one */
        Tcl_Obj **oldv = objv;
        max *= 2;
        objv = (Tcl_Obj **) Tcl_Alloc(max * sizeof(Tcl_Obj *));
        for (i=0; i<objc; i++) {
          objv[i] = oldv[i];
        }
        Tcl_Free((char *) oldv);
      }
    }
    va_end(argList);

    /*
     * Invoke the C procedure
     */
    result = (*info.objProc) (info.objClientData, interp, objc, objv);

    /*
     * Make sure the string value of the result is valid and release our
     * references to the arguments...
     */
    (void) Tcl_GetStringResult(interp);
    for (i=0; i<objc; i++) {
      Tcl_DecrRefCount(objv[i]);
    }
    Tcl_Free((char *) objv);
    return result;
  } /* if(info.isNativeObjectProc) */

  /*
   * If we reach this point, then the requested command is build according to
   * the old string-based approach...
   */
  argv = (char **) Tcl_Alloc(max * sizeof(char *));
  argv[0] = (char *) cmd;
  argc = 1;

  /*
   * Build a vectror out of the rest of the arguments
   */
  while (1) {
    arg = va_arg(argList, char *);
    argv[argc] = arg;
    if (arg == (char *) NULL) {
      break;
    }
    argc++;
    if (argc >= max) {
      /* allocate a bigger vector and copy old one */
      char **oldv = argv;
      max *= 2;
      argv = (char **) Tcl_Alloc(max * sizeof(char *));
      for (i=0; i<argc; i++) {
        argv[i] = oldv[i];
      }
      Tcl_Free((char *) oldv);
    }
  }
  va_end(argList);

  /*
   * Invoke the C procedure
   */
  result = (*info.proc) (info.clientData, interp, argc, (const char **) argv);

  /*
   * Release the memeory
   */
  Tcl_Free((char *) argv);
  return result;
} /* CDM_Exec */

/*
 * ---- CDM_ExtractTextBasedOnAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will accept two parameters, a text object (of type
 *     CDM_ByteSequence or CDM_RawData) and a set of Annotations (of type
 *     CDM_AnnotationSet). It will return a new text object (of type
 *     CDM_ByteSequence or CDM_RawData) that will contain the text portions
 *     from the provided text object that are spanned by Annotations in the
 *     specified Annotation set. The portions of the original text that are not
 *     spanned by Annotations will either be replaced with white space
 *     characters or eliminated from the resulting text, depending on the value
 *     of the "mode" parameter. Valid values for the "mode" parameter are
 *     "CDM_EXTRACT_TEXT_LEAVE_UNSPANNED_TEXT" and
 *     "CDM_EXTRACT_TEXT_REMOVE_UNSPANNED_TEXT". The first value will convert text
 *     portions not spanned into white space characters (either spaces (' ') or
 *     new-line characters ('\n')), while the second value will cause the
 *     elimination of unspanned text portions from the resulting text object.
 *
 *     A common use of this function is when we want to perform a linguistic
 *     task only on part of the text contained on the Document, but we want the
 *     resulting Annotations to span the proper text portions, when added to
 *     the original Document. For example, we want to apply a sentence splitter
 *     on an HTML document. As the sentence splitter has no knowledge of HTML
 *     tags, we want to create a version of the Document text where the HTML
 *     tags have been replaced with spaces. Provided that we have Annotations
 *     that span all the HTML tags, we can create this text by simply executing
 *     the following command:
 *     [[br]][[code]][[center]]
 *     Text = CDM_ExtractTextBasedOnAnnotations(DocText, HTML_Annotations, CDM_EXTRACT_TEXT_LEAVE_UNSPANNED_TEXT);
 *     [[/center]][[/code]]
 *
 *     This function will always create and return a new text object,
 *     even when the Annotation set is empty or the specified text object
 *     through the "Text" parameter has no text.
 *     CDM does not own the returned object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Text: A text object.                 (CDM_ByteSequence or CDM_RawData)
 *     Set:  An Annotation set.             (CDM_AnnotationSet)
 *     mode: One of the values CDM_EXTRACT_TEXT_LEAVE_UNSPANNED_TEXT or CDM_EXTRACT_TEXT_REMOVE_UNSPANNED_TEXT (int)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivalent to tip_ExtractTextBasedOnAnnotations
 *     (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will always create and return a new text object,
 *     even when the Annotation set is empty or the specified text object
 *     through the "Text" parameter has no text.
 *     CDM does not own the returned object: the caller is responsible to use
 *     Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *     associated with it.
 *
 *     In case of an error, NULL will be returned and an error message will be
 *     placed in the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *     tip_ExtractTextBasedOnAnnotations
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequence CDM_ExtractTextBasedOnAnnotations(CDM_ByteSequence Text,
                                              CDM_AnnotationSet Set, int mode) {
  CDM_ByteSequence result;
  CDM_SpanSet SpanSet;
  Tcl_UniChar *initialUniText, *finalUniText;
  Tcl_Obj **Ann, **Span;
  long start, end;
  int items, spans, i, j, status, initialUniTextLen;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_ExtractTextBasedOnAnnotations: Aborting...");
    return NULL;
  }

  /*
   * Is the provided Annotation Set a valid list object?
   */
  if (Text == NULL || Set == NULL) return Tcl_NewStringObj(NULL, 0);
  status = Tcl_ListObjGetElements(CDM_Interp, Set, &items, &Ann);
  if (status != TCL_OK) {
    return NULL;
  }

  /*
   * We want to make an exact copy of the Text object, but with non-white
   * characters substituted with spaces...
   */
  initialUniText = Tcl_GetUnicodeFromObj(Text, &initialUniTextLen);
    if (initialUniTextLen == 0) {
    /* Empty input string, just stop now */
    return Tcl_NewStringObj(NULL, 0);
  }
  finalUniText = (Tcl_UniChar *) Tcl_Alloc((size_t)
                                 sizeof(Tcl_UniChar)*(initialUniTextLen+1));
  /*
   * Iterate over all characters in the original text object and replace all
   * non-white space characters with space...
   */
  if (mode == CDM_EXTRACT_TEXT_REMOVE_UNSPANNED_TEXT) {
    memset((void *) finalUniText, 0,
             (size_t) initialUniTextLen * sizeof(Tcl_UniChar));
  } else {
    for (i = 0; i < initialUniTextLen; i++) {
      if (initialUniText[i] == '\n') {
        finalUniText[i] = '\n';
      } else if (initialUniText[i] == '\t') {
        finalUniText[i] = '\t';
      } else if (initialUniText[i] == '\r') {
        finalUniText[i] = '\r';
      } else {
        finalUniText[i] = ' ';
      }
    }
  }
  finalUniText[initialUniTextLen] = '\0';

  /*
   * Now, in the finalUniText string we keep the white-space version of our
   * text. We have to iterate over the Annotations in the provided set,
   * to restore back the characters that are spanned by the provided
   * Annotations...
   */
  for (i = 0; i < items; i++) {
    SpanSet = CDM_GetSpans(Ann[i]);
    if (SpanSet == NULL) {
      Tcl_Free((char *) finalUniText);
      return NULL;
    }
    /* Iterate over all spans... */
    status = Tcl_ListObjGetElements(CDM_Interp, SpanSet, &spans, &Span);
    if (status != TCL_OK) {
      Tcl_Free((char *) finalUniText);
      return NULL;
    }
    for (j = 0; j < spans; j++) {
      status = CDM_GetSpanOffsets(Span[j], &start, &end);
      if (status != TCL_OK) {
        Tcl_Free((char *) finalUniText);
        return NULL;
      }
      /* Check the range of the span: it should not extend outside our text
       * and the start must be lower than end... */
      if (start < 0 || end < 0 || start > end || end > initialUniTextLen) {
        CDM_ReportError("Span \"%s\" out of range for Annotation with Id=%d!"
                        " The valid range for the provided text is [0, %d].",
                        Tcl_GetString(Span[j]), CDM_GetId(Ann[i]),
                        initialUniTextLen-1);
        Tcl_Free((char *) finalUniText);
        return NULL;
      }
      /* Copy the Unicode characters from the original to the targe string */
      memcpy((void *) (&finalUniText[start]),
             (void *) (&initialUniText[start]),
             (size_t) (end - start) * sizeof(Tcl_UniChar));
    }
  } /* for (i=0; i<items; i++) */

  /*
   * If unspanned text is to be removed, remove all instances of the '\0'
   * character from the string... */
  if (mode == CDM_EXTRACT_TEXT_REMOVE_UNSPANNED_TEXT) {
    initialUniText = (Tcl_UniChar *) Tcl_Alloc((size_t)
                                 sizeof(Tcl_UniChar)*(initialUniTextLen+1));
    for (i = 0, j = 0; i < initialUniTextLen; i++) {
      if (finalUniText[i] != '\0') {
        initialUniText[j] = finalUniText[i];
        j++;
      }
    }
    initialUniText[j] = '\0';
    Tcl_Free((char *) finalUniText);
    finalUniText = initialUniText;
    initialUniTextLen = j;
  }
  result = Tcl_NewUnicodeObj(finalUniText, initialUniTextLen);
  Tcl_Free((char *) finalUniText);
  return result;
} /* CDM_ExtractTextBasedOnAnnotations */

/*
 * ---- CDM_SetGetElements:
 *  DESCRIPTION:
 *   <English:
 *    This function returns a count and a pointer to an array of the elements
 *    (object) in a set object (of type CDM_Set or CDM_AnnotationSet or
 *    CDM_AttributeSet or CDM_ByteSequenceSet or CDM_RawDataSet or CDM_SpanSet).
 *    It returns the count by storing it in the address pointed by the "objcPtr"
 *    parameter (of type int *). Similarly, it returns the array pointer
 *    by storing it in the address pointed by the "objvPtr" parameter (of type
 *    CDM_Object ** or CDM_Annotation ** or
 *    CDM_Attribute ** or CDM_ByteSequence ** or CDM_RawData ** or CDM_Span **).
 *    The memory pointed to is managed by CDM and should
 *    not be freed by the caller.
 *
 *    In case of an error, TCL_ERROR will be returned and an error message will
 *    be left in the current active Tcl interpreter (CDM_Interp).
 *    Otherwise TCL_OK will be returned after storing the count and
 *    array pointer.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set: The set whose elements will be placed in the array. (CDM_AttributeSet)
 *     objcPtr:      A pointer to an integer, where the number of objects placed in the array will be saved. (int *)
 *     objvPtr:      A location where the pointer to the array of objects will be stored. (CDM_Object **)
 *   >
 *  NOTES:
 *   <English:
 *    Usage example:[[br]][[code]][[pre]]
 *    int objc;
 *    CDM_Attribute **objv;
 *    CDM_AttributeSet set = CDM_GetAttributes(Doc);
 *    int status = CDM_SetGetElements(set, objc, objv);
 *    if (status != TCL_OK) {
 *      // An error occured. What is the error message?
 *      Tcl_interp *interp = CDM_GetActiveInterpreter();
 *      if (interp) {
 *        const char* msg = Tcl_GetStringResult(interp);
 *      }
 *    }
 *    // Make an iteration over all Attributes...
 *    for (int i = 0; i < objc, i++) {
 *      // Now, objv[i] is an Attribute...
 *    }
 *    [[/pre]][[/code]]
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a count and a pointer to an array of the elements
 *    (objects) in an set object (of type CDM_Set).
 *    It returns the count by storing it in the address pointed by the "objcPtr"
 *    parameter (of type int *). Similarly, it returns the array pointer
 *    by storing it in the address pointed by the "objvPtr" parameter (of type
 *    CDM_Object **). The memory pointed to is managed by CDM and should
 *    not be freed by the caller.
 *
 *    In case of an error, TCL_ERROR will be returned and an error message
 *    describing the error will be left at the current Tcl interpreter
 *    (CDM_Interp). Otherwise TCL_OK will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_AnnotationSetGetElements, CDM_AttributeSetGetElements,
 *    CDM_SpanSetGetElements
 *   >:&:
 */
CDM_EXPORT
int CDM_SetGetElements(CDM_Set Set, int *objcPtr, CDM_Object **objvPtr) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    return CDM_ReportError("CDM_SetGetElements: Aborting...");
  }

  if (Set == NULL) {
    return CDM_ReportError("CDM_SetGetElements: Invalid Set pointer!");
  }

  return Tcl_ListObjGetElements(CDM_Interp, Set, objcPtr, objvPtr);
} /* CDM_SetGetElements */

/*
 * ---- CDM_GetRange:
 *  DESCRIPTION:
 *   <English:
 *    This function will return a newly created object (of type
 *    CDM_ByteSequence or CDM_RawData) comprised of the characters between
 *    start and end (inclusive).
 *
 *    This function will always create and return a new text object
 *    (of type CDM_ByteSequence or CDM_RawData).
 *    CDM does not own the returned object: the caller is responsible to use
 *    Tcl_DecrRefCount or CDM_Free in order to free the object and the memory
 *    associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     textObj: A text object.              (CDM_ByteSequence or CDM_RawData)
 *     start:   An index to the first character to be extracted.  (long)
 *     end:     An index to the last character to be extracted.   (long)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to "string range" Tcl facility.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return a new text object
 *    (of type CDM_ByteSequence or CDM_RawData) containing the extracted text
 *    range specified by the parameters "start" and "end".
 *
 *    In case of an error, NULL will be return and an error message will be
 *    left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     Tcl_GetRange
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequence CDM_GetRange(CDM_ByteSequence textObj, long start, long end) {
  if (textObj == NULL) {
    CDM_ReportError("CDM_GetRange: Invalid text object (NULL)");
    return NULL;
  }
  return Tcl_GetRange(textObj, (int) start, (int) end);
} /* CDM_GetRange */

/*
 * ---- CDM_ActiveInterpreterResultEmpty:
 *  DESCRIPTION:
 *   <English:
 *    This function will check whether the current active Tcl interpreter
 *    (CDM_Interp) holds the empty object as its value. In such a case, 1
 *    (true) will be returned. If the interpreter does not has the empty
 *    object as its value, 0 (false) will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None: void
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return 1 (true) if the current active Tcl
 *     interpreter (CDM_Interp) has the empty ovject as its value, 0 (false)
 *     otherwise. In case the current active Tcl interpreter (CDM_Interp) is
 *     not a valid interpreter, 0 will be returned as well.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_SetActiveInterpreter, CDM_Interp, CDM_FirstAnnotationContaining
 *   >:&:
 */
CDM_EXPORT
int CDM_ActiveInterpreterResultEmpty(void) {
  CDM_CONST char *cptr;
  CDM_THREAD_SPECIFIC_DATA;
  if (CDM_Interp == NULL) return 0;
  cptr = (CDM_CONST char *) Tcl_GetStringResult(CDM_Interp);
  if (cptr == NULL) return 0;
  return !CDM_Strlen(cptr);
} /* CDM_ActiveInterpreterResultEmpty */

/*
 * ---- CDM_GetActiveInterpreter:
 *  DESCRIPTION:
 *   <English:
 *    This function will return the current CDM interpreter.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None: void
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the current CDM interpreter.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_SetActiveInterpreter, CDM_Interp
 *   >:&:
 */
CDM_EXPORT
Tcl_Interp *CDM_GetActiveInterpreter(void) {
  CDM_THREAD_SPECIFIC_DATA;
  return CDM_Interp;
} /* CDM_GetActiveInterpreter */

/*
 * ---- CDM_SetActiveInterpreter:
 *  DESCRIPTION:
 *   <English:
 *    This function will update the CDM current interpreter to the interpreter
 *    specified throught the "Interp" parameter. The CDM current interpreter
 *    is accessible through the global variable CDM_Interp (of type Tcl_Interp
 *    *). This interpreter can be used in any case a Tcl interpreter is
 *    required (either as a parameter to a Tcl library fucntion or to save an
 *    error message). The value of CDM_Interp should never be changed
 *    directly but only through this function. Finally, the value of the
 *    "Interp" parameter should be a valid Tcl interpreter, preferably not a
 *    safe one.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Interp: A pointer to a valid Tcl Interpreter (Tcl_Interp *)
 *   >
 *  NOTES:
 *   <English:
 *     CDM_Interp will always point to the main Tcl interpreter of the
 *     application after CDM initialises.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value of this function will be a standart Tcl completion code
 *     (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     value of the global variable CDM_Interp cannot be changed, the return
 *     value will be TCL_ERROR. No error message will be left at the
 *     current active Tcl interpreter (CDM_Interp). If the update was done
 *     successfully, TCL_OK will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetActiveInterpreter, CDM_Interp
 *   >:&:
 */
CDM_EXPORT
int CDM_SetActiveInterpreter(Tcl_Interp *Interpreter) {
  CDM_THREAD_SPECIFIC_DATA;
  if (Interpreter) {
    CDM_Interp = Interpreter;
    return TCL_OK;
  }
  return TCL_ERROR;
} /* CDM_SetActiveInterpreter */

/*
 * ---- CDM_SetProcessStatus:
 *  DESCRIPTION:
 *   <English:
 *    Updates the Percent Bar displayed when an Annotator runs...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document: Document Pointer                   (CDM_Document)
 *      Status:   Percent (in range [0 - 100])       (int)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
int CDM_SetProcessStatus(CDM_Document Document, int StatusVal) {
  Tcl_Obj *objv[3];
  register int status, i;
  CDM_THREAD_SPECIFIC_DATA;

  if (Document == NULL) {
    CDM_ReportError("CDM_SetProcessStatus: NULL Document Pointer");
    return -1;
  }
  if (Document->Type != CDM_TYPE_DOCUMENT) {
    CDM_ReportError("CDM_SetProcessStatus: Invalid Document Type");
    return -1;
  }
  if (Document->Interp == NULL) {
    if (CDM_Interp) {
      Document->Interp = CDM_Interp;
    } else {
      CDM_Panic("CDM_SetProcessStatus: Cannot locate a valid interpreter!\n");
    }
  }

  if (StatusVal < 0)   StatusVal = 0;
  if (StatusVal > 100) StatusVal = 100;
  Tcl_ResetResult(Document->Interp);
  objv[0] = CDM_NewStringObj("ggi_wait_update");
  objv[1] = CDM_NewStringObj(
            Tcl_GetCommandName(Document->Interp, Document->Tcl_Cmd));
  objv[2] = CDM_NewLongObj(StatusVal);
  for(i=0; i<3; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(CDM_Interp, 3, objv, TCL_EVAL_GLOBAL);
  for(i=0; i<3; i++) Tcl_DecrRefCount(objv[i]);
  return status;
} /* CDM_SetProcessStatus */

/*
 * ---- CDM_GetCollectionPointer:
 *  DESCRIPTION:
 *   <English:
 *    Retrieves a Collection out of a Tcl Object. If retrieval fails, NULL is
 *    returned. This function can be used in order to find out if a Tcl object
 *    holds a Collection or not.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Obj: The Tcl Object under question          (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
CDM_Collection CDM_GetCollectionPointer(Tcl_Obj *Obj) {
  Tcl_CmdInfo Info;
  CDM_Collection Col;
  char *Text;
  CDM_THREAD_SPECIFIC_DATA;

  if (Obj == NULL) goto error;
  Text = Tcl_GetString(Obj);
  if (strncmp(Text, "Col", 3) != 0) goto error;
  if (Tcl_GetCommandInfo(CDM_Interp, Text, &Info) == 0) goto error;
  Col = (CDM_Collection) Info.objClientData;
  if (CDM_ValidCollectionPointer(Col)) return Col;
error:
  return NULL;
} /* CDM_GetCollectionPointer */

/*
 * ---- CDM_GetCollectionPointer:
 *  DESCRIPTION:
 *   <English:
 *    Retrieves a Collection out of a Tcl Object. If retrieval fails, NULL is
 *    returned. This function can be used in order to find out if a Tcl object
 *    holds a Collection or not.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      interp: The interpreter to report errors, if not NULL (Tcl_Interp *)
 *      Obj: The Tcl Object under question          (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
CDM_Collection CDM_GetCollectionPointer(Tcl_Interp *interp, Tcl_Obj *Obj) {
  Tcl_CmdInfo Info;
  CDM_Collection Col;
  char *Text;
  CDM_THREAD_SPECIFIC_DATA;

  if (Obj == NULL) goto error;
  Text = Tcl_GetString(Obj);
  if (strncmp(Text, "Col", 3) != 0) goto error;
  if (Tcl_GetCommandInfo(CDM_Interp, Text, &Info) == 0) goto error;
  Col = (CDM_Collection) Info.objClientData;
  if (CDM_ValidCollectionPointer(Col)) return Col;
error:
  if (interp != NULL) {
    if (Text) CDM_ReportError(interp, "Invalid Collection object \"%s\"", Text);
    else      CDM_ReportError(interp, "Invalid Collection object \"NULL\"");
  }
  return NULL;
} /* CDM_GetCollectionPointer */

/*
 * ---- CDM_GetDocumentPointer:
 *  DESCRIPTION:
 *   <English:
 *    Retrieves a Document out of a Tcl Object. If retrieval fails, NULL is
 *    returned. This function can be used in order to find out if a Tcl object
 *    holds a Document or not.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Obj: The Tcl Object under question          (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
CDM_Document CDM_GetDocumentPointer(Tcl_Obj *Obj) {
  Tcl_CmdInfo Info;
  CDM_Document Doc;
  char *Text = NULL;
  CDM_THREAD_SPECIFIC_DATA;

  if (Obj == NULL) goto error;
  Text = Tcl_GetString(Obj);
  if (strncmp(Text, "Doc", 3) != 0) goto error;
  if (Tcl_GetCommandInfo(CDM_Interp, Text, &Info) == 0) goto error;
  Doc = (CDM_Document) Info.objClientData;
  if (CDM_ValidDocumentPointer(Doc)) return Doc;
error:
  return NULL;
} /* CDM_GetDocumentPointer */

/*
 * ---- CDM_GetDocumentPointer:
 *  DESCRIPTION:
 *   <English:
 *    Retrieves a Document out of a Tcl Object. If retrieval fails, NULL is
 *    returned. This function can be used in order to find out if a Tcl object
 *    holds a Document or not.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      interp: The interpreter to report errors, if not NULL (Tcl_Interp *)
 *      Obj: The Tcl Object under question                    (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
CDM_Document CDM_GetDocumentPointer(Tcl_Interp *interp, Tcl_Obj *Obj) {
  Tcl_CmdInfo Info;
  CDM_Document Doc;
  char *Text = NULL;
  CDM_THREAD_SPECIFIC_DATA;

  if (Obj == NULL) goto error;
  Text = Tcl_GetString(Obj);
  if (strncmp(Text, "Doc", 3) != 0) goto error;
  if (Tcl_GetCommandInfo(CDM_Interp, Text, &Info) == 0) goto error;
  Doc = (CDM_Document) Info.objClientData;
  if (CDM_ValidDocumentPointer(Doc)) return Doc;
error:
  if (interp != NULL) {
    if (Text) CDM_ReportError(interp, "Invalid Document object \"%s\"", Text);
    else      CDM_ReportError(interp, "Invalid Document object \"NULL\"");
  }
  return NULL;
} /* CDM_GetDocumentPointer */

/*
 * ---- CDM_NewLongObj:
 *  DESCRIPTION:
 *   <English:
 *    This function will create and return an object that holds the specified
 *    by the "Value" parameter long integer. The advantage of this function
 *    compared to [[A HREF="../TclTk/TclLib/IntObj.htm"]]Tcl_NewLongObj[[/A]]
 *    is that CDM_NewLongObj will look into the CDM cache before creating a
 *    new object: If a suitable object is found in the cache, the cached
 *    object is returned. Otherwise, a new object is created, appended to the
 *    CDM cache and returned. As a result, calling this function instead of
 *    [[A HREF="../TclTk/TclLib/IntObj.htm"]]Tcl_NewLongObj[[/A]] is more
 *    memory efficient if it is known that many objects holding the same long
 *    integer are required or that the CDM cache already contains a suitable
 *    object. (Remember that the CDM holds numerous objects holding long
 *    integers used as Annotation Ids and in Spans...)
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Value: The long integer that should be placed in the object. (long)
 *   >
 *  NOTES:
 *   <English:
 *    This function does not have an equivalent in the Tcl API.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return an object (of type Tcl_Obj *) that will
 *    contain as value the specified long integer.
 *
 *    Note that the returned object is property of the CDM and should never be
 *    freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_NewStringObj,
 *    [[A HREF="../TclTk/TclLib/IntObj.htm"]]Tcl_NewLongObj[[/A]]
 *   >:&:
 */
CDM_EXPORT
Tcl_Obj *CDM_NewLongObj(long Value) {
  CDM_THREAD_SPECIFIC_DATA;
  return ObjectCache.FindLong(Value);
} /* CDM_NewLongObj */

/*
 * ---- CDM_NewStringObj:
 *  DESCRIPTION:
 *   <English:
 *    This function will create and return an object that holds the specified
 *    by the "Value" parameter string. The advantage of this function
 *    compared to
 *    [[A HREF="../TclTk/TclLib/StringObj.htm"]]Tcl_NewStringObj[[/A]]
 *    is that CDM_NewStringObj will look into the CDM cache before creating a
 *    new object: If a suitable object is found in the cache, the cached
 *    object is returned. Otherwise, a new object is created, appended to the
 *    CDM cache and returned. As a result, calling this function instead of
 *    [[A HREF="../TclTk/TclLib/StringObj.htm"]]Tcl_NewStringObj[[/A]] is more
 *    memory efficient if it is known that many objects holding the same long
 *    integer are required or that the CDM cache already contains a suitable
 *    object. (Remember that the CDM holds numerous objects holding
 *    strings used as Annotation types, Attribute names or Attribute values...)
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Value: The string that should be placed in the object. (char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function does not have an equivalent in the Tcl API.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return an object (of type Tcl_Obj *) that will
 *    contain as value the specified string.
 *
 *    Note that the returned object is property of the CDM and should never be
 *    freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_NewLongObj
 *    [[A HREF="../TclTk/TclLib/StringObj.htm"]]Tcl_NewStringObj[[/A]]
 *   >:&:
 */
CDM_EXPORT
Tcl_Obj *CDM_NewStringObj(char *Value) {
  CDM_THREAD_SPECIFIC_DATA;
  return ObjectCache.Find(Value);
} /* CDM_NewStringObj */
CDM_EXPORT
Tcl_Obj *CDM_NewStringObj(const char *Value) {
  CDM_THREAD_SPECIFIC_DATA;
  return ObjectCache.Find((char *) Value);
} /* CDM_NewStringObj */
Tcl_Obj *CDM_NewStringObj(const char *Value, int len) {
  CDM_THREAD_SPECIFIC_DATA;
  return ObjectCache.Find((char *) Value);
} /* CDM_NewStringObj */


/***********************************************************
 **  Search Utilities
 ***********************************************************/

/*
 * ---- CDM_Lsearch:
 *  DESCRIPTION:
 *   <English:
 *    Tcl's "lsearch" replacement...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      List:    List to be searched                           (Tcl_Obj *)
 *      patObj:  Object holding search pattern                 (Tcl_Obj *)
 *      mode:    One of CDM_LSEARCH_EXACT, CDM_LSEARCH_GLOB, CDM_LSEARCH_REGEXP                     (int)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
Tcl_Obj *CDM_Lsearch(Tcl_Obj *List, Tcl_Obj *patObj, int mode) {
  int index;
  return CDM_Lsearch(List, patObj, &index, mode);
} /* CDM_Lsearch */


/*
 * ---- CDM_Lsearch:
 *  DESCRIPTION:
 *   <English:
 *    Tcl's "lsearch" replacement...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      List:     List to be searched                             (Tcl_Obj *)
 *      patObj:   Object holding search pattern                   (Tcl_Obj *)
 *      position: Integer pointer to store found object's index   (int)
 *      mode:     One of CDM_LSEARCH_EXACT, CDM_LSEARCH_GLOB, CDM_LSEARCH_REGEXP                       (int)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
Tcl_Obj *CDM_Lsearch(Tcl_Obj *List, Tcl_Obj *patObj, int *position, int mode) {
  char *bytes, *patternBytes;
  int i, match, index, result, listc, length, elemLen;
  Tcl_Obj **listv;
  CDM_THREAD_SPECIFIC_DATA;

  patternBytes = Tcl_GetStringFromObj(patObj, &length);
  /*
   * Make sure the list argument is a list object and get its length and
   * a pointer to its array of element pointers.
   */
  result = Tcl_ListObjGetElements(CDM_Interp, List, &listc, &listv);
  if (result != TCL_OK) {
    return NULL;
  }
  index = -1;
  for (i = 0; i < listc; i++) {
    match = 0;
    bytes = Tcl_GetStringFromObj(listv[i], &elemLen);
    switch (mode) {
      case CDM_LSEARCH_EXACT: {
        if (length == elemLen) {
          match = (memcmp(bytes, patternBytes, (size_t) length) == 0);
        }
        break;
      }
      case CDM_LSEARCH_GLOB: {
        match = Tcl_StringMatch(bytes, patternBytes);
        break;
      }
      case CDM_LSEARCH_REGEXP: {
        match = Tcl_RegExpMatch(CDM_Interp, bytes, Tcl_GetString(patObj));
        if (match < 0) return NULL;
        break;
      }
    }
    if (match != 0) {
      index = i;
      break;
    }
  }
  *position = index;
  if (index < 0) return NULL;
  return listv[i];
} /* CDM_Lsearch */

/*
 * ---- CDM_LappendUnique:
 *  DESCRIPTION:
 *   <English:
 *    This function can be used in order to append a Tcl Object at the end of
 *    a valid Tcl List Object. The new element will be appended only if it is
 *    not already contained in the list.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      List:   List where the element is going to be appended.    (Tcl_Obj *)
 *      Obj:    The element that is to be appended.                (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *     error occurs, the return value will be TCL_ERROR and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp). If all annotators run without errors, TCL_OK will be
 *     returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Lsearch
 *   >:&:
 */
CDM_EXPORT
int CDM_LappendUnique(Tcl_Obj *List, Tcl_Obj *Obj) {
  CDM_THREAD_SPECIFIC_DATA;
  if (CDM_Lsearch(List, Obj, CDM_LSEARCH_EXACT) == NULL) {
    Tcl_ListObjAppendElement(CDM_Interp, List, Obj);
    return TCL_OK;
  }
  return TCL_ERROR;
} /* CDM_LappendUnique */

/*
 * ---- CDM_SplitString:
 *  DESCRIPTION:
 *   <English:
 *    This function will returns a list object (of type Tcl_Obj *) created by
 *    splitting [[i]]string[[/i]] at each character that is in the
 *    [[i]]splitChars[[/i]] argument.
 *    Each element of the result list will consist of the characters from
 *    string that lie between instances of the characters in splitChars. Empty
 *    list elements will be generated if string contains adjacent characters in
 *    splitChars, or if the first or last character of string is in splitChars.
 *    If splitChars is an empty string then each character of string becomes a
 *    separate element of the result list. SplitChars defaults to the
 *    standard white-space characters.
 *
 *    The last argument, [[i]]allowEmptyElements[[/i]] is a boolean flag that
 *    controls whether the resulting list object is allowed to contain empty
 *    elements or not. If > 0, the returned list object can contain empty
 *    elements that can be created if any of the splitting characters is found
 *    just after another splitting character. If this argument is 0 (false),
 *    then no empty elements will be inserted in the returned list object.
 *    This parameter if ommited, defaults to 0 (false).
 *
 *    CDM does not own the returned list object:
 *    the caller is responsible to use
 *    [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *    CDM_Free in order to free the object and the memory
 *    associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      string:     The string to be split.    (char *)
 *      splitChars: A string containing the split characters. Defaults to " \n\t\r". (char *)
 *      allowEmptyElements: An integer flag.    (int)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivalent to the
 *    %%%quot;[[A HREF="../TclTk/TclCmd/split.htm"]]split[[/A]]%%%quot;
 *    Tcl command.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function creates and returns a newly created list object (of type
 *     Tcl_Obj *).
 *     The caller is responsible to free this object through either
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]] or
 *     CDM_Free.
 *
 *     In case of an NULL will be returned and an error message will be
 *     left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]],
 *    [[A HREF="../TclTk/TclCmd/split.htm"]]split[[/A]]
 *   >:&:
 */
CDM_EXPORT
Tcl_Obj *CDM_SplitString(const char *string, const char *splitChars,
                                             int allowEmptyElements) {
  Tcl_UniChar ch;
  int len;
  const char *end;
  int splitCharLen, stringLen;
  Tcl_Obj *listPtr, *objPtr;

  if (splitChars == NULL) {
    splitChars = (char *) " \n\t\r";
    splitCharLen = 4;
  } else {
    splitCharLen = CDM_Strlen(splitChars);
  }

  stringLen = CDM_Strlen(string);
  end = string + stringLen;
  listPtr = Tcl_NewListObj(0, NULL);

  if (stringLen == 0) {
      /*
       * Do nothing.
       */
  } else if (splitCharLen == 0) {
      Tcl_HashTable charReuseTable;
      Tcl_HashEntry *hPtr;
      int isNew;

      /*
       * Handle the special case of splitting on every character.
       *
       * Uses a hash table to ensure that each kind of character has
       * only one Tcl_Obj instance (multiply-referenced) in the
       * final list.  This is a *major* win when splitting on a long
       * string (especially in the megabyte range!) - DKF
       */

      Tcl_InitHashTable(&charReuseTable, TCL_ONE_WORD_KEYS);
      for ( ; string < end; string += len) {
          len = Tcl_UtfToUniChar(string, &ch);
          /* Assume Tcl_UniChar is an integral type... */
          hPtr = Tcl_CreateHashEntry(&charReuseTable, (char*)0 + ch, &isNew);
          if (isNew) {
              objPtr = Tcl_NewStringObj(string, len);
              /* Don't need to fiddle with refcount... */
              Tcl_SetHashValue(hPtr, (ClientData) objPtr);
          } else {
              objPtr = (Tcl_Obj*) Tcl_GetHashValue(hPtr);
          }
          Tcl_ListObjAppendElement(NULL, listPtr, objPtr);
      }
      Tcl_DeleteHashTable(&charReuseTable);
  } else {
      const char *element, *p, *splitEnd;
      int splitLen;
      Tcl_UniChar splitChar;

      /*
       * Normal case: split on any of a given set of characters.
       * Discard instances of the split characters.
       */

      splitEnd = splitChars + splitCharLen;

      if (allowEmptyElements) {
        for (element = string; string < end; string += len) {
          len = Tcl_UtfToUniChar(string, &ch);
          for (p = splitChars; p < splitEnd; p += splitLen) {
            splitLen = Tcl_UtfToUniChar(p, &splitChar);
            if (ch == splitChar) {
              objPtr = Tcl_NewStringObj(element, string - element);
              Tcl_ListObjAppendElement(NULL, listPtr, objPtr);
              element = string + len;
              break;
            }
          }
        }
        objPtr = Tcl_NewStringObj(element, string - element);
        Tcl_ListObjAppendElement(NULL, listPtr, objPtr);
      } else {
        for (element = string; string < end; string += len) {
          len = Tcl_UtfToUniChar(string, &ch);
          for (p = splitChars; p < splitEnd; p += splitLen) {
            splitLen = Tcl_UtfToUniChar(p, &splitChar);
            if (ch == splitChar) {
              if (string - element) {
                objPtr = Tcl_NewStringObj(element, string - element);
                Tcl_ListObjAppendElement(NULL, listPtr, objPtr);
              }
              element = string + len;
              break;
            }
          }
        }
        if (string - element) {
          objPtr = Tcl_NewStringObj(element, string - element);
          Tcl_ListObjAppendElement(NULL, listPtr, objPtr);
        }
      }
  }
  return listPtr;
} /* CDM_SplitString */

CDM_EXPORT
Tcl_Obj *CDM_SplitString(const char *string) {
  return CDM_SplitString(string, NULL, 0);
} /* CDM_SplitString */

CDM_EXPORT
Tcl_Obj *CDM_SplitString(const char *string, int allowEmptyElements) {
  return CDM_SplitString(string, NULL, allowEmptyElements);
} /* CDM_SplitString */

CDM_EXPORT
Tcl_Obj *CDM_SplitString(const char *string, char *splitChars) {
  return CDM_SplitString(string, splitChars, 0);
} /* CDM_SplitString */

/***********************************************************
 **  File Utilities
 ***********************************************************/

/*
 * ---- CDM_DeleteFile:
 *  DESCRIPTION:
 *   <English:
 *    Deletes a file by its filename...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      path:     The filename                        (char *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
int CDM_DeleteFile(char *path) {
  Tcl_Obj *objv[5];
  int status, i;
  CDM_THREAD_SPECIFIC_DATA;

  /* Execute Tcl Command "file delete -force --"... */
  objv[0] = Tcl_NewStringObj("file", 4);
  objv[1] = Tcl_NewStringObj("delete", 6);
  objv[2] = Tcl_NewStringObj("-force", 6);
  objv[3] = Tcl_NewStringObj("--", 2);
  objv[4] = Tcl_NewStringObj(path, -1);
  for (i=0; i<5; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(CDM_Interp, 5, objv, TCL_EVAL_GLOBAL);
  for (i=0; i<5; i++) Tcl_DecrRefCount(objv[i]);
  return status;
} /* CDM_DeleteFile */


/*
 * ---- CDM_FileNameDirname:
 *  DESCRIPTION:
 *   <English:
 *    Given a filename, returns its full path (just like "file dirname" Tcl
 *    command)...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Filename:     The filename                        (char *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
char *CDM_FileNameDirname(char *FileName) {
  Tcl_DString ds;
  char **PathParts = NULL, *DirName = NULL;
  int items;

  Tcl_SplitPath(FileName, &items, (const char ***) &PathParts);
  if (items > 1) {
     Tcl_DStringInit(&ds);
     Tcl_JoinPath(items - 1, PathParts, &ds);
     DirName = CDM_Strdup(Tcl_DStringValue(&ds));
     Tcl_DStringFree(&ds);
  } else if ((items == 0)||(Tcl_GetPathType(PathParts[0])==TCL_PATH_RELATIVE)) {
#ifdef MAC
     DirName = CDM_Strdup(":");
#else
     DirName = CDM_Strdup(":");
#endif
  } else {
     DirName = CDM_Strdup(PathParts[0]);
  }
  Tcl_Free((char *) PathParts);
  return DirName;
} /* CDM_FileNameDirname */


/*
 * ---- CDM_FileNameTail:
 *  DESCRIPTION:
 *   <English:
 *    Given a filename, returns its tail (element after last directory
 *      separator, just like "file tail" Tcl command)...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Filename:     The filename                        (char *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
char *CDM_FileNameTail(char *FileName) {
  char **PathParts = NULL, *Tail = NULL;
  int items;

  Tcl_SplitPath(FileName, &items, (const char ***) &PathParts);
  if (items > 0) {
     if ((items > 1) || (Tcl_GetPathType(PathParts[0]) == TCL_PATH_RELATIVE)) {
       Tail = CDM_Strdup(PathParts[items - 1]);
     }
  }
  Tcl_Free((char *) PathParts);
  return Tail;
} /* CDM_FileNameTail */


/*
 * ---- CDM_GetNativeFileName:
 *  DESCRIPTION:
 *   <English:
 *    Given a filename, returns the platform specific filename...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Filename:     The filename                        (char *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetNativeFileName(char *FileName) {
  Tcl_DString ds;
  char *Native;
  CDM_THREAD_SPECIFIC_DATA;

  /* Get Native File Name */
  Native = (char *) Tcl_TranslateFileName(CDM_Interp, FileName, &ds);
  if (Native != NULL) {
    Native = CDM_Strdup(Native);
  }
  Tcl_DStringFree(&ds);
  return Native;
} /* CDM_GetNativeFileName */

/*
 * ---- CDM_GetIP:
 *  DESCRIPTION:
 *   <English:
 *    This function will return the network IP of the machine running
 *    the current process.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None.
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return a the network IP of the machine running
 *    the current process as a string (of type char *). The returned string is
 *    property of the CDM and the caller is not supposed to modify or
 *    free it.
 *
 *    In case of an error, NULL is returned and an error message is left at
 *    the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    [[A HREF="../TclTk/TclCmd/socket.htm"]]socket[[/A]]
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetIP(void) {
  CDM_THREAD_SPECIFIC_DATA;
  /*
   * In the following object we keep an object holding the IP of
   * the machine we are running on. It is used and managed only by CDM_GetIP.
   */
  if (CDM_MachineIPObject != NULL) {
    Tcl_DecrRefCount(CDM_MachineIPObject);
    CDM_MachineIPObject = NULL;
  }
  /*
   * In order to get the IP, we have to execute the following Tcl code:
   *   set me [socket -server garbage_word -myaddr [info hostname] 0]
   *   set IP [lindex [fconfigure $me -sockname] 0]
   *   close $me
   */
  strcpy(CDM_Temp, "set CDM_Socket [socket -server garbage_word -myaddr "
                   "[info hostname] 0]");
  if (Tcl_Eval(CDM_Interp, CDM_Temp) != TCL_OK) {
    /*
     * Perhaps the host [info hostname] is unreachable. Try also
     * "localhost"...
     */
    strcpy(CDM_Temp, "set CDM_Socket [socket -server garbage_word -myaddr "
                     "localhost 0]");
    if (Tcl_Eval(CDM_Interp, CDM_Temp) != TCL_OK) return NULL;
  }
  /*
   * If we got here, the socket was succesfully opened...
   */
  strcpy(CDM_Temp, "set CDM_IP [lindex [fconfigure $CDM_Socket -sockname] 0];"
                   "close $CDM_Socket");
  if (Tcl_Eval(CDM_Interp, CDM_Temp) != TCL_OK) return NULL;
  CDM_MachineIPObject = Tcl_GetVar2Ex(CDM_Interp, "CDM_IP", NULL, 0);
  if (CDM_MachineIPObject == NULL) return NULL;
  Tcl_IncrRefCount(CDM_MachineIPObject);
  Tcl_UnsetVar(CDM_Interp, "CDM_IP", 0);
  return Tcl_GetString(CDM_MachineIPObject);
} /* CDM_GetIP */

/*
 * ---- CDM_GetPid:
 *  DESCRIPTION:
 *   <English:
 *    This function will return the process identifier (PID) of the current
 *    process.
 *    All process identifiers are returned as decimal numbers.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None.
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivalent to the
 *    [[A HREF="../TclTk/TclCmd/pid.htm"]]pid[[/A]] Tcl command.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return a decimal number (of type int) representing
 *    process identifier (PID) of the current process.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    [[A HREF="../TclTk/TclCmd/pid.htm"]]pid[[/A]]
 *   >:&:
 */
CDM_EXPORT
int CDM_GetPid(void) {
  CDM_THREAD_SPECIFIC_DATA;
  return CDM_Pid;
} /* CDM_GetPid */

/*
 * ---- CDM_GetTempDir:
 *  DESCRIPTION:
 *   <English:
 *    Returns the platform's Temporary Directory...
 *   >
 *  ARGUMENTS:
 *   <English:
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
char *CDM_GetTempDir(void) {
  CDM_THREAD_SPECIFIC_DATA;
  return CDM_TempDir;
} /* CDM_GetTempDir */

/*
 * ---- CDM_GetCurrentThread:
 *  DESCRIPTION:
 *   <English:
 *    This function returns an indentifier for the current thread (of type
 *    Tcl_ThreadID), casted to a (void *) pointer. Note that this function is
 *    a wrapper to the Tcl_GetCurrentThread procedure.
 *   >
 *  ARGUMENTS:
 *   <English:
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
void *CDM_GetCurrentThread(void) {
  return (void *) Tcl_GetCurrentThread();
} /* CDM_GetCurrentThread */

/*
 * ---- CDM_JoinPaths:
 *  DESCRIPTION:
 *   <English:
 *    Returns a full (platform-depented) path given the path pieces...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      a list of "char *" path pieces...
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
char *CDM_JoinPaths TCL_VARARGS_DEF(char *, arg1) {
  Tcl_DString ds;
  va_list argList;
  char *root, **argv, *arg, *result;
  int max = 20, argc;

  root = TCL_VARARGS_START(char *, arg1, argList);
  if (root == NULL) return NULL;
  argv = (char **) Tcl_Alloc(max * sizeof(char *));
  argv[0] = root;
  argc = 1;

  /*
   * Build a vectror out of the rest of the arguments
   */
  while (1) {
    arg = va_arg(argList, char *);
    argv[argc] = arg;
    if (arg == (char *) NULL) {
      break;
    }
    argc++;
    if (argc >= max) {
      /* allocate a bigger vector and copy old one */
      char **oldv = argv;
      int i;
      max *= 2;
      argv = (char **) Tcl_Alloc(max * sizeof(char *));
      for (i=0; i<argc; i++) {
        argv[i] = oldv[i];
      }
      Tcl_Free((char *) oldv);
    }
  }
  va_end(argList);

  Tcl_DStringInit(&ds);
  result = CDM_Strdup((char *) Tcl_JoinPath(argc, argv, &ds));
  Tcl_DStringFree(&ds);
  return result;
} /* CDM_JoinPaths */

/*
 * ---- CDM_JoinSetElements:
 *  DESCRIPTION:
 *   <English:
 *    This function will create and return a newly created object (of type
 *    CDM_ByteSequence or CDM_RawData) whose string will be the
 *    junction of all the set elements.
 *
 *    Note that the returned object is not owned by the CDM. The caller is
 *    responsible to use CDM_Free (or Tcl_Free) in order to free the string and
 *    the memory associated with it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Set: The set whose elements will be joined. (CDM_Set)
 *    joinString: A NULL-terminated string to be used as a delimiter among elements. (const char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivalent to the "join" Tcl command.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    A newly created text object (of type CDM_ByteSequence or CDM_RawData)
 *    containing the merged string. CDM does not own the
 *    returned object: the caller is responsible to use CDM_Free (or
 *    Tcl_Free) in order to free the object when it is not needed any more.
 *
 *    In case of an error NULL will be returned and an error message will be
 *    left at the current active Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Free, Tcl_Free
 *   >:&:
 */
CDM_EXPORT
CDM_ByteSequence CDM_JoinSetElements(CDM_Set Set, const char *joinString) {
  Tcl_Obj **objElements, *strObj = NULL, *value;
  int status, elements, i;
  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Has the user requested an abort?
   */
  if (CDM_AbortCurrentProcess) {
    if (CDM_UnLocked) CDM_AbortCurrentProcess = 0;
    CDM_ReportError("CDM_JoinSetElements: Aborting...");
    return NULL;
  }

  if (Set == NULL) {
    CDM_ReportError("CDM_JoinSetElements: Invalid Set!");
    return NULL;
  }

  if (joinString) {
    strObj = Tcl_NewStringObj(joinString, -1);
    if (strObj) Tcl_IncrRefCount(strObj);
  }

  status = Tcl_ListObjGetElements(CDM_Interp, Set, &elements, &objElements);
  if (status != TCL_OK) return NULL;
  value = Tcl_NewStringObj(NULL, 0);
  if (value == NULL) return NULL;
  elements--;
  for (i=0; i<=elements; i++) {
    Tcl_AppendObjToObj(value, objElements[i]);
    if (strObj && i != elements) Tcl_AppendObjToObj(value, strObj);
  }
  if (strObj) Tcl_DecrRefCount(strObj);
  return value;
} /* CDM_JoinSetElements */

/*
 * ---- CDM_Mkdir:
 *  DESCRIPTION:
 *   <English:
 *    This function will create the directory specified.
 *    This function will create all non-existing parent directories as well as
 *    the specified directory itself.
 *    If an existing directory is specified, then no action is taken and no
 *    error is returned. Trying to overwrite an existing file with a
 *    directory will result in an error.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Directory:     The directory to be created (char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivalent to the "file mkdir directory" Tcl command.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return TCL_OK when the directory has been succesfully
 *    created, or a directory with the same name already exists. In case of an
 *    error, TCL_ERROR will be returned and an error will be left at the
 *    current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
int CDM_Mkdir(char *Directory) {
  Tcl_Obj *objv[3];
  int i, status;
  CDM_THREAD_SPECIFIC_DATA;

  /* Execute Tcl Command "file mkdir"... */
  objv[0] = Tcl_NewStringObj("file", 4);
  objv[1] = Tcl_NewStringObj("mkdir", 5);
  objv[2] = Tcl_NewStringObj(Directory, -1);
  for (i=0; i<3; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(CDM_Interp, 3, objv, TCL_EVAL_GLOBAL);
  for (i=0; i<3; i++) Tcl_DecrRefCount(objv[i]);
  return status;
} /* CDM_Mkdir */

/*
 * ---- CDM_FileExists:
 *  DESCRIPTION:
 *   <English:
 *    This function can be used to check whether a path (either a file or a
 *    directory) exists. This function returns 1 if the provided path exists
 *    and the current user has search privileges for the directories leading
 *    to it, 0 otherwise.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    path: The file or directory to be checked for existence (const char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivalent to the "file exists path" Tcl command.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return 1 when the specified path exists and all the
 *    parent directories are readable from the current user, 0 otherwise.
 *    This function will never report an error. In case of an error, 0 will be
 *    returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
int CDM_FileExists(const char *path) {
  /* The temporary object here will be freed by CDM_FileExists... */
  return CDM_FileExists(Tcl_NewStringObj(path, -1));
} /* CDM_FileExists */

/*
 * ---- CDM_FileExists:
 *  DESCRIPTION:
 *   <English:
 *    This function can be used to check whether a path (either a file or a
 *    directory) exists. This function returns 1 if the provided path exists
 *    and the current user has search privileges for the directories leading
 *    to it, 0 otherwise.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    path: The file or directory to be checked for existence (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivalent to the "file exists path" Tcl command.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return 1 when the specified path exists and all the
 *    parent directories are readable from the current user, 0 otherwise.
 *    This function will never report an error. In case of an error, 0 will be
 *    returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
int CDM_FileExists(Tcl_Obj *path) {
  Tcl_Obj *objv[3];
  int i, status;
  CDM_THREAD_SPECIFIC_DATA;

  /* Execute Tcl Command "file mkdir"... */
  objv[0] = Tcl_NewStringObj("file", 4);
  objv[1] = Tcl_NewStringObj("exists", 6);
  objv[2] = path;
  for (i=0; i<3; i++) Tcl_IncrRefCount(objv[i]);
  status = Tcl_EvalObjv(CDM_Interp, 3, objv, TCL_EVAL_GLOBAL);
  for (i=0; i<3; i++) Tcl_DecrRefCount(objv[i]);
  if (status != TCL_OK) return 0;
  status = 0;
  Tcl_GetBooleanFromObj(NULL, Tcl_GetObjResult(CDM_Interp), &status);
  return status;
} /* CDM_FileExists */

/***********************************************************
 **  Memory Allocation
 ***********************************************************/

/*
 * ---- CDM_Malloc:
 *  DESCRIPTION:
 *   <English:
 *    Allocates memory. This function must be used instead standart "malloc",
 *    especially when building a shared object. It arranges things correctly,
 *    so as the memory allocated be visible by the main program and by the
 *    shared object (a shared library or a dll under windows). The whole CDM
 *    uses this routine for ALL memory allocations...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      size:   Size of requested space.      (size_t)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
char *CDM_Malloc(size_t size) {
  char *ptr;

  ptr = (char *) Tcl_Alloc((int) size);
  if (ptr == NULL) CDM_Panic("CDM_Malloc: Cannot Allocate Memory!\n");
  return ptr;
} /* CDM_Malloc */

/***********************************************************
 **  Duplicating UTF-8 Strings
 ***********************************************************/

/*
 * ---- CDM_Strdup:
 *  DESCRIPTION:
 *   <English:
 *    This function duplicates the provided by the "string" parameter UTF-8
 *    string, just as ANSI C "strdup" duplicates ASCII strings. A new memory
 *    space large enough to hold the provided string will be allocated and the
 *    provided string will be copied to this new allocated buffer. Memory for
 *    the new string is obtained with the use of Tcl_Alloc, and can be freed
 *    by either calling Tcl_Free or CDM_Free. The return value of this
 *    function is a pointer to this newly allocated memory buffer.
 *
 *    CDM applications should use this routine instead of [[b]]strdup[[/b]],
 *    as [[b]]strdup DOES NOT HANDLE CORRECTLY UTF-8 strings[[/b]]!
 *   >
 *  ARGUMENTS:
 *   <English:
 *      string:     The string to be duplicated            (char *)
 *   >
 *  NOTES:
 *   <English:
 *    Note that ANSI C "strdup" may not operate correctly on UTF-8 strings,
 *    as results may vary according to the operating system and
 *    internationalisation environment used.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a pointer (of type char *) to a newly allocated
 *    memory space where the provided UTF-8 string has been copied. Note that
 *    CDM does not own this memory space and the caller is responsible for
 *    freing it by using CDM_Free or Tcl_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Free, Tcl_Free
 *   >:&:
 */
CDM_EXPORT
char *CDM_Strdup(const char *string) {
  char *duplicate;
  int len;
  //Tcl_NumUtfChars(string, -1);
  //duplicate = CDM_Malloc(sizeof(char) * (strlen(string) + 1));
  //strcpy(duplicate, string);

  // The following simple solution, causes many problems. The passed char can
  // be in utf-8 format, so functions like strlen & strdup may not work ok...
  // duplicate = strdup(string);

  len = strlen(string);
  duplicate = CDM_Malloc( (size_t) ((len) + 1) );
  if (len > 0) {
    memcpy((VOID *) duplicate, (VOID *) string, (unsigned) len);
  }
  duplicate[len] = '\0';
  return duplicate;
} /* CDM_Strdup */
CDM_EXPORT
char *CDM_Strdup(char *string) {
  return CDM_Strdup((const char *) string);
};/* CDM_Strdup */

/*
 * ---- CDM_Strdup:
 *  DESCRIPTION:
 *   <English:
 *    This function duplicates the UTF-8 string contained in the (CDM or Tcl)
 *    object specified by the "stringObj" parameter. A new memory
 *    space large enough to hold the string will be allocated and the
 *    string will be copied to this new allocated buffer. Memory for
 *    the new string is obtained with the use of Tcl_Alloc, and can be freed
 *    by either calling Tcl_Free or CDM_Free. The return value of this
 *    function is a pointer to this newly allocated memory buffer.
 *
 *    CDM applications should use this routine instead of [[b]]strdup[[/b]],
 *    as [[b]]strdup DOES NOT HANDLE CORRECTLY UTF-8 strings[[/b]]!
 *   >
 *  ARGUMENTS:
 *   <English:
 *    stringObj: The object containing the string to be duplicated (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *    Note that ANSI C "strdup" may not operate correctly on UTF-8 strings,
 *    as results may vary according to the operating system and
 *    internationalisation environment used.
 *
 *    Also note that although the type of the "stringObj" parameter is
 *    (Tcl_Obj *), this function will accept any of the following types:[[br]]
 *    [[b]] CDM_Annotation, CDM_AnnotationSet, CDM_Attribute, CDM_AttributeSet,
 *    CDM_AttributeName, CDM_AttributeValue, CDM_ByteSequence,
 *    CDM_ByteSequenceSet, CDM_RawData, CDM_RawDataSet, CDM_SpanSet, CDM_Span
 *    (Tcl_Obj *)[[/b]]
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a pointer (of type char *) to a newly allocated
 *    memory space where the provided UTF-8 string has been copied. Note that
 *    CDM does not own this memory space and the caller is responsible for
 *    freing it by using CDM_Free or Tcl_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Free, Tcl_Free
 *   >:&:
 */
CDM_EXPORT
char *CDM_Strdup(Tcl_Obj *stringObj) {
  if (stringObj == NULL) return NULL;
  return CDM_Strdup(Tcl_GetString(stringObj));
} /* CDM_Strdup */

/*
 * ---- CDM_Strlen:
 *  DESCRIPTION:
 *   <English:
 *    Returns the length of a UTF-8 string...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      string:     The string to be lengthed            (char *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
long CDM_Strlen(CDM_CONST char *string) {
  if (string == NULL) return 0;
  return (long) Tcl_NumUtfChars(string, -1);
} /* CDM_Strlen */

/*
 * ---- CDM_Strcmp:
 *  DESCRIPTION:
 *   <English:
 *    Compare two UTF-8 strings...
 *   >
 *  ARGUMENTS:
 *   <English:
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
int CDM_Strcmp(CDM_CONST char *str1, CDM_CONST char *str2) {
  register long str1_len, str2_len, res;

  if (str1 == NULL && str2 != NULL) return -1;
  if (str1 != NULL && str2 == NULL) return  1;
  if (str1 == NULL && str2 == NULL) return  0;
  str1_len = (long) Tcl_NumUtfChars(str1, -1);
  str2_len = (long) Tcl_NumUtfChars(str2, -1);
  res = Tcl_UtfNcmp(str1, str2, (unsigned long)
                                (str1_len<str2_len)?str1_len:str2_len);
  if ((res == 0) && (str1_len != str2_len)) {
    if (str1_len < str2_len) return -1;
    else                     return  1;
  }
  return res;
} /* CDM_Strcmp */

/*
 * ---- CDM_Strcmp:
 *  DESCRIPTION:
 *   <English:
 *    Compare two UTF-8 strings...
 *   >
 *  ARGUMENTS:
 *   <English:
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
int CDM_Strcmp(Tcl_Obj *str1, Tcl_Obj *str2) {

  if (str1 == NULL && str2 != NULL) return -1;
  if (str1 != NULL && str2 == NULL) return  1;
  if (str1 == NULL && str2 == NULL) return  0;

  return CDM_Strcmp(Tcl_GetString(str1), Tcl_GetString(str2));
} /* CDM_Strcmp */

/*
 * ---- CDM_Strncmp:
 *  DESCRIPTION:
 *   <English:
 *    Compare two UTF-8 strings...
 *   >
 *  ARGUMENTS:
 *   <English:
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
int CDM_Strncmp(CDM_CONST char *str1, CDM_CONST char *str2, long len) {
  if (str1 == NULL && str2 != NULL) return -1;
  if (str1 != NULL && str2 == NULL) return  1;
  if (str1 == NULL && str2 == NULL) return  0;
  if (len < 0) {
    register long str1_len, str2_len;
    str1_len = (long) Tcl_NumUtfChars(str1, -1);
    str2_len = (long) Tcl_NumUtfChars(str2, -1);
    len = (str1_len<str2_len) ? str1_len : str2_len;
  }
  return Tcl_UtfNcmp(str1, str2, (unsigned long) len);
} /* CDM_Strncmp */

/*
 * ---- CDM_Isdigit:
 *  DESCRIPTION:
 *   <English:
 *    This function returns true if the character denoted by the "c" parameter
 *    is a digit.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      c: A character to be decided if is a digit or not. (char)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to the ANSI C function "isdigit".
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a value greater than 0, if the character
 *     contained in parameter "c" represents an integer. Else, 0 will be
 *     returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     isdigit (defined in [[font color=#ff00ff]]%%%lt;ctype.h%%%gt;[[/font]])
 *   >:&:
 */
CDM_EXPORT
int CDM_Isdigit(CDM_CONST char c) {
  switch (c) {
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9': {return 1;}
    default:  {return 0;}
  }
  return 0;
} /* CDM_Isdigit */

/*
 * ---- CDM_ExternalToUtf:
 *  DESCRIPTION:
 *   <English:
 *    This function will convert a string (referenced by "str") from a
 *    specified encoding (provided by the "encoding" parameter) to a UTF-8
 *    string.
 *
 *    This function will always create and return a new string buffer,
 *    where the converted string will be placed. The user must call CDM_Free
 *    in order to free the returned value, in case it is not needed any more.
 *
 *    The "encoding" parameter can be NULL. In this case, the system encoding
 *    will be used.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      encoding:     The encoding of the initial string.      (const char *)
 *      str:          A pointer to the string to be converted. (const char *)
 *   >
 *  NOTES:
 *   <English:
 *    Note that the "encoding" paramemter can be NULL.
 *    In this case, the system encoding will be used.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will always create and return a new string buffer,
 *    where the converted string will be placed. The user must call CDM_Free
 *    in order to free the returned value, in case it is not needed any more.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    Tcl_ExternalToUtf, Tcl_ExternalToUtfDString
 *   >:&:
 */
CDM_EXPORT
char *CDM_ExternalToUtf(const char *encoding, const char *str) {
  Tcl_DString Text;
  Tcl_Encoding enc;
  char *result;
  CDM_THREAD_SPECIFIC_DATA;

  if (encoding == NULL) {
    enc = NULL;
  } else {
    enc = Tcl_GetEncoding(CDM_Interp, encoding);
    if (enc == NULL) return NULL; // Tcl_GetEncoding has left an error...
  }

  Tcl_DStringInit(&Text);
  result = CDM_Strdup((char *) Tcl_ExternalToUtfDString(enc, str, -1, &Text));
  Tcl_DStringFree(&Text);
  Tcl_FreeEncoding(enc);
  return result;
} /* CDM_ExternalToUtf */

/*
 * ---- CDM_ExternalToUtf:
 *  DESCRIPTION:
 *   <English:
 *    This function will convert a string (referenced by "str") from a
 *    specified encoding (provided by the "encoding" parameter) to a UTF-8
 *    string.
 *
 *    This function will always create and return a new string buffer,
 *    where the converted string will be placed. The user must call CDM_Free
 *    in order to free the returned value, in case it is not needed any more.
 *
 *    The "encoding" parameter can be NULL. In this case, the system encoding
 *    will be used.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      encoding:     The encoding of the initial string.      (char *)
 *      str:          A pointer to the string to be converted. (const char *)
 *   >
 *  NOTES:
 *   <English:
 *    Note that the "encoding" paramemter can be NULL.
 *    In this case, the system encoding will be used.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will always create and return a new string buffer,
 *    where the converted string will be placed. The user must call CDM_Free
 *    in order to free the returned value, in case it is not needed any more.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    Tcl_ExternalToUtf, Tcl_ExternalToUtfDString
 *   >:&:
 */
CDM_EXPORT
char *CDM_ExternalToUtf(const Tcl_Encoding encoding, const char *str) {
  Tcl_DString Text;
  char *result;

  Tcl_DStringInit(&Text);
  result = CDM_Strdup((char *) Tcl_ExternalToUtfDString(encoding, str,
                                                        -1, &Text));
  Tcl_DStringFree(&Text);
  return result;
} /* CDM_ExternalToUtf */

/*
 * ---- CDM_UtfToExternal:
 *  DESCRIPTION:
 *   <English:
 *    This function will convert a  UTF-8 string (referenced by "str") from
 *    UTF-8 to the specified encoding (provided by the "encoding" parameter).
 *
 *    This function will always create and return a new string buffer,
 *    where the converted string will be placed. The user must call CDM_Free
 *    in order to free the returned value, in case it is not needed any more.
 *
 *    The "encoding" parameter can be NULL. In this case, the system encoding
 *    will be used.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     encoding:     The encoding of the initial string.      (const char *)
 *     str:          A pointer to the string to be converted. (const char *)
 *   >
 *  NOTES:
 *   <English:
 *    Note that the "encoding" paramemter can be NULL.
 *    In this case, the system encoding will be used.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will always create and return a new string buffer,
 *    where the converted string will be placed. The user must call CDM_Free
 *    in order to free the returned value, in case it is not needed any more.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    Tcl_UtfToExternal, Tcl_UtfToExternalDString
 *   >:&:
 */
CDM_EXPORT
char *CDM_UtfToExternal(const char *encoding, const char *str) {
  Tcl_DString Text;
  Tcl_Encoding enc;
  char *result;
  CDM_THREAD_SPECIFIC_DATA;

  enc = Tcl_GetEncoding(CDM_Interp, encoding);
  if (encoding == NULL) return NULL; // Tcl_GetEncoding has left an error...

  Tcl_DStringInit(&Text);
  result = CDM_Strdup((char *) Tcl_UtfToExternalDString(enc, str, -1, &Text));
  Tcl_DStringFree(&Text);
  Tcl_FreeEncoding(enc);
  return result;
} /* CDM_UtfToExternal */

/*
 * ---- CDM_UtfToExternal:
 *  DESCRIPTION:
 *   <English:
 *    This function will convert a UTF-8 string (referenced by "str") from
 *    UTF-8 to the specified encoding (provided by the "encoding" parameter).
 *    A valid encoding can be obtained through the Tcl function
 *    [[A HREF="../TclTk/TclLib/Encoding.htm"]]Tcl_GetEncoding[[/A]].
 *
 *    This function will always create and return a new string buffer,
 *    where the converted string will be placed. The user must call CDM_Free
 *    in order to free the returned value, in case it is not needed any more.
 *
 *    The "encoding" parameter can be NULL. In this case, the system encoding
 *    will be used.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     encoding:     The encoding of the initial string.      (Tcl_Encoding)
 *     str:          A pointer to the string to be converted. (char *)
 *   >
 *  NOTES:
 *   <English:
 *    Note that the "encoding" paramemter can be NULL.
 *    In this case, the system encoding will be used.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will always create and return a new string buffer,
 *    where the converted string will be placed. The user must call CDM_Free
 *    in order to free the returned value, in case it is not needed any more.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    Tcl_UtfToExternal, Tcl_UtfToExternalDString,
 *    [[A HREF="../TclTk/TclLib/Encoding.htm"]]Tcl_GetEncoding[[/A]]
 *   >:&:
 */
CDM_EXPORT
char *CDM_UtfToExternal(const Tcl_Encoding encoding, const char *str) {
  Tcl_DString Text;
  char *result;
  CDM_THREAD_SPECIFIC_DATA;

  Tcl_DStringInit(&Text);
  result = CDM_Strdup((char *) Tcl_UtfToExternalDString(encoding, str,
                                                        -1, &Text));
  Tcl_DStringFree(&Text);
  return result;
} /* CDM_UtfToExternal */

/*
 * ---- CDM_Free:
 *  DESCRIPTION:
 *   <English:
 *    This function frees a UTF-8 string allocated with a call to either
 *    CDM_Strdup, CDM_Malloc or Tcl_Alloc.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      string:     The string to be freed            (char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to Tcl_Free.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function always returns TCL_OK.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Strdup, CDM_Malloc, Tcl_Alloc
 *   >:&:
 */
CDM_EXPORT
int CDM_Free(char *string) {
  if (string == NULL) return TCL_ERROR;
  Tcl_Free(string);
  return TCL_OK;
} /* CDM_Free */

/*
 * ---- CDM_Free:
 *  DESCRIPTION:
 *   <English:
 *    This function destroys and frees the occupied memory of a CDM or Tcl
 *    object.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      obj:     The object to be freed              (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to Tcl_DecrRefCount.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function always returns TCL_OK.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_GetAttribute, CDM_GetAttributes, CDM_AnnotationsAt,
 *    CDM_AnnotationsContaining, CDM_AnnotationsInRange,
 *    CDM_AnnotationsMatchingRange, CDM_GetAnnotations,
 *    CDM_GetAnnotationAttributeNames, CDM_GetAnnotationTypes,
 *    CDM_GetAttributes, CDM_GetByteSequence, CDM_GetRawData,
 *    CDM_NextAnnotations, CDM_SelectAnnotations,
 *    CDM_SelectAnnotationsSorted, CDM_CreateAnnotation,
 *    CDM_CreateAnnotationSet, CDM_MergeAnnotations, CDM_CreateAttribute,
 *    CDM_CreateAttributeSet, CDM_CreateAttributeValue, CDM_CreateSpan,
 *    CDM_CreateSpanSet, Tcl_DecrRefCount
 *   >:&:
 */
CDM_EXPORT
int CDM_Free(Tcl_Obj *obj) {
  if (obj == NULL) return TCL_ERROR;
  Tcl_DecrRefCount(obj);
  return TCL_OK;
} /* CDM_Free */

/*
 * ---- CDM_LinkVar:
 *  DESCRIPTION:
 *   <English:
 *    This function links strings between C++ and Tcl.
 *    CDM_LinkVar uses variable traces to keep the Tcl variable named by
 *    "tcl_var" in sync with the C variable at the address given by "c_var".
 *    Whenever the Tcl variable is red the value of the C variable will be
 *    returned, and whenever the Tcl variable is written the C variable will be
 *    updated to have the same value.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      interp:    A pointer to a Tcl interpreter.    (Tcl_Interp *interp)
 *      tcl_var:   The name of the Tcl variable                  (const char *)
 *      c_var:     The address of the c variable to be linked    (char **)
 *   >
 *  NOTES:
 *   <English:
 *    Note that the memory referenced by "c_var" must have been allocated with
 *    a call to either CDM_Malloc or Tcl_Alloc. "c_var" must never point to a
 *    staticly allocated buffer. "c_var" is allowed to have the value NULL as
 *    initial value. In this case it will be updated to point to a dynamicly
 *    allocated empty string.
 *
 *    If "interp" has the value NULL, the Tcl interpreter referenced by
 *    "CDM_Interp" will be used.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    CDM_LinkVar normally returns TCL_OK; if an error occurs while setting up
 *    the link (e.g. because "tcl_var" is the name of array) then TCL_ERROR is
 *    returned and an error message is placed in the provided Tcl interpreter.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Interp, Tcl_LinkVar
 *   >:&:
 */
CDM_EXPORT
int CDM_LinkVar(Tcl_Interp *interp, const char *tcl_var, char **c_var) {
  char *value;

  if (interp == NULL) {
    CDM_THREAD_SPECIFIC_DATA;
    interp = CDM_Interp;
  }

  /*
   * Get the variable's current value...
   */
  value = (char *) Tcl_GetVar(interp, tcl_var, TCL_GLOBAL_ONLY);
  if (value == NULL) {
    value = strdup("");
  } else {
    value = CDM_Strdup(value);
  }
  *c_var = value;

  return Tcl_LinkVar(interp, tcl_var, (char *) c_var, TCL_LINK_STRING);
} /* CDM_LinkVar */

/*
 * ---- CDM_UnlinkVar:
 *  DESCRIPTION:
 *   <English:
 *    This function removes the link previously set up for the variable given
 *    by "tcl_var_name". If there does not exist a link for "tcl_var_name"
 *    then the procedure has no effect.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      interp:        A pointer to a Tcl interpreter.    (Tcl_Interp *interp)
 *      tcl_var_name:  The name of the Tcl variable.      (const char *)
 *   >
 *  NOTES:
 *   <English:
 *    If "interp" has the value NULL, the Tcl interpreter referenced by
 *    "CDM_Interp" will be used.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    None.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_LinkVar, CDM_Interp, Tcl_LinkVar, Tcl_UnlinkVar
 *   >:&:
 */
CDM_EXPORT
void CDM_UnlinkVar(Tcl_Interp *interp, const char *tcl_var_name) {

  if (interp == NULL) {
    CDM_THREAD_SPECIFIC_DATA;
    interp = CDM_Interp;
  }
  Tcl_UnlinkVar(interp, tcl_var_name);
} /* CDM_UnlinkVar */

/***********************************************************
 **  Error Reporting
 ***********************************************************/

/*
 * ---- CDM_NotReady:
 *  DESCRIPTION:
 *   <English:
 *    This function will print the message
 *    "******** This feature is not implemented yet!\n" to the "stderr" channel.
 *   >
 *  ARGUMENTS:
 *   <English:
 *   >
 *  NOTES:
 *   <English:
 *    Under Microsoft Windows, where "stderr" channel is not available, this
 *    function does nothing.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    None.
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
void CDM_NotReady(void) {
  fprintf(stderr,"******** This feature is not implemented yet!\n");
} /* CDM_NotReady */

/*
 * ---- CDM_ReportError:
 *  DESCRIPTION:
 *   <English:
 *    This function will set the given string as an error to the
 *    current Tcl interpreter (stored in the variable CDM_Interp).
 *    If the variable "CDM_DisplayErrorMessage" is true (has a value != 0),
 *    this function will also try to display a an error dialog, reporting also
 *    the error in a graphical way.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      error:     The error message             (char *)
 *   >
 *  NOTES:
 *   <English:
 *    Note that this function will also display an error dialog, if
 *    "CDM_DisplayErrorMessage" is true.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function always returns TCL_ERROR.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    Tcl_SetResult
 *   >:&:
 */
/*CDM_EXPORT
int CDM_ReportError(char *error) {
  Tcl_CmdInfo infoPtr;
  Tcl_Obj *objv[9];
  int i;
  // Can we display an Error Dialog???
  if (CDM_DisplayErrorMessage &&
      Tcl_GetCommandInfo(CDM_Interp, "tk_messageBox", &infoPtr) != 0) {
    objv[0] = Tcl_NewStringObj("tk_messageBox", -1);
    objv[1] = Tcl_NewStringObj("-icon", -1);
    objv[2] = Tcl_NewStringObj("error", -1);
    objv[3] = Tcl_NewStringObj("-message", -1);
    objv[4] = Tcl_NewStringObj( error, -1);
    objv[5] = Tcl_NewStringObj("-type", -1);
    objv[6] = Tcl_NewStringObj("ok", -1);
    objv[7] = Tcl_NewStringObj("-title", -1);
    objv[8] = Tcl_NewStringObj("CDM Error...", -1);
    for (i=0; i<9; i++) Tcl_IncrRefCount(objv[i]);
    Tcl_EvalObjv(CDM_Interp, 9, objv, TCL_EVAL_GLOBAL);
    for (i=0; i<9; i++) Tcl_DecrRefCount(objv[i]);
  }
  Tcl_SetResult(CDM_Interp, error, TCL_VOLATILE);
  return TCL_ERROR;
}*/ /* CDM_ReportError */

/*
 * ---- CDM_ReportError:
 *  DESCRIPTION:
 *   <English:
 *    This function will set the given string as an error to the
 *    current Tcl interpreter (stored in the variable CDM_Interp).
 *    If the variable "CDM_DisplayErrorMessage" is true (has a value != 0),
 *    this function will also try to display a an error dialog, reporting also
 *    the error in a graphical way.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      interp:     The interpreter to report the error (Tcl_Interp *)
 *      format:     The format string             (char *)
 *        arg1:     Other arguments               (as described by format)
 *   >
 *  NOTES:
 *   <English:
 *    Note that this function will also display an error dialog, if
 *    "CDM_DisplayErrorMessage" is true.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function always returns TCL_ERROR.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    Tcl_SetResult
 *   >:&:
 */
CDM_EXPORT
int CDM_ReportError(Tcl_Interp* interp, const char* arg1, ...) {
  Tcl_CmdInfo infoPtr;
  Tcl_Obj *objv[9];
  const char *fmt;
  va_list args;
  int i;
  CDM_THREAD_SPECIFIC_DATA;

  if (interp == NULL) return TCL_ERROR;
  fmt = TCL_VARARGS_START(const char *, arg1, args);
  if (fmt != CDM_Temp) {
    (void) vsnprintf(CDM_Temp, MAX_TEMP_STRING, fmt, args);
  }
  va_end(args);

  // Can we display an Error Dialog???
  if (CDM_DisplayErrorMessage &&
      Tcl_GetCommandInfo(interp, "tk_messageBox", &infoPtr) != 0) {
    objv[0] = Tcl_NewStringObj("tk_messageBox", -1);
    objv[1] = Tcl_NewStringObj("-icon", -1);
    objv[2] = Tcl_NewStringObj("error", -1);
    objv[3] = Tcl_NewStringObj("-message", -1);
    objv[4] = Tcl_NewStringObj( CDM_Temp, -1);
    objv[5] = Tcl_NewStringObj("-type", -1);
    objv[6] = Tcl_NewStringObj("ok", -1);
    objv[7] = Tcl_NewStringObj("-title", -1);
    objv[8] = Tcl_NewStringObj("CDM Error...", -1);
    for (i=0; i<9; i++) Tcl_IncrRefCount(objv[i]);
    Tcl_EvalObjv(interp, 9, objv, TCL_EVAL_GLOBAL);
    for (i=0; i<9; i++) Tcl_DecrRefCount(objv[i]);
  }
  Tcl_SetResult(interp, CDM_Temp, TCL_VOLATILE);
  return TCL_ERROR;
} /* CDM_ReportError */

CDM_EXPORT
int CDM_ReportError(Tcl_Interp* interp, char * arg1, ...) {
  Tcl_CmdInfo infoPtr;
  Tcl_Obj *objv[9];
  char *fmt;
  va_list args;
  int i;
  CDM_THREAD_SPECIFIC_DATA;

  if (interp == NULL) return TCL_ERROR;
  fmt = TCL_VARARGS_START(char *, arg1, args);
  if (fmt != CDM_Temp) {
    (void) vsnprintf(CDM_Temp, MAX_TEMP_STRING, fmt, args);
  }
  va_end(args);

  // Can we display an Error Dialog???
  if (CDM_DisplayErrorMessage &&
      Tcl_GetCommandInfo(interp, "tk_messageBox", &infoPtr) != 0) {
    objv[0] = Tcl_NewStringObj("tk_messageBox", -1);
    objv[1] = Tcl_NewStringObj("-icon", -1);
    objv[2] = Tcl_NewStringObj("error", -1);
    objv[3] = Tcl_NewStringObj("-message", -1);
    objv[4] = Tcl_NewStringObj( CDM_Temp, -1);
    objv[5] = Tcl_NewStringObj("-type", -1);
    objv[6] = Tcl_NewStringObj("ok", -1);
    objv[7] = Tcl_NewStringObj("-title", -1);
    objv[8] = Tcl_NewStringObj("CDM Error...", -1);
    for (i=0; i<9; i++) Tcl_IncrRefCount(objv[i]);
    Tcl_EvalObjv(interp, 9, objv, TCL_EVAL_GLOBAL);
    for (i=0; i<9; i++) Tcl_DecrRefCount(objv[i]);
  }
  Tcl_SetResult(interp, CDM_Temp, TCL_VOLATILE);
  return TCL_ERROR;
} /* CDM_ReportError */

/*
 * ---- CDM_ReportError:
 *  DESCRIPTION:
 *   <English:
 *    This function will set the given string as an error to the
 *    current Tcl interpreter (stored in the variable CDM_Interp).
 *    If the variable "CDM_DisplayErrorMessage" is true (has a value != 0),
 *    this function will also try to display a an error dialog, reporting also
 *    the error in a graphical way.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      format:     The format string             (char *)
 *        arg1:     Other arguments               (as described by format)
 *   >
 *  NOTES:
 *   <English:
 *    Note that this function will also display an error dialog, if
 *    "CDM_DisplayErrorMessage" is true.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function always returns TCL_ERROR.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    Tcl_SetResult
 *   >:&:
 */
CDM_EXPORT
int CDM_ReportError TCL_VARARGS_DEF(const char *, arg1) {
  Tcl_CmdInfo infoPtr;
  Tcl_Obj *objv[9];
  const char *fmt;
  va_list args;
  int i;
  CDM_THREAD_SPECIFIC_DATA;

  if (CDM_Interp == NULL) return TCL_ERROR;
  fmt = TCL_VARARGS_START(const char *, arg1, args);
  if (fmt != CDM_Temp) {
    (void) vsnprintf(CDM_Temp, MAX_TEMP_STRING, fmt, args);
  }
  va_end(args);

  // Can we display an Error Dialog???
  if (CDM_DisplayErrorMessage &&
      Tcl_GetCommandInfo(CDM_Interp, "tk_messageBox", &infoPtr) != 0) {
    objv[0] = Tcl_NewStringObj("tk_messageBox", -1);
    objv[1] = Tcl_NewStringObj("-icon", -1);
    objv[2] = Tcl_NewStringObj("error", -1);
    objv[3] = Tcl_NewStringObj("-message", -1);
    objv[4] = Tcl_NewStringObj( CDM_Temp, -1);
    objv[5] = Tcl_NewStringObj("-type", -1);
    objv[6] = Tcl_NewStringObj("ok", -1);
    objv[7] = Tcl_NewStringObj("-title", -1);
    objv[8] = Tcl_NewStringObj("CDM Error...", -1);
    for (i=0; i<9; i++) Tcl_IncrRefCount(objv[i]);
    Tcl_EvalObjv(CDM_Interp, 9, objv, TCL_EVAL_GLOBAL);
    for (i=0; i<9; i++) Tcl_DecrRefCount(objv[i]);
  }
  Tcl_SetResult(CDM_Interp, CDM_Temp, TCL_VOLATILE);
  return TCL_ERROR;
} /* CDM_ReportError */

CDM_EXPORT
int CDM_ReportError TCL_VARARGS_DEF(char *, arg1) {
  Tcl_CmdInfo infoPtr;
  Tcl_Obj *objv[9];
  char *fmt;
  va_list args;
  int i;
  CDM_THREAD_SPECIFIC_DATA;

  if (CDM_Interp == NULL) return TCL_ERROR;
  fmt = TCL_VARARGS_START(char *, arg1, args);
  if (fmt != CDM_Temp) {
    (void) vsnprintf(CDM_Temp, MAX_TEMP_STRING, fmt, args);
  }
  va_end(args);

  // Can we display an Error Dialog???
  if (CDM_DisplayErrorMessage &&
      Tcl_GetCommandInfo(CDM_Interp, "tk_messageBox", &infoPtr) != 0) {
    objv[0] = Tcl_NewStringObj("tk_messageBox", -1);
    objv[1] = Tcl_NewStringObj("-icon", -1);
    objv[2] = Tcl_NewStringObj("error", -1);
    objv[3] = Tcl_NewStringObj("-message", -1);
    objv[4] = Tcl_NewStringObj( CDM_Temp, -1);
    objv[5] = Tcl_NewStringObj("-type", -1);
    objv[6] = Tcl_NewStringObj("ok", -1);
    objv[7] = Tcl_NewStringObj("-title", -1);
    objv[8] = Tcl_NewStringObj("CDM Error...", -1);
    for (i=0; i<9; i++) Tcl_IncrRefCount(objv[i]);
    Tcl_EvalObjv(CDM_Interp, 9, objv, TCL_EVAL_GLOBAL);
    for (i=0; i<9; i++) Tcl_DecrRefCount(objv[i]);
  }
  Tcl_SetResult(CDM_Interp, CDM_Temp, TCL_VOLATILE);
  return TCL_ERROR;
} /* CDM_ReportError */

/*
 * ---- CDM_AppendError:
 *  DESCRIPTION:
 *   <English:
 *    This function will append the given string as an error to the value of
 *    current Tcl interpreter (stored in the variable CDM_Interp).
 *   >
 *  ARGUMENTS:
 *   <English:
 *      error:     The error message             (char *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function always returns TCL_ERROR.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    Tcl_AppendResult
 *   >:&:
 */
CDM_EXPORT
int CDM_AppendError TCL_VARARGS_DEF(const char *, arg1) {
  const char *fmt;
  va_list args;
  CDM_THREAD_SPECIFIC_DATA;

  if (CDM_Interp != NULL) {
    fmt = TCL_VARARGS_START(const char *, arg1, args);
    (void) vsnprintf(CDM_Temp, MAX_TEMP_STRING, fmt, args);
    va_end(args);
    Tcl_AppendResult(CDM_Interp, CDM_Temp, (char *) NULL);
  }
  return TCL_ERROR;
} /* CDM_AppendError */

CDM_EXPORT
int CDM_AppendError TCL_VARARGS_DEF(char *, arg1) {
  char *fmt;
  va_list args;
  CDM_THREAD_SPECIFIC_DATA;

  if (CDM_Interp != NULL) {
    fmt = TCL_VARARGS_START(char *, arg1, args);
    (void) vsnprintf(CDM_Temp, MAX_TEMP_STRING, fmt, args);
    va_end(args);
    Tcl_AppendResult(CDM_Interp, CDM_Temp, (char *) NULL);
  }
  return TCL_ERROR;
} /* CDM_AppendError */

/*
 * ---- CDM_Panic:
 *  DESCRIPTION:
 *   <English:
 *    Reportes given string in stderr and aborts application...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      error:     The error message             (char *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
void CDM_Panic(CDM_CONST char *error) {
  CDM_THREAD_SPECIFIC_DATA;
  fprintf(stderr, "%s\n", error);
  CDM_DisplayErrorMessage = 1;
  CDM_ReportError(error);
  abort();
} /* CDM_Panic */

/*
 * ---- CDM_MutexLock:
 *  DESCRIPTION:
 *   <English:
 *    Locks the CDM global Mutex object...
 *   >
 *  ARGUMENTS:
 *   <English:
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
void CDM_MutexLock(void) {
  Tcl_MutexLock(&ellogonCDMMutex);
} /* CDM_MutexLock */

/*
 * ---- CDM_MutexUnlock:
 *  DESCRIPTION:
 *   <English:
 *    Unlocks the CDM global Mutex object...
 *   >
 *  ARGUMENTS:
 *   <English:
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >:&:
 */
CDM_EXPORT
void CDM_MutexUnlock(void) {
  Tcl_MutexUnlock(&ellogonCDMMutex);
} /* CDM_MutexUnlock */

/***********************************************************
 **  Debugging Related Functions...
 ***********************************************************/
/*
 * ---- CDM_Debug:
 *  DESCRIPTION:
 *   <English:
 *     This function can be used in order to display debugging information
 *     under all supported platforms. "CDM_Debug" works in the same way and
 *     accepts the same arguments as the standart C library function "printf".
 *     The only difference is that the message is not displayed to the
 *     "stdout" or "stderr" files of the program. The output is instead
 *     displayed using the "puts" Tcl command, thus offering the ability to
 *     Tcl scripts to redirect the output by renaming the "puts" command, or
 *     see the displayed information under windows in a console, as both
 *     "stdout" and "stderr" files are not available under windows.
 *
 *     The actual Tcl file that the message will be redirected is controled by
 *     the value of the C or Tcl variable "CDM_DebugStdout": if this variable
 *     contains a true value (!= 0), the message will be displayed to the
 *     "stdout" Tcl channel. Atherwise, the "stderr" Tcl channel will be used.
 *
 *     Although the user can call this function directly, it is recommented
 *     that the C/C++ macro "CDM_DEBUG" should be used instead. The greater
 *     advantage of using the macro is that the message will be displayed only
 *     if the code is compiled with debugging enabled (i.e. with the symbol
 *     "DEBUG" defined). If the symbol "DEBUG" is not enabled, calls to
 *     "CDM_DEBUG" will have no effect. The "CDM_DEBUG" macro can be used in all
 *     the places "CDM_Debug" can be used, as it accepts the same arguments.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      format: A format string, describing the arguments in the same way as the first argument of the C library function "printf" (char *)
 *      ...:    A variable set of arguments. These arguments should correspond to the arguments declared through the format string (various)
 *   >
 *  NOTES:
 *   <English:
 *     Note that "CDM_Debug" places a limit to the length of the displayed
 *     message, as its length cannot exceed the 4047 single-byte characters.
 *     If the message (after all the substitutions described by the format
 *     argument have been made) exceeds the length of 4047 characters, it will
 *     be truncuated to this length.
 *
 *     Also note that it is preferable to use the macro "CDM_DEBUG", instead
 *     of calling "CDM_Debug" directly. Such use will enable the
 *     activation/deactivaion of debugging information during compilation,
 *     according whether the symbol "DEBUG" is defined or not.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function does not return anything, even if an error occurs.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_DebugStdout, puts, printf, CDM.h
 *   >:&:
 */
CDM_EXPORT
void CDM_Debug(const char *format, ... ) {
  va_list arguments;
  char string[4048], *ptr;
  CDM_THREAD_SPECIFIC_DATA;

  va_start(arguments, format);
  vsnprintf(string, 4046, format, arguments);
  ptr = CDM_ExternalToUtf((const char *) NULL, string);
  if (CDM_DebugStdout) {
    CDM_Eval(CDM_Interp, "puts", "-nonewline",  ptr, (char *) NULL);
  } else {
    CDM_Eval(CDM_Interp, "puts", "-nonewline", "stderr", ptr, (char *) NULL);
  }
  CDM_Free(ptr);
  va_end(arguments);
} /* CDM_Debug */

#ifdef WINDOWS
CDM_EXPORT
void inline CDM_DebugNull(const char* fmt, ... ) {
} /* CDM_DebugNull */
#endif /* WINDOWS */

/*
 * ---- CDM_Log:
 *  DESCRIPTION:
 *   <English:
 *     This function can be used in order to display debugging information
 *     under all supported platforms. "CDM_Log" works in the same way and
 *     accepts the same arguments as the standart C library function "printf".
 *     The only difference is that the message is not displayed to the
 *     "stdout" or "stderr" files of the program, but in a log file.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      format: A format string, describing the arguments in the same way as the first argument of the C library function "printf" (char *)
 *      ...:    A variable set of arguments. These arguments should correspond to the arguments declared through the format string (various)
 *   >
 *  NOTES:
 *   <English:
 *     Note that "CDM_Log" places a limit to the length of the displayed
 *     message, as its length cannot exceed MAX_TEMP_STRING single-byte
 *     characters.
 *     If the message (after all the substitutions described by the format
 *     argument have been made) exceeds the length of MAX_TEMP_STRING
 *     characters (bytes), it will
 *     be truncuated to this length.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function does not return anything, even if an error occurs.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_LogStdout, puts, printf, CDM.h
 *   >:&:
 */
CDM_EXPORT
void CDM_Log(const char *format, ... ) {
  static FILE *logFile = NULL;
  va_list arguments;
  CDM_THREAD_SPECIFIC_DATA;

  if (logFile == NULL) {
#ifdef WINDOWS
    if (fopen_s(&logFile, "CDM_Log.txt", "w") != 0) return;
#else
    logFile = fopen("CDM_Log.txt", "w");
#endif
    if (logFile == NULL) return;
  }
  va_start(arguments, format);
  if (format != CDM_Temp) {
    vsnprintf(CDM_Temp, MAX_TEMP_STRING, format, arguments);
    fprintf(logFile, "%s", CDM_Temp);
  } else {
    fprintf(logFile, "%s", format);
  }
  va_end(arguments);
  fflush(logFile);
} /* CDM_Log */

/***********************************************************
 **  Pointer Object Related Functions...
 ***********************************************************/

int CDM_DeletePointerFromList(Tcl_Obj *List, void *ptr) {
  int i, index, result, listc;
  Tcl_Obj **listv;
  void *pointer;
  CDM_THREAD_SPECIFIC_DATA;

  result = Tcl_ListObjGetElements(CDM_Interp, List, &listc, &listv);
  if (result != TCL_OK) return TCL_ERROR;
  index = -1;
  for (i = 0; i < listc; i++) {
    pointer = listv[i]->internalRep.otherValuePtr;
    if (pointer == ptr) {index = i; break;}
  }
  if (index > 0) {
    return Tcl_ListObjReplace(CDM_Interp, List, index, 1, 0, NULL);
  }
  return TCL_ERROR;
} /* CDM_DeletePointerFromList */

Tcl_Obj *CDM_NewPointerObj(void *ptr) {
  Tcl_Obj *objPtr;

  objPtr = Tcl_NewObj();
  objPtr->bytes = NULL;

  objPtr->internalRep.otherValuePtr = ptr;
  objPtr->typePtr = &CDM_TclPointerObjType;
  return objPtr;
} /* CDM_NewPointerObj */

void *CDM_GetPointerFromObj(Tcl_Obj *Obj) {
  if (Obj->typePtr != &CDM_TclPointerObjType) return NULL;
  return Obj->internalRep.otherValuePtr;
} /* CDM_GetPointerFromObj */

/*
 * Pointer Object Type Internal Functions...
 */
void CDM_PointerObj_UpdateStringProc(Tcl_Obj *obj) {
  CDM_THREAD_SPECIFIC_DATA;
  snprintf(CDM_Temp, MAX_TEMP_STRING, "%p", obj->internalRep.otherValuePtr);
  obj->bytes = strdup(CDM_Temp);
  obj->length = strlen(obj->bytes);
} /* CDM_PointerObj_UpdateStringProc */
int CDM_PointerObj_SetPointerFromAnyProc(Tcl_Interp *interp, Tcl_Obj *obj) {
  const Tcl_ObjType *oldTypePtr = obj->typePtr;
  char *string;
  void *NewValue;
  int length;

  /*
   * Get the string representation. Make it up-to-date if necessary.
   */
  string = Tcl_GetStringFromObj(obj, &length);
  if (sscanf(string, "%p", &NewValue) != 1) {
    if (interp != NULL) {
      /*
       * Must copy string before resetting the result in case a caller
       * is trying to convert the interpreter's result to a boolean.
       */

      char buf[100];
      sprintf(buf, "expected pointer value but got \"%.50s\"", string);
      Tcl_ResetResult(interp);
      Tcl_AppendToObj(Tcl_GetObjResult(interp), buf, -1);
    }
    return TCL_ERROR;
  }

  /*
   * Free the old internalRep before setting the new one. We do this as
   * late as possible to allow the conversion code, in particular
   * Tcl_GetStringFromObj, to use that old internalRep.
   */
   if ((oldTypePtr != NULL) &&        (oldTypePtr->freeIntRepProc != NULL)) {
      oldTypePtr->freeIntRepProc(obj);
   }
   obj->internalRep.otherValuePtr = NewValue;
   obj->typePtr = &CDM_TclPointerObjType;
   return TCL_OK;
} /* CDM_PointerObj_SetFromAnyProc */

/*
 * ---- CDM_MajorVersion:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the major version of CDM. For example,
 *     if the version of CDM is 1.3, it will return 1.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None: (void)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_MajorVersion (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     An integer, specifying the major version of the CDM library.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MinorVersion, CDM_Version, CDM_PatchLevel, CDM_ReleaseDate,
 *    CDM_ExtendedReleaseDate, tip_MajorVersion
 *   >:&:
 */
CDM_EXPORT
int CDM_MajorVersion(void) {
  return CDM_MAJOR_VERSION;
} /* CDM_MajorVersion */

/*
 * ---- CDM_MinorVersion:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the major version of CDM. For example,
 *     if the version of CDM is 1.3, it will return 3.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None: (void)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_MinorVersion (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     An integer, specifying the minor version of the CDM library.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MajorVersion, CDM_Version, CDM_PatchLevel, CDM_ReleaseDate,
 *    CDM_ExtendedReleaseDate, tip_MinorVersion
 *   >:&:
 */
CDM_EXPORT
int CDM_MinorVersion(void) {
  return CDM_MINOR_VERSION;
} /* CDM_MinorVersion */

/*
 * ---- CDM_Version:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the version of CDM as a string.
 *     For example, if the version of CDM is 1.3, it will return "1.3".
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None: (void)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_Version (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     A string, specifying the version of the CDM library.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MajorVersion, CDM_MinorVersion, CDM_PatchLevel, CDM_ReleaseDate,
 *    CDM_ExtendedReleaseDate, tip_Version
 *   >:&:
 */
CDM_EXPORT
const char *CDM_Version(void) {
  return CDM_VERSION;
} /* CDM_Version */

/*
 * ---- CDM_PatchLevel:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the patch level of the CDM library as a string.
 *     For example, if the version of CDM is 1.3a1, it will return "1.3a1".
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None: (void)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_PatchLevel (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     A string, specifying the patch level of the CDM library.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MajorVersion, CDM_MinorVersion, CDM_Version, CDM_ReleaseDate,
 *    CDM_ExtendedReleaseDate, tip_PatchLevel
 *   >:&:
 */
CDM_EXPORT
const char *CDM_PatchLevel(void) {
  return CDM_PATCH_LEVEL;
} /* CDM_PatchLevel */

/*
 * ---- CDM_ReleaseDate:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the date of the CDM library as a string.
 *     This date corresponds to the release date. The format of the return
 *     string is dd/mm/yyyy.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None: (void)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_ReleaseDate (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     A string, specifying the release date of the CDM library.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MajorVersion, CDM_MinorVersion, CDM_Version, CDM_PatchLevel,
 *    CDM_ExtendedReleaseDate, tip_ReleaseDate
 *   >:&:
 */
CDM_EXPORT
const char *CDM_ReleaseDate(void) {
  return CDM_RELEASE_DATE;
} /* CDM_ReleaseDate */

/*
 * ---- CDM_ExtendedReleaseDate:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the date of the CDM library as a string.
 *     This date corresponds to the release date. The format of the return
 *     string is like in the following example:
 *     "Tue Sep  4 19:08:24 EEST 2001"...
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None: (void)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_ExtendedReleaseDate (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     A string, specifying the release date of the CDM library in an extended
 *     format.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_MajorVersion, CDM_MinorVersion, CDM_Version, CDM_PatchLevel,
 *    CDM_ReleaseDate, tip_ExtendedReleaseDate
 *   >:&:
 */
CDM_EXPORT
const char *CDM_ExtendedReleaseDate(void) {
  return CDM_RELEASE_DATE_STR;
} /* CDM_ExtendedReleaseDate */

extern "C"
/*
 * ---- CDM_Initialise:
 *  DESCRIPTION:
 *   <English:
 *    This function must be called when CDM is embeded inside a C/C++
 *    application, where a Tcl interpreter is not available. This function
 *    will initialise all needed by the CDM internal structures. This function
 *    must be called only once within each application (although calling this
 *    function more than once is safe, as subsequent invocations are ignored).
 *   >
 *  ARGUMENTS:
 *   <English:
 *    appName: A string containing the name of the application. It can also be the empty string "" (char *)
 *   >
 *  NOTES:
 *   <English:
 *    Please note that for embedding CDM under any application, Tcl must be
 *    installed, as CDM uses functions from the Tcl library.
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    If the CDM could not be initialised, TCL_ERROR will be returned. It is
 *    safe to call any CDM function if TCL_ERROR is returned. If CDM is
 *    succesfully initialised, TCL_OK will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Finalise
 *   >:&:
 */
CDM_EXPORT
int CDM_Initialise(const char *appName) {
  CDM_THREAD_SPECIFIC_DATA;
  /* If Cdm_Init has been called, there is no point in initialising again... */
  if (CDM_LibraryInitialised) return TCL_OK;
  if (CDMEmbedInterp != NULL) return TCL_ERROR;
  Tcl_Interp *interp;
  Tcl_FindExecutable(appName);
  interp = Tcl_CreateInterp();
  Tcl_InitMemory(interp);
  if (Tcl_Init(interp) == TCL_ERROR) {
    /* We don't care if Tcl has initialised correctly... */
  }
  /* Neutralise "exit" Tcl command :-) The main application should decide when
   * to terminate the process... */
  Tcl_Eval(interp, "proc exit {args} {}");
  Tcl_SetVar(interp, "argc", "0", TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp, "argv0", "", TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp, "tcl_interactive", "0", TCL_GLOBAL_ONLY);
  CDMEmbedInterp = interp;
  return Cdm_Init(interp);
} /* CDM_Initialise */

extern "C"
/*
 * ---- CDM_Finalise:
 *  DESCRIPTION:
 *   <English:
 *    This function must be called when CDM is embeded inside a C/C++
 *    application, and CDM_Initialise has been used to initialise CDM.
 *    This function will free any resources allocated by CDM_Initialise and
 *    will terminate CDM. It is not safe to call any CDM function after
 *    CDM_Finalise has been called. Of course, CDM can be initialised again,
 *    by calling CDM_Initialise.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    None: (void)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    None.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_Initialise
 *   >:&:
 */
CDM_EXPORT
void CDM_Finalise(void) {
  CDM_THREAD_SPECIFIC_DATA;
  if (CDMEmbedInterp != NULL) {
    Tcl_DeleteInterp(CDMEmbedInterp);
    CDMEmbedInterp = NULL;
  }
} /* CDM_Finalise */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    | E N D    O F    U T I L I T I E S     S E C T I O N S |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 * OCDM Wrapper Functions:
 * The following functions act as a wrapper to tcl functions...
 */
CDM_EXPORT Tcl_Obj *CDMTCL_GetObjResult(Tcl_Interp *interp) {
  return Tcl_GetObjResult(interp);
}
CDM_EXPORT CONST char *CDMTCL_GetString(Tcl_Obj *obj) {
  return Tcl_GetString(obj);
}
CDM_EXPORT CONST char *CDMTCL_GetStringFromObj(Tcl_Obj *obj, int *len) {
  return Tcl_GetStringFromObj(obj, len);
}
CDM_EXPORT Tcl_Obj *CDMTCL_NewStringObj(CONST char *str, int len) {
  return Tcl_NewStringObj(str, len);
}
CDM_EXPORT Tcl_Obj *CDMTCL_NewListObj(int objc, Tcl_Obj *CONST objv[]) {
  return Tcl_NewListObj(objc, objv);
}
CDM_EXPORT
int CDMTCL_ListObjAppendElement(Tcl_Interp *interp,
                                Tcl_Obj *listPtr, Tcl_Obj *objPtr) {
  return Tcl_ListObjAppendElement(interp, listPtr, objPtr);
}
CDM_EXPORT void CDMTCL_IncrRefCount(Tcl_Obj *obj) {Tcl_IncrRefCount(obj);}
CDM_EXPORT void CDMTCL_DecrRefCount(Tcl_Obj *obj) {Tcl_DecrRefCount(obj);}
CDM_EXPORT void CDMTCL_IncrRefCountMutex(Tcl_Obj *obj) {
  CDM_MutexLock(); Tcl_IncrRefCount(obj); CDM_MutexUnlock();
}
CDM_EXPORT void CDMTCL_DecrRefCountMutex(Tcl_Obj *obj) {
  CDM_MutexLock(); Tcl_DecrRefCount(obj); CDM_MutexUnlock();
}
CDM_EXPORT CONST char *CDMTCL_GetCommandName(Tcl_Interp *interp,Tcl_Command t) {
  return Tcl_GetCommandName(interp, t);
}
CDM_EXPORT int CDMTCL_Eval(Tcl_Interp *interp, CONST char *cmd) {
  return Tcl_Eval(interp, cmd);
}
CDM_EXPORT int CDMTCL_ListObjGetElements(Tcl_Obj *obj, int *objcPtr,
                                                           Tcl_Obj ***objvPtr) {
  CDM_THREAD_SPECIFIC_DATA;
  return Tcl_ListObjGetElements(CDM_Interp, obj, objcPtr, objvPtr);
} /* CDMTCL_ListObjGetElements */
/*
 * CDM.cpp - EOF -
 */
