/*****************************************************************************
 *  Cdm_Init.cpp:                                                            *
 * --------------------                                                      *
 *  This file initialises the Tcl bindings of Collection & Document          *
 *  Manager (CDM).                                                           *
 *                                                                           *
 *  Copyright 1998-2010 by:                                                  *
 *  Georgios Petasis,                                                        *
 *  Software and Knowledge Engineering Laboratory,                           *
 *  Institute of Informatics and Telecommunications,                         *
 *  National Center for Scientific Research (N.C.S.R.) "Demokritos",         *
 *  Athens, Greece.                                                          *
 *  E-mail: petasis@iit.demokritos.gr                                        *
 *                                                                           *
 *  This file is part of the "Ellogon" Text Engineering Platform.            *
 *  For licensing information, please refer to the license conditions of     *
 *  "Ellogon" Text Engineering Platform.                                     *
 *****************************************************************************/

#include "Cdm_Init.h"
#include "CDM_License.h"
#include "CDM_ObjectTypes.h"
#include "CDM_CommandSpec.h"

#ifdef CDM_ENABLE_TK_COMMANDS
#include <tk.h>
#endif /* CDM_ENABLE_TK_COMMANDS */

#include "CDM_GlobalObjects.h"
extern Tcl_ThreadDataKey dataKey;

extern Tcl_ObjType CDM_TclPointerObjType;
extern int CDM_XMLCollectionTypeInit(Tcl_Interp *interp);
extern int CDM_GetAttributeTypeFromObject(Tcl_Obj *Type);
#ifdef CDM_ENABLE_LICENSE_MANAGER
#include "Internal/Cdm_InitDefines.h"
#endif /* CDM_ENABLE_LICENSE_MANAGER */

/*
 * Our Copyright Notice!
 */
static const char *CDM_Copyright = 
    "<HTML><H1>CDM - Ellogon Collection and Document Manager</H1>"
    "<br>Copyright &copy; 1998-2001 by "
    "<br><br>Georgios P. Petasis, <a href=\"mailto:petasis@iit.demokritos.gr\">"
    "petasis@iit.demokritos.gr</a>,"
    "<br>Software and Knowledge Engineering Laboratory,<br>Institute of "
    "Informatics and Telecomunications, <br>National Center for Scientific "
    "Research &quot;Demokritos&quot;,<br>Athens, Greece. "
    "<p><center><a href=\"http://www.iit.demokritos.gr/~petasis/Ellogon\"> "
    "http://www.iit.demokritos.gr/~petasis/Ellogon</a></center></p></HTML>";

/*
 * This function is the first function that gets called when the CDM (dynamic)
 * package gets loaded. It registers the new CDM commands with Tcl core...
 */
extern "C" DLLEXPORT int Cdm_Init(Tcl_Interp *Interp) {
  const CDM_CommandSpec *specPtr = CommandSpec;
  int i;

#if defined(USE_TCL_STUBS) || defined(CDM_USE_STUBS)
  if (Tcl_InitStubs(Interp, "8.4", 0) == NULL) {
    return TCL_ERROR;
  }
#endif /* USE_TCL_STUBS || CDM_USE_STUBS */

  CDM_THREAD_SPECIFIC_DATA;

  /*
   * Initialise our thread specific data storage...
   */
  CDM_Interp                   = Interp;
  CDM_UnLocked                 = 1;
  CDM_EnableMemoryCompression  = 1;
  CDM_AbortCurrentProcess      = 0;
  CDM_DebugStdout              = 1;
  CDM_DisplayErrorMessage      = 1;
  CDM_Pid                      = 0;
  CDM_FullGateCompatibility    = 0;
  CDM_EmulateGateGetValue      = 1;
  NumberOfDisplacedAnnotations = 0;
  AttrSetHash_Initialised      = 0;
#ifdef CDM_AYTOCLEAN_COMPRESSEDMEMORYCACHE
  AttrSetHash_NumberOfCalls    = 0;
  AttrSetHash_NumberOfClears   = 0;
#endif /* CDM_AYTOCLEAN_COMPRESSEDMEMORYCACHE */

#ifdef UNIX 
  CDM_CollectionFilename     = (char *) ".CDM_CollectionIndex.txt";
  CDM_CollectionTypeFilename = (char *) "CDM_CollectionType.txt";
#else
  CDM_CollectionFilename     = (char *) ".CDM_CollectionIndex.txt";
  CDM_CollectionTypeFilename = (char *) "CDM_CollectionType.txt";
#endif

  if (!CDM_LibraryInitialised) {

    /* CDM_GenerateLicenseKey */
    if(Tcl_CreateObjCommand(Interp, "CDM_GenerateLicenseKey",
                      (Tcl_ObjCmdProc*) CDM_LicenseKeyObjCmd,
                      (ClientData) 1, (Tcl_CmdDeleteProc *) NULL)  == NULL)
         return TCL_ERROR;
    /* CDM_ValidateLicenseKey */
    if(Tcl_CreateObjCommand(Interp, "CDM_ValidateLicenseKey",
                      (Tcl_ObjCmdProc*) CDM_LicenseKeyObjCmd,
                      (ClientData) 2, (Tcl_CmdDeleteProc *) NULL)  == NULL)
         return TCL_ERROR;
    /* CDM_GetIP */
    if(Tcl_CreateObjCommand(Interp, "CDM_GetIP",
                      (Tcl_ObjCmdProc*) CDM_GetIPObjCmd,
                      (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL)  == NULL)
         return TCL_ERROR;

#ifdef CDM_ENABLE_LICENSE_MANAGER
#include "Internal/Cdm_Init.cpp"
#endif /* CDM_ENABLE_LICENSE_MANAGER */

    /* Create the ::CDM Namespace... */
    strcpy(CDM_Temp, "namespace eval ::CDM {}");
    Tcl_Eval(Interp, CDM_Temp);
    
    ID.Init();
    ObjectCache.Init();
    /* Initialise our AttributeSet hash table... */
    CDM_AttributeSetHashInitialise();
    /* Expose the Memory of Cached Objects to the Tcl level... */
    ObjectCache.ExposeCache(Interp);
    Tcl_RegisterObjType(&CDM_TclPointerObjType);
#ifdef ENABLE_ANNOTATION_OBJECT
    Tcl_RegisterObjType(&cdmAnnotationType);
    Tcl_RegisterObjType(&cdmCompactAnnotationType);
#endif /* ENABLE_ANNOTATION_OBJECT */
    Tcl_SetStringObj(&CDM_Obj_GDM_STRING,       "GDM_STRING",       10);
    Tcl_SetStringObj(&CDM_Obj_GDM_STRING_SET,   "GDM_STRING_SET",   14);
    Tcl_SetStringObj(&CDM_Obj_CDM_IMAGE,        "CDM_IMAGE",         9);
    Tcl_SetStringObj(&CDM_Obj_CDM_BASE64_IMAGE, "CDM_BASE64_IMAGE", 16);
    Tcl_SetStringObj(&CDM_Obj_NULL,             "",                  0);
    /* 
     * Make them "virtual" shared objects. If anybody tries to modify them, an
     * error will be raised!
     */
    for (i = 0; i < 10; i++) {
      Tcl_IncrRefCount(&CDM_Obj_GDM_STRING);
      Tcl_IncrRefCount(&CDM_Obj_GDM_STRING_SET);
      Tcl_IncrRefCount(&CDM_Obj_CDM_IMAGE);
      Tcl_IncrRefCount(&CDM_Obj_CDM_BASE64_IMAGE);
      Tcl_IncrRefCount(&CDM_Obj_NULL);
    }
    
    /* Initialise our registered Collection types structures... */
    for (i=0; i<CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER; i++) {
      CDM_AdditionalCollectionTypes[i].CollectionType    = NULL;
      CDM_AdditionalCollectionTypes[i].loadCollectionPtr = NULL;
      CDM_AdditionalCollectionTypes[i].saveCollectionPtr = NULL;
      CDM_AdditionalCollectionTypes[i].loadDocumentPtr   = NULL;
      CDM_AdditionalCollectionTypes[i].saveDocumentPtr   = NULL;
      CDM_AdditionalCollectionTypes[i].CollectionXIdsPtr = NULL;
      CDM_AdditionalCollectionTypes[i].CollectionIdsPtr  = NULL;
      CDM_AdditionalCollectionTypes[i].DocumentAttrsPtr  = NULL;
    }

    /* Link CDM_DefaultEncoding with Tcl variable CDM_DefaultEncoding */
    CDM_DefaultEncoding = CDM_Strdup("iso8859-1");
    Tcl_LinkVar(Interp, "CDM_DefaultEncoding", (char *) &CDM_DefaultEncoding,
                TCL_LINK_STRING);
    /* Link CDM_CollectionFilename with Tcl variable CDM_CollectionFilename */
    Tcl_LinkVar(Interp,"CDM_CollectionFilename",(char *)&CDM_CollectionFilename,
                TCL_LINK_STRING|TCL_LINK_READ_ONLY);
    /* Link CDM_CollectionTypeFilename with Tcl variable
       CDM_CollectionTypeFilename */
    Tcl_LinkVar(Interp,"CDM_CollectionTypeFilename",
                (char *) &CDM_CollectionTypeFilename,
                TCL_LINK_STRING|TCL_LINK_READ_ONLY);
    /* Link CDM_DisplayErrorMessage in Tcl */
    Tcl_LinkVar(Interp, "CDM_DisplayErrorMessage", 
                (char *) &CDM_DisplayErrorMessage, TCL_LINK_INT);
    /* Link CDM_TempDir in Tcl */
    CDM_TempDir = CDM_Strdup("/tmp");
    Tcl_LinkVar(Interp, "CDM_TempDir", (char *) &CDM_TempDir, TCL_LINK_STRING);
    /* Link CDM_Pid in Tcl */
    Tcl_LinkVar(Interp, "CDM_Pid", (char *) &CDM_Pid, TCL_LINK_INT);
    /* Link CDM_FullGateCompatibility in Tcl */
    Tcl_LinkVar(Interp, "CDM_FullGateCompatibility",
               (char *) &CDM_FullGateCompatibility, TCL_LINK_INT);
    /* Link CDM_EmulateGateGetValue in Tcl */
    Tcl_LinkVar(Interp, "CDM_EmulateGateGetValue",
               (char *) &CDM_EmulateGateGetValue, TCL_LINK_INT);
    /* Link CDM_EnableMemoryCompression in Tcl */
    Tcl_LinkVar(Interp, "CDM_EnableMemoryCompression",
               (char *) &CDM_EnableMemoryCompression, TCL_LINK_INT);
    /* Link CDM_AbortCurrentProcess in Tcl */
    Tcl_LinkVar(Interp, "CDM_AbortCurrentProcess",
               (char *) &CDM_AbortCurrentProcess, TCL_LINK_INT);
    Tcl_LinkVar(Interp, "CDM_DebugStdout",
               (char *) &CDM_DebugStdout, TCL_LINK_INT);
    /* Place a Copyright Notice! */
    Tcl_LinkVar(Interp,"::CDM::CDM_Copyright",
                (char *) &CDM_Copyright,
                TCL_LINK_STRING|TCL_LINK_READ_ONLY);
    CDM_LibraryInitialised = 1;

    /*
     * Finally, register our XML Collection type...
     */
    CDM_XMLCollectionTypeInit(Interp);
  } /* if (!CDM_LibraryInitialised)... */

  /*
   * Register CDM API Commands...
   */
#ifdef CDM_ENABLE_TK_COMMANDS
  if(Tcl_CreateObjCommand(Interp, "CDM_Window", WindowObjCmd, (ClientData) NULL,
                           (Tcl_CmdDeleteProc *) NULL)  == NULL)
       return TCL_ERROR;
#endif /* CDM_ENABLE_TK_COMMANDS */

#ifdef CDM_MEM_DEBUG
  if(Tcl_CreateObjCommand(Interp, "CDM_IsShared", CDM_IsSharedObjCmd,
        (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL)  == NULL)
       return TCL_ERROR;
#endif /* CDM_MEM_DEBUG */
  if(Tcl_CreateObjCommand(Interp, "CDM_FreeTclObjCache",
        CDM_FreeTclObjCacheObjCmd,
        (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL)  == NULL)
       return TCL_ERROR;
  
  for ( ; ; specPtr++) {
    if (specPtr->cmd == NULL) break;
    /* printf("Reg: %s %p\n", specPtr->cmd, specPtr->proc); */
    if(Tcl_CreateObjCommand(Interp, specPtr->cmd, specPtr->proc,
                      (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL) == NULL) {
      return TCL_ERROR;
    }
  }
  
#ifdef CDM_TRAP_SIGNALS
  CDM_InitTrap();
  if(Tcl_CreateObjCommand(Interp, "CDM_Trap", (Tcl_ObjCmdProc*) CDM_TrapObjCmd,
                    (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL)  == NULL)
       return TCL_ERROR;
#else
  if(Tcl_CreateObjCommand(Interp, "CDM_Trap",
                    (Tcl_ObjCmdProc*) CDM_StubCommandObjCmd,
                    (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL)  == NULL)
       return TCL_ERROR;
#endif /* CDM_TRAP_SIGNALS */
  
  /* CDM_MajorVersion */
  if(Tcl_CreateObjCommand(Interp, "CDM_MajorVersion",
                    (Tcl_ObjCmdProc*) CDM_VersionObjCmd,
                    (ClientData) 1, (Tcl_CmdDeleteProc *) NULL)  == NULL)
       return TCL_ERROR;
  /* CDM_MinorVersion */
  if(Tcl_CreateObjCommand(Interp, "CDM_MinorVersion",
                    (Tcl_ObjCmdProc*) CDM_VersionObjCmd,
                    (ClientData) 2, (Tcl_CmdDeleteProc *) NULL)  == NULL)
       return TCL_ERROR;
  /* CDM_Version */
  if(Tcl_CreateObjCommand(Interp, "CDM_Version",
                    (Tcl_ObjCmdProc*) CDM_VersionObjCmd,
                    (ClientData) 3, (Tcl_CmdDeleteProc *) NULL)  == NULL)
       return TCL_ERROR;
  /* CDM_PatchLevel */
  if(Tcl_CreateObjCommand(Interp, "CDM_PatchLevel",
                    (Tcl_ObjCmdProc*) CDM_VersionObjCmd,
                    (ClientData) 4, (Tcl_CmdDeleteProc *) NULL)  == NULL)
       return TCL_ERROR;
  /* CDM_ReleaseDate */
  if(Tcl_CreateObjCommand(Interp, "CDM_ReleaseDate",
                    (Tcl_ObjCmdProc*) CDM_VersionObjCmd,
                    (ClientData) 5, (Tcl_CmdDeleteProc *) NULL)  == NULL)
       return TCL_ERROR;
  /* CDM_ExtendedReleaseDate */
  if(Tcl_CreateObjCommand(Interp, "CDM_ExtendedReleaseDate",
                    (Tcl_ObjCmdProc*) CDM_VersionObjCmd,
                    (ClientData) 6, (Tcl_CmdDeleteProc *) NULL)  == NULL)
       return TCL_ERROR;

  Tcl_PkgProvide(Interp, "CDM", CDM_VERSION);
#if defined(LOG_CONSTRUCTOR_CALLS) || defined(LOG_DESTRUCTOR_CALLS)
  CDM_Log("Cdm_Init: Current Thread Id = %p\n\n", CDM_GetCurrentThread());
#endif /* LOG_CONSTRUCTOR_CALLS || LOG_DESTRUCTOR_CALLS */
  return TCL_OK;
} /* Cdm_Init */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     C O L L E C T I O N            S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*************************** -- NOTE -- ***************************************

 Above every function, there is a simple documentary about the various
 functions.
 This Documentation was done by Afantenos D. Stergos, stergos@iit.demokritos.gr
 and Georgios P. Petasis, petasis@iit.demokritos.gr

************************** -- END NOTE -- ************************************/

/*
 * ---- tip_AnnotateCollection:
 *  DESCRIPTION:
 *   <English:
 *     This function will run the specified annotator over the given
 *     Collection... [Not Implemented]
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  Collection Object. (CDM_Collection)
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *     error occurs, the return value will be TCL_ERROR and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp). If all annotators run without errors, TCL_OK will be
 *     returned.
 *   >
 *   SECTIONS:
 *    <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_AnnotateCollectionObjCmd(CDM_TCL_STD_ARGS) {
  CDM_UPDATE_THREAD_INTERP;
  return TCL_ERROR;
} /* CDM_AnnotateCollectionObjCmd */


/*
 * ---- tip_Close:
 *  DESCRIPTION:
 *   <English:
 *     Closes the given Collection and frees all the memory occupied by the
 *     Collection and its Documents.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  Collection Object to be closed. (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_Close (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    None.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Close
 *   >
 *  SECTIONS:
 *   <Collection
 *   >:&:
 */

/*
 * ---- tip_Close:
 *  DESCRIPTION:
 *   <English:
 *     Closes the given Document and frees all the memory occupied by this
 *     Document.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:  Document Object to be closed. (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_Close (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    None.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Close
 *   >
 *  SECTIONS:
 *    <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_CloseObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Col;
  CDM_Document   Doc;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection/Document"); return TCL_ERROR;}
  Col = CDM_GetCollectionPointer(objv[1]);
  if (Col != NULL) return CDM_Close(Col);
  Doc = CDM_GetDocumentPointer(objv[1]);
  if (Doc != NULL) return CDM_Close(Doc);
  return CDM_ReportError("Invalid Collection/Document Object");
} /* CDM_CloseObjCmd */


/*
 * ---- tip_CreateCollection:
 *  DESCRIPTION:
 *   <English:
 *     This function creates a new Collection object. It accepts three
 *     arguments: the Collection's name which must be the absolute path of
 *     the Collection on disk, a set of Attributes (created with
 *     tip_CreateAtttributeSet) that will be inserted into
 *     the new Collection and the encoding of the Collection. The Collection's
 *     encoding will be the default encoding that will be used when a new
 *     Document will be created in this Collection. Note that each Document can
 *     have a different encoding than the encoding of the container
 *     Collection. The only place that this encoding will be used, is in the
 *     case where a new Document of unspecified encoding will be created.
 *
 *     If the Collection encoding argument is ommited, it defaults to 
 *     "tip_DefaultEncoding". "tip_DefaultEncoding" is a global variable (of
 *     type char *), that has as initial value the system encoding.
 *     This variable is also
 *     exported at the Tcl level under the same name and as a result its value
 *     can easily be changed by the end user.
 *
 *     Note that the new Collection will be created only in memory. Nothing
 *     will be saved on the disk. In order for the Collection to be saved in
 *     disk, the function tip_Sync must be called after the new Collection is
 *     created in memory. This is usually done after all of the desired
 *     Documents have been added to the newly created Collection.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Name:        The Collection's desired disk location (Name) (string)
 *      Attributes:  A set of Attributes (tip_AttributeSet)
 *      Encoding:    The Collection's Encoding (Optional) (string)
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Collection object (of type CDM_Collection) that
 *     can be used as a reference to all functions that require a Collection
 *     argument as input. This object will be valid until the Collection that
 *     is referred by this object is closed (through tip_Close) and the
 *     Collection removed from memory. If this object is used after the 
 *     referred Collection has been closed, an undefined behaviour will occur.
 *
 *     If an error occurs, then an error message will propmpt.
 *
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Close, tip_CreateDocument, tip_CreateAttribute,
 *     tip_CreateAttributeSet, tip_Sync
 *   >
 *  SECTIONS:
 *   <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_CreateCollectionObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Col;
  Tcl_Obj *Result;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc < 3 || objc > 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Name Attributes ?Encoding?");
    return TCL_ERROR;
  }
  if (objc == 3) {
    Col = CDM_CreateCollection(Tcl_GetString(objv[1]), objv[2]);
  } else {
    Col = CDM_CreateCollection(Tcl_GetString(objv[1]), objv[2], 
                               Tcl_GetString(objv[3]));
  }
  if (Col == NULL) return TCL_ERROR;
  Result = CDM_NewStringObj(Tcl_GetCommandName(interp, Col->Tcl_Cmd));
  if (Result == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Result);
  return TCL_OK;
} /* CDM_CreateCollectionObjCmd */


/*
 * ---- tip_CreateDocument:
 *  DESCRIPTION:
 *   <English:
 *     This function creates a new Document object in an existing Collection 
 *     object. It accepts five (or six) arguments: the Collection object in
 *     which the new Document is going to be created (Document's Parent), the
 *     External Id of the Document (which represents the full path in disk of
 *     the original file that contains the Document's text), the text (RawData
 *     or ByteSequence) of the new Document, an initial Annotation set, an
 *     initial Attribute set and (optionally) the encoding of the Document.
 *
 *     The Collection object must be a valid object (corresponding to an open
 *     Collection) returned by either tip_CreateCollection or
 *     tip_OpenCollection.
 *
 *     The External Id must be a valid (absolute) path. A local copy of this
 *     argument will be created for use by this function. The existance of
 *     this path will not be checked. This path value will be
 *     converted automatically to a valid platform dependant path. As a result,
 *     for this value can be safely used the notation used for paths under the
 *     unix operating system. That means that the path:
 *     [[br]][[code]][[center]]/Users/petasis/Collections[[/center]][[/code]]
 *     will be converted interanlly to the following path, assuming windows as
 *     operating system :
 *     [[br]][[code]][[center]]C:\Users\petasis\Collections[[/center]][[/code]][[br]]
 *     
 *     The RawData parameter must contain the desired for the new Document
 *     text, in UTF-8 format. A local copy of this variable will be also
 *     created. If the text is not in UTF-8 (i.e. contains text just red from
 *     a file using the standart C library routines and is in the ISO 8859-7
 *     encoding) then must be converted to UTF using tip_ExternalToUtf. If
 *     the origin of this string is Tcl, then it is already in UTF format, as
 *     Tcl uses ONLY UTF for encoding strings internally.
 *
 *     The Annotations parameter must contain an Annotation set created with
 *     tip_CreateAnnotationSet.
 *
 *     The Attributes parameter must hold a valid Attribute set created with
 *     tip_CreateAttributeSet.
 *
 *     Finally, an optional encoding can be specified. A local copy of this
 *     string will be created. The value of this parameter must be a standart
 *     Tcl encoding value (like iso8859-7 or cp1253). For all available Tcl
 *     encodings please refer to the 
 *     [[A HREF="../TclTk/TclLib/Encoding.htm"]]Tcl manuals[[/A]].
 *     If this parameter is ommitted,
 *     then a default value will be used. This value will be inherited from the
 *     parent Collection object. It is important for the encoding to correctly
 *     describe the text. If the given UTF string cannot be converted to the
 *     Document's encoding, then the text will be filled with the character "?"
 *     in places where the conversion will fail.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  Collection Object.             (CDM_Collection)
 *      ExternalId:  Document's External Id (disk location of original text file).     (string)
 *      RawData:     Document's Text.               (CDM_ByteSequence)
 *      Annotations: Document's Annotations         (CDM_AnnotationSet)
 *      Attributes:  Document's Attributes          (CDM_AttributeSet)
 *      Encoding:    Document's Encoding (Optional) (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_CreateDocument (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Document object (of type CDM_Document) that can
 *     be used as a reference to all functions that require a Document object
 *     as input. This object will be valid until the Collection that holds this
 *     Document object is closed (through tip_Close) and the parent Collection
 *     removed from memory. If this object is used after its parent Collection
 *     has been closed, an undefined behaviour will occur.
 *
 *     If an error occurs, then an error message wil prompt.
 *
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_CreateAnnotationSet, tip_CreateAttributeSet, tip_CreateCollection,
 *     tip_ExternalToUtf, tip_OpenCollection
 *     [[A HREF="../TclTk/TclLib/Encoding.htm"]]The Tcl Encoding manual[[/A]].
 *   >
 *  SECTIONS:
 *    <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_CreateDocumentObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_Document CDM_Ptr;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 6 && objc != 7) {
    Tcl_WrongNumArgs(interp, 1, objv, 
                "Collection XID RawData Annotations Attributes ?Encoding?");
    return TCL_ERROR;
  }
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection == NULL) return TCL_ERROR;
  if (objc == 7) {
    CDM_Ptr = CDM_CreateDocument(Collection, Tcl_GetString(objv[2]),
               objv[3], objv[4], objv[5], Tcl_GetString(objv[6]));
  } else {
    CDM_Ptr = CDM_CreateDocument(Collection, Tcl_GetString(objv[2]),
               objv[3], objv[4], objv[5]);
  }
  if (CDM_Ptr == NULL) return TCL_ERROR;
  Tcl_SetResult(interp, (char *)
          Tcl_GetCommandName(CDM_Ptr->Interp, CDM_Ptr->Tcl_Cmd), TCL_VOLATILE);
  return TCL_OK;
} /* CDM_CreateDocumentObjCmd */



/*
 * ---- tip_Destroy:
 *  DESCRIPTION:
 *   <English:
 *     This function will delete the disk representation of a Collection
 *     object. The Name parameter must be the (absolute) path to the directory
 *     that holds the Collection. Having a Collection object, this information
 *     can be obtained through tip_GetName.
 *
 *     Note that this function will only delete the current disk
 *     representation of a Collection. If the Collection is loaded in memory
 *     and we save the loaded Collection (with the help of tip_Sync), then the
 *     representaion of the Collection in disk will be re-created.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Name:  The Collection's Name                         (string)
 *   >
 *  NOTES:
 *   <English:
 *     Note that this function will delete the whole directory that holds the
 *     Collection, along with ALL the files in that directory. If the user has
 *     placed files inside the Collection's directory, these files will be
 *     also deleted.
 *
 *     This function is equivelant to CDM_Destroy (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value of this function will be a standart Tcl completion code
 *     (of type int) with one of the values TCL_OK and TCL_ERROR. If the 
 *     Collection's directory cannot be removed, the return value will be
 *     TCL_ERROR and an error message describing the error will be left at the
 *     current active Tcl interpreter (tip_Interp). If the deletion was done
 *     successfully, TCL_OK will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetName
 *   >
 *  SECTIONS:
 *    <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_DestroyObjCmd(CDM_TCL_STD_ARGS) {
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Name");
    return TCL_ERROR;
  }
  return CDM_Destroy(Tcl_GetString(objv[1]));
} /* CDM_DestroyObjCmd */


/*
 * ---- tip_FirstDocument:
 *  DESCRIPTION:
 *   <English:
 *    This function will return a Document object representing the first
 *    Document of the specified Collection. It can be used in order
 *    to start an iteration over all Documents of the Collection. This
 *    function will return the first Document in the given Collection and make
 *    the proper arrangements so as subsequent calls to tip_NextDocument
 *    return the following Documents.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_FirstDocument (C++ API).
 *     
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Document object (of type CDM_Document) that can
 *     be used as a reference to all functions that require a Document object
 *     as input. This object will be valid until the Collection that holds this
 *     Document object is closed (through tip_Close) and the parent Collection
 *     removed from memory. If this object is used after its parent Collection
 *     has been closed, an undefined behaviour will occur.
 *
 *     If an error occurs, then an error message will prompt.
 *
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Close, tip_NextDocument
 *   >
 *  SECTIONS:
 *    <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_FirstDocumentObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_Document CDM_Ptr;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection"); return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection == NULL) return TCL_ERROR;
  CDM_Ptr = CDM_FirstDocument(Collection);
  if (CDM_Ptr == NULL) return TCL_ERROR;
  Tcl_SetResult(interp, (char *)
          Tcl_GetCommandName(CDM_Ptr->Interp, CDM_Ptr->Tcl_Cmd), TCL_VOLATILE);
  return TCL_OK;
} /* CDM_FirstDocumentObjCmd */


/*
 * ---- tip_GetAssociatedInfo:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the Tcl object associated by
 *     the user to this Collection object.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     Note that this information is not saved when a Collection is saved. It
 *     exists only while the Collection is open, and is different for each
 *     Collection instance, even if more than one Collection instances have
 *     been created from a single Collection image on the disk.
 *
 *     This function is equivelant to CDM_GetAssociatedInfo (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Tcl object if a Tcl object
 *     has already been associated with this Collection object. Otherwise,
 *     an error message prompts.
 *
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_SetAssociatedInfo, tip_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], Tcl_IncrRefCount
 *   >
 *  SECTIONS:
 *    <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_GetAssociatedInfoObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  Tcl_Obj *CDM_Ptr;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection"); return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection == NULL) return TCL_ERROR;
  CDM_Ptr = CDM_GetAssociatedInfo(Collection);
  if (CDM_Ptr == NULL) CDM_Ptr = &CDM_Obj_NULL;
  Tcl_SetObjResult(interp, CDM_Ptr);
  return TCL_OK;
} /* CDM_GetAssociatedInfoObjCmd */


/*
 * ---- tip_GetAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function will search all the Attributes that the specified
 *     Collection has, for an Attribute that has the same name as the "Name"
 *     parameter. If such an Attribute is found, the Attribute object is
 *     returned. If an Attribute with the specified name does not exist, then an 
 * 
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *      Name:        Attribute name.      (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetAttribute (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Attribute object (of type CDM_Attribute) if an
 *     Attribute named as specified is found. In case that the specified
 *     Attribute does not exist, an error message prompts.
 *
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], Tcl_IncrRefCount
 *   >
 *  SECTIONS:
 *    <Collection Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_GetAttributeObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, 
    "Collection/Document/Annotation AttributeName"); return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection != NULL) {
    ObjPtr = CDM_GetAttribute(Collection, Tcl_GetString(objv[2]));
    if (ObjPtr == NULL) goto no_such_attr;
    Tcl_SetObjResult(interp, ObjPtr);
    return TCL_OK;
  } 
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document != NULL) {
    ObjPtr = CDM_GetAttribute(Document, Tcl_GetString(objv[2]));
    if (ObjPtr == NULL) goto no_such_attr;
    Tcl_SetObjResult(interp, ObjPtr);
    return TCL_OK;
  }
  ObjPtr = CDM_GetAttribute(objv[1], Tcl_GetString(objv[2]));
  if (ObjPtr == NULL) goto no_such_attr;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
no_such_attr:
  Tcl_SetObjResult(interp, (&CDM_Obj_NULL));
  return TCL_OK;
} /* CDM_GetAttributeObjCmd */


/*
 * ---- tip_GetAttributes:
 *  DESCRIPTION:
 *   <English:
 *     This function returns a new object (of type CDM_AttributeSet) that
 *     contains all the Attributes that the specified Collection currently
 *     has. 
 *
 *     This function will always create and return a new Attribute set object,
 *     even when the Collection has no Attributes defined. (In this case an
 *     empty Attribute set will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetAttributes (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Attribute set object (of type CDM_AttributeSet).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_GetAttribute
 *   >
 *  SECTIONS:
 *    <Collection Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_GetAttributesObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection/Document/Annotation");
    return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection != NULL) {
    ObjPtr = CDM_GetAttributes(Collection);
    if (ObjPtr == NULL) return TCL_ERROR;
    Tcl_SetObjResult(interp, ObjPtr);
    return TCL_OK;
  } 
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document != NULL) {
    ObjPtr = CDM_GetAttributes(Document);
    if (ObjPtr == NULL) return TCL_ERROR;
    Tcl_SetObjResult(interp, ObjPtr);
    return TCL_OK;
  }
  ObjPtr = CDM_GetAttributes(objv[1]);
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_GetAttributesObjCmd */


/*
 * ---- tip_GetByExternalId:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Document object that has the same External
 *     Id as the value of the "XID" parameter. The value of the "XID" parameter
 *     is assumed to be encoded using the UTF-8 encoding (thus enabling the 
 *     existance of non Latin characters in the value).
 *     If more than one Documents
 *     having the same External Id are found in the specified Collection, then
 *     the Document that was created first will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object.                     (CDM_Collection)
 *      XID:         The External Id of the desired Document.  (string)
 *   >
 *  NOTES:
 *   <English:
 *     Note that a Collection can have more than one Documents with the same
 *     External Id. These Documents will all have the same External Id but
 *     they will have different Document Ids.
 *
 *     This function will make the proper arrangments so as a call to
 *     tip_NextDocument (or tip_NextDocument) will return the Document that
 *     was created after the one returned by this function.
 *
 *     This function is equivelant to CDM_GetByExternalId (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Document object (of type CDM_Document) that can
 *     be used as a reference to all functions that require a Document object
 *     as input. This object will be valid until the Collection that holds this
 *     Document object is closed (through tip_Close) and the parent Collection
 *     removed from memory. If this object is used after its parent Collection
 *     has been closed, an undefined behaviour will occur.
 *
 *     If a Document with the requested external Id does not exist, or an
 *     error occurs, then an error message will prompt.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Close, tip_GetDocument, tip_NextDocument
 *   >
 *  SECTIONS:
 *   <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_GetByExternalIdObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_Document CDM_Ptr;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection XID"); return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection == NULL) return TCL_ERROR;
  CDM_Ptr = CDM_GetByExternalId(Collection, Tcl_GetString(objv[2]));
  if (CDM_Ptr == NULL) return TCL_ERROR;
  Tcl_SetResult(interp, (char *)
      Tcl_GetCommandName(CDM_Ptr->Interp, CDM_Ptr->Tcl_Cmd), TCL_VOLATILE);
  return TCL_OK;
} /* CDM_GetByExternalIdObjCmd */


/*
 * ---- tip_GetComment:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the Tcl object holding a
 *     comment associated to this this Collection object.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetComment (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Tcl object if a comment
 *     has already been associated with this Collection object. Otherwise,
 *     an error message prompts.
 *
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_SetComment, tip_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], Tcl_IncrRefCount
 *   >
 *  SECTIONS:
 *   <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_GetCommentObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  Tcl_Obj *CDM_Ptr;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection"); return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection == NULL) return TCL_ERROR;
  CDM_Ptr = CDM_GetComment(Collection);
  if (CDM_Ptr == NULL) CDM_Ptr = &CDM_Obj_NULL;
  Tcl_SetObjResult(interp, CDM_Ptr);
  return TCL_OK;
} /* CDM_GetCommentObjCmd */


/*
 * ---- tip_GetDocument:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Document object that has the same Id as 
 *     the value of the "Id" parameter. The value of the "Id" parameter
 *     is assumed to be encoded using the UTF-8 encoding (thus enabling the 
 *     existance of non Latin characters in the value).
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object.                     (CDM_Collection)
 *      Id:          The Id of the desired Document.          (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function will make the proper arrangments so as a call to
 *     tip_NextDocument (or tip_NextDocument) will return the Document that
 *     was created after the one returned by this function.
 *
 *     This function is equivelant to CDM_GetDocument (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Document object (of type CDM_Document) that can
 *     be used as a reference to all functions that require a Document object
 *     as input. This object will be valid until the Collection that holds this
 *     Document object is closed (through tip_Close) and the parent Collection
 *     removed from memory. If this object is used after its parent Collection
 *     has been closed, an undefined behaviour will occur.
 *
 *     If a Document with the requested Id does not exist, or an error occurs,
 *     then an error message will prompt.
 *
 *     Note that the returned object is property of the CDM and should never be
 *     freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Close, tip_GetByExternalId
 *   >
 *  SECTIONS:
 *   <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_GetDocumentObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_Document CDM_Ptr;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 3) {
        Tcl_WrongNumArgs(interp, 1, objv, "Collection Id"); return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection == NULL) return TCL_ERROR;
  CDM_Ptr = CDM_GetDocument(Collection, Tcl_GetString(objv[2]));
  if (CDM_Ptr == NULL) return TCL_ERROR;
  Tcl_SetResult(interp, (char *)
      Tcl_GetCommandName(CDM_Ptr->Interp, CDM_Ptr->Tcl_Cmd), TCL_VOLATILE);
  return TCL_OK;
} /* CDM_GetDocumentObjCmd */



/*
 * ---- tip_GetEncoding:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the encoding of the specified Collection. The
 *     return value will be of type char* and will be owned by CDM. Its value
 *     will be a standart Tcl encoding value (like iso8859-7 or cp1253). For
 *     all available Tcl encodings please refer to the 
 *     [[A HREF="../TclTk/TclLib/Encoding.htm"]]Tcl manuals[[/A]].
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object.           (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetEncoding (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a UTF-8 string that represents the encoding of
 *     the specified Collection.
 *
 *     If an error occurs, then an error message will prompt.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_SetEncoding
 *     [[A HREF="../TclTk/TclLib/Encoding.htm"]]The Tcl Encoding manual[[/A]].
 *   >
 *  SECTIONS:
 *   <Collection Document
 *   >:&:
 */
CDM_CALLBACK int CDM_GetEncodingObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_Document Document;
  char *CPtr;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection/Document"); return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection != NULL) {
    CPtr = CDM_GetEncoding(Collection);
    if (CPtr == NULL) return TCL_ERROR;
    Tcl_SetResult(interp, CPtr, TCL_VOLATILE);
    return TCL_OK;
  }
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document != NULL) {
    CPtr = CDM_GetEncoding(Document);
    if (CPtr == NULL) return TCL_ERROR;
    Tcl_SetResult(interp, CPtr, TCL_VOLATILE);
    return TCL_OK;
  }
  return TCL_ERROR;
} /* CDM_GetEncodingObjCmd */


/*
 * ---- tip_GetOwner:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Owner of the specified Collection. Its
 *     value is not of great importance, and usually defaults to the value 
 *     "CDM". The returned value will be encoded using the UTF-8 encoding
 *     (thus enabling the existance of non Latin characters in the value).
 *
 *     The return value will be of type string.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.           (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetOwner (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a UTF-8 string that represents the Owner of the
 *     specified Collection.
 *
 *     If an error occurs, then an error message will prompt.
 *
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_SetOwner
 *   >
 *  SECTIONS:
 *   <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_GetOwnerObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  char *CPtr;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection"); return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection == NULL) return TCL_ERROR;
  CPtr = CDM_GetOwner(Collection);
  if (CPtr == NULL) return TCL_ERROR;
  Tcl_SetResult(interp, CPtr, TCL_VOLATILE);
  return TCL_OK;
} /* CDM_GetOwnerObjCmd */


/*
 * ---- tip_NextDocument:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the next available Document object. "Next"
 *     means the Document that follows the one previously returned by calls to
 *     the following functions:
 *     tip_FirstDocument, tip_NextDocument, tip_GetDocument,
 *     tip_GetByExternalId.
 *
 *     If previous calls to this function have returned all available
 *     Documents, then an error will prompt and the current active Tcl
 *     interpreter (CDM_Interp) will be reseted. If an error occurs, then an 
 *     error message will prompt again. As a result, the caller must check the
 *     information stored at this Tcl interpreter in order to distinguish
 *     between the normal termination of the Document chain or an error
 *     condition.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.           (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     Note that if this function is called with a Collection object that has
 *     just be created or opened, an undefined behaviour will occur.
 *
 *     This function is equivelant to CDM_NextDocument (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Document object (of type CDM_Document) that can
 *     be used as a reference to all functions that require a Document object
 *     as input. This object will be valid until the Collection that holds this
 *     Document object is closed (through CDM_Close) and the parent Collection
 *     removed from memory. If this object is used after its parent Collection
 *     has been closed, an undefined behaviour will occur.
 *
 *     If an error occurs, an error message will prompt. An error message will
 *     also be prompted if the last Document of the specified Collection has
 *     already been returned by any of the following functions:
 *     tip_FirstDocument, tip_NextDocument, tip_GetDocument,
 *     tip_GetByExternalId.
 *
 *     Also note that the returned object is property of the CDM and should
 *     never be freed or modified in any way by the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Close, tip_FirstDocument, tip_NextDocument, tip_GetDocument,
 *     tip_GetByExternalId
 *   >
 *  SECTIONS:
 *   <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_NextDocumentObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_Document CDM_Ptr;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection"); return TCL_ERROR;}
  Tcl_ResetResult(interp);
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection == NULL) return TCL_ERROR;
  CDM_Ptr = CDM_NextDocument(Collection);
  if (CDM_Ptr == NULL) {
    /*
     * We have to check whether it an error, or simply we have already open
     * all the Documents. In case of an error, the interpreter will have a
     * string, describing the error...
     */
    if (strlen(Tcl_GetString(Tcl_GetObjResult(interp)))) {
      return TCL_ERROR;
    }
    /*
     * Else, the empty string is already in the interpreter, as we have
     * resetted it...
     */
  } else {
    Tcl_SetResult(interp, (char *)
          Tcl_GetCommandName(CDM_Ptr->Interp, CDM_Ptr->Tcl_Cmd), TCL_VOLATILE);
  }
  return TCL_OK;
} /* CDM_NextDocumentObjCmd */



/*
 * ---- tip_OpenCollection:
 *  DESCRIPTION:
 *   <English:
 *     This function will load the Collection from the directory specified by
 *     the "Name" argument. If the target that "Name" points is a directory
 *     that is readable by the current user and contains a valid Collection
 *     specification (either in native or in GATE format), then this
 *     Collection will be loaded from disk to the memory. A new Collection
 *     object (of type CDM_Collection) will be created and returned to the
 *     caller.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Name:  The Name of the Collection (absolute path).   (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_OpenCollection (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will construct and return a Collection object (of type
 *     CDM_Collection) that will represent the Collection stored at the
 *     directory specified by the "Name" parameter. The caller must use this
 *     object in order to refer to the newly opened Collection and must use
 *     the function tip_Close in order to close the Collection and free the
 *     occupied memory if the Collection is not needed any more.
 *
 *     In case of an error an error message will prompt.
 *
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Close
 *   >
 *  SECTIONS:
 *   <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_OpenCollectionObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Col;
  Tcl_Obj *Result;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Name");
    return TCL_ERROR;
  }
  Col = CDM_OpenCollection(Tcl_GetString(objv[1]));
  if (Col == NULL) return TCL_ERROR;
  Result = CDM_NewStringObj(Tcl_GetCommandName(interp, Col->Tcl_Cmd));
  if (Result == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Result);
  return TCL_OK;
} /* CDM_OpenCollectionObjCmd */



/*
 * ---- tip_PutAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function will add a given Attribute to the specified Collection
 *     object. If an Attribute with the same name already exists in the
 *     Collection, then it will be overwritten by the new Attribute. Else, the
 *     new Attribute will be appended to the existing Attribute set of the
 *     specified Collection.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object.       (CDM_Collection)
 *      Attribute:   The Attribute to be added. (CDM_Atrribute)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_PutAttribute (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If an
 *     error occurs, the return value will be TCL_ERROR and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp) an be prompted as well. Else, TCL_OK will be returned.
 *   >
 *  SECTIONS:
 *   <Collection Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_PutAttributeObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_Document Document;
  Tcl_Obj *Obj;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 3 && objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, 
        "Collection/Document/Annotation Attribute ?args?");return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection != NULL) return CDM_PutAttribute(Collection, objv[2]);
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document != NULL)   return CDM_PutAttribute(Document, objv[2]);
  Obj = CDM_PutAttribute(objv[1], objv[2]);
  if (Obj == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Obj);
  return TCL_OK;
} /* CDM_PutAttributeObjCmd */


/*
 * ---- tip_RemoveAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function will remove the Attribute named exactly as the "Name"
 *     parameter. If the requested Attribute does not exist, an error
 *     condition will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.           (CDM_Collection)
 *     Name:        The Name of the Attribute to be removed. (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_RemoveAttribute (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Attribute exists and is successfully removed, then TCL_OK
 *     will be returned. In case of an error, the return value will be
 *     TCL_ERROR and an error message describing the error will be left at the
 *     current Tcl interpreter (CDM_Interp) and be prompted.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_RemoveAttribute
 *   >
 *  SECTIONS:
 *   <Collection Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_RemoveAttributeObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_Document Document;
  Tcl_Obj *Obj;
  char *Name;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, 
          "Collection/Document/Annotation AttributeName");return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  Name = Tcl_GetString(objv[2]);
  if (Collection != NULL) return CDM_RemoveAttribute(Collection, Name);
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document != NULL)   return CDM_RemoveAttribute(Document, Name);
  Obj = CDM_RemoveAttribute(objv[1], Name);
  if (Obj == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Obj);
  return TCL_OK;
} /* CDM_RemoveAttributeObjCmd */


/*
 * ---- tip_RemoveDocument:
 *  DESCRIPTION:
 *   <English:
 *     This function will remove the Document object that has as Id the
 *     specified by the parameter "Id" value. If a Document with the requested
 *     Id does not exist, then an error will be prompted.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.                  (CDM_Collection)
 *      Id:         The Id of the Document to be removed. (string)
 *   >
 *  NOTES:
 *   <English:
 *     If the removed Document object has been previously opened (through 
 *     tip_GetByExternalId, tip_GetDocument, tip_FirstDocument or
 *     tip_NextDocument) then the returned by these functions Document object
 *     will be invalidated. 
 *
 *     This function is equivelant to CDM_RemoveDocument (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Document exists and is successfully removed, then TCL_OK will
 *     be returned. In case of an error, the return value will be TCL_ERROR and
 *     an error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetByExternalId, tip_GetDocument, tip_FirstDocument,
 *     tip_NextDocument
 *   >
 *  SECTIONS:
 *   <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_RemoveDocumentObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection Id"); return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection == NULL) return TCL_ERROR;
  return CDM_RemoveDocument(Collection, Tcl_GetString(objv[2]));
} /* CDM_RemoveDocumentObjCmd */


/*
 * ---- tip_SetAssociatedInfo:
 *  DESCRIPTION:
 *   <English:
 *     This function sets the provided Tcl object as 
 *     associated information to the specified Collection object. This
 *     "associated information" (Tcl object) can contain anything that the
 *     user wants to store ("associate") with this Collection. 
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *      Info:        A Tcl object holding the information that is to be associated with the specified Collection
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetAssociatedInfo (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Collection exists and the information Tcl object is
 *     successfully associated with the Collection, then TCL_OK will
 *     be returned. In case of an error, the return value will be TCL_ERROR and
 *     an error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetAssociatedInfo, tip_SetComment, tip_GetComment, tip_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], Tcl_IncrRefCount
 *   >
 *  SECTIONS:
 *   <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_SetAssociatedInfoObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection Info"); return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection == NULL) return TCL_ERROR;
  return CDM_SetAssociatedInfo(Collection, objv[2]);
} /* CDM_SetAssociatedInfoObjCmd */


/*
 * ---- tip_SetComment:
 *  DESCRIPTION:
 *   <English:
 *     This function sets the provided Tcl object as 
 *     comment to the specified Collection object. This "comment" (Tcl object)
 *     can contain anything that the user wants to store with this Collection.
 *     This comment
 *     object will be saved allong the Collection, if the Collection is saved.
 *     This is the primary difference of comment towards associated
 *     information that can be inserted with 
 *     tip_GetAssociatedInfo/tip_SetAssociatedInfo.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection:  A Collection object. (CDM_Collection)
 *      Info:        A Tcl object holding the information that is to be associated as comment with the specified Collection 
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetComment (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Collection exists and the comment Tcl object is
 *     successfully associated with the Collection, then TCL_OK will
 *     be returned. In case of an error, the return value will be TCL_ERROR and
 *     an error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetComment, tip_GetAssociatedInfo, tip_SetAssociatedInfo, tip_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], Tcl_IncrRefCount
 *   >
 *  SECTIONS:
 *   <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_SetCommentObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection Comment"); return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection == NULL) return TCL_ERROR;
  return CDM_SetComment(Collection, objv[2]);
} /* CDM_SetCommentObjCmd */



/*
 * ---- tip_SetEncoding:
 *  DESCRIPTION:
 *   <English:
 *     This function will change the encoding of a Collection object to the
 *     encoding specified by the "Encoding" parameter.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.                        (CDM_Collection)
 *     Encoding:    The desired new encoding of the Collection. (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_SetEncoding (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a string that holds the new encoding value. 
 *
 *     In case of an error, an error message will eb prompted.
 *
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetEncoding
 *   >
 *  SECTIONS:
 *   <Collection Document
 *   >:&:
 */
CDM_CALLBACK int CDM_SetEncodingObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_Document Document;
  char *CPtr;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection/Document Encoding");
    return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection != NULL) {
    CPtr = CDM_SetEncoding(Collection, Tcl_GetString(objv[2]));
    if (CPtr == NULL) return TCL_ERROR;
    Tcl_SetObjResult(interp, CDM_NewStringObj(CPtr));
    return TCL_OK;
  }
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  CPtr = CDM_SetEncoding(Document, Tcl_GetString(objv[2]));
  if (CPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, CDM_NewStringObj(CPtr));
  return TCL_OK;
} /* CDM_SetEncodingObjCmd */


/*
 * ---- tip_SetName:
 *  DESCRIPTION:
 *   <English:
 *     This function will change the Name of the specified Collection. The
 *     Name of the Collection represents the (absolute) path of the
 *     Collection's disk representation. Changing the Name of a Collection
 *     will result into a new Collection representation in disk if the
 *     Collection is saved (through tip_Sync). Note that this function will
 *     not perform any checks on the value of the "Name" parameter. If an
 *     invalid path is given, or a path that corresponds to an existing file,
 *     no error will be returned by this function. (Although tip_Sync will
 *     return an error if the caller tries to save a Collection with an
 *     invalid Name.) 
 *
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.                (CDM_Collection)
 *     Name:    The desired new Name of the Collection. (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_SetName (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a string
 *     that holds the new Collection Name. 
 *
 *     In case of an error, an error message will be prompted.
 *
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Sync
 *   >
 *  SECTIONS:
 *   <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_SetNameObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  char *CPtr;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection Name"); return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection == NULL) return TCL_ERROR;
  CPtr = CDM_SetName(Collection, Tcl_GetString(objv[2]));
  if (CPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Tcl_NewStringObj(CPtr, -1));
  return TCL_OK;
} /* CDM_SetNameObjCmd */


/*
 * ---- tip_SetOwner:
 *  DESCRIPTION:
 *   <English:
 *     This function will change the Owner of the specified Collection.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.                     (CDM_Collection)
 *     Owner:       The desired new Owner of the Collection. (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_SetOwner (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a string 
 *     that holds the new Collection Owner.
 *
 *     In case of an error, an error message will prompt.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_SetOwner
 *   >
 *  SECTIONS:
 *   <Collection
 *   >:&:
 */
CDM_CALLBACK int CDM_SetOwnerObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  char *CPtr;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection Owner"); return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection == NULL) return TCL_ERROR;
  CPtr = CDM_SetOwner(Collection, Tcl_GetString(objv[2]));
  if (CPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, CDM_NewStringObj(CPtr));
  return TCL_OK;
} /* CDM_SetOwnerObjCmd */


/*
 * ---- tip_Sync:
 *  DESCRIPTION:
 *   <English:
 *     This function will save a Collection object in disk. If the Collection
 *     object has never been saved again, then this function will create the
 *     directory that the Name of the Collection specifies. A representation
 *     of the Collection will be saved in this directory. If the directory
 *     already exists it will be overwritten. If the directory cannot be
 *     created an error condition will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.           (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     The user is not expected to modify directories that hold Collections in
 *     any way. Modifing any of the files under the Collection directory may
 *     result in a corrupted, unreadble Collection.
 *
 *     This function is equivelant to CDM_Sync (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Collection is successfully saved, TCL_OK will
 *     be returned. In case of an error, the return value will be TCL_ERROR and
 *     an error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).     
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_Sync
 *   >
 *  SECTIONS:
 *   <Collection Document
 *   >:&:
 */
CDM_CALLBACK int CDM_SyncObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_Document   Document;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 2) {
     Tcl_WrongNumArgs(interp, 1, objv, "Collection/Document"); 
     return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection != NULL) return CDM_Sync(Collection);
  Document   = CDM_GetDocumentPointer(objv[1]);
  if (Document != NULL) return CDM_Sync(Document);
  return CDM_ReportError("Invalid Collection/Document Pointer object");
} /* CDM_SyncObjCmd */


/*
 * ---- tip_Status:
 *  DESCRIPTION:
 *   <English:
 *    This function will return information about the specified Collection.
 *    This information will be stored as text in a newly created Tcl object.
 *
 *    This function will always create and return a new Tcl string object.
 *    In case of an error, an error message will be prompted.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.           (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a Tcl string object,
 *    containing the information of the specified Collection.
 *    The caller is responsible to free this object through Tcl_DecrRefCount
 *    or tip_Free.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_Free, Tcl_DecrRefCount
 *   >
 *  SECTIONS:
 *   <Collection Document
 *   >:&:
 */
CDM_CALLBACK int CDM_StatusObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Collection;
  CDM_Document   Document;
  Tcl_Obj *StatusObj = NULL;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 2) {
     Tcl_WrongNumArgs(interp, 1, objv, "Collection/Document");
     return TCL_ERROR;}
  Collection = CDM_GetCollectionPointer(objv[1]);
  if (Collection != NULL) {
    StatusObj = CDM_Status(Collection);
  }
  Document   = CDM_GetDocumentPointer(objv[1]);
  if (Document != NULL) {
    StatusObj = CDM_Status(Document);
  }
  if (StatusObj == NULL) {
    return CDM_ReportError("Invalid Collection/Document Pointer object");
  }
  Tcl_SetObjResult(interp, StatusObj);
  return TCL_OK;
} /* CDM_StatusObjCmd */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     D O C U M E N T                S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/
 
 
/*
 * ---- tip_AnnotationsAt:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that have the specified text position (as given
 *     by the parameter "Position") as a start position in any of their spans.
 *     CDM does not own the returned
 *     object: the caller is responsible to use Tcl_DecrRefCount or CDM_Free 
 *     in order to free the object and the memory associated with it.
 *
 *     This function will always create and return a new Annotation set object,
 *     even when the Document has no Annotations defined or when no Annotation
 *     is found that includes the specified text position into its span set.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, an error message will be prompted.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Position:    Position in text.             (int)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_AnnotationsAt (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). The caller is responsible to free this object 
 *     through Tcl_DecrRefCount or tip_Free.
 *
 *     If an error occurs, an error message will be prompted.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_AnnotationsContaining, tip_AnnotationsInRange,
 *     tip_AnnotationsMatchingRange, tip_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]]
 *   >
 *  SECTIONS:
 *   <Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_AnnotationsAtObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  long Integer;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document/AnnotationSet Position");
    return TCL_ERROR;}
  if (Tcl_GetLongFromObj(interp, objv[2], &Integer) != TCL_OK) return TCL_ERROR;
  
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) {
    ObjPtr = CDM_AnnotationsAt(objv[1], Integer);
  } else {
    ObjPtr = CDM_AnnotationsAt(Document, Integer);
  }
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_AnnotationsAtObjCmd */

/*
 * ---- tip_AnnotationsContaining:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that include the specified text position (as given
 *     by the parameter "Position") into their spans. 
 *
 *     This function will always create and return a new Annotation set object,
 *     even when the Document has no Annotations defined or when no Annotation
 *     is found that includes the specified text position into its span.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, an error message will be prompted.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Position:    Position in text.             (int)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_AnnotationsContaining (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). 
 *
 *     If an error occurs, an error message will be prompted.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_AnnotationsAt, tip_AnnotationsInRange, tip_AnnotationsMatchingRange,
 *     tip_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]]
 *   >
 *  SECTIONS:
 *   <Document Annotation
 *   >:&:
 */

/*
 * ---- tip_AnnotationsContaining:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that include the specified text positions (as given
 *     by the parameters "Position1" and "Position2") into their spans. 
 *
 *     This function will always create and return a new Annotation set object,
 *     even when the Document has no Annotations defined or when no Annotation
 *     is found that includes the specified text position into its span.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, an error message will be prompted.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Position1:   Position in text.             (int)
 *     Position2:   Position in text.             (int)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_AnnotationsContaining (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). 
 *
 *     If an error occurs, an error message will be prompted.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_AnnotationsAt, tip_AnnotationsInRange, tip_AnnotationsMatchingRange,
 *     tip_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]]
 *   >
 *  SECTIONS:
 *   <Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_AnnotationsContainingObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  long Long, Long2;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 3 && objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv,
                             "Document/AnnotationSet Position1 ?Position2?");
    return TCL_ERROR;}
  if (Tcl_GetLongFromObj(interp, objv[2], &Long) != TCL_OK) return TCL_ERROR;
  if (objc == 4) {
    if (Tcl_GetLongFromObj(interp, objv[3], &Long2) != TCL_OK) return TCL_ERROR;
  }
  
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) {
    if (objc == 3) {
      ObjPtr = CDM_AnnotationsContaining(objv[1], Long);
    } else {
      ObjPtr = CDM_AnnotationsContaining(objv[1], Long, Long2);
    }
  } else {
    if (objc == 3) {
      ObjPtr = CDM_AnnotationsContaining(Document, Long);
    } else {
      ObjPtr = CDM_AnnotationsContaining(Document, Long, Long2);
    }
  }
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_AnnotationsContainingObjCmd */

/*
 * ---- tip_AnnotationsInRange:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that have at least one span within the specified
 *     range (["Start", "End"]). 
 *
 *     This function will always create and return a new AnnotationSet object,
 *     even when the Document has no Annotations defined or when no Annotation
 *     is found that resides inside the specified range.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, an error message will be prmpted.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     Start:       The start of the desired range.  (int)
 *     End:         The end of the desired range.    (int)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_AnnotationsInRange (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). 
 *
 *     If an error occurs, an error message will be prmpted.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_AnnotationsAt, tip_AnnotationsContaining,
 *     tip_AnnotationsMatchingRange, tip_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]]
 *   >
 *  SECTIONS:
 *   <Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_AnnotationsInRangeObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  long start, end;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document/AnnotationSet start end");
    return TCL_ERROR;}
  if (Tcl_GetLongFromObj(interp, objv[2], &start) != TCL_OK) return TCL_ERROR;
  if (Tcl_GetLongFromObj(interp, objv[3], &end  ) != TCL_OK) return TCL_ERROR;
  
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) {
    ObjPtr = CDM_AnnotationsInRange(objv[1], start, end);
  } else {
    ObjPtr = CDM_AnnotationsInRange(Document, start, end);
  }
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_AnnotationsInRangeObjCmd */


/*
 * ---- tip_AnnotationsMatchingRange:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that have at least one span that 
 *     exactly matches the specified
 *     range (["Start", "End"]). 
 *
 *     This function will always create and return a new AnnotationSet object,
 *     even when the Document has no Annotations defined or when no Annotation
 *     is found that resides inside the specified range.
 *     (In this case an empty Annotation set will be returned.)
 *
 *     In case of an error, an error message will prompt.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     Start:       The start of the desired range.  (int)
 *     End:         The end of the desired range.    (int)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_AnnotationsMatchingRange (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet). 
 *
 *     If an error occurs, an error message will prompt.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_AnnotationsAt, tip_AnnotationsContaining,
 *     tip_AnnotationsInRange, tip_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]]
 *   >
 *  SECTIONS:
 *   <Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_AnnotationsMatchingRangeObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  long start, end;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document/AnnotationSet start end");
    return TCL_ERROR;}
  if (Tcl_GetLongFromObj(interp, objv[2], &start) != TCL_OK) return TCL_ERROR;
  if (Tcl_GetLongFromObj(interp, objv[3], &end  ) != TCL_OK) return TCL_ERROR;
  
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) {
    ObjPtr = CDM_AnnotationsMatchingRange(objv[1], start, end);
  } else {
    ObjPtr = CDM_AnnotationsMatchingRange(Document, start, end);
  }
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_AnnotationsMatchingRangeObjCmd */

/*
 * ---- tip_ByteSequenceInsertString:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the text contained in the specified Document,
 *     by inserting a string at the positions specified by the "pos"
 *     parameter. This position "pos" is the number of characters that have to
 *     be skipped before the string is inserted, where the first character of
 *     the Document text has a position equal to 0. If position is lower than
 *     zero, it is treaded as the [[i]]last[[/i]] character of the Document
 *     text.
 *
 *     For example, if I have the
 *     string "This is a test", the following code:[[br]][[code]][[center]]
 *            tip_ByteSequenceInsertString $Doc 10 "simple "
 *     [[/center]][[/code]][[br]] will modify the text of the Document to
 *     "This is a simple test".
 *
 *     The most important aspect of this function is that it will make the
 *     proper arrangements so as the Annotations contained inside the Document
 *     to be displaced by a proper amound of characters so as Annotations
 *     spanning text ranges after the insertion point to span the same text
 *     portions, after the string has been inserted.
 *
 *     This function will return the new Document text if the insertion was
 *     succesful.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     pos:         The character offset (starting from 0) where the string will be inserted. The string will be inserted at the character pointed by this value, displacing any existing text at this position after the inserted text. (long)
 *     string:      The string to be inserted.       (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_ByteSequenceInsertString (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the new Document text if the insertion was
 *     succesful.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_RawDataInsertString, tip_DisplaceAnnotations,
 *     tip_GetByteSequence, tip_SetByteSequence,
 *     CDM_ByteSequenceInsertString
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */

/*
 * ---- tip_RawDataInsertString:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the text contained in the specified Document,
 *     by inserting a string at the positions specified by the "pos"
 *     parameter. This position "pos" is the number of characters that have to
 *     be skipped before the string is inserted, where the first character of
 *     the Document text has a position equal to 0. If position is lower than
 *     zero, it is treaded as the [[i]]last[[/i]] character of the Document
 *     text.
 *
 *     For example, if I have the
 *     string "This is a test", the following code:[[br]][[code]][[center]]
 *            tip_RawDataInsertString $Doc 10 "simple "
 *     [[/center]][[/code]][[br]] will modify the text of the Document to
 *     "This is a simple test".
 *
 *     The most important aspect of this function is that it will make the
 *     proper arrangements so as the Annotations contained inside the Document
 *     to be displaced by a proper amound of characters so as Annotations
 *     spanning text ranges after the insertion point to span the same text
 *     portions, after the string has been inserted.
 *
 *     This function will return the new Document text if the insertion was
 *     succesful.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     pos:         The character offset (starting from 0) where the string will be inserted. The string will be inserted at the character pointed by this value, displacing any existing text at this position after the inserted text. (long)
 *     string:      The string to be inserted.       (char *)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_RawDataInsertString (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the new Document text if the insertion was
 *     succesful.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_ByteSequenceInsertString, tip_DisplaceAnnotations,
 *     tip_GetRawData, tip_SetRawData,
 *     CDM_RawDataInsertString
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_ByteSequenceInsertStringObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  long Long;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document position string");
    return TCL_ERROR;
  }

  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  if (Tcl_GetLongFromObj(interp, objv[2], &Long) != TCL_OK) return TCL_ERROR;
  ObjPtr = CDM_ByteSequenceInsertString(Document, Long,
                                              Tcl_GetString(objv[3]));
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_ByteSequenceInsertStringObjCmd */

/*
 * ---- tip_ByteSequenceReplace:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the text contained in the specified Document,
 *     by removing a range of consecutive characters from the Document Text,
 *     starting with the character whose index is the value of the parameter
 *     "first" and ending with the character whose index is the value of the 
 *     parameter "last". An index of 0 refers to the first character of the
 *     Document Text, while a negative value (i.e. -1) refers to the last
 *     character of the Document Text. If parameter "newstring" is present,
 *     then it is placed in the removed character range.
 *     If "first" is less than zero then it is treated as if it was zero, and
 *     if "last" is greater than the length of the Document Text
 *     then it is treated as if it was equal to the length of the Document Text.
 *     It is an error if "first" is greater than "last".
 *
 *     The most important aspect of this function is that it will make the
 *     proper arrangements so as the Annotations contained inside the Document
 *     to be displaced by a proper amound of characters so as Annotations
 *     spanning text ranges after the replacement point to span the same text
 *     portions, after the string has replaced the desired characters.
 *
 *     This function will return the new Document text (as an object of type
 *     CDM_ByteSequence) if the replacement was succesful.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     first:       The character offset (starting from 0) of the first character that will be replaced. (long)
 *     last:        The character offset (starting from 0) of the last character that will be replaced. (long) 
 *     newstring:   The string to be placed in the removed character range. (char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function can be also used in order to [[b]]delete[[/b]]
 *    characters from the text contained in the specified Document, if the
 *    value of the "newstring" parameter is an empty string (like "").
 *    In this case, all Annotations objects contained in the specified
 *    Document are searched: if any of them has a Span object in its Span
 *    set whoose offsets span portion of the charactes to be deleted,
 *    then this Span is eliminated from the Span set of the Annotation
 *    object. Note that no Annotation objects are deleted from the Document,
 *    although character deletion can result in Annotations with an empty
 *    Span set.
 *
 *    This function is equivelant to CDM_ByteSequenceReplace (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the new Document text.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_RawDataReplace,
 *     tip_ByteSequenceReplaceCharacters, tip_RawDataReplaceCharacters,
 *     tip_ByteSequenceInsertString, tip_RawDataInsertString,
 *     tip_DisplaceAnnotations,
 *     tip_GetByteSequence, tip_SetByteSequence,
 *     tip_GetRawData, tip_SetRawData,
 *     CDM_ByteSequenceReplace
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */

/*
 * ---- tip_RawDataReplace:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the text contained in the specified Document,
 *     by removing a range of consecutive characters from the Document Text,
 *     starting with the character whose index is the value of the parameter
 *     "first" and ending with the character whose index is the value of the 
 *     parameter "last". An index of 0 refers to the first character of the
 *     Document Text, while a negative value (i.e. -1) refers to the last
 *     character of the Document Text. If parameter "newstring" is present,
 *     then it is placed in the removed character range.
 *     If "first" is less than zero then it is treated as if it was zero, and
 *     if "last" is greater than the length of the Document Text
 *     then it is treated as if it was equal to the length of the Document Text.
 *     It is an error if "first" is greater than "last".
 *
 *     The most important aspect of this function is that it will make the
 *     proper arrangements so as the Annotations contained inside the Document
 *     to be displaced by a proper amound of characters so as Annotations
 *     spanning text ranges after the replacement point to span the same text
 *     portions, after the string has replaced the desired characters.
 *
 *     This function will return the new Document text (as an object of type
 *     CDM_RawData) if the replacement was succesful.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     first:       The character offset (starting from 0) of the first character that will be replaced. (long)
 *     last:        The character offset (starting from 0) of the last character that will be replaced. (long) 
 *     newstring:   The string to be placed in the removed character range. (char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function can be also used in order to [[b]]delete[[/b]]
 *    characters from the text contained in the specified Document, if the
 *    value of the "newstring" parameter is an empty string (like "").
 *    In this case, all Annotations objects contained in the specified
 *    Document are searched: if any of them has a Span object in its Span
 *    set whoose offsets span portion of the charactes to be deleted,
 *    then this Span is eliminated from the Span set of the Annotation
 *    object. Note that no Annotation objects are deleted from the Document,
 *    although character deletion can result in Annotations with an empty
 *    Span set.
 *
 *    This function is equivelant to CDM_RawDataReplace (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the new Document text.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_ByteSequenceReplace,
 *     tip_RawDataReplaceCharacters, tip_ByteSequenceReplaceCharacters,
 *     tip_RawDataInsertString, tip_ByteSequenceInsertString,
 *     tip_DisplaceAnnotations,
 *     tip_GetRawData, tip_SetRawData,
 *     tip_GetByteSequence, tip_SetByteSequence,
 *     CDM_RawDataReplace
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_ByteSequenceReplaceObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  char *CPtr;
  long Long, Long2;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 4 && objc != 5) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document first last ?newstring?");
    return TCL_ERROR;
  }

  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  if (Tcl_GetLongFromObj(interp, objv[2], &Long)  != TCL_OK) return TCL_ERROR;
  if (Tcl_GetLongFromObj(interp, objv[3], &Long2) != TCL_OK) return TCL_ERROR;
  if (objc == 5) CPtr = Tcl_GetString(objv[4]);
  else           CPtr = NULL;
  ObjPtr = CDM_ByteSequenceReplace(Document, Long, Long2, CPtr);
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_ByteSequenceReplaceObjCmd */

/*
 * ---- tip_ByteSequenceReplaceCharacters:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the text contained in the specified Document,
 *     by replacing the specified number of characters at the position
 *     specified by the "position" parameter by the provided string.
 *     This position "position" is the number of characters that have to
 *     be skipped before the characters are replaced by the string,
 *     where the first character of
 *     the Document text has a position equal to 0. If position is lower than
 *     zero, it is treaded as the [[i]]last[[/i]] character of the Document
 *     text. In the case the last character is specified as position, then
 *     the last "number_of_chars_to_replace" characters are replaced.
 *
 *     For example, if I have the
 *     string "This is a test", the following code:[[br]][[code]][[center]]
 *            CDM_ByteSequenceReplaceCharacters(Doc, 10, 4, "simple test");
 *     [[/center]][[/code]][[br]] will modify the text of the Document to
 *     "This is a simple test".
 *
 *     The most important aspect of this function is that it will make the
 *     proper arrangements so as the Annotations contained inside the Document
 *     to be displaced by a proper amound of characters so as Annotations
 *     spanning text ranges after the replacement point to span the same text
 *     portions, after the string has replaced the desired characters.
 *
 *     This function will return the new Document text if the replacement was
 *     succesful.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     pos:         The character offset (starting from 0) where the replacement will happen. The string will be inserted at the character pointed by this value, replacing the specified number of characters. (long)
 *     number_of_chars_to_replace:  An integer specifying the number of characters that will be replaced. (int) 
 *     string:      The string to be inserted.       (char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function can be also used in order to [[b]]delete[[/b]]
 *    characters from the text contained in the specified Document, if the
 *    value of the "string" parameter is an empty string (like "").
 *    In this case, all Annotations objects contained in the specified
 *    Document are searched: if any of them has a Span object in its Span
 *    set whoose offsets span portion of the charactes to be deleted,
 *    then this Span is eliminated from the Span set of the Annotation
 *    object. Note that no Annotation objects are deleted from the Document,
 *    although character deletion can result in Annotations with an empty
 *    Span set.
 *
 *    This function is equivelant to CDM_ByteSequenceReplaceCharacters (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the new Document text if the replacement was
 *     succesful.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_RawDataReplaceCharacters, tip_ByteSequenceInsertString,
 *     tip_DisplaceAnnotations,
 *     tip_GetByteSequence, tip_SetByteSequence,
 *     CDM_ByteSequenceReplaceCharacters
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */

/*
 * ---- tip_RawDataReplaceCharacters:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the text contained in the specified Document,
 *     by replacing the specified number of characters at the position
 *     specified by the "position" parameter by the provided string.
 *     This position "position" is the number of characters that have to
 *     be skipped before the characters are replaced by the string,
 *     where the first character of
 *     the Document text has a position equal to 0. If position is lower than
 *     zero, it is treaded as the [[i]]last[[/i]] character of the Document
 *     text. In the case the last character is specified as position, then
 *     the last "number_of_chars_to_replace" characters are replaced.
 *
 *     For example, if I have the
 *     string "This is a test", the following code:[[br]][[code]][[center]]
 *            CDM_ByteSequenceReplaceCharacters(Doc, 10, 4, "simple test");
 *     [[/center]][[/code]][[br]] will modify the text of the Document to
 *     "This is a simple test".
 *
 *     The most important aspect of this function is that it will make the
 *     proper arrangements so as the Annotations contained inside the Document
 *     to be displaced by a proper amound of characters so as Annotations
 *     spanning text ranges after the replacement point to span the same text
 *     portions, after the string has replaced the desired characters.
 *
 *     This function will return the new Document text if the replacement was
 *     succesful.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.               (CDM_Document)
 *     pos:         The character offset (starting from 0) where the replacement will happen. The string will be inserted at the character pointed by this value, replacing the specified number of characters. (long)
 *     number_of_chars_to_replace:  An integer specifying the number of characters that will be replaced. (int) 
 *     string:      The string to be inserted.       (char *)
 *   >
 *  NOTES:
 *   <English:
 *    This function can be also used in order to [[b]]delete[[/b]]
 *    characters from the text contained in the specified Document, if the
 *    value of the "string" parameter is an empty string (like "").
 *    In this case, all Annotations objects contained in the specified
 *    Document are searched: if any of them has a Span object in its Span
 *    set whoose offsets span portion of the charactes to be deleted,
 *    then this Span is eliminated from the Span set of the Annotation
 *    object. Note that no Annotation objects are deleted from the Document,
 *    although character deletion can result in Annotations with an empty
 *    Span set.
 *
 *    This function is equivelant to CDM_RawDataReplaceCharacters (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the new Document text if the replacement was
 *     succesful.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_ByteSequenceReplaceCharacters, tip_RawDataInsertString,
 *     tip_DisplaceAnnotations,
 *     tip_GetRawData, tip_SetRawData,
 *     CDM_RawDataReplaceCharacters
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_ByteSequenceReplaceCharactersObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  long Long;
  int Integer;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 5) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document "
                      "position number_of_chars_to_replace replacement_string");
    return TCL_ERROR;
  }

  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  if (Tcl_GetLongFromObj(interp, objv[2], &Long) != TCL_OK) return TCL_ERROR;
  if (Tcl_GetIntFromObj(interp, objv[3], &Integer) != TCL_OK) return TCL_ERROR;
  ObjPtr = CDM_ByteSequenceReplaceCharacters(Document, Long, Integer,
                                              Tcl_GetString(objv[4]));
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_ByteSequenceReplaceCharactersObjCmd */

/*
 * ---- tip_CreateVectors:
 *  DESCRIPTION:
 *   <English:
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                          (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_CreateVectors (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_CreateVectors
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_CreateVectorsObjCmd(CDM_TCL_STD_ARGS){
  const char *VectorAnnotation, *VectorAnnConstraints, *VectorFormat,
             *ClassAnnotation, *ClassAttribute,
             *ClassAttributeAbsenceValue, *ClassFail,
             *ClassPrefixStart, *ClassPrefixEnd, *StatusWindow;
  int ClassCanContain, ClassPrefixStartMode, ClassPrefixEndMode;
  CDM_CreateVectorsCellSpec *CellSpec = NULL;
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  int CellNumber, Contained, AnnotationText, Expand, i, index;  
  CDM_UPDATE_THREAD_INTERP;
  
  /* Calculate the number of cells from the arguments */
  if (objc < 15 || ((objc-15)%6)) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document VectorAnnotation "
                     "VectorAnnConstraints VectorFormat ClassAnnotation "
                     "ClassAttribute ClassAttributeAbsenceValue "
                     "ClassFail ClassCanContain "
                     "ClassPrefixStartMode ClassPrefixEndMode "
                     "ClassPrefixStart ClassPrefixEnd StatusWindow "
                     "?CellAnnotation CellAttribute CellFail Contained "
                     "AnnotationText Expand? "
                     "?CellAnnotation CellAttribute CellFail Contained "
                     "AnnotationText Expand? ...");
    return TCL_ERROR;}
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  CellNumber = ((objc-15)/6);
  /* Assign objects */
  VectorAnnotation           = Tcl_GetString(objv[2]);
  VectorAnnConstraints       = Tcl_GetString(objv[3]);
  VectorFormat               = Tcl_GetString(objv[4]);
  ClassAnnotation            = Tcl_GetString(objv[5]);
  ClassAttribute             = Tcl_GetString(objv[6]);
  ClassAttributeAbsenceValue = Tcl_GetString(objv[7]);
  ClassFail                  = Tcl_GetString(objv[8]);
  if (Tcl_GetIntFromObj(interp, objv[9], &ClassCanContain) != TCL_OK)
                                                              return TCL_ERROR;
  if (Tcl_GetIntFromObj(interp, objv[10], &ClassPrefixStartMode) != TCL_OK)
                                                              return TCL_ERROR;
  if (Tcl_GetIntFromObj(interp, objv[11], &ClassPrefixEndMode) != TCL_OK)
                                                              return TCL_ERROR;
  ClassPrefixStart           = Tcl_GetString(objv[12]);
  ClassPrefixEnd             = Tcl_GetString(objv[13]);
  StatusWindow               = Tcl_GetString(objv[14]);
  /* Allocate the CellSpec memory area... */
  if (CellNumber) {
    CellSpec = (CDM_CreateVectorsCellSpec *)
                Tcl_Alloc(CellNumber*sizeof(CDM_CreateVectorsCellSpec));
    if (CellSpec == NULL) return CDM_ReportError("out of memory");
    for (i = 15, index = 0; i < objc; i += 6, index++) {
      CellSpec[index].CellAnnotation = Tcl_GetString(objv[i]);
      CellSpec[index].CellAttribute  = Tcl_GetString(objv[i+1]);
      CellSpec[index].CellFail       = Tcl_GetString(objv[i+2]);
      if (Tcl_GetIntFromObj(interp, objv[i+3], &Contained) != TCL_OK) {
        Tcl_Free((char *) CellSpec);
        return TCL_ERROR;
      }
      if (Tcl_GetIntFromObj(interp, objv[i+4], &AnnotationText) != TCL_OK) {
        Tcl_Free((char *) CellSpec);
        return TCL_ERROR;
      }
      if (Tcl_GetIntFromObj(interp, objv[i+5], &Expand) != TCL_OK) {
        Tcl_Free((char *) CellSpec);
        return TCL_ERROR;
      }
      CellSpec[index].Contained      = Contained;
      CellSpec[index].AnnotationText = AnnotationText;
      CellSpec[index].Expand         = Expand;
      CellSpec[index].tmpObj         = NULL;
      CellSpec[index].CellFailObj    = objv[i+2];
      Tcl_IncrRefCount(CellSpec[index].CellFailObj);
      /*
       * // Uncomment to check what CDM_CreateVectors will get...
       * printf("Index = %d\n", index);
       * printf("\tAnnotation = \"%s\"\n",  CellSpec[index].CellAnnotation);
       * printf("\tAttribute = \"%s\"\n",   CellSpec[index].CellAttribute);
       * printf("\tFail Value = \"%s\"\n",  CellSpec[index].CellFail);
       * printf("\tContained = %d\n",       CellSpec[index].Contained);
       * printf("\tAnnotationText = %d\n",  CellSpec[index].AnnotationText);
       * printf("\tExpand = %d\n",          CellSpec[index].Expand);
      */
    }
  } /* if (CellNumber) */
  ObjPtr = CDM_CreateVectors(Document, VectorAnnotation, VectorAnnConstraints,
                             VectorFormat, ClassAnnotation, ClassAttribute,
                             ClassAttributeAbsenceValue,
                             ClassFail, ClassCanContain,
                             ClassPrefixStartMode, ClassPrefixEndMode,
                             ClassPrefixStart, ClassPrefixEnd, CellNumber,
                             CellSpec, StatusWindow);
  if (CellSpec) {
    for (i = 15, index = 0; i < objc; i += 6, index++) {
      Tcl_DecrRefCount(CellSpec[index].CellFailObj);
    }
    Tcl_Free((char *) CellSpec);
  }
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_CreateVectorsObjCmd */

/*
 * ---- tip_DeleteAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function accepts two arguments: a Document object and a UTF-8 string
 *     that defines the Annotations to be deleted. According to the value
 *     specified by the "Type" parameter, this function will delete
 *     Annotations or Annotation Attributes from the specified Document.
 *
 *     If the value of the "Type" parameter is an empty string, then
 *     ALL the Annotations that the specified Document currently has will
 *     be deleted
 *     from this Document object. The used Annotation Ids will be also reseted.
 *     This means that the first Annotation that will be added to this
 *     Document will take the value "0" as Annotation Id, if it does not
 *     already have an Id. If the
 *     deletion is done successfully, the total number of the deleted
 *     Annotations will be returned.
 *
 *     If Annotations having as type the value of the "Type" parameter exist,
 *     then all these Annotations will be deleted from the Document. If the
 *     deletion is done successfully, the total number of the deleted
 *     Annotations will be returned.
 *
 *     If all Annotations have been searched and no Annotation having as type
 *     the value of the "Type" parameter was found, the value of the "Type"
 *     parameter is examined whether it can be splitted in two words.  If this
 *     is the case (the value of the "Type" parameter contains at least one
 *     space character), then the value is splitted in two parts on the first
 *     space character that is found. The first word will be used as an
 *     Annotation type and the rest of the value (second word) 
 *     will be used as an Attribute
 *     name. All Annotations will be seached. If Annotations having as type
 *     the first word are found, they are examined whether they contain an
 *     Attribute named as the second word. If such an Attribute is found in an
 *     Annotation, it will be deleted from the Annotation. The total number of
 *     deleted Attributes will be returned as the return value of this
 *     function.
 *
 *     If this function fails to delete any Annotation or any Annotation
 *     Attribute, then the value "0" will be deleted. In case of an error, the
 *     value "-1" will be returned, and an error message describing the error
 *     will be left at the current Tcl interpreter.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                          (CDM_Document)
 *     Type:        The type of the Annotations to be deleted.  (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_DeleteAnnotations (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the number of the deleted Annotations or
 *     Annotation Attributes. If nothing was deleted, this function will
 *     return "0". In case of an error, he value "-1" will be returned, and an
 *     error message describing the error will be left at the current Tcl
 *     interpreter.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_FindMaxUsedAnnotationId
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_DeleteAnnotationsObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  int Integer;  
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc < 3 || objc > 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document Type ?Constraints?");
    return TCL_ERROR;}
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  if (objc == 3) {
    Integer = CDM_DeleteAnnotations(Document, Tcl_GetString(objv[2]));
  } else {
    Integer = CDM_DeleteAnnotations(Document, Tcl_GetString(objv[2]), objv[3]);
  }
  if (Integer < 0) return TCL_ERROR;
  ObjPtr = CDM_NewLongObj((long) Integer);
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_DeleteAnnotationsObjCmd */

/*
 * ---- tip_DisplaceAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function displases or "moves" the Annotations of the specified
 *     Document by "displacement" characters. This function will iterate over
 *     all Annotations contained in the Document. For each Annotation, all
 *     spans contained in the span set will be examined: for each offset in
 *     the span (either start or end), if it is equal or greater than the value
 *     provided by the "offset" parameter, then the value of the
 *     "displacement" parameter will be added.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:  A Document Object                               (CDM_Document)
 *     offset:    A character offset after which all offsets are modified (long)
 *     displacement: The number of characters that offsets will be moved  (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_DisplaceAnnotations (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int). As this function ignores any possible errors, the
 *     return value will always be TCL_OK.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_DisplaceAnnotation, CDM_DisplaceAnnotations
 *   >
 *  SECTIONS:
 *   <Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_DisplaceAnnotationsObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  CDM_AnnotationSet Set;
  long offset, displacement;
  int status;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document offset displacement");
    return TCL_ERROR;
  }
  Tcl_ResetResult(interp);
  Document = CDM_GetDocumentPointer(objv[1]);
  status = Tcl_GetLongFromObj(interp, objv[2], &offset);
  if (status != TCL_OK) return status;
  status = Tcl_GetLongFromObj(interp, objv[3], &displacement);
  if (status != TCL_OK) return status;
  if (Document == NULL) {
    Set = CDM_DisplaceAnnotations(objv[1], offset, displacement);
    if (Set == NULL) return TCL_ERROR;
    Tcl_SetObjResult(interp, Set);
  } else {
    status = CDM_DisplaceAnnotations(Document, offset, displacement);
    if (status != TCL_OK) return status;
  }
  return TCL_OK;
} /* CDM_DisplaceAnnotationsObjCmd */

/*
 * ---- tip_FindMaxUsedAnnotationId:
 *  DESCRIPTION:
 *   <English:
 *     This function will examine the Annotations stored in the specified
 *     Document to find the maximum used Annotation Id. This function will
 *     make the proper arrangments so as the next Annotation that will be
 *     added to this Document to receive as Id the following Id after the
 *     identified maximum Ids.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:  A Document Object                   (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_FindMaxUsedAnnotationId (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the Id that will be received by the next 
 *     Annotation that will be added to this Document.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_DeleteAnnotations, CDM_FindMaxUsedAnnotationId
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_FindMaxUsedAnnotationIdObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  long id;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document");
    return TCL_ERROR;
  }
  Tcl_ResetResult(interp);
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  id = CDM_FindMaxUsedAnnotationId(Document);
  if (id < 0) return TCL_ERROR;
  Tcl_SetObjResult(interp, CDM_NewLongObj(id));
  return TCL_OK;
} /* CDM_FindMaxUsedAnnotationIdObjCmd */

/*
 * ---- tip_FirstAnnotationContaining:
 *  DESCRIPTION:
 *   <English:
 *    This function will return a new Annotation object (of type
 *    CDM_Annotation) that will be the first Annotation (with Annotations 
 *    sorted by their Id) of the
 *    specified Document that includes the specified text position (as given
 *    by the parameter "position") or both positions (if parameter "position2"
 *    is also specified) into its spans.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document/AnnotationSet:  A Document or an Annotation set object. (CDM_Document/CDM_AnnotationSet)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_FirstAnnotationContaining (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the first Annotation (contained either in the
 *     Document where Annotations are sorted by their Ids or in the Annotation
 *     set where they are examined in the order defined by the set) whose
 *     specified position(s) is contained into its spans.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_DeleteAnnotations, CDM_FirstAnnotationContaining
 *   >
 *  SECTIONS:
 *   <Document, Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_FirstAnnotationContainingObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  long position1, position2;
  Tcl_Obj *ObjPtr;
  int status;
  CDM_UPDATE_THREAD_INTERP;

  if (objc < 3 || objc > 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document/AnnotationSet position "
                                      "?position2?"); 
    return TCL_ERROR;}
  status = Tcl_GetLongFromObj(interp, objv[2], &position1);
  if (status != TCL_OK) return status;
  Document = CDM_GetDocumentPointer(objv[1]);
  Tcl_ResetResult(interp);
  if (objc == 4) {
    status = Tcl_GetLongFromObj(interp, objv[3], &position2);
    if (status != TCL_OK) return status;
    if (Document != NULL) {
      ObjPtr = CDM_FirstAnnotationContaining(Document, position1, position2);
    } else {
      ObjPtr = CDM_FirstAnnotationContaining(objv[1], position1, position2);
    }
  } else {
    if (Document != NULL) {
      ObjPtr = CDM_FirstAnnotationContaining(Document, position1);
    } else {
      ObjPtr = CDM_FirstAnnotationContaining(objv[1], position1);
    }
  }
  if (ObjPtr == NULL) {
    /* We should report an error only if we have an error message */
    if (!CDM_ActiveInterpreterResultEmpty()) return TCL_ERROR;
  } else {
    Tcl_SetObjResult(interp, ObjPtr);
  }
  return TCL_OK;
} /* CDM_FirstAnnotationContainingObjCmd */

/*
 * ---- tip_GetAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Annotation object that has as Id the value
 *     specified by the "Id" parameter. If the requested Annotation does not
 *     exist, an error will be be prompted.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                     (CDM_Document)
 *     Id:          The Id of the desired Annotation.      (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetAnnotation (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation object (of type CDM_Annotation) if
 *     an Annotation having as Id the value of the "Id" parameter is found.
 *     In case that the specified Annotation does not exist, an error message prompts.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]], [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]]
 *   >
 *  SECTIONS:
 *   <Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_GetAnnotationObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  long Long;
  CDM_THREAD_SPECIFIC_DATA;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document/AnnotationSet Annotation-Id"); 
    return TCL_ERROR;}
  if (Tcl_GetLongFromObj(interp, objv[2], &Long) != TCL_OK) return TCL_ERROR;
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document != NULL) {
    ObjPtr = CDM_GetAnnotation(Document, Long);
  } else {
    ObjPtr = CDM_GetAnnotation(objv[1], Long);
  }
  if (ObjPtr == NULL) return TCL_ERROR;
  if (CDM_FullGateCompatibility) {
    /*
     * Gate wraps the annotation inside a list. Emulate that bug... 
     */
    Tcl_Obj *obj[1];
    obj[0] = ObjPtr;
    ObjPtr = Tcl_NewListObj(1, obj);
  }
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_GetAnnotationObjCmd */

/*
 * ---- tip_GetAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new list object
 *     (of type CDM_AnnotationSet) that will contain all the Annotations that
 *     the specified Document currently has.
 *
 *     Note that this function will always create and return a new Annotation
 *     set object, even when the Document has no Annotations defined.
 *     (In this case an empty Annotation set will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetAnnotations (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type 
 *     CDM_AnnotationSet). 
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_SelectAnnotations, tip_SelectAnnotationsSorted, CDM_GetAnnotations
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_GetAnnotationsObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;

  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collections/Document");
    return TCL_ERROR;
  }
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  ObjPtr = CDM_GetAnnotations(Document);
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_GetAnnotationsObjCmd */

/*
 * ---- tip_GetRawData:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Text of the specified Document. It will
 *     return an object (of type CDM_ByteSequence) that will contain the text
 *     of the Document.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:  A Document object. (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetRawData (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a Text object (of type CDM_ByteSequence) that
 *     will contain the text (or byte sequence or raw data) of the specified
 *     Document.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetByteSequence, CDM_GetRawData
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
 
/*
 * ---- tip_GetByteSequence:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Text of the specified Document. It will
 *     return an object (of type CDM_ByteSequence) that will contain the text
 *     of the Document.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:  A Document object. (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetByteSequence (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return a Text object (of type CDM_ByteSequence) that
 *     will contain the text (or byte sequence or raw data) of the specified
 *     Document.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetRawData, CDM_GetByteSequence
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_GetByteSequenceObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document"); return TCL_ERROR;}
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  ObjPtr = CDM_GetByteSequence(Document);
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_GetByteSequenceObjCmd */

/*
 * ---- tip_GetExternalId:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the external Id (XID) of the specified
 *     Document. The external Id is the (absolute) path of the initial file
 *     that contained the text which was used in order to create the specified
 *     Document. The returned value will be encoded using the UTF-8 encoding
 *     (thus enabling the existance of non Latin characters in the value).
 *
 *     The return value will be of type string.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:    A Document object.            (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetExternalId (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a UTF-8 string that represents the external Id of
 *     the specified Document.
 *
 *     If an error occurs, an error message will be prompted.
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_GetExternalIdObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  char *CPtr;

  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document"); return TCL_ERROR;}
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  CPtr = CDM_GetExternalId(Document);
  if (CPtr == NULL) return TCL_ERROR;
  Tcl_SetResult(interp, CPtr, TCL_VOLATILE);
  return TCL_OK;
} /* CDM_GetExternalIdObjCmd */


/*
 * ---- tip_GetFirstAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will start an iteration over the Annotations of the
 *     specified type that the specified Document has. This function will
 *     return the first Annotation object that has as type the value
 *     specified by the "Type" parameter. If the requested Annotation does not
 *     exist, an error will be prmpted. Subsequent calls to
 *     tip_GetNextAnnotation will return the rest of the Annotations of the
 *     same type, sorted according to their Id.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                     (CDM_Document)
 *     Type:        The type of the desired Annotations.   (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetFirstAnnotation (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation object (of type CDM_Annotation) if
 *     an Annotation having as type the value of the "Type" parameter is found.
 *     In case that the specified Annotation does not exist, an error message will be prompted.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetNextAnnotation, tip_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]], [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]]
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_GetFirstAnnotationObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document Annotation-Type");
    return TCL_ERROR;}
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  ObjPtr = CDM_GetFirstAnnotation(Document, Tcl_GetString(objv[2]));
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_GetAnnotationObjCmd */


/*
 * ---- tip_GetNextAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the next Annotation of the iteration started
 *     with a call to tip_GetFirstAnnotation, or of the Annotation returned by
 *     a previous call to tip_GetNextAnnotation. If no more Annotations of the
 *     same type (which was specified through a call to tip_GetFirstAnnotation)
 *     exist in the specified Document, an error message will be prompted.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                     (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetNextAnnotation (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation object (of type CDM_Annotation) if
 *     an Annotation having as type the value specified by a prevous call to
 *     tip_GetFirstAnnotation exists in the specified Document.
 *     In case that such an Annotation does not exist, An error message will be prmpted.
 *     If that happens, this function cannot be called unless
 *     tip_GetFirstDocument is called again, in order to initialise a new
 *     iteration over Annotations of a specific type.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetFirstAnnotation, tip_Free,
 *     [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_IncrRefCount[[/A]], [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]]
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_GetNextAnnotationObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;

  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document");
    return TCL_ERROR;}
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  ObjPtr = CDM_GetNextAnnotation(Document);
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_GetAnnotationObjCmd */


/*
 * ---- tip_GetParent:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the parent of the specified
 *     Document. The parent of a Document is the Collection object that holds
 *     the specified Document. The return value will be the parent Collection
 *     object (of type CDM_Collection).
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetParent (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return the Collection object that contains the
 *     specified Document.
 *
 *     In case of an error message will be prompted.
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_GetParentObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  CDM_Collection Parent;
  
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document"); return TCL_ERROR;}
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;  
  Parent = CDM_GetParent(Document);
  if (Parent == NULL) return TCL_ERROR;
  Tcl_SetResult(interp, (char *) Tcl_GetCommandName(interp, Parent->Tcl_Cmd),
                TCL_VOLATILE);
  return TCL_OK;
} /* CDM_GetParentObjCmd */

/*
 * ---- tip_NextAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will return a new Annotation set object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that their spans ranges are after
 *     the specified text position (as given by the parameter "Position").
 *
 *     This function will always create and return a new Annotation set object,
 *     even when the Document has no Annotations defined or no Annotations
 *     exist that all their spans are beneath the specified text position.
 *     In this case an empty Annotation set will be returned.
 *
 *     In case of an error, an error message will be prompted.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Position:    Position in text.             (int)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to tip_NextAnnotations (Tcl API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns an Annotation set object (of type
 *     CDM_AnnotationSet).
 *
 *     If an error occurs, an error message will be prompted.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]]
 *   >
 *  SECTIONS:
 *   <Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_NextAnnotationsObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  long Integer;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document/AnnotationSet Position");
    return TCL_ERROR;
  }
  if (Tcl_GetLongFromObj(interp, objv[2], &Integer) != TCL_OK) return TCL_ERROR;
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) {
    ObjPtr = CDM_NextAnnotations(objv[1], Integer);
  } else {
    ObjPtr = CDM_NextAnnotations(Document, Integer);
  }
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_NextAnnotationsObjCmd */

/*
 * ---- tip_AddSpan:
 *  DESCRIPTION:
 *   <English:
 *    This function will add the provided Span object  in the 
 *    Span Set object held internally by the Annotation/Span Set object.
 *    The returned value of this function will always be a new Annotation
 *    object with the provided Span object added.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Ann:    An Annotation object.                            (CDM_Annotation)
 *     span:   The Span object that should be added.            (CDM_Span)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_AddAnnotationSpan/CDM_AddSpan (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a new Annotation object (of type CDM_Annotation)
 *    with the provided span object added in the set of spans held by the
 *    Annotation.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_AddAnnotationSpan, CDM_AddSpan
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */

CDM_CALLBACK int CDM_AddSpanObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Annotation Ann;
  
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Annotation/SpanSet span");
    return TCL_ERROR;
  }
  Tcl_ResetResult(interp);
  /* Has the user provided an Annotation? */
  if (CDM_ValidAnnotation(objv[1])) {
    Ann = CDM_AddAnnotationSpan(objv[1], objv[2]);
  } else {
    Ann = CDM_AddSpan(objv[1], objv[2]);
  }
  if (Ann == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Ann);
  return TCL_OK;
} /* CDM_AddSpanObjCmd */

/*
 * ---- tip_RemoveSpan:
 *  DESCRIPTION:
 *   <English:
 *    This function will remove the Span having as start and end offsets
 *    the values specified through the "start" and "end" parameters
 *    from the specified Annotation, if such a Span object exists.
 *    The returned value of this function will always be a new Annotation
 *    object if a proper Span object was found and removed.
 *    If a Span object with the
 *    proper start and end offsets was not found, the empty string will be
 *    returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Ann:    An Annotation object.                            (CDM_Annotation)
 *     start:  The start byte offset of the Span to be deleted. (long)
 *     end:    The end byte offset of the Span to be deleted.   (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_RemoveSpan (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a new Annotation object (of type CDM_Annotation) if
 *    a Span object with the same offsets as the provided ones is found and
 *    removed from the provided Annotation.
 *    In the case that a proper Span object is not found, the empty string ({})
 *    is returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    CDM_RemoveSpan
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */

CDM_CALLBACK int CDM_RemoveSpanObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Annotation Ann;
  long start, end;
  
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Annotation start end");
    return TCL_ERROR;
  }
  Tcl_ResetResult(interp);
  if (Tcl_GetLongFromObj(interp, objv[2], &start) != TCL_OK) return TCL_ERROR;
  if (Tcl_GetLongFromObj(interp, objv[3], &end)   != TCL_OK) return TCL_ERROR;
  Ann = CDM_RemoveSpan(objv[1], start, end);
  if (Ann == NULL) return TCL_ERROR;
  if (Ann != objv[1]) {
    Tcl_SetObjResult(interp, Ann);
  }
  return TCL_OK;
} /* CDM_RemoveSpanObjCmd */

/*
 * ---- tip_RemoveAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will remove the Annotation object that has as Id the value
 *     specified by the "Id" parameter. If the requested Annotation does not
 *     exist, an error will be peompted.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:  A Document object.                         (CDM_Document)
 *      Id:        The Id of the Annotation to be deleted.    (int)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_RemoveAnnotation (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a standart Tcl completion
 *     code (of type int) with one of the values TCL_OK and TCL_ERROR. If the
 *     requested Annotation exists and is successfully removed, then TCL_OK
 *     will be returned. If the requested Annotation does not exist, or in
 *     case of an error, the return value will be
 *     TCL_ERROR and an error message describing the error will be left at the
 *     current Tcl interpreter (CDM_Interp).
 *   >
 *  SECTIONS:
 *   <Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_RemoveAnnotationObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Document Document;
  Tcl_Obj *ResObj;
  long Long;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document/AnnotationSet Annotation-Id");
    return TCL_ERROR;
  }
  Tcl_ResetResult(interp);
  if (Tcl_GetLongFromObj(interp, objv[2], &Long) != TCL_OK) return TCL_ERROR;
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) {
    ResObj = CDM_RemoveAnnotation(objv[1],  Long);
    if (ResObj == NULL) return TCL_ERROR;
    Tcl_SetObjResult(interp, ResObj);
  } else {
    return   CDM_RemoveAnnotation(Document, Long);
  }
  return TCL_OK;
} /* CDM_RemoveAnnotationObjCmd */


/*
 * ---- tip_SelectAnnotationsSorted:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new list object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that their type is the same as the value of the
 *     "Type" parameter. The Annotations will be sorted according to their
 *     first span range, in ascending order.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                           (CDM_Document)
 *     Type:        The type of the Annotations to be selected.  (string)
 *     ?Constraints?: A boolean expression. (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_SelectAnnotationsSorted (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return an AnnotationSet object, that will
 *     contain all the Annotations of the specified Document that have as type
 *     the value of the "Type" parameter. The Annotations in that object will
 *     be sorted according to their first span in ascending order.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_SelectAnnotations, CDM_SelectAnnotationsSorted
 *   >
 *  SECTIONS:
 *   <Document Annotation
 *   >:&:
 */
 
/*
 * ---- tip_SelectAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new object (of type
 *     CDM_AnnotationSet) that will contain all the Annotations of the
 *     specified Document that their type is the same as the value of the
 *     "Type" parameter. This function will always create and return a new
 *     object, even when no Annotations of the requested type are found or
 *     when the Document has no Annotations defined. In this case an empty
 *     object will be returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Type:        The type of the Annotations to be selected.  (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_SelectAnnotations (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return an AnnotationSet object, that will
 *     contain all the Annotations of the specified Document that have as type
 *     the value of the "Type" parameter.
 *
 *     In case of an error, an error message will be prompted.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_Free, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]]
 *   >
 *  SECTIONS:
 *   <Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_SelectAnnotationsObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  Tcl_Obj *ObjPtr;
  CDM_THREAD_SPECIFIC_DATA;

  if (CDM_FullGateCompatibility && objc != 4) {
    Tcl_WrongNumArgs(interp,1,objv,"Document/AnnotationSet Type Constraints"); 
    return TCL_ERROR;}
  if (objc < 3 || objc > 4) {
    Tcl_WrongNumArgs(interp,1,objv,"Document/AnnotationSet Type ?Constraints?"); 
    return TCL_ERROR;}
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document != NULL) {
    if (objc == 3) {
      ObjPtr = CDM_SelectAnnotations(Document, Tcl_GetString(objv[2]));
    } else {
      ObjPtr = CDM_SelectAnnotations(Document, Tcl_GetString(objv[2]),
                                               Tcl_GetString(objv[3]));
    }
  } else {
    /* Perhaps an Annotation Set... */
    if (CDM_FullGateCompatibility) return TCL_ERROR;
    if (objc == 3) {
      ObjPtr = CDM_SelectAnnotations(objv[1], Tcl_GetString(objv[2]));
    } else {
      ObjPtr = CDM_SelectAnnotations(objv[1], Tcl_GetString(objv[2]),
                                              Tcl_GetString(objv[3]));
    }
  }
  if (ObjPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_SelectAnnotationsObjCmd */


/*
 * ---- tip_SetRawData:
 *  DESCRIPTION:
 *   <English:
 *     This function will change the text of a Document object to the text
 *     specified by the "Text" parameter. The "Text" parameter must be an
 *     object of type CDM_ByteSequence.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Text:        The desired new text object.  (CDM_ByteSequence)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_SetRawData (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     None.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_SetByteSequence, CDM_SetRawData 
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
 
/*
 * ---- tip_SetByteSequence:
 *  DESCRIPTION:
 *   <English:
 *     This function will change the text of a Document object to the text
 *     specified by the "Text" parameter. The "Text" parameter must be an
 *     object of type CDM_ByteSequence.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Text:        The desired new text object.  (CDM_ByteSequence)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_SetByteSequence (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     None.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_SetRawData, CDM_SetByteSequence
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_SetByteSequenceObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document Text"); return TCL_ERROR;}
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  return CDM_SetByteSequence(Document, objv[2]);
} /* CDM_SetByteSequenceObjCmd */


/*
 * ---- tip_SetExternalId:
 *  DESCRIPTION:
 *   <English:
 *     This function will modify the external Id (XID) of the specified
 *     Document. The external Id is the (absolute) path of the initial file
 *     that contained the text which was used in order to create the specified
 *     Document. The value is expected to be encoded using the UTF-8 encoding
 *     (thus enabling the existance of non Latin characters in the value). The
 *     caller can use the function tip_ExternalToUft in order to convert a
 *     string from any supported encoding to UTF.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.                            (CDM_Document)
 *     ExternalId:  The desired new external Id of the Document.  (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_SetEncoding (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     If the external Id of the specified Document is successfully updated,
 *     TCL_OK will be returned. In case of an error, the return value will be
 *     TCL_ERROR and an error message describing the error will be left at the
 *     current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetExternalId, tip_SetExternalId
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_SetExternalIdObjCmd(CDM_TCL_STD_ARGS){
  CDM_Document Document;
  char *CPtr;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document External-Id"); 
    return TCL_ERROR;
  }
  Document = CDM_GetDocumentPointer(objv[1]);
  if (Document == NULL) return TCL_ERROR;
  CPtr = CDM_SetExternalId(Document, Tcl_GetString(objv[2]));
  if (CPtr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Tcl_NewStringObj(CPtr, -1));
  return TCL_OK;
} /* CDM_SetExternalIdObjCmd */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     A T T R I B U T E S            S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/
 
/*
 * ---- tip_CreateAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return an Attribute object
 *     (of type CDM_Attribute). In order to reduce memory requirements, this
 *     function will try to locate if an Attribute with exactly the same
 *     characteristics (same Name and Attribute value of the same type and
 *     value) already exists. In such a case, this (possibly shared) Attribute
 *     object will be returned. Else, a new Attribute object will be created
 *     and returned.
 *
 *     Special attention should be given to the Attribute value object
 *     (specified by the "Value" parameter) so as not to leak memory. This
 *     function internally will always increase and then decrease the
 *     reference count of this object, meaning that this object may be freed
 *     if it has a reference count of 0. If the caller wants to preserve a
 *     reference to this object, so as to free it at a later stage through
 *     tip_Free or [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], Tcl_IncrRefCount should be called be
 *     called before tip_CreateAttribute is called.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Name:   The name of the Attribute.      (string)
 *     Value:  The value of the Attribute.     (CDM_AttributeValue)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_CreateAttribute (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return an Attribute object (of type of
 *     CDM_Attribute). This object may be an old, already in use (shared)
 *     object or a newly created object, according to the specified parameters.
 *
 *     In case of an error, NULL will be returned and an error message
 *     describing the error will be left at the current Tcl interpreter
 *     (CDM_Interp). In such a case, the object described by the "Value"
 *     parameter won't be altered in any way, leaving the responsibility to
 *     free it to the caller.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_CreateAttribute
 *   >
 *  SECTIONS:
 *   <Attribute
 *   >:&:
 */
CDM_CALLBACK int CDM_CreateAttributeObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Attribute Attr;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Name Value");
    return TCL_ERROR;
  }
  Attr = CDM_CreateAttribute(Tcl_GetString(objv[1]), objv[2]);
  if (Attr == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Attr);
  return TCL_OK;
} /* CDM_CreateAttributeObjCmd */


/*
 * ---- tip_CreateAttributeValue:
 *  DESCRIPTION:
 *   <English:
 *    Creates a new Attribute Value...
 *    Use this function when the type is "CDM_STING_SET" to reuse the given
 *    object!
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Type:   Attribute's Type      (int)
 *      Value:  Attribute's Value     (Tcl_Obj *)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >
 *  SECTIONS:
 *   <Attribute
 *   >:&:
 */
CDM_CALLBACK int CDM_CreateAttributeValueObjCmd(CDM_TCL_STD_ARGS) {
  CDM_AttributeValue AttrVal;
  int type;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Type Value");
    return TCL_ERROR;
  }
  type = CDM_GetAttributeTypeFromObject(objv[1]);
  if (type < CDM_STRING) return TCL_ERROR;

  AttrVal = CDM_CreateAttributeValue(type, objv[2]);
  if (AttrVal == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, AttrVal);
  return TCL_OK;
} /* CDM_CreateAttributeValueObjCmd */

/*
 * ---- CDM_CreateAttributeSet:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a new Attribute set object
 *     that will contain a newly created Attribute
 *     object, created according to the values specified as parameters. Note
 *     that this function is equivelant to creating an empty Attribute set
 *     object with tip_CreateAttributeSet, creating a new Attribute
 *     object with tip_CreateAttribute
 *     and adding this Attribute object to the newly created Annotation set.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Name:   Attribute's Name
 *      Value:  Attribute's Value
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_CreateAttributeSet (C API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_CreateAttribute, CDM_CreateAttributeSet
 *   >:&:
 */
CDM_CALLBACK int CDM_CreateAttributeSetObjCmd(CDM_TCL_STD_ARGS) {
  CDM_AttributeSet AttrVal;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 1 && objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "?name value?");
    return TCL_ERROR;
  }
  if (objc == 1) {
    AttrVal = CDM_CreateAttributeSet();
  } else {
    AttrVal = CDM_CreateAttributeSet(Tcl_GetString(objv[1]), objv[2]);
  }
  if (AttrVal == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, AttrVal);
  return TCL_OK;
} /* CDM_CreateAttributeSetObjCmd */

/*
 * ---- tip_GetName:
 *  DESCRIPTION:
 *   <English:
 *    Returns Attribute's Name...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Attribute:   Attribute      (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >
 *  SECTIONS:
 *   <Collection Attribute
 *   >:&:
 */
CDM_CALLBACK int CDM_GetNameObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Col;
  char *Name;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection/Attribute");
    return TCL_ERROR;
  }
  // objv[1] can hold an attribute or a Collection...
  Col = CDM_GetCollectionPointer(objv[1]);
  if (Col != NULL) {
    Name = CDM_GetName(Col);
  } else {
    Name = CDM_GetName(objv[1]);
  }
  if (Name == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, CDM_NewStringObj(Name));
  return TCL_OK;
} /* CDM_GetNameObjCmd */


/*
 * ---- tip_GetValue:
 *  DESCRIPTION:
 *   <English:
 *    Returns Attribute's Value...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Attribute:   Attribute      (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >
 *  SECTIONS:
 *   <Attribute
 *   >:&:
 */
CDM_CALLBACK int CDM_GetValueObjCmd(CDM_TCL_STD_ARGS) {
  CDM_AttributeValue Value;

  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Attribute");
    return TCL_ERROR;
  }
  /*
   * Gate returns only the Attribute's value. Logos on the other side returns
   * also the value's type. Also, if Gate Full compatibilty mode is enabled,
   * emulate the bug of not checking the parameter :-)
   */
  if (CDM_EmulateGateGetValue || CDM_FullGateCompatibility) {
    Tcl_Obj **elements, **ValueElements;
    int result, items;

    result = 
      Tcl_ListObjGetElements(CDM_Interp, objv[1], &items, &elements);
    if (result != TCL_OK || items != 2) {
      if (CDM_FullGateCompatibility) {
        Tcl_SetObjResult(interp, &CDM_Obj_NULL);
        return TCL_OK;
      } else {
        snprintf(CDM_Temp, MAX_TEMP_STRING, "Mallformed Attribute \"%s\"",
                                             Tcl_GetString(objv[1]));
        return CDM_ReportError(CDM_Temp);
      }
    }
    result = 
      Tcl_ListObjGetElements(CDM_Interp, elements[1], &items, &ValueElements);
    if (result != TCL_OK || items != 2) {
      if (CDM_FullGateCompatibility) {
        Tcl_SetObjResult(interp, &CDM_Obj_NULL);
        return TCL_OK;
      } else {
        snprintf(CDM_Temp, MAX_TEMP_STRING, "Mallformed Attribute Value\"%s\"",
                                             Tcl_GetString(elements[1]));
        return CDM_ReportError(CDM_Temp);
      }
    }
    Value = ValueElements[1];
  } else {
    Value = CDM_GetValue(objv[1]);
    if (Value == NULL)  /* return TCL_ERROR; GATE COMPATIBILITY!*/
      return TCL_ERROR; /* Value = &CDM_Obj_NULL; */
  }
  Tcl_SetObjResult(interp, Value);
  return TCL_OK;
} /* CDM_GetValueObjCmd */


/*
 * ---- tip_GetValueType:
 *  DESCRIPTION:
 *   <English:
 *    Returns Attribute's Type...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Attribute:   Attribute      (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >
 *  SECTIONS:
 *   <Attribute
 *   >:&:
 */
CDM_CALLBACK int CDM_GetValueTypeObjCmd(CDM_TCL_STD_ARGS) {
  Tcl_Obj *Type;

  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Attribute");
    return TCL_ERROR;
  }
  Type = CDM_GetValue(objv[1]);
  if (Type == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, CDM_Nth(Type, 0));
  return TCL_OK;
} /* CDM_GetValueTypeObjCmd */



/*
 * ---- tip_GetValueValue:
 *  DESCRIPTION:
 *   <English:
 *    Returns Attribute's Value (without type)...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Attribute:   Attribute      (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >
 *  SECTIONS:
 *   <Attribute
 *   >:&:
 */
CDM_CALLBACK int CDM_GetValueValueObjCmd(CDM_TCL_STD_ARGS) {
  CDM_AttributeValue Value;

  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Attribute");
    return TCL_ERROR;
  }
  Value = CDM_GetValueValue(objv[1]);
  if (Value == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Value);
  return TCL_OK;
} /* CDM_GetValueValueObjCmd */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     A N N O T A T I O N S          S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 * ---- tip_AnnotationContainsAttributeMatchingValue:
 *  DESCRIPTION:
 *   <English:
 *    This function will check the provided Annotation (specified through the
 *    "Annotation" parameter) to locate an Attribute that has the specified
 *    name (through the "AttributeName" parameter). If such an Attribute is
 *    not found in the Annotation, 0 will be returned. If a proper Attribute
 *    is found, its value will be matched against the provided (though the
 *    "ValuePattern" parameter) pattern: if a match is found 1 will be
 *    returned, else this function will return 0.
 *
 *    The pattern specified through the "ValuePattern" parameter must follow
 *    the same rules as the patterns used with the 
 *    %%%quot;[[A HREF="../TclTk/TclCmd/string.htm#M35"
 *    ]]string match[[/A]]%%%quot; Tcl command.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Ann:           The Annotation to be examined             (CDM_Annotation)
 *     AttributeName: The name of the desired Attribute            (string)
 *     ValuePattern:  The pattern the Attribute Value should match (string)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_AnnotationContainsAttributeMatchingValue (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return 1 (true) if an Attribute with the provided
 *    name exists in the specified Annotation, and its value matches the
 *    provided pattern. Else, 0 (false) will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_AnnotationContainsAttributeMatchingValues,
 *    CDM_AnnotationContainsAttributeMatchingValue,
 *    [[A HREF="../TclTk/TclCmd/string.htm#M35"]]string match[[/A]]
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_AnnotationContainsAttributeMatchingValueObjCmd(CDM_TCL_STD_ARGS) {
  int status;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Annotation AttributeName pattern");
    return TCL_ERROR;
  }

  status = CDM_AnnotationContainsAttributeMatchingValue(objv[1],
               Tcl_GetString(objv[2]), Tcl_GetString(objv[3]));
  if (status == -1) return TCL_ERROR;
  Tcl_SetObjResult(interp, Tcl_NewIntObj(status));
  return TCL_OK;
} /* CDM_AnnotationContainsAttributeMatchingValueObjCmd */

/*
 * ---- tip_AnnotationContainsAttributeMatchingValues:
 *  DESCRIPTION:
 *   <English:
 *    This function will check the provided Annotation (specified through the
 *    "Annotation" parameter) to locate an Attribute that has the specified
 *    name (through the "AttributeName" parameter). If such an Attribute is
 *    not found in the Annotation, 0 will be returned. If a proper Attribute
 *    is found, its value will be matched against the set of provided
 *    (though the "ValuePatternsObj" parameter) patterns:
 *    if a match is found with any of the patterns 1 will be
 *    returned, else this function will return 0.
 *
 *    The patterns specified through the "ValuePatternsObj" parameter must
 *    follow the same rules as the patterns used with the 
 *    %%%quot;[[A HREF="../TclTk/TclCmd/string.htm#M35"
 *    ]]string match[[/A]]%%%quot; Tcl command.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Ann:           The Annotation to be examined             (CDM_Annotation)
 *     AttributeName: The name of the desired Attribute               (string)
 *     ValuePatternsObj: The pattern the Attribute Value should match (list)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_AnnotationContainsAttributeMatchingValues (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return 1 (true) if an Attribute with the provided
 *    name exists in the specified Annotation, and its value matches any of the
 *    provided patterns. Else, 0 (false) will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_AnnotationContainsAttributeMatchingValue,
 *    CDM_AnnotationContainsAttributeMatchingValues,
 *    [[A HREF="../TclTk/TclCmd/string.htm#M35"]]string match[[/A]] 
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_AnnotationContainsAttributeMatchingValuesObjCmd(CDM_TCL_STD_ARGS) {
  int status;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Annotation AttributeName patternList");
    return TCL_ERROR;
  }

  status = CDM_AnnotationContainsAttributeMatchingValues(objv[1],
               Tcl_GetString(objv[2]), objv[3]);
  if (status == -1) return TCL_ERROR;
  Tcl_SetObjResult(interp, Tcl_NewIntObj(status));
  return TCL_OK;
} /* CDM_AnnotationContainsAttributeMatchingValuesObjCmd */

/*
 * ---- tip_AddAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will add the given Annotation (specified by the "Ann"
 *     parameter) into the Document specified by the "Document" parameter.
 *     If the Annotation does not already have an Id, a new unique Id will be
 *     assigned to that Annotation and the Annotation will be added to the
 *     current Annotation set of the Document.
 *
 *     If the Annotation already has an Id, then this Id will left intact.
 *     If an Annotation with the same Id already exists in the current
 *     Annotation set, then the new Annotation will replace the old one. If
 *     an Annotation with the same Id does not exist, the given Annotation will
 *     be appended to the existing Annotation set.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:    A Document object.            (CDM_Document)
 *     Ann:         The Annotation to be added.   (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_AddAnnotation (C++ API).
 *
 *     [[FONT COLOR=red]][[B]][[U]]Warning:[[/U]][[/B]][[/FONT]]
 *     [[BR]]This function modifies the given Annotation, if 
 *     MODIFY_SHARED_OBJECTS is defined during compilation of CDM!
 *     (Note that it is not totally safe to enable this option!
 *      This option is disabled by default.)
 *     We decided to offer this option (to modify the Annotation instead of 
 *     creating a copy and modify the copy), in order to be more memory
 *     efficient, as annotation numbers tend to get really large...
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns the Id of the Annotation that was added to the
 *     specified Document object. Whether the Annotation had or had not an Id
 *     makes no difference to the returned value. Note that after using this
 *     function, CDM will keep a reference to the given Annotation. This will
 *     transform the given Annotation into a [[FONT COLOR=red]]shared[[/FONT]]
 *     object, that cannot be manipulated any more by the caller. Always
 *     remember that an attempt to modify a shared object will abort the
 *     execution of the whole platform. If the caller wants to modify a shared
 *     object, the following steps must be followed:
 *     [[LI]] A check should be made in order to find if an object is shared
 *     or not, by using the Tcl function Tcl_IsShared.[[/LI]]
 *     [[LI]] If the object is not shared (i.e. Tcl_IsShared returns an
 *     integer value < 2), the object can be safely modified.[[/LI]]
 *     [[LI]] If the object is shared, a new copy of the object must be
 *     created by using the Tcl function Tcl_DuplicateObj. The new object will
 *     be an exact copy of the original one but it will not be shared. The
 *     returned object can be safely modified.[[/LI]]
 *
 *     In case of an error the return value will be -1 and
 *     an error message describing the error will be left at the current Tcl
 *     interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *     Tcl_IsShared, Tcl_DuplicateObj, tip_AddAnnotation
 *   >
 *  SECTIONS:
 *   <Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_AddAnnotationObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Document Doc;
  Tcl_Obj *ObjPtr;
  long id;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document/AnnotationSet Annotation");
    return TCL_ERROR;
  }

  Doc = CDM_GetDocumentPointer(objv[1]);
  if (Doc == NULL) {
    ObjPtr = CDM_AddAnnotation(objv[1], objv[2]);
    if (ObjPtr == NULL) return TCL_ERROR;
  } else {
    id = CDM_AddAnnotation(Doc, objv[2]);
    if (id < 0) return TCL_ERROR;
    ObjPtr = CDM_NewLongObj(id);
    if (ObjPtr == NULL) 
      CDM_Panic("CDM_AddAnnotation: CDM_NewLongObj returns NULL "
              "- out of memory?\n");
  }
  Tcl_SetObjResult(interp, ObjPtr);
  return TCL_OK;
} /* CDM_AddAnnotationObjCmd */

/*
 * ---- tip_AnnotationContainsPosition:
 *  DESCRIPTION:
 *   <English:
 *    This function will check if the provided Annotation (specified through the
 *    "Annotation" parameter) contains in any of its spans the specified
 *    position.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Ann:    The Annotation to be examined           (CDM_Annotation)
 *    Position: The position to me checked            (long)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to CDM_AnnotationContainsPosition (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return 1 (true) if the Annotation contains the
 *    specified position in any of its spans. Else, 0 (false)
 *    will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_AnnotationMatchesRange, CDM_AnnotationContainsPosition
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_AnnotationContainsPositionObjCmd(CDM_TCL_STD_ARGS) {
  int status;
  long position;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Annotation position");
    return TCL_ERROR;
  }
  if (Tcl_GetLongFromObj(interp, objv[2], &position) !=TCL_OK) return TCL_ERROR;
  status = CDM_AnnotationContainsPosition(objv[1], position);
  if (status == -1) return TCL_ERROR;
  Tcl_SetObjResult(interp, Tcl_NewBooleanObj(status));
  return TCL_OK;
} /* CDM_AnnotationContainsPositionObjCmd */

/*
 * ---- tip_AnnotationMatchesRange:
 *  DESCRIPTION:
 *   <English:
 *    This function will check if the provided Annotation (specified through the
 *    "Annotation" parameter) contains a span that has the specified start and
 *    end offsets.
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Ann:    The Annotation to be examined           (CDM_Annotation)
 *    Start:  The span start to me matched            (long)
 *    End:    The span end to me matched              (long)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to CDM_AnnotationMatchesRange (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will return 1 (true) if the Annotation contains a span
 *    that matches the specified start and end offsets. Else, 0 (false)
 *    will be returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_AnnotationContainsPosition, CDM_AnnotationMatchesRange
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_AnnotationMatchesRangeObjCmd(CDM_TCL_STD_ARGS) {
  int status;
  long position1, position2;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Annotation position1 position2");
    return TCL_ERROR;
  }
  if (Tcl_GetLongFromObj(interp, objv[2], &position1)!=TCL_OK) return TCL_ERROR;
  if (Tcl_GetLongFromObj(interp, objv[3], &position2)!=TCL_OK) return TCL_ERROR;
  status = CDM_AnnotationMatchesRange(objv[1], position1, position2);
  if (status == -1) return TCL_ERROR;
  Tcl_SetObjResult(interp, Tcl_NewBooleanObj(status));
  return TCL_OK;
} /* CDM_AnnotationMatchesRangeObjCmd */

/*
 * ---- tip_CompareAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will compare two Annotations, Ann1 and Ann2, according to
 *     their Span sets.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Ann1:        The first Annotation                (CDM_Annotation)
 *     Ann2:        The second Annotation               (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_AddAnnotation (C++ API).
 *
 *     A common use of this function is to be used with lsort, in order to
 *     sort an Annotation set, in the same way tip_SelectAnnotationsSorted
 *     would have done:
 *     [[br]][[code]][[center]]
 *       lsort -command tip_CompareAnnotations $AnnotationSet
 *     [[/center]][[/code]]
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns -1 if Ann1 is lower than
 *     Ann2, i.e. Ann1 spans a text range that starts earlier than all the
 *     text ranges spanned by the Ann2 Annotation. If the two Annotations span
 *     the same text ranges, zero is returned. Finally, 1
 *     is returned if Ann1 is greater than Ann2, i.e. when Ann2 spans a
 *     text range starts earlier than all the text ranges spanned by the Ann1
 *     Annotation. In case of an error, a value equal or greater than 10 will
 *     be returned and an error message will be left at the current Tcl
 *     interpreter.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_SelectAnnotationsSorted, CDM_CompareAnnotations, 
 *     [[A HREF="../TclTk/TclCmd/lsort.htm"]]lsort[[/A]]
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_CompareAnnotationsObjCmd(CDM_TCL_STD_ARGS) {
  int status;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Annotation1 Annotation2");
    return TCL_ERROR;
  }
  status = CDM_CompareAnnotations(objv[1], objv[2]);
  if (status >= 10) return TCL_ERROR;
  Tcl_SetObjResult(interp, CDM_NewLongObj((long) status));
  return TCL_OK;
} /* CDM_CompareAnnotationsObjCmd */

/*
 * ---- tip_CreateAnnotation:
 *  DESCRIPTION:
 *   <English:
 *    Creates a new Annotation...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Type:        The Annotation's Type                     (string)
 *      Spans:       A Set of Spans that the Annotation Covers (CDM_SpanSet)
 *      Attributes:  A Set of Attributes                  (CDM_AttributeSet)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_CreateAnnotationObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Annotation Ann;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Type Spans Attributes");
    return TCL_ERROR;
  }
  Ann = CDM_CreateAnnotation(Tcl_GetString(objv[1]), objv[2], objv[3]);
  if (Ann == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Ann);
  return TCL_OK;
} /* CDM_CreateAnnotationObjCmd */

/*
 * ---- tip_CreateAnnotationSet:
 *  DESCRIPTION:
 *   <English:
 *    Creates an empty Annotation Set...
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None: - (void)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_CreateAnnotationSetObjCmd(CDM_TCL_STD_ARGS) {
  CDM_AnnotationSet Set;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }
  Set = CDM_CreateAnnotationSet();
  Tcl_SetObjResult(interp, Set);
  return TCL_OK;
} /* CDM_CreateAnnotationSetObjCmd */

/*
 * ---- tip_CompareAnnotationSets:
 *  DESCRIPTION:
 *   <English:
 *    Creates an empty Annotation Set...
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None: - (void)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_CompareAnnotationSetsObjCmd(CDM_TCL_STD_ARGS) {
  double spanOverlapPercent = 100.0;
  int LogInfo = 0, checkValues = 0;
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 6 && objc != 7 && objc != 8) {
    Tcl_WrongNumArgs(interp, 1, objv,
      "AnnotationSet1 AnnotationSet2 AttributesList ValuesList "
      "spanOverlapPercent ?enable_Attribute_value_check ?enableLog? ?");
    return TCL_ERROR;
  }
  if (Tcl_GetDoubleFromObj(interp, objv[5], &spanOverlapPercent) != TCL_OK)
      return TCL_ERROR;
  if (objc > 7) {
    if (Tcl_GetBooleanFromObj(interp, objv[6], &checkValues) != TCL_OK)
      return TCL_ERROR;
  }
  if (objc == 8) {
    if (Tcl_GetBooleanFromObj(interp, objv[7], &LogInfo) != TCL_OK)
      return TCL_ERROR;
  }
 
  return CDM_CompareAnnotationSets(objv[1], objv[2], objv[3], objv[4],
                                   spanOverlapPercent, checkValues, LogInfo);
} /* CDM_CompareAnnotationSetsObjCmd */

/*
 * ---- tip_CreateSpan:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return a Span object
 *     (of type CDM_Span). In order to reduce memory requirements, this
 *     function will try to locate if a Span with exactly the same
 *     characteristics (same start and end values) already exists.
 *     In such a case, this (possibly shared) Span
 *     object will be returned. Else, a new Span object will be created
 *     and returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      start:  Span Start     (int)
 *      end:    Span End       (int)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_CreateAttribute (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >
 *  SECTIONS:
 *   <Span
 *   >:&:
 */
CDM_CALLBACK int CDM_CreateSpanObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Span Span;
  long Start, End;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Start End");
    return TCL_ERROR;
  }
  if (Tcl_GetLongFromObj(interp, objv[1], &Start) != TCL_OK) return TCL_ERROR;
  if (Tcl_GetLongFromObj(interp, objv[2], &End) != TCL_OK) return TCL_ERROR;
  Span=CDM_CreateSpan(Start, End);
  if (Span == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Span);
  return TCL_OK;
} /* CDM_CreateSpanObjCmd */

/*
 * ---- tip_CreateSpanSet:
 *  DESCRIPTION:
 *   <English:
 *     This function will create and return an empty Span set object
 *     (of type CDM_SpanSet). If the two optional parameters "start" and "end"
 *     are present, then a Span object with the specified start and end values
 *     will be created and appended to the Span set object, before it is
 *     returned.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      start:  Span start     (int)
 *      end:    Span end       (int)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to the following Tcl code:
 *    [[br]][[code]][[center]] set span [tip_CreateSpan $start $end][[br]]
 *          set spanSet [tip_CreateSpanSet][[br]]
 *          set spanSet [lappend $spanSet $span]
 *     [[/center]][[/code]]
 *     This function is equivelant to CDM_CreateSpanSet (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return an empty Span set object
 *     (of type CDM_SpanSet). If the two optional parameters "start" and "end"
 *     are present, then a Span object with the specified start and end values
 *     will be created and appended to the Span set object, before it is
 *     returned.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_CreateSpan,
 *     [[A HREF="../TclTk/TclCmd/lappend.htm"]]lappend[[/A]]
 *   >
 *  SECTIONS:
 *   <Span
 *   >:&:
 */
CDM_CALLBACK int CDM_CreateSpanSetObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Span SpanSet;
  long Start, End;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 1 && objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "?start end?");
    return TCL_ERROR;
  }
  if (objc == 1) {
    SpanSet=CDM_CreateSpanSet();
  } else {
    if (Tcl_GetLongFromObj(interp, objv[1], &Start) != TCL_OK) return TCL_ERROR;
    if (Tcl_GetLongFromObj(interp, objv[2], &End) != TCL_OK) return TCL_ERROR;
    SpanSet = CDM_CreateSpanSet(Start, End);
  }
  if (SpanSet == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, SpanSet);
  return TCL_OK;
} /* CDM_CreateSpanObjCmd */

/*
 * ---- tip_DisplaceAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function displases or "moves" the specified Annotation by 
 *     "displacement" characters. This function will iterate over
 *     all spans contained in the Annotation. For each span all the offsets
 *     contained in it (either start or end) will be examined: if any of this
 *     offsets is equal or greater than the value specified through the "offset"
 *     parameter, then the value of the
 *     "displacement" parameter will be added to it.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Annotation:   The Annotation to be displaced             (CDM_Annotation)
 *     offset:    A character offset after which all offsets are modified (long)
 *     displacement: The number of characters that offsets will be moved  (long)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_DisplaceAnnotation (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     The return value from this function will be a new Annotation object, if
 *     any displacement happened. 
 *
 *     If no displacement occured, then the specified Annotation through the
 *     "Annotation" parameter will be returned.
 *
 *     Finally, in case of an error, the empty string will be returned and
 *     a message will be left at the current interpreter.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_DisplaceAnnotations, CDM_DisplaceAnnotation
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_DisplaceAnnotationObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Annotation Ann;
  long offset, displacement;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Annotation offset displacement");
    return TCL_ERROR;
  }
  if (Tcl_GetLongFromObj(interp, objv[2], &offset) != TCL_OK) return TCL_ERROR;
  if (Tcl_GetLongFromObj(interp, objv[3], &displacement) != TCL_OK)
     return TCL_ERROR;
  Ann = CDM_DisplaceAnnotation(objv[1], offset, displacement);
  if (Ann == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Ann);
  return TCL_OK;
} /* CDM_DisplaceAnnotationObjCmd */

/*
 * ---- tip_GetAnnotatedTextRanges:
 *  DESCRIPTION:
 *   <English:
 *    This function will return a set of strings from the provided Text object
 *    that are annotated by the provided Annotation. In simple words, this
 *    function will examine the span set contained in the provided Annotation
 *    and return in a list all text ranges (from the specified text object)
 *    indexed by each one of the spans.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. (In this case an empty
 *    object will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Text:       The text object indexed by the Annotation. (CDM_ByteSequence)
 *    Annotation: The Annotation that indexes the text ranges. (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *    Note that usually the provided text object is the text of a Document,
 *    retrieved through calls to either tip_GetByteSequence or tip_GetRawData.
 *    The text object is expected to contain text in Unicode or UTF-8
 *    representation.
 *
 *    This function is equivelant to CDM_GetAnnotatedTextRanges (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a set of ByteSequence/RawData objects (of type
 *    CDM_ByteSequence or CDM_RawData) containing the text ranges (or
 *    portions) from the "Text" object indexed by the "Annotation" spans.
 *    In order to retrieve the individual text ranges, the user can call
 *    tip_Nth or lindex as usual with set objects.
 *
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. (In this case an empty
 *    object will be returned.)
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_GetByteSequence, tip_GetRawData, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_Free,
 *    tip_GetFirstAnnotatedTextRange
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */

/*
 * ---- tip_GetAnnotatedTextRanges:
 *  DESCRIPTION:
 *   <English:
 *    This function will return a set of strings from the text of the
 *    specified Document object
 *    that are annotated by the provided Annotation. In simple words, this
 *    function will examine the span set contained in the provided Annotation
 *    and return in a list all text ranges (from the text of the
 *    specified Document object) indexed by each one of the spans.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. (In this case an empty
 *    object will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *    Document:   The Document object.                         (CDM_Document)
 *    Annotation: The Annotation that indexes the text ranges. (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to CDM_GetAnnotatedTextRanges (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a set of ByteSequence/RawData objects (of type
 *    CDM_ByteSequence or CDM_RawData) containing the text ranges (or
 *    portions) from the text contained in the "Document" object indexed
 *    by the "Annotation" spans.
 *    In order to retrieve the individual text ranges, the user can call
 *    tip_Nth or lindex as usual with set objects.
 *
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. (In this case an empty
 *    object will be returned.)
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_GetByteSequence, tip_GetRawData, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_Free,
 *    tip_GetFirstAnnotatedTextRange
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_GetAnnotatedTextRangesObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Document Doc;
  CDM_ByteSequenceSet TextSet;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document/Text Annotation");
    return TCL_ERROR;
  }
  Doc = CDM_GetDocumentPointer(objv[1]);
  if (Doc == NULL) {
    TextSet=CDM_GetAnnotatedTextRanges(objv[1], objv[2]);
  } else {
    TextSet=CDM_GetAnnotatedTextRanges(Doc, objv[2]);
  }
  if (TextSet == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, TextSet);
  return TCL_OK;
} /* CDM_GetAnnotatedTextRangesObjCmd */


/*
 * ---- tip_GetFirstAnnotatedTextRange:
 *  DESCRIPTION:
 *   <English:
 *    This function will return a the text range from the provided Text object
 *    that is annotated by the first span of the provided Annotation.
 *    In simple words, this function will retrieve the first span from the
 *    provided Annotation and return the text range (from the specified text
 *    object) that is indexed by this span.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. (In this case an empty
 *    object will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Text:       The text object indexed by the Annotation. (CDM_ByteSequence)
 *     Annotation: The Annotation that indexes the text range. (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *    Note that usually the provided text object is the text of a Document,
 *    retrieved through calls to either CDM_GetByteSequence or CDM_GetRawData.
 *    The text object is expected to contain text in Unicode or UTF-8
 *    representation.
 *
 *    This function is equivelant to CDM_GetFirstAnnotatedTextRange (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a ByteSequence/RawData object (of type
 *    CDM_ByteSequence or CDM_RawData) containing the text range (or
 *    portion) from the "Text" object indexed by the first span of the 
 *    "Annotation" parameter.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. In this case an empty
 *    object (i.e. one containing the string "") will be returned.)
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_GetByteSequence, tip_GetRawData, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_Free,
 *    tip_GetAnnotatedTextRanges
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */

/*
 * ---- tip_GetFirstAnnotatedTextRange:
 *  DESCRIPTION:
 *   <English:
 *    This function will return a the text range from the text of the
 *    specified Document object
 *    that is annotated by the first span of the provided Annotation.
 *    In simple words, this function will retrieve the first span from the
 *    provided Annotation and return the text range (from the text of the
 *    specified Document object) that is indexed by this span.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. (In this case an empty
 *    object will be returned.)
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Document:   The Document object.                         (CDM_Document)
 *     Annotation: The Annotation that indexes the text range. (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to CDM_GetFirstAnnotatedTextRange (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns a ByteSequence/RawData object (of type
 *    CDM_ByteSequence or CDM_RawData) containing the text range (or
 *    portion) from the text contained in the "Document" object indexed
 *    by the first span of the "Annotation" parameter.
 *
 *    Note that this function will always create and return a new object, even
 *    when the Annotations has an empty span set. In this case an empty
 *    object (i.e. one containing the string "") will be returned.)
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_GetByteSequence, tip_GetRawData, [[A HREF="../TclTk/TclLib/Object.htm"]]Tcl_DecrRefCount[[/A]], tip_Free,
 *    tip_GetAnnotatedTextRanges
 *   >
 *  SECTIONS:
 *   <Document
 *   >:&:
 */
CDM_CALLBACK int CDM_GetFirstAnnotatedTextRangeObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Document Doc;
  CDM_ByteSequenceSet Text;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document/Text Annotation");
    return TCL_ERROR;
  }
  Doc = CDM_GetDocumentPointer(objv[1]);
  if (Doc == NULL) {
    Text = CDM_GetFirstAnnotatedTextRange(objv[1], objv[2]);
  } else {
    Text = CDM_GetFirstAnnotatedTextRange(Doc, objv[2]);
  }
  if (Text == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Text);
  return TCL_OK;
} /* CDM_GetFirstAnnotatedTextRangeObjCmd */


/*
 * ---- tip_GetEnd:
 *  DESCRIPTION:
 *   <English:
 *    This function accepts a span (of type CDM_Span) and returns the end of
 *    this span. Each span object is a set of exactly two integers, named as
 *    "start" and "end".
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Span: Span   (CDM_Span)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to CDM_GetEnd (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns the end value of the provided span object. If
 *    the provided object is not a valid span object or some other error
 *    occurs, tip_GetEnd will return -1 and an error message describing the
 *    error will be placed at the current Tcl interpreter.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_CreateSpan, tip_GetStart
 *   >
 *  SECTIONS:
 *   <Span
 *   >:&:
 */
CDM_CALLBACK int CDM_GetEndObjCmd(CDM_TCL_STD_ARGS) {
  long end;

  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Span");
    return TCL_ERROR;
  }
  end = CDM_GetEnd(objv[1]);
  if (end < 0) {
    if (CDM_FullGateCompatibility) {
      /*
       * Gate does not checks if the given Span is valid. Emulate this
       * bug...
       */
      Tcl_SetObjResult(interp, &CDM_Obj_NULL);
      return TCL_OK;
    } else {
      return TCL_ERROR;
    }
  }
  Tcl_SetObjResult(interp, CDM_NewLongObj(end));
  return TCL_OK;
} /* CDM_GetEndObjCmd */

/*
 * ---- tip_GetFirstSpan:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the first span range of the span list
 *     contained in the specified Annotation.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Ann:  The Annotation object.                   (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetFirstSpan (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a Span object (of type CDM_Span).
 *
 *     If an error occurs, then an error message will prompt.
 *
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_GetSpans, tip_Nth
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_GetFirstSpanObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Span Span;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Annotation");
    return TCL_ERROR;
  }
  Span = CDM_GetFirstSpan(objv[1]);
  if (Span == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Span);
  return TCL_OK;
} /* CDM_GetFirstSpanObjCmd */

/*
 * ---- tip_GetFirstSpanOffsets:
 *  DESCRIPTION:
 *   <English:
 *    This function will place the values of both span start and end of
 *    the the Annotation's first span into the two specified variable names.
 *    Each span object is a set of exactly two integers, named as
 *    "start" and "end".
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Annotation: An Annotation object  (CDM_Annotation)
 *     start:  A variable name, where the start value of the specified span will be placed. (string)
 *     end:    A variable name, where the end value of the specified span will be placed. (string)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to the following Tcl code:
 *    [[br]][[code]][[center]]
 *    set span  [tip_GetFirstSpan $Annotation][[br]]
 *    set start [tip_GetStart $span][[br]]
 *    set end   [tip_GetEnd $span]
 *    [[/center]][[/code]]
 *
 *    This function is equivelant to CDM_GetFirstSpanOffsets (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will place the two Offsets contained in the provided Span
 *    object (the start and end of the span) into the two variables
 *    pointed by the two provided variable names.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_GetFirstSpan, tip_GetStart, tip_GetEnd, CDM_GetFirstSpanOffsets
 *   >
 *  SECTIONS:
 *   <Span
 *   >:&:
 */
CDM_CALLBACK int CDM_GetFirstSpanOffsetsObjCmd(CDM_TCL_STD_ARGS) {
  long start, end;
  int status;

  CDM_UPDATE_THREAD_INTERP;
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Annotation start_varname end_varname");
    return TCL_ERROR;
  }
  status = CDM_GetFirstSpanOffsets(objv[1], &start, &end);
  if (status != TCL_OK) return status;

  /*
   * Set the two variables...
   */
  if (Tcl_ObjSetVar2(interp, objv[2], NULL,
                          CDM_NewLongObj(start), TCL_LEAVE_ERR_MSG) == NULL) {
    return TCL_ERROR;
  }
  if (Tcl_ObjSetVar2(interp, objv[3], NULL,
                          CDM_NewLongObj(end), TCL_LEAVE_ERR_MSG) == NULL) {
    return TCL_ERROR;
  }
  return TCL_OK;
} /* CDM_GetFirstSpanOffsetsObjCmd */

/*
 * ---- tip_GetSpans:
 *  DESCRIPTION:
 *   <English:
 *    Returns Annotation's Spans...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Ann:  The Annotation                     (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_GetSpansObjCmd(CDM_TCL_STD_ARGS) {
  CDM_SpanSet Spans;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Annotation");
    return TCL_ERROR;
  }
  Spans = CDM_GetSpans(objv[1]);
  if (Spans == NULL) {
    if (CDM_FullGateCompatibility) {
      /*
       * Gate does not checks if the given Annotation is valid. Emulate this
       * bug...
       */
      Spans = &CDM_Obj_NULL;
    } else {
      return TCL_ERROR;
    }
  }
  Tcl_SetObjResult(interp, Spans);
  return TCL_OK;
} /* CDM_GetSpansObjCmd */

/*
 * ---- tip_GetSpanOffsets:
 *  DESCRIPTION:
 *   <English:
 *    This function will place the values of the both span start and end of
 *    the provided span in the two specified variable names.
 *    Each span object is a set of exactly two integers, named as
 *    "start" and "end".
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Span:   A Span object  (CDM_Span)
 *     start:  A variable name, where the start value of the specified span will be placed. (string)
 *     end:    A variable name, where the end value of the specified span will be placed. (string)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to the following Tcl code:
 *    [[br]][[code]][[center]]set start [tip_GetStart $Span][[br]]
 *                            set end   [tip_GetEnd $Span]
 *     [[/center]][[/code]]
 *
 *    This function is equivelant to CDM_GetSpanOffsets (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function will place the two Offsets contained in the provided Span
 *    object (the start and end of the span) into the two variables
 *    pointed by the two provided variable names.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_CreateSpan, tip_GetStart, tip_GetEnd
 *   >
 *  SECTIONS:
 *   <Span
 *   >:&:
 */
CDM_CALLBACK int CDM_GetSpanOffsetsObjCmd(CDM_TCL_STD_ARGS) {
  long start, end;
  int status;

  CDM_UPDATE_THREAD_INTERP;
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Span start-varname end-varname");
    return TCL_ERROR;
  }
  status = CDM_GetSpanOffsets(objv[1], &start, &end);
  if (status != TCL_OK) return status;

  /*
   * Set the two variables...
   */
  if (Tcl_ObjSetVar2(interp, objv[2], NULL,
                          CDM_NewLongObj(start), TCL_LEAVE_ERR_MSG) == NULL) {
    return TCL_ERROR;
  }
  if (Tcl_ObjSetVar2(interp, objv[3], NULL,
                          CDM_NewLongObj(end), TCL_LEAVE_ERR_MSG) == NULL) {
    return TCL_ERROR;
  }
  return TCL_OK;
} /* CDM_GetSpanOffsetsObjCmd */

/*
 * ---- tip_GetStart:
 *  DESCRIPTION:
 *   <English:
 *    This function accepts a span (of type CDM_Span) and returns the start of
 *    this span. Each span object is a set of exactly two integers, named as
 *    "start" and "end".
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Span: Span   (CDM_Span)
 *   >
 *  NOTES:
 *   <English:
 *    This function is equivelant to CDM_GetStart (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *    This function returns the start value of the provided span object. If
 *    the provided object is not a valid span object or some other error
 *    occurs, CDM_GetStart will return -1 and an error message describing the
 *    error will be placed at the current Tcl interpreter (CDM_Interp).
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_CreateSpan, tip_GetEnd, tip_GetSpanOffsets
 *   >
 *  SECTIONS:
 *   <Span
 *   >:&:
 */
CDM_CALLBACK int CDM_GetStartObjCmd(CDM_TCL_STD_ARGS) {
  long start;

  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Span");
    return TCL_ERROR;
  }
  start = CDM_GetStart(objv[1]);
  if (start < 0) {
    if (CDM_FullGateCompatibility) {
      /*
       * Gate does not checks if the given Span is valid. Emulate this
       * bug...
       */
      Tcl_SetObjResult(interp, &CDM_Obj_NULL);
      return TCL_OK;
    } else {
      return TCL_ERROR;
    }
  }
  Tcl_SetObjResult(interp, CDM_NewLongObj(start));
  return TCL_OK;
} /* CDM_GetStartObjCmd */


/*
 * ---- tip_GetType:
 *    Arguments: 
 *  DESCRIPTION:
 *   <English:
 *    Returns Annotation's Type...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Ann:         Annotation                     (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_GetTypeObjCmd(CDM_TCL_STD_ARGS) {
  char *type;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Annotation");
    return TCL_ERROR;
  }
  type = CDM_GetType(objv[1]);
  if (type == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, ObjectCache.FindAnnotationType(type));
  return TCL_OK;
} /* CDM_GetTypeObjCmd */


/*
 * ---- tip_GetId:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Id of the specified
 *     Document. The Id is the file name of the initial file
 *     that contained the text which was used in order to create the specified
 *     Document. The returned value will be encoded using the UTF-8 encoding
 *     (thus enabling the existance of non Latin characters in the value).
 *
 *     The return value will be of type string.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document:    A Document object.            (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_GetId (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns a UTF-8 string that represents the Id of
 *     the specified Document.
 *
 *     If an error occurs, then an error message will be prompted.
 *
 *   >
 *  SECTIONS:
 *   <Document Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_GetIdObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Document Doc;
  char *cid;
  long id;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document/Annotation");
    return TCL_ERROR;
  }
  Doc = CDM_GetDocumentPointer(objv[1]);
  if (Doc == NULL) {
    id = CDM_GetId(objv[1]);
    if (id < 0) {
      Tcl_ResetResult(interp);
    } else {
      Tcl_SetObjResult(interp, CDM_NewLongObj(id));
    }
  } else {
    cid = CDM_GetId(Doc);
    if (cid == NULL) return TCL_ERROR;
    Tcl_SetResult(interp, cid, TCL_VOLATILE);
  }
  return TCL_OK;
} /* CDM_GetIdObjCmd */


/*
 * ---- tip_Length:
 *  DESCRIPTION:
 *   <English:
 *     This function will return the Length (the number of all the Documents)
 *     of the specified Collection.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Collection:  A Collection object.           (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_Length   (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will return an integer value that will represent the
 *     total number of the Documents that the specified Collection currently
 *     holds.
 *
 *     If an error occurs, then an error message will prompt.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     
 *   >
 *  SECTIONS:
 *   <Collection Annotation Attribute Span
 *   >:&:
 */
CDM_CALLBACK int CDM_LengthObjCmd(CDM_TCL_STD_ARGS) {
  CDM_Collection Col;
  long length;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv,
      "Collection/AnnotationSet/AttributeSet/SpanSet");
    return TCL_ERROR;
  }

  Col = CDM_GetCollectionPointer(objv[1]);
  if (Col == NULL) {
    length = CDM_Length(objv[1]);
  } else {
    length = CDM_Length(Col);
  }
  if (length < 0) return TCL_ERROR;
  Tcl_SetObjResult(interp, CDM_NewLongObj(length));
  return TCL_OK;
} /* CDM_LengthObjCmd */


/*
 * ---- tip_MergeAnnotations:
 *  DESCRIPTION:
 *   <English:
 *    Merge two Annotation Sets...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Set1: First Set  (CDM_AnnotationSet)
 *      Set2: Second Set (CDM_AnnotationSet)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
CDM_CALLBACK int CDM_MergeAnnotationsObjCmd(CDM_TCL_STD_ARGS) {
  CDM_AnnotationSet Set;

  CDM_UPDATE_THREAD_INTERP;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Set1 Set2");
    return TCL_ERROR;
  }
  Set = CDM_MergeAnnotations(objv[1], objv[2]);
  if (Set == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Set);
  return TCL_OK;
} /* CDM_MergeAnnotationsObjCmd */


/*
 * ---- tip_Nth:
 *  DESCRIPTION:
 *   <English:
 *    Return Nth element of a set...
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Set:    Set Pointer      ()
 *      n:      Item number      (int)
 *   >
 *  NOTES:
 *   <English:
 *   >
 *  RETURN_VALUE:
 *   <English:
 *   >
 *  SEE_ALSO:
 *   <English:
 *   >
 *  SECTIONS:
 *   <Annotation Attribute Span
 *   >:&:
 */
CDM_CALLBACK int CDM_NthObjCmd(CDM_TCL_STD_ARGS) {
  Tcl_Obj *Obj;
  long N;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "AnnotationSet/AttributeSet/SpanSet N");
    return TCL_ERROR;
  }
  if (Tcl_GetLongFromObj(interp, objv[2], &N) != TCL_OK) return TCL_ERROR;
  Obj = CDM_Nth(objv[1], N);
  if (Obj == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Obj);
  return TCL_OK;
} /* CDM_NthObjCmd */

/*
 * ---- tip_SortAnnotationSet:
 *  DESCRIPTION:
 *   <English:
 *     This function will sort the Annotations contained in the provided
 *     Annotation set, according to their Span sets.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Set:        The Annotation set to be sorted      (CDM_AnnotationSet)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_SortAnnotationSet (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will create and return a new AnnotationSet object,
 *     that will contain all the Annotations contained in the provided by the
 *     "Set" parameter set, sorted by their span ranges. For more information
 *     about the sorting order, please refer to the CDM_CompareAnnotations
 *     function.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     tip_SelectAnnotationsSorted, tip_CompareAnnotations, 
 *     CDM_SortAnnotationSet,
 *     [[A HREF="../TclTk/TclCmd/lsort.htm"]]lsort[[/A]]
 *   >
 *  SECTIONS:
 *   <Annotation
 *   >:&:
 */
 
/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |      V A L I D A T I O N           S E C T I O N      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/

/*
 * ---- tip_ValidAnnotation:
 *  DESCRIPTION:
 *   <English:
 *     This function will check if the specified as parameter object is a
 *     valid Annotation. A valid Annotation can be used everywhere an object
 *     of type CDM_Annotation is required.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Ann:   Annotation Object    (CDM_Annotation)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_ValidAnnotation (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns 1 if the provided object is a valid Annotation
 *     object, 0 otherwise.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_ValidAnnotation
 *   >
 *  SECTIONS:
 *   <Validation
 *   >:&:
 */
CDM_CALLBACK int CDM_ValidAnnotationObjCmd(CDM_TCL_STD_ARGS) {
  int valid;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Annotation");
    return TCL_ERROR;
  }

  valid = CDM_ValidAnnotation(objv[1]);
  if (valid < 0) return TCL_ERROR;
  Tcl_SetObjResult(interp, CDM_NewLongObj(valid));
  return TCL_OK;
} /* CDM_ValidAnnotationObjCmd */

/*
 * ---- tip_ValidAttribute:
 *  DESCRIPTION:
 *   <English:
 *     This function will check if the specified as parameter object is a
 *     valid Attribute. A valid Attribute can be used everywhere an object
 *     of type CDM_Attribute is required.
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Attribute:   Attribute Object     (CDM_Attribute)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_ValidAttribute (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns 1 if the provided object is a valid Attribute
 *     object, 0 otherwise.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_ValidAttribute
 *   >
 *  SECTIONS:
 *   <Validation
 *   >:&:
 */
CDM_CALLBACK int CDM_ValidAttributeObjCmd(CDM_TCL_STD_ARGS) {
  int valid;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Attribute");
    return TCL_ERROR;
  }

  valid = CDM_ValidAttribute(objv[1]);
  if (valid < 0) return TCL_ERROR;
  Tcl_SetObjResult(interp, CDM_NewLongObj(valid));
  return TCL_OK;
} /* CDM_ValidAttributeObjCmd */

/*
 * ---- tip_ValidCollection:
 *  DESCRIPTION:
 *   <English:
 *    This function will checks whether the specified through the "Collection"
 *    parameter object is a valid Collection object,
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Collection: A Collection object        (CDM_Collection)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_ValidCollectionPointer (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns 1 if the provided object is a valid Collection
 *     object, 0 otherwise.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_ValidCollectionPointer
 *   >
 *  SECTIONS:
 *   <Validation
 *   >:&:
 */
CDM_CALLBACK int CDM_ValidCollectionObjCmd(CDM_TCL_STD_ARGS) {
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Collection");
    return TCL_ERROR;
  }

  if (CDM_GetCollectionPointer(objv[1]) == NULL) {
    Tcl_SetObjResult(interp, CDM_NewLongObj(0));
  } else {
    Tcl_SetObjResult(interp, CDM_NewLongObj(1));
  }
  return TCL_OK;
} /* CDM_ValidCollectionObjCmd */

/*
 * ---- tip_ValidDocument:
 *  DESCRIPTION:
 *   <English:
 *    This function will checks whether the specified through the "Document"
 *    parameter object is a valid Document object,
 *   >
 *  ARGUMENTS:
 *   <English:
 *      Document: A Document object        (CDM_Document)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_ValidDocumentPointer (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function returns 1 if the provided object is a valid Document
 *     object, 0 otherwise.
 *   >
 *  SEE_ALSO:
 *   <English:
 *     CDM_ValidDocumentPointer
 *   >
 *  SECTIONS:
 *   <Validation
 *   >:&:
 */
CDM_CALLBACK int CDM_ValidDocumentObjCmd(CDM_TCL_STD_ARGS) {
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "Document");
    return TCL_ERROR;
  }

  if (CDM_GetDocumentPointer(objv[1]) == NULL) {
    Tcl_SetObjResult(interp, CDM_NewLongObj(0));
  } else {
    Tcl_SetObjResult(interp, CDM_NewLongObj(1));
  }
  return TCL_OK;
} /* CDM_ValidDocumentObjCmd */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |      U T I L I T I E S           R O U T I N E S      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/
#ifdef CDM_ENABLE_TIME_ROUTINES
CDM_CALLBACK int CDM_CdmGetTimeObjCmd(CDM_TCL_STD_ARGS) { 
  Tcl_Time Time;
  Tcl_Obj *ListObjPtr;
  CDM_UPDATE_THREAD_INTERP;

  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }
  TclpGetTime(&Time);
  /*totalMicroSec =
        (stop.sec - start.sec)*1000000 + (stop.usec - start.usec);
    sprintf(buf, "%.0f microseconds per iteration",
        ((count <= 0) ? 0 : totalMicroSec/count));*/
  ListObjPtr = Tcl_NewListObj(0, NULL);
  if (ListObjPtr == NULL) return TCL_ERROR;
  Tcl_ListObjAppendElement(interp, ListObjPtr, Tcl_NewLongObj(Time.sec));
  Tcl_ListObjAppendElement(interp, ListObjPtr, Tcl_NewLongObj(Time.usec));
  Tcl_SetObjResult(interp, ListObjPtr);
  return TCL_OK;
} /* CDM_CdmGetTimeObjCmd */

CDM_CALLBACK int CDM_TimeObjCmd(CDM_TCL_STD_ARGS) {
  register Tcl_Obj *objPtr, *ListPtr;
  register int result;
  double totalMicroSec;
  Tcl_Time start, stop;
  char buf[100];

  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "command");
    return TCL_ERROR;
  }
    
  objPtr = objv[1];
  TclpGetTime(&start);
  result = Tcl_EvalObjEx(interp, objPtr, 0);
  if (result != TCL_OK) return result;
  TclpGetTime(&stop);
  objPtr = Tcl_GetObjResult(interp);
  ListPtr = Tcl_NewListObj(0, NULL);
  if (ListPtr == NULL) 
    return CDM_ReportError("CDM_TimeObjCmd: Tcl_NewListObj returns NULL!");
  
  Tcl_ListObjAppendElement(interp, ListPtr, objPtr);
  
  totalMicroSec = (stop.sec - start.sec)*1000000 + (stop.usec - start.usec);
  sprintf(buf, "%.0f microseconds per iteration", totalMicroSec);
  objPtr = Tcl_NewStringObj(buf, -1);
  Tcl_ListObjAppendElement(interp, ListPtr, objPtr);
  Tcl_SetObjResult(interp, ListPtr);
  return TCL_OK;
} /* CDM_TimeObjCmd */
#endif /* CDM_ENABLE_TIME_ROUTINES */

CDM_CALLBACK int CDM_ClearSavedValuesObjCmd(CDM_TCL_STD_ARGS) {
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }
#ifdef DEBUG_HASHED_VALUES
  CDM_Debug("CDM_ClearSavedValuesObjCmd: Calling CDM_ObjectCache::Clear()!\n");
#endif /* DEBUG_HASHED_VALUES */
  ObjectCache.Clear();
  return TCL_OK;
} /* CDM_ClearSavedValuesObjCmd */

CDM_CALLBACK int CDM_ClearUnusedSavedValuesObjCmd(CDM_TCL_STD_ARGS) {
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }
#ifdef DEBUG_HASHED_VALUES
  CDM_Debug("CDM_ClearUnusedSavedValuesObjCmd: Calling CDM_ObjectCache::RemoveUnusedObjects()!\n");
#endif /* DEBUG_HASHED_VALUES */
  ObjectCache.RemoveUnusedObjects();
  return TCL_OK;
} /* CDM_ClearUnusedSavedValuesObjCmd */

CDM_CALLBACK int CDM_GetIPObjCmd(CDM_TCL_STD_ARGS) {
  char *ip;
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }
  ip = CDM_GetIP();
  if (ip == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Tcl_NewStringObj(ip, -1));
  return TCL_OK;
} /* CDM_GetIPObjCmd */

CDM_CALLBACK int CDM_GetCurrentThreadObjCmd(CDM_TCL_STD_ARGS) {
  void *id;
  char tmp[32];
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }
  id = CDM_GetCurrentThread();
  snprintf(tmp, 31, "%p", id);
  Tcl_SetObjResult(interp, Tcl_NewStringObj(tmp, -1));
  return TCL_OK;
} /* CDM_GetCurrentThreadObjCmd */

CDM_CALLBACK int CDM_EvalObjCmd(CDM_TCL_STD_ARGS) {
  char **argv = NULL, *cmd;
  register int i, max = 20, result;
  CDM_UPDATE_THREAD_INTERP;
  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "command ?args?");
    return TCL_ERROR;
  }
  
  /*
   * Keep a reference to the objects so as not to loose the string
   * pointers...
   */
  for (i=1; i<objc; i++) Tcl_IncrRefCount(objv[i]);
  cmd = Tcl_GetString(objv[1]);
  if (objc > 2) argv = (char **) Tcl_Alloc(max * sizeof(char *));

  i = 2;
  while (i < objc) {
    argv[i-2] = Tcl_GetString(objv[i]);
    if (i-2 >= max) {
      /* allocate a bigger vector and copy old one */
      char **oldv = argv;
      register int j;
      max *= 2;
      argv = (char **) Tcl_Alloc(max * sizeof(char *));
      for (j=0; j<i-1; j++) {
        argv[j] = oldv[j];
      }
      Tcl_Free((char *) oldv);
    }
    i++;
  }
  
  switch (objc) {
    case 2: {result = CDM_Eval(interp, cmd, (char *)NULL);break;}
    case 3: {result = CDM_Eval(interp, cmd, argv[0], (char *)NULL);break;}
    case 4: {result = CDM_Eval(interp,cmd,argv[0],argv[1],(char *)NULL);break;}
    case 5: {result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],(char *)NULL);
             break;}
    case 6: {result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],
             (char *)NULL);break;}
    case 7: {result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             (char *)NULL);break;}
    case 8: {result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],(char *)NULL);break;}
    case 9: {result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],(char *)NULL);break;}
    case 10:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],(char *)NULL);break;}
    case 11:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],(char *)NULL);break;}
    case 12:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],(char *)NULL);break;}
    case 13:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],(char *)NULL);
             break;}
    case 14:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],
             (char *)NULL);break;}
    case 15:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             (char *)NULL);break;}
    case 16:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],(char *)NULL);break;}
    case 17:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],(char *)NULL);break;}
    case 18:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],argv[15],(char *)NULL);break;}
    case 19:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],argv[15],argv[16],(char *)NULL);break;}
    case 20:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],argv[15],argv[16],argv[17],(char *)NULL);break;}
    case 21:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],argv[15],argv[16],argv[17],argv[18],
             (char *)NULL);break;}
    case 22:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],argv[15],argv[16],argv[17],argv[18],argv[19],
             (char *)NULL);break;}
    case 23:{result=CDM_Eval(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],argv[15],argv[16],argv[17],argv[18],argv[19],
             argv[20],(char *)NULL);break;}
    default: {result=CDM_ReportError("cannot handle so many items!");}
  }
  if (argv != NULL) Tcl_Free((char *) argv);
  for (i=1; i<objc; i++) Tcl_IncrRefCount(objv[i]);
  return result;
} /* CDM_EvalObjCmd */

CDM_CALLBACK int CDM_ExecObjCmd(CDM_TCL_STD_ARGS) {
  char **argv = NULL, *cmd;
  register int i, max = 20, result;
  CDM_UPDATE_THREAD_INTERP;
  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "command ?args?");
    return TCL_ERROR;
  }
  
  /*
   * Keep a reference to the objects so as not to loose the string
   * pointers...
   */
  for (i=1; i<objc; i++) Tcl_IncrRefCount(objv[i]);
  cmd = Tcl_GetString(objv[1]);
  if (objc > 2) argv = (char **) Tcl_Alloc(max * sizeof(char *));

  i = 2;
  while (i < objc) {
    argv[i-2] = Tcl_GetString(objv[i]);
    if (i-2 >= max) {
      /* allocate a bigger vector and copy old one */
      char **oldv = argv;
      register int j;
      max *= 2;
      argv = (char **) Tcl_Alloc(max * sizeof(char *));
      for (j=0; j<i-1; j++) {
        argv[j] = oldv[j];
      }
      Tcl_Free((char *) oldv);
    }
    i++;
  }
  
  switch (objc) {
    case 2: {result = CDM_Exec(interp, cmd, (char *)NULL);break;}
    case 3: {result = CDM_Exec(interp, cmd, argv[0], (char *)NULL);break;}
    case 4: {result = CDM_Exec(interp,cmd,argv[0],argv[1],(char *)NULL);break;}
    case 5: {result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],(char *)NULL);
             break;}
    case 6: {result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],
             (char *)NULL);break;}
    case 7: {result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             (char *)NULL);break;}
    case 8: {result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],(char *)NULL);break;}
    case 9: {result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],(char *)NULL);break;}
    case 10:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],(char *)NULL);break;}
    case 11:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],(char *)NULL);break;}
    case 12:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],(char *)NULL);break;}
    case 13:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],(char *)NULL);
             break;}
    case 14:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],
             (char *)NULL);break;}
    case 15:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             (char *)NULL);break;}
    case 16:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],(char *)NULL);break;}
    case 17:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],(char *)NULL);break;}
    case 18:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],argv[15],(char *)NULL);break;}
    case 19:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],argv[15],argv[16],(char *)NULL);break;}
    case 20:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],argv[15],argv[16],argv[17],(char *)NULL);break;}
    case 21:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],argv[15],argv[16],argv[17],argv[18],
             (char *)NULL);break;}
    case 22:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],argv[15],argv[16],argv[17],argv[18],argv[19],
             (char *)NULL);break;}
    case 23:{result=CDM_Exec(interp,cmd,argv[0],argv[1],argv[2],argv[3],argv[4],
             argv[5],argv[6],argv[7],argv[8],argv[9],argv[10],argv[11],argv[12],
             argv[13],argv[14],argv[15],argv[16],argv[17],argv[18],argv[19],
             argv[20],(char *)NULL);break;}
    default: {result=CDM_ReportError("cannot handle so many items!");}
  }
  if (argv != NULL) Tcl_Free((char *) argv);
  for (i=1; i<objc; i++) Tcl_IncrRefCount(objv[i]);
  return result;
} /* CDM_ExecObjCmd */

CDM_CALLBACK int CDM_LappendUniqueObjCmd(CDM_TCL_STD_ARGS) {
  Tcl_Obj *ListObj, *newValuePtr;
  int i, appended = 0;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "varName ?value value ...?");
    return TCL_ERROR;
  }

  newValuePtr = Tcl_ObjGetVar2(interp, objv[1], (Tcl_Obj *) NULL,
                               (TCL_LEAVE_ERR_MSG));
  if (newValuePtr == NULL) {
    /*
     * The variable doesn't exist yet. Just create it with an empty
     * initial value.
     */
    Tcl_Obj *nullObjPtr = Tcl_NewObj();
    newValuePtr = Tcl_ObjSetVar2(interp, objv[1], NULL,
                                 nullObjPtr, TCL_LEAVE_ERR_MSG);
    if (newValuePtr == NULL) {
      Tcl_DecrRefCount(nullObjPtr); /* free unneeded object */
      return TCL_ERROR;
    }
  }

  if (objc > 2) {
    /*
     * We have arguments to append.
     */
    if (Tcl_IsShared(newValuePtr)) newValuePtr = Tcl_DuplicateObj(newValuePtr);
    for (i=2; i<objc; i++) {
      if (CDM_LappendUnique(newValuePtr, objv[i]) == TCL_OK) appended++;
    }
  }
  ListObj = Tcl_ObjSetVar2(interp, objv[1], NULL,
                           newValuePtr, TCL_LEAVE_ERR_MSG);
  if (ListObj == NULL) {
    Tcl_DecrRefCount(newValuePtr); /* free unneeded object */
    return TCL_ERROR;
  }
  Tcl_SetObjResult(interp, Tcl_NewIntObj(appended));
  return TCL_OK;
} /* CDM_LappendUniqueObjCmd */

/*
 * ---- tip_ExtractTextBasedOnAnnotations:
 *  DESCRIPTION:
 *   <English:
 *     This function will accept two parameters, a text object (of type
 *     CDM_ByteSequence or CDM_RawData) and a set of Annotations (of type
 *     CDM_AnnotationSet). It will return a new text object (of type 
 *     CDM_ByteSequence or CDM_RawData) that will contain the text portions
 *     from the provided text object that are spanned by Annotations in the
 *     specified Annotation set. The portions of the original text that are not
 *     spanned by Annotations will either be replaced with white space
 *     characters or eliminated from the resulting text, depending on the value
 *     of the "mode" parameter. Valid values for the "mode" parameter are
 *     "CDM_EXTRACT_TEXT_LEAVE_UNSPANNED_TEXT" and 
 *     "CDM_EXTRACT_TEXT_REMOVE_UNSPANNED_TEXT". The first value will convert text
 *     portions not spanned into white space characters (either spaces (' ') or
 *     new-line characters ('\n')), while the second value will cause the
 *     elimination of unspanned text portions from the resulting text object.
 *
 *     A common use of this function is when we want to perform a linguistic
 *     task only on part of the text contained on the Document, but we want the
 *     resulting Annotations to span the proper text portions, when added to
 *     the original Document. For example, we want to apply a sentence splitter
 *     on an HTML document. As the sentence splitter has no knowledge of HTML
 *     tags, we want to create a version of the Document text where the HTML
 *     tags have been replaced with spaces. Provided that we have Annotations
 *     that span all the HTML tags, we can create this text by simply executing
 *     the following command:
 *     [[br]][[code]][[center]]
 *     set Text [tip_ExtractTextBasedOnAnnotations $DocText $HTML_Annotations]
 *     [[/center]][[/code]]
 *
 *     The value of the "mode" parameter can be ommited. In this case, it
 *     defaults to CDM_EXTRACT_TEXT_LEAVE_UNSPANNED_TEXT.
 *
 *     This function will always create and return a new text object,
 *     even when the Annotation set is empty or the specified text object
 *     through the "Text" parameter has no text. 
 *   >
 *  ARGUMENTS:
 *   <English:
 *     Text: A text object.                 (CDM_ByteSequence or CDM_RawData)
 *     Set:  An Annotation set.             (CDM_AnnotationSet)
 *     mode: One of the values CDM_EXTRACT_TEXT_LEAVE_UNSPANNED_TEXT or CDM_EXTRACT_TEXT_REMOVE_UNSPANNED_TEXT (int)
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivalent to CDM_ExtractTextBasedOnAnnotations
 *     (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     This function will always create and return a new text object,
 *     even when the Annotation set is empty or the specified text object
 *     through the "Text" parameter has no text.
 *   >
 *   SECTIONS:
 *    <Utilities
 *   >:&:
 */
CDM_CALLBACK int CDM_ExtractTextBasedOnAnnotationsObjCmd(CDM_TCL_STD_ARGS) {
  Tcl_Obj *Text;
  int mode = CDM_EXTRACT_TEXT_LEAVE_UNSPANNED_TEXT;
  CDM_UPDATE_THREAD_INTERP;
  
  if (objc != 3 && objc != 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "Text AnnotationSet ?mode?");
    return TCL_ERROR;
  }

  if (objc == 4) {
    if (strcmp("CDM_EXTRACT_TEXT_LEAVE_UNSPANNED_TEXT",
               Tcl_GetString(objv[3])) == 0) {
      mode = CDM_EXTRACT_TEXT_LEAVE_UNSPANNED_TEXT;
    } else if (strcmp("CDM_EXTRACT_TEXT_REMOVE_UNSPANNED_TEXT",
               Tcl_GetString(objv[3])) == 0) {
      mode = CDM_EXTRACT_TEXT_REMOVE_UNSPANNED_TEXT;
    } else {
      return CDM_ReportError("invalid mode \"%s\"!", Tcl_GetString(objv[3]));
    }
  }

  Text = CDM_ExtractTextBasedOnAnnotations(objv[1], objv[2], mode);
  if (Text == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Text);
  return TCL_OK;
} /* CDM_ExtractTextBasedOnAnnotationsObjCmd */

/******************************************************************************
 ******************************************************************************
 ****    +-------------------------------------------------------+         ****
 ****    |     G U I     R E L A T E D      R O U T I N E S      |         ****
 ****    +-------------------------------------------------------+         ****
 ******************************************************************************
 *****************************************************************************/
CDM_CALLBACK int CDM_GetCollectionFileXIDsObjCmd(CDM_TCL_STD_ARGS) {
  Tcl_Obj *Obj;
  int Format = CDM_FORMAT_NATIVE;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2 && objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Filename ?Format?");
    return TCL_ERROR;
  }

  if (objc == 3) {
    if (strcmp(Tcl_GetString(objv[2]), "gate2") == 0) {
      Format = CDM_FORMAT_GATE_2x;
    } else if (strcmp(Tcl_GetString(objv[2]), "gate") == 0){
      Format = CDM_FORMAT_GATE_1x;
    }
  }
  Obj = CDM_GetCollectionFileXIDs(Tcl_GetString(objv[1]), Format);
  if (Obj == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Obj);
  return TCL_OK;
} /* CDM_GetCollectionFileXIDsObjCmd */

CDM_CALLBACK int CDM_GetCollectionFileIDsObjCmd(CDM_TCL_STD_ARGS) {
  Tcl_Obj *Obj;
  int Format = CDM_FORMAT_NATIVE, i;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2 && objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Filename ?Format?");
    return TCL_ERROR;
  }

  if (objc == 3) {
    if (strcmp(Tcl_GetString(objv[2]), "gate2") == 0) {
      Format = CDM_FORMAT_GATE_2x;
    } else if (strcmp(Tcl_GetString(objv[2]), "gate") == 0) {
      Format = CDM_FORMAT_GATE_1x;
    } else if (strcmp(Tcl_GetString(objv[2]), "registered") == 0) {
      Format = CDM_FORMAT_REGISTERED;
    } else {
      /*
       * Probably this is a registered format. See if we can locate it...
       * Search all registered plugins for this type...
       */
      for (i=0; i<CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER; i++) {
        if (CDM_AdditionalCollectionTypes[i].CollectionType == NULL) continue;
        if (CDM_Strcmp(CDM_AdditionalCollectionTypes[i].CollectionType,
                       Tcl_GetString(objv[2])) != 0)                 continue;
        /*
         * We have found the entry matching our type. If the corresponding
         * function pointer is not NULL, called it (and hope it doesn't crash
         * :-))
         */
        Format = i + CDM_FORMAT_REGISTERED;
        break;
      }
    }
  }
  Obj = CDM_GetCollectionFileIDs(Tcl_GetString(objv[1]), Format);
  if (Obj == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Obj);
  return TCL_OK;
} /* CDM_GetCollectionFileIDsObjCmd */

CDM_CALLBACK int CDM_GetDocumentAttributesObjCmd(CDM_TCL_STD_ARGS) {
  Tcl_Obj *Obj;
  int Format = CDM_FORMAT_NATIVE, i;
  
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2 && objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "Filename ?Format?");
    return TCL_ERROR;
  }

  if (objc == 3) {
    if (strcmp(Tcl_GetString(objv[2]), "gate2") == 0) {
      Format = CDM_FORMAT_GATE_2x;
    } else if (strcmp(Tcl_GetString(objv[2]), "gate") == 0){
      Format = CDM_FORMAT_GATE_1x;
    } else if (strcmp(Tcl_GetString(objv[2]), "registered") == 0) {
      Format = CDM_FORMAT_REGISTERED;
    } else {
      /*
       * Probably this is a registered format. See if we can locate it...
       * Search all registered plugins for this type...
       */
      for (i=0; i<CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER; i++) {
        if (CDM_AdditionalCollectionTypes[i].CollectionType == NULL) continue;
        if (CDM_Strcmp(CDM_AdditionalCollectionTypes[i].CollectionType,
                       Tcl_GetString(objv[2])) != 0)                 continue;
        /*
         * We have found the entry matching our type. If the corresponding
         * function pointer is not NULL, called it (and hope it doesn't crash
         * :-))
         */
        Format = i + CDM_FORMAT_REGISTERED;
        break;
      }
    }
  }
  Obj = CDM_GetDocumentAttributes(Tcl_GetString(objv[1]), Format);
  if (Obj == NULL) return TCL_ERROR;
  Tcl_SetObjResult(interp, Obj);
  return TCL_OK;
} /* CDM_GetDocumentAttributesObjCmd */

CDM_CALLBACK int CDM_PresentAnnotationForRawViewerObjCmd(CDM_TCL_STD_ARGS) {
  Tcl_Obj *Obj, *ith_obj, *Result, *Element[4], **obj;
  int elements, i, type;

  CDM_UPDATE_THREAD_INTERP;
  
  Tcl_ResetResult(interp);
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "annotation");
    return TCL_ERROR;
  }

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(objv[1])) {
    Element[0] = CDM_NewLongObj(CDMAnnObj_GetId(objv[1]));
    Element[1] = CDMAnnObj_GetType(objv[1]);
    Element[3] = CDMAnnObj_GetAttributes(objv[1]);
  } else {
#endif /* ENABLE_ANNOTATION_OBJECT */
    if (!CDM_ValidAnnotation(objv[1])) return TCL_ERROR;
    if (Tcl_ListObjGetElements(interp, objv[1], &elements, &obj) != TCL_OK) {
      return TCL_ERROR;
    }
    Element[0] = obj[0];
    Element[1] = obj[1];
    Element[2] = obj[2];
    Element[3] = obj[3];
#ifdef ENABLE_ANNOTATION_OBJECT
  }
#endif /* ENABLE_ANNOTATION_OBJECT */    
  /*
   * Add the ID and the Annotation Type while creating the list object...
   */
  Result = Tcl_NewListObj(2, Element);
  if (Result == NULL) return TCL_ERROR;

#ifdef ENABLE_ANNOTATION_OBJECT
  if (ANNOTATION_OBJ(objv[1])) {
    if (CDMAnnObj_GetStart(objv[1]) != -1) {
      Tcl_ListObjAppendElement(interp, Result,
              CDM_NewLongObj(CDMAnnObj_GetStart(objv[1])));
      Tcl_ListObjAppendElement(interp, Result,
              CDM_NewLongObj(CDMAnnObj_GetEnd(objv[1])));
      Element[2] = NULL;
    } else {
      Element[2] = CDM_GetSpans(objv[1]);
    }
  }
#endif /* ENABLE_ANNOTATION_OBJECT */

  /*
   * Get the Annotation Span set...
   */
  if (Element[2]) {
    Obj = CDM_Nth(Element[2], 0);
    /*
     * Add Start/End of first span...
     */
    ith_obj = CDM_Nth(Obj, 0);
    if (!ith_obj) ith_obj = &CDM_Obj_NULL;
    Tcl_ListObjAppendElement(interp, Result, ith_obj);
    ith_obj = CDM_Nth(Obj, 1);
    if (!ith_obj) ith_obj = &CDM_Obj_NULL;
    Tcl_ListObjAppendElement(interp, Result, ith_obj);
  }

  /*
   * Add Attributes in form "name=value "...
   */
  if (Tcl_ListObjGetElements(interp, Element[3], &elements, &obj)!=TCL_OK) {
    Tcl_DecrRefCount(Result);
    return TCL_ERROR;
  }
  Obj = Tcl_NewStringObj(NULL, 0);
  if (Obj == NULL) {
    Tcl_DecrRefCount(Result);
    return TCL_ERROR;
  }

  for (i=0; i<elements; i++) {
    Tcl_AppendObjToObj(Obj, CDM_GetNameObj(obj[i]));
    Tcl_AppendToObj(Obj, "=", 1);
    type = CDM_GetValueType(obj[i]);
    switch (type) {
      case CDM_STRING: {
        Tcl_AppendObjToObj(Obj, CDM_GetValueValue(obj[i]));
        break;
      }
      case CDM_STRING_SET: {
        Tcl_AppendToObj(Obj, "[", 1);
        Tcl_AppendObjToObj(Obj, CDM_GetValueValue(obj[i]));
        Tcl_AppendToObj(Obj, "]", 1);
        break;
      }
      case CDM_IMAGE: {
        Tcl_AppendObjToObj(Obj, CDM_NewStringObj("[binary image data]"));
        break;
      }
      case CDM_BASE64_IMAGE: {
        Tcl_AppendObjToObj(Obj, CDM_NewStringObj("[image data (base64)]"));
        break;
      }
      default: {
        Tcl_AppendObjToObj(Obj, CDM_GetValueValue(obj[i]));
        break;
      }
    } /* switch (type) */
    if (i != elements-1) {
      Tcl_AppendToObj(Obj, ", ", 2);
    }
  }
  i = Tcl_GetCharLength(Obj);

  /*
   * Don't allow our result string to be too long, as displaying it may crash
   * wish. Restrict it to prevent problems. The user can always use the
   * Annotation properties viewer to see all the information enclosed in its
   * Attributes...
   */
#ifdef UNIX
#define PRESENT_ANNOTATION_MAX_LINE_LENGTH 2048
#else
#define PRESENT_ANNOTATION_MAX_LINE_LENGTH 768
#endif /* UNIX */
  if (i > PRESENT_ANNOTATION_MAX_LINE_LENGTH) {
    Tcl_Obj *trancuated=Tcl_GetRange(Obj, 0,PRESENT_ANNOTATION_MAX_LINE_LENGTH);
    if (trancuated != NULL) {
      Tcl_DecrRefCount(Obj);
      Obj = trancuated;
    } else {
      Tcl_SetObjLength(Obj, PRESENT_ANNOTATION_MAX_LINE_LENGTH);
    }
    Tcl_AppendToObj(Obj, " ... [truncated]", 16);
    i = PRESENT_ANNOTATION_MAX_LINE_LENGTH + 17;
  }
  Tcl_ListObjAppendElement(interp, Result, Obj);

  /*
   * Create the result...
   */
  Obj = Tcl_NewListObj(0, NULL);
  if (Obj == NULL) {
    Tcl_DecrRefCount(Result);
    return TCL_ERROR;
  }
  Tcl_ListObjAppendElement(interp, Obj, CDM_NewLongObj((long) i));
  Tcl_ListObjAppendElement(interp, Obj, Result);
  Tcl_SetObjResult(interp, Obj);
  return TCL_OK;
} /* CDM_PresentAnnotationForRawViewerObjCmd */

#ifdef CDM_MEM_DEBUG
/*
 * Memory inspection commands...
 */

CDM_CALLBACK int CDM_IsSharedObjCmd(CDM_TCL_STD_ARGS) {
  CDM_UPDATE_THREAD_INTERP;
  Tcl_Obj *Obj, *val, **Element;
  int elements, i;
  
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "tcl-object");
    return TCL_ERROR;
  }

  Obj = objv[1];
  if (Tcl_ListObjGetElements(interp, Obj, &elements, &Element) != TCL_OK) {
    return TCL_ERROR;
  }

  Obj = Tcl_NewListObj(0, NULL);
  for (i=0; i<elements; i++) {
    val = Tcl_NewIntObj(Element[i]->refCount);
    if (val == NULL) return TCL_ERROR;
    Tcl_ListObjAppendElement(interp, Obj, val);
  }
  Tcl_SetObjResult(interp, Obj);
  return TCL_OK;
} /* CDM_IsSharedObjCmd */
#endif /* CDM_MEM_DEBUG */

/*
 * CDM_StubCommandObjCmd:
 * This command simply does nothing (except from reseting the result of the
 * Tcl interpreter). Used if we want an easy way to provide commands where
 * their functionality is not available under a specific platform, but the
 * command needs to be there, just for compatibility...
 */
CDM_CALLBACK int CDM_StubCommandObjCmd(CDM_TCL_STD_ARGS) {
  Tcl_ResetResult(interp);
  return TCL_OK;
}; /* CDM_StubCommandObjCmd */

CDM_CALLBACK int CDM_ClearCompressedMemoryCacheObjCmd(CDM_TCL_STD_ARGS) {
  int objects_removed;
  CDM_UPDATE_THREAD_INTERP;
  
  Tcl_ResetResult(interp);
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }
  objects_removed = (int) CDM_ClearUnusedAttributeSetHash();
  Tcl_SetObjResult(interp, Tcl_NewIntObj(objects_removed));
  return TCL_OK;
} /* CDM_ClearCompressedMemoryCacheObjCmd */

CDM_CALLBACK int CDM_CompressedMemoryCacheInfoObjCmd(CDM_TCL_STD_ARGS) {
  CDM_UPDATE_THREAD_INTERP;
  
  Tcl_ResetResult(interp);
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }
#ifdef UNIX
  ObjectCache.Info(stdout);
#endif /* UNIX */
  return TCL_OK;
} /* CDM_CompressedMemoryCacheInfoObjCmd */

CDM_CALLBACK int CDM_CompressedMemoryCacheStatisticsObjCmd(CDM_TCL_STD_ARGS) {
  Tcl_Obj *Result = NULL;
  int index;
  static const char *Methods[] = {
    "strings", "longs", "attribute_values", "attributes", (char *) NULL
  };
  enum methods {
     STRINGS,   LONGS,   ATTRIBUTE_VALUES,   ATTRIBUTES
  };
  
  CDM_UPDATE_THREAD_INTERP;
  Tcl_ResetResult(interp);
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "cache_category");
    return TCL_ERROR;
  }
  
  if (Tcl_GetIndexFromObj(interp, objv[1], (const char **) Methods,
                          "method", 0, &index) != TCL_OK) return TCL_ERROR;
  switch ((enum methods) index) {
    case STRINGS:          {
      Result = CDM_NewLongObj(ObjectCache.GetNumberOfStrings());
      break;
    }
    case LONGS:            {
      Result = CDM_NewLongObj(ObjectCache.GetNumberOfLongs());
      break;
    }
    case ATTRIBUTE_VALUES: {
      Result = CDM_NewLongObj(ObjectCache.GetNumberOfAttributeValues());
      break;
    }
    case ATTRIBUTES:       {
      Result = CDM_NewLongObj(ObjectCache.GetNumberOfAttributes());
      break;
    }
  }
  Tcl_SetObjResult(interp, Result);
  return TCL_OK;
} /* CDM_CompressedMemoryCacheStatisticsObjCmd */

/*
 * --- CDM_GetRegisteredCollectionTypesObjCmd:
 *  This command returns all Collection Types that can be handled by the
 *  application, including native supported types or registered ones...
 */
CDM_CALLBACK int CDM_GetRegisteredCollectionTypesObjCmd(CDM_TCL_STD_ARGS) {
  Tcl_Obj *TypeList;
  int i;
  CDM_THREAD_SPECIFIC_DATA;
  
  Tcl_ResetResult(interp);
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }
  
  TypeList = Tcl_NewListObj(0, NULL);
  if (TypeList == NULL) return TCL_ERROR;
  Tcl_ListObjAppendElement(interp, TypeList, CDM_NewStringObj("native"));
  Tcl_ListObjAppendElement(interp, TypeList, CDM_NewStringObj("gate"));
  for (i=0; i<CDM_MAXIMUM_COLLECTION_PLUGINS_NUMBER; i++) {
    if (CDM_AdditionalCollectionTypes[i].CollectionType == NULL) continue;
    Tcl_ListObjAppendElement(interp, TypeList,
        CDM_NewStringObj(CDM_AdditionalCollectionTypes[i].CollectionType));
  }
  Tcl_SetObjResult(interp, TypeList);
  return TCL_OK;
} /* CDM_GetRegisteredCollectionTypesObjCmd */

/*
 * ---- tip_MajorVersion:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the major version of CDM. For example,
 *     if the version of CDM is 1.3, it will return 1.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None.
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_MajorVersion (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     An integer, specifying the major version of the CDM library.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_MinorVersion, tip_Version, tip_PatchLevel, tip_ReleaseDate,
 *    tip_ExtendedReleaseDate, CDM_MajorVersion
 *   >
 *   SECTIONS:
 *    <Utilities
 *   >:&:
 */

/*
 * ---- tip_MinorVersion:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the major version of CDM. For example,
 *     if the version of CDM is 1.3, it will return 3.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None.
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_MinorVersion (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     An integer, specifying the minor version of the CDM library.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_MajorVersion, tip_Version, tip_PatchLevel, tip_ReleaseDate,
 *    tip_ExtendedReleaseDate, CDM_MinorVersion
 *   >
 *   SECTIONS:
 *    <Utilities
 *   >:&:
 */

/*
 * ---- tip_Version:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the version of CDM as a string.
 *     For example, if the version of CDM is 1.3, it will return "1.3".
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None.
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_Version (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     An integer, specifying the version of the CDM library.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_MajorVersion, tip_MinorVersion, tip_PatchLevel, tip_ReleaseDate,
 *    tip_ExtendedReleaseDate, CDM_Version
 *   >
 *   SECTIONS:
 *    <Utilities
 *   >:&:
 */

/*
 * ---- tip_PatchLevel:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the patch level of the CDM library as a string.
 *     For example, if the version of CDM is 1.3a1, it will return "1.3a1".
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None.
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_PatchLevel (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     A string, specifying the patch level of the CDM library.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_MajorVersion, tip_MinorVersion, tip_Version, tip_ReleaseDate,
 *    tip_ExtendedReleaseDate, CDM_PatchLevel
 *   >
 *   SECTIONS:
 *    <Utilities
 *   >:&:
 */

/*
 * ---- tip_ReleaseDate:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the date of the CDM library as a string.
 *     This date corresponds to the release date. The format of the return
 *     string is dd/mm/yyyy.
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None.
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_ReleaseDate (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     A string, specifying the release date of the CDM library.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_MajorVersion, tip_MinorVersion, tip_Version, tip_PatchLevel,
 *    tip_ExtendedReleaseDate, CDM_ReleaseDate
 *   >
 *   SECTIONS:
 *    <Utilities
 *   >:&:
 */

/*
 * ---- tip_ExtendedReleaseDate:
 *  DESCRIPTION:
 *   <English:
 *     This function returns the date of the CDM library as a string.
 *     This date corresponds to the release date. The format of the return
 *     string is like in the following example:
 *     "Tue Sep  4 19:08:24 EEST 2001"...
 *   >
 *  ARGUMENTS:
 *   <English:
 *     None.
 *   >
 *  NOTES:
 *   <English:
 *     This function is equivelant to CDM_ExtendedReleaseDate (C++ API).
 *   >
 *  RETURN_VALUE:
 *   <English:
 *     A string, specifying the release date of the CDM library in an extended
 *     format.
 *   >
 *  SEE_ALSO:
 *   <English:
 *    tip_MajorVersion, tip_MinorVersion, tip_Version, tip_PatchLevel,
 *    tip_ReleaseDate, CDM_ExtendedReleaseDate
 *   >
 *   SECTIONS:
 *    <Utilities
 *   >:&:
 */

CDM_CALLBACK int CDM_VersionObjCmd(CDM_TCL_STD_ARGS) {
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }
  Tcl_ResetResult(interp);
  switch ((long) clientData) {
    case 1: { /* CDM_MajorVersion */
      Tcl_SetObjResult(interp, Tcl_NewIntObj(CDM_MajorVersion()));
      break;
    }
    case 2: { /* CDM_MinorVersion */
      Tcl_SetObjResult(interp, Tcl_NewIntObj(CDM_MinorVersion()));
      break;
    }
    case 3: { /* CDM_Version */
      Tcl_SetObjResult(interp, Tcl_NewStringObj(CDM_Version(), -1));
      break;
    }
    case 4: { /* CDM_PatchLevel */
      Tcl_SetObjResult(interp, Tcl_NewStringObj(CDM_PatchLevel(), -1));
      break;
    }
    case 5: { /* CDM_ReleaseDate */
      Tcl_SetObjResult(interp, Tcl_NewStringObj(CDM_ReleaseDate(), -1));
      break;
    }
    case 6: { /* CDM_ExtendedReleaseDate */
      Tcl_SetObjResult(interp, Tcl_NewStringObj(CDM_ExtendedReleaseDate(), -1));
      break;
    }
    default: {
      return TCL_ERROR;
    }
  }
  return TCL_OK;
} /* CDM_VersionObjCmd */

CDM_CALLBACK int CDM_LicenseKeyObjCmd(CDM_TCL_STD_ARGS) {
#ifdef CDM_ENABLE_LICENSE_MANAGER
#include "Internal/CDM_LicenseKeyObjCmd.cpp"
#else
  CDM_UPDATE_THREAD_INTERP;
  Tcl_ResetResult(interp);
  switch ((long) clientData) {
    case 1: { /* CDM_GenerateLicenseKey */
      if (objc == 2) Tcl_SetObjResult(interp, Tcl_NewIntObj(0));
      else Tcl_SetObjResult(interp, Tcl_NewStringObj("LICENSES DISABLED", -1));
      return TCL_OK;
    }
    case 2: { /* CDM_ValidateLicenseKey */
      Tcl_SetObjResult(interp, Tcl_NewStringObj("LICENSE_VALID_KEY", -1));
      return TCL_OK;
    }
  }
  return TCL_OK;
#endif /* CDM_ENABLE_LICENSE_MANAGER */
} /* CDM_LicenseKeyObjCmd */

/************************************************************************
 ****       C D M     L I C E N S E      R O U T I N E S             ****
 ************************************************************************/
/* CDM_DeclareLicensingInformation */
CDM_CALLBACK int CDM_DeclareLicensingInformationObjCmd(CDM_TCL_STD_ARGS) {
  
#ifdef CDM_ENABLE_LICENSE_MANAGER
#include "Internal/CDM_DeclareLicensingInformation.cpp"
#else
  Tcl_ResetResult(interp);
  CDM_UPDATE_THREAD_INTERP;
  return TCL_OK;
#endif /* CDM_ENABLE_LICENSE_MANAGER */
} /* CDM_DeclareLicensingInformation */

/* CDM_MutexLockObjCmd */
CDM_CALLBACK int CDM_MutexLockObjCmd(CDM_TCL_STD_ARGS) {
  CDM_UPDATE_THREAD_INTERP;
  CDM_MutexLock();
  return TCL_OK;
} /* CDM_MutexLockObjCmd */

/* CDM_MutexUnlockObjCmd */
CDM_CALLBACK int CDM_MutexUnlockObjCmd(CDM_TCL_STD_ARGS) {
  CDM_UPDATE_THREAD_INTERP;
  CDM_MutexUnlock();
  return TCL_OK;
} /* CDM_MutexUnlockObjCmd */

#ifdef __APPLE__
#include <mach-o/dyld.h>
#endif /* __APPLE__ */
CDM_CALLBACK int CDM_AddLibraryImageObjCmd(CDM_TCL_STD_ARGS) {
  CDM_UPDATE_THREAD_INTERP;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "dynamic-library-native-path");
    return TCL_ERROR;
  }
#ifdef __APPLE__
  NSAddImage(Tcl_GetString(objv[1]), NSADDIMAGE_OPTION_RETURN_ON_ERROR |
                                     NSADDIMAGE_OPTION_WITH_SEARCHING |
                                     NSADDIMAGE_OPTION_MATCH_FILENAME_BY_INSTALLNAME);
#endif /* __APPLE__ */
  return TCL_OK;
} /* CDM_AddLibraryImageObjCmd */

#ifdef CDM_ENABLE_TK_COMMANDS
CDM_CALLBACK int WindowObjCmd(CDM_TCL_STD_ARGS) {
  Tk_Window tkwin;
  Tk_Image  tkimg;
  int width, height;
  Pixmap pixmap;
  unsigned long gcMask;
  XGCValues gcValues;
  GC newGC;
  
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "window image");
    return TCL_ERROR;
  }
  tkwin = Tk_NameToWindow(interp,Tcl_GetString(objv[1]),Tk_MainWindow(interp));
  if (tkwin == NULL) return TCL_ERROR;
  tkimg = Tk_GetImage(interp, tkwin, Tcl_GetString(objv[2]), NULL, NULL);
  if (tkimg == NULL) return TCL_ERROR;
  Tk_MakeWindowExist(tkwin);

  Tk_SizeOfImage(tkimg, &width, &height);
  pixmap = Tk_GetPixmap(Tk_Display(tkwin), Tk_WindowId(tkwin), width, height,
                        Tk_Depth(tkwin));
  Tk_RedrawImage(tkimg, 0, 0, width, height, pixmap, 0, 0);
  gcMask = (GCTile | GCFillStyle);
  gcValues.fill_style = FillTiled;
  gcValues.tile = pixmap;
  newGC = Tk_GetGC(tkwin, gcMask, &gcValues);
  XSetTSOrigin(Tk_Display(tkwin), newGC, 0, 0);
  XFillRectangle(Tk_Display(tkwin), Tk_WindowId(tkwin), newGC, 0, 0,
                 Tk_Width(tkwin), Tk_Height(tkwin));
  //Tk_SetWindowBackgroundPixmap(tkwin, pixmap);
  Tk_FreeGC(Tk_Display(tkwin), newGC);
  Tk_FreePixmap(Tk_Display(tkwin), pixmap);
  Tk_FreeImage(tkimg);
  return TCL_OK;
}
#endif /* CDM_ENABLE_TK_COMMANDS */

/*
 * Experimental Tcl Obj Hash Memory Release...
 */
extern "C" void Tcl_Petasis(void);
CDM_CALLBACK int CDM_FreeTclObjCacheObjCmd(CDM_TCL_STD_ARGS) {
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }
  //Tcl_Petasis();
  return TCL_OK;
}
/*
 * Cdm_Init.cpp - EOF -
 */
